EXPORTS
	pjsip_100rel_attach                      @ 1 NONAME
	pjsip_100rel_create_prack                @ 2 NONAME
	pjsip_100rel_end_session                 @ 3 NONAME
	pjsip_100rel_init_module                 @ 4 NONAME
	pjsip_100rel_is_reliable                 @ 5 NONAME
	pjsip_100rel_on_rx_prack                 @ 6 NONAME
	pjsip_100rel_send_prack                  @ 7 NONAME
	pjsip_100rel_tx_response                 @ 8 NONAME
	pjsip_create_sdp_body                    @ 9 NONAME
	pjsip_dlg_get_inv_session                @ 10 NONAME
	pjsip_get_prack_method                   @ 11 NONAME
	pjsip_get_refer_method                   @ 12 NONAME
	pjsip_inv_answer                         @ 13 NONAME
	pjsip_inv_create_ack                     @ 14 NONAME
	pjsip_inv_create_uac                     @ 15 NONAME
	pjsip_inv_create_uas                     @ 16 NONAME
	pjsip_inv_end_session                    @ 17 NONAME
	pjsip_inv_initial_answer                 @ 18 NONAME
	pjsip_inv_invite                         @ 19 NONAME
	pjsip_inv_reinvite                       @ 20 NONAME
	pjsip_inv_send_msg                       @ 21 NONAME
	pjsip_inv_set_sdp_answer                 @ 22 NONAME
	pjsip_inv_state_name                     @ 23 NONAME
	pjsip_inv_terminate                      @ 24 NONAME
	pjsip_inv_update                         @ 25 NONAME
	pjsip_inv_usage_init                     @ 26 NONAME
	pjsip_inv_usage_instance                 @ 27 NONAME
	pjsip_inv_verify_request                 @ 28 NONAME
	pjsip_prack_method                       @ 29 NONAME
	pjsip_refer_method                       @ 30 NONAME
	pjsip_regc_add_headers                   @ 31 NONAME
	pjsip_regc_create                        @ 32 NONAME
	pjsip_regc_destroy                       @ 33 NONAME
	pjsip_regc_get_info                      @ 34 NONAME
	pjsip_regc_get_pool                      @ 35 NONAME
	pjsip_regc_init                          @ 36 NONAME
	pjsip_regc_register                      @ 37 NONAME
	pjsip_regc_send                          @ 38 NONAME
	pjsip_regc_set_credentials               @ 39 NONAME
	pjsip_regc_set_prefs                     @ 40 NONAME
	pjsip_regc_set_route_set                 @ 41 NONAME
	pjsip_regc_set_transport                 @ 42 NONAME
	pjsip_regc_unregister                    @ 43 NONAME
	pjsip_regc_unregister_all                @ 44 NONAME
	pjsip_regc_update_contact                @ 45 NONAME
	pjsip_regc_update_expires                @ 46 NONAME
	pjsip_replaces_hdr_create                @ 47 NONAME
	pjsip_replaces_init_module               @ 48 NONAME
	pjsip_replaces_verify_request            @ 49 NONAME
	pjsip_xfer_accept                        @ 50 NONAME
	pjsip_xfer_create_uac                    @ 51 NONAME
	pjsip_xfer_create_uas                    @ 52 NONAME
	pjsip_xfer_current_notify                @ 53 NONAME
	pjsip_xfer_init_module                   @ 54 NONAME
	pjsip_xfer_initiate                      @ 55 NONAME
	pjsip_xfer_notify                        @ 56 NONAME
	pjsip_xfer_send_request                  @ 57 NONAME
