EXPORTS
	pjsip_accept_hdr_create                  @ 1 NONAME
	pjsip_accept_hdr_init                    @ 2 NONAME
	pjsip_ack_method                         @ 3 NONAME
	pjsip_allow_hdr_create                   @ 4 NONAME
	pjsip_allow_hdr_init                     @ 5 NONAME
	pjsip_auth_clt_clone                     @ 6 NONAME
	pjsip_auth_clt_get_prefs                 @ 7 NONAME
	pjsip_auth_clt_init                      @ 8 NONAME
	pjsip_auth_clt_init_req                  @ 9 NONAME
	pjsip_auth_clt_reinit_req                @ 10 NONAME
	pjsip_auth_clt_set_credentials           @ 11 NONAME
	pjsip_auth_clt_set_prefs                 @ 12 NONAME
	pjsip_auth_create_digest                 @ 13 NONAME
	pjsip_auth_deinit_parser                 @ 14 NONAME
	pjsip_auth_init_parser                   @ 15 NONAME
	pjsip_auth_srv_challenge                 @ 16 NONAME
	pjsip_auth_srv_init                      @ 17 NONAME
	pjsip_auth_srv_verify                    @ 18 NONAME
	pjsip_authorization_hdr_create           @ 19 NONAME
	pjsip_bye_method                         @ 20 NONAME
	pjsip_calculate_branch_id                @ 21 NONAME
	pjsip_cancel_method                      @ 22 NONAME
	pjsip_cid_hdr_create                     @ 23 NONAME
	pjsip_cid_hdr_init                       @ 24 NONAME
	pjsip_clen_hdr_create                    @ 25 NONAME
	pjsip_clen_hdr_init                      @ 26 NONAME
	pjsip_clone_text_data                    @ 27 NONAME
	pjsip_concat_param_imp                   @ 28 NONAME
	pjsip_contact_hdr_create                 @ 29 NONAME
	pjsip_contact_hdr_init                   @ 30 NONAME
	pjsip_cred_info_dup                      @ 31 NONAME
	pjsip_cseq_hdr_create                    @ 32 NONAME
	pjsip_cseq_hdr_init                      @ 33 NONAME
	pjsip_ctype_hdr_create                   @ 34 NONAME
	pjsip_ctype_hdr_init                     @ 35 NONAME
	pjsip_dlg_add_usage                      @ 36 NONAME
	pjsip_dlg_create_request                 @ 37 NONAME
	pjsip_dlg_create_response                @ 38 NONAME
	pjsip_dlg_create_uac                     @ 39 NONAME
	pjsip_dlg_create_uas                     @ 40 NONAME
	pjsip_dlg_dec_lock                       @ 41 NONAME
	pjsip_dlg_dec_session                    @ 42 NONAME
	pjsip_dlg_fork                           @ 43 NONAME
	pjsip_dlg_get_mod_data                   @ 44 NONAME
	pjsip_dlg_inc_lock                       @ 45 NONAME
	pjsip_dlg_inc_session                    @ 46 NONAME
	pjsip_dlg_modify_response                @ 47 NONAME
	pjsip_dlg_respond                        @ 48 NONAME
	pjsip_dlg_send_request                   @ 49 NONAME
	pjsip_dlg_send_response                  @ 50 NONAME
	pjsip_dlg_set_mod_data                   @ 51 NONAME
	pjsip_dlg_set_route_set                  @ 52 NONAME
	pjsip_dlg_set_transport                  @ 53 NONAME
	pjsip_dlg_terminate                      @ 54 NONAME
	pjsip_dlg_try_inc_lock                   @ 55 NONAME
	pjsip_endpt_acquire_transport            @ 56 NONAME
	pjsip_endpt_add_capability               @ 57 NONAME
	pjsip_endpt_cancel_timer                 @ 58 NONAME
	pjsip_endpt_create                       @ 59 NONAME
	pjsip_endpt_create_ack                   @ 60 NONAME
	pjsip_endpt_create_cancel                @ 61 NONAME
	pjsip_endpt_create_pool                  @ 62 NONAME
	pjsip_endpt_create_request               @ 63 NONAME
	pjsip_endpt_create_request_from_hdr      @ 64 NONAME
	pjsip_endpt_create_request_fwd           @ 65 NONAME
	pjsip_endpt_create_resolver              @ 66 NONAME
	pjsip_endpt_create_response              @ 67 NONAME
	pjsip_endpt_create_response_fwd          @ 68 NONAME
	pjsip_endpt_create_tdata                 @ 69 NONAME
	pjsip_endpt_destroy                      @ 70 NONAME
	pjsip_endpt_dump                         @ 71 NONAME
	pjsip_endpt_get_capability               @ 72 NONAME
	pjsip_endpt_get_ioqueue                  @ 73 NONAME
	pjsip_endpt_get_request_headers          @ 74 NONAME
	pjsip_endpt_get_resolver                 @ 75 NONAME
	pjsip_endpt_get_timer_heap               @ 76 NONAME
	pjsip_endpt_get_tpmgr                    @ 77 NONAME
	pjsip_endpt_handle_events                @ 78 NONAME
	pjsip_endpt_handle_events2               @ 79 NONAME
	pjsip_endpt_has_capability               @ 80 NONAME
	pjsip_endpt_log_error                    @ 81 NONAME
	pjsip_endpt_name                         @ 82 NONAME
	pjsip_endpt_register_module              @ 83 NONAME
	pjsip_endpt_release_pool                 @ 84 NONAME
	pjsip_endpt_resolve                      @ 85 NONAME
	pjsip_endpt_respond                      @ 86 NONAME
	pjsip_endpt_respond_stateless            @ 87 NONAME
	pjsip_endpt_schedule_timer               @ 88 NONAME
	pjsip_endpt_send_raw                     @ 89 NONAME
	pjsip_endpt_send_raw_to_uri              @ 90 NONAME
	pjsip_endpt_send_request                 @ 91 NONAME
	pjsip_endpt_send_request_stateless       @ 92 NONAME
	pjsip_endpt_send_response                @ 93 NONAME
	pjsip_endpt_send_response2               @ 94 NONAME
	pjsip_endpt_set_resolver                 @ 95 NONAME
	pjsip_endpt_unregister_module            @ 96 NONAME
	pjsip_event_str                          @ 97 NONAME
	pjsip_expires_hdr_create                 @ 98 NONAME
	pjsip_expires_hdr_init                   @ 99 NONAME
	pjsip_find_msg                           @ 100 NONAME
	pjsip_from_hdr_create                    @ 101 NONAME
	pjsip_from_hdr_init                      @ 102 NONAME
	pjsip_fromto_hdr_set_from                @ 103 NONAME
	pjsip_fromto_hdr_set_to                  @ 104 NONAME
	pjsip_generic_array_hdr_create           @ 105 NONAME
	pjsip_generic_array_hdr_init             @ 106 NONAME
	pjsip_generic_int_hdr_create             @ 107 NONAME
	pjsip_generic_int_hdr_init               @ 108 NONAME
	pjsip_generic_string_hdr_create          @ 109 NONAME
	pjsip_generic_string_hdr_init            @ 110 NONAME
	pjsip_generic_string_hdr_init2           @ 111 NONAME
	pjsip_get_ack_method                     @ 112 NONAME
	pjsip_get_bye_method                     @ 113 NONAME
	pjsip_get_cancel_method                  @ 114 NONAME
	pjsip_get_invite_method                  @ 115 NONAME
	pjsip_get_options_method                 @ 116 NONAME
	pjsip_get_register_method                @ 117 NONAME
	pjsip_get_request_dest                   @ 118 NONAME
	pjsip_get_response_addr                  @ 119 NONAME
	pjsip_get_status_text                    @ 120 NONAME
	pjsip_hdr_clone                          @ 121 NONAME
	pjsip_hdr_print_on                       @ 122 NONAME
	pjsip_hdr_shallow_clone                  @ 123 NONAME
	pjsip_invite_method                      @ 124 NONAME
	pjsip_loop_set_delay                     @ 125 NONAME
	pjsip_loop_set_discard                   @ 126 NONAME
	pjsip_loop_set_failure                   @ 127 NONAME
	pjsip_loop_set_recv_delay                @ 128 NONAME
	pjsip_loop_set_send_callback_delay       @ 129 NONAME
	pjsip_loop_start                         @ 130 NONAME
	pjsip_max_fwd_hdr_create                 @ 131 NONAME
	pjsip_max_fwd_hdr_init                   @ 132 NONAME
	pjsip_method_cmp                         @ 133 NONAME
	pjsip_method_copy                        @ 134 NONAME
	pjsip_method_creates_dialog              @ 135 NONAME
	pjsip_method_init                        @ 136 NONAME
	pjsip_method_init_np                     @ 137 NONAME
	pjsip_method_set                         @ 138 NONAME
	pjsip_min_expires_hdr_create             @ 139 NONAME
	pjsip_min_expires_hdr_init               @ 140 NONAME
	pjsip_msg_body_clone                     @ 141 NONAME
	pjsip_msg_body_copy                      @ 142 NONAME
	pjsip_msg_body_create                    @ 143 NONAME
	pjsip_msg_clone                          @ 144 NONAME
	pjsip_msg_create                         @ 145 NONAME
	pjsip_msg_find_hdr                       @ 146 NONAME
	pjsip_msg_find_hdr_by_name               @ 147 NONAME
	pjsip_msg_find_remove_hdr                @ 148 NONAME
	pjsip_msg_print                          @ 149 NONAME
	pjsip_name_addr_assign                   @ 150 NONAME
	pjsip_name_addr_create                   @ 151 NONAME
	pjsip_name_addr_init                     @ 152 NONAME
	pjsip_options_method                     @ 153 NONAME
	pjsip_param_cfind                        @ 154 NONAME
	pjsip_param_clone                        @ 155 NONAME
	pjsip_param_find                         @ 156 NONAME
	pjsip_param_print_on                     @ 157 NONAME
	pjsip_param_shallow_clone                @ 158 NONAME
	pjsip_parse_end_hdr_imp                  @ 159 NONAME
	pjsip_parse_hdr                          @ 160 NONAME
	pjsip_parse_msg                          @ 161 NONAME
	pjsip_parse_param_imp                    @ 162 NONAME
	pjsip_parse_rdata                        @ 163 NONAME
	pjsip_parse_status_line                  @ 164 NONAME
	pjsip_parse_uri                          @ 165 NONAME
	pjsip_parse_uri_param_imp                @ 166 NONAME
	pjsip_parser_const                       @ 167 NONAME
	pjsip_print_text_body                    @ 168 NONAME
	pjsip_process_route_set                  @ 169 NONAME
	pjsip_proxy_authenticate_hdr_create      @ 170 NONAME
	pjsip_proxy_authorization_hdr_create     @ 171 NONAME
	pjsip_rdata_get_dlg                      @ 172 NONAME
	pjsip_rdata_get_tsx                      @ 173 NONAME
	pjsip_register_hdr_parser                @ 174 NONAME
	pjsip_register_method                    @ 175 NONAME
	pjsip_register_uri_parser                @ 176 NONAME
	pjsip_require_hdr_create                 @ 177 NONAME
	pjsip_require_hdr_init                   @ 178 NONAME
	pjsip_resolve                            @ 179 NONAME
	pjsip_resolver_create                    @ 180 NONAME
	pjsip_resolver_destroy                   @ 181 NONAME
	pjsip_resolver_get_resolver              @ 182 NONAME
	pjsip_resolver_set_resolver              @ 183 NONAME
	pjsip_retry_after_hdr_create             @ 184 NONAME
	pjsip_retry_after_hdr_init               @ 185 NONAME
	pjsip_role_name                          @ 186 NONAME
	pjsip_route_hdr_create                   @ 187 NONAME
	pjsip_route_hdr_init                     @ 188 NONAME
	pjsip_routing_hdr_set_route              @ 189 NONAME
	pjsip_routing_hdr_set_rr                 @ 190 NONAME
	pjsip_rr_hdr_create                      @ 191 NONAME
	pjsip_rr_hdr_init                        @ 192 NONAME
	pjsip_rx_data_get_info                   @ 193 NONAME
	pjsip_sip_uri_assign                     @ 194 NONAME
	pjsip_sip_uri_create                     @ 195 NONAME
	pjsip_sip_uri_init                       @ 196 NONAME
	pjsip_sip_uri_set_secure                 @ 197 NONAME
	pjsip_strerror                           @ 198 NONAME
	pjsip_supported_hdr_create               @ 199 NONAME
	pjsip_supported_hdr_init                 @ 200 NONAME
	pjsip_tcp_transport_start                @ 201 NONAME
	pjsip_tcp_transport_start2               @ 202 NONAME
	pjsip_tel_nb_cmp                         @ 203 NONAME
	pjsip_tel_uri_create                     @ 204 NONAME
	pjsip_to_hdr_create                      @ 205 NONAME
	pjsip_to_hdr_init                        @ 206 NONAME
	pjsip_tpmgr_acquire_transport            @ 207 NONAME
	pjsip_tpmgr_create                       @ 208 NONAME
	pjsip_tpmgr_destroy                      @ 209 NONAME
	pjsip_tpmgr_dump_transports              @ 210 NONAME
	pjsip_tpmgr_find_local_addr              @ 211 NONAME
	pjsip_tpmgr_get_transport_count          @ 212 NONAME
	pjsip_tpmgr_receive_packet               @ 213 NONAME
	pjsip_tpmgr_register_tpfactory           @ 214 NONAME
	pjsip_tpmgr_send_raw                     @ 215 NONAME
	pjsip_tpmgr_unregister_tpfactory         @ 216 NONAME
	pjsip_tpselector_add_ref                 @ 217 NONAME
	pjsip_tpselector_dec_ref                 @ 218 NONAME
	pjsip_transport_add_ref                  @ 219 NONAME
	pjsip_transport_dec_ref                  @ 220 NONAME
	pjsip_transport_destroy                  @ 221 NONAME
	pjsip_transport_get_default_port_for_type @ 222 NONAME
	pjsip_transport_get_flag_from_type       @ 223 NONAME
	pjsip_transport_get_type_desc            @ 224 NONAME
	pjsip_transport_get_type_from_flag       @ 225 NONAME
	pjsip_transport_get_type_from_name       @ 226 NONAME
	pjsip_transport_get_type_name            @ 227 NONAME
	pjsip_transport_register                 @ 228 NONAME
	pjsip_transport_register_type            @ 229 NONAME
	pjsip_transport_send                     @ 230 NONAME
	pjsip_transport_shutdown                 @ 231 NONAME
	pjsip_transport_type_get_af              @ 232 NONAME
	pjsip_tsx_create_key                     @ 233 NONAME
	pjsip_tsx_create_uac                     @ 234 NONAME
	pjsip_tsx_create_uas                     @ 235 NONAME
	pjsip_tsx_get_dlg                        @ 236 NONAME
	pjsip_tsx_layer_destroy                  @ 237 NONAME
	pjsip_tsx_layer_dump                     @ 238 NONAME
	pjsip_tsx_layer_find_tsx                 @ 239 NONAME
	pjsip_tsx_layer_get_tsx_count            @ 240 NONAME
	pjsip_tsx_layer_init_module              @ 241 NONAME
	pjsip_tsx_layer_instance                 @ 242 NONAME
	pjsip_tsx_recv_msg                       @ 243 NONAME
	pjsip_tsx_retransmit_no_state            @ 244 NONAME
	pjsip_tsx_send_msg                       @ 245 NONAME
	pjsip_tsx_set_transport                  @ 246 NONAME
	pjsip_tsx_state_str                      @ 247 NONAME
	pjsip_tsx_stop_retransmit                @ 248 NONAME
	pjsip_tsx_terminate                      @ 249 NONAME
	pjsip_tx_data_add_ref                    @ 250 NONAME
	pjsip_tx_data_create                     @ 251 NONAME
	pjsip_tx_data_dec_ref                    @ 252 NONAME
	pjsip_tx_data_get_info                   @ 253 NONAME
	pjsip_tx_data_invalidate_msg             @ 254 NONAME
	pjsip_tx_data_is_valid                   @ 255 NONAME
	pjsip_tx_data_set_transport              @ 256 NONAME
	pjsip_ua_destroy                         @ 257 NONAME
	pjsip_ua_dump                            @ 258 NONAME
	pjsip_ua_find_dialog                     @ 259 NONAME
	pjsip_ua_get_dlg_set_count               @ 260 NONAME
	pjsip_ua_get_endpt                       @ 261 NONAME
	pjsip_ua_init_module                     @ 262 NONAME
	pjsip_ua_instance                        @ 263 NONAME
	pjsip_ua_register_dlg                    @ 264 NONAME
	pjsip_ua_unregister_dlg                  @ 265 NONAME
	pjsip_udp_transport_attach               @ 266 NONAME
	pjsip_udp_transport_attach2              @ 267 NONAME
	pjsip_udp_transport_get_socket           @ 268 NONAME
	pjsip_udp_transport_pause                @ 269 NONAME
	pjsip_udp_transport_restart              @ 270 NONAME
	pjsip_udp_transport_start                @ 271 NONAME
	pjsip_udp_transport_start6               @ 272 NONAME
	pjsip_unsupported_hdr_create             @ 273 NONAME
	pjsip_unsupported_hdr_init               @ 274 NONAME
	pjsip_via_hdr_create                     @ 275 NONAME
	pjsip_via_hdr_init                       @ 276 NONAME
	pjsip_warning_hdr_create                 @ 277 NONAME
	pjsip_warning_hdr_create_from_status     @ 278 NONAME
	pjsip_www_authenticate_hdr_create        @ 279 NONAME
