<?xml version="1.0" encoding="UTF-8" ?>
<ContentPage
    xmlns="http://xamarin.com/schemas/2014/forms"
    xmlns:x="http://schemas.microsoft.com/winfx/2009/xaml"
    x:Class="pjsua2xamarin.CallPage">
    <ContentPage.Content>
        
    </ContentPage.Content>
</ContentPage>

