@rem set MWSym2Libraries=1
@rem set EPOCROOT=\Symbian\9.1\S60_3rd\
@rem set EPOCROOT=\Symbian\9.1\S60_3rd_MR_2\
@rem set EPOCROOT=\Symbian\UIQ3SDK\
@rem set EPOCROOT=\symbian\UIQ3.1\
@rem set EPOCROOT=\symbian\9.2\S60_3rd_FP1\
bldmake bldfiles
