/*
========================================================================
 Name        : PjsuaContainerView.h
 Author      : nanang
 Copyright   : Copyright (C) 2013 Teluu Inc. (http://www.teluu.com)
 Description : 
========================================================================
*/
#ifndef PJSUACONTAINERVIEW_H
#define PJSUACONTAINERVIEW_H

// [[[ begin generated region: do not modify [Generated Includes]
#include <aknview.h>
// ]]] end generated region [Generated Includes]


// [[[ begin [Event Handler Includes]
// ]]] end [Event Handler Includes]

// [[[ begin generated region: do not modify [Generated Constants]
// ]]] end generated region [Generated Constants]

// [[[ begin generated region: do not modify [Generated Forward Declarations]
class CPjsuaContainer;
// ]]] end generated region [Generated Forward Declarations]

/**
 * Avkon view class for pjsuaContainerView. It is register with the view server
 * by the AppUi. It owns the container control.
 * @class       CpjsuaContainerView pjsuaContainerView.h
 */                                             
                        
class CpjsuaContainerView : public CAknView
        {
        
        
        // [[[ begin [Public Section]
public:
        // constructors and destructor
        CpjsuaContainerView();
        static CpjsuaContainerView* NewL();
        static CpjsuaContainerView* NewLC();        
        void ConstructL();
        virtual ~CpjsuaContainerView();
                                                
        // from base class CAknView
        TUid Id() const;
        void HandleCommandL( TInt aCommand );
        
        // [[[ begin generated region: do not modify [Generated Methods]
        CPjsuaContainer* CreateContainerL();
        // ]]] end generated region [Generated Methods]
        
        // ]]] end [Public Section]
        
        void PutMessage( const char *msg );
        
        // [[[ begin [Protected Section]
protected:
        // from base class CAknView
        void DoActivateL(
                const TVwsViewId& aPrevViewId,
                TUid aCustomMessageId,
                const TDesC8& aCustomMessage );
        void DoDeactivate();
        void HandleStatusPaneSizeChange();
        
        // [[[ begin generated region: do not modify [Overridden Methods]
        // ]]] end generated region [Overridden Methods]
        
        
        // [[[ begin [User Handlers]
        // ]]] end [User Handlers]
        
        // ]]] end [Protected Section]
        
        
        // [[[ begin [Private Section]
private:
        void SetupStatusPaneL();
        void CleanupStatusPane();
        
        // [[[ begin generated region: do not modify [Generated Instance Variables]
        CPjsuaContainer* iPjsuaContainer;
        // ]]] end generated region [Generated Instance Variables]
        
        // [[[ begin generated region: do not modify [Generated Methods]
        // ]]] end generated region [Generated Methods]
        
        // ]]] end [Private Section]
        
        };

#endif // PJSUACONTAINERVIEW_H
