<?xml version="1.0" encoding="UTF-8" ?>
<ContentPage xmlns="http://schemas.microsoft.com/dotnet/2021/maui"
             xmlns:x="http://schemas.microsoft.com/winfx/2009/xaml"
            x:Class="pjsua2maui.Views.CallPage"
             xmlns:c="clr-namespace:pjsua2maui.Controls">
    <ContentPage.Content>
        <c:CallView>
            

        </c:CallView>
    </ContentPage.Content>
</ContentPage>

