EXPORTS
	echo_supp_cancel_echo                    @ 1 NONAME
	echo_supp_capture                        @ 2 NONAME
	echo_supp_create                         @ 3 NONAME
	echo_supp_destroy                        @ 4 NONAME
	echo_supp_playback                       @ 5 NONAME
	pjmedia_bidirectional_port_create        @ 6 NONAME
	pjmedia_calc_avg_signal                  @ 7 NONAME
	pjmedia_clock_create                     @ 8 NONAME
	pjmedia_clock_destroy                    @ 9 NONAME
	pjmedia_clock_start                      @ 10 NONAME
	pjmedia_clock_stop                       @ 11 NONAME
	pjmedia_clock_wait                       @ 12 NONAME
	pjmedia_codec_g711_deinit                @ 13 NONAME
	pjmedia_codec_g711_init                  @ 14 NONAME
	pjmedia_codec_info_to_id                 @ 15 NONAME
	pjmedia_codec_mgr_alloc_codec            @ 16 NONAME
	pjmedia_codec_mgr_dealloc_codec          @ 17 NONAME
	pjmedia_codec_mgr_enum_codecs            @ 18 NONAME
	pjmedia_codec_mgr_find_codecs_by_id      @ 19 NONAME
	pjmedia_codec_mgr_get_codec_info         @ 20 NONAME
	pjmedia_codec_mgr_get_default_param      @ 21 NONAME
	pjmedia_codec_mgr_init                   @ 22 NONAME
	pjmedia_codec_mgr_register_factory       @ 23 NONAME
	pjmedia_codec_mgr_set_codec_priority     @ 24 NONAME
	pjmedia_codec_mgr_unregister_factory     @ 25 NONAME
	pjmedia_conf_add_passive_port            @ 26 NONAME
	pjmedia_conf_add_port                    @ 27 NONAME
	pjmedia_conf_adjust_rx_level             @ 28 NONAME
	pjmedia_conf_adjust_tx_level             @ 29 NONAME
	pjmedia_conf_configure_port              @ 30 NONAME
	pjmedia_conf_connect_port                @ 31 NONAME
	pjmedia_conf_create                      @ 32 NONAME
	pjmedia_conf_destroy                     @ 33 NONAME
	pjmedia_conf_disconnect_port             @ 34 NONAME
	pjmedia_conf_enum_ports                  @ 35 NONAME
	pjmedia_conf_get_connect_count           @ 36 NONAME
	pjmedia_conf_get_master_port             @ 37 NONAME
	pjmedia_conf_get_port_count              @ 38 NONAME
	pjmedia_conf_get_port_info               @ 39 NONAME
	pjmedia_conf_get_ports_info              @ 40 NONAME
	pjmedia_conf_get_signal_level            @ 41 NONAME
	pjmedia_conf_remove_port                 @ 42 NONAME
	pjmedia_conf_set_port0_name              @ 43 NONAME
	pjmedia_delay_buf_create                 @ 44 NONAME
	pjmedia_delay_buf_get                    @ 45 NONAME
	pjmedia_delay_buf_put                    @ 46 NONAME
	pjmedia_echo_cancel                      @ 47 NONAME
	pjmedia_echo_capture                     @ 48 NONAME
	pjmedia_echo_create                      @ 49 NONAME
	pjmedia_echo_destroy                     @ 50 NONAME
	pjmedia_echo_playback                    @ 51 NONAME
	pjmedia_echo_port_create                 @ 52 NONAME
	pjmedia_endpt_create                     @ 53 NONAME
	pjmedia_endpt_create_pool                @ 54 NONAME
	pjmedia_endpt_create_sdp                 @ 55 NONAME
	pjmedia_endpt_destroy                    @ 56 NONAME
	pjmedia_endpt_dump                       @ 57 NONAME
	pjmedia_endpt_get_codec_mgr              @ 58 NONAME
	pjmedia_endpt_get_ioqueue                @ 59 NONAME
	pjmedia_endpt_get_thread                 @ 60 NONAME
	pjmedia_endpt_get_thread_count           @ 61 NONAME
	pjmedia_ice_create                       @ 62 NONAME
	pjmedia_ice_destroy                      @ 63 NONAME
	pjmedia_ice_get_comp                     @ 64 NONAME
	pjmedia_ice_get_init_status              @ 65 NONAME
	pjmedia_ice_init_ice                     @ 66 NONAME
	pjmedia_ice_modify_sdp                   @ 67 NONAME
	pjmedia_ice_simulate_lost                @ 68 NONAME
	pjmedia_ice_start_ice                    @ 69 NONAME
	pjmedia_ice_start_init                   @ 70 NONAME
	pjmedia_ice_stop_ice                     @ 71 NONAME
	pjmedia_jbuf_create                      @ 72 NONAME
	pjmedia_jbuf_destroy                     @ 73 NONAME
	pjmedia_jbuf_get_frame                   @ 74 NONAME
	pjmedia_jbuf_get_state                   @ 75 NONAME
	pjmedia_jbuf_put_frame                   @ 76 NONAME
	pjmedia_jbuf_reset                       @ 77 NONAME
	pjmedia_jbuf_set_adaptive                @ 78 NONAME
	pjmedia_jbuf_set_fixed                   @ 79 NONAME
	pjmedia_master_port_create               @ 80 NONAME
	pjmedia_master_port_destroy              @ 81 NONAME
	pjmedia_master_port_get_dport            @ 82 NONAME
	pjmedia_master_port_get_uport            @ 83 NONAME
	pjmedia_master_port_set_dport            @ 84 NONAME
	pjmedia_master_port_set_uport            @ 85 NONAME
	pjmedia_master_port_start                @ 86 NONAME
	pjmedia_master_port_stop                 @ 87 NONAME
	pjmedia_mem_capture_create               @ 88 NONAME
	pjmedia_mem_capture_get_size             @ 89 NONAME
	pjmedia_mem_capture_set_eof_cb           @ 90 NONAME
	pjmedia_mem_player_create                @ 91 NONAME
	pjmedia_mem_player_set_eof_cb            @ 92 NONAME
	pjmedia_null_port_create                 @ 93 NONAME
	pjmedia_plc_create                       @ 94 NONAME
	pjmedia_plc_generate                     @ 95 NONAME
	pjmedia_plc_save                         @ 96 NONAME
	pjmedia_port_destroy                     @ 97 NONAME
	pjmedia_port_get_frame                   @ 98 NONAME
	pjmedia_port_info_init                   @ 99 NONAME
	pjmedia_port_put_frame                   @ 100 NONAME
	pjmedia_resample_create                  @ 101 NONAME
	pjmedia_resample_destroy                 @ 102 NONAME
	pjmedia_resample_get_input_size          @ 103 NONAME
	pjmedia_resample_port_create             @ 104 NONAME
	pjmedia_resample_run                     @ 105 NONAME
	pjmedia_rtcp_build_rtcp                  @ 106 NONAME
	pjmedia_rtcp_fini                        @ 107 NONAME
	pjmedia_rtcp_get_ntp_time                @ 108 NONAME
	pjmedia_rtcp_init                        @ 109 NONAME
	pjmedia_rtcp_rx_rtcp                     @ 110 NONAME
	pjmedia_rtcp_rx_rtp                      @ 111 NONAME
	pjmedia_rtcp_tx_rtp                      @ 112 NONAME
	pjmedia_rtp_decode_rtp                   @ 113 NONAME
	pjmedia_rtp_encode_rtp                   @ 114 NONAME
	pjmedia_rtp_session_init                 @ 115 NONAME
	pjmedia_rtp_session_update               @ 116 NONAME
	pjmedia_session_check_dtmf               @ 117 NONAME
	pjmedia_session_create                   @ 118 NONAME
	pjmedia_session_destroy                  @ 119 NONAME
	pjmedia_session_dial_dtmf                @ 120 NONAME
	pjmedia_session_enum_streams             @ 121 NONAME
	pjmedia_session_get_dtmf                 @ 122 NONAME
	pjmedia_session_get_info                 @ 123 NONAME
	pjmedia_session_get_port                 @ 124 NONAME
	pjmedia_session_get_stream_stat          @ 125 NONAME
	pjmedia_session_info_from_sdp            @ 126 NONAME
	pjmedia_session_pause                    @ 127 NONAME
	pjmedia_session_pause_stream             @ 128 NONAME
	pjmedia_session_resume                   @ 129 NONAME
	pjmedia_session_resume_stream            @ 130 NONAME
	pjmedia_session_set_dtmf_callback        @ 131 NONAME
	pjmedia_silence_det_apply                @ 132 NONAME
	pjmedia_silence_det_create               @ 133 NONAME
	pjmedia_silence_det_detect               @ 134 NONAME
	pjmedia_silence_det_disable              @ 135 NONAME
	pjmedia_silence_det_set_adaptive         @ 136 NONAME
	pjmedia_silence_det_set_fixed            @ 137 NONAME
	pjmedia_silence_det_set_name             @ 138 NONAME
	pjmedia_silence_det_set_params           @ 139 NONAME
	pjmedia_snd_port_connect                 @ 140 NONAME
	pjmedia_snd_port_create                  @ 141 NONAME
	pjmedia_snd_port_create_player           @ 142 NONAME
	pjmedia_snd_port_create_rec              @ 143 NONAME
	pjmedia_snd_port_destroy                 @ 144 NONAME
	pjmedia_snd_port_disconnect              @ 145 NONAME
	pjmedia_snd_port_get_ec_tail             @ 146 NONAME
	pjmedia_snd_port_get_port                @ 147 NONAME
	pjmedia_snd_port_get_snd_stream          @ 148 NONAME
	pjmedia_snd_port_set_ec                  @ 149 NONAME
	pjmedia_splitcomb_create                 @ 150 NONAME
	pjmedia_splitcomb_create_rev_channel     @ 151 NONAME
	pjmedia_splitcomb_set_channel            @ 152 NONAME
	pjmedia_stream_check_dtmf                @ 153 NONAME
	pjmedia_stream_create                    @ 154 NONAME
	pjmedia_stream_destroy                   @ 155 NONAME
	pjmedia_stream_dial_dtmf                 @ 156 NONAME
	pjmedia_stream_get_dtmf                  @ 157 NONAME
	pjmedia_stream_get_port                  @ 158 NONAME
	pjmedia_stream_get_stat                  @ 159 NONAME
	pjmedia_stream_get_transport             @ 160 NONAME
	pjmedia_stream_info_from_sdp             @ 161 NONAME
	pjmedia_stream_pause                     @ 162 NONAME
	pjmedia_stream_resume                    @ 163 NONAME
	pjmedia_stream_set_dtmf_callback         @ 164 NONAME
	pjmedia_stream_start                     @ 165 NONAME
	pjmedia_strerror                         @ 166 NONAME
	pjmedia_tonegen_create                   @ 167 NONAME
	pjmedia_tonegen_create2                  @ 168 NONAME
	pjmedia_tonegen_get_digit_map            @ 169 NONAME
	pjmedia_tonegen_is_busy                  @ 170 NONAME
	pjmedia_tonegen_play                     @ 171 NONAME
	pjmedia_tonegen_play_digits              @ 172 NONAME
	pjmedia_tonegen_set_digit_map            @ 173 NONAME
	pjmedia_tonegen_stop                     @ 174 NONAME
	pjmedia_transport_udp_attach             @ 175 NONAME
	pjmedia_transport_udp_close              @ 176 NONAME
	pjmedia_transport_udp_create             @ 177 NONAME
	pjmedia_transport_udp_create2            @ 178 NONAME
	pjmedia_transport_udp_create3            @ 179 NONAME
	pjmedia_transport_udp_get_info           @ 180 NONAME
	pjmedia_transport_udp_simulate_lost      @ 181 NONAME
	pjmedia_wav_player_port_create           @ 182 NONAME
	pjmedia_wav_player_port_get_pos          @ 183 NONAME
	pjmedia_wav_player_port_set_pos          @ 184 NONAME
	pjmedia_wav_player_set_eof_cb            @ 185 NONAME
	pjmedia_wav_playlist_create              @ 186 NONAME
	pjmedia_wav_playlist_set_eof_cb          @ 187 NONAME
	pjmedia_wav_writer_port_create           @ 188 NONAME
	pjmedia_wav_writer_port_get_pos          @ 189 NONAME
	pjmedia_wav_writer_port_set_cb           @ 190 NONAME
	pjmedia_wave_hdr_file_to_host            @ 191 NONAME
	pjmedia_wave_hdr_host_to_file            @ 192 NONAME
