#
import inc_const as const

PJSUA = ["--null-audio --max-calls=1 --no-tcp --auto-answer=183"]

PJSUA_EXPECTS = [[0, "Response msg 488/INVITE", ""]]

PJSUA_CLI_EXPECTS = [[0, "Response msg 488/INVITE", ""]]
