<?xml version="1.0" encoding="UTF-8" ?>
<ContentPage
    xmlns="http://xamarin.com/schemas/2014/forms"
    xmlns:x="http://schemas.microsoft.com/winfx/2009/xaml"
	xmlns:viewModels="clr-namespace:pjsua2xamarin"
    x:Class="pjsua2xamarin.AccountConfigPage"
    Title="Account Config">

    <ContentPage.BindingContext>
        <viewModels:AccountConfigViewModel/>
    </ContentPage.BindingContext>

    <ContentPage.Content>
        <StackLayout VerticalOptions="FillAndExpand">
            <StackLayout VerticalOptions="FillAndExpand"
                         Margin="10"
                         x:DataType="viewModels:AccountConfigViewModel">
                <Label Text="Id:"/>
                <Entry HorizontalOptions="FillAndExpand" Text="{Binding accCfg.idUri}"/>
                <Label Text="Registrar"/>
                <Entry HorizontalOptions="FillAndExpand" Text="{Binding accCfg.registrarUri}"/>
                <Label Text="Proxy"/>
                <Entry HorizontalOptions="FillAndExpand" Text="{Binding accCfg.proxy}"/>
                <Label Text="Username"/>
                <Entry HorizontalOptions="FillAndExpand" Text="{Binding accCfg.username}"/>
                <Label Text="Password"/>
                <Entry HorizontalOptions="FillAndExpand" Text="{Binding accCfg.password}"/>
            </StackLayout>
			<Grid
				ColumnDefinitions="50*, 50*"
				HorizontalOptions="FillAndExpand">

					<Button Text="Ok" Grid.Column="0" Clicked="Ok_Clicked"/>
					<Button Text="Cancel" Grid.Column="1" Clicked="Cancel_Clicked"/>

			</Grid>
        </StackLayout>
    </ContentPage.Content>
</ContentPage>

