/* 
 * Copyright (C) 2008-2011 Teluu Inc. (http://www.teluu.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

#pragma once

namespace PjsuaCLI
{
    namespace BackEnd
    {        
        ref class PjsuaCallback;

        public ref class Globals sealed
        {
        public:
            // Get the single instance of this class
            static property Globals^ Instance
            {
                Globals^ get();
            }
            
            // Get the callback singleton object
            property PjsuaCLI::BackEnd::PjsuaCallback^ PjsuaCallback
            {
                PjsuaCLI::BackEnd::PjsuaCallback^ get();
            }

            /* pjsua method */
            int pjsuaStart();
            void pjsuaDestroy();
            int pjsuaRestart();

        private:
            // Default constructor
            Globals();

            // Destructor
            ~Globals();

            static Globals^ singleton;

            PjsuaCLI::BackEnd::PjsuaCallback^ callback;
        };
    }
}
