#pragma warning disable 1591
//------------------------------------------------------------------------------
// <auto-generated>
//     This code was generated by a tool.
//     Runtime Version:4.0.30319.42000
//
//     Changes to this file may cause incorrect behavior and will be lost if
//     the code is regenerated.
// </auto-generated>
//------------------------------------------------------------------------------

[assembly: global::Android.Runtime.ResourceDesignerAttribute("pjsua2xamarin.Droid.Resource", IsApplication=true)]

namespace pjsua2xamarin.Droid
{
	
	
	[global::System.CodeDom.Compiler.GeneratedCodeAttribute("Xamarin.Android.Build.Tasks", "13.2.0.93")]
	public partial class Resource
	{
		
		static Resource()
		{
			global::Android.Runtime.ResourceIdManager.UpdateIdValues();
		}
		
		public static void UpdateIdValues()
		{
			global::Xamarin.Essentials.Resource.Attribute.alpha = global::pjsua2xamarin.Droid.Resource.Attribute.alpha;
			global::Xamarin.Essentials.Resource.Attribute.font = global::pjsua2xamarin.Droid.Resource.Attribute.font;
			global::Xamarin.Essentials.Resource.Attribute.fontProviderAuthority = global::pjsua2xamarin.Droid.Resource.Attribute.fontProviderAuthority;
			global::Xamarin.Essentials.Resource.Attribute.fontProviderCerts = global::pjsua2xamarin.Droid.Resource.Attribute.fontProviderCerts;
			global::Xamarin.Essentials.Resource.Attribute.fontProviderFetchStrategy = global::pjsua2xamarin.Droid.Resource.Attribute.fontProviderFetchStrategy;
			global::Xamarin.Essentials.Resource.Attribute.fontProviderFetchTimeout = global::pjsua2xamarin.Droid.Resource.Attribute.fontProviderFetchTimeout;
			global::Xamarin.Essentials.Resource.Attribute.fontProviderPackage = global::pjsua2xamarin.Droid.Resource.Attribute.fontProviderPackage;
			global::Xamarin.Essentials.Resource.Attribute.fontProviderQuery = global::pjsua2xamarin.Droid.Resource.Attribute.fontProviderQuery;
			global::Xamarin.Essentials.Resource.Attribute.fontStyle = global::pjsua2xamarin.Droid.Resource.Attribute.fontStyle;
			global::Xamarin.Essentials.Resource.Attribute.fontVariationSettings = global::pjsua2xamarin.Droid.Resource.Attribute.fontVariationSettings;
			global::Xamarin.Essentials.Resource.Attribute.fontWeight = global::pjsua2xamarin.Droid.Resource.Attribute.fontWeight;
			global::Xamarin.Essentials.Resource.Attribute.ttcIndex = global::pjsua2xamarin.Droid.Resource.Attribute.ttcIndex;
			global::Xamarin.Essentials.Resource.Color.androidx_core_ripple_material_light = global::pjsua2xamarin.Droid.Resource.Color.androidx_core_ripple_material_light;
			global::Xamarin.Essentials.Resource.Color.androidx_core_secondary_text_default_material_light = global::pjsua2xamarin.Droid.Resource.Color.androidx_core_secondary_text_default_material_light;
			global::Xamarin.Essentials.Resource.Color.browser_actions_bg_grey = global::pjsua2xamarin.Droid.Resource.Color.browser_actions_bg_grey;
			global::Xamarin.Essentials.Resource.Color.browser_actions_divider_color = global::pjsua2xamarin.Droid.Resource.Color.browser_actions_divider_color;
			global::Xamarin.Essentials.Resource.Color.browser_actions_text_color = global::pjsua2xamarin.Droid.Resource.Color.browser_actions_text_color;
			global::Xamarin.Essentials.Resource.Color.browser_actions_title_color = global::pjsua2xamarin.Droid.Resource.Color.browser_actions_title_color;
			global::Xamarin.Essentials.Resource.Color.notification_action_color_filter = global::pjsua2xamarin.Droid.Resource.Color.notification_action_color_filter;
			global::Xamarin.Essentials.Resource.Color.notification_icon_bg_color = global::pjsua2xamarin.Droid.Resource.Color.notification_icon_bg_color;
			global::Xamarin.Essentials.Resource.Dimension.browser_actions_context_menu_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.browser_actions_context_menu_max_width;
			global::Xamarin.Essentials.Resource.Dimension.browser_actions_context_menu_min_padding = global::pjsua2xamarin.Droid.Resource.Dimension.browser_actions_context_menu_min_padding;
			global::Xamarin.Essentials.Resource.Dimension.compat_button_inset_horizontal_material = global::pjsua2xamarin.Droid.Resource.Dimension.compat_button_inset_horizontal_material;
			global::Xamarin.Essentials.Resource.Dimension.compat_button_inset_vertical_material = global::pjsua2xamarin.Droid.Resource.Dimension.compat_button_inset_vertical_material;
			global::Xamarin.Essentials.Resource.Dimension.compat_button_padding_horizontal_material = global::pjsua2xamarin.Droid.Resource.Dimension.compat_button_padding_horizontal_material;
			global::Xamarin.Essentials.Resource.Dimension.compat_button_padding_vertical_material = global::pjsua2xamarin.Droid.Resource.Dimension.compat_button_padding_vertical_material;
			global::Xamarin.Essentials.Resource.Dimension.compat_control_corner_material = global::pjsua2xamarin.Droid.Resource.Dimension.compat_control_corner_material;
			global::Xamarin.Essentials.Resource.Dimension.compat_notification_large_icon_max_height = global::pjsua2xamarin.Droid.Resource.Dimension.compat_notification_large_icon_max_height;
			global::Xamarin.Essentials.Resource.Dimension.compat_notification_large_icon_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.compat_notification_large_icon_max_width;
			global::Xamarin.Essentials.Resource.Dimension.notification_action_icon_size = global::pjsua2xamarin.Droid.Resource.Dimension.notification_action_icon_size;
			global::Xamarin.Essentials.Resource.Dimension.notification_action_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.notification_action_text_size;
			global::Xamarin.Essentials.Resource.Dimension.notification_big_circle_margin = global::pjsua2xamarin.Droid.Resource.Dimension.notification_big_circle_margin;
			global::Xamarin.Essentials.Resource.Dimension.notification_content_margin_start = global::pjsua2xamarin.Droid.Resource.Dimension.notification_content_margin_start;
			global::Xamarin.Essentials.Resource.Dimension.notification_large_icon_height = global::pjsua2xamarin.Droid.Resource.Dimension.notification_large_icon_height;
			global::Xamarin.Essentials.Resource.Dimension.notification_large_icon_width = global::pjsua2xamarin.Droid.Resource.Dimension.notification_large_icon_width;
			global::Xamarin.Essentials.Resource.Dimension.notification_main_column_padding_top = global::pjsua2xamarin.Droid.Resource.Dimension.notification_main_column_padding_top;
			global::Xamarin.Essentials.Resource.Dimension.notification_media_narrow_margin = global::pjsua2xamarin.Droid.Resource.Dimension.notification_media_narrow_margin;
			global::Xamarin.Essentials.Resource.Dimension.notification_right_icon_size = global::pjsua2xamarin.Droid.Resource.Dimension.notification_right_icon_size;
			global::Xamarin.Essentials.Resource.Dimension.notification_right_side_padding_top = global::pjsua2xamarin.Droid.Resource.Dimension.notification_right_side_padding_top;
			global::Xamarin.Essentials.Resource.Dimension.notification_small_icon_background_padding = global::pjsua2xamarin.Droid.Resource.Dimension.notification_small_icon_background_padding;
			global::Xamarin.Essentials.Resource.Dimension.notification_small_icon_size_as_large = global::pjsua2xamarin.Droid.Resource.Dimension.notification_small_icon_size_as_large;
			global::Xamarin.Essentials.Resource.Dimension.notification_subtext_size = global::pjsua2xamarin.Droid.Resource.Dimension.notification_subtext_size;
			global::Xamarin.Essentials.Resource.Dimension.notification_top_pad = global::pjsua2xamarin.Droid.Resource.Dimension.notification_top_pad;
			global::Xamarin.Essentials.Resource.Dimension.notification_top_pad_large_text = global::pjsua2xamarin.Droid.Resource.Dimension.notification_top_pad_large_text;
			global::Xamarin.Essentials.Resource.Drawable.notification_action_background = global::pjsua2xamarin.Droid.Resource.Drawable.notification_action_background;
			global::Xamarin.Essentials.Resource.Drawable.notification_bg = global::pjsua2xamarin.Droid.Resource.Drawable.notification_bg;
			global::Xamarin.Essentials.Resource.Drawable.notification_bg_low = global::pjsua2xamarin.Droid.Resource.Drawable.notification_bg_low;
			global::Xamarin.Essentials.Resource.Drawable.notification_bg_low_normal = global::pjsua2xamarin.Droid.Resource.Drawable.notification_bg_low_normal;
			global::Xamarin.Essentials.Resource.Drawable.notification_bg_low_pressed = global::pjsua2xamarin.Droid.Resource.Drawable.notification_bg_low_pressed;
			global::Xamarin.Essentials.Resource.Drawable.notification_bg_normal = global::pjsua2xamarin.Droid.Resource.Drawable.notification_bg_normal;
			global::Xamarin.Essentials.Resource.Drawable.notification_bg_normal_pressed = global::pjsua2xamarin.Droid.Resource.Drawable.notification_bg_normal_pressed;
			global::Xamarin.Essentials.Resource.Drawable.notification_icon_background = global::pjsua2xamarin.Droid.Resource.Drawable.notification_icon_background;
			global::Xamarin.Essentials.Resource.Drawable.notification_template_icon_bg = global::pjsua2xamarin.Droid.Resource.Drawable.notification_template_icon_bg;
			global::Xamarin.Essentials.Resource.Drawable.notification_template_icon_low_bg = global::pjsua2xamarin.Droid.Resource.Drawable.notification_template_icon_low_bg;
			global::Xamarin.Essentials.Resource.Drawable.notification_tile_bg = global::pjsua2xamarin.Droid.Resource.Drawable.notification_tile_bg;
			global::Xamarin.Essentials.Resource.Drawable.notify_panel_notification_icon_bg = global::pjsua2xamarin.Droid.Resource.Drawable.notify_panel_notification_icon_bg;
			global::Xamarin.Essentials.Resource.Id.accessibility_action_clickable_span = global::pjsua2xamarin.Droid.Resource.Id.accessibility_action_clickable_span;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_0 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_0;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_1 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_1;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_10 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_10;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_11 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_11;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_12 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_12;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_13 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_13;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_14 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_14;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_15 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_15;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_16 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_16;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_17 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_17;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_18 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_18;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_19 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_19;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_2 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_2;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_20 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_20;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_21 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_21;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_22 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_22;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_23 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_23;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_24 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_24;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_25 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_25;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_26 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_26;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_27 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_27;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_28 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_28;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_29 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_29;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_3 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_3;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_30 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_30;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_31 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_31;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_4 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_4;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_5 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_5;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_6 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_6;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_7 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_7;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_8 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_8;
			global::Xamarin.Essentials.Resource.Id.accessibility_custom_action_9 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_9;
			global::Xamarin.Essentials.Resource.Id.actions = global::pjsua2xamarin.Droid.Resource.Id.actions;
			global::Xamarin.Essentials.Resource.Id.action_container = global::pjsua2xamarin.Droid.Resource.Id.action_container;
			global::Xamarin.Essentials.Resource.Id.action_divider = global::pjsua2xamarin.Droid.Resource.Id.action_divider;
			global::Xamarin.Essentials.Resource.Id.action_image = global::pjsua2xamarin.Droid.Resource.Id.action_image;
			global::Xamarin.Essentials.Resource.Id.action_text = global::pjsua2xamarin.Droid.Resource.Id.action_text;
			global::Xamarin.Essentials.Resource.Id.async = global::pjsua2xamarin.Droid.Resource.Id.async;
			global::Xamarin.Essentials.Resource.Id.blocking = global::pjsua2xamarin.Droid.Resource.Id.blocking;
			global::Xamarin.Essentials.Resource.Id.browser_actions_header_text = global::pjsua2xamarin.Droid.Resource.Id.browser_actions_header_text;
			global::Xamarin.Essentials.Resource.Id.browser_actions_menu_items = global::pjsua2xamarin.Droid.Resource.Id.browser_actions_menu_items;
			global::Xamarin.Essentials.Resource.Id.browser_actions_menu_item_icon = global::pjsua2xamarin.Droid.Resource.Id.browser_actions_menu_item_icon;
			global::Xamarin.Essentials.Resource.Id.browser_actions_menu_item_text = global::pjsua2xamarin.Droid.Resource.Id.browser_actions_menu_item_text;
			global::Xamarin.Essentials.Resource.Id.browser_actions_menu_view = global::pjsua2xamarin.Droid.Resource.Id.browser_actions_menu_view;
			global::Xamarin.Essentials.Resource.Id.chronometer = global::pjsua2xamarin.Droid.Resource.Id.chronometer;
			global::Xamarin.Essentials.Resource.Id.dialog_button = global::pjsua2xamarin.Droid.Resource.Id.dialog_button;
			global::Xamarin.Essentials.Resource.Id.forever = global::pjsua2xamarin.Droid.Resource.Id.forever;
			global::Xamarin.Essentials.Resource.Id.icon = global::pjsua2xamarin.Droid.Resource.Id.icon;
			global::Xamarin.Essentials.Resource.Id.icon_group = global::pjsua2xamarin.Droid.Resource.Id.icon_group;
			global::Xamarin.Essentials.Resource.Id.info = global::pjsua2xamarin.Droid.Resource.Id.info;
			global::Xamarin.Essentials.Resource.Id.italic = global::pjsua2xamarin.Droid.Resource.Id.italic;
			global::Xamarin.Essentials.Resource.Id.line1 = global::pjsua2xamarin.Droid.Resource.Id.line1;
			global::Xamarin.Essentials.Resource.Id.line3 = global::pjsua2xamarin.Droid.Resource.Id.line3;
			global::Xamarin.Essentials.Resource.Id.normal = global::pjsua2xamarin.Droid.Resource.Id.normal;
			global::Xamarin.Essentials.Resource.Id.notification_background = global::pjsua2xamarin.Droid.Resource.Id.notification_background;
			global::Xamarin.Essentials.Resource.Id.notification_main_column = global::pjsua2xamarin.Droid.Resource.Id.notification_main_column;
			global::Xamarin.Essentials.Resource.Id.notification_main_column_container = global::pjsua2xamarin.Droid.Resource.Id.notification_main_column_container;
			global::Xamarin.Essentials.Resource.Id.right_icon = global::pjsua2xamarin.Droid.Resource.Id.right_icon;
			global::Xamarin.Essentials.Resource.Id.right_side = global::pjsua2xamarin.Droid.Resource.Id.right_side;
			global::Xamarin.Essentials.Resource.Id.tag_accessibility_actions = global::pjsua2xamarin.Droid.Resource.Id.tag_accessibility_actions;
			global::Xamarin.Essentials.Resource.Id.tag_accessibility_clickable_spans = global::pjsua2xamarin.Droid.Resource.Id.tag_accessibility_clickable_spans;
			global::Xamarin.Essentials.Resource.Id.tag_accessibility_heading = global::pjsua2xamarin.Droid.Resource.Id.tag_accessibility_heading;
			global::Xamarin.Essentials.Resource.Id.tag_accessibility_pane_title = global::pjsua2xamarin.Droid.Resource.Id.tag_accessibility_pane_title;
			global::Xamarin.Essentials.Resource.Id.tag_screen_reader_focusable = global::pjsua2xamarin.Droid.Resource.Id.tag_screen_reader_focusable;
			global::Xamarin.Essentials.Resource.Id.tag_transition_group = global::pjsua2xamarin.Droid.Resource.Id.tag_transition_group;
			global::Xamarin.Essentials.Resource.Id.tag_unhandled_key_event_manager = global::pjsua2xamarin.Droid.Resource.Id.tag_unhandled_key_event_manager;
			global::Xamarin.Essentials.Resource.Id.tag_unhandled_key_listeners = global::pjsua2xamarin.Droid.Resource.Id.tag_unhandled_key_listeners;
			global::Xamarin.Essentials.Resource.Id.text = global::pjsua2xamarin.Droid.Resource.Id.text;
			global::Xamarin.Essentials.Resource.Id.text2 = global::pjsua2xamarin.Droid.Resource.Id.text2;
			global::Xamarin.Essentials.Resource.Id.time = global::pjsua2xamarin.Droid.Resource.Id.time;
			global::Xamarin.Essentials.Resource.Id.title = global::pjsua2xamarin.Droid.Resource.Id.title;
			global::Xamarin.Essentials.Resource.Id.view_tree_lifecycle_owner = global::pjsua2xamarin.Droid.Resource.Id.view_tree_lifecycle_owner;
			global::Xamarin.Essentials.Resource.Integer.status_bar_notification_info_maxnum = global::pjsua2xamarin.Droid.Resource.Integer.status_bar_notification_info_maxnum;
			global::Xamarin.Essentials.Resource.Layout.browser_actions_context_menu_page = global::pjsua2xamarin.Droid.Resource.Layout.browser_actions_context_menu_page;
			global::Xamarin.Essentials.Resource.Layout.browser_actions_context_menu_row = global::pjsua2xamarin.Droid.Resource.Layout.browser_actions_context_menu_row;
			global::Xamarin.Essentials.Resource.Layout.custom_dialog = global::pjsua2xamarin.Droid.Resource.Layout.custom_dialog;
			global::Xamarin.Essentials.Resource.Layout.notification_action = global::pjsua2xamarin.Droid.Resource.Layout.notification_action;
			global::Xamarin.Essentials.Resource.Layout.notification_action_tombstone = global::pjsua2xamarin.Droid.Resource.Layout.notification_action_tombstone;
			global::Xamarin.Essentials.Resource.Layout.notification_template_custom_big = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_custom_big;
			global::Xamarin.Essentials.Resource.Layout.notification_template_icon_group = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_icon_group;
			global::Xamarin.Essentials.Resource.Layout.notification_template_part_chronometer = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_part_chronometer;
			global::Xamarin.Essentials.Resource.Layout.notification_template_part_time = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_part_time;
			global::Xamarin.Essentials.Resource.String.copy_toast_msg = global::pjsua2xamarin.Droid.Resource.String.copy_toast_msg;
			global::Xamarin.Essentials.Resource.String.fallback_menu_item_copy_link = global::pjsua2xamarin.Droid.Resource.String.fallback_menu_item_copy_link;
			global::Xamarin.Essentials.Resource.String.fallback_menu_item_open_in_browser = global::pjsua2xamarin.Droid.Resource.String.fallback_menu_item_open_in_browser;
			global::Xamarin.Essentials.Resource.String.fallback_menu_item_share_link = global::pjsua2xamarin.Droid.Resource.String.fallback_menu_item_share_link;
			global::Xamarin.Essentials.Resource.String.status_bar_notification_info_overflow = global::pjsua2xamarin.Droid.Resource.String.status_bar_notification_info_overflow;
			global::Xamarin.Essentials.Resource.Style.TextAppearance_Compat_Notification = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification;
			global::Xamarin.Essentials.Resource.Style.TextAppearance_Compat_Notification_Info = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification_Info;
			global::Xamarin.Essentials.Resource.Style.TextAppearance_Compat_Notification_Line2 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification_Line2;
			global::Xamarin.Essentials.Resource.Style.TextAppearance_Compat_Notification_Time = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification_Time;
			global::Xamarin.Essentials.Resource.Style.TextAppearance_Compat_Notification_Title = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification_Title;
			global::Xamarin.Essentials.Resource.Style.Widget_Compat_NotificationActionContainer = global::pjsua2xamarin.Droid.Resource.Style.Widget_Compat_NotificationActionContainer;
			global::Xamarin.Essentials.Resource.Style.Widget_Compat_NotificationActionText = global::pjsua2xamarin.Droid.Resource.Style.Widget_Compat_NotificationActionText;
			global::Xamarin.Essentials.Resource.Styleable.ColorStateListItem = global::pjsua2xamarin.Droid.Resource.Styleable.ColorStateListItem;
			global::Xamarin.Essentials.Resource.Styleable.ColorStateListItem_alpha = global::pjsua2xamarin.Droid.Resource.Styleable.ColorStateListItem_alpha;
			global::Xamarin.Essentials.Resource.Styleable.ColorStateListItem_android_alpha = global::pjsua2xamarin.Droid.Resource.Styleable.ColorStateListItem_android_alpha;
			global::Xamarin.Essentials.Resource.Styleable.ColorStateListItem_android_color = global::pjsua2xamarin.Droid.Resource.Styleable.ColorStateListItem_android_color;
			global::Xamarin.Essentials.Resource.Styleable.FontFamily = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily;
			global::Xamarin.Essentials.Resource.Styleable.FontFamilyFont = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont;
			global::Xamarin.Essentials.Resource.Styleable.FontFamilyFont_android_font = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_android_font;
			global::Xamarin.Essentials.Resource.Styleable.FontFamilyFont_android_fontStyle = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_android_fontStyle;
			global::Xamarin.Essentials.Resource.Styleable.FontFamilyFont_android_fontVariationSettings = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_android_fontVariationSettings;
			global::Xamarin.Essentials.Resource.Styleable.FontFamilyFont_android_fontWeight = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_android_fontWeight;
			global::Xamarin.Essentials.Resource.Styleable.FontFamilyFont_android_ttcIndex = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_android_ttcIndex;
			global::Xamarin.Essentials.Resource.Styleable.FontFamilyFont_font = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_font;
			global::Xamarin.Essentials.Resource.Styleable.FontFamilyFont_fontStyle = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_fontStyle;
			global::Xamarin.Essentials.Resource.Styleable.FontFamilyFont_fontVariationSettings = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_fontVariationSettings;
			global::Xamarin.Essentials.Resource.Styleable.FontFamilyFont_fontWeight = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_fontWeight;
			global::Xamarin.Essentials.Resource.Styleable.FontFamilyFont_ttcIndex = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_ttcIndex;
			global::Xamarin.Essentials.Resource.Styleable.FontFamily_fontProviderAuthority = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily_fontProviderAuthority;
			global::Xamarin.Essentials.Resource.Styleable.FontFamily_fontProviderCerts = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily_fontProviderCerts;
			global::Xamarin.Essentials.Resource.Styleable.FontFamily_fontProviderFetchStrategy = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily_fontProviderFetchStrategy;
			global::Xamarin.Essentials.Resource.Styleable.FontFamily_fontProviderFetchTimeout = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily_fontProviderFetchTimeout;
			global::Xamarin.Essentials.Resource.Styleable.FontFamily_fontProviderPackage = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily_fontProviderPackage;
			global::Xamarin.Essentials.Resource.Styleable.FontFamily_fontProviderQuery = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily_fontProviderQuery;
			global::Xamarin.Essentials.Resource.Styleable.GradientColor = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor;
			global::Xamarin.Essentials.Resource.Styleable.GradientColorItem = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColorItem;
			global::Xamarin.Essentials.Resource.Styleable.GradientColorItem_android_color = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColorItem_android_color;
			global::Xamarin.Essentials.Resource.Styleable.GradientColorItem_android_offset = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColorItem_android_offset;
			global::Xamarin.Essentials.Resource.Styleable.GradientColor_android_centerColor = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_centerColor;
			global::Xamarin.Essentials.Resource.Styleable.GradientColor_android_centerX = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_centerX;
			global::Xamarin.Essentials.Resource.Styleable.GradientColor_android_centerY = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_centerY;
			global::Xamarin.Essentials.Resource.Styleable.GradientColor_android_endColor = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_endColor;
			global::Xamarin.Essentials.Resource.Styleable.GradientColor_android_endX = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_endX;
			global::Xamarin.Essentials.Resource.Styleable.GradientColor_android_endY = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_endY;
			global::Xamarin.Essentials.Resource.Styleable.GradientColor_android_gradientRadius = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_gradientRadius;
			global::Xamarin.Essentials.Resource.Styleable.GradientColor_android_startColor = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_startColor;
			global::Xamarin.Essentials.Resource.Styleable.GradientColor_android_startX = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_startX;
			global::Xamarin.Essentials.Resource.Styleable.GradientColor_android_startY = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_startY;
			global::Xamarin.Essentials.Resource.Styleable.GradientColor_android_tileMode = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_tileMode;
			global::Xamarin.Essentials.Resource.Styleable.GradientColor_android_type = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_type;
			global::Xamarin.Essentials.Resource.Xml.image_share_filepaths = global::pjsua2xamarin.Droid.Resource.Xml.image_share_filepaths;
			global::Xamarin.Essentials.Resource.Xml.xamarin_essentials_fileprovider_file_paths = global::pjsua2xamarin.Droid.Resource.Xml.xamarin_essentials_fileprovider_file_paths;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.abc_fade_in = global::pjsua2xamarin.Droid.Resource.Animation.abc_fade_in;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.abc_fade_out = global::pjsua2xamarin.Droid.Resource.Animation.abc_fade_out;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.abc_grow_fade_in_from_bottom = global::pjsua2xamarin.Droid.Resource.Animation.abc_grow_fade_in_from_bottom;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.abc_popup_enter = global::pjsua2xamarin.Droid.Resource.Animation.abc_popup_enter;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.abc_popup_exit = global::pjsua2xamarin.Droid.Resource.Animation.abc_popup_exit;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.abc_shrink_fade_out_from_bottom = global::pjsua2xamarin.Droid.Resource.Animation.abc_shrink_fade_out_from_bottom;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.abc_slide_in_bottom = global::pjsua2xamarin.Droid.Resource.Animation.abc_slide_in_bottom;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.abc_slide_in_top = global::pjsua2xamarin.Droid.Resource.Animation.abc_slide_in_top;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.abc_slide_out_bottom = global::pjsua2xamarin.Droid.Resource.Animation.abc_slide_out_bottom;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.abc_slide_out_top = global::pjsua2xamarin.Droid.Resource.Animation.abc_slide_out_top;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.abc_tooltip_enter = global::pjsua2xamarin.Droid.Resource.Animation.abc_tooltip_enter;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.abc_tooltip_exit = global::pjsua2xamarin.Droid.Resource.Animation.abc_tooltip_exit;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.btn_checkbox_to_checked_box_inner_merged_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_checkbox_to_checked_box_inner_merged_animation;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.btn_checkbox_to_checked_box_outer_merged_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_checkbox_to_checked_box_outer_merged_animation;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.btn_checkbox_to_checked_icon_null_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_checkbox_to_checked_icon_null_animation;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.btn_checkbox_to_unchecked_box_inner_merged_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_checkbox_to_unchecked_box_inner_merged_animation;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.btn_checkbox_to_unchecked_check_path_merged_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_checkbox_to_unchecked_check_path_merged_animation;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.btn_checkbox_to_unchecked_icon_null_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_checkbox_to_unchecked_icon_null_animation;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.btn_radio_to_off_mtrl_dot_group_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_radio_to_off_mtrl_dot_group_animation;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.btn_radio_to_off_mtrl_ring_outer_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_radio_to_off_mtrl_ring_outer_animation;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.btn_radio_to_off_mtrl_ring_outer_path_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_radio_to_off_mtrl_ring_outer_path_animation;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.btn_radio_to_on_mtrl_dot_group_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_radio_to_on_mtrl_dot_group_animation;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.btn_radio_to_on_mtrl_ring_outer_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_radio_to_on_mtrl_ring_outer_animation;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.btn_radio_to_on_mtrl_ring_outer_path_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_radio_to_on_mtrl_ring_outer_path_animation;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.design_bottom_sheet_slide_in = global::pjsua2xamarin.Droid.Resource.Animation.design_bottom_sheet_slide_in;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.design_bottom_sheet_slide_out = global::pjsua2xamarin.Droid.Resource.Animation.design_bottom_sheet_slide_out;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.design_snackbar_in = global::pjsua2xamarin.Droid.Resource.Animation.design_snackbar_in;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.design_snackbar_out = global::pjsua2xamarin.Droid.Resource.Animation.design_snackbar_out;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.EnterFromLeft = global::pjsua2xamarin.Droid.Resource.Animation.EnterFromLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.EnterFromRight = global::pjsua2xamarin.Droid.Resource.Animation.EnterFromRight;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.ExitToLeft = global::pjsua2xamarin.Droid.Resource.Animation.ExitToLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.ExitToRight = global::pjsua2xamarin.Droid.Resource.Animation.ExitToRight;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.fragment_fast_out_extra_slow_in = global::pjsua2xamarin.Droid.Resource.Animation.fragment_fast_out_extra_slow_in;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.mtrl_bottom_sheet_slide_in = global::pjsua2xamarin.Droid.Resource.Animation.mtrl_bottom_sheet_slide_in;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.mtrl_bottom_sheet_slide_out = global::pjsua2xamarin.Droid.Resource.Animation.mtrl_bottom_sheet_slide_out;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.mtrl_card_lowers_interpolator = global::pjsua2xamarin.Droid.Resource.Animation.mtrl_card_lowers_interpolator;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.nav_default_enter_anim = global::pjsua2xamarin.Droid.Resource.Animation.nav_default_enter_anim;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.nav_default_exit_anim = global::pjsua2xamarin.Droid.Resource.Animation.nav_default_exit_anim;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.nav_default_pop_enter_anim = global::pjsua2xamarin.Droid.Resource.Animation.nav_default_pop_enter_anim;
			global::Xamarin.Forms.Platform.Android.Resource.Animation.nav_default_pop_exit_anim = global::pjsua2xamarin.Droid.Resource.Animation.nav_default_pop_exit_anim;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.design_appbar_state_list_animator = global::pjsua2xamarin.Droid.Resource.Animator.design_appbar_state_list_animator;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.design_fab_hide_motion_spec = global::pjsua2xamarin.Droid.Resource.Animator.design_fab_hide_motion_spec;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.design_fab_show_motion_spec = global::pjsua2xamarin.Droid.Resource.Animator.design_fab_show_motion_spec;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.fragment_close_enter = global::pjsua2xamarin.Droid.Resource.Animator.fragment_close_enter;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.fragment_close_exit = global::pjsua2xamarin.Droid.Resource.Animator.fragment_close_exit;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.fragment_fade_enter = global::pjsua2xamarin.Droid.Resource.Animator.fragment_fade_enter;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.fragment_fade_exit = global::pjsua2xamarin.Droid.Resource.Animator.fragment_fade_exit;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.fragment_open_enter = global::pjsua2xamarin.Droid.Resource.Animator.fragment_open_enter;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.fragment_open_exit = global::pjsua2xamarin.Droid.Resource.Animator.fragment_open_exit;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.linear_indeterminate_line1_head_interpolator = global::pjsua2xamarin.Droid.Resource.Animator.linear_indeterminate_line1_head_interpolator;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.linear_indeterminate_line1_tail_interpolator = global::pjsua2xamarin.Droid.Resource.Animator.linear_indeterminate_line1_tail_interpolator;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.linear_indeterminate_line2_head_interpolator = global::pjsua2xamarin.Droid.Resource.Animator.linear_indeterminate_line2_head_interpolator;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.linear_indeterminate_line2_tail_interpolator = global::pjsua2xamarin.Droid.Resource.Animator.linear_indeterminate_line2_tail_interpolator;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.mtrl_btn_state_list_anim = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_btn_state_list_anim;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.mtrl_btn_unelevated_state_list_anim = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_btn_unelevated_state_list_anim;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.mtrl_card_state_list_anim = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_card_state_list_anim;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.mtrl_chip_state_list_anim = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_chip_state_list_anim;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.mtrl_extended_fab_change_size_collapse_motion_spec = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_extended_fab_change_size_collapse_motion_spec;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.mtrl_extended_fab_change_size_expand_motion_spec = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_extended_fab_change_size_expand_motion_spec;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.mtrl_extended_fab_hide_motion_spec = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_extended_fab_hide_motion_spec;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.mtrl_extended_fab_show_motion_spec = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_extended_fab_show_motion_spec;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.mtrl_extended_fab_state_list_animator = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_extended_fab_state_list_animator;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.mtrl_fab_hide_motion_spec = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_fab_hide_motion_spec;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.mtrl_fab_show_motion_spec = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_fab_show_motion_spec;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.mtrl_fab_transformation_sheet_collapse_spec = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_fab_transformation_sheet_collapse_spec;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.mtrl_fab_transformation_sheet_expand_spec = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_fab_transformation_sheet_expand_spec;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.nav_default_enter_anim = global::pjsua2xamarin.Droid.Resource.Animator.nav_default_enter_anim;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.nav_default_exit_anim = global::pjsua2xamarin.Droid.Resource.Animator.nav_default_exit_anim;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.nav_default_pop_enter_anim = global::pjsua2xamarin.Droid.Resource.Animator.nav_default_pop_enter_anim;
			global::Xamarin.Forms.Platform.Android.Resource.Animator.nav_default_pop_exit_anim = global::pjsua2xamarin.Droid.Resource.Animator.nav_default_pop_exit_anim;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.action = global::pjsua2xamarin.Droid.Resource.Attribute.action;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionBarDivider = global::pjsua2xamarin.Droid.Resource.Attribute.actionBarDivider;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionBarItemBackground = global::pjsua2xamarin.Droid.Resource.Attribute.actionBarItemBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionBarPopupTheme = global::pjsua2xamarin.Droid.Resource.Attribute.actionBarPopupTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionBarSize = global::pjsua2xamarin.Droid.Resource.Attribute.actionBarSize;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionBarSplitStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionBarSplitStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionBarStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionBarStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionBarTabBarStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionBarTabBarStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionBarTabStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionBarTabStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionBarTabTextStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionBarTabTextStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionBarTheme = global::pjsua2xamarin.Droid.Resource.Attribute.actionBarTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionBarWidgetTheme = global::pjsua2xamarin.Droid.Resource.Attribute.actionBarWidgetTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionDropDownStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionDropDownStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionLayout = global::pjsua2xamarin.Droid.Resource.Attribute.actionLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionMenuTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.actionMenuTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionMenuTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.actionMenuTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionModeBackground = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionModeCloseButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeCloseButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionModeCloseContentDescription = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeCloseContentDescription;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionModeCloseDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeCloseDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionModeCopyDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeCopyDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionModeCutDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeCutDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionModeFindDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeFindDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionModePasteDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.actionModePasteDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionModePopupWindowStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionModePopupWindowStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionModeSelectAllDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeSelectAllDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionModeShareDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeShareDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionModeSplitBackground = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeSplitBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionModeStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionModeTheme = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionModeWebSearchDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeWebSearchDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionOverflowButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionOverflowButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionOverflowMenuStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionOverflowMenuStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionProviderClass = global::pjsua2xamarin.Droid.Resource.Attribute.actionProviderClass;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionTextColorAlpha = global::pjsua2xamarin.Droid.Resource.Attribute.actionTextColorAlpha;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.actionViewClass = global::pjsua2xamarin.Droid.Resource.Attribute.actionViewClass;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.activityChooserViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.activityChooserViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.alertDialogButtonGroupStyle = global::pjsua2xamarin.Droid.Resource.Attribute.alertDialogButtonGroupStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.alertDialogCenterButtons = global::pjsua2xamarin.Droid.Resource.Attribute.alertDialogCenterButtons;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.alertDialogStyle = global::pjsua2xamarin.Droid.Resource.Attribute.alertDialogStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.alertDialogTheme = global::pjsua2xamarin.Droid.Resource.Attribute.alertDialogTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.allowStacking = global::pjsua2xamarin.Droid.Resource.Attribute.allowStacking;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.alpha = global::pjsua2xamarin.Droid.Resource.Attribute.alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.alphabeticModifiers = global::pjsua2xamarin.Droid.Resource.Attribute.alphabeticModifiers;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.altSrc = global::pjsua2xamarin.Droid.Resource.Attribute.altSrc;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.animateCircleAngleTo = global::pjsua2xamarin.Droid.Resource.Attribute.animateCircleAngleTo;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.animateRelativeTo = global::pjsua2xamarin.Droid.Resource.Attribute.animateRelativeTo;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.animationMode = global::pjsua2xamarin.Droid.Resource.Attribute.animationMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.appBarLayoutStyle = global::pjsua2xamarin.Droid.Resource.Attribute.appBarLayoutStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.applyMotionScene = global::pjsua2xamarin.Droid.Resource.Attribute.applyMotionScene;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.arcMode = global::pjsua2xamarin.Droid.Resource.Attribute.arcMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.argType = global::pjsua2xamarin.Droid.Resource.Attribute.argType;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.arrowHeadLength = global::pjsua2xamarin.Droid.Resource.Attribute.arrowHeadLength;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.arrowShaftLength = global::pjsua2xamarin.Droid.Resource.Attribute.arrowShaftLength;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.attributeName = global::pjsua2xamarin.Droid.Resource.Attribute.attributeName;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.autoCompleteMode = global::pjsua2xamarin.Droid.Resource.Attribute.autoCompleteMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.autoCompleteTextViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.autoCompleteTextViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.autoSizeMaxTextSize = global::pjsua2xamarin.Droid.Resource.Attribute.autoSizeMaxTextSize;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.autoSizeMinTextSize = global::pjsua2xamarin.Droid.Resource.Attribute.autoSizeMinTextSize;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.autoSizePresetSizes = global::pjsua2xamarin.Droid.Resource.Attribute.autoSizePresetSizes;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.autoSizeStepGranularity = global::pjsua2xamarin.Droid.Resource.Attribute.autoSizeStepGranularity;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.autoSizeTextType = global::pjsua2xamarin.Droid.Resource.Attribute.autoSizeTextType;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.autoTransition = global::pjsua2xamarin.Droid.Resource.Attribute.autoTransition;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.background = global::pjsua2xamarin.Droid.Resource.Attribute.background;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.backgroundColor = global::pjsua2xamarin.Droid.Resource.Attribute.backgroundColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.backgroundInsetBottom = global::pjsua2xamarin.Droid.Resource.Attribute.backgroundInsetBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.backgroundInsetEnd = global::pjsua2xamarin.Droid.Resource.Attribute.backgroundInsetEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.backgroundInsetStart = global::pjsua2xamarin.Droid.Resource.Attribute.backgroundInsetStart;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.backgroundInsetTop = global::pjsua2xamarin.Droid.Resource.Attribute.backgroundInsetTop;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.backgroundOverlayColorAlpha = global::pjsua2xamarin.Droid.Resource.Attribute.backgroundOverlayColorAlpha;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.backgroundSplit = global::pjsua2xamarin.Droid.Resource.Attribute.backgroundSplit;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.backgroundStacked = global::pjsua2xamarin.Droid.Resource.Attribute.backgroundStacked;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.backgroundTint = global::pjsua2xamarin.Droid.Resource.Attribute.backgroundTint;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.backgroundTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.backgroundTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.badgeGravity = global::pjsua2xamarin.Droid.Resource.Attribute.badgeGravity;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.badgeStyle = global::pjsua2xamarin.Droid.Resource.Attribute.badgeStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.badgeTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.badgeTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.barLength = global::pjsua2xamarin.Droid.Resource.Attribute.barLength;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.barrierAllowsGoneWidgets = global::pjsua2xamarin.Droid.Resource.Attribute.barrierAllowsGoneWidgets;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.barrierDirection = global::pjsua2xamarin.Droid.Resource.Attribute.barrierDirection;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.barrierMargin = global::pjsua2xamarin.Droid.Resource.Attribute.barrierMargin;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.behavior_autoHide = global::pjsua2xamarin.Droid.Resource.Attribute.behavior_autoHide;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.behavior_autoShrink = global::pjsua2xamarin.Droid.Resource.Attribute.behavior_autoShrink;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.behavior_draggable = global::pjsua2xamarin.Droid.Resource.Attribute.behavior_draggable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.behavior_expandedOffset = global::pjsua2xamarin.Droid.Resource.Attribute.behavior_expandedOffset;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.behavior_fitToContents = global::pjsua2xamarin.Droid.Resource.Attribute.behavior_fitToContents;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.behavior_halfExpandedRatio = global::pjsua2xamarin.Droid.Resource.Attribute.behavior_halfExpandedRatio;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.behavior_hideable = global::pjsua2xamarin.Droid.Resource.Attribute.behavior_hideable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.behavior_overlapTop = global::pjsua2xamarin.Droid.Resource.Attribute.behavior_overlapTop;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.behavior_peekHeight = global::pjsua2xamarin.Droid.Resource.Attribute.behavior_peekHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.behavior_saveFlags = global::pjsua2xamarin.Droid.Resource.Attribute.behavior_saveFlags;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.behavior_skipCollapsed = global::pjsua2xamarin.Droid.Resource.Attribute.behavior_skipCollapsed;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.blendSrc = global::pjsua2xamarin.Droid.Resource.Attribute.blendSrc;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.borderlessButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.borderlessButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.borderRound = global::pjsua2xamarin.Droid.Resource.Attribute.borderRound;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.borderRoundPercent = global::pjsua2xamarin.Droid.Resource.Attribute.borderRoundPercent;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.borderWidth = global::pjsua2xamarin.Droid.Resource.Attribute.borderWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.bottomAppBarStyle = global::pjsua2xamarin.Droid.Resource.Attribute.bottomAppBarStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.bottomNavigationStyle = global::pjsua2xamarin.Droid.Resource.Attribute.bottomNavigationStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.bottomSheetDialogTheme = global::pjsua2xamarin.Droid.Resource.Attribute.bottomSheetDialogTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.bottomSheetStyle = global::pjsua2xamarin.Droid.Resource.Attribute.bottomSheetStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.boxBackgroundColor = global::pjsua2xamarin.Droid.Resource.Attribute.boxBackgroundColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.boxBackgroundMode = global::pjsua2xamarin.Droid.Resource.Attribute.boxBackgroundMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.boxCollapsedPaddingTop = global::pjsua2xamarin.Droid.Resource.Attribute.boxCollapsedPaddingTop;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.boxCornerRadiusBottomEnd = global::pjsua2xamarin.Droid.Resource.Attribute.boxCornerRadiusBottomEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.boxCornerRadiusBottomStart = global::pjsua2xamarin.Droid.Resource.Attribute.boxCornerRadiusBottomStart;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.boxCornerRadiusTopEnd = global::pjsua2xamarin.Droid.Resource.Attribute.boxCornerRadiusTopEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.boxCornerRadiusTopStart = global::pjsua2xamarin.Droid.Resource.Attribute.boxCornerRadiusTopStart;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.boxStrokeColor = global::pjsua2xamarin.Droid.Resource.Attribute.boxStrokeColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.boxStrokeErrorColor = global::pjsua2xamarin.Droid.Resource.Attribute.boxStrokeErrorColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.boxStrokeWidth = global::pjsua2xamarin.Droid.Resource.Attribute.boxStrokeWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.boxStrokeWidthFocused = global::pjsua2xamarin.Droid.Resource.Attribute.boxStrokeWidthFocused;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.brightness = global::pjsua2xamarin.Droid.Resource.Attribute.brightness;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.buttonBarButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.buttonBarButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.buttonBarNegativeButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.buttonBarNegativeButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.buttonBarNeutralButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.buttonBarNeutralButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.buttonBarPositiveButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.buttonBarPositiveButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.buttonBarStyle = global::pjsua2xamarin.Droid.Resource.Attribute.buttonBarStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.buttonCompat = global::pjsua2xamarin.Droid.Resource.Attribute.buttonCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.buttonGravity = global::pjsua2xamarin.Droid.Resource.Attribute.buttonGravity;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.buttonIconDimen = global::pjsua2xamarin.Droid.Resource.Attribute.buttonIconDimen;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.buttonPanelSideLayout = global::pjsua2xamarin.Droid.Resource.Attribute.buttonPanelSideLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.buttonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.buttonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.buttonStyleSmall = global::pjsua2xamarin.Droid.Resource.Attribute.buttonStyleSmall;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.buttonTint = global::pjsua2xamarin.Droid.Resource.Attribute.buttonTint;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.buttonTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.buttonTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.cardBackgroundColor = global::pjsua2xamarin.Droid.Resource.Attribute.cardBackgroundColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.cardCornerRadius = global::pjsua2xamarin.Droid.Resource.Attribute.cardCornerRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.cardElevation = global::pjsua2xamarin.Droid.Resource.Attribute.cardElevation;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.cardForegroundColor = global::pjsua2xamarin.Droid.Resource.Attribute.cardForegroundColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.cardMaxElevation = global::pjsua2xamarin.Droid.Resource.Attribute.cardMaxElevation;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.cardPreventCornerOverlap = global::pjsua2xamarin.Droid.Resource.Attribute.cardPreventCornerOverlap;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.cardUseCompatPadding = global::pjsua2xamarin.Droid.Resource.Attribute.cardUseCompatPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.cardViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.cardViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.carousel_backwardTransition = global::pjsua2xamarin.Droid.Resource.Attribute.carousel_backwardTransition;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.carousel_emptyViewsBehavior = global::pjsua2xamarin.Droid.Resource.Attribute.carousel_emptyViewsBehavior;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.carousel_firstView = global::pjsua2xamarin.Droid.Resource.Attribute.carousel_firstView;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.carousel_forwardTransition = global::pjsua2xamarin.Droid.Resource.Attribute.carousel_forwardTransition;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.carousel_infinite = global::pjsua2xamarin.Droid.Resource.Attribute.carousel_infinite;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.carousel_nextState = global::pjsua2xamarin.Droid.Resource.Attribute.carousel_nextState;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.carousel_previousState = global::pjsua2xamarin.Droid.Resource.Attribute.carousel_previousState;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.carousel_touchUpMode = global::pjsua2xamarin.Droid.Resource.Attribute.carousel_touchUpMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.carousel_touchUp_dampeningFactor = global::pjsua2xamarin.Droid.Resource.Attribute.carousel_touchUp_dampeningFactor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.carousel_touchUp_velocityThreshold = global::pjsua2xamarin.Droid.Resource.Attribute.carousel_touchUp_velocityThreshold;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.chainUseRtl = global::pjsua2xamarin.Droid.Resource.Attribute.chainUseRtl;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.checkboxStyle = global::pjsua2xamarin.Droid.Resource.Attribute.checkboxStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.checkedButton = global::pjsua2xamarin.Droid.Resource.Attribute.checkedButton;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.checkedChip = global::pjsua2xamarin.Droid.Resource.Attribute.checkedChip;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.checkedIcon = global::pjsua2xamarin.Droid.Resource.Attribute.checkedIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.checkedIconEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.checkedIconEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.checkedIconMargin = global::pjsua2xamarin.Droid.Resource.Attribute.checkedIconMargin;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.checkedIconSize = global::pjsua2xamarin.Droid.Resource.Attribute.checkedIconSize;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.checkedIconTint = global::pjsua2xamarin.Droid.Resource.Attribute.checkedIconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.checkedIconVisible = global::pjsua2xamarin.Droid.Resource.Attribute.checkedIconVisible;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.checkedTextViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.checkedTextViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.chipBackgroundColor = global::pjsua2xamarin.Droid.Resource.Attribute.chipBackgroundColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.chipCornerRadius = global::pjsua2xamarin.Droid.Resource.Attribute.chipCornerRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.chipEndPadding = global::pjsua2xamarin.Droid.Resource.Attribute.chipEndPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.chipGroupStyle = global::pjsua2xamarin.Droid.Resource.Attribute.chipGroupStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.chipIcon = global::pjsua2xamarin.Droid.Resource.Attribute.chipIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.chipIconEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.chipIconEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.chipIconSize = global::pjsua2xamarin.Droid.Resource.Attribute.chipIconSize;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.chipIconTint = global::pjsua2xamarin.Droid.Resource.Attribute.chipIconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.chipIconVisible = global::pjsua2xamarin.Droid.Resource.Attribute.chipIconVisible;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.chipMinHeight = global::pjsua2xamarin.Droid.Resource.Attribute.chipMinHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.chipMinTouchTargetSize = global::pjsua2xamarin.Droid.Resource.Attribute.chipMinTouchTargetSize;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.chipSpacing = global::pjsua2xamarin.Droid.Resource.Attribute.chipSpacing;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.chipSpacingHorizontal = global::pjsua2xamarin.Droid.Resource.Attribute.chipSpacingHorizontal;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.chipSpacingVertical = global::pjsua2xamarin.Droid.Resource.Attribute.chipSpacingVertical;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.chipStandaloneStyle = global::pjsua2xamarin.Droid.Resource.Attribute.chipStandaloneStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.chipStartPadding = global::pjsua2xamarin.Droid.Resource.Attribute.chipStartPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.chipStrokeColor = global::pjsua2xamarin.Droid.Resource.Attribute.chipStrokeColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.chipStrokeWidth = global::pjsua2xamarin.Droid.Resource.Attribute.chipStrokeWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.chipStyle = global::pjsua2xamarin.Droid.Resource.Attribute.chipStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.chipSurfaceColor = global::pjsua2xamarin.Droid.Resource.Attribute.chipSurfaceColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.circleRadius = global::pjsua2xamarin.Droid.Resource.Attribute.circleRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.circularflow_angles = global::pjsua2xamarin.Droid.Resource.Attribute.circularflow_angles;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.circularflow_defaultAngle = global::pjsua2xamarin.Droid.Resource.Attribute.circularflow_defaultAngle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.circularflow_defaultRadius = global::pjsua2xamarin.Droid.Resource.Attribute.circularflow_defaultRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.circularflow_radiusInDP = global::pjsua2xamarin.Droid.Resource.Attribute.circularflow_radiusInDP;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.circularflow_viewCenter = global::pjsua2xamarin.Droid.Resource.Attribute.circularflow_viewCenter;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.circularProgressIndicatorStyle = global::pjsua2xamarin.Droid.Resource.Attribute.circularProgressIndicatorStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.clearsTag = global::pjsua2xamarin.Droid.Resource.Attribute.clearsTag;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.clickAction = global::pjsua2xamarin.Droid.Resource.Attribute.clickAction;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.clockFaceBackgroundColor = global::pjsua2xamarin.Droid.Resource.Attribute.clockFaceBackgroundColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.clockHandColor = global::pjsua2xamarin.Droid.Resource.Attribute.clockHandColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.clockIcon = global::pjsua2xamarin.Droid.Resource.Attribute.clockIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.clockNumberTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.clockNumberTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.closeIcon = global::pjsua2xamarin.Droid.Resource.Attribute.closeIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.closeIconEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.closeIconEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.closeIconEndPadding = global::pjsua2xamarin.Droid.Resource.Attribute.closeIconEndPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.closeIconSize = global::pjsua2xamarin.Droid.Resource.Attribute.closeIconSize;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.closeIconStartPadding = global::pjsua2xamarin.Droid.Resource.Attribute.closeIconStartPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.closeIconTint = global::pjsua2xamarin.Droid.Resource.Attribute.closeIconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.closeIconVisible = global::pjsua2xamarin.Droid.Resource.Attribute.closeIconVisible;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.closeItemLayout = global::pjsua2xamarin.Droid.Resource.Attribute.closeItemLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.collapseContentDescription = global::pjsua2xamarin.Droid.Resource.Attribute.collapseContentDescription;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.collapsedSize = global::pjsua2xamarin.Droid.Resource.Attribute.collapsedSize;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.collapsedTitleGravity = global::pjsua2xamarin.Droid.Resource.Attribute.collapsedTitleGravity;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.collapsedTitleTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.collapsedTitleTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.collapseIcon = global::pjsua2xamarin.Droid.Resource.Attribute.collapseIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.collapsingToolbarLayoutStyle = global::pjsua2xamarin.Droid.Resource.Attribute.collapsingToolbarLayoutStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.collectionViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.collectionViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.color = global::pjsua2xamarin.Droid.Resource.Attribute.color;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.colorAccent = global::pjsua2xamarin.Droid.Resource.Attribute.colorAccent;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.colorBackgroundFloating = global::pjsua2xamarin.Droid.Resource.Attribute.colorBackgroundFloating;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.colorButtonNormal = global::pjsua2xamarin.Droid.Resource.Attribute.colorButtonNormal;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.colorControlActivated = global::pjsua2xamarin.Droid.Resource.Attribute.colorControlActivated;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.colorControlHighlight = global::pjsua2xamarin.Droid.Resource.Attribute.colorControlHighlight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.colorControlNormal = global::pjsua2xamarin.Droid.Resource.Attribute.colorControlNormal;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.colorError = global::pjsua2xamarin.Droid.Resource.Attribute.colorError;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.colorOnBackground = global::pjsua2xamarin.Droid.Resource.Attribute.colorOnBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.colorOnError = global::pjsua2xamarin.Droid.Resource.Attribute.colorOnError;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.colorOnPrimary = global::pjsua2xamarin.Droid.Resource.Attribute.colorOnPrimary;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.colorOnPrimarySurface = global::pjsua2xamarin.Droid.Resource.Attribute.colorOnPrimarySurface;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.colorOnSecondary = global::pjsua2xamarin.Droid.Resource.Attribute.colorOnSecondary;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.colorOnSurface = global::pjsua2xamarin.Droid.Resource.Attribute.colorOnSurface;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.colorPrimary = global::pjsua2xamarin.Droid.Resource.Attribute.colorPrimary;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.colorPrimaryDark = global::pjsua2xamarin.Droid.Resource.Attribute.colorPrimaryDark;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.colorPrimarySurface = global::pjsua2xamarin.Droid.Resource.Attribute.colorPrimarySurface;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.colorPrimaryVariant = global::pjsua2xamarin.Droid.Resource.Attribute.colorPrimaryVariant;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.colorSecondary = global::pjsua2xamarin.Droid.Resource.Attribute.colorSecondary;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.colorSecondaryVariant = global::pjsua2xamarin.Droid.Resource.Attribute.colorSecondaryVariant;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.colorSurface = global::pjsua2xamarin.Droid.Resource.Attribute.colorSurface;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.colorSwitchThumbNormal = global::pjsua2xamarin.Droid.Resource.Attribute.colorSwitchThumbNormal;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.commitIcon = global::pjsua2xamarin.Droid.Resource.Attribute.commitIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.ConstraintRotate = global::pjsua2xamarin.Droid.Resource.Attribute.ConstraintRotate;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.constraints = global::pjsua2xamarin.Droid.Resource.Attribute.constraints;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.constraintSet = global::pjsua2xamarin.Droid.Resource.Attribute.constraintSet;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.constraintSetEnd = global::pjsua2xamarin.Droid.Resource.Attribute.constraintSetEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.constraintSetStart = global::pjsua2xamarin.Droid.Resource.Attribute.constraintSetStart;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.constraint_referenced_ids = global::pjsua2xamarin.Droid.Resource.Attribute.constraint_referenced_ids;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.constraint_referenced_tags = global::pjsua2xamarin.Droid.Resource.Attribute.constraint_referenced_tags;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.content = global::pjsua2xamarin.Droid.Resource.Attribute.content;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.contentDescription = global::pjsua2xamarin.Droid.Resource.Attribute.contentDescription;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.contentInsetEnd = global::pjsua2xamarin.Droid.Resource.Attribute.contentInsetEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.contentInsetEndWithActions = global::pjsua2xamarin.Droid.Resource.Attribute.contentInsetEndWithActions;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.contentInsetLeft = global::pjsua2xamarin.Droid.Resource.Attribute.contentInsetLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.contentInsetRight = global::pjsua2xamarin.Droid.Resource.Attribute.contentInsetRight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.contentInsetStart = global::pjsua2xamarin.Droid.Resource.Attribute.contentInsetStart;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.contentInsetStartWithNavigation = global::pjsua2xamarin.Droid.Resource.Attribute.contentInsetStartWithNavigation;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.contentPadding = global::pjsua2xamarin.Droid.Resource.Attribute.contentPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.contentPaddingBottom = global::pjsua2xamarin.Droid.Resource.Attribute.contentPaddingBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.contentPaddingEnd = global::pjsua2xamarin.Droid.Resource.Attribute.contentPaddingEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.contentPaddingLeft = global::pjsua2xamarin.Droid.Resource.Attribute.contentPaddingLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.contentPaddingRight = global::pjsua2xamarin.Droid.Resource.Attribute.contentPaddingRight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.contentPaddingStart = global::pjsua2xamarin.Droid.Resource.Attribute.contentPaddingStart;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.contentPaddingTop = global::pjsua2xamarin.Droid.Resource.Attribute.contentPaddingTop;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.contentScrim = global::pjsua2xamarin.Droid.Resource.Attribute.contentScrim;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.contrast = global::pjsua2xamarin.Droid.Resource.Attribute.contrast;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.controlBackground = global::pjsua2xamarin.Droid.Resource.Attribute.controlBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.coordinatorLayoutStyle = global::pjsua2xamarin.Droid.Resource.Attribute.coordinatorLayoutStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.cornerFamily = global::pjsua2xamarin.Droid.Resource.Attribute.cornerFamily;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.cornerFamilyBottomLeft = global::pjsua2xamarin.Droid.Resource.Attribute.cornerFamilyBottomLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.cornerFamilyBottomRight = global::pjsua2xamarin.Droid.Resource.Attribute.cornerFamilyBottomRight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.cornerFamilyTopLeft = global::pjsua2xamarin.Droid.Resource.Attribute.cornerFamilyTopLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.cornerFamilyTopRight = global::pjsua2xamarin.Droid.Resource.Attribute.cornerFamilyTopRight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.cornerRadius = global::pjsua2xamarin.Droid.Resource.Attribute.cornerRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.cornerSize = global::pjsua2xamarin.Droid.Resource.Attribute.cornerSize;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.cornerSizeBottomLeft = global::pjsua2xamarin.Droid.Resource.Attribute.cornerSizeBottomLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.cornerSizeBottomRight = global::pjsua2xamarin.Droid.Resource.Attribute.cornerSizeBottomRight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.cornerSizeTopLeft = global::pjsua2xamarin.Droid.Resource.Attribute.cornerSizeTopLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.cornerSizeTopRight = global::pjsua2xamarin.Droid.Resource.Attribute.cornerSizeTopRight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.counterEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.counterEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.counterMaxLength = global::pjsua2xamarin.Droid.Resource.Attribute.counterMaxLength;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.counterOverflowTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.counterOverflowTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.counterOverflowTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.counterOverflowTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.counterTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.counterTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.counterTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.counterTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.crossfade = global::pjsua2xamarin.Droid.Resource.Attribute.crossfade;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.currentState = global::pjsua2xamarin.Droid.Resource.Attribute.currentState;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.curveFit = global::pjsua2xamarin.Droid.Resource.Attribute.curveFit;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.customBoolean = global::pjsua2xamarin.Droid.Resource.Attribute.customBoolean;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.customColorDrawableValue = global::pjsua2xamarin.Droid.Resource.Attribute.customColorDrawableValue;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.customColorValue = global::pjsua2xamarin.Droid.Resource.Attribute.customColorValue;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.customDimension = global::pjsua2xamarin.Droid.Resource.Attribute.customDimension;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.customFloatValue = global::pjsua2xamarin.Droid.Resource.Attribute.customFloatValue;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.customIntegerValue = global::pjsua2xamarin.Droid.Resource.Attribute.customIntegerValue;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.customNavigationLayout = global::pjsua2xamarin.Droid.Resource.Attribute.customNavigationLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.customPixelDimension = global::pjsua2xamarin.Droid.Resource.Attribute.customPixelDimension;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.customReference = global::pjsua2xamarin.Droid.Resource.Attribute.customReference;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.customStringValue = global::pjsua2xamarin.Droid.Resource.Attribute.customStringValue;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.data = global::pjsua2xamarin.Droid.Resource.Attribute.data;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.dataPattern = global::pjsua2xamarin.Droid.Resource.Attribute.dataPattern;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.dayInvalidStyle = global::pjsua2xamarin.Droid.Resource.Attribute.dayInvalidStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.daySelectedStyle = global::pjsua2xamarin.Droid.Resource.Attribute.daySelectedStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.dayStyle = global::pjsua2xamarin.Droid.Resource.Attribute.dayStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.dayTodayStyle = global::pjsua2xamarin.Droid.Resource.Attribute.dayTodayStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.defaultDuration = global::pjsua2xamarin.Droid.Resource.Attribute.defaultDuration;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.defaultQueryHint = global::pjsua2xamarin.Droid.Resource.Attribute.defaultQueryHint;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.defaultState = global::pjsua2xamarin.Droid.Resource.Attribute.defaultState;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.deltaPolarAngle = global::pjsua2xamarin.Droid.Resource.Attribute.deltaPolarAngle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.deltaPolarRadius = global::pjsua2xamarin.Droid.Resource.Attribute.deltaPolarRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.deriveConstraintsFrom = global::pjsua2xamarin.Droid.Resource.Attribute.deriveConstraintsFrom;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.destination = global::pjsua2xamarin.Droid.Resource.Attribute.destination;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.dialogCornerRadius = global::pjsua2xamarin.Droid.Resource.Attribute.dialogCornerRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.dialogPreferredPadding = global::pjsua2xamarin.Droid.Resource.Attribute.dialogPreferredPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.dialogTheme = global::pjsua2xamarin.Droid.Resource.Attribute.dialogTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.displayOptions = global::pjsua2xamarin.Droid.Resource.Attribute.displayOptions;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.divider = global::pjsua2xamarin.Droid.Resource.Attribute.divider;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.dividerHorizontal = global::pjsua2xamarin.Droid.Resource.Attribute.dividerHorizontal;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.dividerPadding = global::pjsua2xamarin.Droid.Resource.Attribute.dividerPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.dividerVertical = global::pjsua2xamarin.Droid.Resource.Attribute.dividerVertical;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.dragDirection = global::pjsua2xamarin.Droid.Resource.Attribute.dragDirection;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.dragScale = global::pjsua2xamarin.Droid.Resource.Attribute.dragScale;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.dragThreshold = global::pjsua2xamarin.Droid.Resource.Attribute.dragThreshold;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.drawableBottomCompat = global::pjsua2xamarin.Droid.Resource.Attribute.drawableBottomCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.drawableEndCompat = global::pjsua2xamarin.Droid.Resource.Attribute.drawableEndCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.drawableLeftCompat = global::pjsua2xamarin.Droid.Resource.Attribute.drawableLeftCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.drawableRightCompat = global::pjsua2xamarin.Droid.Resource.Attribute.drawableRightCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.drawableSize = global::pjsua2xamarin.Droid.Resource.Attribute.drawableSize;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.drawableStartCompat = global::pjsua2xamarin.Droid.Resource.Attribute.drawableStartCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.drawableTint = global::pjsua2xamarin.Droid.Resource.Attribute.drawableTint;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.drawableTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.drawableTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.drawableTopCompat = global::pjsua2xamarin.Droid.Resource.Attribute.drawableTopCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.drawerArrowStyle = global::pjsua2xamarin.Droid.Resource.Attribute.drawerArrowStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.drawerLayoutStyle = global::pjsua2xamarin.Droid.Resource.Attribute.drawerLayoutStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.drawPath = global::pjsua2xamarin.Droid.Resource.Attribute.drawPath;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.dropdownListPreferredItemHeight = global::pjsua2xamarin.Droid.Resource.Attribute.dropdownListPreferredItemHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.dropDownListViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.dropDownListViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.duration = global::pjsua2xamarin.Droid.Resource.Attribute.duration;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.editTextBackground = global::pjsua2xamarin.Droid.Resource.Attribute.editTextBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.editTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.editTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.editTextStyle = global::pjsua2xamarin.Droid.Resource.Attribute.editTextStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.elevation = global::pjsua2xamarin.Droid.Resource.Attribute.elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.elevationOverlayColor = global::pjsua2xamarin.Droid.Resource.Attribute.elevationOverlayColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.elevationOverlayEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.elevationOverlayEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.enableEdgeToEdge = global::pjsua2xamarin.Droid.Resource.Attribute.enableEdgeToEdge;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.endIconCheckable = global::pjsua2xamarin.Droid.Resource.Attribute.endIconCheckable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.endIconContentDescription = global::pjsua2xamarin.Droid.Resource.Attribute.endIconContentDescription;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.endIconDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.endIconDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.endIconMode = global::pjsua2xamarin.Droid.Resource.Attribute.endIconMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.endIconTint = global::pjsua2xamarin.Droid.Resource.Attribute.endIconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.endIconTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.endIconTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.enforceMaterialTheme = global::pjsua2xamarin.Droid.Resource.Attribute.enforceMaterialTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.enforceTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.enforceTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.ensureMinTouchTargetSize = global::pjsua2xamarin.Droid.Resource.Attribute.ensureMinTouchTargetSize;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.enterAnim = global::pjsua2xamarin.Droid.Resource.Attribute.enterAnim;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.errorContentDescription = global::pjsua2xamarin.Droid.Resource.Attribute.errorContentDescription;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.errorEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.errorEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.errorIconDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.errorIconDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.errorIconTint = global::pjsua2xamarin.Droid.Resource.Attribute.errorIconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.errorIconTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.errorIconTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.errorTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.errorTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.errorTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.errorTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.exitAnim = global::pjsua2xamarin.Droid.Resource.Attribute.exitAnim;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.expandActivityOverflowButtonDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.expandActivityOverflowButtonDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.expanded = global::pjsua2xamarin.Droid.Resource.Attribute.expanded;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.expandedHintEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.expandedHintEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.expandedTitleGravity = global::pjsua2xamarin.Droid.Resource.Attribute.expandedTitleGravity;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.expandedTitleMargin = global::pjsua2xamarin.Droid.Resource.Attribute.expandedTitleMargin;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.expandedTitleMarginBottom = global::pjsua2xamarin.Droid.Resource.Attribute.expandedTitleMarginBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.expandedTitleMarginEnd = global::pjsua2xamarin.Droid.Resource.Attribute.expandedTitleMarginEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.expandedTitleMarginStart = global::pjsua2xamarin.Droid.Resource.Attribute.expandedTitleMarginStart;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.expandedTitleMarginTop = global::pjsua2xamarin.Droid.Resource.Attribute.expandedTitleMarginTop;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.expandedTitleTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.expandedTitleTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.extendedFloatingActionButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.extendedFloatingActionButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.extendMotionSpec = global::pjsua2xamarin.Droid.Resource.Attribute.extendMotionSpec;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.extraMultilineHeightEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.extraMultilineHeightEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fabAlignmentMode = global::pjsua2xamarin.Droid.Resource.Attribute.fabAlignmentMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fabAnimationMode = global::pjsua2xamarin.Droid.Resource.Attribute.fabAnimationMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fabCradleMargin = global::pjsua2xamarin.Droid.Resource.Attribute.fabCradleMargin;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fabCradleRoundedCornerRadius = global::pjsua2xamarin.Droid.Resource.Attribute.fabCradleRoundedCornerRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fabCradleVerticalOffset = global::pjsua2xamarin.Droid.Resource.Attribute.fabCradleVerticalOffset;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fabCustomSize = global::pjsua2xamarin.Droid.Resource.Attribute.fabCustomSize;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fabSize = global::pjsua2xamarin.Droid.Resource.Attribute.fabSize;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fastScrollEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.fastScrollEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fastScrollHorizontalThumbDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.fastScrollHorizontalThumbDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fastScrollHorizontalTrackDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.fastScrollHorizontalTrackDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fastScrollVerticalThumbDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.fastScrollVerticalThumbDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fastScrollVerticalTrackDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.fastScrollVerticalTrackDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.firstBaselineToTopHeight = global::pjsua2xamarin.Droid.Resource.Attribute.firstBaselineToTopHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.floatingActionButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.floatingActionButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.flow_firstHorizontalBias = global::pjsua2xamarin.Droid.Resource.Attribute.flow_firstHorizontalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.flow_firstHorizontalStyle = global::pjsua2xamarin.Droid.Resource.Attribute.flow_firstHorizontalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.flow_firstVerticalBias = global::pjsua2xamarin.Droid.Resource.Attribute.flow_firstVerticalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.flow_firstVerticalStyle = global::pjsua2xamarin.Droid.Resource.Attribute.flow_firstVerticalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.flow_horizontalAlign = global::pjsua2xamarin.Droid.Resource.Attribute.flow_horizontalAlign;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.flow_horizontalBias = global::pjsua2xamarin.Droid.Resource.Attribute.flow_horizontalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.flow_horizontalGap = global::pjsua2xamarin.Droid.Resource.Attribute.flow_horizontalGap;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.flow_horizontalStyle = global::pjsua2xamarin.Droid.Resource.Attribute.flow_horizontalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.flow_lastHorizontalBias = global::pjsua2xamarin.Droid.Resource.Attribute.flow_lastHorizontalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.flow_lastHorizontalStyle = global::pjsua2xamarin.Droid.Resource.Attribute.flow_lastHorizontalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.flow_lastVerticalBias = global::pjsua2xamarin.Droid.Resource.Attribute.flow_lastVerticalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.flow_lastVerticalStyle = global::pjsua2xamarin.Droid.Resource.Attribute.flow_lastVerticalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.flow_maxElementsWrap = global::pjsua2xamarin.Droid.Resource.Attribute.flow_maxElementsWrap;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.flow_padding = global::pjsua2xamarin.Droid.Resource.Attribute.flow_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.flow_verticalAlign = global::pjsua2xamarin.Droid.Resource.Attribute.flow_verticalAlign;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.flow_verticalBias = global::pjsua2xamarin.Droid.Resource.Attribute.flow_verticalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.flow_verticalGap = global::pjsua2xamarin.Droid.Resource.Attribute.flow_verticalGap;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.flow_verticalStyle = global::pjsua2xamarin.Droid.Resource.Attribute.flow_verticalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.flow_wrapMode = global::pjsua2xamarin.Droid.Resource.Attribute.flow_wrapMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.font = global::pjsua2xamarin.Droid.Resource.Attribute.font;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fontFamily = global::pjsua2xamarin.Droid.Resource.Attribute.fontFamily;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fontProviderAuthority = global::pjsua2xamarin.Droid.Resource.Attribute.fontProviderAuthority;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fontProviderCerts = global::pjsua2xamarin.Droid.Resource.Attribute.fontProviderCerts;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fontProviderFetchStrategy = global::pjsua2xamarin.Droid.Resource.Attribute.fontProviderFetchStrategy;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fontProviderFetchTimeout = global::pjsua2xamarin.Droid.Resource.Attribute.fontProviderFetchTimeout;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fontProviderPackage = global::pjsua2xamarin.Droid.Resource.Attribute.fontProviderPackage;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fontProviderQuery = global::pjsua2xamarin.Droid.Resource.Attribute.fontProviderQuery;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fontProviderSystemFontFamily = global::pjsua2xamarin.Droid.Resource.Attribute.fontProviderSystemFontFamily;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fontStyle = global::pjsua2xamarin.Droid.Resource.Attribute.fontStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fontVariationSettings = global::pjsua2xamarin.Droid.Resource.Attribute.fontVariationSettings;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.fontWeight = global::pjsua2xamarin.Droid.Resource.Attribute.fontWeight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.forceApplySystemWindowInsetTop = global::pjsua2xamarin.Droid.Resource.Attribute.forceApplySystemWindowInsetTop;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.foregroundInsidePadding = global::pjsua2xamarin.Droid.Resource.Attribute.foregroundInsidePadding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.framePosition = global::pjsua2xamarin.Droid.Resource.Attribute.framePosition;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.gapBetweenBars = global::pjsua2xamarin.Droid.Resource.Attribute.gapBetweenBars;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.gestureInsetBottomIgnored = global::pjsua2xamarin.Droid.Resource.Attribute.gestureInsetBottomIgnored;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.goIcon = global::pjsua2xamarin.Droid.Resource.Attribute.goIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.graph = global::pjsua2xamarin.Droid.Resource.Attribute.graph;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.haloColor = global::pjsua2xamarin.Droid.Resource.Attribute.haloColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.haloRadius = global::pjsua2xamarin.Droid.Resource.Attribute.haloRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.headerLayout = global::pjsua2xamarin.Droid.Resource.Attribute.headerLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.height = global::pjsua2xamarin.Droid.Resource.Attribute.height;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.helperText = global::pjsua2xamarin.Droid.Resource.Attribute.helperText;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.helperTextEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.helperTextEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.helperTextTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.helperTextTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.helperTextTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.helperTextTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.hideAnimationBehavior = global::pjsua2xamarin.Droid.Resource.Attribute.hideAnimationBehavior;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.hideMotionSpec = global::pjsua2xamarin.Droid.Resource.Attribute.hideMotionSpec;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.hideOnContentScroll = global::pjsua2xamarin.Droid.Resource.Attribute.hideOnContentScroll;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.hideOnScroll = global::pjsua2xamarin.Droid.Resource.Attribute.hideOnScroll;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.hintAnimationEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.hintAnimationEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.hintEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.hintEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.hintTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.hintTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.hintTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.hintTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.homeAsUpIndicator = global::pjsua2xamarin.Droid.Resource.Attribute.homeAsUpIndicator;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.homeLayout = global::pjsua2xamarin.Droid.Resource.Attribute.homeLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.horizontalOffset = global::pjsua2xamarin.Droid.Resource.Attribute.horizontalOffset;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.hoveredFocusedTranslationZ = global::pjsua2xamarin.Droid.Resource.Attribute.hoveredFocusedTranslationZ;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.icon = global::pjsua2xamarin.Droid.Resource.Attribute.icon;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.iconEndPadding = global::pjsua2xamarin.Droid.Resource.Attribute.iconEndPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.iconGravity = global::pjsua2xamarin.Droid.Resource.Attribute.iconGravity;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.iconifiedByDefault = global::pjsua2xamarin.Droid.Resource.Attribute.iconifiedByDefault;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.iconPadding = global::pjsua2xamarin.Droid.Resource.Attribute.iconPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.iconSize = global::pjsua2xamarin.Droid.Resource.Attribute.iconSize;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.iconStartPadding = global::pjsua2xamarin.Droid.Resource.Attribute.iconStartPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.iconTint = global::pjsua2xamarin.Droid.Resource.Attribute.iconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.iconTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.iconTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.ifTagNotSet = global::pjsua2xamarin.Droid.Resource.Attribute.ifTagNotSet;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.ifTagSet = global::pjsua2xamarin.Droid.Resource.Attribute.ifTagSet;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.imageButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.imageButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.imagePanX = global::pjsua2xamarin.Droid.Resource.Attribute.imagePanX;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.imagePanY = global::pjsua2xamarin.Droid.Resource.Attribute.imagePanY;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.imageRotate = global::pjsua2xamarin.Droid.Resource.Attribute.imageRotate;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.imageZoom = global::pjsua2xamarin.Droid.Resource.Attribute.imageZoom;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.indeterminateAnimationType = global::pjsua2xamarin.Droid.Resource.Attribute.indeterminateAnimationType;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.indeterminateProgressStyle = global::pjsua2xamarin.Droid.Resource.Attribute.indeterminateProgressStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.indicatorColor = global::pjsua2xamarin.Droid.Resource.Attribute.indicatorColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.indicatorDirectionCircular = global::pjsua2xamarin.Droid.Resource.Attribute.indicatorDirectionCircular;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.indicatorDirectionLinear = global::pjsua2xamarin.Droid.Resource.Attribute.indicatorDirectionLinear;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.indicatorInset = global::pjsua2xamarin.Droid.Resource.Attribute.indicatorInset;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.indicatorSize = global::pjsua2xamarin.Droid.Resource.Attribute.indicatorSize;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.initialActivityCount = global::pjsua2xamarin.Droid.Resource.Attribute.initialActivityCount;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.insetForeground = global::pjsua2xamarin.Droid.Resource.Attribute.insetForeground;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.isLightTheme = global::pjsua2xamarin.Droid.Resource.Attribute.isLightTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.isMaterialTheme = global::pjsua2xamarin.Droid.Resource.Attribute.isMaterialTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemBackground = global::pjsua2xamarin.Droid.Resource.Attribute.itemBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemFillColor = global::pjsua2xamarin.Droid.Resource.Attribute.itemFillColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemHorizontalPadding = global::pjsua2xamarin.Droid.Resource.Attribute.itemHorizontalPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemHorizontalTranslationEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.itemHorizontalTranslationEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemIconPadding = global::pjsua2xamarin.Droid.Resource.Attribute.itemIconPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemIconSize = global::pjsua2xamarin.Droid.Resource.Attribute.itemIconSize;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemIconTint = global::pjsua2xamarin.Droid.Resource.Attribute.itemIconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemMaxLines = global::pjsua2xamarin.Droid.Resource.Attribute.itemMaxLines;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemPadding = global::pjsua2xamarin.Droid.Resource.Attribute.itemPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemRippleColor = global::pjsua2xamarin.Droid.Resource.Attribute.itemRippleColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemShapeAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.itemShapeAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemShapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Attribute.itemShapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemShapeFillColor = global::pjsua2xamarin.Droid.Resource.Attribute.itemShapeFillColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemShapeInsetBottom = global::pjsua2xamarin.Droid.Resource.Attribute.itemShapeInsetBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemShapeInsetEnd = global::pjsua2xamarin.Droid.Resource.Attribute.itemShapeInsetEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemShapeInsetStart = global::pjsua2xamarin.Droid.Resource.Attribute.itemShapeInsetStart;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemShapeInsetTop = global::pjsua2xamarin.Droid.Resource.Attribute.itemShapeInsetTop;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemSpacing = global::pjsua2xamarin.Droid.Resource.Attribute.itemSpacing;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemStrokeColor = global::pjsua2xamarin.Droid.Resource.Attribute.itemStrokeColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemStrokeWidth = global::pjsua2xamarin.Droid.Resource.Attribute.itemStrokeWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.itemTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemTextAppearanceActive = global::pjsua2xamarin.Droid.Resource.Attribute.itemTextAppearanceActive;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemTextAppearanceInactive = global::pjsua2xamarin.Droid.Resource.Attribute.itemTextAppearanceInactive;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.itemTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.itemTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.keyboardIcon = global::pjsua2xamarin.Droid.Resource.Attribute.keyboardIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.keylines = global::pjsua2xamarin.Droid.Resource.Attribute.keylines;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.keyPositionType = global::pjsua2xamarin.Droid.Resource.Attribute.keyPositionType;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.labelBehavior = global::pjsua2xamarin.Droid.Resource.Attribute.labelBehavior;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.labelStyle = global::pjsua2xamarin.Droid.Resource.Attribute.labelStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.labelVisibilityMode = global::pjsua2xamarin.Droid.Resource.Attribute.labelVisibilityMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.lastBaselineToBottomHeight = global::pjsua2xamarin.Droid.Resource.Attribute.lastBaselineToBottomHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.launchSingleTop = global::pjsua2xamarin.Droid.Resource.Attribute.launchSingleTop;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout = global::pjsua2xamarin.Droid.Resource.Attribute.layout;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layoutDescription = global::pjsua2xamarin.Droid.Resource.Attribute.layoutDescription;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layoutDuringTransition = global::pjsua2xamarin.Droid.Resource.Attribute.layoutDuringTransition;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layoutManager = global::pjsua2xamarin.Droid.Resource.Attribute.layoutManager;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_anchor = global::pjsua2xamarin.Droid.Resource.Attribute.layout_anchor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_anchorGravity = global::pjsua2xamarin.Droid.Resource.Attribute.layout_anchorGravity;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_behavior = global::pjsua2xamarin.Droid.Resource.Attribute.layout_behavior;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_collapseMode = global::pjsua2xamarin.Droid.Resource.Attribute.layout_collapseMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_collapseParallaxMultiplier = global::pjsua2xamarin.Droid.Resource.Attribute.layout_collapseParallaxMultiplier;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constrainedHeight = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constrainedHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constrainedWidth = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constrainedWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintBaseline_creator = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintBaseline_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintBaseline_toBaselineOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintBaseline_toBaselineOf;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintBaseline_toBottomOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintBaseline_toBottomOf;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintBaseline_toTopOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintBaseline_toTopOf;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintBottom_creator = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintBottom_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintBottom_toBottomOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintBottom_toBottomOf;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintBottom_toTopOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintBottom_toTopOf;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintCircle = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintCircle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintCircleAngle = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintCircleAngle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintCircleRadius = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintCircleRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintDimensionRatio = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintDimensionRatio;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintEnd_toEndOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintEnd_toEndOf;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintEnd_toStartOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintEnd_toStartOf;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintGuide_begin = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintGuide_begin;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintGuide_end = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintGuide_end;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintGuide_percent = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintGuide_percent;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintHeight = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintHeight_default = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintHeight_default;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintHeight_max = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintHeight_max;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintHeight_min = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintHeight_min;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintHeight_percent = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintHeight_percent;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintHorizontal_bias = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintHorizontal_bias;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintHorizontal_chainStyle = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintHorizontal_chainStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintHorizontal_weight = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintHorizontal_weight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintLeft_creator = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintLeft_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintLeft_toLeftOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintLeft_toLeftOf;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintLeft_toRightOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintLeft_toRightOf;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintRight_creator = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintRight_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintRight_toLeftOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintRight_toLeftOf;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintRight_toRightOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintRight_toRightOf;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintStart_toEndOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintStart_toEndOf;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintStart_toStartOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintStart_toStartOf;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintTag = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintTag;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintTop_creator = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintTop_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintTop_toBottomOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintTop_toBottomOf;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintTop_toTopOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintTop_toTopOf;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintVertical_bias = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintVertical_bias;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintVertical_chainStyle = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintVertical_chainStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintVertical_weight = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintVertical_weight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintWidth = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintWidth_default = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintWidth_default;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintWidth_max = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintWidth_max;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintWidth_min = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintWidth_min;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_constraintWidth_percent = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintWidth_percent;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_dodgeInsetEdges = global::pjsua2xamarin.Droid.Resource.Attribute.layout_dodgeInsetEdges;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_editor_absoluteX = global::pjsua2xamarin.Droid.Resource.Attribute.layout_editor_absoluteX;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_editor_absoluteY = global::pjsua2xamarin.Droid.Resource.Attribute.layout_editor_absoluteY;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_goneMarginBaseline = global::pjsua2xamarin.Droid.Resource.Attribute.layout_goneMarginBaseline;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_goneMarginBottom = global::pjsua2xamarin.Droid.Resource.Attribute.layout_goneMarginBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_goneMarginEnd = global::pjsua2xamarin.Droid.Resource.Attribute.layout_goneMarginEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_goneMarginLeft = global::pjsua2xamarin.Droid.Resource.Attribute.layout_goneMarginLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_goneMarginRight = global::pjsua2xamarin.Droid.Resource.Attribute.layout_goneMarginRight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_goneMarginStart = global::pjsua2xamarin.Droid.Resource.Attribute.layout_goneMarginStart;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_goneMarginTop = global::pjsua2xamarin.Droid.Resource.Attribute.layout_goneMarginTop;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_insetEdge = global::pjsua2xamarin.Droid.Resource.Attribute.layout_insetEdge;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_keyline = global::pjsua2xamarin.Droid.Resource.Attribute.layout_keyline;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_marginBaseline = global::pjsua2xamarin.Droid.Resource.Attribute.layout_marginBaseline;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_optimizationLevel = global::pjsua2xamarin.Droid.Resource.Attribute.layout_optimizationLevel;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_scrollFlags = global::pjsua2xamarin.Droid.Resource.Attribute.layout_scrollFlags;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_scrollInterpolator = global::pjsua2xamarin.Droid.Resource.Attribute.layout_scrollInterpolator;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.layout_wrapBehaviorInParent = global::pjsua2xamarin.Droid.Resource.Attribute.layout_wrapBehaviorInParent;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.liftOnScroll = global::pjsua2xamarin.Droid.Resource.Attribute.liftOnScroll;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.liftOnScrollTargetViewId = global::pjsua2xamarin.Droid.Resource.Attribute.liftOnScrollTargetViewId;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.limitBoundsTo = global::pjsua2xamarin.Droid.Resource.Attribute.limitBoundsTo;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.linearProgressIndicatorStyle = global::pjsua2xamarin.Droid.Resource.Attribute.linearProgressIndicatorStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.lineHeight = global::pjsua2xamarin.Droid.Resource.Attribute.lineHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.lineSpacing = global::pjsua2xamarin.Droid.Resource.Attribute.lineSpacing;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.listChoiceBackgroundIndicator = global::pjsua2xamarin.Droid.Resource.Attribute.listChoiceBackgroundIndicator;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.listChoiceIndicatorMultipleAnimated = global::pjsua2xamarin.Droid.Resource.Attribute.listChoiceIndicatorMultipleAnimated;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.listChoiceIndicatorSingleAnimated = global::pjsua2xamarin.Droid.Resource.Attribute.listChoiceIndicatorSingleAnimated;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.listDividerAlertDialog = global::pjsua2xamarin.Droid.Resource.Attribute.listDividerAlertDialog;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.listItemLayout = global::pjsua2xamarin.Droid.Resource.Attribute.listItemLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.listLayout = global::pjsua2xamarin.Droid.Resource.Attribute.listLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.listMenuViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.listMenuViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.listPopupWindowStyle = global::pjsua2xamarin.Droid.Resource.Attribute.listPopupWindowStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.listPreferredItemHeight = global::pjsua2xamarin.Droid.Resource.Attribute.listPreferredItemHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.listPreferredItemHeightLarge = global::pjsua2xamarin.Droid.Resource.Attribute.listPreferredItemHeightLarge;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.listPreferredItemHeightSmall = global::pjsua2xamarin.Droid.Resource.Attribute.listPreferredItemHeightSmall;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.listPreferredItemPaddingEnd = global::pjsua2xamarin.Droid.Resource.Attribute.listPreferredItemPaddingEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.listPreferredItemPaddingLeft = global::pjsua2xamarin.Droid.Resource.Attribute.listPreferredItemPaddingLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.listPreferredItemPaddingRight = global::pjsua2xamarin.Droid.Resource.Attribute.listPreferredItemPaddingRight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.listPreferredItemPaddingStart = global::pjsua2xamarin.Droid.Resource.Attribute.listPreferredItemPaddingStart;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.logo = global::pjsua2xamarin.Droid.Resource.Attribute.logo;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.logoDescription = global::pjsua2xamarin.Droid.Resource.Attribute.logoDescription;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialAlertDialogBodyTextStyle = global::pjsua2xamarin.Droid.Resource.Attribute.materialAlertDialogBodyTextStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialAlertDialogTheme = global::pjsua2xamarin.Droid.Resource.Attribute.materialAlertDialogTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialAlertDialogTitleIconStyle = global::pjsua2xamarin.Droid.Resource.Attribute.materialAlertDialogTitleIconStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialAlertDialogTitlePanelStyle = global::pjsua2xamarin.Droid.Resource.Attribute.materialAlertDialogTitlePanelStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialAlertDialogTitleTextStyle = global::pjsua2xamarin.Droid.Resource.Attribute.materialAlertDialogTitleTextStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialButtonOutlinedStyle = global::pjsua2xamarin.Droid.Resource.Attribute.materialButtonOutlinedStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.materialButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialButtonToggleGroupStyle = global::pjsua2xamarin.Droid.Resource.Attribute.materialButtonToggleGroupStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialCalendarDay = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarDay;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialCalendarFullscreenTheme = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarFullscreenTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialCalendarHeaderCancelButton = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarHeaderCancelButton;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialCalendarHeaderConfirmButton = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarHeaderConfirmButton;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialCalendarHeaderDivider = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarHeaderDivider;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialCalendarHeaderLayout = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarHeaderLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialCalendarHeaderSelection = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarHeaderSelection;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialCalendarHeaderTitle = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarHeaderTitle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialCalendarHeaderToggleButton = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarHeaderToggleButton;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialCalendarMonth = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarMonth;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialCalendarMonthNavigationButton = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarMonthNavigationButton;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialCalendarStyle = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialCalendarTheme = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialCalendarYearNavigationButton = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarYearNavigationButton;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialCardViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.materialCardViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialCircleRadius = global::pjsua2xamarin.Droid.Resource.Attribute.materialCircleRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialClockStyle = global::pjsua2xamarin.Droid.Resource.Attribute.materialClockStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialThemeOverlay = global::pjsua2xamarin.Droid.Resource.Attribute.materialThemeOverlay;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialTimePickerStyle = global::pjsua2xamarin.Droid.Resource.Attribute.materialTimePickerStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.materialTimePickerTheme = global::pjsua2xamarin.Droid.Resource.Attribute.materialTimePickerTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.maxAcceleration = global::pjsua2xamarin.Droid.Resource.Attribute.maxAcceleration;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.maxActionInlineWidth = global::pjsua2xamarin.Droid.Resource.Attribute.maxActionInlineWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.maxButtonHeight = global::pjsua2xamarin.Droid.Resource.Attribute.maxButtonHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.maxCharacterCount = global::pjsua2xamarin.Droid.Resource.Attribute.maxCharacterCount;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.maxHeight = global::pjsua2xamarin.Droid.Resource.Attribute.maxHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.maxImageSize = global::pjsua2xamarin.Droid.Resource.Attribute.maxImageSize;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.maxLines = global::pjsua2xamarin.Droid.Resource.Attribute.maxLines;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.maxVelocity = global::pjsua2xamarin.Droid.Resource.Attribute.maxVelocity;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.maxWidth = global::pjsua2xamarin.Droid.Resource.Attribute.maxWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.measureWithLargestChild = global::pjsua2xamarin.Droid.Resource.Attribute.measureWithLargestChild;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.menu = global::pjsua2xamarin.Droid.Resource.Attribute.menu;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.menuGravity = global::pjsua2xamarin.Droid.Resource.Attribute.menuGravity;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.methodName = global::pjsua2xamarin.Droid.Resource.Attribute.methodName;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.mimeType = global::pjsua2xamarin.Droid.Resource.Attribute.mimeType;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.minHeight = global::pjsua2xamarin.Droid.Resource.Attribute.minHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.minHideDelay = global::pjsua2xamarin.Droid.Resource.Attribute.minHideDelay;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.minSeparation = global::pjsua2xamarin.Droid.Resource.Attribute.minSeparation;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.minTouchTargetSize = global::pjsua2xamarin.Droid.Resource.Attribute.minTouchTargetSize;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.minWidth = global::pjsua2xamarin.Droid.Resource.Attribute.minWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.mock_diagonalsColor = global::pjsua2xamarin.Droid.Resource.Attribute.mock_diagonalsColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.mock_label = global::pjsua2xamarin.Droid.Resource.Attribute.mock_label;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.mock_labelBackgroundColor = global::pjsua2xamarin.Droid.Resource.Attribute.mock_labelBackgroundColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.mock_labelColor = global::pjsua2xamarin.Droid.Resource.Attribute.mock_labelColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.mock_showDiagonals = global::pjsua2xamarin.Droid.Resource.Attribute.mock_showDiagonals;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.mock_showLabel = global::pjsua2xamarin.Droid.Resource.Attribute.mock_showLabel;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionDebug = global::pjsua2xamarin.Droid.Resource.Attribute.motionDebug;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionDurationLong1 = global::pjsua2xamarin.Droid.Resource.Attribute.motionDurationLong1;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionDurationLong2 = global::pjsua2xamarin.Droid.Resource.Attribute.motionDurationLong2;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionDurationMedium1 = global::pjsua2xamarin.Droid.Resource.Attribute.motionDurationMedium1;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionDurationMedium2 = global::pjsua2xamarin.Droid.Resource.Attribute.motionDurationMedium2;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionDurationShort1 = global::pjsua2xamarin.Droid.Resource.Attribute.motionDurationShort1;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionDurationShort2 = global::pjsua2xamarin.Droid.Resource.Attribute.motionDurationShort2;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionEasingAccelerated = global::pjsua2xamarin.Droid.Resource.Attribute.motionEasingAccelerated;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionEasingDecelerated = global::pjsua2xamarin.Droid.Resource.Attribute.motionEasingDecelerated;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionEasingEmphasized = global::pjsua2xamarin.Droid.Resource.Attribute.motionEasingEmphasized;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionEasingLinear = global::pjsua2xamarin.Droid.Resource.Attribute.motionEasingLinear;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionEasingStandard = global::pjsua2xamarin.Droid.Resource.Attribute.motionEasingStandard;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionEffect_alpha = global::pjsua2xamarin.Droid.Resource.Attribute.motionEffect_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionEffect_end = global::pjsua2xamarin.Droid.Resource.Attribute.motionEffect_end;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionEffect_move = global::pjsua2xamarin.Droid.Resource.Attribute.motionEffect_move;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionEffect_start = global::pjsua2xamarin.Droid.Resource.Attribute.motionEffect_start;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionEffect_strict = global::pjsua2xamarin.Droid.Resource.Attribute.motionEffect_strict;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionEffect_translationX = global::pjsua2xamarin.Droid.Resource.Attribute.motionEffect_translationX;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionEffect_translationY = global::pjsua2xamarin.Droid.Resource.Attribute.motionEffect_translationY;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionEffect_viewTransition = global::pjsua2xamarin.Droid.Resource.Attribute.motionEffect_viewTransition;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionInterpolator = global::pjsua2xamarin.Droid.Resource.Attribute.motionInterpolator;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionPath = global::pjsua2xamarin.Droid.Resource.Attribute.motionPath;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionPathRotate = global::pjsua2xamarin.Droid.Resource.Attribute.motionPathRotate;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionProgress = global::pjsua2xamarin.Droid.Resource.Attribute.motionProgress;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionStagger = global::pjsua2xamarin.Droid.Resource.Attribute.motionStagger;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motionTarget = global::pjsua2xamarin.Droid.Resource.Attribute.motionTarget;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motion_postLayoutCollision = global::pjsua2xamarin.Droid.Resource.Attribute.motion_postLayoutCollision;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.motion_triggerOnCollision = global::pjsua2xamarin.Droid.Resource.Attribute.motion_triggerOnCollision;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.moveWhenScrollAtTop = global::pjsua2xamarin.Droid.Resource.Attribute.moveWhenScrollAtTop;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.multiChoiceItemLayout = global::pjsua2xamarin.Droid.Resource.Attribute.multiChoiceItemLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.navGraph = global::pjsua2xamarin.Droid.Resource.Attribute.navGraph;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.navigationContentDescription = global::pjsua2xamarin.Droid.Resource.Attribute.navigationContentDescription;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.navigationIcon = global::pjsua2xamarin.Droid.Resource.Attribute.navigationIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.navigationIconTint = global::pjsua2xamarin.Droid.Resource.Attribute.navigationIconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.navigationMode = global::pjsua2xamarin.Droid.Resource.Attribute.navigationMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.navigationRailStyle = global::pjsua2xamarin.Droid.Resource.Attribute.navigationRailStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.navigationViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.navigationViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.nestedScrollable = global::pjsua2xamarin.Droid.Resource.Attribute.nestedScrollable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.nestedScrollFlags = global::pjsua2xamarin.Droid.Resource.Attribute.nestedScrollFlags;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.nestedScrollViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.nestedScrollViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.nullable = global::pjsua2xamarin.Droid.Resource.Attribute.nullable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.number = global::pjsua2xamarin.Droid.Resource.Attribute.number;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.numericModifiers = global::pjsua2xamarin.Droid.Resource.Attribute.numericModifiers;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.onCross = global::pjsua2xamarin.Droid.Resource.Attribute.onCross;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.onHide = global::pjsua2xamarin.Droid.Resource.Attribute.onHide;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.onNegativeCross = global::pjsua2xamarin.Droid.Resource.Attribute.onNegativeCross;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.onPositiveCross = global::pjsua2xamarin.Droid.Resource.Attribute.onPositiveCross;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.onShow = global::pjsua2xamarin.Droid.Resource.Attribute.onShow;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.onStateTransition = global::pjsua2xamarin.Droid.Resource.Attribute.onStateTransition;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.onTouchUp = global::pjsua2xamarin.Droid.Resource.Attribute.onTouchUp;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.overlapAnchor = global::pjsua2xamarin.Droid.Resource.Attribute.overlapAnchor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.overlay = global::pjsua2xamarin.Droid.Resource.Attribute.overlay;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.paddingBottomNoButtons = global::pjsua2xamarin.Droid.Resource.Attribute.paddingBottomNoButtons;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.paddingBottomSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Attribute.paddingBottomSystemWindowInsets;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.paddingEnd = global::pjsua2xamarin.Droid.Resource.Attribute.paddingEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.paddingLeftSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Attribute.paddingLeftSystemWindowInsets;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.paddingRightSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Attribute.paddingRightSystemWindowInsets;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.paddingStart = global::pjsua2xamarin.Droid.Resource.Attribute.paddingStart;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.paddingTopNoTitle = global::pjsua2xamarin.Droid.Resource.Attribute.paddingTopNoTitle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.paddingTopSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Attribute.paddingTopSystemWindowInsets;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.panelBackground = global::pjsua2xamarin.Droid.Resource.Attribute.panelBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.panelMenuListTheme = global::pjsua2xamarin.Droid.Resource.Attribute.panelMenuListTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.panelMenuListWidth = global::pjsua2xamarin.Droid.Resource.Attribute.panelMenuListWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.passwordToggleContentDescription = global::pjsua2xamarin.Droid.Resource.Attribute.passwordToggleContentDescription;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.passwordToggleDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.passwordToggleDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.passwordToggleEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.passwordToggleEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.passwordToggleTint = global::pjsua2xamarin.Droid.Resource.Attribute.passwordToggleTint;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.passwordToggleTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.passwordToggleTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.pathMotionArc = global::pjsua2xamarin.Droid.Resource.Attribute.pathMotionArc;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.path_percent = global::pjsua2xamarin.Droid.Resource.Attribute.path_percent;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.percentHeight = global::pjsua2xamarin.Droid.Resource.Attribute.percentHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.percentWidth = global::pjsua2xamarin.Droid.Resource.Attribute.percentWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.percentX = global::pjsua2xamarin.Droid.Resource.Attribute.percentX;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.percentY = global::pjsua2xamarin.Droid.Resource.Attribute.percentY;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.perpendicularPath_percent = global::pjsua2xamarin.Droid.Resource.Attribute.perpendicularPath_percent;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.pivotAnchor = global::pjsua2xamarin.Droid.Resource.Attribute.pivotAnchor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.placeholderText = global::pjsua2xamarin.Droid.Resource.Attribute.placeholderText;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.placeholderTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.placeholderTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.placeholderTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.placeholderTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.placeholder_emptyVisibility = global::pjsua2xamarin.Droid.Resource.Attribute.placeholder_emptyVisibility;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.polarRelativeTo = global::pjsua2xamarin.Droid.Resource.Attribute.polarRelativeTo;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.popEnterAnim = global::pjsua2xamarin.Droid.Resource.Attribute.popEnterAnim;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.popExitAnim = global::pjsua2xamarin.Droid.Resource.Attribute.popExitAnim;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.popupMenuBackground = global::pjsua2xamarin.Droid.Resource.Attribute.popupMenuBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.popupMenuStyle = global::pjsua2xamarin.Droid.Resource.Attribute.popupMenuStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.popupTheme = global::pjsua2xamarin.Droid.Resource.Attribute.popupTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.popUpTo = global::pjsua2xamarin.Droid.Resource.Attribute.popUpTo;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.popUpToInclusive = global::pjsua2xamarin.Droid.Resource.Attribute.popUpToInclusive;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.popupWindowStyle = global::pjsua2xamarin.Droid.Resource.Attribute.popupWindowStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.prefixText = global::pjsua2xamarin.Droid.Resource.Attribute.prefixText;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.prefixTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.prefixTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.prefixTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.prefixTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.preserveIconSpacing = global::pjsua2xamarin.Droid.Resource.Attribute.preserveIconSpacing;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.pressedTranslationZ = global::pjsua2xamarin.Droid.Resource.Attribute.pressedTranslationZ;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.progressBarPadding = global::pjsua2xamarin.Droid.Resource.Attribute.progressBarPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.progressBarStyle = global::pjsua2xamarin.Droid.Resource.Attribute.progressBarStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.quantizeMotionInterpolator = global::pjsua2xamarin.Droid.Resource.Attribute.quantizeMotionInterpolator;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.quantizeMotionPhase = global::pjsua2xamarin.Droid.Resource.Attribute.quantizeMotionPhase;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.quantizeMotionSteps = global::pjsua2xamarin.Droid.Resource.Attribute.quantizeMotionSteps;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.queryBackground = global::pjsua2xamarin.Droid.Resource.Attribute.queryBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.queryHint = global::pjsua2xamarin.Droid.Resource.Attribute.queryHint;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.queryPatterns = global::pjsua2xamarin.Droid.Resource.Attribute.queryPatterns;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.radioButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.radioButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.rangeFillColor = global::pjsua2xamarin.Droid.Resource.Attribute.rangeFillColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.ratingBarStyle = global::pjsua2xamarin.Droid.Resource.Attribute.ratingBarStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.ratingBarStyleIndicator = global::pjsua2xamarin.Droid.Resource.Attribute.ratingBarStyleIndicator;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.ratingBarStyleSmall = global::pjsua2xamarin.Droid.Resource.Attribute.ratingBarStyleSmall;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.reactiveGuide_animateChange = global::pjsua2xamarin.Droid.Resource.Attribute.reactiveGuide_animateChange;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.reactiveGuide_applyToAllConstraintSets = global::pjsua2xamarin.Droid.Resource.Attribute.reactiveGuide_applyToAllConstraintSets;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.reactiveGuide_applyToConstraintSet = global::pjsua2xamarin.Droid.Resource.Attribute.reactiveGuide_applyToConstraintSet;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.reactiveGuide_valueId = global::pjsua2xamarin.Droid.Resource.Attribute.reactiveGuide_valueId;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.recyclerViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.recyclerViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.region_heightLessThan = global::pjsua2xamarin.Droid.Resource.Attribute.region_heightLessThan;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.region_heightMoreThan = global::pjsua2xamarin.Droid.Resource.Attribute.region_heightMoreThan;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.region_widthLessThan = global::pjsua2xamarin.Droid.Resource.Attribute.region_widthLessThan;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.region_widthMoreThan = global::pjsua2xamarin.Droid.Resource.Attribute.region_widthMoreThan;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.reverseLayout = global::pjsua2xamarin.Droid.Resource.Attribute.reverseLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.rippleColor = global::pjsua2xamarin.Droid.Resource.Attribute.rippleColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.rotationCenterId = global::pjsua2xamarin.Droid.Resource.Attribute.rotationCenterId;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.round = global::pjsua2xamarin.Droid.Resource.Attribute.round;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.roundPercent = global::pjsua2xamarin.Droid.Resource.Attribute.roundPercent;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.saturation = global::pjsua2xamarin.Droid.Resource.Attribute.saturation;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.scaleFromTextSize = global::pjsua2xamarin.Droid.Resource.Attribute.scaleFromTextSize;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.scrimAnimationDuration = global::pjsua2xamarin.Droid.Resource.Attribute.scrimAnimationDuration;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.scrimBackground = global::pjsua2xamarin.Droid.Resource.Attribute.scrimBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.scrimVisibleHeightTrigger = global::pjsua2xamarin.Droid.Resource.Attribute.scrimVisibleHeightTrigger;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.scrollViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.scrollViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.searchHintIcon = global::pjsua2xamarin.Droid.Resource.Attribute.searchHintIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.searchIcon = global::pjsua2xamarin.Droid.Resource.Attribute.searchIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.searchViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.searchViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.seekBarStyle = global::pjsua2xamarin.Droid.Resource.Attribute.seekBarStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.selectableItemBackground = global::pjsua2xamarin.Droid.Resource.Attribute.selectableItemBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.selectableItemBackgroundBorderless = global::pjsua2xamarin.Droid.Resource.Attribute.selectableItemBackgroundBorderless;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.selectionRequired = global::pjsua2xamarin.Droid.Resource.Attribute.selectionRequired;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.selectorSize = global::pjsua2xamarin.Droid.Resource.Attribute.selectorSize;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.setsTag = global::pjsua2xamarin.Droid.Resource.Attribute.setsTag;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.shapeAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.shapeAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.shapeAppearanceLargeComponent = global::pjsua2xamarin.Droid.Resource.Attribute.shapeAppearanceLargeComponent;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.shapeAppearanceMediumComponent = global::pjsua2xamarin.Droid.Resource.Attribute.shapeAppearanceMediumComponent;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.shapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Attribute.shapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.shapeAppearanceSmallComponent = global::pjsua2xamarin.Droid.Resource.Attribute.shapeAppearanceSmallComponent;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.SharedValue = global::pjsua2xamarin.Droid.Resource.Attribute.SharedValue;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.SharedValueId = global::pjsua2xamarin.Droid.Resource.Attribute.SharedValueId;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.shortcutMatchRequired = global::pjsua2xamarin.Droid.Resource.Attribute.shortcutMatchRequired;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.showAnimationBehavior = global::pjsua2xamarin.Droid.Resource.Attribute.showAnimationBehavior;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.showAsAction = global::pjsua2xamarin.Droid.Resource.Attribute.showAsAction;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.showDelay = global::pjsua2xamarin.Droid.Resource.Attribute.showDelay;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.showDividers = global::pjsua2xamarin.Droid.Resource.Attribute.showDividers;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.showMotionSpec = global::pjsua2xamarin.Droid.Resource.Attribute.showMotionSpec;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.showPaths = global::pjsua2xamarin.Droid.Resource.Attribute.showPaths;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.showText = global::pjsua2xamarin.Droid.Resource.Attribute.showText;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.showTitle = global::pjsua2xamarin.Droid.Resource.Attribute.showTitle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.shrinkMotionSpec = global::pjsua2xamarin.Droid.Resource.Attribute.shrinkMotionSpec;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.singleChoiceItemLayout = global::pjsua2xamarin.Droid.Resource.Attribute.singleChoiceItemLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.singleLine = global::pjsua2xamarin.Droid.Resource.Attribute.singleLine;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.singleSelection = global::pjsua2xamarin.Droid.Resource.Attribute.singleSelection;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.sizePercent = global::pjsua2xamarin.Droid.Resource.Attribute.sizePercent;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.sliderStyle = global::pjsua2xamarin.Droid.Resource.Attribute.sliderStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.snackbarButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.snackbarButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.snackbarStyle = global::pjsua2xamarin.Droid.Resource.Attribute.snackbarStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.snackbarTextViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.snackbarTextViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.spanCount = global::pjsua2xamarin.Droid.Resource.Attribute.spanCount;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.spinBars = global::pjsua2xamarin.Droid.Resource.Attribute.spinBars;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.spinnerDropDownItemStyle = global::pjsua2xamarin.Droid.Resource.Attribute.spinnerDropDownItemStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.spinnerStyle = global::pjsua2xamarin.Droid.Resource.Attribute.spinnerStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.splitTrack = global::pjsua2xamarin.Droid.Resource.Attribute.splitTrack;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.springBoundary = global::pjsua2xamarin.Droid.Resource.Attribute.springBoundary;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.springDamping = global::pjsua2xamarin.Droid.Resource.Attribute.springDamping;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.springMass = global::pjsua2xamarin.Droid.Resource.Attribute.springMass;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.springStiffness = global::pjsua2xamarin.Droid.Resource.Attribute.springStiffness;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.springStopThreshold = global::pjsua2xamarin.Droid.Resource.Attribute.springStopThreshold;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.srcCompat = global::pjsua2xamarin.Droid.Resource.Attribute.srcCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.stackFromEnd = global::pjsua2xamarin.Droid.Resource.Attribute.stackFromEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.staggered = global::pjsua2xamarin.Droid.Resource.Attribute.staggered;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.startDestination = global::pjsua2xamarin.Droid.Resource.Attribute.startDestination;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.startIconCheckable = global::pjsua2xamarin.Droid.Resource.Attribute.startIconCheckable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.startIconContentDescription = global::pjsua2xamarin.Droid.Resource.Attribute.startIconContentDescription;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.startIconDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.startIconDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.startIconTint = global::pjsua2xamarin.Droid.Resource.Attribute.startIconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.startIconTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.startIconTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.state_above_anchor = global::pjsua2xamarin.Droid.Resource.Attribute.state_above_anchor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.state_collapsed = global::pjsua2xamarin.Droid.Resource.Attribute.state_collapsed;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.state_collapsible = global::pjsua2xamarin.Droid.Resource.Attribute.state_collapsible;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.state_dragged = global::pjsua2xamarin.Droid.Resource.Attribute.state_dragged;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.state_liftable = global::pjsua2xamarin.Droid.Resource.Attribute.state_liftable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.state_lifted = global::pjsua2xamarin.Droid.Resource.Attribute.state_lifted;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.statusBarBackground = global::pjsua2xamarin.Droid.Resource.Attribute.statusBarBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.statusBarForeground = global::pjsua2xamarin.Droid.Resource.Attribute.statusBarForeground;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.statusBarScrim = global::pjsua2xamarin.Droid.Resource.Attribute.statusBarScrim;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.strokeColor = global::pjsua2xamarin.Droid.Resource.Attribute.strokeColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.strokeWidth = global::pjsua2xamarin.Droid.Resource.Attribute.strokeWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.subMenuArrow = global::pjsua2xamarin.Droid.Resource.Attribute.subMenuArrow;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.submitBackground = global::pjsua2xamarin.Droid.Resource.Attribute.submitBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.subtitle = global::pjsua2xamarin.Droid.Resource.Attribute.subtitle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.subtitleCentered = global::pjsua2xamarin.Droid.Resource.Attribute.subtitleCentered;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.subtitleTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.subtitleTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.subtitleTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.subtitleTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.subtitleTextStyle = global::pjsua2xamarin.Droid.Resource.Attribute.subtitleTextStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.suffixText = global::pjsua2xamarin.Droid.Resource.Attribute.suffixText;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.suffixTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.suffixTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.suffixTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.suffixTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.suggestionRowLayout = global::pjsua2xamarin.Droid.Resource.Attribute.suggestionRowLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.swipeRefreshLayoutProgressSpinnerBackgroundColor = global::pjsua2xamarin.Droid.Resource.Attribute.swipeRefreshLayoutProgressSpinnerBackgroundColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.switchMinWidth = global::pjsua2xamarin.Droid.Resource.Attribute.switchMinWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.switchPadding = global::pjsua2xamarin.Droid.Resource.Attribute.switchPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.switchStyle = global::pjsua2xamarin.Droid.Resource.Attribute.switchStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.switchTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.switchTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabBackground = global::pjsua2xamarin.Droid.Resource.Attribute.tabBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabContentStart = global::pjsua2xamarin.Droid.Resource.Attribute.tabContentStart;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabGravity = global::pjsua2xamarin.Droid.Resource.Attribute.tabGravity;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabIconTint = global::pjsua2xamarin.Droid.Resource.Attribute.tabIconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabIconTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.tabIconTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabIndicator = global::pjsua2xamarin.Droid.Resource.Attribute.tabIndicator;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabIndicatorAnimationDuration = global::pjsua2xamarin.Droid.Resource.Attribute.tabIndicatorAnimationDuration;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabIndicatorAnimationMode = global::pjsua2xamarin.Droid.Resource.Attribute.tabIndicatorAnimationMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabIndicatorColor = global::pjsua2xamarin.Droid.Resource.Attribute.tabIndicatorColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabIndicatorFullWidth = global::pjsua2xamarin.Droid.Resource.Attribute.tabIndicatorFullWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabIndicatorGravity = global::pjsua2xamarin.Droid.Resource.Attribute.tabIndicatorGravity;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabIndicatorHeight = global::pjsua2xamarin.Droid.Resource.Attribute.tabIndicatorHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabInlineLabel = global::pjsua2xamarin.Droid.Resource.Attribute.tabInlineLabel;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabMaxWidth = global::pjsua2xamarin.Droid.Resource.Attribute.tabMaxWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabMinWidth = global::pjsua2xamarin.Droid.Resource.Attribute.tabMinWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabMode = global::pjsua2xamarin.Droid.Resource.Attribute.tabMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabPadding = global::pjsua2xamarin.Droid.Resource.Attribute.tabPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabPaddingBottom = global::pjsua2xamarin.Droid.Resource.Attribute.tabPaddingBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabPaddingEnd = global::pjsua2xamarin.Droid.Resource.Attribute.tabPaddingEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabPaddingStart = global::pjsua2xamarin.Droid.Resource.Attribute.tabPaddingStart;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabPaddingTop = global::pjsua2xamarin.Droid.Resource.Attribute.tabPaddingTop;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabRippleColor = global::pjsua2xamarin.Droid.Resource.Attribute.tabRippleColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabSelectedTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.tabSelectedTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabStyle = global::pjsua2xamarin.Droid.Resource.Attribute.tabStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.tabTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.tabTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tabUnboundedRipple = global::pjsua2xamarin.Droid.Resource.Attribute.tabUnboundedRipple;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.targetId = global::pjsua2xamarin.Droid.Resource.Attribute.targetId;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.targetPackage = global::pjsua2xamarin.Droid.Resource.Attribute.targetPackage;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.telltales_tailColor = global::pjsua2xamarin.Droid.Resource.Attribute.telltales_tailColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.telltales_tailScale = global::pjsua2xamarin.Droid.Resource.Attribute.telltales_tailScale;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.telltales_velocityMode = global::pjsua2xamarin.Droid.Resource.Attribute.telltales_velocityMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAllCaps = global::pjsua2xamarin.Droid.Resource.Attribute.textAllCaps;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAppearanceBody1 = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceBody1;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAppearanceBody2 = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceBody2;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAppearanceButton = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceButton;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAppearanceCaption = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceCaption;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAppearanceHeadline1 = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceHeadline1;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAppearanceHeadline2 = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceHeadline2;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAppearanceHeadline3 = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceHeadline3;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAppearanceHeadline4 = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceHeadline4;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAppearanceHeadline5 = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceHeadline5;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAppearanceHeadline6 = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceHeadline6;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAppearanceLargePopupMenu = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceLargePopupMenu;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAppearanceLineHeightEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceLineHeightEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAppearanceListItem = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceListItem;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAppearanceListItemSecondary = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceListItemSecondary;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAppearanceListItemSmall = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceListItemSmall;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAppearanceOverline = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceOverline;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAppearancePopupMenuHeader = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearancePopupMenuHeader;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAppearanceSearchResultSubtitle = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceSearchResultSubtitle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAppearanceSearchResultTitle = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceSearchResultTitle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAppearanceSmallPopupMenu = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceSmallPopupMenu;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAppearanceSubtitle1 = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceSubtitle1;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textAppearanceSubtitle2 = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceSubtitle2;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textBackground = global::pjsua2xamarin.Droid.Resource.Attribute.textBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textBackgroundPanX = global::pjsua2xamarin.Droid.Resource.Attribute.textBackgroundPanX;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textBackgroundPanY = global::pjsua2xamarin.Droid.Resource.Attribute.textBackgroundPanY;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textBackgroundRotate = global::pjsua2xamarin.Droid.Resource.Attribute.textBackgroundRotate;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textBackgroundZoom = global::pjsua2xamarin.Droid.Resource.Attribute.textBackgroundZoom;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textColorAlertDialogListItem = global::pjsua2xamarin.Droid.Resource.Attribute.textColorAlertDialogListItem;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textColorSearchUrl = global::pjsua2xamarin.Droid.Resource.Attribute.textColorSearchUrl;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textEndPadding = global::pjsua2xamarin.Droid.Resource.Attribute.textEndPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textFillColor = global::pjsua2xamarin.Droid.Resource.Attribute.textFillColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textInputLayoutFocusedRectEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.textInputLayoutFocusedRectEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textInputStyle = global::pjsua2xamarin.Droid.Resource.Attribute.textInputStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textLocale = global::pjsua2xamarin.Droid.Resource.Attribute.textLocale;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textOutlineColor = global::pjsua2xamarin.Droid.Resource.Attribute.textOutlineColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textOutlineThickness = global::pjsua2xamarin.Droid.Resource.Attribute.textOutlineThickness;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textPanX = global::pjsua2xamarin.Droid.Resource.Attribute.textPanX;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textPanY = global::pjsua2xamarin.Droid.Resource.Attribute.textPanY;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textStartPadding = global::pjsua2xamarin.Droid.Resource.Attribute.textStartPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textureBlurFactor = global::pjsua2xamarin.Droid.Resource.Attribute.textureBlurFactor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textureEffect = global::pjsua2xamarin.Droid.Resource.Attribute.textureEffect;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textureHeight = global::pjsua2xamarin.Droid.Resource.Attribute.textureHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.textureWidth = global::pjsua2xamarin.Droid.Resource.Attribute.textureWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.theme = global::pjsua2xamarin.Droid.Resource.Attribute.theme;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.themeLineHeight = global::pjsua2xamarin.Droid.Resource.Attribute.themeLineHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.thickness = global::pjsua2xamarin.Droid.Resource.Attribute.thickness;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.thumbColor = global::pjsua2xamarin.Droid.Resource.Attribute.thumbColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.thumbElevation = global::pjsua2xamarin.Droid.Resource.Attribute.thumbElevation;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.thumbRadius = global::pjsua2xamarin.Droid.Resource.Attribute.thumbRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.thumbStrokeColor = global::pjsua2xamarin.Droid.Resource.Attribute.thumbStrokeColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.thumbStrokeWidth = global::pjsua2xamarin.Droid.Resource.Attribute.thumbStrokeWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.thumbTextPadding = global::pjsua2xamarin.Droid.Resource.Attribute.thumbTextPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.thumbTint = global::pjsua2xamarin.Droid.Resource.Attribute.thumbTint;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.thumbTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.thumbTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tickColor = global::pjsua2xamarin.Droid.Resource.Attribute.tickColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tickColorActive = global::pjsua2xamarin.Droid.Resource.Attribute.tickColorActive;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tickColorInactive = global::pjsua2xamarin.Droid.Resource.Attribute.tickColorInactive;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tickMark = global::pjsua2xamarin.Droid.Resource.Attribute.tickMark;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tickMarkTint = global::pjsua2xamarin.Droid.Resource.Attribute.tickMarkTint;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tickMarkTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.tickMarkTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tickVisible = global::pjsua2xamarin.Droid.Resource.Attribute.tickVisible;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tint = global::pjsua2xamarin.Droid.Resource.Attribute.tint;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tintMode = global::pjsua2xamarin.Droid.Resource.Attribute.tintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.title = global::pjsua2xamarin.Droid.Resource.Attribute.title;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.titleCentered = global::pjsua2xamarin.Droid.Resource.Attribute.titleCentered;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.titleCollapseMode = global::pjsua2xamarin.Droid.Resource.Attribute.titleCollapseMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.titleEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.titleEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.titleMargin = global::pjsua2xamarin.Droid.Resource.Attribute.titleMargin;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.titleMarginBottom = global::pjsua2xamarin.Droid.Resource.Attribute.titleMarginBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.titleMarginEnd = global::pjsua2xamarin.Droid.Resource.Attribute.titleMarginEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.titleMargins = global::pjsua2xamarin.Droid.Resource.Attribute.titleMargins;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.titleMarginStart = global::pjsua2xamarin.Droid.Resource.Attribute.titleMarginStart;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.titleMarginTop = global::pjsua2xamarin.Droid.Resource.Attribute.titleMarginTop;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.titleTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.titleTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.titleTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.titleTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.titleTextStyle = global::pjsua2xamarin.Droid.Resource.Attribute.titleTextStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.toolbarId = global::pjsua2xamarin.Droid.Resource.Attribute.toolbarId;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.toolbarNavigationButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.toolbarNavigationButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.toolbarStyle = global::pjsua2xamarin.Droid.Resource.Attribute.toolbarStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tooltipForegroundColor = global::pjsua2xamarin.Droid.Resource.Attribute.tooltipForegroundColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tooltipFrameBackground = global::pjsua2xamarin.Droid.Resource.Attribute.tooltipFrameBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tooltipStyle = global::pjsua2xamarin.Droid.Resource.Attribute.tooltipStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.tooltipText = global::pjsua2xamarin.Droid.Resource.Attribute.tooltipText;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.touchAnchorId = global::pjsua2xamarin.Droid.Resource.Attribute.touchAnchorId;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.touchAnchorSide = global::pjsua2xamarin.Droid.Resource.Attribute.touchAnchorSide;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.touchRegionId = global::pjsua2xamarin.Droid.Resource.Attribute.touchRegionId;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.track = global::pjsua2xamarin.Droid.Resource.Attribute.track;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.trackColor = global::pjsua2xamarin.Droid.Resource.Attribute.trackColor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.trackColorActive = global::pjsua2xamarin.Droid.Resource.Attribute.trackColorActive;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.trackColorInactive = global::pjsua2xamarin.Droid.Resource.Attribute.trackColorInactive;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.trackCornerRadius = global::pjsua2xamarin.Droid.Resource.Attribute.trackCornerRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.trackHeight = global::pjsua2xamarin.Droid.Resource.Attribute.trackHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.trackThickness = global::pjsua2xamarin.Droid.Resource.Attribute.trackThickness;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.trackTint = global::pjsua2xamarin.Droid.Resource.Attribute.trackTint;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.trackTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.trackTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.transformPivotTarget = global::pjsua2xamarin.Droid.Resource.Attribute.transformPivotTarget;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.transitionDisable = global::pjsua2xamarin.Droid.Resource.Attribute.transitionDisable;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.transitionEasing = global::pjsua2xamarin.Droid.Resource.Attribute.transitionEasing;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.transitionFlags = global::pjsua2xamarin.Droid.Resource.Attribute.transitionFlags;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.transitionPathRotate = global::pjsua2xamarin.Droid.Resource.Attribute.transitionPathRotate;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.transitionShapeAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.transitionShapeAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.triggerId = global::pjsua2xamarin.Droid.Resource.Attribute.triggerId;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.triggerReceiver = global::pjsua2xamarin.Droid.Resource.Attribute.triggerReceiver;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.triggerSlack = global::pjsua2xamarin.Droid.Resource.Attribute.triggerSlack;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.ttcIndex = global::pjsua2xamarin.Droid.Resource.Attribute.ttcIndex;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.upDuration = global::pjsua2xamarin.Droid.Resource.Attribute.upDuration;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.uri = global::pjsua2xamarin.Droid.Resource.Attribute.uri;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.useCompatPadding = global::pjsua2xamarin.Droid.Resource.Attribute.useCompatPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.useMaterialThemeColors = global::pjsua2xamarin.Droid.Resource.Attribute.useMaterialThemeColors;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.values = global::pjsua2xamarin.Droid.Resource.Attribute.values;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.verticalOffset = global::pjsua2xamarin.Droid.Resource.Attribute.verticalOffset;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.viewInflaterClass = global::pjsua2xamarin.Droid.Resource.Attribute.viewInflaterClass;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.viewTransitionMode = global::pjsua2xamarin.Droid.Resource.Attribute.viewTransitionMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.viewTransitionOnCross = global::pjsua2xamarin.Droid.Resource.Attribute.viewTransitionOnCross;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.viewTransitionOnNegativeCross = global::pjsua2xamarin.Droid.Resource.Attribute.viewTransitionOnNegativeCross;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.viewTransitionOnPositiveCross = global::pjsua2xamarin.Droid.Resource.Attribute.viewTransitionOnPositiveCross;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.visibilityMode = global::pjsua2xamarin.Droid.Resource.Attribute.visibilityMode;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.voiceIcon = global::pjsua2xamarin.Droid.Resource.Attribute.voiceIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.warmth = global::pjsua2xamarin.Droid.Resource.Attribute.warmth;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.waveDecay = global::pjsua2xamarin.Droid.Resource.Attribute.waveDecay;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.waveOffset = global::pjsua2xamarin.Droid.Resource.Attribute.waveOffset;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.wavePeriod = global::pjsua2xamarin.Droid.Resource.Attribute.wavePeriod;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.wavePhase = global::pjsua2xamarin.Droid.Resource.Attribute.wavePhase;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.waveShape = global::pjsua2xamarin.Droid.Resource.Attribute.waveShape;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.waveVariesBy = global::pjsua2xamarin.Droid.Resource.Attribute.waveVariesBy;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.windowActionBar = global::pjsua2xamarin.Droid.Resource.Attribute.windowActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.windowActionBarOverlay = global::pjsua2xamarin.Droid.Resource.Attribute.windowActionBarOverlay;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.windowActionModeOverlay = global::pjsua2xamarin.Droid.Resource.Attribute.windowActionModeOverlay;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.windowFixedHeightMajor = global::pjsua2xamarin.Droid.Resource.Attribute.windowFixedHeightMajor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.windowFixedHeightMinor = global::pjsua2xamarin.Droid.Resource.Attribute.windowFixedHeightMinor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.windowFixedWidthMajor = global::pjsua2xamarin.Droid.Resource.Attribute.windowFixedWidthMajor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.windowFixedWidthMinor = global::pjsua2xamarin.Droid.Resource.Attribute.windowFixedWidthMinor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.windowMinWidthMajor = global::pjsua2xamarin.Droid.Resource.Attribute.windowMinWidthMajor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.windowMinWidthMinor = global::pjsua2xamarin.Droid.Resource.Attribute.windowMinWidthMinor;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.windowNoTitle = global::pjsua2xamarin.Droid.Resource.Attribute.windowNoTitle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.yearSelectedStyle = global::pjsua2xamarin.Droid.Resource.Attribute.yearSelectedStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.yearStyle = global::pjsua2xamarin.Droid.Resource.Attribute.yearStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Attribute.yearTodayStyle = global::pjsua2xamarin.Droid.Resource.Attribute.yearTodayStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Boolean.abc_action_bar_embed_tabs = global::pjsua2xamarin.Droid.Resource.Boolean.abc_action_bar_embed_tabs;
			global::Xamarin.Forms.Platform.Android.Resource.Boolean.abc_config_actionMenuItemAllCaps = global::pjsua2xamarin.Droid.Resource.Boolean.abc_config_actionMenuItemAllCaps;
			global::Xamarin.Forms.Platform.Android.Resource.Boolean.mtrl_btn_textappearance_all_caps = global::pjsua2xamarin.Droid.Resource.Boolean.mtrl_btn_textappearance_all_caps;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_background_cache_hint_selector_material_dark = global::pjsua2xamarin.Droid.Resource.Color.abc_background_cache_hint_selector_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_background_cache_hint_selector_material_light = global::pjsua2xamarin.Droid.Resource.Color.abc_background_cache_hint_selector_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_btn_colored_borderless_text_material = global::pjsua2xamarin.Droid.Resource.Color.abc_btn_colored_borderless_text_material;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_btn_colored_text_material = global::pjsua2xamarin.Droid.Resource.Color.abc_btn_colored_text_material;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_color_highlight_material = global::pjsua2xamarin.Droid.Resource.Color.abc_color_highlight_material;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_decor_view_status_guard = global::pjsua2xamarin.Droid.Resource.Color.abc_decor_view_status_guard;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_decor_view_status_guard_light = global::pjsua2xamarin.Droid.Resource.Color.abc_decor_view_status_guard_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_hint_foreground_material_dark = global::pjsua2xamarin.Droid.Resource.Color.abc_hint_foreground_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_hint_foreground_material_light = global::pjsua2xamarin.Droid.Resource.Color.abc_hint_foreground_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_primary_text_disable_only_material_dark = global::pjsua2xamarin.Droid.Resource.Color.abc_primary_text_disable_only_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_primary_text_disable_only_material_light = global::pjsua2xamarin.Droid.Resource.Color.abc_primary_text_disable_only_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_primary_text_material_dark = global::pjsua2xamarin.Droid.Resource.Color.abc_primary_text_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_primary_text_material_light = global::pjsua2xamarin.Droid.Resource.Color.abc_primary_text_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_search_url_text = global::pjsua2xamarin.Droid.Resource.Color.abc_search_url_text;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_search_url_text_normal = global::pjsua2xamarin.Droid.Resource.Color.abc_search_url_text_normal;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_search_url_text_pressed = global::pjsua2xamarin.Droid.Resource.Color.abc_search_url_text_pressed;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_search_url_text_selected = global::pjsua2xamarin.Droid.Resource.Color.abc_search_url_text_selected;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_secondary_text_material_dark = global::pjsua2xamarin.Droid.Resource.Color.abc_secondary_text_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_secondary_text_material_light = global::pjsua2xamarin.Droid.Resource.Color.abc_secondary_text_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_tint_btn_checkable = global::pjsua2xamarin.Droid.Resource.Color.abc_tint_btn_checkable;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_tint_default = global::pjsua2xamarin.Droid.Resource.Color.abc_tint_default;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_tint_edittext = global::pjsua2xamarin.Droid.Resource.Color.abc_tint_edittext;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_tint_seek_thumb = global::pjsua2xamarin.Droid.Resource.Color.abc_tint_seek_thumb;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_tint_spinner = global::pjsua2xamarin.Droid.Resource.Color.abc_tint_spinner;
			global::Xamarin.Forms.Platform.Android.Resource.Color.abc_tint_switch_track = global::pjsua2xamarin.Droid.Resource.Color.abc_tint_switch_track;
			global::Xamarin.Forms.Platform.Android.Resource.Color.accent_material_dark = global::pjsua2xamarin.Droid.Resource.Color.accent_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.accent_material_light = global::pjsua2xamarin.Droid.Resource.Color.accent_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.androidx_core_ripple_material_light = global::pjsua2xamarin.Droid.Resource.Color.androidx_core_ripple_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.androidx_core_secondary_text_default_material_light = global::pjsua2xamarin.Droid.Resource.Color.androidx_core_secondary_text_default_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.background_floating_material_dark = global::pjsua2xamarin.Droid.Resource.Color.background_floating_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.background_floating_material_light = global::pjsua2xamarin.Droid.Resource.Color.background_floating_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.background_material_dark = global::pjsua2xamarin.Droid.Resource.Color.background_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.background_material_light = global::pjsua2xamarin.Droid.Resource.Color.background_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.bright_foreground_disabled_material_dark = global::pjsua2xamarin.Droid.Resource.Color.bright_foreground_disabled_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.bright_foreground_disabled_material_light = global::pjsua2xamarin.Droid.Resource.Color.bright_foreground_disabled_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.bright_foreground_inverse_material_dark = global::pjsua2xamarin.Droid.Resource.Color.bright_foreground_inverse_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.bright_foreground_inverse_material_light = global::pjsua2xamarin.Droid.Resource.Color.bright_foreground_inverse_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.bright_foreground_material_dark = global::pjsua2xamarin.Droid.Resource.Color.bright_foreground_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.bright_foreground_material_light = global::pjsua2xamarin.Droid.Resource.Color.bright_foreground_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.button_material_dark = global::pjsua2xamarin.Droid.Resource.Color.button_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.button_material_light = global::pjsua2xamarin.Droid.Resource.Color.button_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.cardview_dark_background = global::pjsua2xamarin.Droid.Resource.Color.cardview_dark_background;
			global::Xamarin.Forms.Platform.Android.Resource.Color.cardview_light_background = global::pjsua2xamarin.Droid.Resource.Color.cardview_light_background;
			global::Xamarin.Forms.Platform.Android.Resource.Color.cardview_shadow_end_color = global::pjsua2xamarin.Droid.Resource.Color.cardview_shadow_end_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.cardview_shadow_start_color = global::pjsua2xamarin.Droid.Resource.Color.cardview_shadow_start_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.checkbox_themeable_attribute_color = global::pjsua2xamarin.Droid.Resource.Color.checkbox_themeable_attribute_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_bottom_navigation_shadow_color = global::pjsua2xamarin.Droid.Resource.Color.design_bottom_navigation_shadow_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_box_stroke_color = global::pjsua2xamarin.Droid.Resource.Color.design_box_stroke_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_dark_default_color_background = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_background;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_dark_default_color_error = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_error;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_dark_default_color_on_background = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_on_background;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_dark_default_color_on_error = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_on_error;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_dark_default_color_on_primary = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_on_primary;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_dark_default_color_on_secondary = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_on_secondary;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_dark_default_color_on_surface = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_on_surface;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_dark_default_color_primary = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_primary;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_dark_default_color_primary_dark = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_primary_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_dark_default_color_primary_variant = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_primary_variant;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_dark_default_color_secondary = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_secondary;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_dark_default_color_secondary_variant = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_secondary_variant;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_dark_default_color_surface = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_surface;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_default_color_background = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_background;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_default_color_error = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_error;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_default_color_on_background = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_on_background;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_default_color_on_error = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_on_error;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_default_color_on_primary = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_on_primary;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_default_color_on_secondary = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_on_secondary;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_default_color_on_surface = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_on_surface;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_default_color_primary = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_primary;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_default_color_primary_dark = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_primary_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_default_color_primary_variant = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_primary_variant;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_default_color_secondary = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_secondary;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_default_color_secondary_variant = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_secondary_variant;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_default_color_surface = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_surface;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_error = global::pjsua2xamarin.Droid.Resource.Color.design_error;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_fab_shadow_end_color = global::pjsua2xamarin.Droid.Resource.Color.design_fab_shadow_end_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_fab_shadow_mid_color = global::pjsua2xamarin.Droid.Resource.Color.design_fab_shadow_mid_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_fab_shadow_start_color = global::pjsua2xamarin.Droid.Resource.Color.design_fab_shadow_start_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_fab_stroke_end_inner_color = global::pjsua2xamarin.Droid.Resource.Color.design_fab_stroke_end_inner_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_fab_stroke_end_outer_color = global::pjsua2xamarin.Droid.Resource.Color.design_fab_stroke_end_outer_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_fab_stroke_top_inner_color = global::pjsua2xamarin.Droid.Resource.Color.design_fab_stroke_top_inner_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_fab_stroke_top_outer_color = global::pjsua2xamarin.Droid.Resource.Color.design_fab_stroke_top_outer_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_icon_tint = global::pjsua2xamarin.Droid.Resource.Color.design_icon_tint;
			global::Xamarin.Forms.Platform.Android.Resource.Color.design_snackbar_background_color = global::pjsua2xamarin.Droid.Resource.Color.design_snackbar_background_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.dim_foreground_disabled_material_dark = global::pjsua2xamarin.Droid.Resource.Color.dim_foreground_disabled_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.dim_foreground_disabled_material_light = global::pjsua2xamarin.Droid.Resource.Color.dim_foreground_disabled_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.dim_foreground_material_dark = global::pjsua2xamarin.Droid.Resource.Color.dim_foreground_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.dim_foreground_material_light = global::pjsua2xamarin.Droid.Resource.Color.dim_foreground_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.error_color_material_dark = global::pjsua2xamarin.Droid.Resource.Color.error_color_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.error_color_material_light = global::pjsua2xamarin.Droid.Resource.Color.error_color_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.foreground_material_dark = global::pjsua2xamarin.Droid.Resource.Color.foreground_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.foreground_material_light = global::pjsua2xamarin.Droid.Resource.Color.foreground_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.highlighted_text_material_dark = global::pjsua2xamarin.Droid.Resource.Color.highlighted_text_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.highlighted_text_material_light = global::pjsua2xamarin.Droid.Resource.Color.highlighted_text_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_blue_grey_800 = global::pjsua2xamarin.Droid.Resource.Color.material_blue_grey_800;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_blue_grey_900 = global::pjsua2xamarin.Droid.Resource.Color.material_blue_grey_900;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_blue_grey_950 = global::pjsua2xamarin.Droid.Resource.Color.material_blue_grey_950;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_cursor_color = global::pjsua2xamarin.Droid.Resource.Color.material_cursor_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_deep_teal_200 = global::pjsua2xamarin.Droid.Resource.Color.material_deep_teal_200;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_deep_teal_500 = global::pjsua2xamarin.Droid.Resource.Color.material_deep_teal_500;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_grey_100 = global::pjsua2xamarin.Droid.Resource.Color.material_grey_100;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_grey_300 = global::pjsua2xamarin.Droid.Resource.Color.material_grey_300;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_grey_50 = global::pjsua2xamarin.Droid.Resource.Color.material_grey_50;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_grey_600 = global::pjsua2xamarin.Droid.Resource.Color.material_grey_600;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_grey_800 = global::pjsua2xamarin.Droid.Resource.Color.material_grey_800;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_grey_850 = global::pjsua2xamarin.Droid.Resource.Color.material_grey_850;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_grey_900 = global::pjsua2xamarin.Droid.Resource.Color.material_grey_900;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_on_background_disabled = global::pjsua2xamarin.Droid.Resource.Color.material_on_background_disabled;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_on_background_emphasis_high_type = global::pjsua2xamarin.Droid.Resource.Color.material_on_background_emphasis_high_type;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_on_background_emphasis_medium = global::pjsua2xamarin.Droid.Resource.Color.material_on_background_emphasis_medium;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_on_primary_disabled = global::pjsua2xamarin.Droid.Resource.Color.material_on_primary_disabled;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_on_primary_emphasis_high_type = global::pjsua2xamarin.Droid.Resource.Color.material_on_primary_emphasis_high_type;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_on_primary_emphasis_medium = global::pjsua2xamarin.Droid.Resource.Color.material_on_primary_emphasis_medium;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_on_surface_disabled = global::pjsua2xamarin.Droid.Resource.Color.material_on_surface_disabled;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_on_surface_emphasis_high_type = global::pjsua2xamarin.Droid.Resource.Color.material_on_surface_emphasis_high_type;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_on_surface_emphasis_medium = global::pjsua2xamarin.Droid.Resource.Color.material_on_surface_emphasis_medium;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_on_surface_stroke = global::pjsua2xamarin.Droid.Resource.Color.material_on_surface_stroke;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_slider_active_tick_marks_color = global::pjsua2xamarin.Droid.Resource.Color.material_slider_active_tick_marks_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_slider_active_track_color = global::pjsua2xamarin.Droid.Resource.Color.material_slider_active_track_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_slider_halo_color = global::pjsua2xamarin.Droid.Resource.Color.material_slider_halo_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_slider_inactive_tick_marks_color = global::pjsua2xamarin.Droid.Resource.Color.material_slider_inactive_tick_marks_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_slider_inactive_track_color = global::pjsua2xamarin.Droid.Resource.Color.material_slider_inactive_track_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_slider_thumb_color = global::pjsua2xamarin.Droid.Resource.Color.material_slider_thumb_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_timepicker_button_background = global::pjsua2xamarin.Droid.Resource.Color.material_timepicker_button_background;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_timepicker_button_stroke = global::pjsua2xamarin.Droid.Resource.Color.material_timepicker_button_stroke;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_timepicker_clockface = global::pjsua2xamarin.Droid.Resource.Color.material_timepicker_clockface;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_timepicker_clock_text_color = global::pjsua2xamarin.Droid.Resource.Color.material_timepicker_clock_text_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.material_timepicker_modebutton_tint = global::pjsua2xamarin.Droid.Resource.Color.material_timepicker_modebutton_tint;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_btn_bg_color_selector = global::pjsua2xamarin.Droid.Resource.Color.mtrl_btn_bg_color_selector;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_btn_ripple_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_btn_ripple_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_btn_stroke_color_selector = global::pjsua2xamarin.Droid.Resource.Color.mtrl_btn_stroke_color_selector;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_btn_text_btn_bg_color_selector = global::pjsua2xamarin.Droid.Resource.Color.mtrl_btn_text_btn_bg_color_selector;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_btn_text_btn_ripple_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_btn_text_btn_ripple_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_btn_text_color_disabled = global::pjsua2xamarin.Droid.Resource.Color.mtrl_btn_text_color_disabled;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_btn_text_color_selector = global::pjsua2xamarin.Droid.Resource.Color.mtrl_btn_text_color_selector;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_btn_transparent_bg_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_btn_transparent_bg_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_calendar_item_stroke_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_calendar_item_stroke_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_calendar_selected_range = global::pjsua2xamarin.Droid.Resource.Color.mtrl_calendar_selected_range;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_card_view_foreground = global::pjsua2xamarin.Droid.Resource.Color.mtrl_card_view_foreground;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_card_view_ripple = global::pjsua2xamarin.Droid.Resource.Color.mtrl_card_view_ripple;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_chip_background_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_chip_background_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_chip_close_icon_tint = global::pjsua2xamarin.Droid.Resource.Color.mtrl_chip_close_icon_tint;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_chip_surface_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_chip_surface_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_chip_text_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_chip_text_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_choice_chip_background_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_choice_chip_background_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_choice_chip_ripple_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_choice_chip_ripple_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_choice_chip_text_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_choice_chip_text_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_error = global::pjsua2xamarin.Droid.Resource.Color.mtrl_error;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_fab_bg_color_selector = global::pjsua2xamarin.Droid.Resource.Color.mtrl_fab_bg_color_selector;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_fab_icon_text_color_selector = global::pjsua2xamarin.Droid.Resource.Color.mtrl_fab_icon_text_color_selector;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_fab_ripple_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_fab_ripple_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_filled_background_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_filled_background_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_filled_icon_tint = global::pjsua2xamarin.Droid.Resource.Color.mtrl_filled_icon_tint;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_filled_stroke_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_filled_stroke_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_indicator_text_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_indicator_text_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_navigation_bar_colored_item_tint = global::pjsua2xamarin.Droid.Resource.Color.mtrl_navigation_bar_colored_item_tint;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_navigation_bar_colored_ripple_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_navigation_bar_colored_ripple_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_navigation_bar_item_tint = global::pjsua2xamarin.Droid.Resource.Color.mtrl_navigation_bar_item_tint;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_navigation_bar_ripple_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_navigation_bar_ripple_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_navigation_item_background_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_navigation_item_background_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_navigation_item_icon_tint = global::pjsua2xamarin.Droid.Resource.Color.mtrl_navigation_item_icon_tint;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_navigation_item_text_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_navigation_item_text_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_on_primary_text_btn_text_color_selector = global::pjsua2xamarin.Droid.Resource.Color.mtrl_on_primary_text_btn_text_color_selector;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_on_surface_ripple_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_on_surface_ripple_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_outlined_icon_tint = global::pjsua2xamarin.Droid.Resource.Color.mtrl_outlined_icon_tint;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_outlined_stroke_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_outlined_stroke_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_popupmenu_overlay_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_popupmenu_overlay_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_scrim_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_scrim_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_tabs_colored_ripple_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_tabs_colored_ripple_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_tabs_icon_color_selector = global::pjsua2xamarin.Droid.Resource.Color.mtrl_tabs_icon_color_selector;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_tabs_icon_color_selector_colored = global::pjsua2xamarin.Droid.Resource.Color.mtrl_tabs_icon_color_selector_colored;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_tabs_legacy_text_color_selector = global::pjsua2xamarin.Droid.Resource.Color.mtrl_tabs_legacy_text_color_selector;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_tabs_ripple_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_tabs_ripple_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_textinput_default_box_stroke_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_textinput_default_box_stroke_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_textinput_disabled_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_textinput_disabled_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_textinput_filled_box_default_background_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_textinput_filled_box_default_background_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_textinput_focused_box_stroke_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_textinput_focused_box_stroke_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_textinput_hovered_box_stroke_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_textinput_hovered_box_stroke_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.mtrl_text_btn_text_color_selector = global::pjsua2xamarin.Droid.Resource.Color.mtrl_text_btn_text_color_selector;
			global::Xamarin.Forms.Platform.Android.Resource.Color.notification_action_color_filter = global::pjsua2xamarin.Droid.Resource.Color.notification_action_color_filter;
			global::Xamarin.Forms.Platform.Android.Resource.Color.notification_icon_bg_color = global::pjsua2xamarin.Droid.Resource.Color.notification_icon_bg_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.notification_material_background_media_default_color = global::pjsua2xamarin.Droid.Resource.Color.notification_material_background_media_default_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.primary_dark_material_dark = global::pjsua2xamarin.Droid.Resource.Color.primary_dark_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.primary_dark_material_light = global::pjsua2xamarin.Droid.Resource.Color.primary_dark_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.primary_material_dark = global::pjsua2xamarin.Droid.Resource.Color.primary_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.primary_material_light = global::pjsua2xamarin.Droid.Resource.Color.primary_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.primary_text_default_material_dark = global::pjsua2xamarin.Droid.Resource.Color.primary_text_default_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.primary_text_default_material_light = global::pjsua2xamarin.Droid.Resource.Color.primary_text_default_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.primary_text_disabled_material_dark = global::pjsua2xamarin.Droid.Resource.Color.primary_text_disabled_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.primary_text_disabled_material_light = global::pjsua2xamarin.Droid.Resource.Color.primary_text_disabled_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.radiobutton_themeable_attribute_color = global::pjsua2xamarin.Droid.Resource.Color.radiobutton_themeable_attribute_color;
			global::Xamarin.Forms.Platform.Android.Resource.Color.ripple_material_dark = global::pjsua2xamarin.Droid.Resource.Color.ripple_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.ripple_material_light = global::pjsua2xamarin.Droid.Resource.Color.ripple_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.secondary_text_default_material_dark = global::pjsua2xamarin.Droid.Resource.Color.secondary_text_default_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.secondary_text_default_material_light = global::pjsua2xamarin.Droid.Resource.Color.secondary_text_default_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.secondary_text_disabled_material_dark = global::pjsua2xamarin.Droid.Resource.Color.secondary_text_disabled_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.secondary_text_disabled_material_light = global::pjsua2xamarin.Droid.Resource.Color.secondary_text_disabled_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.switch_thumb_disabled_material_dark = global::pjsua2xamarin.Droid.Resource.Color.switch_thumb_disabled_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.switch_thumb_disabled_material_light = global::pjsua2xamarin.Droid.Resource.Color.switch_thumb_disabled_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.switch_thumb_material_dark = global::pjsua2xamarin.Droid.Resource.Color.switch_thumb_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.switch_thumb_material_light = global::pjsua2xamarin.Droid.Resource.Color.switch_thumb_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.switch_thumb_normal_material_dark = global::pjsua2xamarin.Droid.Resource.Color.switch_thumb_normal_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.switch_thumb_normal_material_light = global::pjsua2xamarin.Droid.Resource.Color.switch_thumb_normal_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Color.test_mtrl_calendar_day = global::pjsua2xamarin.Droid.Resource.Color.test_mtrl_calendar_day;
			global::Xamarin.Forms.Platform.Android.Resource.Color.test_mtrl_calendar_day_selected = global::pjsua2xamarin.Droid.Resource.Color.test_mtrl_calendar_day_selected;
			global::Xamarin.Forms.Platform.Android.Resource.Color.tooltip_background_dark = global::pjsua2xamarin.Droid.Resource.Color.tooltip_background_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Color.tooltip_background_light = global::pjsua2xamarin.Droid.Resource.Color.tooltip_background_light;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_action_bar_content_inset_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_content_inset_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_action_bar_content_inset_with_nav = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_content_inset_with_nav;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_action_bar_default_height_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_default_height_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_action_bar_default_padding_end_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_default_padding_end_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_action_bar_default_padding_start_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_default_padding_start_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_action_bar_elevation_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_elevation_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_action_bar_icon_vertical_padding_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_icon_vertical_padding_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_action_bar_overflow_padding_end_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_overflow_padding_end_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_action_bar_overflow_padding_start_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_overflow_padding_start_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_action_bar_stacked_max_height = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_stacked_max_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_action_bar_stacked_tab_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_stacked_tab_max_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_action_bar_subtitle_bottom_margin_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_subtitle_bottom_margin_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_action_bar_subtitle_top_margin_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_subtitle_top_margin_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_action_button_min_height_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_button_min_height_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_action_button_min_width_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_button_min_width_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_action_button_min_width_overflow_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_button_min_width_overflow_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_alert_dialog_button_bar_height = global::pjsua2xamarin.Droid.Resource.Dimension.abc_alert_dialog_button_bar_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_alert_dialog_button_dimen = global::pjsua2xamarin.Droid.Resource.Dimension.abc_alert_dialog_button_dimen;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_button_inset_horizontal_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_button_inset_horizontal_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_button_inset_vertical_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_button_inset_vertical_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_button_padding_horizontal_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_button_padding_horizontal_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_button_padding_vertical_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_button_padding_vertical_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_cascading_menus_min_smallest_width = global::pjsua2xamarin.Droid.Resource.Dimension.abc_cascading_menus_min_smallest_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_config_prefDialogWidth = global::pjsua2xamarin.Droid.Resource.Dimension.abc_config_prefDialogWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_control_corner_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_control_corner_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_control_inset_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_control_inset_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_control_padding_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_control_padding_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_dialog_corner_radius_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_corner_radius_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_dialog_fixed_height_major = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_fixed_height_major;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_dialog_fixed_height_minor = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_fixed_height_minor;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_dialog_fixed_width_major = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_fixed_width_major;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_dialog_fixed_width_minor = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_fixed_width_minor;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_dialog_list_padding_bottom_no_buttons = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_list_padding_bottom_no_buttons;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_dialog_list_padding_top_no_title = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_list_padding_top_no_title;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_dialog_min_width_major = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_min_width_major;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_dialog_min_width_minor = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_min_width_minor;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_dialog_padding_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_padding_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_dialog_padding_top_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_padding_top_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_dialog_title_divider_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_title_divider_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_disabled_alpha_material_dark = global::pjsua2xamarin.Droid.Resource.Dimension.abc_disabled_alpha_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_disabled_alpha_material_light = global::pjsua2xamarin.Droid.Resource.Dimension.abc_disabled_alpha_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_dropdownitem_icon_width = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dropdownitem_icon_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_dropdownitem_text_padding_left = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dropdownitem_text_padding_left;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_dropdownitem_text_padding_right = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dropdownitem_text_padding_right;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_edit_text_inset_bottom_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_edit_text_inset_bottom_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_edit_text_inset_horizontal_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_edit_text_inset_horizontal_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_edit_text_inset_top_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_edit_text_inset_top_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_floating_window_z = global::pjsua2xamarin.Droid.Resource.Dimension.abc_floating_window_z;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_list_item_height_large_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_list_item_height_large_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_list_item_height_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_list_item_height_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_list_item_height_small_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_list_item_height_small_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_list_item_padding_horizontal_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_list_item_padding_horizontal_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_panel_menu_list_width = global::pjsua2xamarin.Droid.Resource.Dimension.abc_panel_menu_list_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_progress_bar_height_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_progress_bar_height_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_search_view_preferred_height = global::pjsua2xamarin.Droid.Resource.Dimension.abc_search_view_preferred_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_search_view_preferred_width = global::pjsua2xamarin.Droid.Resource.Dimension.abc_search_view_preferred_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_seekbar_track_background_height_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_seekbar_track_background_height_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_seekbar_track_progress_height_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_seekbar_track_progress_height_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_select_dialog_padding_start_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_select_dialog_padding_start_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_star_big = global::pjsua2xamarin.Droid.Resource.Dimension.abc_star_big;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_star_medium = global::pjsua2xamarin.Droid.Resource.Dimension.abc_star_medium;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_star_small = global::pjsua2xamarin.Droid.Resource.Dimension.abc_star_small;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_switch_padding = global::pjsua2xamarin.Droid.Resource.Dimension.abc_switch_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_text_size_body_1_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_body_1_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_text_size_body_2_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_body_2_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_text_size_button_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_button_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_text_size_caption_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_caption_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_text_size_display_1_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_display_1_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_text_size_display_2_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_display_2_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_text_size_display_3_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_display_3_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_text_size_display_4_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_display_4_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_text_size_headline_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_headline_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_text_size_large_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_large_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_text_size_medium_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_medium_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_text_size_menu_header_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_menu_header_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_text_size_menu_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_menu_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_text_size_small_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_small_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_text_size_subhead_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_subhead_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_text_size_subtitle_material_toolbar = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_subtitle_material_toolbar;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_text_size_title_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_title_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.abc_text_size_title_material_toolbar = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_title_material_toolbar;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.action_bar_size = global::pjsua2xamarin.Droid.Resource.Dimension.action_bar_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.appcompat_dialog_background_inset = global::pjsua2xamarin.Droid.Resource.Dimension.appcompat_dialog_background_inset;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.cardview_compat_inset_shadow = global::pjsua2xamarin.Droid.Resource.Dimension.cardview_compat_inset_shadow;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.cardview_default_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.cardview_default_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.cardview_default_radius = global::pjsua2xamarin.Droid.Resource.Dimension.cardview_default_radius;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.clock_face_margin_start = global::pjsua2xamarin.Droid.Resource.Dimension.clock_face_margin_start;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.compat_button_inset_horizontal_material = global::pjsua2xamarin.Droid.Resource.Dimension.compat_button_inset_horizontal_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.compat_button_inset_vertical_material = global::pjsua2xamarin.Droid.Resource.Dimension.compat_button_inset_vertical_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.compat_button_padding_horizontal_material = global::pjsua2xamarin.Droid.Resource.Dimension.compat_button_padding_horizontal_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.compat_button_padding_vertical_material = global::pjsua2xamarin.Droid.Resource.Dimension.compat_button_padding_vertical_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.compat_control_corner_material = global::pjsua2xamarin.Droid.Resource.Dimension.compat_control_corner_material;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.compat_notification_large_icon_max_height = global::pjsua2xamarin.Droid.Resource.Dimension.compat_notification_large_icon_max_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.compat_notification_large_icon_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.compat_notification_large_icon_max_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.default_dimension = global::pjsua2xamarin.Droid.Resource.Dimension.default_dimension;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.def_drawer_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.def_drawer_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_appbar_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.design_appbar_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_bottom_navigation_active_item_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_active_item_max_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_bottom_navigation_active_item_min_width = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_active_item_min_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_bottom_navigation_active_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_active_text_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_bottom_navigation_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_bottom_navigation_height = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_bottom_navigation_icon_size = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_icon_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_bottom_navigation_item_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_item_max_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_bottom_navigation_item_min_width = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_item_min_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_bottom_navigation_label_padding = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_label_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_bottom_navigation_margin = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_margin;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_bottom_navigation_shadow_height = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_shadow_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_bottom_navigation_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_text_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_bottom_sheet_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_sheet_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_bottom_sheet_modal_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_sheet_modal_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_bottom_sheet_peek_height_min = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_sheet_peek_height_min;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_fab_border_width = global::pjsua2xamarin.Droid.Resource.Dimension.design_fab_border_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_fab_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.design_fab_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_fab_image_size = global::pjsua2xamarin.Droid.Resource.Dimension.design_fab_image_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_fab_size_mini = global::pjsua2xamarin.Droid.Resource.Dimension.design_fab_size_mini;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_fab_size_normal = global::pjsua2xamarin.Droid.Resource.Dimension.design_fab_size_normal;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_fab_translation_z_hovered_focused = global::pjsua2xamarin.Droid.Resource.Dimension.design_fab_translation_z_hovered_focused;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_fab_translation_z_pressed = global::pjsua2xamarin.Droid.Resource.Dimension.design_fab_translation_z_pressed;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_navigation_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.design_navigation_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_navigation_icon_padding = global::pjsua2xamarin.Droid.Resource.Dimension.design_navigation_icon_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_navigation_icon_size = global::pjsua2xamarin.Droid.Resource.Dimension.design_navigation_icon_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_navigation_item_horizontal_padding = global::pjsua2xamarin.Droid.Resource.Dimension.design_navigation_item_horizontal_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_navigation_item_icon_padding = global::pjsua2xamarin.Droid.Resource.Dimension.design_navigation_item_icon_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_navigation_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.design_navigation_max_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_navigation_padding_bottom = global::pjsua2xamarin.Droid.Resource.Dimension.design_navigation_padding_bottom;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_navigation_separator_vertical_padding = global::pjsua2xamarin.Droid.Resource.Dimension.design_navigation_separator_vertical_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_snackbar_action_inline_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.design_snackbar_action_inline_max_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_snackbar_action_text_color_alpha = global::pjsua2xamarin.Droid.Resource.Dimension.design_snackbar_action_text_color_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_snackbar_background_corner_radius = global::pjsua2xamarin.Droid.Resource.Dimension.design_snackbar_background_corner_radius;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_snackbar_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.design_snackbar_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_snackbar_extra_spacing_horizontal = global::pjsua2xamarin.Droid.Resource.Dimension.design_snackbar_extra_spacing_horizontal;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_snackbar_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.design_snackbar_max_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_snackbar_min_width = global::pjsua2xamarin.Droid.Resource.Dimension.design_snackbar_min_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_snackbar_padding_horizontal = global::pjsua2xamarin.Droid.Resource.Dimension.design_snackbar_padding_horizontal;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_snackbar_padding_vertical = global::pjsua2xamarin.Droid.Resource.Dimension.design_snackbar_padding_vertical;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_snackbar_padding_vertical_2lines = global::pjsua2xamarin.Droid.Resource.Dimension.design_snackbar_padding_vertical_2lines;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_snackbar_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.design_snackbar_text_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_tab_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.design_tab_max_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_tab_scrollable_min_width = global::pjsua2xamarin.Droid.Resource.Dimension.design_tab_scrollable_min_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_tab_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.design_tab_text_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_tab_text_size_2line = global::pjsua2xamarin.Droid.Resource.Dimension.design_tab_text_size_2line;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.design_textinput_caption_translate_y = global::pjsua2xamarin.Droid.Resource.Dimension.design_textinput_caption_translate_y;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.disabled_alpha_material_dark = global::pjsua2xamarin.Droid.Resource.Dimension.disabled_alpha_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.disabled_alpha_material_light = global::pjsua2xamarin.Droid.Resource.Dimension.disabled_alpha_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.fastscroll_default_thickness = global::pjsua2xamarin.Droid.Resource.Dimension.fastscroll_default_thickness;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.fastscroll_margin = global::pjsua2xamarin.Droid.Resource.Dimension.fastscroll_margin;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.fastscroll_minimum_range = global::pjsua2xamarin.Droid.Resource.Dimension.fastscroll_minimum_range;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.highlight_alpha_material_colored = global::pjsua2xamarin.Droid.Resource.Dimension.highlight_alpha_material_colored;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.highlight_alpha_material_dark = global::pjsua2xamarin.Droid.Resource.Dimension.highlight_alpha_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.highlight_alpha_material_light = global::pjsua2xamarin.Droid.Resource.Dimension.highlight_alpha_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.hint_alpha_material_dark = global::pjsua2xamarin.Droid.Resource.Dimension.hint_alpha_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.hint_alpha_material_light = global::pjsua2xamarin.Droid.Resource.Dimension.hint_alpha_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.hint_pressed_alpha_material_dark = global::pjsua2xamarin.Droid.Resource.Dimension.hint_pressed_alpha_material_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.hint_pressed_alpha_material_light = global::pjsua2xamarin.Droid.Resource.Dimension.hint_pressed_alpha_material_light;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.item_touch_helper_max_drag_scroll_per_frame = global::pjsua2xamarin.Droid.Resource.Dimension.item_touch_helper_max_drag_scroll_per_frame;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.item_touch_helper_swipe_escape_max_velocity = global::pjsua2xamarin.Droid.Resource.Dimension.item_touch_helper_swipe_escape_max_velocity;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.item_touch_helper_swipe_escape_velocity = global::pjsua2xamarin.Droid.Resource.Dimension.item_touch_helper_swipe_escape_velocity;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_bottom_sheet_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.material_bottom_sheet_max_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_clock_display_padding = global::pjsua2xamarin.Droid.Resource.Dimension.material_clock_display_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_clock_face_margin_top = global::pjsua2xamarin.Droid.Resource.Dimension.material_clock_face_margin_top;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_clock_hand_center_dot_radius = global::pjsua2xamarin.Droid.Resource.Dimension.material_clock_hand_center_dot_radius;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_clock_hand_padding = global::pjsua2xamarin.Droid.Resource.Dimension.material_clock_hand_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_clock_hand_stroke_width = global::pjsua2xamarin.Droid.Resource.Dimension.material_clock_hand_stroke_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_clock_number_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.material_clock_number_text_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_clock_period_toggle_height = global::pjsua2xamarin.Droid.Resource.Dimension.material_clock_period_toggle_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_clock_period_toggle_margin_left = global::pjsua2xamarin.Droid.Resource.Dimension.material_clock_period_toggle_margin_left;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_clock_period_toggle_width = global::pjsua2xamarin.Droid.Resource.Dimension.material_clock_period_toggle_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_clock_size = global::pjsua2xamarin.Droid.Resource.Dimension.material_clock_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_cursor_inset_bottom = global::pjsua2xamarin.Droid.Resource.Dimension.material_cursor_inset_bottom;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_cursor_inset_top = global::pjsua2xamarin.Droid.Resource.Dimension.material_cursor_inset_top;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_cursor_width = global::pjsua2xamarin.Droid.Resource.Dimension.material_cursor_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_emphasis_disabled = global::pjsua2xamarin.Droid.Resource.Dimension.material_emphasis_disabled;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_emphasis_high_type = global::pjsua2xamarin.Droid.Resource.Dimension.material_emphasis_high_type;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_emphasis_medium = global::pjsua2xamarin.Droid.Resource.Dimension.material_emphasis_medium;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_filled_edittext_font_1_3_padding_bottom = global::pjsua2xamarin.Droid.Resource.Dimension.material_filled_edittext_font_1_3_padding_bottom;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_filled_edittext_font_1_3_padding_top = global::pjsua2xamarin.Droid.Resource.Dimension.material_filled_edittext_font_1_3_padding_top;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_filled_edittext_font_2_0_padding_bottom = global::pjsua2xamarin.Droid.Resource.Dimension.material_filled_edittext_font_2_0_padding_bottom;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_filled_edittext_font_2_0_padding_top = global::pjsua2xamarin.Droid.Resource.Dimension.material_filled_edittext_font_2_0_padding_top;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_font_1_3_box_collapsed_padding_top = global::pjsua2xamarin.Droid.Resource.Dimension.material_font_1_3_box_collapsed_padding_top;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_font_2_0_box_collapsed_padding_top = global::pjsua2xamarin.Droid.Resource.Dimension.material_font_2_0_box_collapsed_padding_top;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_helper_text_default_padding_top = global::pjsua2xamarin.Droid.Resource.Dimension.material_helper_text_default_padding_top;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_helper_text_font_1_3_padding_horizontal = global::pjsua2xamarin.Droid.Resource.Dimension.material_helper_text_font_1_3_padding_horizontal;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_helper_text_font_1_3_padding_top = global::pjsua2xamarin.Droid.Resource.Dimension.material_helper_text_font_1_3_padding_top;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_input_text_to_prefix_suffix_padding = global::pjsua2xamarin.Droid.Resource.Dimension.material_input_text_to_prefix_suffix_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_textinput_default_width = global::pjsua2xamarin.Droid.Resource.Dimension.material_textinput_default_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_textinput_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.material_textinput_max_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_textinput_min_width = global::pjsua2xamarin.Droid.Resource.Dimension.material_textinput_min_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_text_view_test_line_height = global::pjsua2xamarin.Droid.Resource.Dimension.material_text_view_test_line_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_text_view_test_line_height_override = global::pjsua2xamarin.Droid.Resource.Dimension.material_text_view_test_line_height_override;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_timepicker_dialog_buttons_margin_top = global::pjsua2xamarin.Droid.Resource.Dimension.material_timepicker_dialog_buttons_margin_top;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_time_picker_minimum_screen_height = global::pjsua2xamarin.Droid.Resource.Dimension.material_time_picker_minimum_screen_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.material_time_picker_minimum_screen_width = global::pjsua2xamarin.Droid.Resource.Dimension.material_time_picker_minimum_screen_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_alert_dialog_background_inset_bottom = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_alert_dialog_background_inset_bottom;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_alert_dialog_background_inset_end = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_alert_dialog_background_inset_end;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_alert_dialog_background_inset_start = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_alert_dialog_background_inset_start;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_alert_dialog_background_inset_top = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_alert_dialog_background_inset_top;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_alert_dialog_picker_background_inset = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_alert_dialog_picker_background_inset;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_badge_horizontal_edge_offset = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_badge_horizontal_edge_offset;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_badge_long_text_horizontal_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_badge_long_text_horizontal_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_badge_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_badge_radius;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_badge_text_horizontal_edge_offset = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_badge_text_horizontal_edge_offset;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_badge_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_badge_text_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_badge_toolbar_action_menu_item_horizontal_offset = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_badge_toolbar_action_menu_item_horizontal_offset;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_badge_toolbar_action_menu_item_vertical_offset = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_badge_toolbar_action_menu_item_vertical_offset;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_badge_with_text_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_badge_with_text_radius;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_bottomappbar_fabOffsetEndMode = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_bottomappbar_fabOffsetEndMode;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_bottomappbar_fab_bottom_margin = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_bottomappbar_fab_bottom_margin;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_bottomappbar_fab_cradle_margin = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_bottomappbar_fab_cradle_margin;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_bottomappbar_fab_cradle_rounded_corner_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_bottomappbar_fab_cradle_rounded_corner_radius;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_bottomappbar_fab_cradle_vertical_offset = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_bottomappbar_fab_cradle_vertical_offset;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_bottomappbar_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_bottomappbar_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_corner_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_corner_radius;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_dialog_btn_min_width = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_dialog_btn_min_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_disabled_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_disabled_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_disabled_z = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_disabled_z;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_focused_z = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_focused_z;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_hovered_z = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_hovered_z;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_icon_btn_padding_left = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_icon_btn_padding_left;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_icon_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_icon_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_inset = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_inset;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_letter_spacing = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_letter_spacing;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_max_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_padding_bottom = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_padding_bottom;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_padding_left = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_padding_left;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_padding_right = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_padding_right;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_padding_top = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_padding_top;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_pressed_z = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_pressed_z;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_snackbar_margin_horizontal = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_snackbar_margin_horizontal;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_stroke_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_stroke_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_text_btn_icon_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_text_btn_icon_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_text_btn_padding_left = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_text_btn_padding_left;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_text_btn_padding_right = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_text_btn_padding_right;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_text_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_btn_z = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_z;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_action_confirm_button_min_width = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_action_confirm_button_min_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_action_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_action_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_action_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_action_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_bottom_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_bottom_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_content_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_content_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_days_of_week_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_days_of_week_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_day_corner = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_day_corner;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_day_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_day_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_day_horizontal_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_day_horizontal_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_day_today_stroke = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_day_today_stroke;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_day_vertical_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_day_vertical_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_day_width = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_day_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_dialog_background_inset = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_dialog_background_inset;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_header_content_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_header_content_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_header_content_padding_fullscreen = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_header_content_padding_fullscreen;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_header_divider_thickness = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_header_divider_thickness;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_header_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_header_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_header_height_fullscreen = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_header_height_fullscreen;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_header_selection_line_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_header_selection_line_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_header_text_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_header_text_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_header_toggle_margin_bottom = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_header_toggle_margin_bottom;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_header_toggle_margin_top = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_header_toggle_margin_top;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_landscape_header_width = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_landscape_header_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_maximum_default_fullscreen_minor_axis = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_maximum_default_fullscreen_minor_axis;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_month_horizontal_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_month_horizontal_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_month_vertical_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_month_vertical_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_navigation_bottom_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_navigation_bottom_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_navigation_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_navigation_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_navigation_top_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_navigation_top_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_pre_l_text_clip_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_pre_l_text_clip_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_selection_baseline_to_top_fullscreen = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_selection_baseline_to_top_fullscreen;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_selection_text_baseline_to_bottom = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_selection_text_baseline_to_bottom;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_selection_text_baseline_to_bottom_fullscreen = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_selection_text_baseline_to_bottom_fullscreen;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_selection_text_baseline_to_top = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_selection_text_baseline_to_top;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_text_input_padding_top = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_text_input_padding_top;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_title_baseline_to_top = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_title_baseline_to_top;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_title_baseline_to_top_fullscreen = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_title_baseline_to_top_fullscreen;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_year_corner = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_year_corner;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_year_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_year_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_year_horizontal_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_year_horizontal_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_year_vertical_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_year_vertical_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_calendar_year_width = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_year_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_card_checked_icon_margin = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_card_checked_icon_margin;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_card_checked_icon_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_card_checked_icon_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_card_corner_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_card_corner_radius;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_card_dragged_z = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_card_dragged_z;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_card_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_card_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_card_spacing = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_card_spacing;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_chip_pressed_translation_z = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_chip_pressed_translation_z;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_chip_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_chip_text_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_edittext_rectangle_top_offset = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_edittext_rectangle_top_offset;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_exposed_dropdown_menu_popup_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_exposed_dropdown_menu_popup_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_exposed_dropdown_menu_popup_vertical_offset = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_exposed_dropdown_menu_popup_vertical_offset;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_exposed_dropdown_menu_popup_vertical_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_exposed_dropdown_menu_popup_vertical_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_extended_fab_bottom_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_bottom_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_extended_fab_corner_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_corner_radius;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_extended_fab_disabled_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_disabled_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_extended_fab_disabled_translation_z = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_disabled_translation_z;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_extended_fab_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_extended_fab_end_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_end_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_extended_fab_end_padding_icon = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_end_padding_icon;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_extended_fab_icon_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_icon_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_extended_fab_icon_text_spacing = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_icon_text_spacing;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_extended_fab_min_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_min_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_extended_fab_min_width = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_min_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_extended_fab_start_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_start_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_extended_fab_start_padding_icon = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_start_padding_icon;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_extended_fab_top_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_top_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_extended_fab_translation_z_base = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_translation_z_base;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_extended_fab_translation_z_hovered_focused = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_translation_z_hovered_focused;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_extended_fab_translation_z_pressed = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_translation_z_pressed;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_fab_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_fab_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_fab_min_touch_target = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_fab_min_touch_target;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_fab_translation_z_hovered_focused = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_fab_translation_z_hovered_focused;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_fab_translation_z_pressed = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_fab_translation_z_pressed;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_high_ripple_default_alpha = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_high_ripple_default_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_high_ripple_focused_alpha = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_high_ripple_focused_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_high_ripple_hovered_alpha = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_high_ripple_hovered_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_high_ripple_pressed_alpha = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_high_ripple_pressed_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_large_touch_target = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_large_touch_target;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_low_ripple_default_alpha = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_low_ripple_default_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_low_ripple_focused_alpha = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_low_ripple_focused_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_low_ripple_hovered_alpha = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_low_ripple_hovered_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_low_ripple_pressed_alpha = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_low_ripple_pressed_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_min_touch_target_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_min_touch_target_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_navigation_bar_item_default_icon_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_bar_item_default_icon_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_navigation_bar_item_default_margin = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_bar_item_default_margin;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_navigation_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_navigation_item_horizontal_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_item_horizontal_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_navigation_item_icon_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_item_icon_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_navigation_item_icon_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_item_icon_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_navigation_item_shape_horizontal_margin = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_item_shape_horizontal_margin;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_navigation_item_shape_vertical_margin = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_item_shape_vertical_margin;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_navigation_rail_active_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_rail_active_text_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_navigation_rail_compact_width = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_rail_compact_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_navigation_rail_default_width = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_rail_default_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_navigation_rail_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_rail_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_navigation_rail_icon_margin = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_rail_icon_margin;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_navigation_rail_icon_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_rail_icon_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_navigation_rail_margin = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_rail_margin;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_navigation_rail_text_bottom_margin = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_rail_text_bottom_margin;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_navigation_rail_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_rail_text_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_progress_circular_inset = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_inset;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_progress_circular_inset_extra_small = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_inset_extra_small;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_progress_circular_inset_medium = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_inset_medium;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_progress_circular_inset_small = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_inset_small;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_progress_circular_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_radius;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_progress_circular_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_progress_circular_size_extra_small = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_size_extra_small;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_progress_circular_size_medium = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_size_medium;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_progress_circular_size_small = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_size_small;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_progress_circular_track_thickness_extra_small = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_track_thickness_extra_small;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_progress_circular_track_thickness_medium = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_track_thickness_medium;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_progress_circular_track_thickness_small = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_track_thickness_small;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_progress_indicator_full_rounded_corner_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_indicator_full_rounded_corner_radius;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_progress_track_thickness = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_track_thickness;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_shape_corner_size_large_component = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_shape_corner_size_large_component;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_shape_corner_size_medium_component = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_shape_corner_size_medium_component;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_shape_corner_size_small_component = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_shape_corner_size_small_component;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_slider_halo_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_slider_halo_radius;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_slider_label_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_slider_label_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_slider_label_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_slider_label_radius;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_slider_label_square_side = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_slider_label_square_side;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_slider_thumb_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_slider_thumb_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_slider_thumb_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_slider_thumb_radius;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_slider_track_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_slider_track_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_slider_track_side_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_slider_track_side_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_slider_track_top = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_slider_track_top;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_slider_widget_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_slider_widget_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_snackbar_action_text_color_alpha = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_snackbar_action_text_color_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_snackbar_background_corner_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_snackbar_background_corner_radius;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_snackbar_background_overlay_color_alpha = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_snackbar_background_overlay_color_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_snackbar_margin = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_snackbar_margin;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_snackbar_message_margin_horizontal = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_snackbar_message_margin_horizontal;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_snackbar_padding_horizontal = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_snackbar_padding_horizontal;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_switch_thumb_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_switch_thumb_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_textinput_box_corner_radius_medium = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_textinput_box_corner_radius_medium;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_textinput_box_corner_radius_small = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_textinput_box_corner_radius_small;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_textinput_box_label_cutout_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_textinput_box_label_cutout_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_textinput_box_stroke_width_default = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_textinput_box_stroke_width_default;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_textinput_box_stroke_width_focused = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_textinput_box_stroke_width_focused;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_textinput_counter_margin_start = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_textinput_counter_margin_start;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_textinput_end_icon_margin_start = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_textinput_end_icon_margin_start;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_textinput_outline_box_expanded_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_textinput_outline_box_expanded_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_textinput_start_icon_margin_end = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_textinput_start_icon_margin_end;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_toolbar_default_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_toolbar_default_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_tooltip_arrowSize = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_tooltip_arrowSize;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_tooltip_cornerSize = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_tooltip_cornerSize;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_tooltip_minHeight = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_tooltip_minHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_tooltip_minWidth = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_tooltip_minWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_tooltip_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_tooltip_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.mtrl_transition_shared_axis_slide_distance = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_transition_shared_axis_slide_distance;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.notification_action_icon_size = global::pjsua2xamarin.Droid.Resource.Dimension.notification_action_icon_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.notification_action_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.notification_action_text_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.notification_big_circle_margin = global::pjsua2xamarin.Droid.Resource.Dimension.notification_big_circle_margin;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.notification_content_margin_start = global::pjsua2xamarin.Droid.Resource.Dimension.notification_content_margin_start;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.notification_large_icon_height = global::pjsua2xamarin.Droid.Resource.Dimension.notification_large_icon_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.notification_large_icon_width = global::pjsua2xamarin.Droid.Resource.Dimension.notification_large_icon_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.notification_main_column_padding_top = global::pjsua2xamarin.Droid.Resource.Dimension.notification_main_column_padding_top;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.notification_media_narrow_margin = global::pjsua2xamarin.Droid.Resource.Dimension.notification_media_narrow_margin;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.notification_right_icon_size = global::pjsua2xamarin.Droid.Resource.Dimension.notification_right_icon_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.notification_right_side_padding_top = global::pjsua2xamarin.Droid.Resource.Dimension.notification_right_side_padding_top;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.notification_small_icon_background_padding = global::pjsua2xamarin.Droid.Resource.Dimension.notification_small_icon_background_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.notification_small_icon_size_as_large = global::pjsua2xamarin.Droid.Resource.Dimension.notification_small_icon_size_as_large;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.notification_subtext_size = global::pjsua2xamarin.Droid.Resource.Dimension.notification_subtext_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.notification_top_pad = global::pjsua2xamarin.Droid.Resource.Dimension.notification_top_pad;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.notification_top_pad_large_text = global::pjsua2xamarin.Droid.Resource.Dimension.notification_top_pad_large_text;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.test_mtrl_calendar_day_cornerSize = global::pjsua2xamarin.Droid.Resource.Dimension.test_mtrl_calendar_day_cornerSize;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.test_navigation_bar_active_item_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.test_navigation_bar_active_item_max_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.test_navigation_bar_active_item_min_width = global::pjsua2xamarin.Droid.Resource.Dimension.test_navigation_bar_active_item_min_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.test_navigation_bar_active_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.test_navigation_bar_active_text_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.test_navigation_bar_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.test_navigation_bar_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.test_navigation_bar_height = global::pjsua2xamarin.Droid.Resource.Dimension.test_navigation_bar_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.test_navigation_bar_icon_size = global::pjsua2xamarin.Droid.Resource.Dimension.test_navigation_bar_icon_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.test_navigation_bar_item_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.test_navigation_bar_item_max_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.test_navigation_bar_item_min_width = global::pjsua2xamarin.Droid.Resource.Dimension.test_navigation_bar_item_min_width;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.test_navigation_bar_label_padding = global::pjsua2xamarin.Droid.Resource.Dimension.test_navigation_bar_label_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.test_navigation_bar_shadow_height = global::pjsua2xamarin.Droid.Resource.Dimension.test_navigation_bar_shadow_height;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.test_navigation_bar_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.test_navigation_bar_text_size;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.tooltip_corner_radius = global::pjsua2xamarin.Droid.Resource.Dimension.tooltip_corner_radius;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.tooltip_horizontal_padding = global::pjsua2xamarin.Droid.Resource.Dimension.tooltip_horizontal_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.tooltip_margin = global::pjsua2xamarin.Droid.Resource.Dimension.tooltip_margin;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.tooltip_precise_anchor_extra_offset = global::pjsua2xamarin.Droid.Resource.Dimension.tooltip_precise_anchor_extra_offset;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.tooltip_precise_anchor_threshold = global::pjsua2xamarin.Droid.Resource.Dimension.tooltip_precise_anchor_threshold;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.tooltip_vertical_padding = global::pjsua2xamarin.Droid.Resource.Dimension.tooltip_vertical_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.tooltip_y_offset_non_touch = global::pjsua2xamarin.Droid.Resource.Dimension.tooltip_y_offset_non_touch;
			global::Xamarin.Forms.Platform.Android.Resource.Dimension.tooltip_y_offset_touch = global::pjsua2xamarin.Droid.Resource.Dimension.tooltip_y_offset_touch;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_ab_share_pack_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ab_share_pack_mtrl_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_action_bar_item_background_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_action_bar_item_background_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_btn_borderless_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_borderless_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_btn_check_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_check_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_btn_check_material_anim = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_check_material_anim;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_btn_check_to_on_mtrl_000 = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_check_to_on_mtrl_000;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_btn_check_to_on_mtrl_015 = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_check_to_on_mtrl_015;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_btn_colored_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_colored_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_btn_default_mtrl_shape = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_default_mtrl_shape;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_btn_radio_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_radio_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_btn_radio_material_anim = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_radio_material_anim;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_btn_radio_to_on_mtrl_000 = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_radio_to_on_mtrl_000;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_btn_radio_to_on_mtrl_015 = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_radio_to_on_mtrl_015;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_btn_switch_to_on_mtrl_00001 = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_switch_to_on_mtrl_00001;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_btn_switch_to_on_mtrl_00012 = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_switch_to_on_mtrl_00012;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_cab_background_internal_bg = global::pjsua2xamarin.Droid.Resource.Drawable.abc_cab_background_internal_bg;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_cab_background_top_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_cab_background_top_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_cab_background_top_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_cab_background_top_mtrl_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_control_background_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_control_background_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_dialog_material_background = global::pjsua2xamarin.Droid.Resource.Drawable.abc_dialog_material_background;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_edit_text_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_edit_text_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_ic_ab_back_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_ab_back_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_ic_arrow_drop_right_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_arrow_drop_right_black_24dp;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_ic_clear_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_clear_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_ic_commit_search_api_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_commit_search_api_mtrl_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_ic_go_search_api_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_go_search_api_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_ic_menu_copy_mtrl_am_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_menu_copy_mtrl_am_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_ic_menu_cut_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_menu_cut_mtrl_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_ic_menu_overflow_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_menu_overflow_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_ic_menu_paste_mtrl_am_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_menu_paste_mtrl_am_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_ic_menu_selectall_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_menu_selectall_mtrl_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_ic_menu_share_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_menu_share_mtrl_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_ic_search_api_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_search_api_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_ic_voice_search_api_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_voice_search_api_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_item_background_holo_dark = global::pjsua2xamarin.Droid.Resource.Drawable.abc_item_background_holo_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_item_background_holo_light = global::pjsua2xamarin.Droid.Resource.Drawable.abc_item_background_holo_light;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_list_divider_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_divider_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_list_divider_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_divider_mtrl_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_list_focused_holo = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_focused_holo;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_list_longpressed_holo = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_longpressed_holo;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_list_pressed_holo_dark = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_pressed_holo_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_list_pressed_holo_light = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_pressed_holo_light;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_list_selector_background_transition_holo_dark = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_selector_background_transition_holo_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_list_selector_background_transition_holo_light = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_selector_background_transition_holo_light;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_list_selector_disabled_holo_dark = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_selector_disabled_holo_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_list_selector_disabled_holo_light = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_selector_disabled_holo_light;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_list_selector_holo_dark = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_selector_holo_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_list_selector_holo_light = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_selector_holo_light;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_menu_hardkey_panel_mtrl_mult = global::pjsua2xamarin.Droid.Resource.Drawable.abc_menu_hardkey_panel_mtrl_mult;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_popup_background_mtrl_mult = global::pjsua2xamarin.Droid.Resource.Drawable.abc_popup_background_mtrl_mult;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_ratingbar_indicator_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ratingbar_indicator_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_ratingbar_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ratingbar_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_ratingbar_small_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ratingbar_small_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_scrubber_control_off_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_scrubber_control_off_mtrl_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_scrubber_control_to_pressed_mtrl_000 = global::pjsua2xamarin.Droid.Resource.Drawable.abc_scrubber_control_to_pressed_mtrl_000;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_scrubber_control_to_pressed_mtrl_005 = global::pjsua2xamarin.Droid.Resource.Drawable.abc_scrubber_control_to_pressed_mtrl_005;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_scrubber_primary_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_scrubber_primary_mtrl_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_scrubber_track_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_scrubber_track_mtrl_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_seekbar_thumb_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_seekbar_thumb_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_seekbar_tick_mark_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_seekbar_tick_mark_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_seekbar_track_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_seekbar_track_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_spinner_mtrl_am_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_spinner_mtrl_am_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_spinner_textfield_background_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_spinner_textfield_background_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_star_black_48dp = global::pjsua2xamarin.Droid.Resource.Drawable.abc_star_black_48dp;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_star_half_black_48dp = global::pjsua2xamarin.Droid.Resource.Drawable.abc_star_half_black_48dp;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_switch_thumb_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_switch_thumb_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_switch_track_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_switch_track_mtrl_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_tab_indicator_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_tab_indicator_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_tab_indicator_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_tab_indicator_mtrl_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_textfield_activated_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_textfield_activated_mtrl_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_textfield_default_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_textfield_default_mtrl_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_textfield_search_activated_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_textfield_search_activated_mtrl_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_textfield_search_default_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_textfield_search_default_mtrl_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_textfield_search_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_textfield_search_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_text_cursor_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_text_cursor_material;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_text_select_handle_left_mtrl = global::pjsua2xamarin.Droid.Resource.Drawable.abc_text_select_handle_left_mtrl;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_text_select_handle_middle_mtrl = global::pjsua2xamarin.Droid.Resource.Drawable.abc_text_select_handle_middle_mtrl;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_text_select_handle_right_mtrl = global::pjsua2xamarin.Droid.Resource.Drawable.abc_text_select_handle_right_mtrl;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.abc_vector_test = global::pjsua2xamarin.Droid.Resource.Drawable.abc_vector_test;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.avd_hide_password = global::pjsua2xamarin.Droid.Resource.Drawable.avd_hide_password;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.avd_show_password = global::pjsua2xamarin.Droid.Resource.Drawable.avd_show_password;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.btn_checkbox_checked_mtrl = global::pjsua2xamarin.Droid.Resource.Drawable.btn_checkbox_checked_mtrl;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.btn_checkbox_checked_to_unchecked_mtrl_animation = global::pjsua2xamarin.Droid.Resource.Drawable.btn_checkbox_checked_to_unchecked_mtrl_animation;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.btn_checkbox_unchecked_mtrl = global::pjsua2xamarin.Droid.Resource.Drawable.btn_checkbox_unchecked_mtrl;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.btn_checkbox_unchecked_to_checked_mtrl_animation = global::pjsua2xamarin.Droid.Resource.Drawable.btn_checkbox_unchecked_to_checked_mtrl_animation;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.btn_radio_off_mtrl = global::pjsua2xamarin.Droid.Resource.Drawable.btn_radio_off_mtrl;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.btn_radio_off_to_on_mtrl_animation = global::pjsua2xamarin.Droid.Resource.Drawable.btn_radio_off_to_on_mtrl_animation;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.btn_radio_on_mtrl = global::pjsua2xamarin.Droid.Resource.Drawable.btn_radio_on_mtrl;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.btn_radio_on_to_off_mtrl_animation = global::pjsua2xamarin.Droid.Resource.Drawable.btn_radio_on_to_off_mtrl_animation;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.design_fab_background = global::pjsua2xamarin.Droid.Resource.Drawable.design_fab_background;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.design_ic_visibility = global::pjsua2xamarin.Droid.Resource.Drawable.design_ic_visibility;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.design_ic_visibility_off = global::pjsua2xamarin.Droid.Resource.Drawable.design_ic_visibility_off;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.design_password_eye = global::pjsua2xamarin.Droid.Resource.Drawable.design_password_eye;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.design_snackbar_background = global::pjsua2xamarin.Droid.Resource.Drawable.design_snackbar_background;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.ic_clock_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.ic_clock_black_24dp;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.ic_keyboard_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.ic_keyboard_black_24dp;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.ic_mtrl_checked_circle = global::pjsua2xamarin.Droid.Resource.Drawable.ic_mtrl_checked_circle;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.ic_mtrl_chip_checked_black = global::pjsua2xamarin.Droid.Resource.Drawable.ic_mtrl_chip_checked_black;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.ic_mtrl_chip_checked_circle = global::pjsua2xamarin.Droid.Resource.Drawable.ic_mtrl_chip_checked_circle;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.ic_mtrl_chip_close_circle = global::pjsua2xamarin.Droid.Resource.Drawable.ic_mtrl_chip_close_circle;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.material_cursor_drawable = global::pjsua2xamarin.Droid.Resource.Drawable.material_cursor_drawable;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.material_ic_calendar_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.material_ic_calendar_black_24dp;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.material_ic_clear_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.material_ic_clear_black_24dp;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.material_ic_edit_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.material_ic_edit_black_24dp;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.material_ic_keyboard_arrow_left_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.material_ic_keyboard_arrow_left_black_24dp;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.material_ic_keyboard_arrow_next_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.material_ic_keyboard_arrow_next_black_24dp;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.material_ic_keyboard_arrow_previous_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.material_ic_keyboard_arrow_previous_black_24dp;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.material_ic_keyboard_arrow_right_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.material_ic_keyboard_arrow_right_black_24dp;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.material_ic_menu_arrow_down_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.material_ic_menu_arrow_down_black_24dp;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.material_ic_menu_arrow_up_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.material_ic_menu_arrow_up_black_24dp;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.mtrl_dialog_background = global::pjsua2xamarin.Droid.Resource.Drawable.mtrl_dialog_background;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.mtrl_dropdown_arrow = global::pjsua2xamarin.Droid.Resource.Drawable.mtrl_dropdown_arrow;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.mtrl_ic_arrow_drop_down = global::pjsua2xamarin.Droid.Resource.Drawable.mtrl_ic_arrow_drop_down;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.mtrl_ic_arrow_drop_up = global::pjsua2xamarin.Droid.Resource.Drawable.mtrl_ic_arrow_drop_up;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.mtrl_ic_cancel = global::pjsua2xamarin.Droid.Resource.Drawable.mtrl_ic_cancel;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.mtrl_ic_error = global::pjsua2xamarin.Droid.Resource.Drawable.mtrl_ic_error;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.mtrl_navigation_bar_item_background = global::pjsua2xamarin.Droid.Resource.Drawable.mtrl_navigation_bar_item_background;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.mtrl_popupmenu_background = global::pjsua2xamarin.Droid.Resource.Drawable.mtrl_popupmenu_background;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.mtrl_popupmenu_background_dark = global::pjsua2xamarin.Droid.Resource.Drawable.mtrl_popupmenu_background_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.mtrl_tabs_default_indicator = global::pjsua2xamarin.Droid.Resource.Drawable.mtrl_tabs_default_indicator;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.navigation_empty_icon = global::pjsua2xamarin.Droid.Resource.Drawable.navigation_empty_icon;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.notification_action_background = global::pjsua2xamarin.Droid.Resource.Drawable.notification_action_background;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.notification_bg = global::pjsua2xamarin.Droid.Resource.Drawable.notification_bg;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.notification_bg_low = global::pjsua2xamarin.Droid.Resource.Drawable.notification_bg_low;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.notification_bg_low_normal = global::pjsua2xamarin.Droid.Resource.Drawable.notification_bg_low_normal;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.notification_bg_low_pressed = global::pjsua2xamarin.Droid.Resource.Drawable.notification_bg_low_pressed;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.notification_bg_normal = global::pjsua2xamarin.Droid.Resource.Drawable.notification_bg_normal;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.notification_bg_normal_pressed = global::pjsua2xamarin.Droid.Resource.Drawable.notification_bg_normal_pressed;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.notification_icon_background = global::pjsua2xamarin.Droid.Resource.Drawable.notification_icon_background;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.notification_template_icon_bg = global::pjsua2xamarin.Droid.Resource.Drawable.notification_template_icon_bg;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.notification_template_icon_low_bg = global::pjsua2xamarin.Droid.Resource.Drawable.notification_template_icon_low_bg;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.notification_tile_bg = global::pjsua2xamarin.Droid.Resource.Drawable.notification_tile_bg;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.notify_panel_notification_icon_bg = global::pjsua2xamarin.Droid.Resource.Drawable.notify_panel_notification_icon_bg;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.test_custom_background = global::pjsua2xamarin.Droid.Resource.Drawable.test_custom_background;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.tooltip_frame_dark = global::pjsua2xamarin.Droid.Resource.Drawable.tooltip_frame_dark;
			global::Xamarin.Forms.Platform.Android.Resource.Drawable.tooltip_frame_light = global::pjsua2xamarin.Droid.Resource.Drawable.tooltip_frame_light;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accelerate = global::pjsua2xamarin.Droid.Resource.Id.accelerate;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_action_clickable_span = global::pjsua2xamarin.Droid.Resource.Id.accessibility_action_clickable_span;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_0 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_0;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_1 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_1;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_10 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_10;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_11 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_11;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_12 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_12;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_13 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_13;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_14 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_14;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_15 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_15;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_16 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_16;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_17 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_17;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_18 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_18;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_19 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_19;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_2 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_2;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_20 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_20;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_21 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_21;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_22 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_22;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_23 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_23;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_24 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_24;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_25 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_25;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_26 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_26;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_27 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_27;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_28 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_28;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_29 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_29;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_3 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_3;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_30 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_30;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_31 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_31;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_4 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_4;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_5 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_5;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_6 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_6;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_7 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_7;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_8 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_8;
			global::Xamarin.Forms.Platform.Android.Resource.Id.accessibility_custom_action_9 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_9;
			global::Xamarin.Forms.Platform.Android.Resource.Id.action0 = global::pjsua2xamarin.Droid.Resource.Id.action0;
			global::Xamarin.Forms.Platform.Android.Resource.Id.actionDown = global::pjsua2xamarin.Droid.Resource.Id.actionDown;
			global::Xamarin.Forms.Platform.Android.Resource.Id.actionDownUp = global::pjsua2xamarin.Droid.Resource.Id.actionDownUp;
			global::Xamarin.Forms.Platform.Android.Resource.Id.actions = global::pjsua2xamarin.Droid.Resource.Id.actions;
			global::Xamarin.Forms.Platform.Android.Resource.Id.actionUp = global::pjsua2xamarin.Droid.Resource.Id.actionUp;
			global::Xamarin.Forms.Platform.Android.Resource.Id.action_bar = global::pjsua2xamarin.Droid.Resource.Id.action_bar;
			global::Xamarin.Forms.Platform.Android.Resource.Id.action_bar_activity_content = global::pjsua2xamarin.Droid.Resource.Id.action_bar_activity_content;
			global::Xamarin.Forms.Platform.Android.Resource.Id.action_bar_container = global::pjsua2xamarin.Droid.Resource.Id.action_bar_container;
			global::Xamarin.Forms.Platform.Android.Resource.Id.action_bar_root = global::pjsua2xamarin.Droid.Resource.Id.action_bar_root;
			global::Xamarin.Forms.Platform.Android.Resource.Id.action_bar_spinner = global::pjsua2xamarin.Droid.Resource.Id.action_bar_spinner;
			global::Xamarin.Forms.Platform.Android.Resource.Id.action_bar_subtitle = global::pjsua2xamarin.Droid.Resource.Id.action_bar_subtitle;
			global::Xamarin.Forms.Platform.Android.Resource.Id.action_bar_title = global::pjsua2xamarin.Droid.Resource.Id.action_bar_title;
			global::Xamarin.Forms.Platform.Android.Resource.Id.action_container = global::pjsua2xamarin.Droid.Resource.Id.action_container;
			global::Xamarin.Forms.Platform.Android.Resource.Id.action_context_bar = global::pjsua2xamarin.Droid.Resource.Id.action_context_bar;
			global::Xamarin.Forms.Platform.Android.Resource.Id.action_divider = global::pjsua2xamarin.Droid.Resource.Id.action_divider;
			global::Xamarin.Forms.Platform.Android.Resource.Id.action_image = global::pjsua2xamarin.Droid.Resource.Id.action_image;
			global::Xamarin.Forms.Platform.Android.Resource.Id.action_menu_divider = global::pjsua2xamarin.Droid.Resource.Id.action_menu_divider;
			global::Xamarin.Forms.Platform.Android.Resource.Id.action_menu_presenter = global::pjsua2xamarin.Droid.Resource.Id.action_menu_presenter;
			global::Xamarin.Forms.Platform.Android.Resource.Id.action_mode_bar = global::pjsua2xamarin.Droid.Resource.Id.action_mode_bar;
			global::Xamarin.Forms.Platform.Android.Resource.Id.action_mode_bar_stub = global::pjsua2xamarin.Droid.Resource.Id.action_mode_bar_stub;
			global::Xamarin.Forms.Platform.Android.Resource.Id.action_mode_close_button = global::pjsua2xamarin.Droid.Resource.Id.action_mode_close_button;
			global::Xamarin.Forms.Platform.Android.Resource.Id.action_text = global::pjsua2xamarin.Droid.Resource.Id.action_text;
			global::Xamarin.Forms.Platform.Android.Resource.Id.activity_chooser_view_content = global::pjsua2xamarin.Droid.Resource.Id.activity_chooser_view_content;
			global::Xamarin.Forms.Platform.Android.Resource.Id.add = global::pjsua2xamarin.Droid.Resource.Id.add;
			global::Xamarin.Forms.Platform.Android.Resource.Id.alertTitle = global::pjsua2xamarin.Droid.Resource.Id.alertTitle;
			global::Xamarin.Forms.Platform.Android.Resource.Id.aligned = global::pjsua2xamarin.Droid.Resource.Id.aligned;
			global::Xamarin.Forms.Platform.Android.Resource.Id.all = global::pjsua2xamarin.Droid.Resource.Id.all;
			global::Xamarin.Forms.Platform.Android.Resource.Id.allStates = global::pjsua2xamarin.Droid.Resource.Id.allStates;
			global::Xamarin.Forms.Platform.Android.Resource.Id.ALT = global::pjsua2xamarin.Droid.Resource.Id.ALT;
			global::Xamarin.Forms.Platform.Android.Resource.Id.always = global::pjsua2xamarin.Droid.Resource.Id.always;
			global::Xamarin.Forms.Platform.Android.Resource.Id.animateToEnd = global::pjsua2xamarin.Droid.Resource.Id.animateToEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Id.animateToStart = global::pjsua2xamarin.Droid.Resource.Id.animateToStart;
			global::Xamarin.Forms.Platform.Android.Resource.Id.anticipate = global::pjsua2xamarin.Droid.Resource.Id.anticipate;
			global::Xamarin.Forms.Platform.Android.Resource.Id.antiClockwise = global::pjsua2xamarin.Droid.Resource.Id.antiClockwise;
			global::Xamarin.Forms.Platform.Android.Resource.Id.arc = global::pjsua2xamarin.Droid.Resource.Id.arc;
			global::Xamarin.Forms.Platform.Android.Resource.Id.asConfigured = global::pjsua2xamarin.Droid.Resource.Id.asConfigured;
			global::Xamarin.Forms.Platform.Android.Resource.Id.async = global::pjsua2xamarin.Droid.Resource.Id.async;
			global::Xamarin.Forms.Platform.Android.Resource.Id.auto = global::pjsua2xamarin.Droid.Resource.Id.auto;
			global::Xamarin.Forms.Platform.Android.Resource.Id.autoComplete = global::pjsua2xamarin.Droid.Resource.Id.autoComplete;
			global::Xamarin.Forms.Platform.Android.Resource.Id.autoCompleteToEnd = global::pjsua2xamarin.Droid.Resource.Id.autoCompleteToEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Id.autoCompleteToStart = global::pjsua2xamarin.Droid.Resource.Id.autoCompleteToStart;
			global::Xamarin.Forms.Platform.Android.Resource.Id.barrier = global::pjsua2xamarin.Droid.Resource.Id.barrier;
			global::Xamarin.Forms.Platform.Android.Resource.Id.baseline = global::pjsua2xamarin.Droid.Resource.Id.baseline;
			global::Xamarin.Forms.Platform.Android.Resource.Id.beginning = global::pjsua2xamarin.Droid.Resource.Id.beginning;
			global::Xamarin.Forms.Platform.Android.Resource.Id.beginOnFirstDraw = global::pjsua2xamarin.Droid.Resource.Id.beginOnFirstDraw;
			global::Xamarin.Forms.Platform.Android.Resource.Id.bestChoice = global::pjsua2xamarin.Droid.Resource.Id.bestChoice;
			global::Xamarin.Forms.Platform.Android.Resource.Id.blocking = global::pjsua2xamarin.Droid.Resource.Id.blocking;
			global::Xamarin.Forms.Platform.Android.Resource.Id.bottom = global::pjsua2xamarin.Droid.Resource.Id.bottom;
			global::Xamarin.Forms.Platform.Android.Resource.Id.bottomtab_navarea = global::pjsua2xamarin.Droid.Resource.Id.bottomtab_navarea;
			global::Xamarin.Forms.Platform.Android.Resource.Id.bottomtab_tabbar = global::pjsua2xamarin.Droid.Resource.Id.bottomtab_tabbar;
			global::Xamarin.Forms.Platform.Android.Resource.Id.BOTTOM_END = global::pjsua2xamarin.Droid.Resource.Id.BOTTOM_END;
			global::Xamarin.Forms.Platform.Android.Resource.Id.BOTTOM_START = global::pjsua2xamarin.Droid.Resource.Id.BOTTOM_START;
			global::Xamarin.Forms.Platform.Android.Resource.Id.bounce = global::pjsua2xamarin.Droid.Resource.Id.bounce;
			global::Xamarin.Forms.Platform.Android.Resource.Id.bounceBoth = global::pjsua2xamarin.Droid.Resource.Id.bounceBoth;
			global::Xamarin.Forms.Platform.Android.Resource.Id.bounceEnd = global::pjsua2xamarin.Droid.Resource.Id.bounceEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Id.bounceStart = global::pjsua2xamarin.Droid.Resource.Id.bounceStart;
			global::Xamarin.Forms.Platform.Android.Resource.Id.buttonPanel = global::pjsua2xamarin.Droid.Resource.Id.buttonPanel;
			global::Xamarin.Forms.Platform.Android.Resource.Id.cache_measures = global::pjsua2xamarin.Droid.Resource.Id.cache_measures;
			global::Xamarin.Forms.Platform.Android.Resource.Id.cancel_action = global::pjsua2xamarin.Droid.Resource.Id.cancel_action;
			global::Xamarin.Forms.Platform.Android.Resource.Id.cancel_button = global::pjsua2xamarin.Droid.Resource.Id.cancel_button;
			global::Xamarin.Forms.Platform.Android.Resource.Id.carryVelocity = global::pjsua2xamarin.Droid.Resource.Id.carryVelocity;
			global::Xamarin.Forms.Platform.Android.Resource.Id.center = global::pjsua2xamarin.Droid.Resource.Id.center;
			global::Xamarin.Forms.Platform.Android.Resource.Id.center_horizontal = global::pjsua2xamarin.Droid.Resource.Id.center_horizontal;
			global::Xamarin.Forms.Platform.Android.Resource.Id.center_vertical = global::pjsua2xamarin.Droid.Resource.Id.center_vertical;
			global::Xamarin.Forms.Platform.Android.Resource.Id.chain = global::pjsua2xamarin.Droid.Resource.Id.chain;
			global::Xamarin.Forms.Platform.Android.Resource.Id.chains = global::pjsua2xamarin.Droid.Resource.Id.chains;
			global::Xamarin.Forms.Platform.Android.Resource.Id.checkbox = global::pjsua2xamarin.Droid.Resource.Id.checkbox;
			global::Xamarin.Forms.Platform.Android.Resource.Id.@checked = global::pjsua2xamarin.Droid.Resource.Id.@checked;
			global::Xamarin.Forms.Platform.Android.Resource.Id.chip = global::pjsua2xamarin.Droid.Resource.Id.chip;
			global::Xamarin.Forms.Platform.Android.Resource.Id.chip1 = global::pjsua2xamarin.Droid.Resource.Id.chip1;
			global::Xamarin.Forms.Platform.Android.Resource.Id.chip2 = global::pjsua2xamarin.Droid.Resource.Id.chip2;
			global::Xamarin.Forms.Platform.Android.Resource.Id.chip3 = global::pjsua2xamarin.Droid.Resource.Id.chip3;
			global::Xamarin.Forms.Platform.Android.Resource.Id.chip_group = global::pjsua2xamarin.Droid.Resource.Id.chip_group;
			global::Xamarin.Forms.Platform.Android.Resource.Id.chronometer = global::pjsua2xamarin.Droid.Resource.Id.chronometer;
			global::Xamarin.Forms.Platform.Android.Resource.Id.circle_center = global::pjsua2xamarin.Droid.Resource.Id.circle_center;
			global::Xamarin.Forms.Platform.Android.Resource.Id.clear_text = global::pjsua2xamarin.Droid.Resource.Id.clear_text;
			global::Xamarin.Forms.Platform.Android.Resource.Id.clip_horizontal = global::pjsua2xamarin.Droid.Resource.Id.clip_horizontal;
			global::Xamarin.Forms.Platform.Android.Resource.Id.clip_vertical = global::pjsua2xamarin.Droid.Resource.Id.clip_vertical;
			global::Xamarin.Forms.Platform.Android.Resource.Id.clockwise = global::pjsua2xamarin.Droid.Resource.Id.clockwise;
			global::Xamarin.Forms.Platform.Android.Resource.Id.closest = global::pjsua2xamarin.Droid.Resource.Id.closest;
			global::Xamarin.Forms.Platform.Android.Resource.Id.collapseActionView = global::pjsua2xamarin.Droid.Resource.Id.collapseActionView;
			global::Xamarin.Forms.Platform.Android.Resource.Id.confirm_button = global::pjsua2xamarin.Droid.Resource.Id.confirm_button;
			global::Xamarin.Forms.Platform.Android.Resource.Id.constraint = global::pjsua2xamarin.Droid.Resource.Id.constraint;
			global::Xamarin.Forms.Platform.Android.Resource.Id.container = global::pjsua2xamarin.Droid.Resource.Id.container;
			global::Xamarin.Forms.Platform.Android.Resource.Id.content = global::pjsua2xamarin.Droid.Resource.Id.content;
			global::Xamarin.Forms.Platform.Android.Resource.Id.contentPanel = global::pjsua2xamarin.Droid.Resource.Id.contentPanel;
			global::Xamarin.Forms.Platform.Android.Resource.Id.contiguous = global::pjsua2xamarin.Droid.Resource.Id.contiguous;
			global::Xamarin.Forms.Platform.Android.Resource.Id.continuousVelocity = global::pjsua2xamarin.Droid.Resource.Id.continuousVelocity;
			global::Xamarin.Forms.Platform.Android.Resource.Id.coordinator = global::pjsua2xamarin.Droid.Resource.Id.coordinator;
			global::Xamarin.Forms.Platform.Android.Resource.Id.cos = global::pjsua2xamarin.Droid.Resource.Id.cos;
			global::Xamarin.Forms.Platform.Android.Resource.Id.counterclockwise = global::pjsua2xamarin.Droid.Resource.Id.counterclockwise;
			global::Xamarin.Forms.Platform.Android.Resource.Id.CTRL = global::pjsua2xamarin.Droid.Resource.Id.CTRL;
			global::Xamarin.Forms.Platform.Android.Resource.Id.currentState = global::pjsua2xamarin.Droid.Resource.Id.currentState;
			global::Xamarin.Forms.Platform.Android.Resource.Id.custom = global::pjsua2xamarin.Droid.Resource.Id.custom;
			global::Xamarin.Forms.Platform.Android.Resource.Id.customPanel = global::pjsua2xamarin.Droid.Resource.Id.customPanel;
			global::Xamarin.Forms.Platform.Android.Resource.Id.cut = global::pjsua2xamarin.Droid.Resource.Id.cut;
			global::Xamarin.Forms.Platform.Android.Resource.Id.date_picker_actions = global::pjsua2xamarin.Droid.Resource.Id.date_picker_actions;
			global::Xamarin.Forms.Platform.Android.Resource.Id.decelerate = global::pjsua2xamarin.Droid.Resource.Id.decelerate;
			global::Xamarin.Forms.Platform.Android.Resource.Id.decelerateAndComplete = global::pjsua2xamarin.Droid.Resource.Id.decelerateAndComplete;
			global::Xamarin.Forms.Platform.Android.Resource.Id.decor_content_parent = global::pjsua2xamarin.Droid.Resource.Id.decor_content_parent;
			global::Xamarin.Forms.Platform.Android.Resource.Id.default_activity_button = global::pjsua2xamarin.Droid.Resource.Id.default_activity_button;
			global::Xamarin.Forms.Platform.Android.Resource.Id.deltaRelative = global::pjsua2xamarin.Droid.Resource.Id.deltaRelative;
			global::Xamarin.Forms.Platform.Android.Resource.Id.dependency_ordering = global::pjsua2xamarin.Droid.Resource.Id.dependency_ordering;
			global::Xamarin.Forms.Platform.Android.Resource.Id.design_bottom_sheet = global::pjsua2xamarin.Droid.Resource.Id.design_bottom_sheet;
			global::Xamarin.Forms.Platform.Android.Resource.Id.design_menu_item_action_area = global::pjsua2xamarin.Droid.Resource.Id.design_menu_item_action_area;
			global::Xamarin.Forms.Platform.Android.Resource.Id.design_menu_item_action_area_stub = global::pjsua2xamarin.Droid.Resource.Id.design_menu_item_action_area_stub;
			global::Xamarin.Forms.Platform.Android.Resource.Id.design_menu_item_text = global::pjsua2xamarin.Droid.Resource.Id.design_menu_item_text;
			global::Xamarin.Forms.Platform.Android.Resource.Id.design_navigation_view = global::pjsua2xamarin.Droid.Resource.Id.design_navigation_view;
			global::Xamarin.Forms.Platform.Android.Resource.Id.dialog_button = global::pjsua2xamarin.Droid.Resource.Id.dialog_button;
			global::Xamarin.Forms.Platform.Android.Resource.Id.dimensions = global::pjsua2xamarin.Droid.Resource.Id.dimensions;
			global::Xamarin.Forms.Platform.Android.Resource.Id.direct = global::pjsua2xamarin.Droid.Resource.Id.direct;
			global::Xamarin.Forms.Platform.Android.Resource.Id.disableHome = global::pjsua2xamarin.Droid.Resource.Id.disableHome;
			global::Xamarin.Forms.Platform.Android.Resource.Id.disableIntraAutoTransition = global::pjsua2xamarin.Droid.Resource.Id.disableIntraAutoTransition;
			global::Xamarin.Forms.Platform.Android.Resource.Id.disablePostScroll = global::pjsua2xamarin.Droid.Resource.Id.disablePostScroll;
			global::Xamarin.Forms.Platform.Android.Resource.Id.disableScroll = global::pjsua2xamarin.Droid.Resource.Id.disableScroll;
			global::Xamarin.Forms.Platform.Android.Resource.Id.disjoint = global::pjsua2xamarin.Droid.Resource.Id.disjoint;
			global::Xamarin.Forms.Platform.Android.Resource.Id.dragAnticlockwise = global::pjsua2xamarin.Droid.Resource.Id.dragAnticlockwise;
			global::Xamarin.Forms.Platform.Android.Resource.Id.dragClockwise = global::pjsua2xamarin.Droid.Resource.Id.dragClockwise;
			global::Xamarin.Forms.Platform.Android.Resource.Id.dragDown = global::pjsua2xamarin.Droid.Resource.Id.dragDown;
			global::Xamarin.Forms.Platform.Android.Resource.Id.dragEnd = global::pjsua2xamarin.Droid.Resource.Id.dragEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Id.dragLeft = global::pjsua2xamarin.Droid.Resource.Id.dragLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Id.dragRight = global::pjsua2xamarin.Droid.Resource.Id.dragRight;
			global::Xamarin.Forms.Platform.Android.Resource.Id.dragStart = global::pjsua2xamarin.Droid.Resource.Id.dragStart;
			global::Xamarin.Forms.Platform.Android.Resource.Id.dragUp = global::pjsua2xamarin.Droid.Resource.Id.dragUp;
			global::Xamarin.Forms.Platform.Android.Resource.Id.dropdown_menu = global::pjsua2xamarin.Droid.Resource.Id.dropdown_menu;
			global::Xamarin.Forms.Platform.Android.Resource.Id.easeIn = global::pjsua2xamarin.Droid.Resource.Id.easeIn;
			global::Xamarin.Forms.Platform.Android.Resource.Id.easeInOut = global::pjsua2xamarin.Droid.Resource.Id.easeInOut;
			global::Xamarin.Forms.Platform.Android.Resource.Id.easeOut = global::pjsua2xamarin.Droid.Resource.Id.easeOut;
			global::Xamarin.Forms.Platform.Android.Resource.Id.east = global::pjsua2xamarin.Droid.Resource.Id.east;
			global::Xamarin.Forms.Platform.Android.Resource.Id.edit_query = global::pjsua2xamarin.Droid.Resource.Id.edit_query;
			global::Xamarin.Forms.Platform.Android.Resource.Id.elastic = global::pjsua2xamarin.Droid.Resource.Id.elastic;
			global::Xamarin.Forms.Platform.Android.Resource.Id.end = global::pjsua2xamarin.Droid.Resource.Id.end;
			global::Xamarin.Forms.Platform.Android.Resource.Id.endToStart = global::pjsua2xamarin.Droid.Resource.Id.endToStart;
			global::Xamarin.Forms.Platform.Android.Resource.Id.end_padder = global::pjsua2xamarin.Droid.Resource.Id.end_padder;
			global::Xamarin.Forms.Platform.Android.Resource.Id.enterAlways = global::pjsua2xamarin.Droid.Resource.Id.enterAlways;
			global::Xamarin.Forms.Platform.Android.Resource.Id.enterAlwaysCollapsed = global::pjsua2xamarin.Droid.Resource.Id.enterAlwaysCollapsed;
			global::Xamarin.Forms.Platform.Android.Resource.Id.exitUntilCollapsed = global::pjsua2xamarin.Droid.Resource.Id.exitUntilCollapsed;
			global::Xamarin.Forms.Platform.Android.Resource.Id.expanded_menu = global::pjsua2xamarin.Droid.Resource.Id.expanded_menu;
			global::Xamarin.Forms.Platform.Android.Resource.Id.expand_activities_button = global::pjsua2xamarin.Droid.Resource.Id.expand_activities_button;
			global::Xamarin.Forms.Platform.Android.Resource.Id.fade = global::pjsua2xamarin.Droid.Resource.Id.fade;
			global::Xamarin.Forms.Platform.Android.Resource.Id.fill = global::pjsua2xamarin.Droid.Resource.Id.fill;
			global::Xamarin.Forms.Platform.Android.Resource.Id.filled = global::pjsua2xamarin.Droid.Resource.Id.filled;
			global::Xamarin.Forms.Platform.Android.Resource.Id.fill_horizontal = global::pjsua2xamarin.Droid.Resource.Id.fill_horizontal;
			global::Xamarin.Forms.Platform.Android.Resource.Id.fill_vertical = global::pjsua2xamarin.Droid.Resource.Id.fill_vertical;
			global::Xamarin.Forms.Platform.Android.Resource.Id.fitToContents = global::pjsua2xamarin.Droid.Resource.Id.fitToContents;
			global::Xamarin.Forms.Platform.Android.Resource.Id.@fixed = global::pjsua2xamarin.Droid.Resource.Id.@fixed;
			global::Xamarin.Forms.Platform.Android.Resource.Id.flip = global::pjsua2xamarin.Droid.Resource.Id.flip;
			global::Xamarin.Forms.Platform.Android.Resource.Id.floating = global::pjsua2xamarin.Droid.Resource.Id.floating;
			global::Xamarin.Forms.Platform.Android.Resource.Id.flyoutcontent_appbar = global::pjsua2xamarin.Droid.Resource.Id.flyoutcontent_appbar;
			global::Xamarin.Forms.Platform.Android.Resource.Id.forever = global::pjsua2xamarin.Droid.Resource.Id.forever;
			global::Xamarin.Forms.Platform.Android.Resource.Id.fragment_container_view_tag = global::pjsua2xamarin.Droid.Resource.Id.fragment_container_view_tag;
			global::Xamarin.Forms.Platform.Android.Resource.Id.frost = global::pjsua2xamarin.Droid.Resource.Id.frost;
			global::Xamarin.Forms.Platform.Android.Resource.Id.FUNCTION = global::pjsua2xamarin.Droid.Resource.Id.FUNCTION;
			global::Xamarin.Forms.Platform.Android.Resource.Id.ghost_view = global::pjsua2xamarin.Droid.Resource.Id.ghost_view;
			global::Xamarin.Forms.Platform.Android.Resource.Id.ghost_view_holder = global::pjsua2xamarin.Droid.Resource.Id.ghost_view_holder;
			global::Xamarin.Forms.Platform.Android.Resource.Id.gone = global::pjsua2xamarin.Droid.Resource.Id.gone;
			global::Xamarin.Forms.Platform.Android.Resource.Id.graph = global::pjsua2xamarin.Droid.Resource.Id.graph;
			global::Xamarin.Forms.Platform.Android.Resource.Id.graph_wrap = global::pjsua2xamarin.Droid.Resource.Id.graph_wrap;
			global::Xamarin.Forms.Platform.Android.Resource.Id.grouping = global::pjsua2xamarin.Droid.Resource.Id.grouping;
			global::Xamarin.Forms.Platform.Android.Resource.Id.groups = global::pjsua2xamarin.Droid.Resource.Id.groups;
			global::Xamarin.Forms.Platform.Android.Resource.Id.group_divider = global::pjsua2xamarin.Droid.Resource.Id.group_divider;
			global::Xamarin.Forms.Platform.Android.Resource.Id.guideline = global::pjsua2xamarin.Droid.Resource.Id.guideline;
			global::Xamarin.Forms.Platform.Android.Resource.Id.header_title = global::pjsua2xamarin.Droid.Resource.Id.header_title;
			global::Xamarin.Forms.Platform.Android.Resource.Id.hideable = global::pjsua2xamarin.Droid.Resource.Id.hideable;
			global::Xamarin.Forms.Platform.Android.Resource.Id.home = global::pjsua2xamarin.Droid.Resource.Id.home;
			global::Xamarin.Forms.Platform.Android.Resource.Id.homeAsUp = global::pjsua2xamarin.Droid.Resource.Id.homeAsUp;
			global::Xamarin.Forms.Platform.Android.Resource.Id.honorRequest = global::pjsua2xamarin.Droid.Resource.Id.honorRequest;
			global::Xamarin.Forms.Platform.Android.Resource.Id.horizontal_only = global::pjsua2xamarin.Droid.Resource.Id.horizontal_only;
			global::Xamarin.Forms.Platform.Android.Resource.Id.icon = global::pjsua2xamarin.Droid.Resource.Id.icon;
			global::Xamarin.Forms.Platform.Android.Resource.Id.icon_group = global::pjsua2xamarin.Droid.Resource.Id.icon_group;
			global::Xamarin.Forms.Platform.Android.Resource.Id.ifRoom = global::pjsua2xamarin.Droid.Resource.Id.ifRoom;
			global::Xamarin.Forms.Platform.Android.Resource.Id.ignore = global::pjsua2xamarin.Droid.Resource.Id.ignore;
			global::Xamarin.Forms.Platform.Android.Resource.Id.ignoreRequest = global::pjsua2xamarin.Droid.Resource.Id.ignoreRequest;
			global::Xamarin.Forms.Platform.Android.Resource.Id.image = global::pjsua2xamarin.Droid.Resource.Id.image;
			global::Xamarin.Forms.Platform.Android.Resource.Id.immediateStop = global::pjsua2xamarin.Droid.Resource.Id.immediateStop;
			global::Xamarin.Forms.Platform.Android.Resource.Id.included = global::pjsua2xamarin.Droid.Resource.Id.included;
			global::Xamarin.Forms.Platform.Android.Resource.Id.info = global::pjsua2xamarin.Droid.Resource.Id.info;
			global::Xamarin.Forms.Platform.Android.Resource.Id.invisible = global::pjsua2xamarin.Droid.Resource.Id.invisible;
			global::Xamarin.Forms.Platform.Android.Resource.Id.inward = global::pjsua2xamarin.Droid.Resource.Id.inward;
			global::Xamarin.Forms.Platform.Android.Resource.Id.italic = global::pjsua2xamarin.Droid.Resource.Id.italic;
			global::Xamarin.Forms.Platform.Android.Resource.Id.item_touch_helper_previous_elevation = global::pjsua2xamarin.Droid.Resource.Id.item_touch_helper_previous_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Id.jumpToEnd = global::pjsua2xamarin.Droid.Resource.Id.jumpToEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Id.jumpToStart = global::pjsua2xamarin.Droid.Resource.Id.jumpToStart;
			global::Xamarin.Forms.Platform.Android.Resource.Id.labeled = global::pjsua2xamarin.Droid.Resource.Id.labeled;
			global::Xamarin.Forms.Platform.Android.Resource.Id.layout = global::pjsua2xamarin.Droid.Resource.Id.layout;
			global::Xamarin.Forms.Platform.Android.Resource.Id.left = global::pjsua2xamarin.Droid.Resource.Id.left;
			global::Xamarin.Forms.Platform.Android.Resource.Id.leftToRight = global::pjsua2xamarin.Droid.Resource.Id.leftToRight;
			global::Xamarin.Forms.Platform.Android.Resource.Id.legacy = global::pjsua2xamarin.Droid.Resource.Id.legacy;
			global::Xamarin.Forms.Platform.Android.Resource.Id.line1 = global::pjsua2xamarin.Droid.Resource.Id.line1;
			global::Xamarin.Forms.Platform.Android.Resource.Id.line3 = global::pjsua2xamarin.Droid.Resource.Id.line3;
			global::Xamarin.Forms.Platform.Android.Resource.Id.linear = global::pjsua2xamarin.Droid.Resource.Id.linear;
			global::Xamarin.Forms.Platform.Android.Resource.Id.listMode = global::pjsua2xamarin.Droid.Resource.Id.listMode;
			global::Xamarin.Forms.Platform.Android.Resource.Id.list_item = global::pjsua2xamarin.Droid.Resource.Id.list_item;
			global::Xamarin.Forms.Platform.Android.Resource.Id.main_appbar = global::pjsua2xamarin.Droid.Resource.Id.main_appbar;
			global::Xamarin.Forms.Platform.Android.Resource.Id.main_tablayout = global::pjsua2xamarin.Droid.Resource.Id.main_tablayout;
			global::Xamarin.Forms.Platform.Android.Resource.Id.main_toolbar = global::pjsua2xamarin.Droid.Resource.Id.main_toolbar;
			global::Xamarin.Forms.Platform.Android.Resource.Id.main_viewpager = global::pjsua2xamarin.Droid.Resource.Id.main_viewpager;
			global::Xamarin.Forms.Platform.Android.Resource.Id.masked = global::pjsua2xamarin.Droid.Resource.Id.masked;
			global::Xamarin.Forms.Platform.Android.Resource.Id.match_constraint = global::pjsua2xamarin.Droid.Resource.Id.match_constraint;
			global::Xamarin.Forms.Platform.Android.Resource.Id.match_parent = global::pjsua2xamarin.Droid.Resource.Id.match_parent;
			global::Xamarin.Forms.Platform.Android.Resource.Id.material_clock_display = global::pjsua2xamarin.Droid.Resource.Id.material_clock_display;
			global::Xamarin.Forms.Platform.Android.Resource.Id.material_clock_face = global::pjsua2xamarin.Droid.Resource.Id.material_clock_face;
			global::Xamarin.Forms.Platform.Android.Resource.Id.material_clock_hand = global::pjsua2xamarin.Droid.Resource.Id.material_clock_hand;
			global::Xamarin.Forms.Platform.Android.Resource.Id.material_clock_period_am_button = global::pjsua2xamarin.Droid.Resource.Id.material_clock_period_am_button;
			global::Xamarin.Forms.Platform.Android.Resource.Id.material_clock_period_pm_button = global::pjsua2xamarin.Droid.Resource.Id.material_clock_period_pm_button;
			global::Xamarin.Forms.Platform.Android.Resource.Id.material_clock_period_toggle = global::pjsua2xamarin.Droid.Resource.Id.material_clock_period_toggle;
			global::Xamarin.Forms.Platform.Android.Resource.Id.material_hour_text_input = global::pjsua2xamarin.Droid.Resource.Id.material_hour_text_input;
			global::Xamarin.Forms.Platform.Android.Resource.Id.material_hour_tv = global::pjsua2xamarin.Droid.Resource.Id.material_hour_tv;
			global::Xamarin.Forms.Platform.Android.Resource.Id.material_label = global::pjsua2xamarin.Droid.Resource.Id.material_label;
			global::Xamarin.Forms.Platform.Android.Resource.Id.material_minute_text_input = global::pjsua2xamarin.Droid.Resource.Id.material_minute_text_input;
			global::Xamarin.Forms.Platform.Android.Resource.Id.material_minute_tv = global::pjsua2xamarin.Droid.Resource.Id.material_minute_tv;
			global::Xamarin.Forms.Platform.Android.Resource.Id.material_textinput_timepicker = global::pjsua2xamarin.Droid.Resource.Id.material_textinput_timepicker;
			global::Xamarin.Forms.Platform.Android.Resource.Id.material_timepicker_cancel_button = global::pjsua2xamarin.Droid.Resource.Id.material_timepicker_cancel_button;
			global::Xamarin.Forms.Platform.Android.Resource.Id.material_timepicker_container = global::pjsua2xamarin.Droid.Resource.Id.material_timepicker_container;
			global::Xamarin.Forms.Platform.Android.Resource.Id.material_timepicker_edit_text = global::pjsua2xamarin.Droid.Resource.Id.material_timepicker_edit_text;
			global::Xamarin.Forms.Platform.Android.Resource.Id.material_timepicker_mode_button = global::pjsua2xamarin.Droid.Resource.Id.material_timepicker_mode_button;
			global::Xamarin.Forms.Platform.Android.Resource.Id.material_timepicker_ok_button = global::pjsua2xamarin.Droid.Resource.Id.material_timepicker_ok_button;
			global::Xamarin.Forms.Platform.Android.Resource.Id.material_timepicker_view = global::pjsua2xamarin.Droid.Resource.Id.material_timepicker_view;
			global::Xamarin.Forms.Platform.Android.Resource.Id.material_value_index = global::pjsua2xamarin.Droid.Resource.Id.material_value_index;
			global::Xamarin.Forms.Platform.Android.Resource.Id.media_actions = global::pjsua2xamarin.Droid.Resource.Id.media_actions;
			global::Xamarin.Forms.Platform.Android.Resource.Id.media_controller_compat_view_tag = global::pjsua2xamarin.Droid.Resource.Id.media_controller_compat_view_tag;
			global::Xamarin.Forms.Platform.Android.Resource.Id.message = global::pjsua2xamarin.Droid.Resource.Id.message;
			global::Xamarin.Forms.Platform.Android.Resource.Id.META = global::pjsua2xamarin.Droid.Resource.Id.META;
			global::Xamarin.Forms.Platform.Android.Resource.Id.middle = global::pjsua2xamarin.Droid.Resource.Id.middle;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mini = global::pjsua2xamarin.Droid.Resource.Id.mini;
			global::Xamarin.Forms.Platform.Android.Resource.Id.month_grid = global::pjsua2xamarin.Droid.Resource.Id.month_grid;
			global::Xamarin.Forms.Platform.Android.Resource.Id.month_navigation_bar = global::pjsua2xamarin.Droid.Resource.Id.month_navigation_bar;
			global::Xamarin.Forms.Platform.Android.Resource.Id.month_navigation_fragment_toggle = global::pjsua2xamarin.Droid.Resource.Id.month_navigation_fragment_toggle;
			global::Xamarin.Forms.Platform.Android.Resource.Id.month_navigation_next = global::pjsua2xamarin.Droid.Resource.Id.month_navigation_next;
			global::Xamarin.Forms.Platform.Android.Resource.Id.month_navigation_previous = global::pjsua2xamarin.Droid.Resource.Id.month_navigation_previous;
			global::Xamarin.Forms.Platform.Android.Resource.Id.month_title = global::pjsua2xamarin.Droid.Resource.Id.month_title;
			global::Xamarin.Forms.Platform.Android.Resource.Id.motion_base = global::pjsua2xamarin.Droid.Resource.Id.motion_base;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_anchor_parent = global::pjsua2xamarin.Droid.Resource.Id.mtrl_anchor_parent;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_calendar_days_of_week = global::pjsua2xamarin.Droid.Resource.Id.mtrl_calendar_days_of_week;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_calendar_day_selector_frame = global::pjsua2xamarin.Droid.Resource.Id.mtrl_calendar_day_selector_frame;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_calendar_frame = global::pjsua2xamarin.Droid.Resource.Id.mtrl_calendar_frame;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_calendar_main_pane = global::pjsua2xamarin.Droid.Resource.Id.mtrl_calendar_main_pane;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_calendar_months = global::pjsua2xamarin.Droid.Resource.Id.mtrl_calendar_months;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_calendar_selection_frame = global::pjsua2xamarin.Droid.Resource.Id.mtrl_calendar_selection_frame;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_calendar_text_input_frame = global::pjsua2xamarin.Droid.Resource.Id.mtrl_calendar_text_input_frame;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_calendar_year_selector_frame = global::pjsua2xamarin.Droid.Resource.Id.mtrl_calendar_year_selector_frame;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_card_checked_layer_id = global::pjsua2xamarin.Droid.Resource.Id.mtrl_card_checked_layer_id;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_child_content_container = global::pjsua2xamarin.Droid.Resource.Id.mtrl_child_content_container;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_internal_children_alpha_tag = global::pjsua2xamarin.Droid.Resource.Id.mtrl_internal_children_alpha_tag;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_motion_snapshot_view = global::pjsua2xamarin.Droid.Resource.Id.mtrl_motion_snapshot_view;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_picker_fullscreen = global::pjsua2xamarin.Droid.Resource.Id.mtrl_picker_fullscreen;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_picker_header = global::pjsua2xamarin.Droid.Resource.Id.mtrl_picker_header;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_picker_header_selection_text = global::pjsua2xamarin.Droid.Resource.Id.mtrl_picker_header_selection_text;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_picker_header_title_and_selection = global::pjsua2xamarin.Droid.Resource.Id.mtrl_picker_header_title_and_selection;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_picker_header_toggle = global::pjsua2xamarin.Droid.Resource.Id.mtrl_picker_header_toggle;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_picker_text_input_date = global::pjsua2xamarin.Droid.Resource.Id.mtrl_picker_text_input_date;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_picker_text_input_range_end = global::pjsua2xamarin.Droid.Resource.Id.mtrl_picker_text_input_range_end;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_picker_text_input_range_start = global::pjsua2xamarin.Droid.Resource.Id.mtrl_picker_text_input_range_start;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_picker_title_text = global::pjsua2xamarin.Droid.Resource.Id.mtrl_picker_title_text;
			global::Xamarin.Forms.Platform.Android.Resource.Id.mtrl_view_tag_bottom_padding = global::pjsua2xamarin.Droid.Resource.Id.mtrl_view_tag_bottom_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Id.multiply = global::pjsua2xamarin.Droid.Resource.Id.multiply;
			global::Xamarin.Forms.Platform.Android.Resource.Id.navigation_bar_item_icon_view = global::pjsua2xamarin.Droid.Resource.Id.navigation_bar_item_icon_view;
			global::Xamarin.Forms.Platform.Android.Resource.Id.navigation_bar_item_labels_group = global::pjsua2xamarin.Droid.Resource.Id.navigation_bar_item_labels_group;
			global::Xamarin.Forms.Platform.Android.Resource.Id.navigation_bar_item_large_label_view = global::pjsua2xamarin.Droid.Resource.Id.navigation_bar_item_large_label_view;
			global::Xamarin.Forms.Platform.Android.Resource.Id.navigation_bar_item_small_label_view = global::pjsua2xamarin.Droid.Resource.Id.navigation_bar_item_small_label_view;
			global::Xamarin.Forms.Platform.Android.Resource.Id.navigation_header_container = global::pjsua2xamarin.Droid.Resource.Id.navigation_header_container;
			global::Xamarin.Forms.Platform.Android.Resource.Id.nav_controller_view_tag = global::pjsua2xamarin.Droid.Resource.Id.nav_controller_view_tag;
			global::Xamarin.Forms.Platform.Android.Resource.Id.never = global::pjsua2xamarin.Droid.Resource.Id.never;
			global::Xamarin.Forms.Platform.Android.Resource.Id.neverCompleteToEnd = global::pjsua2xamarin.Droid.Resource.Id.neverCompleteToEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Id.neverCompleteToStart = global::pjsua2xamarin.Droid.Resource.Id.neverCompleteToStart;
			global::Xamarin.Forms.Platform.Android.Resource.Id.none = global::pjsua2xamarin.Droid.Resource.Id.none;
			global::Xamarin.Forms.Platform.Android.Resource.Id.normal = global::pjsua2xamarin.Droid.Resource.Id.normal;
			global::Xamarin.Forms.Platform.Android.Resource.Id.north = global::pjsua2xamarin.Droid.Resource.Id.north;
			global::Xamarin.Forms.Platform.Android.Resource.Id.noScroll = global::pjsua2xamarin.Droid.Resource.Id.noScroll;
			global::Xamarin.Forms.Platform.Android.Resource.Id.noState = global::pjsua2xamarin.Droid.Resource.Id.noState;
			global::Xamarin.Forms.Platform.Android.Resource.Id.notification_background = global::pjsua2xamarin.Droid.Resource.Id.notification_background;
			global::Xamarin.Forms.Platform.Android.Resource.Id.notification_main_column = global::pjsua2xamarin.Droid.Resource.Id.notification_main_column;
			global::Xamarin.Forms.Platform.Android.Resource.Id.notification_main_column_container = global::pjsua2xamarin.Droid.Resource.Id.notification_main_column_container;
			global::Xamarin.Forms.Platform.Android.Resource.Id.NO_DEBUG = global::pjsua2xamarin.Droid.Resource.Id.NO_DEBUG;
			global::Xamarin.Forms.Platform.Android.Resource.Id.off = global::pjsua2xamarin.Droid.Resource.Id.off;
			global::Xamarin.Forms.Platform.Android.Resource.Id.on = global::pjsua2xamarin.Droid.Resource.Id.on;
			global::Xamarin.Forms.Platform.Android.Resource.Id.outline = global::pjsua2xamarin.Droid.Resource.Id.outline;
			global::Xamarin.Forms.Platform.Android.Resource.Id.outward = global::pjsua2xamarin.Droid.Resource.Id.outward;
			global::Xamarin.Forms.Platform.Android.Resource.Id.overshoot = global::pjsua2xamarin.Droid.Resource.Id.overshoot;
			global::Xamarin.Forms.Platform.Android.Resource.Id.packed = global::pjsua2xamarin.Droid.Resource.Id.packed;
			global::Xamarin.Forms.Platform.Android.Resource.Id.parallax = global::pjsua2xamarin.Droid.Resource.Id.parallax;
			global::Xamarin.Forms.Platform.Android.Resource.Id.parent = global::pjsua2xamarin.Droid.Resource.Id.parent;
			global::Xamarin.Forms.Platform.Android.Resource.Id.parentPanel = global::pjsua2xamarin.Droid.Resource.Id.parentPanel;
			global::Xamarin.Forms.Platform.Android.Resource.Id.parentRelative = global::pjsua2xamarin.Droid.Resource.Id.parentRelative;
			global::Xamarin.Forms.Platform.Android.Resource.Id.parent_matrix = global::pjsua2xamarin.Droid.Resource.Id.parent_matrix;
			global::Xamarin.Forms.Platform.Android.Resource.Id.password_toggle = global::pjsua2xamarin.Droid.Resource.Id.password_toggle;
			global::Xamarin.Forms.Platform.Android.Resource.Id.path = global::pjsua2xamarin.Droid.Resource.Id.path;
			global::Xamarin.Forms.Platform.Android.Resource.Id.pathRelative = global::pjsua2xamarin.Droid.Resource.Id.pathRelative;
			global::Xamarin.Forms.Platform.Android.Resource.Id.peekHeight = global::pjsua2xamarin.Droid.Resource.Id.peekHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Id.percent = global::pjsua2xamarin.Droid.Resource.Id.percent;
			global::Xamarin.Forms.Platform.Android.Resource.Id.pin = global::pjsua2xamarin.Droid.Resource.Id.pin;
			global::Xamarin.Forms.Platform.Android.Resource.Id.position = global::pjsua2xamarin.Droid.Resource.Id.position;
			global::Xamarin.Forms.Platform.Android.Resource.Id.postLayout = global::pjsua2xamarin.Droid.Resource.Id.postLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Id.progress_circular = global::pjsua2xamarin.Droid.Resource.Id.progress_circular;
			global::Xamarin.Forms.Platform.Android.Resource.Id.progress_horizontal = global::pjsua2xamarin.Droid.Resource.Id.progress_horizontal;
			global::Xamarin.Forms.Platform.Android.Resource.Id.radio = global::pjsua2xamarin.Droid.Resource.Id.radio;
			global::Xamarin.Forms.Platform.Android.Resource.Id.ratio = global::pjsua2xamarin.Droid.Resource.Id.ratio;
			global::Xamarin.Forms.Platform.Android.Resource.Id.rectangles = global::pjsua2xamarin.Droid.Resource.Id.rectangles;
			global::Xamarin.Forms.Platform.Android.Resource.Id.reverseSawtooth = global::pjsua2xamarin.Droid.Resource.Id.reverseSawtooth;
			global::Xamarin.Forms.Platform.Android.Resource.Id.right = global::pjsua2xamarin.Droid.Resource.Id.right;
			global::Xamarin.Forms.Platform.Android.Resource.Id.rightToLeft = global::pjsua2xamarin.Droid.Resource.Id.rightToLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Id.right_icon = global::pjsua2xamarin.Droid.Resource.Id.right_icon;
			global::Xamarin.Forms.Platform.Android.Resource.Id.right_side = global::pjsua2xamarin.Droid.Resource.Id.right_side;
			global::Xamarin.Forms.Platform.Android.Resource.Id.rounded = global::pjsua2xamarin.Droid.Resource.Id.rounded;
			global::Xamarin.Forms.Platform.Android.Resource.Id.row_index_key = global::pjsua2xamarin.Droid.Resource.Id.row_index_key;
			global::Xamarin.Forms.Platform.Android.Resource.Id.save_non_transition_alpha = global::pjsua2xamarin.Droid.Resource.Id.save_non_transition_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Id.save_overlay_view = global::pjsua2xamarin.Droid.Resource.Id.save_overlay_view;
			global::Xamarin.Forms.Platform.Android.Resource.Id.sawtooth = global::pjsua2xamarin.Droid.Resource.Id.sawtooth;
			global::Xamarin.Forms.Platform.Android.Resource.Id.scale = global::pjsua2xamarin.Droid.Resource.Id.scale;
			global::Xamarin.Forms.Platform.Android.Resource.Id.screen = global::pjsua2xamarin.Droid.Resource.Id.screen;
			global::Xamarin.Forms.Platform.Android.Resource.Id.scroll = global::pjsua2xamarin.Droid.Resource.Id.scroll;
			global::Xamarin.Forms.Platform.Android.Resource.Id.scrollable = global::pjsua2xamarin.Droid.Resource.Id.scrollable;
			global::Xamarin.Forms.Platform.Android.Resource.Id.scrollIndicatorDown = global::pjsua2xamarin.Droid.Resource.Id.scrollIndicatorDown;
			global::Xamarin.Forms.Platform.Android.Resource.Id.scrollIndicatorUp = global::pjsua2xamarin.Droid.Resource.Id.scrollIndicatorUp;
			global::Xamarin.Forms.Platform.Android.Resource.Id.scrollView = global::pjsua2xamarin.Droid.Resource.Id.scrollView;
			global::Xamarin.Forms.Platform.Android.Resource.Id.search_badge = global::pjsua2xamarin.Droid.Resource.Id.search_badge;
			global::Xamarin.Forms.Platform.Android.Resource.Id.search_bar = global::pjsua2xamarin.Droid.Resource.Id.search_bar;
			global::Xamarin.Forms.Platform.Android.Resource.Id.search_button = global::pjsua2xamarin.Droid.Resource.Id.search_button;
			global::Xamarin.Forms.Platform.Android.Resource.Id.search_close_btn = global::pjsua2xamarin.Droid.Resource.Id.search_close_btn;
			global::Xamarin.Forms.Platform.Android.Resource.Id.search_edit_frame = global::pjsua2xamarin.Droid.Resource.Id.search_edit_frame;
			global::Xamarin.Forms.Platform.Android.Resource.Id.search_go_btn = global::pjsua2xamarin.Droid.Resource.Id.search_go_btn;
			global::Xamarin.Forms.Platform.Android.Resource.Id.search_mag_icon = global::pjsua2xamarin.Droid.Resource.Id.search_mag_icon;
			global::Xamarin.Forms.Platform.Android.Resource.Id.search_plate = global::pjsua2xamarin.Droid.Resource.Id.search_plate;
			global::Xamarin.Forms.Platform.Android.Resource.Id.search_src_text = global::pjsua2xamarin.Droid.Resource.Id.search_src_text;
			global::Xamarin.Forms.Platform.Android.Resource.Id.search_voice_btn = global::pjsua2xamarin.Droid.Resource.Id.search_voice_btn;
			global::Xamarin.Forms.Platform.Android.Resource.Id.selected = global::pjsua2xamarin.Droid.Resource.Id.selected;
			global::Xamarin.Forms.Platform.Android.Resource.Id.selection_type = global::pjsua2xamarin.Droid.Resource.Id.selection_type;
			global::Xamarin.Forms.Platform.Android.Resource.Id.select_dialog_listview = global::pjsua2xamarin.Droid.Resource.Id.select_dialog_listview;
			global::Xamarin.Forms.Platform.Android.Resource.Id.sharedValueSet = global::pjsua2xamarin.Droid.Resource.Id.sharedValueSet;
			global::Xamarin.Forms.Platform.Android.Resource.Id.sharedValueUnset = global::pjsua2xamarin.Droid.Resource.Id.sharedValueUnset;
			global::Xamarin.Forms.Platform.Android.Resource.Id.shellcontent_appbar = global::pjsua2xamarin.Droid.Resource.Id.shellcontent_appbar;
			global::Xamarin.Forms.Platform.Android.Resource.Id.shellcontent_toolbar = global::pjsua2xamarin.Droid.Resource.Id.shellcontent_toolbar;
			global::Xamarin.Forms.Platform.Android.Resource.Id.SHIFT = global::pjsua2xamarin.Droid.Resource.Id.SHIFT;
			global::Xamarin.Forms.Platform.Android.Resource.Id.shortcut = global::pjsua2xamarin.Droid.Resource.Id.shortcut;
			global::Xamarin.Forms.Platform.Android.Resource.Id.showCustom = global::pjsua2xamarin.Droid.Resource.Id.showCustom;
			global::Xamarin.Forms.Platform.Android.Resource.Id.showHome = global::pjsua2xamarin.Droid.Resource.Id.showHome;
			global::Xamarin.Forms.Platform.Android.Resource.Id.showTitle = global::pjsua2xamarin.Droid.Resource.Id.showTitle;
			global::Xamarin.Forms.Platform.Android.Resource.Id.SHOW_ALL = global::pjsua2xamarin.Droid.Resource.Id.SHOW_ALL;
			global::Xamarin.Forms.Platform.Android.Resource.Id.SHOW_PATH = global::pjsua2xamarin.Droid.Resource.Id.SHOW_PATH;
			global::Xamarin.Forms.Platform.Android.Resource.Id.SHOW_PROGRESS = global::pjsua2xamarin.Droid.Resource.Id.SHOW_PROGRESS;
			global::Xamarin.Forms.Platform.Android.Resource.Id.sin = global::pjsua2xamarin.Droid.Resource.Id.sin;
			global::Xamarin.Forms.Platform.Android.Resource.Id.skipCollapsed = global::pjsua2xamarin.Droid.Resource.Id.skipCollapsed;
			global::Xamarin.Forms.Platform.Android.Resource.Id.skipped = global::pjsua2xamarin.Droid.Resource.Id.skipped;
			global::Xamarin.Forms.Platform.Android.Resource.Id.slide = global::pjsua2xamarin.Droid.Resource.Id.slide;
			global::Xamarin.Forms.Platform.Android.Resource.Id.sliding_tabs = global::pjsua2xamarin.Droid.Resource.Id.sliding_tabs;
			global::Xamarin.Forms.Platform.Android.Resource.Id.snackbar_action = global::pjsua2xamarin.Droid.Resource.Id.snackbar_action;
			global::Xamarin.Forms.Platform.Android.Resource.Id.snackbar_text = global::pjsua2xamarin.Droid.Resource.Id.snackbar_text;
			global::Xamarin.Forms.Platform.Android.Resource.Id.snap = global::pjsua2xamarin.Droid.Resource.Id.snap;
			global::Xamarin.Forms.Platform.Android.Resource.Id.snapMargins = global::pjsua2xamarin.Droid.Resource.Id.snapMargins;
			global::Xamarin.Forms.Platform.Android.Resource.Id.south = global::pjsua2xamarin.Droid.Resource.Id.south;
			global::Xamarin.Forms.Platform.Android.Resource.Id.spacer = global::pjsua2xamarin.Droid.Resource.Id.spacer;
			global::Xamarin.Forms.Platform.Android.Resource.Id.special_effects_controller_view_tag = global::pjsua2xamarin.Droid.Resource.Id.special_effects_controller_view_tag;
			global::Xamarin.Forms.Platform.Android.Resource.Id.spline = global::pjsua2xamarin.Droid.Resource.Id.spline;
			global::Xamarin.Forms.Platform.Android.Resource.Id.split_action_bar = global::pjsua2xamarin.Droid.Resource.Id.split_action_bar;
			global::Xamarin.Forms.Platform.Android.Resource.Id.spread = global::pjsua2xamarin.Droid.Resource.Id.spread;
			global::Xamarin.Forms.Platform.Android.Resource.Id.spread_inside = global::pjsua2xamarin.Droid.Resource.Id.spread_inside;
			global::Xamarin.Forms.Platform.Android.Resource.Id.spring = global::pjsua2xamarin.Droid.Resource.Id.spring;
			global::Xamarin.Forms.Platform.Android.Resource.Id.square = global::pjsua2xamarin.Droid.Resource.Id.square;
			global::Xamarin.Forms.Platform.Android.Resource.Id.src_atop = global::pjsua2xamarin.Droid.Resource.Id.src_atop;
			global::Xamarin.Forms.Platform.Android.Resource.Id.src_in = global::pjsua2xamarin.Droid.Resource.Id.src_in;
			global::Xamarin.Forms.Platform.Android.Resource.Id.src_over = global::pjsua2xamarin.Droid.Resource.Id.src_over;
			global::Xamarin.Forms.Platform.Android.Resource.Id.standard = global::pjsua2xamarin.Droid.Resource.Id.standard;
			global::Xamarin.Forms.Platform.Android.Resource.Id.start = global::pjsua2xamarin.Droid.Resource.Id.start;
			global::Xamarin.Forms.Platform.Android.Resource.Id.startHorizontal = global::pjsua2xamarin.Droid.Resource.Id.startHorizontal;
			global::Xamarin.Forms.Platform.Android.Resource.Id.startToEnd = global::pjsua2xamarin.Droid.Resource.Id.startToEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Id.startVertical = global::pjsua2xamarin.Droid.Resource.Id.startVertical;
			global::Xamarin.Forms.Platform.Android.Resource.Id.staticLayout = global::pjsua2xamarin.Droid.Resource.Id.staticLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Id.staticPostLayout = global::pjsua2xamarin.Droid.Resource.Id.staticPostLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Id.status_bar_latest_event_content = global::pjsua2xamarin.Droid.Resource.Id.status_bar_latest_event_content;
			global::Xamarin.Forms.Platform.Android.Resource.Id.stop = global::pjsua2xamarin.Droid.Resource.Id.stop;
			global::Xamarin.Forms.Platform.Android.Resource.Id.stretch = global::pjsua2xamarin.Droid.Resource.Id.stretch;
			global::Xamarin.Forms.Platform.Android.Resource.Id.submenuarrow = global::pjsua2xamarin.Droid.Resource.Id.submenuarrow;
			global::Xamarin.Forms.Platform.Android.Resource.Id.submit_area = global::pjsua2xamarin.Droid.Resource.Id.submit_area;
			global::Xamarin.Forms.Platform.Android.Resource.Id.supportScrollUp = global::pjsua2xamarin.Droid.Resource.Id.supportScrollUp;
			global::Xamarin.Forms.Platform.Android.Resource.Id.SYM = global::pjsua2xamarin.Droid.Resource.Id.SYM;
			global::Xamarin.Forms.Platform.Android.Resource.Id.tabMode = global::pjsua2xamarin.Droid.Resource.Id.tabMode;
			global::Xamarin.Forms.Platform.Android.Resource.Id.tag_accessibility_actions = global::pjsua2xamarin.Droid.Resource.Id.tag_accessibility_actions;
			global::Xamarin.Forms.Platform.Android.Resource.Id.tag_accessibility_clickable_spans = global::pjsua2xamarin.Droid.Resource.Id.tag_accessibility_clickable_spans;
			global::Xamarin.Forms.Platform.Android.Resource.Id.tag_accessibility_heading = global::pjsua2xamarin.Droid.Resource.Id.tag_accessibility_heading;
			global::Xamarin.Forms.Platform.Android.Resource.Id.tag_accessibility_pane_title = global::pjsua2xamarin.Droid.Resource.Id.tag_accessibility_pane_title;
			global::Xamarin.Forms.Platform.Android.Resource.Id.tag_on_apply_window_listener = global::pjsua2xamarin.Droid.Resource.Id.tag_on_apply_window_listener;
			global::Xamarin.Forms.Platform.Android.Resource.Id.tag_on_receive_content_listener = global::pjsua2xamarin.Droid.Resource.Id.tag_on_receive_content_listener;
			global::Xamarin.Forms.Platform.Android.Resource.Id.tag_on_receive_content_mime_types = global::pjsua2xamarin.Droid.Resource.Id.tag_on_receive_content_mime_types;
			global::Xamarin.Forms.Platform.Android.Resource.Id.tag_screen_reader_focusable = global::pjsua2xamarin.Droid.Resource.Id.tag_screen_reader_focusable;
			global::Xamarin.Forms.Platform.Android.Resource.Id.tag_state_description = global::pjsua2xamarin.Droid.Resource.Id.tag_state_description;
			global::Xamarin.Forms.Platform.Android.Resource.Id.tag_transition_group = global::pjsua2xamarin.Droid.Resource.Id.tag_transition_group;
			global::Xamarin.Forms.Platform.Android.Resource.Id.tag_unhandled_key_event_manager = global::pjsua2xamarin.Droid.Resource.Id.tag_unhandled_key_event_manager;
			global::Xamarin.Forms.Platform.Android.Resource.Id.tag_unhandled_key_listeners = global::pjsua2xamarin.Droid.Resource.Id.tag_unhandled_key_listeners;
			global::Xamarin.Forms.Platform.Android.Resource.Id.tag_window_insets_animation_callback = global::pjsua2xamarin.Droid.Resource.Id.tag_window_insets_animation_callback;
			global::Xamarin.Forms.Platform.Android.Resource.Id.test_checkbox_android_button_tint = global::pjsua2xamarin.Droid.Resource.Id.test_checkbox_android_button_tint;
			global::Xamarin.Forms.Platform.Android.Resource.Id.test_checkbox_app_button_tint = global::pjsua2xamarin.Droid.Resource.Id.test_checkbox_app_button_tint;
			global::Xamarin.Forms.Platform.Android.Resource.Id.test_radiobutton_android_button_tint = global::pjsua2xamarin.Droid.Resource.Id.test_radiobutton_android_button_tint;
			global::Xamarin.Forms.Platform.Android.Resource.Id.test_radiobutton_app_button_tint = global::pjsua2xamarin.Droid.Resource.Id.test_radiobutton_app_button_tint;
			global::Xamarin.Forms.Platform.Android.Resource.Id.text = global::pjsua2xamarin.Droid.Resource.Id.text;
			global::Xamarin.Forms.Platform.Android.Resource.Id.text2 = global::pjsua2xamarin.Droid.Resource.Id.text2;
			global::Xamarin.Forms.Platform.Android.Resource.Id.textEnd = global::pjsua2xamarin.Droid.Resource.Id.textEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Id.textinput_counter = global::pjsua2xamarin.Droid.Resource.Id.textinput_counter;
			global::Xamarin.Forms.Platform.Android.Resource.Id.textinput_error = global::pjsua2xamarin.Droid.Resource.Id.textinput_error;
			global::Xamarin.Forms.Platform.Android.Resource.Id.textinput_helper_text = global::pjsua2xamarin.Droid.Resource.Id.textinput_helper_text;
			global::Xamarin.Forms.Platform.Android.Resource.Id.textinput_placeholder = global::pjsua2xamarin.Droid.Resource.Id.textinput_placeholder;
			global::Xamarin.Forms.Platform.Android.Resource.Id.textinput_prefix_text = global::pjsua2xamarin.Droid.Resource.Id.textinput_prefix_text;
			global::Xamarin.Forms.Platform.Android.Resource.Id.textinput_suffix_text = global::pjsua2xamarin.Droid.Resource.Id.textinput_suffix_text;
			global::Xamarin.Forms.Platform.Android.Resource.Id.textSpacerNoButtons = global::pjsua2xamarin.Droid.Resource.Id.textSpacerNoButtons;
			global::Xamarin.Forms.Platform.Android.Resource.Id.textSpacerNoTitle = global::pjsua2xamarin.Droid.Resource.Id.textSpacerNoTitle;
			global::Xamarin.Forms.Platform.Android.Resource.Id.textStart = global::pjsua2xamarin.Droid.Resource.Id.textStart;
			global::Xamarin.Forms.Platform.Android.Resource.Id.textTop = global::pjsua2xamarin.Droid.Resource.Id.textTop;
			global::Xamarin.Forms.Platform.Android.Resource.Id.text_input_end_icon = global::pjsua2xamarin.Droid.Resource.Id.text_input_end_icon;
			global::Xamarin.Forms.Platform.Android.Resource.Id.text_input_error_icon = global::pjsua2xamarin.Droid.Resource.Id.text_input_error_icon;
			global::Xamarin.Forms.Platform.Android.Resource.Id.text_input_start_icon = global::pjsua2xamarin.Droid.Resource.Id.text_input_start_icon;
			global::Xamarin.Forms.Platform.Android.Resource.Id.time = global::pjsua2xamarin.Droid.Resource.Id.time;
			global::Xamarin.Forms.Platform.Android.Resource.Id.title = global::pjsua2xamarin.Droid.Resource.Id.title;
			global::Xamarin.Forms.Platform.Android.Resource.Id.titleDividerNoCustom = global::pjsua2xamarin.Droid.Resource.Id.titleDividerNoCustom;
			global::Xamarin.Forms.Platform.Android.Resource.Id.title_template = global::pjsua2xamarin.Droid.Resource.Id.title_template;
			global::Xamarin.Forms.Platform.Android.Resource.Id.toggle = global::pjsua2xamarin.Droid.Resource.Id.toggle;
			global::Xamarin.Forms.Platform.Android.Resource.Id.toolbar = global::pjsua2xamarin.Droid.Resource.Id.toolbar;
			global::Xamarin.Forms.Platform.Android.Resource.Id.top = global::pjsua2xamarin.Droid.Resource.Id.top;
			global::Xamarin.Forms.Platform.Android.Resource.Id.topPanel = global::pjsua2xamarin.Droid.Resource.Id.topPanel;
			global::Xamarin.Forms.Platform.Android.Resource.Id.TOP_END = global::pjsua2xamarin.Droid.Resource.Id.TOP_END;
			global::Xamarin.Forms.Platform.Android.Resource.Id.TOP_START = global::pjsua2xamarin.Droid.Resource.Id.TOP_START;
			global::Xamarin.Forms.Platform.Android.Resource.Id.touch_outside = global::pjsua2xamarin.Droid.Resource.Id.touch_outside;
			global::Xamarin.Forms.Platform.Android.Resource.Id.transitionToEnd = global::pjsua2xamarin.Droid.Resource.Id.transitionToEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Id.transitionToStart = global::pjsua2xamarin.Droid.Resource.Id.transitionToStart;
			global::Xamarin.Forms.Platform.Android.Resource.Id.transition_current_scene = global::pjsua2xamarin.Droid.Resource.Id.transition_current_scene;
			global::Xamarin.Forms.Platform.Android.Resource.Id.transition_layout_save = global::pjsua2xamarin.Droid.Resource.Id.transition_layout_save;
			global::Xamarin.Forms.Platform.Android.Resource.Id.transition_position = global::pjsua2xamarin.Droid.Resource.Id.transition_position;
			global::Xamarin.Forms.Platform.Android.Resource.Id.transition_scene_layoutid_cache = global::pjsua2xamarin.Droid.Resource.Id.transition_scene_layoutid_cache;
			global::Xamarin.Forms.Platform.Android.Resource.Id.transition_transform = global::pjsua2xamarin.Droid.Resource.Id.transition_transform;
			global::Xamarin.Forms.Platform.Android.Resource.Id.triangle = global::pjsua2xamarin.Droid.Resource.Id.triangle;
			global::Xamarin.Forms.Platform.Android.Resource.Id.@unchecked = global::pjsua2xamarin.Droid.Resource.Id.@unchecked;
			global::Xamarin.Forms.Platform.Android.Resource.Id.uniform = global::pjsua2xamarin.Droid.Resource.Id.uniform;
			global::Xamarin.Forms.Platform.Android.Resource.Id.unlabeled = global::pjsua2xamarin.Droid.Resource.Id.unlabeled;
			global::Xamarin.Forms.Platform.Android.Resource.Id.up = global::pjsua2xamarin.Droid.Resource.Id.up;
			global::Xamarin.Forms.Platform.Android.Resource.Id.useLogo = global::pjsua2xamarin.Droid.Resource.Id.useLogo;
			global::Xamarin.Forms.Platform.Android.Resource.Id.vertical_only = global::pjsua2xamarin.Droid.Resource.Id.vertical_only;
			global::Xamarin.Forms.Platform.Android.Resource.Id.view_offset_helper = global::pjsua2xamarin.Droid.Resource.Id.view_offset_helper;
			global::Xamarin.Forms.Platform.Android.Resource.Id.view_transition = global::pjsua2xamarin.Droid.Resource.Id.view_transition;
			global::Xamarin.Forms.Platform.Android.Resource.Id.view_tree_lifecycle_owner = global::pjsua2xamarin.Droid.Resource.Id.view_tree_lifecycle_owner;
			global::Xamarin.Forms.Platform.Android.Resource.Id.view_tree_saved_state_registry_owner = global::pjsua2xamarin.Droid.Resource.Id.view_tree_saved_state_registry_owner;
			global::Xamarin.Forms.Platform.Android.Resource.Id.view_tree_view_model_store_owner = global::pjsua2xamarin.Droid.Resource.Id.view_tree_view_model_store_owner;
			global::Xamarin.Forms.Platform.Android.Resource.Id.visible = global::pjsua2xamarin.Droid.Resource.Id.visible;
			global::Xamarin.Forms.Platform.Android.Resource.Id.visible_removing_fragment_view_tag = global::pjsua2xamarin.Droid.Resource.Id.visible_removing_fragment_view_tag;
			global::Xamarin.Forms.Platform.Android.Resource.Id.west = global::pjsua2xamarin.Droid.Resource.Id.west;
			global::Xamarin.Forms.Platform.Android.Resource.Id.withinBounds = global::pjsua2xamarin.Droid.Resource.Id.withinBounds;
			global::Xamarin.Forms.Platform.Android.Resource.Id.withText = global::pjsua2xamarin.Droid.Resource.Id.withText;
			global::Xamarin.Forms.Platform.Android.Resource.Id.wrap = global::pjsua2xamarin.Droid.Resource.Id.wrap;
			global::Xamarin.Forms.Platform.Android.Resource.Id.wrap_content = global::pjsua2xamarin.Droid.Resource.Id.wrap_content;
			global::Xamarin.Forms.Platform.Android.Resource.Id.wrap_content_constrained = global::pjsua2xamarin.Droid.Resource.Id.wrap_content_constrained;
			global::Xamarin.Forms.Platform.Android.Resource.Id.x_left = global::pjsua2xamarin.Droid.Resource.Id.x_left;
			global::Xamarin.Forms.Platform.Android.Resource.Id.x_right = global::pjsua2xamarin.Droid.Resource.Id.x_right;
			global::Xamarin.Forms.Platform.Android.Resource.Id.zero_corner_chip = global::pjsua2xamarin.Droid.Resource.Id.zero_corner_chip;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.abc_config_activityDefaultDur = global::pjsua2xamarin.Droid.Resource.Integer.abc_config_activityDefaultDur;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.abc_config_activityShortDur = global::pjsua2xamarin.Droid.Resource.Integer.abc_config_activityShortDur;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.app_bar_elevation_anim_duration = global::pjsua2xamarin.Droid.Resource.Integer.app_bar_elevation_anim_duration;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.bottom_sheet_slide_duration = global::pjsua2xamarin.Droid.Resource.Integer.bottom_sheet_slide_duration;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.cancel_button_image_alpha = global::pjsua2xamarin.Droid.Resource.Integer.cancel_button_image_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.config_navAnimTime = global::pjsua2xamarin.Droid.Resource.Integer.config_navAnimTime;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.config_tooltipAnimTime = global::pjsua2xamarin.Droid.Resource.Integer.config_tooltipAnimTime;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.design_snackbar_text_max_lines = global::pjsua2xamarin.Droid.Resource.Integer.design_snackbar_text_max_lines;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.design_tab_indicator_anim_duration_ms = global::pjsua2xamarin.Droid.Resource.Integer.design_tab_indicator_anim_duration_ms;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.hide_password_duration = global::pjsua2xamarin.Droid.Resource.Integer.hide_password_duration;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.material_motion_duration_long_1 = global::pjsua2xamarin.Droid.Resource.Integer.material_motion_duration_long_1;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.material_motion_duration_long_2 = global::pjsua2xamarin.Droid.Resource.Integer.material_motion_duration_long_2;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.material_motion_duration_medium_1 = global::pjsua2xamarin.Droid.Resource.Integer.material_motion_duration_medium_1;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.material_motion_duration_medium_2 = global::pjsua2xamarin.Droid.Resource.Integer.material_motion_duration_medium_2;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.material_motion_duration_short_1 = global::pjsua2xamarin.Droid.Resource.Integer.material_motion_duration_short_1;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.material_motion_duration_short_2 = global::pjsua2xamarin.Droid.Resource.Integer.material_motion_duration_short_2;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.material_motion_path = global::pjsua2xamarin.Droid.Resource.Integer.material_motion_path;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.mtrl_badge_max_character_count = global::pjsua2xamarin.Droid.Resource.Integer.mtrl_badge_max_character_count;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.mtrl_btn_anim_delay_ms = global::pjsua2xamarin.Droid.Resource.Integer.mtrl_btn_anim_delay_ms;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.mtrl_btn_anim_duration_ms = global::pjsua2xamarin.Droid.Resource.Integer.mtrl_btn_anim_duration_ms;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.mtrl_calendar_header_orientation = global::pjsua2xamarin.Droid.Resource.Integer.mtrl_calendar_header_orientation;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.mtrl_calendar_selection_text_lines = global::pjsua2xamarin.Droid.Resource.Integer.mtrl_calendar_selection_text_lines;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.mtrl_calendar_year_selector_span = global::pjsua2xamarin.Droid.Resource.Integer.mtrl_calendar_year_selector_span;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.mtrl_card_anim_delay_ms = global::pjsua2xamarin.Droid.Resource.Integer.mtrl_card_anim_delay_ms;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.mtrl_card_anim_duration_ms = global::pjsua2xamarin.Droid.Resource.Integer.mtrl_card_anim_duration_ms;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.mtrl_chip_anim_duration = global::pjsua2xamarin.Droid.Resource.Integer.mtrl_chip_anim_duration;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.mtrl_tab_indicator_anim_duration_ms = global::pjsua2xamarin.Droid.Resource.Integer.mtrl_tab_indicator_anim_duration_ms;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.show_password_duration = global::pjsua2xamarin.Droid.Resource.Integer.show_password_duration;
			global::Xamarin.Forms.Platform.Android.Resource.Integer.status_bar_notification_info_maxnum = global::pjsua2xamarin.Droid.Resource.Integer.status_bar_notification_info_maxnum;
			global::Xamarin.Forms.Platform.Android.Resource.Interpolator.btn_checkbox_checked_mtrl_animation_interpolator_0 = global::pjsua2xamarin.Droid.Resource.Interpolator.btn_checkbox_checked_mtrl_animation_interpolator_0;
			global::Xamarin.Forms.Platform.Android.Resource.Interpolator.btn_checkbox_checked_mtrl_animation_interpolator_1 = global::pjsua2xamarin.Droid.Resource.Interpolator.btn_checkbox_checked_mtrl_animation_interpolator_1;
			global::Xamarin.Forms.Platform.Android.Resource.Interpolator.btn_checkbox_unchecked_mtrl_animation_interpolator_0 = global::pjsua2xamarin.Droid.Resource.Interpolator.btn_checkbox_unchecked_mtrl_animation_interpolator_0;
			global::Xamarin.Forms.Platform.Android.Resource.Interpolator.btn_checkbox_unchecked_mtrl_animation_interpolator_1 = global::pjsua2xamarin.Droid.Resource.Interpolator.btn_checkbox_unchecked_mtrl_animation_interpolator_1;
			global::Xamarin.Forms.Platform.Android.Resource.Interpolator.btn_radio_to_off_mtrl_animation_interpolator_0 = global::pjsua2xamarin.Droid.Resource.Interpolator.btn_radio_to_off_mtrl_animation_interpolator_0;
			global::Xamarin.Forms.Platform.Android.Resource.Interpolator.btn_radio_to_on_mtrl_animation_interpolator_0 = global::pjsua2xamarin.Droid.Resource.Interpolator.btn_radio_to_on_mtrl_animation_interpolator_0;
			global::Xamarin.Forms.Platform.Android.Resource.Interpolator.fast_out_slow_in = global::pjsua2xamarin.Droid.Resource.Interpolator.fast_out_slow_in;
			global::Xamarin.Forms.Platform.Android.Resource.Interpolator.mtrl_fast_out_linear_in = global::pjsua2xamarin.Droid.Resource.Interpolator.mtrl_fast_out_linear_in;
			global::Xamarin.Forms.Platform.Android.Resource.Interpolator.mtrl_fast_out_slow_in = global::pjsua2xamarin.Droid.Resource.Interpolator.mtrl_fast_out_slow_in;
			global::Xamarin.Forms.Platform.Android.Resource.Interpolator.mtrl_linear = global::pjsua2xamarin.Droid.Resource.Interpolator.mtrl_linear;
			global::Xamarin.Forms.Platform.Android.Resource.Interpolator.mtrl_linear_out_slow_in = global::pjsua2xamarin.Droid.Resource.Interpolator.mtrl_linear_out_slow_in;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_action_bar_title_item = global::pjsua2xamarin.Droid.Resource.Layout.abc_action_bar_title_item;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_action_bar_up_container = global::pjsua2xamarin.Droid.Resource.Layout.abc_action_bar_up_container;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_action_menu_item_layout = global::pjsua2xamarin.Droid.Resource.Layout.abc_action_menu_item_layout;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_action_menu_layout = global::pjsua2xamarin.Droid.Resource.Layout.abc_action_menu_layout;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_action_mode_bar = global::pjsua2xamarin.Droid.Resource.Layout.abc_action_mode_bar;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_action_mode_close_item_material = global::pjsua2xamarin.Droid.Resource.Layout.abc_action_mode_close_item_material;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_activity_chooser_view = global::pjsua2xamarin.Droid.Resource.Layout.abc_activity_chooser_view;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_activity_chooser_view_list_item = global::pjsua2xamarin.Droid.Resource.Layout.abc_activity_chooser_view_list_item;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_alert_dialog_button_bar_material = global::pjsua2xamarin.Droid.Resource.Layout.abc_alert_dialog_button_bar_material;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_alert_dialog_material = global::pjsua2xamarin.Droid.Resource.Layout.abc_alert_dialog_material;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_alert_dialog_title_material = global::pjsua2xamarin.Droid.Resource.Layout.abc_alert_dialog_title_material;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_cascading_menu_item_layout = global::pjsua2xamarin.Droid.Resource.Layout.abc_cascading_menu_item_layout;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_dialog_title_material = global::pjsua2xamarin.Droid.Resource.Layout.abc_dialog_title_material;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_expanded_menu_layout = global::pjsua2xamarin.Droid.Resource.Layout.abc_expanded_menu_layout;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_list_menu_item_checkbox = global::pjsua2xamarin.Droid.Resource.Layout.abc_list_menu_item_checkbox;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_list_menu_item_icon = global::pjsua2xamarin.Droid.Resource.Layout.abc_list_menu_item_icon;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_list_menu_item_layout = global::pjsua2xamarin.Droid.Resource.Layout.abc_list_menu_item_layout;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_list_menu_item_radio = global::pjsua2xamarin.Droid.Resource.Layout.abc_list_menu_item_radio;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_popup_menu_header_item_layout = global::pjsua2xamarin.Droid.Resource.Layout.abc_popup_menu_header_item_layout;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_popup_menu_item_layout = global::pjsua2xamarin.Droid.Resource.Layout.abc_popup_menu_item_layout;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_screen_content_include = global::pjsua2xamarin.Droid.Resource.Layout.abc_screen_content_include;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_screen_simple = global::pjsua2xamarin.Droid.Resource.Layout.abc_screen_simple;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_screen_simple_overlay_action_mode = global::pjsua2xamarin.Droid.Resource.Layout.abc_screen_simple_overlay_action_mode;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_screen_toolbar = global::pjsua2xamarin.Droid.Resource.Layout.abc_screen_toolbar;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_search_dropdown_item_icons_2line = global::pjsua2xamarin.Droid.Resource.Layout.abc_search_dropdown_item_icons_2line;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_search_view = global::pjsua2xamarin.Droid.Resource.Layout.abc_search_view;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_select_dialog_material = global::pjsua2xamarin.Droid.Resource.Layout.abc_select_dialog_material;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.abc_tooltip = global::pjsua2xamarin.Droid.Resource.Layout.abc_tooltip;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.BottomTabLayout = global::pjsua2xamarin.Droid.Resource.Layout.BottomTabLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.custom_dialog = global::pjsua2xamarin.Droid.Resource.Layout.custom_dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.design_bottom_navigation_item = global::pjsua2xamarin.Droid.Resource.Layout.design_bottom_navigation_item;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.design_bottom_sheet_dialog = global::pjsua2xamarin.Droid.Resource.Layout.design_bottom_sheet_dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.design_layout_snackbar = global::pjsua2xamarin.Droid.Resource.Layout.design_layout_snackbar;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.design_layout_snackbar_include = global::pjsua2xamarin.Droid.Resource.Layout.design_layout_snackbar_include;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.design_layout_tab_icon = global::pjsua2xamarin.Droid.Resource.Layout.design_layout_tab_icon;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.design_layout_tab_text = global::pjsua2xamarin.Droid.Resource.Layout.design_layout_tab_text;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.design_menu_item_action_area = global::pjsua2xamarin.Droid.Resource.Layout.design_menu_item_action_area;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.design_navigation_item = global::pjsua2xamarin.Droid.Resource.Layout.design_navigation_item;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.design_navigation_item_header = global::pjsua2xamarin.Droid.Resource.Layout.design_navigation_item_header;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.design_navigation_item_separator = global::pjsua2xamarin.Droid.Resource.Layout.design_navigation_item_separator;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.design_navigation_item_subheader = global::pjsua2xamarin.Droid.Resource.Layout.design_navigation_item_subheader;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.design_navigation_menu = global::pjsua2xamarin.Droid.Resource.Layout.design_navigation_menu;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.design_navigation_menu_item = global::pjsua2xamarin.Droid.Resource.Layout.design_navigation_menu_item;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.design_text_input_end_icon = global::pjsua2xamarin.Droid.Resource.Layout.design_text_input_end_icon;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.design_text_input_start_icon = global::pjsua2xamarin.Droid.Resource.Layout.design_text_input_start_icon;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.FallbackTabbarDoNotUse = global::pjsua2xamarin.Droid.Resource.Layout.FallbackTabbarDoNotUse;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.FallbackToolbarDoNotUse = global::pjsua2xamarin.Droid.Resource.Layout.FallbackToolbarDoNotUse;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.FlyoutContent = global::pjsua2xamarin.Droid.Resource.Layout.FlyoutContent;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.material_chip_input_combo = global::pjsua2xamarin.Droid.Resource.Layout.material_chip_input_combo;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.material_clockface_textview = global::pjsua2xamarin.Droid.Resource.Layout.material_clockface_textview;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.material_clockface_view = global::pjsua2xamarin.Droid.Resource.Layout.material_clockface_view;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.material_clock_display = global::pjsua2xamarin.Droid.Resource.Layout.material_clock_display;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.material_clock_display_divider = global::pjsua2xamarin.Droid.Resource.Layout.material_clock_display_divider;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.material_clock_period_toggle = global::pjsua2xamarin.Droid.Resource.Layout.material_clock_period_toggle;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.material_clock_period_toggle_land = global::pjsua2xamarin.Droid.Resource.Layout.material_clock_period_toggle_land;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.material_radial_view_group = global::pjsua2xamarin.Droid.Resource.Layout.material_radial_view_group;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.material_textinput_timepicker = global::pjsua2xamarin.Droid.Resource.Layout.material_textinput_timepicker;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.material_timepicker = global::pjsua2xamarin.Droid.Resource.Layout.material_timepicker;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.material_timepicker_dialog = global::pjsua2xamarin.Droid.Resource.Layout.material_timepicker_dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.material_timepicker_textinput_display = global::pjsua2xamarin.Droid.Resource.Layout.material_timepicker_textinput_display;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.material_time_chip = global::pjsua2xamarin.Droid.Resource.Layout.material_time_chip;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.material_time_input = global::pjsua2xamarin.Droid.Resource.Layout.material_time_input;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_alert_dialog = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_alert_dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_alert_dialog_actions = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_alert_dialog_actions;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_alert_dialog_title = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_alert_dialog_title;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_alert_select_dialog_item = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_alert_select_dialog_item;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_alert_select_dialog_multichoice = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_alert_select_dialog_multichoice;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_alert_select_dialog_singlechoice = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_alert_select_dialog_singlechoice;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_calendar_day = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_calendar_day;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_calendar_days_of_week = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_calendar_days_of_week;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_calendar_day_of_week = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_calendar_day_of_week;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_calendar_horizontal = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_calendar_horizontal;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_calendar_month = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_calendar_month;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_calendar_months = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_calendar_months;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_calendar_month_labeled = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_calendar_month_labeled;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_calendar_month_navigation = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_calendar_month_navigation;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_calendar_vertical = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_calendar_vertical;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_calendar_year = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_calendar_year;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_layout_snackbar = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_layout_snackbar;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_layout_snackbar_include = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_layout_snackbar_include;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_navigation_rail_item = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_navigation_rail_item;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_picker_actions = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_picker_actions;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_picker_dialog = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_picker_dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_picker_fullscreen = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_picker_fullscreen;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_picker_header_dialog = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_picker_header_dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_picker_header_fullscreen = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_picker_header_fullscreen;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_picker_header_selection_text = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_picker_header_selection_text;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_picker_header_title_text = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_picker_header_title_text;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_picker_header_toggle = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_picker_header_toggle;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_picker_text_input_date = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_picker_text_input_date;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.mtrl_picker_text_input_date_range = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_picker_text_input_date_range;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.notification_action = global::pjsua2xamarin.Droid.Resource.Layout.notification_action;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.notification_action_tombstone = global::pjsua2xamarin.Droid.Resource.Layout.notification_action_tombstone;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.notification_media_action = global::pjsua2xamarin.Droid.Resource.Layout.notification_media_action;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.notification_media_cancel_action = global::pjsua2xamarin.Droid.Resource.Layout.notification_media_cancel_action;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.notification_template_big_media = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_big_media;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.notification_template_big_media_custom = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_big_media_custom;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.notification_template_big_media_narrow = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_big_media_narrow;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.notification_template_big_media_narrow_custom = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_big_media_narrow_custom;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.notification_template_custom_big = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_custom_big;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.notification_template_icon_group = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_icon_group;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.notification_template_lines_media = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_lines_media;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.notification_template_media = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_media;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.notification_template_media_custom = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_media_custom;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.notification_template_part_chronometer = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_part_chronometer;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.notification_template_part_time = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_part_time;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.RootLayout = global::pjsua2xamarin.Droid.Resource.Layout.RootLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.select_dialog_item_material = global::pjsua2xamarin.Droid.Resource.Layout.select_dialog_item_material;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.select_dialog_multichoice_material = global::pjsua2xamarin.Droid.Resource.Layout.select_dialog_multichoice_material;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.select_dialog_singlechoice_material = global::pjsua2xamarin.Droid.Resource.Layout.select_dialog_singlechoice_material;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.ShellContent = global::pjsua2xamarin.Droid.Resource.Layout.ShellContent;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.support_simple_spinner_dropdown_item = global::pjsua2xamarin.Droid.Resource.Layout.support_simple_spinner_dropdown_item;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.Tabbar = global::pjsua2xamarin.Droid.Resource.Layout.Tabbar;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.test_action_chip = global::pjsua2xamarin.Droid.Resource.Layout.test_action_chip;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.test_chip_zero_corner_radius = global::pjsua2xamarin.Droid.Resource.Layout.test_chip_zero_corner_radius;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.test_design_checkbox = global::pjsua2xamarin.Droid.Resource.Layout.test_design_checkbox;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.test_design_radiobutton = global::pjsua2xamarin.Droid.Resource.Layout.test_design_radiobutton;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.test_navigation_bar_item_layout = global::pjsua2xamarin.Droid.Resource.Layout.test_navigation_bar_item_layout;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.test_reflow_chipgroup = global::pjsua2xamarin.Droid.Resource.Layout.test_reflow_chipgroup;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.test_toolbar = global::pjsua2xamarin.Droid.Resource.Layout.test_toolbar;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.test_toolbar_custom_background = global::pjsua2xamarin.Droid.Resource.Layout.test_toolbar_custom_background;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.test_toolbar_elevation = global::pjsua2xamarin.Droid.Resource.Layout.test_toolbar_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.test_toolbar_surface = global::pjsua2xamarin.Droid.Resource.Layout.test_toolbar_surface;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.text_view_without_line_height = global::pjsua2xamarin.Droid.Resource.Layout.text_view_without_line_height;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.text_view_with_line_height_from_appearance = global::pjsua2xamarin.Droid.Resource.Layout.text_view_with_line_height_from_appearance;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.text_view_with_line_height_from_layout = global::pjsua2xamarin.Droid.Resource.Layout.text_view_with_line_height_from_layout;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.text_view_with_line_height_from_style = global::pjsua2xamarin.Droid.Resource.Layout.text_view_with_line_height_from_style;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.text_view_with_theme_line_height = global::pjsua2xamarin.Droid.Resource.Layout.text_view_with_theme_line_height;
			global::Xamarin.Forms.Platform.Android.Resource.Layout.Toolbar = global::pjsua2xamarin.Droid.Resource.Layout.Toolbar;
			global::Xamarin.Forms.Platform.Android.Resource.Plurals.mtrl_badge_content_description = global::pjsua2xamarin.Droid.Resource.Plurals.mtrl_badge_content_description;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_action_bar_home_description = global::pjsua2xamarin.Droid.Resource.String.abc_action_bar_home_description;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_action_bar_up_description = global::pjsua2xamarin.Droid.Resource.String.abc_action_bar_up_description;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_action_menu_overflow_description = global::pjsua2xamarin.Droid.Resource.String.abc_action_menu_overflow_description;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_action_mode_done = global::pjsua2xamarin.Droid.Resource.String.abc_action_mode_done;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_activitychooserview_choose_application = global::pjsua2xamarin.Droid.Resource.String.abc_activitychooserview_choose_application;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_activity_chooser_view_see_all = global::pjsua2xamarin.Droid.Resource.String.abc_activity_chooser_view_see_all;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_capital_off = global::pjsua2xamarin.Droid.Resource.String.abc_capital_off;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_capital_on = global::pjsua2xamarin.Droid.Resource.String.abc_capital_on;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_menu_alt_shortcut_label = global::pjsua2xamarin.Droid.Resource.String.abc_menu_alt_shortcut_label;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_menu_ctrl_shortcut_label = global::pjsua2xamarin.Droid.Resource.String.abc_menu_ctrl_shortcut_label;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_menu_delete_shortcut_label = global::pjsua2xamarin.Droid.Resource.String.abc_menu_delete_shortcut_label;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_menu_enter_shortcut_label = global::pjsua2xamarin.Droid.Resource.String.abc_menu_enter_shortcut_label;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_menu_function_shortcut_label = global::pjsua2xamarin.Droid.Resource.String.abc_menu_function_shortcut_label;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_menu_meta_shortcut_label = global::pjsua2xamarin.Droid.Resource.String.abc_menu_meta_shortcut_label;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_menu_shift_shortcut_label = global::pjsua2xamarin.Droid.Resource.String.abc_menu_shift_shortcut_label;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_menu_space_shortcut_label = global::pjsua2xamarin.Droid.Resource.String.abc_menu_space_shortcut_label;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_menu_sym_shortcut_label = global::pjsua2xamarin.Droid.Resource.String.abc_menu_sym_shortcut_label;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_prepend_shortcut_label = global::pjsua2xamarin.Droid.Resource.String.abc_prepend_shortcut_label;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_searchview_description_clear = global::pjsua2xamarin.Droid.Resource.String.abc_searchview_description_clear;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_searchview_description_query = global::pjsua2xamarin.Droid.Resource.String.abc_searchview_description_query;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_searchview_description_search = global::pjsua2xamarin.Droid.Resource.String.abc_searchview_description_search;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_searchview_description_submit = global::pjsua2xamarin.Droid.Resource.String.abc_searchview_description_submit;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_searchview_description_voice = global::pjsua2xamarin.Droid.Resource.String.abc_searchview_description_voice;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_search_hint = global::pjsua2xamarin.Droid.Resource.String.abc_search_hint;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_shareactionprovider_share_with = global::pjsua2xamarin.Droid.Resource.String.abc_shareactionprovider_share_with;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_shareactionprovider_share_with_application = global::pjsua2xamarin.Droid.Resource.String.abc_shareactionprovider_share_with_application;
			global::Xamarin.Forms.Platform.Android.Resource.String.abc_toolbar_collapse_description = global::pjsua2xamarin.Droid.Resource.String.abc_toolbar_collapse_description;
			global::Xamarin.Forms.Platform.Android.Resource.String.appbar_scrolling_view_behavior = global::pjsua2xamarin.Droid.Resource.String.appbar_scrolling_view_behavior;
			global::Xamarin.Forms.Platform.Android.Resource.String.bottomsheet_action_expand_halfway = global::pjsua2xamarin.Droid.Resource.String.bottomsheet_action_expand_halfway;
			global::Xamarin.Forms.Platform.Android.Resource.String.bottom_sheet_behavior = global::pjsua2xamarin.Droid.Resource.String.bottom_sheet_behavior;
			global::Xamarin.Forms.Platform.Android.Resource.String.character_counter_content_description = global::pjsua2xamarin.Droid.Resource.String.character_counter_content_description;
			global::Xamarin.Forms.Platform.Android.Resource.String.character_counter_overflowed_content_description = global::pjsua2xamarin.Droid.Resource.String.character_counter_overflowed_content_description;
			global::Xamarin.Forms.Platform.Android.Resource.String.character_counter_pattern = global::pjsua2xamarin.Droid.Resource.String.character_counter_pattern;
			global::Xamarin.Forms.Platform.Android.Resource.String.chip_text = global::pjsua2xamarin.Droid.Resource.String.chip_text;
			global::Xamarin.Forms.Platform.Android.Resource.String.clear_text_end_icon_content_description = global::pjsua2xamarin.Droid.Resource.String.clear_text_end_icon_content_description;
			global::Xamarin.Forms.Platform.Android.Resource.String.error_icon_content_description = global::pjsua2xamarin.Droid.Resource.String.error_icon_content_description;
			global::Xamarin.Forms.Platform.Android.Resource.String.exposed_dropdown_menu_content_description = global::pjsua2xamarin.Droid.Resource.String.exposed_dropdown_menu_content_description;
			global::Xamarin.Forms.Platform.Android.Resource.String.fab_transformation_scrim_behavior = global::pjsua2xamarin.Droid.Resource.String.fab_transformation_scrim_behavior;
			global::Xamarin.Forms.Platform.Android.Resource.String.fab_transformation_sheet_behavior = global::pjsua2xamarin.Droid.Resource.String.fab_transformation_sheet_behavior;
			global::Xamarin.Forms.Platform.Android.Resource.String.hide_bottom_view_on_scroll_behavior = global::pjsua2xamarin.Droid.Resource.String.hide_bottom_view_on_scroll_behavior;
			global::Xamarin.Forms.Platform.Android.Resource.String.icon_content_description = global::pjsua2xamarin.Droid.Resource.String.icon_content_description;
			global::Xamarin.Forms.Platform.Android.Resource.String.item_view_role_description = global::pjsua2xamarin.Droid.Resource.String.item_view_role_description;
			global::Xamarin.Forms.Platform.Android.Resource.String.material_clock_display_divider = global::pjsua2xamarin.Droid.Resource.String.material_clock_display_divider;
			global::Xamarin.Forms.Platform.Android.Resource.String.material_clock_toggle_content_description = global::pjsua2xamarin.Droid.Resource.String.material_clock_toggle_content_description;
			global::Xamarin.Forms.Platform.Android.Resource.String.material_hour_selection = global::pjsua2xamarin.Droid.Resource.String.material_hour_selection;
			global::Xamarin.Forms.Platform.Android.Resource.String.material_hour_suffix = global::pjsua2xamarin.Droid.Resource.String.material_hour_suffix;
			global::Xamarin.Forms.Platform.Android.Resource.String.material_minute_selection = global::pjsua2xamarin.Droid.Resource.String.material_minute_selection;
			global::Xamarin.Forms.Platform.Android.Resource.String.material_minute_suffix = global::pjsua2xamarin.Droid.Resource.String.material_minute_suffix;
			global::Xamarin.Forms.Platform.Android.Resource.String.material_motion_easing_accelerated = global::pjsua2xamarin.Droid.Resource.String.material_motion_easing_accelerated;
			global::Xamarin.Forms.Platform.Android.Resource.String.material_motion_easing_decelerated = global::pjsua2xamarin.Droid.Resource.String.material_motion_easing_decelerated;
			global::Xamarin.Forms.Platform.Android.Resource.String.material_motion_easing_emphasized = global::pjsua2xamarin.Droid.Resource.String.material_motion_easing_emphasized;
			global::Xamarin.Forms.Platform.Android.Resource.String.material_motion_easing_linear = global::pjsua2xamarin.Droid.Resource.String.material_motion_easing_linear;
			global::Xamarin.Forms.Platform.Android.Resource.String.material_motion_easing_standard = global::pjsua2xamarin.Droid.Resource.String.material_motion_easing_standard;
			global::Xamarin.Forms.Platform.Android.Resource.String.material_slider_range_end = global::pjsua2xamarin.Droid.Resource.String.material_slider_range_end;
			global::Xamarin.Forms.Platform.Android.Resource.String.material_slider_range_start = global::pjsua2xamarin.Droid.Resource.String.material_slider_range_start;
			global::Xamarin.Forms.Platform.Android.Resource.String.material_timepicker_am = global::pjsua2xamarin.Droid.Resource.String.material_timepicker_am;
			global::Xamarin.Forms.Platform.Android.Resource.String.material_timepicker_clock_mode_description = global::pjsua2xamarin.Droid.Resource.String.material_timepicker_clock_mode_description;
			global::Xamarin.Forms.Platform.Android.Resource.String.material_timepicker_hour = global::pjsua2xamarin.Droid.Resource.String.material_timepicker_hour;
			global::Xamarin.Forms.Platform.Android.Resource.String.material_timepicker_minute = global::pjsua2xamarin.Droid.Resource.String.material_timepicker_minute;
			global::Xamarin.Forms.Platform.Android.Resource.String.material_timepicker_pm = global::pjsua2xamarin.Droid.Resource.String.material_timepicker_pm;
			global::Xamarin.Forms.Platform.Android.Resource.String.material_timepicker_select_time = global::pjsua2xamarin.Droid.Resource.String.material_timepicker_select_time;
			global::Xamarin.Forms.Platform.Android.Resource.String.material_timepicker_text_input_mode_description = global::pjsua2xamarin.Droid.Resource.String.material_timepicker_text_input_mode_description;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_badge_numberless_content_description = global::pjsua2xamarin.Droid.Resource.String.mtrl_badge_numberless_content_description;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_chip_close_icon_content_description = global::pjsua2xamarin.Droid.Resource.String.mtrl_chip_close_icon_content_description;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_exceed_max_badge_number_content_description = global::pjsua2xamarin.Droid.Resource.String.mtrl_exceed_max_badge_number_content_description;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_exceed_max_badge_number_suffix = global::pjsua2xamarin.Droid.Resource.String.mtrl_exceed_max_badge_number_suffix;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_a11y_next_month = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_a11y_next_month;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_a11y_prev_month = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_a11y_prev_month;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_announce_current_selection = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_announce_current_selection;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_cancel = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_cancel;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_confirm = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_confirm;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_date_header_selected = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_date_header_selected;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_date_header_title = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_date_header_title;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_date_header_unselected = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_date_header_unselected;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_day_of_week_column_header = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_day_of_week_column_header;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_invalid_format = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_invalid_format;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_invalid_format_example = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_invalid_format_example;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_invalid_format_use = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_invalid_format_use;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_invalid_range = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_invalid_range;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_navigate_to_year_description = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_navigate_to_year_description;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_out_of_range = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_out_of_range;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_range_header_only_end_selected = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_range_header_only_end_selected;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_range_header_only_start_selected = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_range_header_only_start_selected;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_range_header_selected = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_range_header_selected;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_range_header_title = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_range_header_title;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_range_header_unselected = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_range_header_unselected;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_save = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_save;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_text_input_date_hint = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_text_input_date_hint;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_text_input_date_range_end_hint = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_text_input_date_range_end_hint;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_text_input_date_range_start_hint = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_text_input_date_range_start_hint;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_text_input_day_abbr = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_text_input_day_abbr;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_text_input_month_abbr = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_text_input_month_abbr;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_text_input_year_abbr = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_text_input_year_abbr;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_toggle_to_calendar_input_mode = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_toggle_to_calendar_input_mode;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_toggle_to_day_selection = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_toggle_to_day_selection;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_toggle_to_text_input_mode = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_toggle_to_text_input_mode;
			global::Xamarin.Forms.Platform.Android.Resource.String.mtrl_picker_toggle_to_year_selection = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_toggle_to_year_selection;
			global::Xamarin.Forms.Platform.Android.Resource.String.nav_app_bar_navigate_up_description = global::pjsua2xamarin.Droid.Resource.String.nav_app_bar_navigate_up_description;
			global::Xamarin.Forms.Platform.Android.Resource.String.nav_app_bar_open_drawer_description = global::pjsua2xamarin.Droid.Resource.String.nav_app_bar_open_drawer_description;
			global::Xamarin.Forms.Platform.Android.Resource.String.overflow_tab_title = global::pjsua2xamarin.Droid.Resource.String.overflow_tab_title;
			global::Xamarin.Forms.Platform.Android.Resource.String.password_toggle_content_description = global::pjsua2xamarin.Droid.Resource.String.password_toggle_content_description;
			global::Xamarin.Forms.Platform.Android.Resource.String.path_password_eye = global::pjsua2xamarin.Droid.Resource.String.path_password_eye;
			global::Xamarin.Forms.Platform.Android.Resource.String.path_password_eye_mask_strike_through = global::pjsua2xamarin.Droid.Resource.String.path_password_eye_mask_strike_through;
			global::Xamarin.Forms.Platform.Android.Resource.String.path_password_eye_mask_visible = global::pjsua2xamarin.Droid.Resource.String.path_password_eye_mask_visible;
			global::Xamarin.Forms.Platform.Android.Resource.String.path_password_strike_through = global::pjsua2xamarin.Droid.Resource.String.path_password_strike_through;
			global::Xamarin.Forms.Platform.Android.Resource.String.search_menu_title = global::pjsua2xamarin.Droid.Resource.String.search_menu_title;
			global::Xamarin.Forms.Platform.Android.Resource.String.status_bar_notification_info_overflow = global::pjsua2xamarin.Droid.Resource.String.status_bar_notification_info_overflow;
			global::Xamarin.Forms.Platform.Android.Resource.Style.AlertDialog_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.AlertDialog_AppCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Style.AlertDialog_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.AlertDialog_AppCompat_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.AndroidThemeColorAccentYellow = global::pjsua2xamarin.Droid.Resource.Style.AndroidThemeColorAccentYellow;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Animation_AppCompat_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Animation_AppCompat_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Animation_AppCompat_DropDownUp = global::pjsua2xamarin.Droid.Resource.Style.Animation_AppCompat_DropDownUp;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Animation_AppCompat_Tooltip = global::pjsua2xamarin.Droid.Resource.Style.Animation_AppCompat_Tooltip;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Animation_Design_BottomSheetDialog = global::pjsua2xamarin.Droid.Resource.Style.Animation_Design_BottomSheetDialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Animation_MaterialComponents_BottomSheetDialog = global::pjsua2xamarin.Droid.Resource.Style.Animation_MaterialComponents_BottomSheetDialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.AppCompatDialogStyle = global::pjsua2xamarin.Droid.Resource.Style.AppCompatDialogStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_AlertDialog_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_AlertDialog_AppCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_AlertDialog_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_AlertDialog_AppCompat_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Animation_AppCompat_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_Animation_AppCompat_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Animation_AppCompat_DropDownUp = global::pjsua2xamarin.Droid.Resource.Style.Base_Animation_AppCompat_DropDownUp;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Animation_AppCompat_Tooltip = global::pjsua2xamarin.Droid.Resource.Style.Base_Animation_AppCompat_Tooltip;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_CardView = global::pjsua2xamarin.Droid.Resource.Style.Base_CardView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_DialogWindowTitleBackground_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_DialogWindowTitleBackground_AppCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_DialogWindowTitle_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_DialogWindowTitle_AppCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_MaterialAlertDialog_MaterialComponents_Title_Icon = global::pjsua2xamarin.Droid.Resource.Style.Base_MaterialAlertDialog_MaterialComponents_Title_Icon;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_MaterialAlertDialog_MaterialComponents_Title_Panel = global::pjsua2xamarin.Droid.Resource.Style.Base_MaterialAlertDialog_MaterialComponents_Title_Panel;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_MaterialAlertDialog_MaterialComponents_Title_Text = global::pjsua2xamarin.Droid.Resource.Style.Base_MaterialAlertDialog_MaterialComponents_Title_Text;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Body1 = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Body1;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Body2 = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Body2;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Button = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Button;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Caption = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Caption;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Display1 = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Display1;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Display2 = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Display2;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Display3 = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Display3;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Display4 = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Display4;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Headline = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Headline;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Large = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Large;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Large_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Large_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Light_Widget_PopupMenu_Large = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Light_Widget_PopupMenu_Large;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Light_Widget_PopupMenu_Small = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Light_Widget_PopupMenu_Small;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Medium = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Medium;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Medium_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Medium_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Menu = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Menu;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_SearchResult = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_SearchResult;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_SearchResult_Subtitle = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_SearchResult_Subtitle;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_SearchResult_Title = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_SearchResult_Title;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Small = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Small;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Small_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Small_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Subhead = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Subhead;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Subhead_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Subhead_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Title = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Title;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Title_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Title_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Tooltip = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Tooltip;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionBar_Menu = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionBar_Menu;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionBar_Subtitle = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionBar_Subtitle;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionBar_Subtitle_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionBar_Subtitle_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionBar_Title = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionBar_Title;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionBar_Title_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionBar_Title_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionMode_Subtitle = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionMode_Subtitle;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionMode_Title = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionMode_Title;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Widget_Button = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_Button;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Widget_Button_Borderless_Colored = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_Button_Borderless_Colored;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Widget_Button_Colored = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_Button_Colored;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Widget_Button_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_Button_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Widget_DropDownItem = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_DropDownItem;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Widget_PopupMenu_Header = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_PopupMenu_Header;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Widget_PopupMenu_Large = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_PopupMenu_Large;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Widget_PopupMenu_Small = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_PopupMenu_Small;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Widget_Switch = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_Switch;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_AppCompat_Widget_TextView_SpinnerItem = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_TextView_SpinnerItem;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_MaterialComponents_Badge = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_MaterialComponents_Badge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_MaterialComponents_Button = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_MaterialComponents_Button;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_MaterialComponents_Headline6 = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_MaterialComponents_Headline6;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_MaterialComponents_Subtitle2 = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_MaterialComponents_Subtitle2;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_Widget_AppCompat_ExpandedMenu_Item = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_Widget_AppCompat_ExpandedMenu_Item;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_Widget_AppCompat_Toolbar_Subtitle = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_Widget_AppCompat_Toolbar_Subtitle;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_TextAppearance_Widget_AppCompat_Toolbar_Title = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_Widget_AppCompat_Toolbar_Title;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_ThemeOverlay_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_AppCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_ThemeOverlay_AppCompat_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_AppCompat_ActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_ThemeOverlay_AppCompat_Dark = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_AppCompat_Dark;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_ThemeOverlay_AppCompat_Dark_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_AppCompat_Dark_ActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_ThemeOverlay_AppCompat_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_AppCompat_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_ThemeOverlay_AppCompat_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_AppCompat_Dialog_Alert;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_ThemeOverlay_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_AppCompat_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_ThemeOverlay_MaterialComponents_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_MaterialComponents_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_ThemeOverlay_MaterialComponents_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_MaterialComponents_Dialog_Alert;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_ThemeOverlay_MaterialComponents_Dialog_Alert_Framework = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_MaterialComponents_Dialog_Alert_Framework;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_ThemeOverlay_MaterialComponents_Light_Dialog_Alert_Framework = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_MaterialComponents_Light_Dialog_Alert_Framework;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_ThemeOverlay_MaterialComponents_MaterialAlertDialog = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_MaterialComponents_MaterialAlertDialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_AppCompat_CompactMenu = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_CompactMenu;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_AppCompat_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_AppCompat_DialogWhenLarge = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_DialogWhenLarge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_AppCompat_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_Dialog_Alert;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_AppCompat_Dialog_FixedSize = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_Dialog_FixedSize;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_AppCompat_Dialog_MinWidth = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_Dialog_MinWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_AppCompat_Light_DarkActionBar = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_Light_DarkActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_AppCompat_Light_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_Light_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_AppCompat_Light_DialogWhenLarge = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_Light_DialogWhenLarge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_AppCompat_Light_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_Light_Dialog_Alert;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_AppCompat_Light_Dialog_FixedSize = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_Light_Dialog_FixedSize;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_AppCompat_Light_Dialog_MinWidth = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_Light_Dialog_MinWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_MaterialComponents = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_MaterialComponents_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_MaterialComponents_CompactMenu = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_CompactMenu;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_MaterialComponents_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_MaterialComponents_DialogWhenLarge = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_DialogWhenLarge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_MaterialComponents_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Dialog_Alert;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_MaterialComponents_Dialog_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Dialog_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_MaterialComponents_Dialog_FixedSize = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Dialog_FixedSize;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_MaterialComponents_Dialog_MinWidth = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Dialog_MinWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_MaterialComponents_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_MaterialComponents_Light_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Light_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_MaterialComponents_Light_DarkActionBar = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Light_DarkActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_MaterialComponents_Light_DarkActionBar_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Light_DarkActionBar_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_MaterialComponents_Light_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Light_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_MaterialComponents_Light_DialogWhenLarge = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Light_DialogWhenLarge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_MaterialComponents_Light_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Light_Dialog_Alert;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_MaterialComponents_Light_Dialog_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Light_Dialog_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_MaterialComponents_Light_Dialog_FixedSize = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Light_Dialog_FixedSize;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Theme_MaterialComponents_Light_Dialog_MinWidth = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Light_Dialog_MinWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V14_ThemeOverlay_MaterialComponents_BottomSheetDialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_ThemeOverlay_MaterialComponents_BottomSheetDialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V14_ThemeOverlay_MaterialComponents_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_ThemeOverlay_MaterialComponents_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V14_ThemeOverlay_MaterialComponents_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_ThemeOverlay_MaterialComponents_Dialog_Alert;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V14_ThemeOverlay_MaterialComponents_MaterialAlertDialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_ThemeOverlay_MaterialComponents_MaterialAlertDialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V14_Theme_MaterialComponents = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_Theme_MaterialComponents;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V14_Theme_MaterialComponents_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_Theme_MaterialComponents_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V14_Theme_MaterialComponents_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_Theme_MaterialComponents_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V14_Theme_MaterialComponents_Dialog_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_Theme_MaterialComponents_Dialog_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V14_Theme_MaterialComponents_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_Theme_MaterialComponents_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V14_Theme_MaterialComponents_Light_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_Theme_MaterialComponents_Light_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V14_Theme_MaterialComponents_Light_DarkActionBar_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_Theme_MaterialComponents_Light_DarkActionBar_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V14_Theme_MaterialComponents_Light_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_Theme_MaterialComponents_Light_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V14_Theme_MaterialComponents_Light_Dialog_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_Theme_MaterialComponents_Light_Dialog_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V21_ThemeOverlay_AppCompat_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V21_ThemeOverlay_AppCompat_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V21_ThemeOverlay_MaterialComponents_BottomSheetDialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V21_ThemeOverlay_MaterialComponents_BottomSheetDialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V21_Theme_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_V21_Theme_AppCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V21_Theme_AppCompat_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V21_Theme_AppCompat_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V21_Theme_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_V21_Theme_AppCompat_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V21_Theme_AppCompat_Light_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V21_Theme_AppCompat_Light_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V21_Theme_MaterialComponents = global::pjsua2xamarin.Droid.Resource.Style.Base_V21_Theme_MaterialComponents;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V21_Theme_MaterialComponents_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V21_Theme_MaterialComponents_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V21_Theme_MaterialComponents_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_V21_Theme_MaterialComponents_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V21_Theme_MaterialComponents_Light_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V21_Theme_MaterialComponents_Light_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V22_Theme_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_V22_Theme_AppCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V22_Theme_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_V22_Theme_AppCompat_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V23_Theme_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_V23_Theme_AppCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V23_Theme_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_V23_Theme_AppCompat_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V26_Theme_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_V26_Theme_AppCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V26_Theme_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_V26_Theme_AppCompat_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V26_Widget_AppCompat_Toolbar = global::pjsua2xamarin.Droid.Resource.Style.Base_V26_Widget_AppCompat_Toolbar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V28_Theme_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_V28_Theme_AppCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V28_Theme_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_V28_Theme_AppCompat_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V7_ThemeOverlay_AppCompat_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V7_ThemeOverlay_AppCompat_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V7_Theme_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_V7_Theme_AppCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V7_Theme_AppCompat_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V7_Theme_AppCompat_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V7_Theme_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_V7_Theme_AppCompat_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V7_Theme_AppCompat_Light_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V7_Theme_AppCompat_Light_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V7_Widget_AppCompat_AutoCompleteTextView = global::pjsua2xamarin.Droid.Resource.Style.Base_V7_Widget_AppCompat_AutoCompleteTextView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V7_Widget_AppCompat_EditText = global::pjsua2xamarin.Droid.Resource.Style.Base_V7_Widget_AppCompat_EditText;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_V7_Widget_AppCompat_Toolbar = global::pjsua2xamarin.Droid.Resource.Style.Base_V7_Widget_AppCompat_Toolbar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_ActionBar_Solid = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ActionBar_Solid;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_ActionBar_TabBar = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ActionBar_TabBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_ActionBar_TabText = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ActionBar_TabText;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_ActionBar_TabView = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ActionBar_TabView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_ActionButton = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ActionButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_ActionButton_CloseMode = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ActionButton_CloseMode;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_ActionButton_Overflow = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ActionButton_Overflow;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_ActionMode = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ActionMode;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_ActivityChooserView = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ActivityChooserView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_AutoCompleteTextView = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_AutoCompleteTextView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_Button = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Button;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_ButtonBar = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ButtonBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_ButtonBar_AlertDialog = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ButtonBar_AlertDialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_Button_Borderless = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Button_Borderless;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_Button_Borderless_Colored = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Button_Borderless_Colored;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_Button_ButtonBar_AlertDialog = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Button_ButtonBar_AlertDialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_Button_Colored = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Button_Colored;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_Button_Small = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Button_Small;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_CompoundButton_CheckBox = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_CompoundButton_CheckBox;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_CompoundButton_RadioButton = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_CompoundButton_RadioButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_CompoundButton_Switch = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_CompoundButton_Switch;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_DrawerArrowToggle = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_DrawerArrowToggle;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_DrawerArrowToggle_Common = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_DrawerArrowToggle_Common;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_DropDownItem_Spinner = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_DropDownItem_Spinner;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_EditText = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_EditText;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_ImageButton = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ImageButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_Light_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Light_ActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_Light_ActionBar_Solid = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Light_ActionBar_Solid;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_Light_ActionBar_TabBar = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Light_ActionBar_TabBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_Light_ActionBar_TabText = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Light_ActionBar_TabText;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_Light_ActionBar_TabText_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Light_ActionBar_TabText_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_Light_ActionBar_TabView = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Light_ActionBar_TabView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_Light_PopupMenu = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Light_PopupMenu;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_Light_PopupMenu_Overflow = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Light_PopupMenu_Overflow;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_ListMenuView = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ListMenuView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_ListPopupWindow = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ListPopupWindow;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_ListView = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ListView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_ListView_DropDown = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ListView_DropDown;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_ListView_Menu = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ListView_Menu;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_PopupMenu = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_PopupMenu;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_PopupMenu_Overflow = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_PopupMenu_Overflow;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_PopupWindow = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_PopupWindow;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_ProgressBar = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ProgressBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_ProgressBar_Horizontal = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ProgressBar_Horizontal;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_RatingBar = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_RatingBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_RatingBar_Indicator = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_RatingBar_Indicator;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_RatingBar_Small = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_RatingBar_Small;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_SearchView = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_SearchView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_SearchView_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_SearchView_ActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_SeekBar = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_SeekBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_SeekBar_Discrete = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_SeekBar_Discrete;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_Spinner = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Spinner;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_Spinner_Underlined = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Spinner_Underlined;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_TextView = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_TextView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_TextView_SpinnerItem = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_TextView_SpinnerItem;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_Toolbar = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Toolbar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_AppCompat_Toolbar_Button_Navigation = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Toolbar_Button_Navigation;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_Design_TabLayout = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_Design_TabLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_MaterialComponents_AutoCompleteTextView = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_AutoCompleteTextView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_MaterialComponents_CheckedTextView = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_CheckedTextView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_MaterialComponents_Chip = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_Chip;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_MaterialComponents_MaterialCalendar_NavigationButton = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_MaterialCalendar_NavigationButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_MaterialComponents_PopupMenu = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_PopupMenu;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_MaterialComponents_PopupMenu_ContextMenu = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_PopupMenu_ContextMenu;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_MaterialComponents_PopupMenu_ListPopupWindow = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_PopupMenu_ListPopupWindow;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_MaterialComponents_PopupMenu_Overflow = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_PopupMenu_Overflow;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_MaterialComponents_Slider = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_Slider;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_MaterialComponents_Snackbar = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_Snackbar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_MaterialComponents_TextInputEditText = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_TextInputEditText;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_MaterialComponents_TextInputLayout = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_TextInputLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Base_Widget_MaterialComponents_TextView = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_TextView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.CardView = global::pjsua2xamarin.Droid.Resource.Style.CardView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.CardView_Dark = global::pjsua2xamarin.Droid.Resource.Style.CardView_Dark;
			global::Xamarin.Forms.Platform.Android.Resource.Style.CardView_Light = global::pjsua2xamarin.Droid.Resource.Style.CardView_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.collectionViewTheme = global::pjsua2xamarin.Droid.Resource.Style.collectionViewTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Style.EmptyTheme = global::pjsua2xamarin.Droid.Resource.Style.EmptyTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Style.MainTheme = global::pjsua2xamarin.Droid.Resource.Style.MainTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Style.MainTheme_Base = global::pjsua2xamarin.Droid.Resource.Style.MainTheme_Base;
			global::Xamarin.Forms.Platform.Android.Resource.Style.MaterialAlertDialog_MaterialComponents = global::pjsua2xamarin.Droid.Resource.Style.MaterialAlertDialog_MaterialComponents;
			global::Xamarin.Forms.Platform.Android.Resource.Style.MaterialAlertDialog_MaterialComponents_Body_Text = global::pjsua2xamarin.Droid.Resource.Style.MaterialAlertDialog_MaterialComponents_Body_Text;
			global::Xamarin.Forms.Platform.Android.Resource.Style.MaterialAlertDialog_MaterialComponents_Picker_Date_Calendar = global::pjsua2xamarin.Droid.Resource.Style.MaterialAlertDialog_MaterialComponents_Picker_Date_Calendar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.MaterialAlertDialog_MaterialComponents_Picker_Date_Spinner = global::pjsua2xamarin.Droid.Resource.Style.MaterialAlertDialog_MaterialComponents_Picker_Date_Spinner;
			global::Xamarin.Forms.Platform.Android.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Icon = global::pjsua2xamarin.Droid.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Icon;
			global::Xamarin.Forms.Platform.Android.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Icon_CenterStacked = global::pjsua2xamarin.Droid.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Icon_CenterStacked;
			global::Xamarin.Forms.Platform.Android.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Panel = global::pjsua2xamarin.Droid.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Panel;
			global::Xamarin.Forms.Platform.Android.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Panel_CenterStacked = global::pjsua2xamarin.Droid.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Panel_CenterStacked;
			global::Xamarin.Forms.Platform.Android.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Text = global::pjsua2xamarin.Droid.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Text;
			global::Xamarin.Forms.Platform.Android.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Text_CenterStacked = global::pjsua2xamarin.Droid.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Text_CenterStacked;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Platform_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Platform_AppCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Platform_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Platform_AppCompat_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Platform_MaterialComponents = global::pjsua2xamarin.Droid.Resource.Style.Platform_MaterialComponents;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Platform_MaterialComponents_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Platform_MaterialComponents_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Platform_MaterialComponents_Light = global::pjsua2xamarin.Droid.Resource.Style.Platform_MaterialComponents_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Platform_MaterialComponents_Light_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Platform_MaterialComponents_Light_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Platform_ThemeOverlay_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Platform_ThemeOverlay_AppCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Platform_ThemeOverlay_AppCompat_Dark = global::pjsua2xamarin.Droid.Resource.Style.Platform_ThemeOverlay_AppCompat_Dark;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Platform_ThemeOverlay_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Platform_ThemeOverlay_AppCompat_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Platform_V21_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Platform_V21_AppCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Platform_V21_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Platform_V21_AppCompat_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Platform_V25_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Platform_V25_AppCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Platform_V25_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Platform_V25_AppCompat_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Platform_Widget_AppCompat_Spinner = global::pjsua2xamarin.Droid.Resource.Style.Platform_Widget_AppCompat_Spinner;
			global::Xamarin.Forms.Platform.Android.Resource.Style.RtlOverlay_DialogWindowTitle_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_DialogWindowTitle_AppCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Style.RtlOverlay_Widget_AppCompat_ActionBar_TitleItem = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_ActionBar_TitleItem;
			global::Xamarin.Forms.Platform.Android.Resource.Style.RtlOverlay_Widget_AppCompat_DialogTitle_Icon = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_DialogTitle_Icon;
			global::Xamarin.Forms.Platform.Android.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem;
			global::Xamarin.Forms.Platform.Android.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem_InternalGroup = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem_InternalGroup;
			global::Xamarin.Forms.Platform.Android.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem_Shortcut = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem_Shortcut;
			global::Xamarin.Forms.Platform.Android.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem_SubmenuArrow = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem_SubmenuArrow;
			global::Xamarin.Forms.Platform.Android.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem_Text = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem_Text;
			global::Xamarin.Forms.Platform.Android.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem_Title = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem_Title;
			global::Xamarin.Forms.Platform.Android.Resource.Style.RtlOverlay_Widget_AppCompat_SearchView_MagIcon = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_SearchView_MagIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Style.RtlOverlay_Widget_AppCompat_Search_DropDown = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_Search_DropDown;
			global::Xamarin.Forms.Platform.Android.Resource.Style.RtlOverlay_Widget_AppCompat_Search_DropDown_Icon1 = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_Search_DropDown_Icon1;
			global::Xamarin.Forms.Platform.Android.Resource.Style.RtlOverlay_Widget_AppCompat_Search_DropDown_Icon2 = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_Search_DropDown_Icon2;
			global::Xamarin.Forms.Platform.Android.Resource.Style.RtlOverlay_Widget_AppCompat_Search_DropDown_Query = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_Search_DropDown_Query;
			global::Xamarin.Forms.Platform.Android.Resource.Style.RtlOverlay_Widget_AppCompat_Search_DropDown_Text = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_Search_DropDown_Text;
			global::Xamarin.Forms.Platform.Android.Resource.Style.RtlUnderlay_Widget_AppCompat_ActionButton = global::pjsua2xamarin.Droid.Resource.Style.RtlUnderlay_Widget_AppCompat_ActionButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.RtlUnderlay_Widget_AppCompat_ActionButton_Overflow = global::pjsua2xamarin.Droid.Resource.Style.RtlUnderlay_Widget_AppCompat_ActionButton_Overflow;
			global::Xamarin.Forms.Platform.Android.Resource.Style.scrollViewScrollBars = global::pjsua2xamarin.Droid.Resource.Style.scrollViewScrollBars;
			global::Xamarin.Forms.Platform.Android.Resource.Style.scrollViewTheme = global::pjsua2xamarin.Droid.Resource.Style.scrollViewTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ShapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ShapeAppearanceOverlay_BottomLeftDifferentCornerSize = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_BottomLeftDifferentCornerSize;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ShapeAppearanceOverlay_BottomRightCut = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_BottomRightCut;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ShapeAppearanceOverlay_Cut = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_Cut;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ShapeAppearanceOverlay_DifferentCornerSize = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_DifferentCornerSize;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_BottomSheet = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_BottomSheet;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_Chip = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_Chip;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_ExtendedFloatingActionButton = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_ExtendedFloatingActionButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_FloatingActionButton = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_FloatingActionButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_MaterialCalendar_Day = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_MaterialCalendar_Day;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_MaterialCalendar_Window_Fullscreen = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_MaterialCalendar_Window_Fullscreen;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_MaterialCalendar_Year = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_MaterialCalendar_Year;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_TextInputLayout_FilledBox = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_TextInputLayout_FilledBox;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ShapeAppearanceOverlay_TopLeftCut = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_TopLeftCut;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ShapeAppearanceOverlay_TopRightDifferentCornerSize = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_TopRightDifferentCornerSize;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ShapeAppearance_MaterialComponents = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearance_MaterialComponents;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ShapeAppearance_MaterialComponents_LargeComponent = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearance_MaterialComponents_LargeComponent;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ShapeAppearance_MaterialComponents_MediumComponent = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearance_MaterialComponents_MediumComponent;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ShapeAppearance_MaterialComponents_SmallComponent = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearance_MaterialComponents_SmallComponent;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ShapeAppearance_MaterialComponents_Test = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearance_MaterialComponents_Test;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ShapeAppearance_MaterialComponents_Tooltip = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearance_MaterialComponents_Tooltip;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TestStyleWithLineHeight = global::pjsua2xamarin.Droid.Resource.Style.TestStyleWithLineHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TestStyleWithLineHeightAppearance = global::pjsua2xamarin.Droid.Resource.Style.TestStyleWithLineHeightAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TestStyleWithoutLineHeight = global::pjsua2xamarin.Droid.Resource.Style.TestStyleWithoutLineHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TestStyleWithThemeLineHeightAttribute = global::pjsua2xamarin.Droid.Resource.Style.TestStyleWithThemeLineHeightAttribute;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TestThemeWithLineHeight = global::pjsua2xamarin.Droid.Resource.Style.TestThemeWithLineHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TestThemeWithLineHeightDisabled = global::pjsua2xamarin.Droid.Resource.Style.TestThemeWithLineHeightDisabled;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Test_ShapeAppearanceOverlay_MaterialComponents_MaterialCalendar_Day = global::pjsua2xamarin.Droid.Resource.Style.Test_ShapeAppearanceOverlay_MaterialComponents_MaterialCalendar_Day;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Test_Theme_MaterialComponents_MaterialCalendar = global::pjsua2xamarin.Droid.Resource.Style.Test_Theme_MaterialComponents_MaterialCalendar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Test_Widget_MaterialComponents_MaterialCalendar = global::pjsua2xamarin.Droid.Resource.Style.Test_Widget_MaterialComponents_MaterialCalendar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Test_Widget_MaterialComponents_MaterialCalendar_Day = global::pjsua2xamarin.Droid.Resource.Style.Test_Widget_MaterialComponents_MaterialCalendar_Day;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Test_Widget_MaterialComponents_MaterialCalendar_Day_Selected = global::pjsua2xamarin.Droid.Resource.Style.Test_Widget_MaterialComponents_MaterialCalendar_Day_Selected;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Body1 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Body1;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Body2 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Body2;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Button = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Button;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Caption = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Caption;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Display1 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Display1;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Display2 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Display2;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Display3 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Display3;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Display4 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Display4;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Headline = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Headline;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Inverse = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Large = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Large;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Large_Inverse = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Large_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Light_SearchResult_Subtitle = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Light_SearchResult_Subtitle;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Light_SearchResult_Title = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Light_SearchResult_Title;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Light_Widget_PopupMenu_Large = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Light_Widget_PopupMenu_Large;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Light_Widget_PopupMenu_Small = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Light_Widget_PopupMenu_Small;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Medium = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Medium;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Medium_Inverse = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Medium_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Menu = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Menu;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_SearchResult_Subtitle = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_SearchResult_Subtitle;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_SearchResult_Title = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_SearchResult_Title;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Small = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Small;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Small_Inverse = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Small_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Subhead = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Subhead;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Subhead_Inverse = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Subhead_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Title = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Title;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Title_Inverse = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Title_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Tooltip = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Tooltip;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Widget_ActionBar_Menu = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_ActionBar_Menu;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Widget_ActionBar_Subtitle = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_ActionBar_Subtitle;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Widget_ActionBar_Subtitle_Inverse = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_ActionBar_Subtitle_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Widget_ActionBar_Title = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_ActionBar_Title;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Widget_ActionBar_Title_Inverse = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_ActionBar_Title_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Widget_ActionMode_Subtitle = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_ActionMode_Subtitle;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Widget_ActionMode_Subtitle_Inverse = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_ActionMode_Subtitle_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Widget_ActionMode_Title = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_ActionMode_Title;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Widget_ActionMode_Title_Inverse = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_ActionMode_Title_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Widget_Button = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_Button;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Widget_Button_Borderless_Colored = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_Button_Borderless_Colored;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Widget_Button_Colored = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_Button_Colored;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Widget_Button_Inverse = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_Button_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Widget_DropDownItem = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_DropDownItem;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Widget_PopupMenu_Header = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_PopupMenu_Header;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Widget_PopupMenu_Large = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_PopupMenu_Large;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Widget_PopupMenu_Small = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_PopupMenu_Small;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Widget_Switch = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_Switch;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_AppCompat_Widget_TextView_SpinnerItem = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_TextView_SpinnerItem;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Compat_Notification = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Compat_Notification_Info = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification_Info;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Compat_Notification_Info_Media = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification_Info_Media;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Compat_Notification_Line2 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification_Line2;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Compat_Notification_Line2_Media = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification_Line2_Media;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Compat_Notification_Media = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification_Media;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Compat_Notification_Time = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification_Time;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Compat_Notification_Time_Media = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification_Time_Media;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Compat_Notification_Title = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification_Title;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Compat_Notification_Title_Media = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification_Title_Media;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Design_CollapsingToolbar_Expanded = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Design_CollapsingToolbar_Expanded;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Design_Counter = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Design_Counter;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Design_Counter_Overflow = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Design_Counter_Overflow;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Design_Error = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Design_Error;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Design_HelperText = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Design_HelperText;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Design_Hint = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Design_Hint;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Design_Placeholder = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Design_Placeholder;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Design_Prefix = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Design_Prefix;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Design_Snackbar_Message = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Design_Snackbar_Message;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Design_Suffix = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Design_Suffix;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Design_Tab = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Design_Tab;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_MaterialComponents_Badge = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Badge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_MaterialComponents_Body1 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Body1;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_MaterialComponents_Body2 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Body2;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_MaterialComponents_Button = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Button;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_MaterialComponents_Caption = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Caption;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_MaterialComponents_Chip = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Chip;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_MaterialComponents_Headline1 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Headline1;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_MaterialComponents_Headline2 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Headline2;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_MaterialComponents_Headline3 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Headline3;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_MaterialComponents_Headline4 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Headline4;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_MaterialComponents_Headline5 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Headline5;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_MaterialComponents_Headline6 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Headline6;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_MaterialComponents_Overline = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Overline;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_MaterialComponents_Subtitle1 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Subtitle1;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_MaterialComponents_Subtitle2 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Subtitle2;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_MaterialComponents_TimePicker_Title = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_TimePicker_Title;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_MaterialComponents_Tooltip = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Tooltip;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Widget_AppCompat_ExpandedMenu_Item = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Widget_AppCompat_ExpandedMenu_Item;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Widget_AppCompat_Toolbar_Subtitle = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Widget_AppCompat_Toolbar_Subtitle;
			global::Xamarin.Forms.Platform.Android.Resource.Style.TextAppearance_Widget_AppCompat_Toolbar_Title = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Widget_AppCompat_Toolbar_Title;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlayColorAccentRed = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlayColorAccentRed;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_AppCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_AppCompat_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_AppCompat_ActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_AppCompat_Dark = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_AppCompat_Dark;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_AppCompat_Dark_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_AppCompat_Dark_ActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_AppCompat_DayNight = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_AppCompat_DayNight;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_AppCompat_DayNight_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_AppCompat_DayNight_ActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_AppCompat_Dialog = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_AppCompat_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_AppCompat_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_AppCompat_Dialog_Alert;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_AppCompat_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_Design_TextInputEditText = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_Design_TextInputEditText;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_ActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_ActionBar_Primary = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_ActionBar_Primary;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_ActionBar_Surface = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_ActionBar_Surface;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_AutoCompleteTextView = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_AutoCompleteTextView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_AutoCompleteTextView_FilledBox = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_AutoCompleteTextView_FilledBox;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_AutoCompleteTextView_FilledBox_Dense = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_AutoCompleteTextView_FilledBox_Dense;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_AutoCompleteTextView_OutlinedBox = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_AutoCompleteTextView_OutlinedBox;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_AutoCompleteTextView_OutlinedBox_Dense = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_AutoCompleteTextView_OutlinedBox_Dense;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_BottomAppBar_Primary = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_BottomAppBar_Primary;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_BottomAppBar_Surface = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_BottomAppBar_Surface;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_BottomSheetDialog = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_BottomSheetDialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_Dark = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_Dark;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_Dark_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_Dark_ActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_DayNight_BottomSheetDialog = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_DayNight_BottomSheetDialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_Dialog = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_Dialog_Alert;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_Dialog_Alert_Framework = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_Dialog_Alert_Framework;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_Light = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_Light_Dialog_Alert_Framework = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_Light_Dialog_Alert_Framework;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Centered = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Centered;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date_Calendar = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date_Calendar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date_Header_Text = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date_Header_Text;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date_Header_Text_Day = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date_Header_Text_Day;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date_Spinner = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date_Spinner;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_MaterialCalendar = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_MaterialCalendar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_MaterialCalendar_Fullscreen = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_MaterialCalendar_Fullscreen;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_TextInputEditText = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_TextInputEditText;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_TextInputEditText_FilledBox = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_TextInputEditText_FilledBox;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_TextInputEditText_FilledBox_Dense = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_TextInputEditText_FilledBox_Dense;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_TextInputEditText_OutlinedBox = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_TextInputEditText_OutlinedBox;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_TextInputEditText_OutlinedBox_Dense = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_TextInputEditText_OutlinedBox_Dense;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_TimePicker = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_TimePicker;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_TimePicker_Display = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_TimePicker_Display;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_Toolbar_Primary = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_Toolbar_Primary;
			global::Xamarin.Forms.Platform.Android.Resource.Style.ThemeOverlay_MaterialComponents_Toolbar_Surface = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_Toolbar_Surface;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_AppCompat_CompactMenu = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_CompactMenu;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_AppCompat_DayNight = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_DayNight;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_AppCompat_DayNight_DarkActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_DayNight_DarkActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_AppCompat_DayNight_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_DayNight_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_AppCompat_DayNight_DialogWhenLarge = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_DayNight_DialogWhenLarge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_AppCompat_DayNight_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_DayNight_Dialog_Alert;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_AppCompat_DayNight_Dialog_MinWidth = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_DayNight_Dialog_MinWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_AppCompat_DayNight_NoActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_DayNight_NoActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_AppCompat_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_AppCompat_DialogWhenLarge = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_DialogWhenLarge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_AppCompat_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_Dialog_Alert;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_AppCompat_Dialog_MinWidth = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_Dialog_MinWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_AppCompat_Empty = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_Empty;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_AppCompat_Light_DarkActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_Light_DarkActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_AppCompat_Light_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_Light_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_AppCompat_Light_DialogWhenLarge = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_Light_DialogWhenLarge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_AppCompat_Light_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_Light_Dialog_Alert;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_AppCompat_Light_Dialog_MinWidth = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_Light_Dialog_MinWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_AppCompat_Light_NoActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_Light_NoActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_AppCompat_NoActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_NoActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_Design = global::pjsua2xamarin.Droid.Resource.Style.Theme_Design;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_Design_BottomSheetDialog = global::pjsua2xamarin.Droid.Resource.Style.Theme_Design_BottomSheetDialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_Design_Light = global::pjsua2xamarin.Droid.Resource.Style.Theme_Design_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_Design_Light_BottomSheetDialog = global::pjsua2xamarin.Droid.Resource.Style.Theme_Design_Light_BottomSheetDialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_Design_Light_NoActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_Design_Light_NoActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_Design_NoActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_Design_NoActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_BottomSheetDialog = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_BottomSheetDialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_CompactMenu = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_CompactMenu;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_DayNight = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_DayNight_BottomSheetDialog = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_BottomSheetDialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_DayNight_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_DayNight_DarkActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_DarkActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_DayNight_DarkActionBar_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_DarkActionBar_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_DayNight_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_DayNight_DialogWhenLarge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_DialogWhenLarge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_Alert;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_Alert_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_Alert_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_FixedSize = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_FixedSize;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_FixedSize_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_FixedSize_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_MinWidth = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_MinWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_MinWidth_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_MinWidth_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_DayNight_NoActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_NoActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_DayNight_NoActionBar_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_NoActionBar_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_DialogWhenLarge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DialogWhenLarge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Dialog_Alert;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Dialog_Alert_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Dialog_Alert_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Dialog_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Dialog_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Dialog_FixedSize = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Dialog_FixedSize;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Dialog_FixedSize_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Dialog_FixedSize_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Dialog_MinWidth = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Dialog_MinWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Dialog_MinWidth_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Dialog_MinWidth_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Light = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Light_BarSize = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_BarSize;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Light_BottomSheetDialog = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_BottomSheetDialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Light_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Light_DarkActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_DarkActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Light_DarkActionBar_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_DarkActionBar_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Light_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Light_DialogWhenLarge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_DialogWhenLarge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Light_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_Dialog_Alert;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Light_Dialog_Alert_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_Dialog_Alert_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Light_Dialog_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_Dialog_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Light_Dialog_FixedSize = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_Dialog_FixedSize;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Light_Dialog_FixedSize_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_Dialog_FixedSize_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Light_Dialog_MinWidth = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_Dialog_MinWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Light_Dialog_MinWidth_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_Dialog_MinWidth_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Light_LargeTouch = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_LargeTouch;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Light_NoActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_NoActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_Light_NoActionBar_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_NoActionBar_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_NoActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_NoActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Theme_MaterialComponents_NoActionBar_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_NoActionBar_Bridge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_ActionBar_Solid = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ActionBar_Solid;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_ActionBar_TabBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ActionBar_TabBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_ActionBar_TabText = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ActionBar_TabText;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_ActionBar_TabView = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ActionBar_TabView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_ActionButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ActionButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_ActionButton_CloseMode = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ActionButton_CloseMode;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_ActionButton_Overflow = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ActionButton_Overflow;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_ActionMode = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ActionMode;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_ActivityChooserView = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ActivityChooserView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_AutoCompleteTextView = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_AutoCompleteTextView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Button = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Button;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_ButtonBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ButtonBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_ButtonBar_AlertDialog = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ButtonBar_AlertDialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Button_Borderless = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Button_Borderless;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Button_Borderless_Colored = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Button_Borderless_Colored;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Button_ButtonBar_AlertDialog = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Button_ButtonBar_AlertDialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Button_Colored = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Button_Colored;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Button_Small = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Button_Small;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_CompoundButton_CheckBox = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_CompoundButton_CheckBox;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_CompoundButton_RadioButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_CompoundButton_RadioButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_CompoundButton_Switch = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_CompoundButton_Switch;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_DrawerArrowToggle = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_DrawerArrowToggle;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_DropDownItem_Spinner = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_DropDownItem_Spinner;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_EditText = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_EditText;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_ImageButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ImageButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Light_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Light_ActionBar_Solid = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionBar_Solid;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Light_ActionBar_Solid_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionBar_Solid_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Light_ActionBar_TabBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionBar_TabBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Light_ActionBar_TabBar_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionBar_TabBar_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Light_ActionBar_TabText = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionBar_TabText;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Light_ActionBar_TabText_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionBar_TabText_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Light_ActionBar_TabView = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionBar_TabView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Light_ActionBar_TabView_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionBar_TabView_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Light_ActionButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Light_ActionButton_CloseMode = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionButton_CloseMode;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Light_ActionButton_Overflow = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionButton_Overflow;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Light_ActionMode_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionMode_Inverse;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Light_ActivityChooserView = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActivityChooserView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Light_AutoCompleteTextView = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_AutoCompleteTextView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Light_DropDownItem_Spinner = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_DropDownItem_Spinner;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Light_ListPopupWindow = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ListPopupWindow;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Light_ListView_DropDown = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ListView_DropDown;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Light_PopupMenu = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_PopupMenu;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Light_PopupMenu_Overflow = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_PopupMenu_Overflow;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Light_SearchView = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_SearchView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Light_Spinner_DropDown_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_Spinner_DropDown_ActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_ListMenuView = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ListMenuView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_ListPopupWindow = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ListPopupWindow;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_ListView = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ListView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_ListView_DropDown = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ListView_DropDown;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_ListView_Menu = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ListView_Menu;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_PopupMenu = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_PopupMenu;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_PopupMenu_Overflow = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_PopupMenu_Overflow;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_PopupWindow = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_PopupWindow;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_ProgressBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ProgressBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_ProgressBar_Horizontal = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ProgressBar_Horizontal;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_RatingBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_RatingBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_RatingBar_Indicator = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_RatingBar_Indicator;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_RatingBar_Small = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_RatingBar_Small;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_SearchView = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_SearchView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_SearchView_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_SearchView_ActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_SeekBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_SeekBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_SeekBar_Discrete = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_SeekBar_Discrete;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Spinner = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Spinner;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Spinner_DropDown = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Spinner_DropDown;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Spinner_DropDown_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Spinner_DropDown_ActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Spinner_Underlined = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Spinner_Underlined;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_TextView = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_TextView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_TextView_SpinnerItem = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_TextView_SpinnerItem;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Toolbar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Toolbar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_AppCompat_Toolbar_Button_Navigation = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Toolbar_Button_Navigation;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_Compat_NotificationActionContainer = global::pjsua2xamarin.Droid.Resource.Style.Widget_Compat_NotificationActionContainer;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_Compat_NotificationActionText = global::pjsua2xamarin.Droid.Resource.Style.Widget_Compat_NotificationActionText;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_Design_AppBarLayout = global::pjsua2xamarin.Droid.Resource.Style.Widget_Design_AppBarLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_Design_BottomNavigationView = global::pjsua2xamarin.Droid.Resource.Style.Widget_Design_BottomNavigationView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_Design_BottomSheet_Modal = global::pjsua2xamarin.Droid.Resource.Style.Widget_Design_BottomSheet_Modal;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_Design_CollapsingToolbar = global::pjsua2xamarin.Droid.Resource.Style.Widget_Design_CollapsingToolbar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_Design_FloatingActionButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_Design_FloatingActionButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_Design_NavigationView = global::pjsua2xamarin.Droid.Resource.Style.Widget_Design_NavigationView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_Design_ScrimInsetsFrameLayout = global::pjsua2xamarin.Droid.Resource.Style.Widget_Design_ScrimInsetsFrameLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_Design_Snackbar = global::pjsua2xamarin.Droid.Resource.Style.Widget_Design_Snackbar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_Design_TabLayout = global::pjsua2xamarin.Droid.Resource.Style.Widget_Design_TabLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_Design_TextInputEditText = global::pjsua2xamarin.Droid.Resource.Style.Widget_Design_TextInputEditText;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_Design_TextInputLayout = global::pjsua2xamarin.Droid.Resource.Style.Widget_Design_TextInputLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_ActionBar_Primary = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_ActionBar_Primary;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_ActionBar_PrimarySurface = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_ActionBar_PrimarySurface;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_ActionBar_Solid = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_ActionBar_Solid;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_ActionBar_Surface = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_ActionBar_Surface;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_AppBarLayout_Primary = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_AppBarLayout_Primary;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_AppBarLayout_PrimarySurface = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_AppBarLayout_PrimarySurface;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_AppBarLayout_Surface = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_AppBarLayout_Surface;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_AutoCompleteTextView_FilledBox = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_AutoCompleteTextView_FilledBox;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_AutoCompleteTextView_FilledBox_Dense = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_AutoCompleteTextView_FilledBox_Dense;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_AutoCompleteTextView_OutlinedBox = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_AutoCompleteTextView_OutlinedBox;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_AutoCompleteTextView_OutlinedBox_Dense = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_AutoCompleteTextView_OutlinedBox_Dense;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Badge = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Badge;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_BottomAppBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_BottomAppBar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_BottomAppBar_Colored = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_BottomAppBar_Colored;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_BottomAppBar_PrimarySurface = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_BottomAppBar_PrimarySurface;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_BottomNavigationView = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_BottomNavigationView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_BottomNavigationView_Colored = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_BottomNavigationView_Colored;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_BottomNavigationView_PrimarySurface = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_BottomNavigationView_PrimarySurface;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_BottomSheet = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_BottomSheet;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_BottomSheet_Modal = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_BottomSheet_Modal;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Button = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Button_Icon = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button_Icon;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Button_OutlinedButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button_OutlinedButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Button_OutlinedButton_Icon = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button_OutlinedButton_Icon;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Button_TextButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button_TextButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Button_TextButton_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button_TextButton_Dialog;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Button_TextButton_Dialog_Flush = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button_TextButton_Dialog_Flush;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Button_TextButton_Dialog_Icon = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button_TextButton_Dialog_Icon;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Button_TextButton_Icon = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button_TextButton_Icon;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Button_TextButton_Snackbar = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button_TextButton_Snackbar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Button_UnelevatedButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button_UnelevatedButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Button_UnelevatedButton_Icon = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button_UnelevatedButton_Icon;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_CardView = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_CardView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_CheckedTextView = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_CheckedTextView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_ChipGroup = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_ChipGroup;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Chip_Action = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Chip_Action;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Chip_Choice = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Chip_Choice;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Chip_Entry = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Chip_Entry;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Chip_Filter = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Chip_Filter;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_CircularProgressIndicator = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_CircularProgressIndicator;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_CircularProgressIndicator_ExtraSmall = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_CircularProgressIndicator_ExtraSmall;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_CircularProgressIndicator_Medium = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_CircularProgressIndicator_Medium;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_CircularProgressIndicator_Small = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_CircularProgressIndicator_Small;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_CollapsingToolbar = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_CollapsingToolbar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_CompoundButton_CheckBox = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_CompoundButton_CheckBox;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_CompoundButton_RadioButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_CompoundButton_RadioButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_CompoundButton_Switch = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_CompoundButton_Switch;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_ExtendedFloatingActionButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_ExtendedFloatingActionButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_ExtendedFloatingActionButton_Icon = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_ExtendedFloatingActionButton_Icon;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_FloatingActionButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_FloatingActionButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Light_ActionBar_Solid = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Light_ActionBar_Solid;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_LinearProgressIndicator = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_LinearProgressIndicator;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialButtonToggleGroup = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialButtonToggleGroup;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialCalendar = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Day = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Day;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialCalendar_DayTextView = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_DayTextView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Day_Invalid = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Day_Invalid;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Day_Selected = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Day_Selected;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Day_Today = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Day_Today;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Fullscreen = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Fullscreen;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderCancelButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderCancelButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderConfirmButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderConfirmButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderDivider = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderDivider;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderLayout = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderSelection = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderSelection;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderSelection_Fullscreen = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderSelection_Fullscreen;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderTitle = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderTitle;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderToggleButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderToggleButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Item = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Item;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialCalendar_MonthNavigationButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_MonthNavigationButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialCalendar_MonthTextView = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_MonthTextView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Year = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Year;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialCalendar_YearNavigationButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_YearNavigationButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Year_Selected = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Year_Selected;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Year_Today = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Year_Today;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_NavigationRailView = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_NavigationRailView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_NavigationRailView_Colored = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_NavigationRailView_Colored;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_NavigationRailView_Colored_Compact = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_NavigationRailView_Colored_Compact;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_NavigationRailView_Compact = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_NavigationRailView_Compact;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_NavigationRailView_PrimarySurface = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_NavigationRailView_PrimarySurface;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_NavigationView = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_NavigationView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_PopupMenu = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_PopupMenu;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_PopupMenu_ContextMenu = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_PopupMenu_ContextMenu;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_PopupMenu_ListPopupWindow = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_PopupMenu_ListPopupWindow;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_PopupMenu_Overflow = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_PopupMenu_Overflow;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_ProgressIndicator = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_ProgressIndicator;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_ShapeableImageView = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_ShapeableImageView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Slider = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Slider;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Snackbar = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Snackbar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Snackbar_FullWidth = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Snackbar_FullWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Snackbar_TextView = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Snackbar_TextView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TabLayout = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TabLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TabLayout_Colored = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TabLayout_Colored;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TabLayout_PrimarySurface = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TabLayout_PrimarySurface;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TextInputEditText_FilledBox = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputEditText_FilledBox;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TextInputEditText_FilledBox_Dense = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputEditText_FilledBox_Dense;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TextInputEditText_OutlinedBox = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputEditText_OutlinedBox;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TextInputEditText_OutlinedBox_Dense = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputEditText_OutlinedBox_Dense;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TextInputLayout_FilledBox = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputLayout_FilledBox;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TextInputLayout_FilledBox_Dense = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputLayout_FilledBox_Dense;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TextInputLayout_FilledBox_Dense_ExposedDropdownMenu = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputLayout_FilledBox_Dense_ExposedDropdownMenu;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TextInputLayout_FilledBox_ExposedDropdownMenu = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputLayout_FilledBox_ExposedDropdownMenu;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TextInputLayout_OutlinedBox = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputLayout_OutlinedBox;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TextInputLayout_OutlinedBox_Dense = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputLayout_OutlinedBox_Dense;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TextInputLayout_OutlinedBox_Dense_ExposedDropdownMenu = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputLayout_OutlinedBox_Dense_ExposedDropdownMenu;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TextInputLayout_OutlinedBox_ExposedDropdownMenu = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputLayout_OutlinedBox_ExposedDropdownMenu;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TextView = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextView;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TimePicker = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TimePicker;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TimePicker_Button = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TimePicker_Button;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TimePicker_Clock = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TimePicker_Clock;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TimePicker_Display = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TimePicker_Display;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TimePicker_Display_TextInputEditText = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TimePicker_Display_TextInputEditText;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TimePicker_ImageButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TimePicker_ImageButton;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_TimePicker_ImageButton_ShapeAppearance = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TimePicker_ImageButton_ShapeAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Toolbar = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Toolbar;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Toolbar_Primary = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Toolbar_Primary;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Toolbar_PrimarySurface = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Toolbar_PrimarySurface;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Toolbar_Surface = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Toolbar_Surface;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_MaterialComponents_Tooltip = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Tooltip;
			global::Xamarin.Forms.Platform.Android.Resource.Style.Widget_Support_CoordinatorLayout = global::pjsua2xamarin.Droid.Resource.Style.Widget_Support_CoordinatorLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBarLayout = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBarLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBarLayout_android_layout_gravity = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBarLayout_android_layout_gravity;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_background = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_background;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_backgroundSplit = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_backgroundSplit;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_backgroundStacked = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_backgroundStacked;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_contentInsetEnd = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_contentInsetEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_contentInsetEndWithActions = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_contentInsetEndWithActions;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_contentInsetLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_contentInsetLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_contentInsetRight = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_contentInsetRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_contentInsetStart = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_contentInsetStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_contentInsetStartWithNavigation = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_contentInsetStartWithNavigation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_customNavigationLayout = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_customNavigationLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_displayOptions = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_displayOptions;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_divider = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_divider;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_height = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_height;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_hideOnContentScroll = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_hideOnContentScroll;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_homeAsUpIndicator = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_homeAsUpIndicator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_homeLayout = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_homeLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_icon = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_icon;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_indeterminateProgressStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_indeterminateProgressStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_itemPadding = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_itemPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_logo = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_logo;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_navigationMode = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_navigationMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_popupTheme = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_popupTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_progressBarPadding = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_progressBarPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_progressBarStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_progressBarStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_subtitle = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_subtitle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_subtitleTextStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_subtitleTextStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_title = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_title;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionBar_titleTextStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_titleTextStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionMenuItemView = global::pjsua2xamarin.Droid.Resource.Styleable.ActionMenuItemView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionMenuItemView_android_minWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ActionMenuItemView_android_minWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionMenuView = global::pjsua2xamarin.Droid.Resource.Styleable.ActionMenuView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionMode = global::pjsua2xamarin.Droid.Resource.Styleable.ActionMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionMode_background = global::pjsua2xamarin.Droid.Resource.Styleable.ActionMode_background;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionMode_backgroundSplit = global::pjsua2xamarin.Droid.Resource.Styleable.ActionMode_backgroundSplit;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionMode_closeItemLayout = global::pjsua2xamarin.Droid.Resource.Styleable.ActionMode_closeItemLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionMode_height = global::pjsua2xamarin.Droid.Resource.Styleable.ActionMode_height;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionMode_subtitleTextStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ActionMode_subtitleTextStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActionMode_titleTextStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ActionMode_titleTextStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActivityChooserView = global::pjsua2xamarin.Droid.Resource.Styleable.ActivityChooserView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActivityChooserView_expandActivityOverflowButtonDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.ActivityChooserView_expandActivityOverflowButtonDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActivityChooserView_initialActivityCount = global::pjsua2xamarin.Droid.Resource.Styleable.ActivityChooserView_initialActivityCount;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActivityNavigator = global::pjsua2xamarin.Droid.Resource.Styleable.ActivityNavigator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActivityNavigator_action = global::pjsua2xamarin.Droid.Resource.Styleable.ActivityNavigator_action;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActivityNavigator_android_name = global::pjsua2xamarin.Droid.Resource.Styleable.ActivityNavigator_android_name;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActivityNavigator_data = global::pjsua2xamarin.Droid.Resource.Styleable.ActivityNavigator_data;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActivityNavigator_dataPattern = global::pjsua2xamarin.Droid.Resource.Styleable.ActivityNavigator_dataPattern;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ActivityNavigator_targetPackage = global::pjsua2xamarin.Droid.Resource.Styleable.ActivityNavigator_targetPackage;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AlertDialog = global::pjsua2xamarin.Droid.Resource.Styleable.AlertDialog;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AlertDialog_android_layout = global::pjsua2xamarin.Droid.Resource.Styleable.AlertDialog_android_layout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AlertDialog_buttonIconDimen = global::pjsua2xamarin.Droid.Resource.Styleable.AlertDialog_buttonIconDimen;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AlertDialog_buttonPanelSideLayout = global::pjsua2xamarin.Droid.Resource.Styleable.AlertDialog_buttonPanelSideLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AlertDialog_listItemLayout = global::pjsua2xamarin.Droid.Resource.Styleable.AlertDialog_listItemLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AlertDialog_listLayout = global::pjsua2xamarin.Droid.Resource.Styleable.AlertDialog_listLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AlertDialog_multiChoiceItemLayout = global::pjsua2xamarin.Droid.Resource.Styleable.AlertDialog_multiChoiceItemLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AlertDialog_showTitle = global::pjsua2xamarin.Droid.Resource.Styleable.AlertDialog_showTitle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AlertDialog_singleChoiceItemLayout = global::pjsua2xamarin.Droid.Resource.Styleable.AlertDialog_singleChoiceItemLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AnimatedStateListDrawableCompat = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AnimatedStateListDrawableCompat_android_constantSize = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableCompat_android_constantSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AnimatedStateListDrawableCompat_android_dither = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableCompat_android_dither;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AnimatedStateListDrawableCompat_android_enterFadeDuration = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableCompat_android_enterFadeDuration;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AnimatedStateListDrawableCompat_android_exitFadeDuration = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableCompat_android_exitFadeDuration;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AnimatedStateListDrawableCompat_android_variablePadding = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableCompat_android_variablePadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AnimatedStateListDrawableCompat_android_visible = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableCompat_android_visible;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AnimatedStateListDrawableItem = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableItem;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AnimatedStateListDrawableItem_android_drawable = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableItem_android_drawable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AnimatedStateListDrawableItem_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableItem_android_id;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AnimatedStateListDrawableTransition = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableTransition;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AnimatedStateListDrawableTransition_android_drawable = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableTransition_android_drawable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AnimatedStateListDrawableTransition_android_fromId = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableTransition_android_fromId;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AnimatedStateListDrawableTransition_android_reversible = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableTransition_android_reversible;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AnimatedStateListDrawableTransition_android_toId = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableTransition_android_toId;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppBarLayout = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppBarLayoutStates = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayoutStates;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppBarLayoutStates_state_collapsed = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayoutStates_state_collapsed;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppBarLayoutStates_state_collapsible = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayoutStates_state_collapsible;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppBarLayoutStates_state_liftable = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayoutStates_state_liftable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppBarLayoutStates_state_lifted = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayoutStates_state_lifted;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppBarLayout_android_background = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout_android_background;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppBarLayout_android_keyboardNavigationCluster = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout_android_keyboardNavigationCluster;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppBarLayout_android_touchscreenBlocksFocus = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout_android_touchscreenBlocksFocus;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppBarLayout_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppBarLayout_expanded = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout_expanded;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppBarLayout_Layout = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout_Layout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppBarLayout_Layout_layout_scrollFlags = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout_Layout_layout_scrollFlags;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppBarLayout_Layout_layout_scrollInterpolator = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout_Layout_layout_scrollInterpolator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppBarLayout_liftOnScroll = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout_liftOnScroll;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppBarLayout_liftOnScrollTargetViewId = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout_liftOnScrollTargetViewId;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppBarLayout_statusBarForeground = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout_statusBarForeground;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatImageView = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatImageView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatImageView_android_src = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatImageView_android_src;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatImageView_srcCompat = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatImageView_srcCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatImageView_tint = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatImageView_tint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatImageView_tintMode = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatImageView_tintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatSeekBar = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatSeekBar;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatSeekBar_android_thumb = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatSeekBar_android_thumb;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatSeekBar_tickMark = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatSeekBar_tickMark;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatSeekBar_tickMarkTint = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatSeekBar_tickMarkTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatSeekBar_tickMarkTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatSeekBar_tickMarkTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextHelper = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextHelper;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextHelper_android_drawableBottom = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextHelper_android_drawableBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextHelper_android_drawableEnd = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextHelper_android_drawableEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextHelper_android_drawableLeft = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextHelper_android_drawableLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextHelper_android_drawableRight = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextHelper_android_drawableRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextHelper_android_drawableStart = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextHelper_android_drawableStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextHelper_android_drawableTop = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextHelper_android_drawableTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextHelper_android_textAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextHelper_android_textAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextView = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextView_android_textAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_android_textAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextView_autoSizeMaxTextSize = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_autoSizeMaxTextSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextView_autoSizeMinTextSize = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_autoSizeMinTextSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextView_autoSizePresetSizes = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_autoSizePresetSizes;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextView_autoSizeStepGranularity = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_autoSizeStepGranularity;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextView_autoSizeTextType = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_autoSizeTextType;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextView_drawableBottomCompat = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_drawableBottomCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextView_drawableEndCompat = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_drawableEndCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextView_drawableLeftCompat = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_drawableLeftCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextView_drawableRightCompat = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_drawableRightCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextView_drawableStartCompat = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_drawableStartCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextView_drawableTint = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_drawableTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextView_drawableTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_drawableTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextView_drawableTopCompat = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_drawableTopCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextView_firstBaselineToTopHeight = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_firstBaselineToTopHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextView_fontFamily = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_fontFamily;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextView_fontVariationSettings = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_fontVariationSettings;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextView_lastBaselineToBottomHeight = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_lastBaselineToBottomHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextView_lineHeight = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_lineHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextView_textAllCaps = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_textAllCaps;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTextView_textLocale = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_textLocale;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionBarDivider = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionBarDivider;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionBarItemBackground = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionBarItemBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionBarPopupTheme = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionBarPopupTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionBarSize = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionBarSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionBarSplitStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionBarSplitStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionBarStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionBarStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionBarTabBarStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionBarTabBarStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionBarTabStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionBarTabStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionBarTabTextStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionBarTabTextStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionBarTheme = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionBarTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionBarWidgetTheme = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionBarWidgetTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionDropDownStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionDropDownStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionMenuTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionMenuTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionMenuTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionMenuTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionModeBackground = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionModeCloseButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeCloseButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionModeCloseContentDescription = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeCloseContentDescription;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionModeCloseDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeCloseDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionModeCopyDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeCopyDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionModeCutDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeCutDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionModeFindDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeFindDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionModePasteDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModePasteDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionModePopupWindowStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModePopupWindowStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionModeSelectAllDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeSelectAllDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionModeShareDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeShareDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionModeSplitBackground = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeSplitBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionModeStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionModeTheme = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionModeWebSearchDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeWebSearchDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionOverflowButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionOverflowButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_actionOverflowMenuStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionOverflowMenuStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_activityChooserViewStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_activityChooserViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_alertDialogButtonGroupStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_alertDialogButtonGroupStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_alertDialogCenterButtons = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_alertDialogCenterButtons;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_alertDialogStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_alertDialogStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_alertDialogTheme = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_alertDialogTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_android_windowAnimationStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_android_windowAnimationStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_android_windowIsFloating = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_android_windowIsFloating;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_autoCompleteTextViewStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_autoCompleteTextViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_borderlessButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_borderlessButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_buttonBarButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_buttonBarButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_buttonBarNegativeButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_buttonBarNegativeButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_buttonBarNeutralButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_buttonBarNeutralButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_buttonBarPositiveButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_buttonBarPositiveButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_buttonBarStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_buttonBarStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_buttonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_buttonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_buttonStyleSmall = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_buttonStyleSmall;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_checkboxStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_checkboxStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_checkedTextViewStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_checkedTextViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_colorAccent = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_colorAccent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_colorBackgroundFloating = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_colorBackgroundFloating;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_colorButtonNormal = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_colorButtonNormal;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_colorControlActivated = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_colorControlActivated;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_colorControlHighlight = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_colorControlHighlight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_colorControlNormal = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_colorControlNormal;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_colorError = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_colorError;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_colorPrimary = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_colorPrimary;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_colorPrimaryDark = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_colorPrimaryDark;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_colorSwitchThumbNormal = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_colorSwitchThumbNormal;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_controlBackground = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_controlBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_dialogCornerRadius = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_dialogCornerRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_dialogPreferredPadding = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_dialogPreferredPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_dialogTheme = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_dialogTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_dividerHorizontal = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_dividerHorizontal;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_dividerVertical = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_dividerVertical;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_dropdownListPreferredItemHeight = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_dropdownListPreferredItemHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_dropDownListViewStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_dropDownListViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_editTextBackground = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_editTextBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_editTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_editTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_editTextStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_editTextStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_homeAsUpIndicator = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_homeAsUpIndicator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_imageButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_imageButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_listChoiceBackgroundIndicator = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listChoiceBackgroundIndicator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_listChoiceIndicatorMultipleAnimated = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listChoiceIndicatorMultipleAnimated;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_listChoiceIndicatorSingleAnimated = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listChoiceIndicatorSingleAnimated;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_listDividerAlertDialog = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listDividerAlertDialog;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_listMenuViewStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listMenuViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_listPopupWindowStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listPopupWindowStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_listPreferredItemHeight = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listPreferredItemHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_listPreferredItemHeightLarge = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listPreferredItemHeightLarge;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_listPreferredItemHeightSmall = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listPreferredItemHeightSmall;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_listPreferredItemPaddingEnd = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listPreferredItemPaddingEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_listPreferredItemPaddingLeft = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listPreferredItemPaddingLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_listPreferredItemPaddingRight = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listPreferredItemPaddingRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_listPreferredItemPaddingStart = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listPreferredItemPaddingStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_panelBackground = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_panelBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_panelMenuListTheme = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_panelMenuListTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_panelMenuListWidth = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_panelMenuListWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_popupMenuStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_popupMenuStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_popupWindowStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_popupWindowStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_radioButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_radioButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_ratingBarStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_ratingBarStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_ratingBarStyleIndicator = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_ratingBarStyleIndicator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_ratingBarStyleSmall = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_ratingBarStyleSmall;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_searchViewStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_searchViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_seekBarStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_seekBarStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_selectableItemBackground = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_selectableItemBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_selectableItemBackgroundBorderless = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_selectableItemBackgroundBorderless;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_spinnerDropDownItemStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_spinnerDropDownItemStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_spinnerStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_spinnerStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_switchStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_switchStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_textAppearanceLargePopupMenu = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_textAppearanceLargePopupMenu;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_textAppearanceListItem = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_textAppearanceListItem;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_textAppearanceListItemSecondary = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_textAppearanceListItemSecondary;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_textAppearanceListItemSmall = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_textAppearanceListItemSmall;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_textAppearancePopupMenuHeader = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_textAppearancePopupMenuHeader;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_textAppearanceSearchResultSubtitle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_textAppearanceSearchResultSubtitle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_textAppearanceSearchResultTitle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_textAppearanceSearchResultTitle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_textAppearanceSmallPopupMenu = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_textAppearanceSmallPopupMenu;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_textColorAlertDialogListItem = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_textColorAlertDialogListItem;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_textColorSearchUrl = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_textColorSearchUrl;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_toolbarNavigationButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_toolbarNavigationButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_toolbarStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_toolbarStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_tooltipForegroundColor = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_tooltipForegroundColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_tooltipFrameBackground = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_tooltipFrameBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_viewInflaterClass = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_viewInflaterClass;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_windowActionBar = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_windowActionBar;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_windowActionBarOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_windowActionBarOverlay;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_windowActionModeOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_windowActionModeOverlay;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_windowFixedHeightMajor = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_windowFixedHeightMajor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_windowFixedHeightMinor = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_windowFixedHeightMinor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_windowFixedWidthMajor = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_windowFixedWidthMajor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_windowFixedWidthMinor = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_windowFixedWidthMinor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_windowMinWidthMajor = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_windowMinWidthMajor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_windowMinWidthMinor = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_windowMinWidthMinor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.AppCompatTheme_windowNoTitle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_windowNoTitle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Badge = global::pjsua2xamarin.Droid.Resource.Styleable.Badge;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Badge_backgroundColor = global::pjsua2xamarin.Droid.Resource.Styleable.Badge_backgroundColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Badge_badgeGravity = global::pjsua2xamarin.Droid.Resource.Styleable.Badge_badgeGravity;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Badge_badgeTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.Badge_badgeTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Badge_horizontalOffset = global::pjsua2xamarin.Droid.Resource.Styleable.Badge_horizontalOffset;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Badge_maxCharacterCount = global::pjsua2xamarin.Droid.Resource.Styleable.Badge_maxCharacterCount;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Badge_number = global::pjsua2xamarin.Droid.Resource.Styleable.Badge_number;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Badge_verticalOffset = global::pjsua2xamarin.Droid.Resource.Styleable.Badge_verticalOffset;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BaseProgressIndicator = global::pjsua2xamarin.Droid.Resource.Styleable.BaseProgressIndicator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BaseProgressIndicator_android_indeterminate = global::pjsua2xamarin.Droid.Resource.Styleable.BaseProgressIndicator_android_indeterminate;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BaseProgressIndicator_hideAnimationBehavior = global::pjsua2xamarin.Droid.Resource.Styleable.BaseProgressIndicator_hideAnimationBehavior;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BaseProgressIndicator_indicatorColor = global::pjsua2xamarin.Droid.Resource.Styleable.BaseProgressIndicator_indicatorColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BaseProgressIndicator_minHideDelay = global::pjsua2xamarin.Droid.Resource.Styleable.BaseProgressIndicator_minHideDelay;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BaseProgressIndicator_showAnimationBehavior = global::pjsua2xamarin.Droid.Resource.Styleable.BaseProgressIndicator_showAnimationBehavior;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BaseProgressIndicator_showDelay = global::pjsua2xamarin.Droid.Resource.Styleable.BaseProgressIndicator_showDelay;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BaseProgressIndicator_trackColor = global::pjsua2xamarin.Droid.Resource.Styleable.BaseProgressIndicator_trackColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BaseProgressIndicator_trackCornerRadius = global::pjsua2xamarin.Droid.Resource.Styleable.BaseProgressIndicator_trackCornerRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BaseProgressIndicator_trackThickness = global::pjsua2xamarin.Droid.Resource.Styleable.BaseProgressIndicator_trackThickness;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomAppBar = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomAppBar_backgroundTint = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar_backgroundTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomAppBar_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomAppBar_fabAlignmentMode = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar_fabAlignmentMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomAppBar_fabAnimationMode = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar_fabAnimationMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomAppBar_fabCradleMargin = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar_fabCradleMargin;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomAppBar_fabCradleRoundedCornerRadius = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar_fabCradleRoundedCornerRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomAppBar_fabCradleVerticalOffset = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar_fabCradleVerticalOffset;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomAppBar_hideOnScroll = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar_hideOnScroll;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomAppBar_paddingBottomSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar_paddingBottomSystemWindowInsets;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomAppBar_paddingLeftSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar_paddingLeftSystemWindowInsets;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomAppBar_paddingRightSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar_paddingRightSystemWindowInsets;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomNavigationView = global::pjsua2xamarin.Droid.Resource.Styleable.BottomNavigationView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomNavigationView_itemHorizontalTranslationEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.BottomNavigationView_itemHorizontalTranslationEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomSheetBehavior_Layout = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomSheetBehavior_Layout_android_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_android_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomSheetBehavior_Layout_android_maxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_android_maxWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomSheetBehavior_Layout_backgroundTint = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_backgroundTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomSheetBehavior_Layout_behavior_draggable = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_behavior_draggable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomSheetBehavior_Layout_behavior_expandedOffset = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_behavior_expandedOffset;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomSheetBehavior_Layout_behavior_fitToContents = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_behavior_fitToContents;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomSheetBehavior_Layout_behavior_halfExpandedRatio = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_behavior_halfExpandedRatio;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomSheetBehavior_Layout_behavior_hideable = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_behavior_hideable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomSheetBehavior_Layout_behavior_peekHeight = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_behavior_peekHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomSheetBehavior_Layout_behavior_saveFlags = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_behavior_saveFlags;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomSheetBehavior_Layout_behavior_skipCollapsed = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_behavior_skipCollapsed;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomSheetBehavior_Layout_gestureInsetBottomIgnored = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_gestureInsetBottomIgnored;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomSheetBehavior_Layout_paddingBottomSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_paddingBottomSystemWindowInsets;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomSheetBehavior_Layout_paddingLeftSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_paddingLeftSystemWindowInsets;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomSheetBehavior_Layout_paddingRightSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_paddingRightSystemWindowInsets;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomSheetBehavior_Layout_paddingTopSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_paddingTopSystemWindowInsets;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomSheetBehavior_Layout_shapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_shapeAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.BottomSheetBehavior_Layout_shapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_shapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ButtonBarLayout = global::pjsua2xamarin.Droid.Resource.Styleable.ButtonBarLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ButtonBarLayout_allowStacking = global::pjsua2xamarin.Droid.Resource.Styleable.ButtonBarLayout_allowStacking;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Capability = global::pjsua2xamarin.Droid.Resource.Styleable.Capability;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Capability_queryPatterns = global::pjsua2xamarin.Droid.Resource.Styleable.Capability_queryPatterns;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Capability_shortcutMatchRequired = global::pjsua2xamarin.Droid.Resource.Styleable.Capability_shortcutMatchRequired;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CardView = global::pjsua2xamarin.Droid.Resource.Styleable.CardView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CardView_android_minHeight = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_android_minHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CardView_android_minWidth = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_android_minWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CardView_cardBackgroundColor = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_cardBackgroundColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CardView_cardCornerRadius = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_cardCornerRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CardView_cardElevation = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_cardElevation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CardView_cardMaxElevation = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_cardMaxElevation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CardView_cardPreventCornerOverlap = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_cardPreventCornerOverlap;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CardView_cardUseCompatPadding = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_cardUseCompatPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CardView_contentPadding = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_contentPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CardView_contentPaddingBottom = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_contentPaddingBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CardView_contentPaddingLeft = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_contentPaddingLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CardView_contentPaddingRight = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_contentPaddingRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CardView_contentPaddingTop = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_contentPaddingTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Carousel = global::pjsua2xamarin.Droid.Resource.Styleable.Carousel;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Carousel_carousel_backwardTransition = global::pjsua2xamarin.Droid.Resource.Styleable.Carousel_carousel_backwardTransition;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Carousel_carousel_emptyViewsBehavior = global::pjsua2xamarin.Droid.Resource.Styleable.Carousel_carousel_emptyViewsBehavior;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Carousel_carousel_firstView = global::pjsua2xamarin.Droid.Resource.Styleable.Carousel_carousel_firstView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Carousel_carousel_forwardTransition = global::pjsua2xamarin.Droid.Resource.Styleable.Carousel_carousel_forwardTransition;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Carousel_carousel_infinite = global::pjsua2xamarin.Droid.Resource.Styleable.Carousel_carousel_infinite;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Carousel_carousel_nextState = global::pjsua2xamarin.Droid.Resource.Styleable.Carousel_carousel_nextState;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Carousel_carousel_previousState = global::pjsua2xamarin.Droid.Resource.Styleable.Carousel_carousel_previousState;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Carousel_carousel_touchUpMode = global::pjsua2xamarin.Droid.Resource.Styleable.Carousel_carousel_touchUpMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Carousel_carousel_touchUp_dampeningFactor = global::pjsua2xamarin.Droid.Resource.Styleable.Carousel_carousel_touchUp_dampeningFactor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Carousel_carousel_touchUp_velocityThreshold = global::pjsua2xamarin.Droid.Resource.Styleable.Carousel_carousel_touchUp_velocityThreshold;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip = global::pjsua2xamarin.Droid.Resource.Styleable.Chip;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ChipGroup = global::pjsua2xamarin.Droid.Resource.Styleable.ChipGroup;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ChipGroup_checkedChip = global::pjsua2xamarin.Droid.Resource.Styleable.ChipGroup_checkedChip;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ChipGroup_chipSpacing = global::pjsua2xamarin.Droid.Resource.Styleable.ChipGroup_chipSpacing;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ChipGroup_chipSpacingHorizontal = global::pjsua2xamarin.Droid.Resource.Styleable.ChipGroup_chipSpacingHorizontal;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ChipGroup_chipSpacingVertical = global::pjsua2xamarin.Droid.Resource.Styleable.ChipGroup_chipSpacingVertical;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ChipGroup_selectionRequired = global::pjsua2xamarin.Droid.Resource.Styleable.ChipGroup_selectionRequired;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ChipGroup_singleLine = global::pjsua2xamarin.Droid.Resource.Styleable.ChipGroup_singleLine;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ChipGroup_singleSelection = global::pjsua2xamarin.Droid.Resource.Styleable.ChipGroup_singleSelection;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_android_checkable = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_android_checkable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_android_ellipsize = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_android_ellipsize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_android_maxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_android_maxWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_android_text = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_android_text;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_android_textAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_android_textAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_android_textColor = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_android_textColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_android_textSize = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_android_textSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_checkedIcon = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_checkedIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_checkedIconEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_checkedIconEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_checkedIconTint = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_checkedIconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_checkedIconVisible = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_checkedIconVisible;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_chipBackgroundColor = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipBackgroundColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_chipCornerRadius = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipCornerRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_chipEndPadding = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipEndPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_chipIcon = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_chipIconEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipIconEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_chipIconSize = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipIconSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_chipIconTint = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipIconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_chipIconVisible = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipIconVisible;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_chipMinHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipMinHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_chipMinTouchTargetSize = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipMinTouchTargetSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_chipStartPadding = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipStartPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_chipStrokeColor = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipStrokeColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_chipStrokeWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipStrokeWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_chipSurfaceColor = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipSurfaceColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_closeIcon = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_closeIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_closeIconEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_closeIconEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_closeIconEndPadding = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_closeIconEndPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_closeIconSize = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_closeIconSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_closeIconStartPadding = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_closeIconStartPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_closeIconTint = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_closeIconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_closeIconVisible = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_closeIconVisible;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_ensureMinTouchTargetSize = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_ensureMinTouchTargetSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_hideMotionSpec = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_hideMotionSpec;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_iconEndPadding = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_iconEndPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_iconStartPadding = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_iconStartPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_rippleColor = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_rippleColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_shapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_shapeAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_shapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_shapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_showMotionSpec = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_showMotionSpec;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_textEndPadding = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_textEndPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Chip_textStartPadding = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_textStartPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CircularProgressIndicator = global::pjsua2xamarin.Droid.Resource.Styleable.CircularProgressIndicator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CircularProgressIndicator_indicatorDirectionCircular = global::pjsua2xamarin.Droid.Resource.Styleable.CircularProgressIndicator_indicatorDirectionCircular;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CircularProgressIndicator_indicatorInset = global::pjsua2xamarin.Droid.Resource.Styleable.CircularProgressIndicator_indicatorInset;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CircularProgressIndicator_indicatorSize = global::pjsua2xamarin.Droid.Resource.Styleable.CircularProgressIndicator_indicatorSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ClockFaceView = global::pjsua2xamarin.Droid.Resource.Styleable.ClockFaceView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ClockFaceView_clockFaceBackgroundColor = global::pjsua2xamarin.Droid.Resource.Styleable.ClockFaceView_clockFaceBackgroundColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ClockFaceView_clockNumberTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.ClockFaceView_clockNumberTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ClockHandView = global::pjsua2xamarin.Droid.Resource.Styleable.ClockHandView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ClockHandView_clockHandColor = global::pjsua2xamarin.Droid.Resource.Styleable.ClockHandView_clockHandColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ClockHandView_materialCircleRadius = global::pjsua2xamarin.Droid.Resource.Styleable.ClockHandView_materialCircleRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ClockHandView_selectorSize = global::pjsua2xamarin.Droid.Resource.Styleable.ClockHandView_selectorSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_collapsedTitleGravity = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_collapsedTitleGravity;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_collapsedTitleTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_collapsedTitleTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_contentScrim = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_contentScrim;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_expandedTitleGravity = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_expandedTitleGravity;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_expandedTitleMargin = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_expandedTitleMargin;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_expandedTitleMarginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_expandedTitleMarginBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_expandedTitleMarginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_expandedTitleMarginEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_expandedTitleMarginStart = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_expandedTitleMarginStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_expandedTitleMarginTop = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_expandedTitleMarginTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_expandedTitleTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_expandedTitleTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_extraMultilineHeightEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_extraMultilineHeightEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_forceApplySystemWindowInsetTop = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_forceApplySystemWindowInsetTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_Layout = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_Layout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_Layout_layout_collapseMode = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_Layout_layout_collapseMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_Layout_layout_collapseParallaxMultiplier = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_Layout_layout_collapseParallaxMultiplier;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_maxLines = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_maxLines;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_scrimAnimationDuration = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_scrimAnimationDuration;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_scrimVisibleHeightTrigger = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_scrimVisibleHeightTrigger;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_statusBarScrim = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_statusBarScrim;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_title = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_title;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_titleCollapseMode = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_titleCollapseMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_titleEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_titleEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CollapsingToolbarLayout_toolbarId = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_toolbarId;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ColorStateListItem = global::pjsua2xamarin.Droid.Resource.Styleable.ColorStateListItem;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ColorStateListItem_alpha = global::pjsua2xamarin.Droid.Resource.Styleable.ColorStateListItem_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ColorStateListItem_android_alpha = global::pjsua2xamarin.Droid.Resource.Styleable.ColorStateListItem_android_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ColorStateListItem_android_color = global::pjsua2xamarin.Droid.Resource.Styleable.ColorStateListItem_android_color;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CompoundButton = global::pjsua2xamarin.Droid.Resource.Styleable.CompoundButton;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CompoundButton_android_button = global::pjsua2xamarin.Droid.Resource.Styleable.CompoundButton_android_button;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CompoundButton_buttonCompat = global::pjsua2xamarin.Droid.Resource.Styleable.CompoundButton_buttonCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CompoundButton_buttonTint = global::pjsua2xamarin.Droid.Resource.Styleable.CompoundButton_buttonTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CompoundButton_buttonTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.CompoundButton_buttonTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_layout_height = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_layout_height;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_layout_margin = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_layout_margin;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginHorizontal = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginHorizontal;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginRight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginStart = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginTop = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginVertical = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginVertical;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_layout_width = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_layout_width;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_maxHeight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_maxHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_maxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_maxWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_minHeight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_minHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_minWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_minWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_orientation = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_orientation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_padding = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_paddingBottom = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_paddingBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_paddingEnd = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_paddingEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_paddingLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_paddingLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_paddingRight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_paddingRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_paddingStart = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_paddingStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_paddingTop = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_paddingTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_android_visibility = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_visibility;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_barrierAllowsGoneWidgets = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_barrierAllowsGoneWidgets;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_barrierDirection = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_barrierDirection;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_barrierMargin = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_barrierMargin;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_chainUseRtl = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_chainUseRtl;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_circularflow_angles = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_circularflow_angles;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_circularflow_defaultAngle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_circularflow_defaultAngle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_circularflow_defaultRadius = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_circularflow_defaultRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_circularflow_radiusInDP = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_circularflow_radiusInDP;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_circularflow_viewCenter = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_circularflow_viewCenter;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_constraintSet = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_constraintSet;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_constraint_referenced_ids = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_constraint_referenced_ids;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_constraint_referenced_tags = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_constraint_referenced_tags;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_flow_firstHorizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_firstHorizontalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_flow_firstHorizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_firstHorizontalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_flow_firstVerticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_firstVerticalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_flow_firstVerticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_firstVerticalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_flow_horizontalAlign = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_horizontalAlign;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_flow_horizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_horizontalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_flow_horizontalGap = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_horizontalGap;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_flow_horizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_horizontalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_flow_lastHorizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_lastHorizontalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_flow_lastHorizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_lastHorizontalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_flow_lastVerticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_lastVerticalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_flow_lastVerticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_lastVerticalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_flow_maxElementsWrap = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_maxElementsWrap;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_flow_verticalAlign = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_verticalAlign;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_flow_verticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_verticalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_flow_verticalGap = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_verticalGap;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_flow_verticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_verticalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_flow_wrapMode = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_wrapMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layoutDescription = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layoutDescription;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constrainedHeight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constrainedHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constrainedWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constrainedWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBaseline_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBaseline_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBaseline_toBaselineOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBaseline_toBaselineOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBaseline_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBaseline_toBottomOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBaseline_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBaseline_toTopOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBottom_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBottom_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBottom_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBottom_toBottomOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBottom_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBottom_toTopOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintCircle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintCircle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintCircleAngle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintCircleAngle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintCircleRadius = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintCircleRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintDimensionRatio = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintDimensionRatio;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintEnd_toEndOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintEnd_toEndOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintEnd_toStartOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintEnd_toStartOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintGuide_begin = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintGuide_begin;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintGuide_end = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintGuide_end;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintGuide_percent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintGuide_percent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHeight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHeight_default = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHeight_default;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHeight_max = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHeight_max;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHeight_min = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHeight_min;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHeight_percent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHeight_percent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHorizontal_bias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHorizontal_bias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHorizontal_chainStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHorizontal_chainStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHorizontal_weight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHorizontal_weight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintLeft_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintLeft_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintLeft_toLeftOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintLeft_toLeftOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintLeft_toRightOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintLeft_toRightOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintRight_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintRight_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintRight_toLeftOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintRight_toLeftOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintRight_toRightOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintRight_toRightOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintStart_toEndOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintStart_toEndOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintStart_toStartOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintStart_toStartOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintTag = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintTag;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintTop_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintTop_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintTop_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintTop_toBottomOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintTop_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintTop_toTopOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintVertical_bias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintVertical_bias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintVertical_chainStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintVertical_chainStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintVertical_weight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintVertical_weight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintWidth_default = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintWidth_default;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintWidth_max = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintWidth_max;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintWidth_min = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintWidth_min;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_constraintWidth_percent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintWidth_percent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_editor_absoluteX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_editor_absoluteX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_editor_absoluteY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_editor_absoluteY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginBaseline = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginBaseline;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginRight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginStart = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginTop = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_marginBaseline = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_marginBaseline;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_optimizationLevel = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_optimizationLevel;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_Layout_layout_wrapBehaviorInParent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_wrapBehaviorInParent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_placeholder = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_placeholder;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_placeholder_content = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_placeholder_content;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_placeholder_placeholder_emptyVisibility = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_placeholder_placeholder_emptyVisibility;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_ReactiveGuide = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_ReactiveGuide;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_ReactiveGuide_reactiveGuide_animateChange = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_ReactiveGuide_reactiveGuide_animateChange;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_ReactiveGuide_reactiveGuide_applyToAllConstraintSets = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_ReactiveGuide_reactiveGuide_applyToAllConstraintSets;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_ReactiveGuide_reactiveGuide_applyToConstraintSet = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_ReactiveGuide_reactiveGuide_applyToConstraintSet;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintLayout_ReactiveGuide_reactiveGuide_valueId = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_ReactiveGuide_reactiveGuide_valueId;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_alpha = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_id;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_layout_height = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_layout_height;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_layout_marginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_layout_marginBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_layout_marginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_layout_marginEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_layout_marginLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_layout_marginLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_layout_marginRight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_layout_marginRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_layout_marginStart = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_layout_marginStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_layout_marginTop = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_layout_marginTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_layout_width = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_layout_width;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_maxHeight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_maxHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_maxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_maxWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_minHeight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_minHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_minWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_minWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_orientation = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_orientation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_rotation = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_rotation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_rotationX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_rotationX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_rotationY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_rotationY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_scaleX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_scaleX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_scaleY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_scaleY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_transformPivotX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_transformPivotX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_transformPivotY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_transformPivotY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_translationX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_translationX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_translationY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_translationY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_translationZ = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_translationZ;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_android_visibility = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_visibility;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_animateCircleAngleTo = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_animateCircleAngleTo;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_animateRelativeTo = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_animateRelativeTo;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_barrierAllowsGoneWidgets = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_barrierAllowsGoneWidgets;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_barrierDirection = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_barrierDirection;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_barrierMargin = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_barrierMargin;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_chainUseRtl = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_chainUseRtl;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_constraint_referenced_ids = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_constraint_referenced_ids;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_drawPath = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_drawPath;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_flow_firstHorizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_firstHorizontalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_flow_firstHorizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_firstHorizontalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_flow_firstVerticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_firstVerticalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_flow_firstVerticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_firstVerticalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_flow_horizontalAlign = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_horizontalAlign;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_flow_horizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_horizontalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_flow_horizontalGap = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_horizontalGap;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_flow_horizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_horizontalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_flow_lastHorizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_lastHorizontalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_flow_lastHorizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_lastHorizontalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_flow_lastVerticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_lastVerticalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_flow_lastVerticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_lastVerticalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_flow_maxElementsWrap = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_maxElementsWrap;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_flow_verticalAlign = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_verticalAlign;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_flow_verticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_verticalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_flow_verticalGap = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_verticalGap;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_flow_verticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_verticalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_flow_wrapMode = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_wrapMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constrainedHeight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constrainedHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constrainedWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constrainedWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintBaseline_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintBaseline_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintBottom_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintBottom_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintCircleAngle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintCircleAngle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintCircleRadius = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintCircleRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintDimensionRatio = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintDimensionRatio;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintGuide_begin = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintGuide_begin;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintGuide_end = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintGuide_end;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintGuide_percent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintGuide_percent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintHeight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintHeight_default = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintHeight_default;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintHeight_max = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintHeight_max;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintHeight_min = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintHeight_min;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintHeight_percent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintHeight_percent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintHorizontal_bias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintHorizontal_bias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintHorizontal_chainStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintHorizontal_chainStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintHorizontal_weight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintHorizontal_weight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintLeft_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintLeft_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintRight_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintRight_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintTag = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintTag;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintTop_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintTop_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintVertical_bias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintVertical_bias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintVertical_chainStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintVertical_chainStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintVertical_weight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintVertical_weight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintWidth_default = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintWidth_default;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintWidth_max = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintWidth_max;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintWidth_min = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintWidth_min;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_constraintWidth_percent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintWidth_percent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_editor_absoluteX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_editor_absoluteX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_editor_absoluteY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_editor_absoluteY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_goneMarginBaseline = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_goneMarginBaseline;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_goneMarginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_goneMarginBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_goneMarginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_goneMarginEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_goneMarginLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_goneMarginLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_goneMarginRight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_goneMarginRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_goneMarginStart = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_goneMarginStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_goneMarginTop = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_goneMarginTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_marginBaseline = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_marginBaseline;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_layout_wrapBehaviorInParent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_wrapBehaviorInParent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_motionProgress = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_motionProgress;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_motionStagger = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_motionStagger;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_motionTarget = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_motionTarget;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_pathMotionArc = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_pathMotionArc;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_pivotAnchor = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_pivotAnchor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_polarRelativeTo = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_polarRelativeTo;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_quantizeMotionInterpolator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_quantizeMotionInterpolator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_quantizeMotionPhase = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_quantizeMotionPhase;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_quantizeMotionSteps = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_quantizeMotionSteps;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_transformPivotTarget = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_transformPivotTarget;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_transitionEasing = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_transitionEasing;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_transitionPathRotate = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_transitionPathRotate;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintOverride_visibilityMode = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_visibilityMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_alpha = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_id;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_layout_height = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_layout_height;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_layout_marginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_layout_marginBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_layout_marginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_layout_marginEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_layout_marginLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_layout_marginLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_layout_marginRight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_layout_marginRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_layout_marginStart = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_layout_marginStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_layout_marginTop = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_layout_marginTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_layout_width = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_layout_width;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_maxHeight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_maxHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_maxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_maxWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_minHeight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_minHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_minWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_minWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_orientation = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_orientation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_pivotX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_pivotX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_pivotY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_pivotY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_rotation = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_rotation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_rotationX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_rotationX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_rotationY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_rotationY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_scaleX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_scaleX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_scaleY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_scaleY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_transformPivotX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_transformPivotX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_transformPivotY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_transformPivotY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_translationX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_translationX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_translationY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_translationY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_translationZ = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_translationZ;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_android_visibility = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_visibility;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_animateCircleAngleTo = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_animateCircleAngleTo;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_animateRelativeTo = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_animateRelativeTo;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_barrierAllowsGoneWidgets = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_barrierAllowsGoneWidgets;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_barrierDirection = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_barrierDirection;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_barrierMargin = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_barrierMargin;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_chainUseRtl = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_chainUseRtl;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_ConstraintRotate = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_ConstraintRotate;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_constraint_referenced_ids = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_constraint_referenced_ids;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_constraint_referenced_tags = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_constraint_referenced_tags;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_deriveConstraintsFrom = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_deriveConstraintsFrom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_drawPath = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_drawPath;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_flow_firstHorizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_firstHorizontalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_flow_firstHorizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_firstHorizontalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_flow_firstVerticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_firstVerticalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_flow_firstVerticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_firstVerticalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_flow_horizontalAlign = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_horizontalAlign;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_flow_horizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_horizontalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_flow_horizontalGap = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_horizontalGap;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_flow_horizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_horizontalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_flow_lastHorizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_lastHorizontalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_flow_lastHorizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_lastHorizontalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_flow_lastVerticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_lastVerticalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_flow_lastVerticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_lastVerticalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_flow_maxElementsWrap = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_maxElementsWrap;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_flow_verticalAlign = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_verticalAlign;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_flow_verticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_verticalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_flow_verticalGap = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_verticalGap;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_flow_verticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_verticalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_flow_wrapMode = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_wrapMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constrainedHeight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constrainedHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constrainedWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constrainedWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintBaseline_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintBaseline_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintBaseline_toBaselineOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintBaseline_toBaselineOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintBaseline_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintBaseline_toBottomOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintBaseline_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintBaseline_toTopOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintBottom_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintBottom_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintBottom_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintBottom_toBottomOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintBottom_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintBottom_toTopOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintCircle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintCircle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintCircleAngle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintCircleAngle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintCircleRadius = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintCircleRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintDimensionRatio = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintDimensionRatio;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintEnd_toEndOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintEnd_toEndOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintEnd_toStartOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintEnd_toStartOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintGuide_begin = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintGuide_begin;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintGuide_end = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintGuide_end;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintGuide_percent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintGuide_percent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintHeight_default = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintHeight_default;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintHeight_max = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintHeight_max;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintHeight_min = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintHeight_min;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintHeight_percent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintHeight_percent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintHorizontal_bias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintHorizontal_bias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintHorizontal_chainStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintHorizontal_chainStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintHorizontal_weight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintHorizontal_weight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintLeft_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintLeft_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintLeft_toLeftOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintLeft_toLeftOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintLeft_toRightOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintLeft_toRightOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintRight_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintRight_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintRight_toLeftOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintRight_toLeftOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintRight_toRightOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintRight_toRightOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintStart_toEndOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintStart_toEndOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintStart_toStartOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintStart_toStartOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintTag = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintTag;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintTop_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintTop_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintTop_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintTop_toBottomOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintTop_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintTop_toTopOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintVertical_bias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintVertical_bias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintVertical_chainStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintVertical_chainStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintVertical_weight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintVertical_weight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintWidth_default = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintWidth_default;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintWidth_max = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintWidth_max;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintWidth_min = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintWidth_min;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_constraintWidth_percent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintWidth_percent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_editor_absoluteX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_editor_absoluteX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_editor_absoluteY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_editor_absoluteY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_goneMarginBaseline = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_goneMarginBaseline;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_goneMarginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_goneMarginBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_goneMarginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_goneMarginEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_goneMarginLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_goneMarginLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_goneMarginRight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_goneMarginRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_goneMarginStart = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_goneMarginStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_goneMarginTop = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_goneMarginTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_marginBaseline = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_marginBaseline;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_layout_wrapBehaviorInParent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_wrapBehaviorInParent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_motionProgress = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_motionProgress;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_motionStagger = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_motionStagger;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_pathMotionArc = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_pathMotionArc;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_pivotAnchor = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_pivotAnchor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_polarRelativeTo = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_polarRelativeTo;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_quantizeMotionSteps = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_quantizeMotionSteps;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_transitionEasing = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_transitionEasing;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ConstraintSet_transitionPathRotate = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_transitionPathRotate;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_alpha = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_id;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_layout_height = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_layout_height;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_layout_marginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_layout_marginBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_layout_marginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_layout_marginEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_layout_marginLeft = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_layout_marginLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_layout_marginRight = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_layout_marginRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_layout_marginStart = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_layout_marginStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_layout_marginTop = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_layout_marginTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_layout_width = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_layout_width;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_maxHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_maxHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_maxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_maxWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_minHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_minHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_minWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_minWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_orientation = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_orientation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_rotation = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_rotation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_rotationX = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_rotationX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_rotationY = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_rotationY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_scaleX = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_scaleX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_scaleY = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_scaleY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_transformPivotX = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_transformPivotX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_transformPivotY = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_transformPivotY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_translationX = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_translationX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_translationY = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_translationY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_translationZ = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_translationZ;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_android_visibility = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_visibility;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_animateCircleAngleTo = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_animateCircleAngleTo;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_animateRelativeTo = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_animateRelativeTo;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_barrierAllowsGoneWidgets = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_barrierAllowsGoneWidgets;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_barrierDirection = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_barrierDirection;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_barrierMargin = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_barrierMargin;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_chainUseRtl = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_chainUseRtl;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_constraint_referenced_ids = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_constraint_referenced_ids;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_constraint_referenced_tags = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_constraint_referenced_tags;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_drawPath = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_drawPath;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_flow_firstHorizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_firstHorizontalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_flow_firstHorizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_firstHorizontalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_flow_firstVerticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_firstVerticalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_flow_firstVerticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_firstVerticalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_flow_horizontalAlign = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_horizontalAlign;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_flow_horizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_horizontalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_flow_horizontalGap = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_horizontalGap;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_flow_horizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_horizontalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_flow_lastHorizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_lastHorizontalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_flow_lastHorizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_lastHorizontalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_flow_lastVerticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_lastVerticalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_flow_lastVerticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_lastVerticalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_flow_maxElementsWrap = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_maxElementsWrap;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_flow_verticalAlign = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_verticalAlign;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_flow_verticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_verticalBias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_flow_verticalGap = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_verticalGap;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_flow_verticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_verticalStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_flow_wrapMode = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_wrapMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constrainedHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constrainedHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constrainedWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constrainedWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintBaseline_creator = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintBaseline_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintBaseline_toBaselineOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintBaseline_toBaselineOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintBaseline_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintBaseline_toBottomOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintBaseline_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintBaseline_toTopOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintBottom_creator = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintBottom_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintBottom_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintBottom_toBottomOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintBottom_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintBottom_toTopOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintCircle = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintCircle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintCircleAngle = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintCircleAngle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintCircleRadius = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintCircleRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintDimensionRatio = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintDimensionRatio;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintEnd_toEndOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintEnd_toEndOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintEnd_toStartOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintEnd_toStartOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintGuide_begin = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintGuide_begin;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintGuide_end = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintGuide_end;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintGuide_percent = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintGuide_percent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintHeight_default = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintHeight_default;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintHeight_max = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintHeight_max;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintHeight_min = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintHeight_min;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintHeight_percent = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintHeight_percent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintHorizontal_bias = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintHorizontal_bias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintHorizontal_chainStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintHorizontal_chainStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintHorizontal_weight = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintHorizontal_weight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintLeft_creator = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintLeft_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintLeft_toLeftOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintLeft_toLeftOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintLeft_toRightOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintLeft_toRightOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintRight_creator = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintRight_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintRight_toLeftOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintRight_toLeftOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintRight_toRightOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintRight_toRightOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintStart_toEndOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintStart_toEndOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintStart_toStartOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintStart_toStartOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintTag = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintTag;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintTop_creator = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintTop_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintTop_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintTop_toBottomOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintTop_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintTop_toTopOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintVertical_bias = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintVertical_bias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintVertical_chainStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintVertical_chainStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintVertical_weight = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintVertical_weight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintWidth_default = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintWidth_default;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintWidth_max = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintWidth_max;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintWidth_min = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintWidth_min;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_constraintWidth_percent = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintWidth_percent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_editor_absoluteX = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_editor_absoluteX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_editor_absoluteY = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_editor_absoluteY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_goneMarginBaseline = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_goneMarginBaseline;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_goneMarginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_goneMarginBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_goneMarginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_goneMarginEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_goneMarginLeft = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_goneMarginLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_goneMarginRight = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_goneMarginRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_goneMarginStart = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_goneMarginStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_goneMarginTop = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_goneMarginTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_marginBaseline = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_marginBaseline;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_layout_wrapBehaviorInParent = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_wrapBehaviorInParent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_motionProgress = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_motionProgress;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_motionStagger = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_motionStagger;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_pathMotionArc = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_pathMotionArc;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_pivotAnchor = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_pivotAnchor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_polarRelativeTo = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_polarRelativeTo;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_quantizeMotionInterpolator = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_quantizeMotionInterpolator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_quantizeMotionPhase = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_quantizeMotionPhase;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_quantizeMotionSteps = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_quantizeMotionSteps;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_transformPivotTarget = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_transformPivotTarget;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_transitionEasing = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_transitionEasing;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_transitionPathRotate = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_transitionPathRotate;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Constraint_visibilityMode = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_visibilityMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CoordinatorLayout = global::pjsua2xamarin.Droid.Resource.Styleable.CoordinatorLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CoordinatorLayout_keylines = global::pjsua2xamarin.Droid.Resource.Styleable.CoordinatorLayout_keylines;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CoordinatorLayout_Layout = global::pjsua2xamarin.Droid.Resource.Styleable.CoordinatorLayout_Layout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CoordinatorLayout_Layout_android_layout_gravity = global::pjsua2xamarin.Droid.Resource.Styleable.CoordinatorLayout_Layout_android_layout_gravity;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CoordinatorLayout_Layout_layout_anchor = global::pjsua2xamarin.Droid.Resource.Styleable.CoordinatorLayout_Layout_layout_anchor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CoordinatorLayout_Layout_layout_anchorGravity = global::pjsua2xamarin.Droid.Resource.Styleable.CoordinatorLayout_Layout_layout_anchorGravity;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CoordinatorLayout_Layout_layout_behavior = global::pjsua2xamarin.Droid.Resource.Styleable.CoordinatorLayout_Layout_layout_behavior;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CoordinatorLayout_Layout_layout_dodgeInsetEdges = global::pjsua2xamarin.Droid.Resource.Styleable.CoordinatorLayout_Layout_layout_dodgeInsetEdges;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CoordinatorLayout_Layout_layout_insetEdge = global::pjsua2xamarin.Droid.Resource.Styleable.CoordinatorLayout_Layout_layout_insetEdge;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CoordinatorLayout_Layout_layout_keyline = global::pjsua2xamarin.Droid.Resource.Styleable.CoordinatorLayout_Layout_layout_keyline;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CoordinatorLayout_statusBarBackground = global::pjsua2xamarin.Droid.Resource.Styleable.CoordinatorLayout_statusBarBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CustomAttribute = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CustomAttribute_attributeName = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute_attributeName;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CustomAttribute_customBoolean = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute_customBoolean;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CustomAttribute_customColorDrawableValue = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute_customColorDrawableValue;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CustomAttribute_customColorValue = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute_customColorValue;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CustomAttribute_customDimension = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute_customDimension;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CustomAttribute_customFloatValue = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute_customFloatValue;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CustomAttribute_customIntegerValue = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute_customIntegerValue;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CustomAttribute_customPixelDimension = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute_customPixelDimension;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CustomAttribute_customReference = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute_customReference;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CustomAttribute_customStringValue = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute_customStringValue;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.CustomAttribute_methodName = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute_methodName;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.DrawerArrowToggle = global::pjsua2xamarin.Droid.Resource.Styleable.DrawerArrowToggle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.DrawerArrowToggle_arrowHeadLength = global::pjsua2xamarin.Droid.Resource.Styleable.DrawerArrowToggle_arrowHeadLength;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.DrawerArrowToggle_arrowShaftLength = global::pjsua2xamarin.Droid.Resource.Styleable.DrawerArrowToggle_arrowShaftLength;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.DrawerArrowToggle_barLength = global::pjsua2xamarin.Droid.Resource.Styleable.DrawerArrowToggle_barLength;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.DrawerArrowToggle_color = global::pjsua2xamarin.Droid.Resource.Styleable.DrawerArrowToggle_color;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.DrawerArrowToggle_drawableSize = global::pjsua2xamarin.Droid.Resource.Styleable.DrawerArrowToggle_drawableSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.DrawerArrowToggle_gapBetweenBars = global::pjsua2xamarin.Droid.Resource.Styleable.DrawerArrowToggle_gapBetweenBars;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.DrawerArrowToggle_spinBars = global::pjsua2xamarin.Droid.Resource.Styleable.DrawerArrowToggle_spinBars;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.DrawerArrowToggle_thickness = global::pjsua2xamarin.Droid.Resource.Styleable.DrawerArrowToggle_thickness;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.DrawerLayout = global::pjsua2xamarin.Droid.Resource.Styleable.DrawerLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.DrawerLayout_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.DrawerLayout_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ExtendedFloatingActionButton = global::pjsua2xamarin.Droid.Resource.Styleable.ExtendedFloatingActionButton;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ExtendedFloatingActionButton_Behavior_Layout = global::pjsua2xamarin.Droid.Resource.Styleable.ExtendedFloatingActionButton_Behavior_Layout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ExtendedFloatingActionButton_Behavior_Layout_behavior_autoHide = global::pjsua2xamarin.Droid.Resource.Styleable.ExtendedFloatingActionButton_Behavior_Layout_behavior_autoHide;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ExtendedFloatingActionButton_Behavior_Layout_behavior_autoShrink = global::pjsua2xamarin.Droid.Resource.Styleable.ExtendedFloatingActionButton_Behavior_Layout_behavior_autoShrink;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ExtendedFloatingActionButton_collapsedSize = global::pjsua2xamarin.Droid.Resource.Styleable.ExtendedFloatingActionButton_collapsedSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ExtendedFloatingActionButton_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.ExtendedFloatingActionButton_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ExtendedFloatingActionButton_extendMotionSpec = global::pjsua2xamarin.Droid.Resource.Styleable.ExtendedFloatingActionButton_extendMotionSpec;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ExtendedFloatingActionButton_hideMotionSpec = global::pjsua2xamarin.Droid.Resource.Styleable.ExtendedFloatingActionButton_hideMotionSpec;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ExtendedFloatingActionButton_showMotionSpec = global::pjsua2xamarin.Droid.Resource.Styleable.ExtendedFloatingActionButton_showMotionSpec;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ExtendedFloatingActionButton_shrinkMotionSpec = global::pjsua2xamarin.Droid.Resource.Styleable.ExtendedFloatingActionButton_shrinkMotionSpec;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FloatingActionButton = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FloatingActionButton_android_enabled = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_android_enabled;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FloatingActionButton_backgroundTint = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_backgroundTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FloatingActionButton_backgroundTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_backgroundTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FloatingActionButton_Behavior_Layout = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_Behavior_Layout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FloatingActionButton_Behavior_Layout_behavior_autoHide = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_Behavior_Layout_behavior_autoHide;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FloatingActionButton_borderWidth = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_borderWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FloatingActionButton_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FloatingActionButton_ensureMinTouchTargetSize = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_ensureMinTouchTargetSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FloatingActionButton_fabCustomSize = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_fabCustomSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FloatingActionButton_fabSize = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_fabSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FloatingActionButton_hideMotionSpec = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_hideMotionSpec;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FloatingActionButton_hoveredFocusedTranslationZ = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_hoveredFocusedTranslationZ;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FloatingActionButton_maxImageSize = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_maxImageSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FloatingActionButton_pressedTranslationZ = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_pressedTranslationZ;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FloatingActionButton_rippleColor = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_rippleColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FloatingActionButton_shapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_shapeAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FloatingActionButton_shapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_shapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FloatingActionButton_showMotionSpec = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_showMotionSpec;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FloatingActionButton_useCompatPadding = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_useCompatPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FlowLayout = global::pjsua2xamarin.Droid.Resource.Styleable.FlowLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FlowLayout_itemSpacing = global::pjsua2xamarin.Droid.Resource.Styleable.FlowLayout_itemSpacing;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FlowLayout_lineSpacing = global::pjsua2xamarin.Droid.Resource.Styleable.FlowLayout_lineSpacing;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FontFamily = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FontFamilyFont = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FontFamilyFont_android_font = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_android_font;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FontFamilyFont_android_fontStyle = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_android_fontStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FontFamilyFont_android_fontVariationSettings = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_android_fontVariationSettings;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FontFamilyFont_android_fontWeight = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_android_fontWeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FontFamilyFont_android_ttcIndex = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_android_ttcIndex;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FontFamilyFont_font = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_font;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FontFamilyFont_fontStyle = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_fontStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FontFamilyFont_fontVariationSettings = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_fontVariationSettings;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FontFamilyFont_fontWeight = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_fontWeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FontFamilyFont_ttcIndex = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_ttcIndex;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FontFamily_fontProviderAuthority = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily_fontProviderAuthority;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FontFamily_fontProviderCerts = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily_fontProviderCerts;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FontFamily_fontProviderFetchStrategy = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily_fontProviderFetchStrategy;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FontFamily_fontProviderFetchTimeout = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily_fontProviderFetchTimeout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FontFamily_fontProviderPackage = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily_fontProviderPackage;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FontFamily_fontProviderQuery = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily_fontProviderQuery;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FontFamily_fontProviderSystemFontFamily = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily_fontProviderSystemFontFamily;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ForegroundLinearLayout = global::pjsua2xamarin.Droid.Resource.Styleable.ForegroundLinearLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ForegroundLinearLayout_android_foreground = global::pjsua2xamarin.Droid.Resource.Styleable.ForegroundLinearLayout_android_foreground;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ForegroundLinearLayout_android_foregroundGravity = global::pjsua2xamarin.Droid.Resource.Styleable.ForegroundLinearLayout_android_foregroundGravity;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ForegroundLinearLayout_foregroundInsidePadding = global::pjsua2xamarin.Droid.Resource.Styleable.ForegroundLinearLayout_foregroundInsidePadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Fragment = global::pjsua2xamarin.Droid.Resource.Styleable.Fragment;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FragmentContainerView = global::pjsua2xamarin.Droid.Resource.Styleable.FragmentContainerView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FragmentContainerView_android_name = global::pjsua2xamarin.Droid.Resource.Styleable.FragmentContainerView_android_name;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.FragmentContainerView_android_tag = global::pjsua2xamarin.Droid.Resource.Styleable.FragmentContainerView_android_tag;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Fragment_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.Fragment_android_id;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Fragment_android_name = global::pjsua2xamarin.Droid.Resource.Styleable.Fragment_android_name;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Fragment_android_tag = global::pjsua2xamarin.Droid.Resource.Styleable.Fragment_android_tag;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.GradientColor = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.GradientColorItem = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColorItem;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.GradientColorItem_android_color = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColorItem_android_color;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.GradientColorItem_android_offset = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColorItem_android_offset;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.GradientColor_android_centerColor = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_centerColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.GradientColor_android_centerX = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_centerX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.GradientColor_android_centerY = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_centerY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.GradientColor_android_endColor = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_endColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.GradientColor_android_endX = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_endX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.GradientColor_android_endY = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_endY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.GradientColor_android_gradientRadius = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_gradientRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.GradientColor_android_startColor = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_startColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.GradientColor_android_startX = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_startX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.GradientColor_android_startY = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_startY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.GradientColor_android_tileMode = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_tileMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.GradientColor_android_type = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_type;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ImageFilterView = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ImageFilterView_altSrc = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_altSrc;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ImageFilterView_blendSrc = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_blendSrc;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ImageFilterView_brightness = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_brightness;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ImageFilterView_contrast = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_contrast;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ImageFilterView_crossfade = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_crossfade;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ImageFilterView_imagePanX = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_imagePanX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ImageFilterView_imagePanY = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_imagePanY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ImageFilterView_imageRotate = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_imageRotate;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ImageFilterView_imageZoom = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_imageZoom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ImageFilterView_overlay = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_overlay;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ImageFilterView_round = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_round;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ImageFilterView_roundPercent = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_roundPercent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ImageFilterView_saturation = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_saturation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ImageFilterView_warmth = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_warmth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.include = global::pjsua2xamarin.Droid.Resource.Styleable.include;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.include_constraintSet = global::pjsua2xamarin.Droid.Resource.Styleable.include_constraintSet;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Insets = global::pjsua2xamarin.Droid.Resource.Styleable.Insets;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Insets_paddingBottomSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Styleable.Insets_paddingBottomSystemWindowInsets;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Insets_paddingLeftSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Styleable.Insets_paddingLeftSystemWindowInsets;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Insets_paddingRightSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Styleable.Insets_paddingRightSystemWindowInsets;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Insets_paddingTopSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Styleable.Insets_paddingTopSystemWindowInsets;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ItemsViewRendererTheme = global::pjsua2xamarin.Droid.Resource.Styleable.ItemsViewRendererTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ItemsViewRendererTheme_collectionViewStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ItemsViewRendererTheme_collectionViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyAttribute = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyAttribute_android_alpha = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyAttribute_android_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyAttribute_android_rotation = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_rotation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyAttribute_android_rotationX = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_rotationX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyAttribute_android_rotationY = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_rotationY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyAttribute_android_scaleX = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_scaleX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyAttribute_android_scaleY = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_scaleY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyAttribute_android_transformPivotX = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_transformPivotX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyAttribute_android_transformPivotY = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_transformPivotY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyAttribute_android_translationX = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_translationX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyAttribute_android_translationY = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_translationY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyAttribute_android_translationZ = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_translationZ;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyAttribute_curveFit = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_curveFit;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyAttribute_framePosition = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_framePosition;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyAttribute_motionProgress = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_motionProgress;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyAttribute_motionTarget = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_motionTarget;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyAttribute_transformPivotTarget = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_transformPivotTarget;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyAttribute_transitionEasing = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_transitionEasing;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyAttribute_transitionPathRotate = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_transitionPathRotate;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyCycle = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyCycle_android_alpha = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_android_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyCycle_android_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_android_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyCycle_android_rotation = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_android_rotation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyCycle_android_rotationX = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_android_rotationX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyCycle_android_rotationY = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_android_rotationY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyCycle_android_scaleX = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_android_scaleX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyCycle_android_scaleY = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_android_scaleY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyCycle_android_translationX = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_android_translationX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyCycle_android_translationY = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_android_translationY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyCycle_android_translationZ = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_android_translationZ;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyCycle_curveFit = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_curveFit;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyCycle_framePosition = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_framePosition;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyCycle_motionProgress = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_motionProgress;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyCycle_motionTarget = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_motionTarget;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyCycle_transitionEasing = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_transitionEasing;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyCycle_transitionPathRotate = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_transitionPathRotate;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyCycle_waveOffset = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_waveOffset;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyCycle_wavePeriod = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_wavePeriod;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyCycle_wavePhase = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_wavePhase;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyCycle_waveShape = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_waveShape;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyCycle_waveVariesBy = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_waveVariesBy;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyFrame = global::pjsua2xamarin.Droid.Resource.Styleable.KeyFrame;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyFramesAcceleration = global::pjsua2xamarin.Droid.Resource.Styleable.KeyFramesAcceleration;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyFramesVelocity = global::pjsua2xamarin.Droid.Resource.Styleable.KeyFramesVelocity;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyPosition = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyPosition_curveFit = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_curveFit;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyPosition_drawPath = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_drawPath;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyPosition_framePosition = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_framePosition;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyPosition_keyPositionType = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_keyPositionType;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyPosition_motionTarget = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_motionTarget;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyPosition_pathMotionArc = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_pathMotionArc;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyPosition_percentHeight = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_percentHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyPosition_percentWidth = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_percentWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyPosition_percentX = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_percentX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyPosition_percentY = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_percentY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyPosition_sizePercent = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_sizePercent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyPosition_transitionEasing = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_transitionEasing;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTimeCycle = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTimeCycle_android_alpha = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_android_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTimeCycle_android_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_android_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTimeCycle_android_rotation = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_android_rotation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTimeCycle_android_rotationX = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_android_rotationX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTimeCycle_android_rotationY = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_android_rotationY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTimeCycle_android_scaleX = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_android_scaleX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTimeCycle_android_scaleY = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_android_scaleY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTimeCycle_android_translationX = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_android_translationX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTimeCycle_android_translationY = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_android_translationY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTimeCycle_android_translationZ = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_android_translationZ;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTimeCycle_curveFit = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_curveFit;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTimeCycle_framePosition = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_framePosition;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTimeCycle_motionProgress = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_motionProgress;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTimeCycle_motionTarget = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_motionTarget;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTimeCycle_transitionEasing = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_transitionEasing;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTimeCycle_transitionPathRotate = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_transitionPathRotate;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTimeCycle_waveDecay = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_waveDecay;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTimeCycle_waveOffset = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_waveOffset;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTimeCycle_wavePeriod = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_wavePeriod;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTimeCycle_wavePhase = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_wavePhase;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTimeCycle_waveShape = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_waveShape;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTrigger = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTrigger_framePosition = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_framePosition;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTrigger_motionTarget = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_motionTarget;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTrigger_motion_postLayoutCollision = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_motion_postLayoutCollision;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTrigger_motion_triggerOnCollision = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_motion_triggerOnCollision;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTrigger_onCross = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_onCross;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTrigger_onNegativeCross = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_onNegativeCross;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTrigger_onPositiveCross = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_onPositiveCross;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTrigger_triggerId = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_triggerId;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTrigger_triggerReceiver = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_triggerReceiver;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTrigger_triggerSlack = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_triggerSlack;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTrigger_viewTransitionOnCross = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_viewTransitionOnCross;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTrigger_viewTransitionOnNegativeCross = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_viewTransitionOnNegativeCross;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.KeyTrigger_viewTransitionOnPositiveCross = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_viewTransitionOnPositiveCross;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout = global::pjsua2xamarin.Droid.Resource.Styleable.Layout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_android_layout_height = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_android_layout_height;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_android_layout_marginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_android_layout_marginBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_android_layout_marginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_android_layout_marginEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_android_layout_marginLeft = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_android_layout_marginLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_android_layout_marginRight = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_android_layout_marginRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_android_layout_marginStart = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_android_layout_marginStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_android_layout_marginTop = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_android_layout_marginTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_android_layout_width = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_android_layout_width;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_android_orientation = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_android_orientation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_barrierAllowsGoneWidgets = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_barrierAllowsGoneWidgets;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_barrierDirection = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_barrierDirection;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_barrierMargin = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_barrierMargin;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_chainUseRtl = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_chainUseRtl;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_constraint_referenced_ids = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_constraint_referenced_ids;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_constraint_referenced_tags = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_constraint_referenced_tags;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constrainedHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constrainedHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constrainedWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constrainedWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintBaseline_creator = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintBaseline_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintBaseline_toBaselineOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintBaseline_toBaselineOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintBaseline_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintBaseline_toBottomOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintBaseline_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintBaseline_toTopOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintBottom_creator = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintBottom_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintBottom_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintBottom_toBottomOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintBottom_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintBottom_toTopOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintCircle = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintCircle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintCircleAngle = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintCircleAngle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintCircleRadius = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintCircleRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintDimensionRatio = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintDimensionRatio;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintEnd_toEndOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintEnd_toEndOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintEnd_toStartOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintEnd_toStartOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintGuide_begin = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintGuide_begin;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintGuide_end = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintGuide_end;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintGuide_percent = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintGuide_percent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintHeight_default = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintHeight_default;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintHeight_max = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintHeight_max;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintHeight_min = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintHeight_min;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintHeight_percent = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintHeight_percent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintHorizontal_bias = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintHorizontal_bias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintHorizontal_chainStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintHorizontal_chainStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintHorizontal_weight = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintHorizontal_weight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintLeft_creator = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintLeft_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintLeft_toLeftOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintLeft_toLeftOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintLeft_toRightOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintLeft_toRightOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintRight_creator = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintRight_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintRight_toLeftOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintRight_toLeftOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintRight_toRightOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintRight_toRightOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintStart_toEndOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintStart_toEndOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintStart_toStartOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintStart_toStartOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintTop_creator = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintTop_creator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintTop_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintTop_toBottomOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintTop_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintTop_toTopOf;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintVertical_bias = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintVertical_bias;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintVertical_chainStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintVertical_chainStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintVertical_weight = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintVertical_weight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintWidth_default = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintWidth_default;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintWidth_max = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintWidth_max;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintWidth_min = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintWidth_min;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_constraintWidth_percent = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintWidth_percent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_editor_absoluteX = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_editor_absoluteX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_editor_absoluteY = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_editor_absoluteY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_goneMarginBaseline = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_goneMarginBaseline;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_goneMarginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_goneMarginBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_goneMarginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_goneMarginEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_goneMarginLeft = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_goneMarginLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_goneMarginRight = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_goneMarginRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_goneMarginStart = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_goneMarginStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_goneMarginTop = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_goneMarginTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_marginBaseline = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_marginBaseline;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_layout_wrapBehaviorInParent = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_wrapBehaviorInParent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_maxHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_maxHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_maxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_maxWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_minHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_minHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Layout_minWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_minWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.LinearLayoutCompat = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.LinearLayoutCompat_android_baselineAligned = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_android_baselineAligned;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.LinearLayoutCompat_android_baselineAlignedChildIndex = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_android_baselineAlignedChildIndex;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.LinearLayoutCompat_android_gravity = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_android_gravity;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.LinearLayoutCompat_android_orientation = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_android_orientation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.LinearLayoutCompat_android_weightSum = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_android_weightSum;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.LinearLayoutCompat_divider = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_divider;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.LinearLayoutCompat_dividerPadding = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_dividerPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.LinearLayoutCompat_Layout = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_Layout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.LinearLayoutCompat_Layout_android_layout_gravity = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_Layout_android_layout_gravity;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.LinearLayoutCompat_Layout_android_layout_height = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_Layout_android_layout_height;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.LinearLayoutCompat_Layout_android_layout_weight = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_Layout_android_layout_weight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.LinearLayoutCompat_Layout_android_layout_width = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_Layout_android_layout_width;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.LinearLayoutCompat_measureWithLargestChild = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_measureWithLargestChild;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.LinearLayoutCompat_showDividers = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_showDividers;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.LinearProgressIndicator = global::pjsua2xamarin.Droid.Resource.Styleable.LinearProgressIndicator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.LinearProgressIndicator_indeterminateAnimationType = global::pjsua2xamarin.Droid.Resource.Styleable.LinearProgressIndicator_indeterminateAnimationType;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.LinearProgressIndicator_indicatorDirectionLinear = global::pjsua2xamarin.Droid.Resource.Styleable.LinearProgressIndicator_indicatorDirectionLinear;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ListPopupWindow = global::pjsua2xamarin.Droid.Resource.Styleable.ListPopupWindow;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ListPopupWindow_android_dropDownHorizontalOffset = global::pjsua2xamarin.Droid.Resource.Styleable.ListPopupWindow_android_dropDownHorizontalOffset;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ListPopupWindow_android_dropDownVerticalOffset = global::pjsua2xamarin.Droid.Resource.Styleable.ListPopupWindow_android_dropDownVerticalOffset;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialAlertDialog = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAlertDialog;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialAlertDialogTheme = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAlertDialogTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialAlertDialogTheme_materialAlertDialogBodyTextStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAlertDialogTheme_materialAlertDialogBodyTextStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialAlertDialogTheme_materialAlertDialogTheme = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAlertDialogTheme_materialAlertDialogTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialAlertDialogTheme_materialAlertDialogTitleIconStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAlertDialogTheme_materialAlertDialogTitleIconStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialAlertDialogTheme_materialAlertDialogTitlePanelStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAlertDialogTheme_materialAlertDialogTitlePanelStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialAlertDialogTheme_materialAlertDialogTitleTextStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAlertDialogTheme_materialAlertDialogTitleTextStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialAlertDialog_backgroundInsetBottom = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAlertDialog_backgroundInsetBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialAlertDialog_backgroundInsetEnd = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAlertDialog_backgroundInsetEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialAlertDialog_backgroundInsetStart = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAlertDialog_backgroundInsetStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialAlertDialog_backgroundInsetTop = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAlertDialog_backgroundInsetTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialAutoCompleteTextView = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAutoCompleteTextView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialAutoCompleteTextView_android_inputType = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAutoCompleteTextView_android_inputType;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButton = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButtonToggleGroup = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButtonToggleGroup;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButtonToggleGroup_checkedButton = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButtonToggleGroup_checkedButton;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButtonToggleGroup_selectionRequired = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButtonToggleGroup_selectionRequired;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButtonToggleGroup_singleSelection = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButtonToggleGroup_singleSelection;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButton_android_background = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_android_background;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButton_android_checkable = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_android_checkable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButton_android_insetBottom = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_android_insetBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButton_android_insetLeft = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_android_insetLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButton_android_insetRight = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_android_insetRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButton_android_insetTop = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_android_insetTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButton_backgroundTint = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_backgroundTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButton_backgroundTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_backgroundTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButton_cornerRadius = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_cornerRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButton_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButton_icon = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_icon;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButton_iconGravity = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_iconGravity;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButton_iconPadding = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_iconPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButton_iconSize = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_iconSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButton_iconTint = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_iconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButton_iconTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_iconTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButton_rippleColor = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_rippleColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButton_shapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_shapeAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButton_shapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_shapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButton_strokeColor = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_strokeColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialButton_strokeWidth = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_strokeWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCalendar = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendar;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCalendarItem = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendarItem;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCalendarItem_android_insetBottom = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendarItem_android_insetBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCalendarItem_android_insetLeft = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendarItem_android_insetLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCalendarItem_android_insetRight = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendarItem_android_insetRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCalendarItem_android_insetTop = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendarItem_android_insetTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCalendarItem_itemFillColor = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendarItem_itemFillColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCalendarItem_itemShapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendarItem_itemShapeAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCalendarItem_itemShapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendarItem_itemShapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCalendarItem_itemStrokeColor = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendarItem_itemStrokeColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCalendarItem_itemStrokeWidth = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendarItem_itemStrokeWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCalendarItem_itemTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendarItem_itemTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCalendar_android_windowFullscreen = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendar_android_windowFullscreen;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCalendar_dayInvalidStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendar_dayInvalidStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCalendar_daySelectedStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendar_daySelectedStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCalendar_dayStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendar_dayStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCalendar_dayTodayStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendar_dayTodayStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCalendar_nestedScrollable = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendar_nestedScrollable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCalendar_rangeFillColor = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendar_rangeFillColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCalendar_yearSelectedStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendar_yearSelectedStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCalendar_yearStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendar_yearStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCalendar_yearTodayStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendar_yearTodayStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCardView = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCardView_android_checkable = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_android_checkable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCardView_cardForegroundColor = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_cardForegroundColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCardView_checkedIcon = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_checkedIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCardView_checkedIconMargin = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_checkedIconMargin;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCardView_checkedIconSize = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_checkedIconSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCardView_checkedIconTint = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_checkedIconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCardView_rippleColor = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_rippleColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCardView_shapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_shapeAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCardView_shapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_shapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCardView_state_dragged = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_state_dragged;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCardView_strokeColor = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_strokeColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCardView_strokeWidth = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_strokeWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCheckBox = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCheckBox;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCheckBox_buttonTint = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCheckBox_buttonTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialCheckBox_useMaterialThemeColors = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCheckBox_useMaterialThemeColors;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialRadioButton = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialRadioButton;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialRadioButton_buttonTint = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialRadioButton_buttonTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialRadioButton_useMaterialThemeColors = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialRadioButton_useMaterialThemeColors;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialShape = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialShape;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialShape_shapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialShape_shapeAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialShape_shapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialShape_shapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialTextAppearance_android_letterSpacing = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialTextAppearance_android_letterSpacing;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialTextAppearance_android_lineHeight = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialTextAppearance_android_lineHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialTextAppearance_lineHeight = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialTextAppearance_lineHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialTextView = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialTextView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialTextView_android_lineHeight = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialTextView_android_lineHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialTextView_android_textAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialTextView_android_textAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialTextView_lineHeight = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialTextView_lineHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialTimePicker = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialTimePicker;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialTimePicker_clockIcon = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialTimePicker_clockIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialTimePicker_keyboardIcon = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialTimePicker_keyboardIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialToolbar = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialToolbar;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialToolbar_navigationIconTint = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialToolbar_navigationIconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialToolbar_subtitleCentered = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialToolbar_subtitleCentered;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MaterialToolbar_titleCentered = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialToolbar_titleCentered;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuGroup = global::pjsua2xamarin.Droid.Resource.Styleable.MenuGroup;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuGroup_android_checkableBehavior = global::pjsua2xamarin.Droid.Resource.Styleable.MenuGroup_android_checkableBehavior;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuGroup_android_enabled = global::pjsua2xamarin.Droid.Resource.Styleable.MenuGroup_android_enabled;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuGroup_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.MenuGroup_android_id;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuGroup_android_menuCategory = global::pjsua2xamarin.Droid.Resource.Styleable.MenuGroup_android_menuCategory;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuGroup_android_orderInCategory = global::pjsua2xamarin.Droid.Resource.Styleable.MenuGroup_android_orderInCategory;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuGroup_android_visible = global::pjsua2xamarin.Droid.Resource.Styleable.MenuGroup_android_visible;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_actionLayout = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_actionLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_actionProviderClass = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_actionProviderClass;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_actionViewClass = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_actionViewClass;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_alphabeticModifiers = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_alphabeticModifiers;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_android_alphabeticShortcut = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_alphabeticShortcut;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_android_checkable = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_checkable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_android_checked = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_checked;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_android_enabled = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_enabled;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_android_icon = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_icon;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_id;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_android_menuCategory = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_menuCategory;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_android_numericShortcut = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_numericShortcut;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_android_onClick = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_onClick;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_android_orderInCategory = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_orderInCategory;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_android_title = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_title;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_android_titleCondensed = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_titleCondensed;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_android_visible = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_visible;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_contentDescription = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_contentDescription;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_iconTint = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_iconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_iconTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_iconTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_numericModifiers = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_numericModifiers;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_showAsAction = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_showAsAction;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuItem_tooltipText = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_tooltipText;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuView = global::pjsua2xamarin.Droid.Resource.Styleable.MenuView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuView_android_headerBackground = global::pjsua2xamarin.Droid.Resource.Styleable.MenuView_android_headerBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuView_android_horizontalDivider = global::pjsua2xamarin.Droid.Resource.Styleable.MenuView_android_horizontalDivider;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuView_android_itemBackground = global::pjsua2xamarin.Droid.Resource.Styleable.MenuView_android_itemBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuView_android_itemIconDisabledAlpha = global::pjsua2xamarin.Droid.Resource.Styleable.MenuView_android_itemIconDisabledAlpha;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuView_android_itemTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.MenuView_android_itemTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuView_android_verticalDivider = global::pjsua2xamarin.Droid.Resource.Styleable.MenuView_android_verticalDivider;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuView_android_windowAnimationStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MenuView_android_windowAnimationStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuView_preserveIconSpacing = global::pjsua2xamarin.Droid.Resource.Styleable.MenuView_preserveIconSpacing;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MenuView_subMenuArrow = global::pjsua2xamarin.Droid.Resource.Styleable.MenuView_subMenuArrow;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MockView = global::pjsua2xamarin.Droid.Resource.Styleable.MockView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MockView_mock_diagonalsColor = global::pjsua2xamarin.Droid.Resource.Styleable.MockView_mock_diagonalsColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MockView_mock_label = global::pjsua2xamarin.Droid.Resource.Styleable.MockView_mock_label;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MockView_mock_labelBackgroundColor = global::pjsua2xamarin.Droid.Resource.Styleable.MockView_mock_labelBackgroundColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MockView_mock_labelColor = global::pjsua2xamarin.Droid.Resource.Styleable.MockView_mock_labelColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MockView_mock_showDiagonals = global::pjsua2xamarin.Droid.Resource.Styleable.MockView_mock_showDiagonals;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MockView_mock_showLabel = global::pjsua2xamarin.Droid.Resource.Styleable.MockView_mock_showLabel;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Motion = global::pjsua2xamarin.Droid.Resource.Styleable.Motion;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionEffect = global::pjsua2xamarin.Droid.Resource.Styleable.MotionEffect;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionEffect_motionEffect_alpha = global::pjsua2xamarin.Droid.Resource.Styleable.MotionEffect_motionEffect_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionEffect_motionEffect_end = global::pjsua2xamarin.Droid.Resource.Styleable.MotionEffect_motionEffect_end;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionEffect_motionEffect_move = global::pjsua2xamarin.Droid.Resource.Styleable.MotionEffect_motionEffect_move;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionEffect_motionEffect_start = global::pjsua2xamarin.Droid.Resource.Styleable.MotionEffect_motionEffect_start;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionEffect_motionEffect_strict = global::pjsua2xamarin.Droid.Resource.Styleable.MotionEffect_motionEffect_strict;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionEffect_motionEffect_translationX = global::pjsua2xamarin.Droid.Resource.Styleable.MotionEffect_motionEffect_translationX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionEffect_motionEffect_translationY = global::pjsua2xamarin.Droid.Resource.Styleable.MotionEffect_motionEffect_translationY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionEffect_motionEffect_viewTransition = global::pjsua2xamarin.Droid.Resource.Styleable.MotionEffect_motionEffect_viewTransition;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionHelper = global::pjsua2xamarin.Droid.Resource.Styleable.MotionHelper;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionHelper_onHide = global::pjsua2xamarin.Droid.Resource.Styleable.MotionHelper_onHide;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionHelper_onShow = global::pjsua2xamarin.Droid.Resource.Styleable.MotionHelper_onShow;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_android_autoSizeTextType = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_android_autoSizeTextType;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_android_fontFamily = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_android_fontFamily;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_android_gravity = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_android_gravity;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_android_shadowRadius = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_android_shadowRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_android_text = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_android_text;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_android_textColor = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_android_textColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_android_textSize = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_android_textSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_android_textStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_android_textStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_android_typeface = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_android_typeface;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_borderRound = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_borderRound;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_borderRoundPercent = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_borderRoundPercent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_scaleFromTextSize = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_scaleFromTextSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_textBackground = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_textBackgroundPanX = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textBackgroundPanX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_textBackgroundPanY = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textBackgroundPanY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_textBackgroundRotate = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textBackgroundRotate;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_textBackgroundZoom = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textBackgroundZoom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_textOutlineColor = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textOutlineColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_textOutlineThickness = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textOutlineThickness;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_textPanX = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textPanX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_textPanY = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textPanY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_textureBlurFactor = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textureBlurFactor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_textureEffect = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textureEffect;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_textureHeight = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textureHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLabel_textureWidth = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textureWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLayout = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLayout_applyMotionScene = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLayout_applyMotionScene;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLayout_currentState = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLayout_currentState;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLayout_layoutDescription = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLayout_layoutDescription;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLayout_motionDebug = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLayout_motionDebug;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLayout_motionProgress = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLayout_motionProgress;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionLayout_showPaths = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLayout_showPaths;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionScene = global::pjsua2xamarin.Droid.Resource.Styleable.MotionScene;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionScene_defaultDuration = global::pjsua2xamarin.Droid.Resource.Styleable.MotionScene_defaultDuration;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionScene_layoutDuringTransition = global::pjsua2xamarin.Droid.Resource.Styleable.MotionScene_layoutDuringTransition;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionTelltales = global::pjsua2xamarin.Droid.Resource.Styleable.MotionTelltales;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionTelltales_telltales_tailColor = global::pjsua2xamarin.Droid.Resource.Styleable.MotionTelltales_telltales_tailColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionTelltales_telltales_tailScale = global::pjsua2xamarin.Droid.Resource.Styleable.MotionTelltales_telltales_tailScale;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.MotionTelltales_telltales_velocityMode = global::pjsua2xamarin.Droid.Resource.Styleable.MotionTelltales_telltales_velocityMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Motion_animateCircleAngleTo = global::pjsua2xamarin.Droid.Resource.Styleable.Motion_animateCircleAngleTo;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Motion_animateRelativeTo = global::pjsua2xamarin.Droid.Resource.Styleable.Motion_animateRelativeTo;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Motion_drawPath = global::pjsua2xamarin.Droid.Resource.Styleable.Motion_drawPath;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Motion_motionPathRotate = global::pjsua2xamarin.Droid.Resource.Styleable.Motion_motionPathRotate;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Motion_motionStagger = global::pjsua2xamarin.Droid.Resource.Styleable.Motion_motionStagger;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Motion_pathMotionArc = global::pjsua2xamarin.Droid.Resource.Styleable.Motion_pathMotionArc;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Motion_quantizeMotionInterpolator = global::pjsua2xamarin.Droid.Resource.Styleable.Motion_quantizeMotionInterpolator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Motion_quantizeMotionPhase = global::pjsua2xamarin.Droid.Resource.Styleable.Motion_quantizeMotionPhase;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Motion_quantizeMotionSteps = global::pjsua2xamarin.Droid.Resource.Styleable.Motion_quantizeMotionSteps;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Motion_transitionEasing = global::pjsua2xamarin.Droid.Resource.Styleable.Motion_transitionEasing;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavAction = global::pjsua2xamarin.Droid.Resource.Styleable.NavAction;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavAction_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.NavAction_android_id;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavAction_destination = global::pjsua2xamarin.Droid.Resource.Styleable.NavAction_destination;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavAction_enterAnim = global::pjsua2xamarin.Droid.Resource.Styleable.NavAction_enterAnim;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavAction_exitAnim = global::pjsua2xamarin.Droid.Resource.Styleable.NavAction_exitAnim;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavAction_launchSingleTop = global::pjsua2xamarin.Droid.Resource.Styleable.NavAction_launchSingleTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavAction_popEnterAnim = global::pjsua2xamarin.Droid.Resource.Styleable.NavAction_popEnterAnim;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavAction_popExitAnim = global::pjsua2xamarin.Droid.Resource.Styleable.NavAction_popExitAnim;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavAction_popUpTo = global::pjsua2xamarin.Droid.Resource.Styleable.NavAction_popUpTo;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavAction_popUpToInclusive = global::pjsua2xamarin.Droid.Resource.Styleable.NavAction_popUpToInclusive;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavArgument = global::pjsua2xamarin.Droid.Resource.Styleable.NavArgument;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavArgument_android_defaultValue = global::pjsua2xamarin.Droid.Resource.Styleable.NavArgument_android_defaultValue;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavArgument_android_name = global::pjsua2xamarin.Droid.Resource.Styleable.NavArgument_android_name;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavArgument_argType = global::pjsua2xamarin.Droid.Resource.Styleable.NavArgument_argType;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavArgument_nullable = global::pjsua2xamarin.Droid.Resource.Styleable.NavArgument_nullable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavDeepLink = global::pjsua2xamarin.Droid.Resource.Styleable.NavDeepLink;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavDeepLink_action = global::pjsua2xamarin.Droid.Resource.Styleable.NavDeepLink_action;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavDeepLink_android_autoVerify = global::pjsua2xamarin.Droid.Resource.Styleable.NavDeepLink_android_autoVerify;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavDeepLink_mimeType = global::pjsua2xamarin.Droid.Resource.Styleable.NavDeepLink_mimeType;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavDeepLink_uri = global::pjsua2xamarin.Droid.Resource.Styleable.NavDeepLink_uri;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavGraphNavigator = global::pjsua2xamarin.Droid.Resource.Styleable.NavGraphNavigator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavGraphNavigator_startDestination = global::pjsua2xamarin.Droid.Resource.Styleable.NavGraphNavigator_startDestination;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavHost = global::pjsua2xamarin.Droid.Resource.Styleable.NavHost;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavHost_navGraph = global::pjsua2xamarin.Droid.Resource.Styleable.NavHost_navGraph;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationBarView = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationBarView_backgroundTint = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView_backgroundTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationBarView_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationBarView_itemBackground = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView_itemBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationBarView_itemIconSize = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView_itemIconSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationBarView_itemIconTint = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView_itemIconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationBarView_itemRippleColor = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView_itemRippleColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationBarView_itemTextAppearanceActive = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView_itemTextAppearanceActive;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationBarView_itemTextAppearanceInactive = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView_itemTextAppearanceInactive;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationBarView_itemTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView_itemTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationBarView_labelVisibilityMode = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView_labelVisibilityMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationBarView_menu = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView_menu;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationRailView = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationRailView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationRailView_headerLayout = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationRailView_headerLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationRailView_menuGravity = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationRailView_menuGravity;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_android_background = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_android_background;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_android_fitsSystemWindows = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_android_fitsSystemWindows;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_android_maxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_android_maxWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_headerLayout = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_headerLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_itemBackground = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_itemHorizontalPadding = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemHorizontalPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_itemIconPadding = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemIconPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_itemIconSize = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemIconSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_itemIconTint = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemIconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_itemMaxLines = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemMaxLines;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_itemShapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemShapeAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_itemShapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemShapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_itemShapeFillColor = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemShapeFillColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_itemShapeInsetBottom = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemShapeInsetBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_itemShapeInsetEnd = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemShapeInsetEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_itemShapeInsetStart = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemShapeInsetStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_itemShapeInsetTop = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemShapeInsetTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_itemTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_itemTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_menu = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_menu;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_shapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_shapeAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavigationView_shapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_shapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Navigator = global::pjsua2xamarin.Droid.Resource.Styleable.Navigator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Navigator_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.Navigator_android_id;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Navigator_android_label = global::pjsua2xamarin.Droid.Resource.Styleable.Navigator_android_label;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavInclude = global::pjsua2xamarin.Droid.Resource.Styleable.NavInclude;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.NavInclude_graph = global::pjsua2xamarin.Droid.Resource.Styleable.NavInclude_graph;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnClick = global::pjsua2xamarin.Droid.Resource.Styleable.OnClick;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnClick_clickAction = global::pjsua2xamarin.Droid.Resource.Styleable.OnClick_clickAction;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnClick_targetId = global::pjsua2xamarin.Droid.Resource.Styleable.OnClick_targetId;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnSwipe = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnSwipe_autoCompleteMode = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_autoCompleteMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnSwipe_dragDirection = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_dragDirection;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnSwipe_dragScale = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_dragScale;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnSwipe_dragThreshold = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_dragThreshold;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnSwipe_limitBoundsTo = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_limitBoundsTo;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnSwipe_maxAcceleration = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_maxAcceleration;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnSwipe_maxVelocity = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_maxVelocity;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnSwipe_moveWhenScrollAtTop = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_moveWhenScrollAtTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnSwipe_nestedScrollFlags = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_nestedScrollFlags;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnSwipe_onTouchUp = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_onTouchUp;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnSwipe_rotationCenterId = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_rotationCenterId;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnSwipe_springBoundary = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_springBoundary;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnSwipe_springDamping = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_springDamping;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnSwipe_springMass = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_springMass;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnSwipe_springStiffness = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_springStiffness;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnSwipe_springStopThreshold = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_springStopThreshold;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnSwipe_touchAnchorId = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_touchAnchorId;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnSwipe_touchAnchorSide = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_touchAnchorSide;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.OnSwipe_touchRegionId = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_touchRegionId;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.PopupWindow = global::pjsua2xamarin.Droid.Resource.Styleable.PopupWindow;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.PopupWindowBackgroundState = global::pjsua2xamarin.Droid.Resource.Styleable.PopupWindowBackgroundState;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.PopupWindowBackgroundState_state_above_anchor = global::pjsua2xamarin.Droid.Resource.Styleable.PopupWindowBackgroundState_state_above_anchor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.PopupWindow_android_popupAnimationStyle = global::pjsua2xamarin.Droid.Resource.Styleable.PopupWindow_android_popupAnimationStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.PopupWindow_android_popupBackground = global::pjsua2xamarin.Droid.Resource.Styleable.PopupWindow_android_popupBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.PopupWindow_overlapAnchor = global::pjsua2xamarin.Droid.Resource.Styleable.PopupWindow_overlapAnchor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.PropertySet = global::pjsua2xamarin.Droid.Resource.Styleable.PropertySet;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.PropertySet_android_alpha = global::pjsua2xamarin.Droid.Resource.Styleable.PropertySet_android_alpha;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.PropertySet_android_visibility = global::pjsua2xamarin.Droid.Resource.Styleable.PropertySet_android_visibility;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.PropertySet_layout_constraintTag = global::pjsua2xamarin.Droid.Resource.Styleable.PropertySet_layout_constraintTag;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.PropertySet_motionProgress = global::pjsua2xamarin.Droid.Resource.Styleable.PropertySet_motionProgress;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.PropertySet_visibilityMode = global::pjsua2xamarin.Droid.Resource.Styleable.PropertySet_visibilityMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.RadialViewGroup = global::pjsua2xamarin.Droid.Resource.Styleable.RadialViewGroup;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.RadialViewGroup_materialCircleRadius = global::pjsua2xamarin.Droid.Resource.Styleable.RadialViewGroup_materialCircleRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.RangeSlider = global::pjsua2xamarin.Droid.Resource.Styleable.RangeSlider;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.RangeSlider_minSeparation = global::pjsua2xamarin.Droid.Resource.Styleable.RangeSlider_minSeparation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.RangeSlider_values = global::pjsua2xamarin.Droid.Resource.Styleable.RangeSlider_values;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.RecycleListView = global::pjsua2xamarin.Droid.Resource.Styleable.RecycleListView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.RecycleListView_paddingBottomNoButtons = global::pjsua2xamarin.Droid.Resource.Styleable.RecycleListView_paddingBottomNoButtons;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.RecycleListView_paddingTopNoTitle = global::pjsua2xamarin.Droid.Resource.Styleable.RecycleListView_paddingTopNoTitle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.RecyclerView = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.RecyclerView_android_clipToPadding = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_android_clipToPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.RecyclerView_android_descendantFocusability = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_android_descendantFocusability;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.RecyclerView_android_orientation = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_android_orientation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.RecyclerView_fastScrollEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_fastScrollEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.RecyclerView_fastScrollHorizontalThumbDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_fastScrollHorizontalThumbDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.RecyclerView_fastScrollHorizontalTrackDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_fastScrollHorizontalTrackDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.RecyclerView_fastScrollVerticalThumbDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_fastScrollVerticalThumbDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.RecyclerView_fastScrollVerticalTrackDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_fastScrollVerticalTrackDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.RecyclerView_layoutManager = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_layoutManager;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.RecyclerView_reverseLayout = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_reverseLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.RecyclerView_spanCount = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_spanCount;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.RecyclerView_stackFromEnd = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_stackFromEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ScrimInsetsFrameLayout = global::pjsua2xamarin.Droid.Resource.Styleable.ScrimInsetsFrameLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ScrimInsetsFrameLayout_insetForeground = global::pjsua2xamarin.Droid.Resource.Styleable.ScrimInsetsFrameLayout_insetForeground;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ScrollingViewBehavior_Layout = global::pjsua2xamarin.Droid.Resource.Styleable.ScrollingViewBehavior_Layout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ScrollingViewBehavior_Layout_behavior_overlapTop = global::pjsua2xamarin.Droid.Resource.Styleable.ScrollingViewBehavior_Layout_behavior_overlapTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ScrollViewRendererTheme = global::pjsua2xamarin.Droid.Resource.Styleable.ScrollViewRendererTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ScrollViewRendererTheme_scrollViewStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ScrollViewRendererTheme_scrollViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SearchView = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SearchView_android_focusable = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_android_focusable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SearchView_android_imeOptions = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_android_imeOptions;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SearchView_android_inputType = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_android_inputType;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SearchView_android_maxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_android_maxWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SearchView_closeIcon = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_closeIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SearchView_commitIcon = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_commitIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SearchView_defaultQueryHint = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_defaultQueryHint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SearchView_goIcon = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_goIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SearchView_iconifiedByDefault = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_iconifiedByDefault;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SearchView_layout = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_layout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SearchView_queryBackground = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_queryBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SearchView_queryHint = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_queryHint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SearchView_searchHintIcon = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_searchHintIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SearchView_searchIcon = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_searchIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SearchView_submitBackground = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_submitBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SearchView_suggestionRowLayout = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_suggestionRowLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SearchView_voiceIcon = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_voiceIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeableImageView = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeableImageView_contentPadding = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView_contentPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeableImageView_contentPaddingBottom = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView_contentPaddingBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeableImageView_contentPaddingEnd = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView_contentPaddingEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeableImageView_contentPaddingLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView_contentPaddingLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeableImageView_contentPaddingRight = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView_contentPaddingRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeableImageView_contentPaddingStart = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView_contentPaddingStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeableImageView_contentPaddingTop = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView_contentPaddingTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeableImageView_shapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView_shapeAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeableImageView_shapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView_shapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeableImageView_strokeColor = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView_strokeColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeableImageView_strokeWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView_strokeWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeAppearance_cornerFamily = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeAppearance_cornerFamily;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeAppearance_cornerFamilyBottomLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeAppearance_cornerFamilyBottomLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeAppearance_cornerFamilyBottomRight = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeAppearance_cornerFamilyBottomRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeAppearance_cornerFamilyTopLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeAppearance_cornerFamilyTopLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeAppearance_cornerFamilyTopRight = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeAppearance_cornerFamilyTopRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeAppearance_cornerSize = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeAppearance_cornerSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeAppearance_cornerSizeBottomLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeAppearance_cornerSizeBottomLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeAppearance_cornerSizeBottomRight = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeAppearance_cornerSizeBottomRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeAppearance_cornerSizeTopLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeAppearance_cornerSizeTopLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ShapeAppearance_cornerSizeTopRight = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeAppearance_cornerSizeTopRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider = global::pjsua2xamarin.Droid.Resource.Styleable.Slider;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider_android_enabled = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_android_enabled;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider_android_stepSize = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_android_stepSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider_android_value = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_android_value;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider_android_valueFrom = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_android_valueFrom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider_android_valueTo = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_android_valueTo;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider_haloColor = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_haloColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider_haloRadius = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_haloRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider_labelBehavior = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_labelBehavior;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider_labelStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_labelStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider_thumbColor = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_thumbColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider_thumbElevation = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_thumbElevation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider_thumbRadius = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_thumbRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider_thumbStrokeColor = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_thumbStrokeColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider_thumbStrokeWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_thumbStrokeWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider_tickColor = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_tickColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider_tickColorActive = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_tickColorActive;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider_tickColorInactive = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_tickColorInactive;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider_tickVisible = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_tickVisible;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider_trackColor = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_trackColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider_trackColorActive = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_trackColorActive;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider_trackColorInactive = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_trackColorInactive;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Slider_trackHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_trackHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Snackbar = global::pjsua2xamarin.Droid.Resource.Styleable.Snackbar;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SnackbarLayout = global::pjsua2xamarin.Droid.Resource.Styleable.SnackbarLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SnackbarLayout_actionTextColorAlpha = global::pjsua2xamarin.Droid.Resource.Styleable.SnackbarLayout_actionTextColorAlpha;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SnackbarLayout_android_maxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.SnackbarLayout_android_maxWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SnackbarLayout_animationMode = global::pjsua2xamarin.Droid.Resource.Styleable.SnackbarLayout_animationMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SnackbarLayout_backgroundOverlayColorAlpha = global::pjsua2xamarin.Droid.Resource.Styleable.SnackbarLayout_backgroundOverlayColorAlpha;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SnackbarLayout_backgroundTint = global::pjsua2xamarin.Droid.Resource.Styleable.SnackbarLayout_backgroundTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SnackbarLayout_backgroundTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.SnackbarLayout_backgroundTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SnackbarLayout_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.SnackbarLayout_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SnackbarLayout_maxActionInlineWidth = global::pjsua2xamarin.Droid.Resource.Styleable.SnackbarLayout_maxActionInlineWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Snackbar_snackbarButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Snackbar_snackbarButtonStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Snackbar_snackbarStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Snackbar_snackbarStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Snackbar_snackbarTextViewStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Snackbar_snackbarTextViewStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Spinner = global::pjsua2xamarin.Droid.Resource.Styleable.Spinner;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Spinner_android_dropDownWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Spinner_android_dropDownWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Spinner_android_entries = global::pjsua2xamarin.Droid.Resource.Styleable.Spinner_android_entries;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Spinner_android_popupBackground = global::pjsua2xamarin.Droid.Resource.Styleable.Spinner_android_popupBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Spinner_android_prompt = global::pjsua2xamarin.Droid.Resource.Styleable.Spinner_android_prompt;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Spinner_popupTheme = global::pjsua2xamarin.Droid.Resource.Styleable.Spinner_popupTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.State = global::pjsua2xamarin.Droid.Resource.Styleable.State;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.StateListDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.StateListDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.StateListDrawableItem = global::pjsua2xamarin.Droid.Resource.Styleable.StateListDrawableItem;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.StateListDrawableItem_android_drawable = global::pjsua2xamarin.Droid.Resource.Styleable.StateListDrawableItem_android_drawable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.StateListDrawable_android_constantSize = global::pjsua2xamarin.Droid.Resource.Styleable.StateListDrawable_android_constantSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.StateListDrawable_android_dither = global::pjsua2xamarin.Droid.Resource.Styleable.StateListDrawable_android_dither;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.StateListDrawable_android_enterFadeDuration = global::pjsua2xamarin.Droid.Resource.Styleable.StateListDrawable_android_enterFadeDuration;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.StateListDrawable_android_exitFadeDuration = global::pjsua2xamarin.Droid.Resource.Styleable.StateListDrawable_android_exitFadeDuration;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.StateListDrawable_android_variablePadding = global::pjsua2xamarin.Droid.Resource.Styleable.StateListDrawable_android_variablePadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.StateListDrawable_android_visible = global::pjsua2xamarin.Droid.Resource.Styleable.StateListDrawable_android_visible;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.StateSet = global::pjsua2xamarin.Droid.Resource.Styleable.StateSet;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.StateSet_defaultState = global::pjsua2xamarin.Droid.Resource.Styleable.StateSet_defaultState;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.State_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.State_android_id;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.State_constraints = global::pjsua2xamarin.Droid.Resource.Styleable.State_constraints;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SwipeRefreshLayout = global::pjsua2xamarin.Droid.Resource.Styleable.SwipeRefreshLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SwipeRefreshLayout_swipeRefreshLayoutProgressSpinnerBackgroundColor = global::pjsua2xamarin.Droid.Resource.Styleable.SwipeRefreshLayout_swipeRefreshLayoutProgressSpinnerBackgroundColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SwitchCompat = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SwitchCompat_android_textOff = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_android_textOff;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SwitchCompat_android_textOn = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_android_textOn;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SwitchCompat_android_thumb = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_android_thumb;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SwitchCompat_showText = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_showText;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SwitchCompat_splitTrack = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_splitTrack;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SwitchCompat_switchMinWidth = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_switchMinWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SwitchCompat_switchPadding = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_switchPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SwitchCompat_switchTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_switchTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SwitchCompat_thumbTextPadding = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_thumbTextPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SwitchCompat_thumbTint = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_thumbTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SwitchCompat_thumbTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_thumbTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SwitchCompat_track = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_track;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SwitchCompat_trackTint = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_trackTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SwitchCompat_trackTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_trackTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SwitchMaterial = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchMaterial;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.SwitchMaterial_useMaterialThemeColors = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchMaterial_useMaterialThemeColors;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabItem = global::pjsua2xamarin.Droid.Resource.Styleable.TabItem;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabItem_android_icon = global::pjsua2xamarin.Droid.Resource.Styleable.TabItem_android_icon;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabItem_android_layout = global::pjsua2xamarin.Droid.Resource.Styleable.TabItem_android_layout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabItem_android_text = global::pjsua2xamarin.Droid.Resource.Styleable.TabItem_android_text;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabBackground = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabBackground;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabContentStart = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabContentStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabGravity = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabGravity;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabIconTint = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabIconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabIconTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabIconTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabIndicator = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabIndicator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabIndicatorAnimationDuration = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabIndicatorAnimationDuration;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabIndicatorAnimationMode = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabIndicatorAnimationMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabIndicatorColor = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabIndicatorColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabIndicatorFullWidth = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabIndicatorFullWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabIndicatorGravity = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabIndicatorGravity;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabIndicatorHeight = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabIndicatorHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabInlineLabel = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabInlineLabel;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabMaxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabMaxWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabMinWidth = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabMinWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabMode = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabPadding = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabPadding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabPaddingBottom = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabPaddingBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabPaddingEnd = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabPaddingEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabPaddingStart = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabPaddingStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabPaddingTop = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabPaddingTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabRippleColor = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabRippleColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabSelectedTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabSelectedTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TabLayout_tabUnboundedRipple = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabUnboundedRipple;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextAppearance_android_fontFamily = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_fontFamily;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextAppearance_android_shadowColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_shadowColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextAppearance_android_shadowDx = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_shadowDx;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextAppearance_android_shadowDy = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_shadowDy;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextAppearance_android_shadowRadius = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_shadowRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextAppearance_android_textColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_textColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextAppearance_android_textColorHint = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_textColorHint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextAppearance_android_textColorLink = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_textColorLink;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextAppearance_android_textFontWeight = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_textFontWeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextAppearance_android_textSize = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_textSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextAppearance_android_textStyle = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_textStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextAppearance_android_typeface = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_typeface;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextAppearance_fontFamily = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_fontFamily;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextAppearance_fontVariationSettings = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_fontVariationSettings;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextAppearance_textAllCaps = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_textAllCaps;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextAppearance_textLocale = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_textLocale;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextEffects = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextEffects_android_fontFamily = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_android_fontFamily;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextEffects_android_shadowColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_android_shadowColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextEffects_android_shadowDx = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_android_shadowDx;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextEffects_android_shadowDy = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_android_shadowDy;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextEffects_android_shadowRadius = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_android_shadowRadius;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextEffects_android_text = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_android_text;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextEffects_android_textSize = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_android_textSize;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextEffects_android_textStyle = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_android_textStyle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextEffects_android_typeface = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_android_typeface;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextEffects_borderRound = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_borderRound;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextEffects_borderRoundPercent = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_borderRoundPercent;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextEffects_textFillColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_textFillColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextEffects_textOutlineColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_textOutlineColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextEffects_textOutlineThickness = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_textOutlineThickness;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputEditText = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputEditText;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputEditText_textInputLayoutFocusedRectEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputEditText_textInputLayoutFocusedRectEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_android_enabled = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_android_enabled;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_android_hint = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_android_hint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_android_maxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_android_maxWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_android_minWidth = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_android_minWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_android_textColorHint = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_android_textColorHint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_boxBackgroundColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_boxBackgroundColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_boxBackgroundMode = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_boxBackgroundMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_boxCollapsedPaddingTop = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_boxCollapsedPaddingTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_boxCornerRadiusBottomEnd = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_boxCornerRadiusBottomEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_boxCornerRadiusBottomStart = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_boxCornerRadiusBottomStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_boxCornerRadiusTopEnd = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_boxCornerRadiusTopEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_boxCornerRadiusTopStart = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_boxCornerRadiusTopStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_boxStrokeColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_boxStrokeColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_boxStrokeErrorColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_boxStrokeErrorColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_boxStrokeWidth = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_boxStrokeWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_boxStrokeWidthFocused = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_boxStrokeWidthFocused;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_counterEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_counterEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_counterMaxLength = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_counterMaxLength;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_counterOverflowTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_counterOverflowTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_counterOverflowTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_counterOverflowTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_counterTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_counterTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_counterTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_counterTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_endIconCheckable = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_endIconCheckable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_endIconContentDescription = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_endIconContentDescription;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_endIconDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_endIconDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_endIconMode = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_endIconMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_endIconTint = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_endIconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_endIconTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_endIconTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_errorContentDescription = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_errorContentDescription;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_errorEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_errorEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_errorIconDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_errorIconDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_errorIconTint = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_errorIconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_errorIconTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_errorIconTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_errorTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_errorTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_errorTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_errorTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_expandedHintEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_expandedHintEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_helperText = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_helperText;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_helperTextEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_helperTextEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_helperTextTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_helperTextTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_helperTextTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_helperTextTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_hintAnimationEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_hintAnimationEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_hintEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_hintEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_hintTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_hintTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_hintTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_hintTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_passwordToggleContentDescription = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_passwordToggleContentDescription;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_passwordToggleDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_passwordToggleDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_passwordToggleEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_passwordToggleEnabled;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_passwordToggleTint = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_passwordToggleTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_passwordToggleTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_passwordToggleTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_placeholderText = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_placeholderText;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_placeholderTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_placeholderTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_placeholderTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_placeholderTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_prefixText = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_prefixText;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_prefixTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_prefixTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_prefixTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_prefixTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_shapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_shapeAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_shapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_shapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_startIconCheckable = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_startIconCheckable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_startIconContentDescription = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_startIconContentDescription;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_startIconDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_startIconDrawable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_startIconTint = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_startIconTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_startIconTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_startIconTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_suffixText = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_suffixText;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_suffixTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_suffixTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.TextInputLayout_suffixTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_suffixTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ThemeEnforcement = global::pjsua2xamarin.Droid.Resource.Styleable.ThemeEnforcement;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ThemeEnforcement_android_textAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.ThemeEnforcement_android_textAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ThemeEnforcement_enforceMaterialTheme = global::pjsua2xamarin.Droid.Resource.Styleable.ThemeEnforcement_enforceMaterialTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ThemeEnforcement_enforceTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.ThemeEnforcement_enforceTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_android_gravity = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_android_gravity;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_android_minHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_android_minHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_buttonGravity = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_buttonGravity;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_collapseContentDescription = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_collapseContentDescription;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_collapseIcon = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_collapseIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_contentInsetEnd = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_contentInsetEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_contentInsetEndWithActions = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_contentInsetEndWithActions;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_contentInsetLeft = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_contentInsetLeft;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_contentInsetRight = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_contentInsetRight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_contentInsetStart = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_contentInsetStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_contentInsetStartWithNavigation = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_contentInsetStartWithNavigation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_logo = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_logo;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_logoDescription = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_logoDescription;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_maxButtonHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_maxButtonHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_menu = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_menu;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_navigationContentDescription = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_navigationContentDescription;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_navigationIcon = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_navigationIcon;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_popupTheme = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_popupTheme;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_subtitle = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_subtitle;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_subtitleTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_subtitleTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_subtitleTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_subtitleTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_title = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_title;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_titleMargin = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_titleMargin;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_titleMarginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_titleMarginBottom;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_titleMarginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_titleMarginEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_titleMargins = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_titleMargins;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_titleMarginStart = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_titleMarginStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_titleMarginTop = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_titleMarginTop;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_titleTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_titleTextAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Toolbar_titleTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_titleTextColor;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Tooltip = global::pjsua2xamarin.Droid.Resource.Styleable.Tooltip;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Tooltip_android_layout_margin = global::pjsua2xamarin.Droid.Resource.Styleable.Tooltip_android_layout_margin;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Tooltip_android_minHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Tooltip_android_minHeight;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Tooltip_android_minWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Tooltip_android_minWidth;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Tooltip_android_padding = global::pjsua2xamarin.Droid.Resource.Styleable.Tooltip_android_padding;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Tooltip_android_text = global::pjsua2xamarin.Droid.Resource.Styleable.Tooltip_android_text;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Tooltip_android_textAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.Tooltip_android_textAppearance;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Tooltip_backgroundTint = global::pjsua2xamarin.Droid.Resource.Styleable.Tooltip_backgroundTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transform = global::pjsua2xamarin.Droid.Resource.Styleable.Transform;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transform_android_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_android_elevation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transform_android_rotation = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_android_rotation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transform_android_rotationX = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_android_rotationX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transform_android_rotationY = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_android_rotationY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transform_android_scaleX = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_android_scaleX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transform_android_scaleY = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_android_scaleY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transform_android_transformPivotX = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_android_transformPivotX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transform_android_transformPivotY = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_android_transformPivotY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transform_android_translationX = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_android_translationX;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transform_android_translationY = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_android_translationY;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transform_android_translationZ = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_android_translationZ;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transform_transformPivotTarget = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_transformPivotTarget;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transition = global::pjsua2xamarin.Droid.Resource.Styleable.Transition;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transition_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.Transition_android_id;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transition_autoTransition = global::pjsua2xamarin.Droid.Resource.Styleable.Transition_autoTransition;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transition_constraintSetEnd = global::pjsua2xamarin.Droid.Resource.Styleable.Transition_constraintSetEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transition_constraintSetStart = global::pjsua2xamarin.Droid.Resource.Styleable.Transition_constraintSetStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transition_duration = global::pjsua2xamarin.Droid.Resource.Styleable.Transition_duration;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transition_layoutDuringTransition = global::pjsua2xamarin.Droid.Resource.Styleable.Transition_layoutDuringTransition;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transition_motionInterpolator = global::pjsua2xamarin.Droid.Resource.Styleable.Transition_motionInterpolator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transition_pathMotionArc = global::pjsua2xamarin.Droid.Resource.Styleable.Transition_pathMotionArc;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transition_staggered = global::pjsua2xamarin.Droid.Resource.Styleable.Transition_staggered;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transition_transitionDisable = global::pjsua2xamarin.Droid.Resource.Styleable.Transition_transitionDisable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Transition_transitionFlags = global::pjsua2xamarin.Droid.Resource.Styleable.Transition_transitionFlags;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Variant = global::pjsua2xamarin.Droid.Resource.Styleable.Variant;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Variant_constraints = global::pjsua2xamarin.Droid.Resource.Styleable.Variant_constraints;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Variant_region_heightLessThan = global::pjsua2xamarin.Droid.Resource.Styleable.Variant_region_heightLessThan;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Variant_region_heightMoreThan = global::pjsua2xamarin.Droid.Resource.Styleable.Variant_region_heightMoreThan;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Variant_region_widthLessThan = global::pjsua2xamarin.Droid.Resource.Styleable.Variant_region_widthLessThan;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.Variant_region_widthMoreThan = global::pjsua2xamarin.Droid.Resource.Styleable.Variant_region_widthMoreThan;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.View = global::pjsua2xamarin.Droid.Resource.Styleable.View;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewBackgroundHelper = global::pjsua2xamarin.Droid.Resource.Styleable.ViewBackgroundHelper;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewBackgroundHelper_android_background = global::pjsua2xamarin.Droid.Resource.Styleable.ViewBackgroundHelper_android_background;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewBackgroundHelper_backgroundTint = global::pjsua2xamarin.Droid.Resource.Styleable.ViewBackgroundHelper_backgroundTint;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewBackgroundHelper_backgroundTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.ViewBackgroundHelper_backgroundTintMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewPager2 = global::pjsua2xamarin.Droid.Resource.Styleable.ViewPager2;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewPager2_android_orientation = global::pjsua2xamarin.Droid.Resource.Styleable.ViewPager2_android_orientation;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewStubCompat = global::pjsua2xamarin.Droid.Resource.Styleable.ViewStubCompat;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewStubCompat_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.ViewStubCompat_android_id;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewStubCompat_android_inflatedId = global::pjsua2xamarin.Droid.Resource.Styleable.ViewStubCompat_android_inflatedId;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewStubCompat_android_layout = global::pjsua2xamarin.Droid.Resource.Styleable.ViewStubCompat_android_layout;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewTransition = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewTransition_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_android_id;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewTransition_clearsTag = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_clearsTag;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewTransition_duration = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_duration;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewTransition_ifTagNotSet = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_ifTagNotSet;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewTransition_ifTagSet = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_ifTagSet;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewTransition_motionInterpolator = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_motionInterpolator;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewTransition_motionTarget = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_motionTarget;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewTransition_onStateTransition = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_onStateTransition;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewTransition_pathMotionArc = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_pathMotionArc;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewTransition_setsTag = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_setsTag;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewTransition_SharedValue = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_SharedValue;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewTransition_SharedValueId = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_SharedValueId;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewTransition_transitionDisable = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_transitionDisable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewTransition_upDuration = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_upDuration;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.ViewTransition_viewTransitionMode = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_viewTransitionMode;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.View_android_focusable = global::pjsua2xamarin.Droid.Resource.Styleable.View_android_focusable;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.View_android_theme = global::pjsua2xamarin.Droid.Resource.Styleable.View_android_theme;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.View_paddingEnd = global::pjsua2xamarin.Droid.Resource.Styleable.View_paddingEnd;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.View_paddingStart = global::pjsua2xamarin.Droid.Resource.Styleable.View_paddingStart;
			global::Xamarin.Forms.Platform.Android.Resource.Styleable.View_theme = global::pjsua2xamarin.Droid.Resource.Styleable.View_theme;
			global::Xamarin.Forms.Platform.Android.Resource.Xml.standalone_badge = global::pjsua2xamarin.Droid.Resource.Xml.standalone_badge;
			global::Xamarin.Forms.Platform.Android.Resource.Xml.standalone_badge_gravity_bottom_end = global::pjsua2xamarin.Droid.Resource.Xml.standalone_badge_gravity_bottom_end;
			global::Xamarin.Forms.Platform.Android.Resource.Xml.standalone_badge_gravity_bottom_start = global::pjsua2xamarin.Droid.Resource.Xml.standalone_badge_gravity_bottom_start;
			global::Xamarin.Forms.Platform.Android.Resource.Xml.standalone_badge_gravity_top_start = global::pjsua2xamarin.Droid.Resource.Xml.standalone_badge_gravity_top_start;
			global::Xamarin.Forms.Platform.Android.Resource.Xml.standalone_badge_offset = global::pjsua2xamarin.Droid.Resource.Xml.standalone_badge_offset;
			global::Xamarin.Forms.Platform.Resource.Animation.abc_fade_in = global::pjsua2xamarin.Droid.Resource.Animation.abc_fade_in;
			global::Xamarin.Forms.Platform.Resource.Animation.abc_fade_out = global::pjsua2xamarin.Droid.Resource.Animation.abc_fade_out;
			global::Xamarin.Forms.Platform.Resource.Animation.abc_grow_fade_in_from_bottom = global::pjsua2xamarin.Droid.Resource.Animation.abc_grow_fade_in_from_bottom;
			global::Xamarin.Forms.Platform.Resource.Animation.abc_popup_enter = global::pjsua2xamarin.Droid.Resource.Animation.abc_popup_enter;
			global::Xamarin.Forms.Platform.Resource.Animation.abc_popup_exit = global::pjsua2xamarin.Droid.Resource.Animation.abc_popup_exit;
			global::Xamarin.Forms.Platform.Resource.Animation.abc_shrink_fade_out_from_bottom = global::pjsua2xamarin.Droid.Resource.Animation.abc_shrink_fade_out_from_bottom;
			global::Xamarin.Forms.Platform.Resource.Animation.abc_slide_in_bottom = global::pjsua2xamarin.Droid.Resource.Animation.abc_slide_in_bottom;
			global::Xamarin.Forms.Platform.Resource.Animation.abc_slide_in_top = global::pjsua2xamarin.Droid.Resource.Animation.abc_slide_in_top;
			global::Xamarin.Forms.Platform.Resource.Animation.abc_slide_out_bottom = global::pjsua2xamarin.Droid.Resource.Animation.abc_slide_out_bottom;
			global::Xamarin.Forms.Platform.Resource.Animation.abc_slide_out_top = global::pjsua2xamarin.Droid.Resource.Animation.abc_slide_out_top;
			global::Xamarin.Forms.Platform.Resource.Animation.abc_tooltip_enter = global::pjsua2xamarin.Droid.Resource.Animation.abc_tooltip_enter;
			global::Xamarin.Forms.Platform.Resource.Animation.abc_tooltip_exit = global::pjsua2xamarin.Droid.Resource.Animation.abc_tooltip_exit;
			global::Xamarin.Forms.Platform.Resource.Animation.btn_checkbox_to_checked_box_inner_merged_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_checkbox_to_checked_box_inner_merged_animation;
			global::Xamarin.Forms.Platform.Resource.Animation.btn_checkbox_to_checked_box_outer_merged_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_checkbox_to_checked_box_outer_merged_animation;
			global::Xamarin.Forms.Platform.Resource.Animation.btn_checkbox_to_checked_icon_null_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_checkbox_to_checked_icon_null_animation;
			global::Xamarin.Forms.Platform.Resource.Animation.btn_checkbox_to_unchecked_box_inner_merged_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_checkbox_to_unchecked_box_inner_merged_animation;
			global::Xamarin.Forms.Platform.Resource.Animation.btn_checkbox_to_unchecked_check_path_merged_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_checkbox_to_unchecked_check_path_merged_animation;
			global::Xamarin.Forms.Platform.Resource.Animation.btn_checkbox_to_unchecked_icon_null_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_checkbox_to_unchecked_icon_null_animation;
			global::Xamarin.Forms.Platform.Resource.Animation.btn_radio_to_off_mtrl_dot_group_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_radio_to_off_mtrl_dot_group_animation;
			global::Xamarin.Forms.Platform.Resource.Animation.btn_radio_to_off_mtrl_ring_outer_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_radio_to_off_mtrl_ring_outer_animation;
			global::Xamarin.Forms.Platform.Resource.Animation.btn_radio_to_off_mtrl_ring_outer_path_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_radio_to_off_mtrl_ring_outer_path_animation;
			global::Xamarin.Forms.Platform.Resource.Animation.btn_radio_to_on_mtrl_dot_group_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_radio_to_on_mtrl_dot_group_animation;
			global::Xamarin.Forms.Platform.Resource.Animation.btn_radio_to_on_mtrl_ring_outer_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_radio_to_on_mtrl_ring_outer_animation;
			global::Xamarin.Forms.Platform.Resource.Animation.btn_radio_to_on_mtrl_ring_outer_path_animation = global::pjsua2xamarin.Droid.Resource.Animation.btn_radio_to_on_mtrl_ring_outer_path_animation;
			global::Xamarin.Forms.Platform.Resource.Animation.design_bottom_sheet_slide_in = global::pjsua2xamarin.Droid.Resource.Animation.design_bottom_sheet_slide_in;
			global::Xamarin.Forms.Platform.Resource.Animation.design_bottom_sheet_slide_out = global::pjsua2xamarin.Droid.Resource.Animation.design_bottom_sheet_slide_out;
			global::Xamarin.Forms.Platform.Resource.Animation.design_snackbar_in = global::pjsua2xamarin.Droid.Resource.Animation.design_snackbar_in;
			global::Xamarin.Forms.Platform.Resource.Animation.design_snackbar_out = global::pjsua2xamarin.Droid.Resource.Animation.design_snackbar_out;
			global::Xamarin.Forms.Platform.Resource.Animation.EnterFromLeft = global::pjsua2xamarin.Droid.Resource.Animation.EnterFromLeft;
			global::Xamarin.Forms.Platform.Resource.Animation.EnterFromRight = global::pjsua2xamarin.Droid.Resource.Animation.EnterFromRight;
			global::Xamarin.Forms.Platform.Resource.Animation.ExitToLeft = global::pjsua2xamarin.Droid.Resource.Animation.ExitToLeft;
			global::Xamarin.Forms.Platform.Resource.Animation.ExitToRight = global::pjsua2xamarin.Droid.Resource.Animation.ExitToRight;
			global::Xamarin.Forms.Platform.Resource.Animation.fragment_fast_out_extra_slow_in = global::pjsua2xamarin.Droid.Resource.Animation.fragment_fast_out_extra_slow_in;
			global::Xamarin.Forms.Platform.Resource.Animation.mtrl_bottom_sheet_slide_in = global::pjsua2xamarin.Droid.Resource.Animation.mtrl_bottom_sheet_slide_in;
			global::Xamarin.Forms.Platform.Resource.Animation.mtrl_bottom_sheet_slide_out = global::pjsua2xamarin.Droid.Resource.Animation.mtrl_bottom_sheet_slide_out;
			global::Xamarin.Forms.Platform.Resource.Animation.mtrl_card_lowers_interpolator = global::pjsua2xamarin.Droid.Resource.Animation.mtrl_card_lowers_interpolator;
			global::Xamarin.Forms.Platform.Resource.Animation.nav_default_enter_anim = global::pjsua2xamarin.Droid.Resource.Animation.nav_default_enter_anim;
			global::Xamarin.Forms.Platform.Resource.Animation.nav_default_exit_anim = global::pjsua2xamarin.Droid.Resource.Animation.nav_default_exit_anim;
			global::Xamarin.Forms.Platform.Resource.Animation.nav_default_pop_enter_anim = global::pjsua2xamarin.Droid.Resource.Animation.nav_default_pop_enter_anim;
			global::Xamarin.Forms.Platform.Resource.Animation.nav_default_pop_exit_anim = global::pjsua2xamarin.Droid.Resource.Animation.nav_default_pop_exit_anim;
			global::Xamarin.Forms.Platform.Resource.Animator.design_appbar_state_list_animator = global::pjsua2xamarin.Droid.Resource.Animator.design_appbar_state_list_animator;
			global::Xamarin.Forms.Platform.Resource.Animator.design_fab_hide_motion_spec = global::pjsua2xamarin.Droid.Resource.Animator.design_fab_hide_motion_spec;
			global::Xamarin.Forms.Platform.Resource.Animator.design_fab_show_motion_spec = global::pjsua2xamarin.Droid.Resource.Animator.design_fab_show_motion_spec;
			global::Xamarin.Forms.Platform.Resource.Animator.fragment_close_enter = global::pjsua2xamarin.Droid.Resource.Animator.fragment_close_enter;
			global::Xamarin.Forms.Platform.Resource.Animator.fragment_close_exit = global::pjsua2xamarin.Droid.Resource.Animator.fragment_close_exit;
			global::Xamarin.Forms.Platform.Resource.Animator.fragment_fade_enter = global::pjsua2xamarin.Droid.Resource.Animator.fragment_fade_enter;
			global::Xamarin.Forms.Platform.Resource.Animator.fragment_fade_exit = global::pjsua2xamarin.Droid.Resource.Animator.fragment_fade_exit;
			global::Xamarin.Forms.Platform.Resource.Animator.fragment_open_enter = global::pjsua2xamarin.Droid.Resource.Animator.fragment_open_enter;
			global::Xamarin.Forms.Platform.Resource.Animator.fragment_open_exit = global::pjsua2xamarin.Droid.Resource.Animator.fragment_open_exit;
			global::Xamarin.Forms.Platform.Resource.Animator.linear_indeterminate_line1_head_interpolator = global::pjsua2xamarin.Droid.Resource.Animator.linear_indeterminate_line1_head_interpolator;
			global::Xamarin.Forms.Platform.Resource.Animator.linear_indeterminate_line1_tail_interpolator = global::pjsua2xamarin.Droid.Resource.Animator.linear_indeterminate_line1_tail_interpolator;
			global::Xamarin.Forms.Platform.Resource.Animator.linear_indeterminate_line2_head_interpolator = global::pjsua2xamarin.Droid.Resource.Animator.linear_indeterminate_line2_head_interpolator;
			global::Xamarin.Forms.Platform.Resource.Animator.linear_indeterminate_line2_tail_interpolator = global::pjsua2xamarin.Droid.Resource.Animator.linear_indeterminate_line2_tail_interpolator;
			global::Xamarin.Forms.Platform.Resource.Animator.mtrl_btn_state_list_anim = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_btn_state_list_anim;
			global::Xamarin.Forms.Platform.Resource.Animator.mtrl_btn_unelevated_state_list_anim = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_btn_unelevated_state_list_anim;
			global::Xamarin.Forms.Platform.Resource.Animator.mtrl_card_state_list_anim = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_card_state_list_anim;
			global::Xamarin.Forms.Platform.Resource.Animator.mtrl_chip_state_list_anim = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_chip_state_list_anim;
			global::Xamarin.Forms.Platform.Resource.Animator.mtrl_extended_fab_change_size_collapse_motion_spec = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_extended_fab_change_size_collapse_motion_spec;
			global::Xamarin.Forms.Platform.Resource.Animator.mtrl_extended_fab_change_size_expand_motion_spec = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_extended_fab_change_size_expand_motion_spec;
			global::Xamarin.Forms.Platform.Resource.Animator.mtrl_extended_fab_hide_motion_spec = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_extended_fab_hide_motion_spec;
			global::Xamarin.Forms.Platform.Resource.Animator.mtrl_extended_fab_show_motion_spec = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_extended_fab_show_motion_spec;
			global::Xamarin.Forms.Platform.Resource.Animator.mtrl_extended_fab_state_list_animator = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_extended_fab_state_list_animator;
			global::Xamarin.Forms.Platform.Resource.Animator.mtrl_fab_hide_motion_spec = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_fab_hide_motion_spec;
			global::Xamarin.Forms.Platform.Resource.Animator.mtrl_fab_show_motion_spec = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_fab_show_motion_spec;
			global::Xamarin.Forms.Platform.Resource.Animator.mtrl_fab_transformation_sheet_collapse_spec = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_fab_transformation_sheet_collapse_spec;
			global::Xamarin.Forms.Platform.Resource.Animator.mtrl_fab_transformation_sheet_expand_spec = global::pjsua2xamarin.Droid.Resource.Animator.mtrl_fab_transformation_sheet_expand_spec;
			global::Xamarin.Forms.Platform.Resource.Animator.nav_default_enter_anim = global::pjsua2xamarin.Droid.Resource.Animator.nav_default_enter_anim;
			global::Xamarin.Forms.Platform.Resource.Animator.nav_default_exit_anim = global::pjsua2xamarin.Droid.Resource.Animator.nav_default_exit_anim;
			global::Xamarin.Forms.Platform.Resource.Animator.nav_default_pop_enter_anim = global::pjsua2xamarin.Droid.Resource.Animator.nav_default_pop_enter_anim;
			global::Xamarin.Forms.Platform.Resource.Animator.nav_default_pop_exit_anim = global::pjsua2xamarin.Droid.Resource.Animator.nav_default_pop_exit_anim;
			global::Xamarin.Forms.Platform.Resource.Attribute.action = global::pjsua2xamarin.Droid.Resource.Attribute.action;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionBarDivider = global::pjsua2xamarin.Droid.Resource.Attribute.actionBarDivider;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionBarItemBackground = global::pjsua2xamarin.Droid.Resource.Attribute.actionBarItemBackground;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionBarPopupTheme = global::pjsua2xamarin.Droid.Resource.Attribute.actionBarPopupTheme;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionBarSize = global::pjsua2xamarin.Droid.Resource.Attribute.actionBarSize;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionBarSplitStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionBarSplitStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionBarStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionBarStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionBarTabBarStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionBarTabBarStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionBarTabStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionBarTabStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionBarTabTextStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionBarTabTextStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionBarTheme = global::pjsua2xamarin.Droid.Resource.Attribute.actionBarTheme;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionBarWidgetTheme = global::pjsua2xamarin.Droid.Resource.Attribute.actionBarWidgetTheme;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionDropDownStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionDropDownStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionLayout = global::pjsua2xamarin.Droid.Resource.Attribute.actionLayout;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionMenuTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.actionMenuTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionMenuTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.actionMenuTextColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionModeBackground = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeBackground;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionModeCloseButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeCloseButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionModeCloseContentDescription = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeCloseContentDescription;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionModeCloseDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeCloseDrawable;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionModeCopyDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeCopyDrawable;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionModeCutDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeCutDrawable;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionModeFindDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeFindDrawable;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionModePasteDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.actionModePasteDrawable;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionModePopupWindowStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionModePopupWindowStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionModeSelectAllDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeSelectAllDrawable;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionModeShareDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeShareDrawable;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionModeSplitBackground = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeSplitBackground;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionModeStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionModeTheme = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeTheme;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionModeWebSearchDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.actionModeWebSearchDrawable;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionOverflowButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionOverflowButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionOverflowMenuStyle = global::pjsua2xamarin.Droid.Resource.Attribute.actionOverflowMenuStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionProviderClass = global::pjsua2xamarin.Droid.Resource.Attribute.actionProviderClass;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionTextColorAlpha = global::pjsua2xamarin.Droid.Resource.Attribute.actionTextColorAlpha;
			global::Xamarin.Forms.Platform.Resource.Attribute.actionViewClass = global::pjsua2xamarin.Droid.Resource.Attribute.actionViewClass;
			global::Xamarin.Forms.Platform.Resource.Attribute.activityChooserViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.activityChooserViewStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.alertDialogButtonGroupStyle = global::pjsua2xamarin.Droid.Resource.Attribute.alertDialogButtonGroupStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.alertDialogCenterButtons = global::pjsua2xamarin.Droid.Resource.Attribute.alertDialogCenterButtons;
			global::Xamarin.Forms.Platform.Resource.Attribute.alertDialogStyle = global::pjsua2xamarin.Droid.Resource.Attribute.alertDialogStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.alertDialogTheme = global::pjsua2xamarin.Droid.Resource.Attribute.alertDialogTheme;
			global::Xamarin.Forms.Platform.Resource.Attribute.allowStacking = global::pjsua2xamarin.Droid.Resource.Attribute.allowStacking;
			global::Xamarin.Forms.Platform.Resource.Attribute.alpha = global::pjsua2xamarin.Droid.Resource.Attribute.alpha;
			global::Xamarin.Forms.Platform.Resource.Attribute.alphabeticModifiers = global::pjsua2xamarin.Droid.Resource.Attribute.alphabeticModifiers;
			global::Xamarin.Forms.Platform.Resource.Attribute.altSrc = global::pjsua2xamarin.Droid.Resource.Attribute.altSrc;
			global::Xamarin.Forms.Platform.Resource.Attribute.animateCircleAngleTo = global::pjsua2xamarin.Droid.Resource.Attribute.animateCircleAngleTo;
			global::Xamarin.Forms.Platform.Resource.Attribute.animateRelativeTo = global::pjsua2xamarin.Droid.Resource.Attribute.animateRelativeTo;
			global::Xamarin.Forms.Platform.Resource.Attribute.animationMode = global::pjsua2xamarin.Droid.Resource.Attribute.animationMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.appBarLayoutStyle = global::pjsua2xamarin.Droid.Resource.Attribute.appBarLayoutStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.applyMotionScene = global::pjsua2xamarin.Droid.Resource.Attribute.applyMotionScene;
			global::Xamarin.Forms.Platform.Resource.Attribute.arcMode = global::pjsua2xamarin.Droid.Resource.Attribute.arcMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.argType = global::pjsua2xamarin.Droid.Resource.Attribute.argType;
			global::Xamarin.Forms.Platform.Resource.Attribute.arrowHeadLength = global::pjsua2xamarin.Droid.Resource.Attribute.arrowHeadLength;
			global::Xamarin.Forms.Platform.Resource.Attribute.arrowShaftLength = global::pjsua2xamarin.Droid.Resource.Attribute.arrowShaftLength;
			global::Xamarin.Forms.Platform.Resource.Attribute.attributeName = global::pjsua2xamarin.Droid.Resource.Attribute.attributeName;
			global::Xamarin.Forms.Platform.Resource.Attribute.autoCompleteMode = global::pjsua2xamarin.Droid.Resource.Attribute.autoCompleteMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.autoCompleteTextViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.autoCompleteTextViewStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.autoSizeMaxTextSize = global::pjsua2xamarin.Droid.Resource.Attribute.autoSizeMaxTextSize;
			global::Xamarin.Forms.Platform.Resource.Attribute.autoSizeMinTextSize = global::pjsua2xamarin.Droid.Resource.Attribute.autoSizeMinTextSize;
			global::Xamarin.Forms.Platform.Resource.Attribute.autoSizePresetSizes = global::pjsua2xamarin.Droid.Resource.Attribute.autoSizePresetSizes;
			global::Xamarin.Forms.Platform.Resource.Attribute.autoSizeStepGranularity = global::pjsua2xamarin.Droid.Resource.Attribute.autoSizeStepGranularity;
			global::Xamarin.Forms.Platform.Resource.Attribute.autoSizeTextType = global::pjsua2xamarin.Droid.Resource.Attribute.autoSizeTextType;
			global::Xamarin.Forms.Platform.Resource.Attribute.autoTransition = global::pjsua2xamarin.Droid.Resource.Attribute.autoTransition;
			global::Xamarin.Forms.Platform.Resource.Attribute.background = global::pjsua2xamarin.Droid.Resource.Attribute.background;
			global::Xamarin.Forms.Platform.Resource.Attribute.backgroundColor = global::pjsua2xamarin.Droid.Resource.Attribute.backgroundColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.backgroundInsetBottom = global::pjsua2xamarin.Droid.Resource.Attribute.backgroundInsetBottom;
			global::Xamarin.Forms.Platform.Resource.Attribute.backgroundInsetEnd = global::pjsua2xamarin.Droid.Resource.Attribute.backgroundInsetEnd;
			global::Xamarin.Forms.Platform.Resource.Attribute.backgroundInsetStart = global::pjsua2xamarin.Droid.Resource.Attribute.backgroundInsetStart;
			global::Xamarin.Forms.Platform.Resource.Attribute.backgroundInsetTop = global::pjsua2xamarin.Droid.Resource.Attribute.backgroundInsetTop;
			global::Xamarin.Forms.Platform.Resource.Attribute.backgroundOverlayColorAlpha = global::pjsua2xamarin.Droid.Resource.Attribute.backgroundOverlayColorAlpha;
			global::Xamarin.Forms.Platform.Resource.Attribute.backgroundSplit = global::pjsua2xamarin.Droid.Resource.Attribute.backgroundSplit;
			global::Xamarin.Forms.Platform.Resource.Attribute.backgroundStacked = global::pjsua2xamarin.Droid.Resource.Attribute.backgroundStacked;
			global::Xamarin.Forms.Platform.Resource.Attribute.backgroundTint = global::pjsua2xamarin.Droid.Resource.Attribute.backgroundTint;
			global::Xamarin.Forms.Platform.Resource.Attribute.backgroundTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.backgroundTintMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.badgeGravity = global::pjsua2xamarin.Droid.Resource.Attribute.badgeGravity;
			global::Xamarin.Forms.Platform.Resource.Attribute.badgeStyle = global::pjsua2xamarin.Droid.Resource.Attribute.badgeStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.badgeTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.badgeTextColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.barLength = global::pjsua2xamarin.Droid.Resource.Attribute.barLength;
			global::Xamarin.Forms.Platform.Resource.Attribute.barrierAllowsGoneWidgets = global::pjsua2xamarin.Droid.Resource.Attribute.barrierAllowsGoneWidgets;
			global::Xamarin.Forms.Platform.Resource.Attribute.barrierDirection = global::pjsua2xamarin.Droid.Resource.Attribute.barrierDirection;
			global::Xamarin.Forms.Platform.Resource.Attribute.barrierMargin = global::pjsua2xamarin.Droid.Resource.Attribute.barrierMargin;
			global::Xamarin.Forms.Platform.Resource.Attribute.behavior_autoHide = global::pjsua2xamarin.Droid.Resource.Attribute.behavior_autoHide;
			global::Xamarin.Forms.Platform.Resource.Attribute.behavior_autoShrink = global::pjsua2xamarin.Droid.Resource.Attribute.behavior_autoShrink;
			global::Xamarin.Forms.Platform.Resource.Attribute.behavior_draggable = global::pjsua2xamarin.Droid.Resource.Attribute.behavior_draggable;
			global::Xamarin.Forms.Platform.Resource.Attribute.behavior_expandedOffset = global::pjsua2xamarin.Droid.Resource.Attribute.behavior_expandedOffset;
			global::Xamarin.Forms.Platform.Resource.Attribute.behavior_fitToContents = global::pjsua2xamarin.Droid.Resource.Attribute.behavior_fitToContents;
			global::Xamarin.Forms.Platform.Resource.Attribute.behavior_halfExpandedRatio = global::pjsua2xamarin.Droid.Resource.Attribute.behavior_halfExpandedRatio;
			global::Xamarin.Forms.Platform.Resource.Attribute.behavior_hideable = global::pjsua2xamarin.Droid.Resource.Attribute.behavior_hideable;
			global::Xamarin.Forms.Platform.Resource.Attribute.behavior_overlapTop = global::pjsua2xamarin.Droid.Resource.Attribute.behavior_overlapTop;
			global::Xamarin.Forms.Platform.Resource.Attribute.behavior_peekHeight = global::pjsua2xamarin.Droid.Resource.Attribute.behavior_peekHeight;
			global::Xamarin.Forms.Platform.Resource.Attribute.behavior_saveFlags = global::pjsua2xamarin.Droid.Resource.Attribute.behavior_saveFlags;
			global::Xamarin.Forms.Platform.Resource.Attribute.behavior_skipCollapsed = global::pjsua2xamarin.Droid.Resource.Attribute.behavior_skipCollapsed;
			global::Xamarin.Forms.Platform.Resource.Attribute.blendSrc = global::pjsua2xamarin.Droid.Resource.Attribute.blendSrc;
			global::Xamarin.Forms.Platform.Resource.Attribute.borderlessButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.borderlessButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.borderRound = global::pjsua2xamarin.Droid.Resource.Attribute.borderRound;
			global::Xamarin.Forms.Platform.Resource.Attribute.borderRoundPercent = global::pjsua2xamarin.Droid.Resource.Attribute.borderRoundPercent;
			global::Xamarin.Forms.Platform.Resource.Attribute.borderWidth = global::pjsua2xamarin.Droid.Resource.Attribute.borderWidth;
			global::Xamarin.Forms.Platform.Resource.Attribute.bottomAppBarStyle = global::pjsua2xamarin.Droid.Resource.Attribute.bottomAppBarStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.bottomNavigationStyle = global::pjsua2xamarin.Droid.Resource.Attribute.bottomNavigationStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.bottomSheetDialogTheme = global::pjsua2xamarin.Droid.Resource.Attribute.bottomSheetDialogTheme;
			global::Xamarin.Forms.Platform.Resource.Attribute.bottomSheetStyle = global::pjsua2xamarin.Droid.Resource.Attribute.bottomSheetStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.boxBackgroundColor = global::pjsua2xamarin.Droid.Resource.Attribute.boxBackgroundColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.boxBackgroundMode = global::pjsua2xamarin.Droid.Resource.Attribute.boxBackgroundMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.boxCollapsedPaddingTop = global::pjsua2xamarin.Droid.Resource.Attribute.boxCollapsedPaddingTop;
			global::Xamarin.Forms.Platform.Resource.Attribute.boxCornerRadiusBottomEnd = global::pjsua2xamarin.Droid.Resource.Attribute.boxCornerRadiusBottomEnd;
			global::Xamarin.Forms.Platform.Resource.Attribute.boxCornerRadiusBottomStart = global::pjsua2xamarin.Droid.Resource.Attribute.boxCornerRadiusBottomStart;
			global::Xamarin.Forms.Platform.Resource.Attribute.boxCornerRadiusTopEnd = global::pjsua2xamarin.Droid.Resource.Attribute.boxCornerRadiusTopEnd;
			global::Xamarin.Forms.Platform.Resource.Attribute.boxCornerRadiusTopStart = global::pjsua2xamarin.Droid.Resource.Attribute.boxCornerRadiusTopStart;
			global::Xamarin.Forms.Platform.Resource.Attribute.boxStrokeColor = global::pjsua2xamarin.Droid.Resource.Attribute.boxStrokeColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.boxStrokeErrorColor = global::pjsua2xamarin.Droid.Resource.Attribute.boxStrokeErrorColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.boxStrokeWidth = global::pjsua2xamarin.Droid.Resource.Attribute.boxStrokeWidth;
			global::Xamarin.Forms.Platform.Resource.Attribute.boxStrokeWidthFocused = global::pjsua2xamarin.Droid.Resource.Attribute.boxStrokeWidthFocused;
			global::Xamarin.Forms.Platform.Resource.Attribute.brightness = global::pjsua2xamarin.Droid.Resource.Attribute.brightness;
			global::Xamarin.Forms.Platform.Resource.Attribute.buttonBarButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.buttonBarButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.buttonBarNegativeButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.buttonBarNegativeButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.buttonBarNeutralButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.buttonBarNeutralButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.buttonBarPositiveButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.buttonBarPositiveButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.buttonBarStyle = global::pjsua2xamarin.Droid.Resource.Attribute.buttonBarStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.buttonCompat = global::pjsua2xamarin.Droid.Resource.Attribute.buttonCompat;
			global::Xamarin.Forms.Platform.Resource.Attribute.buttonGravity = global::pjsua2xamarin.Droid.Resource.Attribute.buttonGravity;
			global::Xamarin.Forms.Platform.Resource.Attribute.buttonIconDimen = global::pjsua2xamarin.Droid.Resource.Attribute.buttonIconDimen;
			global::Xamarin.Forms.Platform.Resource.Attribute.buttonPanelSideLayout = global::pjsua2xamarin.Droid.Resource.Attribute.buttonPanelSideLayout;
			global::Xamarin.Forms.Platform.Resource.Attribute.buttonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.buttonStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.buttonStyleSmall = global::pjsua2xamarin.Droid.Resource.Attribute.buttonStyleSmall;
			global::Xamarin.Forms.Platform.Resource.Attribute.buttonTint = global::pjsua2xamarin.Droid.Resource.Attribute.buttonTint;
			global::Xamarin.Forms.Platform.Resource.Attribute.buttonTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.buttonTintMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.cardBackgroundColor = global::pjsua2xamarin.Droid.Resource.Attribute.cardBackgroundColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.cardCornerRadius = global::pjsua2xamarin.Droid.Resource.Attribute.cardCornerRadius;
			global::Xamarin.Forms.Platform.Resource.Attribute.cardElevation = global::pjsua2xamarin.Droid.Resource.Attribute.cardElevation;
			global::Xamarin.Forms.Platform.Resource.Attribute.cardForegroundColor = global::pjsua2xamarin.Droid.Resource.Attribute.cardForegroundColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.cardMaxElevation = global::pjsua2xamarin.Droid.Resource.Attribute.cardMaxElevation;
			global::Xamarin.Forms.Platform.Resource.Attribute.cardPreventCornerOverlap = global::pjsua2xamarin.Droid.Resource.Attribute.cardPreventCornerOverlap;
			global::Xamarin.Forms.Platform.Resource.Attribute.cardUseCompatPadding = global::pjsua2xamarin.Droid.Resource.Attribute.cardUseCompatPadding;
			global::Xamarin.Forms.Platform.Resource.Attribute.cardViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.cardViewStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.carousel_backwardTransition = global::pjsua2xamarin.Droid.Resource.Attribute.carousel_backwardTransition;
			global::Xamarin.Forms.Platform.Resource.Attribute.carousel_emptyViewsBehavior = global::pjsua2xamarin.Droid.Resource.Attribute.carousel_emptyViewsBehavior;
			global::Xamarin.Forms.Platform.Resource.Attribute.carousel_firstView = global::pjsua2xamarin.Droid.Resource.Attribute.carousel_firstView;
			global::Xamarin.Forms.Platform.Resource.Attribute.carousel_forwardTransition = global::pjsua2xamarin.Droid.Resource.Attribute.carousel_forwardTransition;
			global::Xamarin.Forms.Platform.Resource.Attribute.carousel_infinite = global::pjsua2xamarin.Droid.Resource.Attribute.carousel_infinite;
			global::Xamarin.Forms.Platform.Resource.Attribute.carousel_nextState = global::pjsua2xamarin.Droid.Resource.Attribute.carousel_nextState;
			global::Xamarin.Forms.Platform.Resource.Attribute.carousel_previousState = global::pjsua2xamarin.Droid.Resource.Attribute.carousel_previousState;
			global::Xamarin.Forms.Platform.Resource.Attribute.carousel_touchUpMode = global::pjsua2xamarin.Droid.Resource.Attribute.carousel_touchUpMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.carousel_touchUp_dampeningFactor = global::pjsua2xamarin.Droid.Resource.Attribute.carousel_touchUp_dampeningFactor;
			global::Xamarin.Forms.Platform.Resource.Attribute.carousel_touchUp_velocityThreshold = global::pjsua2xamarin.Droid.Resource.Attribute.carousel_touchUp_velocityThreshold;
			global::Xamarin.Forms.Platform.Resource.Attribute.chainUseRtl = global::pjsua2xamarin.Droid.Resource.Attribute.chainUseRtl;
			global::Xamarin.Forms.Platform.Resource.Attribute.checkboxStyle = global::pjsua2xamarin.Droid.Resource.Attribute.checkboxStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.checkedButton = global::pjsua2xamarin.Droid.Resource.Attribute.checkedButton;
			global::Xamarin.Forms.Platform.Resource.Attribute.checkedChip = global::pjsua2xamarin.Droid.Resource.Attribute.checkedChip;
			global::Xamarin.Forms.Platform.Resource.Attribute.checkedIcon = global::pjsua2xamarin.Droid.Resource.Attribute.checkedIcon;
			global::Xamarin.Forms.Platform.Resource.Attribute.checkedIconEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.checkedIconEnabled;
			global::Xamarin.Forms.Platform.Resource.Attribute.checkedIconMargin = global::pjsua2xamarin.Droid.Resource.Attribute.checkedIconMargin;
			global::Xamarin.Forms.Platform.Resource.Attribute.checkedIconSize = global::pjsua2xamarin.Droid.Resource.Attribute.checkedIconSize;
			global::Xamarin.Forms.Platform.Resource.Attribute.checkedIconTint = global::pjsua2xamarin.Droid.Resource.Attribute.checkedIconTint;
			global::Xamarin.Forms.Platform.Resource.Attribute.checkedIconVisible = global::pjsua2xamarin.Droid.Resource.Attribute.checkedIconVisible;
			global::Xamarin.Forms.Platform.Resource.Attribute.checkedTextViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.checkedTextViewStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.chipBackgroundColor = global::pjsua2xamarin.Droid.Resource.Attribute.chipBackgroundColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.chipCornerRadius = global::pjsua2xamarin.Droid.Resource.Attribute.chipCornerRadius;
			global::Xamarin.Forms.Platform.Resource.Attribute.chipEndPadding = global::pjsua2xamarin.Droid.Resource.Attribute.chipEndPadding;
			global::Xamarin.Forms.Platform.Resource.Attribute.chipGroupStyle = global::pjsua2xamarin.Droid.Resource.Attribute.chipGroupStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.chipIcon = global::pjsua2xamarin.Droid.Resource.Attribute.chipIcon;
			global::Xamarin.Forms.Platform.Resource.Attribute.chipIconEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.chipIconEnabled;
			global::Xamarin.Forms.Platform.Resource.Attribute.chipIconSize = global::pjsua2xamarin.Droid.Resource.Attribute.chipIconSize;
			global::Xamarin.Forms.Platform.Resource.Attribute.chipIconTint = global::pjsua2xamarin.Droid.Resource.Attribute.chipIconTint;
			global::Xamarin.Forms.Platform.Resource.Attribute.chipIconVisible = global::pjsua2xamarin.Droid.Resource.Attribute.chipIconVisible;
			global::Xamarin.Forms.Platform.Resource.Attribute.chipMinHeight = global::pjsua2xamarin.Droid.Resource.Attribute.chipMinHeight;
			global::Xamarin.Forms.Platform.Resource.Attribute.chipMinTouchTargetSize = global::pjsua2xamarin.Droid.Resource.Attribute.chipMinTouchTargetSize;
			global::Xamarin.Forms.Platform.Resource.Attribute.chipSpacing = global::pjsua2xamarin.Droid.Resource.Attribute.chipSpacing;
			global::Xamarin.Forms.Platform.Resource.Attribute.chipSpacingHorizontal = global::pjsua2xamarin.Droid.Resource.Attribute.chipSpacingHorizontal;
			global::Xamarin.Forms.Platform.Resource.Attribute.chipSpacingVertical = global::pjsua2xamarin.Droid.Resource.Attribute.chipSpacingVertical;
			global::Xamarin.Forms.Platform.Resource.Attribute.chipStandaloneStyle = global::pjsua2xamarin.Droid.Resource.Attribute.chipStandaloneStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.chipStartPadding = global::pjsua2xamarin.Droid.Resource.Attribute.chipStartPadding;
			global::Xamarin.Forms.Platform.Resource.Attribute.chipStrokeColor = global::pjsua2xamarin.Droid.Resource.Attribute.chipStrokeColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.chipStrokeWidth = global::pjsua2xamarin.Droid.Resource.Attribute.chipStrokeWidth;
			global::Xamarin.Forms.Platform.Resource.Attribute.chipStyle = global::pjsua2xamarin.Droid.Resource.Attribute.chipStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.chipSurfaceColor = global::pjsua2xamarin.Droid.Resource.Attribute.chipSurfaceColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.circleRadius = global::pjsua2xamarin.Droid.Resource.Attribute.circleRadius;
			global::Xamarin.Forms.Platform.Resource.Attribute.circularflow_angles = global::pjsua2xamarin.Droid.Resource.Attribute.circularflow_angles;
			global::Xamarin.Forms.Platform.Resource.Attribute.circularflow_defaultAngle = global::pjsua2xamarin.Droid.Resource.Attribute.circularflow_defaultAngle;
			global::Xamarin.Forms.Platform.Resource.Attribute.circularflow_defaultRadius = global::pjsua2xamarin.Droid.Resource.Attribute.circularflow_defaultRadius;
			global::Xamarin.Forms.Platform.Resource.Attribute.circularflow_radiusInDP = global::pjsua2xamarin.Droid.Resource.Attribute.circularflow_radiusInDP;
			global::Xamarin.Forms.Platform.Resource.Attribute.circularflow_viewCenter = global::pjsua2xamarin.Droid.Resource.Attribute.circularflow_viewCenter;
			global::Xamarin.Forms.Platform.Resource.Attribute.circularProgressIndicatorStyle = global::pjsua2xamarin.Droid.Resource.Attribute.circularProgressIndicatorStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.clearsTag = global::pjsua2xamarin.Droid.Resource.Attribute.clearsTag;
			global::Xamarin.Forms.Platform.Resource.Attribute.clickAction = global::pjsua2xamarin.Droid.Resource.Attribute.clickAction;
			global::Xamarin.Forms.Platform.Resource.Attribute.clockFaceBackgroundColor = global::pjsua2xamarin.Droid.Resource.Attribute.clockFaceBackgroundColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.clockHandColor = global::pjsua2xamarin.Droid.Resource.Attribute.clockHandColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.clockIcon = global::pjsua2xamarin.Droid.Resource.Attribute.clockIcon;
			global::Xamarin.Forms.Platform.Resource.Attribute.clockNumberTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.clockNumberTextColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.closeIcon = global::pjsua2xamarin.Droid.Resource.Attribute.closeIcon;
			global::Xamarin.Forms.Platform.Resource.Attribute.closeIconEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.closeIconEnabled;
			global::Xamarin.Forms.Platform.Resource.Attribute.closeIconEndPadding = global::pjsua2xamarin.Droid.Resource.Attribute.closeIconEndPadding;
			global::Xamarin.Forms.Platform.Resource.Attribute.closeIconSize = global::pjsua2xamarin.Droid.Resource.Attribute.closeIconSize;
			global::Xamarin.Forms.Platform.Resource.Attribute.closeIconStartPadding = global::pjsua2xamarin.Droid.Resource.Attribute.closeIconStartPadding;
			global::Xamarin.Forms.Platform.Resource.Attribute.closeIconTint = global::pjsua2xamarin.Droid.Resource.Attribute.closeIconTint;
			global::Xamarin.Forms.Platform.Resource.Attribute.closeIconVisible = global::pjsua2xamarin.Droid.Resource.Attribute.closeIconVisible;
			global::Xamarin.Forms.Platform.Resource.Attribute.closeItemLayout = global::pjsua2xamarin.Droid.Resource.Attribute.closeItemLayout;
			global::Xamarin.Forms.Platform.Resource.Attribute.collapseContentDescription = global::pjsua2xamarin.Droid.Resource.Attribute.collapseContentDescription;
			global::Xamarin.Forms.Platform.Resource.Attribute.collapsedSize = global::pjsua2xamarin.Droid.Resource.Attribute.collapsedSize;
			global::Xamarin.Forms.Platform.Resource.Attribute.collapsedTitleGravity = global::pjsua2xamarin.Droid.Resource.Attribute.collapsedTitleGravity;
			global::Xamarin.Forms.Platform.Resource.Attribute.collapsedTitleTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.collapsedTitleTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Attribute.collapseIcon = global::pjsua2xamarin.Droid.Resource.Attribute.collapseIcon;
			global::Xamarin.Forms.Platform.Resource.Attribute.collapsingToolbarLayoutStyle = global::pjsua2xamarin.Droid.Resource.Attribute.collapsingToolbarLayoutStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.collectionViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.collectionViewStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.color = global::pjsua2xamarin.Droid.Resource.Attribute.color;
			global::Xamarin.Forms.Platform.Resource.Attribute.colorAccent = global::pjsua2xamarin.Droid.Resource.Attribute.colorAccent;
			global::Xamarin.Forms.Platform.Resource.Attribute.colorBackgroundFloating = global::pjsua2xamarin.Droid.Resource.Attribute.colorBackgroundFloating;
			global::Xamarin.Forms.Platform.Resource.Attribute.colorButtonNormal = global::pjsua2xamarin.Droid.Resource.Attribute.colorButtonNormal;
			global::Xamarin.Forms.Platform.Resource.Attribute.colorControlActivated = global::pjsua2xamarin.Droid.Resource.Attribute.colorControlActivated;
			global::Xamarin.Forms.Platform.Resource.Attribute.colorControlHighlight = global::pjsua2xamarin.Droid.Resource.Attribute.colorControlHighlight;
			global::Xamarin.Forms.Platform.Resource.Attribute.colorControlNormal = global::pjsua2xamarin.Droid.Resource.Attribute.colorControlNormal;
			global::Xamarin.Forms.Platform.Resource.Attribute.colorError = global::pjsua2xamarin.Droid.Resource.Attribute.colorError;
			global::Xamarin.Forms.Platform.Resource.Attribute.colorOnBackground = global::pjsua2xamarin.Droid.Resource.Attribute.colorOnBackground;
			global::Xamarin.Forms.Platform.Resource.Attribute.colorOnError = global::pjsua2xamarin.Droid.Resource.Attribute.colorOnError;
			global::Xamarin.Forms.Platform.Resource.Attribute.colorOnPrimary = global::pjsua2xamarin.Droid.Resource.Attribute.colorOnPrimary;
			global::Xamarin.Forms.Platform.Resource.Attribute.colorOnPrimarySurface = global::pjsua2xamarin.Droid.Resource.Attribute.colorOnPrimarySurface;
			global::Xamarin.Forms.Platform.Resource.Attribute.colorOnSecondary = global::pjsua2xamarin.Droid.Resource.Attribute.colorOnSecondary;
			global::Xamarin.Forms.Platform.Resource.Attribute.colorOnSurface = global::pjsua2xamarin.Droid.Resource.Attribute.colorOnSurface;
			global::Xamarin.Forms.Platform.Resource.Attribute.colorPrimary = global::pjsua2xamarin.Droid.Resource.Attribute.colorPrimary;
			global::Xamarin.Forms.Platform.Resource.Attribute.colorPrimaryDark = global::pjsua2xamarin.Droid.Resource.Attribute.colorPrimaryDark;
			global::Xamarin.Forms.Platform.Resource.Attribute.colorPrimarySurface = global::pjsua2xamarin.Droid.Resource.Attribute.colorPrimarySurface;
			global::Xamarin.Forms.Platform.Resource.Attribute.colorPrimaryVariant = global::pjsua2xamarin.Droid.Resource.Attribute.colorPrimaryVariant;
			global::Xamarin.Forms.Platform.Resource.Attribute.colorSecondary = global::pjsua2xamarin.Droid.Resource.Attribute.colorSecondary;
			global::Xamarin.Forms.Platform.Resource.Attribute.colorSecondaryVariant = global::pjsua2xamarin.Droid.Resource.Attribute.colorSecondaryVariant;
			global::Xamarin.Forms.Platform.Resource.Attribute.colorSurface = global::pjsua2xamarin.Droid.Resource.Attribute.colorSurface;
			global::Xamarin.Forms.Platform.Resource.Attribute.colorSwitchThumbNormal = global::pjsua2xamarin.Droid.Resource.Attribute.colorSwitchThumbNormal;
			global::Xamarin.Forms.Platform.Resource.Attribute.commitIcon = global::pjsua2xamarin.Droid.Resource.Attribute.commitIcon;
			global::Xamarin.Forms.Platform.Resource.Attribute.ConstraintRotate = global::pjsua2xamarin.Droid.Resource.Attribute.ConstraintRotate;
			global::Xamarin.Forms.Platform.Resource.Attribute.constraints = global::pjsua2xamarin.Droid.Resource.Attribute.constraints;
			global::Xamarin.Forms.Platform.Resource.Attribute.constraintSet = global::pjsua2xamarin.Droid.Resource.Attribute.constraintSet;
			global::Xamarin.Forms.Platform.Resource.Attribute.constraintSetEnd = global::pjsua2xamarin.Droid.Resource.Attribute.constraintSetEnd;
			global::Xamarin.Forms.Platform.Resource.Attribute.constraintSetStart = global::pjsua2xamarin.Droid.Resource.Attribute.constraintSetStart;
			global::Xamarin.Forms.Platform.Resource.Attribute.constraint_referenced_ids = global::pjsua2xamarin.Droid.Resource.Attribute.constraint_referenced_ids;
			global::Xamarin.Forms.Platform.Resource.Attribute.constraint_referenced_tags = global::pjsua2xamarin.Droid.Resource.Attribute.constraint_referenced_tags;
			global::Xamarin.Forms.Platform.Resource.Attribute.content = global::pjsua2xamarin.Droid.Resource.Attribute.content;
			global::Xamarin.Forms.Platform.Resource.Attribute.contentDescription = global::pjsua2xamarin.Droid.Resource.Attribute.contentDescription;
			global::Xamarin.Forms.Platform.Resource.Attribute.contentInsetEnd = global::pjsua2xamarin.Droid.Resource.Attribute.contentInsetEnd;
			global::Xamarin.Forms.Platform.Resource.Attribute.contentInsetEndWithActions = global::pjsua2xamarin.Droid.Resource.Attribute.contentInsetEndWithActions;
			global::Xamarin.Forms.Platform.Resource.Attribute.contentInsetLeft = global::pjsua2xamarin.Droid.Resource.Attribute.contentInsetLeft;
			global::Xamarin.Forms.Platform.Resource.Attribute.contentInsetRight = global::pjsua2xamarin.Droid.Resource.Attribute.contentInsetRight;
			global::Xamarin.Forms.Platform.Resource.Attribute.contentInsetStart = global::pjsua2xamarin.Droid.Resource.Attribute.contentInsetStart;
			global::Xamarin.Forms.Platform.Resource.Attribute.contentInsetStartWithNavigation = global::pjsua2xamarin.Droid.Resource.Attribute.contentInsetStartWithNavigation;
			global::Xamarin.Forms.Platform.Resource.Attribute.contentPadding = global::pjsua2xamarin.Droid.Resource.Attribute.contentPadding;
			global::Xamarin.Forms.Platform.Resource.Attribute.contentPaddingBottom = global::pjsua2xamarin.Droid.Resource.Attribute.contentPaddingBottom;
			global::Xamarin.Forms.Platform.Resource.Attribute.contentPaddingEnd = global::pjsua2xamarin.Droid.Resource.Attribute.contentPaddingEnd;
			global::Xamarin.Forms.Platform.Resource.Attribute.contentPaddingLeft = global::pjsua2xamarin.Droid.Resource.Attribute.contentPaddingLeft;
			global::Xamarin.Forms.Platform.Resource.Attribute.contentPaddingRight = global::pjsua2xamarin.Droid.Resource.Attribute.contentPaddingRight;
			global::Xamarin.Forms.Platform.Resource.Attribute.contentPaddingStart = global::pjsua2xamarin.Droid.Resource.Attribute.contentPaddingStart;
			global::Xamarin.Forms.Platform.Resource.Attribute.contentPaddingTop = global::pjsua2xamarin.Droid.Resource.Attribute.contentPaddingTop;
			global::Xamarin.Forms.Platform.Resource.Attribute.contentScrim = global::pjsua2xamarin.Droid.Resource.Attribute.contentScrim;
			global::Xamarin.Forms.Platform.Resource.Attribute.contrast = global::pjsua2xamarin.Droid.Resource.Attribute.contrast;
			global::Xamarin.Forms.Platform.Resource.Attribute.controlBackground = global::pjsua2xamarin.Droid.Resource.Attribute.controlBackground;
			global::Xamarin.Forms.Platform.Resource.Attribute.coordinatorLayoutStyle = global::pjsua2xamarin.Droid.Resource.Attribute.coordinatorLayoutStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.cornerFamily = global::pjsua2xamarin.Droid.Resource.Attribute.cornerFamily;
			global::Xamarin.Forms.Platform.Resource.Attribute.cornerFamilyBottomLeft = global::pjsua2xamarin.Droid.Resource.Attribute.cornerFamilyBottomLeft;
			global::Xamarin.Forms.Platform.Resource.Attribute.cornerFamilyBottomRight = global::pjsua2xamarin.Droid.Resource.Attribute.cornerFamilyBottomRight;
			global::Xamarin.Forms.Platform.Resource.Attribute.cornerFamilyTopLeft = global::pjsua2xamarin.Droid.Resource.Attribute.cornerFamilyTopLeft;
			global::Xamarin.Forms.Platform.Resource.Attribute.cornerFamilyTopRight = global::pjsua2xamarin.Droid.Resource.Attribute.cornerFamilyTopRight;
			global::Xamarin.Forms.Platform.Resource.Attribute.cornerRadius = global::pjsua2xamarin.Droid.Resource.Attribute.cornerRadius;
			global::Xamarin.Forms.Platform.Resource.Attribute.cornerSize = global::pjsua2xamarin.Droid.Resource.Attribute.cornerSize;
			global::Xamarin.Forms.Platform.Resource.Attribute.cornerSizeBottomLeft = global::pjsua2xamarin.Droid.Resource.Attribute.cornerSizeBottomLeft;
			global::Xamarin.Forms.Platform.Resource.Attribute.cornerSizeBottomRight = global::pjsua2xamarin.Droid.Resource.Attribute.cornerSizeBottomRight;
			global::Xamarin.Forms.Platform.Resource.Attribute.cornerSizeTopLeft = global::pjsua2xamarin.Droid.Resource.Attribute.cornerSizeTopLeft;
			global::Xamarin.Forms.Platform.Resource.Attribute.cornerSizeTopRight = global::pjsua2xamarin.Droid.Resource.Attribute.cornerSizeTopRight;
			global::Xamarin.Forms.Platform.Resource.Attribute.counterEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.counterEnabled;
			global::Xamarin.Forms.Platform.Resource.Attribute.counterMaxLength = global::pjsua2xamarin.Droid.Resource.Attribute.counterMaxLength;
			global::Xamarin.Forms.Platform.Resource.Attribute.counterOverflowTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.counterOverflowTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Attribute.counterOverflowTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.counterOverflowTextColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.counterTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.counterTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Attribute.counterTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.counterTextColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.crossfade = global::pjsua2xamarin.Droid.Resource.Attribute.crossfade;
			global::Xamarin.Forms.Platform.Resource.Attribute.currentState = global::pjsua2xamarin.Droid.Resource.Attribute.currentState;
			global::Xamarin.Forms.Platform.Resource.Attribute.curveFit = global::pjsua2xamarin.Droid.Resource.Attribute.curveFit;
			global::Xamarin.Forms.Platform.Resource.Attribute.customBoolean = global::pjsua2xamarin.Droid.Resource.Attribute.customBoolean;
			global::Xamarin.Forms.Platform.Resource.Attribute.customColorDrawableValue = global::pjsua2xamarin.Droid.Resource.Attribute.customColorDrawableValue;
			global::Xamarin.Forms.Platform.Resource.Attribute.customColorValue = global::pjsua2xamarin.Droid.Resource.Attribute.customColorValue;
			global::Xamarin.Forms.Platform.Resource.Attribute.customDimension = global::pjsua2xamarin.Droid.Resource.Attribute.customDimension;
			global::Xamarin.Forms.Platform.Resource.Attribute.customFloatValue = global::pjsua2xamarin.Droid.Resource.Attribute.customFloatValue;
			global::Xamarin.Forms.Platform.Resource.Attribute.customIntegerValue = global::pjsua2xamarin.Droid.Resource.Attribute.customIntegerValue;
			global::Xamarin.Forms.Platform.Resource.Attribute.customNavigationLayout = global::pjsua2xamarin.Droid.Resource.Attribute.customNavigationLayout;
			global::Xamarin.Forms.Platform.Resource.Attribute.customPixelDimension = global::pjsua2xamarin.Droid.Resource.Attribute.customPixelDimension;
			global::Xamarin.Forms.Platform.Resource.Attribute.customReference = global::pjsua2xamarin.Droid.Resource.Attribute.customReference;
			global::Xamarin.Forms.Platform.Resource.Attribute.customStringValue = global::pjsua2xamarin.Droid.Resource.Attribute.customStringValue;
			global::Xamarin.Forms.Platform.Resource.Attribute.data = global::pjsua2xamarin.Droid.Resource.Attribute.data;
			global::Xamarin.Forms.Platform.Resource.Attribute.dataPattern = global::pjsua2xamarin.Droid.Resource.Attribute.dataPattern;
			global::Xamarin.Forms.Platform.Resource.Attribute.dayInvalidStyle = global::pjsua2xamarin.Droid.Resource.Attribute.dayInvalidStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.daySelectedStyle = global::pjsua2xamarin.Droid.Resource.Attribute.daySelectedStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.dayStyle = global::pjsua2xamarin.Droid.Resource.Attribute.dayStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.dayTodayStyle = global::pjsua2xamarin.Droid.Resource.Attribute.dayTodayStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.defaultDuration = global::pjsua2xamarin.Droid.Resource.Attribute.defaultDuration;
			global::Xamarin.Forms.Platform.Resource.Attribute.defaultQueryHint = global::pjsua2xamarin.Droid.Resource.Attribute.defaultQueryHint;
			global::Xamarin.Forms.Platform.Resource.Attribute.defaultState = global::pjsua2xamarin.Droid.Resource.Attribute.defaultState;
			global::Xamarin.Forms.Platform.Resource.Attribute.deltaPolarAngle = global::pjsua2xamarin.Droid.Resource.Attribute.deltaPolarAngle;
			global::Xamarin.Forms.Platform.Resource.Attribute.deltaPolarRadius = global::pjsua2xamarin.Droid.Resource.Attribute.deltaPolarRadius;
			global::Xamarin.Forms.Platform.Resource.Attribute.deriveConstraintsFrom = global::pjsua2xamarin.Droid.Resource.Attribute.deriveConstraintsFrom;
			global::Xamarin.Forms.Platform.Resource.Attribute.destination = global::pjsua2xamarin.Droid.Resource.Attribute.destination;
			global::Xamarin.Forms.Platform.Resource.Attribute.dialogCornerRadius = global::pjsua2xamarin.Droid.Resource.Attribute.dialogCornerRadius;
			global::Xamarin.Forms.Platform.Resource.Attribute.dialogPreferredPadding = global::pjsua2xamarin.Droid.Resource.Attribute.dialogPreferredPadding;
			global::Xamarin.Forms.Platform.Resource.Attribute.dialogTheme = global::pjsua2xamarin.Droid.Resource.Attribute.dialogTheme;
			global::Xamarin.Forms.Platform.Resource.Attribute.displayOptions = global::pjsua2xamarin.Droid.Resource.Attribute.displayOptions;
			global::Xamarin.Forms.Platform.Resource.Attribute.divider = global::pjsua2xamarin.Droid.Resource.Attribute.divider;
			global::Xamarin.Forms.Platform.Resource.Attribute.dividerHorizontal = global::pjsua2xamarin.Droid.Resource.Attribute.dividerHorizontal;
			global::Xamarin.Forms.Platform.Resource.Attribute.dividerPadding = global::pjsua2xamarin.Droid.Resource.Attribute.dividerPadding;
			global::Xamarin.Forms.Platform.Resource.Attribute.dividerVertical = global::pjsua2xamarin.Droid.Resource.Attribute.dividerVertical;
			global::Xamarin.Forms.Platform.Resource.Attribute.dragDirection = global::pjsua2xamarin.Droid.Resource.Attribute.dragDirection;
			global::Xamarin.Forms.Platform.Resource.Attribute.dragScale = global::pjsua2xamarin.Droid.Resource.Attribute.dragScale;
			global::Xamarin.Forms.Platform.Resource.Attribute.dragThreshold = global::pjsua2xamarin.Droid.Resource.Attribute.dragThreshold;
			global::Xamarin.Forms.Platform.Resource.Attribute.drawableBottomCompat = global::pjsua2xamarin.Droid.Resource.Attribute.drawableBottomCompat;
			global::Xamarin.Forms.Platform.Resource.Attribute.drawableEndCompat = global::pjsua2xamarin.Droid.Resource.Attribute.drawableEndCompat;
			global::Xamarin.Forms.Platform.Resource.Attribute.drawableLeftCompat = global::pjsua2xamarin.Droid.Resource.Attribute.drawableLeftCompat;
			global::Xamarin.Forms.Platform.Resource.Attribute.drawableRightCompat = global::pjsua2xamarin.Droid.Resource.Attribute.drawableRightCompat;
			global::Xamarin.Forms.Platform.Resource.Attribute.drawableSize = global::pjsua2xamarin.Droid.Resource.Attribute.drawableSize;
			global::Xamarin.Forms.Platform.Resource.Attribute.drawableStartCompat = global::pjsua2xamarin.Droid.Resource.Attribute.drawableStartCompat;
			global::Xamarin.Forms.Platform.Resource.Attribute.drawableTint = global::pjsua2xamarin.Droid.Resource.Attribute.drawableTint;
			global::Xamarin.Forms.Platform.Resource.Attribute.drawableTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.drawableTintMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.drawableTopCompat = global::pjsua2xamarin.Droid.Resource.Attribute.drawableTopCompat;
			global::Xamarin.Forms.Platform.Resource.Attribute.drawerArrowStyle = global::pjsua2xamarin.Droid.Resource.Attribute.drawerArrowStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.drawerLayoutStyle = global::pjsua2xamarin.Droid.Resource.Attribute.drawerLayoutStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.drawPath = global::pjsua2xamarin.Droid.Resource.Attribute.drawPath;
			global::Xamarin.Forms.Platform.Resource.Attribute.dropdownListPreferredItemHeight = global::pjsua2xamarin.Droid.Resource.Attribute.dropdownListPreferredItemHeight;
			global::Xamarin.Forms.Platform.Resource.Attribute.dropDownListViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.dropDownListViewStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.duration = global::pjsua2xamarin.Droid.Resource.Attribute.duration;
			global::Xamarin.Forms.Platform.Resource.Attribute.editTextBackground = global::pjsua2xamarin.Droid.Resource.Attribute.editTextBackground;
			global::Xamarin.Forms.Platform.Resource.Attribute.editTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.editTextColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.editTextStyle = global::pjsua2xamarin.Droid.Resource.Attribute.editTextStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.elevation = global::pjsua2xamarin.Droid.Resource.Attribute.elevation;
			global::Xamarin.Forms.Platform.Resource.Attribute.elevationOverlayColor = global::pjsua2xamarin.Droid.Resource.Attribute.elevationOverlayColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.elevationOverlayEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.elevationOverlayEnabled;
			global::Xamarin.Forms.Platform.Resource.Attribute.enableEdgeToEdge = global::pjsua2xamarin.Droid.Resource.Attribute.enableEdgeToEdge;
			global::Xamarin.Forms.Platform.Resource.Attribute.endIconCheckable = global::pjsua2xamarin.Droid.Resource.Attribute.endIconCheckable;
			global::Xamarin.Forms.Platform.Resource.Attribute.endIconContentDescription = global::pjsua2xamarin.Droid.Resource.Attribute.endIconContentDescription;
			global::Xamarin.Forms.Platform.Resource.Attribute.endIconDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.endIconDrawable;
			global::Xamarin.Forms.Platform.Resource.Attribute.endIconMode = global::pjsua2xamarin.Droid.Resource.Attribute.endIconMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.endIconTint = global::pjsua2xamarin.Droid.Resource.Attribute.endIconTint;
			global::Xamarin.Forms.Platform.Resource.Attribute.endIconTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.endIconTintMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.enforceMaterialTheme = global::pjsua2xamarin.Droid.Resource.Attribute.enforceMaterialTheme;
			global::Xamarin.Forms.Platform.Resource.Attribute.enforceTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.enforceTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Attribute.ensureMinTouchTargetSize = global::pjsua2xamarin.Droid.Resource.Attribute.ensureMinTouchTargetSize;
			global::Xamarin.Forms.Platform.Resource.Attribute.enterAnim = global::pjsua2xamarin.Droid.Resource.Attribute.enterAnim;
			global::Xamarin.Forms.Platform.Resource.Attribute.errorContentDescription = global::pjsua2xamarin.Droid.Resource.Attribute.errorContentDescription;
			global::Xamarin.Forms.Platform.Resource.Attribute.errorEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.errorEnabled;
			global::Xamarin.Forms.Platform.Resource.Attribute.errorIconDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.errorIconDrawable;
			global::Xamarin.Forms.Platform.Resource.Attribute.errorIconTint = global::pjsua2xamarin.Droid.Resource.Attribute.errorIconTint;
			global::Xamarin.Forms.Platform.Resource.Attribute.errorIconTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.errorIconTintMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.errorTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.errorTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Attribute.errorTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.errorTextColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.exitAnim = global::pjsua2xamarin.Droid.Resource.Attribute.exitAnim;
			global::Xamarin.Forms.Platform.Resource.Attribute.expandActivityOverflowButtonDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.expandActivityOverflowButtonDrawable;
			global::Xamarin.Forms.Platform.Resource.Attribute.expanded = global::pjsua2xamarin.Droid.Resource.Attribute.expanded;
			global::Xamarin.Forms.Platform.Resource.Attribute.expandedHintEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.expandedHintEnabled;
			global::Xamarin.Forms.Platform.Resource.Attribute.expandedTitleGravity = global::pjsua2xamarin.Droid.Resource.Attribute.expandedTitleGravity;
			global::Xamarin.Forms.Platform.Resource.Attribute.expandedTitleMargin = global::pjsua2xamarin.Droid.Resource.Attribute.expandedTitleMargin;
			global::Xamarin.Forms.Platform.Resource.Attribute.expandedTitleMarginBottom = global::pjsua2xamarin.Droid.Resource.Attribute.expandedTitleMarginBottom;
			global::Xamarin.Forms.Platform.Resource.Attribute.expandedTitleMarginEnd = global::pjsua2xamarin.Droid.Resource.Attribute.expandedTitleMarginEnd;
			global::Xamarin.Forms.Platform.Resource.Attribute.expandedTitleMarginStart = global::pjsua2xamarin.Droid.Resource.Attribute.expandedTitleMarginStart;
			global::Xamarin.Forms.Platform.Resource.Attribute.expandedTitleMarginTop = global::pjsua2xamarin.Droid.Resource.Attribute.expandedTitleMarginTop;
			global::Xamarin.Forms.Platform.Resource.Attribute.expandedTitleTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.expandedTitleTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Attribute.extendedFloatingActionButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.extendedFloatingActionButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.extendMotionSpec = global::pjsua2xamarin.Droid.Resource.Attribute.extendMotionSpec;
			global::Xamarin.Forms.Platform.Resource.Attribute.extraMultilineHeightEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.extraMultilineHeightEnabled;
			global::Xamarin.Forms.Platform.Resource.Attribute.fabAlignmentMode = global::pjsua2xamarin.Droid.Resource.Attribute.fabAlignmentMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.fabAnimationMode = global::pjsua2xamarin.Droid.Resource.Attribute.fabAnimationMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.fabCradleMargin = global::pjsua2xamarin.Droid.Resource.Attribute.fabCradleMargin;
			global::Xamarin.Forms.Platform.Resource.Attribute.fabCradleRoundedCornerRadius = global::pjsua2xamarin.Droid.Resource.Attribute.fabCradleRoundedCornerRadius;
			global::Xamarin.Forms.Platform.Resource.Attribute.fabCradleVerticalOffset = global::pjsua2xamarin.Droid.Resource.Attribute.fabCradleVerticalOffset;
			global::Xamarin.Forms.Platform.Resource.Attribute.fabCustomSize = global::pjsua2xamarin.Droid.Resource.Attribute.fabCustomSize;
			global::Xamarin.Forms.Platform.Resource.Attribute.fabSize = global::pjsua2xamarin.Droid.Resource.Attribute.fabSize;
			global::Xamarin.Forms.Platform.Resource.Attribute.fastScrollEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.fastScrollEnabled;
			global::Xamarin.Forms.Platform.Resource.Attribute.fastScrollHorizontalThumbDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.fastScrollHorizontalThumbDrawable;
			global::Xamarin.Forms.Platform.Resource.Attribute.fastScrollHorizontalTrackDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.fastScrollHorizontalTrackDrawable;
			global::Xamarin.Forms.Platform.Resource.Attribute.fastScrollVerticalThumbDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.fastScrollVerticalThumbDrawable;
			global::Xamarin.Forms.Platform.Resource.Attribute.fastScrollVerticalTrackDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.fastScrollVerticalTrackDrawable;
			global::Xamarin.Forms.Platform.Resource.Attribute.firstBaselineToTopHeight = global::pjsua2xamarin.Droid.Resource.Attribute.firstBaselineToTopHeight;
			global::Xamarin.Forms.Platform.Resource.Attribute.floatingActionButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.floatingActionButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.flow_firstHorizontalBias = global::pjsua2xamarin.Droid.Resource.Attribute.flow_firstHorizontalBias;
			global::Xamarin.Forms.Platform.Resource.Attribute.flow_firstHorizontalStyle = global::pjsua2xamarin.Droid.Resource.Attribute.flow_firstHorizontalStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.flow_firstVerticalBias = global::pjsua2xamarin.Droid.Resource.Attribute.flow_firstVerticalBias;
			global::Xamarin.Forms.Platform.Resource.Attribute.flow_firstVerticalStyle = global::pjsua2xamarin.Droid.Resource.Attribute.flow_firstVerticalStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.flow_horizontalAlign = global::pjsua2xamarin.Droid.Resource.Attribute.flow_horizontalAlign;
			global::Xamarin.Forms.Platform.Resource.Attribute.flow_horizontalBias = global::pjsua2xamarin.Droid.Resource.Attribute.flow_horizontalBias;
			global::Xamarin.Forms.Platform.Resource.Attribute.flow_horizontalGap = global::pjsua2xamarin.Droid.Resource.Attribute.flow_horizontalGap;
			global::Xamarin.Forms.Platform.Resource.Attribute.flow_horizontalStyle = global::pjsua2xamarin.Droid.Resource.Attribute.flow_horizontalStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.flow_lastHorizontalBias = global::pjsua2xamarin.Droid.Resource.Attribute.flow_lastHorizontalBias;
			global::Xamarin.Forms.Platform.Resource.Attribute.flow_lastHorizontalStyle = global::pjsua2xamarin.Droid.Resource.Attribute.flow_lastHorizontalStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.flow_lastVerticalBias = global::pjsua2xamarin.Droid.Resource.Attribute.flow_lastVerticalBias;
			global::Xamarin.Forms.Platform.Resource.Attribute.flow_lastVerticalStyle = global::pjsua2xamarin.Droid.Resource.Attribute.flow_lastVerticalStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.flow_maxElementsWrap = global::pjsua2xamarin.Droid.Resource.Attribute.flow_maxElementsWrap;
			global::Xamarin.Forms.Platform.Resource.Attribute.flow_padding = global::pjsua2xamarin.Droid.Resource.Attribute.flow_padding;
			global::Xamarin.Forms.Platform.Resource.Attribute.flow_verticalAlign = global::pjsua2xamarin.Droid.Resource.Attribute.flow_verticalAlign;
			global::Xamarin.Forms.Platform.Resource.Attribute.flow_verticalBias = global::pjsua2xamarin.Droid.Resource.Attribute.flow_verticalBias;
			global::Xamarin.Forms.Platform.Resource.Attribute.flow_verticalGap = global::pjsua2xamarin.Droid.Resource.Attribute.flow_verticalGap;
			global::Xamarin.Forms.Platform.Resource.Attribute.flow_verticalStyle = global::pjsua2xamarin.Droid.Resource.Attribute.flow_verticalStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.flow_wrapMode = global::pjsua2xamarin.Droid.Resource.Attribute.flow_wrapMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.font = global::pjsua2xamarin.Droid.Resource.Attribute.font;
			global::Xamarin.Forms.Platform.Resource.Attribute.fontFamily = global::pjsua2xamarin.Droid.Resource.Attribute.fontFamily;
			global::Xamarin.Forms.Platform.Resource.Attribute.fontProviderAuthority = global::pjsua2xamarin.Droid.Resource.Attribute.fontProviderAuthority;
			global::Xamarin.Forms.Platform.Resource.Attribute.fontProviderCerts = global::pjsua2xamarin.Droid.Resource.Attribute.fontProviderCerts;
			global::Xamarin.Forms.Platform.Resource.Attribute.fontProviderFetchStrategy = global::pjsua2xamarin.Droid.Resource.Attribute.fontProviderFetchStrategy;
			global::Xamarin.Forms.Platform.Resource.Attribute.fontProviderFetchTimeout = global::pjsua2xamarin.Droid.Resource.Attribute.fontProviderFetchTimeout;
			global::Xamarin.Forms.Platform.Resource.Attribute.fontProviderPackage = global::pjsua2xamarin.Droid.Resource.Attribute.fontProviderPackage;
			global::Xamarin.Forms.Platform.Resource.Attribute.fontProviderQuery = global::pjsua2xamarin.Droid.Resource.Attribute.fontProviderQuery;
			global::Xamarin.Forms.Platform.Resource.Attribute.fontProviderSystemFontFamily = global::pjsua2xamarin.Droid.Resource.Attribute.fontProviderSystemFontFamily;
			global::Xamarin.Forms.Platform.Resource.Attribute.fontStyle = global::pjsua2xamarin.Droid.Resource.Attribute.fontStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.fontVariationSettings = global::pjsua2xamarin.Droid.Resource.Attribute.fontVariationSettings;
			global::Xamarin.Forms.Platform.Resource.Attribute.fontWeight = global::pjsua2xamarin.Droid.Resource.Attribute.fontWeight;
			global::Xamarin.Forms.Platform.Resource.Attribute.forceApplySystemWindowInsetTop = global::pjsua2xamarin.Droid.Resource.Attribute.forceApplySystemWindowInsetTop;
			global::Xamarin.Forms.Platform.Resource.Attribute.foregroundInsidePadding = global::pjsua2xamarin.Droid.Resource.Attribute.foregroundInsidePadding;
			global::Xamarin.Forms.Platform.Resource.Attribute.framePosition = global::pjsua2xamarin.Droid.Resource.Attribute.framePosition;
			global::Xamarin.Forms.Platform.Resource.Attribute.gapBetweenBars = global::pjsua2xamarin.Droid.Resource.Attribute.gapBetweenBars;
			global::Xamarin.Forms.Platform.Resource.Attribute.gestureInsetBottomIgnored = global::pjsua2xamarin.Droid.Resource.Attribute.gestureInsetBottomIgnored;
			global::Xamarin.Forms.Platform.Resource.Attribute.goIcon = global::pjsua2xamarin.Droid.Resource.Attribute.goIcon;
			global::Xamarin.Forms.Platform.Resource.Attribute.graph = global::pjsua2xamarin.Droid.Resource.Attribute.graph;
			global::Xamarin.Forms.Platform.Resource.Attribute.haloColor = global::pjsua2xamarin.Droid.Resource.Attribute.haloColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.haloRadius = global::pjsua2xamarin.Droid.Resource.Attribute.haloRadius;
			global::Xamarin.Forms.Platform.Resource.Attribute.headerLayout = global::pjsua2xamarin.Droid.Resource.Attribute.headerLayout;
			global::Xamarin.Forms.Platform.Resource.Attribute.height = global::pjsua2xamarin.Droid.Resource.Attribute.height;
			global::Xamarin.Forms.Platform.Resource.Attribute.helperText = global::pjsua2xamarin.Droid.Resource.Attribute.helperText;
			global::Xamarin.Forms.Platform.Resource.Attribute.helperTextEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.helperTextEnabled;
			global::Xamarin.Forms.Platform.Resource.Attribute.helperTextTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.helperTextTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Attribute.helperTextTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.helperTextTextColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.hideAnimationBehavior = global::pjsua2xamarin.Droid.Resource.Attribute.hideAnimationBehavior;
			global::Xamarin.Forms.Platform.Resource.Attribute.hideMotionSpec = global::pjsua2xamarin.Droid.Resource.Attribute.hideMotionSpec;
			global::Xamarin.Forms.Platform.Resource.Attribute.hideOnContentScroll = global::pjsua2xamarin.Droid.Resource.Attribute.hideOnContentScroll;
			global::Xamarin.Forms.Platform.Resource.Attribute.hideOnScroll = global::pjsua2xamarin.Droid.Resource.Attribute.hideOnScroll;
			global::Xamarin.Forms.Platform.Resource.Attribute.hintAnimationEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.hintAnimationEnabled;
			global::Xamarin.Forms.Platform.Resource.Attribute.hintEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.hintEnabled;
			global::Xamarin.Forms.Platform.Resource.Attribute.hintTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.hintTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Attribute.hintTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.hintTextColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.homeAsUpIndicator = global::pjsua2xamarin.Droid.Resource.Attribute.homeAsUpIndicator;
			global::Xamarin.Forms.Platform.Resource.Attribute.homeLayout = global::pjsua2xamarin.Droid.Resource.Attribute.homeLayout;
			global::Xamarin.Forms.Platform.Resource.Attribute.horizontalOffset = global::pjsua2xamarin.Droid.Resource.Attribute.horizontalOffset;
			global::Xamarin.Forms.Platform.Resource.Attribute.hoveredFocusedTranslationZ = global::pjsua2xamarin.Droid.Resource.Attribute.hoveredFocusedTranslationZ;
			global::Xamarin.Forms.Platform.Resource.Attribute.icon = global::pjsua2xamarin.Droid.Resource.Attribute.icon;
			global::Xamarin.Forms.Platform.Resource.Attribute.iconEndPadding = global::pjsua2xamarin.Droid.Resource.Attribute.iconEndPadding;
			global::Xamarin.Forms.Platform.Resource.Attribute.iconGravity = global::pjsua2xamarin.Droid.Resource.Attribute.iconGravity;
			global::Xamarin.Forms.Platform.Resource.Attribute.iconifiedByDefault = global::pjsua2xamarin.Droid.Resource.Attribute.iconifiedByDefault;
			global::Xamarin.Forms.Platform.Resource.Attribute.iconPadding = global::pjsua2xamarin.Droid.Resource.Attribute.iconPadding;
			global::Xamarin.Forms.Platform.Resource.Attribute.iconSize = global::pjsua2xamarin.Droid.Resource.Attribute.iconSize;
			global::Xamarin.Forms.Platform.Resource.Attribute.iconStartPadding = global::pjsua2xamarin.Droid.Resource.Attribute.iconStartPadding;
			global::Xamarin.Forms.Platform.Resource.Attribute.iconTint = global::pjsua2xamarin.Droid.Resource.Attribute.iconTint;
			global::Xamarin.Forms.Platform.Resource.Attribute.iconTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.iconTintMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.ifTagNotSet = global::pjsua2xamarin.Droid.Resource.Attribute.ifTagNotSet;
			global::Xamarin.Forms.Platform.Resource.Attribute.ifTagSet = global::pjsua2xamarin.Droid.Resource.Attribute.ifTagSet;
			global::Xamarin.Forms.Platform.Resource.Attribute.imageButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.imageButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.imagePanX = global::pjsua2xamarin.Droid.Resource.Attribute.imagePanX;
			global::Xamarin.Forms.Platform.Resource.Attribute.imagePanY = global::pjsua2xamarin.Droid.Resource.Attribute.imagePanY;
			global::Xamarin.Forms.Platform.Resource.Attribute.imageRotate = global::pjsua2xamarin.Droid.Resource.Attribute.imageRotate;
			global::Xamarin.Forms.Platform.Resource.Attribute.imageZoom = global::pjsua2xamarin.Droid.Resource.Attribute.imageZoom;
			global::Xamarin.Forms.Platform.Resource.Attribute.indeterminateAnimationType = global::pjsua2xamarin.Droid.Resource.Attribute.indeterminateAnimationType;
			global::Xamarin.Forms.Platform.Resource.Attribute.indeterminateProgressStyle = global::pjsua2xamarin.Droid.Resource.Attribute.indeterminateProgressStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.indicatorColor = global::pjsua2xamarin.Droid.Resource.Attribute.indicatorColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.indicatorDirectionCircular = global::pjsua2xamarin.Droid.Resource.Attribute.indicatorDirectionCircular;
			global::Xamarin.Forms.Platform.Resource.Attribute.indicatorDirectionLinear = global::pjsua2xamarin.Droid.Resource.Attribute.indicatorDirectionLinear;
			global::Xamarin.Forms.Platform.Resource.Attribute.indicatorInset = global::pjsua2xamarin.Droid.Resource.Attribute.indicatorInset;
			global::Xamarin.Forms.Platform.Resource.Attribute.indicatorSize = global::pjsua2xamarin.Droid.Resource.Attribute.indicatorSize;
			global::Xamarin.Forms.Platform.Resource.Attribute.initialActivityCount = global::pjsua2xamarin.Droid.Resource.Attribute.initialActivityCount;
			global::Xamarin.Forms.Platform.Resource.Attribute.insetForeground = global::pjsua2xamarin.Droid.Resource.Attribute.insetForeground;
			global::Xamarin.Forms.Platform.Resource.Attribute.isLightTheme = global::pjsua2xamarin.Droid.Resource.Attribute.isLightTheme;
			global::Xamarin.Forms.Platform.Resource.Attribute.isMaterialTheme = global::pjsua2xamarin.Droid.Resource.Attribute.isMaterialTheme;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemBackground = global::pjsua2xamarin.Droid.Resource.Attribute.itemBackground;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemFillColor = global::pjsua2xamarin.Droid.Resource.Attribute.itemFillColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemHorizontalPadding = global::pjsua2xamarin.Droid.Resource.Attribute.itemHorizontalPadding;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemHorizontalTranslationEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.itemHorizontalTranslationEnabled;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemIconPadding = global::pjsua2xamarin.Droid.Resource.Attribute.itemIconPadding;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemIconSize = global::pjsua2xamarin.Droid.Resource.Attribute.itemIconSize;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemIconTint = global::pjsua2xamarin.Droid.Resource.Attribute.itemIconTint;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemMaxLines = global::pjsua2xamarin.Droid.Resource.Attribute.itemMaxLines;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemPadding = global::pjsua2xamarin.Droid.Resource.Attribute.itemPadding;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemRippleColor = global::pjsua2xamarin.Droid.Resource.Attribute.itemRippleColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemShapeAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.itemShapeAppearance;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemShapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Attribute.itemShapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemShapeFillColor = global::pjsua2xamarin.Droid.Resource.Attribute.itemShapeFillColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemShapeInsetBottom = global::pjsua2xamarin.Droid.Resource.Attribute.itemShapeInsetBottom;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemShapeInsetEnd = global::pjsua2xamarin.Droid.Resource.Attribute.itemShapeInsetEnd;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemShapeInsetStart = global::pjsua2xamarin.Droid.Resource.Attribute.itemShapeInsetStart;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemShapeInsetTop = global::pjsua2xamarin.Droid.Resource.Attribute.itemShapeInsetTop;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemSpacing = global::pjsua2xamarin.Droid.Resource.Attribute.itemSpacing;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemStrokeColor = global::pjsua2xamarin.Droid.Resource.Attribute.itemStrokeColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemStrokeWidth = global::pjsua2xamarin.Droid.Resource.Attribute.itemStrokeWidth;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.itemTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemTextAppearanceActive = global::pjsua2xamarin.Droid.Resource.Attribute.itemTextAppearanceActive;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemTextAppearanceInactive = global::pjsua2xamarin.Droid.Resource.Attribute.itemTextAppearanceInactive;
			global::Xamarin.Forms.Platform.Resource.Attribute.itemTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.itemTextColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.keyboardIcon = global::pjsua2xamarin.Droid.Resource.Attribute.keyboardIcon;
			global::Xamarin.Forms.Platform.Resource.Attribute.keylines = global::pjsua2xamarin.Droid.Resource.Attribute.keylines;
			global::Xamarin.Forms.Platform.Resource.Attribute.keyPositionType = global::pjsua2xamarin.Droid.Resource.Attribute.keyPositionType;
			global::Xamarin.Forms.Platform.Resource.Attribute.labelBehavior = global::pjsua2xamarin.Droid.Resource.Attribute.labelBehavior;
			global::Xamarin.Forms.Platform.Resource.Attribute.labelStyle = global::pjsua2xamarin.Droid.Resource.Attribute.labelStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.labelVisibilityMode = global::pjsua2xamarin.Droid.Resource.Attribute.labelVisibilityMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.lastBaselineToBottomHeight = global::pjsua2xamarin.Droid.Resource.Attribute.lastBaselineToBottomHeight;
			global::Xamarin.Forms.Platform.Resource.Attribute.launchSingleTop = global::pjsua2xamarin.Droid.Resource.Attribute.launchSingleTop;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout = global::pjsua2xamarin.Droid.Resource.Attribute.layout;
			global::Xamarin.Forms.Platform.Resource.Attribute.layoutDescription = global::pjsua2xamarin.Droid.Resource.Attribute.layoutDescription;
			global::Xamarin.Forms.Platform.Resource.Attribute.layoutDuringTransition = global::pjsua2xamarin.Droid.Resource.Attribute.layoutDuringTransition;
			global::Xamarin.Forms.Platform.Resource.Attribute.layoutManager = global::pjsua2xamarin.Droid.Resource.Attribute.layoutManager;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_anchor = global::pjsua2xamarin.Droid.Resource.Attribute.layout_anchor;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_anchorGravity = global::pjsua2xamarin.Droid.Resource.Attribute.layout_anchorGravity;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_behavior = global::pjsua2xamarin.Droid.Resource.Attribute.layout_behavior;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_collapseMode = global::pjsua2xamarin.Droid.Resource.Attribute.layout_collapseMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_collapseParallaxMultiplier = global::pjsua2xamarin.Droid.Resource.Attribute.layout_collapseParallaxMultiplier;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constrainedHeight = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constrainedHeight;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constrainedWidth = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constrainedWidth;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintBaseline_creator = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintBaseline_creator;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintBaseline_toBaselineOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintBaseline_toBaselineOf;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintBaseline_toBottomOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintBaseline_toBottomOf;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintBaseline_toTopOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintBaseline_toTopOf;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintBottom_creator = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintBottom_creator;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintBottom_toBottomOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintBottom_toBottomOf;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintBottom_toTopOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintBottom_toTopOf;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintCircle = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintCircle;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintCircleAngle = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintCircleAngle;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintCircleRadius = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintCircleRadius;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintDimensionRatio = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintDimensionRatio;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintEnd_toEndOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintEnd_toEndOf;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintEnd_toStartOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintEnd_toStartOf;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintGuide_begin = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintGuide_begin;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintGuide_end = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintGuide_end;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintGuide_percent = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintGuide_percent;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintHeight = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintHeight;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintHeight_default = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintHeight_default;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintHeight_max = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintHeight_max;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintHeight_min = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintHeight_min;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintHeight_percent = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintHeight_percent;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintHorizontal_bias = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintHorizontal_bias;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintHorizontal_chainStyle = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintHorizontal_chainStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintHorizontal_weight = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintHorizontal_weight;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintLeft_creator = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintLeft_creator;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintLeft_toLeftOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintLeft_toLeftOf;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintLeft_toRightOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintLeft_toRightOf;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintRight_creator = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintRight_creator;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintRight_toLeftOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintRight_toLeftOf;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintRight_toRightOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintRight_toRightOf;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintStart_toEndOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintStart_toEndOf;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintStart_toStartOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintStart_toStartOf;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintTag = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintTag;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintTop_creator = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintTop_creator;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintTop_toBottomOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintTop_toBottomOf;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintTop_toTopOf = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintTop_toTopOf;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintVertical_bias = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintVertical_bias;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintVertical_chainStyle = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintVertical_chainStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintVertical_weight = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintVertical_weight;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintWidth = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintWidth;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintWidth_default = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintWidth_default;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintWidth_max = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintWidth_max;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintWidth_min = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintWidth_min;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_constraintWidth_percent = global::pjsua2xamarin.Droid.Resource.Attribute.layout_constraintWidth_percent;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_dodgeInsetEdges = global::pjsua2xamarin.Droid.Resource.Attribute.layout_dodgeInsetEdges;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_editor_absoluteX = global::pjsua2xamarin.Droid.Resource.Attribute.layout_editor_absoluteX;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_editor_absoluteY = global::pjsua2xamarin.Droid.Resource.Attribute.layout_editor_absoluteY;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_goneMarginBaseline = global::pjsua2xamarin.Droid.Resource.Attribute.layout_goneMarginBaseline;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_goneMarginBottom = global::pjsua2xamarin.Droid.Resource.Attribute.layout_goneMarginBottom;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_goneMarginEnd = global::pjsua2xamarin.Droid.Resource.Attribute.layout_goneMarginEnd;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_goneMarginLeft = global::pjsua2xamarin.Droid.Resource.Attribute.layout_goneMarginLeft;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_goneMarginRight = global::pjsua2xamarin.Droid.Resource.Attribute.layout_goneMarginRight;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_goneMarginStart = global::pjsua2xamarin.Droid.Resource.Attribute.layout_goneMarginStart;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_goneMarginTop = global::pjsua2xamarin.Droid.Resource.Attribute.layout_goneMarginTop;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_insetEdge = global::pjsua2xamarin.Droid.Resource.Attribute.layout_insetEdge;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_keyline = global::pjsua2xamarin.Droid.Resource.Attribute.layout_keyline;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_marginBaseline = global::pjsua2xamarin.Droid.Resource.Attribute.layout_marginBaseline;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_optimizationLevel = global::pjsua2xamarin.Droid.Resource.Attribute.layout_optimizationLevel;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_scrollFlags = global::pjsua2xamarin.Droid.Resource.Attribute.layout_scrollFlags;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_scrollInterpolator = global::pjsua2xamarin.Droid.Resource.Attribute.layout_scrollInterpolator;
			global::Xamarin.Forms.Platform.Resource.Attribute.layout_wrapBehaviorInParent = global::pjsua2xamarin.Droid.Resource.Attribute.layout_wrapBehaviorInParent;
			global::Xamarin.Forms.Platform.Resource.Attribute.liftOnScroll = global::pjsua2xamarin.Droid.Resource.Attribute.liftOnScroll;
			global::Xamarin.Forms.Platform.Resource.Attribute.liftOnScrollTargetViewId = global::pjsua2xamarin.Droid.Resource.Attribute.liftOnScrollTargetViewId;
			global::Xamarin.Forms.Platform.Resource.Attribute.limitBoundsTo = global::pjsua2xamarin.Droid.Resource.Attribute.limitBoundsTo;
			global::Xamarin.Forms.Platform.Resource.Attribute.linearProgressIndicatorStyle = global::pjsua2xamarin.Droid.Resource.Attribute.linearProgressIndicatorStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.lineHeight = global::pjsua2xamarin.Droid.Resource.Attribute.lineHeight;
			global::Xamarin.Forms.Platform.Resource.Attribute.lineSpacing = global::pjsua2xamarin.Droid.Resource.Attribute.lineSpacing;
			global::Xamarin.Forms.Platform.Resource.Attribute.listChoiceBackgroundIndicator = global::pjsua2xamarin.Droid.Resource.Attribute.listChoiceBackgroundIndicator;
			global::Xamarin.Forms.Platform.Resource.Attribute.listChoiceIndicatorMultipleAnimated = global::pjsua2xamarin.Droid.Resource.Attribute.listChoiceIndicatorMultipleAnimated;
			global::Xamarin.Forms.Platform.Resource.Attribute.listChoiceIndicatorSingleAnimated = global::pjsua2xamarin.Droid.Resource.Attribute.listChoiceIndicatorSingleAnimated;
			global::Xamarin.Forms.Platform.Resource.Attribute.listDividerAlertDialog = global::pjsua2xamarin.Droid.Resource.Attribute.listDividerAlertDialog;
			global::Xamarin.Forms.Platform.Resource.Attribute.listItemLayout = global::pjsua2xamarin.Droid.Resource.Attribute.listItemLayout;
			global::Xamarin.Forms.Platform.Resource.Attribute.listLayout = global::pjsua2xamarin.Droid.Resource.Attribute.listLayout;
			global::Xamarin.Forms.Platform.Resource.Attribute.listMenuViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.listMenuViewStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.listPopupWindowStyle = global::pjsua2xamarin.Droid.Resource.Attribute.listPopupWindowStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.listPreferredItemHeight = global::pjsua2xamarin.Droid.Resource.Attribute.listPreferredItemHeight;
			global::Xamarin.Forms.Platform.Resource.Attribute.listPreferredItemHeightLarge = global::pjsua2xamarin.Droid.Resource.Attribute.listPreferredItemHeightLarge;
			global::Xamarin.Forms.Platform.Resource.Attribute.listPreferredItemHeightSmall = global::pjsua2xamarin.Droid.Resource.Attribute.listPreferredItemHeightSmall;
			global::Xamarin.Forms.Platform.Resource.Attribute.listPreferredItemPaddingEnd = global::pjsua2xamarin.Droid.Resource.Attribute.listPreferredItemPaddingEnd;
			global::Xamarin.Forms.Platform.Resource.Attribute.listPreferredItemPaddingLeft = global::pjsua2xamarin.Droid.Resource.Attribute.listPreferredItemPaddingLeft;
			global::Xamarin.Forms.Platform.Resource.Attribute.listPreferredItemPaddingRight = global::pjsua2xamarin.Droid.Resource.Attribute.listPreferredItemPaddingRight;
			global::Xamarin.Forms.Platform.Resource.Attribute.listPreferredItemPaddingStart = global::pjsua2xamarin.Droid.Resource.Attribute.listPreferredItemPaddingStart;
			global::Xamarin.Forms.Platform.Resource.Attribute.logo = global::pjsua2xamarin.Droid.Resource.Attribute.logo;
			global::Xamarin.Forms.Platform.Resource.Attribute.logoDescription = global::pjsua2xamarin.Droid.Resource.Attribute.logoDescription;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialAlertDialogBodyTextStyle = global::pjsua2xamarin.Droid.Resource.Attribute.materialAlertDialogBodyTextStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialAlertDialogTheme = global::pjsua2xamarin.Droid.Resource.Attribute.materialAlertDialogTheme;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialAlertDialogTitleIconStyle = global::pjsua2xamarin.Droid.Resource.Attribute.materialAlertDialogTitleIconStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialAlertDialogTitlePanelStyle = global::pjsua2xamarin.Droid.Resource.Attribute.materialAlertDialogTitlePanelStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialAlertDialogTitleTextStyle = global::pjsua2xamarin.Droid.Resource.Attribute.materialAlertDialogTitleTextStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialButtonOutlinedStyle = global::pjsua2xamarin.Droid.Resource.Attribute.materialButtonOutlinedStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.materialButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialButtonToggleGroupStyle = global::pjsua2xamarin.Droid.Resource.Attribute.materialButtonToggleGroupStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialCalendarDay = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarDay;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialCalendarFullscreenTheme = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarFullscreenTheme;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialCalendarHeaderCancelButton = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarHeaderCancelButton;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialCalendarHeaderConfirmButton = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarHeaderConfirmButton;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialCalendarHeaderDivider = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarHeaderDivider;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialCalendarHeaderLayout = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarHeaderLayout;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialCalendarHeaderSelection = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarHeaderSelection;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialCalendarHeaderTitle = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarHeaderTitle;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialCalendarHeaderToggleButton = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarHeaderToggleButton;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialCalendarMonth = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarMonth;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialCalendarMonthNavigationButton = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarMonthNavigationButton;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialCalendarStyle = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialCalendarTheme = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarTheme;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialCalendarYearNavigationButton = global::pjsua2xamarin.Droid.Resource.Attribute.materialCalendarYearNavigationButton;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialCardViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.materialCardViewStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialCircleRadius = global::pjsua2xamarin.Droid.Resource.Attribute.materialCircleRadius;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialClockStyle = global::pjsua2xamarin.Droid.Resource.Attribute.materialClockStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialThemeOverlay = global::pjsua2xamarin.Droid.Resource.Attribute.materialThemeOverlay;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialTimePickerStyle = global::pjsua2xamarin.Droid.Resource.Attribute.materialTimePickerStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.materialTimePickerTheme = global::pjsua2xamarin.Droid.Resource.Attribute.materialTimePickerTheme;
			global::Xamarin.Forms.Platform.Resource.Attribute.maxAcceleration = global::pjsua2xamarin.Droid.Resource.Attribute.maxAcceleration;
			global::Xamarin.Forms.Platform.Resource.Attribute.maxActionInlineWidth = global::pjsua2xamarin.Droid.Resource.Attribute.maxActionInlineWidth;
			global::Xamarin.Forms.Platform.Resource.Attribute.maxButtonHeight = global::pjsua2xamarin.Droid.Resource.Attribute.maxButtonHeight;
			global::Xamarin.Forms.Platform.Resource.Attribute.maxCharacterCount = global::pjsua2xamarin.Droid.Resource.Attribute.maxCharacterCount;
			global::Xamarin.Forms.Platform.Resource.Attribute.maxHeight = global::pjsua2xamarin.Droid.Resource.Attribute.maxHeight;
			global::Xamarin.Forms.Platform.Resource.Attribute.maxImageSize = global::pjsua2xamarin.Droid.Resource.Attribute.maxImageSize;
			global::Xamarin.Forms.Platform.Resource.Attribute.maxLines = global::pjsua2xamarin.Droid.Resource.Attribute.maxLines;
			global::Xamarin.Forms.Platform.Resource.Attribute.maxVelocity = global::pjsua2xamarin.Droid.Resource.Attribute.maxVelocity;
			global::Xamarin.Forms.Platform.Resource.Attribute.maxWidth = global::pjsua2xamarin.Droid.Resource.Attribute.maxWidth;
			global::Xamarin.Forms.Platform.Resource.Attribute.measureWithLargestChild = global::pjsua2xamarin.Droid.Resource.Attribute.measureWithLargestChild;
			global::Xamarin.Forms.Platform.Resource.Attribute.menu = global::pjsua2xamarin.Droid.Resource.Attribute.menu;
			global::Xamarin.Forms.Platform.Resource.Attribute.menuGravity = global::pjsua2xamarin.Droid.Resource.Attribute.menuGravity;
			global::Xamarin.Forms.Platform.Resource.Attribute.methodName = global::pjsua2xamarin.Droid.Resource.Attribute.methodName;
			global::Xamarin.Forms.Platform.Resource.Attribute.mimeType = global::pjsua2xamarin.Droid.Resource.Attribute.mimeType;
			global::Xamarin.Forms.Platform.Resource.Attribute.minHeight = global::pjsua2xamarin.Droid.Resource.Attribute.minHeight;
			global::Xamarin.Forms.Platform.Resource.Attribute.minHideDelay = global::pjsua2xamarin.Droid.Resource.Attribute.minHideDelay;
			global::Xamarin.Forms.Platform.Resource.Attribute.minSeparation = global::pjsua2xamarin.Droid.Resource.Attribute.minSeparation;
			global::Xamarin.Forms.Platform.Resource.Attribute.minTouchTargetSize = global::pjsua2xamarin.Droid.Resource.Attribute.minTouchTargetSize;
			global::Xamarin.Forms.Platform.Resource.Attribute.minWidth = global::pjsua2xamarin.Droid.Resource.Attribute.minWidth;
			global::Xamarin.Forms.Platform.Resource.Attribute.mock_diagonalsColor = global::pjsua2xamarin.Droid.Resource.Attribute.mock_diagonalsColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.mock_label = global::pjsua2xamarin.Droid.Resource.Attribute.mock_label;
			global::Xamarin.Forms.Platform.Resource.Attribute.mock_labelBackgroundColor = global::pjsua2xamarin.Droid.Resource.Attribute.mock_labelBackgroundColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.mock_labelColor = global::pjsua2xamarin.Droid.Resource.Attribute.mock_labelColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.mock_showDiagonals = global::pjsua2xamarin.Droid.Resource.Attribute.mock_showDiagonals;
			global::Xamarin.Forms.Platform.Resource.Attribute.mock_showLabel = global::pjsua2xamarin.Droid.Resource.Attribute.mock_showLabel;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionDebug = global::pjsua2xamarin.Droid.Resource.Attribute.motionDebug;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionDurationLong1 = global::pjsua2xamarin.Droid.Resource.Attribute.motionDurationLong1;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionDurationLong2 = global::pjsua2xamarin.Droid.Resource.Attribute.motionDurationLong2;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionDurationMedium1 = global::pjsua2xamarin.Droid.Resource.Attribute.motionDurationMedium1;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionDurationMedium2 = global::pjsua2xamarin.Droid.Resource.Attribute.motionDurationMedium2;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionDurationShort1 = global::pjsua2xamarin.Droid.Resource.Attribute.motionDurationShort1;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionDurationShort2 = global::pjsua2xamarin.Droid.Resource.Attribute.motionDurationShort2;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionEasingAccelerated = global::pjsua2xamarin.Droid.Resource.Attribute.motionEasingAccelerated;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionEasingDecelerated = global::pjsua2xamarin.Droid.Resource.Attribute.motionEasingDecelerated;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionEasingEmphasized = global::pjsua2xamarin.Droid.Resource.Attribute.motionEasingEmphasized;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionEasingLinear = global::pjsua2xamarin.Droid.Resource.Attribute.motionEasingLinear;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionEasingStandard = global::pjsua2xamarin.Droid.Resource.Attribute.motionEasingStandard;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionEffect_alpha = global::pjsua2xamarin.Droid.Resource.Attribute.motionEffect_alpha;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionEffect_end = global::pjsua2xamarin.Droid.Resource.Attribute.motionEffect_end;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionEffect_move = global::pjsua2xamarin.Droid.Resource.Attribute.motionEffect_move;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionEffect_start = global::pjsua2xamarin.Droid.Resource.Attribute.motionEffect_start;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionEffect_strict = global::pjsua2xamarin.Droid.Resource.Attribute.motionEffect_strict;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionEffect_translationX = global::pjsua2xamarin.Droid.Resource.Attribute.motionEffect_translationX;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionEffect_translationY = global::pjsua2xamarin.Droid.Resource.Attribute.motionEffect_translationY;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionEffect_viewTransition = global::pjsua2xamarin.Droid.Resource.Attribute.motionEffect_viewTransition;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionInterpolator = global::pjsua2xamarin.Droid.Resource.Attribute.motionInterpolator;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionPath = global::pjsua2xamarin.Droid.Resource.Attribute.motionPath;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionPathRotate = global::pjsua2xamarin.Droid.Resource.Attribute.motionPathRotate;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionProgress = global::pjsua2xamarin.Droid.Resource.Attribute.motionProgress;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionStagger = global::pjsua2xamarin.Droid.Resource.Attribute.motionStagger;
			global::Xamarin.Forms.Platform.Resource.Attribute.motionTarget = global::pjsua2xamarin.Droid.Resource.Attribute.motionTarget;
			global::Xamarin.Forms.Platform.Resource.Attribute.motion_postLayoutCollision = global::pjsua2xamarin.Droid.Resource.Attribute.motion_postLayoutCollision;
			global::Xamarin.Forms.Platform.Resource.Attribute.motion_triggerOnCollision = global::pjsua2xamarin.Droid.Resource.Attribute.motion_triggerOnCollision;
			global::Xamarin.Forms.Platform.Resource.Attribute.moveWhenScrollAtTop = global::pjsua2xamarin.Droid.Resource.Attribute.moveWhenScrollAtTop;
			global::Xamarin.Forms.Platform.Resource.Attribute.multiChoiceItemLayout = global::pjsua2xamarin.Droid.Resource.Attribute.multiChoiceItemLayout;
			global::Xamarin.Forms.Platform.Resource.Attribute.navGraph = global::pjsua2xamarin.Droid.Resource.Attribute.navGraph;
			global::Xamarin.Forms.Platform.Resource.Attribute.navigationContentDescription = global::pjsua2xamarin.Droid.Resource.Attribute.navigationContentDescription;
			global::Xamarin.Forms.Platform.Resource.Attribute.navigationIcon = global::pjsua2xamarin.Droid.Resource.Attribute.navigationIcon;
			global::Xamarin.Forms.Platform.Resource.Attribute.navigationIconTint = global::pjsua2xamarin.Droid.Resource.Attribute.navigationIconTint;
			global::Xamarin.Forms.Platform.Resource.Attribute.navigationMode = global::pjsua2xamarin.Droid.Resource.Attribute.navigationMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.navigationRailStyle = global::pjsua2xamarin.Droid.Resource.Attribute.navigationRailStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.navigationViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.navigationViewStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.nestedScrollable = global::pjsua2xamarin.Droid.Resource.Attribute.nestedScrollable;
			global::Xamarin.Forms.Platform.Resource.Attribute.nestedScrollFlags = global::pjsua2xamarin.Droid.Resource.Attribute.nestedScrollFlags;
			global::Xamarin.Forms.Platform.Resource.Attribute.nestedScrollViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.nestedScrollViewStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.nullable = global::pjsua2xamarin.Droid.Resource.Attribute.nullable;
			global::Xamarin.Forms.Platform.Resource.Attribute.number = global::pjsua2xamarin.Droid.Resource.Attribute.number;
			global::Xamarin.Forms.Platform.Resource.Attribute.numericModifiers = global::pjsua2xamarin.Droid.Resource.Attribute.numericModifiers;
			global::Xamarin.Forms.Platform.Resource.Attribute.onCross = global::pjsua2xamarin.Droid.Resource.Attribute.onCross;
			global::Xamarin.Forms.Platform.Resource.Attribute.onHide = global::pjsua2xamarin.Droid.Resource.Attribute.onHide;
			global::Xamarin.Forms.Platform.Resource.Attribute.onNegativeCross = global::pjsua2xamarin.Droid.Resource.Attribute.onNegativeCross;
			global::Xamarin.Forms.Platform.Resource.Attribute.onPositiveCross = global::pjsua2xamarin.Droid.Resource.Attribute.onPositiveCross;
			global::Xamarin.Forms.Platform.Resource.Attribute.onShow = global::pjsua2xamarin.Droid.Resource.Attribute.onShow;
			global::Xamarin.Forms.Platform.Resource.Attribute.onStateTransition = global::pjsua2xamarin.Droid.Resource.Attribute.onStateTransition;
			global::Xamarin.Forms.Platform.Resource.Attribute.onTouchUp = global::pjsua2xamarin.Droid.Resource.Attribute.onTouchUp;
			global::Xamarin.Forms.Platform.Resource.Attribute.overlapAnchor = global::pjsua2xamarin.Droid.Resource.Attribute.overlapAnchor;
			global::Xamarin.Forms.Platform.Resource.Attribute.overlay = global::pjsua2xamarin.Droid.Resource.Attribute.overlay;
			global::Xamarin.Forms.Platform.Resource.Attribute.paddingBottomNoButtons = global::pjsua2xamarin.Droid.Resource.Attribute.paddingBottomNoButtons;
			global::Xamarin.Forms.Platform.Resource.Attribute.paddingBottomSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Attribute.paddingBottomSystemWindowInsets;
			global::Xamarin.Forms.Platform.Resource.Attribute.paddingEnd = global::pjsua2xamarin.Droid.Resource.Attribute.paddingEnd;
			global::Xamarin.Forms.Platform.Resource.Attribute.paddingLeftSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Attribute.paddingLeftSystemWindowInsets;
			global::Xamarin.Forms.Platform.Resource.Attribute.paddingRightSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Attribute.paddingRightSystemWindowInsets;
			global::Xamarin.Forms.Platform.Resource.Attribute.paddingStart = global::pjsua2xamarin.Droid.Resource.Attribute.paddingStart;
			global::Xamarin.Forms.Platform.Resource.Attribute.paddingTopNoTitle = global::pjsua2xamarin.Droid.Resource.Attribute.paddingTopNoTitle;
			global::Xamarin.Forms.Platform.Resource.Attribute.paddingTopSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Attribute.paddingTopSystemWindowInsets;
			global::Xamarin.Forms.Platform.Resource.Attribute.panelBackground = global::pjsua2xamarin.Droid.Resource.Attribute.panelBackground;
			global::Xamarin.Forms.Platform.Resource.Attribute.panelMenuListTheme = global::pjsua2xamarin.Droid.Resource.Attribute.panelMenuListTheme;
			global::Xamarin.Forms.Platform.Resource.Attribute.panelMenuListWidth = global::pjsua2xamarin.Droid.Resource.Attribute.panelMenuListWidth;
			global::Xamarin.Forms.Platform.Resource.Attribute.passwordToggleContentDescription = global::pjsua2xamarin.Droid.Resource.Attribute.passwordToggleContentDescription;
			global::Xamarin.Forms.Platform.Resource.Attribute.passwordToggleDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.passwordToggleDrawable;
			global::Xamarin.Forms.Platform.Resource.Attribute.passwordToggleEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.passwordToggleEnabled;
			global::Xamarin.Forms.Platform.Resource.Attribute.passwordToggleTint = global::pjsua2xamarin.Droid.Resource.Attribute.passwordToggleTint;
			global::Xamarin.Forms.Platform.Resource.Attribute.passwordToggleTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.passwordToggleTintMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.pathMotionArc = global::pjsua2xamarin.Droid.Resource.Attribute.pathMotionArc;
			global::Xamarin.Forms.Platform.Resource.Attribute.path_percent = global::pjsua2xamarin.Droid.Resource.Attribute.path_percent;
			global::Xamarin.Forms.Platform.Resource.Attribute.percentHeight = global::pjsua2xamarin.Droid.Resource.Attribute.percentHeight;
			global::Xamarin.Forms.Platform.Resource.Attribute.percentWidth = global::pjsua2xamarin.Droid.Resource.Attribute.percentWidth;
			global::Xamarin.Forms.Platform.Resource.Attribute.percentX = global::pjsua2xamarin.Droid.Resource.Attribute.percentX;
			global::Xamarin.Forms.Platform.Resource.Attribute.percentY = global::pjsua2xamarin.Droid.Resource.Attribute.percentY;
			global::Xamarin.Forms.Platform.Resource.Attribute.perpendicularPath_percent = global::pjsua2xamarin.Droid.Resource.Attribute.perpendicularPath_percent;
			global::Xamarin.Forms.Platform.Resource.Attribute.pivotAnchor = global::pjsua2xamarin.Droid.Resource.Attribute.pivotAnchor;
			global::Xamarin.Forms.Platform.Resource.Attribute.placeholderText = global::pjsua2xamarin.Droid.Resource.Attribute.placeholderText;
			global::Xamarin.Forms.Platform.Resource.Attribute.placeholderTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.placeholderTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Attribute.placeholderTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.placeholderTextColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.placeholder_emptyVisibility = global::pjsua2xamarin.Droid.Resource.Attribute.placeholder_emptyVisibility;
			global::Xamarin.Forms.Platform.Resource.Attribute.polarRelativeTo = global::pjsua2xamarin.Droid.Resource.Attribute.polarRelativeTo;
			global::Xamarin.Forms.Platform.Resource.Attribute.popEnterAnim = global::pjsua2xamarin.Droid.Resource.Attribute.popEnterAnim;
			global::Xamarin.Forms.Platform.Resource.Attribute.popExitAnim = global::pjsua2xamarin.Droid.Resource.Attribute.popExitAnim;
			global::Xamarin.Forms.Platform.Resource.Attribute.popupMenuBackground = global::pjsua2xamarin.Droid.Resource.Attribute.popupMenuBackground;
			global::Xamarin.Forms.Platform.Resource.Attribute.popupMenuStyle = global::pjsua2xamarin.Droid.Resource.Attribute.popupMenuStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.popupTheme = global::pjsua2xamarin.Droid.Resource.Attribute.popupTheme;
			global::Xamarin.Forms.Platform.Resource.Attribute.popUpTo = global::pjsua2xamarin.Droid.Resource.Attribute.popUpTo;
			global::Xamarin.Forms.Platform.Resource.Attribute.popUpToInclusive = global::pjsua2xamarin.Droid.Resource.Attribute.popUpToInclusive;
			global::Xamarin.Forms.Platform.Resource.Attribute.popupWindowStyle = global::pjsua2xamarin.Droid.Resource.Attribute.popupWindowStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.prefixText = global::pjsua2xamarin.Droid.Resource.Attribute.prefixText;
			global::Xamarin.Forms.Platform.Resource.Attribute.prefixTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.prefixTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Attribute.prefixTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.prefixTextColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.preserveIconSpacing = global::pjsua2xamarin.Droid.Resource.Attribute.preserveIconSpacing;
			global::Xamarin.Forms.Platform.Resource.Attribute.pressedTranslationZ = global::pjsua2xamarin.Droid.Resource.Attribute.pressedTranslationZ;
			global::Xamarin.Forms.Platform.Resource.Attribute.progressBarPadding = global::pjsua2xamarin.Droid.Resource.Attribute.progressBarPadding;
			global::Xamarin.Forms.Platform.Resource.Attribute.progressBarStyle = global::pjsua2xamarin.Droid.Resource.Attribute.progressBarStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.quantizeMotionInterpolator = global::pjsua2xamarin.Droid.Resource.Attribute.quantizeMotionInterpolator;
			global::Xamarin.Forms.Platform.Resource.Attribute.quantizeMotionPhase = global::pjsua2xamarin.Droid.Resource.Attribute.quantizeMotionPhase;
			global::Xamarin.Forms.Platform.Resource.Attribute.quantizeMotionSteps = global::pjsua2xamarin.Droid.Resource.Attribute.quantizeMotionSteps;
			global::Xamarin.Forms.Platform.Resource.Attribute.queryBackground = global::pjsua2xamarin.Droid.Resource.Attribute.queryBackground;
			global::Xamarin.Forms.Platform.Resource.Attribute.queryHint = global::pjsua2xamarin.Droid.Resource.Attribute.queryHint;
			global::Xamarin.Forms.Platform.Resource.Attribute.queryPatterns = global::pjsua2xamarin.Droid.Resource.Attribute.queryPatterns;
			global::Xamarin.Forms.Platform.Resource.Attribute.radioButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.radioButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.rangeFillColor = global::pjsua2xamarin.Droid.Resource.Attribute.rangeFillColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.ratingBarStyle = global::pjsua2xamarin.Droid.Resource.Attribute.ratingBarStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.ratingBarStyleIndicator = global::pjsua2xamarin.Droid.Resource.Attribute.ratingBarStyleIndicator;
			global::Xamarin.Forms.Platform.Resource.Attribute.ratingBarStyleSmall = global::pjsua2xamarin.Droid.Resource.Attribute.ratingBarStyleSmall;
			global::Xamarin.Forms.Platform.Resource.Attribute.reactiveGuide_animateChange = global::pjsua2xamarin.Droid.Resource.Attribute.reactiveGuide_animateChange;
			global::Xamarin.Forms.Platform.Resource.Attribute.reactiveGuide_applyToAllConstraintSets = global::pjsua2xamarin.Droid.Resource.Attribute.reactiveGuide_applyToAllConstraintSets;
			global::Xamarin.Forms.Platform.Resource.Attribute.reactiveGuide_applyToConstraintSet = global::pjsua2xamarin.Droid.Resource.Attribute.reactiveGuide_applyToConstraintSet;
			global::Xamarin.Forms.Platform.Resource.Attribute.reactiveGuide_valueId = global::pjsua2xamarin.Droid.Resource.Attribute.reactiveGuide_valueId;
			global::Xamarin.Forms.Platform.Resource.Attribute.recyclerViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.recyclerViewStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.region_heightLessThan = global::pjsua2xamarin.Droid.Resource.Attribute.region_heightLessThan;
			global::Xamarin.Forms.Platform.Resource.Attribute.region_heightMoreThan = global::pjsua2xamarin.Droid.Resource.Attribute.region_heightMoreThan;
			global::Xamarin.Forms.Platform.Resource.Attribute.region_widthLessThan = global::pjsua2xamarin.Droid.Resource.Attribute.region_widthLessThan;
			global::Xamarin.Forms.Platform.Resource.Attribute.region_widthMoreThan = global::pjsua2xamarin.Droid.Resource.Attribute.region_widthMoreThan;
			global::Xamarin.Forms.Platform.Resource.Attribute.reverseLayout = global::pjsua2xamarin.Droid.Resource.Attribute.reverseLayout;
			global::Xamarin.Forms.Platform.Resource.Attribute.rippleColor = global::pjsua2xamarin.Droid.Resource.Attribute.rippleColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.rotationCenterId = global::pjsua2xamarin.Droid.Resource.Attribute.rotationCenterId;
			global::Xamarin.Forms.Platform.Resource.Attribute.round = global::pjsua2xamarin.Droid.Resource.Attribute.round;
			global::Xamarin.Forms.Platform.Resource.Attribute.roundPercent = global::pjsua2xamarin.Droid.Resource.Attribute.roundPercent;
			global::Xamarin.Forms.Platform.Resource.Attribute.saturation = global::pjsua2xamarin.Droid.Resource.Attribute.saturation;
			global::Xamarin.Forms.Platform.Resource.Attribute.scaleFromTextSize = global::pjsua2xamarin.Droid.Resource.Attribute.scaleFromTextSize;
			global::Xamarin.Forms.Platform.Resource.Attribute.scrimAnimationDuration = global::pjsua2xamarin.Droid.Resource.Attribute.scrimAnimationDuration;
			global::Xamarin.Forms.Platform.Resource.Attribute.scrimBackground = global::pjsua2xamarin.Droid.Resource.Attribute.scrimBackground;
			global::Xamarin.Forms.Platform.Resource.Attribute.scrimVisibleHeightTrigger = global::pjsua2xamarin.Droid.Resource.Attribute.scrimVisibleHeightTrigger;
			global::Xamarin.Forms.Platform.Resource.Attribute.scrollViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.scrollViewStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.searchHintIcon = global::pjsua2xamarin.Droid.Resource.Attribute.searchHintIcon;
			global::Xamarin.Forms.Platform.Resource.Attribute.searchIcon = global::pjsua2xamarin.Droid.Resource.Attribute.searchIcon;
			global::Xamarin.Forms.Platform.Resource.Attribute.searchViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.searchViewStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.seekBarStyle = global::pjsua2xamarin.Droid.Resource.Attribute.seekBarStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.selectableItemBackground = global::pjsua2xamarin.Droid.Resource.Attribute.selectableItemBackground;
			global::Xamarin.Forms.Platform.Resource.Attribute.selectableItemBackgroundBorderless = global::pjsua2xamarin.Droid.Resource.Attribute.selectableItemBackgroundBorderless;
			global::Xamarin.Forms.Platform.Resource.Attribute.selectionRequired = global::pjsua2xamarin.Droid.Resource.Attribute.selectionRequired;
			global::Xamarin.Forms.Platform.Resource.Attribute.selectorSize = global::pjsua2xamarin.Droid.Resource.Attribute.selectorSize;
			global::Xamarin.Forms.Platform.Resource.Attribute.setsTag = global::pjsua2xamarin.Droid.Resource.Attribute.setsTag;
			global::Xamarin.Forms.Platform.Resource.Attribute.shapeAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.shapeAppearance;
			global::Xamarin.Forms.Platform.Resource.Attribute.shapeAppearanceLargeComponent = global::pjsua2xamarin.Droid.Resource.Attribute.shapeAppearanceLargeComponent;
			global::Xamarin.Forms.Platform.Resource.Attribute.shapeAppearanceMediumComponent = global::pjsua2xamarin.Droid.Resource.Attribute.shapeAppearanceMediumComponent;
			global::Xamarin.Forms.Platform.Resource.Attribute.shapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Attribute.shapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Resource.Attribute.shapeAppearanceSmallComponent = global::pjsua2xamarin.Droid.Resource.Attribute.shapeAppearanceSmallComponent;
			global::Xamarin.Forms.Platform.Resource.Attribute.SharedValue = global::pjsua2xamarin.Droid.Resource.Attribute.SharedValue;
			global::Xamarin.Forms.Platform.Resource.Attribute.SharedValueId = global::pjsua2xamarin.Droid.Resource.Attribute.SharedValueId;
			global::Xamarin.Forms.Platform.Resource.Attribute.shortcutMatchRequired = global::pjsua2xamarin.Droid.Resource.Attribute.shortcutMatchRequired;
			global::Xamarin.Forms.Platform.Resource.Attribute.showAnimationBehavior = global::pjsua2xamarin.Droid.Resource.Attribute.showAnimationBehavior;
			global::Xamarin.Forms.Platform.Resource.Attribute.showAsAction = global::pjsua2xamarin.Droid.Resource.Attribute.showAsAction;
			global::Xamarin.Forms.Platform.Resource.Attribute.showDelay = global::pjsua2xamarin.Droid.Resource.Attribute.showDelay;
			global::Xamarin.Forms.Platform.Resource.Attribute.showDividers = global::pjsua2xamarin.Droid.Resource.Attribute.showDividers;
			global::Xamarin.Forms.Platform.Resource.Attribute.showMotionSpec = global::pjsua2xamarin.Droid.Resource.Attribute.showMotionSpec;
			global::Xamarin.Forms.Platform.Resource.Attribute.showPaths = global::pjsua2xamarin.Droid.Resource.Attribute.showPaths;
			global::Xamarin.Forms.Platform.Resource.Attribute.showText = global::pjsua2xamarin.Droid.Resource.Attribute.showText;
			global::Xamarin.Forms.Platform.Resource.Attribute.showTitle = global::pjsua2xamarin.Droid.Resource.Attribute.showTitle;
			global::Xamarin.Forms.Platform.Resource.Attribute.shrinkMotionSpec = global::pjsua2xamarin.Droid.Resource.Attribute.shrinkMotionSpec;
			global::Xamarin.Forms.Platform.Resource.Attribute.singleChoiceItemLayout = global::pjsua2xamarin.Droid.Resource.Attribute.singleChoiceItemLayout;
			global::Xamarin.Forms.Platform.Resource.Attribute.singleLine = global::pjsua2xamarin.Droid.Resource.Attribute.singleLine;
			global::Xamarin.Forms.Platform.Resource.Attribute.singleSelection = global::pjsua2xamarin.Droid.Resource.Attribute.singleSelection;
			global::Xamarin.Forms.Platform.Resource.Attribute.sizePercent = global::pjsua2xamarin.Droid.Resource.Attribute.sizePercent;
			global::Xamarin.Forms.Platform.Resource.Attribute.sliderStyle = global::pjsua2xamarin.Droid.Resource.Attribute.sliderStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.snackbarButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.snackbarButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.snackbarStyle = global::pjsua2xamarin.Droid.Resource.Attribute.snackbarStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.snackbarTextViewStyle = global::pjsua2xamarin.Droid.Resource.Attribute.snackbarTextViewStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.spanCount = global::pjsua2xamarin.Droid.Resource.Attribute.spanCount;
			global::Xamarin.Forms.Platform.Resource.Attribute.spinBars = global::pjsua2xamarin.Droid.Resource.Attribute.spinBars;
			global::Xamarin.Forms.Platform.Resource.Attribute.spinnerDropDownItemStyle = global::pjsua2xamarin.Droid.Resource.Attribute.spinnerDropDownItemStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.spinnerStyle = global::pjsua2xamarin.Droid.Resource.Attribute.spinnerStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.splitTrack = global::pjsua2xamarin.Droid.Resource.Attribute.splitTrack;
			global::Xamarin.Forms.Platform.Resource.Attribute.springBoundary = global::pjsua2xamarin.Droid.Resource.Attribute.springBoundary;
			global::Xamarin.Forms.Platform.Resource.Attribute.springDamping = global::pjsua2xamarin.Droid.Resource.Attribute.springDamping;
			global::Xamarin.Forms.Platform.Resource.Attribute.springMass = global::pjsua2xamarin.Droid.Resource.Attribute.springMass;
			global::Xamarin.Forms.Platform.Resource.Attribute.springStiffness = global::pjsua2xamarin.Droid.Resource.Attribute.springStiffness;
			global::Xamarin.Forms.Platform.Resource.Attribute.springStopThreshold = global::pjsua2xamarin.Droid.Resource.Attribute.springStopThreshold;
			global::Xamarin.Forms.Platform.Resource.Attribute.srcCompat = global::pjsua2xamarin.Droid.Resource.Attribute.srcCompat;
			global::Xamarin.Forms.Platform.Resource.Attribute.stackFromEnd = global::pjsua2xamarin.Droid.Resource.Attribute.stackFromEnd;
			global::Xamarin.Forms.Platform.Resource.Attribute.staggered = global::pjsua2xamarin.Droid.Resource.Attribute.staggered;
			global::Xamarin.Forms.Platform.Resource.Attribute.startDestination = global::pjsua2xamarin.Droid.Resource.Attribute.startDestination;
			global::Xamarin.Forms.Platform.Resource.Attribute.startIconCheckable = global::pjsua2xamarin.Droid.Resource.Attribute.startIconCheckable;
			global::Xamarin.Forms.Platform.Resource.Attribute.startIconContentDescription = global::pjsua2xamarin.Droid.Resource.Attribute.startIconContentDescription;
			global::Xamarin.Forms.Platform.Resource.Attribute.startIconDrawable = global::pjsua2xamarin.Droid.Resource.Attribute.startIconDrawable;
			global::Xamarin.Forms.Platform.Resource.Attribute.startIconTint = global::pjsua2xamarin.Droid.Resource.Attribute.startIconTint;
			global::Xamarin.Forms.Platform.Resource.Attribute.startIconTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.startIconTintMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.state_above_anchor = global::pjsua2xamarin.Droid.Resource.Attribute.state_above_anchor;
			global::Xamarin.Forms.Platform.Resource.Attribute.state_collapsed = global::pjsua2xamarin.Droid.Resource.Attribute.state_collapsed;
			global::Xamarin.Forms.Platform.Resource.Attribute.state_collapsible = global::pjsua2xamarin.Droid.Resource.Attribute.state_collapsible;
			global::Xamarin.Forms.Platform.Resource.Attribute.state_dragged = global::pjsua2xamarin.Droid.Resource.Attribute.state_dragged;
			global::Xamarin.Forms.Platform.Resource.Attribute.state_liftable = global::pjsua2xamarin.Droid.Resource.Attribute.state_liftable;
			global::Xamarin.Forms.Platform.Resource.Attribute.state_lifted = global::pjsua2xamarin.Droid.Resource.Attribute.state_lifted;
			global::Xamarin.Forms.Platform.Resource.Attribute.statusBarBackground = global::pjsua2xamarin.Droid.Resource.Attribute.statusBarBackground;
			global::Xamarin.Forms.Platform.Resource.Attribute.statusBarForeground = global::pjsua2xamarin.Droid.Resource.Attribute.statusBarForeground;
			global::Xamarin.Forms.Platform.Resource.Attribute.statusBarScrim = global::pjsua2xamarin.Droid.Resource.Attribute.statusBarScrim;
			global::Xamarin.Forms.Platform.Resource.Attribute.strokeColor = global::pjsua2xamarin.Droid.Resource.Attribute.strokeColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.strokeWidth = global::pjsua2xamarin.Droid.Resource.Attribute.strokeWidth;
			global::Xamarin.Forms.Platform.Resource.Attribute.subMenuArrow = global::pjsua2xamarin.Droid.Resource.Attribute.subMenuArrow;
			global::Xamarin.Forms.Platform.Resource.Attribute.submitBackground = global::pjsua2xamarin.Droid.Resource.Attribute.submitBackground;
			global::Xamarin.Forms.Platform.Resource.Attribute.subtitle = global::pjsua2xamarin.Droid.Resource.Attribute.subtitle;
			global::Xamarin.Forms.Platform.Resource.Attribute.subtitleCentered = global::pjsua2xamarin.Droid.Resource.Attribute.subtitleCentered;
			global::Xamarin.Forms.Platform.Resource.Attribute.subtitleTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.subtitleTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Attribute.subtitleTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.subtitleTextColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.subtitleTextStyle = global::pjsua2xamarin.Droid.Resource.Attribute.subtitleTextStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.suffixText = global::pjsua2xamarin.Droid.Resource.Attribute.suffixText;
			global::Xamarin.Forms.Platform.Resource.Attribute.suffixTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.suffixTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Attribute.suffixTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.suffixTextColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.suggestionRowLayout = global::pjsua2xamarin.Droid.Resource.Attribute.suggestionRowLayout;
			global::Xamarin.Forms.Platform.Resource.Attribute.swipeRefreshLayoutProgressSpinnerBackgroundColor = global::pjsua2xamarin.Droid.Resource.Attribute.swipeRefreshLayoutProgressSpinnerBackgroundColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.switchMinWidth = global::pjsua2xamarin.Droid.Resource.Attribute.switchMinWidth;
			global::Xamarin.Forms.Platform.Resource.Attribute.switchPadding = global::pjsua2xamarin.Droid.Resource.Attribute.switchPadding;
			global::Xamarin.Forms.Platform.Resource.Attribute.switchStyle = global::pjsua2xamarin.Droid.Resource.Attribute.switchStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.switchTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.switchTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabBackground = global::pjsua2xamarin.Droid.Resource.Attribute.tabBackground;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabContentStart = global::pjsua2xamarin.Droid.Resource.Attribute.tabContentStart;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabGravity = global::pjsua2xamarin.Droid.Resource.Attribute.tabGravity;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabIconTint = global::pjsua2xamarin.Droid.Resource.Attribute.tabIconTint;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabIconTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.tabIconTintMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabIndicator = global::pjsua2xamarin.Droid.Resource.Attribute.tabIndicator;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabIndicatorAnimationDuration = global::pjsua2xamarin.Droid.Resource.Attribute.tabIndicatorAnimationDuration;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabIndicatorAnimationMode = global::pjsua2xamarin.Droid.Resource.Attribute.tabIndicatorAnimationMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabIndicatorColor = global::pjsua2xamarin.Droid.Resource.Attribute.tabIndicatorColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabIndicatorFullWidth = global::pjsua2xamarin.Droid.Resource.Attribute.tabIndicatorFullWidth;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabIndicatorGravity = global::pjsua2xamarin.Droid.Resource.Attribute.tabIndicatorGravity;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabIndicatorHeight = global::pjsua2xamarin.Droid.Resource.Attribute.tabIndicatorHeight;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabInlineLabel = global::pjsua2xamarin.Droid.Resource.Attribute.tabInlineLabel;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabMaxWidth = global::pjsua2xamarin.Droid.Resource.Attribute.tabMaxWidth;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabMinWidth = global::pjsua2xamarin.Droid.Resource.Attribute.tabMinWidth;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabMode = global::pjsua2xamarin.Droid.Resource.Attribute.tabMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabPadding = global::pjsua2xamarin.Droid.Resource.Attribute.tabPadding;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabPaddingBottom = global::pjsua2xamarin.Droid.Resource.Attribute.tabPaddingBottom;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabPaddingEnd = global::pjsua2xamarin.Droid.Resource.Attribute.tabPaddingEnd;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabPaddingStart = global::pjsua2xamarin.Droid.Resource.Attribute.tabPaddingStart;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabPaddingTop = global::pjsua2xamarin.Droid.Resource.Attribute.tabPaddingTop;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabRippleColor = global::pjsua2xamarin.Droid.Resource.Attribute.tabRippleColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabSelectedTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.tabSelectedTextColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabStyle = global::pjsua2xamarin.Droid.Resource.Attribute.tabStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.tabTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.tabTextColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.tabUnboundedRipple = global::pjsua2xamarin.Droid.Resource.Attribute.tabUnboundedRipple;
			global::Xamarin.Forms.Platform.Resource.Attribute.targetId = global::pjsua2xamarin.Droid.Resource.Attribute.targetId;
			global::Xamarin.Forms.Platform.Resource.Attribute.targetPackage = global::pjsua2xamarin.Droid.Resource.Attribute.targetPackage;
			global::Xamarin.Forms.Platform.Resource.Attribute.telltales_tailColor = global::pjsua2xamarin.Droid.Resource.Attribute.telltales_tailColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.telltales_tailScale = global::pjsua2xamarin.Droid.Resource.Attribute.telltales_tailScale;
			global::Xamarin.Forms.Platform.Resource.Attribute.telltales_velocityMode = global::pjsua2xamarin.Droid.Resource.Attribute.telltales_velocityMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAllCaps = global::pjsua2xamarin.Droid.Resource.Attribute.textAllCaps;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAppearanceBody1 = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceBody1;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAppearanceBody2 = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceBody2;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAppearanceButton = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceButton;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAppearanceCaption = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceCaption;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAppearanceHeadline1 = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceHeadline1;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAppearanceHeadline2 = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceHeadline2;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAppearanceHeadline3 = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceHeadline3;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAppearanceHeadline4 = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceHeadline4;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAppearanceHeadline5 = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceHeadline5;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAppearanceHeadline6 = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceHeadline6;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAppearanceLargePopupMenu = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceLargePopupMenu;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAppearanceLineHeightEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceLineHeightEnabled;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAppearanceListItem = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceListItem;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAppearanceListItemSecondary = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceListItemSecondary;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAppearanceListItemSmall = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceListItemSmall;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAppearanceOverline = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceOverline;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAppearancePopupMenuHeader = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearancePopupMenuHeader;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAppearanceSearchResultSubtitle = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceSearchResultSubtitle;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAppearanceSearchResultTitle = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceSearchResultTitle;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAppearanceSmallPopupMenu = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceSmallPopupMenu;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAppearanceSubtitle1 = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceSubtitle1;
			global::Xamarin.Forms.Platform.Resource.Attribute.textAppearanceSubtitle2 = global::pjsua2xamarin.Droid.Resource.Attribute.textAppearanceSubtitle2;
			global::Xamarin.Forms.Platform.Resource.Attribute.textBackground = global::pjsua2xamarin.Droid.Resource.Attribute.textBackground;
			global::Xamarin.Forms.Platform.Resource.Attribute.textBackgroundPanX = global::pjsua2xamarin.Droid.Resource.Attribute.textBackgroundPanX;
			global::Xamarin.Forms.Platform.Resource.Attribute.textBackgroundPanY = global::pjsua2xamarin.Droid.Resource.Attribute.textBackgroundPanY;
			global::Xamarin.Forms.Platform.Resource.Attribute.textBackgroundRotate = global::pjsua2xamarin.Droid.Resource.Attribute.textBackgroundRotate;
			global::Xamarin.Forms.Platform.Resource.Attribute.textBackgroundZoom = global::pjsua2xamarin.Droid.Resource.Attribute.textBackgroundZoom;
			global::Xamarin.Forms.Platform.Resource.Attribute.textColorAlertDialogListItem = global::pjsua2xamarin.Droid.Resource.Attribute.textColorAlertDialogListItem;
			global::Xamarin.Forms.Platform.Resource.Attribute.textColorSearchUrl = global::pjsua2xamarin.Droid.Resource.Attribute.textColorSearchUrl;
			global::Xamarin.Forms.Platform.Resource.Attribute.textEndPadding = global::pjsua2xamarin.Droid.Resource.Attribute.textEndPadding;
			global::Xamarin.Forms.Platform.Resource.Attribute.textFillColor = global::pjsua2xamarin.Droid.Resource.Attribute.textFillColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.textInputLayoutFocusedRectEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.textInputLayoutFocusedRectEnabled;
			global::Xamarin.Forms.Platform.Resource.Attribute.textInputStyle = global::pjsua2xamarin.Droid.Resource.Attribute.textInputStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.textLocale = global::pjsua2xamarin.Droid.Resource.Attribute.textLocale;
			global::Xamarin.Forms.Platform.Resource.Attribute.textOutlineColor = global::pjsua2xamarin.Droid.Resource.Attribute.textOutlineColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.textOutlineThickness = global::pjsua2xamarin.Droid.Resource.Attribute.textOutlineThickness;
			global::Xamarin.Forms.Platform.Resource.Attribute.textPanX = global::pjsua2xamarin.Droid.Resource.Attribute.textPanX;
			global::Xamarin.Forms.Platform.Resource.Attribute.textPanY = global::pjsua2xamarin.Droid.Resource.Attribute.textPanY;
			global::Xamarin.Forms.Platform.Resource.Attribute.textStartPadding = global::pjsua2xamarin.Droid.Resource.Attribute.textStartPadding;
			global::Xamarin.Forms.Platform.Resource.Attribute.textureBlurFactor = global::pjsua2xamarin.Droid.Resource.Attribute.textureBlurFactor;
			global::Xamarin.Forms.Platform.Resource.Attribute.textureEffect = global::pjsua2xamarin.Droid.Resource.Attribute.textureEffect;
			global::Xamarin.Forms.Platform.Resource.Attribute.textureHeight = global::pjsua2xamarin.Droid.Resource.Attribute.textureHeight;
			global::Xamarin.Forms.Platform.Resource.Attribute.textureWidth = global::pjsua2xamarin.Droid.Resource.Attribute.textureWidth;
			global::Xamarin.Forms.Platform.Resource.Attribute.theme = global::pjsua2xamarin.Droid.Resource.Attribute.theme;
			global::Xamarin.Forms.Platform.Resource.Attribute.themeLineHeight = global::pjsua2xamarin.Droid.Resource.Attribute.themeLineHeight;
			global::Xamarin.Forms.Platform.Resource.Attribute.thickness = global::pjsua2xamarin.Droid.Resource.Attribute.thickness;
			global::Xamarin.Forms.Platform.Resource.Attribute.thumbColor = global::pjsua2xamarin.Droid.Resource.Attribute.thumbColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.thumbElevation = global::pjsua2xamarin.Droid.Resource.Attribute.thumbElevation;
			global::Xamarin.Forms.Platform.Resource.Attribute.thumbRadius = global::pjsua2xamarin.Droid.Resource.Attribute.thumbRadius;
			global::Xamarin.Forms.Platform.Resource.Attribute.thumbStrokeColor = global::pjsua2xamarin.Droid.Resource.Attribute.thumbStrokeColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.thumbStrokeWidth = global::pjsua2xamarin.Droid.Resource.Attribute.thumbStrokeWidth;
			global::Xamarin.Forms.Platform.Resource.Attribute.thumbTextPadding = global::pjsua2xamarin.Droid.Resource.Attribute.thumbTextPadding;
			global::Xamarin.Forms.Platform.Resource.Attribute.thumbTint = global::pjsua2xamarin.Droid.Resource.Attribute.thumbTint;
			global::Xamarin.Forms.Platform.Resource.Attribute.thumbTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.thumbTintMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.tickColor = global::pjsua2xamarin.Droid.Resource.Attribute.tickColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.tickColorActive = global::pjsua2xamarin.Droid.Resource.Attribute.tickColorActive;
			global::Xamarin.Forms.Platform.Resource.Attribute.tickColorInactive = global::pjsua2xamarin.Droid.Resource.Attribute.tickColorInactive;
			global::Xamarin.Forms.Platform.Resource.Attribute.tickMark = global::pjsua2xamarin.Droid.Resource.Attribute.tickMark;
			global::Xamarin.Forms.Platform.Resource.Attribute.tickMarkTint = global::pjsua2xamarin.Droid.Resource.Attribute.tickMarkTint;
			global::Xamarin.Forms.Platform.Resource.Attribute.tickMarkTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.tickMarkTintMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.tickVisible = global::pjsua2xamarin.Droid.Resource.Attribute.tickVisible;
			global::Xamarin.Forms.Platform.Resource.Attribute.tint = global::pjsua2xamarin.Droid.Resource.Attribute.tint;
			global::Xamarin.Forms.Platform.Resource.Attribute.tintMode = global::pjsua2xamarin.Droid.Resource.Attribute.tintMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.title = global::pjsua2xamarin.Droid.Resource.Attribute.title;
			global::Xamarin.Forms.Platform.Resource.Attribute.titleCentered = global::pjsua2xamarin.Droid.Resource.Attribute.titleCentered;
			global::Xamarin.Forms.Platform.Resource.Attribute.titleCollapseMode = global::pjsua2xamarin.Droid.Resource.Attribute.titleCollapseMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.titleEnabled = global::pjsua2xamarin.Droid.Resource.Attribute.titleEnabled;
			global::Xamarin.Forms.Platform.Resource.Attribute.titleMargin = global::pjsua2xamarin.Droid.Resource.Attribute.titleMargin;
			global::Xamarin.Forms.Platform.Resource.Attribute.titleMarginBottom = global::pjsua2xamarin.Droid.Resource.Attribute.titleMarginBottom;
			global::Xamarin.Forms.Platform.Resource.Attribute.titleMarginEnd = global::pjsua2xamarin.Droid.Resource.Attribute.titleMarginEnd;
			global::Xamarin.Forms.Platform.Resource.Attribute.titleMargins = global::pjsua2xamarin.Droid.Resource.Attribute.titleMargins;
			global::Xamarin.Forms.Platform.Resource.Attribute.titleMarginStart = global::pjsua2xamarin.Droid.Resource.Attribute.titleMarginStart;
			global::Xamarin.Forms.Platform.Resource.Attribute.titleMarginTop = global::pjsua2xamarin.Droid.Resource.Attribute.titleMarginTop;
			global::Xamarin.Forms.Platform.Resource.Attribute.titleTextAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.titleTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Attribute.titleTextColor = global::pjsua2xamarin.Droid.Resource.Attribute.titleTextColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.titleTextStyle = global::pjsua2xamarin.Droid.Resource.Attribute.titleTextStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.toolbarId = global::pjsua2xamarin.Droid.Resource.Attribute.toolbarId;
			global::Xamarin.Forms.Platform.Resource.Attribute.toolbarNavigationButtonStyle = global::pjsua2xamarin.Droid.Resource.Attribute.toolbarNavigationButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.toolbarStyle = global::pjsua2xamarin.Droid.Resource.Attribute.toolbarStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.tooltipForegroundColor = global::pjsua2xamarin.Droid.Resource.Attribute.tooltipForegroundColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.tooltipFrameBackground = global::pjsua2xamarin.Droid.Resource.Attribute.tooltipFrameBackground;
			global::Xamarin.Forms.Platform.Resource.Attribute.tooltipStyle = global::pjsua2xamarin.Droid.Resource.Attribute.tooltipStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.tooltipText = global::pjsua2xamarin.Droid.Resource.Attribute.tooltipText;
			global::Xamarin.Forms.Platform.Resource.Attribute.touchAnchorId = global::pjsua2xamarin.Droid.Resource.Attribute.touchAnchorId;
			global::Xamarin.Forms.Platform.Resource.Attribute.touchAnchorSide = global::pjsua2xamarin.Droid.Resource.Attribute.touchAnchorSide;
			global::Xamarin.Forms.Platform.Resource.Attribute.touchRegionId = global::pjsua2xamarin.Droid.Resource.Attribute.touchRegionId;
			global::Xamarin.Forms.Platform.Resource.Attribute.track = global::pjsua2xamarin.Droid.Resource.Attribute.track;
			global::Xamarin.Forms.Platform.Resource.Attribute.trackColor = global::pjsua2xamarin.Droid.Resource.Attribute.trackColor;
			global::Xamarin.Forms.Platform.Resource.Attribute.trackColorActive = global::pjsua2xamarin.Droid.Resource.Attribute.trackColorActive;
			global::Xamarin.Forms.Platform.Resource.Attribute.trackColorInactive = global::pjsua2xamarin.Droid.Resource.Attribute.trackColorInactive;
			global::Xamarin.Forms.Platform.Resource.Attribute.trackCornerRadius = global::pjsua2xamarin.Droid.Resource.Attribute.trackCornerRadius;
			global::Xamarin.Forms.Platform.Resource.Attribute.trackHeight = global::pjsua2xamarin.Droid.Resource.Attribute.trackHeight;
			global::Xamarin.Forms.Platform.Resource.Attribute.trackThickness = global::pjsua2xamarin.Droid.Resource.Attribute.trackThickness;
			global::Xamarin.Forms.Platform.Resource.Attribute.trackTint = global::pjsua2xamarin.Droid.Resource.Attribute.trackTint;
			global::Xamarin.Forms.Platform.Resource.Attribute.trackTintMode = global::pjsua2xamarin.Droid.Resource.Attribute.trackTintMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.transformPivotTarget = global::pjsua2xamarin.Droid.Resource.Attribute.transformPivotTarget;
			global::Xamarin.Forms.Platform.Resource.Attribute.transitionDisable = global::pjsua2xamarin.Droid.Resource.Attribute.transitionDisable;
			global::Xamarin.Forms.Platform.Resource.Attribute.transitionEasing = global::pjsua2xamarin.Droid.Resource.Attribute.transitionEasing;
			global::Xamarin.Forms.Platform.Resource.Attribute.transitionFlags = global::pjsua2xamarin.Droid.Resource.Attribute.transitionFlags;
			global::Xamarin.Forms.Platform.Resource.Attribute.transitionPathRotate = global::pjsua2xamarin.Droid.Resource.Attribute.transitionPathRotate;
			global::Xamarin.Forms.Platform.Resource.Attribute.transitionShapeAppearance = global::pjsua2xamarin.Droid.Resource.Attribute.transitionShapeAppearance;
			global::Xamarin.Forms.Platform.Resource.Attribute.triggerId = global::pjsua2xamarin.Droid.Resource.Attribute.triggerId;
			global::Xamarin.Forms.Platform.Resource.Attribute.triggerReceiver = global::pjsua2xamarin.Droid.Resource.Attribute.triggerReceiver;
			global::Xamarin.Forms.Platform.Resource.Attribute.triggerSlack = global::pjsua2xamarin.Droid.Resource.Attribute.triggerSlack;
			global::Xamarin.Forms.Platform.Resource.Attribute.ttcIndex = global::pjsua2xamarin.Droid.Resource.Attribute.ttcIndex;
			global::Xamarin.Forms.Platform.Resource.Attribute.upDuration = global::pjsua2xamarin.Droid.Resource.Attribute.upDuration;
			global::Xamarin.Forms.Platform.Resource.Attribute.uri = global::pjsua2xamarin.Droid.Resource.Attribute.uri;
			global::Xamarin.Forms.Platform.Resource.Attribute.useCompatPadding = global::pjsua2xamarin.Droid.Resource.Attribute.useCompatPadding;
			global::Xamarin.Forms.Platform.Resource.Attribute.useMaterialThemeColors = global::pjsua2xamarin.Droid.Resource.Attribute.useMaterialThemeColors;
			global::Xamarin.Forms.Platform.Resource.Attribute.values = global::pjsua2xamarin.Droid.Resource.Attribute.values;
			global::Xamarin.Forms.Platform.Resource.Attribute.verticalOffset = global::pjsua2xamarin.Droid.Resource.Attribute.verticalOffset;
			global::Xamarin.Forms.Platform.Resource.Attribute.viewInflaterClass = global::pjsua2xamarin.Droid.Resource.Attribute.viewInflaterClass;
			global::Xamarin.Forms.Platform.Resource.Attribute.viewTransitionMode = global::pjsua2xamarin.Droid.Resource.Attribute.viewTransitionMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.viewTransitionOnCross = global::pjsua2xamarin.Droid.Resource.Attribute.viewTransitionOnCross;
			global::Xamarin.Forms.Platform.Resource.Attribute.viewTransitionOnNegativeCross = global::pjsua2xamarin.Droid.Resource.Attribute.viewTransitionOnNegativeCross;
			global::Xamarin.Forms.Platform.Resource.Attribute.viewTransitionOnPositiveCross = global::pjsua2xamarin.Droid.Resource.Attribute.viewTransitionOnPositiveCross;
			global::Xamarin.Forms.Platform.Resource.Attribute.visibilityMode = global::pjsua2xamarin.Droid.Resource.Attribute.visibilityMode;
			global::Xamarin.Forms.Platform.Resource.Attribute.voiceIcon = global::pjsua2xamarin.Droid.Resource.Attribute.voiceIcon;
			global::Xamarin.Forms.Platform.Resource.Attribute.warmth = global::pjsua2xamarin.Droid.Resource.Attribute.warmth;
			global::Xamarin.Forms.Platform.Resource.Attribute.waveDecay = global::pjsua2xamarin.Droid.Resource.Attribute.waveDecay;
			global::Xamarin.Forms.Platform.Resource.Attribute.waveOffset = global::pjsua2xamarin.Droid.Resource.Attribute.waveOffset;
			global::Xamarin.Forms.Platform.Resource.Attribute.wavePeriod = global::pjsua2xamarin.Droid.Resource.Attribute.wavePeriod;
			global::Xamarin.Forms.Platform.Resource.Attribute.wavePhase = global::pjsua2xamarin.Droid.Resource.Attribute.wavePhase;
			global::Xamarin.Forms.Platform.Resource.Attribute.waveShape = global::pjsua2xamarin.Droid.Resource.Attribute.waveShape;
			global::Xamarin.Forms.Platform.Resource.Attribute.waveVariesBy = global::pjsua2xamarin.Droid.Resource.Attribute.waveVariesBy;
			global::Xamarin.Forms.Platform.Resource.Attribute.windowActionBar = global::pjsua2xamarin.Droid.Resource.Attribute.windowActionBar;
			global::Xamarin.Forms.Platform.Resource.Attribute.windowActionBarOverlay = global::pjsua2xamarin.Droid.Resource.Attribute.windowActionBarOverlay;
			global::Xamarin.Forms.Platform.Resource.Attribute.windowActionModeOverlay = global::pjsua2xamarin.Droid.Resource.Attribute.windowActionModeOverlay;
			global::Xamarin.Forms.Platform.Resource.Attribute.windowFixedHeightMajor = global::pjsua2xamarin.Droid.Resource.Attribute.windowFixedHeightMajor;
			global::Xamarin.Forms.Platform.Resource.Attribute.windowFixedHeightMinor = global::pjsua2xamarin.Droid.Resource.Attribute.windowFixedHeightMinor;
			global::Xamarin.Forms.Platform.Resource.Attribute.windowFixedWidthMajor = global::pjsua2xamarin.Droid.Resource.Attribute.windowFixedWidthMajor;
			global::Xamarin.Forms.Platform.Resource.Attribute.windowFixedWidthMinor = global::pjsua2xamarin.Droid.Resource.Attribute.windowFixedWidthMinor;
			global::Xamarin.Forms.Platform.Resource.Attribute.windowMinWidthMajor = global::pjsua2xamarin.Droid.Resource.Attribute.windowMinWidthMajor;
			global::Xamarin.Forms.Platform.Resource.Attribute.windowMinWidthMinor = global::pjsua2xamarin.Droid.Resource.Attribute.windowMinWidthMinor;
			global::Xamarin.Forms.Platform.Resource.Attribute.windowNoTitle = global::pjsua2xamarin.Droid.Resource.Attribute.windowNoTitle;
			global::Xamarin.Forms.Platform.Resource.Attribute.yearSelectedStyle = global::pjsua2xamarin.Droid.Resource.Attribute.yearSelectedStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.yearStyle = global::pjsua2xamarin.Droid.Resource.Attribute.yearStyle;
			global::Xamarin.Forms.Platform.Resource.Attribute.yearTodayStyle = global::pjsua2xamarin.Droid.Resource.Attribute.yearTodayStyle;
			global::Xamarin.Forms.Platform.Resource.Boolean.abc_action_bar_embed_tabs = global::pjsua2xamarin.Droid.Resource.Boolean.abc_action_bar_embed_tabs;
			global::Xamarin.Forms.Platform.Resource.Boolean.abc_config_actionMenuItemAllCaps = global::pjsua2xamarin.Droid.Resource.Boolean.abc_config_actionMenuItemAllCaps;
			global::Xamarin.Forms.Platform.Resource.Boolean.mtrl_btn_textappearance_all_caps = global::pjsua2xamarin.Droid.Resource.Boolean.mtrl_btn_textappearance_all_caps;
			global::Xamarin.Forms.Platform.Resource.Color.abc_background_cache_hint_selector_material_dark = global::pjsua2xamarin.Droid.Resource.Color.abc_background_cache_hint_selector_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.abc_background_cache_hint_selector_material_light = global::pjsua2xamarin.Droid.Resource.Color.abc_background_cache_hint_selector_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.abc_btn_colored_borderless_text_material = global::pjsua2xamarin.Droid.Resource.Color.abc_btn_colored_borderless_text_material;
			global::Xamarin.Forms.Platform.Resource.Color.abc_btn_colored_text_material = global::pjsua2xamarin.Droid.Resource.Color.abc_btn_colored_text_material;
			global::Xamarin.Forms.Platform.Resource.Color.abc_color_highlight_material = global::pjsua2xamarin.Droid.Resource.Color.abc_color_highlight_material;
			global::Xamarin.Forms.Platform.Resource.Color.abc_decor_view_status_guard = global::pjsua2xamarin.Droid.Resource.Color.abc_decor_view_status_guard;
			global::Xamarin.Forms.Platform.Resource.Color.abc_decor_view_status_guard_light = global::pjsua2xamarin.Droid.Resource.Color.abc_decor_view_status_guard_light;
			global::Xamarin.Forms.Platform.Resource.Color.abc_hint_foreground_material_dark = global::pjsua2xamarin.Droid.Resource.Color.abc_hint_foreground_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.abc_hint_foreground_material_light = global::pjsua2xamarin.Droid.Resource.Color.abc_hint_foreground_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.abc_primary_text_disable_only_material_dark = global::pjsua2xamarin.Droid.Resource.Color.abc_primary_text_disable_only_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.abc_primary_text_disable_only_material_light = global::pjsua2xamarin.Droid.Resource.Color.abc_primary_text_disable_only_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.abc_primary_text_material_dark = global::pjsua2xamarin.Droid.Resource.Color.abc_primary_text_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.abc_primary_text_material_light = global::pjsua2xamarin.Droid.Resource.Color.abc_primary_text_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.abc_search_url_text = global::pjsua2xamarin.Droid.Resource.Color.abc_search_url_text;
			global::Xamarin.Forms.Platform.Resource.Color.abc_search_url_text_normal = global::pjsua2xamarin.Droid.Resource.Color.abc_search_url_text_normal;
			global::Xamarin.Forms.Platform.Resource.Color.abc_search_url_text_pressed = global::pjsua2xamarin.Droid.Resource.Color.abc_search_url_text_pressed;
			global::Xamarin.Forms.Platform.Resource.Color.abc_search_url_text_selected = global::pjsua2xamarin.Droid.Resource.Color.abc_search_url_text_selected;
			global::Xamarin.Forms.Platform.Resource.Color.abc_secondary_text_material_dark = global::pjsua2xamarin.Droid.Resource.Color.abc_secondary_text_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.abc_secondary_text_material_light = global::pjsua2xamarin.Droid.Resource.Color.abc_secondary_text_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.abc_tint_btn_checkable = global::pjsua2xamarin.Droid.Resource.Color.abc_tint_btn_checkable;
			global::Xamarin.Forms.Platform.Resource.Color.abc_tint_default = global::pjsua2xamarin.Droid.Resource.Color.abc_tint_default;
			global::Xamarin.Forms.Platform.Resource.Color.abc_tint_edittext = global::pjsua2xamarin.Droid.Resource.Color.abc_tint_edittext;
			global::Xamarin.Forms.Platform.Resource.Color.abc_tint_seek_thumb = global::pjsua2xamarin.Droid.Resource.Color.abc_tint_seek_thumb;
			global::Xamarin.Forms.Platform.Resource.Color.abc_tint_spinner = global::pjsua2xamarin.Droid.Resource.Color.abc_tint_spinner;
			global::Xamarin.Forms.Platform.Resource.Color.abc_tint_switch_track = global::pjsua2xamarin.Droid.Resource.Color.abc_tint_switch_track;
			global::Xamarin.Forms.Platform.Resource.Color.accent_material_dark = global::pjsua2xamarin.Droid.Resource.Color.accent_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.accent_material_light = global::pjsua2xamarin.Droid.Resource.Color.accent_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.androidx_core_ripple_material_light = global::pjsua2xamarin.Droid.Resource.Color.androidx_core_ripple_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.androidx_core_secondary_text_default_material_light = global::pjsua2xamarin.Droid.Resource.Color.androidx_core_secondary_text_default_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.background_floating_material_dark = global::pjsua2xamarin.Droid.Resource.Color.background_floating_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.background_floating_material_light = global::pjsua2xamarin.Droid.Resource.Color.background_floating_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.background_material_dark = global::pjsua2xamarin.Droid.Resource.Color.background_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.background_material_light = global::pjsua2xamarin.Droid.Resource.Color.background_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.bright_foreground_disabled_material_dark = global::pjsua2xamarin.Droid.Resource.Color.bright_foreground_disabled_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.bright_foreground_disabled_material_light = global::pjsua2xamarin.Droid.Resource.Color.bright_foreground_disabled_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.bright_foreground_inverse_material_dark = global::pjsua2xamarin.Droid.Resource.Color.bright_foreground_inverse_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.bright_foreground_inverse_material_light = global::pjsua2xamarin.Droid.Resource.Color.bright_foreground_inverse_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.bright_foreground_material_dark = global::pjsua2xamarin.Droid.Resource.Color.bright_foreground_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.bright_foreground_material_light = global::pjsua2xamarin.Droid.Resource.Color.bright_foreground_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.button_material_dark = global::pjsua2xamarin.Droid.Resource.Color.button_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.button_material_light = global::pjsua2xamarin.Droid.Resource.Color.button_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.cardview_dark_background = global::pjsua2xamarin.Droid.Resource.Color.cardview_dark_background;
			global::Xamarin.Forms.Platform.Resource.Color.cardview_light_background = global::pjsua2xamarin.Droid.Resource.Color.cardview_light_background;
			global::Xamarin.Forms.Platform.Resource.Color.cardview_shadow_end_color = global::pjsua2xamarin.Droid.Resource.Color.cardview_shadow_end_color;
			global::Xamarin.Forms.Platform.Resource.Color.cardview_shadow_start_color = global::pjsua2xamarin.Droid.Resource.Color.cardview_shadow_start_color;
			global::Xamarin.Forms.Platform.Resource.Color.checkbox_themeable_attribute_color = global::pjsua2xamarin.Droid.Resource.Color.checkbox_themeable_attribute_color;
			global::Xamarin.Forms.Platform.Resource.Color.design_bottom_navigation_shadow_color = global::pjsua2xamarin.Droid.Resource.Color.design_bottom_navigation_shadow_color;
			global::Xamarin.Forms.Platform.Resource.Color.design_box_stroke_color = global::pjsua2xamarin.Droid.Resource.Color.design_box_stroke_color;
			global::Xamarin.Forms.Platform.Resource.Color.design_dark_default_color_background = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_background;
			global::Xamarin.Forms.Platform.Resource.Color.design_dark_default_color_error = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_error;
			global::Xamarin.Forms.Platform.Resource.Color.design_dark_default_color_on_background = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_on_background;
			global::Xamarin.Forms.Platform.Resource.Color.design_dark_default_color_on_error = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_on_error;
			global::Xamarin.Forms.Platform.Resource.Color.design_dark_default_color_on_primary = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_on_primary;
			global::Xamarin.Forms.Platform.Resource.Color.design_dark_default_color_on_secondary = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_on_secondary;
			global::Xamarin.Forms.Platform.Resource.Color.design_dark_default_color_on_surface = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_on_surface;
			global::Xamarin.Forms.Platform.Resource.Color.design_dark_default_color_primary = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_primary;
			global::Xamarin.Forms.Platform.Resource.Color.design_dark_default_color_primary_dark = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_primary_dark;
			global::Xamarin.Forms.Platform.Resource.Color.design_dark_default_color_primary_variant = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_primary_variant;
			global::Xamarin.Forms.Platform.Resource.Color.design_dark_default_color_secondary = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_secondary;
			global::Xamarin.Forms.Platform.Resource.Color.design_dark_default_color_secondary_variant = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_secondary_variant;
			global::Xamarin.Forms.Platform.Resource.Color.design_dark_default_color_surface = global::pjsua2xamarin.Droid.Resource.Color.design_dark_default_color_surface;
			global::Xamarin.Forms.Platform.Resource.Color.design_default_color_background = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_background;
			global::Xamarin.Forms.Platform.Resource.Color.design_default_color_error = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_error;
			global::Xamarin.Forms.Platform.Resource.Color.design_default_color_on_background = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_on_background;
			global::Xamarin.Forms.Platform.Resource.Color.design_default_color_on_error = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_on_error;
			global::Xamarin.Forms.Platform.Resource.Color.design_default_color_on_primary = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_on_primary;
			global::Xamarin.Forms.Platform.Resource.Color.design_default_color_on_secondary = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_on_secondary;
			global::Xamarin.Forms.Platform.Resource.Color.design_default_color_on_surface = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_on_surface;
			global::Xamarin.Forms.Platform.Resource.Color.design_default_color_primary = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_primary;
			global::Xamarin.Forms.Platform.Resource.Color.design_default_color_primary_dark = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_primary_dark;
			global::Xamarin.Forms.Platform.Resource.Color.design_default_color_primary_variant = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_primary_variant;
			global::Xamarin.Forms.Platform.Resource.Color.design_default_color_secondary = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_secondary;
			global::Xamarin.Forms.Platform.Resource.Color.design_default_color_secondary_variant = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_secondary_variant;
			global::Xamarin.Forms.Platform.Resource.Color.design_default_color_surface = global::pjsua2xamarin.Droid.Resource.Color.design_default_color_surface;
			global::Xamarin.Forms.Platform.Resource.Color.design_error = global::pjsua2xamarin.Droid.Resource.Color.design_error;
			global::Xamarin.Forms.Platform.Resource.Color.design_fab_shadow_end_color = global::pjsua2xamarin.Droid.Resource.Color.design_fab_shadow_end_color;
			global::Xamarin.Forms.Platform.Resource.Color.design_fab_shadow_mid_color = global::pjsua2xamarin.Droid.Resource.Color.design_fab_shadow_mid_color;
			global::Xamarin.Forms.Platform.Resource.Color.design_fab_shadow_start_color = global::pjsua2xamarin.Droid.Resource.Color.design_fab_shadow_start_color;
			global::Xamarin.Forms.Platform.Resource.Color.design_fab_stroke_end_inner_color = global::pjsua2xamarin.Droid.Resource.Color.design_fab_stroke_end_inner_color;
			global::Xamarin.Forms.Platform.Resource.Color.design_fab_stroke_end_outer_color = global::pjsua2xamarin.Droid.Resource.Color.design_fab_stroke_end_outer_color;
			global::Xamarin.Forms.Platform.Resource.Color.design_fab_stroke_top_inner_color = global::pjsua2xamarin.Droid.Resource.Color.design_fab_stroke_top_inner_color;
			global::Xamarin.Forms.Platform.Resource.Color.design_fab_stroke_top_outer_color = global::pjsua2xamarin.Droid.Resource.Color.design_fab_stroke_top_outer_color;
			global::Xamarin.Forms.Platform.Resource.Color.design_icon_tint = global::pjsua2xamarin.Droid.Resource.Color.design_icon_tint;
			global::Xamarin.Forms.Platform.Resource.Color.design_snackbar_background_color = global::pjsua2xamarin.Droid.Resource.Color.design_snackbar_background_color;
			global::Xamarin.Forms.Platform.Resource.Color.dim_foreground_disabled_material_dark = global::pjsua2xamarin.Droid.Resource.Color.dim_foreground_disabled_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.dim_foreground_disabled_material_light = global::pjsua2xamarin.Droid.Resource.Color.dim_foreground_disabled_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.dim_foreground_material_dark = global::pjsua2xamarin.Droid.Resource.Color.dim_foreground_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.dim_foreground_material_light = global::pjsua2xamarin.Droid.Resource.Color.dim_foreground_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.error_color_material_dark = global::pjsua2xamarin.Droid.Resource.Color.error_color_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.error_color_material_light = global::pjsua2xamarin.Droid.Resource.Color.error_color_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.foreground_material_dark = global::pjsua2xamarin.Droid.Resource.Color.foreground_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.foreground_material_light = global::pjsua2xamarin.Droid.Resource.Color.foreground_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.highlighted_text_material_dark = global::pjsua2xamarin.Droid.Resource.Color.highlighted_text_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.highlighted_text_material_light = global::pjsua2xamarin.Droid.Resource.Color.highlighted_text_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.material_blue_grey_800 = global::pjsua2xamarin.Droid.Resource.Color.material_blue_grey_800;
			global::Xamarin.Forms.Platform.Resource.Color.material_blue_grey_900 = global::pjsua2xamarin.Droid.Resource.Color.material_blue_grey_900;
			global::Xamarin.Forms.Platform.Resource.Color.material_blue_grey_950 = global::pjsua2xamarin.Droid.Resource.Color.material_blue_grey_950;
			global::Xamarin.Forms.Platform.Resource.Color.material_cursor_color = global::pjsua2xamarin.Droid.Resource.Color.material_cursor_color;
			global::Xamarin.Forms.Platform.Resource.Color.material_deep_teal_200 = global::pjsua2xamarin.Droid.Resource.Color.material_deep_teal_200;
			global::Xamarin.Forms.Platform.Resource.Color.material_deep_teal_500 = global::pjsua2xamarin.Droid.Resource.Color.material_deep_teal_500;
			global::Xamarin.Forms.Platform.Resource.Color.material_grey_100 = global::pjsua2xamarin.Droid.Resource.Color.material_grey_100;
			global::Xamarin.Forms.Platform.Resource.Color.material_grey_300 = global::pjsua2xamarin.Droid.Resource.Color.material_grey_300;
			global::Xamarin.Forms.Platform.Resource.Color.material_grey_50 = global::pjsua2xamarin.Droid.Resource.Color.material_grey_50;
			global::Xamarin.Forms.Platform.Resource.Color.material_grey_600 = global::pjsua2xamarin.Droid.Resource.Color.material_grey_600;
			global::Xamarin.Forms.Platform.Resource.Color.material_grey_800 = global::pjsua2xamarin.Droid.Resource.Color.material_grey_800;
			global::Xamarin.Forms.Platform.Resource.Color.material_grey_850 = global::pjsua2xamarin.Droid.Resource.Color.material_grey_850;
			global::Xamarin.Forms.Platform.Resource.Color.material_grey_900 = global::pjsua2xamarin.Droid.Resource.Color.material_grey_900;
			global::Xamarin.Forms.Platform.Resource.Color.material_on_background_disabled = global::pjsua2xamarin.Droid.Resource.Color.material_on_background_disabled;
			global::Xamarin.Forms.Platform.Resource.Color.material_on_background_emphasis_high_type = global::pjsua2xamarin.Droid.Resource.Color.material_on_background_emphasis_high_type;
			global::Xamarin.Forms.Platform.Resource.Color.material_on_background_emphasis_medium = global::pjsua2xamarin.Droid.Resource.Color.material_on_background_emphasis_medium;
			global::Xamarin.Forms.Platform.Resource.Color.material_on_primary_disabled = global::pjsua2xamarin.Droid.Resource.Color.material_on_primary_disabled;
			global::Xamarin.Forms.Platform.Resource.Color.material_on_primary_emphasis_high_type = global::pjsua2xamarin.Droid.Resource.Color.material_on_primary_emphasis_high_type;
			global::Xamarin.Forms.Platform.Resource.Color.material_on_primary_emphasis_medium = global::pjsua2xamarin.Droid.Resource.Color.material_on_primary_emphasis_medium;
			global::Xamarin.Forms.Platform.Resource.Color.material_on_surface_disabled = global::pjsua2xamarin.Droid.Resource.Color.material_on_surface_disabled;
			global::Xamarin.Forms.Platform.Resource.Color.material_on_surface_emphasis_high_type = global::pjsua2xamarin.Droid.Resource.Color.material_on_surface_emphasis_high_type;
			global::Xamarin.Forms.Platform.Resource.Color.material_on_surface_emphasis_medium = global::pjsua2xamarin.Droid.Resource.Color.material_on_surface_emphasis_medium;
			global::Xamarin.Forms.Platform.Resource.Color.material_on_surface_stroke = global::pjsua2xamarin.Droid.Resource.Color.material_on_surface_stroke;
			global::Xamarin.Forms.Platform.Resource.Color.material_slider_active_tick_marks_color = global::pjsua2xamarin.Droid.Resource.Color.material_slider_active_tick_marks_color;
			global::Xamarin.Forms.Platform.Resource.Color.material_slider_active_track_color = global::pjsua2xamarin.Droid.Resource.Color.material_slider_active_track_color;
			global::Xamarin.Forms.Platform.Resource.Color.material_slider_halo_color = global::pjsua2xamarin.Droid.Resource.Color.material_slider_halo_color;
			global::Xamarin.Forms.Platform.Resource.Color.material_slider_inactive_tick_marks_color = global::pjsua2xamarin.Droid.Resource.Color.material_slider_inactive_tick_marks_color;
			global::Xamarin.Forms.Platform.Resource.Color.material_slider_inactive_track_color = global::pjsua2xamarin.Droid.Resource.Color.material_slider_inactive_track_color;
			global::Xamarin.Forms.Platform.Resource.Color.material_slider_thumb_color = global::pjsua2xamarin.Droid.Resource.Color.material_slider_thumb_color;
			global::Xamarin.Forms.Platform.Resource.Color.material_timepicker_button_background = global::pjsua2xamarin.Droid.Resource.Color.material_timepicker_button_background;
			global::Xamarin.Forms.Platform.Resource.Color.material_timepicker_button_stroke = global::pjsua2xamarin.Droid.Resource.Color.material_timepicker_button_stroke;
			global::Xamarin.Forms.Platform.Resource.Color.material_timepicker_clockface = global::pjsua2xamarin.Droid.Resource.Color.material_timepicker_clockface;
			global::Xamarin.Forms.Platform.Resource.Color.material_timepicker_clock_text_color = global::pjsua2xamarin.Droid.Resource.Color.material_timepicker_clock_text_color;
			global::Xamarin.Forms.Platform.Resource.Color.material_timepicker_modebutton_tint = global::pjsua2xamarin.Droid.Resource.Color.material_timepicker_modebutton_tint;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_btn_bg_color_selector = global::pjsua2xamarin.Droid.Resource.Color.mtrl_btn_bg_color_selector;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_btn_ripple_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_btn_ripple_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_btn_stroke_color_selector = global::pjsua2xamarin.Droid.Resource.Color.mtrl_btn_stroke_color_selector;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_btn_text_btn_bg_color_selector = global::pjsua2xamarin.Droid.Resource.Color.mtrl_btn_text_btn_bg_color_selector;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_btn_text_btn_ripple_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_btn_text_btn_ripple_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_btn_text_color_disabled = global::pjsua2xamarin.Droid.Resource.Color.mtrl_btn_text_color_disabled;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_btn_text_color_selector = global::pjsua2xamarin.Droid.Resource.Color.mtrl_btn_text_color_selector;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_btn_transparent_bg_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_btn_transparent_bg_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_calendar_item_stroke_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_calendar_item_stroke_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_calendar_selected_range = global::pjsua2xamarin.Droid.Resource.Color.mtrl_calendar_selected_range;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_card_view_foreground = global::pjsua2xamarin.Droid.Resource.Color.mtrl_card_view_foreground;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_card_view_ripple = global::pjsua2xamarin.Droid.Resource.Color.mtrl_card_view_ripple;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_chip_background_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_chip_background_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_chip_close_icon_tint = global::pjsua2xamarin.Droid.Resource.Color.mtrl_chip_close_icon_tint;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_chip_surface_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_chip_surface_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_chip_text_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_chip_text_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_choice_chip_background_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_choice_chip_background_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_choice_chip_ripple_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_choice_chip_ripple_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_choice_chip_text_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_choice_chip_text_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_error = global::pjsua2xamarin.Droid.Resource.Color.mtrl_error;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_fab_bg_color_selector = global::pjsua2xamarin.Droid.Resource.Color.mtrl_fab_bg_color_selector;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_fab_icon_text_color_selector = global::pjsua2xamarin.Droid.Resource.Color.mtrl_fab_icon_text_color_selector;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_fab_ripple_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_fab_ripple_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_filled_background_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_filled_background_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_filled_icon_tint = global::pjsua2xamarin.Droid.Resource.Color.mtrl_filled_icon_tint;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_filled_stroke_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_filled_stroke_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_indicator_text_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_indicator_text_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_navigation_bar_colored_item_tint = global::pjsua2xamarin.Droid.Resource.Color.mtrl_navigation_bar_colored_item_tint;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_navigation_bar_colored_ripple_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_navigation_bar_colored_ripple_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_navigation_bar_item_tint = global::pjsua2xamarin.Droid.Resource.Color.mtrl_navigation_bar_item_tint;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_navigation_bar_ripple_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_navigation_bar_ripple_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_navigation_item_background_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_navigation_item_background_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_navigation_item_icon_tint = global::pjsua2xamarin.Droid.Resource.Color.mtrl_navigation_item_icon_tint;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_navigation_item_text_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_navigation_item_text_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_on_primary_text_btn_text_color_selector = global::pjsua2xamarin.Droid.Resource.Color.mtrl_on_primary_text_btn_text_color_selector;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_on_surface_ripple_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_on_surface_ripple_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_outlined_icon_tint = global::pjsua2xamarin.Droid.Resource.Color.mtrl_outlined_icon_tint;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_outlined_stroke_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_outlined_stroke_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_popupmenu_overlay_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_popupmenu_overlay_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_scrim_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_scrim_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_tabs_colored_ripple_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_tabs_colored_ripple_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_tabs_icon_color_selector = global::pjsua2xamarin.Droid.Resource.Color.mtrl_tabs_icon_color_selector;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_tabs_icon_color_selector_colored = global::pjsua2xamarin.Droid.Resource.Color.mtrl_tabs_icon_color_selector_colored;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_tabs_legacy_text_color_selector = global::pjsua2xamarin.Droid.Resource.Color.mtrl_tabs_legacy_text_color_selector;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_tabs_ripple_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_tabs_ripple_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_textinput_default_box_stroke_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_textinput_default_box_stroke_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_textinput_disabled_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_textinput_disabled_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_textinput_filled_box_default_background_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_textinput_filled_box_default_background_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_textinput_focused_box_stroke_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_textinput_focused_box_stroke_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_textinput_hovered_box_stroke_color = global::pjsua2xamarin.Droid.Resource.Color.mtrl_textinput_hovered_box_stroke_color;
			global::Xamarin.Forms.Platform.Resource.Color.mtrl_text_btn_text_color_selector = global::pjsua2xamarin.Droid.Resource.Color.mtrl_text_btn_text_color_selector;
			global::Xamarin.Forms.Platform.Resource.Color.notification_action_color_filter = global::pjsua2xamarin.Droid.Resource.Color.notification_action_color_filter;
			global::Xamarin.Forms.Platform.Resource.Color.notification_icon_bg_color = global::pjsua2xamarin.Droid.Resource.Color.notification_icon_bg_color;
			global::Xamarin.Forms.Platform.Resource.Color.notification_material_background_media_default_color = global::pjsua2xamarin.Droid.Resource.Color.notification_material_background_media_default_color;
			global::Xamarin.Forms.Platform.Resource.Color.primary_dark_material_dark = global::pjsua2xamarin.Droid.Resource.Color.primary_dark_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.primary_dark_material_light = global::pjsua2xamarin.Droid.Resource.Color.primary_dark_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.primary_material_dark = global::pjsua2xamarin.Droid.Resource.Color.primary_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.primary_material_light = global::pjsua2xamarin.Droid.Resource.Color.primary_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.primary_text_default_material_dark = global::pjsua2xamarin.Droid.Resource.Color.primary_text_default_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.primary_text_default_material_light = global::pjsua2xamarin.Droid.Resource.Color.primary_text_default_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.primary_text_disabled_material_dark = global::pjsua2xamarin.Droid.Resource.Color.primary_text_disabled_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.primary_text_disabled_material_light = global::pjsua2xamarin.Droid.Resource.Color.primary_text_disabled_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.radiobutton_themeable_attribute_color = global::pjsua2xamarin.Droid.Resource.Color.radiobutton_themeable_attribute_color;
			global::Xamarin.Forms.Platform.Resource.Color.ripple_material_dark = global::pjsua2xamarin.Droid.Resource.Color.ripple_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.ripple_material_light = global::pjsua2xamarin.Droid.Resource.Color.ripple_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.secondary_text_default_material_dark = global::pjsua2xamarin.Droid.Resource.Color.secondary_text_default_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.secondary_text_default_material_light = global::pjsua2xamarin.Droid.Resource.Color.secondary_text_default_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.secondary_text_disabled_material_dark = global::pjsua2xamarin.Droid.Resource.Color.secondary_text_disabled_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.secondary_text_disabled_material_light = global::pjsua2xamarin.Droid.Resource.Color.secondary_text_disabled_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.switch_thumb_disabled_material_dark = global::pjsua2xamarin.Droid.Resource.Color.switch_thumb_disabled_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.switch_thumb_disabled_material_light = global::pjsua2xamarin.Droid.Resource.Color.switch_thumb_disabled_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.switch_thumb_material_dark = global::pjsua2xamarin.Droid.Resource.Color.switch_thumb_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.switch_thumb_material_light = global::pjsua2xamarin.Droid.Resource.Color.switch_thumb_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.switch_thumb_normal_material_dark = global::pjsua2xamarin.Droid.Resource.Color.switch_thumb_normal_material_dark;
			global::Xamarin.Forms.Platform.Resource.Color.switch_thumb_normal_material_light = global::pjsua2xamarin.Droid.Resource.Color.switch_thumb_normal_material_light;
			global::Xamarin.Forms.Platform.Resource.Color.test_mtrl_calendar_day = global::pjsua2xamarin.Droid.Resource.Color.test_mtrl_calendar_day;
			global::Xamarin.Forms.Platform.Resource.Color.test_mtrl_calendar_day_selected = global::pjsua2xamarin.Droid.Resource.Color.test_mtrl_calendar_day_selected;
			global::Xamarin.Forms.Platform.Resource.Color.tooltip_background_dark = global::pjsua2xamarin.Droid.Resource.Color.tooltip_background_dark;
			global::Xamarin.Forms.Platform.Resource.Color.tooltip_background_light = global::pjsua2xamarin.Droid.Resource.Color.tooltip_background_light;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_action_bar_content_inset_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_content_inset_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_action_bar_content_inset_with_nav = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_content_inset_with_nav;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_action_bar_default_height_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_default_height_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_action_bar_default_padding_end_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_default_padding_end_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_action_bar_default_padding_start_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_default_padding_start_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_action_bar_elevation_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_elevation_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_action_bar_icon_vertical_padding_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_icon_vertical_padding_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_action_bar_overflow_padding_end_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_overflow_padding_end_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_action_bar_overflow_padding_start_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_overflow_padding_start_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_action_bar_stacked_max_height = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_stacked_max_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_action_bar_stacked_tab_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_stacked_tab_max_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_action_bar_subtitle_bottom_margin_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_subtitle_bottom_margin_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_action_bar_subtitle_top_margin_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_bar_subtitle_top_margin_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_action_button_min_height_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_button_min_height_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_action_button_min_width_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_button_min_width_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_action_button_min_width_overflow_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_action_button_min_width_overflow_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_alert_dialog_button_bar_height = global::pjsua2xamarin.Droid.Resource.Dimension.abc_alert_dialog_button_bar_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_alert_dialog_button_dimen = global::pjsua2xamarin.Droid.Resource.Dimension.abc_alert_dialog_button_dimen;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_button_inset_horizontal_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_button_inset_horizontal_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_button_inset_vertical_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_button_inset_vertical_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_button_padding_horizontal_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_button_padding_horizontal_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_button_padding_vertical_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_button_padding_vertical_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_cascading_menus_min_smallest_width = global::pjsua2xamarin.Droid.Resource.Dimension.abc_cascading_menus_min_smallest_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_config_prefDialogWidth = global::pjsua2xamarin.Droid.Resource.Dimension.abc_config_prefDialogWidth;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_control_corner_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_control_corner_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_control_inset_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_control_inset_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_control_padding_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_control_padding_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_dialog_corner_radius_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_corner_radius_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_dialog_fixed_height_major = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_fixed_height_major;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_dialog_fixed_height_minor = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_fixed_height_minor;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_dialog_fixed_width_major = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_fixed_width_major;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_dialog_fixed_width_minor = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_fixed_width_minor;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_dialog_list_padding_bottom_no_buttons = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_list_padding_bottom_no_buttons;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_dialog_list_padding_top_no_title = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_list_padding_top_no_title;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_dialog_min_width_major = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_min_width_major;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_dialog_min_width_minor = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_min_width_minor;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_dialog_padding_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_padding_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_dialog_padding_top_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_padding_top_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_dialog_title_divider_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dialog_title_divider_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_disabled_alpha_material_dark = global::pjsua2xamarin.Droid.Resource.Dimension.abc_disabled_alpha_material_dark;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_disabled_alpha_material_light = global::pjsua2xamarin.Droid.Resource.Dimension.abc_disabled_alpha_material_light;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_dropdownitem_icon_width = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dropdownitem_icon_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_dropdownitem_text_padding_left = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dropdownitem_text_padding_left;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_dropdownitem_text_padding_right = global::pjsua2xamarin.Droid.Resource.Dimension.abc_dropdownitem_text_padding_right;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_edit_text_inset_bottom_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_edit_text_inset_bottom_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_edit_text_inset_horizontal_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_edit_text_inset_horizontal_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_edit_text_inset_top_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_edit_text_inset_top_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_floating_window_z = global::pjsua2xamarin.Droid.Resource.Dimension.abc_floating_window_z;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_list_item_height_large_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_list_item_height_large_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_list_item_height_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_list_item_height_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_list_item_height_small_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_list_item_height_small_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_list_item_padding_horizontal_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_list_item_padding_horizontal_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_panel_menu_list_width = global::pjsua2xamarin.Droid.Resource.Dimension.abc_panel_menu_list_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_progress_bar_height_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_progress_bar_height_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_search_view_preferred_height = global::pjsua2xamarin.Droid.Resource.Dimension.abc_search_view_preferred_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_search_view_preferred_width = global::pjsua2xamarin.Droid.Resource.Dimension.abc_search_view_preferred_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_seekbar_track_background_height_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_seekbar_track_background_height_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_seekbar_track_progress_height_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_seekbar_track_progress_height_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_select_dialog_padding_start_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_select_dialog_padding_start_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_star_big = global::pjsua2xamarin.Droid.Resource.Dimension.abc_star_big;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_star_medium = global::pjsua2xamarin.Droid.Resource.Dimension.abc_star_medium;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_star_small = global::pjsua2xamarin.Droid.Resource.Dimension.abc_star_small;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_switch_padding = global::pjsua2xamarin.Droid.Resource.Dimension.abc_switch_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_text_size_body_1_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_body_1_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_text_size_body_2_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_body_2_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_text_size_button_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_button_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_text_size_caption_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_caption_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_text_size_display_1_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_display_1_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_text_size_display_2_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_display_2_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_text_size_display_3_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_display_3_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_text_size_display_4_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_display_4_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_text_size_headline_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_headline_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_text_size_large_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_large_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_text_size_medium_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_medium_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_text_size_menu_header_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_menu_header_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_text_size_menu_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_menu_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_text_size_small_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_small_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_text_size_subhead_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_subhead_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_text_size_subtitle_material_toolbar = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_subtitle_material_toolbar;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_text_size_title_material = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_title_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.abc_text_size_title_material_toolbar = global::pjsua2xamarin.Droid.Resource.Dimension.abc_text_size_title_material_toolbar;
			global::Xamarin.Forms.Platform.Resource.Dimension.action_bar_size = global::pjsua2xamarin.Droid.Resource.Dimension.action_bar_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.appcompat_dialog_background_inset = global::pjsua2xamarin.Droid.Resource.Dimension.appcompat_dialog_background_inset;
			global::Xamarin.Forms.Platform.Resource.Dimension.cardview_compat_inset_shadow = global::pjsua2xamarin.Droid.Resource.Dimension.cardview_compat_inset_shadow;
			global::Xamarin.Forms.Platform.Resource.Dimension.cardview_default_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.cardview_default_elevation;
			global::Xamarin.Forms.Platform.Resource.Dimension.cardview_default_radius = global::pjsua2xamarin.Droid.Resource.Dimension.cardview_default_radius;
			global::Xamarin.Forms.Platform.Resource.Dimension.clock_face_margin_start = global::pjsua2xamarin.Droid.Resource.Dimension.clock_face_margin_start;
			global::Xamarin.Forms.Platform.Resource.Dimension.compat_button_inset_horizontal_material = global::pjsua2xamarin.Droid.Resource.Dimension.compat_button_inset_horizontal_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.compat_button_inset_vertical_material = global::pjsua2xamarin.Droid.Resource.Dimension.compat_button_inset_vertical_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.compat_button_padding_horizontal_material = global::pjsua2xamarin.Droid.Resource.Dimension.compat_button_padding_horizontal_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.compat_button_padding_vertical_material = global::pjsua2xamarin.Droid.Resource.Dimension.compat_button_padding_vertical_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.compat_control_corner_material = global::pjsua2xamarin.Droid.Resource.Dimension.compat_control_corner_material;
			global::Xamarin.Forms.Platform.Resource.Dimension.compat_notification_large_icon_max_height = global::pjsua2xamarin.Droid.Resource.Dimension.compat_notification_large_icon_max_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.compat_notification_large_icon_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.compat_notification_large_icon_max_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.default_dimension = global::pjsua2xamarin.Droid.Resource.Dimension.default_dimension;
			global::Xamarin.Forms.Platform.Resource.Dimension.def_drawer_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.def_drawer_elevation;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_appbar_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.design_appbar_elevation;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_bottom_navigation_active_item_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_active_item_max_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_bottom_navigation_active_item_min_width = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_active_item_min_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_bottom_navigation_active_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_active_text_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_bottom_navigation_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_elevation;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_bottom_navigation_height = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_bottom_navigation_icon_size = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_icon_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_bottom_navigation_item_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_item_max_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_bottom_navigation_item_min_width = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_item_min_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_bottom_navigation_label_padding = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_label_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_bottom_navigation_margin = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_margin;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_bottom_navigation_shadow_height = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_shadow_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_bottom_navigation_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_navigation_text_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_bottom_sheet_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_sheet_elevation;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_bottom_sheet_modal_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_sheet_modal_elevation;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_bottom_sheet_peek_height_min = global::pjsua2xamarin.Droid.Resource.Dimension.design_bottom_sheet_peek_height_min;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_fab_border_width = global::pjsua2xamarin.Droid.Resource.Dimension.design_fab_border_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_fab_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.design_fab_elevation;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_fab_image_size = global::pjsua2xamarin.Droid.Resource.Dimension.design_fab_image_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_fab_size_mini = global::pjsua2xamarin.Droid.Resource.Dimension.design_fab_size_mini;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_fab_size_normal = global::pjsua2xamarin.Droid.Resource.Dimension.design_fab_size_normal;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_fab_translation_z_hovered_focused = global::pjsua2xamarin.Droid.Resource.Dimension.design_fab_translation_z_hovered_focused;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_fab_translation_z_pressed = global::pjsua2xamarin.Droid.Resource.Dimension.design_fab_translation_z_pressed;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_navigation_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.design_navigation_elevation;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_navigation_icon_padding = global::pjsua2xamarin.Droid.Resource.Dimension.design_navigation_icon_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_navigation_icon_size = global::pjsua2xamarin.Droid.Resource.Dimension.design_navigation_icon_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_navigation_item_horizontal_padding = global::pjsua2xamarin.Droid.Resource.Dimension.design_navigation_item_horizontal_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_navigation_item_icon_padding = global::pjsua2xamarin.Droid.Resource.Dimension.design_navigation_item_icon_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_navigation_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.design_navigation_max_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_navigation_padding_bottom = global::pjsua2xamarin.Droid.Resource.Dimension.design_navigation_padding_bottom;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_navigation_separator_vertical_padding = global::pjsua2xamarin.Droid.Resource.Dimension.design_navigation_separator_vertical_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_snackbar_action_inline_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.design_snackbar_action_inline_max_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_snackbar_action_text_color_alpha = global::pjsua2xamarin.Droid.Resource.Dimension.design_snackbar_action_text_color_alpha;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_snackbar_background_corner_radius = global::pjsua2xamarin.Droid.Resource.Dimension.design_snackbar_background_corner_radius;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_snackbar_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.design_snackbar_elevation;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_snackbar_extra_spacing_horizontal = global::pjsua2xamarin.Droid.Resource.Dimension.design_snackbar_extra_spacing_horizontal;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_snackbar_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.design_snackbar_max_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_snackbar_min_width = global::pjsua2xamarin.Droid.Resource.Dimension.design_snackbar_min_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_snackbar_padding_horizontal = global::pjsua2xamarin.Droid.Resource.Dimension.design_snackbar_padding_horizontal;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_snackbar_padding_vertical = global::pjsua2xamarin.Droid.Resource.Dimension.design_snackbar_padding_vertical;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_snackbar_padding_vertical_2lines = global::pjsua2xamarin.Droid.Resource.Dimension.design_snackbar_padding_vertical_2lines;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_snackbar_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.design_snackbar_text_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_tab_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.design_tab_max_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_tab_scrollable_min_width = global::pjsua2xamarin.Droid.Resource.Dimension.design_tab_scrollable_min_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_tab_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.design_tab_text_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_tab_text_size_2line = global::pjsua2xamarin.Droid.Resource.Dimension.design_tab_text_size_2line;
			global::Xamarin.Forms.Platform.Resource.Dimension.design_textinput_caption_translate_y = global::pjsua2xamarin.Droid.Resource.Dimension.design_textinput_caption_translate_y;
			global::Xamarin.Forms.Platform.Resource.Dimension.disabled_alpha_material_dark = global::pjsua2xamarin.Droid.Resource.Dimension.disabled_alpha_material_dark;
			global::Xamarin.Forms.Platform.Resource.Dimension.disabled_alpha_material_light = global::pjsua2xamarin.Droid.Resource.Dimension.disabled_alpha_material_light;
			global::Xamarin.Forms.Platform.Resource.Dimension.fastscroll_default_thickness = global::pjsua2xamarin.Droid.Resource.Dimension.fastscroll_default_thickness;
			global::Xamarin.Forms.Platform.Resource.Dimension.fastscroll_margin = global::pjsua2xamarin.Droid.Resource.Dimension.fastscroll_margin;
			global::Xamarin.Forms.Platform.Resource.Dimension.fastscroll_minimum_range = global::pjsua2xamarin.Droid.Resource.Dimension.fastscroll_minimum_range;
			global::Xamarin.Forms.Platform.Resource.Dimension.highlight_alpha_material_colored = global::pjsua2xamarin.Droid.Resource.Dimension.highlight_alpha_material_colored;
			global::Xamarin.Forms.Platform.Resource.Dimension.highlight_alpha_material_dark = global::pjsua2xamarin.Droid.Resource.Dimension.highlight_alpha_material_dark;
			global::Xamarin.Forms.Platform.Resource.Dimension.highlight_alpha_material_light = global::pjsua2xamarin.Droid.Resource.Dimension.highlight_alpha_material_light;
			global::Xamarin.Forms.Platform.Resource.Dimension.hint_alpha_material_dark = global::pjsua2xamarin.Droid.Resource.Dimension.hint_alpha_material_dark;
			global::Xamarin.Forms.Platform.Resource.Dimension.hint_alpha_material_light = global::pjsua2xamarin.Droid.Resource.Dimension.hint_alpha_material_light;
			global::Xamarin.Forms.Platform.Resource.Dimension.hint_pressed_alpha_material_dark = global::pjsua2xamarin.Droid.Resource.Dimension.hint_pressed_alpha_material_dark;
			global::Xamarin.Forms.Platform.Resource.Dimension.hint_pressed_alpha_material_light = global::pjsua2xamarin.Droid.Resource.Dimension.hint_pressed_alpha_material_light;
			global::Xamarin.Forms.Platform.Resource.Dimension.item_touch_helper_max_drag_scroll_per_frame = global::pjsua2xamarin.Droid.Resource.Dimension.item_touch_helper_max_drag_scroll_per_frame;
			global::Xamarin.Forms.Platform.Resource.Dimension.item_touch_helper_swipe_escape_max_velocity = global::pjsua2xamarin.Droid.Resource.Dimension.item_touch_helper_swipe_escape_max_velocity;
			global::Xamarin.Forms.Platform.Resource.Dimension.item_touch_helper_swipe_escape_velocity = global::pjsua2xamarin.Droid.Resource.Dimension.item_touch_helper_swipe_escape_velocity;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_bottom_sheet_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.material_bottom_sheet_max_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_clock_display_padding = global::pjsua2xamarin.Droid.Resource.Dimension.material_clock_display_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_clock_face_margin_top = global::pjsua2xamarin.Droid.Resource.Dimension.material_clock_face_margin_top;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_clock_hand_center_dot_radius = global::pjsua2xamarin.Droid.Resource.Dimension.material_clock_hand_center_dot_radius;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_clock_hand_padding = global::pjsua2xamarin.Droid.Resource.Dimension.material_clock_hand_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_clock_hand_stroke_width = global::pjsua2xamarin.Droid.Resource.Dimension.material_clock_hand_stroke_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_clock_number_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.material_clock_number_text_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_clock_period_toggle_height = global::pjsua2xamarin.Droid.Resource.Dimension.material_clock_period_toggle_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_clock_period_toggle_margin_left = global::pjsua2xamarin.Droid.Resource.Dimension.material_clock_period_toggle_margin_left;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_clock_period_toggle_width = global::pjsua2xamarin.Droid.Resource.Dimension.material_clock_period_toggle_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_clock_size = global::pjsua2xamarin.Droid.Resource.Dimension.material_clock_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_cursor_inset_bottom = global::pjsua2xamarin.Droid.Resource.Dimension.material_cursor_inset_bottom;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_cursor_inset_top = global::pjsua2xamarin.Droid.Resource.Dimension.material_cursor_inset_top;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_cursor_width = global::pjsua2xamarin.Droid.Resource.Dimension.material_cursor_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_emphasis_disabled = global::pjsua2xamarin.Droid.Resource.Dimension.material_emphasis_disabled;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_emphasis_high_type = global::pjsua2xamarin.Droid.Resource.Dimension.material_emphasis_high_type;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_emphasis_medium = global::pjsua2xamarin.Droid.Resource.Dimension.material_emphasis_medium;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_filled_edittext_font_1_3_padding_bottom = global::pjsua2xamarin.Droid.Resource.Dimension.material_filled_edittext_font_1_3_padding_bottom;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_filled_edittext_font_1_3_padding_top = global::pjsua2xamarin.Droid.Resource.Dimension.material_filled_edittext_font_1_3_padding_top;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_filled_edittext_font_2_0_padding_bottom = global::pjsua2xamarin.Droid.Resource.Dimension.material_filled_edittext_font_2_0_padding_bottom;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_filled_edittext_font_2_0_padding_top = global::pjsua2xamarin.Droid.Resource.Dimension.material_filled_edittext_font_2_0_padding_top;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_font_1_3_box_collapsed_padding_top = global::pjsua2xamarin.Droid.Resource.Dimension.material_font_1_3_box_collapsed_padding_top;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_font_2_0_box_collapsed_padding_top = global::pjsua2xamarin.Droid.Resource.Dimension.material_font_2_0_box_collapsed_padding_top;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_helper_text_default_padding_top = global::pjsua2xamarin.Droid.Resource.Dimension.material_helper_text_default_padding_top;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_helper_text_font_1_3_padding_horizontal = global::pjsua2xamarin.Droid.Resource.Dimension.material_helper_text_font_1_3_padding_horizontal;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_helper_text_font_1_3_padding_top = global::pjsua2xamarin.Droid.Resource.Dimension.material_helper_text_font_1_3_padding_top;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_input_text_to_prefix_suffix_padding = global::pjsua2xamarin.Droid.Resource.Dimension.material_input_text_to_prefix_suffix_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_textinput_default_width = global::pjsua2xamarin.Droid.Resource.Dimension.material_textinput_default_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_textinput_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.material_textinput_max_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_textinput_min_width = global::pjsua2xamarin.Droid.Resource.Dimension.material_textinput_min_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_text_view_test_line_height = global::pjsua2xamarin.Droid.Resource.Dimension.material_text_view_test_line_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_text_view_test_line_height_override = global::pjsua2xamarin.Droid.Resource.Dimension.material_text_view_test_line_height_override;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_timepicker_dialog_buttons_margin_top = global::pjsua2xamarin.Droid.Resource.Dimension.material_timepicker_dialog_buttons_margin_top;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_time_picker_minimum_screen_height = global::pjsua2xamarin.Droid.Resource.Dimension.material_time_picker_minimum_screen_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.material_time_picker_minimum_screen_width = global::pjsua2xamarin.Droid.Resource.Dimension.material_time_picker_minimum_screen_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_alert_dialog_background_inset_bottom = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_alert_dialog_background_inset_bottom;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_alert_dialog_background_inset_end = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_alert_dialog_background_inset_end;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_alert_dialog_background_inset_start = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_alert_dialog_background_inset_start;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_alert_dialog_background_inset_top = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_alert_dialog_background_inset_top;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_alert_dialog_picker_background_inset = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_alert_dialog_picker_background_inset;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_badge_horizontal_edge_offset = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_badge_horizontal_edge_offset;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_badge_long_text_horizontal_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_badge_long_text_horizontal_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_badge_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_badge_radius;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_badge_text_horizontal_edge_offset = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_badge_text_horizontal_edge_offset;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_badge_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_badge_text_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_badge_toolbar_action_menu_item_horizontal_offset = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_badge_toolbar_action_menu_item_horizontal_offset;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_badge_toolbar_action_menu_item_vertical_offset = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_badge_toolbar_action_menu_item_vertical_offset;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_badge_with_text_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_badge_with_text_radius;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_bottomappbar_fabOffsetEndMode = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_bottomappbar_fabOffsetEndMode;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_bottomappbar_fab_bottom_margin = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_bottomappbar_fab_bottom_margin;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_bottomappbar_fab_cradle_margin = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_bottomappbar_fab_cradle_margin;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_bottomappbar_fab_cradle_rounded_corner_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_bottomappbar_fab_cradle_rounded_corner_radius;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_bottomappbar_fab_cradle_vertical_offset = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_bottomappbar_fab_cradle_vertical_offset;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_bottomappbar_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_bottomappbar_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_corner_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_corner_radius;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_dialog_btn_min_width = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_dialog_btn_min_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_disabled_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_disabled_elevation;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_disabled_z = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_disabled_z;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_elevation;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_focused_z = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_focused_z;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_hovered_z = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_hovered_z;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_icon_btn_padding_left = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_icon_btn_padding_left;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_icon_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_icon_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_inset = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_inset;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_letter_spacing = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_letter_spacing;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_max_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_padding_bottom = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_padding_bottom;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_padding_left = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_padding_left;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_padding_right = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_padding_right;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_padding_top = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_padding_top;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_pressed_z = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_pressed_z;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_snackbar_margin_horizontal = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_snackbar_margin_horizontal;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_stroke_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_stroke_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_text_btn_icon_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_text_btn_icon_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_text_btn_padding_left = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_text_btn_padding_left;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_text_btn_padding_right = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_text_btn_padding_right;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_text_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_btn_z = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_btn_z;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_action_confirm_button_min_width = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_action_confirm_button_min_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_action_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_action_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_action_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_action_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_bottom_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_bottom_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_content_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_content_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_days_of_week_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_days_of_week_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_day_corner = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_day_corner;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_day_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_day_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_day_horizontal_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_day_horizontal_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_day_today_stroke = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_day_today_stroke;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_day_vertical_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_day_vertical_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_day_width = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_day_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_dialog_background_inset = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_dialog_background_inset;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_header_content_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_header_content_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_header_content_padding_fullscreen = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_header_content_padding_fullscreen;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_header_divider_thickness = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_header_divider_thickness;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_header_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_header_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_header_height_fullscreen = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_header_height_fullscreen;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_header_selection_line_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_header_selection_line_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_header_text_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_header_text_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_header_toggle_margin_bottom = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_header_toggle_margin_bottom;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_header_toggle_margin_top = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_header_toggle_margin_top;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_landscape_header_width = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_landscape_header_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_maximum_default_fullscreen_minor_axis = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_maximum_default_fullscreen_minor_axis;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_month_horizontal_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_month_horizontal_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_month_vertical_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_month_vertical_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_navigation_bottom_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_navigation_bottom_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_navigation_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_navigation_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_navigation_top_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_navigation_top_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_pre_l_text_clip_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_pre_l_text_clip_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_selection_baseline_to_top_fullscreen = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_selection_baseline_to_top_fullscreen;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_selection_text_baseline_to_bottom = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_selection_text_baseline_to_bottom;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_selection_text_baseline_to_bottom_fullscreen = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_selection_text_baseline_to_bottom_fullscreen;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_selection_text_baseline_to_top = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_selection_text_baseline_to_top;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_text_input_padding_top = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_text_input_padding_top;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_title_baseline_to_top = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_title_baseline_to_top;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_title_baseline_to_top_fullscreen = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_title_baseline_to_top_fullscreen;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_year_corner = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_year_corner;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_year_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_year_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_year_horizontal_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_year_horizontal_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_year_vertical_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_year_vertical_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_calendar_year_width = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_calendar_year_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_card_checked_icon_margin = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_card_checked_icon_margin;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_card_checked_icon_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_card_checked_icon_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_card_corner_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_card_corner_radius;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_card_dragged_z = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_card_dragged_z;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_card_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_card_elevation;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_card_spacing = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_card_spacing;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_chip_pressed_translation_z = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_chip_pressed_translation_z;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_chip_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_chip_text_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_edittext_rectangle_top_offset = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_edittext_rectangle_top_offset;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_exposed_dropdown_menu_popup_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_exposed_dropdown_menu_popup_elevation;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_exposed_dropdown_menu_popup_vertical_offset = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_exposed_dropdown_menu_popup_vertical_offset;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_exposed_dropdown_menu_popup_vertical_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_exposed_dropdown_menu_popup_vertical_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_extended_fab_bottom_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_bottom_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_extended_fab_corner_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_corner_radius;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_extended_fab_disabled_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_disabled_elevation;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_extended_fab_disabled_translation_z = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_disabled_translation_z;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_extended_fab_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_elevation;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_extended_fab_end_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_end_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_extended_fab_end_padding_icon = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_end_padding_icon;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_extended_fab_icon_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_icon_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_extended_fab_icon_text_spacing = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_icon_text_spacing;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_extended_fab_min_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_min_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_extended_fab_min_width = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_min_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_extended_fab_start_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_start_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_extended_fab_start_padding_icon = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_start_padding_icon;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_extended_fab_top_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_top_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_extended_fab_translation_z_base = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_translation_z_base;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_extended_fab_translation_z_hovered_focused = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_translation_z_hovered_focused;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_extended_fab_translation_z_pressed = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_extended_fab_translation_z_pressed;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_fab_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_fab_elevation;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_fab_min_touch_target = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_fab_min_touch_target;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_fab_translation_z_hovered_focused = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_fab_translation_z_hovered_focused;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_fab_translation_z_pressed = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_fab_translation_z_pressed;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_high_ripple_default_alpha = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_high_ripple_default_alpha;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_high_ripple_focused_alpha = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_high_ripple_focused_alpha;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_high_ripple_hovered_alpha = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_high_ripple_hovered_alpha;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_high_ripple_pressed_alpha = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_high_ripple_pressed_alpha;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_large_touch_target = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_large_touch_target;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_low_ripple_default_alpha = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_low_ripple_default_alpha;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_low_ripple_focused_alpha = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_low_ripple_focused_alpha;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_low_ripple_hovered_alpha = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_low_ripple_hovered_alpha;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_low_ripple_pressed_alpha = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_low_ripple_pressed_alpha;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_min_touch_target_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_min_touch_target_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_navigation_bar_item_default_icon_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_bar_item_default_icon_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_navigation_bar_item_default_margin = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_bar_item_default_margin;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_navigation_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_elevation;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_navigation_item_horizontal_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_item_horizontal_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_navigation_item_icon_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_item_icon_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_navigation_item_icon_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_item_icon_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_navigation_item_shape_horizontal_margin = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_item_shape_horizontal_margin;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_navigation_item_shape_vertical_margin = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_item_shape_vertical_margin;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_navigation_rail_active_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_rail_active_text_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_navigation_rail_compact_width = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_rail_compact_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_navigation_rail_default_width = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_rail_default_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_navigation_rail_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_rail_elevation;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_navigation_rail_icon_margin = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_rail_icon_margin;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_navigation_rail_icon_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_rail_icon_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_navigation_rail_margin = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_rail_margin;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_navigation_rail_text_bottom_margin = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_rail_text_bottom_margin;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_navigation_rail_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_navigation_rail_text_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_progress_circular_inset = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_inset;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_progress_circular_inset_extra_small = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_inset_extra_small;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_progress_circular_inset_medium = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_inset_medium;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_progress_circular_inset_small = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_inset_small;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_progress_circular_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_radius;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_progress_circular_size = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_progress_circular_size_extra_small = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_size_extra_small;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_progress_circular_size_medium = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_size_medium;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_progress_circular_size_small = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_size_small;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_progress_circular_track_thickness_extra_small = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_track_thickness_extra_small;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_progress_circular_track_thickness_medium = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_track_thickness_medium;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_progress_circular_track_thickness_small = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_circular_track_thickness_small;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_progress_indicator_full_rounded_corner_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_indicator_full_rounded_corner_radius;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_progress_track_thickness = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_progress_track_thickness;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_shape_corner_size_large_component = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_shape_corner_size_large_component;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_shape_corner_size_medium_component = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_shape_corner_size_medium_component;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_shape_corner_size_small_component = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_shape_corner_size_small_component;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_slider_halo_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_slider_halo_radius;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_slider_label_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_slider_label_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_slider_label_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_slider_label_radius;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_slider_label_square_side = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_slider_label_square_side;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_slider_thumb_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_slider_thumb_elevation;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_slider_thumb_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_slider_thumb_radius;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_slider_track_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_slider_track_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_slider_track_side_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_slider_track_side_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_slider_track_top = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_slider_track_top;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_slider_widget_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_slider_widget_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_snackbar_action_text_color_alpha = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_snackbar_action_text_color_alpha;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_snackbar_background_corner_radius = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_snackbar_background_corner_radius;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_snackbar_background_overlay_color_alpha = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_snackbar_background_overlay_color_alpha;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_snackbar_margin = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_snackbar_margin;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_snackbar_message_margin_horizontal = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_snackbar_message_margin_horizontal;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_snackbar_padding_horizontal = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_snackbar_padding_horizontal;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_switch_thumb_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_switch_thumb_elevation;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_textinput_box_corner_radius_medium = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_textinput_box_corner_radius_medium;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_textinput_box_corner_radius_small = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_textinput_box_corner_radius_small;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_textinput_box_label_cutout_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_textinput_box_label_cutout_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_textinput_box_stroke_width_default = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_textinput_box_stroke_width_default;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_textinput_box_stroke_width_focused = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_textinput_box_stroke_width_focused;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_textinput_counter_margin_start = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_textinput_counter_margin_start;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_textinput_end_icon_margin_start = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_textinput_end_icon_margin_start;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_textinput_outline_box_expanded_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_textinput_outline_box_expanded_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_textinput_start_icon_margin_end = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_textinput_start_icon_margin_end;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_toolbar_default_height = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_toolbar_default_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_tooltip_arrowSize = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_tooltip_arrowSize;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_tooltip_cornerSize = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_tooltip_cornerSize;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_tooltip_minHeight = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_tooltip_minHeight;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_tooltip_minWidth = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_tooltip_minWidth;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_tooltip_padding = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_tooltip_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.mtrl_transition_shared_axis_slide_distance = global::pjsua2xamarin.Droid.Resource.Dimension.mtrl_transition_shared_axis_slide_distance;
			global::Xamarin.Forms.Platform.Resource.Dimension.notification_action_icon_size = global::pjsua2xamarin.Droid.Resource.Dimension.notification_action_icon_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.notification_action_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.notification_action_text_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.notification_big_circle_margin = global::pjsua2xamarin.Droid.Resource.Dimension.notification_big_circle_margin;
			global::Xamarin.Forms.Platform.Resource.Dimension.notification_content_margin_start = global::pjsua2xamarin.Droid.Resource.Dimension.notification_content_margin_start;
			global::Xamarin.Forms.Platform.Resource.Dimension.notification_large_icon_height = global::pjsua2xamarin.Droid.Resource.Dimension.notification_large_icon_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.notification_large_icon_width = global::pjsua2xamarin.Droid.Resource.Dimension.notification_large_icon_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.notification_main_column_padding_top = global::pjsua2xamarin.Droid.Resource.Dimension.notification_main_column_padding_top;
			global::Xamarin.Forms.Platform.Resource.Dimension.notification_media_narrow_margin = global::pjsua2xamarin.Droid.Resource.Dimension.notification_media_narrow_margin;
			global::Xamarin.Forms.Platform.Resource.Dimension.notification_right_icon_size = global::pjsua2xamarin.Droid.Resource.Dimension.notification_right_icon_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.notification_right_side_padding_top = global::pjsua2xamarin.Droid.Resource.Dimension.notification_right_side_padding_top;
			global::Xamarin.Forms.Platform.Resource.Dimension.notification_small_icon_background_padding = global::pjsua2xamarin.Droid.Resource.Dimension.notification_small_icon_background_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.notification_small_icon_size_as_large = global::pjsua2xamarin.Droid.Resource.Dimension.notification_small_icon_size_as_large;
			global::Xamarin.Forms.Platform.Resource.Dimension.notification_subtext_size = global::pjsua2xamarin.Droid.Resource.Dimension.notification_subtext_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.notification_top_pad = global::pjsua2xamarin.Droid.Resource.Dimension.notification_top_pad;
			global::Xamarin.Forms.Platform.Resource.Dimension.notification_top_pad_large_text = global::pjsua2xamarin.Droid.Resource.Dimension.notification_top_pad_large_text;
			global::Xamarin.Forms.Platform.Resource.Dimension.test_mtrl_calendar_day_cornerSize = global::pjsua2xamarin.Droid.Resource.Dimension.test_mtrl_calendar_day_cornerSize;
			global::Xamarin.Forms.Platform.Resource.Dimension.test_navigation_bar_active_item_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.test_navigation_bar_active_item_max_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.test_navigation_bar_active_item_min_width = global::pjsua2xamarin.Droid.Resource.Dimension.test_navigation_bar_active_item_min_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.test_navigation_bar_active_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.test_navigation_bar_active_text_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.test_navigation_bar_elevation = global::pjsua2xamarin.Droid.Resource.Dimension.test_navigation_bar_elevation;
			global::Xamarin.Forms.Platform.Resource.Dimension.test_navigation_bar_height = global::pjsua2xamarin.Droid.Resource.Dimension.test_navigation_bar_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.test_navigation_bar_icon_size = global::pjsua2xamarin.Droid.Resource.Dimension.test_navigation_bar_icon_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.test_navigation_bar_item_max_width = global::pjsua2xamarin.Droid.Resource.Dimension.test_navigation_bar_item_max_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.test_navigation_bar_item_min_width = global::pjsua2xamarin.Droid.Resource.Dimension.test_navigation_bar_item_min_width;
			global::Xamarin.Forms.Platform.Resource.Dimension.test_navigation_bar_label_padding = global::pjsua2xamarin.Droid.Resource.Dimension.test_navigation_bar_label_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.test_navigation_bar_shadow_height = global::pjsua2xamarin.Droid.Resource.Dimension.test_navigation_bar_shadow_height;
			global::Xamarin.Forms.Platform.Resource.Dimension.test_navigation_bar_text_size = global::pjsua2xamarin.Droid.Resource.Dimension.test_navigation_bar_text_size;
			global::Xamarin.Forms.Platform.Resource.Dimension.tooltip_corner_radius = global::pjsua2xamarin.Droid.Resource.Dimension.tooltip_corner_radius;
			global::Xamarin.Forms.Platform.Resource.Dimension.tooltip_horizontal_padding = global::pjsua2xamarin.Droid.Resource.Dimension.tooltip_horizontal_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.tooltip_margin = global::pjsua2xamarin.Droid.Resource.Dimension.tooltip_margin;
			global::Xamarin.Forms.Platform.Resource.Dimension.tooltip_precise_anchor_extra_offset = global::pjsua2xamarin.Droid.Resource.Dimension.tooltip_precise_anchor_extra_offset;
			global::Xamarin.Forms.Platform.Resource.Dimension.tooltip_precise_anchor_threshold = global::pjsua2xamarin.Droid.Resource.Dimension.tooltip_precise_anchor_threshold;
			global::Xamarin.Forms.Platform.Resource.Dimension.tooltip_vertical_padding = global::pjsua2xamarin.Droid.Resource.Dimension.tooltip_vertical_padding;
			global::Xamarin.Forms.Platform.Resource.Dimension.tooltip_y_offset_non_touch = global::pjsua2xamarin.Droid.Resource.Dimension.tooltip_y_offset_non_touch;
			global::Xamarin.Forms.Platform.Resource.Dimension.tooltip_y_offset_touch = global::pjsua2xamarin.Droid.Resource.Dimension.tooltip_y_offset_touch;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_ab_share_pack_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ab_share_pack_mtrl_alpha;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_action_bar_item_background_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_action_bar_item_background_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_btn_borderless_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_borderless_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_btn_check_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_check_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_btn_check_material_anim = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_check_material_anim;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_btn_check_to_on_mtrl_000 = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_check_to_on_mtrl_000;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_btn_check_to_on_mtrl_015 = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_check_to_on_mtrl_015;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_btn_colored_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_colored_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_btn_default_mtrl_shape = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_default_mtrl_shape;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_btn_radio_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_radio_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_btn_radio_material_anim = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_radio_material_anim;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_btn_radio_to_on_mtrl_000 = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_radio_to_on_mtrl_000;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_btn_radio_to_on_mtrl_015 = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_radio_to_on_mtrl_015;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_btn_switch_to_on_mtrl_00001 = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_switch_to_on_mtrl_00001;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_btn_switch_to_on_mtrl_00012 = global::pjsua2xamarin.Droid.Resource.Drawable.abc_btn_switch_to_on_mtrl_00012;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_cab_background_internal_bg = global::pjsua2xamarin.Droid.Resource.Drawable.abc_cab_background_internal_bg;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_cab_background_top_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_cab_background_top_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_cab_background_top_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_cab_background_top_mtrl_alpha;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_control_background_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_control_background_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_dialog_material_background = global::pjsua2xamarin.Droid.Resource.Drawable.abc_dialog_material_background;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_edit_text_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_edit_text_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_ic_ab_back_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_ab_back_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_ic_arrow_drop_right_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_arrow_drop_right_black_24dp;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_ic_clear_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_clear_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_ic_commit_search_api_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_commit_search_api_mtrl_alpha;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_ic_go_search_api_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_go_search_api_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_ic_menu_copy_mtrl_am_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_menu_copy_mtrl_am_alpha;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_ic_menu_cut_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_menu_cut_mtrl_alpha;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_ic_menu_overflow_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_menu_overflow_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_ic_menu_paste_mtrl_am_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_menu_paste_mtrl_am_alpha;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_ic_menu_selectall_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_menu_selectall_mtrl_alpha;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_ic_menu_share_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_menu_share_mtrl_alpha;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_ic_search_api_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_search_api_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_ic_voice_search_api_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ic_voice_search_api_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_item_background_holo_dark = global::pjsua2xamarin.Droid.Resource.Drawable.abc_item_background_holo_dark;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_item_background_holo_light = global::pjsua2xamarin.Droid.Resource.Drawable.abc_item_background_holo_light;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_list_divider_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_divider_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_list_divider_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_divider_mtrl_alpha;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_list_focused_holo = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_focused_holo;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_list_longpressed_holo = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_longpressed_holo;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_list_pressed_holo_dark = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_pressed_holo_dark;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_list_pressed_holo_light = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_pressed_holo_light;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_list_selector_background_transition_holo_dark = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_selector_background_transition_holo_dark;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_list_selector_background_transition_holo_light = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_selector_background_transition_holo_light;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_list_selector_disabled_holo_dark = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_selector_disabled_holo_dark;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_list_selector_disabled_holo_light = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_selector_disabled_holo_light;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_list_selector_holo_dark = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_selector_holo_dark;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_list_selector_holo_light = global::pjsua2xamarin.Droid.Resource.Drawable.abc_list_selector_holo_light;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_menu_hardkey_panel_mtrl_mult = global::pjsua2xamarin.Droid.Resource.Drawable.abc_menu_hardkey_panel_mtrl_mult;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_popup_background_mtrl_mult = global::pjsua2xamarin.Droid.Resource.Drawable.abc_popup_background_mtrl_mult;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_ratingbar_indicator_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ratingbar_indicator_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_ratingbar_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ratingbar_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_ratingbar_small_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_ratingbar_small_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_scrubber_control_off_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_scrubber_control_off_mtrl_alpha;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_scrubber_control_to_pressed_mtrl_000 = global::pjsua2xamarin.Droid.Resource.Drawable.abc_scrubber_control_to_pressed_mtrl_000;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_scrubber_control_to_pressed_mtrl_005 = global::pjsua2xamarin.Droid.Resource.Drawable.abc_scrubber_control_to_pressed_mtrl_005;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_scrubber_primary_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_scrubber_primary_mtrl_alpha;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_scrubber_track_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_scrubber_track_mtrl_alpha;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_seekbar_thumb_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_seekbar_thumb_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_seekbar_tick_mark_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_seekbar_tick_mark_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_seekbar_track_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_seekbar_track_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_spinner_mtrl_am_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_spinner_mtrl_am_alpha;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_spinner_textfield_background_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_spinner_textfield_background_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_star_black_48dp = global::pjsua2xamarin.Droid.Resource.Drawable.abc_star_black_48dp;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_star_half_black_48dp = global::pjsua2xamarin.Droid.Resource.Drawable.abc_star_half_black_48dp;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_switch_thumb_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_switch_thumb_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_switch_track_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_switch_track_mtrl_alpha;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_tab_indicator_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_tab_indicator_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_tab_indicator_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_tab_indicator_mtrl_alpha;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_textfield_activated_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_textfield_activated_mtrl_alpha;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_textfield_default_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_textfield_default_mtrl_alpha;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_textfield_search_activated_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_textfield_search_activated_mtrl_alpha;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_textfield_search_default_mtrl_alpha = global::pjsua2xamarin.Droid.Resource.Drawable.abc_textfield_search_default_mtrl_alpha;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_textfield_search_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_textfield_search_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_text_cursor_material = global::pjsua2xamarin.Droid.Resource.Drawable.abc_text_cursor_material;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_text_select_handle_left_mtrl = global::pjsua2xamarin.Droid.Resource.Drawable.abc_text_select_handle_left_mtrl;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_text_select_handle_middle_mtrl = global::pjsua2xamarin.Droid.Resource.Drawable.abc_text_select_handle_middle_mtrl;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_text_select_handle_right_mtrl = global::pjsua2xamarin.Droid.Resource.Drawable.abc_text_select_handle_right_mtrl;
			global::Xamarin.Forms.Platform.Resource.Drawable.abc_vector_test = global::pjsua2xamarin.Droid.Resource.Drawable.abc_vector_test;
			global::Xamarin.Forms.Platform.Resource.Drawable.avd_hide_password = global::pjsua2xamarin.Droid.Resource.Drawable.avd_hide_password;
			global::Xamarin.Forms.Platform.Resource.Drawable.avd_show_password = global::pjsua2xamarin.Droid.Resource.Drawable.avd_show_password;
			global::Xamarin.Forms.Platform.Resource.Drawable.btn_checkbox_checked_mtrl = global::pjsua2xamarin.Droid.Resource.Drawable.btn_checkbox_checked_mtrl;
			global::Xamarin.Forms.Platform.Resource.Drawable.btn_checkbox_checked_to_unchecked_mtrl_animation = global::pjsua2xamarin.Droid.Resource.Drawable.btn_checkbox_checked_to_unchecked_mtrl_animation;
			global::Xamarin.Forms.Platform.Resource.Drawable.btn_checkbox_unchecked_mtrl = global::pjsua2xamarin.Droid.Resource.Drawable.btn_checkbox_unchecked_mtrl;
			global::Xamarin.Forms.Platform.Resource.Drawable.btn_checkbox_unchecked_to_checked_mtrl_animation = global::pjsua2xamarin.Droid.Resource.Drawable.btn_checkbox_unchecked_to_checked_mtrl_animation;
			global::Xamarin.Forms.Platform.Resource.Drawable.btn_radio_off_mtrl = global::pjsua2xamarin.Droid.Resource.Drawable.btn_radio_off_mtrl;
			global::Xamarin.Forms.Platform.Resource.Drawable.btn_radio_off_to_on_mtrl_animation = global::pjsua2xamarin.Droid.Resource.Drawable.btn_radio_off_to_on_mtrl_animation;
			global::Xamarin.Forms.Platform.Resource.Drawable.btn_radio_on_mtrl = global::pjsua2xamarin.Droid.Resource.Drawable.btn_radio_on_mtrl;
			global::Xamarin.Forms.Platform.Resource.Drawable.btn_radio_on_to_off_mtrl_animation = global::pjsua2xamarin.Droid.Resource.Drawable.btn_radio_on_to_off_mtrl_animation;
			global::Xamarin.Forms.Platform.Resource.Drawable.design_fab_background = global::pjsua2xamarin.Droid.Resource.Drawable.design_fab_background;
			global::Xamarin.Forms.Platform.Resource.Drawable.design_ic_visibility = global::pjsua2xamarin.Droid.Resource.Drawable.design_ic_visibility;
			global::Xamarin.Forms.Platform.Resource.Drawable.design_ic_visibility_off = global::pjsua2xamarin.Droid.Resource.Drawable.design_ic_visibility_off;
			global::Xamarin.Forms.Platform.Resource.Drawable.design_password_eye = global::pjsua2xamarin.Droid.Resource.Drawable.design_password_eye;
			global::Xamarin.Forms.Platform.Resource.Drawable.design_snackbar_background = global::pjsua2xamarin.Droid.Resource.Drawable.design_snackbar_background;
			global::Xamarin.Forms.Platform.Resource.Drawable.ic_clock_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.ic_clock_black_24dp;
			global::Xamarin.Forms.Platform.Resource.Drawable.ic_keyboard_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.ic_keyboard_black_24dp;
			global::Xamarin.Forms.Platform.Resource.Drawable.ic_mtrl_checked_circle = global::pjsua2xamarin.Droid.Resource.Drawable.ic_mtrl_checked_circle;
			global::Xamarin.Forms.Platform.Resource.Drawable.ic_mtrl_chip_checked_black = global::pjsua2xamarin.Droid.Resource.Drawable.ic_mtrl_chip_checked_black;
			global::Xamarin.Forms.Platform.Resource.Drawable.ic_mtrl_chip_checked_circle = global::pjsua2xamarin.Droid.Resource.Drawable.ic_mtrl_chip_checked_circle;
			global::Xamarin.Forms.Platform.Resource.Drawable.ic_mtrl_chip_close_circle = global::pjsua2xamarin.Droid.Resource.Drawable.ic_mtrl_chip_close_circle;
			global::Xamarin.Forms.Platform.Resource.Drawable.material_cursor_drawable = global::pjsua2xamarin.Droid.Resource.Drawable.material_cursor_drawable;
			global::Xamarin.Forms.Platform.Resource.Drawable.material_ic_calendar_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.material_ic_calendar_black_24dp;
			global::Xamarin.Forms.Platform.Resource.Drawable.material_ic_clear_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.material_ic_clear_black_24dp;
			global::Xamarin.Forms.Platform.Resource.Drawable.material_ic_edit_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.material_ic_edit_black_24dp;
			global::Xamarin.Forms.Platform.Resource.Drawable.material_ic_keyboard_arrow_left_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.material_ic_keyboard_arrow_left_black_24dp;
			global::Xamarin.Forms.Platform.Resource.Drawable.material_ic_keyboard_arrow_next_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.material_ic_keyboard_arrow_next_black_24dp;
			global::Xamarin.Forms.Platform.Resource.Drawable.material_ic_keyboard_arrow_previous_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.material_ic_keyboard_arrow_previous_black_24dp;
			global::Xamarin.Forms.Platform.Resource.Drawable.material_ic_keyboard_arrow_right_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.material_ic_keyboard_arrow_right_black_24dp;
			global::Xamarin.Forms.Platform.Resource.Drawable.material_ic_menu_arrow_down_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.material_ic_menu_arrow_down_black_24dp;
			global::Xamarin.Forms.Platform.Resource.Drawable.material_ic_menu_arrow_up_black_24dp = global::pjsua2xamarin.Droid.Resource.Drawable.material_ic_menu_arrow_up_black_24dp;
			global::Xamarin.Forms.Platform.Resource.Drawable.mtrl_dialog_background = global::pjsua2xamarin.Droid.Resource.Drawable.mtrl_dialog_background;
			global::Xamarin.Forms.Platform.Resource.Drawable.mtrl_dropdown_arrow = global::pjsua2xamarin.Droid.Resource.Drawable.mtrl_dropdown_arrow;
			global::Xamarin.Forms.Platform.Resource.Drawable.mtrl_ic_arrow_drop_down = global::pjsua2xamarin.Droid.Resource.Drawable.mtrl_ic_arrow_drop_down;
			global::Xamarin.Forms.Platform.Resource.Drawable.mtrl_ic_arrow_drop_up = global::pjsua2xamarin.Droid.Resource.Drawable.mtrl_ic_arrow_drop_up;
			global::Xamarin.Forms.Platform.Resource.Drawable.mtrl_ic_cancel = global::pjsua2xamarin.Droid.Resource.Drawable.mtrl_ic_cancel;
			global::Xamarin.Forms.Platform.Resource.Drawable.mtrl_ic_error = global::pjsua2xamarin.Droid.Resource.Drawable.mtrl_ic_error;
			global::Xamarin.Forms.Platform.Resource.Drawable.mtrl_navigation_bar_item_background = global::pjsua2xamarin.Droid.Resource.Drawable.mtrl_navigation_bar_item_background;
			global::Xamarin.Forms.Platform.Resource.Drawable.mtrl_popupmenu_background = global::pjsua2xamarin.Droid.Resource.Drawable.mtrl_popupmenu_background;
			global::Xamarin.Forms.Platform.Resource.Drawable.mtrl_popupmenu_background_dark = global::pjsua2xamarin.Droid.Resource.Drawable.mtrl_popupmenu_background_dark;
			global::Xamarin.Forms.Platform.Resource.Drawable.mtrl_tabs_default_indicator = global::pjsua2xamarin.Droid.Resource.Drawable.mtrl_tabs_default_indicator;
			global::Xamarin.Forms.Platform.Resource.Drawable.navigation_empty_icon = global::pjsua2xamarin.Droid.Resource.Drawable.navigation_empty_icon;
			global::Xamarin.Forms.Platform.Resource.Drawable.notification_action_background = global::pjsua2xamarin.Droid.Resource.Drawable.notification_action_background;
			global::Xamarin.Forms.Platform.Resource.Drawable.notification_bg = global::pjsua2xamarin.Droid.Resource.Drawable.notification_bg;
			global::Xamarin.Forms.Platform.Resource.Drawable.notification_bg_low = global::pjsua2xamarin.Droid.Resource.Drawable.notification_bg_low;
			global::Xamarin.Forms.Platform.Resource.Drawable.notification_bg_low_normal = global::pjsua2xamarin.Droid.Resource.Drawable.notification_bg_low_normal;
			global::Xamarin.Forms.Platform.Resource.Drawable.notification_bg_low_pressed = global::pjsua2xamarin.Droid.Resource.Drawable.notification_bg_low_pressed;
			global::Xamarin.Forms.Platform.Resource.Drawable.notification_bg_normal = global::pjsua2xamarin.Droid.Resource.Drawable.notification_bg_normal;
			global::Xamarin.Forms.Platform.Resource.Drawable.notification_bg_normal_pressed = global::pjsua2xamarin.Droid.Resource.Drawable.notification_bg_normal_pressed;
			global::Xamarin.Forms.Platform.Resource.Drawable.notification_icon_background = global::pjsua2xamarin.Droid.Resource.Drawable.notification_icon_background;
			global::Xamarin.Forms.Platform.Resource.Drawable.notification_template_icon_bg = global::pjsua2xamarin.Droid.Resource.Drawable.notification_template_icon_bg;
			global::Xamarin.Forms.Platform.Resource.Drawable.notification_template_icon_low_bg = global::pjsua2xamarin.Droid.Resource.Drawable.notification_template_icon_low_bg;
			global::Xamarin.Forms.Platform.Resource.Drawable.notification_tile_bg = global::pjsua2xamarin.Droid.Resource.Drawable.notification_tile_bg;
			global::Xamarin.Forms.Platform.Resource.Drawable.notify_panel_notification_icon_bg = global::pjsua2xamarin.Droid.Resource.Drawable.notify_panel_notification_icon_bg;
			global::Xamarin.Forms.Platform.Resource.Drawable.test_custom_background = global::pjsua2xamarin.Droid.Resource.Drawable.test_custom_background;
			global::Xamarin.Forms.Platform.Resource.Drawable.tooltip_frame_dark = global::pjsua2xamarin.Droid.Resource.Drawable.tooltip_frame_dark;
			global::Xamarin.Forms.Platform.Resource.Drawable.tooltip_frame_light = global::pjsua2xamarin.Droid.Resource.Drawable.tooltip_frame_light;
			global::Xamarin.Forms.Platform.Resource.Id.accelerate = global::pjsua2xamarin.Droid.Resource.Id.accelerate;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_action_clickable_span = global::pjsua2xamarin.Droid.Resource.Id.accessibility_action_clickable_span;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_0 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_0;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_1 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_1;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_10 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_10;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_11 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_11;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_12 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_12;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_13 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_13;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_14 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_14;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_15 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_15;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_16 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_16;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_17 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_17;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_18 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_18;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_19 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_19;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_2 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_2;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_20 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_20;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_21 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_21;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_22 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_22;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_23 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_23;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_24 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_24;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_25 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_25;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_26 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_26;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_27 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_27;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_28 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_28;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_29 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_29;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_3 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_3;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_30 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_30;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_31 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_31;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_4 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_4;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_5 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_5;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_6 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_6;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_7 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_7;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_8 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_8;
			global::Xamarin.Forms.Platform.Resource.Id.accessibility_custom_action_9 = global::pjsua2xamarin.Droid.Resource.Id.accessibility_custom_action_9;
			global::Xamarin.Forms.Platform.Resource.Id.action0 = global::pjsua2xamarin.Droid.Resource.Id.action0;
			global::Xamarin.Forms.Platform.Resource.Id.actionDown = global::pjsua2xamarin.Droid.Resource.Id.actionDown;
			global::Xamarin.Forms.Platform.Resource.Id.actionDownUp = global::pjsua2xamarin.Droid.Resource.Id.actionDownUp;
			global::Xamarin.Forms.Platform.Resource.Id.actions = global::pjsua2xamarin.Droid.Resource.Id.actions;
			global::Xamarin.Forms.Platform.Resource.Id.actionUp = global::pjsua2xamarin.Droid.Resource.Id.actionUp;
			global::Xamarin.Forms.Platform.Resource.Id.action_bar = global::pjsua2xamarin.Droid.Resource.Id.action_bar;
			global::Xamarin.Forms.Platform.Resource.Id.action_bar_activity_content = global::pjsua2xamarin.Droid.Resource.Id.action_bar_activity_content;
			global::Xamarin.Forms.Platform.Resource.Id.action_bar_container = global::pjsua2xamarin.Droid.Resource.Id.action_bar_container;
			global::Xamarin.Forms.Platform.Resource.Id.action_bar_root = global::pjsua2xamarin.Droid.Resource.Id.action_bar_root;
			global::Xamarin.Forms.Platform.Resource.Id.action_bar_spinner = global::pjsua2xamarin.Droid.Resource.Id.action_bar_spinner;
			global::Xamarin.Forms.Platform.Resource.Id.action_bar_subtitle = global::pjsua2xamarin.Droid.Resource.Id.action_bar_subtitle;
			global::Xamarin.Forms.Platform.Resource.Id.action_bar_title = global::pjsua2xamarin.Droid.Resource.Id.action_bar_title;
			global::Xamarin.Forms.Platform.Resource.Id.action_container = global::pjsua2xamarin.Droid.Resource.Id.action_container;
			global::Xamarin.Forms.Platform.Resource.Id.action_context_bar = global::pjsua2xamarin.Droid.Resource.Id.action_context_bar;
			global::Xamarin.Forms.Platform.Resource.Id.action_divider = global::pjsua2xamarin.Droid.Resource.Id.action_divider;
			global::Xamarin.Forms.Platform.Resource.Id.action_image = global::pjsua2xamarin.Droid.Resource.Id.action_image;
			global::Xamarin.Forms.Platform.Resource.Id.action_menu_divider = global::pjsua2xamarin.Droid.Resource.Id.action_menu_divider;
			global::Xamarin.Forms.Platform.Resource.Id.action_menu_presenter = global::pjsua2xamarin.Droid.Resource.Id.action_menu_presenter;
			global::Xamarin.Forms.Platform.Resource.Id.action_mode_bar = global::pjsua2xamarin.Droid.Resource.Id.action_mode_bar;
			global::Xamarin.Forms.Platform.Resource.Id.action_mode_bar_stub = global::pjsua2xamarin.Droid.Resource.Id.action_mode_bar_stub;
			global::Xamarin.Forms.Platform.Resource.Id.action_mode_close_button = global::pjsua2xamarin.Droid.Resource.Id.action_mode_close_button;
			global::Xamarin.Forms.Platform.Resource.Id.action_text = global::pjsua2xamarin.Droid.Resource.Id.action_text;
			global::Xamarin.Forms.Platform.Resource.Id.activity_chooser_view_content = global::pjsua2xamarin.Droid.Resource.Id.activity_chooser_view_content;
			global::Xamarin.Forms.Platform.Resource.Id.add = global::pjsua2xamarin.Droid.Resource.Id.add;
			global::Xamarin.Forms.Platform.Resource.Id.alertTitle = global::pjsua2xamarin.Droid.Resource.Id.alertTitle;
			global::Xamarin.Forms.Platform.Resource.Id.aligned = global::pjsua2xamarin.Droid.Resource.Id.aligned;
			global::Xamarin.Forms.Platform.Resource.Id.all = global::pjsua2xamarin.Droid.Resource.Id.all;
			global::Xamarin.Forms.Platform.Resource.Id.allStates = global::pjsua2xamarin.Droid.Resource.Id.allStates;
			global::Xamarin.Forms.Platform.Resource.Id.ALT = global::pjsua2xamarin.Droid.Resource.Id.ALT;
			global::Xamarin.Forms.Platform.Resource.Id.always = global::pjsua2xamarin.Droid.Resource.Id.always;
			global::Xamarin.Forms.Platform.Resource.Id.animateToEnd = global::pjsua2xamarin.Droid.Resource.Id.animateToEnd;
			global::Xamarin.Forms.Platform.Resource.Id.animateToStart = global::pjsua2xamarin.Droid.Resource.Id.animateToStart;
			global::Xamarin.Forms.Platform.Resource.Id.anticipate = global::pjsua2xamarin.Droid.Resource.Id.anticipate;
			global::Xamarin.Forms.Platform.Resource.Id.antiClockwise = global::pjsua2xamarin.Droid.Resource.Id.antiClockwise;
			global::Xamarin.Forms.Platform.Resource.Id.arc = global::pjsua2xamarin.Droid.Resource.Id.arc;
			global::Xamarin.Forms.Platform.Resource.Id.asConfigured = global::pjsua2xamarin.Droid.Resource.Id.asConfigured;
			global::Xamarin.Forms.Platform.Resource.Id.async = global::pjsua2xamarin.Droid.Resource.Id.async;
			global::Xamarin.Forms.Platform.Resource.Id.auto = global::pjsua2xamarin.Droid.Resource.Id.auto;
			global::Xamarin.Forms.Platform.Resource.Id.autoComplete = global::pjsua2xamarin.Droid.Resource.Id.autoComplete;
			global::Xamarin.Forms.Platform.Resource.Id.autoCompleteToEnd = global::pjsua2xamarin.Droid.Resource.Id.autoCompleteToEnd;
			global::Xamarin.Forms.Platform.Resource.Id.autoCompleteToStart = global::pjsua2xamarin.Droid.Resource.Id.autoCompleteToStart;
			global::Xamarin.Forms.Platform.Resource.Id.barrier = global::pjsua2xamarin.Droid.Resource.Id.barrier;
			global::Xamarin.Forms.Platform.Resource.Id.baseline = global::pjsua2xamarin.Droid.Resource.Id.baseline;
			global::Xamarin.Forms.Platform.Resource.Id.beginning = global::pjsua2xamarin.Droid.Resource.Id.beginning;
			global::Xamarin.Forms.Platform.Resource.Id.beginOnFirstDraw = global::pjsua2xamarin.Droid.Resource.Id.beginOnFirstDraw;
			global::Xamarin.Forms.Platform.Resource.Id.bestChoice = global::pjsua2xamarin.Droid.Resource.Id.bestChoice;
			global::Xamarin.Forms.Platform.Resource.Id.blocking = global::pjsua2xamarin.Droid.Resource.Id.blocking;
			global::Xamarin.Forms.Platform.Resource.Id.bottom = global::pjsua2xamarin.Droid.Resource.Id.bottom;
			global::Xamarin.Forms.Platform.Resource.Id.bottomtab_navarea = global::pjsua2xamarin.Droid.Resource.Id.bottomtab_navarea;
			global::Xamarin.Forms.Platform.Resource.Id.bottomtab_tabbar = global::pjsua2xamarin.Droid.Resource.Id.bottomtab_tabbar;
			global::Xamarin.Forms.Platform.Resource.Id.BOTTOM_END = global::pjsua2xamarin.Droid.Resource.Id.BOTTOM_END;
			global::Xamarin.Forms.Platform.Resource.Id.BOTTOM_START = global::pjsua2xamarin.Droid.Resource.Id.BOTTOM_START;
			global::Xamarin.Forms.Platform.Resource.Id.bounce = global::pjsua2xamarin.Droid.Resource.Id.bounce;
			global::Xamarin.Forms.Platform.Resource.Id.bounceBoth = global::pjsua2xamarin.Droid.Resource.Id.bounceBoth;
			global::Xamarin.Forms.Platform.Resource.Id.bounceEnd = global::pjsua2xamarin.Droid.Resource.Id.bounceEnd;
			global::Xamarin.Forms.Platform.Resource.Id.bounceStart = global::pjsua2xamarin.Droid.Resource.Id.bounceStart;
			global::Xamarin.Forms.Platform.Resource.Id.buttonPanel = global::pjsua2xamarin.Droid.Resource.Id.buttonPanel;
			global::Xamarin.Forms.Platform.Resource.Id.cache_measures = global::pjsua2xamarin.Droid.Resource.Id.cache_measures;
			global::Xamarin.Forms.Platform.Resource.Id.cancel_action = global::pjsua2xamarin.Droid.Resource.Id.cancel_action;
			global::Xamarin.Forms.Platform.Resource.Id.cancel_button = global::pjsua2xamarin.Droid.Resource.Id.cancel_button;
			global::Xamarin.Forms.Platform.Resource.Id.carryVelocity = global::pjsua2xamarin.Droid.Resource.Id.carryVelocity;
			global::Xamarin.Forms.Platform.Resource.Id.center = global::pjsua2xamarin.Droid.Resource.Id.center;
			global::Xamarin.Forms.Platform.Resource.Id.center_horizontal = global::pjsua2xamarin.Droid.Resource.Id.center_horizontal;
			global::Xamarin.Forms.Platform.Resource.Id.center_vertical = global::pjsua2xamarin.Droid.Resource.Id.center_vertical;
			global::Xamarin.Forms.Platform.Resource.Id.chain = global::pjsua2xamarin.Droid.Resource.Id.chain;
			global::Xamarin.Forms.Platform.Resource.Id.chains = global::pjsua2xamarin.Droid.Resource.Id.chains;
			global::Xamarin.Forms.Platform.Resource.Id.checkbox = global::pjsua2xamarin.Droid.Resource.Id.checkbox;
			global::Xamarin.Forms.Platform.Resource.Id.@checked = global::pjsua2xamarin.Droid.Resource.Id.@checked;
			global::Xamarin.Forms.Platform.Resource.Id.chip = global::pjsua2xamarin.Droid.Resource.Id.chip;
			global::Xamarin.Forms.Platform.Resource.Id.chip1 = global::pjsua2xamarin.Droid.Resource.Id.chip1;
			global::Xamarin.Forms.Platform.Resource.Id.chip2 = global::pjsua2xamarin.Droid.Resource.Id.chip2;
			global::Xamarin.Forms.Platform.Resource.Id.chip3 = global::pjsua2xamarin.Droid.Resource.Id.chip3;
			global::Xamarin.Forms.Platform.Resource.Id.chip_group = global::pjsua2xamarin.Droid.Resource.Id.chip_group;
			global::Xamarin.Forms.Platform.Resource.Id.chronometer = global::pjsua2xamarin.Droid.Resource.Id.chronometer;
			global::Xamarin.Forms.Platform.Resource.Id.circle_center = global::pjsua2xamarin.Droid.Resource.Id.circle_center;
			global::Xamarin.Forms.Platform.Resource.Id.clear_text = global::pjsua2xamarin.Droid.Resource.Id.clear_text;
			global::Xamarin.Forms.Platform.Resource.Id.clip_horizontal = global::pjsua2xamarin.Droid.Resource.Id.clip_horizontal;
			global::Xamarin.Forms.Platform.Resource.Id.clip_vertical = global::pjsua2xamarin.Droid.Resource.Id.clip_vertical;
			global::Xamarin.Forms.Platform.Resource.Id.clockwise = global::pjsua2xamarin.Droid.Resource.Id.clockwise;
			global::Xamarin.Forms.Platform.Resource.Id.closest = global::pjsua2xamarin.Droid.Resource.Id.closest;
			global::Xamarin.Forms.Platform.Resource.Id.collapseActionView = global::pjsua2xamarin.Droid.Resource.Id.collapseActionView;
			global::Xamarin.Forms.Platform.Resource.Id.confirm_button = global::pjsua2xamarin.Droid.Resource.Id.confirm_button;
			global::Xamarin.Forms.Platform.Resource.Id.constraint = global::pjsua2xamarin.Droid.Resource.Id.constraint;
			global::Xamarin.Forms.Platform.Resource.Id.container = global::pjsua2xamarin.Droid.Resource.Id.container;
			global::Xamarin.Forms.Platform.Resource.Id.content = global::pjsua2xamarin.Droid.Resource.Id.content;
			global::Xamarin.Forms.Platform.Resource.Id.contentPanel = global::pjsua2xamarin.Droid.Resource.Id.contentPanel;
			global::Xamarin.Forms.Platform.Resource.Id.contiguous = global::pjsua2xamarin.Droid.Resource.Id.contiguous;
			global::Xamarin.Forms.Platform.Resource.Id.continuousVelocity = global::pjsua2xamarin.Droid.Resource.Id.continuousVelocity;
			global::Xamarin.Forms.Platform.Resource.Id.coordinator = global::pjsua2xamarin.Droid.Resource.Id.coordinator;
			global::Xamarin.Forms.Platform.Resource.Id.cos = global::pjsua2xamarin.Droid.Resource.Id.cos;
			global::Xamarin.Forms.Platform.Resource.Id.counterclockwise = global::pjsua2xamarin.Droid.Resource.Id.counterclockwise;
			global::Xamarin.Forms.Platform.Resource.Id.CTRL = global::pjsua2xamarin.Droid.Resource.Id.CTRL;
			global::Xamarin.Forms.Platform.Resource.Id.currentState = global::pjsua2xamarin.Droid.Resource.Id.currentState;
			global::Xamarin.Forms.Platform.Resource.Id.custom = global::pjsua2xamarin.Droid.Resource.Id.custom;
			global::Xamarin.Forms.Platform.Resource.Id.customPanel = global::pjsua2xamarin.Droid.Resource.Id.customPanel;
			global::Xamarin.Forms.Platform.Resource.Id.cut = global::pjsua2xamarin.Droid.Resource.Id.cut;
			global::Xamarin.Forms.Platform.Resource.Id.date_picker_actions = global::pjsua2xamarin.Droid.Resource.Id.date_picker_actions;
			global::Xamarin.Forms.Platform.Resource.Id.decelerate = global::pjsua2xamarin.Droid.Resource.Id.decelerate;
			global::Xamarin.Forms.Platform.Resource.Id.decelerateAndComplete = global::pjsua2xamarin.Droid.Resource.Id.decelerateAndComplete;
			global::Xamarin.Forms.Platform.Resource.Id.decor_content_parent = global::pjsua2xamarin.Droid.Resource.Id.decor_content_parent;
			global::Xamarin.Forms.Platform.Resource.Id.default_activity_button = global::pjsua2xamarin.Droid.Resource.Id.default_activity_button;
			global::Xamarin.Forms.Platform.Resource.Id.deltaRelative = global::pjsua2xamarin.Droid.Resource.Id.deltaRelative;
			global::Xamarin.Forms.Platform.Resource.Id.dependency_ordering = global::pjsua2xamarin.Droid.Resource.Id.dependency_ordering;
			global::Xamarin.Forms.Platform.Resource.Id.design_bottom_sheet = global::pjsua2xamarin.Droid.Resource.Id.design_bottom_sheet;
			global::Xamarin.Forms.Platform.Resource.Id.design_menu_item_action_area = global::pjsua2xamarin.Droid.Resource.Id.design_menu_item_action_area;
			global::Xamarin.Forms.Platform.Resource.Id.design_menu_item_action_area_stub = global::pjsua2xamarin.Droid.Resource.Id.design_menu_item_action_area_stub;
			global::Xamarin.Forms.Platform.Resource.Id.design_menu_item_text = global::pjsua2xamarin.Droid.Resource.Id.design_menu_item_text;
			global::Xamarin.Forms.Platform.Resource.Id.design_navigation_view = global::pjsua2xamarin.Droid.Resource.Id.design_navigation_view;
			global::Xamarin.Forms.Platform.Resource.Id.dialog_button = global::pjsua2xamarin.Droid.Resource.Id.dialog_button;
			global::Xamarin.Forms.Platform.Resource.Id.dimensions = global::pjsua2xamarin.Droid.Resource.Id.dimensions;
			global::Xamarin.Forms.Platform.Resource.Id.direct = global::pjsua2xamarin.Droid.Resource.Id.direct;
			global::Xamarin.Forms.Platform.Resource.Id.disableHome = global::pjsua2xamarin.Droid.Resource.Id.disableHome;
			global::Xamarin.Forms.Platform.Resource.Id.disableIntraAutoTransition = global::pjsua2xamarin.Droid.Resource.Id.disableIntraAutoTransition;
			global::Xamarin.Forms.Platform.Resource.Id.disablePostScroll = global::pjsua2xamarin.Droid.Resource.Id.disablePostScroll;
			global::Xamarin.Forms.Platform.Resource.Id.disableScroll = global::pjsua2xamarin.Droid.Resource.Id.disableScroll;
			global::Xamarin.Forms.Platform.Resource.Id.disjoint = global::pjsua2xamarin.Droid.Resource.Id.disjoint;
			global::Xamarin.Forms.Platform.Resource.Id.dragAnticlockwise = global::pjsua2xamarin.Droid.Resource.Id.dragAnticlockwise;
			global::Xamarin.Forms.Platform.Resource.Id.dragClockwise = global::pjsua2xamarin.Droid.Resource.Id.dragClockwise;
			global::Xamarin.Forms.Platform.Resource.Id.dragDown = global::pjsua2xamarin.Droid.Resource.Id.dragDown;
			global::Xamarin.Forms.Platform.Resource.Id.dragEnd = global::pjsua2xamarin.Droid.Resource.Id.dragEnd;
			global::Xamarin.Forms.Platform.Resource.Id.dragLeft = global::pjsua2xamarin.Droid.Resource.Id.dragLeft;
			global::Xamarin.Forms.Platform.Resource.Id.dragRight = global::pjsua2xamarin.Droid.Resource.Id.dragRight;
			global::Xamarin.Forms.Platform.Resource.Id.dragStart = global::pjsua2xamarin.Droid.Resource.Id.dragStart;
			global::Xamarin.Forms.Platform.Resource.Id.dragUp = global::pjsua2xamarin.Droid.Resource.Id.dragUp;
			global::Xamarin.Forms.Platform.Resource.Id.dropdown_menu = global::pjsua2xamarin.Droid.Resource.Id.dropdown_menu;
			global::Xamarin.Forms.Platform.Resource.Id.easeIn = global::pjsua2xamarin.Droid.Resource.Id.easeIn;
			global::Xamarin.Forms.Platform.Resource.Id.easeInOut = global::pjsua2xamarin.Droid.Resource.Id.easeInOut;
			global::Xamarin.Forms.Platform.Resource.Id.easeOut = global::pjsua2xamarin.Droid.Resource.Id.easeOut;
			global::Xamarin.Forms.Platform.Resource.Id.east = global::pjsua2xamarin.Droid.Resource.Id.east;
			global::Xamarin.Forms.Platform.Resource.Id.edit_query = global::pjsua2xamarin.Droid.Resource.Id.edit_query;
			global::Xamarin.Forms.Platform.Resource.Id.elastic = global::pjsua2xamarin.Droid.Resource.Id.elastic;
			global::Xamarin.Forms.Platform.Resource.Id.end = global::pjsua2xamarin.Droid.Resource.Id.end;
			global::Xamarin.Forms.Platform.Resource.Id.endToStart = global::pjsua2xamarin.Droid.Resource.Id.endToStart;
			global::Xamarin.Forms.Platform.Resource.Id.end_padder = global::pjsua2xamarin.Droid.Resource.Id.end_padder;
			global::Xamarin.Forms.Platform.Resource.Id.enterAlways = global::pjsua2xamarin.Droid.Resource.Id.enterAlways;
			global::Xamarin.Forms.Platform.Resource.Id.enterAlwaysCollapsed = global::pjsua2xamarin.Droid.Resource.Id.enterAlwaysCollapsed;
			global::Xamarin.Forms.Platform.Resource.Id.exitUntilCollapsed = global::pjsua2xamarin.Droid.Resource.Id.exitUntilCollapsed;
			global::Xamarin.Forms.Platform.Resource.Id.expanded_menu = global::pjsua2xamarin.Droid.Resource.Id.expanded_menu;
			global::Xamarin.Forms.Platform.Resource.Id.expand_activities_button = global::pjsua2xamarin.Droid.Resource.Id.expand_activities_button;
			global::Xamarin.Forms.Platform.Resource.Id.fade = global::pjsua2xamarin.Droid.Resource.Id.fade;
			global::Xamarin.Forms.Platform.Resource.Id.fill = global::pjsua2xamarin.Droid.Resource.Id.fill;
			global::Xamarin.Forms.Platform.Resource.Id.filled = global::pjsua2xamarin.Droid.Resource.Id.filled;
			global::Xamarin.Forms.Platform.Resource.Id.fill_horizontal = global::pjsua2xamarin.Droid.Resource.Id.fill_horizontal;
			global::Xamarin.Forms.Platform.Resource.Id.fill_vertical = global::pjsua2xamarin.Droid.Resource.Id.fill_vertical;
			global::Xamarin.Forms.Platform.Resource.Id.fitToContents = global::pjsua2xamarin.Droid.Resource.Id.fitToContents;
			global::Xamarin.Forms.Platform.Resource.Id.@fixed = global::pjsua2xamarin.Droid.Resource.Id.@fixed;
			global::Xamarin.Forms.Platform.Resource.Id.flip = global::pjsua2xamarin.Droid.Resource.Id.flip;
			global::Xamarin.Forms.Platform.Resource.Id.floating = global::pjsua2xamarin.Droid.Resource.Id.floating;
			global::Xamarin.Forms.Platform.Resource.Id.flyoutcontent_appbar = global::pjsua2xamarin.Droid.Resource.Id.flyoutcontent_appbar;
			global::Xamarin.Forms.Platform.Resource.Id.forever = global::pjsua2xamarin.Droid.Resource.Id.forever;
			global::Xamarin.Forms.Platform.Resource.Id.fragment_container_view_tag = global::pjsua2xamarin.Droid.Resource.Id.fragment_container_view_tag;
			global::Xamarin.Forms.Platform.Resource.Id.frost = global::pjsua2xamarin.Droid.Resource.Id.frost;
			global::Xamarin.Forms.Platform.Resource.Id.FUNCTION = global::pjsua2xamarin.Droid.Resource.Id.FUNCTION;
			global::Xamarin.Forms.Platform.Resource.Id.ghost_view = global::pjsua2xamarin.Droid.Resource.Id.ghost_view;
			global::Xamarin.Forms.Platform.Resource.Id.ghost_view_holder = global::pjsua2xamarin.Droid.Resource.Id.ghost_view_holder;
			global::Xamarin.Forms.Platform.Resource.Id.gone = global::pjsua2xamarin.Droid.Resource.Id.gone;
			global::Xamarin.Forms.Platform.Resource.Id.graph = global::pjsua2xamarin.Droid.Resource.Id.graph;
			global::Xamarin.Forms.Platform.Resource.Id.graph_wrap = global::pjsua2xamarin.Droid.Resource.Id.graph_wrap;
			global::Xamarin.Forms.Platform.Resource.Id.grouping = global::pjsua2xamarin.Droid.Resource.Id.grouping;
			global::Xamarin.Forms.Platform.Resource.Id.groups = global::pjsua2xamarin.Droid.Resource.Id.groups;
			global::Xamarin.Forms.Platform.Resource.Id.group_divider = global::pjsua2xamarin.Droid.Resource.Id.group_divider;
			global::Xamarin.Forms.Platform.Resource.Id.guideline = global::pjsua2xamarin.Droid.Resource.Id.guideline;
			global::Xamarin.Forms.Platform.Resource.Id.header_title = global::pjsua2xamarin.Droid.Resource.Id.header_title;
			global::Xamarin.Forms.Platform.Resource.Id.hideable = global::pjsua2xamarin.Droid.Resource.Id.hideable;
			global::Xamarin.Forms.Platform.Resource.Id.home = global::pjsua2xamarin.Droid.Resource.Id.home;
			global::Xamarin.Forms.Platform.Resource.Id.homeAsUp = global::pjsua2xamarin.Droid.Resource.Id.homeAsUp;
			global::Xamarin.Forms.Platform.Resource.Id.honorRequest = global::pjsua2xamarin.Droid.Resource.Id.honorRequest;
			global::Xamarin.Forms.Platform.Resource.Id.horizontal_only = global::pjsua2xamarin.Droid.Resource.Id.horizontal_only;
			global::Xamarin.Forms.Platform.Resource.Id.icon = global::pjsua2xamarin.Droid.Resource.Id.icon;
			global::Xamarin.Forms.Platform.Resource.Id.icon_group = global::pjsua2xamarin.Droid.Resource.Id.icon_group;
			global::Xamarin.Forms.Platform.Resource.Id.ifRoom = global::pjsua2xamarin.Droid.Resource.Id.ifRoom;
			global::Xamarin.Forms.Platform.Resource.Id.ignore = global::pjsua2xamarin.Droid.Resource.Id.ignore;
			global::Xamarin.Forms.Platform.Resource.Id.ignoreRequest = global::pjsua2xamarin.Droid.Resource.Id.ignoreRequest;
			global::Xamarin.Forms.Platform.Resource.Id.image = global::pjsua2xamarin.Droid.Resource.Id.image;
			global::Xamarin.Forms.Platform.Resource.Id.immediateStop = global::pjsua2xamarin.Droid.Resource.Id.immediateStop;
			global::Xamarin.Forms.Platform.Resource.Id.included = global::pjsua2xamarin.Droid.Resource.Id.included;
			global::Xamarin.Forms.Platform.Resource.Id.info = global::pjsua2xamarin.Droid.Resource.Id.info;
			global::Xamarin.Forms.Platform.Resource.Id.invisible = global::pjsua2xamarin.Droid.Resource.Id.invisible;
			global::Xamarin.Forms.Platform.Resource.Id.inward = global::pjsua2xamarin.Droid.Resource.Id.inward;
			global::Xamarin.Forms.Platform.Resource.Id.italic = global::pjsua2xamarin.Droid.Resource.Id.italic;
			global::Xamarin.Forms.Platform.Resource.Id.item_touch_helper_previous_elevation = global::pjsua2xamarin.Droid.Resource.Id.item_touch_helper_previous_elevation;
			global::Xamarin.Forms.Platform.Resource.Id.jumpToEnd = global::pjsua2xamarin.Droid.Resource.Id.jumpToEnd;
			global::Xamarin.Forms.Platform.Resource.Id.jumpToStart = global::pjsua2xamarin.Droid.Resource.Id.jumpToStart;
			global::Xamarin.Forms.Platform.Resource.Id.labeled = global::pjsua2xamarin.Droid.Resource.Id.labeled;
			global::Xamarin.Forms.Platform.Resource.Id.layout = global::pjsua2xamarin.Droid.Resource.Id.layout;
			global::Xamarin.Forms.Platform.Resource.Id.left = global::pjsua2xamarin.Droid.Resource.Id.left;
			global::Xamarin.Forms.Platform.Resource.Id.leftToRight = global::pjsua2xamarin.Droid.Resource.Id.leftToRight;
			global::Xamarin.Forms.Platform.Resource.Id.legacy = global::pjsua2xamarin.Droid.Resource.Id.legacy;
			global::Xamarin.Forms.Platform.Resource.Id.line1 = global::pjsua2xamarin.Droid.Resource.Id.line1;
			global::Xamarin.Forms.Platform.Resource.Id.line3 = global::pjsua2xamarin.Droid.Resource.Id.line3;
			global::Xamarin.Forms.Platform.Resource.Id.linear = global::pjsua2xamarin.Droid.Resource.Id.linear;
			global::Xamarin.Forms.Platform.Resource.Id.listMode = global::pjsua2xamarin.Droid.Resource.Id.listMode;
			global::Xamarin.Forms.Platform.Resource.Id.list_item = global::pjsua2xamarin.Droid.Resource.Id.list_item;
			global::Xamarin.Forms.Platform.Resource.Id.main_appbar = global::pjsua2xamarin.Droid.Resource.Id.main_appbar;
			global::Xamarin.Forms.Platform.Resource.Id.main_tablayout = global::pjsua2xamarin.Droid.Resource.Id.main_tablayout;
			global::Xamarin.Forms.Platform.Resource.Id.main_toolbar = global::pjsua2xamarin.Droid.Resource.Id.main_toolbar;
			global::Xamarin.Forms.Platform.Resource.Id.main_viewpager = global::pjsua2xamarin.Droid.Resource.Id.main_viewpager;
			global::Xamarin.Forms.Platform.Resource.Id.masked = global::pjsua2xamarin.Droid.Resource.Id.masked;
			global::Xamarin.Forms.Platform.Resource.Id.match_constraint = global::pjsua2xamarin.Droid.Resource.Id.match_constraint;
			global::Xamarin.Forms.Platform.Resource.Id.match_parent = global::pjsua2xamarin.Droid.Resource.Id.match_parent;
			global::Xamarin.Forms.Platform.Resource.Id.material_clock_display = global::pjsua2xamarin.Droid.Resource.Id.material_clock_display;
			global::Xamarin.Forms.Platform.Resource.Id.material_clock_face = global::pjsua2xamarin.Droid.Resource.Id.material_clock_face;
			global::Xamarin.Forms.Platform.Resource.Id.material_clock_hand = global::pjsua2xamarin.Droid.Resource.Id.material_clock_hand;
			global::Xamarin.Forms.Platform.Resource.Id.material_clock_period_am_button = global::pjsua2xamarin.Droid.Resource.Id.material_clock_period_am_button;
			global::Xamarin.Forms.Platform.Resource.Id.material_clock_period_pm_button = global::pjsua2xamarin.Droid.Resource.Id.material_clock_period_pm_button;
			global::Xamarin.Forms.Platform.Resource.Id.material_clock_period_toggle = global::pjsua2xamarin.Droid.Resource.Id.material_clock_period_toggle;
			global::Xamarin.Forms.Platform.Resource.Id.material_hour_text_input = global::pjsua2xamarin.Droid.Resource.Id.material_hour_text_input;
			global::Xamarin.Forms.Platform.Resource.Id.material_hour_tv = global::pjsua2xamarin.Droid.Resource.Id.material_hour_tv;
			global::Xamarin.Forms.Platform.Resource.Id.material_label = global::pjsua2xamarin.Droid.Resource.Id.material_label;
			global::Xamarin.Forms.Platform.Resource.Id.material_minute_text_input = global::pjsua2xamarin.Droid.Resource.Id.material_minute_text_input;
			global::Xamarin.Forms.Platform.Resource.Id.material_minute_tv = global::pjsua2xamarin.Droid.Resource.Id.material_minute_tv;
			global::Xamarin.Forms.Platform.Resource.Id.material_textinput_timepicker = global::pjsua2xamarin.Droid.Resource.Id.material_textinput_timepicker;
			global::Xamarin.Forms.Platform.Resource.Id.material_timepicker_cancel_button = global::pjsua2xamarin.Droid.Resource.Id.material_timepicker_cancel_button;
			global::Xamarin.Forms.Platform.Resource.Id.material_timepicker_container = global::pjsua2xamarin.Droid.Resource.Id.material_timepicker_container;
			global::Xamarin.Forms.Platform.Resource.Id.material_timepicker_edit_text = global::pjsua2xamarin.Droid.Resource.Id.material_timepicker_edit_text;
			global::Xamarin.Forms.Platform.Resource.Id.material_timepicker_mode_button = global::pjsua2xamarin.Droid.Resource.Id.material_timepicker_mode_button;
			global::Xamarin.Forms.Platform.Resource.Id.material_timepicker_ok_button = global::pjsua2xamarin.Droid.Resource.Id.material_timepicker_ok_button;
			global::Xamarin.Forms.Platform.Resource.Id.material_timepicker_view = global::pjsua2xamarin.Droid.Resource.Id.material_timepicker_view;
			global::Xamarin.Forms.Platform.Resource.Id.material_value_index = global::pjsua2xamarin.Droid.Resource.Id.material_value_index;
			global::Xamarin.Forms.Platform.Resource.Id.media_actions = global::pjsua2xamarin.Droid.Resource.Id.media_actions;
			global::Xamarin.Forms.Platform.Resource.Id.media_controller_compat_view_tag = global::pjsua2xamarin.Droid.Resource.Id.media_controller_compat_view_tag;
			global::Xamarin.Forms.Platform.Resource.Id.message = global::pjsua2xamarin.Droid.Resource.Id.message;
			global::Xamarin.Forms.Platform.Resource.Id.META = global::pjsua2xamarin.Droid.Resource.Id.META;
			global::Xamarin.Forms.Platform.Resource.Id.middle = global::pjsua2xamarin.Droid.Resource.Id.middle;
			global::Xamarin.Forms.Platform.Resource.Id.mini = global::pjsua2xamarin.Droid.Resource.Id.mini;
			global::Xamarin.Forms.Platform.Resource.Id.month_grid = global::pjsua2xamarin.Droid.Resource.Id.month_grid;
			global::Xamarin.Forms.Platform.Resource.Id.month_navigation_bar = global::pjsua2xamarin.Droid.Resource.Id.month_navigation_bar;
			global::Xamarin.Forms.Platform.Resource.Id.month_navigation_fragment_toggle = global::pjsua2xamarin.Droid.Resource.Id.month_navigation_fragment_toggle;
			global::Xamarin.Forms.Platform.Resource.Id.month_navigation_next = global::pjsua2xamarin.Droid.Resource.Id.month_navigation_next;
			global::Xamarin.Forms.Platform.Resource.Id.month_navigation_previous = global::pjsua2xamarin.Droid.Resource.Id.month_navigation_previous;
			global::Xamarin.Forms.Platform.Resource.Id.month_title = global::pjsua2xamarin.Droid.Resource.Id.month_title;
			global::Xamarin.Forms.Platform.Resource.Id.motion_base = global::pjsua2xamarin.Droid.Resource.Id.motion_base;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_anchor_parent = global::pjsua2xamarin.Droid.Resource.Id.mtrl_anchor_parent;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_calendar_days_of_week = global::pjsua2xamarin.Droid.Resource.Id.mtrl_calendar_days_of_week;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_calendar_day_selector_frame = global::pjsua2xamarin.Droid.Resource.Id.mtrl_calendar_day_selector_frame;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_calendar_frame = global::pjsua2xamarin.Droid.Resource.Id.mtrl_calendar_frame;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_calendar_main_pane = global::pjsua2xamarin.Droid.Resource.Id.mtrl_calendar_main_pane;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_calendar_months = global::pjsua2xamarin.Droid.Resource.Id.mtrl_calendar_months;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_calendar_selection_frame = global::pjsua2xamarin.Droid.Resource.Id.mtrl_calendar_selection_frame;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_calendar_text_input_frame = global::pjsua2xamarin.Droid.Resource.Id.mtrl_calendar_text_input_frame;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_calendar_year_selector_frame = global::pjsua2xamarin.Droid.Resource.Id.mtrl_calendar_year_selector_frame;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_card_checked_layer_id = global::pjsua2xamarin.Droid.Resource.Id.mtrl_card_checked_layer_id;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_child_content_container = global::pjsua2xamarin.Droid.Resource.Id.mtrl_child_content_container;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_internal_children_alpha_tag = global::pjsua2xamarin.Droid.Resource.Id.mtrl_internal_children_alpha_tag;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_motion_snapshot_view = global::pjsua2xamarin.Droid.Resource.Id.mtrl_motion_snapshot_view;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_picker_fullscreen = global::pjsua2xamarin.Droid.Resource.Id.mtrl_picker_fullscreen;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_picker_header = global::pjsua2xamarin.Droid.Resource.Id.mtrl_picker_header;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_picker_header_selection_text = global::pjsua2xamarin.Droid.Resource.Id.mtrl_picker_header_selection_text;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_picker_header_title_and_selection = global::pjsua2xamarin.Droid.Resource.Id.mtrl_picker_header_title_and_selection;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_picker_header_toggle = global::pjsua2xamarin.Droid.Resource.Id.mtrl_picker_header_toggle;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_picker_text_input_date = global::pjsua2xamarin.Droid.Resource.Id.mtrl_picker_text_input_date;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_picker_text_input_range_end = global::pjsua2xamarin.Droid.Resource.Id.mtrl_picker_text_input_range_end;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_picker_text_input_range_start = global::pjsua2xamarin.Droid.Resource.Id.mtrl_picker_text_input_range_start;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_picker_title_text = global::pjsua2xamarin.Droid.Resource.Id.mtrl_picker_title_text;
			global::Xamarin.Forms.Platform.Resource.Id.mtrl_view_tag_bottom_padding = global::pjsua2xamarin.Droid.Resource.Id.mtrl_view_tag_bottom_padding;
			global::Xamarin.Forms.Platform.Resource.Id.multiply = global::pjsua2xamarin.Droid.Resource.Id.multiply;
			global::Xamarin.Forms.Platform.Resource.Id.navigation_bar_item_icon_view = global::pjsua2xamarin.Droid.Resource.Id.navigation_bar_item_icon_view;
			global::Xamarin.Forms.Platform.Resource.Id.navigation_bar_item_labels_group = global::pjsua2xamarin.Droid.Resource.Id.navigation_bar_item_labels_group;
			global::Xamarin.Forms.Platform.Resource.Id.navigation_bar_item_large_label_view = global::pjsua2xamarin.Droid.Resource.Id.navigation_bar_item_large_label_view;
			global::Xamarin.Forms.Platform.Resource.Id.navigation_bar_item_small_label_view = global::pjsua2xamarin.Droid.Resource.Id.navigation_bar_item_small_label_view;
			global::Xamarin.Forms.Platform.Resource.Id.navigation_header_container = global::pjsua2xamarin.Droid.Resource.Id.navigation_header_container;
			global::Xamarin.Forms.Platform.Resource.Id.nav_controller_view_tag = global::pjsua2xamarin.Droid.Resource.Id.nav_controller_view_tag;
			global::Xamarin.Forms.Platform.Resource.Id.never = global::pjsua2xamarin.Droid.Resource.Id.never;
			global::Xamarin.Forms.Platform.Resource.Id.neverCompleteToEnd = global::pjsua2xamarin.Droid.Resource.Id.neverCompleteToEnd;
			global::Xamarin.Forms.Platform.Resource.Id.neverCompleteToStart = global::pjsua2xamarin.Droid.Resource.Id.neverCompleteToStart;
			global::Xamarin.Forms.Platform.Resource.Id.none = global::pjsua2xamarin.Droid.Resource.Id.none;
			global::Xamarin.Forms.Platform.Resource.Id.normal = global::pjsua2xamarin.Droid.Resource.Id.normal;
			global::Xamarin.Forms.Platform.Resource.Id.north = global::pjsua2xamarin.Droid.Resource.Id.north;
			global::Xamarin.Forms.Platform.Resource.Id.noScroll = global::pjsua2xamarin.Droid.Resource.Id.noScroll;
			global::Xamarin.Forms.Platform.Resource.Id.noState = global::pjsua2xamarin.Droid.Resource.Id.noState;
			global::Xamarin.Forms.Platform.Resource.Id.notification_background = global::pjsua2xamarin.Droid.Resource.Id.notification_background;
			global::Xamarin.Forms.Platform.Resource.Id.notification_main_column = global::pjsua2xamarin.Droid.Resource.Id.notification_main_column;
			global::Xamarin.Forms.Platform.Resource.Id.notification_main_column_container = global::pjsua2xamarin.Droid.Resource.Id.notification_main_column_container;
			global::Xamarin.Forms.Platform.Resource.Id.NO_DEBUG = global::pjsua2xamarin.Droid.Resource.Id.NO_DEBUG;
			global::Xamarin.Forms.Platform.Resource.Id.off = global::pjsua2xamarin.Droid.Resource.Id.off;
			global::Xamarin.Forms.Platform.Resource.Id.on = global::pjsua2xamarin.Droid.Resource.Id.on;
			global::Xamarin.Forms.Platform.Resource.Id.outline = global::pjsua2xamarin.Droid.Resource.Id.outline;
			global::Xamarin.Forms.Platform.Resource.Id.outward = global::pjsua2xamarin.Droid.Resource.Id.outward;
			global::Xamarin.Forms.Platform.Resource.Id.overshoot = global::pjsua2xamarin.Droid.Resource.Id.overshoot;
			global::Xamarin.Forms.Platform.Resource.Id.packed = global::pjsua2xamarin.Droid.Resource.Id.packed;
			global::Xamarin.Forms.Platform.Resource.Id.parallax = global::pjsua2xamarin.Droid.Resource.Id.parallax;
			global::Xamarin.Forms.Platform.Resource.Id.parent = global::pjsua2xamarin.Droid.Resource.Id.parent;
			global::Xamarin.Forms.Platform.Resource.Id.parentPanel = global::pjsua2xamarin.Droid.Resource.Id.parentPanel;
			global::Xamarin.Forms.Platform.Resource.Id.parentRelative = global::pjsua2xamarin.Droid.Resource.Id.parentRelative;
			global::Xamarin.Forms.Platform.Resource.Id.parent_matrix = global::pjsua2xamarin.Droid.Resource.Id.parent_matrix;
			global::Xamarin.Forms.Platform.Resource.Id.password_toggle = global::pjsua2xamarin.Droid.Resource.Id.password_toggle;
			global::Xamarin.Forms.Platform.Resource.Id.path = global::pjsua2xamarin.Droid.Resource.Id.path;
			global::Xamarin.Forms.Platform.Resource.Id.pathRelative = global::pjsua2xamarin.Droid.Resource.Id.pathRelative;
			global::Xamarin.Forms.Platform.Resource.Id.peekHeight = global::pjsua2xamarin.Droid.Resource.Id.peekHeight;
			global::Xamarin.Forms.Platform.Resource.Id.percent = global::pjsua2xamarin.Droid.Resource.Id.percent;
			global::Xamarin.Forms.Platform.Resource.Id.pin = global::pjsua2xamarin.Droid.Resource.Id.pin;
			global::Xamarin.Forms.Platform.Resource.Id.position = global::pjsua2xamarin.Droid.Resource.Id.position;
			global::Xamarin.Forms.Platform.Resource.Id.postLayout = global::pjsua2xamarin.Droid.Resource.Id.postLayout;
			global::Xamarin.Forms.Platform.Resource.Id.progress_circular = global::pjsua2xamarin.Droid.Resource.Id.progress_circular;
			global::Xamarin.Forms.Platform.Resource.Id.progress_horizontal = global::pjsua2xamarin.Droid.Resource.Id.progress_horizontal;
			global::Xamarin.Forms.Platform.Resource.Id.radio = global::pjsua2xamarin.Droid.Resource.Id.radio;
			global::Xamarin.Forms.Platform.Resource.Id.ratio = global::pjsua2xamarin.Droid.Resource.Id.ratio;
			global::Xamarin.Forms.Platform.Resource.Id.rectangles = global::pjsua2xamarin.Droid.Resource.Id.rectangles;
			global::Xamarin.Forms.Platform.Resource.Id.reverseSawtooth = global::pjsua2xamarin.Droid.Resource.Id.reverseSawtooth;
			global::Xamarin.Forms.Platform.Resource.Id.right = global::pjsua2xamarin.Droid.Resource.Id.right;
			global::Xamarin.Forms.Platform.Resource.Id.rightToLeft = global::pjsua2xamarin.Droid.Resource.Id.rightToLeft;
			global::Xamarin.Forms.Platform.Resource.Id.right_icon = global::pjsua2xamarin.Droid.Resource.Id.right_icon;
			global::Xamarin.Forms.Platform.Resource.Id.right_side = global::pjsua2xamarin.Droid.Resource.Id.right_side;
			global::Xamarin.Forms.Platform.Resource.Id.rounded = global::pjsua2xamarin.Droid.Resource.Id.rounded;
			global::Xamarin.Forms.Platform.Resource.Id.row_index_key = global::pjsua2xamarin.Droid.Resource.Id.row_index_key;
			global::Xamarin.Forms.Platform.Resource.Id.save_non_transition_alpha = global::pjsua2xamarin.Droid.Resource.Id.save_non_transition_alpha;
			global::Xamarin.Forms.Platform.Resource.Id.save_overlay_view = global::pjsua2xamarin.Droid.Resource.Id.save_overlay_view;
			global::Xamarin.Forms.Platform.Resource.Id.sawtooth = global::pjsua2xamarin.Droid.Resource.Id.sawtooth;
			global::Xamarin.Forms.Platform.Resource.Id.scale = global::pjsua2xamarin.Droid.Resource.Id.scale;
			global::Xamarin.Forms.Platform.Resource.Id.screen = global::pjsua2xamarin.Droid.Resource.Id.screen;
			global::Xamarin.Forms.Platform.Resource.Id.scroll = global::pjsua2xamarin.Droid.Resource.Id.scroll;
			global::Xamarin.Forms.Platform.Resource.Id.scrollable = global::pjsua2xamarin.Droid.Resource.Id.scrollable;
			global::Xamarin.Forms.Platform.Resource.Id.scrollIndicatorDown = global::pjsua2xamarin.Droid.Resource.Id.scrollIndicatorDown;
			global::Xamarin.Forms.Platform.Resource.Id.scrollIndicatorUp = global::pjsua2xamarin.Droid.Resource.Id.scrollIndicatorUp;
			global::Xamarin.Forms.Platform.Resource.Id.scrollView = global::pjsua2xamarin.Droid.Resource.Id.scrollView;
			global::Xamarin.Forms.Platform.Resource.Id.search_badge = global::pjsua2xamarin.Droid.Resource.Id.search_badge;
			global::Xamarin.Forms.Platform.Resource.Id.search_bar = global::pjsua2xamarin.Droid.Resource.Id.search_bar;
			global::Xamarin.Forms.Platform.Resource.Id.search_button = global::pjsua2xamarin.Droid.Resource.Id.search_button;
			global::Xamarin.Forms.Platform.Resource.Id.search_close_btn = global::pjsua2xamarin.Droid.Resource.Id.search_close_btn;
			global::Xamarin.Forms.Platform.Resource.Id.search_edit_frame = global::pjsua2xamarin.Droid.Resource.Id.search_edit_frame;
			global::Xamarin.Forms.Platform.Resource.Id.search_go_btn = global::pjsua2xamarin.Droid.Resource.Id.search_go_btn;
			global::Xamarin.Forms.Platform.Resource.Id.search_mag_icon = global::pjsua2xamarin.Droid.Resource.Id.search_mag_icon;
			global::Xamarin.Forms.Platform.Resource.Id.search_plate = global::pjsua2xamarin.Droid.Resource.Id.search_plate;
			global::Xamarin.Forms.Platform.Resource.Id.search_src_text = global::pjsua2xamarin.Droid.Resource.Id.search_src_text;
			global::Xamarin.Forms.Platform.Resource.Id.search_voice_btn = global::pjsua2xamarin.Droid.Resource.Id.search_voice_btn;
			global::Xamarin.Forms.Platform.Resource.Id.selected = global::pjsua2xamarin.Droid.Resource.Id.selected;
			global::Xamarin.Forms.Platform.Resource.Id.selection_type = global::pjsua2xamarin.Droid.Resource.Id.selection_type;
			global::Xamarin.Forms.Platform.Resource.Id.select_dialog_listview = global::pjsua2xamarin.Droid.Resource.Id.select_dialog_listview;
			global::Xamarin.Forms.Platform.Resource.Id.sharedValueSet = global::pjsua2xamarin.Droid.Resource.Id.sharedValueSet;
			global::Xamarin.Forms.Platform.Resource.Id.sharedValueUnset = global::pjsua2xamarin.Droid.Resource.Id.sharedValueUnset;
			global::Xamarin.Forms.Platform.Resource.Id.shellcontent_appbar = global::pjsua2xamarin.Droid.Resource.Id.shellcontent_appbar;
			global::Xamarin.Forms.Platform.Resource.Id.shellcontent_toolbar = global::pjsua2xamarin.Droid.Resource.Id.shellcontent_toolbar;
			global::Xamarin.Forms.Platform.Resource.Id.SHIFT = global::pjsua2xamarin.Droid.Resource.Id.SHIFT;
			global::Xamarin.Forms.Platform.Resource.Id.shortcut = global::pjsua2xamarin.Droid.Resource.Id.shortcut;
			global::Xamarin.Forms.Platform.Resource.Id.showCustom = global::pjsua2xamarin.Droid.Resource.Id.showCustom;
			global::Xamarin.Forms.Platform.Resource.Id.showHome = global::pjsua2xamarin.Droid.Resource.Id.showHome;
			global::Xamarin.Forms.Platform.Resource.Id.showTitle = global::pjsua2xamarin.Droid.Resource.Id.showTitle;
			global::Xamarin.Forms.Platform.Resource.Id.SHOW_ALL = global::pjsua2xamarin.Droid.Resource.Id.SHOW_ALL;
			global::Xamarin.Forms.Platform.Resource.Id.SHOW_PATH = global::pjsua2xamarin.Droid.Resource.Id.SHOW_PATH;
			global::Xamarin.Forms.Platform.Resource.Id.SHOW_PROGRESS = global::pjsua2xamarin.Droid.Resource.Id.SHOW_PROGRESS;
			global::Xamarin.Forms.Platform.Resource.Id.sin = global::pjsua2xamarin.Droid.Resource.Id.sin;
			global::Xamarin.Forms.Platform.Resource.Id.skipCollapsed = global::pjsua2xamarin.Droid.Resource.Id.skipCollapsed;
			global::Xamarin.Forms.Platform.Resource.Id.skipped = global::pjsua2xamarin.Droid.Resource.Id.skipped;
			global::Xamarin.Forms.Platform.Resource.Id.slide = global::pjsua2xamarin.Droid.Resource.Id.slide;
			global::Xamarin.Forms.Platform.Resource.Id.sliding_tabs = global::pjsua2xamarin.Droid.Resource.Id.sliding_tabs;
			global::Xamarin.Forms.Platform.Resource.Id.snackbar_action = global::pjsua2xamarin.Droid.Resource.Id.snackbar_action;
			global::Xamarin.Forms.Platform.Resource.Id.snackbar_text = global::pjsua2xamarin.Droid.Resource.Id.snackbar_text;
			global::Xamarin.Forms.Platform.Resource.Id.snap = global::pjsua2xamarin.Droid.Resource.Id.snap;
			global::Xamarin.Forms.Platform.Resource.Id.snapMargins = global::pjsua2xamarin.Droid.Resource.Id.snapMargins;
			global::Xamarin.Forms.Platform.Resource.Id.south = global::pjsua2xamarin.Droid.Resource.Id.south;
			global::Xamarin.Forms.Platform.Resource.Id.spacer = global::pjsua2xamarin.Droid.Resource.Id.spacer;
			global::Xamarin.Forms.Platform.Resource.Id.special_effects_controller_view_tag = global::pjsua2xamarin.Droid.Resource.Id.special_effects_controller_view_tag;
			global::Xamarin.Forms.Platform.Resource.Id.spline = global::pjsua2xamarin.Droid.Resource.Id.spline;
			global::Xamarin.Forms.Platform.Resource.Id.split_action_bar = global::pjsua2xamarin.Droid.Resource.Id.split_action_bar;
			global::Xamarin.Forms.Platform.Resource.Id.spread = global::pjsua2xamarin.Droid.Resource.Id.spread;
			global::Xamarin.Forms.Platform.Resource.Id.spread_inside = global::pjsua2xamarin.Droid.Resource.Id.spread_inside;
			global::Xamarin.Forms.Platform.Resource.Id.spring = global::pjsua2xamarin.Droid.Resource.Id.spring;
			global::Xamarin.Forms.Platform.Resource.Id.square = global::pjsua2xamarin.Droid.Resource.Id.square;
			global::Xamarin.Forms.Platform.Resource.Id.src_atop = global::pjsua2xamarin.Droid.Resource.Id.src_atop;
			global::Xamarin.Forms.Platform.Resource.Id.src_in = global::pjsua2xamarin.Droid.Resource.Id.src_in;
			global::Xamarin.Forms.Platform.Resource.Id.src_over = global::pjsua2xamarin.Droid.Resource.Id.src_over;
			global::Xamarin.Forms.Platform.Resource.Id.standard = global::pjsua2xamarin.Droid.Resource.Id.standard;
			global::Xamarin.Forms.Platform.Resource.Id.start = global::pjsua2xamarin.Droid.Resource.Id.start;
			global::Xamarin.Forms.Platform.Resource.Id.startHorizontal = global::pjsua2xamarin.Droid.Resource.Id.startHorizontal;
			global::Xamarin.Forms.Platform.Resource.Id.startToEnd = global::pjsua2xamarin.Droid.Resource.Id.startToEnd;
			global::Xamarin.Forms.Platform.Resource.Id.startVertical = global::pjsua2xamarin.Droid.Resource.Id.startVertical;
			global::Xamarin.Forms.Platform.Resource.Id.staticLayout = global::pjsua2xamarin.Droid.Resource.Id.staticLayout;
			global::Xamarin.Forms.Platform.Resource.Id.staticPostLayout = global::pjsua2xamarin.Droid.Resource.Id.staticPostLayout;
			global::Xamarin.Forms.Platform.Resource.Id.status_bar_latest_event_content = global::pjsua2xamarin.Droid.Resource.Id.status_bar_latest_event_content;
			global::Xamarin.Forms.Platform.Resource.Id.stop = global::pjsua2xamarin.Droid.Resource.Id.stop;
			global::Xamarin.Forms.Platform.Resource.Id.stretch = global::pjsua2xamarin.Droid.Resource.Id.stretch;
			global::Xamarin.Forms.Platform.Resource.Id.submenuarrow = global::pjsua2xamarin.Droid.Resource.Id.submenuarrow;
			global::Xamarin.Forms.Platform.Resource.Id.submit_area = global::pjsua2xamarin.Droid.Resource.Id.submit_area;
			global::Xamarin.Forms.Platform.Resource.Id.supportScrollUp = global::pjsua2xamarin.Droid.Resource.Id.supportScrollUp;
			global::Xamarin.Forms.Platform.Resource.Id.SYM = global::pjsua2xamarin.Droid.Resource.Id.SYM;
			global::Xamarin.Forms.Platform.Resource.Id.tabMode = global::pjsua2xamarin.Droid.Resource.Id.tabMode;
			global::Xamarin.Forms.Platform.Resource.Id.tag_accessibility_actions = global::pjsua2xamarin.Droid.Resource.Id.tag_accessibility_actions;
			global::Xamarin.Forms.Platform.Resource.Id.tag_accessibility_clickable_spans = global::pjsua2xamarin.Droid.Resource.Id.tag_accessibility_clickable_spans;
			global::Xamarin.Forms.Platform.Resource.Id.tag_accessibility_heading = global::pjsua2xamarin.Droid.Resource.Id.tag_accessibility_heading;
			global::Xamarin.Forms.Platform.Resource.Id.tag_accessibility_pane_title = global::pjsua2xamarin.Droid.Resource.Id.tag_accessibility_pane_title;
			global::Xamarin.Forms.Platform.Resource.Id.tag_on_apply_window_listener = global::pjsua2xamarin.Droid.Resource.Id.tag_on_apply_window_listener;
			global::Xamarin.Forms.Platform.Resource.Id.tag_on_receive_content_listener = global::pjsua2xamarin.Droid.Resource.Id.tag_on_receive_content_listener;
			global::Xamarin.Forms.Platform.Resource.Id.tag_on_receive_content_mime_types = global::pjsua2xamarin.Droid.Resource.Id.tag_on_receive_content_mime_types;
			global::Xamarin.Forms.Platform.Resource.Id.tag_screen_reader_focusable = global::pjsua2xamarin.Droid.Resource.Id.tag_screen_reader_focusable;
			global::Xamarin.Forms.Platform.Resource.Id.tag_state_description = global::pjsua2xamarin.Droid.Resource.Id.tag_state_description;
			global::Xamarin.Forms.Platform.Resource.Id.tag_transition_group = global::pjsua2xamarin.Droid.Resource.Id.tag_transition_group;
			global::Xamarin.Forms.Platform.Resource.Id.tag_unhandled_key_event_manager = global::pjsua2xamarin.Droid.Resource.Id.tag_unhandled_key_event_manager;
			global::Xamarin.Forms.Platform.Resource.Id.tag_unhandled_key_listeners = global::pjsua2xamarin.Droid.Resource.Id.tag_unhandled_key_listeners;
			global::Xamarin.Forms.Platform.Resource.Id.tag_window_insets_animation_callback = global::pjsua2xamarin.Droid.Resource.Id.tag_window_insets_animation_callback;
			global::Xamarin.Forms.Platform.Resource.Id.test_checkbox_android_button_tint = global::pjsua2xamarin.Droid.Resource.Id.test_checkbox_android_button_tint;
			global::Xamarin.Forms.Platform.Resource.Id.test_checkbox_app_button_tint = global::pjsua2xamarin.Droid.Resource.Id.test_checkbox_app_button_tint;
			global::Xamarin.Forms.Platform.Resource.Id.test_radiobutton_android_button_tint = global::pjsua2xamarin.Droid.Resource.Id.test_radiobutton_android_button_tint;
			global::Xamarin.Forms.Platform.Resource.Id.test_radiobutton_app_button_tint = global::pjsua2xamarin.Droid.Resource.Id.test_radiobutton_app_button_tint;
			global::Xamarin.Forms.Platform.Resource.Id.text = global::pjsua2xamarin.Droid.Resource.Id.text;
			global::Xamarin.Forms.Platform.Resource.Id.text2 = global::pjsua2xamarin.Droid.Resource.Id.text2;
			global::Xamarin.Forms.Platform.Resource.Id.textEnd = global::pjsua2xamarin.Droid.Resource.Id.textEnd;
			global::Xamarin.Forms.Platform.Resource.Id.textinput_counter = global::pjsua2xamarin.Droid.Resource.Id.textinput_counter;
			global::Xamarin.Forms.Platform.Resource.Id.textinput_error = global::pjsua2xamarin.Droid.Resource.Id.textinput_error;
			global::Xamarin.Forms.Platform.Resource.Id.textinput_helper_text = global::pjsua2xamarin.Droid.Resource.Id.textinput_helper_text;
			global::Xamarin.Forms.Platform.Resource.Id.textinput_placeholder = global::pjsua2xamarin.Droid.Resource.Id.textinput_placeholder;
			global::Xamarin.Forms.Platform.Resource.Id.textinput_prefix_text = global::pjsua2xamarin.Droid.Resource.Id.textinput_prefix_text;
			global::Xamarin.Forms.Platform.Resource.Id.textinput_suffix_text = global::pjsua2xamarin.Droid.Resource.Id.textinput_suffix_text;
			global::Xamarin.Forms.Platform.Resource.Id.textSpacerNoButtons = global::pjsua2xamarin.Droid.Resource.Id.textSpacerNoButtons;
			global::Xamarin.Forms.Platform.Resource.Id.textSpacerNoTitle = global::pjsua2xamarin.Droid.Resource.Id.textSpacerNoTitle;
			global::Xamarin.Forms.Platform.Resource.Id.textStart = global::pjsua2xamarin.Droid.Resource.Id.textStart;
			global::Xamarin.Forms.Platform.Resource.Id.textTop = global::pjsua2xamarin.Droid.Resource.Id.textTop;
			global::Xamarin.Forms.Platform.Resource.Id.text_input_end_icon = global::pjsua2xamarin.Droid.Resource.Id.text_input_end_icon;
			global::Xamarin.Forms.Platform.Resource.Id.text_input_error_icon = global::pjsua2xamarin.Droid.Resource.Id.text_input_error_icon;
			global::Xamarin.Forms.Platform.Resource.Id.text_input_start_icon = global::pjsua2xamarin.Droid.Resource.Id.text_input_start_icon;
			global::Xamarin.Forms.Platform.Resource.Id.time = global::pjsua2xamarin.Droid.Resource.Id.time;
			global::Xamarin.Forms.Platform.Resource.Id.title = global::pjsua2xamarin.Droid.Resource.Id.title;
			global::Xamarin.Forms.Platform.Resource.Id.titleDividerNoCustom = global::pjsua2xamarin.Droid.Resource.Id.titleDividerNoCustom;
			global::Xamarin.Forms.Platform.Resource.Id.title_template = global::pjsua2xamarin.Droid.Resource.Id.title_template;
			global::Xamarin.Forms.Platform.Resource.Id.toggle = global::pjsua2xamarin.Droid.Resource.Id.toggle;
			global::Xamarin.Forms.Platform.Resource.Id.toolbar = global::pjsua2xamarin.Droid.Resource.Id.toolbar;
			global::Xamarin.Forms.Platform.Resource.Id.top = global::pjsua2xamarin.Droid.Resource.Id.top;
			global::Xamarin.Forms.Platform.Resource.Id.topPanel = global::pjsua2xamarin.Droid.Resource.Id.topPanel;
			global::Xamarin.Forms.Platform.Resource.Id.TOP_END = global::pjsua2xamarin.Droid.Resource.Id.TOP_END;
			global::Xamarin.Forms.Platform.Resource.Id.TOP_START = global::pjsua2xamarin.Droid.Resource.Id.TOP_START;
			global::Xamarin.Forms.Platform.Resource.Id.touch_outside = global::pjsua2xamarin.Droid.Resource.Id.touch_outside;
			global::Xamarin.Forms.Platform.Resource.Id.transitionToEnd = global::pjsua2xamarin.Droid.Resource.Id.transitionToEnd;
			global::Xamarin.Forms.Platform.Resource.Id.transitionToStart = global::pjsua2xamarin.Droid.Resource.Id.transitionToStart;
			global::Xamarin.Forms.Platform.Resource.Id.transition_current_scene = global::pjsua2xamarin.Droid.Resource.Id.transition_current_scene;
			global::Xamarin.Forms.Platform.Resource.Id.transition_layout_save = global::pjsua2xamarin.Droid.Resource.Id.transition_layout_save;
			global::Xamarin.Forms.Platform.Resource.Id.transition_position = global::pjsua2xamarin.Droid.Resource.Id.transition_position;
			global::Xamarin.Forms.Platform.Resource.Id.transition_scene_layoutid_cache = global::pjsua2xamarin.Droid.Resource.Id.transition_scene_layoutid_cache;
			global::Xamarin.Forms.Platform.Resource.Id.transition_transform = global::pjsua2xamarin.Droid.Resource.Id.transition_transform;
			global::Xamarin.Forms.Platform.Resource.Id.triangle = global::pjsua2xamarin.Droid.Resource.Id.triangle;
			global::Xamarin.Forms.Platform.Resource.Id.@unchecked = global::pjsua2xamarin.Droid.Resource.Id.@unchecked;
			global::Xamarin.Forms.Platform.Resource.Id.uniform = global::pjsua2xamarin.Droid.Resource.Id.uniform;
			global::Xamarin.Forms.Platform.Resource.Id.unlabeled = global::pjsua2xamarin.Droid.Resource.Id.unlabeled;
			global::Xamarin.Forms.Platform.Resource.Id.up = global::pjsua2xamarin.Droid.Resource.Id.up;
			global::Xamarin.Forms.Platform.Resource.Id.useLogo = global::pjsua2xamarin.Droid.Resource.Id.useLogo;
			global::Xamarin.Forms.Platform.Resource.Id.vertical_only = global::pjsua2xamarin.Droid.Resource.Id.vertical_only;
			global::Xamarin.Forms.Platform.Resource.Id.view_offset_helper = global::pjsua2xamarin.Droid.Resource.Id.view_offset_helper;
			global::Xamarin.Forms.Platform.Resource.Id.view_transition = global::pjsua2xamarin.Droid.Resource.Id.view_transition;
			global::Xamarin.Forms.Platform.Resource.Id.view_tree_lifecycle_owner = global::pjsua2xamarin.Droid.Resource.Id.view_tree_lifecycle_owner;
			global::Xamarin.Forms.Platform.Resource.Id.view_tree_saved_state_registry_owner = global::pjsua2xamarin.Droid.Resource.Id.view_tree_saved_state_registry_owner;
			global::Xamarin.Forms.Platform.Resource.Id.view_tree_view_model_store_owner = global::pjsua2xamarin.Droid.Resource.Id.view_tree_view_model_store_owner;
			global::Xamarin.Forms.Platform.Resource.Id.visible = global::pjsua2xamarin.Droid.Resource.Id.visible;
			global::Xamarin.Forms.Platform.Resource.Id.visible_removing_fragment_view_tag = global::pjsua2xamarin.Droid.Resource.Id.visible_removing_fragment_view_tag;
			global::Xamarin.Forms.Platform.Resource.Id.west = global::pjsua2xamarin.Droid.Resource.Id.west;
			global::Xamarin.Forms.Platform.Resource.Id.withinBounds = global::pjsua2xamarin.Droid.Resource.Id.withinBounds;
			global::Xamarin.Forms.Platform.Resource.Id.withText = global::pjsua2xamarin.Droid.Resource.Id.withText;
			global::Xamarin.Forms.Platform.Resource.Id.wrap = global::pjsua2xamarin.Droid.Resource.Id.wrap;
			global::Xamarin.Forms.Platform.Resource.Id.wrap_content = global::pjsua2xamarin.Droid.Resource.Id.wrap_content;
			global::Xamarin.Forms.Platform.Resource.Id.wrap_content_constrained = global::pjsua2xamarin.Droid.Resource.Id.wrap_content_constrained;
			global::Xamarin.Forms.Platform.Resource.Id.x_left = global::pjsua2xamarin.Droid.Resource.Id.x_left;
			global::Xamarin.Forms.Platform.Resource.Id.x_right = global::pjsua2xamarin.Droid.Resource.Id.x_right;
			global::Xamarin.Forms.Platform.Resource.Id.zero_corner_chip = global::pjsua2xamarin.Droid.Resource.Id.zero_corner_chip;
			global::Xamarin.Forms.Platform.Resource.Integer.abc_config_activityDefaultDur = global::pjsua2xamarin.Droid.Resource.Integer.abc_config_activityDefaultDur;
			global::Xamarin.Forms.Platform.Resource.Integer.abc_config_activityShortDur = global::pjsua2xamarin.Droid.Resource.Integer.abc_config_activityShortDur;
			global::Xamarin.Forms.Platform.Resource.Integer.app_bar_elevation_anim_duration = global::pjsua2xamarin.Droid.Resource.Integer.app_bar_elevation_anim_duration;
			global::Xamarin.Forms.Platform.Resource.Integer.bottom_sheet_slide_duration = global::pjsua2xamarin.Droid.Resource.Integer.bottom_sheet_slide_duration;
			global::Xamarin.Forms.Platform.Resource.Integer.cancel_button_image_alpha = global::pjsua2xamarin.Droid.Resource.Integer.cancel_button_image_alpha;
			global::Xamarin.Forms.Platform.Resource.Integer.config_navAnimTime = global::pjsua2xamarin.Droid.Resource.Integer.config_navAnimTime;
			global::Xamarin.Forms.Platform.Resource.Integer.config_tooltipAnimTime = global::pjsua2xamarin.Droid.Resource.Integer.config_tooltipAnimTime;
			global::Xamarin.Forms.Platform.Resource.Integer.design_snackbar_text_max_lines = global::pjsua2xamarin.Droid.Resource.Integer.design_snackbar_text_max_lines;
			global::Xamarin.Forms.Platform.Resource.Integer.design_tab_indicator_anim_duration_ms = global::pjsua2xamarin.Droid.Resource.Integer.design_tab_indicator_anim_duration_ms;
			global::Xamarin.Forms.Platform.Resource.Integer.hide_password_duration = global::pjsua2xamarin.Droid.Resource.Integer.hide_password_duration;
			global::Xamarin.Forms.Platform.Resource.Integer.material_motion_duration_long_1 = global::pjsua2xamarin.Droid.Resource.Integer.material_motion_duration_long_1;
			global::Xamarin.Forms.Platform.Resource.Integer.material_motion_duration_long_2 = global::pjsua2xamarin.Droid.Resource.Integer.material_motion_duration_long_2;
			global::Xamarin.Forms.Platform.Resource.Integer.material_motion_duration_medium_1 = global::pjsua2xamarin.Droid.Resource.Integer.material_motion_duration_medium_1;
			global::Xamarin.Forms.Platform.Resource.Integer.material_motion_duration_medium_2 = global::pjsua2xamarin.Droid.Resource.Integer.material_motion_duration_medium_2;
			global::Xamarin.Forms.Platform.Resource.Integer.material_motion_duration_short_1 = global::pjsua2xamarin.Droid.Resource.Integer.material_motion_duration_short_1;
			global::Xamarin.Forms.Platform.Resource.Integer.material_motion_duration_short_2 = global::pjsua2xamarin.Droid.Resource.Integer.material_motion_duration_short_2;
			global::Xamarin.Forms.Platform.Resource.Integer.material_motion_path = global::pjsua2xamarin.Droid.Resource.Integer.material_motion_path;
			global::Xamarin.Forms.Platform.Resource.Integer.mtrl_badge_max_character_count = global::pjsua2xamarin.Droid.Resource.Integer.mtrl_badge_max_character_count;
			global::Xamarin.Forms.Platform.Resource.Integer.mtrl_btn_anim_delay_ms = global::pjsua2xamarin.Droid.Resource.Integer.mtrl_btn_anim_delay_ms;
			global::Xamarin.Forms.Platform.Resource.Integer.mtrl_btn_anim_duration_ms = global::pjsua2xamarin.Droid.Resource.Integer.mtrl_btn_anim_duration_ms;
			global::Xamarin.Forms.Platform.Resource.Integer.mtrl_calendar_header_orientation = global::pjsua2xamarin.Droid.Resource.Integer.mtrl_calendar_header_orientation;
			global::Xamarin.Forms.Platform.Resource.Integer.mtrl_calendar_selection_text_lines = global::pjsua2xamarin.Droid.Resource.Integer.mtrl_calendar_selection_text_lines;
			global::Xamarin.Forms.Platform.Resource.Integer.mtrl_calendar_year_selector_span = global::pjsua2xamarin.Droid.Resource.Integer.mtrl_calendar_year_selector_span;
			global::Xamarin.Forms.Platform.Resource.Integer.mtrl_card_anim_delay_ms = global::pjsua2xamarin.Droid.Resource.Integer.mtrl_card_anim_delay_ms;
			global::Xamarin.Forms.Platform.Resource.Integer.mtrl_card_anim_duration_ms = global::pjsua2xamarin.Droid.Resource.Integer.mtrl_card_anim_duration_ms;
			global::Xamarin.Forms.Platform.Resource.Integer.mtrl_chip_anim_duration = global::pjsua2xamarin.Droid.Resource.Integer.mtrl_chip_anim_duration;
			global::Xamarin.Forms.Platform.Resource.Integer.mtrl_tab_indicator_anim_duration_ms = global::pjsua2xamarin.Droid.Resource.Integer.mtrl_tab_indicator_anim_duration_ms;
			global::Xamarin.Forms.Platform.Resource.Integer.show_password_duration = global::pjsua2xamarin.Droid.Resource.Integer.show_password_duration;
			global::Xamarin.Forms.Platform.Resource.Integer.status_bar_notification_info_maxnum = global::pjsua2xamarin.Droid.Resource.Integer.status_bar_notification_info_maxnum;
			global::Xamarin.Forms.Platform.Resource.Interpolator.btn_checkbox_checked_mtrl_animation_interpolator_0 = global::pjsua2xamarin.Droid.Resource.Interpolator.btn_checkbox_checked_mtrl_animation_interpolator_0;
			global::Xamarin.Forms.Platform.Resource.Interpolator.btn_checkbox_checked_mtrl_animation_interpolator_1 = global::pjsua2xamarin.Droid.Resource.Interpolator.btn_checkbox_checked_mtrl_animation_interpolator_1;
			global::Xamarin.Forms.Platform.Resource.Interpolator.btn_checkbox_unchecked_mtrl_animation_interpolator_0 = global::pjsua2xamarin.Droid.Resource.Interpolator.btn_checkbox_unchecked_mtrl_animation_interpolator_0;
			global::Xamarin.Forms.Platform.Resource.Interpolator.btn_checkbox_unchecked_mtrl_animation_interpolator_1 = global::pjsua2xamarin.Droid.Resource.Interpolator.btn_checkbox_unchecked_mtrl_animation_interpolator_1;
			global::Xamarin.Forms.Platform.Resource.Interpolator.btn_radio_to_off_mtrl_animation_interpolator_0 = global::pjsua2xamarin.Droid.Resource.Interpolator.btn_radio_to_off_mtrl_animation_interpolator_0;
			global::Xamarin.Forms.Platform.Resource.Interpolator.btn_radio_to_on_mtrl_animation_interpolator_0 = global::pjsua2xamarin.Droid.Resource.Interpolator.btn_radio_to_on_mtrl_animation_interpolator_0;
			global::Xamarin.Forms.Platform.Resource.Interpolator.fast_out_slow_in = global::pjsua2xamarin.Droid.Resource.Interpolator.fast_out_slow_in;
			global::Xamarin.Forms.Platform.Resource.Interpolator.mtrl_fast_out_linear_in = global::pjsua2xamarin.Droid.Resource.Interpolator.mtrl_fast_out_linear_in;
			global::Xamarin.Forms.Platform.Resource.Interpolator.mtrl_fast_out_slow_in = global::pjsua2xamarin.Droid.Resource.Interpolator.mtrl_fast_out_slow_in;
			global::Xamarin.Forms.Platform.Resource.Interpolator.mtrl_linear = global::pjsua2xamarin.Droid.Resource.Interpolator.mtrl_linear;
			global::Xamarin.Forms.Platform.Resource.Interpolator.mtrl_linear_out_slow_in = global::pjsua2xamarin.Droid.Resource.Interpolator.mtrl_linear_out_slow_in;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_action_bar_title_item = global::pjsua2xamarin.Droid.Resource.Layout.abc_action_bar_title_item;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_action_bar_up_container = global::pjsua2xamarin.Droid.Resource.Layout.abc_action_bar_up_container;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_action_menu_item_layout = global::pjsua2xamarin.Droid.Resource.Layout.abc_action_menu_item_layout;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_action_menu_layout = global::pjsua2xamarin.Droid.Resource.Layout.abc_action_menu_layout;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_action_mode_bar = global::pjsua2xamarin.Droid.Resource.Layout.abc_action_mode_bar;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_action_mode_close_item_material = global::pjsua2xamarin.Droid.Resource.Layout.abc_action_mode_close_item_material;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_activity_chooser_view = global::pjsua2xamarin.Droid.Resource.Layout.abc_activity_chooser_view;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_activity_chooser_view_list_item = global::pjsua2xamarin.Droid.Resource.Layout.abc_activity_chooser_view_list_item;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_alert_dialog_button_bar_material = global::pjsua2xamarin.Droid.Resource.Layout.abc_alert_dialog_button_bar_material;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_alert_dialog_material = global::pjsua2xamarin.Droid.Resource.Layout.abc_alert_dialog_material;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_alert_dialog_title_material = global::pjsua2xamarin.Droid.Resource.Layout.abc_alert_dialog_title_material;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_cascading_menu_item_layout = global::pjsua2xamarin.Droid.Resource.Layout.abc_cascading_menu_item_layout;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_dialog_title_material = global::pjsua2xamarin.Droid.Resource.Layout.abc_dialog_title_material;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_expanded_menu_layout = global::pjsua2xamarin.Droid.Resource.Layout.abc_expanded_menu_layout;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_list_menu_item_checkbox = global::pjsua2xamarin.Droid.Resource.Layout.abc_list_menu_item_checkbox;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_list_menu_item_icon = global::pjsua2xamarin.Droid.Resource.Layout.abc_list_menu_item_icon;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_list_menu_item_layout = global::pjsua2xamarin.Droid.Resource.Layout.abc_list_menu_item_layout;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_list_menu_item_radio = global::pjsua2xamarin.Droid.Resource.Layout.abc_list_menu_item_radio;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_popup_menu_header_item_layout = global::pjsua2xamarin.Droid.Resource.Layout.abc_popup_menu_header_item_layout;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_popup_menu_item_layout = global::pjsua2xamarin.Droid.Resource.Layout.abc_popup_menu_item_layout;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_screen_content_include = global::pjsua2xamarin.Droid.Resource.Layout.abc_screen_content_include;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_screen_simple = global::pjsua2xamarin.Droid.Resource.Layout.abc_screen_simple;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_screen_simple_overlay_action_mode = global::pjsua2xamarin.Droid.Resource.Layout.abc_screen_simple_overlay_action_mode;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_screen_toolbar = global::pjsua2xamarin.Droid.Resource.Layout.abc_screen_toolbar;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_search_dropdown_item_icons_2line = global::pjsua2xamarin.Droid.Resource.Layout.abc_search_dropdown_item_icons_2line;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_search_view = global::pjsua2xamarin.Droid.Resource.Layout.abc_search_view;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_select_dialog_material = global::pjsua2xamarin.Droid.Resource.Layout.abc_select_dialog_material;
			global::Xamarin.Forms.Platform.Resource.Layout.abc_tooltip = global::pjsua2xamarin.Droid.Resource.Layout.abc_tooltip;
			global::Xamarin.Forms.Platform.Resource.Layout.BottomTabLayout = global::pjsua2xamarin.Droid.Resource.Layout.BottomTabLayout;
			global::Xamarin.Forms.Platform.Resource.Layout.custom_dialog = global::pjsua2xamarin.Droid.Resource.Layout.custom_dialog;
			global::Xamarin.Forms.Platform.Resource.Layout.design_bottom_navigation_item = global::pjsua2xamarin.Droid.Resource.Layout.design_bottom_navigation_item;
			global::Xamarin.Forms.Platform.Resource.Layout.design_bottom_sheet_dialog = global::pjsua2xamarin.Droid.Resource.Layout.design_bottom_sheet_dialog;
			global::Xamarin.Forms.Platform.Resource.Layout.design_layout_snackbar = global::pjsua2xamarin.Droid.Resource.Layout.design_layout_snackbar;
			global::Xamarin.Forms.Platform.Resource.Layout.design_layout_snackbar_include = global::pjsua2xamarin.Droid.Resource.Layout.design_layout_snackbar_include;
			global::Xamarin.Forms.Platform.Resource.Layout.design_layout_tab_icon = global::pjsua2xamarin.Droid.Resource.Layout.design_layout_tab_icon;
			global::Xamarin.Forms.Platform.Resource.Layout.design_layout_tab_text = global::pjsua2xamarin.Droid.Resource.Layout.design_layout_tab_text;
			global::Xamarin.Forms.Platform.Resource.Layout.design_menu_item_action_area = global::pjsua2xamarin.Droid.Resource.Layout.design_menu_item_action_area;
			global::Xamarin.Forms.Platform.Resource.Layout.design_navigation_item = global::pjsua2xamarin.Droid.Resource.Layout.design_navigation_item;
			global::Xamarin.Forms.Platform.Resource.Layout.design_navigation_item_header = global::pjsua2xamarin.Droid.Resource.Layout.design_navigation_item_header;
			global::Xamarin.Forms.Platform.Resource.Layout.design_navigation_item_separator = global::pjsua2xamarin.Droid.Resource.Layout.design_navigation_item_separator;
			global::Xamarin.Forms.Platform.Resource.Layout.design_navigation_item_subheader = global::pjsua2xamarin.Droid.Resource.Layout.design_navigation_item_subheader;
			global::Xamarin.Forms.Platform.Resource.Layout.design_navigation_menu = global::pjsua2xamarin.Droid.Resource.Layout.design_navigation_menu;
			global::Xamarin.Forms.Platform.Resource.Layout.design_navigation_menu_item = global::pjsua2xamarin.Droid.Resource.Layout.design_navigation_menu_item;
			global::Xamarin.Forms.Platform.Resource.Layout.design_text_input_end_icon = global::pjsua2xamarin.Droid.Resource.Layout.design_text_input_end_icon;
			global::Xamarin.Forms.Platform.Resource.Layout.design_text_input_start_icon = global::pjsua2xamarin.Droid.Resource.Layout.design_text_input_start_icon;
			global::Xamarin.Forms.Platform.Resource.Layout.FallbackTabbarDoNotUse = global::pjsua2xamarin.Droid.Resource.Layout.FallbackTabbarDoNotUse;
			global::Xamarin.Forms.Platform.Resource.Layout.FallbackToolbarDoNotUse = global::pjsua2xamarin.Droid.Resource.Layout.FallbackToolbarDoNotUse;
			global::Xamarin.Forms.Platform.Resource.Layout.FlyoutContent = global::pjsua2xamarin.Droid.Resource.Layout.FlyoutContent;
			global::Xamarin.Forms.Platform.Resource.Layout.material_chip_input_combo = global::pjsua2xamarin.Droid.Resource.Layout.material_chip_input_combo;
			global::Xamarin.Forms.Platform.Resource.Layout.material_clockface_textview = global::pjsua2xamarin.Droid.Resource.Layout.material_clockface_textview;
			global::Xamarin.Forms.Platform.Resource.Layout.material_clockface_view = global::pjsua2xamarin.Droid.Resource.Layout.material_clockface_view;
			global::Xamarin.Forms.Platform.Resource.Layout.material_clock_display = global::pjsua2xamarin.Droid.Resource.Layout.material_clock_display;
			global::Xamarin.Forms.Platform.Resource.Layout.material_clock_display_divider = global::pjsua2xamarin.Droid.Resource.Layout.material_clock_display_divider;
			global::Xamarin.Forms.Platform.Resource.Layout.material_clock_period_toggle = global::pjsua2xamarin.Droid.Resource.Layout.material_clock_period_toggle;
			global::Xamarin.Forms.Platform.Resource.Layout.material_clock_period_toggle_land = global::pjsua2xamarin.Droid.Resource.Layout.material_clock_period_toggle_land;
			global::Xamarin.Forms.Platform.Resource.Layout.material_radial_view_group = global::pjsua2xamarin.Droid.Resource.Layout.material_radial_view_group;
			global::Xamarin.Forms.Platform.Resource.Layout.material_textinput_timepicker = global::pjsua2xamarin.Droid.Resource.Layout.material_textinput_timepicker;
			global::Xamarin.Forms.Platform.Resource.Layout.material_timepicker = global::pjsua2xamarin.Droid.Resource.Layout.material_timepicker;
			global::Xamarin.Forms.Platform.Resource.Layout.material_timepicker_dialog = global::pjsua2xamarin.Droid.Resource.Layout.material_timepicker_dialog;
			global::Xamarin.Forms.Platform.Resource.Layout.material_timepicker_textinput_display = global::pjsua2xamarin.Droid.Resource.Layout.material_timepicker_textinput_display;
			global::Xamarin.Forms.Platform.Resource.Layout.material_time_chip = global::pjsua2xamarin.Droid.Resource.Layout.material_time_chip;
			global::Xamarin.Forms.Platform.Resource.Layout.material_time_input = global::pjsua2xamarin.Droid.Resource.Layout.material_time_input;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_alert_dialog = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_alert_dialog;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_alert_dialog_actions = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_alert_dialog_actions;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_alert_dialog_title = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_alert_dialog_title;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_alert_select_dialog_item = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_alert_select_dialog_item;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_alert_select_dialog_multichoice = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_alert_select_dialog_multichoice;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_alert_select_dialog_singlechoice = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_alert_select_dialog_singlechoice;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_calendar_day = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_calendar_day;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_calendar_days_of_week = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_calendar_days_of_week;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_calendar_day_of_week = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_calendar_day_of_week;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_calendar_horizontal = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_calendar_horizontal;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_calendar_month = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_calendar_month;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_calendar_months = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_calendar_months;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_calendar_month_labeled = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_calendar_month_labeled;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_calendar_month_navigation = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_calendar_month_navigation;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_calendar_vertical = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_calendar_vertical;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_calendar_year = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_calendar_year;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_layout_snackbar = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_layout_snackbar;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_layout_snackbar_include = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_layout_snackbar_include;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_navigation_rail_item = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_navigation_rail_item;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_picker_actions = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_picker_actions;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_picker_dialog = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_picker_dialog;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_picker_fullscreen = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_picker_fullscreen;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_picker_header_dialog = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_picker_header_dialog;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_picker_header_fullscreen = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_picker_header_fullscreen;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_picker_header_selection_text = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_picker_header_selection_text;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_picker_header_title_text = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_picker_header_title_text;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_picker_header_toggle = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_picker_header_toggle;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_picker_text_input_date = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_picker_text_input_date;
			global::Xamarin.Forms.Platform.Resource.Layout.mtrl_picker_text_input_date_range = global::pjsua2xamarin.Droid.Resource.Layout.mtrl_picker_text_input_date_range;
			global::Xamarin.Forms.Platform.Resource.Layout.notification_action = global::pjsua2xamarin.Droid.Resource.Layout.notification_action;
			global::Xamarin.Forms.Platform.Resource.Layout.notification_action_tombstone = global::pjsua2xamarin.Droid.Resource.Layout.notification_action_tombstone;
			global::Xamarin.Forms.Platform.Resource.Layout.notification_media_action = global::pjsua2xamarin.Droid.Resource.Layout.notification_media_action;
			global::Xamarin.Forms.Platform.Resource.Layout.notification_media_cancel_action = global::pjsua2xamarin.Droid.Resource.Layout.notification_media_cancel_action;
			global::Xamarin.Forms.Platform.Resource.Layout.notification_template_big_media = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_big_media;
			global::Xamarin.Forms.Platform.Resource.Layout.notification_template_big_media_custom = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_big_media_custom;
			global::Xamarin.Forms.Platform.Resource.Layout.notification_template_big_media_narrow = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_big_media_narrow;
			global::Xamarin.Forms.Platform.Resource.Layout.notification_template_big_media_narrow_custom = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_big_media_narrow_custom;
			global::Xamarin.Forms.Platform.Resource.Layout.notification_template_custom_big = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_custom_big;
			global::Xamarin.Forms.Platform.Resource.Layout.notification_template_icon_group = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_icon_group;
			global::Xamarin.Forms.Platform.Resource.Layout.notification_template_lines_media = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_lines_media;
			global::Xamarin.Forms.Platform.Resource.Layout.notification_template_media = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_media;
			global::Xamarin.Forms.Platform.Resource.Layout.notification_template_media_custom = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_media_custom;
			global::Xamarin.Forms.Platform.Resource.Layout.notification_template_part_chronometer = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_part_chronometer;
			global::Xamarin.Forms.Platform.Resource.Layout.notification_template_part_time = global::pjsua2xamarin.Droid.Resource.Layout.notification_template_part_time;
			global::Xamarin.Forms.Platform.Resource.Layout.RootLayout = global::pjsua2xamarin.Droid.Resource.Layout.RootLayout;
			global::Xamarin.Forms.Platform.Resource.Layout.select_dialog_item_material = global::pjsua2xamarin.Droid.Resource.Layout.select_dialog_item_material;
			global::Xamarin.Forms.Platform.Resource.Layout.select_dialog_multichoice_material = global::pjsua2xamarin.Droid.Resource.Layout.select_dialog_multichoice_material;
			global::Xamarin.Forms.Platform.Resource.Layout.select_dialog_singlechoice_material = global::pjsua2xamarin.Droid.Resource.Layout.select_dialog_singlechoice_material;
			global::Xamarin.Forms.Platform.Resource.Layout.ShellContent = global::pjsua2xamarin.Droid.Resource.Layout.ShellContent;
			global::Xamarin.Forms.Platform.Resource.Layout.support_simple_spinner_dropdown_item = global::pjsua2xamarin.Droid.Resource.Layout.support_simple_spinner_dropdown_item;
			global::Xamarin.Forms.Platform.Resource.Layout.Tabbar = global::pjsua2xamarin.Droid.Resource.Layout.Tabbar;
			global::Xamarin.Forms.Platform.Resource.Layout.test_action_chip = global::pjsua2xamarin.Droid.Resource.Layout.test_action_chip;
			global::Xamarin.Forms.Platform.Resource.Layout.test_chip_zero_corner_radius = global::pjsua2xamarin.Droid.Resource.Layout.test_chip_zero_corner_radius;
			global::Xamarin.Forms.Platform.Resource.Layout.test_design_checkbox = global::pjsua2xamarin.Droid.Resource.Layout.test_design_checkbox;
			global::Xamarin.Forms.Platform.Resource.Layout.test_design_radiobutton = global::pjsua2xamarin.Droid.Resource.Layout.test_design_radiobutton;
			global::Xamarin.Forms.Platform.Resource.Layout.test_navigation_bar_item_layout = global::pjsua2xamarin.Droid.Resource.Layout.test_navigation_bar_item_layout;
			global::Xamarin.Forms.Platform.Resource.Layout.test_reflow_chipgroup = global::pjsua2xamarin.Droid.Resource.Layout.test_reflow_chipgroup;
			global::Xamarin.Forms.Platform.Resource.Layout.test_toolbar = global::pjsua2xamarin.Droid.Resource.Layout.test_toolbar;
			global::Xamarin.Forms.Platform.Resource.Layout.test_toolbar_custom_background = global::pjsua2xamarin.Droid.Resource.Layout.test_toolbar_custom_background;
			global::Xamarin.Forms.Platform.Resource.Layout.test_toolbar_elevation = global::pjsua2xamarin.Droid.Resource.Layout.test_toolbar_elevation;
			global::Xamarin.Forms.Platform.Resource.Layout.test_toolbar_surface = global::pjsua2xamarin.Droid.Resource.Layout.test_toolbar_surface;
			global::Xamarin.Forms.Platform.Resource.Layout.text_view_without_line_height = global::pjsua2xamarin.Droid.Resource.Layout.text_view_without_line_height;
			global::Xamarin.Forms.Platform.Resource.Layout.text_view_with_line_height_from_appearance = global::pjsua2xamarin.Droid.Resource.Layout.text_view_with_line_height_from_appearance;
			global::Xamarin.Forms.Platform.Resource.Layout.text_view_with_line_height_from_layout = global::pjsua2xamarin.Droid.Resource.Layout.text_view_with_line_height_from_layout;
			global::Xamarin.Forms.Platform.Resource.Layout.text_view_with_line_height_from_style = global::pjsua2xamarin.Droid.Resource.Layout.text_view_with_line_height_from_style;
			global::Xamarin.Forms.Platform.Resource.Layout.text_view_with_theme_line_height = global::pjsua2xamarin.Droid.Resource.Layout.text_view_with_theme_line_height;
			global::Xamarin.Forms.Platform.Resource.Layout.Toolbar = global::pjsua2xamarin.Droid.Resource.Layout.Toolbar;
			global::Xamarin.Forms.Platform.Resource.Plurals.mtrl_badge_content_description = global::pjsua2xamarin.Droid.Resource.Plurals.mtrl_badge_content_description;
			global::Xamarin.Forms.Platform.Resource.String.abc_action_bar_home_description = global::pjsua2xamarin.Droid.Resource.String.abc_action_bar_home_description;
			global::Xamarin.Forms.Platform.Resource.String.abc_action_bar_up_description = global::pjsua2xamarin.Droid.Resource.String.abc_action_bar_up_description;
			global::Xamarin.Forms.Platform.Resource.String.abc_action_menu_overflow_description = global::pjsua2xamarin.Droid.Resource.String.abc_action_menu_overflow_description;
			global::Xamarin.Forms.Platform.Resource.String.abc_action_mode_done = global::pjsua2xamarin.Droid.Resource.String.abc_action_mode_done;
			global::Xamarin.Forms.Platform.Resource.String.abc_activitychooserview_choose_application = global::pjsua2xamarin.Droid.Resource.String.abc_activitychooserview_choose_application;
			global::Xamarin.Forms.Platform.Resource.String.abc_activity_chooser_view_see_all = global::pjsua2xamarin.Droid.Resource.String.abc_activity_chooser_view_see_all;
			global::Xamarin.Forms.Platform.Resource.String.abc_capital_off = global::pjsua2xamarin.Droid.Resource.String.abc_capital_off;
			global::Xamarin.Forms.Platform.Resource.String.abc_capital_on = global::pjsua2xamarin.Droid.Resource.String.abc_capital_on;
			global::Xamarin.Forms.Platform.Resource.String.abc_menu_alt_shortcut_label = global::pjsua2xamarin.Droid.Resource.String.abc_menu_alt_shortcut_label;
			global::Xamarin.Forms.Platform.Resource.String.abc_menu_ctrl_shortcut_label = global::pjsua2xamarin.Droid.Resource.String.abc_menu_ctrl_shortcut_label;
			global::Xamarin.Forms.Platform.Resource.String.abc_menu_delete_shortcut_label = global::pjsua2xamarin.Droid.Resource.String.abc_menu_delete_shortcut_label;
			global::Xamarin.Forms.Platform.Resource.String.abc_menu_enter_shortcut_label = global::pjsua2xamarin.Droid.Resource.String.abc_menu_enter_shortcut_label;
			global::Xamarin.Forms.Platform.Resource.String.abc_menu_function_shortcut_label = global::pjsua2xamarin.Droid.Resource.String.abc_menu_function_shortcut_label;
			global::Xamarin.Forms.Platform.Resource.String.abc_menu_meta_shortcut_label = global::pjsua2xamarin.Droid.Resource.String.abc_menu_meta_shortcut_label;
			global::Xamarin.Forms.Platform.Resource.String.abc_menu_shift_shortcut_label = global::pjsua2xamarin.Droid.Resource.String.abc_menu_shift_shortcut_label;
			global::Xamarin.Forms.Platform.Resource.String.abc_menu_space_shortcut_label = global::pjsua2xamarin.Droid.Resource.String.abc_menu_space_shortcut_label;
			global::Xamarin.Forms.Platform.Resource.String.abc_menu_sym_shortcut_label = global::pjsua2xamarin.Droid.Resource.String.abc_menu_sym_shortcut_label;
			global::Xamarin.Forms.Platform.Resource.String.abc_prepend_shortcut_label = global::pjsua2xamarin.Droid.Resource.String.abc_prepend_shortcut_label;
			global::Xamarin.Forms.Platform.Resource.String.abc_searchview_description_clear = global::pjsua2xamarin.Droid.Resource.String.abc_searchview_description_clear;
			global::Xamarin.Forms.Platform.Resource.String.abc_searchview_description_query = global::pjsua2xamarin.Droid.Resource.String.abc_searchview_description_query;
			global::Xamarin.Forms.Platform.Resource.String.abc_searchview_description_search = global::pjsua2xamarin.Droid.Resource.String.abc_searchview_description_search;
			global::Xamarin.Forms.Platform.Resource.String.abc_searchview_description_submit = global::pjsua2xamarin.Droid.Resource.String.abc_searchview_description_submit;
			global::Xamarin.Forms.Platform.Resource.String.abc_searchview_description_voice = global::pjsua2xamarin.Droid.Resource.String.abc_searchview_description_voice;
			global::Xamarin.Forms.Platform.Resource.String.abc_search_hint = global::pjsua2xamarin.Droid.Resource.String.abc_search_hint;
			global::Xamarin.Forms.Platform.Resource.String.abc_shareactionprovider_share_with = global::pjsua2xamarin.Droid.Resource.String.abc_shareactionprovider_share_with;
			global::Xamarin.Forms.Platform.Resource.String.abc_shareactionprovider_share_with_application = global::pjsua2xamarin.Droid.Resource.String.abc_shareactionprovider_share_with_application;
			global::Xamarin.Forms.Platform.Resource.String.abc_toolbar_collapse_description = global::pjsua2xamarin.Droid.Resource.String.abc_toolbar_collapse_description;
			global::Xamarin.Forms.Platform.Resource.String.appbar_scrolling_view_behavior = global::pjsua2xamarin.Droid.Resource.String.appbar_scrolling_view_behavior;
			global::Xamarin.Forms.Platform.Resource.String.bottomsheet_action_expand_halfway = global::pjsua2xamarin.Droid.Resource.String.bottomsheet_action_expand_halfway;
			global::Xamarin.Forms.Platform.Resource.String.bottom_sheet_behavior = global::pjsua2xamarin.Droid.Resource.String.bottom_sheet_behavior;
			global::Xamarin.Forms.Platform.Resource.String.character_counter_content_description = global::pjsua2xamarin.Droid.Resource.String.character_counter_content_description;
			global::Xamarin.Forms.Platform.Resource.String.character_counter_overflowed_content_description = global::pjsua2xamarin.Droid.Resource.String.character_counter_overflowed_content_description;
			global::Xamarin.Forms.Platform.Resource.String.character_counter_pattern = global::pjsua2xamarin.Droid.Resource.String.character_counter_pattern;
			global::Xamarin.Forms.Platform.Resource.String.chip_text = global::pjsua2xamarin.Droid.Resource.String.chip_text;
			global::Xamarin.Forms.Platform.Resource.String.clear_text_end_icon_content_description = global::pjsua2xamarin.Droid.Resource.String.clear_text_end_icon_content_description;
			global::Xamarin.Forms.Platform.Resource.String.error_icon_content_description = global::pjsua2xamarin.Droid.Resource.String.error_icon_content_description;
			global::Xamarin.Forms.Platform.Resource.String.exposed_dropdown_menu_content_description = global::pjsua2xamarin.Droid.Resource.String.exposed_dropdown_menu_content_description;
			global::Xamarin.Forms.Platform.Resource.String.fab_transformation_scrim_behavior = global::pjsua2xamarin.Droid.Resource.String.fab_transformation_scrim_behavior;
			global::Xamarin.Forms.Platform.Resource.String.fab_transformation_sheet_behavior = global::pjsua2xamarin.Droid.Resource.String.fab_transformation_sheet_behavior;
			global::Xamarin.Forms.Platform.Resource.String.hide_bottom_view_on_scroll_behavior = global::pjsua2xamarin.Droid.Resource.String.hide_bottom_view_on_scroll_behavior;
			global::Xamarin.Forms.Platform.Resource.String.icon_content_description = global::pjsua2xamarin.Droid.Resource.String.icon_content_description;
			global::Xamarin.Forms.Platform.Resource.String.item_view_role_description = global::pjsua2xamarin.Droid.Resource.String.item_view_role_description;
			global::Xamarin.Forms.Platform.Resource.String.material_clock_display_divider = global::pjsua2xamarin.Droid.Resource.String.material_clock_display_divider;
			global::Xamarin.Forms.Platform.Resource.String.material_clock_toggle_content_description = global::pjsua2xamarin.Droid.Resource.String.material_clock_toggle_content_description;
			global::Xamarin.Forms.Platform.Resource.String.material_hour_selection = global::pjsua2xamarin.Droid.Resource.String.material_hour_selection;
			global::Xamarin.Forms.Platform.Resource.String.material_hour_suffix = global::pjsua2xamarin.Droid.Resource.String.material_hour_suffix;
			global::Xamarin.Forms.Platform.Resource.String.material_minute_selection = global::pjsua2xamarin.Droid.Resource.String.material_minute_selection;
			global::Xamarin.Forms.Platform.Resource.String.material_minute_suffix = global::pjsua2xamarin.Droid.Resource.String.material_minute_suffix;
			global::Xamarin.Forms.Platform.Resource.String.material_motion_easing_accelerated = global::pjsua2xamarin.Droid.Resource.String.material_motion_easing_accelerated;
			global::Xamarin.Forms.Platform.Resource.String.material_motion_easing_decelerated = global::pjsua2xamarin.Droid.Resource.String.material_motion_easing_decelerated;
			global::Xamarin.Forms.Platform.Resource.String.material_motion_easing_emphasized = global::pjsua2xamarin.Droid.Resource.String.material_motion_easing_emphasized;
			global::Xamarin.Forms.Platform.Resource.String.material_motion_easing_linear = global::pjsua2xamarin.Droid.Resource.String.material_motion_easing_linear;
			global::Xamarin.Forms.Platform.Resource.String.material_motion_easing_standard = global::pjsua2xamarin.Droid.Resource.String.material_motion_easing_standard;
			global::Xamarin.Forms.Platform.Resource.String.material_slider_range_end = global::pjsua2xamarin.Droid.Resource.String.material_slider_range_end;
			global::Xamarin.Forms.Platform.Resource.String.material_slider_range_start = global::pjsua2xamarin.Droid.Resource.String.material_slider_range_start;
			global::Xamarin.Forms.Platform.Resource.String.material_timepicker_am = global::pjsua2xamarin.Droid.Resource.String.material_timepicker_am;
			global::Xamarin.Forms.Platform.Resource.String.material_timepicker_clock_mode_description = global::pjsua2xamarin.Droid.Resource.String.material_timepicker_clock_mode_description;
			global::Xamarin.Forms.Platform.Resource.String.material_timepicker_hour = global::pjsua2xamarin.Droid.Resource.String.material_timepicker_hour;
			global::Xamarin.Forms.Platform.Resource.String.material_timepicker_minute = global::pjsua2xamarin.Droid.Resource.String.material_timepicker_minute;
			global::Xamarin.Forms.Platform.Resource.String.material_timepicker_pm = global::pjsua2xamarin.Droid.Resource.String.material_timepicker_pm;
			global::Xamarin.Forms.Platform.Resource.String.material_timepicker_select_time = global::pjsua2xamarin.Droid.Resource.String.material_timepicker_select_time;
			global::Xamarin.Forms.Platform.Resource.String.material_timepicker_text_input_mode_description = global::pjsua2xamarin.Droid.Resource.String.material_timepicker_text_input_mode_description;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_badge_numberless_content_description = global::pjsua2xamarin.Droid.Resource.String.mtrl_badge_numberless_content_description;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_chip_close_icon_content_description = global::pjsua2xamarin.Droid.Resource.String.mtrl_chip_close_icon_content_description;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_exceed_max_badge_number_content_description = global::pjsua2xamarin.Droid.Resource.String.mtrl_exceed_max_badge_number_content_description;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_exceed_max_badge_number_suffix = global::pjsua2xamarin.Droid.Resource.String.mtrl_exceed_max_badge_number_suffix;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_a11y_next_month = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_a11y_next_month;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_a11y_prev_month = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_a11y_prev_month;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_announce_current_selection = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_announce_current_selection;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_cancel = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_cancel;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_confirm = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_confirm;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_date_header_selected = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_date_header_selected;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_date_header_title = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_date_header_title;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_date_header_unselected = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_date_header_unselected;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_day_of_week_column_header = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_day_of_week_column_header;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_invalid_format = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_invalid_format;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_invalid_format_example = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_invalid_format_example;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_invalid_format_use = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_invalid_format_use;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_invalid_range = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_invalid_range;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_navigate_to_year_description = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_navigate_to_year_description;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_out_of_range = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_out_of_range;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_range_header_only_end_selected = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_range_header_only_end_selected;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_range_header_only_start_selected = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_range_header_only_start_selected;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_range_header_selected = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_range_header_selected;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_range_header_title = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_range_header_title;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_range_header_unselected = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_range_header_unselected;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_save = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_save;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_text_input_date_hint = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_text_input_date_hint;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_text_input_date_range_end_hint = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_text_input_date_range_end_hint;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_text_input_date_range_start_hint = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_text_input_date_range_start_hint;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_text_input_day_abbr = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_text_input_day_abbr;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_text_input_month_abbr = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_text_input_month_abbr;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_text_input_year_abbr = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_text_input_year_abbr;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_toggle_to_calendar_input_mode = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_toggle_to_calendar_input_mode;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_toggle_to_day_selection = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_toggle_to_day_selection;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_toggle_to_text_input_mode = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_toggle_to_text_input_mode;
			global::Xamarin.Forms.Platform.Resource.String.mtrl_picker_toggle_to_year_selection = global::pjsua2xamarin.Droid.Resource.String.mtrl_picker_toggle_to_year_selection;
			global::Xamarin.Forms.Platform.Resource.String.nav_app_bar_navigate_up_description = global::pjsua2xamarin.Droid.Resource.String.nav_app_bar_navigate_up_description;
			global::Xamarin.Forms.Platform.Resource.String.nav_app_bar_open_drawer_description = global::pjsua2xamarin.Droid.Resource.String.nav_app_bar_open_drawer_description;
			global::Xamarin.Forms.Platform.Resource.String.overflow_tab_title = global::pjsua2xamarin.Droid.Resource.String.overflow_tab_title;
			global::Xamarin.Forms.Platform.Resource.String.password_toggle_content_description = global::pjsua2xamarin.Droid.Resource.String.password_toggle_content_description;
			global::Xamarin.Forms.Platform.Resource.String.path_password_eye = global::pjsua2xamarin.Droid.Resource.String.path_password_eye;
			global::Xamarin.Forms.Platform.Resource.String.path_password_eye_mask_strike_through = global::pjsua2xamarin.Droid.Resource.String.path_password_eye_mask_strike_through;
			global::Xamarin.Forms.Platform.Resource.String.path_password_eye_mask_visible = global::pjsua2xamarin.Droid.Resource.String.path_password_eye_mask_visible;
			global::Xamarin.Forms.Platform.Resource.String.path_password_strike_through = global::pjsua2xamarin.Droid.Resource.String.path_password_strike_through;
			global::Xamarin.Forms.Platform.Resource.String.search_menu_title = global::pjsua2xamarin.Droid.Resource.String.search_menu_title;
			global::Xamarin.Forms.Platform.Resource.String.status_bar_notification_info_overflow = global::pjsua2xamarin.Droid.Resource.String.status_bar_notification_info_overflow;
			global::Xamarin.Forms.Platform.Resource.Style.AlertDialog_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.AlertDialog_AppCompat;
			global::Xamarin.Forms.Platform.Resource.Style.AlertDialog_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.AlertDialog_AppCompat_Light;
			global::Xamarin.Forms.Platform.Resource.Style.AndroidThemeColorAccentYellow = global::pjsua2xamarin.Droid.Resource.Style.AndroidThemeColorAccentYellow;
			global::Xamarin.Forms.Platform.Resource.Style.Animation_AppCompat_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Animation_AppCompat_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Animation_AppCompat_DropDownUp = global::pjsua2xamarin.Droid.Resource.Style.Animation_AppCompat_DropDownUp;
			global::Xamarin.Forms.Platform.Resource.Style.Animation_AppCompat_Tooltip = global::pjsua2xamarin.Droid.Resource.Style.Animation_AppCompat_Tooltip;
			global::Xamarin.Forms.Platform.Resource.Style.Animation_Design_BottomSheetDialog = global::pjsua2xamarin.Droid.Resource.Style.Animation_Design_BottomSheetDialog;
			global::Xamarin.Forms.Platform.Resource.Style.Animation_MaterialComponents_BottomSheetDialog = global::pjsua2xamarin.Droid.Resource.Style.Animation_MaterialComponents_BottomSheetDialog;
			global::Xamarin.Forms.Platform.Resource.Style.AppCompatDialogStyle = global::pjsua2xamarin.Droid.Resource.Style.AppCompatDialogStyle;
			global::Xamarin.Forms.Platform.Resource.Style.Base_AlertDialog_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_AlertDialog_AppCompat;
			global::Xamarin.Forms.Platform.Resource.Style.Base_AlertDialog_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_AlertDialog_AppCompat_Light;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Animation_AppCompat_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_Animation_AppCompat_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Animation_AppCompat_DropDownUp = global::pjsua2xamarin.Droid.Resource.Style.Base_Animation_AppCompat_DropDownUp;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Animation_AppCompat_Tooltip = global::pjsua2xamarin.Droid.Resource.Style.Base_Animation_AppCompat_Tooltip;
			global::Xamarin.Forms.Platform.Resource.Style.Base_CardView = global::pjsua2xamarin.Droid.Resource.Style.Base_CardView;
			global::Xamarin.Forms.Platform.Resource.Style.Base_DialogWindowTitleBackground_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_DialogWindowTitleBackground_AppCompat;
			global::Xamarin.Forms.Platform.Resource.Style.Base_DialogWindowTitle_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_DialogWindowTitle_AppCompat;
			global::Xamarin.Forms.Platform.Resource.Style.Base_MaterialAlertDialog_MaterialComponents_Title_Icon = global::pjsua2xamarin.Droid.Resource.Style.Base_MaterialAlertDialog_MaterialComponents_Title_Icon;
			global::Xamarin.Forms.Platform.Resource.Style.Base_MaterialAlertDialog_MaterialComponents_Title_Panel = global::pjsua2xamarin.Droid.Resource.Style.Base_MaterialAlertDialog_MaterialComponents_Title_Panel;
			global::Xamarin.Forms.Platform.Resource.Style.Base_MaterialAlertDialog_MaterialComponents_Title_Text = global::pjsua2xamarin.Droid.Resource.Style.Base_MaterialAlertDialog_MaterialComponents_Title_Text;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Body1 = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Body1;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Body2 = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Body2;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Button = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Button;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Caption = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Caption;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Display1 = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Display1;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Display2 = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Display2;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Display3 = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Display3;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Display4 = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Display4;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Headline = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Headline;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Large = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Large;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Large_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Large_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Light_Widget_PopupMenu_Large = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Light_Widget_PopupMenu_Large;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Light_Widget_PopupMenu_Small = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Light_Widget_PopupMenu_Small;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Medium = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Medium;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Medium_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Medium_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Menu = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Menu;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_SearchResult = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_SearchResult;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_SearchResult_Subtitle = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_SearchResult_Subtitle;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_SearchResult_Title = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_SearchResult_Title;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Small = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Small;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Small_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Small_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Subhead = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Subhead;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Subhead_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Subhead_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Title = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Title;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Title_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Title_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Tooltip = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Tooltip;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionBar_Menu = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionBar_Menu;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionBar_Subtitle = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionBar_Subtitle;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionBar_Subtitle_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionBar_Subtitle_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionBar_Title = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionBar_Title;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionBar_Title_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionBar_Title_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionMode_Subtitle = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionMode_Subtitle;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionMode_Title = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_ActionMode_Title;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Widget_Button = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_Button;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Widget_Button_Borderless_Colored = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_Button_Borderless_Colored;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Widget_Button_Colored = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_Button_Colored;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Widget_Button_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_Button_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Widget_DropDownItem = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_DropDownItem;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Widget_PopupMenu_Header = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_PopupMenu_Header;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Widget_PopupMenu_Large = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_PopupMenu_Large;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Widget_PopupMenu_Small = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_PopupMenu_Small;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Widget_Switch = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_Switch;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_AppCompat_Widget_TextView_SpinnerItem = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_AppCompat_Widget_TextView_SpinnerItem;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_MaterialComponents_Badge = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_MaterialComponents_Badge;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_MaterialComponents_Button = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_MaterialComponents_Button;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_MaterialComponents_Headline6 = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_MaterialComponents_Headline6;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_MaterialComponents_Subtitle2 = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_MaterialComponents_Subtitle2;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_Widget_AppCompat_ExpandedMenu_Item = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_Widget_AppCompat_ExpandedMenu_Item;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_Widget_AppCompat_Toolbar_Subtitle = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_Widget_AppCompat_Toolbar_Subtitle;
			global::Xamarin.Forms.Platform.Resource.Style.Base_TextAppearance_Widget_AppCompat_Toolbar_Title = global::pjsua2xamarin.Droid.Resource.Style.Base_TextAppearance_Widget_AppCompat_Toolbar_Title;
			global::Xamarin.Forms.Platform.Resource.Style.Base_ThemeOverlay_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_AppCompat;
			global::Xamarin.Forms.Platform.Resource.Style.Base_ThemeOverlay_AppCompat_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_AppCompat_ActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Base_ThemeOverlay_AppCompat_Dark = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_AppCompat_Dark;
			global::Xamarin.Forms.Platform.Resource.Style.Base_ThemeOverlay_AppCompat_Dark_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_AppCompat_Dark_ActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Base_ThemeOverlay_AppCompat_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_AppCompat_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_ThemeOverlay_AppCompat_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_AppCompat_Dialog_Alert;
			global::Xamarin.Forms.Platform.Resource.Style.Base_ThemeOverlay_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_AppCompat_Light;
			global::Xamarin.Forms.Platform.Resource.Style.Base_ThemeOverlay_MaterialComponents_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_MaterialComponents_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_ThemeOverlay_MaterialComponents_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_MaterialComponents_Dialog_Alert;
			global::Xamarin.Forms.Platform.Resource.Style.Base_ThemeOverlay_MaterialComponents_Dialog_Alert_Framework = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_MaterialComponents_Dialog_Alert_Framework;
			global::Xamarin.Forms.Platform.Resource.Style.Base_ThemeOverlay_MaterialComponents_Light_Dialog_Alert_Framework = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_MaterialComponents_Light_Dialog_Alert_Framework;
			global::Xamarin.Forms.Platform.Resource.Style.Base_ThemeOverlay_MaterialComponents_MaterialAlertDialog = global::pjsua2xamarin.Droid.Resource.Style.Base_ThemeOverlay_MaterialComponents_MaterialAlertDialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_AppCompat_CompactMenu = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_CompactMenu;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_AppCompat_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_AppCompat_DialogWhenLarge = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_DialogWhenLarge;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_AppCompat_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_Dialog_Alert;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_AppCompat_Dialog_FixedSize = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_Dialog_FixedSize;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_AppCompat_Dialog_MinWidth = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_Dialog_MinWidth;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_Light;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_AppCompat_Light_DarkActionBar = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_Light_DarkActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_AppCompat_Light_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_Light_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_AppCompat_Light_DialogWhenLarge = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_Light_DialogWhenLarge;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_AppCompat_Light_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_Light_Dialog_Alert;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_AppCompat_Light_Dialog_FixedSize = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_Light_Dialog_FixedSize;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_AppCompat_Light_Dialog_MinWidth = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_AppCompat_Light_Dialog_MinWidth;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_MaterialComponents = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_MaterialComponents_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_MaterialComponents_CompactMenu = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_CompactMenu;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_MaterialComponents_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_MaterialComponents_DialogWhenLarge = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_DialogWhenLarge;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_MaterialComponents_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Dialog_Alert;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_MaterialComponents_Dialog_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Dialog_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_MaterialComponents_Dialog_FixedSize = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Dialog_FixedSize;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_MaterialComponents_Dialog_MinWidth = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Dialog_MinWidth;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_MaterialComponents_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Light;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_MaterialComponents_Light_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Light_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_MaterialComponents_Light_DarkActionBar = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Light_DarkActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_MaterialComponents_Light_DarkActionBar_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Light_DarkActionBar_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_MaterialComponents_Light_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Light_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_MaterialComponents_Light_DialogWhenLarge = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Light_DialogWhenLarge;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_MaterialComponents_Light_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Light_Dialog_Alert;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_MaterialComponents_Light_Dialog_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Light_Dialog_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_MaterialComponents_Light_Dialog_FixedSize = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Light_Dialog_FixedSize;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Theme_MaterialComponents_Light_Dialog_MinWidth = global::pjsua2xamarin.Droid.Resource.Style.Base_Theme_MaterialComponents_Light_Dialog_MinWidth;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V14_ThemeOverlay_MaterialComponents_BottomSheetDialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_ThemeOverlay_MaterialComponents_BottomSheetDialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V14_ThemeOverlay_MaterialComponents_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_ThemeOverlay_MaterialComponents_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V14_ThemeOverlay_MaterialComponents_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_ThemeOverlay_MaterialComponents_Dialog_Alert;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V14_ThemeOverlay_MaterialComponents_MaterialAlertDialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_ThemeOverlay_MaterialComponents_MaterialAlertDialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V14_Theme_MaterialComponents = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_Theme_MaterialComponents;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V14_Theme_MaterialComponents_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_Theme_MaterialComponents_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V14_Theme_MaterialComponents_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_Theme_MaterialComponents_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V14_Theme_MaterialComponents_Dialog_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_Theme_MaterialComponents_Dialog_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V14_Theme_MaterialComponents_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_Theme_MaterialComponents_Light;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V14_Theme_MaterialComponents_Light_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_Theme_MaterialComponents_Light_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V14_Theme_MaterialComponents_Light_DarkActionBar_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_Theme_MaterialComponents_Light_DarkActionBar_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V14_Theme_MaterialComponents_Light_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_Theme_MaterialComponents_Light_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V14_Theme_MaterialComponents_Light_Dialog_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Base_V14_Theme_MaterialComponents_Light_Dialog_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V21_ThemeOverlay_AppCompat_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V21_ThemeOverlay_AppCompat_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V21_ThemeOverlay_MaterialComponents_BottomSheetDialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V21_ThemeOverlay_MaterialComponents_BottomSheetDialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V21_Theme_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_V21_Theme_AppCompat;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V21_Theme_AppCompat_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V21_Theme_AppCompat_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V21_Theme_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_V21_Theme_AppCompat_Light;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V21_Theme_AppCompat_Light_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V21_Theme_AppCompat_Light_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V21_Theme_MaterialComponents = global::pjsua2xamarin.Droid.Resource.Style.Base_V21_Theme_MaterialComponents;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V21_Theme_MaterialComponents_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V21_Theme_MaterialComponents_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V21_Theme_MaterialComponents_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_V21_Theme_MaterialComponents_Light;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V21_Theme_MaterialComponents_Light_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V21_Theme_MaterialComponents_Light_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V22_Theme_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_V22_Theme_AppCompat;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V22_Theme_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_V22_Theme_AppCompat_Light;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V23_Theme_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_V23_Theme_AppCompat;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V23_Theme_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_V23_Theme_AppCompat_Light;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V26_Theme_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_V26_Theme_AppCompat;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V26_Theme_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_V26_Theme_AppCompat_Light;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V26_Widget_AppCompat_Toolbar = global::pjsua2xamarin.Droid.Resource.Style.Base_V26_Widget_AppCompat_Toolbar;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V28_Theme_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_V28_Theme_AppCompat;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V28_Theme_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_V28_Theme_AppCompat_Light;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V7_ThemeOverlay_AppCompat_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V7_ThemeOverlay_AppCompat_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V7_Theme_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Base_V7_Theme_AppCompat;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V7_Theme_AppCompat_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V7_Theme_AppCompat_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V7_Theme_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Base_V7_Theme_AppCompat_Light;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V7_Theme_AppCompat_Light_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Base_V7_Theme_AppCompat_Light_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V7_Widget_AppCompat_AutoCompleteTextView = global::pjsua2xamarin.Droid.Resource.Style.Base_V7_Widget_AppCompat_AutoCompleteTextView;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V7_Widget_AppCompat_EditText = global::pjsua2xamarin.Droid.Resource.Style.Base_V7_Widget_AppCompat_EditText;
			global::Xamarin.Forms.Platform.Resource.Style.Base_V7_Widget_AppCompat_Toolbar = global::pjsua2xamarin.Droid.Resource.Style.Base_V7_Widget_AppCompat_Toolbar;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_ActionBar_Solid = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ActionBar_Solid;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_ActionBar_TabBar = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ActionBar_TabBar;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_ActionBar_TabText = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ActionBar_TabText;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_ActionBar_TabView = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ActionBar_TabView;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_ActionButton = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ActionButton;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_ActionButton_CloseMode = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ActionButton_CloseMode;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_ActionButton_Overflow = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ActionButton_Overflow;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_ActionMode = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ActionMode;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_ActivityChooserView = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ActivityChooserView;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_AutoCompleteTextView = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_AutoCompleteTextView;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_Button = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Button;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_ButtonBar = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ButtonBar;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_ButtonBar_AlertDialog = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ButtonBar_AlertDialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_Button_Borderless = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Button_Borderless;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_Button_Borderless_Colored = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Button_Borderless_Colored;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_Button_ButtonBar_AlertDialog = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Button_ButtonBar_AlertDialog;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_Button_Colored = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Button_Colored;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_Button_Small = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Button_Small;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_CompoundButton_CheckBox = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_CompoundButton_CheckBox;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_CompoundButton_RadioButton = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_CompoundButton_RadioButton;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_CompoundButton_Switch = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_CompoundButton_Switch;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_DrawerArrowToggle = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_DrawerArrowToggle;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_DrawerArrowToggle_Common = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_DrawerArrowToggle_Common;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_DropDownItem_Spinner = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_DropDownItem_Spinner;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_EditText = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_EditText;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_ImageButton = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ImageButton;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_Light_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Light_ActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_Light_ActionBar_Solid = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Light_ActionBar_Solid;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_Light_ActionBar_TabBar = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Light_ActionBar_TabBar;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_Light_ActionBar_TabText = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Light_ActionBar_TabText;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_Light_ActionBar_TabText_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Light_ActionBar_TabText_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_Light_ActionBar_TabView = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Light_ActionBar_TabView;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_Light_PopupMenu = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Light_PopupMenu;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_Light_PopupMenu_Overflow = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Light_PopupMenu_Overflow;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_ListMenuView = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ListMenuView;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_ListPopupWindow = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ListPopupWindow;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_ListView = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ListView;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_ListView_DropDown = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ListView_DropDown;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_ListView_Menu = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ListView_Menu;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_PopupMenu = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_PopupMenu;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_PopupMenu_Overflow = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_PopupMenu_Overflow;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_PopupWindow = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_PopupWindow;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_ProgressBar = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ProgressBar;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_ProgressBar_Horizontal = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_ProgressBar_Horizontal;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_RatingBar = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_RatingBar;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_RatingBar_Indicator = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_RatingBar_Indicator;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_RatingBar_Small = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_RatingBar_Small;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_SearchView = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_SearchView;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_SearchView_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_SearchView_ActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_SeekBar = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_SeekBar;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_SeekBar_Discrete = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_SeekBar_Discrete;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_Spinner = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Spinner;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_Spinner_Underlined = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Spinner_Underlined;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_TextView = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_TextView;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_TextView_SpinnerItem = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_TextView_SpinnerItem;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_Toolbar = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Toolbar;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_AppCompat_Toolbar_Button_Navigation = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_AppCompat_Toolbar_Button_Navigation;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_Design_TabLayout = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_Design_TabLayout;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_MaterialComponents_AutoCompleteTextView = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_AutoCompleteTextView;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_MaterialComponents_CheckedTextView = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_CheckedTextView;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_MaterialComponents_Chip = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_Chip;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_MaterialComponents_MaterialCalendar_NavigationButton = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_MaterialCalendar_NavigationButton;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_MaterialComponents_PopupMenu = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_PopupMenu;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_MaterialComponents_PopupMenu_ContextMenu = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_PopupMenu_ContextMenu;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_MaterialComponents_PopupMenu_ListPopupWindow = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_PopupMenu_ListPopupWindow;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_MaterialComponents_PopupMenu_Overflow = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_PopupMenu_Overflow;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_MaterialComponents_Slider = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_Slider;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_MaterialComponents_Snackbar = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_Snackbar;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_MaterialComponents_TextInputEditText = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_TextInputEditText;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_MaterialComponents_TextInputLayout = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_TextInputLayout;
			global::Xamarin.Forms.Platform.Resource.Style.Base_Widget_MaterialComponents_TextView = global::pjsua2xamarin.Droid.Resource.Style.Base_Widget_MaterialComponents_TextView;
			global::Xamarin.Forms.Platform.Resource.Style.CardView = global::pjsua2xamarin.Droid.Resource.Style.CardView;
			global::Xamarin.Forms.Platform.Resource.Style.CardView_Dark = global::pjsua2xamarin.Droid.Resource.Style.CardView_Dark;
			global::Xamarin.Forms.Platform.Resource.Style.CardView_Light = global::pjsua2xamarin.Droid.Resource.Style.CardView_Light;
			global::Xamarin.Forms.Platform.Resource.Style.collectionViewTheme = global::pjsua2xamarin.Droid.Resource.Style.collectionViewTheme;
			global::Xamarin.Forms.Platform.Resource.Style.EmptyTheme = global::pjsua2xamarin.Droid.Resource.Style.EmptyTheme;
			global::Xamarin.Forms.Platform.Resource.Style.MainTheme = global::pjsua2xamarin.Droid.Resource.Style.MainTheme;
			global::Xamarin.Forms.Platform.Resource.Style.MainTheme_Base = global::pjsua2xamarin.Droid.Resource.Style.MainTheme_Base;
			global::Xamarin.Forms.Platform.Resource.Style.MaterialAlertDialog_MaterialComponents = global::pjsua2xamarin.Droid.Resource.Style.MaterialAlertDialog_MaterialComponents;
			global::Xamarin.Forms.Platform.Resource.Style.MaterialAlertDialog_MaterialComponents_Body_Text = global::pjsua2xamarin.Droid.Resource.Style.MaterialAlertDialog_MaterialComponents_Body_Text;
			global::Xamarin.Forms.Platform.Resource.Style.MaterialAlertDialog_MaterialComponents_Picker_Date_Calendar = global::pjsua2xamarin.Droid.Resource.Style.MaterialAlertDialog_MaterialComponents_Picker_Date_Calendar;
			global::Xamarin.Forms.Platform.Resource.Style.MaterialAlertDialog_MaterialComponents_Picker_Date_Spinner = global::pjsua2xamarin.Droid.Resource.Style.MaterialAlertDialog_MaterialComponents_Picker_Date_Spinner;
			global::Xamarin.Forms.Platform.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Icon = global::pjsua2xamarin.Droid.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Icon;
			global::Xamarin.Forms.Platform.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Icon_CenterStacked = global::pjsua2xamarin.Droid.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Icon_CenterStacked;
			global::Xamarin.Forms.Platform.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Panel = global::pjsua2xamarin.Droid.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Panel;
			global::Xamarin.Forms.Platform.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Panel_CenterStacked = global::pjsua2xamarin.Droid.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Panel_CenterStacked;
			global::Xamarin.Forms.Platform.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Text = global::pjsua2xamarin.Droid.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Text;
			global::Xamarin.Forms.Platform.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Text_CenterStacked = global::pjsua2xamarin.Droid.Resource.Style.MaterialAlertDialog_MaterialComponents_Title_Text_CenterStacked;
			global::Xamarin.Forms.Platform.Resource.Style.Platform_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Platform_AppCompat;
			global::Xamarin.Forms.Platform.Resource.Style.Platform_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Platform_AppCompat_Light;
			global::Xamarin.Forms.Platform.Resource.Style.Platform_MaterialComponents = global::pjsua2xamarin.Droid.Resource.Style.Platform_MaterialComponents;
			global::Xamarin.Forms.Platform.Resource.Style.Platform_MaterialComponents_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Platform_MaterialComponents_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Platform_MaterialComponents_Light = global::pjsua2xamarin.Droid.Resource.Style.Platform_MaterialComponents_Light;
			global::Xamarin.Forms.Platform.Resource.Style.Platform_MaterialComponents_Light_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Platform_MaterialComponents_Light_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Platform_ThemeOverlay_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Platform_ThemeOverlay_AppCompat;
			global::Xamarin.Forms.Platform.Resource.Style.Platform_ThemeOverlay_AppCompat_Dark = global::pjsua2xamarin.Droid.Resource.Style.Platform_ThemeOverlay_AppCompat_Dark;
			global::Xamarin.Forms.Platform.Resource.Style.Platform_ThemeOverlay_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Platform_ThemeOverlay_AppCompat_Light;
			global::Xamarin.Forms.Platform.Resource.Style.Platform_V21_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Platform_V21_AppCompat;
			global::Xamarin.Forms.Platform.Resource.Style.Platform_V21_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Platform_V21_AppCompat_Light;
			global::Xamarin.Forms.Platform.Resource.Style.Platform_V25_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Platform_V25_AppCompat;
			global::Xamarin.Forms.Platform.Resource.Style.Platform_V25_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Platform_V25_AppCompat_Light;
			global::Xamarin.Forms.Platform.Resource.Style.Platform_Widget_AppCompat_Spinner = global::pjsua2xamarin.Droid.Resource.Style.Platform_Widget_AppCompat_Spinner;
			global::Xamarin.Forms.Platform.Resource.Style.RtlOverlay_DialogWindowTitle_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_DialogWindowTitle_AppCompat;
			global::Xamarin.Forms.Platform.Resource.Style.RtlOverlay_Widget_AppCompat_ActionBar_TitleItem = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_ActionBar_TitleItem;
			global::Xamarin.Forms.Platform.Resource.Style.RtlOverlay_Widget_AppCompat_DialogTitle_Icon = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_DialogTitle_Icon;
			global::Xamarin.Forms.Platform.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem;
			global::Xamarin.Forms.Platform.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem_InternalGroup = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem_InternalGroup;
			global::Xamarin.Forms.Platform.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem_Shortcut = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem_Shortcut;
			global::Xamarin.Forms.Platform.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem_SubmenuArrow = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem_SubmenuArrow;
			global::Xamarin.Forms.Platform.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem_Text = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem_Text;
			global::Xamarin.Forms.Platform.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem_Title = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_PopupMenuItem_Title;
			global::Xamarin.Forms.Platform.Resource.Style.RtlOverlay_Widget_AppCompat_SearchView_MagIcon = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_SearchView_MagIcon;
			global::Xamarin.Forms.Platform.Resource.Style.RtlOverlay_Widget_AppCompat_Search_DropDown = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_Search_DropDown;
			global::Xamarin.Forms.Platform.Resource.Style.RtlOverlay_Widget_AppCompat_Search_DropDown_Icon1 = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_Search_DropDown_Icon1;
			global::Xamarin.Forms.Platform.Resource.Style.RtlOverlay_Widget_AppCompat_Search_DropDown_Icon2 = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_Search_DropDown_Icon2;
			global::Xamarin.Forms.Platform.Resource.Style.RtlOverlay_Widget_AppCompat_Search_DropDown_Query = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_Search_DropDown_Query;
			global::Xamarin.Forms.Platform.Resource.Style.RtlOverlay_Widget_AppCompat_Search_DropDown_Text = global::pjsua2xamarin.Droid.Resource.Style.RtlOverlay_Widget_AppCompat_Search_DropDown_Text;
			global::Xamarin.Forms.Platform.Resource.Style.RtlUnderlay_Widget_AppCompat_ActionButton = global::pjsua2xamarin.Droid.Resource.Style.RtlUnderlay_Widget_AppCompat_ActionButton;
			global::Xamarin.Forms.Platform.Resource.Style.RtlUnderlay_Widget_AppCompat_ActionButton_Overflow = global::pjsua2xamarin.Droid.Resource.Style.RtlUnderlay_Widget_AppCompat_ActionButton_Overflow;
			global::Xamarin.Forms.Platform.Resource.Style.scrollViewScrollBars = global::pjsua2xamarin.Droid.Resource.Style.scrollViewScrollBars;
			global::Xamarin.Forms.Platform.Resource.Style.scrollViewTheme = global::pjsua2xamarin.Droid.Resource.Style.scrollViewTheme;
			global::Xamarin.Forms.Platform.Resource.Style.ShapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Resource.Style.ShapeAppearanceOverlay_BottomLeftDifferentCornerSize = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_BottomLeftDifferentCornerSize;
			global::Xamarin.Forms.Platform.Resource.Style.ShapeAppearanceOverlay_BottomRightCut = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_BottomRightCut;
			global::Xamarin.Forms.Platform.Resource.Style.ShapeAppearanceOverlay_Cut = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_Cut;
			global::Xamarin.Forms.Platform.Resource.Style.ShapeAppearanceOverlay_DifferentCornerSize = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_DifferentCornerSize;
			global::Xamarin.Forms.Platform.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_BottomSheet = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_BottomSheet;
			global::Xamarin.Forms.Platform.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_Chip = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_Chip;
			global::Xamarin.Forms.Platform.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_ExtendedFloatingActionButton = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_ExtendedFloatingActionButton;
			global::Xamarin.Forms.Platform.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_FloatingActionButton = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_FloatingActionButton;
			global::Xamarin.Forms.Platform.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_MaterialCalendar_Day = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_MaterialCalendar_Day;
			global::Xamarin.Forms.Platform.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_MaterialCalendar_Window_Fullscreen = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_MaterialCalendar_Window_Fullscreen;
			global::Xamarin.Forms.Platform.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_MaterialCalendar_Year = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_MaterialCalendar_Year;
			global::Xamarin.Forms.Platform.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_TextInputLayout_FilledBox = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_MaterialComponents_TextInputLayout_FilledBox;
			global::Xamarin.Forms.Platform.Resource.Style.ShapeAppearanceOverlay_TopLeftCut = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_TopLeftCut;
			global::Xamarin.Forms.Platform.Resource.Style.ShapeAppearanceOverlay_TopRightDifferentCornerSize = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearanceOverlay_TopRightDifferentCornerSize;
			global::Xamarin.Forms.Platform.Resource.Style.ShapeAppearance_MaterialComponents = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearance_MaterialComponents;
			global::Xamarin.Forms.Platform.Resource.Style.ShapeAppearance_MaterialComponents_LargeComponent = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearance_MaterialComponents_LargeComponent;
			global::Xamarin.Forms.Platform.Resource.Style.ShapeAppearance_MaterialComponents_MediumComponent = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearance_MaterialComponents_MediumComponent;
			global::Xamarin.Forms.Platform.Resource.Style.ShapeAppearance_MaterialComponents_SmallComponent = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearance_MaterialComponents_SmallComponent;
			global::Xamarin.Forms.Platform.Resource.Style.ShapeAppearance_MaterialComponents_Test = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearance_MaterialComponents_Test;
			global::Xamarin.Forms.Platform.Resource.Style.ShapeAppearance_MaterialComponents_Tooltip = global::pjsua2xamarin.Droid.Resource.Style.ShapeAppearance_MaterialComponents_Tooltip;
			global::Xamarin.Forms.Platform.Resource.Style.TestStyleWithLineHeight = global::pjsua2xamarin.Droid.Resource.Style.TestStyleWithLineHeight;
			global::Xamarin.Forms.Platform.Resource.Style.TestStyleWithLineHeightAppearance = global::pjsua2xamarin.Droid.Resource.Style.TestStyleWithLineHeightAppearance;
			global::Xamarin.Forms.Platform.Resource.Style.TestStyleWithoutLineHeight = global::pjsua2xamarin.Droid.Resource.Style.TestStyleWithoutLineHeight;
			global::Xamarin.Forms.Platform.Resource.Style.TestStyleWithThemeLineHeightAttribute = global::pjsua2xamarin.Droid.Resource.Style.TestStyleWithThemeLineHeightAttribute;
			global::Xamarin.Forms.Platform.Resource.Style.TestThemeWithLineHeight = global::pjsua2xamarin.Droid.Resource.Style.TestThemeWithLineHeight;
			global::Xamarin.Forms.Platform.Resource.Style.TestThemeWithLineHeightDisabled = global::pjsua2xamarin.Droid.Resource.Style.TestThemeWithLineHeightDisabled;
			global::Xamarin.Forms.Platform.Resource.Style.Test_ShapeAppearanceOverlay_MaterialComponents_MaterialCalendar_Day = global::pjsua2xamarin.Droid.Resource.Style.Test_ShapeAppearanceOverlay_MaterialComponents_MaterialCalendar_Day;
			global::Xamarin.Forms.Platform.Resource.Style.Test_Theme_MaterialComponents_MaterialCalendar = global::pjsua2xamarin.Droid.Resource.Style.Test_Theme_MaterialComponents_MaterialCalendar;
			global::Xamarin.Forms.Platform.Resource.Style.Test_Widget_MaterialComponents_MaterialCalendar = global::pjsua2xamarin.Droid.Resource.Style.Test_Widget_MaterialComponents_MaterialCalendar;
			global::Xamarin.Forms.Platform.Resource.Style.Test_Widget_MaterialComponents_MaterialCalendar_Day = global::pjsua2xamarin.Droid.Resource.Style.Test_Widget_MaterialComponents_MaterialCalendar_Day;
			global::Xamarin.Forms.Platform.Resource.Style.Test_Widget_MaterialComponents_MaterialCalendar_Day_Selected = global::pjsua2xamarin.Droid.Resource.Style.Test_Widget_MaterialComponents_MaterialCalendar_Day_Selected;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Body1 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Body1;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Body2 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Body2;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Button = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Button;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Caption = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Caption;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Display1 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Display1;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Display2 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Display2;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Display3 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Display3;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Display4 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Display4;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Headline = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Headline;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Inverse = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Large = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Large;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Large_Inverse = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Large_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Light_SearchResult_Subtitle = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Light_SearchResult_Subtitle;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Light_SearchResult_Title = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Light_SearchResult_Title;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Light_Widget_PopupMenu_Large = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Light_Widget_PopupMenu_Large;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Light_Widget_PopupMenu_Small = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Light_Widget_PopupMenu_Small;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Medium = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Medium;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Medium_Inverse = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Medium_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Menu = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Menu;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_SearchResult_Subtitle = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_SearchResult_Subtitle;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_SearchResult_Title = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_SearchResult_Title;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Small = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Small;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Small_Inverse = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Small_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Subhead = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Subhead;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Subhead_Inverse = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Subhead_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Title = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Title;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Title_Inverse = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Title_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Tooltip = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Tooltip;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Widget_ActionBar_Menu = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_ActionBar_Menu;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Widget_ActionBar_Subtitle = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_ActionBar_Subtitle;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Widget_ActionBar_Subtitle_Inverse = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_ActionBar_Subtitle_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Widget_ActionBar_Title = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_ActionBar_Title;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Widget_ActionBar_Title_Inverse = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_ActionBar_Title_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Widget_ActionMode_Subtitle = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_ActionMode_Subtitle;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Widget_ActionMode_Subtitle_Inverse = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_ActionMode_Subtitle_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Widget_ActionMode_Title = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_ActionMode_Title;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Widget_ActionMode_Title_Inverse = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_ActionMode_Title_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Widget_Button = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_Button;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Widget_Button_Borderless_Colored = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_Button_Borderless_Colored;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Widget_Button_Colored = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_Button_Colored;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Widget_Button_Inverse = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_Button_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Widget_DropDownItem = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_DropDownItem;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Widget_PopupMenu_Header = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_PopupMenu_Header;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Widget_PopupMenu_Large = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_PopupMenu_Large;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Widget_PopupMenu_Small = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_PopupMenu_Small;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Widget_Switch = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_Switch;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_AppCompat_Widget_TextView_SpinnerItem = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_AppCompat_Widget_TextView_SpinnerItem;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Compat_Notification = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Compat_Notification_Info = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification_Info;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Compat_Notification_Info_Media = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification_Info_Media;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Compat_Notification_Line2 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification_Line2;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Compat_Notification_Line2_Media = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification_Line2_Media;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Compat_Notification_Media = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification_Media;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Compat_Notification_Time = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification_Time;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Compat_Notification_Time_Media = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification_Time_Media;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Compat_Notification_Title = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification_Title;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Compat_Notification_Title_Media = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Compat_Notification_Title_Media;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Design_CollapsingToolbar_Expanded = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Design_CollapsingToolbar_Expanded;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Design_Counter = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Design_Counter;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Design_Counter_Overflow = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Design_Counter_Overflow;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Design_Error = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Design_Error;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Design_HelperText = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Design_HelperText;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Design_Hint = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Design_Hint;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Design_Placeholder = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Design_Placeholder;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Design_Prefix = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Design_Prefix;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Design_Snackbar_Message = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Design_Snackbar_Message;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Design_Suffix = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Design_Suffix;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Design_Tab = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Design_Tab;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_MaterialComponents_Badge = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Badge;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_MaterialComponents_Body1 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Body1;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_MaterialComponents_Body2 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Body2;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_MaterialComponents_Button = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Button;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_MaterialComponents_Caption = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Caption;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_MaterialComponents_Chip = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Chip;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_MaterialComponents_Headline1 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Headline1;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_MaterialComponents_Headline2 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Headline2;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_MaterialComponents_Headline3 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Headline3;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_MaterialComponents_Headline4 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Headline4;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_MaterialComponents_Headline5 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Headline5;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_MaterialComponents_Headline6 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Headline6;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_MaterialComponents_Overline = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Overline;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_MaterialComponents_Subtitle1 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Subtitle1;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_MaterialComponents_Subtitle2 = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Subtitle2;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_MaterialComponents_TimePicker_Title = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_TimePicker_Title;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_MaterialComponents_Tooltip = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_MaterialComponents_Tooltip;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Widget_AppCompat_ExpandedMenu_Item = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Widget_AppCompat_ExpandedMenu_Item;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Widget_AppCompat_Toolbar_Subtitle = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Widget_AppCompat_Toolbar_Subtitle;
			global::Xamarin.Forms.Platform.Resource.Style.TextAppearance_Widget_AppCompat_Toolbar_Title = global::pjsua2xamarin.Droid.Resource.Style.TextAppearance_Widget_AppCompat_Toolbar_Title;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlayColorAccentRed = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlayColorAccentRed;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_AppCompat;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_AppCompat_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_AppCompat_ActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_AppCompat_Dark = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_AppCompat_Dark;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_AppCompat_Dark_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_AppCompat_Dark_ActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_AppCompat_DayNight = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_AppCompat_DayNight;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_AppCompat_DayNight_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_AppCompat_DayNight_ActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_AppCompat_Dialog = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_AppCompat_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_AppCompat_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_AppCompat_Dialog_Alert;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_AppCompat_Light;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_Design_TextInputEditText = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_Design_TextInputEditText;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_ActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_ActionBar_Primary = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_ActionBar_Primary;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_ActionBar_Surface = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_ActionBar_Surface;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_AutoCompleteTextView = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_AutoCompleteTextView;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_AutoCompleteTextView_FilledBox = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_AutoCompleteTextView_FilledBox;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_AutoCompleteTextView_FilledBox_Dense = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_AutoCompleteTextView_FilledBox_Dense;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_AutoCompleteTextView_OutlinedBox = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_AutoCompleteTextView_OutlinedBox;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_AutoCompleteTextView_OutlinedBox_Dense = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_AutoCompleteTextView_OutlinedBox_Dense;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_BottomAppBar_Primary = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_BottomAppBar_Primary;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_BottomAppBar_Surface = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_BottomAppBar_Surface;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_BottomSheetDialog = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_BottomSheetDialog;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_Dark = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_Dark;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_Dark_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_Dark_ActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_DayNight_BottomSheetDialog = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_DayNight_BottomSheetDialog;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_Dialog = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_Dialog_Alert;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_Dialog_Alert_Framework = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_Dialog_Alert_Framework;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_Light = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_Light;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_Light_Dialog_Alert_Framework = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_Light_Dialog_Alert_Framework;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Centered = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Centered;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date_Calendar = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date_Calendar;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date_Header_Text = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date_Header_Text;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date_Header_Text_Day = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date_Header_Text_Day;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date_Spinner = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date_Spinner;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_MaterialCalendar = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_MaterialCalendar;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_MaterialCalendar_Fullscreen = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_MaterialCalendar_Fullscreen;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_TextInputEditText = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_TextInputEditText;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_TextInputEditText_FilledBox = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_TextInputEditText_FilledBox;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_TextInputEditText_FilledBox_Dense = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_TextInputEditText_FilledBox_Dense;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_TextInputEditText_OutlinedBox = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_TextInputEditText_OutlinedBox;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_TextInputEditText_OutlinedBox_Dense = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_TextInputEditText_OutlinedBox_Dense;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_TimePicker = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_TimePicker;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_TimePicker_Display = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_TimePicker_Display;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_Toolbar_Primary = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_Toolbar_Primary;
			global::Xamarin.Forms.Platform.Resource.Style.ThemeOverlay_MaterialComponents_Toolbar_Surface = global::pjsua2xamarin.Droid.Resource.Style.ThemeOverlay_MaterialComponents_Toolbar_Surface;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_AppCompat = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_AppCompat_CompactMenu = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_CompactMenu;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_AppCompat_DayNight = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_DayNight;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_AppCompat_DayNight_DarkActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_DayNight_DarkActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_AppCompat_DayNight_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_DayNight_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_AppCompat_DayNight_DialogWhenLarge = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_DayNight_DialogWhenLarge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_AppCompat_DayNight_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_DayNight_Dialog_Alert;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_AppCompat_DayNight_Dialog_MinWidth = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_DayNight_Dialog_MinWidth;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_AppCompat_DayNight_NoActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_DayNight_NoActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_AppCompat_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_AppCompat_DialogWhenLarge = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_DialogWhenLarge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_AppCompat_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_Dialog_Alert;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_AppCompat_Dialog_MinWidth = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_Dialog_MinWidth;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_AppCompat_Empty = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_Empty;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_AppCompat_Light = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_Light;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_AppCompat_Light_DarkActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_Light_DarkActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_AppCompat_Light_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_Light_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_AppCompat_Light_DialogWhenLarge = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_Light_DialogWhenLarge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_AppCompat_Light_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_Light_Dialog_Alert;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_AppCompat_Light_Dialog_MinWidth = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_Light_Dialog_MinWidth;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_AppCompat_Light_NoActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_Light_NoActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_AppCompat_NoActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_AppCompat_NoActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_Design = global::pjsua2xamarin.Droid.Resource.Style.Theme_Design;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_Design_BottomSheetDialog = global::pjsua2xamarin.Droid.Resource.Style.Theme_Design_BottomSheetDialog;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_Design_Light = global::pjsua2xamarin.Droid.Resource.Style.Theme_Design_Light;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_Design_Light_BottomSheetDialog = global::pjsua2xamarin.Droid.Resource.Style.Theme_Design_Light_BottomSheetDialog;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_Design_Light_NoActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_Design_Light_NoActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_Design_NoActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_Design_NoActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_BottomSheetDialog = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_BottomSheetDialog;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_CompactMenu = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_CompactMenu;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_DayNight = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_DayNight_BottomSheetDialog = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_BottomSheetDialog;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_DayNight_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_DayNight_DarkActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_DarkActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_DayNight_DarkActionBar_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_DarkActionBar_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_DayNight_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_DayNight_DialogWhenLarge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_DialogWhenLarge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_Alert;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_Alert_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_Alert_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_FixedSize = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_FixedSize;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_FixedSize_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_FixedSize_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_MinWidth = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_MinWidth;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_MinWidth_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_Dialog_MinWidth_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_DayNight_NoActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_NoActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_DayNight_NoActionBar_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DayNight_NoActionBar_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_DialogWhenLarge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_DialogWhenLarge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Dialog_Alert;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Dialog_Alert_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Dialog_Alert_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Dialog_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Dialog_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Dialog_FixedSize = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Dialog_FixedSize;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Dialog_FixedSize_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Dialog_FixedSize_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Dialog_MinWidth = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Dialog_MinWidth;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Dialog_MinWidth_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Dialog_MinWidth_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Light = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Light_BarSize = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_BarSize;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Light_BottomSheetDialog = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_BottomSheetDialog;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Light_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Light_DarkActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_DarkActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Light_DarkActionBar_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_DarkActionBar_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Light_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Light_DialogWhenLarge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_DialogWhenLarge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Light_Dialog_Alert = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_Dialog_Alert;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Light_Dialog_Alert_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_Dialog_Alert_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Light_Dialog_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_Dialog_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Light_Dialog_FixedSize = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_Dialog_FixedSize;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Light_Dialog_FixedSize_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_Dialog_FixedSize_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Light_Dialog_MinWidth = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_Dialog_MinWidth;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Light_Dialog_MinWidth_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_Dialog_MinWidth_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Light_LargeTouch = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_LargeTouch;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Light_NoActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_NoActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_Light_NoActionBar_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_Light_NoActionBar_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_NoActionBar = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_NoActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Theme_MaterialComponents_NoActionBar_Bridge = global::pjsua2xamarin.Droid.Resource.Style.Theme_MaterialComponents_NoActionBar_Bridge;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_ActionBar_Solid = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ActionBar_Solid;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_ActionBar_TabBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ActionBar_TabBar;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_ActionBar_TabText = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ActionBar_TabText;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_ActionBar_TabView = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ActionBar_TabView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_ActionButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ActionButton;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_ActionButton_CloseMode = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ActionButton_CloseMode;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_ActionButton_Overflow = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ActionButton_Overflow;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_ActionMode = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ActionMode;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_ActivityChooserView = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ActivityChooserView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_AutoCompleteTextView = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_AutoCompleteTextView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Button = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Button;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_ButtonBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ButtonBar;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_ButtonBar_AlertDialog = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ButtonBar_AlertDialog;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Button_Borderless = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Button_Borderless;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Button_Borderless_Colored = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Button_Borderless_Colored;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Button_ButtonBar_AlertDialog = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Button_ButtonBar_AlertDialog;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Button_Colored = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Button_Colored;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Button_Small = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Button_Small;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_CompoundButton_CheckBox = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_CompoundButton_CheckBox;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_CompoundButton_RadioButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_CompoundButton_RadioButton;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_CompoundButton_Switch = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_CompoundButton_Switch;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_DrawerArrowToggle = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_DrawerArrowToggle;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_DropDownItem_Spinner = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_DropDownItem_Spinner;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_EditText = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_EditText;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_ImageButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ImageButton;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Light_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Light_ActionBar_Solid = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionBar_Solid;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Light_ActionBar_Solid_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionBar_Solid_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Light_ActionBar_TabBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionBar_TabBar;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Light_ActionBar_TabBar_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionBar_TabBar_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Light_ActionBar_TabText = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionBar_TabText;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Light_ActionBar_TabText_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionBar_TabText_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Light_ActionBar_TabView = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionBar_TabView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Light_ActionBar_TabView_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionBar_TabView_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Light_ActionButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionButton;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Light_ActionButton_CloseMode = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionButton_CloseMode;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Light_ActionButton_Overflow = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionButton_Overflow;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Light_ActionMode_Inverse = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActionMode_Inverse;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Light_ActivityChooserView = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ActivityChooserView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Light_AutoCompleteTextView = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_AutoCompleteTextView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Light_DropDownItem_Spinner = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_DropDownItem_Spinner;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Light_ListPopupWindow = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ListPopupWindow;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Light_ListView_DropDown = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_ListView_DropDown;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Light_PopupMenu = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_PopupMenu;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Light_PopupMenu_Overflow = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_PopupMenu_Overflow;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Light_SearchView = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_SearchView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Light_Spinner_DropDown_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Light_Spinner_DropDown_ActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_ListMenuView = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ListMenuView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_ListPopupWindow = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ListPopupWindow;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_ListView = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ListView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_ListView_DropDown = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ListView_DropDown;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_ListView_Menu = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ListView_Menu;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_PopupMenu = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_PopupMenu;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_PopupMenu_Overflow = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_PopupMenu_Overflow;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_PopupWindow = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_PopupWindow;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_ProgressBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ProgressBar;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_ProgressBar_Horizontal = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_ProgressBar_Horizontal;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_RatingBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_RatingBar;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_RatingBar_Indicator = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_RatingBar_Indicator;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_RatingBar_Small = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_RatingBar_Small;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_SearchView = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_SearchView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_SearchView_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_SearchView_ActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_SeekBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_SeekBar;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_SeekBar_Discrete = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_SeekBar_Discrete;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Spinner = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Spinner;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Spinner_DropDown = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Spinner_DropDown;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Spinner_DropDown_ActionBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Spinner_DropDown_ActionBar;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Spinner_Underlined = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Spinner_Underlined;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_TextView = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_TextView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_TextView_SpinnerItem = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_TextView_SpinnerItem;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Toolbar = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Toolbar;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_AppCompat_Toolbar_Button_Navigation = global::pjsua2xamarin.Droid.Resource.Style.Widget_AppCompat_Toolbar_Button_Navigation;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_Compat_NotificationActionContainer = global::pjsua2xamarin.Droid.Resource.Style.Widget_Compat_NotificationActionContainer;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_Compat_NotificationActionText = global::pjsua2xamarin.Droid.Resource.Style.Widget_Compat_NotificationActionText;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_Design_AppBarLayout = global::pjsua2xamarin.Droid.Resource.Style.Widget_Design_AppBarLayout;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_Design_BottomNavigationView = global::pjsua2xamarin.Droid.Resource.Style.Widget_Design_BottomNavigationView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_Design_BottomSheet_Modal = global::pjsua2xamarin.Droid.Resource.Style.Widget_Design_BottomSheet_Modal;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_Design_CollapsingToolbar = global::pjsua2xamarin.Droid.Resource.Style.Widget_Design_CollapsingToolbar;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_Design_FloatingActionButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_Design_FloatingActionButton;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_Design_NavigationView = global::pjsua2xamarin.Droid.Resource.Style.Widget_Design_NavigationView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_Design_ScrimInsetsFrameLayout = global::pjsua2xamarin.Droid.Resource.Style.Widget_Design_ScrimInsetsFrameLayout;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_Design_Snackbar = global::pjsua2xamarin.Droid.Resource.Style.Widget_Design_Snackbar;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_Design_TabLayout = global::pjsua2xamarin.Droid.Resource.Style.Widget_Design_TabLayout;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_Design_TextInputEditText = global::pjsua2xamarin.Droid.Resource.Style.Widget_Design_TextInputEditText;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_Design_TextInputLayout = global::pjsua2xamarin.Droid.Resource.Style.Widget_Design_TextInputLayout;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_ActionBar_Primary = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_ActionBar_Primary;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_ActionBar_PrimarySurface = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_ActionBar_PrimarySurface;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_ActionBar_Solid = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_ActionBar_Solid;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_ActionBar_Surface = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_ActionBar_Surface;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_AppBarLayout_Primary = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_AppBarLayout_Primary;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_AppBarLayout_PrimarySurface = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_AppBarLayout_PrimarySurface;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_AppBarLayout_Surface = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_AppBarLayout_Surface;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_AutoCompleteTextView_FilledBox = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_AutoCompleteTextView_FilledBox;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_AutoCompleteTextView_FilledBox_Dense = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_AutoCompleteTextView_FilledBox_Dense;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_AutoCompleteTextView_OutlinedBox = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_AutoCompleteTextView_OutlinedBox;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_AutoCompleteTextView_OutlinedBox_Dense = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_AutoCompleteTextView_OutlinedBox_Dense;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Badge = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Badge;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_BottomAppBar = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_BottomAppBar;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_BottomAppBar_Colored = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_BottomAppBar_Colored;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_BottomAppBar_PrimarySurface = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_BottomAppBar_PrimarySurface;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_BottomNavigationView = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_BottomNavigationView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_BottomNavigationView_Colored = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_BottomNavigationView_Colored;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_BottomNavigationView_PrimarySurface = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_BottomNavigationView_PrimarySurface;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_BottomSheet = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_BottomSheet;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_BottomSheet_Modal = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_BottomSheet_Modal;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Button = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Button_Icon = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button_Icon;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Button_OutlinedButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button_OutlinedButton;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Button_OutlinedButton_Icon = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button_OutlinedButton_Icon;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Button_TextButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button_TextButton;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Button_TextButton_Dialog = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button_TextButton_Dialog;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Button_TextButton_Dialog_Flush = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button_TextButton_Dialog_Flush;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Button_TextButton_Dialog_Icon = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button_TextButton_Dialog_Icon;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Button_TextButton_Icon = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button_TextButton_Icon;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Button_TextButton_Snackbar = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button_TextButton_Snackbar;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Button_UnelevatedButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button_UnelevatedButton;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Button_UnelevatedButton_Icon = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Button_UnelevatedButton_Icon;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_CardView = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_CardView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_CheckedTextView = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_CheckedTextView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_ChipGroup = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_ChipGroup;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Chip_Action = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Chip_Action;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Chip_Choice = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Chip_Choice;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Chip_Entry = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Chip_Entry;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Chip_Filter = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Chip_Filter;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_CircularProgressIndicator = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_CircularProgressIndicator;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_CircularProgressIndicator_ExtraSmall = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_CircularProgressIndicator_ExtraSmall;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_CircularProgressIndicator_Medium = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_CircularProgressIndicator_Medium;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_CircularProgressIndicator_Small = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_CircularProgressIndicator_Small;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_CollapsingToolbar = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_CollapsingToolbar;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_CompoundButton_CheckBox = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_CompoundButton_CheckBox;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_CompoundButton_RadioButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_CompoundButton_RadioButton;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_CompoundButton_Switch = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_CompoundButton_Switch;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_ExtendedFloatingActionButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_ExtendedFloatingActionButton;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_ExtendedFloatingActionButton_Icon = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_ExtendedFloatingActionButton_Icon;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_FloatingActionButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_FloatingActionButton;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Light_ActionBar_Solid = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Light_ActionBar_Solid;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_LinearProgressIndicator = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_LinearProgressIndicator;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialButtonToggleGroup = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialButtonToggleGroup;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialCalendar = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Day = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Day;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialCalendar_DayTextView = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_DayTextView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Day_Invalid = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Day_Invalid;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Day_Selected = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Day_Selected;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Day_Today = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Day_Today;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Fullscreen = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Fullscreen;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderCancelButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderCancelButton;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderConfirmButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderConfirmButton;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderDivider = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderDivider;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderLayout = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderLayout;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderSelection = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderSelection;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderSelection_Fullscreen = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderSelection_Fullscreen;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderTitle = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderTitle;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderToggleButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_HeaderToggleButton;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Item = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Item;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialCalendar_MonthNavigationButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_MonthNavigationButton;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialCalendar_MonthTextView = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_MonthTextView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Year = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Year;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialCalendar_YearNavigationButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_YearNavigationButton;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Year_Selected = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Year_Selected;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Year_Today = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_MaterialCalendar_Year_Today;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_NavigationRailView = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_NavigationRailView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_NavigationRailView_Colored = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_NavigationRailView_Colored;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_NavigationRailView_Colored_Compact = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_NavigationRailView_Colored_Compact;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_NavigationRailView_Compact = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_NavigationRailView_Compact;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_NavigationRailView_PrimarySurface = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_NavigationRailView_PrimarySurface;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_NavigationView = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_NavigationView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_PopupMenu = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_PopupMenu;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_PopupMenu_ContextMenu = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_PopupMenu_ContextMenu;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_PopupMenu_ListPopupWindow = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_PopupMenu_ListPopupWindow;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_PopupMenu_Overflow = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_PopupMenu_Overflow;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_ProgressIndicator = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_ProgressIndicator;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_ShapeableImageView = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_ShapeableImageView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Slider = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Slider;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Snackbar = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Snackbar;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Snackbar_FullWidth = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Snackbar_FullWidth;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Snackbar_TextView = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Snackbar_TextView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TabLayout = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TabLayout;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TabLayout_Colored = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TabLayout_Colored;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TabLayout_PrimarySurface = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TabLayout_PrimarySurface;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TextInputEditText_FilledBox = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputEditText_FilledBox;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TextInputEditText_FilledBox_Dense = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputEditText_FilledBox_Dense;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TextInputEditText_OutlinedBox = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputEditText_OutlinedBox;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TextInputEditText_OutlinedBox_Dense = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputEditText_OutlinedBox_Dense;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TextInputLayout_FilledBox = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputLayout_FilledBox;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TextInputLayout_FilledBox_Dense = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputLayout_FilledBox_Dense;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TextInputLayout_FilledBox_Dense_ExposedDropdownMenu = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputLayout_FilledBox_Dense_ExposedDropdownMenu;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TextInputLayout_FilledBox_ExposedDropdownMenu = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputLayout_FilledBox_ExposedDropdownMenu;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TextInputLayout_OutlinedBox = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputLayout_OutlinedBox;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TextInputLayout_OutlinedBox_Dense = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputLayout_OutlinedBox_Dense;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TextInputLayout_OutlinedBox_Dense_ExposedDropdownMenu = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputLayout_OutlinedBox_Dense_ExposedDropdownMenu;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TextInputLayout_OutlinedBox_ExposedDropdownMenu = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextInputLayout_OutlinedBox_ExposedDropdownMenu;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TextView = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TextView;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TimePicker = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TimePicker;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TimePicker_Button = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TimePicker_Button;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TimePicker_Clock = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TimePicker_Clock;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TimePicker_Display = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TimePicker_Display;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TimePicker_Display_TextInputEditText = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TimePicker_Display_TextInputEditText;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TimePicker_ImageButton = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TimePicker_ImageButton;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_TimePicker_ImageButton_ShapeAppearance = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_TimePicker_ImageButton_ShapeAppearance;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Toolbar = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Toolbar;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Toolbar_Primary = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Toolbar_Primary;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Toolbar_PrimarySurface = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Toolbar_PrimarySurface;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Toolbar_Surface = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Toolbar_Surface;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_MaterialComponents_Tooltip = global::pjsua2xamarin.Droid.Resource.Style.Widget_MaterialComponents_Tooltip;
			global::Xamarin.Forms.Platform.Resource.Style.Widget_Support_CoordinatorLayout = global::pjsua2xamarin.Droid.Resource.Style.Widget_Support_CoordinatorLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBarLayout = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBarLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBarLayout_android_layout_gravity = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBarLayout_android_layout_gravity;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_background = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_background;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_backgroundSplit = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_backgroundSplit;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_backgroundStacked = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_backgroundStacked;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_contentInsetEnd = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_contentInsetEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_contentInsetEndWithActions = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_contentInsetEndWithActions;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_contentInsetLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_contentInsetLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_contentInsetRight = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_contentInsetRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_contentInsetStart = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_contentInsetStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_contentInsetStartWithNavigation = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_contentInsetStartWithNavigation;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_customNavigationLayout = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_customNavigationLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_displayOptions = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_displayOptions;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_divider = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_divider;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_elevation;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_height = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_height;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_hideOnContentScroll = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_hideOnContentScroll;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_homeAsUpIndicator = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_homeAsUpIndicator;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_homeLayout = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_homeLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_icon = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_icon;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_indeterminateProgressStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_indeterminateProgressStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_itemPadding = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_itemPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_logo = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_logo;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_navigationMode = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_navigationMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_popupTheme = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_popupTheme;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_progressBarPadding = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_progressBarPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_progressBarStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_progressBarStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_subtitle = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_subtitle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_subtitleTextStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_subtitleTextStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_title = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_title;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionBar_titleTextStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ActionBar_titleTextStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionMenuItemView = global::pjsua2xamarin.Droid.Resource.Styleable.ActionMenuItemView;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionMenuItemView_android_minWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ActionMenuItemView_android_minWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionMenuView = global::pjsua2xamarin.Droid.Resource.Styleable.ActionMenuView;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionMode = global::pjsua2xamarin.Droid.Resource.Styleable.ActionMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionMode_background = global::pjsua2xamarin.Droid.Resource.Styleable.ActionMode_background;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionMode_backgroundSplit = global::pjsua2xamarin.Droid.Resource.Styleable.ActionMode_backgroundSplit;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionMode_closeItemLayout = global::pjsua2xamarin.Droid.Resource.Styleable.ActionMode_closeItemLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionMode_height = global::pjsua2xamarin.Droid.Resource.Styleable.ActionMode_height;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionMode_subtitleTextStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ActionMode_subtitleTextStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActionMode_titleTextStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ActionMode_titleTextStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActivityChooserView = global::pjsua2xamarin.Droid.Resource.Styleable.ActivityChooserView;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActivityChooserView_expandActivityOverflowButtonDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.ActivityChooserView_expandActivityOverflowButtonDrawable;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActivityChooserView_initialActivityCount = global::pjsua2xamarin.Droid.Resource.Styleable.ActivityChooserView_initialActivityCount;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActivityNavigator = global::pjsua2xamarin.Droid.Resource.Styleable.ActivityNavigator;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActivityNavigator_action = global::pjsua2xamarin.Droid.Resource.Styleable.ActivityNavigator_action;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActivityNavigator_android_name = global::pjsua2xamarin.Droid.Resource.Styleable.ActivityNavigator_android_name;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActivityNavigator_data = global::pjsua2xamarin.Droid.Resource.Styleable.ActivityNavigator_data;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActivityNavigator_dataPattern = global::pjsua2xamarin.Droid.Resource.Styleable.ActivityNavigator_dataPattern;
			global::Xamarin.Forms.Platform.Resource.Styleable.ActivityNavigator_targetPackage = global::pjsua2xamarin.Droid.Resource.Styleable.ActivityNavigator_targetPackage;
			global::Xamarin.Forms.Platform.Resource.Styleable.AlertDialog = global::pjsua2xamarin.Droid.Resource.Styleable.AlertDialog;
			global::Xamarin.Forms.Platform.Resource.Styleable.AlertDialog_android_layout = global::pjsua2xamarin.Droid.Resource.Styleable.AlertDialog_android_layout;
			global::Xamarin.Forms.Platform.Resource.Styleable.AlertDialog_buttonIconDimen = global::pjsua2xamarin.Droid.Resource.Styleable.AlertDialog_buttonIconDimen;
			global::Xamarin.Forms.Platform.Resource.Styleable.AlertDialog_buttonPanelSideLayout = global::pjsua2xamarin.Droid.Resource.Styleable.AlertDialog_buttonPanelSideLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.AlertDialog_listItemLayout = global::pjsua2xamarin.Droid.Resource.Styleable.AlertDialog_listItemLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.AlertDialog_listLayout = global::pjsua2xamarin.Droid.Resource.Styleable.AlertDialog_listLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.AlertDialog_multiChoiceItemLayout = global::pjsua2xamarin.Droid.Resource.Styleable.AlertDialog_multiChoiceItemLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.AlertDialog_showTitle = global::pjsua2xamarin.Droid.Resource.Styleable.AlertDialog_showTitle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AlertDialog_singleChoiceItemLayout = global::pjsua2xamarin.Droid.Resource.Styleable.AlertDialog_singleChoiceItemLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.AnimatedStateListDrawableCompat = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableCompat;
			global::Xamarin.Forms.Platform.Resource.Styleable.AnimatedStateListDrawableCompat_android_constantSize = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableCompat_android_constantSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.AnimatedStateListDrawableCompat_android_dither = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableCompat_android_dither;
			global::Xamarin.Forms.Platform.Resource.Styleable.AnimatedStateListDrawableCompat_android_enterFadeDuration = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableCompat_android_enterFadeDuration;
			global::Xamarin.Forms.Platform.Resource.Styleable.AnimatedStateListDrawableCompat_android_exitFadeDuration = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableCompat_android_exitFadeDuration;
			global::Xamarin.Forms.Platform.Resource.Styleable.AnimatedStateListDrawableCompat_android_variablePadding = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableCompat_android_variablePadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.AnimatedStateListDrawableCompat_android_visible = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableCompat_android_visible;
			global::Xamarin.Forms.Platform.Resource.Styleable.AnimatedStateListDrawableItem = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableItem;
			global::Xamarin.Forms.Platform.Resource.Styleable.AnimatedStateListDrawableItem_android_drawable = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableItem_android_drawable;
			global::Xamarin.Forms.Platform.Resource.Styleable.AnimatedStateListDrawableItem_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableItem_android_id;
			global::Xamarin.Forms.Platform.Resource.Styleable.AnimatedStateListDrawableTransition = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableTransition;
			global::Xamarin.Forms.Platform.Resource.Styleable.AnimatedStateListDrawableTransition_android_drawable = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableTransition_android_drawable;
			global::Xamarin.Forms.Platform.Resource.Styleable.AnimatedStateListDrawableTransition_android_fromId = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableTransition_android_fromId;
			global::Xamarin.Forms.Platform.Resource.Styleable.AnimatedStateListDrawableTransition_android_reversible = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableTransition_android_reversible;
			global::Xamarin.Forms.Platform.Resource.Styleable.AnimatedStateListDrawableTransition_android_toId = global::pjsua2xamarin.Droid.Resource.Styleable.AnimatedStateListDrawableTransition_android_toId;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppBarLayout = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppBarLayoutStates = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayoutStates;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppBarLayoutStates_state_collapsed = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayoutStates_state_collapsed;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppBarLayoutStates_state_collapsible = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayoutStates_state_collapsible;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppBarLayoutStates_state_liftable = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayoutStates_state_liftable;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppBarLayoutStates_state_lifted = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayoutStates_state_lifted;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppBarLayout_android_background = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout_android_background;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppBarLayout_android_keyboardNavigationCluster = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout_android_keyboardNavigationCluster;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppBarLayout_android_touchscreenBlocksFocus = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout_android_touchscreenBlocksFocus;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppBarLayout_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout_elevation;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppBarLayout_expanded = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout_expanded;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppBarLayout_Layout = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout_Layout;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppBarLayout_Layout_layout_scrollFlags = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout_Layout_layout_scrollFlags;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppBarLayout_Layout_layout_scrollInterpolator = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout_Layout_layout_scrollInterpolator;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppBarLayout_liftOnScroll = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout_liftOnScroll;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppBarLayout_liftOnScrollTargetViewId = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout_liftOnScrollTargetViewId;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppBarLayout_statusBarForeground = global::pjsua2xamarin.Droid.Resource.Styleable.AppBarLayout_statusBarForeground;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatImageView = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatImageView;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatImageView_android_src = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatImageView_android_src;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatImageView_srcCompat = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatImageView_srcCompat;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatImageView_tint = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatImageView_tint;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatImageView_tintMode = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatImageView_tintMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatSeekBar = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatSeekBar;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatSeekBar_android_thumb = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatSeekBar_android_thumb;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatSeekBar_tickMark = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatSeekBar_tickMark;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatSeekBar_tickMarkTint = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatSeekBar_tickMarkTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatSeekBar_tickMarkTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatSeekBar_tickMarkTintMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextHelper = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextHelper;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextHelper_android_drawableBottom = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextHelper_android_drawableBottom;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextHelper_android_drawableEnd = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextHelper_android_drawableEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextHelper_android_drawableLeft = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextHelper_android_drawableLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextHelper_android_drawableRight = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextHelper_android_drawableRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextHelper_android_drawableStart = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextHelper_android_drawableStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextHelper_android_drawableTop = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextHelper_android_drawableTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextHelper_android_textAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextHelper_android_textAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextView = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextView_android_textAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_android_textAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextView_autoSizeMaxTextSize = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_autoSizeMaxTextSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextView_autoSizeMinTextSize = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_autoSizeMinTextSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextView_autoSizePresetSizes = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_autoSizePresetSizes;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextView_autoSizeStepGranularity = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_autoSizeStepGranularity;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextView_autoSizeTextType = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_autoSizeTextType;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextView_drawableBottomCompat = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_drawableBottomCompat;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextView_drawableEndCompat = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_drawableEndCompat;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextView_drawableLeftCompat = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_drawableLeftCompat;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextView_drawableRightCompat = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_drawableRightCompat;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextView_drawableStartCompat = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_drawableStartCompat;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextView_drawableTint = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_drawableTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextView_drawableTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_drawableTintMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextView_drawableTopCompat = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_drawableTopCompat;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextView_firstBaselineToTopHeight = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_firstBaselineToTopHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextView_fontFamily = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_fontFamily;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextView_fontVariationSettings = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_fontVariationSettings;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextView_lastBaselineToBottomHeight = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_lastBaselineToBottomHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextView_lineHeight = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_lineHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextView_textAllCaps = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_textAllCaps;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTextView_textLocale = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTextView_textLocale;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionBarDivider = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionBarDivider;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionBarItemBackground = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionBarItemBackground;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionBarPopupTheme = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionBarPopupTheme;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionBarSize = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionBarSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionBarSplitStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionBarSplitStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionBarStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionBarStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionBarTabBarStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionBarTabBarStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionBarTabStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionBarTabStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionBarTabTextStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionBarTabTextStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionBarTheme = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionBarTheme;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionBarWidgetTheme = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionBarWidgetTheme;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionDropDownStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionDropDownStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionMenuTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionMenuTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionMenuTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionMenuTextColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionModeBackground = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeBackground;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionModeCloseButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeCloseButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionModeCloseContentDescription = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeCloseContentDescription;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionModeCloseDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeCloseDrawable;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionModeCopyDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeCopyDrawable;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionModeCutDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeCutDrawable;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionModeFindDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeFindDrawable;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionModePasteDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModePasteDrawable;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionModePopupWindowStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModePopupWindowStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionModeSelectAllDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeSelectAllDrawable;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionModeShareDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeShareDrawable;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionModeSplitBackground = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeSplitBackground;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionModeStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionModeTheme = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeTheme;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionModeWebSearchDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionModeWebSearchDrawable;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionOverflowButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionOverflowButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_actionOverflowMenuStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_actionOverflowMenuStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_activityChooserViewStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_activityChooserViewStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_alertDialogButtonGroupStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_alertDialogButtonGroupStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_alertDialogCenterButtons = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_alertDialogCenterButtons;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_alertDialogStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_alertDialogStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_alertDialogTheme = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_alertDialogTheme;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_android_windowAnimationStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_android_windowAnimationStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_android_windowIsFloating = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_android_windowIsFloating;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_autoCompleteTextViewStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_autoCompleteTextViewStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_borderlessButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_borderlessButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_buttonBarButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_buttonBarButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_buttonBarNegativeButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_buttonBarNegativeButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_buttonBarNeutralButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_buttonBarNeutralButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_buttonBarPositiveButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_buttonBarPositiveButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_buttonBarStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_buttonBarStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_buttonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_buttonStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_buttonStyleSmall = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_buttonStyleSmall;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_checkboxStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_checkboxStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_checkedTextViewStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_checkedTextViewStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_colorAccent = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_colorAccent;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_colorBackgroundFloating = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_colorBackgroundFloating;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_colorButtonNormal = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_colorButtonNormal;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_colorControlActivated = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_colorControlActivated;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_colorControlHighlight = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_colorControlHighlight;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_colorControlNormal = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_colorControlNormal;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_colorError = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_colorError;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_colorPrimary = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_colorPrimary;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_colorPrimaryDark = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_colorPrimaryDark;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_colorSwitchThumbNormal = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_colorSwitchThumbNormal;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_controlBackground = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_controlBackground;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_dialogCornerRadius = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_dialogCornerRadius;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_dialogPreferredPadding = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_dialogPreferredPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_dialogTheme = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_dialogTheme;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_dividerHorizontal = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_dividerHorizontal;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_dividerVertical = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_dividerVertical;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_dropdownListPreferredItemHeight = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_dropdownListPreferredItemHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_dropDownListViewStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_dropDownListViewStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_editTextBackground = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_editTextBackground;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_editTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_editTextColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_editTextStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_editTextStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_homeAsUpIndicator = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_homeAsUpIndicator;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_imageButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_imageButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_listChoiceBackgroundIndicator = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listChoiceBackgroundIndicator;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_listChoiceIndicatorMultipleAnimated = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listChoiceIndicatorMultipleAnimated;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_listChoiceIndicatorSingleAnimated = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listChoiceIndicatorSingleAnimated;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_listDividerAlertDialog = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listDividerAlertDialog;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_listMenuViewStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listMenuViewStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_listPopupWindowStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listPopupWindowStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_listPreferredItemHeight = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listPreferredItemHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_listPreferredItemHeightLarge = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listPreferredItemHeightLarge;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_listPreferredItemHeightSmall = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listPreferredItemHeightSmall;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_listPreferredItemPaddingEnd = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listPreferredItemPaddingEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_listPreferredItemPaddingLeft = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listPreferredItemPaddingLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_listPreferredItemPaddingRight = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listPreferredItemPaddingRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_listPreferredItemPaddingStart = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_listPreferredItemPaddingStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_panelBackground = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_panelBackground;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_panelMenuListTheme = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_panelMenuListTheme;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_panelMenuListWidth = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_panelMenuListWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_popupMenuStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_popupMenuStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_popupWindowStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_popupWindowStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_radioButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_radioButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_ratingBarStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_ratingBarStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_ratingBarStyleIndicator = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_ratingBarStyleIndicator;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_ratingBarStyleSmall = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_ratingBarStyleSmall;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_searchViewStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_searchViewStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_seekBarStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_seekBarStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_selectableItemBackground = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_selectableItemBackground;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_selectableItemBackgroundBorderless = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_selectableItemBackgroundBorderless;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_spinnerDropDownItemStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_spinnerDropDownItemStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_spinnerStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_spinnerStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_switchStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_switchStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_textAppearanceLargePopupMenu = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_textAppearanceLargePopupMenu;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_textAppearanceListItem = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_textAppearanceListItem;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_textAppearanceListItemSecondary = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_textAppearanceListItemSecondary;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_textAppearanceListItemSmall = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_textAppearanceListItemSmall;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_textAppearancePopupMenuHeader = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_textAppearancePopupMenuHeader;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_textAppearanceSearchResultSubtitle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_textAppearanceSearchResultSubtitle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_textAppearanceSearchResultTitle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_textAppearanceSearchResultTitle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_textAppearanceSmallPopupMenu = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_textAppearanceSmallPopupMenu;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_textColorAlertDialogListItem = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_textColorAlertDialogListItem;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_textColorSearchUrl = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_textColorSearchUrl;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_toolbarNavigationButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_toolbarNavigationButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_toolbarStyle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_toolbarStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_tooltipForegroundColor = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_tooltipForegroundColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_tooltipFrameBackground = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_tooltipFrameBackground;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_viewInflaterClass = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_viewInflaterClass;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_windowActionBar = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_windowActionBar;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_windowActionBarOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_windowActionBarOverlay;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_windowActionModeOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_windowActionModeOverlay;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_windowFixedHeightMajor = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_windowFixedHeightMajor;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_windowFixedHeightMinor = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_windowFixedHeightMinor;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_windowFixedWidthMajor = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_windowFixedWidthMajor;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_windowFixedWidthMinor = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_windowFixedWidthMinor;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_windowMinWidthMajor = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_windowMinWidthMajor;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_windowMinWidthMinor = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_windowMinWidthMinor;
			global::Xamarin.Forms.Platform.Resource.Styleable.AppCompatTheme_windowNoTitle = global::pjsua2xamarin.Droid.Resource.Styleable.AppCompatTheme_windowNoTitle;
			global::Xamarin.Forms.Platform.Resource.Styleable.Badge = global::pjsua2xamarin.Droid.Resource.Styleable.Badge;
			global::Xamarin.Forms.Platform.Resource.Styleable.Badge_backgroundColor = global::pjsua2xamarin.Droid.Resource.Styleable.Badge_backgroundColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.Badge_badgeGravity = global::pjsua2xamarin.Droid.Resource.Styleable.Badge_badgeGravity;
			global::Xamarin.Forms.Platform.Resource.Styleable.Badge_badgeTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.Badge_badgeTextColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.Badge_horizontalOffset = global::pjsua2xamarin.Droid.Resource.Styleable.Badge_horizontalOffset;
			global::Xamarin.Forms.Platform.Resource.Styleable.Badge_maxCharacterCount = global::pjsua2xamarin.Droid.Resource.Styleable.Badge_maxCharacterCount;
			global::Xamarin.Forms.Platform.Resource.Styleable.Badge_number = global::pjsua2xamarin.Droid.Resource.Styleable.Badge_number;
			global::Xamarin.Forms.Platform.Resource.Styleable.Badge_verticalOffset = global::pjsua2xamarin.Droid.Resource.Styleable.Badge_verticalOffset;
			global::Xamarin.Forms.Platform.Resource.Styleable.BaseProgressIndicator = global::pjsua2xamarin.Droid.Resource.Styleable.BaseProgressIndicator;
			global::Xamarin.Forms.Platform.Resource.Styleable.BaseProgressIndicator_android_indeterminate = global::pjsua2xamarin.Droid.Resource.Styleable.BaseProgressIndicator_android_indeterminate;
			global::Xamarin.Forms.Platform.Resource.Styleable.BaseProgressIndicator_hideAnimationBehavior = global::pjsua2xamarin.Droid.Resource.Styleable.BaseProgressIndicator_hideAnimationBehavior;
			global::Xamarin.Forms.Platform.Resource.Styleable.BaseProgressIndicator_indicatorColor = global::pjsua2xamarin.Droid.Resource.Styleable.BaseProgressIndicator_indicatorColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.BaseProgressIndicator_minHideDelay = global::pjsua2xamarin.Droid.Resource.Styleable.BaseProgressIndicator_minHideDelay;
			global::Xamarin.Forms.Platform.Resource.Styleable.BaseProgressIndicator_showAnimationBehavior = global::pjsua2xamarin.Droid.Resource.Styleable.BaseProgressIndicator_showAnimationBehavior;
			global::Xamarin.Forms.Platform.Resource.Styleable.BaseProgressIndicator_showDelay = global::pjsua2xamarin.Droid.Resource.Styleable.BaseProgressIndicator_showDelay;
			global::Xamarin.Forms.Platform.Resource.Styleable.BaseProgressIndicator_trackColor = global::pjsua2xamarin.Droid.Resource.Styleable.BaseProgressIndicator_trackColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.BaseProgressIndicator_trackCornerRadius = global::pjsua2xamarin.Droid.Resource.Styleable.BaseProgressIndicator_trackCornerRadius;
			global::Xamarin.Forms.Platform.Resource.Styleable.BaseProgressIndicator_trackThickness = global::pjsua2xamarin.Droid.Resource.Styleable.BaseProgressIndicator_trackThickness;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomAppBar = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomAppBar_backgroundTint = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar_backgroundTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomAppBar_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar_elevation;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomAppBar_fabAlignmentMode = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar_fabAlignmentMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomAppBar_fabAnimationMode = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar_fabAnimationMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomAppBar_fabCradleMargin = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar_fabCradleMargin;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomAppBar_fabCradleRoundedCornerRadius = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar_fabCradleRoundedCornerRadius;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomAppBar_fabCradleVerticalOffset = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar_fabCradleVerticalOffset;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomAppBar_hideOnScroll = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar_hideOnScroll;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomAppBar_paddingBottomSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar_paddingBottomSystemWindowInsets;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomAppBar_paddingLeftSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar_paddingLeftSystemWindowInsets;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomAppBar_paddingRightSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Styleable.BottomAppBar_paddingRightSystemWindowInsets;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomNavigationView = global::pjsua2xamarin.Droid.Resource.Styleable.BottomNavigationView;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomNavigationView_itemHorizontalTranslationEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.BottomNavigationView_itemHorizontalTranslationEnabled;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomSheetBehavior_Layout = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomSheetBehavior_Layout_android_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_android_elevation;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomSheetBehavior_Layout_android_maxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_android_maxWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomSheetBehavior_Layout_backgroundTint = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_backgroundTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomSheetBehavior_Layout_behavior_draggable = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_behavior_draggable;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomSheetBehavior_Layout_behavior_expandedOffset = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_behavior_expandedOffset;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomSheetBehavior_Layout_behavior_fitToContents = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_behavior_fitToContents;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomSheetBehavior_Layout_behavior_halfExpandedRatio = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_behavior_halfExpandedRatio;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomSheetBehavior_Layout_behavior_hideable = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_behavior_hideable;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomSheetBehavior_Layout_behavior_peekHeight = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_behavior_peekHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomSheetBehavior_Layout_behavior_saveFlags = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_behavior_saveFlags;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomSheetBehavior_Layout_behavior_skipCollapsed = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_behavior_skipCollapsed;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomSheetBehavior_Layout_gestureInsetBottomIgnored = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_gestureInsetBottomIgnored;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomSheetBehavior_Layout_paddingBottomSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_paddingBottomSystemWindowInsets;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomSheetBehavior_Layout_paddingLeftSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_paddingLeftSystemWindowInsets;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomSheetBehavior_Layout_paddingRightSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_paddingRightSystemWindowInsets;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomSheetBehavior_Layout_paddingTopSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_paddingTopSystemWindowInsets;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomSheetBehavior_Layout_shapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_shapeAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.BottomSheetBehavior_Layout_shapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.BottomSheetBehavior_Layout_shapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Resource.Styleable.ButtonBarLayout = global::pjsua2xamarin.Droid.Resource.Styleable.ButtonBarLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.ButtonBarLayout_allowStacking = global::pjsua2xamarin.Droid.Resource.Styleable.ButtonBarLayout_allowStacking;
			global::Xamarin.Forms.Platform.Resource.Styleable.Capability = global::pjsua2xamarin.Droid.Resource.Styleable.Capability;
			global::Xamarin.Forms.Platform.Resource.Styleable.Capability_queryPatterns = global::pjsua2xamarin.Droid.Resource.Styleable.Capability_queryPatterns;
			global::Xamarin.Forms.Platform.Resource.Styleable.Capability_shortcutMatchRequired = global::pjsua2xamarin.Droid.Resource.Styleable.Capability_shortcutMatchRequired;
			global::Xamarin.Forms.Platform.Resource.Styleable.CardView = global::pjsua2xamarin.Droid.Resource.Styleable.CardView;
			global::Xamarin.Forms.Platform.Resource.Styleable.CardView_android_minHeight = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_android_minHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.CardView_android_minWidth = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_android_minWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.CardView_cardBackgroundColor = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_cardBackgroundColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.CardView_cardCornerRadius = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_cardCornerRadius;
			global::Xamarin.Forms.Platform.Resource.Styleable.CardView_cardElevation = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_cardElevation;
			global::Xamarin.Forms.Platform.Resource.Styleable.CardView_cardMaxElevation = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_cardMaxElevation;
			global::Xamarin.Forms.Platform.Resource.Styleable.CardView_cardPreventCornerOverlap = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_cardPreventCornerOverlap;
			global::Xamarin.Forms.Platform.Resource.Styleable.CardView_cardUseCompatPadding = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_cardUseCompatPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.CardView_contentPadding = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_contentPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.CardView_contentPaddingBottom = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_contentPaddingBottom;
			global::Xamarin.Forms.Platform.Resource.Styleable.CardView_contentPaddingLeft = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_contentPaddingLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.CardView_contentPaddingRight = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_contentPaddingRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.CardView_contentPaddingTop = global::pjsua2xamarin.Droid.Resource.Styleable.CardView_contentPaddingTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.Carousel = global::pjsua2xamarin.Droid.Resource.Styleable.Carousel;
			global::Xamarin.Forms.Platform.Resource.Styleable.Carousel_carousel_backwardTransition = global::pjsua2xamarin.Droid.Resource.Styleable.Carousel_carousel_backwardTransition;
			global::Xamarin.Forms.Platform.Resource.Styleable.Carousel_carousel_emptyViewsBehavior = global::pjsua2xamarin.Droid.Resource.Styleable.Carousel_carousel_emptyViewsBehavior;
			global::Xamarin.Forms.Platform.Resource.Styleable.Carousel_carousel_firstView = global::pjsua2xamarin.Droid.Resource.Styleable.Carousel_carousel_firstView;
			global::Xamarin.Forms.Platform.Resource.Styleable.Carousel_carousel_forwardTransition = global::pjsua2xamarin.Droid.Resource.Styleable.Carousel_carousel_forwardTransition;
			global::Xamarin.Forms.Platform.Resource.Styleable.Carousel_carousel_infinite = global::pjsua2xamarin.Droid.Resource.Styleable.Carousel_carousel_infinite;
			global::Xamarin.Forms.Platform.Resource.Styleable.Carousel_carousel_nextState = global::pjsua2xamarin.Droid.Resource.Styleable.Carousel_carousel_nextState;
			global::Xamarin.Forms.Platform.Resource.Styleable.Carousel_carousel_previousState = global::pjsua2xamarin.Droid.Resource.Styleable.Carousel_carousel_previousState;
			global::Xamarin.Forms.Platform.Resource.Styleable.Carousel_carousel_touchUpMode = global::pjsua2xamarin.Droid.Resource.Styleable.Carousel_carousel_touchUpMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.Carousel_carousel_touchUp_dampeningFactor = global::pjsua2xamarin.Droid.Resource.Styleable.Carousel_carousel_touchUp_dampeningFactor;
			global::Xamarin.Forms.Platform.Resource.Styleable.Carousel_carousel_touchUp_velocityThreshold = global::pjsua2xamarin.Droid.Resource.Styleable.Carousel_carousel_touchUp_velocityThreshold;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip = global::pjsua2xamarin.Droid.Resource.Styleable.Chip;
			global::Xamarin.Forms.Platform.Resource.Styleable.ChipGroup = global::pjsua2xamarin.Droid.Resource.Styleable.ChipGroup;
			global::Xamarin.Forms.Platform.Resource.Styleable.ChipGroup_checkedChip = global::pjsua2xamarin.Droid.Resource.Styleable.ChipGroup_checkedChip;
			global::Xamarin.Forms.Platform.Resource.Styleable.ChipGroup_chipSpacing = global::pjsua2xamarin.Droid.Resource.Styleable.ChipGroup_chipSpacing;
			global::Xamarin.Forms.Platform.Resource.Styleable.ChipGroup_chipSpacingHorizontal = global::pjsua2xamarin.Droid.Resource.Styleable.ChipGroup_chipSpacingHorizontal;
			global::Xamarin.Forms.Platform.Resource.Styleable.ChipGroup_chipSpacingVertical = global::pjsua2xamarin.Droid.Resource.Styleable.ChipGroup_chipSpacingVertical;
			global::Xamarin.Forms.Platform.Resource.Styleable.ChipGroup_selectionRequired = global::pjsua2xamarin.Droid.Resource.Styleable.ChipGroup_selectionRequired;
			global::Xamarin.Forms.Platform.Resource.Styleable.ChipGroup_singleLine = global::pjsua2xamarin.Droid.Resource.Styleable.ChipGroup_singleLine;
			global::Xamarin.Forms.Platform.Resource.Styleable.ChipGroup_singleSelection = global::pjsua2xamarin.Droid.Resource.Styleable.ChipGroup_singleSelection;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_android_checkable = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_android_checkable;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_android_ellipsize = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_android_ellipsize;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_android_maxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_android_maxWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_android_text = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_android_text;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_android_textAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_android_textAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_android_textColor = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_android_textColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_android_textSize = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_android_textSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_checkedIcon = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_checkedIcon;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_checkedIconEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_checkedIconEnabled;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_checkedIconTint = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_checkedIconTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_checkedIconVisible = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_checkedIconVisible;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_chipBackgroundColor = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipBackgroundColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_chipCornerRadius = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipCornerRadius;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_chipEndPadding = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipEndPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_chipIcon = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipIcon;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_chipIconEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipIconEnabled;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_chipIconSize = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipIconSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_chipIconTint = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipIconTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_chipIconVisible = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipIconVisible;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_chipMinHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipMinHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_chipMinTouchTargetSize = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipMinTouchTargetSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_chipStartPadding = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipStartPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_chipStrokeColor = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipStrokeColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_chipStrokeWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipStrokeWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_chipSurfaceColor = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_chipSurfaceColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_closeIcon = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_closeIcon;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_closeIconEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_closeIconEnabled;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_closeIconEndPadding = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_closeIconEndPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_closeIconSize = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_closeIconSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_closeIconStartPadding = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_closeIconStartPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_closeIconTint = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_closeIconTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_closeIconVisible = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_closeIconVisible;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_ensureMinTouchTargetSize = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_ensureMinTouchTargetSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_hideMotionSpec = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_hideMotionSpec;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_iconEndPadding = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_iconEndPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_iconStartPadding = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_iconStartPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_rippleColor = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_rippleColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_shapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_shapeAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_shapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_shapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_showMotionSpec = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_showMotionSpec;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_textEndPadding = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_textEndPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.Chip_textStartPadding = global::pjsua2xamarin.Droid.Resource.Styleable.Chip_textStartPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.CircularProgressIndicator = global::pjsua2xamarin.Droid.Resource.Styleable.CircularProgressIndicator;
			global::Xamarin.Forms.Platform.Resource.Styleable.CircularProgressIndicator_indicatorDirectionCircular = global::pjsua2xamarin.Droid.Resource.Styleable.CircularProgressIndicator_indicatorDirectionCircular;
			global::Xamarin.Forms.Platform.Resource.Styleable.CircularProgressIndicator_indicatorInset = global::pjsua2xamarin.Droid.Resource.Styleable.CircularProgressIndicator_indicatorInset;
			global::Xamarin.Forms.Platform.Resource.Styleable.CircularProgressIndicator_indicatorSize = global::pjsua2xamarin.Droid.Resource.Styleable.CircularProgressIndicator_indicatorSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.ClockFaceView = global::pjsua2xamarin.Droid.Resource.Styleable.ClockFaceView;
			global::Xamarin.Forms.Platform.Resource.Styleable.ClockFaceView_clockFaceBackgroundColor = global::pjsua2xamarin.Droid.Resource.Styleable.ClockFaceView_clockFaceBackgroundColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.ClockFaceView_clockNumberTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.ClockFaceView_clockNumberTextColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.ClockHandView = global::pjsua2xamarin.Droid.Resource.Styleable.ClockHandView;
			global::Xamarin.Forms.Platform.Resource.Styleable.ClockHandView_clockHandColor = global::pjsua2xamarin.Droid.Resource.Styleable.ClockHandView_clockHandColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.ClockHandView_materialCircleRadius = global::pjsua2xamarin.Droid.Resource.Styleable.ClockHandView_materialCircleRadius;
			global::Xamarin.Forms.Platform.Resource.Styleable.ClockHandView_selectorSize = global::pjsua2xamarin.Droid.Resource.Styleable.ClockHandView_selectorSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_collapsedTitleGravity = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_collapsedTitleGravity;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_collapsedTitleTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_collapsedTitleTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_contentScrim = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_contentScrim;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_expandedTitleGravity = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_expandedTitleGravity;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_expandedTitleMargin = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_expandedTitleMargin;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_expandedTitleMarginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_expandedTitleMarginBottom;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_expandedTitleMarginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_expandedTitleMarginEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_expandedTitleMarginStart = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_expandedTitleMarginStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_expandedTitleMarginTop = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_expandedTitleMarginTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_expandedTitleTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_expandedTitleTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_extraMultilineHeightEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_extraMultilineHeightEnabled;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_forceApplySystemWindowInsetTop = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_forceApplySystemWindowInsetTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_Layout = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_Layout;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_Layout_layout_collapseMode = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_Layout_layout_collapseMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_Layout_layout_collapseParallaxMultiplier = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_Layout_layout_collapseParallaxMultiplier;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_maxLines = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_maxLines;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_scrimAnimationDuration = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_scrimAnimationDuration;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_scrimVisibleHeightTrigger = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_scrimVisibleHeightTrigger;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_statusBarScrim = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_statusBarScrim;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_title = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_title;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_titleCollapseMode = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_titleCollapseMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_titleEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_titleEnabled;
			global::Xamarin.Forms.Platform.Resource.Styleable.CollapsingToolbarLayout_toolbarId = global::pjsua2xamarin.Droid.Resource.Styleable.CollapsingToolbarLayout_toolbarId;
			global::Xamarin.Forms.Platform.Resource.Styleable.ColorStateListItem = global::pjsua2xamarin.Droid.Resource.Styleable.ColorStateListItem;
			global::Xamarin.Forms.Platform.Resource.Styleable.ColorStateListItem_alpha = global::pjsua2xamarin.Droid.Resource.Styleable.ColorStateListItem_alpha;
			global::Xamarin.Forms.Platform.Resource.Styleable.ColorStateListItem_android_alpha = global::pjsua2xamarin.Droid.Resource.Styleable.ColorStateListItem_android_alpha;
			global::Xamarin.Forms.Platform.Resource.Styleable.ColorStateListItem_android_color = global::pjsua2xamarin.Droid.Resource.Styleable.ColorStateListItem_android_color;
			global::Xamarin.Forms.Platform.Resource.Styleable.CompoundButton = global::pjsua2xamarin.Droid.Resource.Styleable.CompoundButton;
			global::Xamarin.Forms.Platform.Resource.Styleable.CompoundButton_android_button = global::pjsua2xamarin.Droid.Resource.Styleable.CompoundButton_android_button;
			global::Xamarin.Forms.Platform.Resource.Styleable.CompoundButton_buttonCompat = global::pjsua2xamarin.Droid.Resource.Styleable.CompoundButton_buttonCompat;
			global::Xamarin.Forms.Platform.Resource.Styleable.CompoundButton_buttonTint = global::pjsua2xamarin.Droid.Resource.Styleable.CompoundButton_buttonTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.CompoundButton_buttonTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.CompoundButton_buttonTintMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_elevation;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_layout_height = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_layout_height;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_layout_margin = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_layout_margin;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginBottom;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginHorizontal = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginHorizontal;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginRight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginStart = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginTop = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginVertical = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_layout_marginVertical;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_layout_width = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_layout_width;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_maxHeight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_maxHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_maxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_maxWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_minHeight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_minHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_minWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_minWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_orientation = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_orientation;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_padding = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_padding;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_paddingBottom = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_paddingBottom;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_paddingEnd = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_paddingEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_paddingLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_paddingLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_paddingRight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_paddingRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_paddingStart = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_paddingStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_paddingTop = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_paddingTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_android_visibility = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_android_visibility;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_barrierAllowsGoneWidgets = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_barrierAllowsGoneWidgets;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_barrierDirection = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_barrierDirection;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_barrierMargin = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_barrierMargin;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_chainUseRtl = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_chainUseRtl;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_circularflow_angles = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_circularflow_angles;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_circularflow_defaultAngle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_circularflow_defaultAngle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_circularflow_defaultRadius = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_circularflow_defaultRadius;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_circularflow_radiusInDP = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_circularflow_radiusInDP;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_circularflow_viewCenter = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_circularflow_viewCenter;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_constraintSet = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_constraintSet;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_constraint_referenced_ids = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_constraint_referenced_ids;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_constraint_referenced_tags = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_constraint_referenced_tags;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_flow_firstHorizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_firstHorizontalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_flow_firstHorizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_firstHorizontalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_flow_firstVerticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_firstVerticalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_flow_firstVerticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_firstVerticalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_flow_horizontalAlign = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_horizontalAlign;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_flow_horizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_horizontalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_flow_horizontalGap = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_horizontalGap;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_flow_horizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_horizontalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_flow_lastHorizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_lastHorizontalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_flow_lastHorizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_lastHorizontalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_flow_lastVerticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_lastVerticalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_flow_lastVerticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_lastVerticalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_flow_maxElementsWrap = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_maxElementsWrap;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_flow_verticalAlign = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_verticalAlign;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_flow_verticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_verticalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_flow_verticalGap = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_verticalGap;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_flow_verticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_verticalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_flow_wrapMode = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_flow_wrapMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layoutDescription = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layoutDescription;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constrainedHeight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constrainedHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constrainedWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constrainedWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBaseline_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBaseline_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBaseline_toBaselineOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBaseline_toBaselineOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBaseline_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBaseline_toBottomOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBaseline_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBaseline_toTopOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBottom_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBottom_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBottom_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBottom_toBottomOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBottom_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintBottom_toTopOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintCircle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintCircle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintCircleAngle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintCircleAngle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintCircleRadius = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintCircleRadius;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintDimensionRatio = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintDimensionRatio;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintEnd_toEndOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintEnd_toEndOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintEnd_toStartOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintEnd_toStartOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintGuide_begin = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintGuide_begin;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintGuide_end = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintGuide_end;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintGuide_percent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintGuide_percent;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHeight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHeight_default = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHeight_default;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHeight_max = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHeight_max;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHeight_min = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHeight_min;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHeight_percent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHeight_percent;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHorizontal_bias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHorizontal_bias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHorizontal_chainStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHorizontal_chainStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHorizontal_weight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintHorizontal_weight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintLeft_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintLeft_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintLeft_toLeftOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintLeft_toLeftOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintLeft_toRightOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintLeft_toRightOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintRight_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintRight_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintRight_toLeftOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintRight_toLeftOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintRight_toRightOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintRight_toRightOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintStart_toEndOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintStart_toEndOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintStart_toStartOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintStart_toStartOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintTag = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintTag;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintTop_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintTop_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintTop_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintTop_toBottomOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintTop_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintTop_toTopOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintVertical_bias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintVertical_bias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintVertical_chainStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintVertical_chainStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintVertical_weight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintVertical_weight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintWidth_default = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintWidth_default;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintWidth_max = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintWidth_max;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintWidth_min = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintWidth_min;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_constraintWidth_percent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_constraintWidth_percent;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_editor_absoluteX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_editor_absoluteX;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_editor_absoluteY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_editor_absoluteY;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginBaseline = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginBaseline;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginBottom;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginRight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginStart = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginTop = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_goneMarginTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_marginBaseline = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_marginBaseline;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_optimizationLevel = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_optimizationLevel;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_Layout_layout_wrapBehaviorInParent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_Layout_layout_wrapBehaviorInParent;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_placeholder = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_placeholder;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_placeholder_content = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_placeholder_content;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_placeholder_placeholder_emptyVisibility = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_placeholder_placeholder_emptyVisibility;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_ReactiveGuide = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_ReactiveGuide;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_ReactiveGuide_reactiveGuide_animateChange = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_ReactiveGuide_reactiveGuide_animateChange;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_ReactiveGuide_reactiveGuide_applyToAllConstraintSets = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_ReactiveGuide_reactiveGuide_applyToAllConstraintSets;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_ReactiveGuide_reactiveGuide_applyToConstraintSet = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_ReactiveGuide_reactiveGuide_applyToConstraintSet;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintLayout_ReactiveGuide_reactiveGuide_valueId = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintLayout_ReactiveGuide_reactiveGuide_valueId;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_alpha = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_alpha;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_elevation;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_id;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_layout_height = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_layout_height;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_layout_marginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_layout_marginBottom;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_layout_marginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_layout_marginEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_layout_marginLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_layout_marginLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_layout_marginRight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_layout_marginRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_layout_marginStart = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_layout_marginStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_layout_marginTop = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_layout_marginTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_layout_width = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_layout_width;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_maxHeight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_maxHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_maxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_maxWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_minHeight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_minHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_minWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_minWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_orientation = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_orientation;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_rotation = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_rotation;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_rotationX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_rotationX;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_rotationY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_rotationY;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_scaleX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_scaleX;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_scaleY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_scaleY;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_transformPivotX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_transformPivotX;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_transformPivotY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_transformPivotY;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_translationX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_translationX;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_translationY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_translationY;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_translationZ = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_translationZ;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_android_visibility = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_android_visibility;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_animateCircleAngleTo = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_animateCircleAngleTo;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_animateRelativeTo = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_animateRelativeTo;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_barrierAllowsGoneWidgets = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_barrierAllowsGoneWidgets;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_barrierDirection = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_barrierDirection;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_barrierMargin = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_barrierMargin;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_chainUseRtl = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_chainUseRtl;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_constraint_referenced_ids = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_constraint_referenced_ids;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_drawPath = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_drawPath;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_flow_firstHorizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_firstHorizontalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_flow_firstHorizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_firstHorizontalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_flow_firstVerticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_firstVerticalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_flow_firstVerticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_firstVerticalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_flow_horizontalAlign = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_horizontalAlign;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_flow_horizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_horizontalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_flow_horizontalGap = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_horizontalGap;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_flow_horizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_horizontalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_flow_lastHorizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_lastHorizontalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_flow_lastHorizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_lastHorizontalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_flow_lastVerticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_lastVerticalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_flow_lastVerticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_lastVerticalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_flow_maxElementsWrap = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_maxElementsWrap;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_flow_verticalAlign = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_verticalAlign;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_flow_verticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_verticalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_flow_verticalGap = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_verticalGap;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_flow_verticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_verticalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_flow_wrapMode = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_flow_wrapMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constrainedHeight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constrainedHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constrainedWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constrainedWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintBaseline_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintBaseline_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintBottom_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintBottom_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintCircleAngle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintCircleAngle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintCircleRadius = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintCircleRadius;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintDimensionRatio = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintDimensionRatio;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintGuide_begin = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintGuide_begin;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintGuide_end = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintGuide_end;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintGuide_percent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintGuide_percent;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintHeight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintHeight_default = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintHeight_default;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintHeight_max = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintHeight_max;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintHeight_min = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintHeight_min;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintHeight_percent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintHeight_percent;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintHorizontal_bias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintHorizontal_bias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintHorizontal_chainStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintHorizontal_chainStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintHorizontal_weight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintHorizontal_weight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintLeft_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintLeft_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintRight_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintRight_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintTag = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintTag;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintTop_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintTop_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintVertical_bias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintVertical_bias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintVertical_chainStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintVertical_chainStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintVertical_weight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintVertical_weight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintWidth_default = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintWidth_default;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintWidth_max = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintWidth_max;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintWidth_min = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintWidth_min;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_constraintWidth_percent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_constraintWidth_percent;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_editor_absoluteX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_editor_absoluteX;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_editor_absoluteY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_editor_absoluteY;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_goneMarginBaseline = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_goneMarginBaseline;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_goneMarginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_goneMarginBottom;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_goneMarginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_goneMarginEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_goneMarginLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_goneMarginLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_goneMarginRight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_goneMarginRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_goneMarginStart = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_goneMarginStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_goneMarginTop = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_goneMarginTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_marginBaseline = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_marginBaseline;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_layout_wrapBehaviorInParent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_layout_wrapBehaviorInParent;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_motionProgress = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_motionProgress;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_motionStagger = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_motionStagger;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_motionTarget = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_motionTarget;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_pathMotionArc = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_pathMotionArc;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_pivotAnchor = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_pivotAnchor;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_polarRelativeTo = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_polarRelativeTo;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_quantizeMotionInterpolator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_quantizeMotionInterpolator;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_quantizeMotionPhase = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_quantizeMotionPhase;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_quantizeMotionSteps = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_quantizeMotionSteps;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_transformPivotTarget = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_transformPivotTarget;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_transitionEasing = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_transitionEasing;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_transitionPathRotate = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_transitionPathRotate;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintOverride_visibilityMode = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintOverride_visibilityMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_alpha = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_alpha;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_elevation;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_id;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_layout_height = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_layout_height;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_layout_marginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_layout_marginBottom;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_layout_marginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_layout_marginEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_layout_marginLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_layout_marginLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_layout_marginRight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_layout_marginRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_layout_marginStart = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_layout_marginStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_layout_marginTop = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_layout_marginTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_layout_width = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_layout_width;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_maxHeight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_maxHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_maxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_maxWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_minHeight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_minHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_minWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_minWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_orientation = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_orientation;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_pivotX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_pivotX;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_pivotY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_pivotY;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_rotation = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_rotation;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_rotationX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_rotationX;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_rotationY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_rotationY;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_scaleX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_scaleX;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_scaleY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_scaleY;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_transformPivotX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_transformPivotX;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_transformPivotY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_transformPivotY;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_translationX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_translationX;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_translationY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_translationY;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_translationZ = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_translationZ;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_android_visibility = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_android_visibility;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_animateCircleAngleTo = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_animateCircleAngleTo;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_animateRelativeTo = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_animateRelativeTo;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_barrierAllowsGoneWidgets = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_barrierAllowsGoneWidgets;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_barrierDirection = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_barrierDirection;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_barrierMargin = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_barrierMargin;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_chainUseRtl = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_chainUseRtl;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_ConstraintRotate = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_ConstraintRotate;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_constraint_referenced_ids = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_constraint_referenced_ids;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_constraint_referenced_tags = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_constraint_referenced_tags;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_deriveConstraintsFrom = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_deriveConstraintsFrom;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_drawPath = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_drawPath;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_flow_firstHorizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_firstHorizontalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_flow_firstHorizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_firstHorizontalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_flow_firstVerticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_firstVerticalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_flow_firstVerticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_firstVerticalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_flow_horizontalAlign = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_horizontalAlign;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_flow_horizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_horizontalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_flow_horizontalGap = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_horizontalGap;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_flow_horizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_horizontalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_flow_lastHorizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_lastHorizontalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_flow_lastHorizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_lastHorizontalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_flow_lastVerticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_lastVerticalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_flow_lastVerticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_lastVerticalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_flow_maxElementsWrap = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_maxElementsWrap;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_flow_verticalAlign = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_verticalAlign;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_flow_verticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_verticalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_flow_verticalGap = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_verticalGap;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_flow_verticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_verticalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_flow_wrapMode = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_flow_wrapMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constrainedHeight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constrainedHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constrainedWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constrainedWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintBaseline_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintBaseline_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintBaseline_toBaselineOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintBaseline_toBaselineOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintBaseline_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintBaseline_toBottomOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintBaseline_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintBaseline_toTopOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintBottom_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintBottom_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintBottom_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintBottom_toBottomOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintBottom_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintBottom_toTopOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintCircle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintCircle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintCircleAngle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintCircleAngle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintCircleRadius = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintCircleRadius;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintDimensionRatio = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintDimensionRatio;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintEnd_toEndOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintEnd_toEndOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintEnd_toStartOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintEnd_toStartOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintGuide_begin = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintGuide_begin;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintGuide_end = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintGuide_end;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintGuide_percent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintGuide_percent;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintHeight_default = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintHeight_default;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintHeight_max = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintHeight_max;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintHeight_min = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintHeight_min;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintHeight_percent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintHeight_percent;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintHorizontal_bias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintHorizontal_bias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintHorizontal_chainStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintHorizontal_chainStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintHorizontal_weight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintHorizontal_weight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintLeft_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintLeft_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintLeft_toLeftOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintLeft_toLeftOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintLeft_toRightOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintLeft_toRightOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintRight_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintRight_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintRight_toLeftOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintRight_toLeftOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintRight_toRightOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintRight_toRightOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintStart_toEndOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintStart_toEndOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintStart_toStartOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintStart_toStartOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintTag = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintTag;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintTop_creator = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintTop_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintTop_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintTop_toBottomOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintTop_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintTop_toTopOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintVertical_bias = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintVertical_bias;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintVertical_chainStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintVertical_chainStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintVertical_weight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintVertical_weight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintWidth_default = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintWidth_default;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintWidth_max = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintWidth_max;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintWidth_min = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintWidth_min;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_constraintWidth_percent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_constraintWidth_percent;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_editor_absoluteX = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_editor_absoluteX;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_editor_absoluteY = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_editor_absoluteY;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_goneMarginBaseline = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_goneMarginBaseline;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_goneMarginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_goneMarginBottom;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_goneMarginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_goneMarginEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_goneMarginLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_goneMarginLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_goneMarginRight = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_goneMarginRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_goneMarginStart = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_goneMarginStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_goneMarginTop = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_goneMarginTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_marginBaseline = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_marginBaseline;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_layout_wrapBehaviorInParent = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_layout_wrapBehaviorInParent;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_motionProgress = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_motionProgress;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_motionStagger = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_motionStagger;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_pathMotionArc = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_pathMotionArc;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_pivotAnchor = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_pivotAnchor;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_polarRelativeTo = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_polarRelativeTo;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_quantizeMotionSteps = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_quantizeMotionSteps;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_transitionEasing = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_transitionEasing;
			global::Xamarin.Forms.Platform.Resource.Styleable.ConstraintSet_transitionPathRotate = global::pjsua2xamarin.Droid.Resource.Styleable.ConstraintSet_transitionPathRotate;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_alpha = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_alpha;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_elevation;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_id;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_layout_height = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_layout_height;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_layout_marginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_layout_marginBottom;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_layout_marginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_layout_marginEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_layout_marginLeft = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_layout_marginLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_layout_marginRight = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_layout_marginRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_layout_marginStart = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_layout_marginStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_layout_marginTop = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_layout_marginTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_layout_width = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_layout_width;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_maxHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_maxHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_maxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_maxWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_minHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_minHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_minWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_minWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_orientation = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_orientation;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_rotation = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_rotation;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_rotationX = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_rotationX;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_rotationY = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_rotationY;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_scaleX = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_scaleX;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_scaleY = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_scaleY;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_transformPivotX = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_transformPivotX;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_transformPivotY = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_transformPivotY;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_translationX = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_translationX;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_translationY = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_translationY;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_translationZ = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_translationZ;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_android_visibility = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_android_visibility;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_animateCircleAngleTo = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_animateCircleAngleTo;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_animateRelativeTo = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_animateRelativeTo;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_barrierAllowsGoneWidgets = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_barrierAllowsGoneWidgets;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_barrierDirection = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_barrierDirection;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_barrierMargin = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_barrierMargin;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_chainUseRtl = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_chainUseRtl;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_constraint_referenced_ids = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_constraint_referenced_ids;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_constraint_referenced_tags = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_constraint_referenced_tags;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_drawPath = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_drawPath;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_flow_firstHorizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_firstHorizontalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_flow_firstHorizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_firstHorizontalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_flow_firstVerticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_firstVerticalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_flow_firstVerticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_firstVerticalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_flow_horizontalAlign = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_horizontalAlign;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_flow_horizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_horizontalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_flow_horizontalGap = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_horizontalGap;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_flow_horizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_horizontalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_flow_lastHorizontalBias = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_lastHorizontalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_flow_lastHorizontalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_lastHorizontalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_flow_lastVerticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_lastVerticalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_flow_lastVerticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_lastVerticalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_flow_maxElementsWrap = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_maxElementsWrap;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_flow_verticalAlign = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_verticalAlign;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_flow_verticalBias = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_verticalBias;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_flow_verticalGap = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_verticalGap;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_flow_verticalStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_verticalStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_flow_wrapMode = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_flow_wrapMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constrainedHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constrainedHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constrainedWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constrainedWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintBaseline_creator = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintBaseline_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintBaseline_toBaselineOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintBaseline_toBaselineOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintBaseline_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintBaseline_toBottomOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintBaseline_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintBaseline_toTopOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintBottom_creator = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintBottom_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintBottom_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintBottom_toBottomOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintBottom_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintBottom_toTopOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintCircle = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintCircle;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintCircleAngle = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintCircleAngle;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintCircleRadius = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintCircleRadius;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintDimensionRatio = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintDimensionRatio;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintEnd_toEndOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintEnd_toEndOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintEnd_toStartOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintEnd_toStartOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintGuide_begin = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintGuide_begin;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintGuide_end = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintGuide_end;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintGuide_percent = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintGuide_percent;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintHeight_default = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintHeight_default;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintHeight_max = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintHeight_max;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintHeight_min = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintHeight_min;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintHeight_percent = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintHeight_percent;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintHorizontal_bias = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintHorizontal_bias;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintHorizontal_chainStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintHorizontal_chainStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintHorizontal_weight = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintHorizontal_weight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintLeft_creator = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintLeft_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintLeft_toLeftOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintLeft_toLeftOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintLeft_toRightOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintLeft_toRightOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintRight_creator = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintRight_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintRight_toLeftOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintRight_toLeftOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintRight_toRightOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintRight_toRightOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintStart_toEndOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintStart_toEndOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintStart_toStartOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintStart_toStartOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintTag = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintTag;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintTop_creator = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintTop_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintTop_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintTop_toBottomOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintTop_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintTop_toTopOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintVertical_bias = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintVertical_bias;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintVertical_chainStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintVertical_chainStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintVertical_weight = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintVertical_weight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintWidth_default = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintWidth_default;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintWidth_max = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintWidth_max;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintWidth_min = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintWidth_min;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_constraintWidth_percent = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_constraintWidth_percent;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_editor_absoluteX = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_editor_absoluteX;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_editor_absoluteY = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_editor_absoluteY;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_goneMarginBaseline = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_goneMarginBaseline;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_goneMarginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_goneMarginBottom;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_goneMarginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_goneMarginEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_goneMarginLeft = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_goneMarginLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_goneMarginRight = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_goneMarginRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_goneMarginStart = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_goneMarginStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_goneMarginTop = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_goneMarginTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_marginBaseline = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_marginBaseline;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_layout_wrapBehaviorInParent = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_layout_wrapBehaviorInParent;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_motionProgress = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_motionProgress;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_motionStagger = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_motionStagger;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_pathMotionArc = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_pathMotionArc;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_pivotAnchor = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_pivotAnchor;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_polarRelativeTo = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_polarRelativeTo;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_quantizeMotionInterpolator = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_quantizeMotionInterpolator;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_quantizeMotionPhase = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_quantizeMotionPhase;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_quantizeMotionSteps = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_quantizeMotionSteps;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_transformPivotTarget = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_transformPivotTarget;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_transitionEasing = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_transitionEasing;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_transitionPathRotate = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_transitionPathRotate;
			global::Xamarin.Forms.Platform.Resource.Styleable.Constraint_visibilityMode = global::pjsua2xamarin.Droid.Resource.Styleable.Constraint_visibilityMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.CoordinatorLayout = global::pjsua2xamarin.Droid.Resource.Styleable.CoordinatorLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.CoordinatorLayout_keylines = global::pjsua2xamarin.Droid.Resource.Styleable.CoordinatorLayout_keylines;
			global::Xamarin.Forms.Platform.Resource.Styleable.CoordinatorLayout_Layout = global::pjsua2xamarin.Droid.Resource.Styleable.CoordinatorLayout_Layout;
			global::Xamarin.Forms.Platform.Resource.Styleable.CoordinatorLayout_Layout_android_layout_gravity = global::pjsua2xamarin.Droid.Resource.Styleable.CoordinatorLayout_Layout_android_layout_gravity;
			global::Xamarin.Forms.Platform.Resource.Styleable.CoordinatorLayout_Layout_layout_anchor = global::pjsua2xamarin.Droid.Resource.Styleable.CoordinatorLayout_Layout_layout_anchor;
			global::Xamarin.Forms.Platform.Resource.Styleable.CoordinatorLayout_Layout_layout_anchorGravity = global::pjsua2xamarin.Droid.Resource.Styleable.CoordinatorLayout_Layout_layout_anchorGravity;
			global::Xamarin.Forms.Platform.Resource.Styleable.CoordinatorLayout_Layout_layout_behavior = global::pjsua2xamarin.Droid.Resource.Styleable.CoordinatorLayout_Layout_layout_behavior;
			global::Xamarin.Forms.Platform.Resource.Styleable.CoordinatorLayout_Layout_layout_dodgeInsetEdges = global::pjsua2xamarin.Droid.Resource.Styleable.CoordinatorLayout_Layout_layout_dodgeInsetEdges;
			global::Xamarin.Forms.Platform.Resource.Styleable.CoordinatorLayout_Layout_layout_insetEdge = global::pjsua2xamarin.Droid.Resource.Styleable.CoordinatorLayout_Layout_layout_insetEdge;
			global::Xamarin.Forms.Platform.Resource.Styleable.CoordinatorLayout_Layout_layout_keyline = global::pjsua2xamarin.Droid.Resource.Styleable.CoordinatorLayout_Layout_layout_keyline;
			global::Xamarin.Forms.Platform.Resource.Styleable.CoordinatorLayout_statusBarBackground = global::pjsua2xamarin.Droid.Resource.Styleable.CoordinatorLayout_statusBarBackground;
			global::Xamarin.Forms.Platform.Resource.Styleable.CustomAttribute = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute;
			global::Xamarin.Forms.Platform.Resource.Styleable.CustomAttribute_attributeName = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute_attributeName;
			global::Xamarin.Forms.Platform.Resource.Styleable.CustomAttribute_customBoolean = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute_customBoolean;
			global::Xamarin.Forms.Platform.Resource.Styleable.CustomAttribute_customColorDrawableValue = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute_customColorDrawableValue;
			global::Xamarin.Forms.Platform.Resource.Styleable.CustomAttribute_customColorValue = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute_customColorValue;
			global::Xamarin.Forms.Platform.Resource.Styleable.CustomAttribute_customDimension = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute_customDimension;
			global::Xamarin.Forms.Platform.Resource.Styleable.CustomAttribute_customFloatValue = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute_customFloatValue;
			global::Xamarin.Forms.Platform.Resource.Styleable.CustomAttribute_customIntegerValue = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute_customIntegerValue;
			global::Xamarin.Forms.Platform.Resource.Styleable.CustomAttribute_customPixelDimension = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute_customPixelDimension;
			global::Xamarin.Forms.Platform.Resource.Styleable.CustomAttribute_customReference = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute_customReference;
			global::Xamarin.Forms.Platform.Resource.Styleable.CustomAttribute_customStringValue = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute_customStringValue;
			global::Xamarin.Forms.Platform.Resource.Styleable.CustomAttribute_methodName = global::pjsua2xamarin.Droid.Resource.Styleable.CustomAttribute_methodName;
			global::Xamarin.Forms.Platform.Resource.Styleable.DrawerArrowToggle = global::pjsua2xamarin.Droid.Resource.Styleable.DrawerArrowToggle;
			global::Xamarin.Forms.Platform.Resource.Styleable.DrawerArrowToggle_arrowHeadLength = global::pjsua2xamarin.Droid.Resource.Styleable.DrawerArrowToggle_arrowHeadLength;
			global::Xamarin.Forms.Platform.Resource.Styleable.DrawerArrowToggle_arrowShaftLength = global::pjsua2xamarin.Droid.Resource.Styleable.DrawerArrowToggle_arrowShaftLength;
			global::Xamarin.Forms.Platform.Resource.Styleable.DrawerArrowToggle_barLength = global::pjsua2xamarin.Droid.Resource.Styleable.DrawerArrowToggle_barLength;
			global::Xamarin.Forms.Platform.Resource.Styleable.DrawerArrowToggle_color = global::pjsua2xamarin.Droid.Resource.Styleable.DrawerArrowToggle_color;
			global::Xamarin.Forms.Platform.Resource.Styleable.DrawerArrowToggle_drawableSize = global::pjsua2xamarin.Droid.Resource.Styleable.DrawerArrowToggle_drawableSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.DrawerArrowToggle_gapBetweenBars = global::pjsua2xamarin.Droid.Resource.Styleable.DrawerArrowToggle_gapBetweenBars;
			global::Xamarin.Forms.Platform.Resource.Styleable.DrawerArrowToggle_spinBars = global::pjsua2xamarin.Droid.Resource.Styleable.DrawerArrowToggle_spinBars;
			global::Xamarin.Forms.Platform.Resource.Styleable.DrawerArrowToggle_thickness = global::pjsua2xamarin.Droid.Resource.Styleable.DrawerArrowToggle_thickness;
			global::Xamarin.Forms.Platform.Resource.Styleable.DrawerLayout = global::pjsua2xamarin.Droid.Resource.Styleable.DrawerLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.DrawerLayout_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.DrawerLayout_elevation;
			global::Xamarin.Forms.Platform.Resource.Styleable.ExtendedFloatingActionButton = global::pjsua2xamarin.Droid.Resource.Styleable.ExtendedFloatingActionButton;
			global::Xamarin.Forms.Platform.Resource.Styleable.ExtendedFloatingActionButton_Behavior_Layout = global::pjsua2xamarin.Droid.Resource.Styleable.ExtendedFloatingActionButton_Behavior_Layout;
			global::Xamarin.Forms.Platform.Resource.Styleable.ExtendedFloatingActionButton_Behavior_Layout_behavior_autoHide = global::pjsua2xamarin.Droid.Resource.Styleable.ExtendedFloatingActionButton_Behavior_Layout_behavior_autoHide;
			global::Xamarin.Forms.Platform.Resource.Styleable.ExtendedFloatingActionButton_Behavior_Layout_behavior_autoShrink = global::pjsua2xamarin.Droid.Resource.Styleable.ExtendedFloatingActionButton_Behavior_Layout_behavior_autoShrink;
			global::Xamarin.Forms.Platform.Resource.Styleable.ExtendedFloatingActionButton_collapsedSize = global::pjsua2xamarin.Droid.Resource.Styleable.ExtendedFloatingActionButton_collapsedSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.ExtendedFloatingActionButton_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.ExtendedFloatingActionButton_elevation;
			global::Xamarin.Forms.Platform.Resource.Styleable.ExtendedFloatingActionButton_extendMotionSpec = global::pjsua2xamarin.Droid.Resource.Styleable.ExtendedFloatingActionButton_extendMotionSpec;
			global::Xamarin.Forms.Platform.Resource.Styleable.ExtendedFloatingActionButton_hideMotionSpec = global::pjsua2xamarin.Droid.Resource.Styleable.ExtendedFloatingActionButton_hideMotionSpec;
			global::Xamarin.Forms.Platform.Resource.Styleable.ExtendedFloatingActionButton_showMotionSpec = global::pjsua2xamarin.Droid.Resource.Styleable.ExtendedFloatingActionButton_showMotionSpec;
			global::Xamarin.Forms.Platform.Resource.Styleable.ExtendedFloatingActionButton_shrinkMotionSpec = global::pjsua2xamarin.Droid.Resource.Styleable.ExtendedFloatingActionButton_shrinkMotionSpec;
			global::Xamarin.Forms.Platform.Resource.Styleable.FloatingActionButton = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton;
			global::Xamarin.Forms.Platform.Resource.Styleable.FloatingActionButton_android_enabled = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_android_enabled;
			global::Xamarin.Forms.Platform.Resource.Styleable.FloatingActionButton_backgroundTint = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_backgroundTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.FloatingActionButton_backgroundTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_backgroundTintMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.FloatingActionButton_Behavior_Layout = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_Behavior_Layout;
			global::Xamarin.Forms.Platform.Resource.Styleable.FloatingActionButton_Behavior_Layout_behavior_autoHide = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_Behavior_Layout_behavior_autoHide;
			global::Xamarin.Forms.Platform.Resource.Styleable.FloatingActionButton_borderWidth = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_borderWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.FloatingActionButton_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_elevation;
			global::Xamarin.Forms.Platform.Resource.Styleable.FloatingActionButton_ensureMinTouchTargetSize = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_ensureMinTouchTargetSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.FloatingActionButton_fabCustomSize = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_fabCustomSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.FloatingActionButton_fabSize = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_fabSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.FloatingActionButton_hideMotionSpec = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_hideMotionSpec;
			global::Xamarin.Forms.Platform.Resource.Styleable.FloatingActionButton_hoveredFocusedTranslationZ = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_hoveredFocusedTranslationZ;
			global::Xamarin.Forms.Platform.Resource.Styleable.FloatingActionButton_maxImageSize = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_maxImageSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.FloatingActionButton_pressedTranslationZ = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_pressedTranslationZ;
			global::Xamarin.Forms.Platform.Resource.Styleable.FloatingActionButton_rippleColor = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_rippleColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.FloatingActionButton_shapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_shapeAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.FloatingActionButton_shapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_shapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Resource.Styleable.FloatingActionButton_showMotionSpec = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_showMotionSpec;
			global::Xamarin.Forms.Platform.Resource.Styleable.FloatingActionButton_useCompatPadding = global::pjsua2xamarin.Droid.Resource.Styleable.FloatingActionButton_useCompatPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.FlowLayout = global::pjsua2xamarin.Droid.Resource.Styleable.FlowLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.FlowLayout_itemSpacing = global::pjsua2xamarin.Droid.Resource.Styleable.FlowLayout_itemSpacing;
			global::Xamarin.Forms.Platform.Resource.Styleable.FlowLayout_lineSpacing = global::pjsua2xamarin.Droid.Resource.Styleable.FlowLayout_lineSpacing;
			global::Xamarin.Forms.Platform.Resource.Styleable.FontFamily = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily;
			global::Xamarin.Forms.Platform.Resource.Styleable.FontFamilyFont = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont;
			global::Xamarin.Forms.Platform.Resource.Styleable.FontFamilyFont_android_font = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_android_font;
			global::Xamarin.Forms.Platform.Resource.Styleable.FontFamilyFont_android_fontStyle = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_android_fontStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.FontFamilyFont_android_fontVariationSettings = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_android_fontVariationSettings;
			global::Xamarin.Forms.Platform.Resource.Styleable.FontFamilyFont_android_fontWeight = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_android_fontWeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.FontFamilyFont_android_ttcIndex = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_android_ttcIndex;
			global::Xamarin.Forms.Platform.Resource.Styleable.FontFamilyFont_font = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_font;
			global::Xamarin.Forms.Platform.Resource.Styleable.FontFamilyFont_fontStyle = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_fontStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.FontFamilyFont_fontVariationSettings = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_fontVariationSettings;
			global::Xamarin.Forms.Platform.Resource.Styleable.FontFamilyFont_fontWeight = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_fontWeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.FontFamilyFont_ttcIndex = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamilyFont_ttcIndex;
			global::Xamarin.Forms.Platform.Resource.Styleable.FontFamily_fontProviderAuthority = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily_fontProviderAuthority;
			global::Xamarin.Forms.Platform.Resource.Styleable.FontFamily_fontProviderCerts = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily_fontProviderCerts;
			global::Xamarin.Forms.Platform.Resource.Styleable.FontFamily_fontProviderFetchStrategy = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily_fontProviderFetchStrategy;
			global::Xamarin.Forms.Platform.Resource.Styleable.FontFamily_fontProviderFetchTimeout = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily_fontProviderFetchTimeout;
			global::Xamarin.Forms.Platform.Resource.Styleable.FontFamily_fontProviderPackage = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily_fontProviderPackage;
			global::Xamarin.Forms.Platform.Resource.Styleable.FontFamily_fontProviderQuery = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily_fontProviderQuery;
			global::Xamarin.Forms.Platform.Resource.Styleable.FontFamily_fontProviderSystemFontFamily = global::pjsua2xamarin.Droid.Resource.Styleable.FontFamily_fontProviderSystemFontFamily;
			global::Xamarin.Forms.Platform.Resource.Styleable.ForegroundLinearLayout = global::pjsua2xamarin.Droid.Resource.Styleable.ForegroundLinearLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.ForegroundLinearLayout_android_foreground = global::pjsua2xamarin.Droid.Resource.Styleable.ForegroundLinearLayout_android_foreground;
			global::Xamarin.Forms.Platform.Resource.Styleable.ForegroundLinearLayout_android_foregroundGravity = global::pjsua2xamarin.Droid.Resource.Styleable.ForegroundLinearLayout_android_foregroundGravity;
			global::Xamarin.Forms.Platform.Resource.Styleable.ForegroundLinearLayout_foregroundInsidePadding = global::pjsua2xamarin.Droid.Resource.Styleable.ForegroundLinearLayout_foregroundInsidePadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.Fragment = global::pjsua2xamarin.Droid.Resource.Styleable.Fragment;
			global::Xamarin.Forms.Platform.Resource.Styleable.FragmentContainerView = global::pjsua2xamarin.Droid.Resource.Styleable.FragmentContainerView;
			global::Xamarin.Forms.Platform.Resource.Styleable.FragmentContainerView_android_name = global::pjsua2xamarin.Droid.Resource.Styleable.FragmentContainerView_android_name;
			global::Xamarin.Forms.Platform.Resource.Styleable.FragmentContainerView_android_tag = global::pjsua2xamarin.Droid.Resource.Styleable.FragmentContainerView_android_tag;
			global::Xamarin.Forms.Platform.Resource.Styleable.Fragment_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.Fragment_android_id;
			global::Xamarin.Forms.Platform.Resource.Styleable.Fragment_android_name = global::pjsua2xamarin.Droid.Resource.Styleable.Fragment_android_name;
			global::Xamarin.Forms.Platform.Resource.Styleable.Fragment_android_tag = global::pjsua2xamarin.Droid.Resource.Styleable.Fragment_android_tag;
			global::Xamarin.Forms.Platform.Resource.Styleable.GradientColor = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.GradientColorItem = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColorItem;
			global::Xamarin.Forms.Platform.Resource.Styleable.GradientColorItem_android_color = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColorItem_android_color;
			global::Xamarin.Forms.Platform.Resource.Styleable.GradientColorItem_android_offset = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColorItem_android_offset;
			global::Xamarin.Forms.Platform.Resource.Styleable.GradientColor_android_centerColor = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_centerColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.GradientColor_android_centerX = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_centerX;
			global::Xamarin.Forms.Platform.Resource.Styleable.GradientColor_android_centerY = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_centerY;
			global::Xamarin.Forms.Platform.Resource.Styleable.GradientColor_android_endColor = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_endColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.GradientColor_android_endX = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_endX;
			global::Xamarin.Forms.Platform.Resource.Styleable.GradientColor_android_endY = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_endY;
			global::Xamarin.Forms.Platform.Resource.Styleable.GradientColor_android_gradientRadius = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_gradientRadius;
			global::Xamarin.Forms.Platform.Resource.Styleable.GradientColor_android_startColor = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_startColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.GradientColor_android_startX = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_startX;
			global::Xamarin.Forms.Platform.Resource.Styleable.GradientColor_android_startY = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_startY;
			global::Xamarin.Forms.Platform.Resource.Styleable.GradientColor_android_tileMode = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_tileMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.GradientColor_android_type = global::pjsua2xamarin.Droid.Resource.Styleable.GradientColor_android_type;
			global::Xamarin.Forms.Platform.Resource.Styleable.ImageFilterView = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView;
			global::Xamarin.Forms.Platform.Resource.Styleable.ImageFilterView_altSrc = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_altSrc;
			global::Xamarin.Forms.Platform.Resource.Styleable.ImageFilterView_blendSrc = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_blendSrc;
			global::Xamarin.Forms.Platform.Resource.Styleable.ImageFilterView_brightness = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_brightness;
			global::Xamarin.Forms.Platform.Resource.Styleable.ImageFilterView_contrast = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_contrast;
			global::Xamarin.Forms.Platform.Resource.Styleable.ImageFilterView_crossfade = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_crossfade;
			global::Xamarin.Forms.Platform.Resource.Styleable.ImageFilterView_imagePanX = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_imagePanX;
			global::Xamarin.Forms.Platform.Resource.Styleable.ImageFilterView_imagePanY = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_imagePanY;
			global::Xamarin.Forms.Platform.Resource.Styleable.ImageFilterView_imageRotate = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_imageRotate;
			global::Xamarin.Forms.Platform.Resource.Styleable.ImageFilterView_imageZoom = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_imageZoom;
			global::Xamarin.Forms.Platform.Resource.Styleable.ImageFilterView_overlay = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_overlay;
			global::Xamarin.Forms.Platform.Resource.Styleable.ImageFilterView_round = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_round;
			global::Xamarin.Forms.Platform.Resource.Styleable.ImageFilterView_roundPercent = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_roundPercent;
			global::Xamarin.Forms.Platform.Resource.Styleable.ImageFilterView_saturation = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_saturation;
			global::Xamarin.Forms.Platform.Resource.Styleable.ImageFilterView_warmth = global::pjsua2xamarin.Droid.Resource.Styleable.ImageFilterView_warmth;
			global::Xamarin.Forms.Platform.Resource.Styleable.include = global::pjsua2xamarin.Droid.Resource.Styleable.include;
			global::Xamarin.Forms.Platform.Resource.Styleable.include_constraintSet = global::pjsua2xamarin.Droid.Resource.Styleable.include_constraintSet;
			global::Xamarin.Forms.Platform.Resource.Styleable.Insets = global::pjsua2xamarin.Droid.Resource.Styleable.Insets;
			global::Xamarin.Forms.Platform.Resource.Styleable.Insets_paddingBottomSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Styleable.Insets_paddingBottomSystemWindowInsets;
			global::Xamarin.Forms.Platform.Resource.Styleable.Insets_paddingLeftSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Styleable.Insets_paddingLeftSystemWindowInsets;
			global::Xamarin.Forms.Platform.Resource.Styleable.Insets_paddingRightSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Styleable.Insets_paddingRightSystemWindowInsets;
			global::Xamarin.Forms.Platform.Resource.Styleable.Insets_paddingTopSystemWindowInsets = global::pjsua2xamarin.Droid.Resource.Styleable.Insets_paddingTopSystemWindowInsets;
			global::Xamarin.Forms.Platform.Resource.Styleable.ItemsViewRendererTheme = global::pjsua2xamarin.Droid.Resource.Styleable.ItemsViewRendererTheme;
			global::Xamarin.Forms.Platform.Resource.Styleable.ItemsViewRendererTheme_collectionViewStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ItemsViewRendererTheme_collectionViewStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyAttribute = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyAttribute_android_alpha = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_alpha;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyAttribute_android_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_elevation;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyAttribute_android_rotation = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_rotation;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyAttribute_android_rotationX = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_rotationX;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyAttribute_android_rotationY = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_rotationY;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyAttribute_android_scaleX = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_scaleX;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyAttribute_android_scaleY = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_scaleY;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyAttribute_android_transformPivotX = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_transformPivotX;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyAttribute_android_transformPivotY = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_transformPivotY;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyAttribute_android_translationX = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_translationX;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyAttribute_android_translationY = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_translationY;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyAttribute_android_translationZ = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_android_translationZ;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyAttribute_curveFit = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_curveFit;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyAttribute_framePosition = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_framePosition;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyAttribute_motionProgress = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_motionProgress;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyAttribute_motionTarget = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_motionTarget;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyAttribute_transformPivotTarget = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_transformPivotTarget;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyAttribute_transitionEasing = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_transitionEasing;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyAttribute_transitionPathRotate = global::pjsua2xamarin.Droid.Resource.Styleable.KeyAttribute_transitionPathRotate;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyCycle = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyCycle_android_alpha = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_android_alpha;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyCycle_android_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_android_elevation;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyCycle_android_rotation = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_android_rotation;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyCycle_android_rotationX = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_android_rotationX;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyCycle_android_rotationY = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_android_rotationY;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyCycle_android_scaleX = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_android_scaleX;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyCycle_android_scaleY = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_android_scaleY;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyCycle_android_translationX = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_android_translationX;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyCycle_android_translationY = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_android_translationY;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyCycle_android_translationZ = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_android_translationZ;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyCycle_curveFit = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_curveFit;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyCycle_framePosition = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_framePosition;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyCycle_motionProgress = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_motionProgress;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyCycle_motionTarget = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_motionTarget;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyCycle_transitionEasing = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_transitionEasing;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyCycle_transitionPathRotate = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_transitionPathRotate;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyCycle_waveOffset = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_waveOffset;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyCycle_wavePeriod = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_wavePeriod;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyCycle_wavePhase = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_wavePhase;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyCycle_waveShape = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_waveShape;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyCycle_waveVariesBy = global::pjsua2xamarin.Droid.Resource.Styleable.KeyCycle_waveVariesBy;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyFrame = global::pjsua2xamarin.Droid.Resource.Styleable.KeyFrame;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyFramesAcceleration = global::pjsua2xamarin.Droid.Resource.Styleable.KeyFramesAcceleration;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyFramesVelocity = global::pjsua2xamarin.Droid.Resource.Styleable.KeyFramesVelocity;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyPosition = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyPosition_curveFit = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_curveFit;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyPosition_drawPath = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_drawPath;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyPosition_framePosition = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_framePosition;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyPosition_keyPositionType = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_keyPositionType;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyPosition_motionTarget = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_motionTarget;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyPosition_pathMotionArc = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_pathMotionArc;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyPosition_percentHeight = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_percentHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyPosition_percentWidth = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_percentWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyPosition_percentX = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_percentX;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyPosition_percentY = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_percentY;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyPosition_sizePercent = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_sizePercent;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyPosition_transitionEasing = global::pjsua2xamarin.Droid.Resource.Styleable.KeyPosition_transitionEasing;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTimeCycle = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTimeCycle_android_alpha = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_android_alpha;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTimeCycle_android_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_android_elevation;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTimeCycle_android_rotation = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_android_rotation;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTimeCycle_android_rotationX = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_android_rotationX;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTimeCycle_android_rotationY = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_android_rotationY;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTimeCycle_android_scaleX = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_android_scaleX;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTimeCycle_android_scaleY = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_android_scaleY;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTimeCycle_android_translationX = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_android_translationX;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTimeCycle_android_translationY = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_android_translationY;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTimeCycle_android_translationZ = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_android_translationZ;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTimeCycle_curveFit = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_curveFit;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTimeCycle_framePosition = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_framePosition;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTimeCycle_motionProgress = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_motionProgress;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTimeCycle_motionTarget = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_motionTarget;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTimeCycle_transitionEasing = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_transitionEasing;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTimeCycle_transitionPathRotate = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_transitionPathRotate;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTimeCycle_waveDecay = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_waveDecay;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTimeCycle_waveOffset = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_waveOffset;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTimeCycle_wavePeriod = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_wavePeriod;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTimeCycle_wavePhase = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_wavePhase;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTimeCycle_waveShape = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTimeCycle_waveShape;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTrigger = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTrigger_framePosition = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_framePosition;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTrigger_motionTarget = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_motionTarget;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTrigger_motion_postLayoutCollision = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_motion_postLayoutCollision;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTrigger_motion_triggerOnCollision = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_motion_triggerOnCollision;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTrigger_onCross = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_onCross;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTrigger_onNegativeCross = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_onNegativeCross;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTrigger_onPositiveCross = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_onPositiveCross;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTrigger_triggerId = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_triggerId;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTrigger_triggerReceiver = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_triggerReceiver;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTrigger_triggerSlack = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_triggerSlack;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTrigger_viewTransitionOnCross = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_viewTransitionOnCross;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTrigger_viewTransitionOnNegativeCross = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_viewTransitionOnNegativeCross;
			global::Xamarin.Forms.Platform.Resource.Styleable.KeyTrigger_viewTransitionOnPositiveCross = global::pjsua2xamarin.Droid.Resource.Styleable.KeyTrigger_viewTransitionOnPositiveCross;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout = global::pjsua2xamarin.Droid.Resource.Styleable.Layout;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_android_layout_height = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_android_layout_height;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_android_layout_marginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_android_layout_marginBottom;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_android_layout_marginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_android_layout_marginEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_android_layout_marginLeft = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_android_layout_marginLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_android_layout_marginRight = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_android_layout_marginRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_android_layout_marginStart = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_android_layout_marginStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_android_layout_marginTop = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_android_layout_marginTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_android_layout_width = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_android_layout_width;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_android_orientation = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_android_orientation;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_barrierAllowsGoneWidgets = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_barrierAllowsGoneWidgets;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_barrierDirection = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_barrierDirection;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_barrierMargin = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_barrierMargin;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_chainUseRtl = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_chainUseRtl;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_constraint_referenced_ids = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_constraint_referenced_ids;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_constraint_referenced_tags = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_constraint_referenced_tags;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constrainedHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constrainedHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constrainedWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constrainedWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintBaseline_creator = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintBaseline_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintBaseline_toBaselineOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintBaseline_toBaselineOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintBaseline_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintBaseline_toBottomOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintBaseline_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintBaseline_toTopOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintBottom_creator = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintBottom_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintBottom_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintBottom_toBottomOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintBottom_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintBottom_toTopOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintCircle = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintCircle;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintCircleAngle = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintCircleAngle;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintCircleRadius = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintCircleRadius;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintDimensionRatio = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintDimensionRatio;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintEnd_toEndOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintEnd_toEndOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintEnd_toStartOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintEnd_toStartOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintGuide_begin = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintGuide_begin;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintGuide_end = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintGuide_end;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintGuide_percent = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintGuide_percent;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintHeight_default = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintHeight_default;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintHeight_max = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintHeight_max;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintHeight_min = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintHeight_min;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintHeight_percent = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintHeight_percent;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintHorizontal_bias = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintHorizontal_bias;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintHorizontal_chainStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintHorizontal_chainStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintHorizontal_weight = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintHorizontal_weight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintLeft_creator = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintLeft_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintLeft_toLeftOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintLeft_toLeftOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintLeft_toRightOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintLeft_toRightOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintRight_creator = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintRight_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintRight_toLeftOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintRight_toLeftOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintRight_toRightOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintRight_toRightOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintStart_toEndOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintStart_toEndOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintStart_toStartOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintStart_toStartOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintTop_creator = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintTop_creator;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintTop_toBottomOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintTop_toBottomOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintTop_toTopOf = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintTop_toTopOf;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintVertical_bias = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintVertical_bias;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintVertical_chainStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintVertical_chainStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintVertical_weight = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintVertical_weight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintWidth_default = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintWidth_default;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintWidth_max = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintWidth_max;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintWidth_min = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintWidth_min;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_constraintWidth_percent = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_constraintWidth_percent;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_editor_absoluteX = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_editor_absoluteX;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_editor_absoluteY = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_editor_absoluteY;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_goneMarginBaseline = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_goneMarginBaseline;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_goneMarginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_goneMarginBottom;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_goneMarginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_goneMarginEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_goneMarginLeft = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_goneMarginLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_goneMarginRight = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_goneMarginRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_goneMarginStart = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_goneMarginStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_goneMarginTop = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_goneMarginTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_marginBaseline = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_marginBaseline;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_layout_wrapBehaviorInParent = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_layout_wrapBehaviorInParent;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_maxHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_maxHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_maxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_maxWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_minHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_minHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Layout_minWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Layout_minWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.LinearLayoutCompat = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat;
			global::Xamarin.Forms.Platform.Resource.Styleable.LinearLayoutCompat_android_baselineAligned = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_android_baselineAligned;
			global::Xamarin.Forms.Platform.Resource.Styleable.LinearLayoutCompat_android_baselineAlignedChildIndex = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_android_baselineAlignedChildIndex;
			global::Xamarin.Forms.Platform.Resource.Styleable.LinearLayoutCompat_android_gravity = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_android_gravity;
			global::Xamarin.Forms.Platform.Resource.Styleable.LinearLayoutCompat_android_orientation = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_android_orientation;
			global::Xamarin.Forms.Platform.Resource.Styleable.LinearLayoutCompat_android_weightSum = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_android_weightSum;
			global::Xamarin.Forms.Platform.Resource.Styleable.LinearLayoutCompat_divider = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_divider;
			global::Xamarin.Forms.Platform.Resource.Styleable.LinearLayoutCompat_dividerPadding = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_dividerPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.LinearLayoutCompat_Layout = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_Layout;
			global::Xamarin.Forms.Platform.Resource.Styleable.LinearLayoutCompat_Layout_android_layout_gravity = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_Layout_android_layout_gravity;
			global::Xamarin.Forms.Platform.Resource.Styleable.LinearLayoutCompat_Layout_android_layout_height = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_Layout_android_layout_height;
			global::Xamarin.Forms.Platform.Resource.Styleable.LinearLayoutCompat_Layout_android_layout_weight = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_Layout_android_layout_weight;
			global::Xamarin.Forms.Platform.Resource.Styleable.LinearLayoutCompat_Layout_android_layout_width = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_Layout_android_layout_width;
			global::Xamarin.Forms.Platform.Resource.Styleable.LinearLayoutCompat_measureWithLargestChild = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_measureWithLargestChild;
			global::Xamarin.Forms.Platform.Resource.Styleable.LinearLayoutCompat_showDividers = global::pjsua2xamarin.Droid.Resource.Styleable.LinearLayoutCompat_showDividers;
			global::Xamarin.Forms.Platform.Resource.Styleable.LinearProgressIndicator = global::pjsua2xamarin.Droid.Resource.Styleable.LinearProgressIndicator;
			global::Xamarin.Forms.Platform.Resource.Styleable.LinearProgressIndicator_indeterminateAnimationType = global::pjsua2xamarin.Droid.Resource.Styleable.LinearProgressIndicator_indeterminateAnimationType;
			global::Xamarin.Forms.Platform.Resource.Styleable.LinearProgressIndicator_indicatorDirectionLinear = global::pjsua2xamarin.Droid.Resource.Styleable.LinearProgressIndicator_indicatorDirectionLinear;
			global::Xamarin.Forms.Platform.Resource.Styleable.ListPopupWindow = global::pjsua2xamarin.Droid.Resource.Styleable.ListPopupWindow;
			global::Xamarin.Forms.Platform.Resource.Styleable.ListPopupWindow_android_dropDownHorizontalOffset = global::pjsua2xamarin.Droid.Resource.Styleable.ListPopupWindow_android_dropDownHorizontalOffset;
			global::Xamarin.Forms.Platform.Resource.Styleable.ListPopupWindow_android_dropDownVerticalOffset = global::pjsua2xamarin.Droid.Resource.Styleable.ListPopupWindow_android_dropDownVerticalOffset;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialAlertDialog = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAlertDialog;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialAlertDialogTheme = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAlertDialogTheme;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialAlertDialogTheme_materialAlertDialogBodyTextStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAlertDialogTheme_materialAlertDialogBodyTextStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialAlertDialogTheme_materialAlertDialogTheme = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAlertDialogTheme_materialAlertDialogTheme;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialAlertDialogTheme_materialAlertDialogTitleIconStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAlertDialogTheme_materialAlertDialogTitleIconStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialAlertDialogTheme_materialAlertDialogTitlePanelStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAlertDialogTheme_materialAlertDialogTitlePanelStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialAlertDialogTheme_materialAlertDialogTitleTextStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAlertDialogTheme_materialAlertDialogTitleTextStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialAlertDialog_backgroundInsetBottom = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAlertDialog_backgroundInsetBottom;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialAlertDialog_backgroundInsetEnd = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAlertDialog_backgroundInsetEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialAlertDialog_backgroundInsetStart = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAlertDialog_backgroundInsetStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialAlertDialog_backgroundInsetTop = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAlertDialog_backgroundInsetTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialAutoCompleteTextView = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAutoCompleteTextView;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialAutoCompleteTextView_android_inputType = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialAutoCompleteTextView_android_inputType;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButton = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButtonToggleGroup = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButtonToggleGroup;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButtonToggleGroup_checkedButton = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButtonToggleGroup_checkedButton;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButtonToggleGroup_selectionRequired = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButtonToggleGroup_selectionRequired;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButtonToggleGroup_singleSelection = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButtonToggleGroup_singleSelection;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButton_android_background = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_android_background;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButton_android_checkable = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_android_checkable;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButton_android_insetBottom = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_android_insetBottom;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButton_android_insetLeft = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_android_insetLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButton_android_insetRight = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_android_insetRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButton_android_insetTop = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_android_insetTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButton_backgroundTint = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_backgroundTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButton_backgroundTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_backgroundTintMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButton_cornerRadius = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_cornerRadius;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButton_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_elevation;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButton_icon = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_icon;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButton_iconGravity = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_iconGravity;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButton_iconPadding = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_iconPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButton_iconSize = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_iconSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButton_iconTint = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_iconTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButton_iconTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_iconTintMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButton_rippleColor = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_rippleColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButton_shapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_shapeAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButton_shapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_shapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButton_strokeColor = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_strokeColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialButton_strokeWidth = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialButton_strokeWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCalendar = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendar;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCalendarItem = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendarItem;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCalendarItem_android_insetBottom = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendarItem_android_insetBottom;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCalendarItem_android_insetLeft = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendarItem_android_insetLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCalendarItem_android_insetRight = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendarItem_android_insetRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCalendarItem_android_insetTop = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendarItem_android_insetTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCalendarItem_itemFillColor = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendarItem_itemFillColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCalendarItem_itemShapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendarItem_itemShapeAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCalendarItem_itemShapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendarItem_itemShapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCalendarItem_itemStrokeColor = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendarItem_itemStrokeColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCalendarItem_itemStrokeWidth = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendarItem_itemStrokeWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCalendarItem_itemTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendarItem_itemTextColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCalendar_android_windowFullscreen = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendar_android_windowFullscreen;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCalendar_dayInvalidStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendar_dayInvalidStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCalendar_daySelectedStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendar_daySelectedStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCalendar_dayStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendar_dayStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCalendar_dayTodayStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendar_dayTodayStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCalendar_nestedScrollable = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendar_nestedScrollable;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCalendar_rangeFillColor = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendar_rangeFillColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCalendar_yearSelectedStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendar_yearSelectedStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCalendar_yearStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendar_yearStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCalendar_yearTodayStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCalendar_yearTodayStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCardView = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCardView_android_checkable = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_android_checkable;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCardView_cardForegroundColor = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_cardForegroundColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCardView_checkedIcon = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_checkedIcon;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCardView_checkedIconMargin = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_checkedIconMargin;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCardView_checkedIconSize = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_checkedIconSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCardView_checkedIconTint = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_checkedIconTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCardView_rippleColor = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_rippleColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCardView_shapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_shapeAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCardView_shapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_shapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCardView_state_dragged = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_state_dragged;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCardView_strokeColor = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_strokeColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCardView_strokeWidth = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCardView_strokeWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCheckBox = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCheckBox;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCheckBox_buttonTint = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCheckBox_buttonTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialCheckBox_useMaterialThemeColors = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialCheckBox_useMaterialThemeColors;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialRadioButton = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialRadioButton;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialRadioButton_buttonTint = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialRadioButton_buttonTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialRadioButton_useMaterialThemeColors = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialRadioButton_useMaterialThemeColors;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialShape = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialShape;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialShape_shapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialShape_shapeAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialShape_shapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialShape_shapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialTextAppearance_android_letterSpacing = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialTextAppearance_android_letterSpacing;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialTextAppearance_android_lineHeight = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialTextAppearance_android_lineHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialTextAppearance_lineHeight = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialTextAppearance_lineHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialTextView = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialTextView;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialTextView_android_lineHeight = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialTextView_android_lineHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialTextView_android_textAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialTextView_android_textAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialTextView_lineHeight = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialTextView_lineHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialTimePicker = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialTimePicker;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialTimePicker_clockIcon = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialTimePicker_clockIcon;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialTimePicker_keyboardIcon = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialTimePicker_keyboardIcon;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialToolbar = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialToolbar;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialToolbar_navigationIconTint = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialToolbar_navigationIconTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialToolbar_subtitleCentered = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialToolbar_subtitleCentered;
			global::Xamarin.Forms.Platform.Resource.Styleable.MaterialToolbar_titleCentered = global::pjsua2xamarin.Droid.Resource.Styleable.MaterialToolbar_titleCentered;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuGroup = global::pjsua2xamarin.Droid.Resource.Styleable.MenuGroup;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuGroup_android_checkableBehavior = global::pjsua2xamarin.Droid.Resource.Styleable.MenuGroup_android_checkableBehavior;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuGroup_android_enabled = global::pjsua2xamarin.Droid.Resource.Styleable.MenuGroup_android_enabled;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuGroup_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.MenuGroup_android_id;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuGroup_android_menuCategory = global::pjsua2xamarin.Droid.Resource.Styleable.MenuGroup_android_menuCategory;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuGroup_android_orderInCategory = global::pjsua2xamarin.Droid.Resource.Styleable.MenuGroup_android_orderInCategory;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuGroup_android_visible = global::pjsua2xamarin.Droid.Resource.Styleable.MenuGroup_android_visible;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_actionLayout = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_actionLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_actionProviderClass = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_actionProviderClass;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_actionViewClass = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_actionViewClass;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_alphabeticModifiers = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_alphabeticModifiers;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_android_alphabeticShortcut = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_alphabeticShortcut;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_android_checkable = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_checkable;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_android_checked = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_checked;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_android_enabled = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_enabled;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_android_icon = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_icon;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_id;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_android_menuCategory = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_menuCategory;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_android_numericShortcut = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_numericShortcut;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_android_onClick = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_onClick;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_android_orderInCategory = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_orderInCategory;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_android_title = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_title;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_android_titleCondensed = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_titleCondensed;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_android_visible = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_android_visible;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_contentDescription = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_contentDescription;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_iconTint = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_iconTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_iconTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_iconTintMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_numericModifiers = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_numericModifiers;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_showAsAction = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_showAsAction;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuItem_tooltipText = global::pjsua2xamarin.Droid.Resource.Styleable.MenuItem_tooltipText;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuView = global::pjsua2xamarin.Droid.Resource.Styleable.MenuView;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuView_android_headerBackground = global::pjsua2xamarin.Droid.Resource.Styleable.MenuView_android_headerBackground;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuView_android_horizontalDivider = global::pjsua2xamarin.Droid.Resource.Styleable.MenuView_android_horizontalDivider;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuView_android_itemBackground = global::pjsua2xamarin.Droid.Resource.Styleable.MenuView_android_itemBackground;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuView_android_itemIconDisabledAlpha = global::pjsua2xamarin.Droid.Resource.Styleable.MenuView_android_itemIconDisabledAlpha;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuView_android_itemTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.MenuView_android_itemTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuView_android_verticalDivider = global::pjsua2xamarin.Droid.Resource.Styleable.MenuView_android_verticalDivider;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuView_android_windowAnimationStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MenuView_android_windowAnimationStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuView_preserveIconSpacing = global::pjsua2xamarin.Droid.Resource.Styleable.MenuView_preserveIconSpacing;
			global::Xamarin.Forms.Platform.Resource.Styleable.MenuView_subMenuArrow = global::pjsua2xamarin.Droid.Resource.Styleable.MenuView_subMenuArrow;
			global::Xamarin.Forms.Platform.Resource.Styleable.MockView = global::pjsua2xamarin.Droid.Resource.Styleable.MockView;
			global::Xamarin.Forms.Platform.Resource.Styleable.MockView_mock_diagonalsColor = global::pjsua2xamarin.Droid.Resource.Styleable.MockView_mock_diagonalsColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.MockView_mock_label = global::pjsua2xamarin.Droid.Resource.Styleable.MockView_mock_label;
			global::Xamarin.Forms.Platform.Resource.Styleable.MockView_mock_labelBackgroundColor = global::pjsua2xamarin.Droid.Resource.Styleable.MockView_mock_labelBackgroundColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.MockView_mock_labelColor = global::pjsua2xamarin.Droid.Resource.Styleable.MockView_mock_labelColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.MockView_mock_showDiagonals = global::pjsua2xamarin.Droid.Resource.Styleable.MockView_mock_showDiagonals;
			global::Xamarin.Forms.Platform.Resource.Styleable.MockView_mock_showLabel = global::pjsua2xamarin.Droid.Resource.Styleable.MockView_mock_showLabel;
			global::Xamarin.Forms.Platform.Resource.Styleable.Motion = global::pjsua2xamarin.Droid.Resource.Styleable.Motion;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionEffect = global::pjsua2xamarin.Droid.Resource.Styleable.MotionEffect;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionEffect_motionEffect_alpha = global::pjsua2xamarin.Droid.Resource.Styleable.MotionEffect_motionEffect_alpha;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionEffect_motionEffect_end = global::pjsua2xamarin.Droid.Resource.Styleable.MotionEffect_motionEffect_end;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionEffect_motionEffect_move = global::pjsua2xamarin.Droid.Resource.Styleable.MotionEffect_motionEffect_move;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionEffect_motionEffect_start = global::pjsua2xamarin.Droid.Resource.Styleable.MotionEffect_motionEffect_start;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionEffect_motionEffect_strict = global::pjsua2xamarin.Droid.Resource.Styleable.MotionEffect_motionEffect_strict;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionEffect_motionEffect_translationX = global::pjsua2xamarin.Droid.Resource.Styleable.MotionEffect_motionEffect_translationX;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionEffect_motionEffect_translationY = global::pjsua2xamarin.Droid.Resource.Styleable.MotionEffect_motionEffect_translationY;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionEffect_motionEffect_viewTransition = global::pjsua2xamarin.Droid.Resource.Styleable.MotionEffect_motionEffect_viewTransition;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionHelper = global::pjsua2xamarin.Droid.Resource.Styleable.MotionHelper;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionHelper_onHide = global::pjsua2xamarin.Droid.Resource.Styleable.MotionHelper_onHide;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionHelper_onShow = global::pjsua2xamarin.Droid.Resource.Styleable.MotionHelper_onShow;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_android_autoSizeTextType = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_android_autoSizeTextType;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_android_fontFamily = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_android_fontFamily;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_android_gravity = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_android_gravity;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_android_shadowRadius = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_android_shadowRadius;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_android_text = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_android_text;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_android_textColor = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_android_textColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_android_textSize = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_android_textSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_android_textStyle = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_android_textStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_android_typeface = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_android_typeface;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_borderRound = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_borderRound;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_borderRoundPercent = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_borderRoundPercent;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_scaleFromTextSize = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_scaleFromTextSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_textBackground = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textBackground;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_textBackgroundPanX = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textBackgroundPanX;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_textBackgroundPanY = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textBackgroundPanY;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_textBackgroundRotate = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textBackgroundRotate;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_textBackgroundZoom = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textBackgroundZoom;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_textOutlineColor = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textOutlineColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_textOutlineThickness = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textOutlineThickness;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_textPanX = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textPanX;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_textPanY = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textPanY;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_textureBlurFactor = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textureBlurFactor;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_textureEffect = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textureEffect;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_textureHeight = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textureHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLabel_textureWidth = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLabel_textureWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLayout = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLayout_applyMotionScene = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLayout_applyMotionScene;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLayout_currentState = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLayout_currentState;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLayout_layoutDescription = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLayout_layoutDescription;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLayout_motionDebug = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLayout_motionDebug;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLayout_motionProgress = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLayout_motionProgress;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionLayout_showPaths = global::pjsua2xamarin.Droid.Resource.Styleable.MotionLayout_showPaths;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionScene = global::pjsua2xamarin.Droid.Resource.Styleable.MotionScene;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionScene_defaultDuration = global::pjsua2xamarin.Droid.Resource.Styleable.MotionScene_defaultDuration;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionScene_layoutDuringTransition = global::pjsua2xamarin.Droid.Resource.Styleable.MotionScene_layoutDuringTransition;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionTelltales = global::pjsua2xamarin.Droid.Resource.Styleable.MotionTelltales;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionTelltales_telltales_tailColor = global::pjsua2xamarin.Droid.Resource.Styleable.MotionTelltales_telltales_tailColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionTelltales_telltales_tailScale = global::pjsua2xamarin.Droid.Resource.Styleable.MotionTelltales_telltales_tailScale;
			global::Xamarin.Forms.Platform.Resource.Styleable.MotionTelltales_telltales_velocityMode = global::pjsua2xamarin.Droid.Resource.Styleable.MotionTelltales_telltales_velocityMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.Motion_animateCircleAngleTo = global::pjsua2xamarin.Droid.Resource.Styleable.Motion_animateCircleAngleTo;
			global::Xamarin.Forms.Platform.Resource.Styleable.Motion_animateRelativeTo = global::pjsua2xamarin.Droid.Resource.Styleable.Motion_animateRelativeTo;
			global::Xamarin.Forms.Platform.Resource.Styleable.Motion_drawPath = global::pjsua2xamarin.Droid.Resource.Styleable.Motion_drawPath;
			global::Xamarin.Forms.Platform.Resource.Styleable.Motion_motionPathRotate = global::pjsua2xamarin.Droid.Resource.Styleable.Motion_motionPathRotate;
			global::Xamarin.Forms.Platform.Resource.Styleable.Motion_motionStagger = global::pjsua2xamarin.Droid.Resource.Styleable.Motion_motionStagger;
			global::Xamarin.Forms.Platform.Resource.Styleable.Motion_pathMotionArc = global::pjsua2xamarin.Droid.Resource.Styleable.Motion_pathMotionArc;
			global::Xamarin.Forms.Platform.Resource.Styleable.Motion_quantizeMotionInterpolator = global::pjsua2xamarin.Droid.Resource.Styleable.Motion_quantizeMotionInterpolator;
			global::Xamarin.Forms.Platform.Resource.Styleable.Motion_quantizeMotionPhase = global::pjsua2xamarin.Droid.Resource.Styleable.Motion_quantizeMotionPhase;
			global::Xamarin.Forms.Platform.Resource.Styleable.Motion_quantizeMotionSteps = global::pjsua2xamarin.Droid.Resource.Styleable.Motion_quantizeMotionSteps;
			global::Xamarin.Forms.Platform.Resource.Styleable.Motion_transitionEasing = global::pjsua2xamarin.Droid.Resource.Styleable.Motion_transitionEasing;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavAction = global::pjsua2xamarin.Droid.Resource.Styleable.NavAction;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavAction_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.NavAction_android_id;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavAction_destination = global::pjsua2xamarin.Droid.Resource.Styleable.NavAction_destination;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavAction_enterAnim = global::pjsua2xamarin.Droid.Resource.Styleable.NavAction_enterAnim;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavAction_exitAnim = global::pjsua2xamarin.Droid.Resource.Styleable.NavAction_exitAnim;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavAction_launchSingleTop = global::pjsua2xamarin.Droid.Resource.Styleable.NavAction_launchSingleTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavAction_popEnterAnim = global::pjsua2xamarin.Droid.Resource.Styleable.NavAction_popEnterAnim;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavAction_popExitAnim = global::pjsua2xamarin.Droid.Resource.Styleable.NavAction_popExitAnim;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavAction_popUpTo = global::pjsua2xamarin.Droid.Resource.Styleable.NavAction_popUpTo;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavAction_popUpToInclusive = global::pjsua2xamarin.Droid.Resource.Styleable.NavAction_popUpToInclusive;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavArgument = global::pjsua2xamarin.Droid.Resource.Styleable.NavArgument;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavArgument_android_defaultValue = global::pjsua2xamarin.Droid.Resource.Styleable.NavArgument_android_defaultValue;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavArgument_android_name = global::pjsua2xamarin.Droid.Resource.Styleable.NavArgument_android_name;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavArgument_argType = global::pjsua2xamarin.Droid.Resource.Styleable.NavArgument_argType;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavArgument_nullable = global::pjsua2xamarin.Droid.Resource.Styleable.NavArgument_nullable;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavDeepLink = global::pjsua2xamarin.Droid.Resource.Styleable.NavDeepLink;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavDeepLink_action = global::pjsua2xamarin.Droid.Resource.Styleable.NavDeepLink_action;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavDeepLink_android_autoVerify = global::pjsua2xamarin.Droid.Resource.Styleable.NavDeepLink_android_autoVerify;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavDeepLink_mimeType = global::pjsua2xamarin.Droid.Resource.Styleable.NavDeepLink_mimeType;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavDeepLink_uri = global::pjsua2xamarin.Droid.Resource.Styleable.NavDeepLink_uri;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavGraphNavigator = global::pjsua2xamarin.Droid.Resource.Styleable.NavGraphNavigator;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavGraphNavigator_startDestination = global::pjsua2xamarin.Droid.Resource.Styleable.NavGraphNavigator_startDestination;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavHost = global::pjsua2xamarin.Droid.Resource.Styleable.NavHost;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavHost_navGraph = global::pjsua2xamarin.Droid.Resource.Styleable.NavHost_navGraph;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationBarView = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationBarView_backgroundTint = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView_backgroundTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationBarView_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView_elevation;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationBarView_itemBackground = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView_itemBackground;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationBarView_itemIconSize = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView_itemIconSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationBarView_itemIconTint = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView_itemIconTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationBarView_itemRippleColor = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView_itemRippleColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationBarView_itemTextAppearanceActive = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView_itemTextAppearanceActive;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationBarView_itemTextAppearanceInactive = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView_itemTextAppearanceInactive;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationBarView_itemTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView_itemTextColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationBarView_labelVisibilityMode = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView_labelVisibilityMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationBarView_menu = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationBarView_menu;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationRailView = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationRailView;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationRailView_headerLayout = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationRailView_headerLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationRailView_menuGravity = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationRailView_menuGravity;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_android_background = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_android_background;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_android_fitsSystemWindows = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_android_fitsSystemWindows;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_android_maxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_android_maxWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_elevation;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_headerLayout = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_headerLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_itemBackground = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemBackground;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_itemHorizontalPadding = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemHorizontalPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_itemIconPadding = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemIconPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_itemIconSize = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemIconSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_itemIconTint = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemIconTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_itemMaxLines = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemMaxLines;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_itemShapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemShapeAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_itemShapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemShapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_itemShapeFillColor = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemShapeFillColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_itemShapeInsetBottom = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemShapeInsetBottom;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_itemShapeInsetEnd = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemShapeInsetEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_itemShapeInsetStart = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemShapeInsetStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_itemShapeInsetTop = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemShapeInsetTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_itemTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_itemTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_itemTextColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_menu = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_menu;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_shapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_shapeAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavigationView_shapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.NavigationView_shapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Resource.Styleable.Navigator = global::pjsua2xamarin.Droid.Resource.Styleable.Navigator;
			global::Xamarin.Forms.Platform.Resource.Styleable.Navigator_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.Navigator_android_id;
			global::Xamarin.Forms.Platform.Resource.Styleable.Navigator_android_label = global::pjsua2xamarin.Droid.Resource.Styleable.Navigator_android_label;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavInclude = global::pjsua2xamarin.Droid.Resource.Styleable.NavInclude;
			global::Xamarin.Forms.Platform.Resource.Styleable.NavInclude_graph = global::pjsua2xamarin.Droid.Resource.Styleable.NavInclude_graph;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnClick = global::pjsua2xamarin.Droid.Resource.Styleable.OnClick;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnClick_clickAction = global::pjsua2xamarin.Droid.Resource.Styleable.OnClick_clickAction;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnClick_targetId = global::pjsua2xamarin.Droid.Resource.Styleable.OnClick_targetId;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnSwipe = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnSwipe_autoCompleteMode = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_autoCompleteMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnSwipe_dragDirection = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_dragDirection;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnSwipe_dragScale = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_dragScale;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnSwipe_dragThreshold = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_dragThreshold;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnSwipe_limitBoundsTo = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_limitBoundsTo;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnSwipe_maxAcceleration = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_maxAcceleration;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnSwipe_maxVelocity = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_maxVelocity;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnSwipe_moveWhenScrollAtTop = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_moveWhenScrollAtTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnSwipe_nestedScrollFlags = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_nestedScrollFlags;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnSwipe_onTouchUp = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_onTouchUp;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnSwipe_rotationCenterId = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_rotationCenterId;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnSwipe_springBoundary = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_springBoundary;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnSwipe_springDamping = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_springDamping;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnSwipe_springMass = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_springMass;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnSwipe_springStiffness = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_springStiffness;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnSwipe_springStopThreshold = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_springStopThreshold;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnSwipe_touchAnchorId = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_touchAnchorId;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnSwipe_touchAnchorSide = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_touchAnchorSide;
			global::Xamarin.Forms.Platform.Resource.Styleable.OnSwipe_touchRegionId = global::pjsua2xamarin.Droid.Resource.Styleable.OnSwipe_touchRegionId;
			global::Xamarin.Forms.Platform.Resource.Styleable.PopupWindow = global::pjsua2xamarin.Droid.Resource.Styleable.PopupWindow;
			global::Xamarin.Forms.Platform.Resource.Styleable.PopupWindowBackgroundState = global::pjsua2xamarin.Droid.Resource.Styleable.PopupWindowBackgroundState;
			global::Xamarin.Forms.Platform.Resource.Styleable.PopupWindowBackgroundState_state_above_anchor = global::pjsua2xamarin.Droid.Resource.Styleable.PopupWindowBackgroundState_state_above_anchor;
			global::Xamarin.Forms.Platform.Resource.Styleable.PopupWindow_android_popupAnimationStyle = global::pjsua2xamarin.Droid.Resource.Styleable.PopupWindow_android_popupAnimationStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.PopupWindow_android_popupBackground = global::pjsua2xamarin.Droid.Resource.Styleable.PopupWindow_android_popupBackground;
			global::Xamarin.Forms.Platform.Resource.Styleable.PopupWindow_overlapAnchor = global::pjsua2xamarin.Droid.Resource.Styleable.PopupWindow_overlapAnchor;
			global::Xamarin.Forms.Platform.Resource.Styleable.PropertySet = global::pjsua2xamarin.Droid.Resource.Styleable.PropertySet;
			global::Xamarin.Forms.Platform.Resource.Styleable.PropertySet_android_alpha = global::pjsua2xamarin.Droid.Resource.Styleable.PropertySet_android_alpha;
			global::Xamarin.Forms.Platform.Resource.Styleable.PropertySet_android_visibility = global::pjsua2xamarin.Droid.Resource.Styleable.PropertySet_android_visibility;
			global::Xamarin.Forms.Platform.Resource.Styleable.PropertySet_layout_constraintTag = global::pjsua2xamarin.Droid.Resource.Styleable.PropertySet_layout_constraintTag;
			global::Xamarin.Forms.Platform.Resource.Styleable.PropertySet_motionProgress = global::pjsua2xamarin.Droid.Resource.Styleable.PropertySet_motionProgress;
			global::Xamarin.Forms.Platform.Resource.Styleable.PropertySet_visibilityMode = global::pjsua2xamarin.Droid.Resource.Styleable.PropertySet_visibilityMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.RadialViewGroup = global::pjsua2xamarin.Droid.Resource.Styleable.RadialViewGroup;
			global::Xamarin.Forms.Platform.Resource.Styleable.RadialViewGroup_materialCircleRadius = global::pjsua2xamarin.Droid.Resource.Styleable.RadialViewGroup_materialCircleRadius;
			global::Xamarin.Forms.Platform.Resource.Styleable.RangeSlider = global::pjsua2xamarin.Droid.Resource.Styleable.RangeSlider;
			global::Xamarin.Forms.Platform.Resource.Styleable.RangeSlider_minSeparation = global::pjsua2xamarin.Droid.Resource.Styleable.RangeSlider_minSeparation;
			global::Xamarin.Forms.Platform.Resource.Styleable.RangeSlider_values = global::pjsua2xamarin.Droid.Resource.Styleable.RangeSlider_values;
			global::Xamarin.Forms.Platform.Resource.Styleable.RecycleListView = global::pjsua2xamarin.Droid.Resource.Styleable.RecycleListView;
			global::Xamarin.Forms.Platform.Resource.Styleable.RecycleListView_paddingBottomNoButtons = global::pjsua2xamarin.Droid.Resource.Styleable.RecycleListView_paddingBottomNoButtons;
			global::Xamarin.Forms.Platform.Resource.Styleable.RecycleListView_paddingTopNoTitle = global::pjsua2xamarin.Droid.Resource.Styleable.RecycleListView_paddingTopNoTitle;
			global::Xamarin.Forms.Platform.Resource.Styleable.RecyclerView = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView;
			global::Xamarin.Forms.Platform.Resource.Styleable.RecyclerView_android_clipToPadding = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_android_clipToPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.RecyclerView_android_descendantFocusability = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_android_descendantFocusability;
			global::Xamarin.Forms.Platform.Resource.Styleable.RecyclerView_android_orientation = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_android_orientation;
			global::Xamarin.Forms.Platform.Resource.Styleable.RecyclerView_fastScrollEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_fastScrollEnabled;
			global::Xamarin.Forms.Platform.Resource.Styleable.RecyclerView_fastScrollHorizontalThumbDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_fastScrollHorizontalThumbDrawable;
			global::Xamarin.Forms.Platform.Resource.Styleable.RecyclerView_fastScrollHorizontalTrackDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_fastScrollHorizontalTrackDrawable;
			global::Xamarin.Forms.Platform.Resource.Styleable.RecyclerView_fastScrollVerticalThumbDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_fastScrollVerticalThumbDrawable;
			global::Xamarin.Forms.Platform.Resource.Styleable.RecyclerView_fastScrollVerticalTrackDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_fastScrollVerticalTrackDrawable;
			global::Xamarin.Forms.Platform.Resource.Styleable.RecyclerView_layoutManager = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_layoutManager;
			global::Xamarin.Forms.Platform.Resource.Styleable.RecyclerView_reverseLayout = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_reverseLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.RecyclerView_spanCount = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_spanCount;
			global::Xamarin.Forms.Platform.Resource.Styleable.RecyclerView_stackFromEnd = global::pjsua2xamarin.Droid.Resource.Styleable.RecyclerView_stackFromEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.ScrimInsetsFrameLayout = global::pjsua2xamarin.Droid.Resource.Styleable.ScrimInsetsFrameLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.ScrimInsetsFrameLayout_insetForeground = global::pjsua2xamarin.Droid.Resource.Styleable.ScrimInsetsFrameLayout_insetForeground;
			global::Xamarin.Forms.Platform.Resource.Styleable.ScrollingViewBehavior_Layout = global::pjsua2xamarin.Droid.Resource.Styleable.ScrollingViewBehavior_Layout;
			global::Xamarin.Forms.Platform.Resource.Styleable.ScrollingViewBehavior_Layout_behavior_overlapTop = global::pjsua2xamarin.Droid.Resource.Styleable.ScrollingViewBehavior_Layout_behavior_overlapTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.ScrollViewRendererTheme = global::pjsua2xamarin.Droid.Resource.Styleable.ScrollViewRendererTheme;
			global::Xamarin.Forms.Platform.Resource.Styleable.ScrollViewRendererTheme_scrollViewStyle = global::pjsua2xamarin.Droid.Resource.Styleable.ScrollViewRendererTheme_scrollViewStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.SearchView = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView;
			global::Xamarin.Forms.Platform.Resource.Styleable.SearchView_android_focusable = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_android_focusable;
			global::Xamarin.Forms.Platform.Resource.Styleable.SearchView_android_imeOptions = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_android_imeOptions;
			global::Xamarin.Forms.Platform.Resource.Styleable.SearchView_android_inputType = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_android_inputType;
			global::Xamarin.Forms.Platform.Resource.Styleable.SearchView_android_maxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_android_maxWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.SearchView_closeIcon = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_closeIcon;
			global::Xamarin.Forms.Platform.Resource.Styleable.SearchView_commitIcon = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_commitIcon;
			global::Xamarin.Forms.Platform.Resource.Styleable.SearchView_defaultQueryHint = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_defaultQueryHint;
			global::Xamarin.Forms.Platform.Resource.Styleable.SearchView_goIcon = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_goIcon;
			global::Xamarin.Forms.Platform.Resource.Styleable.SearchView_iconifiedByDefault = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_iconifiedByDefault;
			global::Xamarin.Forms.Platform.Resource.Styleable.SearchView_layout = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_layout;
			global::Xamarin.Forms.Platform.Resource.Styleable.SearchView_queryBackground = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_queryBackground;
			global::Xamarin.Forms.Platform.Resource.Styleable.SearchView_queryHint = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_queryHint;
			global::Xamarin.Forms.Platform.Resource.Styleable.SearchView_searchHintIcon = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_searchHintIcon;
			global::Xamarin.Forms.Platform.Resource.Styleable.SearchView_searchIcon = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_searchIcon;
			global::Xamarin.Forms.Platform.Resource.Styleable.SearchView_submitBackground = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_submitBackground;
			global::Xamarin.Forms.Platform.Resource.Styleable.SearchView_suggestionRowLayout = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_suggestionRowLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.SearchView_voiceIcon = global::pjsua2xamarin.Droid.Resource.Styleable.SearchView_voiceIcon;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeableImageView = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeableImageView_contentPadding = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView_contentPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeableImageView_contentPaddingBottom = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView_contentPaddingBottom;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeableImageView_contentPaddingEnd = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView_contentPaddingEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeableImageView_contentPaddingLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView_contentPaddingLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeableImageView_contentPaddingRight = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView_contentPaddingRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeableImageView_contentPaddingStart = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView_contentPaddingStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeableImageView_contentPaddingTop = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView_contentPaddingTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeableImageView_shapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView_shapeAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeableImageView_shapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView_shapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeableImageView_strokeColor = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView_strokeColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeableImageView_strokeWidth = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeableImageView_strokeWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeAppearance_cornerFamily = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeAppearance_cornerFamily;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeAppearance_cornerFamilyBottomLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeAppearance_cornerFamilyBottomLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeAppearance_cornerFamilyBottomRight = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeAppearance_cornerFamilyBottomRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeAppearance_cornerFamilyTopLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeAppearance_cornerFamilyTopLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeAppearance_cornerFamilyTopRight = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeAppearance_cornerFamilyTopRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeAppearance_cornerSize = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeAppearance_cornerSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeAppearance_cornerSizeBottomLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeAppearance_cornerSizeBottomLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeAppearance_cornerSizeBottomRight = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeAppearance_cornerSizeBottomRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeAppearance_cornerSizeTopLeft = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeAppearance_cornerSizeTopLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.ShapeAppearance_cornerSizeTopRight = global::pjsua2xamarin.Droid.Resource.Styleable.ShapeAppearance_cornerSizeTopRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider = global::pjsua2xamarin.Droid.Resource.Styleable.Slider;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider_android_enabled = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_android_enabled;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider_android_stepSize = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_android_stepSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider_android_value = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_android_value;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider_android_valueFrom = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_android_valueFrom;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider_android_valueTo = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_android_valueTo;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider_haloColor = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_haloColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider_haloRadius = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_haloRadius;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider_labelBehavior = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_labelBehavior;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider_labelStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_labelStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider_thumbColor = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_thumbColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider_thumbElevation = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_thumbElevation;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider_thumbRadius = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_thumbRadius;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider_thumbStrokeColor = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_thumbStrokeColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider_thumbStrokeWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_thumbStrokeWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider_tickColor = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_tickColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider_tickColorActive = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_tickColorActive;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider_tickColorInactive = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_tickColorInactive;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider_tickVisible = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_tickVisible;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider_trackColor = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_trackColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider_trackColorActive = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_trackColorActive;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider_trackColorInactive = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_trackColorInactive;
			global::Xamarin.Forms.Platform.Resource.Styleable.Slider_trackHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Slider_trackHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Snackbar = global::pjsua2xamarin.Droid.Resource.Styleable.Snackbar;
			global::Xamarin.Forms.Platform.Resource.Styleable.SnackbarLayout = global::pjsua2xamarin.Droid.Resource.Styleable.SnackbarLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.SnackbarLayout_actionTextColorAlpha = global::pjsua2xamarin.Droid.Resource.Styleable.SnackbarLayout_actionTextColorAlpha;
			global::Xamarin.Forms.Platform.Resource.Styleable.SnackbarLayout_android_maxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.SnackbarLayout_android_maxWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.SnackbarLayout_animationMode = global::pjsua2xamarin.Droid.Resource.Styleable.SnackbarLayout_animationMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.SnackbarLayout_backgroundOverlayColorAlpha = global::pjsua2xamarin.Droid.Resource.Styleable.SnackbarLayout_backgroundOverlayColorAlpha;
			global::Xamarin.Forms.Platform.Resource.Styleable.SnackbarLayout_backgroundTint = global::pjsua2xamarin.Droid.Resource.Styleable.SnackbarLayout_backgroundTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.SnackbarLayout_backgroundTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.SnackbarLayout_backgroundTintMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.SnackbarLayout_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.SnackbarLayout_elevation;
			global::Xamarin.Forms.Platform.Resource.Styleable.SnackbarLayout_maxActionInlineWidth = global::pjsua2xamarin.Droid.Resource.Styleable.SnackbarLayout_maxActionInlineWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.Snackbar_snackbarButtonStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Snackbar_snackbarButtonStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.Snackbar_snackbarStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Snackbar_snackbarStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.Snackbar_snackbarTextViewStyle = global::pjsua2xamarin.Droid.Resource.Styleable.Snackbar_snackbarTextViewStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.Spinner = global::pjsua2xamarin.Droid.Resource.Styleable.Spinner;
			global::Xamarin.Forms.Platform.Resource.Styleable.Spinner_android_dropDownWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Spinner_android_dropDownWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.Spinner_android_entries = global::pjsua2xamarin.Droid.Resource.Styleable.Spinner_android_entries;
			global::Xamarin.Forms.Platform.Resource.Styleable.Spinner_android_popupBackground = global::pjsua2xamarin.Droid.Resource.Styleable.Spinner_android_popupBackground;
			global::Xamarin.Forms.Platform.Resource.Styleable.Spinner_android_prompt = global::pjsua2xamarin.Droid.Resource.Styleable.Spinner_android_prompt;
			global::Xamarin.Forms.Platform.Resource.Styleable.Spinner_popupTheme = global::pjsua2xamarin.Droid.Resource.Styleable.Spinner_popupTheme;
			global::Xamarin.Forms.Platform.Resource.Styleable.State = global::pjsua2xamarin.Droid.Resource.Styleable.State;
			global::Xamarin.Forms.Platform.Resource.Styleable.StateListDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.StateListDrawable;
			global::Xamarin.Forms.Platform.Resource.Styleable.StateListDrawableItem = global::pjsua2xamarin.Droid.Resource.Styleable.StateListDrawableItem;
			global::Xamarin.Forms.Platform.Resource.Styleable.StateListDrawableItem_android_drawable = global::pjsua2xamarin.Droid.Resource.Styleable.StateListDrawableItem_android_drawable;
			global::Xamarin.Forms.Platform.Resource.Styleable.StateListDrawable_android_constantSize = global::pjsua2xamarin.Droid.Resource.Styleable.StateListDrawable_android_constantSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.StateListDrawable_android_dither = global::pjsua2xamarin.Droid.Resource.Styleable.StateListDrawable_android_dither;
			global::Xamarin.Forms.Platform.Resource.Styleable.StateListDrawable_android_enterFadeDuration = global::pjsua2xamarin.Droid.Resource.Styleable.StateListDrawable_android_enterFadeDuration;
			global::Xamarin.Forms.Platform.Resource.Styleable.StateListDrawable_android_exitFadeDuration = global::pjsua2xamarin.Droid.Resource.Styleable.StateListDrawable_android_exitFadeDuration;
			global::Xamarin.Forms.Platform.Resource.Styleable.StateListDrawable_android_variablePadding = global::pjsua2xamarin.Droid.Resource.Styleable.StateListDrawable_android_variablePadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.StateListDrawable_android_visible = global::pjsua2xamarin.Droid.Resource.Styleable.StateListDrawable_android_visible;
			global::Xamarin.Forms.Platform.Resource.Styleable.StateSet = global::pjsua2xamarin.Droid.Resource.Styleable.StateSet;
			global::Xamarin.Forms.Platform.Resource.Styleable.StateSet_defaultState = global::pjsua2xamarin.Droid.Resource.Styleable.StateSet_defaultState;
			global::Xamarin.Forms.Platform.Resource.Styleable.State_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.State_android_id;
			global::Xamarin.Forms.Platform.Resource.Styleable.State_constraints = global::pjsua2xamarin.Droid.Resource.Styleable.State_constraints;
			global::Xamarin.Forms.Platform.Resource.Styleable.SwipeRefreshLayout = global::pjsua2xamarin.Droid.Resource.Styleable.SwipeRefreshLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.SwipeRefreshLayout_swipeRefreshLayoutProgressSpinnerBackgroundColor = global::pjsua2xamarin.Droid.Resource.Styleable.SwipeRefreshLayout_swipeRefreshLayoutProgressSpinnerBackgroundColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.SwitchCompat = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat;
			global::Xamarin.Forms.Platform.Resource.Styleable.SwitchCompat_android_textOff = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_android_textOff;
			global::Xamarin.Forms.Platform.Resource.Styleable.SwitchCompat_android_textOn = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_android_textOn;
			global::Xamarin.Forms.Platform.Resource.Styleable.SwitchCompat_android_thumb = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_android_thumb;
			global::Xamarin.Forms.Platform.Resource.Styleable.SwitchCompat_showText = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_showText;
			global::Xamarin.Forms.Platform.Resource.Styleable.SwitchCompat_splitTrack = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_splitTrack;
			global::Xamarin.Forms.Platform.Resource.Styleable.SwitchCompat_switchMinWidth = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_switchMinWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.SwitchCompat_switchPadding = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_switchPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.SwitchCompat_switchTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_switchTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.SwitchCompat_thumbTextPadding = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_thumbTextPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.SwitchCompat_thumbTint = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_thumbTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.SwitchCompat_thumbTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_thumbTintMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.SwitchCompat_track = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_track;
			global::Xamarin.Forms.Platform.Resource.Styleable.SwitchCompat_trackTint = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_trackTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.SwitchCompat_trackTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchCompat_trackTintMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.SwitchMaterial = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchMaterial;
			global::Xamarin.Forms.Platform.Resource.Styleable.SwitchMaterial_useMaterialThemeColors = global::pjsua2xamarin.Droid.Resource.Styleable.SwitchMaterial_useMaterialThemeColors;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabItem = global::pjsua2xamarin.Droid.Resource.Styleable.TabItem;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabItem_android_icon = global::pjsua2xamarin.Droid.Resource.Styleable.TabItem_android_icon;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabItem_android_layout = global::pjsua2xamarin.Droid.Resource.Styleable.TabItem_android_layout;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabItem_android_text = global::pjsua2xamarin.Droid.Resource.Styleable.TabItem_android_text;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabBackground = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabBackground;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabContentStart = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabContentStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabGravity = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabGravity;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabIconTint = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabIconTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabIconTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabIconTintMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabIndicator = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabIndicator;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabIndicatorAnimationDuration = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabIndicatorAnimationDuration;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabIndicatorAnimationMode = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabIndicatorAnimationMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabIndicatorColor = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabIndicatorColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabIndicatorFullWidth = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabIndicatorFullWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabIndicatorGravity = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabIndicatorGravity;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabIndicatorHeight = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabIndicatorHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabInlineLabel = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabInlineLabel;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabMaxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabMaxWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabMinWidth = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabMinWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabMode = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabPadding = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabPadding;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabPaddingBottom = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabPaddingBottom;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabPaddingEnd = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabPaddingEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabPaddingStart = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabPaddingStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabPaddingTop = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabPaddingTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabRippleColor = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabRippleColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabSelectedTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabSelectedTextColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabTextColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.TabLayout_tabUnboundedRipple = global::pjsua2xamarin.Droid.Resource.Styleable.TabLayout_tabUnboundedRipple;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextAppearance_android_fontFamily = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_fontFamily;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextAppearance_android_shadowColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_shadowColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextAppearance_android_shadowDx = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_shadowDx;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextAppearance_android_shadowDy = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_shadowDy;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextAppearance_android_shadowRadius = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_shadowRadius;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextAppearance_android_textColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_textColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextAppearance_android_textColorHint = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_textColorHint;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextAppearance_android_textColorLink = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_textColorLink;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextAppearance_android_textFontWeight = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_textFontWeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextAppearance_android_textSize = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_textSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextAppearance_android_textStyle = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_textStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextAppearance_android_typeface = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_android_typeface;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextAppearance_fontFamily = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_fontFamily;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextAppearance_fontVariationSettings = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_fontVariationSettings;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextAppearance_textAllCaps = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_textAllCaps;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextAppearance_textLocale = global::pjsua2xamarin.Droid.Resource.Styleable.TextAppearance_textLocale;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextEffects = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextEffects_android_fontFamily = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_android_fontFamily;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextEffects_android_shadowColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_android_shadowColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextEffects_android_shadowDx = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_android_shadowDx;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextEffects_android_shadowDy = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_android_shadowDy;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextEffects_android_shadowRadius = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_android_shadowRadius;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextEffects_android_text = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_android_text;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextEffects_android_textSize = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_android_textSize;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextEffects_android_textStyle = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_android_textStyle;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextEffects_android_typeface = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_android_typeface;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextEffects_borderRound = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_borderRound;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextEffects_borderRoundPercent = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_borderRoundPercent;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextEffects_textFillColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_textFillColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextEffects_textOutlineColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_textOutlineColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextEffects_textOutlineThickness = global::pjsua2xamarin.Droid.Resource.Styleable.TextEffects_textOutlineThickness;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputEditText = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputEditText;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputEditText_textInputLayoutFocusedRectEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputEditText_textInputLayoutFocusedRectEnabled;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_android_enabled = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_android_enabled;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_android_hint = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_android_hint;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_android_maxWidth = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_android_maxWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_android_minWidth = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_android_minWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_android_textColorHint = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_android_textColorHint;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_boxBackgroundColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_boxBackgroundColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_boxBackgroundMode = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_boxBackgroundMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_boxCollapsedPaddingTop = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_boxCollapsedPaddingTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_boxCornerRadiusBottomEnd = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_boxCornerRadiusBottomEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_boxCornerRadiusBottomStart = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_boxCornerRadiusBottomStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_boxCornerRadiusTopEnd = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_boxCornerRadiusTopEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_boxCornerRadiusTopStart = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_boxCornerRadiusTopStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_boxStrokeColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_boxStrokeColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_boxStrokeErrorColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_boxStrokeErrorColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_boxStrokeWidth = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_boxStrokeWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_boxStrokeWidthFocused = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_boxStrokeWidthFocused;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_counterEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_counterEnabled;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_counterMaxLength = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_counterMaxLength;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_counterOverflowTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_counterOverflowTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_counterOverflowTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_counterOverflowTextColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_counterTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_counterTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_counterTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_counterTextColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_endIconCheckable = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_endIconCheckable;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_endIconContentDescription = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_endIconContentDescription;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_endIconDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_endIconDrawable;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_endIconMode = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_endIconMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_endIconTint = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_endIconTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_endIconTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_endIconTintMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_errorContentDescription = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_errorContentDescription;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_errorEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_errorEnabled;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_errorIconDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_errorIconDrawable;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_errorIconTint = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_errorIconTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_errorIconTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_errorIconTintMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_errorTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_errorTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_errorTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_errorTextColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_expandedHintEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_expandedHintEnabled;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_helperText = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_helperText;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_helperTextEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_helperTextEnabled;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_helperTextTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_helperTextTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_helperTextTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_helperTextTextColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_hintAnimationEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_hintAnimationEnabled;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_hintEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_hintEnabled;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_hintTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_hintTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_hintTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_hintTextColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_passwordToggleContentDescription = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_passwordToggleContentDescription;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_passwordToggleDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_passwordToggleDrawable;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_passwordToggleEnabled = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_passwordToggleEnabled;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_passwordToggleTint = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_passwordToggleTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_passwordToggleTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_passwordToggleTintMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_placeholderText = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_placeholderText;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_placeholderTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_placeholderTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_placeholderTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_placeholderTextColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_prefixText = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_prefixText;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_prefixTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_prefixTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_prefixTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_prefixTextColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_shapeAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_shapeAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_shapeAppearanceOverlay = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_shapeAppearanceOverlay;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_startIconCheckable = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_startIconCheckable;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_startIconContentDescription = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_startIconContentDescription;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_startIconDrawable = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_startIconDrawable;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_startIconTint = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_startIconTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_startIconTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_startIconTintMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_suffixText = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_suffixText;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_suffixTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_suffixTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.TextInputLayout_suffixTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.TextInputLayout_suffixTextColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.ThemeEnforcement = global::pjsua2xamarin.Droid.Resource.Styleable.ThemeEnforcement;
			global::Xamarin.Forms.Platform.Resource.Styleable.ThemeEnforcement_android_textAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.ThemeEnforcement_android_textAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.ThemeEnforcement_enforceMaterialTheme = global::pjsua2xamarin.Droid.Resource.Styleable.ThemeEnforcement_enforceMaterialTheme;
			global::Xamarin.Forms.Platform.Resource.Styleable.ThemeEnforcement_enforceTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.ThemeEnforcement_enforceTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_android_gravity = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_android_gravity;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_android_minHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_android_minHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_buttonGravity = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_buttonGravity;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_collapseContentDescription = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_collapseContentDescription;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_collapseIcon = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_collapseIcon;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_contentInsetEnd = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_contentInsetEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_contentInsetEndWithActions = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_contentInsetEndWithActions;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_contentInsetLeft = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_contentInsetLeft;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_contentInsetRight = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_contentInsetRight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_contentInsetStart = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_contentInsetStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_contentInsetStartWithNavigation = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_contentInsetStartWithNavigation;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_logo = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_logo;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_logoDescription = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_logoDescription;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_maxButtonHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_maxButtonHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_menu = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_menu;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_navigationContentDescription = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_navigationContentDescription;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_navigationIcon = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_navigationIcon;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_popupTheme = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_popupTheme;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_subtitle = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_subtitle;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_subtitleTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_subtitleTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_subtitleTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_subtitleTextColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_title = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_title;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_titleMargin = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_titleMargin;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_titleMarginBottom = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_titleMarginBottom;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_titleMarginEnd = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_titleMarginEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_titleMargins = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_titleMargins;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_titleMarginStart = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_titleMarginStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_titleMarginTop = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_titleMarginTop;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_titleTextAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_titleTextAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.Toolbar_titleTextColor = global::pjsua2xamarin.Droid.Resource.Styleable.Toolbar_titleTextColor;
			global::Xamarin.Forms.Platform.Resource.Styleable.Tooltip = global::pjsua2xamarin.Droid.Resource.Styleable.Tooltip;
			global::Xamarin.Forms.Platform.Resource.Styleable.Tooltip_android_layout_margin = global::pjsua2xamarin.Droid.Resource.Styleable.Tooltip_android_layout_margin;
			global::Xamarin.Forms.Platform.Resource.Styleable.Tooltip_android_minHeight = global::pjsua2xamarin.Droid.Resource.Styleable.Tooltip_android_minHeight;
			global::Xamarin.Forms.Platform.Resource.Styleable.Tooltip_android_minWidth = global::pjsua2xamarin.Droid.Resource.Styleable.Tooltip_android_minWidth;
			global::Xamarin.Forms.Platform.Resource.Styleable.Tooltip_android_padding = global::pjsua2xamarin.Droid.Resource.Styleable.Tooltip_android_padding;
			global::Xamarin.Forms.Platform.Resource.Styleable.Tooltip_android_text = global::pjsua2xamarin.Droid.Resource.Styleable.Tooltip_android_text;
			global::Xamarin.Forms.Platform.Resource.Styleable.Tooltip_android_textAppearance = global::pjsua2xamarin.Droid.Resource.Styleable.Tooltip_android_textAppearance;
			global::Xamarin.Forms.Platform.Resource.Styleable.Tooltip_backgroundTint = global::pjsua2xamarin.Droid.Resource.Styleable.Tooltip_backgroundTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transform = global::pjsua2xamarin.Droid.Resource.Styleable.Transform;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transform_android_elevation = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_android_elevation;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transform_android_rotation = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_android_rotation;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transform_android_rotationX = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_android_rotationX;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transform_android_rotationY = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_android_rotationY;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transform_android_scaleX = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_android_scaleX;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transform_android_scaleY = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_android_scaleY;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transform_android_transformPivotX = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_android_transformPivotX;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transform_android_transformPivotY = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_android_transformPivotY;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transform_android_translationX = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_android_translationX;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transform_android_translationY = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_android_translationY;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transform_android_translationZ = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_android_translationZ;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transform_transformPivotTarget = global::pjsua2xamarin.Droid.Resource.Styleable.Transform_transformPivotTarget;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transition = global::pjsua2xamarin.Droid.Resource.Styleable.Transition;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transition_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.Transition_android_id;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transition_autoTransition = global::pjsua2xamarin.Droid.Resource.Styleable.Transition_autoTransition;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transition_constraintSetEnd = global::pjsua2xamarin.Droid.Resource.Styleable.Transition_constraintSetEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transition_constraintSetStart = global::pjsua2xamarin.Droid.Resource.Styleable.Transition_constraintSetStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transition_duration = global::pjsua2xamarin.Droid.Resource.Styleable.Transition_duration;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transition_layoutDuringTransition = global::pjsua2xamarin.Droid.Resource.Styleable.Transition_layoutDuringTransition;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transition_motionInterpolator = global::pjsua2xamarin.Droid.Resource.Styleable.Transition_motionInterpolator;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transition_pathMotionArc = global::pjsua2xamarin.Droid.Resource.Styleable.Transition_pathMotionArc;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transition_staggered = global::pjsua2xamarin.Droid.Resource.Styleable.Transition_staggered;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transition_transitionDisable = global::pjsua2xamarin.Droid.Resource.Styleable.Transition_transitionDisable;
			global::Xamarin.Forms.Platform.Resource.Styleable.Transition_transitionFlags = global::pjsua2xamarin.Droid.Resource.Styleable.Transition_transitionFlags;
			global::Xamarin.Forms.Platform.Resource.Styleable.Variant = global::pjsua2xamarin.Droid.Resource.Styleable.Variant;
			global::Xamarin.Forms.Platform.Resource.Styleable.Variant_constraints = global::pjsua2xamarin.Droid.Resource.Styleable.Variant_constraints;
			global::Xamarin.Forms.Platform.Resource.Styleable.Variant_region_heightLessThan = global::pjsua2xamarin.Droid.Resource.Styleable.Variant_region_heightLessThan;
			global::Xamarin.Forms.Platform.Resource.Styleable.Variant_region_heightMoreThan = global::pjsua2xamarin.Droid.Resource.Styleable.Variant_region_heightMoreThan;
			global::Xamarin.Forms.Platform.Resource.Styleable.Variant_region_widthLessThan = global::pjsua2xamarin.Droid.Resource.Styleable.Variant_region_widthLessThan;
			global::Xamarin.Forms.Platform.Resource.Styleable.Variant_region_widthMoreThan = global::pjsua2xamarin.Droid.Resource.Styleable.Variant_region_widthMoreThan;
			global::Xamarin.Forms.Platform.Resource.Styleable.View = global::pjsua2xamarin.Droid.Resource.Styleable.View;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewBackgroundHelper = global::pjsua2xamarin.Droid.Resource.Styleable.ViewBackgroundHelper;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewBackgroundHelper_android_background = global::pjsua2xamarin.Droid.Resource.Styleable.ViewBackgroundHelper_android_background;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewBackgroundHelper_backgroundTint = global::pjsua2xamarin.Droid.Resource.Styleable.ViewBackgroundHelper_backgroundTint;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewBackgroundHelper_backgroundTintMode = global::pjsua2xamarin.Droid.Resource.Styleable.ViewBackgroundHelper_backgroundTintMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewPager2 = global::pjsua2xamarin.Droid.Resource.Styleable.ViewPager2;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewPager2_android_orientation = global::pjsua2xamarin.Droid.Resource.Styleable.ViewPager2_android_orientation;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewStubCompat = global::pjsua2xamarin.Droid.Resource.Styleable.ViewStubCompat;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewStubCompat_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.ViewStubCompat_android_id;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewStubCompat_android_inflatedId = global::pjsua2xamarin.Droid.Resource.Styleable.ViewStubCompat_android_inflatedId;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewStubCompat_android_layout = global::pjsua2xamarin.Droid.Resource.Styleable.ViewStubCompat_android_layout;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewTransition = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewTransition_android_id = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_android_id;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewTransition_clearsTag = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_clearsTag;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewTransition_duration = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_duration;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewTransition_ifTagNotSet = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_ifTagNotSet;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewTransition_ifTagSet = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_ifTagSet;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewTransition_motionInterpolator = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_motionInterpolator;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewTransition_motionTarget = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_motionTarget;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewTransition_onStateTransition = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_onStateTransition;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewTransition_pathMotionArc = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_pathMotionArc;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewTransition_setsTag = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_setsTag;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewTransition_SharedValue = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_SharedValue;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewTransition_SharedValueId = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_SharedValueId;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewTransition_transitionDisable = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_transitionDisable;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewTransition_upDuration = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_upDuration;
			global::Xamarin.Forms.Platform.Resource.Styleable.ViewTransition_viewTransitionMode = global::pjsua2xamarin.Droid.Resource.Styleable.ViewTransition_viewTransitionMode;
			global::Xamarin.Forms.Platform.Resource.Styleable.View_android_focusable = global::pjsua2xamarin.Droid.Resource.Styleable.View_android_focusable;
			global::Xamarin.Forms.Platform.Resource.Styleable.View_android_theme = global::pjsua2xamarin.Droid.Resource.Styleable.View_android_theme;
			global::Xamarin.Forms.Platform.Resource.Styleable.View_paddingEnd = global::pjsua2xamarin.Droid.Resource.Styleable.View_paddingEnd;
			global::Xamarin.Forms.Platform.Resource.Styleable.View_paddingStart = global::pjsua2xamarin.Droid.Resource.Styleable.View_paddingStart;
			global::Xamarin.Forms.Platform.Resource.Styleable.View_theme = global::pjsua2xamarin.Droid.Resource.Styleable.View_theme;
			global::Xamarin.Forms.Platform.Resource.Xml.standalone_badge = global::pjsua2xamarin.Droid.Resource.Xml.standalone_badge;
			global::Xamarin.Forms.Platform.Resource.Xml.standalone_badge_gravity_bottom_end = global::pjsua2xamarin.Droid.Resource.Xml.standalone_badge_gravity_bottom_end;
			global::Xamarin.Forms.Platform.Resource.Xml.standalone_badge_gravity_bottom_start = global::pjsua2xamarin.Droid.Resource.Xml.standalone_badge_gravity_bottom_start;
			global::Xamarin.Forms.Platform.Resource.Xml.standalone_badge_gravity_top_start = global::pjsua2xamarin.Droid.Resource.Xml.standalone_badge_gravity_top_start;
			global::Xamarin.Forms.Platform.Resource.Xml.standalone_badge_offset = global::pjsua2xamarin.Droid.Resource.Xml.standalone_badge_offset;
		}
		
		public partial class Animation
		{
			
			// aapt resource value: 0x7F010000
			public const int abc_fade_in = 2130771968;
			
			// aapt resource value: 0x7F010001
			public const int abc_fade_out = 2130771969;
			
			// aapt resource value: 0x7F010002
			public const int abc_grow_fade_in_from_bottom = 2130771970;
			
			// aapt resource value: 0x7F010003
			public const int abc_popup_enter = 2130771971;
			
			// aapt resource value: 0x7F010004
			public const int abc_popup_exit = 2130771972;
			
			// aapt resource value: 0x7F010005
			public const int abc_shrink_fade_out_from_bottom = 2130771973;
			
			// aapt resource value: 0x7F010006
			public const int abc_slide_in_bottom = 2130771974;
			
			// aapt resource value: 0x7F010007
			public const int abc_slide_in_top = 2130771975;
			
			// aapt resource value: 0x7F010008
			public const int abc_slide_out_bottom = 2130771976;
			
			// aapt resource value: 0x7F010009
			public const int abc_slide_out_top = 2130771977;
			
			// aapt resource value: 0x7F01000A
			public const int abc_tooltip_enter = 2130771978;
			
			// aapt resource value: 0x7F01000B
			public const int abc_tooltip_exit = 2130771979;
			
			// aapt resource value: 0x7F01000C
			public const int btn_checkbox_to_checked_box_inner_merged_animation = 2130771980;
			
			// aapt resource value: 0x7F01000D
			public const int btn_checkbox_to_checked_box_outer_merged_animation = 2130771981;
			
			// aapt resource value: 0x7F01000E
			public const int btn_checkbox_to_checked_icon_null_animation = 2130771982;
			
			// aapt resource value: 0x7F01000F
			public const int btn_checkbox_to_unchecked_box_inner_merged_animation = 2130771983;
			
			// aapt resource value: 0x7F010010
			public const int btn_checkbox_to_unchecked_check_path_merged_animation = 2130771984;
			
			// aapt resource value: 0x7F010011
			public const int btn_checkbox_to_unchecked_icon_null_animation = 2130771985;
			
			// aapt resource value: 0x7F010012
			public const int btn_radio_to_off_mtrl_dot_group_animation = 2130771986;
			
			// aapt resource value: 0x7F010013
			public const int btn_radio_to_off_mtrl_ring_outer_animation = 2130771987;
			
			// aapt resource value: 0x7F010014
			public const int btn_radio_to_off_mtrl_ring_outer_path_animation = 2130771988;
			
			// aapt resource value: 0x7F010015
			public const int btn_radio_to_on_mtrl_dot_group_animation = 2130771989;
			
			// aapt resource value: 0x7F010016
			public const int btn_radio_to_on_mtrl_ring_outer_animation = 2130771990;
			
			// aapt resource value: 0x7F010017
			public const int btn_radio_to_on_mtrl_ring_outer_path_animation = 2130771991;
			
			// aapt resource value: 0x7F010018
			public const int design_bottom_sheet_slide_in = 2130771992;
			
			// aapt resource value: 0x7F010019
			public const int design_bottom_sheet_slide_out = 2130771993;
			
			// aapt resource value: 0x7F01001A
			public const int design_snackbar_in = 2130771994;
			
			// aapt resource value: 0x7F01001B
			public const int design_snackbar_out = 2130771995;
			
			// aapt resource value: 0x7F01001C
			public const int EnterFromLeft = 2130771996;
			
			// aapt resource value: 0x7F01001D
			public const int EnterFromRight = 2130771997;
			
			// aapt resource value: 0x7F01001E
			public const int ExitToLeft = 2130771998;
			
			// aapt resource value: 0x7F01001F
			public const int ExitToRight = 2130771999;
			
			// aapt resource value: 0x7F010020
			public const int fragment_fast_out_extra_slow_in = 2130772000;
			
			// aapt resource value: 0x7F010021
			public const int mtrl_bottom_sheet_slide_in = 2130772001;
			
			// aapt resource value: 0x7F010022
			public const int mtrl_bottom_sheet_slide_out = 2130772002;
			
			// aapt resource value: 0x7F010023
			public const int mtrl_card_lowers_interpolator = 2130772003;
			
			// aapt resource value: 0x7F010024
			public const int nav_default_enter_anim = 2130772004;
			
			// aapt resource value: 0x7F010025
			public const int nav_default_exit_anim = 2130772005;
			
			// aapt resource value: 0x7F010026
			public const int nav_default_pop_enter_anim = 2130772006;
			
			// aapt resource value: 0x7F010027
			public const int nav_default_pop_exit_anim = 2130772007;
			
			static Animation()
			{
				global::Android.Runtime.ResourceIdManager.UpdateIdValues();
			}
			
			private Animation()
			{
			}
		}
		
		public partial class Animator
		{
			
			// aapt resource value: 0x7F020000
			public const int design_appbar_state_list_animator = 2130837504;
			
			// aapt resource value: 0x7F020001
			public const int design_fab_hide_motion_spec = 2130837505;
			
			// aapt resource value: 0x7F020002
			public const int design_fab_show_motion_spec = 2130837506;
			
			// aapt resource value: 0x7F020003
			public const int fragment_close_enter = 2130837507;
			
			// aapt resource value: 0x7F020004
			public const int fragment_close_exit = 2130837508;
			
			// aapt resource value: 0x7F020005
			public const int fragment_fade_enter = 2130837509;
			
			// aapt resource value: 0x7F020006
			public const int fragment_fade_exit = 2130837510;
			
			// aapt resource value: 0x7F020007
			public const int fragment_open_enter = 2130837511;
			
			// aapt resource value: 0x7F020008
			public const int fragment_open_exit = 2130837512;
			
			// aapt resource value: 0x7F020009
			public const int linear_indeterminate_line1_head_interpolator = 2130837513;
			
			// aapt resource value: 0x7F02000A
			public const int linear_indeterminate_line1_tail_interpolator = 2130837514;
			
			// aapt resource value: 0x7F02000B
			public const int linear_indeterminate_line2_head_interpolator = 2130837515;
			
			// aapt resource value: 0x7F02000C
			public const int linear_indeterminate_line2_tail_interpolator = 2130837516;
			
			// aapt resource value: 0x7F02000D
			public const int mtrl_btn_state_list_anim = 2130837517;
			
			// aapt resource value: 0x7F02000E
			public const int mtrl_btn_unelevated_state_list_anim = 2130837518;
			
			// aapt resource value: 0x7F02000F
			public const int mtrl_card_state_list_anim = 2130837519;
			
			// aapt resource value: 0x7F020010
			public const int mtrl_chip_state_list_anim = 2130837520;
			
			// aapt resource value: 0x7F020011
			public const int mtrl_extended_fab_change_size_collapse_motion_spec = 2130837521;
			
			// aapt resource value: 0x7F020012
			public const int mtrl_extended_fab_change_size_expand_motion_spec = 2130837522;
			
			// aapt resource value: 0x7F020013
			public const int mtrl_extended_fab_hide_motion_spec = 2130837523;
			
			// aapt resource value: 0x7F020014
			public const int mtrl_extended_fab_show_motion_spec = 2130837524;
			
			// aapt resource value: 0x7F020015
			public const int mtrl_extended_fab_state_list_animator = 2130837525;
			
			// aapt resource value: 0x7F020016
			public const int mtrl_fab_hide_motion_spec = 2130837526;
			
			// aapt resource value: 0x7F020017
			public const int mtrl_fab_show_motion_spec = 2130837527;
			
			// aapt resource value: 0x7F020018
			public const int mtrl_fab_transformation_sheet_collapse_spec = 2130837528;
			
			// aapt resource value: 0x7F020019
			public const int mtrl_fab_transformation_sheet_expand_spec = 2130837529;
			
			// aapt resource value: 0x7F02001A
			public const int nav_default_enter_anim = 2130837530;
			
			// aapt resource value: 0x7F02001B
			public const int nav_default_exit_anim = 2130837531;
			
			// aapt resource value: 0x7F02001C
			public const int nav_default_pop_enter_anim = 2130837532;
			
			// aapt resource value: 0x7F02001D
			public const int nav_default_pop_exit_anim = 2130837533;
			
			static Animator()
			{
				global::Android.Runtime.ResourceIdManager.UpdateIdValues();
			}
			
			private Animator()
			{
			}
		}
		
		public partial class Attribute
		{
			
			// aapt resource value: 0x7F030003
			public const int action = 2130903043;
			
			// aapt resource value: 0x7F030004
			public const int actionBarDivider = 2130903044;
			
			// aapt resource value: 0x7F030005
			public const int actionBarItemBackground = 2130903045;
			
			// aapt resource value: 0x7F030006
			public const int actionBarPopupTheme = 2130903046;
			
			// aapt resource value: 0x7F030007
			public const int actionBarSize = 2130903047;
			
			// aapt resource value: 0x7F030008
			public const int actionBarSplitStyle = 2130903048;
			
			// aapt resource value: 0x7F030009
			public const int actionBarStyle = 2130903049;
			
			// aapt resource value: 0x7F03000A
			public const int actionBarTabBarStyle = 2130903050;
			
			// aapt resource value: 0x7F03000B
			public const int actionBarTabStyle = 2130903051;
			
			// aapt resource value: 0x7F03000C
			public const int actionBarTabTextStyle = 2130903052;
			
			// aapt resource value: 0x7F03000D
			public const int actionBarTheme = 2130903053;
			
			// aapt resource value: 0x7F03000E
			public const int actionBarWidgetTheme = 2130903054;
			
			// aapt resource value: 0x7F03000F
			public const int actionButtonStyle = 2130903055;
			
			// aapt resource value: 0x7F030010
			public const int actionDropDownStyle = 2130903056;
			
			// aapt resource value: 0x7F030011
			public const int actionLayout = 2130903057;
			
			// aapt resource value: 0x7F030012
			public const int actionMenuTextAppearance = 2130903058;
			
			// aapt resource value: 0x7F030013
			public const int actionMenuTextColor = 2130903059;
			
			// aapt resource value: 0x7F030014
			public const int actionModeBackground = 2130903060;
			
			// aapt resource value: 0x7F030015
			public const int actionModeCloseButtonStyle = 2130903061;
			
			// aapt resource value: 0x7F030016
			public const int actionModeCloseContentDescription = 2130903062;
			
			// aapt resource value: 0x7F030017
			public const int actionModeCloseDrawable = 2130903063;
			
			// aapt resource value: 0x7F030018
			public const int actionModeCopyDrawable = 2130903064;
			
			// aapt resource value: 0x7F030019
			public const int actionModeCutDrawable = 2130903065;
			
			// aapt resource value: 0x7F03001A
			public const int actionModeFindDrawable = 2130903066;
			
			// aapt resource value: 0x7F03001B
			public const int actionModePasteDrawable = 2130903067;
			
			// aapt resource value: 0x7F03001C
			public const int actionModePopupWindowStyle = 2130903068;
			
			// aapt resource value: 0x7F03001D
			public const int actionModeSelectAllDrawable = 2130903069;
			
			// aapt resource value: 0x7F03001E
			public const int actionModeShareDrawable = 2130903070;
			
			// aapt resource value: 0x7F03001F
			public const int actionModeSplitBackground = 2130903071;
			
			// aapt resource value: 0x7F030020
			public const int actionModeStyle = 2130903072;
			
			// aapt resource value: 0x7F030021
			public const int actionModeTheme = 2130903073;
			
			// aapt resource value: 0x7F030022
			public const int actionModeWebSearchDrawable = 2130903074;
			
			// aapt resource value: 0x7F030023
			public const int actionOverflowButtonStyle = 2130903075;
			
			// aapt resource value: 0x7F030024
			public const int actionOverflowMenuStyle = 2130903076;
			
			// aapt resource value: 0x7F030025
			public const int actionProviderClass = 2130903077;
			
			// aapt resource value: 0x7F030026
			public const int actionTextColorAlpha = 2130903078;
			
			// aapt resource value: 0x7F030027
			public const int actionViewClass = 2130903079;
			
			// aapt resource value: 0x7F030028
			public const int activityChooserViewStyle = 2130903080;
			
			// aapt resource value: 0x7F030029
			public const int adjustable = 2130903081;
			
			// aapt resource value: 0x7F03002A
			public const int alertDialogButtonGroupStyle = 2130903082;
			
			// aapt resource value: 0x7F03002B
			public const int alertDialogCenterButtons = 2130903083;
			
			// aapt resource value: 0x7F03002C
			public const int alertDialogStyle = 2130903084;
			
			// aapt resource value: 0x7F03002D
			public const int alertDialogTheme = 2130903085;
			
			// aapt resource value: 0x7F03002E
			public const int allowDividerAbove = 2130903086;
			
			// aapt resource value: 0x7F03002F
			public const int allowDividerAfterLastItem = 2130903087;
			
			// aapt resource value: 0x7F030030
			public const int allowDividerBelow = 2130903088;
			
			// aapt resource value: 0x7F030031
			public const int allowStacking = 2130903089;
			
			// aapt resource value: 0x7F030032
			public const int alpha = 2130903090;
			
			// aapt resource value: 0x7F030033
			public const int alphabeticModifiers = 2130903091;
			
			// aapt resource value: 0x7F030034
			public const int altSrc = 2130903092;
			
			// aapt resource value: 0x7F030035
			public const int animateCircleAngleTo = 2130903093;
			
			// aapt resource value: 0x7F030036
			public const int animateRelativeTo = 2130903094;
			
			// aapt resource value: 0x7F030037
			public const int animationMode = 2130903095;
			
			// aapt resource value: 0x7F030038
			public const int appBarLayoutStyle = 2130903096;
			
			// aapt resource value: 0x7F030039
			public const int applyMotionScene = 2130903097;
			
			// aapt resource value: 0x7F03003A
			public const int arcMode = 2130903098;
			
			// aapt resource value: 0x7F03003B
			public const int argType = 2130903099;
			
			// aapt resource value: 0x7F03003C
			public const int arrowHeadLength = 2130903100;
			
			// aapt resource value: 0x7F03003D
			public const int arrowShaftLength = 2130903101;
			
			// aapt resource value: 0x7F03003E
			public const int attributeName = 2130903102;
			
			// aapt resource value: 0x7F03003F
			public const int autoCompleteMode = 2130903103;
			
			// aapt resource value: 0x7F030040
			public const int autoCompleteTextViewStyle = 2130903104;
			
			// aapt resource value: 0x7F030041
			public const int autoSizeMaxTextSize = 2130903105;
			
			// aapt resource value: 0x7F030042
			public const int autoSizeMinTextSize = 2130903106;
			
			// aapt resource value: 0x7F030043
			public const int autoSizePresetSizes = 2130903107;
			
			// aapt resource value: 0x7F030044
			public const int autoSizeStepGranularity = 2130903108;
			
			// aapt resource value: 0x7F030045
			public const int autoSizeTextType = 2130903109;
			
			// aapt resource value: 0x7F030046
			public const int autoTransition = 2130903110;
			
			// aapt resource value: 0x7F030047
			public const int background = 2130903111;
			
			// aapt resource value: 0x7F030048
			public const int backgroundColor = 2130903112;
			
			// aapt resource value: 0x7F030049
			public const int backgroundInsetBottom = 2130903113;
			
			// aapt resource value: 0x7F03004A
			public const int backgroundInsetEnd = 2130903114;
			
			// aapt resource value: 0x7F03004B
			public const int backgroundInsetStart = 2130903115;
			
			// aapt resource value: 0x7F03004C
			public const int backgroundInsetTop = 2130903116;
			
			// aapt resource value: 0x7F03004D
			public const int backgroundOverlayColorAlpha = 2130903117;
			
			// aapt resource value: 0x7F03004E
			public const int backgroundSplit = 2130903118;
			
			// aapt resource value: 0x7F03004F
			public const int backgroundStacked = 2130903119;
			
			// aapt resource value: 0x7F030050
			public const int backgroundTint = 2130903120;
			
			// aapt resource value: 0x7F030051
			public const int backgroundTintMode = 2130903121;
			
			// aapt resource value: 0x7F030052
			public const int badgeGravity = 2130903122;
			
			// aapt resource value: 0x7F030053
			public const int badgeStyle = 2130903123;
			
			// aapt resource value: 0x7F030054
			public const int badgeTextColor = 2130903124;
			
			// aapt resource value: 0x7F030055
			public const int barLength = 2130903125;
			
			// aapt resource value: 0x7F030056
			public const int barrierAllowsGoneWidgets = 2130903126;
			
			// aapt resource value: 0x7F030057
			public const int barrierDirection = 2130903127;
			
			// aapt resource value: 0x7F030058
			public const int barrierMargin = 2130903128;
			
			// aapt resource value: 0x7F030059
			public const int behavior_autoHide = 2130903129;
			
			// aapt resource value: 0x7F03005A
			public const int behavior_autoShrink = 2130903130;
			
			// aapt resource value: 0x7F03005B
			public const int behavior_draggable = 2130903131;
			
			// aapt resource value: 0x7F03005C
			public const int behavior_expandedOffset = 2130903132;
			
			// aapt resource value: 0x7F03005D
			public const int behavior_fitToContents = 2130903133;
			
			// aapt resource value: 0x7F03005E
			public const int behavior_halfExpandedRatio = 2130903134;
			
			// aapt resource value: 0x7F03005F
			public const int behavior_hideable = 2130903135;
			
			// aapt resource value: 0x7F030060
			public const int behavior_overlapTop = 2130903136;
			
			// aapt resource value: 0x7F030061
			public const int behavior_peekHeight = 2130903137;
			
			// aapt resource value: 0x7F030062
			public const int behavior_saveFlags = 2130903138;
			
			// aapt resource value: 0x7F030063
			public const int behavior_skipCollapsed = 2130903139;
			
			// aapt resource value: 0x7F030064
			public const int blendSrc = 2130903140;
			
			// aapt resource value: 0x7F030068
			public const int borderlessButtonStyle = 2130903144;
			
			// aapt resource value: 0x7F030065
			public const int borderRound = 2130903141;
			
			// aapt resource value: 0x7F030066
			public const int borderRoundPercent = 2130903142;
			
			// aapt resource value: 0x7F030067
			public const int borderWidth = 2130903143;
			
			// aapt resource value: 0x7F030069
			public const int bottomAppBarStyle = 2130903145;
			
			// aapt resource value: 0x7F03006A
			public const int bottomNavigationStyle = 2130903146;
			
			// aapt resource value: 0x7F03006B
			public const int bottomSheetDialogTheme = 2130903147;
			
			// aapt resource value: 0x7F03006C
			public const int bottomSheetStyle = 2130903148;
			
			// aapt resource value: 0x7F03006D
			public const int boxBackgroundColor = 2130903149;
			
			// aapt resource value: 0x7F03006E
			public const int boxBackgroundMode = 2130903150;
			
			// aapt resource value: 0x7F03006F
			public const int boxCollapsedPaddingTop = 2130903151;
			
			// aapt resource value: 0x7F030070
			public const int boxCornerRadiusBottomEnd = 2130903152;
			
			// aapt resource value: 0x7F030071
			public const int boxCornerRadiusBottomStart = 2130903153;
			
			// aapt resource value: 0x7F030072
			public const int boxCornerRadiusTopEnd = 2130903154;
			
			// aapt resource value: 0x7F030073
			public const int boxCornerRadiusTopStart = 2130903155;
			
			// aapt resource value: 0x7F030074
			public const int boxStrokeColor = 2130903156;
			
			// aapt resource value: 0x7F030075
			public const int boxStrokeErrorColor = 2130903157;
			
			// aapt resource value: 0x7F030076
			public const int boxStrokeWidth = 2130903158;
			
			// aapt resource value: 0x7F030077
			public const int boxStrokeWidthFocused = 2130903159;
			
			// aapt resource value: 0x7F030078
			public const int brightness = 2130903160;
			
			// aapt resource value: 0x7F030079
			public const int buttonBarButtonStyle = 2130903161;
			
			// aapt resource value: 0x7F03007A
			public const int buttonBarNegativeButtonStyle = 2130903162;
			
			// aapt resource value: 0x7F03007B
			public const int buttonBarNeutralButtonStyle = 2130903163;
			
			// aapt resource value: 0x7F03007C
			public const int buttonBarPositiveButtonStyle = 2130903164;
			
			// aapt resource value: 0x7F03007D
			public const int buttonBarStyle = 2130903165;
			
			// aapt resource value: 0x7F03007E
			public const int buttonCompat = 2130903166;
			
			// aapt resource value: 0x7F03007F
			public const int buttonGravity = 2130903167;
			
			// aapt resource value: 0x7F030080
			public const int buttonIconDimen = 2130903168;
			
			// aapt resource value: 0x7F030081
			public const int buttonPanelSideLayout = 2130903169;
			
			// aapt resource value: 0x7F030082
			public const int buttonStyle = 2130903170;
			
			// aapt resource value: 0x7F030083
			public const int buttonStyleSmall = 2130903171;
			
			// aapt resource value: 0x7F030084
			public const int buttonTint = 2130903172;
			
			// aapt resource value: 0x7F030085
			public const int buttonTintMode = 2130903173;
			
			// aapt resource value: 0x7F030086
			public const int cardBackgroundColor = 2130903174;
			
			// aapt resource value: 0x7F030087
			public const int cardCornerRadius = 2130903175;
			
			// aapt resource value: 0x7F030088
			public const int cardElevation = 2130903176;
			
			// aapt resource value: 0x7F030089
			public const int cardForegroundColor = 2130903177;
			
			// aapt resource value: 0x7F03008A
			public const int cardMaxElevation = 2130903178;
			
			// aapt resource value: 0x7F03008B
			public const int cardPreventCornerOverlap = 2130903179;
			
			// aapt resource value: 0x7F03008C
			public const int cardUseCompatPadding = 2130903180;
			
			// aapt resource value: 0x7F03008D
			public const int cardViewStyle = 2130903181;
			
			// aapt resource value: 0x7F03008E
			public const int carousel_backwardTransition = 2130903182;
			
			// aapt resource value: 0x7F03008F
			public const int carousel_emptyViewsBehavior = 2130903183;
			
			// aapt resource value: 0x7F030090
			public const int carousel_firstView = 2130903184;
			
			// aapt resource value: 0x7F030091
			public const int carousel_forwardTransition = 2130903185;
			
			// aapt resource value: 0x7F030092
			public const int carousel_infinite = 2130903186;
			
			// aapt resource value: 0x7F030093
			public const int carousel_nextState = 2130903187;
			
			// aapt resource value: 0x7F030094
			public const int carousel_previousState = 2130903188;
			
			// aapt resource value: 0x7F030095
			public const int carousel_touchUpMode = 2130903189;
			
			// aapt resource value: 0x7F030096
			public const int carousel_touchUp_dampeningFactor = 2130903190;
			
			// aapt resource value: 0x7F030097
			public const int carousel_touchUp_velocityThreshold = 2130903191;
			
			// aapt resource value: 0x7F030098
			public const int chainUseRtl = 2130903192;
			
			// aapt resource value: 0x7F030099
			public const int checkBoxPreferenceStyle = 2130903193;
			
			// aapt resource value: 0x7F03009A
			public const int checkboxStyle = 2130903194;
			
			// aapt resource value: 0x7F03009B
			public const int checkedButton = 2130903195;
			
			// aapt resource value: 0x7F03009C
			public const int checkedChip = 2130903196;
			
			// aapt resource value: 0x7F03009D
			public const int checkedIcon = 2130903197;
			
			// aapt resource value: 0x7F03009E
			public const int checkedIconEnabled = 2130903198;
			
			// aapt resource value: 0x7F03009F
			public const int checkedIconMargin = 2130903199;
			
			// aapt resource value: 0x7F0300A0
			public const int checkedIconSize = 2130903200;
			
			// aapt resource value: 0x7F0300A1
			public const int checkedIconTint = 2130903201;
			
			// aapt resource value: 0x7F0300A2
			public const int checkedIconVisible = 2130903202;
			
			// aapt resource value: 0x7F0300A3
			public const int checkedTextViewStyle = 2130903203;
			
			// aapt resource value: 0x7F0300A4
			public const int chipBackgroundColor = 2130903204;
			
			// aapt resource value: 0x7F0300A5
			public const int chipCornerRadius = 2130903205;
			
			// aapt resource value: 0x7F0300A6
			public const int chipEndPadding = 2130903206;
			
			// aapt resource value: 0x7F0300A7
			public const int chipGroupStyle = 2130903207;
			
			// aapt resource value: 0x7F0300A8
			public const int chipIcon = 2130903208;
			
			// aapt resource value: 0x7F0300A9
			public const int chipIconEnabled = 2130903209;
			
			// aapt resource value: 0x7F0300AA
			public const int chipIconSize = 2130903210;
			
			// aapt resource value: 0x7F0300AB
			public const int chipIconTint = 2130903211;
			
			// aapt resource value: 0x7F0300AC
			public const int chipIconVisible = 2130903212;
			
			// aapt resource value: 0x7F0300AD
			public const int chipMinHeight = 2130903213;
			
			// aapt resource value: 0x7F0300AE
			public const int chipMinTouchTargetSize = 2130903214;
			
			// aapt resource value: 0x7F0300AF
			public const int chipSpacing = 2130903215;
			
			// aapt resource value: 0x7F0300B0
			public const int chipSpacingHorizontal = 2130903216;
			
			// aapt resource value: 0x7F0300B1
			public const int chipSpacingVertical = 2130903217;
			
			// aapt resource value: 0x7F0300B2
			public const int chipStandaloneStyle = 2130903218;
			
			// aapt resource value: 0x7F0300B3
			public const int chipStartPadding = 2130903219;
			
			// aapt resource value: 0x7F0300B4
			public const int chipStrokeColor = 2130903220;
			
			// aapt resource value: 0x7F0300B5
			public const int chipStrokeWidth = 2130903221;
			
			// aapt resource value: 0x7F0300B6
			public const int chipStyle = 2130903222;
			
			// aapt resource value: 0x7F0300B7
			public const int chipSurfaceColor = 2130903223;
			
			// aapt resource value: 0x7F0300B8
			public const int circleRadius = 2130903224;
			
			// aapt resource value: 0x7F0300BA
			public const int circularflow_angles = 2130903226;
			
			// aapt resource value: 0x7F0300BB
			public const int circularflow_defaultAngle = 2130903227;
			
			// aapt resource value: 0x7F0300BC
			public const int circularflow_defaultRadius = 2130903228;
			
			// aapt resource value: 0x7F0300BD
			public const int circularflow_radiusInDP = 2130903229;
			
			// aapt resource value: 0x7F0300BE
			public const int circularflow_viewCenter = 2130903230;
			
			// aapt resource value: 0x7F0300B9
			public const int circularProgressIndicatorStyle = 2130903225;
			
			// aapt resource value: 0x7F0300BF
			public const int clearsTag = 2130903231;
			
			// aapt resource value: 0x7F0300C0
			public const int clickAction = 2130903232;
			
			// aapt resource value: 0x7F0300C1
			public const int clockFaceBackgroundColor = 2130903233;
			
			// aapt resource value: 0x7F0300C2
			public const int clockHandColor = 2130903234;
			
			// aapt resource value: 0x7F0300C3
			public const int clockIcon = 2130903235;
			
			// aapt resource value: 0x7F0300C4
			public const int clockNumberTextColor = 2130903236;
			
			// aapt resource value: 0x7F0300C5
			public const int closeIcon = 2130903237;
			
			// aapt resource value: 0x7F0300C6
			public const int closeIconEnabled = 2130903238;
			
			// aapt resource value: 0x7F0300C7
			public const int closeIconEndPadding = 2130903239;
			
			// aapt resource value: 0x7F0300C8
			public const int closeIconSize = 2130903240;
			
			// aapt resource value: 0x7F0300C9
			public const int closeIconStartPadding = 2130903241;
			
			// aapt resource value: 0x7F0300CA
			public const int closeIconTint = 2130903242;
			
			// aapt resource value: 0x7F0300CB
			public const int closeIconVisible = 2130903243;
			
			// aapt resource value: 0x7F0300CC
			public const int closeItemLayout = 2130903244;
			
			// aapt resource value: 0x7F0300CD
			public const int collapseContentDescription = 2130903245;
			
			// aapt resource value: 0x7F0300CF
			public const int collapsedSize = 2130903247;
			
			// aapt resource value: 0x7F0300D0
			public const int collapsedTitleGravity = 2130903248;
			
			// aapt resource value: 0x7F0300D1
			public const int collapsedTitleTextAppearance = 2130903249;
			
			// aapt resource value: 0x7F0300CE
			public const int collapseIcon = 2130903246;
			
			// aapt resource value: 0x7F0300D2
			public const int collapsingToolbarLayoutStyle = 2130903250;
			
			// aapt resource value: 0x7F0300D3
			public const int collectionViewStyle = 2130903251;
			
			// aapt resource value: 0x7F0300D4
			public const int color = 2130903252;
			
			// aapt resource value: 0x7F0300D5
			public const int colorAccent = 2130903253;
			
			// aapt resource value: 0x7F0300D6
			public const int colorBackgroundFloating = 2130903254;
			
			// aapt resource value: 0x7F0300D7
			public const int colorButtonNormal = 2130903255;
			
			// aapt resource value: 0x7F0300D8
			public const int colorControlActivated = 2130903256;
			
			// aapt resource value: 0x7F0300D9
			public const int colorControlHighlight = 2130903257;
			
			// aapt resource value: 0x7F0300DA
			public const int colorControlNormal = 2130903258;
			
			// aapt resource value: 0x7F0300DB
			public const int colorError = 2130903259;
			
			// aapt resource value: 0x7F0300DC
			public const int colorOnBackground = 2130903260;
			
			// aapt resource value: 0x7F0300DD
			public const int colorOnError = 2130903261;
			
			// aapt resource value: 0x7F0300DE
			public const int colorOnPrimary = 2130903262;
			
			// aapt resource value: 0x7F0300DF
			public const int colorOnPrimarySurface = 2130903263;
			
			// aapt resource value: 0x7F0300E0
			public const int colorOnSecondary = 2130903264;
			
			// aapt resource value: 0x7F0300E1
			public const int colorOnSurface = 2130903265;
			
			// aapt resource value: 0x7F0300E2
			public const int colorPrimary = 2130903266;
			
			// aapt resource value: 0x7F0300E3
			public const int colorPrimaryDark = 2130903267;
			
			// aapt resource value: 0x7F0300E4
			public const int colorPrimarySurface = 2130903268;
			
			// aapt resource value: 0x7F0300E5
			public const int colorPrimaryVariant = 2130903269;
			
			// aapt resource value: 0x7F0300E6
			public const int colorSecondary = 2130903270;
			
			// aapt resource value: 0x7F0300E7
			public const int colorSecondaryVariant = 2130903271;
			
			// aapt resource value: 0x7F0300E8
			public const int colorSurface = 2130903272;
			
			// aapt resource value: 0x7F0300E9
			public const int colorSwitchThumbNormal = 2130903273;
			
			// aapt resource value: 0x7F0300EA
			public const int commitIcon = 2130903274;
			
			// aapt resource value: 0x7F030000
			public const int ConstraintRotate = 2130903040;
			
			// aapt resource value: 0x7F0300F0
			public const int constraints = 2130903280;
			
			// aapt resource value: 0x7F0300EB
			public const int constraintSet = 2130903275;
			
			// aapt resource value: 0x7F0300EC
			public const int constraintSetEnd = 2130903276;
			
			// aapt resource value: 0x7F0300ED
			public const int constraintSetStart = 2130903277;
			
			// aapt resource value: 0x7F0300EE
			public const int constraint_referenced_ids = 2130903278;
			
			// aapt resource value: 0x7F0300EF
			public const int constraint_referenced_tags = 2130903279;
			
			// aapt resource value: 0x7F0300F1
			public const int content = 2130903281;
			
			// aapt resource value: 0x7F0300F2
			public const int contentDescription = 2130903282;
			
			// aapt resource value: 0x7F0300F3
			public const int contentInsetEnd = 2130903283;
			
			// aapt resource value: 0x7F0300F4
			public const int contentInsetEndWithActions = 2130903284;
			
			// aapt resource value: 0x7F0300F5
			public const int contentInsetLeft = 2130903285;
			
			// aapt resource value: 0x7F0300F6
			public const int contentInsetRight = 2130903286;
			
			// aapt resource value: 0x7F0300F7
			public const int contentInsetStart = 2130903287;
			
			// aapt resource value: 0x7F0300F8
			public const int contentInsetStartWithNavigation = 2130903288;
			
			// aapt resource value: 0x7F0300F9
			public const int contentPadding = 2130903289;
			
			// aapt resource value: 0x7F0300FA
			public const int contentPaddingBottom = 2130903290;
			
			// aapt resource value: 0x7F0300FB
			public const int contentPaddingEnd = 2130903291;
			
			// aapt resource value: 0x7F0300FC
			public const int contentPaddingLeft = 2130903292;
			
			// aapt resource value: 0x7F0300FD
			public const int contentPaddingRight = 2130903293;
			
			// aapt resource value: 0x7F0300FE
			public const int contentPaddingStart = 2130903294;
			
			// aapt resource value: 0x7F0300FF
			public const int contentPaddingTop = 2130903295;
			
			// aapt resource value: 0x7F030100
			public const int contentScrim = 2130903296;
			
			// aapt resource value: 0x7F030101
			public const int contrast = 2130903297;
			
			// aapt resource value: 0x7F030102
			public const int controlBackground = 2130903298;
			
			// aapt resource value: 0x7F030103
			public const int coordinatorLayoutStyle = 2130903299;
			
			// aapt resource value: 0x7F030104
			public const int cornerFamily = 2130903300;
			
			// aapt resource value: 0x7F030105
			public const int cornerFamilyBottomLeft = 2130903301;
			
			// aapt resource value: 0x7F030106
			public const int cornerFamilyBottomRight = 2130903302;
			
			// aapt resource value: 0x7F030107
			public const int cornerFamilyTopLeft = 2130903303;
			
			// aapt resource value: 0x7F030108
			public const int cornerFamilyTopRight = 2130903304;
			
			// aapt resource value: 0x7F030109
			public const int cornerRadius = 2130903305;
			
			// aapt resource value: 0x7F03010A
			public const int cornerSize = 2130903306;
			
			// aapt resource value: 0x7F03010B
			public const int cornerSizeBottomLeft = 2130903307;
			
			// aapt resource value: 0x7F03010C
			public const int cornerSizeBottomRight = 2130903308;
			
			// aapt resource value: 0x7F03010D
			public const int cornerSizeTopLeft = 2130903309;
			
			// aapt resource value: 0x7F03010E
			public const int cornerSizeTopRight = 2130903310;
			
			// aapt resource value: 0x7F03010F
			public const int counterEnabled = 2130903311;
			
			// aapt resource value: 0x7F030110
			public const int counterMaxLength = 2130903312;
			
			// aapt resource value: 0x7F030111
			public const int counterOverflowTextAppearance = 2130903313;
			
			// aapt resource value: 0x7F030112
			public const int counterOverflowTextColor = 2130903314;
			
			// aapt resource value: 0x7F030113
			public const int counterTextAppearance = 2130903315;
			
			// aapt resource value: 0x7F030114
			public const int counterTextColor = 2130903316;
			
			// aapt resource value: 0x7F030115
			public const int crossfade = 2130903317;
			
			// aapt resource value: 0x7F030116
			public const int currentState = 2130903318;
			
			// aapt resource value: 0x7F030117
			public const int curveFit = 2130903319;
			
			// aapt resource value: 0x7F030118
			public const int customBoolean = 2130903320;
			
			// aapt resource value: 0x7F030119
			public const int customColorDrawableValue = 2130903321;
			
			// aapt resource value: 0x7F03011A
			public const int customColorValue = 2130903322;
			
			// aapt resource value: 0x7F03011B
			public const int customDimension = 2130903323;
			
			// aapt resource value: 0x7F03011C
			public const int customFloatValue = 2130903324;
			
			// aapt resource value: 0x7F03011D
			public const int customIntegerValue = 2130903325;
			
			// aapt resource value: 0x7F03011E
			public const int customNavigationLayout = 2130903326;
			
			// aapt resource value: 0x7F03011F
			public const int customPixelDimension = 2130903327;
			
			// aapt resource value: 0x7F030120
			public const int customReference = 2130903328;
			
			// aapt resource value: 0x7F030121
			public const int customStringValue = 2130903329;
			
			// aapt resource value: 0x7F030122
			public const int data = 2130903330;
			
			// aapt resource value: 0x7F030123
			public const int dataPattern = 2130903331;
			
			// aapt resource value: 0x7F030124
			public const int dayInvalidStyle = 2130903332;
			
			// aapt resource value: 0x7F030125
			public const int daySelectedStyle = 2130903333;
			
			// aapt resource value: 0x7F030126
			public const int dayStyle = 2130903334;
			
			// aapt resource value: 0x7F030127
			public const int dayTodayStyle = 2130903335;
			
			// aapt resource value: 0x7F030128
			public const int defaultDuration = 2130903336;
			
			// aapt resource value: 0x7F030129
			public const int defaultQueryHint = 2130903337;
			
			// aapt resource value: 0x7F03012A
			public const int defaultState = 2130903338;
			
			// aapt resource value: 0x7F03012B
			public const int defaultValue = 2130903339;
			
			// aapt resource value: 0x7F03012C
			public const int deltaPolarAngle = 2130903340;
			
			// aapt resource value: 0x7F03012D
			public const int deltaPolarRadius = 2130903341;
			
			// aapt resource value: 0x7F03012E
			public const int dependency = 2130903342;
			
			// aapt resource value: 0x7F03012F
			public const int deriveConstraintsFrom = 2130903343;
			
			// aapt resource value: 0x7F030130
			public const int destination = 2130903344;
			
			// aapt resource value: 0x7F030131
			public const int dialogCornerRadius = 2130903345;
			
			// aapt resource value: 0x7F030132
			public const int dialogIcon = 2130903346;
			
			// aapt resource value: 0x7F030133
			public const int dialogLayout = 2130903347;
			
			// aapt resource value: 0x7F030134
			public const int dialogMessage = 2130903348;
			
			// aapt resource value: 0x7F030135
			public const int dialogPreferenceStyle = 2130903349;
			
			// aapt resource value: 0x7F030136
			public const int dialogPreferredPadding = 2130903350;
			
			// aapt resource value: 0x7F030137
			public const int dialogTheme = 2130903351;
			
			// aapt resource value: 0x7F030138
			public const int dialogTitle = 2130903352;
			
			// aapt resource value: 0x7F030139
			public const int disableDependentsState = 2130903353;
			
			// aapt resource value: 0x7F03013A
			public const int displayOptions = 2130903354;
			
			// aapt resource value: 0x7F03013B
			public const int divider = 2130903355;
			
			// aapt resource value: 0x7F03013C
			public const int dividerHorizontal = 2130903356;
			
			// aapt resource value: 0x7F03013D
			public const int dividerPadding = 2130903357;
			
			// aapt resource value: 0x7F03013E
			public const int dividerVertical = 2130903358;
			
			// aapt resource value: 0x7F03013F
			public const int dragDirection = 2130903359;
			
			// aapt resource value: 0x7F030140
			public const int dragScale = 2130903360;
			
			// aapt resource value: 0x7F030141
			public const int dragThreshold = 2130903361;
			
			// aapt resource value: 0x7F030143
			public const int drawableBottomCompat = 2130903363;
			
			// aapt resource value: 0x7F030144
			public const int drawableEndCompat = 2130903364;
			
			// aapt resource value: 0x7F030145
			public const int drawableLeftCompat = 2130903365;
			
			// aapt resource value: 0x7F030146
			public const int drawableRightCompat = 2130903366;
			
			// aapt resource value: 0x7F030147
			public const int drawableSize = 2130903367;
			
			// aapt resource value: 0x7F030148
			public const int drawableStartCompat = 2130903368;
			
			// aapt resource value: 0x7F030149
			public const int drawableTint = 2130903369;
			
			// aapt resource value: 0x7F03014A
			public const int drawableTintMode = 2130903370;
			
			// aapt resource value: 0x7F03014B
			public const int drawableTopCompat = 2130903371;
			
			// aapt resource value: 0x7F03014C
			public const int drawerArrowStyle = 2130903372;
			
			// aapt resource value: 0x7F03014D
			public const int drawerLayoutStyle = 2130903373;
			
			// aapt resource value: 0x7F030142
			public const int drawPath = 2130903362;
			
			// aapt resource value: 0x7F03014F
			public const int dropdownListPreferredItemHeight = 2130903375;
			
			// aapt resource value: 0x7F03014E
			public const int dropDownListViewStyle = 2130903374;
			
			// aapt resource value: 0x7F030150
			public const int dropdownPreferenceStyle = 2130903376;
			
			// aapt resource value: 0x7F030151
			public const int duration = 2130903377;
			
			// aapt resource value: 0x7F030152
			public const int editTextBackground = 2130903378;
			
			// aapt resource value: 0x7F030153
			public const int editTextColor = 2130903379;
			
			// aapt resource value: 0x7F030154
			public const int editTextPreferenceStyle = 2130903380;
			
			// aapt resource value: 0x7F030155
			public const int editTextStyle = 2130903381;
			
			// aapt resource value: 0x7F030156
			public const int elevation = 2130903382;
			
			// aapt resource value: 0x7F030157
			public const int elevationOverlayColor = 2130903383;
			
			// aapt resource value: 0x7F030158
			public const int elevationOverlayEnabled = 2130903384;
			
			// aapt resource value: 0x7F030159
			public const int enableCopying = 2130903385;
			
			// aapt resource value: 0x7F03015B
			public const int enabled = 2130903387;
			
			// aapt resource value: 0x7F03015A
			public const int enableEdgeToEdge = 2130903386;
			
			// aapt resource value: 0x7F03015C
			public const int endIconCheckable = 2130903388;
			
			// aapt resource value: 0x7F03015D
			public const int endIconContentDescription = 2130903389;
			
			// aapt resource value: 0x7F03015E
			public const int endIconDrawable = 2130903390;
			
			// aapt resource value: 0x7F03015F
			public const int endIconMode = 2130903391;
			
			// aapt resource value: 0x7F030160
			public const int endIconTint = 2130903392;
			
			// aapt resource value: 0x7F030161
			public const int endIconTintMode = 2130903393;
			
			// aapt resource value: 0x7F030162
			public const int enforceMaterialTheme = 2130903394;
			
			// aapt resource value: 0x7F030163
			public const int enforceTextAppearance = 2130903395;
			
			// aapt resource value: 0x7F030164
			public const int ensureMinTouchTargetSize = 2130903396;
			
			// aapt resource value: 0x7F030165
			public const int enterAnim = 2130903397;
			
			// aapt resource value: 0x7F030166
			public const int entries = 2130903398;
			
			// aapt resource value: 0x7F030167
			public const int entryValues = 2130903399;
			
			// aapt resource value: 0x7F030168
			public const int errorContentDescription = 2130903400;
			
			// aapt resource value: 0x7F030169
			public const int errorEnabled = 2130903401;
			
			// aapt resource value: 0x7F03016A
			public const int errorIconDrawable = 2130903402;
			
			// aapt resource value: 0x7F03016B
			public const int errorIconTint = 2130903403;
			
			// aapt resource value: 0x7F03016C
			public const int errorIconTintMode = 2130903404;
			
			// aapt resource value: 0x7F03016D
			public const int errorTextAppearance = 2130903405;
			
			// aapt resource value: 0x7F03016E
			public const int errorTextColor = 2130903406;
			
			// aapt resource value: 0x7F03016F
			public const int exitAnim = 2130903407;
			
			// aapt resource value: 0x7F030170
			public const int expandActivityOverflowButtonDrawable = 2130903408;
			
			// aapt resource value: 0x7F030171
			public const int expanded = 2130903409;
			
			// aapt resource value: 0x7F030172
			public const int expandedHintEnabled = 2130903410;
			
			// aapt resource value: 0x7F030173
			public const int expandedTitleGravity = 2130903411;
			
			// aapt resource value: 0x7F030174
			public const int expandedTitleMargin = 2130903412;
			
			// aapt resource value: 0x7F030175
			public const int expandedTitleMarginBottom = 2130903413;
			
			// aapt resource value: 0x7F030176
			public const int expandedTitleMarginEnd = 2130903414;
			
			// aapt resource value: 0x7F030177
			public const int expandedTitleMarginStart = 2130903415;
			
			// aapt resource value: 0x7F030178
			public const int expandedTitleMarginTop = 2130903416;
			
			// aapt resource value: 0x7F030179
			public const int expandedTitleTextAppearance = 2130903417;
			
			// aapt resource value: 0x7F03017B
			public const int extendedFloatingActionButtonStyle = 2130903419;
			
			// aapt resource value: 0x7F03017A
			public const int extendMotionSpec = 2130903418;
			
			// aapt resource value: 0x7F03017C
			public const int extraMultilineHeightEnabled = 2130903420;
			
			// aapt resource value: 0x7F03017D
			public const int fabAlignmentMode = 2130903421;
			
			// aapt resource value: 0x7F03017E
			public const int fabAnimationMode = 2130903422;
			
			// aapt resource value: 0x7F03017F
			public const int fabCradleMargin = 2130903423;
			
			// aapt resource value: 0x7F030180
			public const int fabCradleRoundedCornerRadius = 2130903424;
			
			// aapt resource value: 0x7F030181
			public const int fabCradleVerticalOffset = 2130903425;
			
			// aapt resource value: 0x7F030182
			public const int fabCustomSize = 2130903426;
			
			// aapt resource value: 0x7F030183
			public const int fabSize = 2130903427;
			
			// aapt resource value: 0x7F030184
			public const int fastScrollEnabled = 2130903428;
			
			// aapt resource value: 0x7F030185
			public const int fastScrollHorizontalThumbDrawable = 2130903429;
			
			// aapt resource value: 0x7F030186
			public const int fastScrollHorizontalTrackDrawable = 2130903430;
			
			// aapt resource value: 0x7F030187
			public const int fastScrollVerticalThumbDrawable = 2130903431;
			
			// aapt resource value: 0x7F030188
			public const int fastScrollVerticalTrackDrawable = 2130903432;
			
			// aapt resource value: 0x7F030189
			public const int firstBaselineToTopHeight = 2130903433;
			
			// aapt resource value: 0x7F03018A
			public const int floatingActionButtonStyle = 2130903434;
			
			// aapt resource value: 0x7F03018B
			public const int flow_firstHorizontalBias = 2130903435;
			
			// aapt resource value: 0x7F03018C
			public const int flow_firstHorizontalStyle = 2130903436;
			
			// aapt resource value: 0x7F03018D
			public const int flow_firstVerticalBias = 2130903437;
			
			// aapt resource value: 0x7F03018E
			public const int flow_firstVerticalStyle = 2130903438;
			
			// aapt resource value: 0x7F03018F
			public const int flow_horizontalAlign = 2130903439;
			
			// aapt resource value: 0x7F030190
			public const int flow_horizontalBias = 2130903440;
			
			// aapt resource value: 0x7F030191
			public const int flow_horizontalGap = 2130903441;
			
			// aapt resource value: 0x7F030192
			public const int flow_horizontalStyle = 2130903442;
			
			// aapt resource value: 0x7F030193
			public const int flow_lastHorizontalBias = 2130903443;
			
			// aapt resource value: 0x7F030194
			public const int flow_lastHorizontalStyle = 2130903444;
			
			// aapt resource value: 0x7F030195
			public const int flow_lastVerticalBias = 2130903445;
			
			// aapt resource value: 0x7F030196
			public const int flow_lastVerticalStyle = 2130903446;
			
			// aapt resource value: 0x7F030197
			public const int flow_maxElementsWrap = 2130903447;
			
			// aapt resource value: 0x7F030198
			public const int flow_padding = 2130903448;
			
			// aapt resource value: 0x7F030199
			public const int flow_verticalAlign = 2130903449;
			
			// aapt resource value: 0x7F03019A
			public const int flow_verticalBias = 2130903450;
			
			// aapt resource value: 0x7F03019B
			public const int flow_verticalGap = 2130903451;
			
			// aapt resource value: 0x7F03019C
			public const int flow_verticalStyle = 2130903452;
			
			// aapt resource value: 0x7F03019D
			public const int flow_wrapMode = 2130903453;
			
			// aapt resource value: 0x7F03019E
			public const int font = 2130903454;
			
			// aapt resource value: 0x7F03019F
			public const int fontFamily = 2130903455;
			
			// aapt resource value: 0x7F0301A0
			public const int fontProviderAuthority = 2130903456;
			
			// aapt resource value: 0x7F0301A1
			public const int fontProviderCerts = 2130903457;
			
			// aapt resource value: 0x7F0301A2
			public const int fontProviderFetchStrategy = 2130903458;
			
			// aapt resource value: 0x7F0301A3
			public const int fontProviderFetchTimeout = 2130903459;
			
			// aapt resource value: 0x7F0301A4
			public const int fontProviderPackage = 2130903460;
			
			// aapt resource value: 0x7F0301A5
			public const int fontProviderQuery = 2130903461;
			
			// aapt resource value: 0x7F0301A6
			public const int fontProviderSystemFontFamily = 2130903462;
			
			// aapt resource value: 0x7F0301A7
			public const int fontStyle = 2130903463;
			
			// aapt resource value: 0x7F0301A8
			public const int fontVariationSettings = 2130903464;
			
			// aapt resource value: 0x7F0301A9
			public const int fontWeight = 2130903465;
			
			// aapt resource value: 0x7F0301AA
			public const int forceApplySystemWindowInsetTop = 2130903466;
			
			// aapt resource value: 0x7F0301AB
			public const int foregroundInsidePadding = 2130903467;
			
			// aapt resource value: 0x7F0301AC
			public const int fragment = 2130903468;
			
			// aapt resource value: 0x7F0301AD
			public const int framePosition = 2130903469;
			
			// aapt resource value: 0x7F0301AE
			public const int gapBetweenBars = 2130903470;
			
			// aapt resource value: 0x7F0301AF
			public const int gestureInsetBottomIgnored = 2130903471;
			
			// aapt resource value: 0x7F0301B0
			public const int goIcon = 2130903472;
			
			// aapt resource value: 0x7F0301B1
			public const int graph = 2130903473;
			
			// aapt resource value: 0x7F0301B2
			public const int haloColor = 2130903474;
			
			// aapt resource value: 0x7F0301B3
			public const int haloRadius = 2130903475;
			
			// aapt resource value: 0x7F0301B4
			public const int headerLayout = 2130903476;
			
			// aapt resource value: 0x7F0301B5
			public const int height = 2130903477;
			
			// aapt resource value: 0x7F0301B6
			public const int helperText = 2130903478;
			
			// aapt resource value: 0x7F0301B7
			public const int helperTextEnabled = 2130903479;
			
			// aapt resource value: 0x7F0301B8
			public const int helperTextTextAppearance = 2130903480;
			
			// aapt resource value: 0x7F0301B9
			public const int helperTextTextColor = 2130903481;
			
			// aapt resource value: 0x7F0301BA
			public const int hideAnimationBehavior = 2130903482;
			
			// aapt resource value: 0x7F0301BB
			public const int hideMotionSpec = 2130903483;
			
			// aapt resource value: 0x7F0301BC
			public const int hideOnContentScroll = 2130903484;
			
			// aapt resource value: 0x7F0301BD
			public const int hideOnScroll = 2130903485;
			
			// aapt resource value: 0x7F0301BE
			public const int hintAnimationEnabled = 2130903486;
			
			// aapt resource value: 0x7F0301BF
			public const int hintEnabled = 2130903487;
			
			// aapt resource value: 0x7F0301C0
			public const int hintTextAppearance = 2130903488;
			
			// aapt resource value: 0x7F0301C1
			public const int hintTextColor = 2130903489;
			
			// aapt resource value: 0x7F0301C2
			public const int homeAsUpIndicator = 2130903490;
			
			// aapt resource value: 0x7F0301C3
			public const int homeLayout = 2130903491;
			
			// aapt resource value: 0x7F0301C4
			public const int horizontalOffset = 2130903492;
			
			// aapt resource value: 0x7F0301C5
			public const int hoveredFocusedTranslationZ = 2130903493;
			
			// aapt resource value: 0x7F0301C6
			public const int icon = 2130903494;
			
			// aapt resource value: 0x7F0301C7
			public const int iconEndPadding = 2130903495;
			
			// aapt resource value: 0x7F0301C8
			public const int iconGravity = 2130903496;
			
			// aapt resource value: 0x7F0301CF
			public const int iconifiedByDefault = 2130903503;
			
			// aapt resource value: 0x7F0301C9
			public const int iconPadding = 2130903497;
			
			// aapt resource value: 0x7F0301CA
			public const int iconSize = 2130903498;
			
			// aapt resource value: 0x7F0301CB
			public const int iconSpaceReserved = 2130903499;
			
			// aapt resource value: 0x7F0301CC
			public const int iconStartPadding = 2130903500;
			
			// aapt resource value: 0x7F0301CD
			public const int iconTint = 2130903501;
			
			// aapt resource value: 0x7F0301CE
			public const int iconTintMode = 2130903502;
			
			// aapt resource value: 0x7F0301D0
			public const int ifTagNotSet = 2130903504;
			
			// aapt resource value: 0x7F0301D1
			public const int ifTagSet = 2130903505;
			
			// aapt resource value: 0x7F0301D2
			public const int imageButtonStyle = 2130903506;
			
			// aapt resource value: 0x7F0301D3
			public const int imagePanX = 2130903507;
			
			// aapt resource value: 0x7F0301D4
			public const int imagePanY = 2130903508;
			
			// aapt resource value: 0x7F0301D5
			public const int imageRotate = 2130903509;
			
			// aapt resource value: 0x7F0301D6
			public const int imageZoom = 2130903510;
			
			// aapt resource value: 0x7F0301D7
			public const int indeterminateAnimationType = 2130903511;
			
			// aapt resource value: 0x7F0301D8
			public const int indeterminateProgressStyle = 2130903512;
			
			// aapt resource value: 0x7F0301D9
			public const int indicatorColor = 2130903513;
			
			// aapt resource value: 0x7F0301DA
			public const int indicatorDirectionCircular = 2130903514;
			
			// aapt resource value: 0x7F0301DB
			public const int indicatorDirectionLinear = 2130903515;
			
			// aapt resource value: 0x7F0301DC
			public const int indicatorInset = 2130903516;
			
			// aapt resource value: 0x7F0301DD
			public const int indicatorSize = 2130903517;
			
			// aapt resource value: 0x7F0301DE
			public const int initialActivityCount = 2130903518;
			
			// aapt resource value: 0x7F0301DF
			public const int initialExpandedChildrenCount = 2130903519;
			
			// aapt resource value: 0x7F0301E0
			public const int insetForeground = 2130903520;
			
			// aapt resource value: 0x7F0301E1
			public const int isLightTheme = 2130903521;
			
			// aapt resource value: 0x7F0301E2
			public const int isMaterialTheme = 2130903522;
			
			// aapt resource value: 0x7F0301E3
			public const int isPreferenceVisible = 2130903523;
			
			// aapt resource value: 0x7F0301E4
			public const int itemBackground = 2130903524;
			
			// aapt resource value: 0x7F0301E5
			public const int itemFillColor = 2130903525;
			
			// aapt resource value: 0x7F0301E6
			public const int itemHorizontalPadding = 2130903526;
			
			// aapt resource value: 0x7F0301E7
			public const int itemHorizontalTranslationEnabled = 2130903527;
			
			// aapt resource value: 0x7F0301E8
			public const int itemIconPadding = 2130903528;
			
			// aapt resource value: 0x7F0301E9
			public const int itemIconSize = 2130903529;
			
			// aapt resource value: 0x7F0301EA
			public const int itemIconTint = 2130903530;
			
			// aapt resource value: 0x7F0301EB
			public const int itemMaxLines = 2130903531;
			
			// aapt resource value: 0x7F0301EC
			public const int itemPadding = 2130903532;
			
			// aapt resource value: 0x7F0301ED
			public const int itemRippleColor = 2130903533;
			
			// aapt resource value: 0x7F0301EE
			public const int itemShapeAppearance = 2130903534;
			
			// aapt resource value: 0x7F0301EF
			public const int itemShapeAppearanceOverlay = 2130903535;
			
			// aapt resource value: 0x7F0301F0
			public const int itemShapeFillColor = 2130903536;
			
			// aapt resource value: 0x7F0301F1
			public const int itemShapeInsetBottom = 2130903537;
			
			// aapt resource value: 0x7F0301F2
			public const int itemShapeInsetEnd = 2130903538;
			
			// aapt resource value: 0x7F0301F3
			public const int itemShapeInsetStart = 2130903539;
			
			// aapt resource value: 0x7F0301F4
			public const int itemShapeInsetTop = 2130903540;
			
			// aapt resource value: 0x7F0301F5
			public const int itemSpacing = 2130903541;
			
			// aapt resource value: 0x7F0301F6
			public const int itemStrokeColor = 2130903542;
			
			// aapt resource value: 0x7F0301F7
			public const int itemStrokeWidth = 2130903543;
			
			// aapt resource value: 0x7F0301F8
			public const int itemTextAppearance = 2130903544;
			
			// aapt resource value: 0x7F0301F9
			public const int itemTextAppearanceActive = 2130903545;
			
			// aapt resource value: 0x7F0301FA
			public const int itemTextAppearanceInactive = 2130903546;
			
			// aapt resource value: 0x7F0301FB
			public const int itemTextColor = 2130903547;
			
			// aapt resource value: 0x7F0301FC
			public const int key = 2130903548;
			
			// aapt resource value: 0x7F0301FE
			public const int keyboardIcon = 2130903550;
			
			// aapt resource value: 0x7F0301FF
			public const int keylines = 2130903551;
			
			// aapt resource value: 0x7F0301FD
			public const int keyPositionType = 2130903549;
			
			// aapt resource value: 0x7F030200
			public const int labelBehavior = 2130903552;
			
			// aapt resource value: 0x7F030201
			public const int labelStyle = 2130903553;
			
			// aapt resource value: 0x7F030202
			public const int labelVisibilityMode = 2130903554;
			
			// aapt resource value: 0x7F030203
			public const int lastBaselineToBottomHeight = 2130903555;
			
			// aapt resource value: 0x7F030204
			public const int launchSingleTop = 2130903556;
			
			// aapt resource value: 0x7F030205
			public const int layout = 2130903557;
			
			// aapt resource value: 0x7F030206
			public const int layoutDescription = 2130903558;
			
			// aapt resource value: 0x7F030207
			public const int layoutDuringTransition = 2130903559;
			
			// aapt resource value: 0x7F030208
			public const int layoutManager = 2130903560;
			
			// aapt resource value: 0x7F030209
			public const int layout_anchor = 2130903561;
			
			// aapt resource value: 0x7F03020A
			public const int layout_anchorGravity = 2130903562;
			
			// aapt resource value: 0x7F03020B
			public const int layout_behavior = 2130903563;
			
			// aapt resource value: 0x7F03020C
			public const int layout_collapseMode = 2130903564;
			
			// aapt resource value: 0x7F03020D
			public const int layout_collapseParallaxMultiplier = 2130903565;
			
			// aapt resource value: 0x7F03020E
			public const int layout_constrainedHeight = 2130903566;
			
			// aapt resource value: 0x7F03020F
			public const int layout_constrainedWidth = 2130903567;
			
			// aapt resource value: 0x7F030210
			public const int layout_constraintBaseline_creator = 2130903568;
			
			// aapt resource value: 0x7F030211
			public const int layout_constraintBaseline_toBaselineOf = 2130903569;
			
			// aapt resource value: 0x7F030212
			public const int layout_constraintBaseline_toBottomOf = 2130903570;
			
			// aapt resource value: 0x7F030213
			public const int layout_constraintBaseline_toTopOf = 2130903571;
			
			// aapt resource value: 0x7F030214
			public const int layout_constraintBottom_creator = 2130903572;
			
			// aapt resource value: 0x7F030215
			public const int layout_constraintBottom_toBottomOf = 2130903573;
			
			// aapt resource value: 0x7F030216
			public const int layout_constraintBottom_toTopOf = 2130903574;
			
			// aapt resource value: 0x7F030217
			public const int layout_constraintCircle = 2130903575;
			
			// aapt resource value: 0x7F030218
			public const int layout_constraintCircleAngle = 2130903576;
			
			// aapt resource value: 0x7F030219
			public const int layout_constraintCircleRadius = 2130903577;
			
			// aapt resource value: 0x7F03021A
			public const int layout_constraintDimensionRatio = 2130903578;
			
			// aapt resource value: 0x7F03021B
			public const int layout_constraintEnd_toEndOf = 2130903579;
			
			// aapt resource value: 0x7F03021C
			public const int layout_constraintEnd_toStartOf = 2130903580;
			
			// aapt resource value: 0x7F03021D
			public const int layout_constraintGuide_begin = 2130903581;
			
			// aapt resource value: 0x7F03021E
			public const int layout_constraintGuide_end = 2130903582;
			
			// aapt resource value: 0x7F03021F
			public const int layout_constraintGuide_percent = 2130903583;
			
			// aapt resource value: 0x7F030220
			public const int layout_constraintHeight = 2130903584;
			
			// aapt resource value: 0x7F030221
			public const int layout_constraintHeight_default = 2130903585;
			
			// aapt resource value: 0x7F030222
			public const int layout_constraintHeight_max = 2130903586;
			
			// aapt resource value: 0x7F030223
			public const int layout_constraintHeight_min = 2130903587;
			
			// aapt resource value: 0x7F030224
			public const int layout_constraintHeight_percent = 2130903588;
			
			// aapt resource value: 0x7F030225
			public const int layout_constraintHorizontal_bias = 2130903589;
			
			// aapt resource value: 0x7F030226
			public const int layout_constraintHorizontal_chainStyle = 2130903590;
			
			// aapt resource value: 0x7F030227
			public const int layout_constraintHorizontal_weight = 2130903591;
			
			// aapt resource value: 0x7F030228
			public const int layout_constraintLeft_creator = 2130903592;
			
			// aapt resource value: 0x7F030229
			public const int layout_constraintLeft_toLeftOf = 2130903593;
			
			// aapt resource value: 0x7F03022A
			public const int layout_constraintLeft_toRightOf = 2130903594;
			
			// aapt resource value: 0x7F03022B
			public const int layout_constraintRight_creator = 2130903595;
			
			// aapt resource value: 0x7F03022C
			public const int layout_constraintRight_toLeftOf = 2130903596;
			
			// aapt resource value: 0x7F03022D
			public const int layout_constraintRight_toRightOf = 2130903597;
			
			// aapt resource value: 0x7F03022E
			public const int layout_constraintStart_toEndOf = 2130903598;
			
			// aapt resource value: 0x7F03022F
			public const int layout_constraintStart_toStartOf = 2130903599;
			
			// aapt resource value: 0x7F030230
			public const int layout_constraintTag = 2130903600;
			
			// aapt resource value: 0x7F030231
			public const int layout_constraintTop_creator = 2130903601;
			
			// aapt resource value: 0x7F030232
			public const int layout_constraintTop_toBottomOf = 2130903602;
			
			// aapt resource value: 0x7F030233
			public const int layout_constraintTop_toTopOf = 2130903603;
			
			// aapt resource value: 0x7F030234
			public const int layout_constraintVertical_bias = 2130903604;
			
			// aapt resource value: 0x7F030235
			public const int layout_constraintVertical_chainStyle = 2130903605;
			
			// aapt resource value: 0x7F030236
			public const int layout_constraintVertical_weight = 2130903606;
			
			// aapt resource value: 0x7F030237
			public const int layout_constraintWidth = 2130903607;
			
			// aapt resource value: 0x7F030238
			public const int layout_constraintWidth_default = 2130903608;
			
			// aapt resource value: 0x7F030239
			public const int layout_constraintWidth_max = 2130903609;
			
			// aapt resource value: 0x7F03023A
			public const int layout_constraintWidth_min = 2130903610;
			
			// aapt resource value: 0x7F03023B
			public const int layout_constraintWidth_percent = 2130903611;
			
			// aapt resource value: 0x7F03023C
			public const int layout_dodgeInsetEdges = 2130903612;
			
			// aapt resource value: 0x7F03023D
			public const int layout_editor_absoluteX = 2130903613;
			
			// aapt resource value: 0x7F03023E
			public const int layout_editor_absoluteY = 2130903614;
			
			// aapt resource value: 0x7F03023F
			public const int layout_goneMarginBaseline = 2130903615;
			
			// aapt resource value: 0x7F030240
			public const int layout_goneMarginBottom = 2130903616;
			
			// aapt resource value: 0x7F030241
			public const int layout_goneMarginEnd = 2130903617;
			
			// aapt resource value: 0x7F030242
			public const int layout_goneMarginLeft = 2130903618;
			
			// aapt resource value: 0x7F030243
			public const int layout_goneMarginRight = 2130903619;
			
			// aapt resource value: 0x7F030244
			public const int layout_goneMarginStart = 2130903620;
			
			// aapt resource value: 0x7F030245
			public const int layout_goneMarginTop = 2130903621;
			
			// aapt resource value: 0x7F030246
			public const int layout_insetEdge = 2130903622;
			
			// aapt resource value: 0x7F030247
			public const int layout_keyline = 2130903623;
			
			// aapt resource value: 0x7F030248
			public const int layout_marginBaseline = 2130903624;
			
			// aapt resource value: 0x7F030249
			public const int layout_optimizationLevel = 2130903625;
			
			// aapt resource value: 0x7F03024A
			public const int layout_scrollFlags = 2130903626;
			
			// aapt resource value: 0x7F03024B
			public const int layout_scrollInterpolator = 2130903627;
			
			// aapt resource value: 0x7F03024C
			public const int layout_wrapBehaviorInParent = 2130903628;
			
			// aapt resource value: 0x7F03024D
			public const int liftOnScroll = 2130903629;
			
			// aapt resource value: 0x7F03024E
			public const int liftOnScrollTargetViewId = 2130903630;
			
			// aapt resource value: 0x7F03024F
			public const int limitBoundsTo = 2130903631;
			
			// aapt resource value: 0x7F030252
			public const int linearProgressIndicatorStyle = 2130903634;
			
			// aapt resource value: 0x7F030250
			public const int lineHeight = 2130903632;
			
			// aapt resource value: 0x7F030251
			public const int lineSpacing = 2130903633;
			
			// aapt resource value: 0x7F030253
			public const int listChoiceBackgroundIndicator = 2130903635;
			
			// aapt resource value: 0x7F030254
			public const int listChoiceIndicatorMultipleAnimated = 2130903636;
			
			// aapt resource value: 0x7F030255
			public const int listChoiceIndicatorSingleAnimated = 2130903637;
			
			// aapt resource value: 0x7F030256
			public const int listDividerAlertDialog = 2130903638;
			
			// aapt resource value: 0x7F030257
			public const int listItemLayout = 2130903639;
			
			// aapt resource value: 0x7F030258
			public const int listLayout = 2130903640;
			
			// aapt resource value: 0x7F030259
			public const int listMenuViewStyle = 2130903641;
			
			// aapt resource value: 0x7F03025A
			public const int listPopupWindowStyle = 2130903642;
			
			// aapt resource value: 0x7F03025B
			public const int listPreferredItemHeight = 2130903643;
			
			// aapt resource value: 0x7F03025C
			public const int listPreferredItemHeightLarge = 2130903644;
			
			// aapt resource value: 0x7F03025D
			public const int listPreferredItemHeightSmall = 2130903645;
			
			// aapt resource value: 0x7F03025E
			public const int listPreferredItemPaddingEnd = 2130903646;
			
			// aapt resource value: 0x7F03025F
			public const int listPreferredItemPaddingLeft = 2130903647;
			
			// aapt resource value: 0x7F030260
			public const int listPreferredItemPaddingRight = 2130903648;
			
			// aapt resource value: 0x7F030261
			public const int listPreferredItemPaddingStart = 2130903649;
			
			// aapt resource value: 0x7F030262
			public const int logo = 2130903650;
			
			// aapt resource value: 0x7F030263
			public const int logoDescription = 2130903651;
			
			// aapt resource value: 0x7F030264
			public const int materialAlertDialogBodyTextStyle = 2130903652;
			
			// aapt resource value: 0x7F030265
			public const int materialAlertDialogTheme = 2130903653;
			
			// aapt resource value: 0x7F030266
			public const int materialAlertDialogTitleIconStyle = 2130903654;
			
			// aapt resource value: 0x7F030267
			public const int materialAlertDialogTitlePanelStyle = 2130903655;
			
			// aapt resource value: 0x7F030268
			public const int materialAlertDialogTitleTextStyle = 2130903656;
			
			// aapt resource value: 0x7F030269
			public const int materialButtonOutlinedStyle = 2130903657;
			
			// aapt resource value: 0x7F03026A
			public const int materialButtonStyle = 2130903658;
			
			// aapt resource value: 0x7F03026B
			public const int materialButtonToggleGroupStyle = 2130903659;
			
			// aapt resource value: 0x7F03026C
			public const int materialCalendarDay = 2130903660;
			
			// aapt resource value: 0x7F03026D
			public const int materialCalendarFullscreenTheme = 2130903661;
			
			// aapt resource value: 0x7F03026E
			public const int materialCalendarHeaderCancelButton = 2130903662;
			
			// aapt resource value: 0x7F03026F
			public const int materialCalendarHeaderConfirmButton = 2130903663;
			
			// aapt resource value: 0x7F030270
			public const int materialCalendarHeaderDivider = 2130903664;
			
			// aapt resource value: 0x7F030271
			public const int materialCalendarHeaderLayout = 2130903665;
			
			// aapt resource value: 0x7F030272
			public const int materialCalendarHeaderSelection = 2130903666;
			
			// aapt resource value: 0x7F030273
			public const int materialCalendarHeaderTitle = 2130903667;
			
			// aapt resource value: 0x7F030274
			public const int materialCalendarHeaderToggleButton = 2130903668;
			
			// aapt resource value: 0x7F030275
			public const int materialCalendarMonth = 2130903669;
			
			// aapt resource value: 0x7F030276
			public const int materialCalendarMonthNavigationButton = 2130903670;
			
			// aapt resource value: 0x7F030277
			public const int materialCalendarStyle = 2130903671;
			
			// aapt resource value: 0x7F030278
			public const int materialCalendarTheme = 2130903672;
			
			// aapt resource value: 0x7F030279
			public const int materialCalendarYearNavigationButton = 2130903673;
			
			// aapt resource value: 0x7F03027A
			public const int materialCardViewStyle = 2130903674;
			
			// aapt resource value: 0x7F03027B
			public const int materialCircleRadius = 2130903675;
			
			// aapt resource value: 0x7F03027C
			public const int materialClockStyle = 2130903676;
			
			// aapt resource value: 0x7F03027D
			public const int materialThemeOverlay = 2130903677;
			
			// aapt resource value: 0x7F03027E
			public const int materialTimePickerStyle = 2130903678;
			
			// aapt resource value: 0x7F03027F
			public const int materialTimePickerTheme = 2130903679;
			
			// aapt resource value: 0x7F030280
			public const int maxAcceleration = 2130903680;
			
			// aapt resource value: 0x7F030281
			public const int maxActionInlineWidth = 2130903681;
			
			// aapt resource value: 0x7F030282
			public const int maxButtonHeight = 2130903682;
			
			// aapt resource value: 0x7F030283
			public const int maxCharacterCount = 2130903683;
			
			// aapt resource value: 0x7F030284
			public const int maxHeight = 2130903684;
			
			// aapt resource value: 0x7F030285
			public const int maxImageSize = 2130903685;
			
			// aapt resource value: 0x7F030286
			public const int maxLines = 2130903686;
			
			// aapt resource value: 0x7F030287
			public const int maxVelocity = 2130903687;
			
			// aapt resource value: 0x7F030288
			public const int maxWidth = 2130903688;
			
			// aapt resource value: 0x7F030289
			public const int measureWithLargestChild = 2130903689;
			
			// aapt resource value: 0x7F03028A
			public const int menu = 2130903690;
			
			// aapt resource value: 0x7F03028B
			public const int menuGravity = 2130903691;
			
			// aapt resource value: 0x7F03028C
			public const int methodName = 2130903692;
			
			// aapt resource value: 0x7F03028D
			public const int mimeType = 2130903693;
			
			// aapt resource value: 0x7F03028E
			public const int min = 2130903694;
			
			// aapt resource value: 0x7F03028F
			public const int minHeight = 2130903695;
			
			// aapt resource value: 0x7F030290
			public const int minHideDelay = 2130903696;
			
			// aapt resource value: 0x7F030291
			public const int minSeparation = 2130903697;
			
			// aapt resource value: 0x7F030292
			public const int minTouchTargetSize = 2130903698;
			
			// aapt resource value: 0x7F030293
			public const int minWidth = 2130903699;
			
			// aapt resource value: 0x7F030294
			public const int mock_diagonalsColor = 2130903700;
			
			// aapt resource value: 0x7F030295
			public const int mock_label = 2130903701;
			
			// aapt resource value: 0x7F030296
			public const int mock_labelBackgroundColor = 2130903702;
			
			// aapt resource value: 0x7F030297
			public const int mock_labelColor = 2130903703;
			
			// aapt resource value: 0x7F030298
			public const int mock_showDiagonals = 2130903704;
			
			// aapt resource value: 0x7F030299
			public const int mock_showLabel = 2130903705;
			
			// aapt resource value: 0x7F03029A
			public const int motionDebug = 2130903706;
			
			// aapt resource value: 0x7F03029B
			public const int motionDurationLong1 = 2130903707;
			
			// aapt resource value: 0x7F03029C
			public const int motionDurationLong2 = 2130903708;
			
			// aapt resource value: 0x7F03029D
			public const int motionDurationMedium1 = 2130903709;
			
			// aapt resource value: 0x7F03029E
			public const int motionDurationMedium2 = 2130903710;
			
			// aapt resource value: 0x7F03029F
			public const int motionDurationShort1 = 2130903711;
			
			// aapt resource value: 0x7F0302A0
			public const int motionDurationShort2 = 2130903712;
			
			// aapt resource value: 0x7F0302A1
			public const int motionEasingAccelerated = 2130903713;
			
			// aapt resource value: 0x7F0302A2
			public const int motionEasingDecelerated = 2130903714;
			
			// aapt resource value: 0x7F0302A3
			public const int motionEasingEmphasized = 2130903715;
			
			// aapt resource value: 0x7F0302A4
			public const int motionEasingLinear = 2130903716;
			
			// aapt resource value: 0x7F0302A5
			public const int motionEasingStandard = 2130903717;
			
			// aapt resource value: 0x7F0302A6
			public const int motionEffect_alpha = 2130903718;
			
			// aapt resource value: 0x7F0302A7
			public const int motionEffect_end = 2130903719;
			
			// aapt resource value: 0x7F0302A8
			public const int motionEffect_move = 2130903720;
			
			// aapt resource value: 0x7F0302A9
			public const int motionEffect_start = 2130903721;
			
			// aapt resource value: 0x7F0302AA
			public const int motionEffect_strict = 2130903722;
			
			// aapt resource value: 0x7F0302AB
			public const int motionEffect_translationX = 2130903723;
			
			// aapt resource value: 0x7F0302AC
			public const int motionEffect_translationY = 2130903724;
			
			// aapt resource value: 0x7F0302AD
			public const int motionEffect_viewTransition = 2130903725;
			
			// aapt resource value: 0x7F0302AE
			public const int motionInterpolator = 2130903726;
			
			// aapt resource value: 0x7F0302AF
			public const int motionPath = 2130903727;
			
			// aapt resource value: 0x7F0302B0
			public const int motionPathRotate = 2130903728;
			
			// aapt resource value: 0x7F0302B1
			public const int motionProgress = 2130903729;
			
			// aapt resource value: 0x7F0302B2
			public const int motionStagger = 2130903730;
			
			// aapt resource value: 0x7F0302B3
			public const int motionTarget = 2130903731;
			
			// aapt resource value: 0x7F0302B4
			public const int motion_postLayoutCollision = 2130903732;
			
			// aapt resource value: 0x7F0302B5
			public const int motion_triggerOnCollision = 2130903733;
			
			// aapt resource value: 0x7F0302B6
			public const int moveWhenScrollAtTop = 2130903734;
			
			// aapt resource value: 0x7F0302B7
			public const int multiChoiceItemLayout = 2130903735;
			
			// aapt resource value: 0x7F0302B8
			public const int navGraph = 2130903736;
			
			// aapt resource value: 0x7F0302B9
			public const int navigationContentDescription = 2130903737;
			
			// aapt resource value: 0x7F0302BA
			public const int navigationIcon = 2130903738;
			
			// aapt resource value: 0x7F0302BB
			public const int navigationIconTint = 2130903739;
			
			// aapt resource value: 0x7F0302BC
			public const int navigationMode = 2130903740;
			
			// aapt resource value: 0x7F0302BD
			public const int navigationRailStyle = 2130903741;
			
			// aapt resource value: 0x7F0302BE
			public const int navigationViewStyle = 2130903742;
			
			// aapt resource value: 0x7F0302BF
			public const int negativeButtonText = 2130903743;
			
			// aapt resource value: 0x7F0302C2
			public const int nestedScrollable = 2130903746;
			
			// aapt resource value: 0x7F0302C0
			public const int nestedScrollFlags = 2130903744;
			
			// aapt resource value: 0x7F0302C1
			public const int nestedScrollViewStyle = 2130903745;
			
			// aapt resource value: 0x7F0302C3
			public const int nullable = 2130903747;
			
			// aapt resource value: 0x7F0302C4
			public const int number = 2130903748;
			
			// aapt resource value: 0x7F0302C5
			public const int numericModifiers = 2130903749;
			
			// aapt resource value: 0x7F0302C6
			public const int onCross = 2130903750;
			
			// aapt resource value: 0x7F0302C7
			public const int onHide = 2130903751;
			
			// aapt resource value: 0x7F0302C8
			public const int onNegativeCross = 2130903752;
			
			// aapt resource value: 0x7F0302C9
			public const int onPositiveCross = 2130903753;
			
			// aapt resource value: 0x7F0302CA
			public const int onShow = 2130903754;
			
			// aapt resource value: 0x7F0302CB
			public const int onStateTransition = 2130903755;
			
			// aapt resource value: 0x7F0302CC
			public const int onTouchUp = 2130903756;
			
			// aapt resource value: 0x7F0302CD
			public const int order = 2130903757;
			
			// aapt resource value: 0x7F0302CE
			public const int orderingFromXml = 2130903758;
			
			// aapt resource value: 0x7F0302CF
			public const int overlapAnchor = 2130903759;
			
			// aapt resource value: 0x7F0302D0
			public const int overlay = 2130903760;
			
			// aapt resource value: 0x7F0302D1
			public const int paddingBottomNoButtons = 2130903761;
			
			// aapt resource value: 0x7F0302D2
			public const int paddingBottomSystemWindowInsets = 2130903762;
			
			// aapt resource value: 0x7F0302D3
			public const int paddingEnd = 2130903763;
			
			// aapt resource value: 0x7F0302D4
			public const int paddingLeftSystemWindowInsets = 2130903764;
			
			// aapt resource value: 0x7F0302D5
			public const int paddingRightSystemWindowInsets = 2130903765;
			
			// aapt resource value: 0x7F0302D6
			public const int paddingStart = 2130903766;
			
			// aapt resource value: 0x7F0302D7
			public const int paddingTopNoTitle = 2130903767;
			
			// aapt resource value: 0x7F0302D8
			public const int paddingTopSystemWindowInsets = 2130903768;
			
			// aapt resource value: 0x7F0302D9
			public const int panelBackground = 2130903769;
			
			// aapt resource value: 0x7F0302DA
			public const int panelMenuListTheme = 2130903770;
			
			// aapt resource value: 0x7F0302DB
			public const int panelMenuListWidth = 2130903771;
			
			// aapt resource value: 0x7F0302DC
			public const int passwordToggleContentDescription = 2130903772;
			
			// aapt resource value: 0x7F0302DD
			public const int passwordToggleDrawable = 2130903773;
			
			// aapt resource value: 0x7F0302DE
			public const int passwordToggleEnabled = 2130903774;
			
			// aapt resource value: 0x7F0302DF
			public const int passwordToggleTint = 2130903775;
			
			// aapt resource value: 0x7F0302E0
			public const int passwordToggleTintMode = 2130903776;
			
			// aapt resource value: 0x7F0302E1
			public const int pathMotionArc = 2130903777;
			
			// aapt resource value: 0x7F0302E2
			public const int path_percent = 2130903778;
			
			// aapt resource value: 0x7F0302E3
			public const int percentHeight = 2130903779;
			
			// aapt resource value: 0x7F0302E4
			public const int percentWidth = 2130903780;
			
			// aapt resource value: 0x7F0302E5
			public const int percentX = 2130903781;
			
			// aapt resource value: 0x7F0302E6
			public const int percentY = 2130903782;
			
			// aapt resource value: 0x7F0302E7
			public const int perpendicularPath_percent = 2130903783;
			
			// aapt resource value: 0x7F0302E8
			public const int persistent = 2130903784;
			
			// aapt resource value: 0x7F0302E9
			public const int pivotAnchor = 2130903785;
			
			// aapt resource value: 0x7F0302EA
			public const int placeholderText = 2130903786;
			
			// aapt resource value: 0x7F0302EB
			public const int placeholderTextAppearance = 2130903787;
			
			// aapt resource value: 0x7F0302EC
			public const int placeholderTextColor = 2130903788;
			
			// aapt resource value: 0x7F0302ED
			public const int placeholder_emptyVisibility = 2130903789;
			
			// aapt resource value: 0x7F0302EE
			public const int polarRelativeTo = 2130903790;
			
			// aapt resource value: 0x7F0302EF
			public const int popEnterAnim = 2130903791;
			
			// aapt resource value: 0x7F0302F0
			public const int popExitAnim = 2130903792;
			
			// aapt resource value: 0x7F0302F3
			public const int popupMenuBackground = 2130903795;
			
			// aapt resource value: 0x7F0302F4
			public const int popupMenuStyle = 2130903796;
			
			// aapt resource value: 0x7F0302F5
			public const int popupTheme = 2130903797;
			
			// aapt resource value: 0x7F0302F1
			public const int popUpTo = 2130903793;
			
			// aapt resource value: 0x7F0302F2
			public const int popUpToInclusive = 2130903794;
			
			// aapt resource value: 0x7F0302F6
			public const int popupWindowStyle = 2130903798;
			
			// aapt resource value: 0x7F0302F7
			public const int positiveButtonText = 2130903799;
			
			// aapt resource value: 0x7F0302F8
			public const int preferenceCategoryStyle = 2130903800;
			
			// aapt resource value: 0x7F0302F9
			public const int preferenceCategoryTitleTextAppearance = 2130903801;
			
			// aapt resource value: 0x7F0302FA
			public const int preferenceFragmentCompatStyle = 2130903802;
			
			// aapt resource value: 0x7F0302FB
			public const int preferenceFragmentListStyle = 2130903803;
			
			// aapt resource value: 0x7F0302FC
			public const int preferenceFragmentStyle = 2130903804;
			
			// aapt resource value: 0x7F0302FD
			public const int preferenceInformationStyle = 2130903805;
			
			// aapt resource value: 0x7F0302FE
			public const int preferenceScreenStyle = 2130903806;
			
			// aapt resource value: 0x7F0302FF
			public const int preferenceStyle = 2130903807;
			
			// aapt resource value: 0x7F030300
			public const int preferenceTheme = 2130903808;
			
			// aapt resource value: 0x7F030301
			public const int prefixText = 2130903809;
			
			// aapt resource value: 0x7F030302
			public const int prefixTextAppearance = 2130903810;
			
			// aapt resource value: 0x7F030303
			public const int prefixTextColor = 2130903811;
			
			// aapt resource value: 0x7F030304
			public const int preserveIconSpacing = 2130903812;
			
			// aapt resource value: 0x7F030305
			public const int pressedTranslationZ = 2130903813;
			
			// aapt resource value: 0x7F030306
			public const int progressBarPadding = 2130903814;
			
			// aapt resource value: 0x7F030307
			public const int progressBarStyle = 2130903815;
			
			// aapt resource value: 0x7F030308
			public const int quantizeMotionInterpolator = 2130903816;
			
			// aapt resource value: 0x7F030309
			public const int quantizeMotionPhase = 2130903817;
			
			// aapt resource value: 0x7F03030A
			public const int quantizeMotionSteps = 2130903818;
			
			// aapt resource value: 0x7F03030B
			public const int queryBackground = 2130903819;
			
			// aapt resource value: 0x7F03030C
			public const int queryHint = 2130903820;
			
			// aapt resource value: 0x7F03030D
			public const int queryPatterns = 2130903821;
			
			// aapt resource value: 0x7F03030E
			public const int radioButtonStyle = 2130903822;
			
			// aapt resource value: 0x7F03030F
			public const int rangeFillColor = 2130903823;
			
			// aapt resource value: 0x7F030310
			public const int ratingBarStyle = 2130903824;
			
			// aapt resource value: 0x7F030311
			public const int ratingBarStyleIndicator = 2130903825;
			
			// aapt resource value: 0x7F030312
			public const int ratingBarStyleSmall = 2130903826;
			
			// aapt resource value: 0x7F030313
			public const int reactiveGuide_animateChange = 2130903827;
			
			// aapt resource value: 0x7F030314
			public const int reactiveGuide_applyToAllConstraintSets = 2130903828;
			
			// aapt resource value: 0x7F030315
			public const int reactiveGuide_applyToConstraintSet = 2130903829;
			
			// aapt resource value: 0x7F030316
			public const int reactiveGuide_valueId = 2130903830;
			
			// aapt resource value: 0x7F030317
			public const int recyclerViewStyle = 2130903831;
			
			// aapt resource value: 0x7F030318
			public const int region_heightLessThan = 2130903832;
			
			// aapt resource value: 0x7F030319
			public const int region_heightMoreThan = 2130903833;
			
			// aapt resource value: 0x7F03031A
			public const int region_widthLessThan = 2130903834;
			
			// aapt resource value: 0x7F03031B
			public const int region_widthMoreThan = 2130903835;
			
			// aapt resource value: 0x7F03031C
			public const int reverseLayout = 2130903836;
			
			// aapt resource value: 0x7F03031D
			public const int rippleColor = 2130903837;
			
			// aapt resource value: 0x7F03031E
			public const int rotationCenterId = 2130903838;
			
			// aapt resource value: 0x7F03031F
			public const int round = 2130903839;
			
			// aapt resource value: 0x7F030320
			public const int roundPercent = 2130903840;
			
			// aapt resource value: 0x7F030321
			public const int saturation = 2130903841;
			
			// aapt resource value: 0x7F030322
			public const int scaleFromTextSize = 2130903842;
			
			// aapt resource value: 0x7F030323
			public const int scrimAnimationDuration = 2130903843;
			
			// aapt resource value: 0x7F030324
			public const int scrimBackground = 2130903844;
			
			// aapt resource value: 0x7F030325
			public const int scrimVisibleHeightTrigger = 2130903845;
			
			// aapt resource value: 0x7F030326
			public const int scrollViewStyle = 2130903846;
			
			// aapt resource value: 0x7F030327
			public const int searchHintIcon = 2130903847;
			
			// aapt resource value: 0x7F030328
			public const int searchIcon = 2130903848;
			
			// aapt resource value: 0x7F030329
			public const int searchViewStyle = 2130903849;
			
			// aapt resource value: 0x7F03032A
			public const int seekBarIncrement = 2130903850;
			
			// aapt resource value: 0x7F03032B
			public const int seekBarPreferenceStyle = 2130903851;
			
			// aapt resource value: 0x7F03032C
			public const int seekBarStyle = 2130903852;
			
			// aapt resource value: 0x7F03032D
			public const int selectable = 2130903853;
			
			// aapt resource value: 0x7F03032E
			public const int selectableItemBackground = 2130903854;
			
			// aapt resource value: 0x7F03032F
			public const int selectableItemBackgroundBorderless = 2130903855;
			
			// aapt resource value: 0x7F030330
			public const int selectionRequired = 2130903856;
			
			// aapt resource value: 0x7F030331
			public const int selectorSize = 2130903857;
			
			// aapt resource value: 0x7F030332
			public const int setsTag = 2130903858;
			
			// aapt resource value: 0x7F030333
			public const int shapeAppearance = 2130903859;
			
			// aapt resource value: 0x7F030334
			public const int shapeAppearanceLargeComponent = 2130903860;
			
			// aapt resource value: 0x7F030335
			public const int shapeAppearanceMediumComponent = 2130903861;
			
			// aapt resource value: 0x7F030336
			public const int shapeAppearanceOverlay = 2130903862;
			
			// aapt resource value: 0x7F030337
			public const int shapeAppearanceSmallComponent = 2130903863;
			
			// aapt resource value: 0x7F030001
			public const int SharedValue = 2130903041;
			
			// aapt resource value: 0x7F030002
			public const int SharedValueId = 2130903042;
			
			// aapt resource value: 0x7F030338
			public const int shortcutMatchRequired = 2130903864;
			
			// aapt resource value: 0x7F030339
			public const int shouldDisableView = 2130903865;
			
			// aapt resource value: 0x7F03033A
			public const int showAnimationBehavior = 2130903866;
			
			// aapt resource value: 0x7F03033B
			public const int showAsAction = 2130903867;
			
			// aapt resource value: 0x7F03033C
			public const int showDelay = 2130903868;
			
			// aapt resource value: 0x7F03033D
			public const int showDividers = 2130903869;
			
			// aapt resource value: 0x7F03033E
			public const int showMotionSpec = 2130903870;
			
			// aapt resource value: 0x7F03033F
			public const int showPaths = 2130903871;
			
			// aapt resource value: 0x7F030340
			public const int showSeekBarValue = 2130903872;
			
			// aapt resource value: 0x7F030341
			public const int showText = 2130903873;
			
			// aapt resource value: 0x7F030342
			public const int showTitle = 2130903874;
			
			// aapt resource value: 0x7F030343
			public const int shrinkMotionSpec = 2130903875;
			
			// aapt resource value: 0x7F030344
			public const int singleChoiceItemLayout = 2130903876;
			
			// aapt resource value: 0x7F030345
			public const int singleLine = 2130903877;
			
			// aapt resource value: 0x7F030346
			public const int singleLineTitle = 2130903878;
			
			// aapt resource value: 0x7F030347
			public const int singleSelection = 2130903879;
			
			// aapt resource value: 0x7F030348
			public const int sizePercent = 2130903880;
			
			// aapt resource value: 0x7F030349
			public const int sliderStyle = 2130903881;
			
			// aapt resource value: 0x7F03034A
			public const int snackbarButtonStyle = 2130903882;
			
			// aapt resource value: 0x7F03034B
			public const int snackbarStyle = 2130903883;
			
			// aapt resource value: 0x7F03034C
			public const int snackbarTextViewStyle = 2130903884;
			
			// aapt resource value: 0x7F03034D
			public const int spanCount = 2130903885;
			
			// aapt resource value: 0x7F03034E
			public const int spinBars = 2130903886;
			
			// aapt resource value: 0x7F03034F
			public const int spinnerDropDownItemStyle = 2130903887;
			
			// aapt resource value: 0x7F030350
			public const int spinnerStyle = 2130903888;
			
			// aapt resource value: 0x7F030351
			public const int splitTrack = 2130903889;
			
			// aapt resource value: 0x7F030352
			public const int springBoundary = 2130903890;
			
			// aapt resource value: 0x7F030353
			public const int springDamping = 2130903891;
			
			// aapt resource value: 0x7F030354
			public const int springMass = 2130903892;
			
			// aapt resource value: 0x7F030355
			public const int springStiffness = 2130903893;
			
			// aapt resource value: 0x7F030356
			public const int springStopThreshold = 2130903894;
			
			// aapt resource value: 0x7F030357
			public const int srcCompat = 2130903895;
			
			// aapt resource value: 0x7F030358
			public const int stackFromEnd = 2130903896;
			
			// aapt resource value: 0x7F030359
			public const int staggered = 2130903897;
			
			// aapt resource value: 0x7F03035A
			public const int startDestination = 2130903898;
			
			// aapt resource value: 0x7F03035B
			public const int startIconCheckable = 2130903899;
			
			// aapt resource value: 0x7F03035C
			public const int startIconContentDescription = 2130903900;
			
			// aapt resource value: 0x7F03035D
			public const int startIconDrawable = 2130903901;
			
			// aapt resource value: 0x7F03035E
			public const int startIconTint = 2130903902;
			
			// aapt resource value: 0x7F03035F
			public const int startIconTintMode = 2130903903;
			
			// aapt resource value: 0x7F030360
			public const int state_above_anchor = 2130903904;
			
			// aapt resource value: 0x7F030361
			public const int state_collapsed = 2130903905;
			
			// aapt resource value: 0x7F030362
			public const int state_collapsible = 2130903906;
			
			// aapt resource value: 0x7F030363
			public const int state_dragged = 2130903907;
			
			// aapt resource value: 0x7F030364
			public const int state_liftable = 2130903908;
			
			// aapt resource value: 0x7F030365
			public const int state_lifted = 2130903909;
			
			// aapt resource value: 0x7F030366
			public const int statusBarBackground = 2130903910;
			
			// aapt resource value: 0x7F030367
			public const int statusBarForeground = 2130903911;
			
			// aapt resource value: 0x7F030368
			public const int statusBarScrim = 2130903912;
			
			// aapt resource value: 0x7F030369
			public const int strokeColor = 2130903913;
			
			// aapt resource value: 0x7F03036A
			public const int strokeWidth = 2130903914;
			
			// aapt resource value: 0x7F03036B
			public const int subMenuArrow = 2130903915;
			
			// aapt resource value: 0x7F03036C
			public const int submitBackground = 2130903916;
			
			// aapt resource value: 0x7F03036D
			public const int subtitle = 2130903917;
			
			// aapt resource value: 0x7F03036E
			public const int subtitleCentered = 2130903918;
			
			// aapt resource value: 0x7F03036F
			public const int subtitleTextAppearance = 2130903919;
			
			// aapt resource value: 0x7F030370
			public const int subtitleTextColor = 2130903920;
			
			// aapt resource value: 0x7F030371
			public const int subtitleTextStyle = 2130903921;
			
			// aapt resource value: 0x7F030372
			public const int suffixText = 2130903922;
			
			// aapt resource value: 0x7F030373
			public const int suffixTextAppearance = 2130903923;
			
			// aapt resource value: 0x7F030374
			public const int suffixTextColor = 2130903924;
			
			// aapt resource value: 0x7F030375
			public const int suggestionRowLayout = 2130903925;
			
			// aapt resource value: 0x7F030376
			public const int summary = 2130903926;
			
			// aapt resource value: 0x7F030377
			public const int summaryOff = 2130903927;
			
			// aapt resource value: 0x7F030378
			public const int summaryOn = 2130903928;
			
			// aapt resource value: 0x7F030379
			public const int swipeRefreshLayoutProgressSpinnerBackgroundColor = 2130903929;
			
			// aapt resource value: 0x7F03037A
			public const int switchMinWidth = 2130903930;
			
			// aapt resource value: 0x7F03037B
			public const int switchPadding = 2130903931;
			
			// aapt resource value: 0x7F03037C
			public const int switchPreferenceCompatStyle = 2130903932;
			
			// aapt resource value: 0x7F03037D
			public const int switchPreferenceStyle = 2130903933;
			
			// aapt resource value: 0x7F03037E
			public const int switchStyle = 2130903934;
			
			// aapt resource value: 0x7F03037F
			public const int switchTextAppearance = 2130903935;
			
			// aapt resource value: 0x7F030380
			public const int switchTextOff = 2130903936;
			
			// aapt resource value: 0x7F030381
			public const int switchTextOn = 2130903937;
			
			// aapt resource value: 0x7F030382
			public const int tabBackground = 2130903938;
			
			// aapt resource value: 0x7F030383
			public const int tabContentStart = 2130903939;
			
			// aapt resource value: 0x7F030384
			public const int tabGravity = 2130903940;
			
			// aapt resource value: 0x7F030385
			public const int tabIconTint = 2130903941;
			
			// aapt resource value: 0x7F030386
			public const int tabIconTintMode = 2130903942;
			
			// aapt resource value: 0x7F030387
			public const int tabIndicator = 2130903943;
			
			// aapt resource value: 0x7F030388
			public const int tabIndicatorAnimationDuration = 2130903944;
			
			// aapt resource value: 0x7F030389
			public const int tabIndicatorAnimationMode = 2130903945;
			
			// aapt resource value: 0x7F03038A
			public const int tabIndicatorColor = 2130903946;
			
			// aapt resource value: 0x7F03038B
			public const int tabIndicatorFullWidth = 2130903947;
			
			// aapt resource value: 0x7F03038C
			public const int tabIndicatorGravity = 2130903948;
			
			// aapt resource value: 0x7F03038D
			public const int tabIndicatorHeight = 2130903949;
			
			// aapt resource value: 0x7F03038E
			public const int tabInlineLabel = 2130903950;
			
			// aapt resource value: 0x7F03038F
			public const int tabMaxWidth = 2130903951;
			
			// aapt resource value: 0x7F030390
			public const int tabMinWidth = 2130903952;
			
			// aapt resource value: 0x7F030391
			public const int tabMode = 2130903953;
			
			// aapt resource value: 0x7F030392
			public const int tabPadding = 2130903954;
			
			// aapt resource value: 0x7F030393
			public const int tabPaddingBottom = 2130903955;
			
			// aapt resource value: 0x7F030394
			public const int tabPaddingEnd = 2130903956;
			
			// aapt resource value: 0x7F030395
			public const int tabPaddingStart = 2130903957;
			
			// aapt resource value: 0x7F030396
			public const int tabPaddingTop = 2130903958;
			
			// aapt resource value: 0x7F030397
			public const int tabRippleColor = 2130903959;
			
			// aapt resource value: 0x7F030398
			public const int tabSelectedTextColor = 2130903960;
			
			// aapt resource value: 0x7F030399
			public const int tabStyle = 2130903961;
			
			// aapt resource value: 0x7F03039A
			public const int tabTextAppearance = 2130903962;
			
			// aapt resource value: 0x7F03039B
			public const int tabTextColor = 2130903963;
			
			// aapt resource value: 0x7F03039C
			public const int tabUnboundedRipple = 2130903964;
			
			// aapt resource value: 0x7F03039D
			public const int targetId = 2130903965;
			
			// aapt resource value: 0x7F03039E
			public const int targetPackage = 2130903966;
			
			// aapt resource value: 0x7F03039F
			public const int telltales_tailColor = 2130903967;
			
			// aapt resource value: 0x7F0303A0
			public const int telltales_tailScale = 2130903968;
			
			// aapt resource value: 0x7F0303A1
			public const int telltales_velocityMode = 2130903969;
			
			// aapt resource value: 0x7F0303A2
			public const int textAllCaps = 2130903970;
			
			// aapt resource value: 0x7F0303A3
			public const int textAppearanceBody1 = 2130903971;
			
			// aapt resource value: 0x7F0303A4
			public const int textAppearanceBody2 = 2130903972;
			
			// aapt resource value: 0x7F0303A5
			public const int textAppearanceButton = 2130903973;
			
			// aapt resource value: 0x7F0303A6
			public const int textAppearanceCaption = 2130903974;
			
			// aapt resource value: 0x7F0303A7
			public const int textAppearanceHeadline1 = 2130903975;
			
			// aapt resource value: 0x7F0303A8
			public const int textAppearanceHeadline2 = 2130903976;
			
			// aapt resource value: 0x7F0303A9
			public const int textAppearanceHeadline3 = 2130903977;
			
			// aapt resource value: 0x7F0303AA
			public const int textAppearanceHeadline4 = 2130903978;
			
			// aapt resource value: 0x7F0303AB
			public const int textAppearanceHeadline5 = 2130903979;
			
			// aapt resource value: 0x7F0303AC
			public const int textAppearanceHeadline6 = 2130903980;
			
			// aapt resource value: 0x7F0303AD
			public const int textAppearanceLargePopupMenu = 2130903981;
			
			// aapt resource value: 0x7F0303AE
			public const int textAppearanceLineHeightEnabled = 2130903982;
			
			// aapt resource value: 0x7F0303AF
			public const int textAppearanceListItem = 2130903983;
			
			// aapt resource value: 0x7F0303B0
			public const int textAppearanceListItemSecondary = 2130903984;
			
			// aapt resource value: 0x7F0303B1
			public const int textAppearanceListItemSmall = 2130903985;
			
			// aapt resource value: 0x7F0303B2
			public const int textAppearanceOverline = 2130903986;
			
			// aapt resource value: 0x7F0303B3
			public const int textAppearancePopupMenuHeader = 2130903987;
			
			// aapt resource value: 0x7F0303B4
			public const int textAppearanceSearchResultSubtitle = 2130903988;
			
			// aapt resource value: 0x7F0303B5
			public const int textAppearanceSearchResultTitle = 2130903989;
			
			// aapt resource value: 0x7F0303B6
			public const int textAppearanceSmallPopupMenu = 2130903990;
			
			// aapt resource value: 0x7F0303B7
			public const int textAppearanceSubtitle1 = 2130903991;
			
			// aapt resource value: 0x7F0303B8
			public const int textAppearanceSubtitle2 = 2130903992;
			
			// aapt resource value: 0x7F0303B9
			public const int textBackground = 2130903993;
			
			// aapt resource value: 0x7F0303BA
			public const int textBackgroundPanX = 2130903994;
			
			// aapt resource value: 0x7F0303BB
			public const int textBackgroundPanY = 2130903995;
			
			// aapt resource value: 0x7F0303BC
			public const int textBackgroundRotate = 2130903996;
			
			// aapt resource value: 0x7F0303BD
			public const int textBackgroundZoom = 2130903997;
			
			// aapt resource value: 0x7F0303BE
			public const int textColorAlertDialogListItem = 2130903998;
			
			// aapt resource value: 0x7F0303BF
			public const int textColorSearchUrl = 2130903999;
			
			// aapt resource value: 0x7F0303C0
			public const int textEndPadding = 2130904000;
			
			// aapt resource value: 0x7F0303C1
			public const int textFillColor = 2130904001;
			
			// aapt resource value: 0x7F0303C2
			public const int textInputLayoutFocusedRectEnabled = 2130904002;
			
			// aapt resource value: 0x7F0303C3
			public const int textInputStyle = 2130904003;
			
			// aapt resource value: 0x7F0303C4
			public const int textLocale = 2130904004;
			
			// aapt resource value: 0x7F0303C5
			public const int textOutlineColor = 2130904005;
			
			// aapt resource value: 0x7F0303C6
			public const int textOutlineThickness = 2130904006;
			
			// aapt resource value: 0x7F0303C7
			public const int textPanX = 2130904007;
			
			// aapt resource value: 0x7F0303C8
			public const int textPanY = 2130904008;
			
			// aapt resource value: 0x7F0303C9
			public const int textStartPadding = 2130904009;
			
			// aapt resource value: 0x7F0303CA
			public const int textureBlurFactor = 2130904010;
			
			// aapt resource value: 0x7F0303CB
			public const int textureEffect = 2130904011;
			
			// aapt resource value: 0x7F0303CC
			public const int textureHeight = 2130904012;
			
			// aapt resource value: 0x7F0303CD
			public const int textureWidth = 2130904013;
			
			// aapt resource value: 0x7F0303CE
			public const int theme = 2130904014;
			
			// aapt resource value: 0x7F0303CF
			public const int themeLineHeight = 2130904015;
			
			// aapt resource value: 0x7F0303D0
			public const int thickness = 2130904016;
			
			// aapt resource value: 0x7F0303D1
			public const int thumbColor = 2130904017;
			
			// aapt resource value: 0x7F0303D2
			public const int thumbElevation = 2130904018;
			
			// aapt resource value: 0x7F0303D3
			public const int thumbRadius = 2130904019;
			
			// aapt resource value: 0x7F0303D4
			public const int thumbStrokeColor = 2130904020;
			
			// aapt resource value: 0x7F0303D5
			public const int thumbStrokeWidth = 2130904021;
			
			// aapt resource value: 0x7F0303D6
			public const int thumbTextPadding = 2130904022;
			
			// aapt resource value: 0x7F0303D7
			public const int thumbTint = 2130904023;
			
			// aapt resource value: 0x7F0303D8
			public const int thumbTintMode = 2130904024;
			
			// aapt resource value: 0x7F0303D9
			public const int tickColor = 2130904025;
			
			// aapt resource value: 0x7F0303DA
			public const int tickColorActive = 2130904026;
			
			// aapt resource value: 0x7F0303DB
			public const int tickColorInactive = 2130904027;
			
			// aapt resource value: 0x7F0303DC
			public const int tickMark = 2130904028;
			
			// aapt resource value: 0x7F0303DD
			public const int tickMarkTint = 2130904029;
			
			// aapt resource value: 0x7F0303DE
			public const int tickMarkTintMode = 2130904030;
			
			// aapt resource value: 0x7F0303DF
			public const int tickVisible = 2130904031;
			
			// aapt resource value: 0x7F0303E0
			public const int tint = 2130904032;
			
			// aapt resource value: 0x7F0303E1
			public const int tintMode = 2130904033;
			
			// aapt resource value: 0x7F0303E2
			public const int title = 2130904034;
			
			// aapt resource value: 0x7F0303E3
			public const int titleCentered = 2130904035;
			
			// aapt resource value: 0x7F0303E4
			public const int titleCollapseMode = 2130904036;
			
			// aapt resource value: 0x7F0303E5
			public const int titleEnabled = 2130904037;
			
			// aapt resource value: 0x7F0303E6
			public const int titleMargin = 2130904038;
			
			// aapt resource value: 0x7F0303E7
			public const int titleMarginBottom = 2130904039;
			
			// aapt resource value: 0x7F0303E8
			public const int titleMarginEnd = 2130904040;
			
			// aapt resource value: 0x7F0303EB
			public const int titleMargins = 2130904043;
			
			// aapt resource value: 0x7F0303E9
			public const int titleMarginStart = 2130904041;
			
			// aapt resource value: 0x7F0303EA
			public const int titleMarginTop = 2130904042;
			
			// aapt resource value: 0x7F0303EC
			public const int titleTextAppearance = 2130904044;
			
			// aapt resource value: 0x7F0303ED
			public const int titleTextColor = 2130904045;
			
			// aapt resource value: 0x7F0303EE
			public const int titleTextStyle = 2130904046;
			
			// aapt resource value: 0x7F0303EF
			public const int toolbarId = 2130904047;
			
			// aapt resource value: 0x7F0303F0
			public const int toolbarNavigationButtonStyle = 2130904048;
			
			// aapt resource value: 0x7F0303F1
			public const int toolbarStyle = 2130904049;
			
			// aapt resource value: 0x7F0303F2
			public const int tooltipForegroundColor = 2130904050;
			
			// aapt resource value: 0x7F0303F3
			public const int tooltipFrameBackground = 2130904051;
			
			// aapt resource value: 0x7F0303F4
			public const int tooltipStyle = 2130904052;
			
			// aapt resource value: 0x7F0303F5
			public const int tooltipText = 2130904053;
			
			// aapt resource value: 0x7F0303F6
			public const int touchAnchorId = 2130904054;
			
			// aapt resource value: 0x7F0303F7
			public const int touchAnchorSide = 2130904055;
			
			// aapt resource value: 0x7F0303F8
			public const int touchRegionId = 2130904056;
			
			// aapt resource value: 0x7F0303F9
			public const int track = 2130904057;
			
			// aapt resource value: 0x7F0303FA
			public const int trackColor = 2130904058;
			
			// aapt resource value: 0x7F0303FB
			public const int trackColorActive = 2130904059;
			
			// aapt resource value: 0x7F0303FC
			public const int trackColorInactive = 2130904060;
			
			// aapt resource value: 0x7F0303FD
			public const int trackCornerRadius = 2130904061;
			
			// aapt resource value: 0x7F0303FE
			public const int trackHeight = 2130904062;
			
			// aapt resource value: 0x7F0303FF
			public const int trackThickness = 2130904063;
			
			// aapt resource value: 0x7F030400
			public const int trackTint = 2130904064;
			
			// aapt resource value: 0x7F030401
			public const int trackTintMode = 2130904065;
			
			// aapt resource value: 0x7F030402
			public const int transformPivotTarget = 2130904066;
			
			// aapt resource value: 0x7F030403
			public const int transitionDisable = 2130904067;
			
			// aapt resource value: 0x7F030404
			public const int transitionEasing = 2130904068;
			
			// aapt resource value: 0x7F030405
			public const int transitionFlags = 2130904069;
			
			// aapt resource value: 0x7F030406
			public const int transitionPathRotate = 2130904070;
			
			// aapt resource value: 0x7F030407
			public const int transitionShapeAppearance = 2130904071;
			
			// aapt resource value: 0x7F030408
			public const int triggerId = 2130904072;
			
			// aapt resource value: 0x7F030409
			public const int triggerReceiver = 2130904073;
			
			// aapt resource value: 0x7F03040A
			public const int triggerSlack = 2130904074;
			
			// aapt resource value: 0x7F03040B
			public const int ttcIndex = 2130904075;
			
			// aapt resource value: 0x7F03040D
			public const int updatesContinuously = 2130904077;
			
			// aapt resource value: 0x7F03040C
			public const int upDuration = 2130904076;
			
			// aapt resource value: 0x7F03040E
			public const int uri = 2130904078;
			
			// aapt resource value: 0x7F03040F
			public const int useCompatPadding = 2130904079;
			
			// aapt resource value: 0x7F030410
			public const int useMaterialThemeColors = 2130904080;
			
			// aapt resource value: 0x7F030411
			public const int useSimpleSummaryProvider = 2130904081;
			
			// aapt resource value: 0x7F030412
			public const int values = 2130904082;
			
			// aapt resource value: 0x7F030413
			public const int verticalOffset = 2130904083;
			
			// aapt resource value: 0x7F030414
			public const int viewInflaterClass = 2130904084;
			
			// aapt resource value: 0x7F030415
			public const int viewTransitionMode = 2130904085;
			
			// aapt resource value: 0x7F030416
			public const int viewTransitionOnCross = 2130904086;
			
			// aapt resource value: 0x7F030417
			public const int viewTransitionOnNegativeCross = 2130904087;
			
			// aapt resource value: 0x7F030418
			public const int viewTransitionOnPositiveCross = 2130904088;
			
			// aapt resource value: 0x7F030419
			public const int visibilityMode = 2130904089;
			
			// aapt resource value: 0x7F03041A
			public const int voiceIcon = 2130904090;
			
			// aapt resource value: 0x7F03041B
			public const int warmth = 2130904091;
			
			// aapt resource value: 0x7F03041C
			public const int waveDecay = 2130904092;
			
			// aapt resource value: 0x7F03041D
			public const int waveOffset = 2130904093;
			
			// aapt resource value: 0x7F03041E
			public const int wavePeriod = 2130904094;
			
			// aapt resource value: 0x7F03041F
			public const int wavePhase = 2130904095;
			
			// aapt resource value: 0x7F030420
			public const int waveShape = 2130904096;
			
			// aapt resource value: 0x7F030421
			public const int waveVariesBy = 2130904097;
			
			// aapt resource value: 0x7F030422
			public const int widgetLayout = 2130904098;
			
			// aapt resource value: 0x7F030423
			public const int windowActionBar = 2130904099;
			
			// aapt resource value: 0x7F030424
			public const int windowActionBarOverlay = 2130904100;
			
			// aapt resource value: 0x7F030425
			public const int windowActionModeOverlay = 2130904101;
			
			// aapt resource value: 0x7F030426
			public const int windowFixedHeightMajor = 2130904102;
			
			// aapt resource value: 0x7F030427
			public const int windowFixedHeightMinor = 2130904103;
			
			// aapt resource value: 0x7F030428
			public const int windowFixedWidthMajor = 2130904104;
			
			// aapt resource value: 0x7F030429
			public const int windowFixedWidthMinor = 2130904105;
			
			// aapt resource value: 0x7F03042A
			public const int windowMinWidthMajor = 2130904106;
			
			// aapt resource value: 0x7F03042B
			public const int windowMinWidthMinor = 2130904107;
			
			// aapt resource value: 0x7F03042C
			public const int windowNoTitle = 2130904108;
			
			// aapt resource value: 0x7F03042D
			public const int yearSelectedStyle = 2130904109;
			
			// aapt resource value: 0x7F03042E
			public const int yearStyle = 2130904110;
			
			// aapt resource value: 0x7F03042F
			public const int yearTodayStyle = 2130904111;
			
			static Attribute()
			{
				global::Android.Runtime.ResourceIdManager.UpdateIdValues();
			}
			
			private Attribute()
			{
			}
		}
		
		public partial class Boolean
		{
			
			// aapt resource value: 0x7F040000
			public const int abc_action_bar_embed_tabs = 2130968576;
			
			// aapt resource value: 0x7F040001
			public const int abc_config_actionMenuItemAllCaps = 2130968577;
			
			// aapt resource value: 0x7F040002
			public const int config_materialPreferenceIconSpaceReserved = 2130968578;
			
			// aapt resource value: 0x7F040003
			public const int mtrl_btn_textappearance_all_caps = 2130968579;
			
			static Boolean()
			{
				global::Android.Runtime.ResourceIdManager.UpdateIdValues();
			}
			
			private Boolean()
			{
			}
		}
		
		public partial class Color
		{
			
			// aapt resource value: 0x7F050000
			public const int abc_background_cache_hint_selector_material_dark = 2131034112;
			
			// aapt resource value: 0x7F050001
			public const int abc_background_cache_hint_selector_material_light = 2131034113;
			
			// aapt resource value: 0x7F050002
			public const int abc_btn_colored_borderless_text_material = 2131034114;
			
			// aapt resource value: 0x7F050003
			public const int abc_btn_colored_text_material = 2131034115;
			
			// aapt resource value: 0x7F050004
			public const int abc_color_highlight_material = 2131034116;
			
			// aapt resource value: 0x7F050005
			public const int abc_decor_view_status_guard = 2131034117;
			
			// aapt resource value: 0x7F050006
			public const int abc_decor_view_status_guard_light = 2131034118;
			
			// aapt resource value: 0x7F050007
			public const int abc_hint_foreground_material_dark = 2131034119;
			
			// aapt resource value: 0x7F050008
			public const int abc_hint_foreground_material_light = 2131034120;
			
			// aapt resource value: 0x7F050009
			public const int abc_primary_text_disable_only_material_dark = 2131034121;
			
			// aapt resource value: 0x7F05000A
			public const int abc_primary_text_disable_only_material_light = 2131034122;
			
			// aapt resource value: 0x7F05000B
			public const int abc_primary_text_material_dark = 2131034123;
			
			// aapt resource value: 0x7F05000C
			public const int abc_primary_text_material_light = 2131034124;
			
			// aapt resource value: 0x7F05000D
			public const int abc_search_url_text = 2131034125;
			
			// aapt resource value: 0x7F05000E
			public const int abc_search_url_text_normal = 2131034126;
			
			// aapt resource value: 0x7F05000F
			public const int abc_search_url_text_pressed = 2131034127;
			
			// aapt resource value: 0x7F050010
			public const int abc_search_url_text_selected = 2131034128;
			
			// aapt resource value: 0x7F050011
			public const int abc_secondary_text_material_dark = 2131034129;
			
			// aapt resource value: 0x7F050012
			public const int abc_secondary_text_material_light = 2131034130;
			
			// aapt resource value: 0x7F050013
			public const int abc_tint_btn_checkable = 2131034131;
			
			// aapt resource value: 0x7F050014
			public const int abc_tint_default = 2131034132;
			
			// aapt resource value: 0x7F050015
			public const int abc_tint_edittext = 2131034133;
			
			// aapt resource value: 0x7F050016
			public const int abc_tint_seek_thumb = 2131034134;
			
			// aapt resource value: 0x7F050017
			public const int abc_tint_spinner = 2131034135;
			
			// aapt resource value: 0x7F050018
			public const int abc_tint_switch_track = 2131034136;
			
			// aapt resource value: 0x7F050019
			public const int accent_material_dark = 2131034137;
			
			// aapt resource value: 0x7F05001A
			public const int accent_material_light = 2131034138;
			
			// aapt resource value: 0x7F05001B
			public const int androidx_core_ripple_material_light = 2131034139;
			
			// aapt resource value: 0x7F05001C
			public const int androidx_core_secondary_text_default_material_light = 2131034140;
			
			// aapt resource value: 0x7F05001D
			public const int background_floating_material_dark = 2131034141;
			
			// aapt resource value: 0x7F05001E
			public const int background_floating_material_light = 2131034142;
			
			// aapt resource value: 0x7F05001F
			public const int background_material_dark = 2131034143;
			
			// aapt resource value: 0x7F050020
			public const int background_material_light = 2131034144;
			
			// aapt resource value: 0x7F050021
			public const int bright_foreground_disabled_material_dark = 2131034145;
			
			// aapt resource value: 0x7F050022
			public const int bright_foreground_disabled_material_light = 2131034146;
			
			// aapt resource value: 0x7F050023
			public const int bright_foreground_inverse_material_dark = 2131034147;
			
			// aapt resource value: 0x7F050024
			public const int bright_foreground_inverse_material_light = 2131034148;
			
			// aapt resource value: 0x7F050025
			public const int bright_foreground_material_dark = 2131034149;
			
			// aapt resource value: 0x7F050026
			public const int bright_foreground_material_light = 2131034150;
			
			// aapt resource value: 0x7F050027
			public const int browser_actions_bg_grey = 2131034151;
			
			// aapt resource value: 0x7F050028
			public const int browser_actions_divider_color = 2131034152;
			
			// aapt resource value: 0x7F050029
			public const int browser_actions_text_color = 2131034153;
			
			// aapt resource value: 0x7F05002A
			public const int browser_actions_title_color = 2131034154;
			
			// aapt resource value: 0x7F05002B
			public const int button_material_dark = 2131034155;
			
			// aapt resource value: 0x7F05002C
			public const int button_material_light = 2131034156;
			
			// aapt resource value: 0x7F05002D
			public const int cardview_dark_background = 2131034157;
			
			// aapt resource value: 0x7F05002E
			public const int cardview_light_background = 2131034158;
			
			// aapt resource value: 0x7F05002F
			public const int cardview_shadow_end_color = 2131034159;
			
			// aapt resource value: 0x7F050030
			public const int cardview_shadow_start_color = 2131034160;
			
			// aapt resource value: 0x7F050031
			public const int checkbox_themeable_attribute_color = 2131034161;
			
			// aapt resource value: 0x7F050032
			public const int colorAccent = 2131034162;
			
			// aapt resource value: 0x7F050033
			public const int colorPrimary = 2131034163;
			
			// aapt resource value: 0x7F050034
			public const int colorPrimaryDark = 2131034164;
			
			// aapt resource value: 0x7F050035
			public const int design_bottom_navigation_shadow_color = 2131034165;
			
			// aapt resource value: 0x7F050036
			public const int design_box_stroke_color = 2131034166;
			
			// aapt resource value: 0x7F050037
			public const int design_dark_default_color_background = 2131034167;
			
			// aapt resource value: 0x7F050038
			public const int design_dark_default_color_error = 2131034168;
			
			// aapt resource value: 0x7F050039
			public const int design_dark_default_color_on_background = 2131034169;
			
			// aapt resource value: 0x7F05003A
			public const int design_dark_default_color_on_error = 2131034170;
			
			// aapt resource value: 0x7F05003B
			public const int design_dark_default_color_on_primary = 2131034171;
			
			// aapt resource value: 0x7F05003C
			public const int design_dark_default_color_on_secondary = 2131034172;
			
			// aapt resource value: 0x7F05003D
			public const int design_dark_default_color_on_surface = 2131034173;
			
			// aapt resource value: 0x7F05003E
			public const int design_dark_default_color_primary = 2131034174;
			
			// aapt resource value: 0x7F05003F
			public const int design_dark_default_color_primary_dark = 2131034175;
			
			// aapt resource value: 0x7F050040
			public const int design_dark_default_color_primary_variant = 2131034176;
			
			// aapt resource value: 0x7F050041
			public const int design_dark_default_color_secondary = 2131034177;
			
			// aapt resource value: 0x7F050042
			public const int design_dark_default_color_secondary_variant = 2131034178;
			
			// aapt resource value: 0x7F050043
			public const int design_dark_default_color_surface = 2131034179;
			
			// aapt resource value: 0x7F050044
			public const int design_default_color_background = 2131034180;
			
			// aapt resource value: 0x7F050045
			public const int design_default_color_error = 2131034181;
			
			// aapt resource value: 0x7F050046
			public const int design_default_color_on_background = 2131034182;
			
			// aapt resource value: 0x7F050047
			public const int design_default_color_on_error = 2131034183;
			
			// aapt resource value: 0x7F050048
			public const int design_default_color_on_primary = 2131034184;
			
			// aapt resource value: 0x7F050049
			public const int design_default_color_on_secondary = 2131034185;
			
			// aapt resource value: 0x7F05004A
			public const int design_default_color_on_surface = 2131034186;
			
			// aapt resource value: 0x7F05004B
			public const int design_default_color_primary = 2131034187;
			
			// aapt resource value: 0x7F05004C
			public const int design_default_color_primary_dark = 2131034188;
			
			// aapt resource value: 0x7F05004D
			public const int design_default_color_primary_variant = 2131034189;
			
			// aapt resource value: 0x7F05004E
			public const int design_default_color_secondary = 2131034190;
			
			// aapt resource value: 0x7F05004F
			public const int design_default_color_secondary_variant = 2131034191;
			
			// aapt resource value: 0x7F050050
			public const int design_default_color_surface = 2131034192;
			
			// aapt resource value: 0x7F050051
			public const int design_error = 2131034193;
			
			// aapt resource value: 0x7F050052
			public const int design_fab_shadow_end_color = 2131034194;
			
			// aapt resource value: 0x7F050053
			public const int design_fab_shadow_mid_color = 2131034195;
			
			// aapt resource value: 0x7F050054
			public const int design_fab_shadow_start_color = 2131034196;
			
			// aapt resource value: 0x7F050055
			public const int design_fab_stroke_end_inner_color = 2131034197;
			
			// aapt resource value: 0x7F050056
			public const int design_fab_stroke_end_outer_color = 2131034198;
			
			// aapt resource value: 0x7F050057
			public const int design_fab_stroke_top_inner_color = 2131034199;
			
			// aapt resource value: 0x7F050058
			public const int design_fab_stroke_top_outer_color = 2131034200;
			
			// aapt resource value: 0x7F050059
			public const int design_icon_tint = 2131034201;
			
			// aapt resource value: 0x7F05005A
			public const int design_snackbar_background_color = 2131034202;
			
			// aapt resource value: 0x7F05005B
			public const int dim_foreground_disabled_material_dark = 2131034203;
			
			// aapt resource value: 0x7F05005C
			public const int dim_foreground_disabled_material_light = 2131034204;
			
			// aapt resource value: 0x7F05005D
			public const int dim_foreground_material_dark = 2131034205;
			
			// aapt resource value: 0x7F05005E
			public const int dim_foreground_material_light = 2131034206;
			
			// aapt resource value: 0x7F05005F
			public const int error_color_material_dark = 2131034207;
			
			// aapt resource value: 0x7F050060
			public const int error_color_material_light = 2131034208;
			
			// aapt resource value: 0x7F050061
			public const int foreground_material_dark = 2131034209;
			
			// aapt resource value: 0x7F050062
			public const int foreground_material_light = 2131034210;
			
			// aapt resource value: 0x7F050063
			public const int highlighted_text_material_dark = 2131034211;
			
			// aapt resource value: 0x7F050064
			public const int highlighted_text_material_light = 2131034212;
			
			// aapt resource value: 0x7F050065
			public const int launcher_background = 2131034213;
			
			// aapt resource value: 0x7F050066
			public const int material_blue_grey_800 = 2131034214;
			
			// aapt resource value: 0x7F050067
			public const int material_blue_grey_900 = 2131034215;
			
			// aapt resource value: 0x7F050068
			public const int material_blue_grey_950 = 2131034216;
			
			// aapt resource value: 0x7F050069
			public const int material_cursor_color = 2131034217;
			
			// aapt resource value: 0x7F05006A
			public const int material_deep_teal_200 = 2131034218;
			
			// aapt resource value: 0x7F05006B
			public const int material_deep_teal_500 = 2131034219;
			
			// aapt resource value: 0x7F05006C
			public const int material_grey_100 = 2131034220;
			
			// aapt resource value: 0x7F05006D
			public const int material_grey_300 = 2131034221;
			
			// aapt resource value: 0x7F05006E
			public const int material_grey_50 = 2131034222;
			
			// aapt resource value: 0x7F05006F
			public const int material_grey_600 = 2131034223;
			
			// aapt resource value: 0x7F050070
			public const int material_grey_800 = 2131034224;
			
			// aapt resource value: 0x7F050071
			public const int material_grey_850 = 2131034225;
			
			// aapt resource value: 0x7F050072
			public const int material_grey_900 = 2131034226;
			
			// aapt resource value: 0x7F050073
			public const int material_on_background_disabled = 2131034227;
			
			// aapt resource value: 0x7F050074
			public const int material_on_background_emphasis_high_type = 2131034228;
			
			// aapt resource value: 0x7F050075
			public const int material_on_background_emphasis_medium = 2131034229;
			
			// aapt resource value: 0x7F050076
			public const int material_on_primary_disabled = 2131034230;
			
			// aapt resource value: 0x7F050077
			public const int material_on_primary_emphasis_high_type = 2131034231;
			
			// aapt resource value: 0x7F050078
			public const int material_on_primary_emphasis_medium = 2131034232;
			
			// aapt resource value: 0x7F050079
			public const int material_on_surface_disabled = 2131034233;
			
			// aapt resource value: 0x7F05007A
			public const int material_on_surface_emphasis_high_type = 2131034234;
			
			// aapt resource value: 0x7F05007B
			public const int material_on_surface_emphasis_medium = 2131034235;
			
			// aapt resource value: 0x7F05007C
			public const int material_on_surface_stroke = 2131034236;
			
			// aapt resource value: 0x7F05007D
			public const int material_slider_active_tick_marks_color = 2131034237;
			
			// aapt resource value: 0x7F05007E
			public const int material_slider_active_track_color = 2131034238;
			
			// aapt resource value: 0x7F05007F
			public const int material_slider_halo_color = 2131034239;
			
			// aapt resource value: 0x7F050080
			public const int material_slider_inactive_tick_marks_color = 2131034240;
			
			// aapt resource value: 0x7F050081
			public const int material_slider_inactive_track_color = 2131034241;
			
			// aapt resource value: 0x7F050082
			public const int material_slider_thumb_color = 2131034242;
			
			// aapt resource value: 0x7F050083
			public const int material_timepicker_button_background = 2131034243;
			
			// aapt resource value: 0x7F050084
			public const int material_timepicker_button_stroke = 2131034244;
			
			// aapt resource value: 0x7F050086
			public const int material_timepicker_clockface = 2131034246;
			
			// aapt resource value: 0x7F050085
			public const int material_timepicker_clock_text_color = 2131034245;
			
			// aapt resource value: 0x7F050087
			public const int material_timepicker_modebutton_tint = 2131034247;
			
			// aapt resource value: 0x7F050088
			public const int mtrl_btn_bg_color_selector = 2131034248;
			
			// aapt resource value: 0x7F050089
			public const int mtrl_btn_ripple_color = 2131034249;
			
			// aapt resource value: 0x7F05008A
			public const int mtrl_btn_stroke_color_selector = 2131034250;
			
			// aapt resource value: 0x7F05008B
			public const int mtrl_btn_text_btn_bg_color_selector = 2131034251;
			
			// aapt resource value: 0x7F05008C
			public const int mtrl_btn_text_btn_ripple_color = 2131034252;
			
			// aapt resource value: 0x7F05008D
			public const int mtrl_btn_text_color_disabled = 2131034253;
			
			// aapt resource value: 0x7F05008E
			public const int mtrl_btn_text_color_selector = 2131034254;
			
			// aapt resource value: 0x7F05008F
			public const int mtrl_btn_transparent_bg_color = 2131034255;
			
			// aapt resource value: 0x7F050090
			public const int mtrl_calendar_item_stroke_color = 2131034256;
			
			// aapt resource value: 0x7F050091
			public const int mtrl_calendar_selected_range = 2131034257;
			
			// aapt resource value: 0x7F050092
			public const int mtrl_card_view_foreground = 2131034258;
			
			// aapt resource value: 0x7F050093
			public const int mtrl_card_view_ripple = 2131034259;
			
			// aapt resource value: 0x7F050094
			public const int mtrl_chip_background_color = 2131034260;
			
			// aapt resource value: 0x7F050095
			public const int mtrl_chip_close_icon_tint = 2131034261;
			
			// aapt resource value: 0x7F050096
			public const int mtrl_chip_surface_color = 2131034262;
			
			// aapt resource value: 0x7F050097
			public const int mtrl_chip_text_color = 2131034263;
			
			// aapt resource value: 0x7F050098
			public const int mtrl_choice_chip_background_color = 2131034264;
			
			// aapt resource value: 0x7F050099
			public const int mtrl_choice_chip_ripple_color = 2131034265;
			
			// aapt resource value: 0x7F05009A
			public const int mtrl_choice_chip_text_color = 2131034266;
			
			// aapt resource value: 0x7F05009B
			public const int mtrl_error = 2131034267;
			
			// aapt resource value: 0x7F05009C
			public const int mtrl_fab_bg_color_selector = 2131034268;
			
			// aapt resource value: 0x7F05009D
			public const int mtrl_fab_icon_text_color_selector = 2131034269;
			
			// aapt resource value: 0x7F05009E
			public const int mtrl_fab_ripple_color = 2131034270;
			
			// aapt resource value: 0x7F05009F
			public const int mtrl_filled_background_color = 2131034271;
			
			// aapt resource value: 0x7F0500A0
			public const int mtrl_filled_icon_tint = 2131034272;
			
			// aapt resource value: 0x7F0500A1
			public const int mtrl_filled_stroke_color = 2131034273;
			
			// aapt resource value: 0x7F0500A2
			public const int mtrl_indicator_text_color = 2131034274;
			
			// aapt resource value: 0x7F0500A3
			public const int mtrl_navigation_bar_colored_item_tint = 2131034275;
			
			// aapt resource value: 0x7F0500A4
			public const int mtrl_navigation_bar_colored_ripple_color = 2131034276;
			
			// aapt resource value: 0x7F0500A5
			public const int mtrl_navigation_bar_item_tint = 2131034277;
			
			// aapt resource value: 0x7F0500A6
			public const int mtrl_navigation_bar_ripple_color = 2131034278;
			
			// aapt resource value: 0x7F0500A7
			public const int mtrl_navigation_item_background_color = 2131034279;
			
			// aapt resource value: 0x7F0500A8
			public const int mtrl_navigation_item_icon_tint = 2131034280;
			
			// aapt resource value: 0x7F0500A9
			public const int mtrl_navigation_item_text_color = 2131034281;
			
			// aapt resource value: 0x7F0500AA
			public const int mtrl_on_primary_text_btn_text_color_selector = 2131034282;
			
			// aapt resource value: 0x7F0500AB
			public const int mtrl_on_surface_ripple_color = 2131034283;
			
			// aapt resource value: 0x7F0500AC
			public const int mtrl_outlined_icon_tint = 2131034284;
			
			// aapt resource value: 0x7F0500AD
			public const int mtrl_outlined_stroke_color = 2131034285;
			
			// aapt resource value: 0x7F0500AE
			public const int mtrl_popupmenu_overlay_color = 2131034286;
			
			// aapt resource value: 0x7F0500AF
			public const int mtrl_scrim_color = 2131034287;
			
			// aapt resource value: 0x7F0500B0
			public const int mtrl_tabs_colored_ripple_color = 2131034288;
			
			// aapt resource value: 0x7F0500B1
			public const int mtrl_tabs_icon_color_selector = 2131034289;
			
			// aapt resource value: 0x7F0500B2
			public const int mtrl_tabs_icon_color_selector_colored = 2131034290;
			
			// aapt resource value: 0x7F0500B3
			public const int mtrl_tabs_legacy_text_color_selector = 2131034291;
			
			// aapt resource value: 0x7F0500B4
			public const int mtrl_tabs_ripple_color = 2131034292;
			
			// aapt resource value: 0x7F0500B6
			public const int mtrl_textinput_default_box_stroke_color = 2131034294;
			
			// aapt resource value: 0x7F0500B7
			public const int mtrl_textinput_disabled_color = 2131034295;
			
			// aapt resource value: 0x7F0500B8
			public const int mtrl_textinput_filled_box_default_background_color = 2131034296;
			
			// aapt resource value: 0x7F0500B9
			public const int mtrl_textinput_focused_box_stroke_color = 2131034297;
			
			// aapt resource value: 0x7F0500BA
			public const int mtrl_textinput_hovered_box_stroke_color = 2131034298;
			
			// aapt resource value: 0x7F0500B5
			public const int mtrl_text_btn_text_color_selector = 2131034293;
			
			// aapt resource value: 0x7F0500BB
			public const int notification_action_color_filter = 2131034299;
			
			// aapt resource value: 0x7F0500BC
			public const int notification_icon_bg_color = 2131034300;
			
			// aapt resource value: 0x7F0500BD
			public const int notification_material_background_media_default_color = 2131034301;
			
			// aapt resource value: 0x7F0500BE
			public const int preference_fallback_accent_color = 2131034302;
			
			// aapt resource value: 0x7F0500BF
			public const int primary_dark_material_dark = 2131034303;
			
			// aapt resource value: 0x7F0500C0
			public const int primary_dark_material_light = 2131034304;
			
			// aapt resource value: 0x7F0500C1
			public const int primary_material_dark = 2131034305;
			
			// aapt resource value: 0x7F0500C2
			public const int primary_material_light = 2131034306;
			
			// aapt resource value: 0x7F0500C3
			public const int primary_text_default_material_dark = 2131034307;
			
			// aapt resource value: 0x7F0500C4
			public const int primary_text_default_material_light = 2131034308;
			
			// aapt resource value: 0x7F0500C5
			public const int primary_text_disabled_material_dark = 2131034309;
			
			// aapt resource value: 0x7F0500C6
			public const int primary_text_disabled_material_light = 2131034310;
			
			// aapt resource value: 0x7F0500C7
			public const int radiobutton_themeable_attribute_color = 2131034311;
			
			// aapt resource value: 0x7F0500C8
			public const int ripple_material_dark = 2131034312;
			
			// aapt resource value: 0x7F0500C9
			public const int ripple_material_light = 2131034313;
			
			// aapt resource value: 0x7F0500CA
			public const int secondary_text_default_material_dark = 2131034314;
			
			// aapt resource value: 0x7F0500CB
			public const int secondary_text_default_material_light = 2131034315;
			
			// aapt resource value: 0x7F0500CC
			public const int secondary_text_disabled_material_dark = 2131034316;
			
			// aapt resource value: 0x7F0500CD
			public const int secondary_text_disabled_material_light = 2131034317;
			
			// aapt resource value: 0x7F0500CE
			public const int switch_thumb_disabled_material_dark = 2131034318;
			
			// aapt resource value: 0x7F0500CF
			public const int switch_thumb_disabled_material_light = 2131034319;
			
			// aapt resource value: 0x7F0500D0
			public const int switch_thumb_material_dark = 2131034320;
			
			// aapt resource value: 0x7F0500D1
			public const int switch_thumb_material_light = 2131034321;
			
			// aapt resource value: 0x7F0500D2
			public const int switch_thumb_normal_material_dark = 2131034322;
			
			// aapt resource value: 0x7F0500D3
			public const int switch_thumb_normal_material_light = 2131034323;
			
			// aapt resource value: 0x7F0500D4
			public const int test_mtrl_calendar_day = 2131034324;
			
			// aapt resource value: 0x7F0500D5
			public const int test_mtrl_calendar_day_selected = 2131034325;
			
			// aapt resource value: 0x7F0500D6
			public const int tooltip_background_dark = 2131034326;
			
			// aapt resource value: 0x7F0500D7
			public const int tooltip_background_light = 2131034327;
			
			static Color()
			{
				global::Android.Runtime.ResourceIdManager.UpdateIdValues();
			}
			
			private Color()
			{
			}
		}
		
		public partial class Dimension
		{
			
			// aapt resource value: 0x7F060000
			public const int abc_action_bar_content_inset_material = 2131099648;
			
			// aapt resource value: 0x7F060001
			public const int abc_action_bar_content_inset_with_nav = 2131099649;
			
			// aapt resource value: 0x7F060002
			public const int abc_action_bar_default_height_material = 2131099650;
			
			// aapt resource value: 0x7F060003
			public const int abc_action_bar_default_padding_end_material = 2131099651;
			
			// aapt resource value: 0x7F060004
			public const int abc_action_bar_default_padding_start_material = 2131099652;
			
			// aapt resource value: 0x7F060005
			public const int abc_action_bar_elevation_material = 2131099653;
			
			// aapt resource value: 0x7F060006
			public const int abc_action_bar_icon_vertical_padding_material = 2131099654;
			
			// aapt resource value: 0x7F060007
			public const int abc_action_bar_overflow_padding_end_material = 2131099655;
			
			// aapt resource value: 0x7F060008
			public const int abc_action_bar_overflow_padding_start_material = 2131099656;
			
			// aapt resource value: 0x7F060009
			public const int abc_action_bar_stacked_max_height = 2131099657;
			
			// aapt resource value: 0x7F06000A
			public const int abc_action_bar_stacked_tab_max_width = 2131099658;
			
			// aapt resource value: 0x7F06000B
			public const int abc_action_bar_subtitle_bottom_margin_material = 2131099659;
			
			// aapt resource value: 0x7F06000C
			public const int abc_action_bar_subtitle_top_margin_material = 2131099660;
			
			// aapt resource value: 0x7F06000D
			public const int abc_action_button_min_height_material = 2131099661;
			
			// aapt resource value: 0x7F06000E
			public const int abc_action_button_min_width_material = 2131099662;
			
			// aapt resource value: 0x7F06000F
			public const int abc_action_button_min_width_overflow_material = 2131099663;
			
			// aapt resource value: 0x7F060010
			public const int abc_alert_dialog_button_bar_height = 2131099664;
			
			// aapt resource value: 0x7F060011
			public const int abc_alert_dialog_button_dimen = 2131099665;
			
			// aapt resource value: 0x7F060012
			public const int abc_button_inset_horizontal_material = 2131099666;
			
			// aapt resource value: 0x7F060013
			public const int abc_button_inset_vertical_material = 2131099667;
			
			// aapt resource value: 0x7F060014
			public const int abc_button_padding_horizontal_material = 2131099668;
			
			// aapt resource value: 0x7F060015
			public const int abc_button_padding_vertical_material = 2131099669;
			
			// aapt resource value: 0x7F060016
			public const int abc_cascading_menus_min_smallest_width = 2131099670;
			
			// aapt resource value: 0x7F060017
			public const int abc_config_prefDialogWidth = 2131099671;
			
			// aapt resource value: 0x7F060018
			public const int abc_control_corner_material = 2131099672;
			
			// aapt resource value: 0x7F060019
			public const int abc_control_inset_material = 2131099673;
			
			// aapt resource value: 0x7F06001A
			public const int abc_control_padding_material = 2131099674;
			
			// aapt resource value: 0x7F06001B
			public const int abc_dialog_corner_radius_material = 2131099675;
			
			// aapt resource value: 0x7F06001C
			public const int abc_dialog_fixed_height_major = 2131099676;
			
			// aapt resource value: 0x7F06001D
			public const int abc_dialog_fixed_height_minor = 2131099677;
			
			// aapt resource value: 0x7F06001E
			public const int abc_dialog_fixed_width_major = 2131099678;
			
			// aapt resource value: 0x7F06001F
			public const int abc_dialog_fixed_width_minor = 2131099679;
			
			// aapt resource value: 0x7F060020
			public const int abc_dialog_list_padding_bottom_no_buttons = 2131099680;
			
			// aapt resource value: 0x7F060021
			public const int abc_dialog_list_padding_top_no_title = 2131099681;
			
			// aapt resource value: 0x7F060022
			public const int abc_dialog_min_width_major = 2131099682;
			
			// aapt resource value: 0x7F060023
			public const int abc_dialog_min_width_minor = 2131099683;
			
			// aapt resource value: 0x7F060024
			public const int abc_dialog_padding_material = 2131099684;
			
			// aapt resource value: 0x7F060025
			public const int abc_dialog_padding_top_material = 2131099685;
			
			// aapt resource value: 0x7F060026
			public const int abc_dialog_title_divider_material = 2131099686;
			
			// aapt resource value: 0x7F060027
			public const int abc_disabled_alpha_material_dark = 2131099687;
			
			// aapt resource value: 0x7F060028
			public const int abc_disabled_alpha_material_light = 2131099688;
			
			// aapt resource value: 0x7F060029
			public const int abc_dropdownitem_icon_width = 2131099689;
			
			// aapt resource value: 0x7F06002A
			public const int abc_dropdownitem_text_padding_left = 2131099690;
			
			// aapt resource value: 0x7F06002B
			public const int abc_dropdownitem_text_padding_right = 2131099691;
			
			// aapt resource value: 0x7F06002C
			public const int abc_edit_text_inset_bottom_material = 2131099692;
			
			// aapt resource value: 0x7F06002D
			public const int abc_edit_text_inset_horizontal_material = 2131099693;
			
			// aapt resource value: 0x7F06002E
			public const int abc_edit_text_inset_top_material = 2131099694;
			
			// aapt resource value: 0x7F06002F
			public const int abc_floating_window_z = 2131099695;
			
			// aapt resource value: 0x7F060030
			public const int abc_list_item_height_large_material = 2131099696;
			
			// aapt resource value: 0x7F060031
			public const int abc_list_item_height_material = 2131099697;
			
			// aapt resource value: 0x7F060032
			public const int abc_list_item_height_small_material = 2131099698;
			
			// aapt resource value: 0x7F060033
			public const int abc_list_item_padding_horizontal_material = 2131099699;
			
			// aapt resource value: 0x7F060034
			public const int abc_panel_menu_list_width = 2131099700;
			
			// aapt resource value: 0x7F060035
			public const int abc_progress_bar_height_material = 2131099701;
			
			// aapt resource value: 0x7F060036
			public const int abc_search_view_preferred_height = 2131099702;
			
			// aapt resource value: 0x7F060037
			public const int abc_search_view_preferred_width = 2131099703;
			
			// aapt resource value: 0x7F060038
			public const int abc_seekbar_track_background_height_material = 2131099704;
			
			// aapt resource value: 0x7F060039
			public const int abc_seekbar_track_progress_height_material = 2131099705;
			
			// aapt resource value: 0x7F06003A
			public const int abc_select_dialog_padding_start_material = 2131099706;
			
			// aapt resource value: 0x7F06003B
			public const int abc_star_big = 2131099707;
			
			// aapt resource value: 0x7F06003C
			public const int abc_star_medium = 2131099708;
			
			// aapt resource value: 0x7F06003D
			public const int abc_star_small = 2131099709;
			
			// aapt resource value: 0x7F06003E
			public const int abc_switch_padding = 2131099710;
			
			// aapt resource value: 0x7F06003F
			public const int abc_text_size_body_1_material = 2131099711;
			
			// aapt resource value: 0x7F060040
			public const int abc_text_size_body_2_material = 2131099712;
			
			// aapt resource value: 0x7F060041
			public const int abc_text_size_button_material = 2131099713;
			
			// aapt resource value: 0x7F060042
			public const int abc_text_size_caption_material = 2131099714;
			
			// aapt resource value: 0x7F060043
			public const int abc_text_size_display_1_material = 2131099715;
			
			// aapt resource value: 0x7F060044
			public const int abc_text_size_display_2_material = 2131099716;
			
			// aapt resource value: 0x7F060045
			public const int abc_text_size_display_3_material = 2131099717;
			
			// aapt resource value: 0x7F060046
			public const int abc_text_size_display_4_material = 2131099718;
			
			// aapt resource value: 0x7F060047
			public const int abc_text_size_headline_material = 2131099719;
			
			// aapt resource value: 0x7F060048
			public const int abc_text_size_large_material = 2131099720;
			
			// aapt resource value: 0x7F060049
			public const int abc_text_size_medium_material = 2131099721;
			
			// aapt resource value: 0x7F06004A
			public const int abc_text_size_menu_header_material = 2131099722;
			
			// aapt resource value: 0x7F06004B
			public const int abc_text_size_menu_material = 2131099723;
			
			// aapt resource value: 0x7F06004C
			public const int abc_text_size_small_material = 2131099724;
			
			// aapt resource value: 0x7F06004D
			public const int abc_text_size_subhead_material = 2131099725;
			
			// aapt resource value: 0x7F06004E
			public const int abc_text_size_subtitle_material_toolbar = 2131099726;
			
			// aapt resource value: 0x7F06004F
			public const int abc_text_size_title_material = 2131099727;
			
			// aapt resource value: 0x7F060050
			public const int abc_text_size_title_material_toolbar = 2131099728;
			
			// aapt resource value: 0x7F060051
			public const int action_bar_size = 2131099729;
			
			// aapt resource value: 0x7F060052
			public const int appcompat_dialog_background_inset = 2131099730;
			
			// aapt resource value: 0x7F060053
			public const int browser_actions_context_menu_max_width = 2131099731;
			
			// aapt resource value: 0x7F060054
			public const int browser_actions_context_menu_min_padding = 2131099732;
			
			// aapt resource value: 0x7F060055
			public const int cardview_compat_inset_shadow = 2131099733;
			
			// aapt resource value: 0x7F060056
			public const int cardview_default_elevation = 2131099734;
			
			// aapt resource value: 0x7F060057
			public const int cardview_default_radius = 2131099735;
			
			// aapt resource value: 0x7F060058
			public const int clock_face_margin_start = 2131099736;
			
			// aapt resource value: 0x7F060059
			public const int compat_button_inset_horizontal_material = 2131099737;
			
			// aapt resource value: 0x7F06005A
			public const int compat_button_inset_vertical_material = 2131099738;
			
			// aapt resource value: 0x7F06005B
			public const int compat_button_padding_horizontal_material = 2131099739;
			
			// aapt resource value: 0x7F06005C
			public const int compat_button_padding_vertical_material = 2131099740;
			
			// aapt resource value: 0x7F06005D
			public const int compat_control_corner_material = 2131099741;
			
			// aapt resource value: 0x7F06005E
			public const int compat_notification_large_icon_max_height = 2131099742;
			
			// aapt resource value: 0x7F06005F
			public const int compat_notification_large_icon_max_width = 2131099743;
			
			// aapt resource value: 0x7F060061
			public const int default_dimension = 2131099745;
			
			// aapt resource value: 0x7F060060
			public const int def_drawer_elevation = 2131099744;
			
			// aapt resource value: 0x7F060062
			public const int design_appbar_elevation = 2131099746;
			
			// aapt resource value: 0x7F060063
			public const int design_bottom_navigation_active_item_max_width = 2131099747;
			
			// aapt resource value: 0x7F060064
			public const int design_bottom_navigation_active_item_min_width = 2131099748;
			
			// aapt resource value: 0x7F060065
			public const int design_bottom_navigation_active_text_size = 2131099749;
			
			// aapt resource value: 0x7F060066
			public const int design_bottom_navigation_elevation = 2131099750;
			
			// aapt resource value: 0x7F060067
			public const int design_bottom_navigation_height = 2131099751;
			
			// aapt resource value: 0x7F060068
			public const int design_bottom_navigation_icon_size = 2131099752;
			
			// aapt resource value: 0x7F060069
			public const int design_bottom_navigation_item_max_width = 2131099753;
			
			// aapt resource value: 0x7F06006A
			public const int design_bottom_navigation_item_min_width = 2131099754;
			
			// aapt resource value: 0x7F06006B
			public const int design_bottom_navigation_label_padding = 2131099755;
			
			// aapt resource value: 0x7F06006C
			public const int design_bottom_navigation_margin = 2131099756;
			
			// aapt resource value: 0x7F06006D
			public const int design_bottom_navigation_shadow_height = 2131099757;
			
			// aapt resource value: 0x7F06006E
			public const int design_bottom_navigation_text_size = 2131099758;
			
			// aapt resource value: 0x7F06006F
			public const int design_bottom_sheet_elevation = 2131099759;
			
			// aapt resource value: 0x7F060070
			public const int design_bottom_sheet_modal_elevation = 2131099760;
			
			// aapt resource value: 0x7F060071
			public const int design_bottom_sheet_peek_height_min = 2131099761;
			
			// aapt resource value: 0x7F060072
			public const int design_fab_border_width = 2131099762;
			
			// aapt resource value: 0x7F060073
			public const int design_fab_elevation = 2131099763;
			
			// aapt resource value: 0x7F060074
			public const int design_fab_image_size = 2131099764;
			
			// aapt resource value: 0x7F060075
			public const int design_fab_size_mini = 2131099765;
			
			// aapt resource value: 0x7F060076
			public const int design_fab_size_normal = 2131099766;
			
			// aapt resource value: 0x7F060077
			public const int design_fab_translation_z_hovered_focused = 2131099767;
			
			// aapt resource value: 0x7F060078
			public const int design_fab_translation_z_pressed = 2131099768;
			
			// aapt resource value: 0x7F060079
			public const int design_navigation_elevation = 2131099769;
			
			// aapt resource value: 0x7F06007A
			public const int design_navigation_icon_padding = 2131099770;
			
			// aapt resource value: 0x7F06007B
			public const int design_navigation_icon_size = 2131099771;
			
			// aapt resource value: 0x7F06007C
			public const int design_navigation_item_horizontal_padding = 2131099772;
			
			// aapt resource value: 0x7F06007D
			public const int design_navigation_item_icon_padding = 2131099773;
			
			// aapt resource value: 0x7F06007E
			public const int design_navigation_max_width = 2131099774;
			
			// aapt resource value: 0x7F06007F
			public const int design_navigation_padding_bottom = 2131099775;
			
			// aapt resource value: 0x7F060080
			public const int design_navigation_separator_vertical_padding = 2131099776;
			
			// aapt resource value: 0x7F060081
			public const int design_snackbar_action_inline_max_width = 2131099777;
			
			// aapt resource value: 0x7F060082
			public const int design_snackbar_action_text_color_alpha = 2131099778;
			
			// aapt resource value: 0x7F060083
			public const int design_snackbar_background_corner_radius = 2131099779;
			
			// aapt resource value: 0x7F060084
			public const int design_snackbar_elevation = 2131099780;
			
			// aapt resource value: 0x7F060085
			public const int design_snackbar_extra_spacing_horizontal = 2131099781;
			
			// aapt resource value: 0x7F060086
			public const int design_snackbar_max_width = 2131099782;
			
			// aapt resource value: 0x7F060087
			public const int design_snackbar_min_width = 2131099783;
			
			// aapt resource value: 0x7F060088
			public const int design_snackbar_padding_horizontal = 2131099784;
			
			// aapt resource value: 0x7F060089
			public const int design_snackbar_padding_vertical = 2131099785;
			
			// aapt resource value: 0x7F06008A
			public const int design_snackbar_padding_vertical_2lines = 2131099786;
			
			// aapt resource value: 0x7F06008B
			public const int design_snackbar_text_size = 2131099787;
			
			// aapt resource value: 0x7F06008C
			public const int design_tab_max_width = 2131099788;
			
			// aapt resource value: 0x7F06008D
			public const int design_tab_scrollable_min_width = 2131099789;
			
			// aapt resource value: 0x7F06008E
			public const int design_tab_text_size = 2131099790;
			
			// aapt resource value: 0x7F06008F
			public const int design_tab_text_size_2line = 2131099791;
			
			// aapt resource value: 0x7F060090
			public const int design_textinput_caption_translate_y = 2131099792;
			
			// aapt resource value: 0x7F060091
			public const int disabled_alpha_material_dark = 2131099793;
			
			// aapt resource value: 0x7F060092
			public const int disabled_alpha_material_light = 2131099794;
			
			// aapt resource value: 0x7F060093
			public const int fastscroll_default_thickness = 2131099795;
			
			// aapt resource value: 0x7F060094
			public const int fastscroll_margin = 2131099796;
			
			// aapt resource value: 0x7F060095
			public const int fastscroll_minimum_range = 2131099797;
			
			// aapt resource value: 0x7F060096
			public const int highlight_alpha_material_colored = 2131099798;
			
			// aapt resource value: 0x7F060097
			public const int highlight_alpha_material_dark = 2131099799;
			
			// aapt resource value: 0x7F060098
			public const int highlight_alpha_material_light = 2131099800;
			
			// aapt resource value: 0x7F060099
			public const int hint_alpha_material_dark = 2131099801;
			
			// aapt resource value: 0x7F06009A
			public const int hint_alpha_material_light = 2131099802;
			
			// aapt resource value: 0x7F06009B
			public const int hint_pressed_alpha_material_dark = 2131099803;
			
			// aapt resource value: 0x7F06009C
			public const int hint_pressed_alpha_material_light = 2131099804;
			
			// aapt resource value: 0x7F06009D
			public const int item_touch_helper_max_drag_scroll_per_frame = 2131099805;
			
			// aapt resource value: 0x7F06009E
			public const int item_touch_helper_swipe_escape_max_velocity = 2131099806;
			
			// aapt resource value: 0x7F06009F
			public const int item_touch_helper_swipe_escape_velocity = 2131099807;
			
			// aapt resource value: 0x7F0600A0
			public const int material_bottom_sheet_max_width = 2131099808;
			
			// aapt resource value: 0x7F0600A1
			public const int material_clock_display_padding = 2131099809;
			
			// aapt resource value: 0x7F0600A2
			public const int material_clock_face_margin_top = 2131099810;
			
			// aapt resource value: 0x7F0600A3
			public const int material_clock_hand_center_dot_radius = 2131099811;
			
			// aapt resource value: 0x7F0600A4
			public const int material_clock_hand_padding = 2131099812;
			
			// aapt resource value: 0x7F0600A5
			public const int material_clock_hand_stroke_width = 2131099813;
			
			// aapt resource value: 0x7F0600A6
			public const int material_clock_number_text_size = 2131099814;
			
			// aapt resource value: 0x7F0600A7
			public const int material_clock_period_toggle_height = 2131099815;
			
			// aapt resource value: 0x7F0600A8
			public const int material_clock_period_toggle_margin_left = 2131099816;
			
			// aapt resource value: 0x7F0600A9
			public const int material_clock_period_toggle_width = 2131099817;
			
			// aapt resource value: 0x7F0600AA
			public const int material_clock_size = 2131099818;
			
			// aapt resource value: 0x7F0600AB
			public const int material_cursor_inset_bottom = 2131099819;
			
			// aapt resource value: 0x7F0600AC
			public const int material_cursor_inset_top = 2131099820;
			
			// aapt resource value: 0x7F0600AD
			public const int material_cursor_width = 2131099821;
			
			// aapt resource value: 0x7F0600AE
			public const int material_emphasis_disabled = 2131099822;
			
			// aapt resource value: 0x7F0600AF
			public const int material_emphasis_high_type = 2131099823;
			
			// aapt resource value: 0x7F0600B0
			public const int material_emphasis_medium = 2131099824;
			
			// aapt resource value: 0x7F0600B1
			public const int material_filled_edittext_font_1_3_padding_bottom = 2131099825;
			
			// aapt resource value: 0x7F0600B2
			public const int material_filled_edittext_font_1_3_padding_top = 2131099826;
			
			// aapt resource value: 0x7F0600B3
			public const int material_filled_edittext_font_2_0_padding_bottom = 2131099827;
			
			// aapt resource value: 0x7F0600B4
			public const int material_filled_edittext_font_2_0_padding_top = 2131099828;
			
			// aapt resource value: 0x7F0600B5
			public const int material_font_1_3_box_collapsed_padding_top = 2131099829;
			
			// aapt resource value: 0x7F0600B6
			public const int material_font_2_0_box_collapsed_padding_top = 2131099830;
			
			// aapt resource value: 0x7F0600B7
			public const int material_helper_text_default_padding_top = 2131099831;
			
			// aapt resource value: 0x7F0600B8
			public const int material_helper_text_font_1_3_padding_horizontal = 2131099832;
			
			// aapt resource value: 0x7F0600B9
			public const int material_helper_text_font_1_3_padding_top = 2131099833;
			
			// aapt resource value: 0x7F0600BA
			public const int material_input_text_to_prefix_suffix_padding = 2131099834;
			
			// aapt resource value: 0x7F0600BD
			public const int material_textinput_default_width = 2131099837;
			
			// aapt resource value: 0x7F0600BE
			public const int material_textinput_max_width = 2131099838;
			
			// aapt resource value: 0x7F0600BF
			public const int material_textinput_min_width = 2131099839;
			
			// aapt resource value: 0x7F0600BB
			public const int material_text_view_test_line_height = 2131099835;
			
			// aapt resource value: 0x7F0600BC
			public const int material_text_view_test_line_height_override = 2131099836;
			
			// aapt resource value: 0x7F0600C2
			public const int material_timepicker_dialog_buttons_margin_top = 2131099842;
			
			// aapt resource value: 0x7F0600C0
			public const int material_time_picker_minimum_screen_height = 2131099840;
			
			// aapt resource value: 0x7F0600C1
			public const int material_time_picker_minimum_screen_width = 2131099841;
			
			// aapt resource value: 0x7F0600C3
			public const int mtrl_alert_dialog_background_inset_bottom = 2131099843;
			
			// aapt resource value: 0x7F0600C4
			public const int mtrl_alert_dialog_background_inset_end = 2131099844;
			
			// aapt resource value: 0x7F0600C5
			public const int mtrl_alert_dialog_background_inset_start = 2131099845;
			
			// aapt resource value: 0x7F0600C6
			public const int mtrl_alert_dialog_background_inset_top = 2131099846;
			
			// aapt resource value: 0x7F0600C7
			public const int mtrl_alert_dialog_picker_background_inset = 2131099847;
			
			// aapt resource value: 0x7F0600C8
			public const int mtrl_badge_horizontal_edge_offset = 2131099848;
			
			// aapt resource value: 0x7F0600C9
			public const int mtrl_badge_long_text_horizontal_padding = 2131099849;
			
			// aapt resource value: 0x7F0600CA
			public const int mtrl_badge_radius = 2131099850;
			
			// aapt resource value: 0x7F0600CB
			public const int mtrl_badge_text_horizontal_edge_offset = 2131099851;
			
			// aapt resource value: 0x7F0600CC
			public const int mtrl_badge_text_size = 2131099852;
			
			// aapt resource value: 0x7F0600CD
			public const int mtrl_badge_toolbar_action_menu_item_horizontal_offset = 2131099853;
			
			// aapt resource value: 0x7F0600CE
			public const int mtrl_badge_toolbar_action_menu_item_vertical_offset = 2131099854;
			
			// aapt resource value: 0x7F0600CF
			public const int mtrl_badge_with_text_radius = 2131099855;
			
			// aapt resource value: 0x7F0600D0
			public const int mtrl_bottomappbar_fabOffsetEndMode = 2131099856;
			
			// aapt resource value: 0x7F0600D1
			public const int mtrl_bottomappbar_fab_bottom_margin = 2131099857;
			
			// aapt resource value: 0x7F0600D2
			public const int mtrl_bottomappbar_fab_cradle_margin = 2131099858;
			
			// aapt resource value: 0x7F0600D3
			public const int mtrl_bottomappbar_fab_cradle_rounded_corner_radius = 2131099859;
			
			// aapt resource value: 0x7F0600D4
			public const int mtrl_bottomappbar_fab_cradle_vertical_offset = 2131099860;
			
			// aapt resource value: 0x7F0600D5
			public const int mtrl_bottomappbar_height = 2131099861;
			
			// aapt resource value: 0x7F0600D6
			public const int mtrl_btn_corner_radius = 2131099862;
			
			// aapt resource value: 0x7F0600D7
			public const int mtrl_btn_dialog_btn_min_width = 2131099863;
			
			// aapt resource value: 0x7F0600D8
			public const int mtrl_btn_disabled_elevation = 2131099864;
			
			// aapt resource value: 0x7F0600D9
			public const int mtrl_btn_disabled_z = 2131099865;
			
			// aapt resource value: 0x7F0600DA
			public const int mtrl_btn_elevation = 2131099866;
			
			// aapt resource value: 0x7F0600DB
			public const int mtrl_btn_focused_z = 2131099867;
			
			// aapt resource value: 0x7F0600DC
			public const int mtrl_btn_hovered_z = 2131099868;
			
			// aapt resource value: 0x7F0600DD
			public const int mtrl_btn_icon_btn_padding_left = 2131099869;
			
			// aapt resource value: 0x7F0600DE
			public const int mtrl_btn_icon_padding = 2131099870;
			
			// aapt resource value: 0x7F0600DF
			public const int mtrl_btn_inset = 2131099871;
			
			// aapt resource value: 0x7F0600E0
			public const int mtrl_btn_letter_spacing = 2131099872;
			
			// aapt resource value: 0x7F0600E1
			public const int mtrl_btn_max_width = 2131099873;
			
			// aapt resource value: 0x7F0600E2
			public const int mtrl_btn_padding_bottom = 2131099874;
			
			// aapt resource value: 0x7F0600E3
			public const int mtrl_btn_padding_left = 2131099875;
			
			// aapt resource value: 0x7F0600E4
			public const int mtrl_btn_padding_right = 2131099876;
			
			// aapt resource value: 0x7F0600E5
			public const int mtrl_btn_padding_top = 2131099877;
			
			// aapt resource value: 0x7F0600E6
			public const int mtrl_btn_pressed_z = 2131099878;
			
			// aapt resource value: 0x7F0600E7
			public const int mtrl_btn_snackbar_margin_horizontal = 2131099879;
			
			// aapt resource value: 0x7F0600E8
			public const int mtrl_btn_stroke_size = 2131099880;
			
			// aapt resource value: 0x7F0600E9
			public const int mtrl_btn_text_btn_icon_padding = 2131099881;
			
			// aapt resource value: 0x7F0600EA
			public const int mtrl_btn_text_btn_padding_left = 2131099882;
			
			// aapt resource value: 0x7F0600EB
			public const int mtrl_btn_text_btn_padding_right = 2131099883;
			
			// aapt resource value: 0x7F0600EC
			public const int mtrl_btn_text_size = 2131099884;
			
			// aapt resource value: 0x7F0600ED
			public const int mtrl_btn_z = 2131099885;
			
			// aapt resource value: 0x7F0600EE
			public const int mtrl_calendar_action_confirm_button_min_width = 2131099886;
			
			// aapt resource value: 0x7F0600EF
			public const int mtrl_calendar_action_height = 2131099887;
			
			// aapt resource value: 0x7F0600F0
			public const int mtrl_calendar_action_padding = 2131099888;
			
			// aapt resource value: 0x7F0600F1
			public const int mtrl_calendar_bottom_padding = 2131099889;
			
			// aapt resource value: 0x7F0600F2
			public const int mtrl_calendar_content_padding = 2131099890;
			
			// aapt resource value: 0x7F0600F9
			public const int mtrl_calendar_days_of_week_height = 2131099897;
			
			// aapt resource value: 0x7F0600F3
			public const int mtrl_calendar_day_corner = 2131099891;
			
			// aapt resource value: 0x7F0600F4
			public const int mtrl_calendar_day_height = 2131099892;
			
			// aapt resource value: 0x7F0600F5
			public const int mtrl_calendar_day_horizontal_padding = 2131099893;
			
			// aapt resource value: 0x7F0600F6
			public const int mtrl_calendar_day_today_stroke = 2131099894;
			
			// aapt resource value: 0x7F0600F7
			public const int mtrl_calendar_day_vertical_padding = 2131099895;
			
			// aapt resource value: 0x7F0600F8
			public const int mtrl_calendar_day_width = 2131099896;
			
			// aapt resource value: 0x7F0600FA
			public const int mtrl_calendar_dialog_background_inset = 2131099898;
			
			// aapt resource value: 0x7F0600FB
			public const int mtrl_calendar_header_content_padding = 2131099899;
			
			// aapt resource value: 0x7F0600FC
			public const int mtrl_calendar_header_content_padding_fullscreen = 2131099900;
			
			// aapt resource value: 0x7F0600FD
			public const int mtrl_calendar_header_divider_thickness = 2131099901;
			
			// aapt resource value: 0x7F0600FE
			public const int mtrl_calendar_header_height = 2131099902;
			
			// aapt resource value: 0x7F0600FF
			public const int mtrl_calendar_header_height_fullscreen = 2131099903;
			
			// aapt resource value: 0x7F060100
			public const int mtrl_calendar_header_selection_line_height = 2131099904;
			
			// aapt resource value: 0x7F060101
			public const int mtrl_calendar_header_text_padding = 2131099905;
			
			// aapt resource value: 0x7F060102
			public const int mtrl_calendar_header_toggle_margin_bottom = 2131099906;
			
			// aapt resource value: 0x7F060103
			public const int mtrl_calendar_header_toggle_margin_top = 2131099907;
			
			// aapt resource value: 0x7F060104
			public const int mtrl_calendar_landscape_header_width = 2131099908;
			
			// aapt resource value: 0x7F060105
			public const int mtrl_calendar_maximum_default_fullscreen_minor_axis = 2131099909;
			
			// aapt resource value: 0x7F060106
			public const int mtrl_calendar_month_horizontal_padding = 2131099910;
			
			// aapt resource value: 0x7F060107
			public const int mtrl_calendar_month_vertical_padding = 2131099911;
			
			// aapt resource value: 0x7F060108
			public const int mtrl_calendar_navigation_bottom_padding = 2131099912;
			
			// aapt resource value: 0x7F060109
			public const int mtrl_calendar_navigation_height = 2131099913;
			
			// aapt resource value: 0x7F06010A
			public const int mtrl_calendar_navigation_top_padding = 2131099914;
			
			// aapt resource value: 0x7F06010B
			public const int mtrl_calendar_pre_l_text_clip_padding = 2131099915;
			
			// aapt resource value: 0x7F06010C
			public const int mtrl_calendar_selection_baseline_to_top_fullscreen = 2131099916;
			
			// aapt resource value: 0x7F06010D
			public const int mtrl_calendar_selection_text_baseline_to_bottom = 2131099917;
			
			// aapt resource value: 0x7F06010E
			public const int mtrl_calendar_selection_text_baseline_to_bottom_fullscreen = 2131099918;
			
			// aapt resource value: 0x7F06010F
			public const int mtrl_calendar_selection_text_baseline_to_top = 2131099919;
			
			// aapt resource value: 0x7F060110
			public const int mtrl_calendar_text_input_padding_top = 2131099920;
			
			// aapt resource value: 0x7F060111
			public const int mtrl_calendar_title_baseline_to_top = 2131099921;
			
			// aapt resource value: 0x7F060112
			public const int mtrl_calendar_title_baseline_to_top_fullscreen = 2131099922;
			
			// aapt resource value: 0x7F060113
			public const int mtrl_calendar_year_corner = 2131099923;
			
			// aapt resource value: 0x7F060114
			public const int mtrl_calendar_year_height = 2131099924;
			
			// aapt resource value: 0x7F060115
			public const int mtrl_calendar_year_horizontal_padding = 2131099925;
			
			// aapt resource value: 0x7F060116
			public const int mtrl_calendar_year_vertical_padding = 2131099926;
			
			// aapt resource value: 0x7F060117
			public const int mtrl_calendar_year_width = 2131099927;
			
			// aapt resource value: 0x7F060118
			public const int mtrl_card_checked_icon_margin = 2131099928;
			
			// aapt resource value: 0x7F060119
			public const int mtrl_card_checked_icon_size = 2131099929;
			
			// aapt resource value: 0x7F06011A
			public const int mtrl_card_corner_radius = 2131099930;
			
			// aapt resource value: 0x7F06011B
			public const int mtrl_card_dragged_z = 2131099931;
			
			// aapt resource value: 0x7F06011C
			public const int mtrl_card_elevation = 2131099932;
			
			// aapt resource value: 0x7F06011D
			public const int mtrl_card_spacing = 2131099933;
			
			// aapt resource value: 0x7F06011E
			public const int mtrl_chip_pressed_translation_z = 2131099934;
			
			// aapt resource value: 0x7F06011F
			public const int mtrl_chip_text_size = 2131099935;
			
			// aapt resource value: 0x7F060120
			public const int mtrl_edittext_rectangle_top_offset = 2131099936;
			
			// aapt resource value: 0x7F060121
			public const int mtrl_exposed_dropdown_menu_popup_elevation = 2131099937;
			
			// aapt resource value: 0x7F060122
			public const int mtrl_exposed_dropdown_menu_popup_vertical_offset = 2131099938;
			
			// aapt resource value: 0x7F060123
			public const int mtrl_exposed_dropdown_menu_popup_vertical_padding = 2131099939;
			
			// aapt resource value: 0x7F060124
			public const int mtrl_extended_fab_bottom_padding = 2131099940;
			
			// aapt resource value: 0x7F060125
			public const int mtrl_extended_fab_corner_radius = 2131099941;
			
			// aapt resource value: 0x7F060126
			public const int mtrl_extended_fab_disabled_elevation = 2131099942;
			
			// aapt resource value: 0x7F060127
			public const int mtrl_extended_fab_disabled_translation_z = 2131099943;
			
			// aapt resource value: 0x7F060128
			public const int mtrl_extended_fab_elevation = 2131099944;
			
			// aapt resource value: 0x7F060129
			public const int mtrl_extended_fab_end_padding = 2131099945;
			
			// aapt resource value: 0x7F06012A
			public const int mtrl_extended_fab_end_padding_icon = 2131099946;
			
			// aapt resource value: 0x7F06012B
			public const int mtrl_extended_fab_icon_size = 2131099947;
			
			// aapt resource value: 0x7F06012C
			public const int mtrl_extended_fab_icon_text_spacing = 2131099948;
			
			// aapt resource value: 0x7F06012D
			public const int mtrl_extended_fab_min_height = 2131099949;
			
			// aapt resource value: 0x7F06012E
			public const int mtrl_extended_fab_min_width = 2131099950;
			
			// aapt resource value: 0x7F06012F
			public const int mtrl_extended_fab_start_padding = 2131099951;
			
			// aapt resource value: 0x7F060130
			public const int mtrl_extended_fab_start_padding_icon = 2131099952;
			
			// aapt resource value: 0x7F060131
			public const int mtrl_extended_fab_top_padding = 2131099953;
			
			// aapt resource value: 0x7F060132
			public const int mtrl_extended_fab_translation_z_base = 2131099954;
			
			// aapt resource value: 0x7F060133
			public const int mtrl_extended_fab_translation_z_hovered_focused = 2131099955;
			
			// aapt resource value: 0x7F060134
			public const int mtrl_extended_fab_translation_z_pressed = 2131099956;
			
			// aapt resource value: 0x7F060135
			public const int mtrl_fab_elevation = 2131099957;
			
			// aapt resource value: 0x7F060136
			public const int mtrl_fab_min_touch_target = 2131099958;
			
			// aapt resource value: 0x7F060137
			public const int mtrl_fab_translation_z_hovered_focused = 2131099959;
			
			// aapt resource value: 0x7F060138
			public const int mtrl_fab_translation_z_pressed = 2131099960;
			
			// aapt resource value: 0x7F060139
			public const int mtrl_high_ripple_default_alpha = 2131099961;
			
			// aapt resource value: 0x7F06013A
			public const int mtrl_high_ripple_focused_alpha = 2131099962;
			
			// aapt resource value: 0x7F06013B
			public const int mtrl_high_ripple_hovered_alpha = 2131099963;
			
			// aapt resource value: 0x7F06013C
			public const int mtrl_high_ripple_pressed_alpha = 2131099964;
			
			// aapt resource value: 0x7F06013D
			public const int mtrl_large_touch_target = 2131099965;
			
			// aapt resource value: 0x7F06013E
			public const int mtrl_low_ripple_default_alpha = 2131099966;
			
			// aapt resource value: 0x7F06013F
			public const int mtrl_low_ripple_focused_alpha = 2131099967;
			
			// aapt resource value: 0x7F060140
			public const int mtrl_low_ripple_hovered_alpha = 2131099968;
			
			// aapt resource value: 0x7F060141
			public const int mtrl_low_ripple_pressed_alpha = 2131099969;
			
			// aapt resource value: 0x7F060142
			public const int mtrl_min_touch_target_size = 2131099970;
			
			// aapt resource value: 0x7F060143
			public const int mtrl_navigation_bar_item_default_icon_size = 2131099971;
			
			// aapt resource value: 0x7F060144
			public const int mtrl_navigation_bar_item_default_margin = 2131099972;
			
			// aapt resource value: 0x7F060145
			public const int mtrl_navigation_elevation = 2131099973;
			
			// aapt resource value: 0x7F060146
			public const int mtrl_navigation_item_horizontal_padding = 2131099974;
			
			// aapt resource value: 0x7F060147
			public const int mtrl_navigation_item_icon_padding = 2131099975;
			
			// aapt resource value: 0x7F060148
			public const int mtrl_navigation_item_icon_size = 2131099976;
			
			// aapt resource value: 0x7F060149
			public const int mtrl_navigation_item_shape_horizontal_margin = 2131099977;
			
			// aapt resource value: 0x7F06014A
			public const int mtrl_navigation_item_shape_vertical_margin = 2131099978;
			
			// aapt resource value: 0x7F06014B
			public const int mtrl_navigation_rail_active_text_size = 2131099979;
			
			// aapt resource value: 0x7F06014C
			public const int mtrl_navigation_rail_compact_width = 2131099980;
			
			// aapt resource value: 0x7F06014D
			public const int mtrl_navigation_rail_default_width = 2131099981;
			
			// aapt resource value: 0x7F06014E
			public const int mtrl_navigation_rail_elevation = 2131099982;
			
			// aapt resource value: 0x7F06014F
			public const int mtrl_navigation_rail_icon_margin = 2131099983;
			
			// aapt resource value: 0x7F060150
			public const int mtrl_navigation_rail_icon_size = 2131099984;
			
			// aapt resource value: 0x7F060151
			public const int mtrl_navigation_rail_margin = 2131099985;
			
			// aapt resource value: 0x7F060152
			public const int mtrl_navigation_rail_text_bottom_margin = 2131099986;
			
			// aapt resource value: 0x7F060153
			public const int mtrl_navigation_rail_text_size = 2131099987;
			
			// aapt resource value: 0x7F060154
			public const int mtrl_progress_circular_inset = 2131099988;
			
			// aapt resource value: 0x7F060155
			public const int mtrl_progress_circular_inset_extra_small = 2131099989;
			
			// aapt resource value: 0x7F060156
			public const int mtrl_progress_circular_inset_medium = 2131099990;
			
			// aapt resource value: 0x7F060157
			public const int mtrl_progress_circular_inset_small = 2131099991;
			
			// aapt resource value: 0x7F060158
			public const int mtrl_progress_circular_radius = 2131099992;
			
			// aapt resource value: 0x7F060159
			public const int mtrl_progress_circular_size = 2131099993;
			
			// aapt resource value: 0x7F06015A
			public const int mtrl_progress_circular_size_extra_small = 2131099994;
			
			// aapt resource value: 0x7F06015B
			public const int mtrl_progress_circular_size_medium = 2131099995;
			
			// aapt resource value: 0x7F06015C
			public const int mtrl_progress_circular_size_small = 2131099996;
			
			// aapt resource value: 0x7F06015D
			public const int mtrl_progress_circular_track_thickness_extra_small = 2131099997;
			
			// aapt resource value: 0x7F06015E
			public const int mtrl_progress_circular_track_thickness_medium = 2131099998;
			
			// aapt resource value: 0x7F06015F
			public const int mtrl_progress_circular_track_thickness_small = 2131099999;
			
			// aapt resource value: 0x7F060160
			public const int mtrl_progress_indicator_full_rounded_corner_radius = 2131100000;
			
			// aapt resource value: 0x7F060161
			public const int mtrl_progress_track_thickness = 2131100001;
			
			// aapt resource value: 0x7F060162
			public const int mtrl_shape_corner_size_large_component = 2131100002;
			
			// aapt resource value: 0x7F060163
			public const int mtrl_shape_corner_size_medium_component = 2131100003;
			
			// aapt resource value: 0x7F060164
			public const int mtrl_shape_corner_size_small_component = 2131100004;
			
			// aapt resource value: 0x7F060165
			public const int mtrl_slider_halo_radius = 2131100005;
			
			// aapt resource value: 0x7F060166
			public const int mtrl_slider_label_padding = 2131100006;
			
			// aapt resource value: 0x7F060167
			public const int mtrl_slider_label_radius = 2131100007;
			
			// aapt resource value: 0x7F060168
			public const int mtrl_slider_label_square_side = 2131100008;
			
			// aapt resource value: 0x7F060169
			public const int mtrl_slider_thumb_elevation = 2131100009;
			
			// aapt resource value: 0x7F06016A
			public const int mtrl_slider_thumb_radius = 2131100010;
			
			// aapt resource value: 0x7F06016B
			public const int mtrl_slider_track_height = 2131100011;
			
			// aapt resource value: 0x7F06016C
			public const int mtrl_slider_track_side_padding = 2131100012;
			
			// aapt resource value: 0x7F06016D
			public const int mtrl_slider_track_top = 2131100013;
			
			// aapt resource value: 0x7F06016E
			public const int mtrl_slider_widget_height = 2131100014;
			
			// aapt resource value: 0x7F06016F
			public const int mtrl_snackbar_action_text_color_alpha = 2131100015;
			
			// aapt resource value: 0x7F060170
			public const int mtrl_snackbar_background_corner_radius = 2131100016;
			
			// aapt resource value: 0x7F060171
			public const int mtrl_snackbar_background_overlay_color_alpha = 2131100017;
			
			// aapt resource value: 0x7F060172
			public const int mtrl_snackbar_margin = 2131100018;
			
			// aapt resource value: 0x7F060173
			public const int mtrl_snackbar_message_margin_horizontal = 2131100019;
			
			// aapt resource value: 0x7F060174
			public const int mtrl_snackbar_padding_horizontal = 2131100020;
			
			// aapt resource value: 0x7F060175
			public const int mtrl_switch_thumb_elevation = 2131100021;
			
			// aapt resource value: 0x7F060176
			public const int mtrl_textinput_box_corner_radius_medium = 2131100022;
			
			// aapt resource value: 0x7F060177
			public const int mtrl_textinput_box_corner_radius_small = 2131100023;
			
			// aapt resource value: 0x7F060178
			public const int mtrl_textinput_box_label_cutout_padding = 2131100024;
			
			// aapt resource value: 0x7F060179
			public const int mtrl_textinput_box_stroke_width_default = 2131100025;
			
			// aapt resource value: 0x7F06017A
			public const int mtrl_textinput_box_stroke_width_focused = 2131100026;
			
			// aapt resource value: 0x7F06017B
			public const int mtrl_textinput_counter_margin_start = 2131100027;
			
			// aapt resource value: 0x7F06017C
			public const int mtrl_textinput_end_icon_margin_start = 2131100028;
			
			// aapt resource value: 0x7F06017D
			public const int mtrl_textinput_outline_box_expanded_padding = 2131100029;
			
			// aapt resource value: 0x7F06017E
			public const int mtrl_textinput_start_icon_margin_end = 2131100030;
			
			// aapt resource value: 0x7F06017F
			public const int mtrl_toolbar_default_height = 2131100031;
			
			// aapt resource value: 0x7F060180
			public const int mtrl_tooltip_arrowSize = 2131100032;
			
			// aapt resource value: 0x7F060181
			public const int mtrl_tooltip_cornerSize = 2131100033;
			
			// aapt resource value: 0x7F060182
			public const int mtrl_tooltip_minHeight = 2131100034;
			
			// aapt resource value: 0x7F060183
			public const int mtrl_tooltip_minWidth = 2131100035;
			
			// aapt resource value: 0x7F060184
			public const int mtrl_tooltip_padding = 2131100036;
			
			// aapt resource value: 0x7F060185
			public const int mtrl_transition_shared_axis_slide_distance = 2131100037;
			
			// aapt resource value: 0x7F060186
			public const int notification_action_icon_size = 2131100038;
			
			// aapt resource value: 0x7F060187
			public const int notification_action_text_size = 2131100039;
			
			// aapt resource value: 0x7F060188
			public const int notification_big_circle_margin = 2131100040;
			
			// aapt resource value: 0x7F060189
			public const int notification_content_margin_start = 2131100041;
			
			// aapt resource value: 0x7F06018A
			public const int notification_large_icon_height = 2131100042;
			
			// aapt resource value: 0x7F06018B
			public const int notification_large_icon_width = 2131100043;
			
			// aapt resource value: 0x7F06018C
			public const int notification_main_column_padding_top = 2131100044;
			
			// aapt resource value: 0x7F06018D
			public const int notification_media_narrow_margin = 2131100045;
			
			// aapt resource value: 0x7F06018E
			public const int notification_right_icon_size = 2131100046;
			
			// aapt resource value: 0x7F06018F
			public const int notification_right_side_padding_top = 2131100047;
			
			// aapt resource value: 0x7F060190
			public const int notification_small_icon_background_padding = 2131100048;
			
			// aapt resource value: 0x7F060191
			public const int notification_small_icon_size_as_large = 2131100049;
			
			// aapt resource value: 0x7F060192
			public const int notification_subtext_size = 2131100050;
			
			// aapt resource value: 0x7F060193
			public const int notification_top_pad = 2131100051;
			
			// aapt resource value: 0x7F060194
			public const int notification_top_pad_large_text = 2131100052;
			
			// aapt resource value: 0x7F060195
			public const int preference_dropdown_padding_start = 2131100053;
			
			// aapt resource value: 0x7F060196
			public const int preference_icon_minWidth = 2131100054;
			
			// aapt resource value: 0x7F060197
			public const int preference_seekbar_padding_horizontal = 2131100055;
			
			// aapt resource value: 0x7F060198
			public const int preference_seekbar_padding_vertical = 2131100056;
			
			// aapt resource value: 0x7F060199
			public const int preference_seekbar_value_minWidth = 2131100057;
			
			// aapt resource value: 0x7F06019A
			public const int test_mtrl_calendar_day_cornerSize = 2131100058;
			
			// aapt resource value: 0x7F06019B
			public const int test_navigation_bar_active_item_max_width = 2131100059;
			
			// aapt resource value: 0x7F06019C
			public const int test_navigation_bar_active_item_min_width = 2131100060;
			
			// aapt resource value: 0x7F06019D
			public const int test_navigation_bar_active_text_size = 2131100061;
			
			// aapt resource value: 0x7F06019E
			public const int test_navigation_bar_elevation = 2131100062;
			
			// aapt resource value: 0x7F06019F
			public const int test_navigation_bar_height = 2131100063;
			
			// aapt resource value: 0x7F0601A0
			public const int test_navigation_bar_icon_size = 2131100064;
			
			// aapt resource value: 0x7F0601A1
			public const int test_navigation_bar_item_max_width = 2131100065;
			
			// aapt resource value: 0x7F0601A2
			public const int test_navigation_bar_item_min_width = 2131100066;
			
			// aapt resource value: 0x7F0601A3
			public const int test_navigation_bar_label_padding = 2131100067;
			
			// aapt resource value: 0x7F0601A4
			public const int test_navigation_bar_shadow_height = 2131100068;
			
			// aapt resource value: 0x7F0601A5
			public const int test_navigation_bar_text_size = 2131100069;
			
			// aapt resource value: 0x7F0601A6
			public const int tooltip_corner_radius = 2131100070;
			
			// aapt resource value: 0x7F0601A7
			public const int tooltip_horizontal_padding = 2131100071;
			
			// aapt resource value: 0x7F0601A8
			public const int tooltip_margin = 2131100072;
			
			// aapt resource value: 0x7F0601A9
			public const int tooltip_precise_anchor_extra_offset = 2131100073;
			
			// aapt resource value: 0x7F0601AA
			public const int tooltip_precise_anchor_threshold = 2131100074;
			
			// aapt resource value: 0x7F0601AB
			public const int tooltip_vertical_padding = 2131100075;
			
			// aapt resource value: 0x7F0601AC
			public const int tooltip_y_offset_non_touch = 2131100076;
			
			// aapt resource value: 0x7F0601AD
			public const int tooltip_y_offset_touch = 2131100077;
			
			static Dimension()
			{
				global::Android.Runtime.ResourceIdManager.UpdateIdValues();
			}
			
			private Dimension()
			{
			}
		}
		
		public partial class Drawable
		{
			
			// aapt resource value: 0x7F070006
			public const int abc_ab_share_pack_mtrl_alpha = 2131165190;
			
			// aapt resource value: 0x7F070007
			public const int abc_action_bar_item_background_material = 2131165191;
			
			// aapt resource value: 0x7F070008
			public const int abc_btn_borderless_material = 2131165192;
			
			// aapt resource value: 0x7F070009
			public const int abc_btn_check_material = 2131165193;
			
			// aapt resource value: 0x7F07000A
			public const int abc_btn_check_material_anim = 2131165194;
			
			// aapt resource value: 0x7F07000B
			public const int abc_btn_check_to_on_mtrl_000 = 2131165195;
			
			// aapt resource value: 0x7F07000C
			public const int abc_btn_check_to_on_mtrl_015 = 2131165196;
			
			// aapt resource value: 0x7F07000D
			public const int abc_btn_colored_material = 2131165197;
			
			// aapt resource value: 0x7F07000E
			public const int abc_btn_default_mtrl_shape = 2131165198;
			
			// aapt resource value: 0x7F07000F
			public const int abc_btn_radio_material = 2131165199;
			
			// aapt resource value: 0x7F070010
			public const int abc_btn_radio_material_anim = 2131165200;
			
			// aapt resource value: 0x7F070011
			public const int abc_btn_radio_to_on_mtrl_000 = 2131165201;
			
			// aapt resource value: 0x7F070012
			public const int abc_btn_radio_to_on_mtrl_015 = 2131165202;
			
			// aapt resource value: 0x7F070013
			public const int abc_btn_switch_to_on_mtrl_00001 = 2131165203;
			
			// aapt resource value: 0x7F070014
			public const int abc_btn_switch_to_on_mtrl_00012 = 2131165204;
			
			// aapt resource value: 0x7F070015
			public const int abc_cab_background_internal_bg = 2131165205;
			
			// aapt resource value: 0x7F070016
			public const int abc_cab_background_top_material = 2131165206;
			
			// aapt resource value: 0x7F070017
			public const int abc_cab_background_top_mtrl_alpha = 2131165207;
			
			// aapt resource value: 0x7F070018
			public const int abc_control_background_material = 2131165208;
			
			// aapt resource value: 0x7F070019
			public const int abc_dialog_material_background = 2131165209;
			
			// aapt resource value: 0x7F07001A
			public const int abc_edit_text_material = 2131165210;
			
			// aapt resource value: 0x7F07001B
			public const int abc_ic_ab_back_material = 2131165211;
			
			// aapt resource value: 0x7F07001C
			public const int abc_ic_arrow_drop_right_black_24dp = 2131165212;
			
			// aapt resource value: 0x7F07001D
			public const int abc_ic_clear_material = 2131165213;
			
			// aapt resource value: 0x7F07001E
			public const int abc_ic_commit_search_api_mtrl_alpha = 2131165214;
			
			// aapt resource value: 0x7F07001F
			public const int abc_ic_go_search_api_material = 2131165215;
			
			// aapt resource value: 0x7F070020
			public const int abc_ic_menu_copy_mtrl_am_alpha = 2131165216;
			
			// aapt resource value: 0x7F070021
			public const int abc_ic_menu_cut_mtrl_alpha = 2131165217;
			
			// aapt resource value: 0x7F070022
			public const int abc_ic_menu_overflow_material = 2131165218;
			
			// aapt resource value: 0x7F070023
			public const int abc_ic_menu_paste_mtrl_am_alpha = 2131165219;
			
			// aapt resource value: 0x7F070024
			public const int abc_ic_menu_selectall_mtrl_alpha = 2131165220;
			
			// aapt resource value: 0x7F070025
			public const int abc_ic_menu_share_mtrl_alpha = 2131165221;
			
			// aapt resource value: 0x7F070026
			public const int abc_ic_search_api_material = 2131165222;
			
			// aapt resource value: 0x7F070027
			public const int abc_ic_voice_search_api_material = 2131165223;
			
			// aapt resource value: 0x7F070028
			public const int abc_item_background_holo_dark = 2131165224;
			
			// aapt resource value: 0x7F070029
			public const int abc_item_background_holo_light = 2131165225;
			
			// aapt resource value: 0x7F07002A
			public const int abc_list_divider_material = 2131165226;
			
			// aapt resource value: 0x7F07002B
			public const int abc_list_divider_mtrl_alpha = 2131165227;
			
			// aapt resource value: 0x7F07002C
			public const int abc_list_focused_holo = 2131165228;
			
			// aapt resource value: 0x7F07002D
			public const int abc_list_longpressed_holo = 2131165229;
			
			// aapt resource value: 0x7F07002E
			public const int abc_list_pressed_holo_dark = 2131165230;
			
			// aapt resource value: 0x7F07002F
			public const int abc_list_pressed_holo_light = 2131165231;
			
			// aapt resource value: 0x7F070030
			public const int abc_list_selector_background_transition_holo_dark = 2131165232;
			
			// aapt resource value: 0x7F070031
			public const int abc_list_selector_background_transition_holo_light = 2131165233;
			
			// aapt resource value: 0x7F070032
			public const int abc_list_selector_disabled_holo_dark = 2131165234;
			
			// aapt resource value: 0x7F070033
			public const int abc_list_selector_disabled_holo_light = 2131165235;
			
			// aapt resource value: 0x7F070034
			public const int abc_list_selector_holo_dark = 2131165236;
			
			// aapt resource value: 0x7F070035
			public const int abc_list_selector_holo_light = 2131165237;
			
			// aapt resource value: 0x7F070036
			public const int abc_menu_hardkey_panel_mtrl_mult = 2131165238;
			
			// aapt resource value: 0x7F070037
			public const int abc_popup_background_mtrl_mult = 2131165239;
			
			// aapt resource value: 0x7F070038
			public const int abc_ratingbar_indicator_material = 2131165240;
			
			// aapt resource value: 0x7F070039
			public const int abc_ratingbar_material = 2131165241;
			
			// aapt resource value: 0x7F07003A
			public const int abc_ratingbar_small_material = 2131165242;
			
			// aapt resource value: 0x7F07003B
			public const int abc_scrubber_control_off_mtrl_alpha = 2131165243;
			
			// aapt resource value: 0x7F07003C
			public const int abc_scrubber_control_to_pressed_mtrl_000 = 2131165244;
			
			// aapt resource value: 0x7F07003D
			public const int abc_scrubber_control_to_pressed_mtrl_005 = 2131165245;
			
			// aapt resource value: 0x7F07003E
			public const int abc_scrubber_primary_mtrl_alpha = 2131165246;
			
			// aapt resource value: 0x7F07003F
			public const int abc_scrubber_track_mtrl_alpha = 2131165247;
			
			// aapt resource value: 0x7F070040
			public const int abc_seekbar_thumb_material = 2131165248;
			
			// aapt resource value: 0x7F070041
			public const int abc_seekbar_tick_mark_material = 2131165249;
			
			// aapt resource value: 0x7F070042
			public const int abc_seekbar_track_material = 2131165250;
			
			// aapt resource value: 0x7F070043
			public const int abc_spinner_mtrl_am_alpha = 2131165251;
			
			// aapt resource value: 0x7F070044
			public const int abc_spinner_textfield_background_material = 2131165252;
			
			// aapt resource value: 0x7F070045
			public const int abc_star_black_48dp = 2131165253;
			
			// aapt resource value: 0x7F070046
			public const int abc_star_half_black_48dp = 2131165254;
			
			// aapt resource value: 0x7F070047
			public const int abc_switch_thumb_material = 2131165255;
			
			// aapt resource value: 0x7F070048
			public const int abc_switch_track_mtrl_alpha = 2131165256;
			
			// aapt resource value: 0x7F070049
			public const int abc_tab_indicator_material = 2131165257;
			
			// aapt resource value: 0x7F07004A
			public const int abc_tab_indicator_mtrl_alpha = 2131165258;
			
			// aapt resource value: 0x7F07004F
			public const int abc_textfield_activated_mtrl_alpha = 2131165263;
			
			// aapt resource value: 0x7F070050
			public const int abc_textfield_default_mtrl_alpha = 2131165264;
			
			// aapt resource value: 0x7F070051
			public const int abc_textfield_search_activated_mtrl_alpha = 2131165265;
			
			// aapt resource value: 0x7F070052
			public const int abc_textfield_search_default_mtrl_alpha = 2131165266;
			
			// aapt resource value: 0x7F070053
			public const int abc_textfield_search_material = 2131165267;
			
			// aapt resource value: 0x7F07004B
			public const int abc_text_cursor_material = 2131165259;
			
			// aapt resource value: 0x7F07004C
			public const int abc_text_select_handle_left_mtrl = 2131165260;
			
			// aapt resource value: 0x7F07004D
			public const int abc_text_select_handle_middle_mtrl = 2131165261;
			
			// aapt resource value: 0x7F07004E
			public const int abc_text_select_handle_right_mtrl = 2131165262;
			
			// aapt resource value: 0x7F070054
			public const int abc_vector_test = 2131165268;
			
			// aapt resource value: 0x7F070055
			public const int avd_hide_password = 2131165269;
			
			// aapt resource value: 0x7F070056
			public const int avd_show_password = 2131165270;
			
			// aapt resource value: 0x7F070057
			public const int btn_checkbox_checked_mtrl = 2131165271;
			
			// aapt resource value: 0x7F070058
			public const int btn_checkbox_checked_to_unchecked_mtrl_animation = 2131165272;
			
			// aapt resource value: 0x7F070059
			public const int btn_checkbox_unchecked_mtrl = 2131165273;
			
			// aapt resource value: 0x7F07005A
			public const int btn_checkbox_unchecked_to_checked_mtrl_animation = 2131165274;
			
			// aapt resource value: 0x7F07005B
			public const int btn_radio_off_mtrl = 2131165275;
			
			// aapt resource value: 0x7F07005C
			public const int btn_radio_off_to_on_mtrl_animation = 2131165276;
			
			// aapt resource value: 0x7F07005D
			public const int btn_radio_on_mtrl = 2131165277;
			
			// aapt resource value: 0x7F07005E
			public const int btn_radio_on_to_off_mtrl_animation = 2131165278;
			
			// aapt resource value: 0x7F07005F
			public const int design_fab_background = 2131165279;
			
			// aapt resource value: 0x7F070060
			public const int design_ic_visibility = 2131165280;
			
			// aapt resource value: 0x7F070061
			public const int design_ic_visibility_off = 2131165281;
			
			// aapt resource value: 0x7F070062
			public const int design_password_eye = 2131165282;
			
			// aapt resource value: 0x7F070063
			public const int design_snackbar_background = 2131165283;
			
			// aapt resource value: 0x7F070064
			public const int ic_arrow_down_24dp = 2131165284;
			
			// aapt resource value: 0x7F070065
			public const int ic_clock_black_24dp = 2131165285;
			
			// aapt resource value: 0x7F070066
			public const int ic_keyboard_black_24dp = 2131165286;
			
			// aapt resource value: 0x7F070067
			public const int ic_mtrl_checked_circle = 2131165287;
			
			// aapt resource value: 0x7F070068
			public const int ic_mtrl_chip_checked_black = 2131165288;
			
			// aapt resource value: 0x7F070069
			public const int ic_mtrl_chip_checked_circle = 2131165289;
			
			// aapt resource value: 0x7F07006A
			public const int ic_mtrl_chip_close_circle = 2131165290;
			
			// aapt resource value: 0x7F07006B
			public const int material_cursor_drawable = 2131165291;
			
			// aapt resource value: 0x7F07006C
			public const int material_ic_calendar_black_24dp = 2131165292;
			
			// aapt resource value: 0x7F07006D
			public const int material_ic_clear_black_24dp = 2131165293;
			
			// aapt resource value: 0x7F07006E
			public const int material_ic_edit_black_24dp = 2131165294;
			
			// aapt resource value: 0x7F07006F
			public const int material_ic_keyboard_arrow_left_black_24dp = 2131165295;
			
			// aapt resource value: 0x7F070070
			public const int material_ic_keyboard_arrow_next_black_24dp = 2131165296;
			
			// aapt resource value: 0x7F070071
			public const int material_ic_keyboard_arrow_previous_black_24dp = 2131165297;
			
			// aapt resource value: 0x7F070072
			public const int material_ic_keyboard_arrow_right_black_24dp = 2131165298;
			
			// aapt resource value: 0x7F070073
			public const int material_ic_menu_arrow_down_black_24dp = 2131165299;
			
			// aapt resource value: 0x7F070074
			public const int material_ic_menu_arrow_up_black_24dp = 2131165300;
			
			// aapt resource value: 0x7F070075
			public const int mtrl_dialog_background = 2131165301;
			
			// aapt resource value: 0x7F070076
			public const int mtrl_dropdown_arrow = 2131165302;
			
			// aapt resource value: 0x7F070077
			public const int mtrl_ic_arrow_drop_down = 2131165303;
			
			// aapt resource value: 0x7F070078
			public const int mtrl_ic_arrow_drop_up = 2131165304;
			
			// aapt resource value: 0x7F070079
			public const int mtrl_ic_cancel = 2131165305;
			
			// aapt resource value: 0x7F07007A
			public const int mtrl_ic_error = 2131165306;
			
			// aapt resource value: 0x7F07007B
			public const int mtrl_navigation_bar_item_background = 2131165307;
			
			// aapt resource value: 0x7F07007C
			public const int mtrl_popupmenu_background = 2131165308;
			
			// aapt resource value: 0x7F07007D
			public const int mtrl_popupmenu_background_dark = 2131165309;
			
			// aapt resource value: 0x7F07007E
			public const int mtrl_tabs_default_indicator = 2131165310;
			
			// aapt resource value: 0x7F07007F
			public const int navigation_empty_icon = 2131165311;
			
			// aapt resource value: 0x7F070080
			public const int notification_action_background = 2131165312;
			
			// aapt resource value: 0x7F070081
			public const int notification_bg = 2131165313;
			
			// aapt resource value: 0x7F070082
			public const int notification_bg_low = 2131165314;
			
			// aapt resource value: 0x7F070083
			public const int notification_bg_low_normal = 2131165315;
			
			// aapt resource value: 0x7F070084
			public const int notification_bg_low_pressed = 2131165316;
			
			// aapt resource value: 0x7F070085
			public const int notification_bg_normal = 2131165317;
			
			// aapt resource value: 0x7F070086
			public const int notification_bg_normal_pressed = 2131165318;
			
			// aapt resource value: 0x7F070087
			public const int notification_icon_background = 2131165319;
			
			// aapt resource value: 0x7F070088
			public const int notification_template_icon_bg = 2131165320;
			
			// aapt resource value: 0x7F070089
			public const int notification_template_icon_low_bg = 2131165321;
			
			// aapt resource value: 0x7F07008A
			public const int notification_tile_bg = 2131165322;
			
			// aapt resource value: 0x7F07008B
			public const int notify_panel_notification_icon_bg = 2131165323;
			
			// aapt resource value: 0x7F07008C
			public const int preference_list_divider_material = 2131165324;
			
			// aapt resource value: 0x7F07008D
			public const int test_custom_background = 2131165325;
			
			// aapt resource value: 0x7F07008E
			public const int tooltip_frame_dark = 2131165326;
			
			// aapt resource value: 0x7F07008F
			public const int tooltip_frame_light = 2131165327;
			
			static Drawable()
			{
				global::Android.Runtime.ResourceIdManager.UpdateIdValues();
			}
			
			private Drawable()
			{
			}
		}
		
		public partial class Id
		{
			
			// aapt resource value: 0x7F08000E
			public const int accelerate = 2131230734;
			
			// aapt resource value: 0x7F08000F
			public const int acceptCallButton = 2131230735;
			
			// aapt resource value: 0x7F080010
			public const int accessibility_action_clickable_span = 2131230736;
			
			// aapt resource value: 0x7F080011
			public const int accessibility_custom_action_0 = 2131230737;
			
			// aapt resource value: 0x7F080012
			public const int accessibility_custom_action_1 = 2131230738;
			
			// aapt resource value: 0x7F080013
			public const int accessibility_custom_action_10 = 2131230739;
			
			// aapt resource value: 0x7F080014
			public const int accessibility_custom_action_11 = 2131230740;
			
			// aapt resource value: 0x7F080015
			public const int accessibility_custom_action_12 = 2131230741;
			
			// aapt resource value: 0x7F080016
			public const int accessibility_custom_action_13 = 2131230742;
			
			// aapt resource value: 0x7F080017
			public const int accessibility_custom_action_14 = 2131230743;
			
			// aapt resource value: 0x7F080018
			public const int accessibility_custom_action_15 = 2131230744;
			
			// aapt resource value: 0x7F080019
			public const int accessibility_custom_action_16 = 2131230745;
			
			// aapt resource value: 0x7F08001A
			public const int accessibility_custom_action_17 = 2131230746;
			
			// aapt resource value: 0x7F08001B
			public const int accessibility_custom_action_18 = 2131230747;
			
			// aapt resource value: 0x7F08001C
			public const int accessibility_custom_action_19 = 2131230748;
			
			// aapt resource value: 0x7F08001D
			public const int accessibility_custom_action_2 = 2131230749;
			
			// aapt resource value: 0x7F08001E
			public const int accessibility_custom_action_20 = 2131230750;
			
			// aapt resource value: 0x7F08001F
			public const int accessibility_custom_action_21 = 2131230751;
			
			// aapt resource value: 0x7F080020
			public const int accessibility_custom_action_22 = 2131230752;
			
			// aapt resource value: 0x7F080021
			public const int accessibility_custom_action_23 = 2131230753;
			
			// aapt resource value: 0x7F080022
			public const int accessibility_custom_action_24 = 2131230754;
			
			// aapt resource value: 0x7F080023
			public const int accessibility_custom_action_25 = 2131230755;
			
			// aapt resource value: 0x7F080024
			public const int accessibility_custom_action_26 = 2131230756;
			
			// aapt resource value: 0x7F080025
			public const int accessibility_custom_action_27 = 2131230757;
			
			// aapt resource value: 0x7F080026
			public const int accessibility_custom_action_28 = 2131230758;
			
			// aapt resource value: 0x7F080027
			public const int accessibility_custom_action_29 = 2131230759;
			
			// aapt resource value: 0x7F080028
			public const int accessibility_custom_action_3 = 2131230760;
			
			// aapt resource value: 0x7F080029
			public const int accessibility_custom_action_30 = 2131230761;
			
			// aapt resource value: 0x7F08002A
			public const int accessibility_custom_action_31 = 2131230762;
			
			// aapt resource value: 0x7F08002B
			public const int accessibility_custom_action_4 = 2131230763;
			
			// aapt resource value: 0x7F08002C
			public const int accessibility_custom_action_5 = 2131230764;
			
			// aapt resource value: 0x7F08002D
			public const int accessibility_custom_action_6 = 2131230765;
			
			// aapt resource value: 0x7F08002E
			public const int accessibility_custom_action_7 = 2131230766;
			
			// aapt resource value: 0x7F08002F
			public const int accessibility_custom_action_8 = 2131230767;
			
			// aapt resource value: 0x7F080030
			public const int accessibility_custom_action_9 = 2131230768;
			
			// aapt resource value: 0x7F080031
			public const int action0 = 2131230769;
			
			// aapt resource value: 0x7F080032
			public const int actionDown = 2131230770;
			
			// aapt resource value: 0x7F080033
			public const int actionDownUp = 2131230771;
			
			// aapt resource value: 0x7F080046
			public const int actions = 2131230790;
			
			// aapt resource value: 0x7F080034
			public const int actionUp = 2131230772;
			
			// aapt resource value: 0x7F080035
			public const int action_bar = 2131230773;
			
			// aapt resource value: 0x7F080036
			public const int action_bar_activity_content = 2131230774;
			
			// aapt resource value: 0x7F080037
			public const int action_bar_container = 2131230775;
			
			// aapt resource value: 0x7F080038
			public const int action_bar_root = 2131230776;
			
			// aapt resource value: 0x7F080039
			public const int action_bar_spinner = 2131230777;
			
			// aapt resource value: 0x7F08003A
			public const int action_bar_subtitle = 2131230778;
			
			// aapt resource value: 0x7F08003B
			public const int action_bar_title = 2131230779;
			
			// aapt resource value: 0x7F08003C
			public const int action_container = 2131230780;
			
			// aapt resource value: 0x7F08003D
			public const int action_context_bar = 2131230781;
			
			// aapt resource value: 0x7F08003E
			public const int action_divider = 2131230782;
			
			// aapt resource value: 0x7F08003F
			public const int action_image = 2131230783;
			
			// aapt resource value: 0x7F080040
			public const int action_menu_divider = 2131230784;
			
			// aapt resource value: 0x7F080041
			public const int action_menu_presenter = 2131230785;
			
			// aapt resource value: 0x7F080042
			public const int action_mode_bar = 2131230786;
			
			// aapt resource value: 0x7F080043
			public const int action_mode_bar_stub = 2131230787;
			
			// aapt resource value: 0x7F080044
			public const int action_mode_close_button = 2131230788;
			
			// aapt resource value: 0x7F080045
			public const int action_text = 2131230789;
			
			// aapt resource value: 0x7F080047
			public const int activity_chooser_view_content = 2131230791;
			
			// aapt resource value: 0x7F080048
			public const int add = 2131230792;
			
			// aapt resource value: 0x7F080049
			public const int alertTitle = 2131230793;
			
			// aapt resource value: 0x7F08004A
			public const int aligned = 2131230794;
			
			// aapt resource value: 0x7F08004B
			public const int all = 2131230795;
			
			// aapt resource value: 0x7F08004C
			public const int allStates = 2131230796;
			
			// aapt resource value: 0x7F080000
			public const int ALT = 2131230720;
			
			// aapt resource value: 0x7F08004D
			public const int always = 2131230797;
			
			// aapt resource value: 0x7F08004E
			public const int animateToEnd = 2131230798;
			
			// aapt resource value: 0x7F08004F
			public const int animateToStart = 2131230799;
			
			// aapt resource value: 0x7F080051
			public const int anticipate = 2131230801;
			
			// aapt resource value: 0x7F080050
			public const int antiClockwise = 2131230800;
			
			// aapt resource value: 0x7F080052
			public const int arc = 2131230802;
			
			// aapt resource value: 0x7F080053
			public const int asConfigured = 2131230803;
			
			// aapt resource value: 0x7F080054
			public const int async = 2131230804;
			
			// aapt resource value: 0x7F080055
			public const int auto = 2131230805;
			
			// aapt resource value: 0x7F080056
			public const int autoComplete = 2131230806;
			
			// aapt resource value: 0x7F080057
			public const int autoCompleteToEnd = 2131230807;
			
			// aapt resource value: 0x7F080058
			public const int autoCompleteToStart = 2131230808;
			
			// aapt resource value: 0x7F080059
			public const int barrier = 2131230809;
			
			// aapt resource value: 0x7F08005A
			public const int baseline = 2131230810;
			
			// aapt resource value: 0x7F08005C
			public const int beginning = 2131230812;
			
			// aapt resource value: 0x7F08005B
			public const int beginOnFirstDraw = 2131230811;
			
			// aapt resource value: 0x7F08005D
			public const int bestChoice = 2131230813;
			
			// aapt resource value: 0x7F08005E
			public const int blocking = 2131230814;
			
			// aapt resource value: 0x7F08005F
			public const int bottom = 2131230815;
			
			// aapt resource value: 0x7F080060
			public const int bottomtab_navarea = 2131230816;
			
			// aapt resource value: 0x7F080061
			public const int bottomtab_tabbar = 2131230817;
			
			// aapt resource value: 0x7F080001
			public const int BOTTOM_END = 2131230721;
			
			// aapt resource value: 0x7F080002
			public const int BOTTOM_START = 2131230722;
			
			// aapt resource value: 0x7F080062
			public const int bounce = 2131230818;
			
			// aapt resource value: 0x7F080063
			public const int bounceBoth = 2131230819;
			
			// aapt resource value: 0x7F080064
			public const int bounceEnd = 2131230820;
			
			// aapt resource value: 0x7F080065
			public const int bounceStart = 2131230821;
			
			// aapt resource value: 0x7F080066
			public const int browser_actions_header_text = 2131230822;
			
			// aapt resource value: 0x7F080069
			public const int browser_actions_menu_items = 2131230825;
			
			// aapt resource value: 0x7F080067
			public const int browser_actions_menu_item_icon = 2131230823;
			
			// aapt resource value: 0x7F080068
			public const int browser_actions_menu_item_text = 2131230824;
			
			// aapt resource value: 0x7F08006A
			public const int browser_actions_menu_view = 2131230826;
			
			// aapt resource value: 0x7F08006B
			public const int buttonPanel = 2131230827;
			
			// aapt resource value: 0x7F08006C
			public const int cache_measures = 2131230828;
			
			// aapt resource value: 0x7F08006D
			public const int cancel_action = 2131230829;
			
			// aapt resource value: 0x7F08006E
			public const int cancel_button = 2131230830;
			
			// aapt resource value: 0x7F08006F
			public const int carryVelocity = 2131230831;
			
			// aapt resource value: 0x7F080070
			public const int center = 2131230832;
			
			// aapt resource value: 0x7F080071
			public const int center_horizontal = 2131230833;
			
			// aapt resource value: 0x7F080072
			public const int center_vertical = 2131230834;
			
			// aapt resource value: 0x7F080073
			public const int chain = 2131230835;
			
			// aapt resource value: 0x7F080074
			public const int chains = 2131230836;
			
			// aapt resource value: 0x7F080075
			public const int checkbox = 2131230837;
			
			// aapt resource value: 0x7F080076
			public const int @checked = 2131230838;
			
			// aapt resource value: 0x7F080077
			public const int chip = 2131230839;
			
			// aapt resource value: 0x7F080078
			public const int chip1 = 2131230840;
			
			// aapt resource value: 0x7F080079
			public const int chip2 = 2131230841;
			
			// aapt resource value: 0x7F08007A
			public const int chip3 = 2131230842;
			
			// aapt resource value: 0x7F08007B
			public const int chip_group = 2131230843;
			
			// aapt resource value: 0x7F08007C
			public const int chronometer = 2131230844;
			
			// aapt resource value: 0x7F08007D
			public const int circle_center = 2131230845;
			
			// aapt resource value: 0x7F08007E
			public const int clear_text = 2131230846;
			
			// aapt resource value: 0x7F08007F
			public const int clip_horizontal = 2131230847;
			
			// aapt resource value: 0x7F080080
			public const int clip_vertical = 2131230848;
			
			// aapt resource value: 0x7F080081
			public const int clockwise = 2131230849;
			
			// aapt resource value: 0x7F080082
			public const int closest = 2131230850;
			
			// aapt resource value: 0x7F080083
			public const int collapseActionView = 2131230851;
			
			// aapt resource value: 0x7F080084
			public const int confirm_button = 2131230852;
			
			// aapt resource value: 0x7F080085
			public const int constraint = 2131230853;
			
			// aapt resource value: 0x7F080086
			public const int container = 2131230854;
			
			// aapt resource value: 0x7F080087
			public const int content = 2131230855;
			
			// aapt resource value: 0x7F080088
			public const int contentPanel = 2131230856;
			
			// aapt resource value: 0x7F080089
			public const int contiguous = 2131230857;
			
			// aapt resource value: 0x7F08008A
			public const int continuousVelocity = 2131230858;
			
			// aapt resource value: 0x7F08008B
			public const int coordinator = 2131230859;
			
			// aapt resource value: 0x7F08008C
			public const int cos = 2131230860;
			
			// aapt resource value: 0x7F08008D
			public const int counterclockwise = 2131230861;
			
			// aapt resource value: 0x7F080003
			public const int CTRL = 2131230723;
			
			// aapt resource value: 0x7F08008E
			public const int currentState = 2131230862;
			
			// aapt resource value: 0x7F08008F
			public const int custom = 2131230863;
			
			// aapt resource value: 0x7F080090
			public const int customPanel = 2131230864;
			
			// aapt resource value: 0x7F080091
			public const int cut = 2131230865;
			
			// aapt resource value: 0x7F080092
			public const int date_picker_actions = 2131230866;
			
			// aapt resource value: 0x7F080093
			public const int decelerate = 2131230867;
			
			// aapt resource value: 0x7F080094
			public const int decelerateAndComplete = 2131230868;
			
			// aapt resource value: 0x7F080095
			public const int decor_content_parent = 2131230869;
			
			// aapt resource value: 0x7F080096
			public const int default_activity_button = 2131230870;
			
			// aapt resource value: 0x7F080097
			public const int deltaRelative = 2131230871;
			
			// aapt resource value: 0x7F080098
			public const int dependency_ordering = 2131230872;
			
			// aapt resource value: 0x7F080099
			public const int design_bottom_sheet = 2131230873;
			
			// aapt resource value: 0x7F08009A
			public const int design_menu_item_action_area = 2131230874;
			
			// aapt resource value: 0x7F08009B
			public const int design_menu_item_action_area_stub = 2131230875;
			
			// aapt resource value: 0x7F08009C
			public const int design_menu_item_text = 2131230876;
			
			// aapt resource value: 0x7F08009D
			public const int design_navigation_view = 2131230877;
			
			// aapt resource value: 0x7F08009E
			public const int dialog_button = 2131230878;
			
			// aapt resource value: 0x7F08009F
			public const int dimensions = 2131230879;
			
			// aapt resource value: 0x7F0800A0
			public const int direct = 2131230880;
			
			// aapt resource value: 0x7F0800A1
			public const int disableHome = 2131230881;
			
			// aapt resource value: 0x7F0800A2
			public const int disableIntraAutoTransition = 2131230882;
			
			// aapt resource value: 0x7F0800A3
			public const int disablePostScroll = 2131230883;
			
			// aapt resource value: 0x7F0800A4
			public const int disableScroll = 2131230884;
			
			// aapt resource value: 0x7F0800A5
			public const int disjoint = 2131230885;
			
			// aapt resource value: 0x7F0800A6
			public const int dragAnticlockwise = 2131230886;
			
			// aapt resource value: 0x7F0800A7
			public const int dragClockwise = 2131230887;
			
			// aapt resource value: 0x7F0800A8
			public const int dragDown = 2131230888;
			
			// aapt resource value: 0x7F0800A9
			public const int dragEnd = 2131230889;
			
			// aapt resource value: 0x7F0800AA
			public const int dragLeft = 2131230890;
			
			// aapt resource value: 0x7F0800AB
			public const int dragRight = 2131230891;
			
			// aapt resource value: 0x7F0800AC
			public const int dragStart = 2131230892;
			
			// aapt resource value: 0x7F0800AD
			public const int dragUp = 2131230893;
			
			// aapt resource value: 0x7F0800AE
			public const int dropdown_menu = 2131230894;
			
			// aapt resource value: 0x7F0800AF
			public const int easeIn = 2131230895;
			
			// aapt resource value: 0x7F0800B0
			public const int easeInOut = 2131230896;
			
			// aapt resource value: 0x7F0800B1
			public const int easeOut = 2131230897;
			
			// aapt resource value: 0x7F0800B2
			public const int east = 2131230898;
			
			// aapt resource value: 0x7F0800B3
			public const int edit_query = 2131230899;
			
			// aapt resource value: 0x7F0800B4
			public const int elastic = 2131230900;
			
			// aapt resource value: 0x7F0800B5
			public const int end = 2131230901;
			
			// aapt resource value: 0x7F0800B6
			public const int endToStart = 2131230902;
			
			// aapt resource value: 0x7F0800B7
			public const int end_padder = 2131230903;
			
			// aapt resource value: 0x7F0800B8
			public const int enterAlways = 2131230904;
			
			// aapt resource value: 0x7F0800B9
			public const int enterAlwaysCollapsed = 2131230905;
			
			// aapt resource value: 0x7F0800BA
			public const int exitUntilCollapsed = 2131230906;
			
			// aapt resource value: 0x7F0800BC
			public const int expanded_menu = 2131230908;
			
			// aapt resource value: 0x7F0800BB
			public const int expand_activities_button = 2131230907;
			
			// aapt resource value: 0x7F0800BD
			public const int fade = 2131230909;
			
			// aapt resource value: 0x7F0800BE
			public const int fill = 2131230910;
			
			// aapt resource value: 0x7F0800C1
			public const int filled = 2131230913;
			
			// aapt resource value: 0x7F0800BF
			public const int fill_horizontal = 2131230911;
			
			// aapt resource value: 0x7F0800C0
			public const int fill_vertical = 2131230912;
			
			// aapt resource value: 0x7F0800C2
			public const int fitToContents = 2131230914;
			
			// aapt resource value: 0x7F0800C3
			public const int @fixed = 2131230915;
			
			// aapt resource value: 0x7F0800C4
			public const int flip = 2131230916;
			
			// aapt resource value: 0x7F0800C5
			public const int floating = 2131230917;
			
			// aapt resource value: 0x7F0800C6
			public const int flyoutcontent_appbar = 2131230918;
			
			// aapt resource value: 0x7F0800C7
			public const int forever = 2131230919;
			
			// aapt resource value: 0x7F0800C8
			public const int fragment_container_view_tag = 2131230920;
			
			// aapt resource value: 0x7F0800C9
			public const int frost = 2131230921;
			
			// aapt resource value: 0x7F080004
			public const int FUNCTION = 2131230724;
			
			// aapt resource value: 0x7F0800CA
			public const int ghost_view = 2131230922;
			
			// aapt resource value: 0x7F0800CB
			public const int ghost_view_holder = 2131230923;
			
			// aapt resource value: 0x7F0800CC
			public const int gone = 2131230924;
			
			// aapt resource value: 0x7F0800CD
			public const int graph = 2131230925;
			
			// aapt resource value: 0x7F0800CE
			public const int graph_wrap = 2131230926;
			
			// aapt resource value: 0x7F0800D0
			public const int grouping = 2131230928;
			
			// aapt resource value: 0x7F0800D1
			public const int groups = 2131230929;
			
			// aapt resource value: 0x7F0800CF
			public const int group_divider = 2131230927;
			
			// aapt resource value: 0x7F0800D2
			public const int guideline = 2131230930;
			
			// aapt resource value: 0x7F0800D3
			public const int hangupCallButton = 2131230931;
			
			// aapt resource value: 0x7F0800D4
			public const int header_title = 2131230932;
			
			// aapt resource value: 0x7F0800D5
			public const int hideable = 2131230933;
			
			// aapt resource value: 0x7F0800D6
			public const int home = 2131230934;
			
			// aapt resource value: 0x7F0800D7
			public const int homeAsUp = 2131230935;
			
			// aapt resource value: 0x7F0800D8
			public const int honorRequest = 2131230936;
			
			// aapt resource value: 0x7F0800D9
			public const int horizontal_only = 2131230937;
			
			// aapt resource value: 0x7F0800DA
			public const int icon = 2131230938;
			
			// aapt resource value: 0x7F0800DB
			public const int icon_frame = 2131230939;
			
			// aapt resource value: 0x7F0800DC
			public const int icon_group = 2131230940;
			
			// aapt resource value: 0x7F0800DD
			public const int ifRoom = 2131230941;
			
			// aapt resource value: 0x7F0800DE
			public const int ignore = 2131230942;
			
			// aapt resource value: 0x7F0800DF
			public const int ignoreRequest = 2131230943;
			
			// aapt resource value: 0x7F0800E0
			public const int image = 2131230944;
			
			// aapt resource value: 0x7F0800E1
			public const int immediateStop = 2131230945;
			
			// aapt resource value: 0x7F0800E2
			public const int included = 2131230946;
			
			// aapt resource value: 0x7F0800E3
			public const int incomingVideoView = 2131230947;
			
			// aapt resource value: 0x7F0800E4
			public const int info = 2131230948;
			
			// aapt resource value: 0x7F0800E5
			public const int invisible = 2131230949;
			
			// aapt resource value: 0x7F0800E6
			public const int inward = 2131230950;
			
			// aapt resource value: 0x7F0800E7
			public const int italic = 2131230951;
			
			// aapt resource value: 0x7F0800E8
			public const int item_touch_helper_previous_elevation = 2131230952;
			
			// aapt resource value: 0x7F0800E9
			public const int jumpToEnd = 2131230953;
			
			// aapt resource value: 0x7F0800EA
			public const int jumpToStart = 2131230954;
			
			// aapt resource value: 0x7F0800EB
			public const int labeled = 2131230955;
			
			// aapt resource value: 0x7F0800EC
			public const int layout = 2131230956;
			
			// aapt resource value: 0x7F0800ED
			public const int left = 2131230957;
			
			// aapt resource value: 0x7F0800EE
			public const int leftToRight = 2131230958;
			
			// aapt resource value: 0x7F0800EF
			public const int legacy = 2131230959;
			
			// aapt resource value: 0x7F0800F0
			public const int line1 = 2131230960;
			
			// aapt resource value: 0x7F0800F1
			public const int line3 = 2131230961;
			
			// aapt resource value: 0x7F0800F2
			public const int linear = 2131230962;
			
			// aapt resource value: 0x7F0800F3
			public const int listMode = 2131230963;
			
			// aapt resource value: 0x7F0800F4
			public const int list_item = 2131230964;
			
			// aapt resource value: 0x7F0800F5
			public const int main_appbar = 2131230965;
			
			// aapt resource value: 0x7F0800F6
			public const int main_tablayout = 2131230966;
			
			// aapt resource value: 0x7F0800F7
			public const int main_toolbar = 2131230967;
			
			// aapt resource value: 0x7F0800F8
			public const int main_viewpager = 2131230968;
			
			// aapt resource value: 0x7F0800F9
			public const int masked = 2131230969;
			
			// aapt resource value: 0x7F0800FA
			public const int match_constraint = 2131230970;
			
			// aapt resource value: 0x7F0800FB
			public const int match_parent = 2131230971;
			
			// aapt resource value: 0x7F0800FC
			public const int material_clock_display = 2131230972;
			
			// aapt resource value: 0x7F0800FD
			public const int material_clock_face = 2131230973;
			
			// aapt resource value: 0x7F0800FE
			public const int material_clock_hand = 2131230974;
			
			// aapt resource value: 0x7F0800FF
			public const int material_clock_period_am_button = 2131230975;
			
			// aapt resource value: 0x7F080100
			public const int material_clock_period_pm_button = 2131230976;
			
			// aapt resource value: 0x7F080101
			public const int material_clock_period_toggle = 2131230977;
			
			// aapt resource value: 0x7F080102
			public const int material_hour_text_input = 2131230978;
			
			// aapt resource value: 0x7F080103
			public const int material_hour_tv = 2131230979;
			
			// aapt resource value: 0x7F080104
			public const int material_label = 2131230980;
			
			// aapt resource value: 0x7F080105
			public const int material_minute_text_input = 2131230981;
			
			// aapt resource value: 0x7F080106
			public const int material_minute_tv = 2131230982;
			
			// aapt resource value: 0x7F080107
			public const int material_textinput_timepicker = 2131230983;
			
			// aapt resource value: 0x7F080108
			public const int material_timepicker_cancel_button = 2131230984;
			
			// aapt resource value: 0x7F080109
			public const int material_timepicker_container = 2131230985;
			
			// aapt resource value: 0x7F08010A
			public const int material_timepicker_edit_text = 2131230986;
			
			// aapt resource value: 0x7F08010B
			public const int material_timepicker_mode_button = 2131230987;
			
			// aapt resource value: 0x7F08010C
			public const int material_timepicker_ok_button = 2131230988;
			
			// aapt resource value: 0x7F08010D
			public const int material_timepicker_view = 2131230989;
			
			// aapt resource value: 0x7F08010E
			public const int material_value_index = 2131230990;
			
			// aapt resource value: 0x7F08010F
			public const int media_actions = 2131230991;
			
			// aapt resource value: 0x7F080110
			public const int media_controller_compat_view_tag = 2131230992;
			
			// aapt resource value: 0x7F080111
			public const int message = 2131230993;
			
			// aapt resource value: 0x7F080005
			public const int META = 2131230725;
			
			// aapt resource value: 0x7F080112
			public const int middle = 2131230994;
			
			// aapt resource value: 0x7F080113
			public const int mini = 2131230995;
			
			// aapt resource value: 0x7F080114
			public const int month_grid = 2131230996;
			
			// aapt resource value: 0x7F080115
			public const int month_navigation_bar = 2131230997;
			
			// aapt resource value: 0x7F080116
			public const int month_navigation_fragment_toggle = 2131230998;
			
			// aapt resource value: 0x7F080117
			public const int month_navigation_next = 2131230999;
			
			// aapt resource value: 0x7F080118
			public const int month_navigation_previous = 2131231000;
			
			// aapt resource value: 0x7F080119
			public const int month_title = 2131231001;
			
			// aapt resource value: 0x7F08011A
			public const int motion_base = 2131231002;
			
			// aapt resource value: 0x7F08011B
			public const int mtrl_anchor_parent = 2131231003;
			
			// aapt resource value: 0x7F08011D
			public const int mtrl_calendar_days_of_week = 2131231005;
			
			// aapt resource value: 0x7F08011C
			public const int mtrl_calendar_day_selector_frame = 2131231004;
			
			// aapt resource value: 0x7F08011E
			public const int mtrl_calendar_frame = 2131231006;
			
			// aapt resource value: 0x7F08011F
			public const int mtrl_calendar_main_pane = 2131231007;
			
			// aapt resource value: 0x7F080120
			public const int mtrl_calendar_months = 2131231008;
			
			// aapt resource value: 0x7F080121
			public const int mtrl_calendar_selection_frame = 2131231009;
			
			// aapt resource value: 0x7F080122
			public const int mtrl_calendar_text_input_frame = 2131231010;
			
			// aapt resource value: 0x7F080123
			public const int mtrl_calendar_year_selector_frame = 2131231011;
			
			// aapt resource value: 0x7F080124
			public const int mtrl_card_checked_layer_id = 2131231012;
			
			// aapt resource value: 0x7F080125
			public const int mtrl_child_content_container = 2131231013;
			
			// aapt resource value: 0x7F080126
			public const int mtrl_internal_children_alpha_tag = 2131231014;
			
			// aapt resource value: 0x7F080127
			public const int mtrl_motion_snapshot_view = 2131231015;
			
			// aapt resource value: 0x7F080128
			public const int mtrl_picker_fullscreen = 2131231016;
			
			// aapt resource value: 0x7F080129
			public const int mtrl_picker_header = 2131231017;
			
			// aapt resource value: 0x7F08012A
			public const int mtrl_picker_header_selection_text = 2131231018;
			
			// aapt resource value: 0x7F08012B
			public const int mtrl_picker_header_title_and_selection = 2131231019;
			
			// aapt resource value: 0x7F08012C
			public const int mtrl_picker_header_toggle = 2131231020;
			
			// aapt resource value: 0x7F08012D
			public const int mtrl_picker_text_input_date = 2131231021;
			
			// aapt resource value: 0x7F08012E
			public const int mtrl_picker_text_input_range_end = 2131231022;
			
			// aapt resource value: 0x7F08012F
			public const int mtrl_picker_text_input_range_start = 2131231023;
			
			// aapt resource value: 0x7F080130
			public const int mtrl_picker_title_text = 2131231024;
			
			// aapt resource value: 0x7F080131
			public const int mtrl_view_tag_bottom_padding = 2131231025;
			
			// aapt resource value: 0x7F080132
			public const int multiply = 2131231026;
			
			// aapt resource value: 0x7F080134
			public const int navigation_bar_item_icon_view = 2131231028;
			
			// aapt resource value: 0x7F080135
			public const int navigation_bar_item_labels_group = 2131231029;
			
			// aapt resource value: 0x7F080136
			public const int navigation_bar_item_large_label_view = 2131231030;
			
			// aapt resource value: 0x7F080137
			public const int navigation_bar_item_small_label_view = 2131231031;
			
			// aapt resource value: 0x7F080138
			public const int navigation_header_container = 2131231032;
			
			// aapt resource value: 0x7F080133
			public const int nav_controller_view_tag = 2131231027;
			
			// aapt resource value: 0x7F080139
			public const int never = 2131231033;
			
			// aapt resource value: 0x7F08013A
			public const int neverCompleteToEnd = 2131231034;
			
			// aapt resource value: 0x7F08013B
			public const int neverCompleteToStart = 2131231035;
			
			// aapt resource value: 0x7F08013E
			public const int none = 2131231038;
			
			// aapt resource value: 0x7F08013F
			public const int normal = 2131231039;
			
			// aapt resource value: 0x7F080140
			public const int north = 2131231040;
			
			// aapt resource value: 0x7F08013C
			public const int noScroll = 2131231036;
			
			// aapt resource value: 0x7F08013D
			public const int noState = 2131231037;
			
			// aapt resource value: 0x7F080141
			public const int notification_background = 2131231041;
			
			// aapt resource value: 0x7F080142
			public const int notification_main_column = 2131231042;
			
			// aapt resource value: 0x7F080143
			public const int notification_main_column_container = 2131231043;
			
			// aapt resource value: 0x7F080006
			public const int NO_DEBUG = 2131230726;
			
			// aapt resource value: 0x7F080144
			public const int off = 2131231044;
			
			// aapt resource value: 0x7F080145
			public const int on = 2131231045;
			
			// aapt resource value: 0x7F080146
			public const int outline = 2131231046;
			
			// aapt resource value: 0x7F080147
			public const int outward = 2131231047;
			
			// aapt resource value: 0x7F080148
			public const int overshoot = 2131231048;
			
			// aapt resource value: 0x7F080149
			public const int packed = 2131231049;
			
			// aapt resource value: 0x7F08014A
			public const int parallax = 2131231050;
			
			// aapt resource value: 0x7F08014B
			public const int parent = 2131231051;
			
			// aapt resource value: 0x7F08014C
			public const int parentPanel = 2131231052;
			
			// aapt resource value: 0x7F08014D
			public const int parentRelative = 2131231053;
			
			// aapt resource value: 0x7F08014E
			public const int parent_matrix = 2131231054;
			
			// aapt resource value: 0x7F08014F
			public const int password_toggle = 2131231055;
			
			// aapt resource value: 0x7F080150
			public const int path = 2131231056;
			
			// aapt resource value: 0x7F080151
			public const int pathRelative = 2131231057;
			
			// aapt resource value: 0x7F080152
			public const int peekHeight = 2131231058;
			
			// aapt resource value: 0x7F080153
			public const int peerTxt = 2131231059;
			
			// aapt resource value: 0x7F080154
			public const int percent = 2131231060;
			
			// aapt resource value: 0x7F080155
			public const int pin = 2131231061;
			
			// aapt resource value: 0x7F080156
			public const int position = 2131231062;
			
			// aapt resource value: 0x7F080157
			public const int postLayout = 2131231063;
			
			// aapt resource value: 0x7F080158
			public const int progress_circular = 2131231064;
			
			// aapt resource value: 0x7F080159
			public const int progress_horizontal = 2131231065;
			
			// aapt resource value: 0x7F08015A
			public const int radio = 2131231066;
			
			// aapt resource value: 0x7F08015B
			public const int ratio = 2131231067;
			
			// aapt resource value: 0x7F08015C
			public const int rectangles = 2131231068;
			
			// aapt resource value: 0x7F08015D
			public const int recycler_view = 2131231069;
			
			// aapt resource value: 0x7F08015E
			public const int reverseSawtooth = 2131231070;
			
			// aapt resource value: 0x7F08015F
			public const int right = 2131231071;
			
			// aapt resource value: 0x7F080160
			public const int rightToLeft = 2131231072;
			
			// aapt resource value: 0x7F080161
			public const int right_icon = 2131231073;
			
			// aapt resource value: 0x7F080162
			public const int right_side = 2131231074;
			
			// aapt resource value: 0x7F080163
			public const int rounded = 2131231075;
			
			// aapt resource value: 0x7F080164
			public const int row_index_key = 2131231076;
			
			// aapt resource value: 0x7F080165
			public const int save_non_transition_alpha = 2131231077;
			
			// aapt resource value: 0x7F080166
			public const int save_overlay_view = 2131231078;
			
			// aapt resource value: 0x7F080167
			public const int sawtooth = 2131231079;
			
			// aapt resource value: 0x7F080168
			public const int scale = 2131231080;
			
			// aapt resource value: 0x7F080169
			public const int screen = 2131231081;
			
			// aapt resource value: 0x7F08016A
			public const int scroll = 2131231082;
			
			// aapt resource value: 0x7F08016E
			public const int scrollable = 2131231086;
			
			// aapt resource value: 0x7F08016B
			public const int scrollIndicatorDown = 2131231083;
			
			// aapt resource value: 0x7F08016C
			public const int scrollIndicatorUp = 2131231084;
			
			// aapt resource value: 0x7F08016D
			public const int scrollView = 2131231085;
			
			// aapt resource value: 0x7F08016F
			public const int search_badge = 2131231087;
			
			// aapt resource value: 0x7F080170
			public const int search_bar = 2131231088;
			
			// aapt resource value: 0x7F080171
			public const int search_button = 2131231089;
			
			// aapt resource value: 0x7F080172
			public const int search_close_btn = 2131231090;
			
			// aapt resource value: 0x7F080173
			public const int search_edit_frame = 2131231091;
			
			// aapt resource value: 0x7F080174
			public const int search_go_btn = 2131231092;
			
			// aapt resource value: 0x7F080175
			public const int search_mag_icon = 2131231093;
			
			// aapt resource value: 0x7F080176
			public const int search_plate = 2131231094;
			
			// aapt resource value: 0x7F080177
			public const int search_src_text = 2131231095;
			
			// aapt resource value: 0x7F080178
			public const int search_voice_btn = 2131231096;
			
			// aapt resource value: 0x7F080179
			public const int seekbar = 2131231097;
			
			// aapt resource value: 0x7F08017A
			public const int seekbar_value = 2131231098;
			
			// aapt resource value: 0x7F08017C
			public const int selected = 2131231100;
			
			// aapt resource value: 0x7F08017D
			public const int selection_type = 2131231101;
			
			// aapt resource value: 0x7F08017B
			public const int select_dialog_listview = 2131231099;
			
			// aapt resource value: 0x7F08017E
			public const int sharedValueSet = 2131231102;
			
			// aapt resource value: 0x7F08017F
			public const int sharedValueUnset = 2131231103;
			
			// aapt resource value: 0x7F080180
			public const int shellcontent_appbar = 2131231104;
			
			// aapt resource value: 0x7F080181
			public const int shellcontent_toolbar = 2131231105;
			
			// aapt resource value: 0x7F080007
			public const int SHIFT = 2131230727;
			
			// aapt resource value: 0x7F080182
			public const int shortcut = 2131231106;
			
			// aapt resource value: 0x7F080183
			public const int showCustom = 2131231107;
			
			// aapt resource value: 0x7F080184
			public const int showHome = 2131231108;
			
			// aapt resource value: 0x7F080185
			public const int showTitle = 2131231109;
			
			// aapt resource value: 0x7F080008
			public const int SHOW_ALL = 2131230728;
			
			// aapt resource value: 0x7F080009
			public const int SHOW_PATH = 2131230729;
			
			// aapt resource value: 0x7F08000A
			public const int SHOW_PROGRESS = 2131230730;
			
			// aapt resource value: 0x7F080186
			public const int sin = 2131231110;
			
			// aapt resource value: 0x7F080187
			public const int skipCollapsed = 2131231111;
			
			// aapt resource value: 0x7F080188
			public const int skipped = 2131231112;
			
			// aapt resource value: 0x7F080189
			public const int slide = 2131231113;
			
			// aapt resource value: 0x7F08018A
			public const int sliding_tabs = 2131231114;
			
			// aapt resource value: 0x7F08018B
			public const int snackbar_action = 2131231115;
			
			// aapt resource value: 0x7F08018C
			public const int snackbar_text = 2131231116;
			
			// aapt resource value: 0x7F08018D
			public const int snap = 2131231117;
			
			// aapt resource value: 0x7F08018E
			public const int snapMargins = 2131231118;
			
			// aapt resource value: 0x7F08018F
			public const int south = 2131231119;
			
			// aapt resource value: 0x7F080190
			public const int spacer = 2131231120;
			
			// aapt resource value: 0x7F080191
			public const int special_effects_controller_view_tag = 2131231121;
			
			// aapt resource value: 0x7F080192
			public const int spinner = 2131231122;
			
			// aapt resource value: 0x7F080193
			public const int spline = 2131231123;
			
			// aapt resource value: 0x7F080194
			public const int split_action_bar = 2131231124;
			
			// aapt resource value: 0x7F080195
			public const int spread = 2131231125;
			
			// aapt resource value: 0x7F080196
			public const int spread_inside = 2131231126;
			
			// aapt resource value: 0x7F080197
			public const int spring = 2131231127;
			
			// aapt resource value: 0x7F080198
			public const int square = 2131231128;
			
			// aapt resource value: 0x7F080199
			public const int src_atop = 2131231129;
			
			// aapt resource value: 0x7F08019A
			public const int src_in = 2131231130;
			
			// aapt resource value: 0x7F08019B
			public const int src_over = 2131231131;
			
			// aapt resource value: 0x7F08019C
			public const int standard = 2131231132;
			
			// aapt resource value: 0x7F08019D
			public const int start = 2131231133;
			
			// aapt resource value: 0x7F08019E
			public const int startHorizontal = 2131231134;
			
			// aapt resource value: 0x7F08019F
			public const int startToEnd = 2131231135;
			
			// aapt resource value: 0x7F0801A0
			public const int startVertical = 2131231136;
			
			// aapt resource value: 0x7F0801A1
			public const int staticLayout = 2131231137;
			
			// aapt resource value: 0x7F0801A2
			public const int staticPostLayout = 2131231138;
			
			// aapt resource value: 0x7F0801A3
			public const int statusTxt = 2131231139;
			
			// aapt resource value: 0x7F0801A4
			public const int status_bar_latest_event_content = 2131231140;
			
			// aapt resource value: 0x7F0801A5
			public const int stop = 2131231141;
			
			// aapt resource value: 0x7F0801A6
			public const int stretch = 2131231142;
			
			// aapt resource value: 0x7F0801A7
			public const int submenuarrow = 2131231143;
			
			// aapt resource value: 0x7F0801A8
			public const int submit_area = 2131231144;
			
			// aapt resource value: 0x7F0801A9
			public const int supportScrollUp = 2131231145;
			
			// aapt resource value: 0x7F0801AA
			public const int switchWidget = 2131231146;
			
			// aapt resource value: 0x7F08000B
			public const int SYM = 2131230731;
			
			// aapt resource value: 0x7F0801AB
			public const int tabMode = 2131231147;
			
			// aapt resource value: 0x7F0801AC
			public const int tag_accessibility_actions = 2131231148;
			
			// aapt resource value: 0x7F0801AD
			public const int tag_accessibility_clickable_spans = 2131231149;
			
			// aapt resource value: 0x7F0801AE
			public const int tag_accessibility_heading = 2131231150;
			
			// aapt resource value: 0x7F0801AF
			public const int tag_accessibility_pane_title = 2131231151;
			
			// aapt resource value: 0x7F0801B0
			public const int tag_on_apply_window_listener = 2131231152;
			
			// aapt resource value: 0x7F0801B1
			public const int tag_on_receive_content_listener = 2131231153;
			
			// aapt resource value: 0x7F0801B2
			public const int tag_on_receive_content_mime_types = 2131231154;
			
			// aapt resource value: 0x7F0801B3
			public const int tag_screen_reader_focusable = 2131231155;
			
			// aapt resource value: 0x7F0801B4
			public const int tag_state_description = 2131231156;
			
			// aapt resource value: 0x7F0801B5
			public const int tag_transition_group = 2131231157;
			
			// aapt resource value: 0x7F0801B6
			public const int tag_unhandled_key_event_manager = 2131231158;
			
			// aapt resource value: 0x7F0801B7
			public const int tag_unhandled_key_listeners = 2131231159;
			
			// aapt resource value: 0x7F0801B8
			public const int tag_window_insets_animation_callback = 2131231160;
			
			// aapt resource value: 0x7F0801B9
			public const int test_checkbox_android_button_tint = 2131231161;
			
			// aapt resource value: 0x7F0801BA
			public const int test_checkbox_app_button_tint = 2131231162;
			
			// aapt resource value: 0x7F0801BB
			public const int test_radiobutton_android_button_tint = 2131231163;
			
			// aapt resource value: 0x7F0801BC
			public const int test_radiobutton_app_button_tint = 2131231164;
			
			// aapt resource value: 0x7F0801BD
			public const int text = 2131231165;
			
			// aapt resource value: 0x7F0801BE
			public const int text2 = 2131231166;
			
			// aapt resource value: 0x7F0801BF
			public const int textEnd = 2131231167;
			
			// aapt resource value: 0x7F0801C7
			public const int textinput_counter = 2131231175;
			
			// aapt resource value: 0x7F0801C8
			public const int textinput_error = 2131231176;
			
			// aapt resource value: 0x7F0801C9
			public const int textinput_helper_text = 2131231177;
			
			// aapt resource value: 0x7F0801CA
			public const int textinput_placeholder = 2131231178;
			
			// aapt resource value: 0x7F0801CB
			public const int textinput_prefix_text = 2131231179;
			
			// aapt resource value: 0x7F0801CC
			public const int textinput_suffix_text = 2131231180;
			
			// aapt resource value: 0x7F0801C0
			public const int textSpacerNoButtons = 2131231168;
			
			// aapt resource value: 0x7F0801C1
			public const int textSpacerNoTitle = 2131231169;
			
			// aapt resource value: 0x7F0801C2
			public const int textStart = 2131231170;
			
			// aapt resource value: 0x7F0801C3
			public const int textTop = 2131231171;
			
			// aapt resource value: 0x7F0801C4
			public const int text_input_end_icon = 2131231172;
			
			// aapt resource value: 0x7F0801C5
			public const int text_input_error_icon = 2131231173;
			
			// aapt resource value: 0x7F0801C6
			public const int text_input_start_icon = 2131231174;
			
			// aapt resource value: 0x7F0801CD
			public const int time = 2131231181;
			
			// aapt resource value: 0x7F0801CE
			public const int title = 2131231182;
			
			// aapt resource value: 0x7F0801CF
			public const int titleDividerNoCustom = 2131231183;
			
			// aapt resource value: 0x7F0801D0
			public const int title_template = 2131231184;
			
			// aapt resource value: 0x7F0801D1
			public const int toggle = 2131231185;
			
			// aapt resource value: 0x7F0801D2
			public const int toolbar = 2131231186;
			
			// aapt resource value: 0x7F0801D3
			public const int top = 2131231187;
			
			// aapt resource value: 0x7F0801D4
			public const int topPanel = 2131231188;
			
			// aapt resource value: 0x7F08000C
			public const int TOP_END = 2131230732;
			
			// aapt resource value: 0x7F08000D
			public const int TOP_START = 2131230733;
			
			// aapt resource value: 0x7F0801D5
			public const int touch_outside = 2131231189;
			
			// aapt resource value: 0x7F0801D6
			public const int transitionToEnd = 2131231190;
			
			// aapt resource value: 0x7F0801D7
			public const int transitionToStart = 2131231191;
			
			// aapt resource value: 0x7F0801D8
			public const int transition_current_scene = 2131231192;
			
			// aapt resource value: 0x7F0801D9
			public const int transition_layout_save = 2131231193;
			
			// aapt resource value: 0x7F0801DA
			public const int transition_position = 2131231194;
			
			// aapt resource value: 0x7F0801DB
			public const int transition_scene_layoutid_cache = 2131231195;
			
			// aapt resource value: 0x7F0801DC
			public const int transition_transform = 2131231196;
			
			// aapt resource value: 0x7F0801DD
			public const int triangle = 2131231197;
			
			// aapt resource value: 0x7F0801DE
			public const int @unchecked = 2131231198;
			
			// aapt resource value: 0x7F0801DF
			public const int uniform = 2131231199;
			
			// aapt resource value: 0x7F0801E0
			public const int unlabeled = 2131231200;
			
			// aapt resource value: 0x7F0801E1
			public const int up = 2131231201;
			
			// aapt resource value: 0x7F0801E2
			public const int useLogo = 2131231202;
			
			// aapt resource value: 0x7F0801E3
			public const int vertical_only = 2131231203;
			
			// aapt resource value: 0x7F0801E4
			public const int view_offset_helper = 2131231204;
			
			// aapt resource value: 0x7F0801E5
			public const int view_transition = 2131231205;
			
			// aapt resource value: 0x7F0801E6
			public const int view_tree_lifecycle_owner = 2131231206;
			
			// aapt resource value: 0x7F0801E7
			public const int view_tree_saved_state_registry_owner = 2131231207;
			
			// aapt resource value: 0x7F0801E8
			public const int view_tree_view_model_store_owner = 2131231208;
			
			// aapt resource value: 0x7F0801E9
			public const int visible = 2131231209;
			
			// aapt resource value: 0x7F0801EA
			public const int visible_removing_fragment_view_tag = 2131231210;
			
			// aapt resource value: 0x7F0801EB
			public const int west = 2131231211;
			
			// aapt resource value: 0x7F0801ED
			public const int withinBounds = 2131231213;
			
			// aapt resource value: 0x7F0801EC
			public const int withText = 2131231212;
			
			// aapt resource value: 0x7F0801EE
			public const int wrap = 2131231214;
			
			// aapt resource value: 0x7F0801EF
			public const int wrap_content = 2131231215;
			
			// aapt resource value: 0x7F0801F0
			public const int wrap_content_constrained = 2131231216;
			
			// aapt resource value: 0x7F0801F1
			public const int x_left = 2131231217;
			
			// aapt resource value: 0x7F0801F2
			public const int x_right = 2131231218;
			
			// aapt resource value: 0x7F0801F3
			public const int zero_corner_chip = 2131231219;
			
			static Id()
			{
				global::Android.Runtime.ResourceIdManager.UpdateIdValues();
			}
			
			private Id()
			{
			}
		}
		
		public partial class Integer
		{
			
			// aapt resource value: 0x7F090000
			public const int abc_config_activityDefaultDur = 2131296256;
			
			// aapt resource value: 0x7F090001
			public const int abc_config_activityShortDur = 2131296257;
			
			// aapt resource value: 0x7F090002
			public const int app_bar_elevation_anim_duration = 2131296258;
			
			// aapt resource value: 0x7F090003
			public const int bottom_sheet_slide_duration = 2131296259;
			
			// aapt resource value: 0x7F090004
			public const int cancel_button_image_alpha = 2131296260;
			
			// aapt resource value: 0x7F090005
			public const int config_navAnimTime = 2131296261;
			
			// aapt resource value: 0x7F090006
			public const int config_tooltipAnimTime = 2131296262;
			
			// aapt resource value: 0x7F090007
			public const int design_snackbar_text_max_lines = 2131296263;
			
			// aapt resource value: 0x7F090008
			public const int design_tab_indicator_anim_duration_ms = 2131296264;
			
			// aapt resource value: 0x7F090009
			public const int hide_password_duration = 2131296265;
			
			// aapt resource value: 0x7F09000A
			public const int material_motion_duration_long_1 = 2131296266;
			
			// aapt resource value: 0x7F09000B
			public const int material_motion_duration_long_2 = 2131296267;
			
			// aapt resource value: 0x7F09000C
			public const int material_motion_duration_medium_1 = 2131296268;
			
			// aapt resource value: 0x7F09000D
			public const int material_motion_duration_medium_2 = 2131296269;
			
			// aapt resource value: 0x7F09000E
			public const int material_motion_duration_short_1 = 2131296270;
			
			// aapt resource value: 0x7F09000F
			public const int material_motion_duration_short_2 = 2131296271;
			
			// aapt resource value: 0x7F090010
			public const int material_motion_path = 2131296272;
			
			// aapt resource value: 0x7F090011
			public const int mtrl_badge_max_character_count = 2131296273;
			
			// aapt resource value: 0x7F090012
			public const int mtrl_btn_anim_delay_ms = 2131296274;
			
			// aapt resource value: 0x7F090013
			public const int mtrl_btn_anim_duration_ms = 2131296275;
			
			// aapt resource value: 0x7F090014
			public const int mtrl_calendar_header_orientation = 2131296276;
			
			// aapt resource value: 0x7F090015
			public const int mtrl_calendar_selection_text_lines = 2131296277;
			
			// aapt resource value: 0x7F090016
			public const int mtrl_calendar_year_selector_span = 2131296278;
			
			// aapt resource value: 0x7F090017
			public const int mtrl_card_anim_delay_ms = 2131296279;
			
			// aapt resource value: 0x7F090018
			public const int mtrl_card_anim_duration_ms = 2131296280;
			
			// aapt resource value: 0x7F090019
			public const int mtrl_chip_anim_duration = 2131296281;
			
			// aapt resource value: 0x7F09001A
			public const int mtrl_tab_indicator_anim_duration_ms = 2131296282;
			
			// aapt resource value: 0x7F09001B
			public const int show_password_duration = 2131296283;
			
			// aapt resource value: 0x7F09001C
			public const int status_bar_notification_info_maxnum = 2131296284;
			
			static Integer()
			{
				global::Android.Runtime.ResourceIdManager.UpdateIdValues();
			}
			
			private Integer()
			{
			}
		}
		
		public partial class Interpolator
		{
			
			// aapt resource value: 0x7F0A0000
			public const int btn_checkbox_checked_mtrl_animation_interpolator_0 = 2131361792;
			
			// aapt resource value: 0x7F0A0001
			public const int btn_checkbox_checked_mtrl_animation_interpolator_1 = 2131361793;
			
			// aapt resource value: 0x7F0A0002
			public const int btn_checkbox_unchecked_mtrl_animation_interpolator_0 = 2131361794;
			
			// aapt resource value: 0x7F0A0003
			public const int btn_checkbox_unchecked_mtrl_animation_interpolator_1 = 2131361795;
			
			// aapt resource value: 0x7F0A0004
			public const int btn_radio_to_off_mtrl_animation_interpolator_0 = 2131361796;
			
			// aapt resource value: 0x7F0A0005
			public const int btn_radio_to_on_mtrl_animation_interpolator_0 = 2131361797;
			
			// aapt resource value: 0x7F0A0006
			public const int fast_out_slow_in = 2131361798;
			
			// aapt resource value: 0x7F0A0007
			public const int mtrl_fast_out_linear_in = 2131361799;
			
			// aapt resource value: 0x7F0A0008
			public const int mtrl_fast_out_slow_in = 2131361800;
			
			// aapt resource value: 0x7F0A0009
			public const int mtrl_linear = 2131361801;
			
			// aapt resource value: 0x7F0A000A
			public const int mtrl_linear_out_slow_in = 2131361802;
			
			static Interpolator()
			{
				global::Android.Runtime.ResourceIdManager.UpdateIdValues();
			}
			
			private Interpolator()
			{
			}
		}
		
		public partial class Layout
		{
			
			// aapt resource value: 0x7F0B0000
			public const int abc_action_bar_title_item = 2131427328;
			
			// aapt resource value: 0x7F0B0001
			public const int abc_action_bar_up_container = 2131427329;
			
			// aapt resource value: 0x7F0B0002
			public const int abc_action_menu_item_layout = 2131427330;
			
			// aapt resource value: 0x7F0B0003
			public const int abc_action_menu_layout = 2131427331;
			
			// aapt resource value: 0x7F0B0004
			public const int abc_action_mode_bar = 2131427332;
			
			// aapt resource value: 0x7F0B0005
			public const int abc_action_mode_close_item_material = 2131427333;
			
			// aapt resource value: 0x7F0B0006
			public const int abc_activity_chooser_view = 2131427334;
			
			// aapt resource value: 0x7F0B0007
			public const int abc_activity_chooser_view_list_item = 2131427335;
			
			// aapt resource value: 0x7F0B0008
			public const int abc_alert_dialog_button_bar_material = 2131427336;
			
			// aapt resource value: 0x7F0B0009
			public const int abc_alert_dialog_material = 2131427337;
			
			// aapt resource value: 0x7F0B000A
			public const int abc_alert_dialog_title_material = 2131427338;
			
			// aapt resource value: 0x7F0B000B
			public const int abc_cascading_menu_item_layout = 2131427339;
			
			// aapt resource value: 0x7F0B000C
			public const int abc_dialog_title_material = 2131427340;
			
			// aapt resource value: 0x7F0B000D
			public const int abc_expanded_menu_layout = 2131427341;
			
			// aapt resource value: 0x7F0B000E
			public const int abc_list_menu_item_checkbox = 2131427342;
			
			// aapt resource value: 0x7F0B000F
			public const int abc_list_menu_item_icon = 2131427343;
			
			// aapt resource value: 0x7F0B0010
			public const int abc_list_menu_item_layout = 2131427344;
			
			// aapt resource value: 0x7F0B0011
			public const int abc_list_menu_item_radio = 2131427345;
			
			// aapt resource value: 0x7F0B0012
			public const int abc_popup_menu_header_item_layout = 2131427346;
			
			// aapt resource value: 0x7F0B0013
			public const int abc_popup_menu_item_layout = 2131427347;
			
			// aapt resource value: 0x7F0B0014
			public const int abc_screen_content_include = 2131427348;
			
			// aapt resource value: 0x7F0B0015
			public const int abc_screen_simple = 2131427349;
			
			// aapt resource value: 0x7F0B0016
			public const int abc_screen_simple_overlay_action_mode = 2131427350;
			
			// aapt resource value: 0x7F0B0017
			public const int abc_screen_toolbar = 2131427351;
			
			// aapt resource value: 0x7F0B0018
			public const int abc_search_dropdown_item_icons_2line = 2131427352;
			
			// aapt resource value: 0x7F0B0019
			public const int abc_search_view = 2131427353;
			
			// aapt resource value: 0x7F0B001A
			public const int abc_select_dialog_material = 2131427354;
			
			// aapt resource value: 0x7F0B001B
			public const int abc_tooltip = 2131427355;
			
			// aapt resource value: 0x7F0B001C
			public const int activity_call = 2131427356;
			
			// aapt resource value: 0x7F0B001D
			public const int BottomTabLayout = 2131427357;
			
			// aapt resource value: 0x7F0B001E
			public const int browser_actions_context_menu_page = 2131427358;
			
			// aapt resource value: 0x7F0B001F
			public const int browser_actions_context_menu_row = 2131427359;
			
			// aapt resource value: 0x7F0B0020
			public const int custom_dialog = 2131427360;
			
			// aapt resource value: 0x7F0B0021
			public const int design_bottom_navigation_item = 2131427361;
			
			// aapt resource value: 0x7F0B0022
			public const int design_bottom_sheet_dialog = 2131427362;
			
			// aapt resource value: 0x7F0B0023
			public const int design_layout_snackbar = 2131427363;
			
			// aapt resource value: 0x7F0B0024
			public const int design_layout_snackbar_include = 2131427364;
			
			// aapt resource value: 0x7F0B0025
			public const int design_layout_tab_icon = 2131427365;
			
			// aapt resource value: 0x7F0B0026
			public const int design_layout_tab_text = 2131427366;
			
			// aapt resource value: 0x7F0B0027
			public const int design_menu_item_action_area = 2131427367;
			
			// aapt resource value: 0x7F0B0028
			public const int design_navigation_item = 2131427368;
			
			// aapt resource value: 0x7F0B0029
			public const int design_navigation_item_header = 2131427369;
			
			// aapt resource value: 0x7F0B002A
			public const int design_navigation_item_separator = 2131427370;
			
			// aapt resource value: 0x7F0B002B
			public const int design_navigation_item_subheader = 2131427371;
			
			// aapt resource value: 0x7F0B002C
			public const int design_navigation_menu = 2131427372;
			
			// aapt resource value: 0x7F0B002D
			public const int design_navigation_menu_item = 2131427373;
			
			// aapt resource value: 0x7F0B002E
			public const int design_text_input_end_icon = 2131427374;
			
			// aapt resource value: 0x7F0B002F
			public const int design_text_input_start_icon = 2131427375;
			
			// aapt resource value: 0x7F0B0030
			public const int expand_button = 2131427376;
			
			// aapt resource value: 0x7F0B0031
			public const int FallbackTabbarDoNotUse = 2131427377;
			
			// aapt resource value: 0x7F0B0032
			public const int FallbackToolbarDoNotUse = 2131427378;
			
			// aapt resource value: 0x7F0B0033
			public const int FlyoutContent = 2131427379;
			
			// aapt resource value: 0x7F0B0034
			public const int image_frame = 2131427380;
			
			// aapt resource value: 0x7F0B0035
			public const int material_chip_input_combo = 2131427381;
			
			// aapt resource value: 0x7F0B003A
			public const int material_clockface_textview = 2131427386;
			
			// aapt resource value: 0x7F0B003B
			public const int material_clockface_view = 2131427387;
			
			// aapt resource value: 0x7F0B0036
			public const int material_clock_display = 2131427382;
			
			// aapt resource value: 0x7F0B0037
			public const int material_clock_display_divider = 2131427383;
			
			// aapt resource value: 0x7F0B0038
			public const int material_clock_period_toggle = 2131427384;
			
			// aapt resource value: 0x7F0B0039
			public const int material_clock_period_toggle_land = 2131427385;
			
			// aapt resource value: 0x7F0B003C
			public const int material_radial_view_group = 2131427388;
			
			// aapt resource value: 0x7F0B003D
			public const int material_textinput_timepicker = 2131427389;
			
			// aapt resource value: 0x7F0B0040
			public const int material_timepicker = 2131427392;
			
			// aapt resource value: 0x7F0B0041
			public const int material_timepicker_dialog = 2131427393;
			
			// aapt resource value: 0x7F0B0042
			public const int material_timepicker_textinput_display = 2131427394;
			
			// aapt resource value: 0x7F0B003E
			public const int material_time_chip = 2131427390;
			
			// aapt resource value: 0x7F0B003F
			public const int material_time_input = 2131427391;
			
			// aapt resource value: 0x7F0B0043
			public const int mtrl_alert_dialog = 2131427395;
			
			// aapt resource value: 0x7F0B0044
			public const int mtrl_alert_dialog_actions = 2131427396;
			
			// aapt resource value: 0x7F0B0045
			public const int mtrl_alert_dialog_title = 2131427397;
			
			// aapt resource value: 0x7F0B0046
			public const int mtrl_alert_select_dialog_item = 2131427398;
			
			// aapt resource value: 0x7F0B0047
			public const int mtrl_alert_select_dialog_multichoice = 2131427399;
			
			// aapt resource value: 0x7F0B0048
			public const int mtrl_alert_select_dialog_singlechoice = 2131427400;
			
			// aapt resource value: 0x7F0B0049
			public const int mtrl_calendar_day = 2131427401;
			
			// aapt resource value: 0x7F0B004B
			public const int mtrl_calendar_days_of_week = 2131427403;
			
			// aapt resource value: 0x7F0B004A
			public const int mtrl_calendar_day_of_week = 2131427402;
			
			// aapt resource value: 0x7F0B004C
			public const int mtrl_calendar_horizontal = 2131427404;
			
			// aapt resource value: 0x7F0B004D
			public const int mtrl_calendar_month = 2131427405;
			
			// aapt resource value: 0x7F0B0050
			public const int mtrl_calendar_months = 2131427408;
			
			// aapt resource value: 0x7F0B004E
			public const int mtrl_calendar_month_labeled = 2131427406;
			
			// aapt resource value: 0x7F0B004F
			public const int mtrl_calendar_month_navigation = 2131427407;
			
			// aapt resource value: 0x7F0B0051
			public const int mtrl_calendar_vertical = 2131427409;
			
			// aapt resource value: 0x7F0B0052
			public const int mtrl_calendar_year = 2131427410;
			
			// aapt resource value: 0x7F0B0053
			public const int mtrl_layout_snackbar = 2131427411;
			
			// aapt resource value: 0x7F0B0054
			public const int mtrl_layout_snackbar_include = 2131427412;
			
			// aapt resource value: 0x7F0B0055
			public const int mtrl_navigation_rail_item = 2131427413;
			
			// aapt resource value: 0x7F0B0056
			public const int mtrl_picker_actions = 2131427414;
			
			// aapt resource value: 0x7F0B0057
			public const int mtrl_picker_dialog = 2131427415;
			
			// aapt resource value: 0x7F0B0058
			public const int mtrl_picker_fullscreen = 2131427416;
			
			// aapt resource value: 0x7F0B0059
			public const int mtrl_picker_header_dialog = 2131427417;
			
			// aapt resource value: 0x7F0B005A
			public const int mtrl_picker_header_fullscreen = 2131427418;
			
			// aapt resource value: 0x7F0B005B
			public const int mtrl_picker_header_selection_text = 2131427419;
			
			// aapt resource value: 0x7F0B005C
			public const int mtrl_picker_header_title_text = 2131427420;
			
			// aapt resource value: 0x7F0B005D
			public const int mtrl_picker_header_toggle = 2131427421;
			
			// aapt resource value: 0x7F0B005E
			public const int mtrl_picker_text_input_date = 2131427422;
			
			// aapt resource value: 0x7F0B005F
			public const int mtrl_picker_text_input_date_range = 2131427423;
			
			// aapt resource value: 0x7F0B0060
			public const int notification_action = 2131427424;
			
			// aapt resource value: 0x7F0B0061
			public const int notification_action_tombstone = 2131427425;
			
			// aapt resource value: 0x7F0B0062
			public const int notification_media_action = 2131427426;
			
			// aapt resource value: 0x7F0B0063
			public const int notification_media_cancel_action = 2131427427;
			
			// aapt resource value: 0x7F0B0064
			public const int notification_template_big_media = 2131427428;
			
			// aapt resource value: 0x7F0B0065
			public const int notification_template_big_media_custom = 2131427429;
			
			// aapt resource value: 0x7F0B0066
			public const int notification_template_big_media_narrow = 2131427430;
			
			// aapt resource value: 0x7F0B0067
			public const int notification_template_big_media_narrow_custom = 2131427431;
			
			// aapt resource value: 0x7F0B0068
			public const int notification_template_custom_big = 2131427432;
			
			// aapt resource value: 0x7F0B0069
			public const int notification_template_icon_group = 2131427433;
			
			// aapt resource value: 0x7F0B006A
			public const int notification_template_lines_media = 2131427434;
			
			// aapt resource value: 0x7F0B006B
			public const int notification_template_media = 2131427435;
			
			// aapt resource value: 0x7F0B006C
			public const int notification_template_media_custom = 2131427436;
			
			// aapt resource value: 0x7F0B006D
			public const int notification_template_part_chronometer = 2131427437;
			
			// aapt resource value: 0x7F0B006E
			public const int notification_template_part_time = 2131427438;
			
			// aapt resource value: 0x7F0B006F
			public const int preference = 2131427439;
			
			// aapt resource value: 0x7F0B0070
			public const int preference_category = 2131427440;
			
			// aapt resource value: 0x7F0B0071
			public const int preference_category_material = 2131427441;
			
			// aapt resource value: 0x7F0B0072
			public const int preference_dialog_edittext = 2131427442;
			
			// aapt resource value: 0x7F0B0073
			public const int preference_dropdown = 2131427443;
			
			// aapt resource value: 0x7F0B0074
			public const int preference_dropdown_material = 2131427444;
			
			// aapt resource value: 0x7F0B0075
			public const int preference_information = 2131427445;
			
			// aapt resource value: 0x7F0B0076
			public const int preference_information_material = 2131427446;
			
			// aapt resource value: 0x7F0B0077
			public const int preference_list_fragment = 2131427447;
			
			// aapt resource value: 0x7F0B0078
			public const int preference_material = 2131427448;
			
			// aapt resource value: 0x7F0B0079
			public const int preference_recyclerview = 2131427449;
			
			// aapt resource value: 0x7F0B007A
			public const int preference_widget_checkbox = 2131427450;
			
			// aapt resource value: 0x7F0B007B
			public const int preference_widget_seekbar = 2131427451;
			
			// aapt resource value: 0x7F0B007C
			public const int preference_widget_seekbar_material = 2131427452;
			
			// aapt resource value: 0x7F0B007D
			public const int preference_widget_switch = 2131427453;
			
			// aapt resource value: 0x7F0B007E
			public const int preference_widget_switch_compat = 2131427454;
			
			// aapt resource value: 0x7F0B007F
			public const int RootLayout = 2131427455;
			
			// aapt resource value: 0x7F0B0080
			public const int select_dialog_item_material = 2131427456;
			
			// aapt resource value: 0x7F0B0081
			public const int select_dialog_multichoice_material = 2131427457;
			
			// aapt resource value: 0x7F0B0082
			public const int select_dialog_singlechoice_material = 2131427458;
			
			// aapt resource value: 0x7F0B0083
			public const int ShellContent = 2131427459;
			
			// aapt resource value: 0x7F0B0084
			public const int support_simple_spinner_dropdown_item = 2131427460;
			
			// aapt resource value: 0x7F0B0085
			public const int Tabbar = 2131427461;
			
			// aapt resource value: 0x7F0B0086
			public const int test_action_chip = 2131427462;
			
			// aapt resource value: 0x7F0B0087
			public const int test_chip_zero_corner_radius = 2131427463;
			
			// aapt resource value: 0x7F0B0088
			public const int test_design_checkbox = 2131427464;
			
			// aapt resource value: 0x7F0B0089
			public const int test_design_radiobutton = 2131427465;
			
			// aapt resource value: 0x7F0B008A
			public const int test_navigation_bar_item_layout = 2131427466;
			
			// aapt resource value: 0x7F0B008B
			public const int test_reflow_chipgroup = 2131427467;
			
			// aapt resource value: 0x7F0B008C
			public const int test_toolbar = 2131427468;
			
			// aapt resource value: 0x7F0B008D
			public const int test_toolbar_custom_background = 2131427469;
			
			// aapt resource value: 0x7F0B008E
			public const int test_toolbar_elevation = 2131427470;
			
			// aapt resource value: 0x7F0B008F
			public const int test_toolbar_surface = 2131427471;
			
			// aapt resource value: 0x7F0B0094
			public const int text_view_without_line_height = 2131427476;
			
			// aapt resource value: 0x7F0B0090
			public const int text_view_with_line_height_from_appearance = 2131427472;
			
			// aapt resource value: 0x7F0B0091
			public const int text_view_with_line_height_from_layout = 2131427473;
			
			// aapt resource value: 0x7F0B0092
			public const int text_view_with_line_height_from_style = 2131427474;
			
			// aapt resource value: 0x7F0B0093
			public const int text_view_with_theme_line_height = 2131427475;
			
			// aapt resource value: 0x7F0B0095
			public const int Toolbar = 2131427477;
			
			static Layout()
			{
				global::Android.Runtime.ResourceIdManager.UpdateIdValues();
			}
			
			private Layout()
			{
			}
		}
		
		public partial class Mipmap
		{
			
			// aapt resource value: 0x7F0C0000
			public const int icon = 2131492864;
			
			// aapt resource value: 0x7F0C0001
			public const int icon_round = 2131492865;
			
			// aapt resource value: 0x7F0C0002
			public const int launcher_foreground = 2131492866;
			
			static Mipmap()
			{
				global::Android.Runtime.ResourceIdManager.UpdateIdValues();
			}
			
			private Mipmap()
			{
			}
		}
		
		public partial class Plurals
		{
			
			// aapt resource value: 0x7F0D0000
			public const int mtrl_badge_content_description = 2131558400;
			
			static Plurals()
			{
				global::Android.Runtime.ResourceIdManager.UpdateIdValues();
			}
			
			private Plurals()
			{
			}
		}
		
		public partial class String
		{
			
			// aapt resource value: 0x7F0E0000
			public const int abc_action_bar_home_description = 2131623936;
			
			// aapt resource value: 0x7F0E0001
			public const int abc_action_bar_up_description = 2131623937;
			
			// aapt resource value: 0x7F0E0002
			public const int abc_action_menu_overflow_description = 2131623938;
			
			// aapt resource value: 0x7F0E0003
			public const int abc_action_mode_done = 2131623939;
			
			// aapt resource value: 0x7F0E0005
			public const int abc_activitychooserview_choose_application = 2131623941;
			
			// aapt resource value: 0x7F0E0004
			public const int abc_activity_chooser_view_see_all = 2131623940;
			
			// aapt resource value: 0x7F0E0006
			public const int abc_capital_off = 2131623942;
			
			// aapt resource value: 0x7F0E0007
			public const int abc_capital_on = 2131623943;
			
			// aapt resource value: 0x7F0E0008
			public const int abc_menu_alt_shortcut_label = 2131623944;
			
			// aapt resource value: 0x7F0E0009
			public const int abc_menu_ctrl_shortcut_label = 2131623945;
			
			// aapt resource value: 0x7F0E000A
			public const int abc_menu_delete_shortcut_label = 2131623946;
			
			// aapt resource value: 0x7F0E000B
			public const int abc_menu_enter_shortcut_label = 2131623947;
			
			// aapt resource value: 0x7F0E000C
			public const int abc_menu_function_shortcut_label = 2131623948;
			
			// aapt resource value: 0x7F0E000D
			public const int abc_menu_meta_shortcut_label = 2131623949;
			
			// aapt resource value: 0x7F0E000E
			public const int abc_menu_shift_shortcut_label = 2131623950;
			
			// aapt resource value: 0x7F0E000F
			public const int abc_menu_space_shortcut_label = 2131623951;
			
			// aapt resource value: 0x7F0E0010
			public const int abc_menu_sym_shortcut_label = 2131623952;
			
			// aapt resource value: 0x7F0E0011
			public const int abc_prepend_shortcut_label = 2131623953;
			
			// aapt resource value: 0x7F0E0013
			public const int abc_searchview_description_clear = 2131623955;
			
			// aapt resource value: 0x7F0E0014
			public const int abc_searchview_description_query = 2131623956;
			
			// aapt resource value: 0x7F0E0015
			public const int abc_searchview_description_search = 2131623957;
			
			// aapt resource value: 0x7F0E0016
			public const int abc_searchview_description_submit = 2131623958;
			
			// aapt resource value: 0x7F0E0017
			public const int abc_searchview_description_voice = 2131623959;
			
			// aapt resource value: 0x7F0E0012
			public const int abc_search_hint = 2131623954;
			
			// aapt resource value: 0x7F0E0018
			public const int abc_shareactionprovider_share_with = 2131623960;
			
			// aapt resource value: 0x7F0E0019
			public const int abc_shareactionprovider_share_with_application = 2131623961;
			
			// aapt resource value: 0x7F0E001A
			public const int abc_toolbar_collapse_description = 2131623962;
			
			// aapt resource value: 0x7F0E001B
			public const int appbar_scrolling_view_behavior = 2131623963;
			
			// aapt resource value: 0x7F0E001D
			public const int bottomsheet_action_expand_halfway = 2131623965;
			
			// aapt resource value: 0x7F0E001C
			public const int bottom_sheet_behavior = 2131623964;
			
			// aapt resource value: 0x7F0E001E
			public const int character_counter_content_description = 2131623966;
			
			// aapt resource value: 0x7F0E001F
			public const int character_counter_overflowed_content_description = 2131623967;
			
			// aapt resource value: 0x7F0E0020
			public const int character_counter_pattern = 2131623968;
			
			// aapt resource value: 0x7F0E0021
			public const int chip_text = 2131623969;
			
			// aapt resource value: 0x7F0E0022
			public const int clear_text_end_icon_content_description = 2131623970;
			
			// aapt resource value: 0x7F0E0023
			public const int copy = 2131623971;
			
			// aapt resource value: 0x7F0E0024
			public const int copy_toast_msg = 2131623972;
			
			// aapt resource value: 0x7F0E0025
			public const int error_icon_content_description = 2131623973;
			
			// aapt resource value: 0x7F0E0026
			public const int expand_button_title = 2131623974;
			
			// aapt resource value: 0x7F0E0027
			public const int exposed_dropdown_menu_content_description = 2131623975;
			
			// aapt resource value: 0x7F0E0028
			public const int fab_transformation_scrim_behavior = 2131623976;
			
			// aapt resource value: 0x7F0E0029
			public const int fab_transformation_sheet_behavior = 2131623977;
			
			// aapt resource value: 0x7F0E002A
			public const int fallback_menu_item_copy_link = 2131623978;
			
			// aapt resource value: 0x7F0E002B
			public const int fallback_menu_item_open_in_browser = 2131623979;
			
			// aapt resource value: 0x7F0E002C
			public const int fallback_menu_item_share_link = 2131623980;
			
			// aapt resource value: 0x7F0E002D
			public const int hide_bottom_view_on_scroll_behavior = 2131623981;
			
			// aapt resource value: 0x7F0E002E
			public const int icon_content_description = 2131623982;
			
			// aapt resource value: 0x7F0E002F
			public const int item_view_role_description = 2131623983;
			
			// aapt resource value: 0x7F0E0030
			public const int material_clock_display_divider = 2131623984;
			
			// aapt resource value: 0x7F0E0031
			public const int material_clock_toggle_content_description = 2131623985;
			
			// aapt resource value: 0x7F0E0032
			public const int material_hour_selection = 2131623986;
			
			// aapt resource value: 0x7F0E0033
			public const int material_hour_suffix = 2131623987;
			
			// aapt resource value: 0x7F0E0034
			public const int material_minute_selection = 2131623988;
			
			// aapt resource value: 0x7F0E0035
			public const int material_minute_suffix = 2131623989;
			
			// aapt resource value: 0x7F0E0036
			public const int material_motion_easing_accelerated = 2131623990;
			
			// aapt resource value: 0x7F0E0037
			public const int material_motion_easing_decelerated = 2131623991;
			
			// aapt resource value: 0x7F0E0038
			public const int material_motion_easing_emphasized = 2131623992;
			
			// aapt resource value: 0x7F0E0039
			public const int material_motion_easing_linear = 2131623993;
			
			// aapt resource value: 0x7F0E003A
			public const int material_motion_easing_standard = 2131623994;
			
			// aapt resource value: 0x7F0E003B
			public const int material_slider_range_end = 2131623995;
			
			// aapt resource value: 0x7F0E003C
			public const int material_slider_range_start = 2131623996;
			
			// aapt resource value: 0x7F0E003D
			public const int material_timepicker_am = 2131623997;
			
			// aapt resource value: 0x7F0E003E
			public const int material_timepicker_clock_mode_description = 2131623998;
			
			// aapt resource value: 0x7F0E003F
			public const int material_timepicker_hour = 2131623999;
			
			// aapt resource value: 0x7F0E0040
			public const int material_timepicker_minute = 2131624000;
			
			// aapt resource value: 0x7F0E0041
			public const int material_timepicker_pm = 2131624001;
			
			// aapt resource value: 0x7F0E0042
			public const int material_timepicker_select_time = 2131624002;
			
			// aapt resource value: 0x7F0E0043
			public const int material_timepicker_text_input_mode_description = 2131624003;
			
			// aapt resource value: 0x7F0E0044
			public const int mtrl_badge_numberless_content_description = 2131624004;
			
			// aapt resource value: 0x7F0E0045
			public const int mtrl_chip_close_icon_content_description = 2131624005;
			
			// aapt resource value: 0x7F0E0046
			public const int mtrl_exceed_max_badge_number_content_description = 2131624006;
			
			// aapt resource value: 0x7F0E0047
			public const int mtrl_exceed_max_badge_number_suffix = 2131624007;
			
			// aapt resource value: 0x7F0E0048
			public const int mtrl_picker_a11y_next_month = 2131624008;
			
			// aapt resource value: 0x7F0E0049
			public const int mtrl_picker_a11y_prev_month = 2131624009;
			
			// aapt resource value: 0x7F0E004A
			public const int mtrl_picker_announce_current_selection = 2131624010;
			
			// aapt resource value: 0x7F0E004B
			public const int mtrl_picker_cancel = 2131624011;
			
			// aapt resource value: 0x7F0E004C
			public const int mtrl_picker_confirm = 2131624012;
			
			// aapt resource value: 0x7F0E004D
			public const int mtrl_picker_date_header_selected = 2131624013;
			
			// aapt resource value: 0x7F0E004E
			public const int mtrl_picker_date_header_title = 2131624014;
			
			// aapt resource value: 0x7F0E004F
			public const int mtrl_picker_date_header_unselected = 2131624015;
			
			// aapt resource value: 0x7F0E0050
			public const int mtrl_picker_day_of_week_column_header = 2131624016;
			
			// aapt resource value: 0x7F0E0051
			public const int mtrl_picker_invalid_format = 2131624017;
			
			// aapt resource value: 0x7F0E0052
			public const int mtrl_picker_invalid_format_example = 2131624018;
			
			// aapt resource value: 0x7F0E0053
			public const int mtrl_picker_invalid_format_use = 2131624019;
			
			// aapt resource value: 0x7F0E0054
			public const int mtrl_picker_invalid_range = 2131624020;
			
			// aapt resource value: 0x7F0E0055
			public const int mtrl_picker_navigate_to_year_description = 2131624021;
			
			// aapt resource value: 0x7F0E0056
			public const int mtrl_picker_out_of_range = 2131624022;
			
			// aapt resource value: 0x7F0E0057
			public const int mtrl_picker_range_header_only_end_selected = 2131624023;
			
			// aapt resource value: 0x7F0E0058
			public const int mtrl_picker_range_header_only_start_selected = 2131624024;
			
			// aapt resource value: 0x7F0E0059
			public const int mtrl_picker_range_header_selected = 2131624025;
			
			// aapt resource value: 0x7F0E005A
			public const int mtrl_picker_range_header_title = 2131624026;
			
			// aapt resource value: 0x7F0E005B
			public const int mtrl_picker_range_header_unselected = 2131624027;
			
			// aapt resource value: 0x7F0E005C
			public const int mtrl_picker_save = 2131624028;
			
			// aapt resource value: 0x7F0E005D
			public const int mtrl_picker_text_input_date_hint = 2131624029;
			
			// aapt resource value: 0x7F0E005E
			public const int mtrl_picker_text_input_date_range_end_hint = 2131624030;
			
			// aapt resource value: 0x7F0E005F
			public const int mtrl_picker_text_input_date_range_start_hint = 2131624031;
			
			// aapt resource value: 0x7F0E0060
			public const int mtrl_picker_text_input_day_abbr = 2131624032;
			
			// aapt resource value: 0x7F0E0061
			public const int mtrl_picker_text_input_month_abbr = 2131624033;
			
			// aapt resource value: 0x7F0E0062
			public const int mtrl_picker_text_input_year_abbr = 2131624034;
			
			// aapt resource value: 0x7F0E0063
			public const int mtrl_picker_toggle_to_calendar_input_mode = 2131624035;
			
			// aapt resource value: 0x7F0E0064
			public const int mtrl_picker_toggle_to_day_selection = 2131624036;
			
			// aapt resource value: 0x7F0E0065
			public const int mtrl_picker_toggle_to_text_input_mode = 2131624037;
			
			// aapt resource value: 0x7F0E0066
			public const int mtrl_picker_toggle_to_year_selection = 2131624038;
			
			// aapt resource value: 0x7F0E0067
			public const int nav_app_bar_navigate_up_description = 2131624039;
			
			// aapt resource value: 0x7F0E0068
			public const int nav_app_bar_open_drawer_description = 2131624040;
			
			// aapt resource value: 0x7F0E0069
			public const int not_set = 2131624041;
			
			// aapt resource value: 0x7F0E006A
			public const int overflow_tab_title = 2131624042;
			
			// aapt resource value: 0x7F0E006B
			public const int password_toggle_content_description = 2131624043;
			
			// aapt resource value: 0x7F0E006C
			public const int path_password_eye = 2131624044;
			
			// aapt resource value: 0x7F0E006D
			public const int path_password_eye_mask_strike_through = 2131624045;
			
			// aapt resource value: 0x7F0E006E
			public const int path_password_eye_mask_visible = 2131624046;
			
			// aapt resource value: 0x7F0E006F
			public const int path_password_strike_through = 2131624047;
			
			// aapt resource value: 0x7F0E0070
			public const int preference_copied = 2131624048;
			
			// aapt resource value: 0x7F0E0071
			public const int search_menu_title = 2131624049;
			
			// aapt resource value: 0x7F0E0072
			public const int status_bar_notification_info_overflow = 2131624050;
			
			// aapt resource value: 0x7F0E0073
			public const int summary_collapsed_preference_list = 2131624051;
			
			// aapt resource value: 0x7F0E0074
			public const int v7_preference_off = 2131624052;
			
			// aapt resource value: 0x7F0E0075
			public const int v7_preference_on = 2131624053;
			
			static String()
			{
				global::Android.Runtime.ResourceIdManager.UpdateIdValues();
			}
			
			private String()
			{
			}
		}
		
		public partial class Style
		{
			
			// aapt resource value: 0x7F0F0000
			public const int AlertDialog_AppCompat = 2131689472;
			
			// aapt resource value: 0x7F0F0001
			public const int AlertDialog_AppCompat_Light = 2131689473;
			
			// aapt resource value: 0x7F0F0002
			public const int AndroidThemeColorAccentYellow = 2131689474;
			
			// aapt resource value: 0x7F0F0003
			public const int Animation_AppCompat_Dialog = 2131689475;
			
			// aapt resource value: 0x7F0F0004
			public const int Animation_AppCompat_DropDownUp = 2131689476;
			
			// aapt resource value: 0x7F0F0005
			public const int Animation_AppCompat_Tooltip = 2131689477;
			
			// aapt resource value: 0x7F0F0006
			public const int Animation_Design_BottomSheetDialog = 2131689478;
			
			// aapt resource value: 0x7F0F0007
			public const int Animation_MaterialComponents_BottomSheetDialog = 2131689479;
			
			// aapt resource value: 0x7F0F0008
			public const int AppCompatDialogStyle = 2131689480;
			
			// aapt resource value: 0x7F0F0009
			public const int Base_AlertDialog_AppCompat = 2131689481;
			
			// aapt resource value: 0x7F0F000A
			public const int Base_AlertDialog_AppCompat_Light = 2131689482;
			
			// aapt resource value: 0x7F0F000B
			public const int Base_Animation_AppCompat_Dialog = 2131689483;
			
			// aapt resource value: 0x7F0F000C
			public const int Base_Animation_AppCompat_DropDownUp = 2131689484;
			
			// aapt resource value: 0x7F0F000D
			public const int Base_Animation_AppCompat_Tooltip = 2131689485;
			
			// aapt resource value: 0x7F0F000E
			public const int Base_CardView = 2131689486;
			
			// aapt resource value: 0x7F0F0010
			public const int Base_DialogWindowTitleBackground_AppCompat = 2131689488;
			
			// aapt resource value: 0x7F0F000F
			public const int Base_DialogWindowTitle_AppCompat = 2131689487;
			
			// aapt resource value: 0x7F0F0011
			public const int Base_MaterialAlertDialog_MaterialComponents_Title_Icon = 2131689489;
			
			// aapt resource value: 0x7F0F0012
			public const int Base_MaterialAlertDialog_MaterialComponents_Title_Panel = 2131689490;
			
			// aapt resource value: 0x7F0F0013
			public const int Base_MaterialAlertDialog_MaterialComponents_Title_Text = 2131689491;
			
			// aapt resource value: 0x7F0F0014
			public const int Base_TextAppearance_AppCompat = 2131689492;
			
			// aapt resource value: 0x7F0F0015
			public const int Base_TextAppearance_AppCompat_Body1 = 2131689493;
			
			// aapt resource value: 0x7F0F0016
			public const int Base_TextAppearance_AppCompat_Body2 = 2131689494;
			
			// aapt resource value: 0x7F0F0017
			public const int Base_TextAppearance_AppCompat_Button = 2131689495;
			
			// aapt resource value: 0x7F0F0018
			public const int Base_TextAppearance_AppCompat_Caption = 2131689496;
			
			// aapt resource value: 0x7F0F0019
			public const int Base_TextAppearance_AppCompat_Display1 = 2131689497;
			
			// aapt resource value: 0x7F0F001A
			public const int Base_TextAppearance_AppCompat_Display2 = 2131689498;
			
			// aapt resource value: 0x7F0F001B
			public const int Base_TextAppearance_AppCompat_Display3 = 2131689499;
			
			// aapt resource value: 0x7F0F001C
			public const int Base_TextAppearance_AppCompat_Display4 = 2131689500;
			
			// aapt resource value: 0x7F0F001D
			public const int Base_TextAppearance_AppCompat_Headline = 2131689501;
			
			// aapt resource value: 0x7F0F001E
			public const int Base_TextAppearance_AppCompat_Inverse = 2131689502;
			
			// aapt resource value: 0x7F0F001F
			public const int Base_TextAppearance_AppCompat_Large = 2131689503;
			
			// aapt resource value: 0x7F0F0020
			public const int Base_TextAppearance_AppCompat_Large_Inverse = 2131689504;
			
			// aapt resource value: 0x7F0F0021
			public const int Base_TextAppearance_AppCompat_Light_Widget_PopupMenu_Large = 2131689505;
			
			// aapt resource value: 0x7F0F0022
			public const int Base_TextAppearance_AppCompat_Light_Widget_PopupMenu_Small = 2131689506;
			
			// aapt resource value: 0x7F0F0023
			public const int Base_TextAppearance_AppCompat_Medium = 2131689507;
			
			// aapt resource value: 0x7F0F0024
			public const int Base_TextAppearance_AppCompat_Medium_Inverse = 2131689508;
			
			// aapt resource value: 0x7F0F0025
			public const int Base_TextAppearance_AppCompat_Menu = 2131689509;
			
			// aapt resource value: 0x7F0F0026
			public const int Base_TextAppearance_AppCompat_SearchResult = 2131689510;
			
			// aapt resource value: 0x7F0F0027
			public const int Base_TextAppearance_AppCompat_SearchResult_Subtitle = 2131689511;
			
			// aapt resource value: 0x7F0F0028
			public const int Base_TextAppearance_AppCompat_SearchResult_Title = 2131689512;
			
			// aapt resource value: 0x7F0F0029
			public const int Base_TextAppearance_AppCompat_Small = 2131689513;
			
			// aapt resource value: 0x7F0F002A
			public const int Base_TextAppearance_AppCompat_Small_Inverse = 2131689514;
			
			// aapt resource value: 0x7F0F002B
			public const int Base_TextAppearance_AppCompat_Subhead = 2131689515;
			
			// aapt resource value: 0x7F0F002C
			public const int Base_TextAppearance_AppCompat_Subhead_Inverse = 2131689516;
			
			// aapt resource value: 0x7F0F002D
			public const int Base_TextAppearance_AppCompat_Title = 2131689517;
			
			// aapt resource value: 0x7F0F002E
			public const int Base_TextAppearance_AppCompat_Title_Inverse = 2131689518;
			
			// aapt resource value: 0x7F0F002F
			public const int Base_TextAppearance_AppCompat_Tooltip = 2131689519;
			
			// aapt resource value: 0x7F0F0030
			public const int Base_TextAppearance_AppCompat_Widget_ActionBar_Menu = 2131689520;
			
			// aapt resource value: 0x7F0F0031
			public const int Base_TextAppearance_AppCompat_Widget_ActionBar_Subtitle = 2131689521;
			
			// aapt resource value: 0x7F0F0032
			public const int Base_TextAppearance_AppCompat_Widget_ActionBar_Subtitle_Inverse = 2131689522;
			
			// aapt resource value: 0x7F0F0033
			public const int Base_TextAppearance_AppCompat_Widget_ActionBar_Title = 2131689523;
			
			// aapt resource value: 0x7F0F0034
			public const int Base_TextAppearance_AppCompat_Widget_ActionBar_Title_Inverse = 2131689524;
			
			// aapt resource value: 0x7F0F0035
			public const int Base_TextAppearance_AppCompat_Widget_ActionMode_Subtitle = 2131689525;
			
			// aapt resource value: 0x7F0F0036
			public const int Base_TextAppearance_AppCompat_Widget_ActionMode_Title = 2131689526;
			
			// aapt resource value: 0x7F0F0037
			public const int Base_TextAppearance_AppCompat_Widget_Button = 2131689527;
			
			// aapt resource value: 0x7F0F0038
			public const int Base_TextAppearance_AppCompat_Widget_Button_Borderless_Colored = 2131689528;
			
			// aapt resource value: 0x7F0F0039
			public const int Base_TextAppearance_AppCompat_Widget_Button_Colored = 2131689529;
			
			// aapt resource value: 0x7F0F003A
			public const int Base_TextAppearance_AppCompat_Widget_Button_Inverse = 2131689530;
			
			// aapt resource value: 0x7F0F003B
			public const int Base_TextAppearance_AppCompat_Widget_DropDownItem = 2131689531;
			
			// aapt resource value: 0x7F0F003C
			public const int Base_TextAppearance_AppCompat_Widget_PopupMenu_Header = 2131689532;
			
			// aapt resource value: 0x7F0F003D
			public const int Base_TextAppearance_AppCompat_Widget_PopupMenu_Large = 2131689533;
			
			// aapt resource value: 0x7F0F003E
			public const int Base_TextAppearance_AppCompat_Widget_PopupMenu_Small = 2131689534;
			
			// aapt resource value: 0x7F0F003F
			public const int Base_TextAppearance_AppCompat_Widget_Switch = 2131689535;
			
			// aapt resource value: 0x7F0F0040
			public const int Base_TextAppearance_AppCompat_Widget_TextView_SpinnerItem = 2131689536;
			
			// aapt resource value: 0x7F0F0041
			public const int Base_TextAppearance_MaterialComponents_Badge = 2131689537;
			
			// aapt resource value: 0x7F0F0042
			public const int Base_TextAppearance_MaterialComponents_Button = 2131689538;
			
			// aapt resource value: 0x7F0F0043
			public const int Base_TextAppearance_MaterialComponents_Headline6 = 2131689539;
			
			// aapt resource value: 0x7F0F0044
			public const int Base_TextAppearance_MaterialComponents_Subtitle2 = 2131689540;
			
			// aapt resource value: 0x7F0F0045
			public const int Base_TextAppearance_Widget_AppCompat_ExpandedMenu_Item = 2131689541;
			
			// aapt resource value: 0x7F0F0046
			public const int Base_TextAppearance_Widget_AppCompat_Toolbar_Subtitle = 2131689542;
			
			// aapt resource value: 0x7F0F0047
			public const int Base_TextAppearance_Widget_AppCompat_Toolbar_Title = 2131689543;
			
			// aapt resource value: 0x7F0F0069
			public const int Base_ThemeOverlay_AppCompat = 2131689577;
			
			// aapt resource value: 0x7F0F006A
			public const int Base_ThemeOverlay_AppCompat_ActionBar = 2131689578;
			
			// aapt resource value: 0x7F0F006B
			public const int Base_ThemeOverlay_AppCompat_Dark = 2131689579;
			
			// aapt resource value: 0x7F0F006C
			public const int Base_ThemeOverlay_AppCompat_Dark_ActionBar = 2131689580;
			
			// aapt resource value: 0x7F0F006D
			public const int Base_ThemeOverlay_AppCompat_Dialog = 2131689581;
			
			// aapt resource value: 0x7F0F006E
			public const int Base_ThemeOverlay_AppCompat_Dialog_Alert = 2131689582;
			
			// aapt resource value: 0x7F0F006F
			public const int Base_ThemeOverlay_AppCompat_Light = 2131689583;
			
			// aapt resource value: 0x7F0F0070
			public const int Base_ThemeOverlay_MaterialComponents_Dialog = 2131689584;
			
			// aapt resource value: 0x7F0F0071
			public const int Base_ThemeOverlay_MaterialComponents_Dialog_Alert = 2131689585;
			
			// aapt resource value: 0x7F0F0072
			public const int Base_ThemeOverlay_MaterialComponents_Dialog_Alert_Framework = 2131689586;
			
			// aapt resource value: 0x7F0F0073
			public const int Base_ThemeOverlay_MaterialComponents_Light_Dialog_Alert_Framework = 2131689587;
			
			// aapt resource value: 0x7F0F0074
			public const int Base_ThemeOverlay_MaterialComponents_MaterialAlertDialog = 2131689588;
			
			// aapt resource value: 0x7F0F0048
			public const int Base_Theme_AppCompat = 2131689544;
			
			// aapt resource value: 0x7F0F0049
			public const int Base_Theme_AppCompat_CompactMenu = 2131689545;
			
			// aapt resource value: 0x7F0F004A
			public const int Base_Theme_AppCompat_Dialog = 2131689546;
			
			// aapt resource value: 0x7F0F004E
			public const int Base_Theme_AppCompat_DialogWhenLarge = 2131689550;
			
			// aapt resource value: 0x7F0F004B
			public const int Base_Theme_AppCompat_Dialog_Alert = 2131689547;
			
			// aapt resource value: 0x7F0F004C
			public const int Base_Theme_AppCompat_Dialog_FixedSize = 2131689548;
			
			// aapt resource value: 0x7F0F004D
			public const int Base_Theme_AppCompat_Dialog_MinWidth = 2131689549;
			
			// aapt resource value: 0x7F0F004F
			public const int Base_Theme_AppCompat_Light = 2131689551;
			
			// aapt resource value: 0x7F0F0050
			public const int Base_Theme_AppCompat_Light_DarkActionBar = 2131689552;
			
			// aapt resource value: 0x7F0F0051
			public const int Base_Theme_AppCompat_Light_Dialog = 2131689553;
			
			// aapt resource value: 0x7F0F0055
			public const int Base_Theme_AppCompat_Light_DialogWhenLarge = 2131689557;
			
			// aapt resource value: 0x7F0F0052
			public const int Base_Theme_AppCompat_Light_Dialog_Alert = 2131689554;
			
			// aapt resource value: 0x7F0F0053
			public const int Base_Theme_AppCompat_Light_Dialog_FixedSize = 2131689555;
			
			// aapt resource value: 0x7F0F0054
			public const int Base_Theme_AppCompat_Light_Dialog_MinWidth = 2131689556;
			
			// aapt resource value: 0x7F0F0056
			public const int Base_Theme_MaterialComponents = 2131689558;
			
			// aapt resource value: 0x7F0F0057
			public const int Base_Theme_MaterialComponents_Bridge = 2131689559;
			
			// aapt resource value: 0x7F0F0058
			public const int Base_Theme_MaterialComponents_CompactMenu = 2131689560;
			
			// aapt resource value: 0x7F0F0059
			public const int Base_Theme_MaterialComponents_Dialog = 2131689561;
			
			// aapt resource value: 0x7F0F005E
			public const int Base_Theme_MaterialComponents_DialogWhenLarge = 2131689566;
			
			// aapt resource value: 0x7F0F005A
			public const int Base_Theme_MaterialComponents_Dialog_Alert = 2131689562;
			
			// aapt resource value: 0x7F0F005B
			public const int Base_Theme_MaterialComponents_Dialog_Bridge = 2131689563;
			
			// aapt resource value: 0x7F0F005C
			public const int Base_Theme_MaterialComponents_Dialog_FixedSize = 2131689564;
			
			// aapt resource value: 0x7F0F005D
			public const int Base_Theme_MaterialComponents_Dialog_MinWidth = 2131689565;
			
			// aapt resource value: 0x7F0F005F
			public const int Base_Theme_MaterialComponents_Light = 2131689567;
			
			// aapt resource value: 0x7F0F0060
			public const int Base_Theme_MaterialComponents_Light_Bridge = 2131689568;
			
			// aapt resource value: 0x7F0F0061
			public const int Base_Theme_MaterialComponents_Light_DarkActionBar = 2131689569;
			
			// aapt resource value: 0x7F0F0062
			public const int Base_Theme_MaterialComponents_Light_DarkActionBar_Bridge = 2131689570;
			
			// aapt resource value: 0x7F0F0063
			public const int Base_Theme_MaterialComponents_Light_Dialog = 2131689571;
			
			// aapt resource value: 0x7F0F0068
			public const int Base_Theme_MaterialComponents_Light_DialogWhenLarge = 2131689576;
			
			// aapt resource value: 0x7F0F0064
			public const int Base_Theme_MaterialComponents_Light_Dialog_Alert = 2131689572;
			
			// aapt resource value: 0x7F0F0065
			public const int Base_Theme_MaterialComponents_Light_Dialog_Bridge = 2131689573;
			
			// aapt resource value: 0x7F0F0066
			public const int Base_Theme_MaterialComponents_Light_Dialog_FixedSize = 2131689574;
			
			// aapt resource value: 0x7F0F0067
			public const int Base_Theme_MaterialComponents_Light_Dialog_MinWidth = 2131689575;
			
			// aapt resource value: 0x7F0F007E
			public const int Base_V14_ThemeOverlay_MaterialComponents_BottomSheetDialog = 2131689598;
			
			// aapt resource value: 0x7F0F007F
			public const int Base_V14_ThemeOverlay_MaterialComponents_Dialog = 2131689599;
			
			// aapt resource value: 0x7F0F0080
			public const int Base_V14_ThemeOverlay_MaterialComponents_Dialog_Alert = 2131689600;
			
			// aapt resource value: 0x7F0F0081
			public const int Base_V14_ThemeOverlay_MaterialComponents_MaterialAlertDialog = 2131689601;
			
			// aapt resource value: 0x7F0F0075
			public const int Base_V14_Theme_MaterialComponents = 2131689589;
			
			// aapt resource value: 0x7F0F0076
			public const int Base_V14_Theme_MaterialComponents_Bridge = 2131689590;
			
			// aapt resource value: 0x7F0F0077
			public const int Base_V14_Theme_MaterialComponents_Dialog = 2131689591;
			
			// aapt resource value: 0x7F0F0078
			public const int Base_V14_Theme_MaterialComponents_Dialog_Bridge = 2131689592;
			
			// aapt resource value: 0x7F0F0079
			public const int Base_V14_Theme_MaterialComponents_Light = 2131689593;
			
			// aapt resource value: 0x7F0F007A
			public const int Base_V14_Theme_MaterialComponents_Light_Bridge = 2131689594;
			
			// aapt resource value: 0x7F0F007B
			public const int Base_V14_Theme_MaterialComponents_Light_DarkActionBar_Bridge = 2131689595;
			
			// aapt resource value: 0x7F0F007C
			public const int Base_V14_Theme_MaterialComponents_Light_Dialog = 2131689596;
			
			// aapt resource value: 0x7F0F007D
			public const int Base_V14_Theme_MaterialComponents_Light_Dialog_Bridge = 2131689597;
			
			// aapt resource value: 0x7F0F008A
			public const int Base_V21_ThemeOverlay_AppCompat_Dialog = 2131689610;
			
			// aapt resource value: 0x7F0F008B
			public const int Base_V21_ThemeOverlay_MaterialComponents_BottomSheetDialog = 2131689611;
			
			// aapt resource value: 0x7F0F0082
			public const int Base_V21_Theme_AppCompat = 2131689602;
			
			// aapt resource value: 0x7F0F0083
			public const int Base_V21_Theme_AppCompat_Dialog = 2131689603;
			
			// aapt resource value: 0x7F0F0084
			public const int Base_V21_Theme_AppCompat_Light = 2131689604;
			
			// aapt resource value: 0x7F0F0085
			public const int Base_V21_Theme_AppCompat_Light_Dialog = 2131689605;
			
			// aapt resource value: 0x7F0F0086
			public const int Base_V21_Theme_MaterialComponents = 2131689606;
			
			// aapt resource value: 0x7F0F0087
			public const int Base_V21_Theme_MaterialComponents_Dialog = 2131689607;
			
			// aapt resource value: 0x7F0F0088
			public const int Base_V21_Theme_MaterialComponents_Light = 2131689608;
			
			// aapt resource value: 0x7F0F0089
			public const int Base_V21_Theme_MaterialComponents_Light_Dialog = 2131689609;
			
			// aapt resource value: 0x7F0F008C
			public const int Base_V22_Theme_AppCompat = 2131689612;
			
			// aapt resource value: 0x7F0F008D
			public const int Base_V22_Theme_AppCompat_Light = 2131689613;
			
			// aapt resource value: 0x7F0F008E
			public const int Base_V23_Theme_AppCompat = 2131689614;
			
			// aapt resource value: 0x7F0F008F
			public const int Base_V23_Theme_AppCompat_Light = 2131689615;
			
			// aapt resource value: 0x7F0F0090
			public const int Base_V26_Theme_AppCompat = 2131689616;
			
			// aapt resource value: 0x7F0F0091
			public const int Base_V26_Theme_AppCompat_Light = 2131689617;
			
			// aapt resource value: 0x7F0F0092
			public const int Base_V26_Widget_AppCompat_Toolbar = 2131689618;
			
			// aapt resource value: 0x7F0F0093
			public const int Base_V28_Theme_AppCompat = 2131689619;
			
			// aapt resource value: 0x7F0F0094
			public const int Base_V28_Theme_AppCompat_Light = 2131689620;
			
			// aapt resource value: 0x7F0F0099
			public const int Base_V7_ThemeOverlay_AppCompat_Dialog = 2131689625;
			
			// aapt resource value: 0x7F0F0095
			public const int Base_V7_Theme_AppCompat = 2131689621;
			
			// aapt resource value: 0x7F0F0096
			public const int Base_V7_Theme_AppCompat_Dialog = 2131689622;
			
			// aapt resource value: 0x7F0F0097
			public const int Base_V7_Theme_AppCompat_Light = 2131689623;
			
			// aapt resource value: 0x7F0F0098
			public const int Base_V7_Theme_AppCompat_Light_Dialog = 2131689624;
			
			// aapt resource value: 0x7F0F009A
			public const int Base_V7_Widget_AppCompat_AutoCompleteTextView = 2131689626;
			
			// aapt resource value: 0x7F0F009B
			public const int Base_V7_Widget_AppCompat_EditText = 2131689627;
			
			// aapt resource value: 0x7F0F009C
			public const int Base_V7_Widget_AppCompat_Toolbar = 2131689628;
			
			// aapt resource value: 0x7F0F009D
			public const int Base_Widget_AppCompat_ActionBar = 2131689629;
			
			// aapt resource value: 0x7F0F009E
			public const int Base_Widget_AppCompat_ActionBar_Solid = 2131689630;
			
			// aapt resource value: 0x7F0F009F
			public const int Base_Widget_AppCompat_ActionBar_TabBar = 2131689631;
			
			// aapt resource value: 0x7F0F00A0
			public const int Base_Widget_AppCompat_ActionBar_TabText = 2131689632;
			
			// aapt resource value: 0x7F0F00A1
			public const int Base_Widget_AppCompat_ActionBar_TabView = 2131689633;
			
			// aapt resource value: 0x7F0F00A2
			public const int Base_Widget_AppCompat_ActionButton = 2131689634;
			
			// aapt resource value: 0x7F0F00A3
			public const int Base_Widget_AppCompat_ActionButton_CloseMode = 2131689635;
			
			// aapt resource value: 0x7F0F00A4
			public const int Base_Widget_AppCompat_ActionButton_Overflow = 2131689636;
			
			// aapt resource value: 0x7F0F00A5
			public const int Base_Widget_AppCompat_ActionMode = 2131689637;
			
			// aapt resource value: 0x7F0F00A6
			public const int Base_Widget_AppCompat_ActivityChooserView = 2131689638;
			
			// aapt resource value: 0x7F0F00A7
			public const int Base_Widget_AppCompat_AutoCompleteTextView = 2131689639;
			
			// aapt resource value: 0x7F0F00A8
			public const int Base_Widget_AppCompat_Button = 2131689640;
			
			// aapt resource value: 0x7F0F00AE
			public const int Base_Widget_AppCompat_ButtonBar = 2131689646;
			
			// aapt resource value: 0x7F0F00AF
			public const int Base_Widget_AppCompat_ButtonBar_AlertDialog = 2131689647;
			
			// aapt resource value: 0x7F0F00A9
			public const int Base_Widget_AppCompat_Button_Borderless = 2131689641;
			
			// aapt resource value: 0x7F0F00AA
			public const int Base_Widget_AppCompat_Button_Borderless_Colored = 2131689642;
			
			// aapt resource value: 0x7F0F00AB
			public const int Base_Widget_AppCompat_Button_ButtonBar_AlertDialog = 2131689643;
			
			// aapt resource value: 0x7F0F00AC
			public const int Base_Widget_AppCompat_Button_Colored = 2131689644;
			
			// aapt resource value: 0x7F0F00AD
			public const int Base_Widget_AppCompat_Button_Small = 2131689645;
			
			// aapt resource value: 0x7F0F00B0
			public const int Base_Widget_AppCompat_CompoundButton_CheckBox = 2131689648;
			
			// aapt resource value: 0x7F0F00B1
			public const int Base_Widget_AppCompat_CompoundButton_RadioButton = 2131689649;
			
			// aapt resource value: 0x7F0F00B2
			public const int Base_Widget_AppCompat_CompoundButton_Switch = 2131689650;
			
			// aapt resource value: 0x7F0F00B3
			public const int Base_Widget_AppCompat_DrawerArrowToggle = 2131689651;
			
			// aapt resource value: 0x7F0F00B4
			public const int Base_Widget_AppCompat_DrawerArrowToggle_Common = 2131689652;
			
			// aapt resource value: 0x7F0F00B5
			public const int Base_Widget_AppCompat_DropDownItem_Spinner = 2131689653;
			
			// aapt resource value: 0x7F0F00B6
			public const int Base_Widget_AppCompat_EditText = 2131689654;
			
			// aapt resource value: 0x7F0F00B7
			public const int Base_Widget_AppCompat_ImageButton = 2131689655;
			
			// aapt resource value: 0x7F0F00B8
			public const int Base_Widget_AppCompat_Light_ActionBar = 2131689656;
			
			// aapt resource value: 0x7F0F00B9
			public const int Base_Widget_AppCompat_Light_ActionBar_Solid = 2131689657;
			
			// aapt resource value: 0x7F0F00BA
			public const int Base_Widget_AppCompat_Light_ActionBar_TabBar = 2131689658;
			
			// aapt resource value: 0x7F0F00BB
			public const int Base_Widget_AppCompat_Light_ActionBar_TabText = 2131689659;
			
			// aapt resource value: 0x7F0F00BC
			public const int Base_Widget_AppCompat_Light_ActionBar_TabText_Inverse = 2131689660;
			
			// aapt resource value: 0x7F0F00BD
			public const int Base_Widget_AppCompat_Light_ActionBar_TabView = 2131689661;
			
			// aapt resource value: 0x7F0F00BE
			public const int Base_Widget_AppCompat_Light_PopupMenu = 2131689662;
			
			// aapt resource value: 0x7F0F00BF
			public const int Base_Widget_AppCompat_Light_PopupMenu_Overflow = 2131689663;
			
			// aapt resource value: 0x7F0F00C0
			public const int Base_Widget_AppCompat_ListMenuView = 2131689664;
			
			// aapt resource value: 0x7F0F00C1
			public const int Base_Widget_AppCompat_ListPopupWindow = 2131689665;
			
			// aapt resource value: 0x7F0F00C2
			public const int Base_Widget_AppCompat_ListView = 2131689666;
			
			// aapt resource value: 0x7F0F00C3
			public const int Base_Widget_AppCompat_ListView_DropDown = 2131689667;
			
			// aapt resource value: 0x7F0F00C4
			public const int Base_Widget_AppCompat_ListView_Menu = 2131689668;
			
			// aapt resource value: 0x7F0F00C5
			public const int Base_Widget_AppCompat_PopupMenu = 2131689669;
			
			// aapt resource value: 0x7F0F00C6
			public const int Base_Widget_AppCompat_PopupMenu_Overflow = 2131689670;
			
			// aapt resource value: 0x7F0F00C7
			public const int Base_Widget_AppCompat_PopupWindow = 2131689671;
			
			// aapt resource value: 0x7F0F00C8
			public const int Base_Widget_AppCompat_ProgressBar = 2131689672;
			
			// aapt resource value: 0x7F0F00C9
			public const int Base_Widget_AppCompat_ProgressBar_Horizontal = 2131689673;
			
			// aapt resource value: 0x7F0F00CA
			public const int Base_Widget_AppCompat_RatingBar = 2131689674;
			
			// aapt resource value: 0x7F0F00CB
			public const int Base_Widget_AppCompat_RatingBar_Indicator = 2131689675;
			
			// aapt resource value: 0x7F0F00CC
			public const int Base_Widget_AppCompat_RatingBar_Small = 2131689676;
			
			// aapt resource value: 0x7F0F00CD
			public const int Base_Widget_AppCompat_SearchView = 2131689677;
			
			// aapt resource value: 0x7F0F00CE
			public const int Base_Widget_AppCompat_SearchView_ActionBar = 2131689678;
			
			// aapt resource value: 0x7F0F00CF
			public const int Base_Widget_AppCompat_SeekBar = 2131689679;
			
			// aapt resource value: 0x7F0F00D0
			public const int Base_Widget_AppCompat_SeekBar_Discrete = 2131689680;
			
			// aapt resource value: 0x7F0F00D1
			public const int Base_Widget_AppCompat_Spinner = 2131689681;
			
			// aapt resource value: 0x7F0F00D2
			public const int Base_Widget_AppCompat_Spinner_Underlined = 2131689682;
			
			// aapt resource value: 0x7F0F00D3
			public const int Base_Widget_AppCompat_TextView = 2131689683;
			
			// aapt resource value: 0x7F0F00D4
			public const int Base_Widget_AppCompat_TextView_SpinnerItem = 2131689684;
			
			// aapt resource value: 0x7F0F00D5
			public const int Base_Widget_AppCompat_Toolbar = 2131689685;
			
			// aapt resource value: 0x7F0F00D6
			public const int Base_Widget_AppCompat_Toolbar_Button_Navigation = 2131689686;
			
			// aapt resource value: 0x7F0F00D7
			public const int Base_Widget_Design_TabLayout = 2131689687;
			
			// aapt resource value: 0x7F0F00D8
			public const int Base_Widget_MaterialComponents_AutoCompleteTextView = 2131689688;
			
			// aapt resource value: 0x7F0F00D9
			public const int Base_Widget_MaterialComponents_CheckedTextView = 2131689689;
			
			// aapt resource value: 0x7F0F00DA
			public const int Base_Widget_MaterialComponents_Chip = 2131689690;
			
			// aapt resource value: 0x7F0F00DB
			public const int Base_Widget_MaterialComponents_MaterialCalendar_NavigationButton = 2131689691;
			
			// aapt resource value: 0x7F0F00DC
			public const int Base_Widget_MaterialComponents_PopupMenu = 2131689692;
			
			// aapt resource value: 0x7F0F00DD
			public const int Base_Widget_MaterialComponents_PopupMenu_ContextMenu = 2131689693;
			
			// aapt resource value: 0x7F0F00DE
			public const int Base_Widget_MaterialComponents_PopupMenu_ListPopupWindow = 2131689694;
			
			// aapt resource value: 0x7F0F00DF
			public const int Base_Widget_MaterialComponents_PopupMenu_Overflow = 2131689695;
			
			// aapt resource value: 0x7F0F00E0
			public const int Base_Widget_MaterialComponents_Slider = 2131689696;
			
			// aapt resource value: 0x7F0F00E1
			public const int Base_Widget_MaterialComponents_Snackbar = 2131689697;
			
			// aapt resource value: 0x7F0F00E2
			public const int Base_Widget_MaterialComponents_TextInputEditText = 2131689698;
			
			// aapt resource value: 0x7F0F00E3
			public const int Base_Widget_MaterialComponents_TextInputLayout = 2131689699;
			
			// aapt resource value: 0x7F0F00E4
			public const int Base_Widget_MaterialComponents_TextView = 2131689700;
			
			// aapt resource value: 0x7F0F00E5
			public const int CardView = 2131689701;
			
			// aapt resource value: 0x7F0F00E6
			public const int CardView_Dark = 2131689702;
			
			// aapt resource value: 0x7F0F00E7
			public const int CardView_Light = 2131689703;
			
			// aapt resource value: 0x7F0F02F8
			public const int collectionViewTheme = 2131690232;
			
			// aapt resource value: 0x7F0F00E8
			public const int EmptyTheme = 2131689704;
			
			// aapt resource value: 0x7F0F00E9
			public const int MainTheme = 2131689705;
			
			// aapt resource value: 0x7F0F00EA
			public const int MainTheme_Base = 2131689706;
			
			// aapt resource value: 0x7F0F00EB
			public const int MaterialAlertDialog_MaterialComponents = 2131689707;
			
			// aapt resource value: 0x7F0F00EC
			public const int MaterialAlertDialog_MaterialComponents_Body_Text = 2131689708;
			
			// aapt resource value: 0x7F0F00ED
			public const int MaterialAlertDialog_MaterialComponents_Picker_Date_Calendar = 2131689709;
			
			// aapt resource value: 0x7F0F00EE
			public const int MaterialAlertDialog_MaterialComponents_Picker_Date_Spinner = 2131689710;
			
			// aapt resource value: 0x7F0F00EF
			public const int MaterialAlertDialog_MaterialComponents_Title_Icon = 2131689711;
			
			// aapt resource value: 0x7F0F00F0
			public const int MaterialAlertDialog_MaterialComponents_Title_Icon_CenterStacked = 2131689712;
			
			// aapt resource value: 0x7F0F00F1
			public const int MaterialAlertDialog_MaterialComponents_Title_Panel = 2131689713;
			
			// aapt resource value: 0x7F0F00F2
			public const int MaterialAlertDialog_MaterialComponents_Title_Panel_CenterStacked = 2131689714;
			
			// aapt resource value: 0x7F0F00F3
			public const int MaterialAlertDialog_MaterialComponents_Title_Text = 2131689715;
			
			// aapt resource value: 0x7F0F00F4
			public const int MaterialAlertDialog_MaterialComponents_Title_Text_CenterStacked = 2131689716;
			
			// aapt resource value: 0x7F0F00F5
			public const int Platform_AppCompat = 2131689717;
			
			// aapt resource value: 0x7F0F00F6
			public const int Platform_AppCompat_Light = 2131689718;
			
			// aapt resource value: 0x7F0F00F7
			public const int Platform_MaterialComponents = 2131689719;
			
			// aapt resource value: 0x7F0F00F8
			public const int Platform_MaterialComponents_Dialog = 2131689720;
			
			// aapt resource value: 0x7F0F00F9
			public const int Platform_MaterialComponents_Light = 2131689721;
			
			// aapt resource value: 0x7F0F00FA
			public const int Platform_MaterialComponents_Light_Dialog = 2131689722;
			
			// aapt resource value: 0x7F0F00FB
			public const int Platform_ThemeOverlay_AppCompat = 2131689723;
			
			// aapt resource value: 0x7F0F00FC
			public const int Platform_ThemeOverlay_AppCompat_Dark = 2131689724;
			
			// aapt resource value: 0x7F0F00FD
			public const int Platform_ThemeOverlay_AppCompat_Light = 2131689725;
			
			// aapt resource value: 0x7F0F00FE
			public const int Platform_V21_AppCompat = 2131689726;
			
			// aapt resource value: 0x7F0F00FF
			public const int Platform_V21_AppCompat_Light = 2131689727;
			
			// aapt resource value: 0x7F0F0100
			public const int Platform_V25_AppCompat = 2131689728;
			
			// aapt resource value: 0x7F0F0101
			public const int Platform_V25_AppCompat_Light = 2131689729;
			
			// aapt resource value: 0x7F0F0102
			public const int Platform_Widget_AppCompat_Spinner = 2131689730;
			
			// aapt resource value: 0x7F0F0103
			public const int Preference = 2131689731;
			
			// aapt resource value: 0x7F0F0119
			public const int PreferenceCategoryTitleTextStyle = 2131689753;
			
			// aapt resource value: 0x7F0F011A
			public const int PreferenceFragment = 2131689754;
			
			// aapt resource value: 0x7F0F011C
			public const int PreferenceFragmentList = 2131689756;
			
			// aapt resource value: 0x7F0F011D
			public const int PreferenceFragmentList_Material = 2131689757;
			
			// aapt resource value: 0x7F0F011B
			public const int PreferenceFragment_Material = 2131689755;
			
			// aapt resource value: 0x7F0F011E
			public const int PreferenceSummaryTextStyle = 2131689758;
			
			// aapt resource value: 0x7F0F011F
			public const int PreferenceThemeOverlay = 2131689759;
			
			// aapt resource value: 0x7F0F0120
			public const int PreferenceThemeOverlay_v14 = 2131689760;
			
			// aapt resource value: 0x7F0F0121
			public const int PreferenceThemeOverlay_v14_Material = 2131689761;
			
			// aapt resource value: 0x7F0F0104
			public const int Preference_Category = 2131689732;
			
			// aapt resource value: 0x7F0F0105
			public const int Preference_Category_Material = 2131689733;
			
			// aapt resource value: 0x7F0F0106
			public const int Preference_CheckBoxPreference = 2131689734;
			
			// aapt resource value: 0x7F0F0107
			public const int Preference_CheckBoxPreference_Material = 2131689735;
			
			// aapt resource value: 0x7F0F0108
			public const int Preference_DialogPreference = 2131689736;
			
			// aapt resource value: 0x7F0F0109
			public const int Preference_DialogPreference_EditTextPreference = 2131689737;
			
			// aapt resource value: 0x7F0F010A
			public const int Preference_DialogPreference_EditTextPreference_Material = 2131689738;
			
			// aapt resource value: 0x7F0F010B
			public const int Preference_DialogPreference_Material = 2131689739;
			
			// aapt resource value: 0x7F0F010C
			public const int Preference_DropDown = 2131689740;
			
			// aapt resource value: 0x7F0F010D
			public const int Preference_DropDown_Material = 2131689741;
			
			// aapt resource value: 0x7F0F010E
			public const int Preference_Information = 2131689742;
			
			// aapt resource value: 0x7F0F010F
			public const int Preference_Information_Material = 2131689743;
			
			// aapt resource value: 0x7F0F0110
			public const int Preference_Material = 2131689744;
			
			// aapt resource value: 0x7F0F0111
			public const int Preference_PreferenceScreen = 2131689745;
			
			// aapt resource value: 0x7F0F0112
			public const int Preference_PreferenceScreen_Material = 2131689746;
			
			// aapt resource value: 0x7F0F0113
			public const int Preference_SeekBarPreference = 2131689747;
			
			// aapt resource value: 0x7F0F0114
			public const int Preference_SeekBarPreference_Material = 2131689748;
			
			// aapt resource value: 0x7F0F0115
			public const int Preference_SwitchPreference = 2131689749;
			
			// aapt resource value: 0x7F0F0117
			public const int Preference_SwitchPreferenceCompat = 2131689751;
			
			// aapt resource value: 0x7F0F0118
			public const int Preference_SwitchPreferenceCompat_Material = 2131689752;
			
			// aapt resource value: 0x7F0F0116
			public const int Preference_SwitchPreference_Material = 2131689750;
			
			// aapt resource value: 0x7F0F0122
			public const int RtlOverlay_DialogWindowTitle_AppCompat = 2131689762;
			
			// aapt resource value: 0x7F0F0123
			public const int RtlOverlay_Widget_AppCompat_ActionBar_TitleItem = 2131689763;
			
			// aapt resource value: 0x7F0F0124
			public const int RtlOverlay_Widget_AppCompat_DialogTitle_Icon = 2131689764;
			
			// aapt resource value: 0x7F0F0125
			public const int RtlOverlay_Widget_AppCompat_PopupMenuItem = 2131689765;
			
			// aapt resource value: 0x7F0F0126
			public const int RtlOverlay_Widget_AppCompat_PopupMenuItem_InternalGroup = 2131689766;
			
			// aapt resource value: 0x7F0F0127
			public const int RtlOverlay_Widget_AppCompat_PopupMenuItem_Shortcut = 2131689767;
			
			// aapt resource value: 0x7F0F0128
			public const int RtlOverlay_Widget_AppCompat_PopupMenuItem_SubmenuArrow = 2131689768;
			
			// aapt resource value: 0x7F0F0129
			public const int RtlOverlay_Widget_AppCompat_PopupMenuItem_Text = 2131689769;
			
			// aapt resource value: 0x7F0F012A
			public const int RtlOverlay_Widget_AppCompat_PopupMenuItem_Title = 2131689770;
			
			// aapt resource value: 0x7F0F0130
			public const int RtlOverlay_Widget_AppCompat_SearchView_MagIcon = 2131689776;
			
			// aapt resource value: 0x7F0F012B
			public const int RtlOverlay_Widget_AppCompat_Search_DropDown = 2131689771;
			
			// aapt resource value: 0x7F0F012C
			public const int RtlOverlay_Widget_AppCompat_Search_DropDown_Icon1 = 2131689772;
			
			// aapt resource value: 0x7F0F012D
			public const int RtlOverlay_Widget_AppCompat_Search_DropDown_Icon2 = 2131689773;
			
			// aapt resource value: 0x7F0F012E
			public const int RtlOverlay_Widget_AppCompat_Search_DropDown_Query = 2131689774;
			
			// aapt resource value: 0x7F0F012F
			public const int RtlOverlay_Widget_AppCompat_Search_DropDown_Text = 2131689775;
			
			// aapt resource value: 0x7F0F0131
			public const int RtlUnderlay_Widget_AppCompat_ActionButton = 2131689777;
			
			// aapt resource value: 0x7F0F0132
			public const int RtlUnderlay_Widget_AppCompat_ActionButton_Overflow = 2131689778;
			
			// aapt resource value: 0x7F0F02F9
			public const int scrollViewScrollBars = 2131690233;
			
			// aapt resource value: 0x7F0F02FA
			public const int scrollViewTheme = 2131690234;
			
			// aapt resource value: 0x7F0F0139
			public const int ShapeAppearanceOverlay = 2131689785;
			
			// aapt resource value: 0x7F0F013A
			public const int ShapeAppearanceOverlay_BottomLeftDifferentCornerSize = 2131689786;
			
			// aapt resource value: 0x7F0F013B
			public const int ShapeAppearanceOverlay_BottomRightCut = 2131689787;
			
			// aapt resource value: 0x7F0F013C
			public const int ShapeAppearanceOverlay_Cut = 2131689788;
			
			// aapt resource value: 0x7F0F013D
			public const int ShapeAppearanceOverlay_DifferentCornerSize = 2131689789;
			
			// aapt resource value: 0x7F0F013E
			public const int ShapeAppearanceOverlay_MaterialComponents_BottomSheet = 2131689790;
			
			// aapt resource value: 0x7F0F013F
			public const int ShapeAppearanceOverlay_MaterialComponents_Chip = 2131689791;
			
			// aapt resource value: 0x7F0F0140
			public const int ShapeAppearanceOverlay_MaterialComponents_ExtendedFloatingActionButton = 2131689792;
			
			// aapt resource value: 0x7F0F0141
			public const int ShapeAppearanceOverlay_MaterialComponents_FloatingActionButton = 2131689793;
			
			// aapt resource value: 0x7F0F0142
			public const int ShapeAppearanceOverlay_MaterialComponents_MaterialCalendar_Day = 2131689794;
			
			// aapt resource value: 0x7F0F0143
			public const int ShapeAppearanceOverlay_MaterialComponents_MaterialCalendar_Window_Fullscreen = 2131689795;
			
			// aapt resource value: 0x7F0F0144
			public const int ShapeAppearanceOverlay_MaterialComponents_MaterialCalendar_Year = 2131689796;
			
			// aapt resource value: 0x7F0F0145
			public const int ShapeAppearanceOverlay_MaterialComponents_TextInputLayout_FilledBox = 2131689797;
			
			// aapt resource value: 0x7F0F0146
			public const int ShapeAppearanceOverlay_TopLeftCut = 2131689798;
			
			// aapt resource value: 0x7F0F0147
			public const int ShapeAppearanceOverlay_TopRightDifferentCornerSize = 2131689799;
			
			// aapt resource value: 0x7F0F0133
			public const int ShapeAppearance_MaterialComponents = 2131689779;
			
			// aapt resource value: 0x7F0F0134
			public const int ShapeAppearance_MaterialComponents_LargeComponent = 2131689780;
			
			// aapt resource value: 0x7F0F0135
			public const int ShapeAppearance_MaterialComponents_MediumComponent = 2131689781;
			
			// aapt resource value: 0x7F0F0136
			public const int ShapeAppearance_MaterialComponents_SmallComponent = 2131689782;
			
			// aapt resource value: 0x7F0F0137
			public const int ShapeAppearance_MaterialComponents_Test = 2131689783;
			
			// aapt resource value: 0x7F0F0138
			public const int ShapeAppearance_MaterialComponents_Tooltip = 2131689784;
			
			// aapt resource value: 0x7F0F014D
			public const int TestStyleWithLineHeight = 2131689805;
			
			// aapt resource value: 0x7F0F014E
			public const int TestStyleWithLineHeightAppearance = 2131689806;
			
			// aapt resource value: 0x7F0F0150
			public const int TestStyleWithoutLineHeight = 2131689808;
			
			// aapt resource value: 0x7F0F014F
			public const int TestStyleWithThemeLineHeightAttribute = 2131689807;
			
			// aapt resource value: 0x7F0F0151
			public const int TestThemeWithLineHeight = 2131689809;
			
			// aapt resource value: 0x7F0F0152
			public const int TestThemeWithLineHeightDisabled = 2131689810;
			
			// aapt resource value: 0x7F0F0148
			public const int Test_ShapeAppearanceOverlay_MaterialComponents_MaterialCalendar_Day = 2131689800;
			
			// aapt resource value: 0x7F0F0149
			public const int Test_Theme_MaterialComponents_MaterialCalendar = 2131689801;
			
			// aapt resource value: 0x7F0F014A
			public const int Test_Widget_MaterialComponents_MaterialCalendar = 2131689802;
			
			// aapt resource value: 0x7F0F014B
			public const int Test_Widget_MaterialComponents_MaterialCalendar_Day = 2131689803;
			
			// aapt resource value: 0x7F0F014C
			public const int Test_Widget_MaterialComponents_MaterialCalendar_Day_Selected = 2131689804;
			
			// aapt resource value: 0x7F0F0153
			public const int TextAppearance_AppCompat = 2131689811;
			
			// aapt resource value: 0x7F0F0154
			public const int TextAppearance_AppCompat_Body1 = 2131689812;
			
			// aapt resource value: 0x7F0F0155
			public const int TextAppearance_AppCompat_Body2 = 2131689813;
			
			// aapt resource value: 0x7F0F0156
			public const int TextAppearance_AppCompat_Button = 2131689814;
			
			// aapt resource value: 0x7F0F0157
			public const int TextAppearance_AppCompat_Caption = 2131689815;
			
			// aapt resource value: 0x7F0F0158
			public const int TextAppearance_AppCompat_Display1 = 2131689816;
			
			// aapt resource value: 0x7F0F0159
			public const int TextAppearance_AppCompat_Display2 = 2131689817;
			
			// aapt resource value: 0x7F0F015A
			public const int TextAppearance_AppCompat_Display3 = 2131689818;
			
			// aapt resource value: 0x7F0F015B
			public const int TextAppearance_AppCompat_Display4 = 2131689819;
			
			// aapt resource value: 0x7F0F015C
			public const int TextAppearance_AppCompat_Headline = 2131689820;
			
			// aapt resource value: 0x7F0F015D
			public const int TextAppearance_AppCompat_Inverse = 2131689821;
			
			// aapt resource value: 0x7F0F015E
			public const int TextAppearance_AppCompat_Large = 2131689822;
			
			// aapt resource value: 0x7F0F015F
			public const int TextAppearance_AppCompat_Large_Inverse = 2131689823;
			
			// aapt resource value: 0x7F0F0160
			public const int TextAppearance_AppCompat_Light_SearchResult_Subtitle = 2131689824;
			
			// aapt resource value: 0x7F0F0161
			public const int TextAppearance_AppCompat_Light_SearchResult_Title = 2131689825;
			
			// aapt resource value: 0x7F0F0162
			public const int TextAppearance_AppCompat_Light_Widget_PopupMenu_Large = 2131689826;
			
			// aapt resource value: 0x7F0F0163
			public const int TextAppearance_AppCompat_Light_Widget_PopupMenu_Small = 2131689827;
			
			// aapt resource value: 0x7F0F0164
			public const int TextAppearance_AppCompat_Medium = 2131689828;
			
			// aapt resource value: 0x7F0F0165
			public const int TextAppearance_AppCompat_Medium_Inverse = 2131689829;
			
			// aapt resource value: 0x7F0F0166
			public const int TextAppearance_AppCompat_Menu = 2131689830;
			
			// aapt resource value: 0x7F0F0167
			public const int TextAppearance_AppCompat_SearchResult_Subtitle = 2131689831;
			
			// aapt resource value: 0x7F0F0168
			public const int TextAppearance_AppCompat_SearchResult_Title = 2131689832;
			
			// aapt resource value: 0x7F0F0169
			public const int TextAppearance_AppCompat_Small = 2131689833;
			
			// aapt resource value: 0x7F0F016A
			public const int TextAppearance_AppCompat_Small_Inverse = 2131689834;
			
			// aapt resource value: 0x7F0F016B
			public const int TextAppearance_AppCompat_Subhead = 2131689835;
			
			// aapt resource value: 0x7F0F016C
			public const int TextAppearance_AppCompat_Subhead_Inverse = 2131689836;
			
			// aapt resource value: 0x7F0F016D
			public const int TextAppearance_AppCompat_Title = 2131689837;
			
			// aapt resource value: 0x7F0F016E
			public const int TextAppearance_AppCompat_Title_Inverse = 2131689838;
			
			// aapt resource value: 0x7F0F016F
			public const int TextAppearance_AppCompat_Tooltip = 2131689839;
			
			// aapt resource value: 0x7F0F0170
			public const int TextAppearance_AppCompat_Widget_ActionBar_Menu = 2131689840;
			
			// aapt resource value: 0x7F0F0171
			public const int TextAppearance_AppCompat_Widget_ActionBar_Subtitle = 2131689841;
			
			// aapt resource value: 0x7F0F0172
			public const int TextAppearance_AppCompat_Widget_ActionBar_Subtitle_Inverse = 2131689842;
			
			// aapt resource value: 0x7F0F0173
			public const int TextAppearance_AppCompat_Widget_ActionBar_Title = 2131689843;
			
			// aapt resource value: 0x7F0F0174
			public const int TextAppearance_AppCompat_Widget_ActionBar_Title_Inverse = 2131689844;
			
			// aapt resource value: 0x7F0F0175
			public const int TextAppearance_AppCompat_Widget_ActionMode_Subtitle = 2131689845;
			
			// aapt resource value: 0x7F0F0176
			public const int TextAppearance_AppCompat_Widget_ActionMode_Subtitle_Inverse = 2131689846;
			
			// aapt resource value: 0x7F0F0177
			public const int TextAppearance_AppCompat_Widget_ActionMode_Title = 2131689847;
			
			// aapt resource value: 0x7F0F0178
			public const int TextAppearance_AppCompat_Widget_ActionMode_Title_Inverse = 2131689848;
			
			// aapt resource value: 0x7F0F0179
			public const int TextAppearance_AppCompat_Widget_Button = 2131689849;
			
			// aapt resource value: 0x7F0F017A
			public const int TextAppearance_AppCompat_Widget_Button_Borderless_Colored = 2131689850;
			
			// aapt resource value: 0x7F0F017B
			public const int TextAppearance_AppCompat_Widget_Button_Colored = 2131689851;
			
			// aapt resource value: 0x7F0F017C
			public const int TextAppearance_AppCompat_Widget_Button_Inverse = 2131689852;
			
			// aapt resource value: 0x7F0F017D
			public const int TextAppearance_AppCompat_Widget_DropDownItem = 2131689853;
			
			// aapt resource value: 0x7F0F017E
			public const int TextAppearance_AppCompat_Widget_PopupMenu_Header = 2131689854;
			
			// aapt resource value: 0x7F0F017F
			public const int TextAppearance_AppCompat_Widget_PopupMenu_Large = 2131689855;
			
			// aapt resource value: 0x7F0F0180
			public const int TextAppearance_AppCompat_Widget_PopupMenu_Small = 2131689856;
			
			// aapt resource value: 0x7F0F0181
			public const int TextAppearance_AppCompat_Widget_Switch = 2131689857;
			
			// aapt resource value: 0x7F0F0182
			public const int TextAppearance_AppCompat_Widget_TextView_SpinnerItem = 2131689858;
			
			// aapt resource value: 0x7F0F0183
			public const int TextAppearance_Compat_Notification = 2131689859;
			
			// aapt resource value: 0x7F0F0184
			public const int TextAppearance_Compat_Notification_Info = 2131689860;
			
			// aapt resource value: 0x7F0F0185
			public const int TextAppearance_Compat_Notification_Info_Media = 2131689861;
			
			// aapt resource value: 0x7F0F0186
			public const int TextAppearance_Compat_Notification_Line2 = 2131689862;
			
			// aapt resource value: 0x7F0F0187
			public const int TextAppearance_Compat_Notification_Line2_Media = 2131689863;
			
			// aapt resource value: 0x7F0F0188
			public const int TextAppearance_Compat_Notification_Media = 2131689864;
			
			// aapt resource value: 0x7F0F0189
			public const int TextAppearance_Compat_Notification_Time = 2131689865;
			
			// aapt resource value: 0x7F0F018A
			public const int TextAppearance_Compat_Notification_Time_Media = 2131689866;
			
			// aapt resource value: 0x7F0F018B
			public const int TextAppearance_Compat_Notification_Title = 2131689867;
			
			// aapt resource value: 0x7F0F018C
			public const int TextAppearance_Compat_Notification_Title_Media = 2131689868;
			
			// aapt resource value: 0x7F0F018D
			public const int TextAppearance_Design_CollapsingToolbar_Expanded = 2131689869;
			
			// aapt resource value: 0x7F0F018E
			public const int TextAppearance_Design_Counter = 2131689870;
			
			// aapt resource value: 0x7F0F018F
			public const int TextAppearance_Design_Counter_Overflow = 2131689871;
			
			// aapt resource value: 0x7F0F0190
			public const int TextAppearance_Design_Error = 2131689872;
			
			// aapt resource value: 0x7F0F0191
			public const int TextAppearance_Design_HelperText = 2131689873;
			
			// aapt resource value: 0x7F0F0192
			public const int TextAppearance_Design_Hint = 2131689874;
			
			// aapt resource value: 0x7F0F0193
			public const int TextAppearance_Design_Placeholder = 2131689875;
			
			// aapt resource value: 0x7F0F0194
			public const int TextAppearance_Design_Prefix = 2131689876;
			
			// aapt resource value: 0x7F0F0195
			public const int TextAppearance_Design_Snackbar_Message = 2131689877;
			
			// aapt resource value: 0x7F0F0196
			public const int TextAppearance_Design_Suffix = 2131689878;
			
			// aapt resource value: 0x7F0F0197
			public const int TextAppearance_Design_Tab = 2131689879;
			
			// aapt resource value: 0x7F0F0198
			public const int TextAppearance_MaterialComponents_Badge = 2131689880;
			
			// aapt resource value: 0x7F0F0199
			public const int TextAppearance_MaterialComponents_Body1 = 2131689881;
			
			// aapt resource value: 0x7F0F019A
			public const int TextAppearance_MaterialComponents_Body2 = 2131689882;
			
			// aapt resource value: 0x7F0F019B
			public const int TextAppearance_MaterialComponents_Button = 2131689883;
			
			// aapt resource value: 0x7F0F019C
			public const int TextAppearance_MaterialComponents_Caption = 2131689884;
			
			// aapt resource value: 0x7F0F019D
			public const int TextAppearance_MaterialComponents_Chip = 2131689885;
			
			// aapt resource value: 0x7F0F019E
			public const int TextAppearance_MaterialComponents_Headline1 = 2131689886;
			
			// aapt resource value: 0x7F0F019F
			public const int TextAppearance_MaterialComponents_Headline2 = 2131689887;
			
			// aapt resource value: 0x7F0F01A0
			public const int TextAppearance_MaterialComponents_Headline3 = 2131689888;
			
			// aapt resource value: 0x7F0F01A1
			public const int TextAppearance_MaterialComponents_Headline4 = 2131689889;
			
			// aapt resource value: 0x7F0F01A2
			public const int TextAppearance_MaterialComponents_Headline5 = 2131689890;
			
			// aapt resource value: 0x7F0F01A3
			public const int TextAppearance_MaterialComponents_Headline6 = 2131689891;
			
			// aapt resource value: 0x7F0F01A4
			public const int TextAppearance_MaterialComponents_Overline = 2131689892;
			
			// aapt resource value: 0x7F0F01A5
			public const int TextAppearance_MaterialComponents_Subtitle1 = 2131689893;
			
			// aapt resource value: 0x7F0F01A6
			public const int TextAppearance_MaterialComponents_Subtitle2 = 2131689894;
			
			// aapt resource value: 0x7F0F01A7
			public const int TextAppearance_MaterialComponents_TimePicker_Title = 2131689895;
			
			// aapt resource value: 0x7F0F01A8
			public const int TextAppearance_MaterialComponents_Tooltip = 2131689896;
			
			// aapt resource value: 0x7F0F01A9
			public const int TextAppearance_Widget_AppCompat_ExpandedMenu_Item = 2131689897;
			
			// aapt resource value: 0x7F0F01AA
			public const int TextAppearance_Widget_AppCompat_Toolbar_Subtitle = 2131689898;
			
			// aapt resource value: 0x7F0F01AB
			public const int TextAppearance_Widget_AppCompat_Toolbar_Title = 2131689899;
			
			// aapt resource value: 0x7F0F0229
			public const int ThemeOverlayColorAccentRed = 2131690025;
			
			// aapt resource value: 0x7F0F01F9
			public const int ThemeOverlay_AppCompat = 2131689977;
			
			// aapt resource value: 0x7F0F01FA
			public const int ThemeOverlay_AppCompat_ActionBar = 2131689978;
			
			// aapt resource value: 0x7F0F01FB
			public const int ThemeOverlay_AppCompat_Dark = 2131689979;
			
			// aapt resource value: 0x7F0F01FC
			public const int ThemeOverlay_AppCompat_Dark_ActionBar = 2131689980;
			
			// aapt resource value: 0x7F0F01FD
			public const int ThemeOverlay_AppCompat_DayNight = 2131689981;
			
			// aapt resource value: 0x7F0F01FE
			public const int ThemeOverlay_AppCompat_DayNight_ActionBar = 2131689982;
			
			// aapt resource value: 0x7F0F01FF
			public const int ThemeOverlay_AppCompat_Dialog = 2131689983;
			
			// aapt resource value: 0x7F0F0200
			public const int ThemeOverlay_AppCompat_Dialog_Alert = 2131689984;
			
			// aapt resource value: 0x7F0F0201
			public const int ThemeOverlay_AppCompat_Light = 2131689985;
			
			// aapt resource value: 0x7F0F0202
			public const int ThemeOverlay_Design_TextInputEditText = 2131689986;
			
			// aapt resource value: 0x7F0F0203
			public const int ThemeOverlay_MaterialComponents = 2131689987;
			
			// aapt resource value: 0x7F0F0204
			public const int ThemeOverlay_MaterialComponents_ActionBar = 2131689988;
			
			// aapt resource value: 0x7F0F0205
			public const int ThemeOverlay_MaterialComponents_ActionBar_Primary = 2131689989;
			
			// aapt resource value: 0x7F0F0206
			public const int ThemeOverlay_MaterialComponents_ActionBar_Surface = 2131689990;
			
			// aapt resource value: 0x7F0F0207
			public const int ThemeOverlay_MaterialComponents_AutoCompleteTextView = 2131689991;
			
			// aapt resource value: 0x7F0F0208
			public const int ThemeOverlay_MaterialComponents_AutoCompleteTextView_FilledBox = 2131689992;
			
			// aapt resource value: 0x7F0F0209
			public const int ThemeOverlay_MaterialComponents_AutoCompleteTextView_FilledBox_Dense = 2131689993;
			
			// aapt resource value: 0x7F0F020A
			public const int ThemeOverlay_MaterialComponents_AutoCompleteTextView_OutlinedBox = 2131689994;
			
			// aapt resource value: 0x7F0F020B
			public const int ThemeOverlay_MaterialComponents_AutoCompleteTextView_OutlinedBox_Dense = 2131689995;
			
			// aapt resource value: 0x7F0F020C
			public const int ThemeOverlay_MaterialComponents_BottomAppBar_Primary = 2131689996;
			
			// aapt resource value: 0x7F0F020D
			public const int ThemeOverlay_MaterialComponents_BottomAppBar_Surface = 2131689997;
			
			// aapt resource value: 0x7F0F020E
			public const int ThemeOverlay_MaterialComponents_BottomSheetDialog = 2131689998;
			
			// aapt resource value: 0x7F0F020F
			public const int ThemeOverlay_MaterialComponents_Dark = 2131689999;
			
			// aapt resource value: 0x7F0F0210
			public const int ThemeOverlay_MaterialComponents_Dark_ActionBar = 2131690000;
			
			// aapt resource value: 0x7F0F0211
			public const int ThemeOverlay_MaterialComponents_DayNight_BottomSheetDialog = 2131690001;
			
			// aapt resource value: 0x7F0F0212
			public const int ThemeOverlay_MaterialComponents_Dialog = 2131690002;
			
			// aapt resource value: 0x7F0F0213
			public const int ThemeOverlay_MaterialComponents_Dialog_Alert = 2131690003;
			
			// aapt resource value: 0x7F0F0214
			public const int ThemeOverlay_MaterialComponents_Dialog_Alert_Framework = 2131690004;
			
			// aapt resource value: 0x7F0F0215
			public const int ThemeOverlay_MaterialComponents_Light = 2131690005;
			
			// aapt resource value: 0x7F0F0216
			public const int ThemeOverlay_MaterialComponents_Light_Dialog_Alert_Framework = 2131690006;
			
			// aapt resource value: 0x7F0F0217
			public const int ThemeOverlay_MaterialComponents_MaterialAlertDialog = 2131690007;
			
			// aapt resource value: 0x7F0F0218
			public const int ThemeOverlay_MaterialComponents_MaterialAlertDialog_Centered = 2131690008;
			
			// aapt resource value: 0x7F0F0219
			public const int ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date = 2131690009;
			
			// aapt resource value: 0x7F0F021A
			public const int ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date_Calendar = 2131690010;
			
			// aapt resource value: 0x7F0F021B
			public const int ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date_Header_Text = 2131690011;
			
			// aapt resource value: 0x7F0F021C
			public const int ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date_Header_Text_Day = 2131690012;
			
			// aapt resource value: 0x7F0F021D
			public const int ThemeOverlay_MaterialComponents_MaterialAlertDialog_Picker_Date_Spinner = 2131690013;
			
			// aapt resource value: 0x7F0F021E
			public const int ThemeOverlay_MaterialComponents_MaterialCalendar = 2131690014;
			
			// aapt resource value: 0x7F0F021F
			public const int ThemeOverlay_MaterialComponents_MaterialCalendar_Fullscreen = 2131690015;
			
			// aapt resource value: 0x7F0F0220
			public const int ThemeOverlay_MaterialComponents_TextInputEditText = 2131690016;
			
			// aapt resource value: 0x7F0F0221
			public const int ThemeOverlay_MaterialComponents_TextInputEditText_FilledBox = 2131690017;
			
			// aapt resource value: 0x7F0F0222
			public const int ThemeOverlay_MaterialComponents_TextInputEditText_FilledBox_Dense = 2131690018;
			
			// aapt resource value: 0x7F0F0223
			public const int ThemeOverlay_MaterialComponents_TextInputEditText_OutlinedBox = 2131690019;
			
			// aapt resource value: 0x7F0F0224
			public const int ThemeOverlay_MaterialComponents_TextInputEditText_OutlinedBox_Dense = 2131690020;
			
			// aapt resource value: 0x7F0F0225
			public const int ThemeOverlay_MaterialComponents_TimePicker = 2131690021;
			
			// aapt resource value: 0x7F0F0226
			public const int ThemeOverlay_MaterialComponents_TimePicker_Display = 2131690022;
			
			// aapt resource value: 0x7F0F0227
			public const int ThemeOverlay_MaterialComponents_Toolbar_Primary = 2131690023;
			
			// aapt resource value: 0x7F0F0228
			public const int ThemeOverlay_MaterialComponents_Toolbar_Surface = 2131690024;
			
			// aapt resource value: 0x7F0F01AC
			public const int Theme_AppCompat = 2131689900;
			
			// aapt resource value: 0x7F0F01AD
			public const int Theme_AppCompat_CompactMenu = 2131689901;
			
			// aapt resource value: 0x7F0F01AE
			public const int Theme_AppCompat_DayNight = 2131689902;
			
			// aapt resource value: 0x7F0F01AF
			public const int Theme_AppCompat_DayNight_DarkActionBar = 2131689903;
			
			// aapt resource value: 0x7F0F01B0
			public const int Theme_AppCompat_DayNight_Dialog = 2131689904;
			
			// aapt resource value: 0x7F0F01B3
			public const int Theme_AppCompat_DayNight_DialogWhenLarge = 2131689907;
			
			// aapt resource value: 0x7F0F01B1
			public const int Theme_AppCompat_DayNight_Dialog_Alert = 2131689905;
			
			// aapt resource value: 0x7F0F01B2
			public const int Theme_AppCompat_DayNight_Dialog_MinWidth = 2131689906;
			
			// aapt resource value: 0x7F0F01B4
			public const int Theme_AppCompat_DayNight_NoActionBar = 2131689908;
			
			// aapt resource value: 0x7F0F01B5
			public const int Theme_AppCompat_Dialog = 2131689909;
			
			// aapt resource value: 0x7F0F01B8
			public const int Theme_AppCompat_DialogWhenLarge = 2131689912;
			
			// aapt resource value: 0x7F0F01B6
			public const int Theme_AppCompat_Dialog_Alert = 2131689910;
			
			// aapt resource value: 0x7F0F01B7
			public const int Theme_AppCompat_Dialog_MinWidth = 2131689911;
			
			// aapt resource value: 0x7F0F01B9
			public const int Theme_AppCompat_Empty = 2131689913;
			
			// aapt resource value: 0x7F0F01BA
			public const int Theme_AppCompat_Light = 2131689914;
			
			// aapt resource value: 0x7F0F01BB
			public const int Theme_AppCompat_Light_DarkActionBar = 2131689915;
			
			// aapt resource value: 0x7F0F01BC
			public const int Theme_AppCompat_Light_Dialog = 2131689916;
			
			// aapt resource value: 0x7F0F01BF
			public const int Theme_AppCompat_Light_DialogWhenLarge = 2131689919;
			
			// aapt resource value: 0x7F0F01BD
			public const int Theme_AppCompat_Light_Dialog_Alert = 2131689917;
			
			// aapt resource value: 0x7F0F01BE
			public const int Theme_AppCompat_Light_Dialog_MinWidth = 2131689918;
			
			// aapt resource value: 0x7F0F01C0
			public const int Theme_AppCompat_Light_NoActionBar = 2131689920;
			
			// aapt resource value: 0x7F0F01C1
			public const int Theme_AppCompat_NoActionBar = 2131689921;
			
			// aapt resource value: 0x7F0F01C2
			public const int Theme_Design = 2131689922;
			
			// aapt resource value: 0x7F0F01C3
			public const int Theme_Design_BottomSheetDialog = 2131689923;
			
			// aapt resource value: 0x7F0F01C4
			public const int Theme_Design_Light = 2131689924;
			
			// aapt resource value: 0x7F0F01C5
			public const int Theme_Design_Light_BottomSheetDialog = 2131689925;
			
			// aapt resource value: 0x7F0F01C6
			public const int Theme_Design_Light_NoActionBar = 2131689926;
			
			// aapt resource value: 0x7F0F01C7
			public const int Theme_Design_NoActionBar = 2131689927;
			
			// aapt resource value: 0x7F0F01C8
			public const int Theme_MaterialComponents = 2131689928;
			
			// aapt resource value: 0x7F0F01C9
			public const int Theme_MaterialComponents_BottomSheetDialog = 2131689929;
			
			// aapt resource value: 0x7F0F01CA
			public const int Theme_MaterialComponents_Bridge = 2131689930;
			
			// aapt resource value: 0x7F0F01CB
			public const int Theme_MaterialComponents_CompactMenu = 2131689931;
			
			// aapt resource value: 0x7F0F01CC
			public const int Theme_MaterialComponents_DayNight = 2131689932;
			
			// aapt resource value: 0x7F0F01CD
			public const int Theme_MaterialComponents_DayNight_BottomSheetDialog = 2131689933;
			
			// aapt resource value: 0x7F0F01CE
			public const int Theme_MaterialComponents_DayNight_Bridge = 2131689934;
			
			// aapt resource value: 0x7F0F01CF
			public const int Theme_MaterialComponents_DayNight_DarkActionBar = 2131689935;
			
			// aapt resource value: 0x7F0F01D0
			public const int Theme_MaterialComponents_DayNight_DarkActionBar_Bridge = 2131689936;
			
			// aapt resource value: 0x7F0F01D1
			public const int Theme_MaterialComponents_DayNight_Dialog = 2131689937;
			
			// aapt resource value: 0x7F0F01D9
			public const int Theme_MaterialComponents_DayNight_DialogWhenLarge = 2131689945;
			
			// aapt resource value: 0x7F0F01D2
			public const int Theme_MaterialComponents_DayNight_Dialog_Alert = 2131689938;
			
			// aapt resource value: 0x7F0F01D3
			public const int Theme_MaterialComponents_DayNight_Dialog_Alert_Bridge = 2131689939;
			
			// aapt resource value: 0x7F0F01D4
			public const int Theme_MaterialComponents_DayNight_Dialog_Bridge = 2131689940;
			
			// aapt resource value: 0x7F0F01D5
			public const int Theme_MaterialComponents_DayNight_Dialog_FixedSize = 2131689941;
			
			// aapt resource value: 0x7F0F01D6
			public const int Theme_MaterialComponents_DayNight_Dialog_FixedSize_Bridge = 2131689942;
			
			// aapt resource value: 0x7F0F01D7
			public const int Theme_MaterialComponents_DayNight_Dialog_MinWidth = 2131689943;
			
			// aapt resource value: 0x7F0F01D8
			public const int Theme_MaterialComponents_DayNight_Dialog_MinWidth_Bridge = 2131689944;
			
			// aapt resource value: 0x7F0F01DA
			public const int Theme_MaterialComponents_DayNight_NoActionBar = 2131689946;
			
			// aapt resource value: 0x7F0F01DB
			public const int Theme_MaterialComponents_DayNight_NoActionBar_Bridge = 2131689947;
			
			// aapt resource value: 0x7F0F01DC
			public const int Theme_MaterialComponents_Dialog = 2131689948;
			
			// aapt resource value: 0x7F0F01E4
			public const int Theme_MaterialComponents_DialogWhenLarge = 2131689956;
			
			// aapt resource value: 0x7F0F01DD
			public const int Theme_MaterialComponents_Dialog_Alert = 2131689949;
			
			// aapt resource value: 0x7F0F01DE
			public const int Theme_MaterialComponents_Dialog_Alert_Bridge = 2131689950;
			
			// aapt resource value: 0x7F0F01DF
			public const int Theme_MaterialComponents_Dialog_Bridge = 2131689951;
			
			// aapt resource value: 0x7F0F01E0
			public const int Theme_MaterialComponents_Dialog_FixedSize = 2131689952;
			
			// aapt resource value: 0x7F0F01E1
			public const int Theme_MaterialComponents_Dialog_FixedSize_Bridge = 2131689953;
			
			// aapt resource value: 0x7F0F01E2
			public const int Theme_MaterialComponents_Dialog_MinWidth = 2131689954;
			
			// aapt resource value: 0x7F0F01E3
			public const int Theme_MaterialComponents_Dialog_MinWidth_Bridge = 2131689955;
			
			// aapt resource value: 0x7F0F01E5
			public const int Theme_MaterialComponents_Light = 2131689957;
			
			// aapt resource value: 0x7F0F01E6
			public const int Theme_MaterialComponents_Light_BarSize = 2131689958;
			
			// aapt resource value: 0x7F0F01E7
			public const int Theme_MaterialComponents_Light_BottomSheetDialog = 2131689959;
			
			// aapt resource value: 0x7F0F01E8
			public const int Theme_MaterialComponents_Light_Bridge = 2131689960;
			
			// aapt resource value: 0x7F0F01E9
			public const int Theme_MaterialComponents_Light_DarkActionBar = 2131689961;
			
			// aapt resource value: 0x7F0F01EA
			public const int Theme_MaterialComponents_Light_DarkActionBar_Bridge = 2131689962;
			
			// aapt resource value: 0x7F0F01EB
			public const int Theme_MaterialComponents_Light_Dialog = 2131689963;
			
			// aapt resource value: 0x7F0F01F3
			public const int Theme_MaterialComponents_Light_DialogWhenLarge = 2131689971;
			
			// aapt resource value: 0x7F0F01EC
			public const int Theme_MaterialComponents_Light_Dialog_Alert = 2131689964;
			
			// aapt resource value: 0x7F0F01ED
			public const int Theme_MaterialComponents_Light_Dialog_Alert_Bridge = 2131689965;
			
			// aapt resource value: 0x7F0F01EE
			public const int Theme_MaterialComponents_Light_Dialog_Bridge = 2131689966;
			
			// aapt resource value: 0x7F0F01EF
			public const int Theme_MaterialComponents_Light_Dialog_FixedSize = 2131689967;
			
			// aapt resource value: 0x7F0F01F0
			public const int Theme_MaterialComponents_Light_Dialog_FixedSize_Bridge = 2131689968;
			
			// aapt resource value: 0x7F0F01F1
			public const int Theme_MaterialComponents_Light_Dialog_MinWidth = 2131689969;
			
			// aapt resource value: 0x7F0F01F2
			public const int Theme_MaterialComponents_Light_Dialog_MinWidth_Bridge = 2131689970;
			
			// aapt resource value: 0x7F0F01F4
			public const int Theme_MaterialComponents_Light_LargeTouch = 2131689972;
			
			// aapt resource value: 0x7F0F01F5
			public const int Theme_MaterialComponents_Light_NoActionBar = 2131689973;
			
			// aapt resource value: 0x7F0F01F6
			public const int Theme_MaterialComponents_Light_NoActionBar_Bridge = 2131689974;
			
			// aapt resource value: 0x7F0F01F7
			public const int Theme_MaterialComponents_NoActionBar = 2131689975;
			
			// aapt resource value: 0x7F0F01F8
			public const int Theme_MaterialComponents_NoActionBar_Bridge = 2131689976;
			
			// aapt resource value: 0x7F0F022A
			public const int Widget_AppCompat_ActionBar = 2131690026;
			
			// aapt resource value: 0x7F0F022B
			public const int Widget_AppCompat_ActionBar_Solid = 2131690027;
			
			// aapt resource value: 0x7F0F022C
			public const int Widget_AppCompat_ActionBar_TabBar = 2131690028;
			
			// aapt resource value: 0x7F0F022D
			public const int Widget_AppCompat_ActionBar_TabText = 2131690029;
			
			// aapt resource value: 0x7F0F022E
			public const int Widget_AppCompat_ActionBar_TabView = 2131690030;
			
			// aapt resource value: 0x7F0F022F
			public const int Widget_AppCompat_ActionButton = 2131690031;
			
			// aapt resource value: 0x7F0F0230
			public const int Widget_AppCompat_ActionButton_CloseMode = 2131690032;
			
			// aapt resource value: 0x7F0F0231
			public const int Widget_AppCompat_ActionButton_Overflow = 2131690033;
			
			// aapt resource value: 0x7F0F0232
			public const int Widget_AppCompat_ActionMode = 2131690034;
			
			// aapt resource value: 0x7F0F0233
			public const int Widget_AppCompat_ActivityChooserView = 2131690035;
			
			// aapt resource value: 0x7F0F0234
			public const int Widget_AppCompat_AutoCompleteTextView = 2131690036;
			
			// aapt resource value: 0x7F0F0235
			public const int Widget_AppCompat_Button = 2131690037;
			
			// aapt resource value: 0x7F0F023B
			public const int Widget_AppCompat_ButtonBar = 2131690043;
			
			// aapt resource value: 0x7F0F023C
			public const int Widget_AppCompat_ButtonBar_AlertDialog = 2131690044;
			
			// aapt resource value: 0x7F0F0236
			public const int Widget_AppCompat_Button_Borderless = 2131690038;
			
			// aapt resource value: 0x7F0F0237
			public const int Widget_AppCompat_Button_Borderless_Colored = 2131690039;
			
			// aapt resource value: 0x7F0F0238
			public const int Widget_AppCompat_Button_ButtonBar_AlertDialog = 2131690040;
			
			// aapt resource value: 0x7F0F0239
			public const int Widget_AppCompat_Button_Colored = 2131690041;
			
			// aapt resource value: 0x7F0F023A
			public const int Widget_AppCompat_Button_Small = 2131690042;
			
			// aapt resource value: 0x7F0F023D
			public const int Widget_AppCompat_CompoundButton_CheckBox = 2131690045;
			
			// aapt resource value: 0x7F0F023E
			public const int Widget_AppCompat_CompoundButton_RadioButton = 2131690046;
			
			// aapt resource value: 0x7F0F023F
			public const int Widget_AppCompat_CompoundButton_Switch = 2131690047;
			
			// aapt resource value: 0x7F0F0240
			public const int Widget_AppCompat_DrawerArrowToggle = 2131690048;
			
			// aapt resource value: 0x7F0F0241
			public const int Widget_AppCompat_DropDownItem_Spinner = 2131690049;
			
			// aapt resource value: 0x7F0F0242
			public const int Widget_AppCompat_EditText = 2131690050;
			
			// aapt resource value: 0x7F0F0243
			public const int Widget_AppCompat_ImageButton = 2131690051;
			
			// aapt resource value: 0x7F0F0244
			public const int Widget_AppCompat_Light_ActionBar = 2131690052;
			
			// aapt resource value: 0x7F0F0245
			public const int Widget_AppCompat_Light_ActionBar_Solid = 2131690053;
			
			// aapt resource value: 0x7F0F0246
			public const int Widget_AppCompat_Light_ActionBar_Solid_Inverse = 2131690054;
			
			// aapt resource value: 0x7F0F0247
			public const int Widget_AppCompat_Light_ActionBar_TabBar = 2131690055;
			
			// aapt resource value: 0x7F0F0248
			public const int Widget_AppCompat_Light_ActionBar_TabBar_Inverse = 2131690056;
			
			// aapt resource value: 0x7F0F0249
			public const int Widget_AppCompat_Light_ActionBar_TabText = 2131690057;
			
			// aapt resource value: 0x7F0F024A
			public const int Widget_AppCompat_Light_ActionBar_TabText_Inverse = 2131690058;
			
			// aapt resource value: 0x7F0F024B
			public const int Widget_AppCompat_Light_ActionBar_TabView = 2131690059;
			
			// aapt resource value: 0x7F0F024C
			public const int Widget_AppCompat_Light_ActionBar_TabView_Inverse = 2131690060;
			
			// aapt resource value: 0x7F0F024D
			public const int Widget_AppCompat_Light_ActionButton = 2131690061;
			
			// aapt resource value: 0x7F0F024E
			public const int Widget_AppCompat_Light_ActionButton_CloseMode = 2131690062;
			
			// aapt resource value: 0x7F0F024F
			public const int Widget_AppCompat_Light_ActionButton_Overflow = 2131690063;
			
			// aapt resource value: 0x7F0F0250
			public const int Widget_AppCompat_Light_ActionMode_Inverse = 2131690064;
			
			// aapt resource value: 0x7F0F0251
			public const int Widget_AppCompat_Light_ActivityChooserView = 2131690065;
			
			// aapt resource value: 0x7F0F0252
			public const int Widget_AppCompat_Light_AutoCompleteTextView = 2131690066;
			
			// aapt resource value: 0x7F0F0253
			public const int Widget_AppCompat_Light_DropDownItem_Spinner = 2131690067;
			
			// aapt resource value: 0x7F0F0254
			public const int Widget_AppCompat_Light_ListPopupWindow = 2131690068;
			
			// aapt resource value: 0x7F0F0255
			public const int Widget_AppCompat_Light_ListView_DropDown = 2131690069;
			
			// aapt resource value: 0x7F0F0256
			public const int Widget_AppCompat_Light_PopupMenu = 2131690070;
			
			// aapt resource value: 0x7F0F0257
			public const int Widget_AppCompat_Light_PopupMenu_Overflow = 2131690071;
			
			// aapt resource value: 0x7F0F0258
			public const int Widget_AppCompat_Light_SearchView = 2131690072;
			
			// aapt resource value: 0x7F0F0259
			public const int Widget_AppCompat_Light_Spinner_DropDown_ActionBar = 2131690073;
			
			// aapt resource value: 0x7F0F025A
			public const int Widget_AppCompat_ListMenuView = 2131690074;
			
			// aapt resource value: 0x7F0F025B
			public const int Widget_AppCompat_ListPopupWindow = 2131690075;
			
			// aapt resource value: 0x7F0F025C
			public const int Widget_AppCompat_ListView = 2131690076;
			
			// aapt resource value: 0x7F0F025D
			public const int Widget_AppCompat_ListView_DropDown = 2131690077;
			
			// aapt resource value: 0x7F0F025E
			public const int Widget_AppCompat_ListView_Menu = 2131690078;
			
			// aapt resource value: 0x7F0F025F
			public const int Widget_AppCompat_PopupMenu = 2131690079;
			
			// aapt resource value: 0x7F0F0260
			public const int Widget_AppCompat_PopupMenu_Overflow = 2131690080;
			
			// aapt resource value: 0x7F0F0261
			public const int Widget_AppCompat_PopupWindow = 2131690081;
			
			// aapt resource value: 0x7F0F0262
			public const int Widget_AppCompat_ProgressBar = 2131690082;
			
			// aapt resource value: 0x7F0F0263
			public const int Widget_AppCompat_ProgressBar_Horizontal = 2131690083;
			
			// aapt resource value: 0x7F0F0264
			public const int Widget_AppCompat_RatingBar = 2131690084;
			
			// aapt resource value: 0x7F0F0265
			public const int Widget_AppCompat_RatingBar_Indicator = 2131690085;
			
			// aapt resource value: 0x7F0F0266
			public const int Widget_AppCompat_RatingBar_Small = 2131690086;
			
			// aapt resource value: 0x7F0F0267
			public const int Widget_AppCompat_SearchView = 2131690087;
			
			// aapt resource value: 0x7F0F0268
			public const int Widget_AppCompat_SearchView_ActionBar = 2131690088;
			
			// aapt resource value: 0x7F0F0269
			public const int Widget_AppCompat_SeekBar = 2131690089;
			
			// aapt resource value: 0x7F0F026A
			public const int Widget_AppCompat_SeekBar_Discrete = 2131690090;
			
			// aapt resource value: 0x7F0F026B
			public const int Widget_AppCompat_Spinner = 2131690091;
			
			// aapt resource value: 0x7F0F026C
			public const int Widget_AppCompat_Spinner_DropDown = 2131690092;
			
			// aapt resource value: 0x7F0F026D
			public const int Widget_AppCompat_Spinner_DropDown_ActionBar = 2131690093;
			
			// aapt resource value: 0x7F0F026E
			public const int Widget_AppCompat_Spinner_Underlined = 2131690094;
			
			// aapt resource value: 0x7F0F026F
			public const int Widget_AppCompat_TextView = 2131690095;
			
			// aapt resource value: 0x7F0F0270
			public const int Widget_AppCompat_TextView_SpinnerItem = 2131690096;
			
			// aapt resource value: 0x7F0F0271
			public const int Widget_AppCompat_Toolbar = 2131690097;
			
			// aapt resource value: 0x7F0F0272
			public const int Widget_AppCompat_Toolbar_Button_Navigation = 2131690098;
			
			// aapt resource value: 0x7F0F0273
			public const int Widget_Compat_NotificationActionContainer = 2131690099;
			
			// aapt resource value: 0x7F0F0274
			public const int Widget_Compat_NotificationActionText = 2131690100;
			
			// aapt resource value: 0x7F0F0275
			public const int Widget_Design_AppBarLayout = 2131690101;
			
			// aapt resource value: 0x7F0F0276
			public const int Widget_Design_BottomNavigationView = 2131690102;
			
			// aapt resource value: 0x7F0F0277
			public const int Widget_Design_BottomSheet_Modal = 2131690103;
			
			// aapt resource value: 0x7F0F0278
			public const int Widget_Design_CollapsingToolbar = 2131690104;
			
			// aapt resource value: 0x7F0F0279
			public const int Widget_Design_FloatingActionButton = 2131690105;
			
			// aapt resource value: 0x7F0F027A
			public const int Widget_Design_NavigationView = 2131690106;
			
			// aapt resource value: 0x7F0F027B
			public const int Widget_Design_ScrimInsetsFrameLayout = 2131690107;
			
			// aapt resource value: 0x7F0F027C
			public const int Widget_Design_Snackbar = 2131690108;
			
			// aapt resource value: 0x7F0F027D
			public const int Widget_Design_TabLayout = 2131690109;
			
			// aapt resource value: 0x7F0F027E
			public const int Widget_Design_TextInputEditText = 2131690110;
			
			// aapt resource value: 0x7F0F027F
			public const int Widget_Design_TextInputLayout = 2131690111;
			
			// aapt resource value: 0x7F0F0280
			public const int Widget_MaterialComponents_ActionBar_Primary = 2131690112;
			
			// aapt resource value: 0x7F0F0281
			public const int Widget_MaterialComponents_ActionBar_PrimarySurface = 2131690113;
			
			// aapt resource value: 0x7F0F0282
			public const int Widget_MaterialComponents_ActionBar_Solid = 2131690114;
			
			// aapt resource value: 0x7F0F0283
			public const int Widget_MaterialComponents_ActionBar_Surface = 2131690115;
			
			// aapt resource value: 0x7F0F0284
			public const int Widget_MaterialComponents_AppBarLayout_Primary = 2131690116;
			
			// aapt resource value: 0x7F0F0285
			public const int Widget_MaterialComponents_AppBarLayout_PrimarySurface = 2131690117;
			
			// aapt resource value: 0x7F0F0286
			public const int Widget_MaterialComponents_AppBarLayout_Surface = 2131690118;
			
			// aapt resource value: 0x7F0F0287
			public const int Widget_MaterialComponents_AutoCompleteTextView_FilledBox = 2131690119;
			
			// aapt resource value: 0x7F0F0288
			public const int Widget_MaterialComponents_AutoCompleteTextView_FilledBox_Dense = 2131690120;
			
			// aapt resource value: 0x7F0F0289
			public const int Widget_MaterialComponents_AutoCompleteTextView_OutlinedBox = 2131690121;
			
			// aapt resource value: 0x7F0F028A
			public const int Widget_MaterialComponents_AutoCompleteTextView_OutlinedBox_Dense = 2131690122;
			
			// aapt resource value: 0x7F0F028B
			public const int Widget_MaterialComponents_Badge = 2131690123;
			
			// aapt resource value: 0x7F0F028C
			public const int Widget_MaterialComponents_BottomAppBar = 2131690124;
			
			// aapt resource value: 0x7F0F028D
			public const int Widget_MaterialComponents_BottomAppBar_Colored = 2131690125;
			
			// aapt resource value: 0x7F0F028E
			public const int Widget_MaterialComponents_BottomAppBar_PrimarySurface = 2131690126;
			
			// aapt resource value: 0x7F0F028F
			public const int Widget_MaterialComponents_BottomNavigationView = 2131690127;
			
			// aapt resource value: 0x7F0F0290
			public const int Widget_MaterialComponents_BottomNavigationView_Colored = 2131690128;
			
			// aapt resource value: 0x7F0F0291
			public const int Widget_MaterialComponents_BottomNavigationView_PrimarySurface = 2131690129;
			
			// aapt resource value: 0x7F0F0292
			public const int Widget_MaterialComponents_BottomSheet = 2131690130;
			
			// aapt resource value: 0x7F0F0293
			public const int Widget_MaterialComponents_BottomSheet_Modal = 2131690131;
			
			// aapt resource value: 0x7F0F0294
			public const int Widget_MaterialComponents_Button = 2131690132;
			
			// aapt resource value: 0x7F0F0295
			public const int Widget_MaterialComponents_Button_Icon = 2131690133;
			
			// aapt resource value: 0x7F0F0296
			public const int Widget_MaterialComponents_Button_OutlinedButton = 2131690134;
			
			// aapt resource value: 0x7F0F0297
			public const int Widget_MaterialComponents_Button_OutlinedButton_Icon = 2131690135;
			
			// aapt resource value: 0x7F0F0298
			public const int Widget_MaterialComponents_Button_TextButton = 2131690136;
			
			// aapt resource value: 0x7F0F0299
			public const int Widget_MaterialComponents_Button_TextButton_Dialog = 2131690137;
			
			// aapt resource value: 0x7F0F029A
			public const int Widget_MaterialComponents_Button_TextButton_Dialog_Flush = 2131690138;
			
			// aapt resource value: 0x7F0F029B
			public const int Widget_MaterialComponents_Button_TextButton_Dialog_Icon = 2131690139;
			
			// aapt resource value: 0x7F0F029C
			public const int Widget_MaterialComponents_Button_TextButton_Icon = 2131690140;
			
			// aapt resource value: 0x7F0F029D
			public const int Widget_MaterialComponents_Button_TextButton_Snackbar = 2131690141;
			
			// aapt resource value: 0x7F0F029E
			public const int Widget_MaterialComponents_Button_UnelevatedButton = 2131690142;
			
			// aapt resource value: 0x7F0F029F
			public const int Widget_MaterialComponents_Button_UnelevatedButton_Icon = 2131690143;
			
			// aapt resource value: 0x7F0F02A0
			public const int Widget_MaterialComponents_CardView = 2131690144;
			
			// aapt resource value: 0x7F0F02A1
			public const int Widget_MaterialComponents_CheckedTextView = 2131690145;
			
			// aapt resource value: 0x7F0F02A6
			public const int Widget_MaterialComponents_ChipGroup = 2131690150;
			
			// aapt resource value: 0x7F0F02A2
			public const int Widget_MaterialComponents_Chip_Action = 2131690146;
			
			// aapt resource value: 0x7F0F02A3
			public const int Widget_MaterialComponents_Chip_Choice = 2131690147;
			
			// aapt resource value: 0x7F0F02A4
			public const int Widget_MaterialComponents_Chip_Entry = 2131690148;
			
			// aapt resource value: 0x7F0F02A5
			public const int Widget_MaterialComponents_Chip_Filter = 2131690149;
			
			// aapt resource value: 0x7F0F02A7
			public const int Widget_MaterialComponents_CircularProgressIndicator = 2131690151;
			
			// aapt resource value: 0x7F0F02A8
			public const int Widget_MaterialComponents_CircularProgressIndicator_ExtraSmall = 2131690152;
			
			// aapt resource value: 0x7F0F02A9
			public const int Widget_MaterialComponents_CircularProgressIndicator_Medium = 2131690153;
			
			// aapt resource value: 0x7F0F02AA
			public const int Widget_MaterialComponents_CircularProgressIndicator_Small = 2131690154;
			
			// aapt resource value: 0x7F0F02AB
			public const int Widget_MaterialComponents_CollapsingToolbar = 2131690155;
			
			// aapt resource value: 0x7F0F02AC
			public const int Widget_MaterialComponents_CompoundButton_CheckBox = 2131690156;
			
			// aapt resource value: 0x7F0F02AD
			public const int Widget_MaterialComponents_CompoundButton_RadioButton = 2131690157;
			
			// aapt resource value: 0x7F0F02AE
			public const int Widget_MaterialComponents_CompoundButton_Switch = 2131690158;
			
			// aapt resource value: 0x7F0F02AF
			public const int Widget_MaterialComponents_ExtendedFloatingActionButton = 2131690159;
			
			// aapt resource value: 0x7F0F02B0
			public const int Widget_MaterialComponents_ExtendedFloatingActionButton_Icon = 2131690160;
			
			// aapt resource value: 0x7F0F02B1
			public const int Widget_MaterialComponents_FloatingActionButton = 2131690161;
			
			// aapt resource value: 0x7F0F02B2
			public const int Widget_MaterialComponents_Light_ActionBar_Solid = 2131690162;
			
			// aapt resource value: 0x7F0F02B3
			public const int Widget_MaterialComponents_LinearProgressIndicator = 2131690163;
			
			// aapt resource value: 0x7F0F02B4
			public const int Widget_MaterialComponents_MaterialButtonToggleGroup = 2131690164;
			
			// aapt resource value: 0x7F0F02B5
			public const int Widget_MaterialComponents_MaterialCalendar = 2131690165;
			
			// aapt resource value: 0x7F0F02B6
			public const int Widget_MaterialComponents_MaterialCalendar_Day = 2131690166;
			
			// aapt resource value: 0x7F0F02BA
			public const int Widget_MaterialComponents_MaterialCalendar_DayTextView = 2131690170;
			
			// aapt resource value: 0x7F0F02B7
			public const int Widget_MaterialComponents_MaterialCalendar_Day_Invalid = 2131690167;
			
			// aapt resource value: 0x7F0F02B8
			public const int Widget_MaterialComponents_MaterialCalendar_Day_Selected = 2131690168;
			
			// aapt resource value: 0x7F0F02B9
			public const int Widget_MaterialComponents_MaterialCalendar_Day_Today = 2131690169;
			
			// aapt resource value: 0x7F0F02BB
			public const int Widget_MaterialComponents_MaterialCalendar_Fullscreen = 2131690171;
			
			// aapt resource value: 0x7F0F02BC
			public const int Widget_MaterialComponents_MaterialCalendar_HeaderCancelButton = 2131690172;
			
			// aapt resource value: 0x7F0F02BD
			public const int Widget_MaterialComponents_MaterialCalendar_HeaderConfirmButton = 2131690173;
			
			// aapt resource value: 0x7F0F02BE
			public const int Widget_MaterialComponents_MaterialCalendar_HeaderDivider = 2131690174;
			
			// aapt resource value: 0x7F0F02BF
			public const int Widget_MaterialComponents_MaterialCalendar_HeaderLayout = 2131690175;
			
			// aapt resource value: 0x7F0F02C0
			public const int Widget_MaterialComponents_MaterialCalendar_HeaderSelection = 2131690176;
			
			// aapt resource value: 0x7F0F02C1
			public const int Widget_MaterialComponents_MaterialCalendar_HeaderSelection_Fullscreen = 2131690177;
			
			// aapt resource value: 0x7F0F02C2
			public const int Widget_MaterialComponents_MaterialCalendar_HeaderTitle = 2131690178;
			
			// aapt resource value: 0x7F0F02C3
			public const int Widget_MaterialComponents_MaterialCalendar_HeaderToggleButton = 2131690179;
			
			// aapt resource value: 0x7F0F02C4
			public const int Widget_MaterialComponents_MaterialCalendar_Item = 2131690180;
			
			// aapt resource value: 0x7F0F02C5
			public const int Widget_MaterialComponents_MaterialCalendar_MonthNavigationButton = 2131690181;
			
			// aapt resource value: 0x7F0F02C6
			public const int Widget_MaterialComponents_MaterialCalendar_MonthTextView = 2131690182;
			
			// aapt resource value: 0x7F0F02C7
			public const int Widget_MaterialComponents_MaterialCalendar_Year = 2131690183;
			
			// aapt resource value: 0x7F0F02CA
			public const int Widget_MaterialComponents_MaterialCalendar_YearNavigationButton = 2131690186;
			
			// aapt resource value: 0x7F0F02C8
			public const int Widget_MaterialComponents_MaterialCalendar_Year_Selected = 2131690184;
			
			// aapt resource value: 0x7F0F02C9
			public const int Widget_MaterialComponents_MaterialCalendar_Year_Today = 2131690185;
			
			// aapt resource value: 0x7F0F02CB
			public const int Widget_MaterialComponents_NavigationRailView = 2131690187;
			
			// aapt resource value: 0x7F0F02CC
			public const int Widget_MaterialComponents_NavigationRailView_Colored = 2131690188;
			
			// aapt resource value: 0x7F0F02CD
			public const int Widget_MaterialComponents_NavigationRailView_Colored_Compact = 2131690189;
			
			// aapt resource value: 0x7F0F02CE
			public const int Widget_MaterialComponents_NavigationRailView_Compact = 2131690190;
			
			// aapt resource value: 0x7F0F02CF
			public const int Widget_MaterialComponents_NavigationRailView_PrimarySurface = 2131690191;
			
			// aapt resource value: 0x7F0F02D0
			public const int Widget_MaterialComponents_NavigationView = 2131690192;
			
			// aapt resource value: 0x7F0F02D1
			public const int Widget_MaterialComponents_PopupMenu = 2131690193;
			
			// aapt resource value: 0x7F0F02D2
			public const int Widget_MaterialComponents_PopupMenu_ContextMenu = 2131690194;
			
			// aapt resource value: 0x7F0F02D3
			public const int Widget_MaterialComponents_PopupMenu_ListPopupWindow = 2131690195;
			
			// aapt resource value: 0x7F0F02D4
			public const int Widget_MaterialComponents_PopupMenu_Overflow = 2131690196;
			
			// aapt resource value: 0x7F0F02D5
			public const int Widget_MaterialComponents_ProgressIndicator = 2131690197;
			
			// aapt resource value: 0x7F0F02D6
			public const int Widget_MaterialComponents_ShapeableImageView = 2131690198;
			
			// aapt resource value: 0x7F0F02D7
			public const int Widget_MaterialComponents_Slider = 2131690199;
			
			// aapt resource value: 0x7F0F02D8
			public const int Widget_MaterialComponents_Snackbar = 2131690200;
			
			// aapt resource value: 0x7F0F02D9
			public const int Widget_MaterialComponents_Snackbar_FullWidth = 2131690201;
			
			// aapt resource value: 0x7F0F02DA
			public const int Widget_MaterialComponents_Snackbar_TextView = 2131690202;
			
			// aapt resource value: 0x7F0F02DB
			public const int Widget_MaterialComponents_TabLayout = 2131690203;
			
			// aapt resource value: 0x7F0F02DC
			public const int Widget_MaterialComponents_TabLayout_Colored = 2131690204;
			
			// aapt resource value: 0x7F0F02DD
			public const int Widget_MaterialComponents_TabLayout_PrimarySurface = 2131690205;
			
			// aapt resource value: 0x7F0F02DE
			public const int Widget_MaterialComponents_TextInputEditText_FilledBox = 2131690206;
			
			// aapt resource value: 0x7F0F02DF
			public const int Widget_MaterialComponents_TextInputEditText_FilledBox_Dense = 2131690207;
			
			// aapt resource value: 0x7F0F02E0
			public const int Widget_MaterialComponents_TextInputEditText_OutlinedBox = 2131690208;
			
			// aapt resource value: 0x7F0F02E1
			public const int Widget_MaterialComponents_TextInputEditText_OutlinedBox_Dense = 2131690209;
			
			// aapt resource value: 0x7F0F02E2
			public const int Widget_MaterialComponents_TextInputLayout_FilledBox = 2131690210;
			
			// aapt resource value: 0x7F0F02E3
			public const int Widget_MaterialComponents_TextInputLayout_FilledBox_Dense = 2131690211;
			
			// aapt resource value: 0x7F0F02E4
			public const int Widget_MaterialComponents_TextInputLayout_FilledBox_Dense_ExposedDropdownMenu = 2131690212;
			
			// aapt resource value: 0x7F0F02E5
			public const int Widget_MaterialComponents_TextInputLayout_FilledBox_ExposedDropdownMenu = 2131690213;
			
			// aapt resource value: 0x7F0F02E6
			public const int Widget_MaterialComponents_TextInputLayout_OutlinedBox = 2131690214;
			
			// aapt resource value: 0x7F0F02E7
			public const int Widget_MaterialComponents_TextInputLayout_OutlinedBox_Dense = 2131690215;
			
			// aapt resource value: 0x7F0F02E8
			public const int Widget_MaterialComponents_TextInputLayout_OutlinedBox_Dense_ExposedDropdownMenu = 2131690216;
			
			// aapt resource value: 0x7F0F02E9
			public const int Widget_MaterialComponents_TextInputLayout_OutlinedBox_ExposedDropdownMenu = 2131690217;
			
			// aapt resource value: 0x7F0F02EA
			public const int Widget_MaterialComponents_TextView = 2131690218;
			
			// aapt resource value: 0x7F0F02EB
			public const int Widget_MaterialComponents_TimePicker = 2131690219;
			
			// aapt resource value: 0x7F0F02EC
			public const int Widget_MaterialComponents_TimePicker_Button = 2131690220;
			
			// aapt resource value: 0x7F0F02ED
			public const int Widget_MaterialComponents_TimePicker_Clock = 2131690221;
			
			// aapt resource value: 0x7F0F02EE
			public const int Widget_MaterialComponents_TimePicker_Display = 2131690222;
			
			// aapt resource value: 0x7F0F02EF
			public const int Widget_MaterialComponents_TimePicker_Display_TextInputEditText = 2131690223;
			
			// aapt resource value: 0x7F0F02F0
			public const int Widget_MaterialComponents_TimePicker_ImageButton = 2131690224;
			
			// aapt resource value: 0x7F0F02F1
			public const int Widget_MaterialComponents_TimePicker_ImageButton_ShapeAppearance = 2131690225;
			
			// aapt resource value: 0x7F0F02F2
			public const int Widget_MaterialComponents_Toolbar = 2131690226;
			
			// aapt resource value: 0x7F0F02F3
			public const int Widget_MaterialComponents_Toolbar_Primary = 2131690227;
			
			// aapt resource value: 0x7F0F02F4
			public const int Widget_MaterialComponents_Toolbar_PrimarySurface = 2131690228;
			
			// aapt resource value: 0x7F0F02F5
			public const int Widget_MaterialComponents_Toolbar_Surface = 2131690229;
			
			// aapt resource value: 0x7F0F02F6
			public const int Widget_MaterialComponents_Tooltip = 2131690230;
			
			// aapt resource value: 0x7F0F02F7
			public const int Widget_Support_CoordinatorLayout = 2131690231;
			
			static Style()
			{
				global::Android.Runtime.ResourceIdManager.UpdateIdValues();
			}
			
			private Style()
			{
			}
		}
		
		public partial class Styleable
		{
			
			// aapt resource value: { 0x7F030047,0x7F03004E,0x7F03004F,0x7F0300F3,0x7F0300F4,0x7F0300F5,0x7F0300F6,0x7F0300F7,0x7F0300F8,0x7F03011E,0x7F03013A,0x7F03013B,0x7F030156,0x7F0301B5,0x7F0301BC,0x7F0301C2,0x7F0301C3,0x7F0301C6,0x7F0301D8,0x7F0301EC,0x7F030262,0x7F0302BC,0x7F0302F5,0x7F030306,0x7F030307,0x7F03036D,0x7F030371,0x7F0303E2,0x7F0303EE }
			public static int[] ActionBar = new int[] {
					2130903111,
					2130903118,
					2130903119,
					2130903283,
					2130903284,
					2130903285,
					2130903286,
					2130903287,
					2130903288,
					2130903326,
					2130903354,
					2130903355,
					2130903382,
					2130903477,
					2130903484,
					2130903490,
					2130903491,
					2130903494,
					2130903512,
					2130903532,
					2130903650,
					2130903740,
					2130903797,
					2130903814,
					2130903815,
					2130903917,
					2130903921,
					2130904034,
					2130904046};
			
			// aapt resource value: { 0x10100B3 }
			public static int[] ActionBarLayout = new int[] {
					16842931};
			
			// aapt resource value: 0
			public const int ActionBarLayout_android_layout_gravity = 0;
			
			// aapt resource value: 0
			public const int ActionBar_background = 0;
			
			// aapt resource value: 1
			public const int ActionBar_backgroundSplit = 1;
			
			// aapt resource value: 2
			public const int ActionBar_backgroundStacked = 2;
			
			// aapt resource value: 3
			public const int ActionBar_contentInsetEnd = 3;
			
			// aapt resource value: 4
			public const int ActionBar_contentInsetEndWithActions = 4;
			
			// aapt resource value: 5
			public const int ActionBar_contentInsetLeft = 5;
			
			// aapt resource value: 6
			public const int ActionBar_contentInsetRight = 6;
			
			// aapt resource value: 7
			public const int ActionBar_contentInsetStart = 7;
			
			// aapt resource value: 8
			public const int ActionBar_contentInsetStartWithNavigation = 8;
			
			// aapt resource value: 9
			public const int ActionBar_customNavigationLayout = 9;
			
			// aapt resource value: 10
			public const int ActionBar_displayOptions = 10;
			
			// aapt resource value: 11
			public const int ActionBar_divider = 11;
			
			// aapt resource value: 12
			public const int ActionBar_elevation = 12;
			
			// aapt resource value: 13
			public const int ActionBar_height = 13;
			
			// aapt resource value: 14
			public const int ActionBar_hideOnContentScroll = 14;
			
			// aapt resource value: 15
			public const int ActionBar_homeAsUpIndicator = 15;
			
			// aapt resource value: 16
			public const int ActionBar_homeLayout = 16;
			
			// aapt resource value: 17
			public const int ActionBar_icon = 17;
			
			// aapt resource value: 18
			public const int ActionBar_indeterminateProgressStyle = 18;
			
			// aapt resource value: 19
			public const int ActionBar_itemPadding = 19;
			
			// aapt resource value: 20
			public const int ActionBar_logo = 20;
			
			// aapt resource value: 21
			public const int ActionBar_navigationMode = 21;
			
			// aapt resource value: 22
			public const int ActionBar_popupTheme = 22;
			
			// aapt resource value: 23
			public const int ActionBar_progressBarPadding = 23;
			
			// aapt resource value: 24
			public const int ActionBar_progressBarStyle = 24;
			
			// aapt resource value: 25
			public const int ActionBar_subtitle = 25;
			
			// aapt resource value: 26
			public const int ActionBar_subtitleTextStyle = 26;
			
			// aapt resource value: 27
			public const int ActionBar_title = 27;
			
			// aapt resource value: 28
			public const int ActionBar_titleTextStyle = 28;
			
			// aapt resource value: { 0x101013F }
			public static int[] ActionMenuItemView = new int[] {
					16843071};
			
			// aapt resource value: 0
			public const int ActionMenuItemView_android_minWidth = 0;
			
			// aapt resource value: { 0xFFFFFFFF }
			public static int[] ActionMenuView = new int[] {
					-1};
			
			// aapt resource value: { 0x7F030047,0x7F03004E,0x7F0300CC,0x7F0301B5,0x7F030371,0x7F0303EE }
			public static int[] ActionMode = new int[] {
					2130903111,
					2130903118,
					2130903244,
					2130903477,
					2130903921,
					2130904046};
			
			// aapt resource value: 0
			public const int ActionMode_background = 0;
			
			// aapt resource value: 1
			public const int ActionMode_backgroundSplit = 1;
			
			// aapt resource value: 2
			public const int ActionMode_closeItemLayout = 2;
			
			// aapt resource value: 3
			public const int ActionMode_height = 3;
			
			// aapt resource value: 4
			public const int ActionMode_subtitleTextStyle = 4;
			
			// aapt resource value: 5
			public const int ActionMode_titleTextStyle = 5;
			
			// aapt resource value: { 0x7F030170,0x7F0301DE }
			public static int[] ActivityChooserView = new int[] {
					2130903408,
					2130903518};
			
			// aapt resource value: 0
			public const int ActivityChooserView_expandActivityOverflowButtonDrawable = 0;
			
			// aapt resource value: 1
			public const int ActivityChooserView_initialActivityCount = 1;
			
			// aapt resource value: { 0x1010003,0x7F030003,0x7F030122,0x7F030123,0x7F03039E }
			public static int[] ActivityNavigator = new int[] {
					16842755,
					2130903043,
					2130903330,
					2130903331,
					2130903966};
			
			// aapt resource value: 1
			public const int ActivityNavigator_action = 1;
			
			// aapt resource value: 0
			public const int ActivityNavigator_android_name = 0;
			
			// aapt resource value: 2
			public const int ActivityNavigator_data = 2;
			
			// aapt resource value: 3
			public const int ActivityNavigator_dataPattern = 3;
			
			// aapt resource value: 4
			public const int ActivityNavigator_targetPackage = 4;
			
			// aapt resource value: { 0x10100F2,0x7F030080,0x7F030081,0x7F030257,0x7F030258,0x7F0302B7,0x7F030342,0x7F030344 }
			public static int[] AlertDialog = new int[] {
					16842994,
					2130903168,
					2130903169,
					2130903639,
					2130903640,
					2130903735,
					2130903874,
					2130903876};
			
			// aapt resource value: 0
			public const int AlertDialog_android_layout = 0;
			
			// aapt resource value: 1
			public const int AlertDialog_buttonIconDimen = 1;
			
			// aapt resource value: 2
			public const int AlertDialog_buttonPanelSideLayout = 2;
			
			// aapt resource value: 3
			public const int AlertDialog_listItemLayout = 3;
			
			// aapt resource value: 4
			public const int AlertDialog_listLayout = 4;
			
			// aapt resource value: 5
			public const int AlertDialog_multiChoiceItemLayout = 5;
			
			// aapt resource value: 6
			public const int AlertDialog_showTitle = 6;
			
			// aapt resource value: 7
			public const int AlertDialog_singleChoiceItemLayout = 7;
			
			// aapt resource value: { 0x101011C,0x1010194,0x1010195,0x1010196,0x101030C,0x101030D }
			public static int[] AnimatedStateListDrawableCompat = new int[] {
					16843036,
					16843156,
					16843157,
					16843158,
					16843532,
					16843533};
			
			// aapt resource value: 3
			public const int AnimatedStateListDrawableCompat_android_constantSize = 3;
			
			// aapt resource value: 0
			public const int AnimatedStateListDrawableCompat_android_dither = 0;
			
			// aapt resource value: 4
			public const int AnimatedStateListDrawableCompat_android_enterFadeDuration = 4;
			
			// aapt resource value: 5
			public const int AnimatedStateListDrawableCompat_android_exitFadeDuration = 5;
			
			// aapt resource value: 2
			public const int AnimatedStateListDrawableCompat_android_variablePadding = 2;
			
			// aapt resource value: 1
			public const int AnimatedStateListDrawableCompat_android_visible = 1;
			
			// aapt resource value: { 0x10100D0,0x1010199 }
			public static int[] AnimatedStateListDrawableItem = new int[] {
					16842960,
					16843161};
			
			// aapt resource value: 1
			public const int AnimatedStateListDrawableItem_android_drawable = 1;
			
			// aapt resource value: 0
			public const int AnimatedStateListDrawableItem_android_id = 0;
			
			// aapt resource value: { 0x1010199,0x1010449,0x101044A,0x101044B }
			public static int[] AnimatedStateListDrawableTransition = new int[] {
					16843161,
					16843849,
					16843850,
					16843851};
			
			// aapt resource value: 0
			public const int AnimatedStateListDrawableTransition_android_drawable = 0;
			
			// aapt resource value: 2
			public const int AnimatedStateListDrawableTransition_android_fromId = 2;
			
			// aapt resource value: 3
			public const int AnimatedStateListDrawableTransition_android_reversible = 3;
			
			// aapt resource value: 1
			public const int AnimatedStateListDrawableTransition_android_toId = 1;
			
			// aapt resource value: { 0x10100D4,0x101048F,0x1010540,0x7F030156,0x7F030171,0x7F03024D,0x7F03024E,0x7F030367 }
			public static int[] AppBarLayout = new int[] {
					16842964,
					16843919,
					16844096,
					2130903382,
					2130903409,
					2130903629,
					2130903630,
					2130903911};
			
			// aapt resource value: { 0x7F030361,0x7F030362,0x7F030364,0x7F030365 }
			public static int[] AppBarLayoutStates = new int[] {
					2130903905,
					2130903906,
					2130903908,
					2130903909};
			
			// aapt resource value: 0
			public const int AppBarLayoutStates_state_collapsed = 0;
			
			// aapt resource value: 1
			public const int AppBarLayoutStates_state_collapsible = 1;
			
			// aapt resource value: 2
			public const int AppBarLayoutStates_state_liftable = 2;
			
			// aapt resource value: 3
			public const int AppBarLayoutStates_state_lifted = 3;
			
			// aapt resource value: 0
			public const int AppBarLayout_android_background = 0;
			
			// aapt resource value: 2
			public const int AppBarLayout_android_keyboardNavigationCluster = 2;
			
			// aapt resource value: 1
			public const int AppBarLayout_android_touchscreenBlocksFocus = 1;
			
			// aapt resource value: 3
			public const int AppBarLayout_elevation = 3;
			
			// aapt resource value: 4
			public const int AppBarLayout_expanded = 4;
			
			// aapt resource value: { 0x7F03024A,0x7F03024B }
			public static int[] AppBarLayout_Layout = new int[] {
					2130903626,
					2130903627};
			
			// aapt resource value: 0
			public const int AppBarLayout_Layout_layout_scrollFlags = 0;
			
			// aapt resource value: 1
			public const int AppBarLayout_Layout_layout_scrollInterpolator = 1;
			
			// aapt resource value: 5
			public const int AppBarLayout_liftOnScroll = 5;
			
			// aapt resource value: 6
			public const int AppBarLayout_liftOnScrollTargetViewId = 6;
			
			// aapt resource value: 7
			public const int AppBarLayout_statusBarForeground = 7;
			
			// aapt resource value: { 0x1010119,0x7F030357,0x7F0303E0,0x7F0303E1 }
			public static int[] AppCompatImageView = new int[] {
					16843033,
					2130903895,
					2130904032,
					2130904033};
			
			// aapt resource value: 0
			public const int AppCompatImageView_android_src = 0;
			
			// aapt resource value: 1
			public const int AppCompatImageView_srcCompat = 1;
			
			// aapt resource value: 2
			public const int AppCompatImageView_tint = 2;
			
			// aapt resource value: 3
			public const int AppCompatImageView_tintMode = 3;
			
			// aapt resource value: { 0x1010142,0x7F0303DC,0x7F0303DD,0x7F0303DE }
			public static int[] AppCompatSeekBar = new int[] {
					16843074,
					2130904028,
					2130904029,
					2130904030};
			
			// aapt resource value: 0
			public const int AppCompatSeekBar_android_thumb = 0;
			
			// aapt resource value: 1
			public const int AppCompatSeekBar_tickMark = 1;
			
			// aapt resource value: 2
			public const int AppCompatSeekBar_tickMarkTint = 2;
			
			// aapt resource value: 3
			public const int AppCompatSeekBar_tickMarkTintMode = 3;
			
			// aapt resource value: { 0x1010034,0x101016D,0x101016E,0x101016F,0x1010170,0x1010392,0x1010393 }
			public static int[] AppCompatTextHelper = new int[] {
					16842804,
					16843117,
					16843118,
					16843119,
					16843120,
					16843666,
					16843667};
			
			// aapt resource value: 2
			public const int AppCompatTextHelper_android_drawableBottom = 2;
			
			// aapt resource value: 6
			public const int AppCompatTextHelper_android_drawableEnd = 6;
			
			// aapt resource value: 3
			public const int AppCompatTextHelper_android_drawableLeft = 3;
			
			// aapt resource value: 4
			public const int AppCompatTextHelper_android_drawableRight = 4;
			
			// aapt resource value: 5
			public const int AppCompatTextHelper_android_drawableStart = 5;
			
			// aapt resource value: 1
			public const int AppCompatTextHelper_android_drawableTop = 1;
			
			// aapt resource value: 0
			public const int AppCompatTextHelper_android_textAppearance = 0;
			
			// aapt resource value: { 0x1010034,0x7F030041,0x7F030042,0x7F030043,0x7F030044,0x7F030045,0x7F030143,0x7F030144,0x7F030145,0x7F030146,0x7F030148,0x7F030149,0x7F03014A,0x7F03014B,0x7F030189,0x7F03019F,0x7F0301A8,0x7F030203,0x7F030250,0x7F0303A2,0x7F0303C4 }
			public static int[] AppCompatTextView = new int[] {
					16842804,
					2130903105,
					2130903106,
					2130903107,
					2130903108,
					2130903109,
					2130903363,
					2130903364,
					2130903365,
					2130903366,
					2130903368,
					2130903369,
					2130903370,
					2130903371,
					2130903433,
					2130903455,
					2130903464,
					2130903555,
					2130903632,
					2130903970,
					2130904004};
			
			// aapt resource value: 0
			public const int AppCompatTextView_android_textAppearance = 0;
			
			// aapt resource value: 1
			public const int AppCompatTextView_autoSizeMaxTextSize = 1;
			
			// aapt resource value: 2
			public const int AppCompatTextView_autoSizeMinTextSize = 2;
			
			// aapt resource value: 3
			public const int AppCompatTextView_autoSizePresetSizes = 3;
			
			// aapt resource value: 4
			public const int AppCompatTextView_autoSizeStepGranularity = 4;
			
			// aapt resource value: 5
			public const int AppCompatTextView_autoSizeTextType = 5;
			
			// aapt resource value: 6
			public const int AppCompatTextView_drawableBottomCompat = 6;
			
			// aapt resource value: 7
			public const int AppCompatTextView_drawableEndCompat = 7;
			
			// aapt resource value: 8
			public const int AppCompatTextView_drawableLeftCompat = 8;
			
			// aapt resource value: 9
			public const int AppCompatTextView_drawableRightCompat = 9;
			
			// aapt resource value: 10
			public const int AppCompatTextView_drawableStartCompat = 10;
			
			// aapt resource value: 11
			public const int AppCompatTextView_drawableTint = 11;
			
			// aapt resource value: 12
			public const int AppCompatTextView_drawableTintMode = 12;
			
			// aapt resource value: 13
			public const int AppCompatTextView_drawableTopCompat = 13;
			
			// aapt resource value: 14
			public const int AppCompatTextView_firstBaselineToTopHeight = 14;
			
			// aapt resource value: 15
			public const int AppCompatTextView_fontFamily = 15;
			
			// aapt resource value: 16
			public const int AppCompatTextView_fontVariationSettings = 16;
			
			// aapt resource value: 17
			public const int AppCompatTextView_lastBaselineToBottomHeight = 17;
			
			// aapt resource value: 18
			public const int AppCompatTextView_lineHeight = 18;
			
			// aapt resource value: 19
			public const int AppCompatTextView_textAllCaps = 19;
			
			// aapt resource value: 20
			public const int AppCompatTextView_textLocale = 20;
			
			// aapt resource value: { 0x1010057,0x10100AE,0x7F030004,0x7F030005,0x7F030006,0x7F030007,0x7F030008,0x7F030009,0x7F03000A,0x7F03000B,0x7F03000C,0x7F03000D,0x7F03000E,0x7F03000F,0x7F030010,0x7F030012,0x7F030013,0x7F030014,0x7F030015,0x7F030016,0x7F030017,0x7F030018,0x7F030019,0x7F03001A,0x7F03001B,0x7F03001C,0x7F03001D,0x7F03001E,0x7F03001F,0x7F030020,0x7F030021,0x7F030022,0x7F030023,0x7F030024,0x7F030028,0x7F03002A,0x7F03002B,0x7F03002C,0x7F03002D,0x7F030040,0x7F030068,0x7F030079,0x7F03007A,0x7F03007B,0x7F03007C,0x7F03007D,0x7F030082,0x7F030083,0x7F03009A,0x7F0300A3,0x7F0300D5,0x7F0300D6,0x7F0300D7,0x7F0300D8,0x7F0300D9,0x7F0300DA,0x7F0300DB,0x7F0300E2,0x7F0300E3,0x7F0300E9,0x7F030102,0x7F030131,0x7F030136,0x7F030137,0x7F03013C,0x7F03013E,0x7F03014E,0x7F03014F,0x7F030152,0x7F030153,0x7F030155,0x7F0301C2,0x7F0301D2,0x7F030253,0x7F030254,0x7F030255,0x7F030256,0x7F030259,0x7F03025A,0x7F03025B,0x7F03025C,0x7F03025D,0x7F03025E,0x7F03025F,0x7F030260,0x7F030261,0x7F0302D9,0x7F0302DA,0x7F0302DB,0x7F0302F4,0x7F0302F6,0x7F03030E,0x7F030310,0x7F030311,0x7F030312,0x7F030329,0x7F03032C,0x7F03032E,0x7F03032F,0x7F03034F,0x7F030350,0x7F03037E,0x7F0303AD,0x7F0303AF,0x7F0303B0,0x7F0303B1,0x7F0303B3,0x7F0303B4,0x7F0303B5,0x7F0303B6,0x7F0303BE,0x7F0303BF,0x7F0303F0,0x7F0303F1,0x7F0303F2,0x7F0303F3,0x7F030414,0x7F030423,0x7F030424,0x7F030425,0x7F030426,0x7F030427,0x7F030428,0x7F030429,0x7F03042A,0x7F03042B,0x7F03042C }
			public static int[] AppCompatTheme = new int[] {
					16842839,
					16842926,
					2130903044,
					2130903045,
					2130903046,
					2130903047,
					2130903048,
					2130903049,
					2130903050,
					2130903051,
					2130903052,
					2130903053,
					2130903054,
					2130903055,
					2130903056,
					2130903058,
					2130903059,
					2130903060,
					2130903061,
					2130903062,
					2130903063,
					2130903064,
					2130903065,
					2130903066,
					2130903067,
					2130903068,
					2130903069,
					2130903070,
					2130903071,
					2130903072,
					2130903073,
					2130903074,
					2130903075,
					2130903076,
					2130903080,
					2130903082,
					2130903083,
					2130903084,
					2130903085,
					2130903104,
					2130903144,
					2130903161,
					2130903162,
					2130903163,
					2130903164,
					2130903165,
					2130903170,
					2130903171,
					2130903194,
					2130903203,
					2130903253,
					2130903254,
					2130903255,
					2130903256,
					2130903257,
					2130903258,
					2130903259,
					2130903266,
					2130903267,
					2130903273,
					2130903298,
					2130903345,
					2130903350,
					2130903351,
					2130903356,
					2130903358,
					2130903374,
					2130903375,
					2130903378,
					2130903379,
					2130903381,
					2130903490,
					2130903506,
					2130903635,
					2130903636,
					2130903637,
					2130903638,
					2130903641,
					2130903642,
					2130903643,
					2130903644,
					2130903645,
					2130903646,
					2130903647,
					2130903648,
					2130903649,
					2130903769,
					2130903770,
					2130903771,
					2130903796,
					2130903798,
					2130903822,
					2130903824,
					2130903825,
					2130903826,
					2130903849,
					2130903852,
					2130903854,
					2130903855,
					2130903887,
					2130903888,
					2130903934,
					2130903981,
					2130903983,
					2130903984,
					2130903985,
					2130903987,
					2130903988,
					2130903989,
					2130903990,
					2130903998,
					2130903999,
					2130904048,
					2130904049,
					2130904050,
					2130904051,
					2130904084,
					2130904099,
					2130904100,
					2130904101,
					2130904102,
					2130904103,
					2130904104,
					2130904105,
					2130904106,
					2130904107,
					2130904108};
			
			// aapt resource value: 2
			public const int AppCompatTheme_actionBarDivider = 2;
			
			// aapt resource value: 3
			public const int AppCompatTheme_actionBarItemBackground = 3;
			
			// aapt resource value: 4
			public const int AppCompatTheme_actionBarPopupTheme = 4;
			
			// aapt resource value: 5
			public const int AppCompatTheme_actionBarSize = 5;
			
			// aapt resource value: 6
			public const int AppCompatTheme_actionBarSplitStyle = 6;
			
			// aapt resource value: 7
			public const int AppCompatTheme_actionBarStyle = 7;
			
			// aapt resource value: 8
			public const int AppCompatTheme_actionBarTabBarStyle = 8;
			
			// aapt resource value: 9
			public const int AppCompatTheme_actionBarTabStyle = 9;
			
			// aapt resource value: 10
			public const int AppCompatTheme_actionBarTabTextStyle = 10;
			
			// aapt resource value: 11
			public const int AppCompatTheme_actionBarTheme = 11;
			
			// aapt resource value: 12
			public const int AppCompatTheme_actionBarWidgetTheme = 12;
			
			// aapt resource value: 13
			public const int AppCompatTheme_actionButtonStyle = 13;
			
			// aapt resource value: 14
			public const int AppCompatTheme_actionDropDownStyle = 14;
			
			// aapt resource value: 15
			public const int AppCompatTheme_actionMenuTextAppearance = 15;
			
			// aapt resource value: 16
			public const int AppCompatTheme_actionMenuTextColor = 16;
			
			// aapt resource value: 17
			public const int AppCompatTheme_actionModeBackground = 17;
			
			// aapt resource value: 18
			public const int AppCompatTheme_actionModeCloseButtonStyle = 18;
			
			// aapt resource value: 19
			public const int AppCompatTheme_actionModeCloseContentDescription = 19;
			
			// aapt resource value: 20
			public const int AppCompatTheme_actionModeCloseDrawable = 20;
			
			// aapt resource value: 21
			public const int AppCompatTheme_actionModeCopyDrawable = 21;
			
			// aapt resource value: 22
			public const int AppCompatTheme_actionModeCutDrawable = 22;
			
			// aapt resource value: 23
			public const int AppCompatTheme_actionModeFindDrawable = 23;
			
			// aapt resource value: 24
			public const int AppCompatTheme_actionModePasteDrawable = 24;
			
			// aapt resource value: 25
			public const int AppCompatTheme_actionModePopupWindowStyle = 25;
			
			// aapt resource value: 26
			public const int AppCompatTheme_actionModeSelectAllDrawable = 26;
			
			// aapt resource value: 27
			public const int AppCompatTheme_actionModeShareDrawable = 27;
			
			// aapt resource value: 28
			public const int AppCompatTheme_actionModeSplitBackground = 28;
			
			// aapt resource value: 29
			public const int AppCompatTheme_actionModeStyle = 29;
			
			// aapt resource value: 30
			public const int AppCompatTheme_actionModeTheme = 30;
			
			// aapt resource value: 31
			public const int AppCompatTheme_actionModeWebSearchDrawable = 31;
			
			// aapt resource value: 32
			public const int AppCompatTheme_actionOverflowButtonStyle = 32;
			
			// aapt resource value: 33
			public const int AppCompatTheme_actionOverflowMenuStyle = 33;
			
			// aapt resource value: 34
			public const int AppCompatTheme_activityChooserViewStyle = 34;
			
			// aapt resource value: 35
			public const int AppCompatTheme_alertDialogButtonGroupStyle = 35;
			
			// aapt resource value: 36
			public const int AppCompatTheme_alertDialogCenterButtons = 36;
			
			// aapt resource value: 37
			public const int AppCompatTheme_alertDialogStyle = 37;
			
			// aapt resource value: 38
			public const int AppCompatTheme_alertDialogTheme = 38;
			
			// aapt resource value: 1
			public const int AppCompatTheme_android_windowAnimationStyle = 1;
			
			// aapt resource value: 0
			public const int AppCompatTheme_android_windowIsFloating = 0;
			
			// aapt resource value: 39
			public const int AppCompatTheme_autoCompleteTextViewStyle = 39;
			
			// aapt resource value: 40
			public const int AppCompatTheme_borderlessButtonStyle = 40;
			
			// aapt resource value: 41
			public const int AppCompatTheme_buttonBarButtonStyle = 41;
			
			// aapt resource value: 42
			public const int AppCompatTheme_buttonBarNegativeButtonStyle = 42;
			
			// aapt resource value: 43
			public const int AppCompatTheme_buttonBarNeutralButtonStyle = 43;
			
			// aapt resource value: 44
			public const int AppCompatTheme_buttonBarPositiveButtonStyle = 44;
			
			// aapt resource value: 45
			public const int AppCompatTheme_buttonBarStyle = 45;
			
			// aapt resource value: 46
			public const int AppCompatTheme_buttonStyle = 46;
			
			// aapt resource value: 47
			public const int AppCompatTheme_buttonStyleSmall = 47;
			
			// aapt resource value: 48
			public const int AppCompatTheme_checkboxStyle = 48;
			
			// aapt resource value: 49
			public const int AppCompatTheme_checkedTextViewStyle = 49;
			
			// aapt resource value: 50
			public const int AppCompatTheme_colorAccent = 50;
			
			// aapt resource value: 51
			public const int AppCompatTheme_colorBackgroundFloating = 51;
			
			// aapt resource value: 52
			public const int AppCompatTheme_colorButtonNormal = 52;
			
			// aapt resource value: 53
			public const int AppCompatTheme_colorControlActivated = 53;
			
			// aapt resource value: 54
			public const int AppCompatTheme_colorControlHighlight = 54;
			
			// aapt resource value: 55
			public const int AppCompatTheme_colorControlNormal = 55;
			
			// aapt resource value: 56
			public const int AppCompatTheme_colorError = 56;
			
			// aapt resource value: 57
			public const int AppCompatTheme_colorPrimary = 57;
			
			// aapt resource value: 58
			public const int AppCompatTheme_colorPrimaryDark = 58;
			
			// aapt resource value: 59
			public const int AppCompatTheme_colorSwitchThumbNormal = 59;
			
			// aapt resource value: 60
			public const int AppCompatTheme_controlBackground = 60;
			
			// aapt resource value: 61
			public const int AppCompatTheme_dialogCornerRadius = 61;
			
			// aapt resource value: 62
			public const int AppCompatTheme_dialogPreferredPadding = 62;
			
			// aapt resource value: 63
			public const int AppCompatTheme_dialogTheme = 63;
			
			// aapt resource value: 64
			public const int AppCompatTheme_dividerHorizontal = 64;
			
			// aapt resource value: 65
			public const int AppCompatTheme_dividerVertical = 65;
			
			// aapt resource value: 67
			public const int AppCompatTheme_dropdownListPreferredItemHeight = 67;
			
			// aapt resource value: 66
			public const int AppCompatTheme_dropDownListViewStyle = 66;
			
			// aapt resource value: 68
			public const int AppCompatTheme_editTextBackground = 68;
			
			// aapt resource value: 69
			public const int AppCompatTheme_editTextColor = 69;
			
			// aapt resource value: 70
			public const int AppCompatTheme_editTextStyle = 70;
			
			// aapt resource value: 71
			public const int AppCompatTheme_homeAsUpIndicator = 71;
			
			// aapt resource value: 72
			public const int AppCompatTheme_imageButtonStyle = 72;
			
			// aapt resource value: 73
			public const int AppCompatTheme_listChoiceBackgroundIndicator = 73;
			
			// aapt resource value: 74
			public const int AppCompatTheme_listChoiceIndicatorMultipleAnimated = 74;
			
			// aapt resource value: 75
			public const int AppCompatTheme_listChoiceIndicatorSingleAnimated = 75;
			
			// aapt resource value: 76
			public const int AppCompatTheme_listDividerAlertDialog = 76;
			
			// aapt resource value: 77
			public const int AppCompatTheme_listMenuViewStyle = 77;
			
			// aapt resource value: 78
			public const int AppCompatTheme_listPopupWindowStyle = 78;
			
			// aapt resource value: 79
			public const int AppCompatTheme_listPreferredItemHeight = 79;
			
			// aapt resource value: 80
			public const int AppCompatTheme_listPreferredItemHeightLarge = 80;
			
			// aapt resource value: 81
			public const int AppCompatTheme_listPreferredItemHeightSmall = 81;
			
			// aapt resource value: 82
			public const int AppCompatTheme_listPreferredItemPaddingEnd = 82;
			
			// aapt resource value: 83
			public const int AppCompatTheme_listPreferredItemPaddingLeft = 83;
			
			// aapt resource value: 84
			public const int AppCompatTheme_listPreferredItemPaddingRight = 84;
			
			// aapt resource value: 85
			public const int AppCompatTheme_listPreferredItemPaddingStart = 85;
			
			// aapt resource value: 86
			public const int AppCompatTheme_panelBackground = 86;
			
			// aapt resource value: 87
			public const int AppCompatTheme_panelMenuListTheme = 87;
			
			// aapt resource value: 88
			public const int AppCompatTheme_panelMenuListWidth = 88;
			
			// aapt resource value: 89
			public const int AppCompatTheme_popupMenuStyle = 89;
			
			// aapt resource value: 90
			public const int AppCompatTheme_popupWindowStyle = 90;
			
			// aapt resource value: 91
			public const int AppCompatTheme_radioButtonStyle = 91;
			
			// aapt resource value: 92
			public const int AppCompatTheme_ratingBarStyle = 92;
			
			// aapt resource value: 93
			public const int AppCompatTheme_ratingBarStyleIndicator = 93;
			
			// aapt resource value: 94
			public const int AppCompatTheme_ratingBarStyleSmall = 94;
			
			// aapt resource value: 95
			public const int AppCompatTheme_searchViewStyle = 95;
			
			// aapt resource value: 96
			public const int AppCompatTheme_seekBarStyle = 96;
			
			// aapt resource value: 97
			public const int AppCompatTheme_selectableItemBackground = 97;
			
			// aapt resource value: 98
			public const int AppCompatTheme_selectableItemBackgroundBorderless = 98;
			
			// aapt resource value: 99
			public const int AppCompatTheme_spinnerDropDownItemStyle = 99;
			
			// aapt resource value: 100
			public const int AppCompatTheme_spinnerStyle = 100;
			
			// aapt resource value: 101
			public const int AppCompatTheme_switchStyle = 101;
			
			// aapt resource value: 102
			public const int AppCompatTheme_textAppearanceLargePopupMenu = 102;
			
			// aapt resource value: 103
			public const int AppCompatTheme_textAppearanceListItem = 103;
			
			// aapt resource value: 104
			public const int AppCompatTheme_textAppearanceListItemSecondary = 104;
			
			// aapt resource value: 105
			public const int AppCompatTheme_textAppearanceListItemSmall = 105;
			
			// aapt resource value: 106
			public const int AppCompatTheme_textAppearancePopupMenuHeader = 106;
			
			// aapt resource value: 107
			public const int AppCompatTheme_textAppearanceSearchResultSubtitle = 107;
			
			// aapt resource value: 108
			public const int AppCompatTheme_textAppearanceSearchResultTitle = 108;
			
			// aapt resource value: 109
			public const int AppCompatTheme_textAppearanceSmallPopupMenu = 109;
			
			// aapt resource value: 110
			public const int AppCompatTheme_textColorAlertDialogListItem = 110;
			
			// aapt resource value: 111
			public const int AppCompatTheme_textColorSearchUrl = 111;
			
			// aapt resource value: 112
			public const int AppCompatTheme_toolbarNavigationButtonStyle = 112;
			
			// aapt resource value: 113
			public const int AppCompatTheme_toolbarStyle = 113;
			
			// aapt resource value: 114
			public const int AppCompatTheme_tooltipForegroundColor = 114;
			
			// aapt resource value: 115
			public const int AppCompatTheme_tooltipFrameBackground = 115;
			
			// aapt resource value: 116
			public const int AppCompatTheme_viewInflaterClass = 116;
			
			// aapt resource value: 117
			public const int AppCompatTheme_windowActionBar = 117;
			
			// aapt resource value: 118
			public const int AppCompatTheme_windowActionBarOverlay = 118;
			
			// aapt resource value: 119
			public const int AppCompatTheme_windowActionModeOverlay = 119;
			
			// aapt resource value: 120
			public const int AppCompatTheme_windowFixedHeightMajor = 120;
			
			// aapt resource value: 121
			public const int AppCompatTheme_windowFixedHeightMinor = 121;
			
			// aapt resource value: 122
			public const int AppCompatTheme_windowFixedWidthMajor = 122;
			
			// aapt resource value: 123
			public const int AppCompatTheme_windowFixedWidthMinor = 123;
			
			// aapt resource value: 124
			public const int AppCompatTheme_windowMinWidthMajor = 124;
			
			// aapt resource value: 125
			public const int AppCompatTheme_windowMinWidthMinor = 125;
			
			// aapt resource value: 126
			public const int AppCompatTheme_windowNoTitle = 126;
			
			// aapt resource value: { 0x101030E,0x7F03032E }
			public static int[] BackgroundStyle = new int[] {
					16843534,
					2130903854};
			
			// aapt resource value: 0
			public const int BackgroundStyle_android_selectableItemBackground = 0;
			
			// aapt resource value: 1
			public const int BackgroundStyle_selectableItemBackground = 1;
			
			// aapt resource value: { 0x7F030048,0x7F030052,0x7F030054,0x7F0301C4,0x7F030283,0x7F0302C4,0x7F030413 }
			public static int[] Badge = new int[] {
					2130903112,
					2130903122,
					2130903124,
					2130903492,
					2130903683,
					2130903748,
					2130904083};
			
			// aapt resource value: 0
			public const int Badge_backgroundColor = 0;
			
			// aapt resource value: 1
			public const int Badge_badgeGravity = 1;
			
			// aapt resource value: 2
			public const int Badge_badgeTextColor = 2;
			
			// aapt resource value: 3
			public const int Badge_horizontalOffset = 3;
			
			// aapt resource value: 4
			public const int Badge_maxCharacterCount = 4;
			
			// aapt resource value: 5
			public const int Badge_number = 5;
			
			// aapt resource value: 6
			public const int Badge_verticalOffset = 6;
			
			// aapt resource value: { 0x1010139,0x7F0301BA,0x7F0301D9,0x7F030290,0x7F03033A,0x7F03033C,0x7F0303FA,0x7F0303FD,0x7F0303FF }
			public static int[] BaseProgressIndicator = new int[] {
					16843065,
					2130903482,
					2130903513,
					2130903696,
					2130903866,
					2130903868,
					2130904058,
					2130904061,
					2130904063};
			
			// aapt resource value: 0
			public const int BaseProgressIndicator_android_indeterminate = 0;
			
			// aapt resource value: 1
			public const int BaseProgressIndicator_hideAnimationBehavior = 1;
			
			// aapt resource value: 2
			public const int BaseProgressIndicator_indicatorColor = 2;
			
			// aapt resource value: 3
			public const int BaseProgressIndicator_minHideDelay = 3;
			
			// aapt resource value: 4
			public const int BaseProgressIndicator_showAnimationBehavior = 4;
			
			// aapt resource value: 5
			public const int BaseProgressIndicator_showDelay = 5;
			
			// aapt resource value: 6
			public const int BaseProgressIndicator_trackColor = 6;
			
			// aapt resource value: 7
			public const int BaseProgressIndicator_trackCornerRadius = 7;
			
			// aapt resource value: 8
			public const int BaseProgressIndicator_trackThickness = 8;
			
			// aapt resource value: { 0x7F030050,0x7F030156,0x7F03017D,0x7F03017E,0x7F03017F,0x7F030180,0x7F030181,0x7F0301BD,0x7F0302D2,0x7F0302D4,0x7F0302D5 }
			public static int[] BottomAppBar = new int[] {
					2130903120,
					2130903382,
					2130903421,
					2130903422,
					2130903423,
					2130903424,
					2130903425,
					2130903485,
					2130903762,
					2130903764,
					2130903765};
			
			// aapt resource value: 0
			public const int BottomAppBar_backgroundTint = 0;
			
			// aapt resource value: 1
			public const int BottomAppBar_elevation = 1;
			
			// aapt resource value: 2
			public const int BottomAppBar_fabAlignmentMode = 2;
			
			// aapt resource value: 3
			public const int BottomAppBar_fabAnimationMode = 3;
			
			// aapt resource value: 4
			public const int BottomAppBar_fabCradleMargin = 4;
			
			// aapt resource value: 5
			public const int BottomAppBar_fabCradleRoundedCornerRadius = 5;
			
			// aapt resource value: 6
			public const int BottomAppBar_fabCradleVerticalOffset = 6;
			
			// aapt resource value: 7
			public const int BottomAppBar_hideOnScroll = 7;
			
			// aapt resource value: 8
			public const int BottomAppBar_paddingBottomSystemWindowInsets = 8;
			
			// aapt resource value: 9
			public const int BottomAppBar_paddingLeftSystemWindowInsets = 9;
			
			// aapt resource value: 10
			public const int BottomAppBar_paddingRightSystemWindowInsets = 10;
			
			// aapt resource value: { 0x7F0301E7 }
			public static int[] BottomNavigationView = new int[] {
					2130903527};
			
			// aapt resource value: 0
			public const int BottomNavigationView_itemHorizontalTranslationEnabled = 0;
			
			// aapt resource value: { 0x101011F,0x1010440,0x7F030050,0x7F03005B,0x7F03005C,0x7F03005D,0x7F03005E,0x7F03005F,0x7F030061,0x7F030062,0x7F030063,0x7F0301AF,0x7F0302D2,0x7F0302D4,0x7F0302D5,0x7F0302D8,0x7F030333,0x7F030336 }
			public static int[] BottomSheetBehavior_Layout = new int[] {
					16843039,
					16843840,
					2130903120,
					2130903131,
					2130903132,
					2130903133,
					2130903134,
					2130903135,
					2130903137,
					2130903138,
					2130903139,
					2130903471,
					2130903762,
					2130903764,
					2130903765,
					2130903768,
					2130903859,
					2130903862};
			
			// aapt resource value: 1
			public const int BottomSheetBehavior_Layout_android_elevation = 1;
			
			// aapt resource value: 0
			public const int BottomSheetBehavior_Layout_android_maxWidth = 0;
			
			// aapt resource value: 2
			public const int BottomSheetBehavior_Layout_backgroundTint = 2;
			
			// aapt resource value: 3
			public const int BottomSheetBehavior_Layout_behavior_draggable = 3;
			
			// aapt resource value: 4
			public const int BottomSheetBehavior_Layout_behavior_expandedOffset = 4;
			
			// aapt resource value: 5
			public const int BottomSheetBehavior_Layout_behavior_fitToContents = 5;
			
			// aapt resource value: 6
			public const int BottomSheetBehavior_Layout_behavior_halfExpandedRatio = 6;
			
			// aapt resource value: 7
			public const int BottomSheetBehavior_Layout_behavior_hideable = 7;
			
			// aapt resource value: 8
			public const int BottomSheetBehavior_Layout_behavior_peekHeight = 8;
			
			// aapt resource value: 9
			public const int BottomSheetBehavior_Layout_behavior_saveFlags = 9;
			
			// aapt resource value: 10
			public const int BottomSheetBehavior_Layout_behavior_skipCollapsed = 10;
			
			// aapt resource value: 11
			public const int BottomSheetBehavior_Layout_gestureInsetBottomIgnored = 11;
			
			// aapt resource value: 12
			public const int BottomSheetBehavior_Layout_paddingBottomSystemWindowInsets = 12;
			
			// aapt resource value: 13
			public const int BottomSheetBehavior_Layout_paddingLeftSystemWindowInsets = 13;
			
			// aapt resource value: 14
			public const int BottomSheetBehavior_Layout_paddingRightSystemWindowInsets = 14;
			
			// aapt resource value: 15
			public const int BottomSheetBehavior_Layout_paddingTopSystemWindowInsets = 15;
			
			// aapt resource value: 16
			public const int BottomSheetBehavior_Layout_shapeAppearance = 16;
			
			// aapt resource value: 17
			public const int BottomSheetBehavior_Layout_shapeAppearanceOverlay = 17;
			
			// aapt resource value: { 0x7F030031 }
			public static int[] ButtonBarLayout = new int[] {
					2130903089};
			
			// aapt resource value: 0
			public const int ButtonBarLayout_allowStacking = 0;
			
			// aapt resource value: { 0x7F03030D,0x7F030338 }
			public static int[] Capability = new int[] {
					2130903821,
					2130903864};
			
			// aapt resource value: 0
			public const int Capability_queryPatterns = 0;
			
			// aapt resource value: 1
			public const int Capability_shortcutMatchRequired = 1;
			
			// aapt resource value: { 0x101013F,0x1010140,0x7F030086,0x7F030087,0x7F030088,0x7F03008A,0x7F03008B,0x7F03008C,0x7F0300F9,0x7F0300FA,0x7F0300FC,0x7F0300FD,0x7F0300FF }
			public static int[] CardView = new int[] {
					16843071,
					16843072,
					2130903174,
					2130903175,
					2130903176,
					2130903178,
					2130903179,
					2130903180,
					2130903289,
					2130903290,
					2130903292,
					2130903293,
					2130903295};
			
			// aapt resource value: 1
			public const int CardView_android_minHeight = 1;
			
			// aapt resource value: 0
			public const int CardView_android_minWidth = 0;
			
			// aapt resource value: 2
			public const int CardView_cardBackgroundColor = 2;
			
			// aapt resource value: 3
			public const int CardView_cardCornerRadius = 3;
			
			// aapt resource value: 4
			public const int CardView_cardElevation = 4;
			
			// aapt resource value: 5
			public const int CardView_cardMaxElevation = 5;
			
			// aapt resource value: 6
			public const int CardView_cardPreventCornerOverlap = 6;
			
			// aapt resource value: 7
			public const int CardView_cardUseCompatPadding = 7;
			
			// aapt resource value: 8
			public const int CardView_contentPadding = 8;
			
			// aapt resource value: 9
			public const int CardView_contentPaddingBottom = 9;
			
			// aapt resource value: 10
			public const int CardView_contentPaddingLeft = 10;
			
			// aapt resource value: 11
			public const int CardView_contentPaddingRight = 11;
			
			// aapt resource value: 12
			public const int CardView_contentPaddingTop = 12;
			
			// aapt resource value: { 0x7F03008E,0x7F03008F,0x7F030090,0x7F030091,0x7F030092,0x7F030093,0x7F030094,0x7F030095,0x7F030096,0x7F030097 }
			public static int[] Carousel = new int[] {
					2130903182,
					2130903183,
					2130903184,
					2130903185,
					2130903186,
					2130903187,
					2130903188,
					2130903189,
					2130903190,
					2130903191};
			
			// aapt resource value: 0
			public const int Carousel_carousel_backwardTransition = 0;
			
			// aapt resource value: 1
			public const int Carousel_carousel_emptyViewsBehavior = 1;
			
			// aapt resource value: 2
			public const int Carousel_carousel_firstView = 2;
			
			// aapt resource value: 3
			public const int Carousel_carousel_forwardTransition = 3;
			
			// aapt resource value: 4
			public const int Carousel_carousel_infinite = 4;
			
			// aapt resource value: 5
			public const int Carousel_carousel_nextState = 5;
			
			// aapt resource value: 6
			public const int Carousel_carousel_previousState = 6;
			
			// aapt resource value: 7
			public const int Carousel_carousel_touchUpMode = 7;
			
			// aapt resource value: 8
			public const int Carousel_carousel_touchUp_dampeningFactor = 8;
			
			// aapt resource value: 9
			public const int Carousel_carousel_touchUp_velocityThreshold = 9;
			
			// aapt resource value: { 0x10101EF,0x10101F0,0x10101F1,0x7F030139,0x7F030377,0x7F030378 }
			public static int[] CheckBoxPreference = new int[] {
					16843247,
					16843248,
					16843249,
					2130903353,
					2130903927,
					2130903928};
			
			// aapt resource value: 2
			public const int CheckBoxPreference_android_disableDependentsState = 2;
			
			// aapt resource value: 1
			public const int CheckBoxPreference_android_summaryOff = 1;
			
			// aapt resource value: 0
			public const int CheckBoxPreference_android_summaryOn = 0;
			
			// aapt resource value: 3
			public const int CheckBoxPreference_disableDependentsState = 3;
			
			// aapt resource value: 4
			public const int CheckBoxPreference_summaryOff = 4;
			
			// aapt resource value: 5
			public const int CheckBoxPreference_summaryOn = 5;
			
			// aapt resource value: { 0x1010034,0x1010095,0x1010098,0x10100AB,0x101011F,0x101014F,0x10101E5,0x7F03009D,0x7F03009E,0x7F0300A1,0x7F0300A2,0x7F0300A4,0x7F0300A5,0x7F0300A6,0x7F0300A8,0x7F0300A9,0x7F0300AA,0x7F0300AB,0x7F0300AC,0x7F0300AD,0x7F0300AE,0x7F0300B3,0x7F0300B4,0x7F0300B5,0x7F0300B7,0x7F0300C5,0x7F0300C6,0x7F0300C7,0x7F0300C8,0x7F0300C9,0x7F0300CA,0x7F0300CB,0x7F030164,0x7F0301BB,0x7F0301C7,0x7F0301CC,0x7F03031D,0x7F030333,0x7F030336,0x7F03033E,0x7F0303C0,0x7F0303C9 }
			public static int[] Chip = new int[] {
					16842804,
					16842901,
					16842904,
					16842923,
					16843039,
					16843087,
					16843237,
					2130903197,
					2130903198,
					2130903201,
					2130903202,
					2130903204,
					2130903205,
					2130903206,
					2130903208,
					2130903209,
					2130903210,
					2130903211,
					2130903212,
					2130903213,
					2130903214,
					2130903219,
					2130903220,
					2130903221,
					2130903223,
					2130903237,
					2130903238,
					2130903239,
					2130903240,
					2130903241,
					2130903242,
					2130903243,
					2130903396,
					2130903483,
					2130903495,
					2130903500,
					2130903837,
					2130903859,
					2130903862,
					2130903870,
					2130904000,
					2130904009};
			
			// aapt resource value: { 0x7F03009C,0x7F0300AF,0x7F0300B0,0x7F0300B1,0x7F030330,0x7F030345,0x7F030347 }
			public static int[] ChipGroup = new int[] {
					2130903196,
					2130903215,
					2130903216,
					2130903217,
					2130903856,
					2130903877,
					2130903879};
			
			// aapt resource value: 0
			public const int ChipGroup_checkedChip = 0;
			
			// aapt resource value: 1
			public const int ChipGroup_chipSpacing = 1;
			
			// aapt resource value: 2
			public const int ChipGroup_chipSpacingHorizontal = 2;
			
			// aapt resource value: 3
			public const int ChipGroup_chipSpacingVertical = 3;
			
			// aapt resource value: 4
			public const int ChipGroup_selectionRequired = 4;
			
			// aapt resource value: 5
			public const int ChipGroup_singleLine = 5;
			
			// aapt resource value: 6
			public const int ChipGroup_singleSelection = 6;
			
			// aapt resource value: 6
			public const int Chip_android_checkable = 6;
			
			// aapt resource value: 3
			public const int Chip_android_ellipsize = 3;
			
			// aapt resource value: 4
			public const int Chip_android_maxWidth = 4;
			
			// aapt resource value: 5
			public const int Chip_android_text = 5;
			
			// aapt resource value: 0
			public const int Chip_android_textAppearance = 0;
			
			// aapt resource value: 2
			public const int Chip_android_textColor = 2;
			
			// aapt resource value: 1
			public const int Chip_android_textSize = 1;
			
			// aapt resource value: 7
			public const int Chip_checkedIcon = 7;
			
			// aapt resource value: 8
			public const int Chip_checkedIconEnabled = 8;
			
			// aapt resource value: 9
			public const int Chip_checkedIconTint = 9;
			
			// aapt resource value: 10
			public const int Chip_checkedIconVisible = 10;
			
			// aapt resource value: 11
			public const int Chip_chipBackgroundColor = 11;
			
			// aapt resource value: 12
			public const int Chip_chipCornerRadius = 12;
			
			// aapt resource value: 13
			public const int Chip_chipEndPadding = 13;
			
			// aapt resource value: 14
			public const int Chip_chipIcon = 14;
			
			// aapt resource value: 15
			public const int Chip_chipIconEnabled = 15;
			
			// aapt resource value: 16
			public const int Chip_chipIconSize = 16;
			
			// aapt resource value: 17
			public const int Chip_chipIconTint = 17;
			
			// aapt resource value: 18
			public const int Chip_chipIconVisible = 18;
			
			// aapt resource value: 19
			public const int Chip_chipMinHeight = 19;
			
			// aapt resource value: 20
			public const int Chip_chipMinTouchTargetSize = 20;
			
			// aapt resource value: 21
			public const int Chip_chipStartPadding = 21;
			
			// aapt resource value: 22
			public const int Chip_chipStrokeColor = 22;
			
			// aapt resource value: 23
			public const int Chip_chipStrokeWidth = 23;
			
			// aapt resource value: 24
			public const int Chip_chipSurfaceColor = 24;
			
			// aapt resource value: 25
			public const int Chip_closeIcon = 25;
			
			// aapt resource value: 26
			public const int Chip_closeIconEnabled = 26;
			
			// aapt resource value: 27
			public const int Chip_closeIconEndPadding = 27;
			
			// aapt resource value: 28
			public const int Chip_closeIconSize = 28;
			
			// aapt resource value: 29
			public const int Chip_closeIconStartPadding = 29;
			
			// aapt resource value: 30
			public const int Chip_closeIconTint = 30;
			
			// aapt resource value: 31
			public const int Chip_closeIconVisible = 31;
			
			// aapt resource value: 32
			public const int Chip_ensureMinTouchTargetSize = 32;
			
			// aapt resource value: 33
			public const int Chip_hideMotionSpec = 33;
			
			// aapt resource value: 34
			public const int Chip_iconEndPadding = 34;
			
			// aapt resource value: 35
			public const int Chip_iconStartPadding = 35;
			
			// aapt resource value: 36
			public const int Chip_rippleColor = 36;
			
			// aapt resource value: 37
			public const int Chip_shapeAppearance = 37;
			
			// aapt resource value: 38
			public const int Chip_shapeAppearanceOverlay = 38;
			
			// aapt resource value: 39
			public const int Chip_showMotionSpec = 39;
			
			// aapt resource value: 40
			public const int Chip_textEndPadding = 40;
			
			// aapt resource value: 41
			public const int Chip_textStartPadding = 41;
			
			// aapt resource value: { 0x7F0301DA,0x7F0301DC,0x7F0301DD }
			public static int[] CircularProgressIndicator = new int[] {
					2130903514,
					2130903516,
					2130903517};
			
			// aapt resource value: 0
			public const int CircularProgressIndicator_indicatorDirectionCircular = 0;
			
			// aapt resource value: 1
			public const int CircularProgressIndicator_indicatorInset = 1;
			
			// aapt resource value: 2
			public const int CircularProgressIndicator_indicatorSize = 2;
			
			// aapt resource value: { 0x7F0300C1,0x7F0300C4 }
			public static int[] ClockFaceView = new int[] {
					2130903233,
					2130903236};
			
			// aapt resource value: 0
			public const int ClockFaceView_clockFaceBackgroundColor = 0;
			
			// aapt resource value: 1
			public const int ClockFaceView_clockNumberTextColor = 1;
			
			// aapt resource value: { 0x7F0300C2,0x7F03027B,0x7F030331 }
			public static int[] ClockHandView = new int[] {
					2130903234,
					2130903675,
					2130903857};
			
			// aapt resource value: 0
			public const int ClockHandView_clockHandColor = 0;
			
			// aapt resource value: 1
			public const int ClockHandView_materialCircleRadius = 1;
			
			// aapt resource value: 2
			public const int ClockHandView_selectorSize = 2;
			
			// aapt resource value: { 0x7F0300D0,0x7F0300D1,0x7F030100,0x7F030173,0x7F030174,0x7F030175,0x7F030176,0x7F030177,0x7F030178,0x7F030179,0x7F03017C,0x7F0301AA,0x7F030286,0x7F030323,0x7F030325,0x7F030368,0x7F0303E2,0x7F0303E4,0x7F0303E5,0x7F0303EF }
			public static int[] CollapsingToolbarLayout = new int[] {
					2130903248,
					2130903249,
					2130903296,
					2130903411,
					2130903412,
					2130903413,
					2130903414,
					2130903415,
					2130903416,
					2130903417,
					2130903420,
					2130903466,
					2130903686,
					2130903843,
					2130903845,
					2130903912,
					2130904034,
					2130904036,
					2130904037,
					2130904047};
			
			// aapt resource value: 0
			public const int CollapsingToolbarLayout_collapsedTitleGravity = 0;
			
			// aapt resource value: 1
			public const int CollapsingToolbarLayout_collapsedTitleTextAppearance = 1;
			
			// aapt resource value: 2
			public const int CollapsingToolbarLayout_contentScrim = 2;
			
			// aapt resource value: 3
			public const int CollapsingToolbarLayout_expandedTitleGravity = 3;
			
			// aapt resource value: 4
			public const int CollapsingToolbarLayout_expandedTitleMargin = 4;
			
			// aapt resource value: 5
			public const int CollapsingToolbarLayout_expandedTitleMarginBottom = 5;
			
			// aapt resource value: 6
			public const int CollapsingToolbarLayout_expandedTitleMarginEnd = 6;
			
			// aapt resource value: 7
			public const int CollapsingToolbarLayout_expandedTitleMarginStart = 7;
			
			// aapt resource value: 8
			public const int CollapsingToolbarLayout_expandedTitleMarginTop = 8;
			
			// aapt resource value: 9
			public const int CollapsingToolbarLayout_expandedTitleTextAppearance = 9;
			
			// aapt resource value: 10
			public const int CollapsingToolbarLayout_extraMultilineHeightEnabled = 10;
			
			// aapt resource value: 11
			public const int CollapsingToolbarLayout_forceApplySystemWindowInsetTop = 11;
			
			// aapt resource value: { 0x7F03020C,0x7F03020D }
			public static int[] CollapsingToolbarLayout_Layout = new int[] {
					2130903564,
					2130903565};
			
			// aapt resource value: 0
			public const int CollapsingToolbarLayout_Layout_layout_collapseMode = 0;
			
			// aapt resource value: 1
			public const int CollapsingToolbarLayout_Layout_layout_collapseParallaxMultiplier = 1;
			
			// aapt resource value: 12
			public const int CollapsingToolbarLayout_maxLines = 12;
			
			// aapt resource value: 13
			public const int CollapsingToolbarLayout_scrimAnimationDuration = 13;
			
			// aapt resource value: 14
			public const int CollapsingToolbarLayout_scrimVisibleHeightTrigger = 14;
			
			// aapt resource value: 15
			public const int CollapsingToolbarLayout_statusBarScrim = 15;
			
			// aapt resource value: 16
			public const int CollapsingToolbarLayout_title = 16;
			
			// aapt resource value: 17
			public const int CollapsingToolbarLayout_titleCollapseMode = 17;
			
			// aapt resource value: 18
			public const int CollapsingToolbarLayout_titleEnabled = 18;
			
			// aapt resource value: 19
			public const int CollapsingToolbarLayout_toolbarId = 19;
			
			// aapt resource value: { 0x10101A5,0x101031F,0x7F030032 }
			public static int[] ColorStateListItem = new int[] {
					16843173,
					16843551,
					2130903090};
			
			// aapt resource value: 2
			public const int ColorStateListItem_alpha = 2;
			
			// aapt resource value: 1
			public const int ColorStateListItem_android_alpha = 1;
			
			// aapt resource value: 0
			public const int ColorStateListItem_android_color = 0;
			
			// aapt resource value: { 0x1010107,0x7F03007E,0x7F030084,0x7F030085 }
			public static int[] CompoundButton = new int[] {
					16843015,
					2130903166,
					2130903172,
					2130903173};
			
			// aapt resource value: 0
			public const int CompoundButton_android_button = 0;
			
			// aapt resource value: 1
			public const int CompoundButton_buttonCompat = 1;
			
			// aapt resource value: 2
			public const int CompoundButton_buttonTint = 2;
			
			// aapt resource value: 3
			public const int CompoundButton_buttonTintMode = 3;
			
			// aapt resource value: { 0x10100C4,0x10100D0,0x10100DC,0x10100F4,0x10100F5,0x10100F7,0x10100F8,0x10100F9,0x10100FA,0x101011F,0x1010120,0x101013F,0x1010140,0x101031F,0x1010320,0x1010321,0x1010322,0x1010323,0x1010324,0x1010325,0x1010326,0x1010327,0x1010328,0x10103B5,0x10103B6,0x10103FA,0x1010440,0x7F030035,0x7F030036,0x7F030056,0x7F030057,0x7F030058,0x7F030098,0x7F0300EE,0x7F0300EF,0x7F030142,0x7F03018B,0x7F03018C,0x7F03018D,0x7F03018E,0x7F03018F,0x7F030190,0x7F030191,0x7F030192,0x7F030193,0x7F030194,0x7F030195,0x7F030196,0x7F030197,0x7F030199,0x7F03019A,0x7F03019B,0x7F03019C,0x7F03019D,0x7F03020E,0x7F03020F,0x7F030210,0x7F030211,0x7F030212,0x7F030213,0x7F030214,0x7F030215,0x7F030216,0x7F030217,0x7F030218,0x7F030219,0x7F03021A,0x7F03021B,0x7F03021C,0x7F03021D,0x7F03021E,0x7F03021F,0x7F030220,0x7F030221,0x7F030222,0x7F030223,0x7F030224,0x7F030225,0x7F030226,0x7F030227,0x7F030228,0x7F030229,0x7F03022A,0x7F03022B,0x7F03022C,0x7F03022D,0x7F03022E,0x7F03022F,0x7F030230,0x7F030231,0x7F030232,0x7F030233,0x7F030234,0x7F030235,0x7F030236,0x7F030237,0x7F030238,0x7F030239,0x7F03023A,0x7F03023B,0x7F03023D,0x7F03023E,0x7F03023F,0x7F030240,0x7F030241,0x7F030242,0x7F030243,0x7F030244,0x7F030245,0x7F030248,0x7F03024C,0x7F0302B1,0x7F0302B2,0x7F0302E1,0x7F0302E9,0x7F0302EE,0x7F030308,0x7F030309,0x7F03030A,0x7F030402,0x7F030404,0x7F030406,0x7F030419 }
			public static int[] Constraint = new int[] {
					16842948,
					16842960,
					16842972,
					16842996,
					16842997,
					16842999,
					16843000,
					16843001,
					16843002,
					16843039,
					16843040,
					16843071,
					16843072,
					16843551,
					16843552,
					16843553,
					16843554,
					16843555,
					16843556,
					16843557,
					16843558,
					16843559,
					16843560,
					16843701,
					16843702,
					16843770,
					16843840,
					2130903093,
					2130903094,
					2130903126,
					2130903127,
					2130903128,
					2130903192,
					2130903278,
					2130903279,
					2130903362,
					2130903435,
					2130903436,
					2130903437,
					2130903438,
					2130903439,
					2130903440,
					2130903441,
					2130903442,
					2130903443,
					2130903444,
					2130903445,
					2130903446,
					2130903447,
					2130903449,
					2130903450,
					2130903451,
					2130903452,
					2130903453,
					2130903566,
					2130903567,
					2130903568,
					2130903569,
					2130903570,
					2130903571,
					2130903572,
					2130903573,
					2130903574,
					2130903575,
					2130903576,
					2130903577,
					2130903578,
					2130903579,
					2130903580,
					2130903581,
					2130903582,
					2130903583,
					2130903584,
					2130903585,
					2130903586,
					2130903587,
					2130903588,
					2130903589,
					2130903590,
					2130903591,
					2130903592,
					2130903593,
					2130903594,
					2130903595,
					2130903596,
					2130903597,
					2130903598,
					2130903599,
					2130903600,
					2130903601,
					2130903602,
					2130903603,
					2130903604,
					2130903605,
					2130903606,
					2130903607,
					2130903608,
					2130903609,
					2130903610,
					2130903611,
					2130903613,
					2130903614,
					2130903615,
					2130903616,
					2130903617,
					2130903618,
					2130903619,
					2130903620,
					2130903621,
					2130903624,
					2130903628,
					2130903729,
					2130903730,
					2130903777,
					2130903785,
					2130903790,
					2130903816,
					2130903817,
					2130903818,
					2130904066,
					2130904068,
					2130904070,
					2130904089};
			
			// aapt resource value: { 0x10100C4,0x10100D5,0x10100D6,0x10100D7,0x10100D8,0x10100D9,0x10100DC,0x10100F4,0x10100F5,0x10100F6,0x10100F7,0x10100F8,0x10100F9,0x10100FA,0x101011F,0x1010120,0x101013F,0x1010140,0x10103B3,0x10103B4,0x10103B5,0x10103B6,0x1010440,0x101053B,0x101053C,0x7F030056,0x7F030057,0x7F030058,0x7F030098,0x7F0300BA,0x7F0300BB,0x7F0300BC,0x7F0300BD,0x7F0300BE,0x7F0300EB,0x7F0300EE,0x7F0300EF,0x7F03018B,0x7F03018C,0x7F03018D,0x7F03018E,0x7F03018F,0x7F030190,0x7F030191,0x7F030192,0x7F030193,0x7F030194,0x7F030195,0x7F030196,0x7F030197,0x7F030199,0x7F03019A,0x7F03019B,0x7F03019C,0x7F03019D,0x7F030206,0x7F03020E,0x7F03020F,0x7F030210,0x7F030211,0x7F030212,0x7F030213,0x7F030214,0x7F030215,0x7F030216,0x7F030217,0x7F030218,0x7F030219,0x7F03021A,0x7F03021B,0x7F03021C,0x7F03021D,0x7F03021E,0x7F03021F,0x7F030220,0x7F030221,0x7F030222,0x7F030223,0x7F030224,0x7F030225,0x7F030226,0x7F030227,0x7F030228,0x7F030229,0x7F03022A,0x7F03022B,0x7F03022C,0x7F03022D,0x7F03022E,0x7F03022F,0x7F030230,0x7F030231,0x7F030232,0x7F030233,0x7F030234,0x7F030235,0x7F030236,0x7F030237,0x7F030238,0x7F030239,0x7F03023A,0x7F03023B,0x7F03023D,0x7F03023E,0x7F03023F,0x7F030240,0x7F030241,0x7F030242,0x7F030243,0x7F030244,0x7F030245,0x7F030248,0x7F030249,0x7F03024C }
			public static int[] ConstraintLayout_Layout = new int[] {
					16842948,
					16842965,
					16842966,
					16842967,
					16842968,
					16842969,
					16842972,
					16842996,
					16842997,
					16842998,
					16842999,
					16843000,
					16843001,
					16843002,
					16843039,
					16843040,
					16843071,
					16843072,
					16843699,
					16843700,
					16843701,
					16843702,
					16843840,
					16844091,
					16844092,
					2130903126,
					2130903127,
					2130903128,
					2130903192,
					2130903226,
					2130903227,
					2130903228,
					2130903229,
					2130903230,
					2130903275,
					2130903278,
					2130903279,
					2130903435,
					2130903436,
					2130903437,
					2130903438,
					2130903439,
					2130903440,
					2130903441,
					2130903442,
					2130903443,
					2130903444,
					2130903445,
					2130903446,
					2130903447,
					2130903449,
					2130903450,
					2130903451,
					2130903452,
					2130903453,
					2130903558,
					2130903566,
					2130903567,
					2130903568,
					2130903569,
					2130903570,
					2130903571,
					2130903572,
					2130903573,
					2130903574,
					2130903575,
					2130903576,
					2130903577,
					2130903578,
					2130903579,
					2130903580,
					2130903581,
					2130903582,
					2130903583,
					2130903584,
					2130903585,
					2130903586,
					2130903587,
					2130903588,
					2130903589,
					2130903590,
					2130903591,
					2130903592,
					2130903593,
					2130903594,
					2130903595,
					2130903596,
					2130903597,
					2130903598,
					2130903599,
					2130903600,
					2130903601,
					2130903602,
					2130903603,
					2130903604,
					2130903605,
					2130903606,
					2130903607,
					2130903608,
					2130903609,
					2130903610,
					2130903611,
					2130903613,
					2130903614,
					2130903615,
					2130903616,
					2130903617,
					2130903618,
					2130903619,
					2130903620,
					2130903621,
					2130903624,
					2130903625,
					2130903628};
			
			// aapt resource value: 22
			public const int ConstraintLayout_Layout_android_elevation = 22;
			
			// aapt resource value: 8
			public const int ConstraintLayout_Layout_android_layout_height = 8;
			
			// aapt resource value: 9
			public const int ConstraintLayout_Layout_android_layout_margin = 9;
			
			// aapt resource value: 13
			public const int ConstraintLayout_Layout_android_layout_marginBottom = 13;
			
			// aapt resource value: 21
			public const int ConstraintLayout_Layout_android_layout_marginEnd = 21;
			
			// aapt resource value: 23
			public const int ConstraintLayout_Layout_android_layout_marginHorizontal = 23;
			
			// aapt resource value: 10
			public const int ConstraintLayout_Layout_android_layout_marginLeft = 10;
			
			// aapt resource value: 12
			public const int ConstraintLayout_Layout_android_layout_marginRight = 12;
			
			// aapt resource value: 20
			public const int ConstraintLayout_Layout_android_layout_marginStart = 20;
			
			// aapt resource value: 11
			public const int ConstraintLayout_Layout_android_layout_marginTop = 11;
			
			// aapt resource value: 24
			public const int ConstraintLayout_Layout_android_layout_marginVertical = 24;
			
			// aapt resource value: 7
			public const int ConstraintLayout_Layout_android_layout_width = 7;
			
			// aapt resource value: 15
			public const int ConstraintLayout_Layout_android_maxHeight = 15;
			
			// aapt resource value: 14
			public const int ConstraintLayout_Layout_android_maxWidth = 14;
			
			// aapt resource value: 17
			public const int ConstraintLayout_Layout_android_minHeight = 17;
			
			// aapt resource value: 16
			public const int ConstraintLayout_Layout_android_minWidth = 16;
			
			// aapt resource value: 0
			public const int ConstraintLayout_Layout_android_orientation = 0;
			
			// aapt resource value: 1
			public const int ConstraintLayout_Layout_android_padding = 1;
			
			// aapt resource value: 5
			public const int ConstraintLayout_Layout_android_paddingBottom = 5;
			
			// aapt resource value: 19
			public const int ConstraintLayout_Layout_android_paddingEnd = 19;
			
			// aapt resource value: 2
			public const int ConstraintLayout_Layout_android_paddingLeft = 2;
			
			// aapt resource value: 4
			public const int ConstraintLayout_Layout_android_paddingRight = 4;
			
			// aapt resource value: 18
			public const int ConstraintLayout_Layout_android_paddingStart = 18;
			
			// aapt resource value: 3
			public const int ConstraintLayout_Layout_android_paddingTop = 3;
			
			// aapt resource value: 6
			public const int ConstraintLayout_Layout_android_visibility = 6;
			
			// aapt resource value: 25
			public const int ConstraintLayout_Layout_barrierAllowsGoneWidgets = 25;
			
			// aapt resource value: 26
			public const int ConstraintLayout_Layout_barrierDirection = 26;
			
			// aapt resource value: 27
			public const int ConstraintLayout_Layout_barrierMargin = 27;
			
			// aapt resource value: 28
			public const int ConstraintLayout_Layout_chainUseRtl = 28;
			
			// aapt resource value: 29
			public const int ConstraintLayout_Layout_circularflow_angles = 29;
			
			// aapt resource value: 30
			public const int ConstraintLayout_Layout_circularflow_defaultAngle = 30;
			
			// aapt resource value: 31
			public const int ConstraintLayout_Layout_circularflow_defaultRadius = 31;
			
			// aapt resource value: 32
			public const int ConstraintLayout_Layout_circularflow_radiusInDP = 32;
			
			// aapt resource value: 33
			public const int ConstraintLayout_Layout_circularflow_viewCenter = 33;
			
			// aapt resource value: 34
			public const int ConstraintLayout_Layout_constraintSet = 34;
			
			// aapt resource value: 35
			public const int ConstraintLayout_Layout_constraint_referenced_ids = 35;
			
			// aapt resource value: 36
			public const int ConstraintLayout_Layout_constraint_referenced_tags = 36;
			
			// aapt resource value: 37
			public const int ConstraintLayout_Layout_flow_firstHorizontalBias = 37;
			
			// aapt resource value: 38
			public const int ConstraintLayout_Layout_flow_firstHorizontalStyle = 38;
			
			// aapt resource value: 39
			public const int ConstraintLayout_Layout_flow_firstVerticalBias = 39;
			
			// aapt resource value: 40
			public const int ConstraintLayout_Layout_flow_firstVerticalStyle = 40;
			
			// aapt resource value: 41
			public const int ConstraintLayout_Layout_flow_horizontalAlign = 41;
			
			// aapt resource value: 42
			public const int ConstraintLayout_Layout_flow_horizontalBias = 42;
			
			// aapt resource value: 43
			public const int ConstraintLayout_Layout_flow_horizontalGap = 43;
			
			// aapt resource value: 44
			public const int ConstraintLayout_Layout_flow_horizontalStyle = 44;
			
			// aapt resource value: 45
			public const int ConstraintLayout_Layout_flow_lastHorizontalBias = 45;
			
			// aapt resource value: 46
			public const int ConstraintLayout_Layout_flow_lastHorizontalStyle = 46;
			
			// aapt resource value: 47
			public const int ConstraintLayout_Layout_flow_lastVerticalBias = 47;
			
			// aapt resource value: 48
			public const int ConstraintLayout_Layout_flow_lastVerticalStyle = 48;
			
			// aapt resource value: 49
			public const int ConstraintLayout_Layout_flow_maxElementsWrap = 49;
			
			// aapt resource value: 50
			public const int ConstraintLayout_Layout_flow_verticalAlign = 50;
			
			// aapt resource value: 51
			public const int ConstraintLayout_Layout_flow_verticalBias = 51;
			
			// aapt resource value: 52
			public const int ConstraintLayout_Layout_flow_verticalGap = 52;
			
			// aapt resource value: 53
			public const int ConstraintLayout_Layout_flow_verticalStyle = 53;
			
			// aapt resource value: 54
			public const int ConstraintLayout_Layout_flow_wrapMode = 54;
			
			// aapt resource value: 55
			public const int ConstraintLayout_Layout_layoutDescription = 55;
			
			// aapt resource value: 56
			public const int ConstraintLayout_Layout_layout_constrainedHeight = 56;
			
			// aapt resource value: 57
			public const int ConstraintLayout_Layout_layout_constrainedWidth = 57;
			
			// aapt resource value: 58
			public const int ConstraintLayout_Layout_layout_constraintBaseline_creator = 58;
			
			// aapt resource value: 59
			public const int ConstraintLayout_Layout_layout_constraintBaseline_toBaselineOf = 59;
			
			// aapt resource value: 60
			public const int ConstraintLayout_Layout_layout_constraintBaseline_toBottomOf = 60;
			
			// aapt resource value: 61
			public const int ConstraintLayout_Layout_layout_constraintBaseline_toTopOf = 61;
			
			// aapt resource value: 62
			public const int ConstraintLayout_Layout_layout_constraintBottom_creator = 62;
			
			// aapt resource value: 63
			public const int ConstraintLayout_Layout_layout_constraintBottom_toBottomOf = 63;
			
			// aapt resource value: 64
			public const int ConstraintLayout_Layout_layout_constraintBottom_toTopOf = 64;
			
			// aapt resource value: 65
			public const int ConstraintLayout_Layout_layout_constraintCircle = 65;
			
			// aapt resource value: 66
			public const int ConstraintLayout_Layout_layout_constraintCircleAngle = 66;
			
			// aapt resource value: 67
			public const int ConstraintLayout_Layout_layout_constraintCircleRadius = 67;
			
			// aapt resource value: 68
			public const int ConstraintLayout_Layout_layout_constraintDimensionRatio = 68;
			
			// aapt resource value: 69
			public const int ConstraintLayout_Layout_layout_constraintEnd_toEndOf = 69;
			
			// aapt resource value: 70
			public const int ConstraintLayout_Layout_layout_constraintEnd_toStartOf = 70;
			
			// aapt resource value: 71
			public const int ConstraintLayout_Layout_layout_constraintGuide_begin = 71;
			
			// aapt resource value: 72
			public const int ConstraintLayout_Layout_layout_constraintGuide_end = 72;
			
			// aapt resource value: 73
			public const int ConstraintLayout_Layout_layout_constraintGuide_percent = 73;
			
			// aapt resource value: 74
			public const int ConstraintLayout_Layout_layout_constraintHeight = 74;
			
			// aapt resource value: 75
			public const int ConstraintLayout_Layout_layout_constraintHeight_default = 75;
			
			// aapt resource value: 76
			public const int ConstraintLayout_Layout_layout_constraintHeight_max = 76;
			
			// aapt resource value: 77
			public const int ConstraintLayout_Layout_layout_constraintHeight_min = 77;
			
			// aapt resource value: 78
			public const int ConstraintLayout_Layout_layout_constraintHeight_percent = 78;
			
			// aapt resource value: 79
			public const int ConstraintLayout_Layout_layout_constraintHorizontal_bias = 79;
			
			// aapt resource value: 80
			public const int ConstraintLayout_Layout_layout_constraintHorizontal_chainStyle = 80;
			
			// aapt resource value: 81
			public const int ConstraintLayout_Layout_layout_constraintHorizontal_weight = 81;
			
			// aapt resource value: 82
			public const int ConstraintLayout_Layout_layout_constraintLeft_creator = 82;
			
			// aapt resource value: 83
			public const int ConstraintLayout_Layout_layout_constraintLeft_toLeftOf = 83;
			
			// aapt resource value: 84
			public const int ConstraintLayout_Layout_layout_constraintLeft_toRightOf = 84;
			
			// aapt resource value: 85
			public const int ConstraintLayout_Layout_layout_constraintRight_creator = 85;
			
			// aapt resource value: 86
			public const int ConstraintLayout_Layout_layout_constraintRight_toLeftOf = 86;
			
			// aapt resource value: 87
			public const int ConstraintLayout_Layout_layout_constraintRight_toRightOf = 87;
			
			// aapt resource value: 88
			public const int ConstraintLayout_Layout_layout_constraintStart_toEndOf = 88;
			
			// aapt resource value: 89
			public const int ConstraintLayout_Layout_layout_constraintStart_toStartOf = 89;
			
			// aapt resource value: 90
			public const int ConstraintLayout_Layout_layout_constraintTag = 90;
			
			// aapt resource value: 91
			public const int ConstraintLayout_Layout_layout_constraintTop_creator = 91;
			
			// aapt resource value: 92
			public const int ConstraintLayout_Layout_layout_constraintTop_toBottomOf = 92;
			
			// aapt resource value: 93
			public const int ConstraintLayout_Layout_layout_constraintTop_toTopOf = 93;
			
			// aapt resource value: 94
			public const int ConstraintLayout_Layout_layout_constraintVertical_bias = 94;
			
			// aapt resource value: 95
			public const int ConstraintLayout_Layout_layout_constraintVertical_chainStyle = 95;
			
			// aapt resource value: 96
			public const int ConstraintLayout_Layout_layout_constraintVertical_weight = 96;
			
			// aapt resource value: 97
			public const int ConstraintLayout_Layout_layout_constraintWidth = 97;
			
			// aapt resource value: 98
			public const int ConstraintLayout_Layout_layout_constraintWidth_default = 98;
			
			// aapt resource value: 99
			public const int ConstraintLayout_Layout_layout_constraintWidth_max = 99;
			
			// aapt resource value: 100
			public const int ConstraintLayout_Layout_layout_constraintWidth_min = 100;
			
			// aapt resource value: 101
			public const int ConstraintLayout_Layout_layout_constraintWidth_percent = 101;
			
			// aapt resource value: 102
			public const int ConstraintLayout_Layout_layout_editor_absoluteX = 102;
			
			// aapt resource value: 103
			public const int ConstraintLayout_Layout_layout_editor_absoluteY = 103;
			
			// aapt resource value: 104
			public const int ConstraintLayout_Layout_layout_goneMarginBaseline = 104;
			
			// aapt resource value: 105
			public const int ConstraintLayout_Layout_layout_goneMarginBottom = 105;
			
			// aapt resource value: 106
			public const int ConstraintLayout_Layout_layout_goneMarginEnd = 106;
			
			// aapt resource value: 107
			public const int ConstraintLayout_Layout_layout_goneMarginLeft = 107;
			
			// aapt resource value: 108
			public const int ConstraintLayout_Layout_layout_goneMarginRight = 108;
			
			// aapt resource value: 109
			public const int ConstraintLayout_Layout_layout_goneMarginStart = 109;
			
			// aapt resource value: 110
			public const int ConstraintLayout_Layout_layout_goneMarginTop = 110;
			
			// aapt resource value: 111
			public const int ConstraintLayout_Layout_layout_marginBaseline = 111;
			
			// aapt resource value: 112
			public const int ConstraintLayout_Layout_layout_optimizationLevel = 112;
			
			// aapt resource value: 113
			public const int ConstraintLayout_Layout_layout_wrapBehaviorInParent = 113;
			
			// aapt resource value: { 0x7F0300F1,0x7F0302ED }
			public static int[] ConstraintLayout_placeholder = new int[] {
					2130903281,
					2130903789};
			
			// aapt resource value: 0
			public const int ConstraintLayout_placeholder_content = 0;
			
			// aapt resource value: 1
			public const int ConstraintLayout_placeholder_placeholder_emptyVisibility = 1;
			
			// aapt resource value: { 0x7F030313,0x7F030314,0x7F030315,0x7F030316 }
			public static int[] ConstraintLayout_ReactiveGuide = new int[] {
					2130903827,
					2130903828,
					2130903829,
					2130903830};
			
			// aapt resource value: 0
			public const int ConstraintLayout_ReactiveGuide_reactiveGuide_animateChange = 0;
			
			// aapt resource value: 1
			public const int ConstraintLayout_ReactiveGuide_reactiveGuide_applyToAllConstraintSets = 1;
			
			// aapt resource value: 2
			public const int ConstraintLayout_ReactiveGuide_reactiveGuide_applyToConstraintSet = 2;
			
			// aapt resource value: 3
			public const int ConstraintLayout_ReactiveGuide_reactiveGuide_valueId = 3;
			
			// aapt resource value: { 0x10100C4,0x10100D0,0x10100DC,0x10100F4,0x10100F5,0x10100F7,0x10100F8,0x10100F9,0x10100FA,0x101011F,0x1010120,0x101013F,0x1010140,0x101031F,0x1010320,0x1010321,0x1010322,0x1010323,0x1010324,0x1010325,0x1010326,0x1010327,0x1010328,0x10103B5,0x10103B6,0x10103FA,0x1010440,0x7F030035,0x7F030036,0x7F030056,0x7F030057,0x7F030058,0x7F030098,0x7F0300EE,0x7F030142,0x7F03018B,0x7F03018C,0x7F03018D,0x7F03018E,0x7F03018F,0x7F030190,0x7F030191,0x7F030192,0x7F030193,0x7F030194,0x7F030195,0x7F030196,0x7F030197,0x7F030199,0x7F03019A,0x7F03019B,0x7F03019C,0x7F03019D,0x7F03020E,0x7F03020F,0x7F030210,0x7F030214,0x7F030218,0x7F030219,0x7F03021A,0x7F03021D,0x7F03021E,0x7F03021F,0x7F030220,0x7F030221,0x7F030222,0x7F030223,0x7F030224,0x7F030225,0x7F030226,0x7F030227,0x7F030228,0x7F03022B,0x7F030230,0x7F030231,0x7F030234,0x7F030235,0x7F030236,0x7F030237,0x7F030238,0x7F030239,0x7F03023A,0x7F03023B,0x7F03023D,0x7F03023E,0x7F03023F,0x7F030240,0x7F030241,0x7F030242,0x7F030243,0x7F030244,0x7F030245,0x7F030248,0x7F03024C,0x7F0302B1,0x7F0302B2,0x7F0302B3,0x7F0302E1,0x7F0302E9,0x7F0302EE,0x7F030308,0x7F030309,0x7F03030A,0x7F030402,0x7F030404,0x7F030406,0x7F030419 }
			public static int[] ConstraintOverride = new int[] {
					16842948,
					16842960,
					16842972,
					16842996,
					16842997,
					16842999,
					16843000,
					16843001,
					16843002,
					16843039,
					16843040,
					16843071,
					16843072,
					16843551,
					16843552,
					16843553,
					16843554,
					16843555,
					16843556,
					16843557,
					16843558,
					16843559,
					16843560,
					16843701,
					16843702,
					16843770,
					16843840,
					2130903093,
					2130903094,
					2130903126,
					2130903127,
					2130903128,
					2130903192,
					2130903278,
					2130903362,
					2130903435,
					2130903436,
					2130903437,
					2130903438,
					2130903439,
					2130903440,
					2130903441,
					2130903442,
					2130903443,
					2130903444,
					2130903445,
					2130903446,
					2130903447,
					2130903449,
					2130903450,
					2130903451,
					2130903452,
					2130903453,
					2130903566,
					2130903567,
					2130903568,
					2130903572,
					2130903576,
					2130903577,
					2130903578,
					2130903581,
					2130903582,
					2130903583,
					2130903584,
					2130903585,
					2130903586,
					2130903587,
					2130903588,
					2130903589,
					2130903590,
					2130903591,
					2130903592,
					2130903595,
					2130903600,
					2130903601,
					2130903604,
					2130903605,
					2130903606,
					2130903607,
					2130903608,
					2130903609,
					2130903610,
					2130903611,
					2130903613,
					2130903614,
					2130903615,
					2130903616,
					2130903617,
					2130903618,
					2130903619,
					2130903620,
					2130903621,
					2130903624,
					2130903628,
					2130903729,
					2130903730,
					2130903731,
					2130903777,
					2130903785,
					2130903790,
					2130903816,
					2130903817,
					2130903818,
					2130904066,
					2130904068,
					2130904070,
					2130904089};
			
			// aapt resource value: 13
			public const int ConstraintOverride_android_alpha = 13;
			
			// aapt resource value: 26
			public const int ConstraintOverride_android_elevation = 26;
			
			// aapt resource value: 1
			public const int ConstraintOverride_android_id = 1;
			
			// aapt resource value: 4
			public const int ConstraintOverride_android_layout_height = 4;
			
			// aapt resource value: 8
			public const int ConstraintOverride_android_layout_marginBottom = 8;
			
			// aapt resource value: 24
			public const int ConstraintOverride_android_layout_marginEnd = 24;
			
			// aapt resource value: 5
			public const int ConstraintOverride_android_layout_marginLeft = 5;
			
			// aapt resource value: 7
			public const int ConstraintOverride_android_layout_marginRight = 7;
			
			// aapt resource value: 23
			public const int ConstraintOverride_android_layout_marginStart = 23;
			
			// aapt resource value: 6
			public const int ConstraintOverride_android_layout_marginTop = 6;
			
			// aapt resource value: 3
			public const int ConstraintOverride_android_layout_width = 3;
			
			// aapt resource value: 10
			public const int ConstraintOverride_android_maxHeight = 10;
			
			// aapt resource value: 9
			public const int ConstraintOverride_android_maxWidth = 9;
			
			// aapt resource value: 12
			public const int ConstraintOverride_android_minHeight = 12;
			
			// aapt resource value: 11
			public const int ConstraintOverride_android_minWidth = 11;
			
			// aapt resource value: 0
			public const int ConstraintOverride_android_orientation = 0;
			
			// aapt resource value: 20
			public const int ConstraintOverride_android_rotation = 20;
			
			// aapt resource value: 21
			public const int ConstraintOverride_android_rotationX = 21;
			
			// aapt resource value: 22
			public const int ConstraintOverride_android_rotationY = 22;
			
			// aapt resource value: 18
			public const int ConstraintOverride_android_scaleX = 18;
			
			// aapt resource value: 19
			public const int ConstraintOverride_android_scaleY = 19;
			
			// aapt resource value: 14
			public const int ConstraintOverride_android_transformPivotX = 14;
			
			// aapt resource value: 15
			public const int ConstraintOverride_android_transformPivotY = 15;
			
			// aapt resource value: 16
			public const int ConstraintOverride_android_translationX = 16;
			
			// aapt resource value: 17
			public const int ConstraintOverride_android_translationY = 17;
			
			// aapt resource value: 25
			public const int ConstraintOverride_android_translationZ = 25;
			
			// aapt resource value: 2
			public const int ConstraintOverride_android_visibility = 2;
			
			// aapt resource value: 27
			public const int ConstraintOverride_animateCircleAngleTo = 27;
			
			// aapt resource value: 28
			public const int ConstraintOverride_animateRelativeTo = 28;
			
			// aapt resource value: 29
			public const int ConstraintOverride_barrierAllowsGoneWidgets = 29;
			
			// aapt resource value: 30
			public const int ConstraintOverride_barrierDirection = 30;
			
			// aapt resource value: 31
			public const int ConstraintOverride_barrierMargin = 31;
			
			// aapt resource value: 32
			public const int ConstraintOverride_chainUseRtl = 32;
			
			// aapt resource value: 33
			public const int ConstraintOverride_constraint_referenced_ids = 33;
			
			// aapt resource value: 34
			public const int ConstraintOverride_drawPath = 34;
			
			// aapt resource value: 35
			public const int ConstraintOverride_flow_firstHorizontalBias = 35;
			
			// aapt resource value: 36
			public const int ConstraintOverride_flow_firstHorizontalStyle = 36;
			
			// aapt resource value: 37
			public const int ConstraintOverride_flow_firstVerticalBias = 37;
			
			// aapt resource value: 38
			public const int ConstraintOverride_flow_firstVerticalStyle = 38;
			
			// aapt resource value: 39
			public const int ConstraintOverride_flow_horizontalAlign = 39;
			
			// aapt resource value: 40
			public const int ConstraintOverride_flow_horizontalBias = 40;
			
			// aapt resource value: 41
			public const int ConstraintOverride_flow_horizontalGap = 41;
			
			// aapt resource value: 42
			public const int ConstraintOverride_flow_horizontalStyle = 42;
			
			// aapt resource value: 43
			public const int ConstraintOverride_flow_lastHorizontalBias = 43;
			
			// aapt resource value: 44
			public const int ConstraintOverride_flow_lastHorizontalStyle = 44;
			
			// aapt resource value: 45
			public const int ConstraintOverride_flow_lastVerticalBias = 45;
			
			// aapt resource value: 46
			public const int ConstraintOverride_flow_lastVerticalStyle = 46;
			
			// aapt resource value: 47
			public const int ConstraintOverride_flow_maxElementsWrap = 47;
			
			// aapt resource value: 48
			public const int ConstraintOverride_flow_verticalAlign = 48;
			
			// aapt resource value: 49
			public const int ConstraintOverride_flow_verticalBias = 49;
			
			// aapt resource value: 50
			public const int ConstraintOverride_flow_verticalGap = 50;
			
			// aapt resource value: 51
			public const int ConstraintOverride_flow_verticalStyle = 51;
			
			// aapt resource value: 52
			public const int ConstraintOverride_flow_wrapMode = 52;
			
			// aapt resource value: 53
			public const int ConstraintOverride_layout_constrainedHeight = 53;
			
			// aapt resource value: 54
			public const int ConstraintOverride_layout_constrainedWidth = 54;
			
			// aapt resource value: 55
			public const int ConstraintOverride_layout_constraintBaseline_creator = 55;
			
			// aapt resource value: 56
			public const int ConstraintOverride_layout_constraintBottom_creator = 56;
			
			// aapt resource value: 57
			public const int ConstraintOverride_layout_constraintCircleAngle = 57;
			
			// aapt resource value: 58
			public const int ConstraintOverride_layout_constraintCircleRadius = 58;
			
			// aapt resource value: 59
			public const int ConstraintOverride_layout_constraintDimensionRatio = 59;
			
			// aapt resource value: 60
			public const int ConstraintOverride_layout_constraintGuide_begin = 60;
			
			// aapt resource value: 61
			public const int ConstraintOverride_layout_constraintGuide_end = 61;
			
			// aapt resource value: 62
			public const int ConstraintOverride_layout_constraintGuide_percent = 62;
			
			// aapt resource value: 63
			public const int ConstraintOverride_layout_constraintHeight = 63;
			
			// aapt resource value: 64
			public const int ConstraintOverride_layout_constraintHeight_default = 64;
			
			// aapt resource value: 65
			public const int ConstraintOverride_layout_constraintHeight_max = 65;
			
			// aapt resource value: 66
			public const int ConstraintOverride_layout_constraintHeight_min = 66;
			
			// aapt resource value: 67
			public const int ConstraintOverride_layout_constraintHeight_percent = 67;
			
			// aapt resource value: 68
			public const int ConstraintOverride_layout_constraintHorizontal_bias = 68;
			
			// aapt resource value: 69
			public const int ConstraintOverride_layout_constraintHorizontal_chainStyle = 69;
			
			// aapt resource value: 70
			public const int ConstraintOverride_layout_constraintHorizontal_weight = 70;
			
			// aapt resource value: 71
			public const int ConstraintOverride_layout_constraintLeft_creator = 71;
			
			// aapt resource value: 72
			public const int ConstraintOverride_layout_constraintRight_creator = 72;
			
			// aapt resource value: 73
			public const int ConstraintOverride_layout_constraintTag = 73;
			
			// aapt resource value: 74
			public const int ConstraintOverride_layout_constraintTop_creator = 74;
			
			// aapt resource value: 75
			public const int ConstraintOverride_layout_constraintVertical_bias = 75;
			
			// aapt resource value: 76
			public const int ConstraintOverride_layout_constraintVertical_chainStyle = 76;
			
			// aapt resource value: 77
			public const int ConstraintOverride_layout_constraintVertical_weight = 77;
			
			// aapt resource value: 78
			public const int ConstraintOverride_layout_constraintWidth = 78;
			
			// aapt resource value: 79
			public const int ConstraintOverride_layout_constraintWidth_default = 79;
			
			// aapt resource value: 80
			public const int ConstraintOverride_layout_constraintWidth_max = 80;
			
			// aapt resource value: 81
			public const int ConstraintOverride_layout_constraintWidth_min = 81;
			
			// aapt resource value: 82
			public const int ConstraintOverride_layout_constraintWidth_percent = 82;
			
			// aapt resource value: 83
			public const int ConstraintOverride_layout_editor_absoluteX = 83;
			
			// aapt resource value: 84
			public const int ConstraintOverride_layout_editor_absoluteY = 84;
			
			// aapt resource value: 85
			public const int ConstraintOverride_layout_goneMarginBaseline = 85;
			
			// aapt resource value: 86
			public const int ConstraintOverride_layout_goneMarginBottom = 86;
			
			// aapt resource value: 87
			public const int ConstraintOverride_layout_goneMarginEnd = 87;
			
			// aapt resource value: 88
			public const int ConstraintOverride_layout_goneMarginLeft = 88;
			
			// aapt resource value: 89
			public const int ConstraintOverride_layout_goneMarginRight = 89;
			
			// aapt resource value: 90
			public const int ConstraintOverride_layout_goneMarginStart = 90;
			
			// aapt resource value: 91
			public const int ConstraintOverride_layout_goneMarginTop = 91;
			
			// aapt resource value: 92
			public const int ConstraintOverride_layout_marginBaseline = 92;
			
			// aapt resource value: 93
			public const int ConstraintOverride_layout_wrapBehaviorInParent = 93;
			
			// aapt resource value: 94
			public const int ConstraintOverride_motionProgress = 94;
			
			// aapt resource value: 95
			public const int ConstraintOverride_motionStagger = 95;
			
			// aapt resource value: 96
			public const int ConstraintOverride_motionTarget = 96;
			
			// aapt resource value: 97
			public const int ConstraintOverride_pathMotionArc = 97;
			
			// aapt resource value: 98
			public const int ConstraintOverride_pivotAnchor = 98;
			
			// aapt resource value: 99
			public const int ConstraintOverride_polarRelativeTo = 99;
			
			// aapt resource value: 100
			public const int ConstraintOverride_quantizeMotionInterpolator = 100;
			
			// aapt resource value: 101
			public const int ConstraintOverride_quantizeMotionPhase = 101;
			
			// aapt resource value: 102
			public const int ConstraintOverride_quantizeMotionSteps = 102;
			
			// aapt resource value: 103
			public const int ConstraintOverride_transformPivotTarget = 103;
			
			// aapt resource value: 104
			public const int ConstraintOverride_transitionEasing = 104;
			
			// aapt resource value: 105
			public const int ConstraintOverride_transitionPathRotate = 105;
			
			// aapt resource value: 106
			public const int ConstraintOverride_visibilityMode = 106;
			
			// aapt resource value: { 0x10100C4,0x10100D0,0x10100DC,0x10100F4,0x10100F5,0x10100F7,0x10100F8,0x10100F9,0x10100FA,0x101011F,0x1010120,0x101013F,0x1010140,0x10101B5,0x10101B6,0x101031F,0x1010320,0x1010321,0x1010322,0x1010323,0x1010324,0x1010325,0x1010326,0x1010327,0x1010328,0x10103B5,0x10103B6,0x10103FA,0x1010440,0x7F030000,0x7F030035,0x7F030036,0x7F030056,0x7F030057,0x7F030058,0x7F030098,0x7F0300EE,0x7F0300EF,0x7F03012F,0x7F030142,0x7F03018B,0x7F03018C,0x7F03018D,0x7F03018E,0x7F03018F,0x7F030190,0x7F030191,0x7F030192,0x7F030193,0x7F030194,0x7F030195,0x7F030196,0x7F030197,0x7F030199,0x7F03019A,0x7F03019B,0x7F03019C,0x7F03019D,0x7F03020E,0x7F03020F,0x7F030210,0x7F030211,0x7F030212,0x7F030213,0x7F030214,0x7F030215,0x7F030216,0x7F030217,0x7F030218,0x7F030219,0x7F03021A,0x7F03021B,0x7F03021C,0x7F03021D,0x7F03021E,0x7F03021F,0x7F030221,0x7F030222,0x7F030223,0x7F030224,0x7F030225,0x7F030226,0x7F030227,0x7F030228,0x7F030229,0x7F03022A,0x7F03022B,0x7F03022C,0x7F03022D,0x7F03022E,0x7F03022F,0x7F030230,0x7F030231,0x7F030232,0x7F030233,0x7F030234,0x7F030235,0x7F030236,0x7F030238,0x7F030239,0x7F03023A,0x7F03023B,0x7F03023D,0x7F03023E,0x7F03023F,0x7F030240,0x7F030241,0x7F030242,0x7F030243,0x7F030244,0x7F030245,0x7F030248,0x7F03024C,0x7F0302B1,0x7F0302B2,0x7F0302E1,0x7F0302E9,0x7F0302EE,0x7F03030A,0x7F030404,0x7F030406 }
			public static int[] ConstraintSet = new int[] {
					16842948,
					16842960,
					16842972,
					16842996,
					16842997,
					16842999,
					16843000,
					16843001,
					16843002,
					16843039,
					16843040,
					16843071,
					16843072,
					16843189,
					16843190,
					16843551,
					16843552,
					16843553,
					16843554,
					16843555,
					16843556,
					16843557,
					16843558,
					16843559,
					16843560,
					16843701,
					16843702,
					16843770,
					16843840,
					2130903040,
					2130903093,
					2130903094,
					2130903126,
					2130903127,
					2130903128,
					2130903192,
					2130903278,
					2130903279,
					2130903343,
					2130903362,
					2130903435,
					2130903436,
					2130903437,
					2130903438,
					2130903439,
					2130903440,
					2130903441,
					2130903442,
					2130903443,
					2130903444,
					2130903445,
					2130903446,
					2130903447,
					2130903449,
					2130903450,
					2130903451,
					2130903452,
					2130903453,
					2130903566,
					2130903567,
					2130903568,
					2130903569,
					2130903570,
					2130903571,
					2130903572,
					2130903573,
					2130903574,
					2130903575,
					2130903576,
					2130903577,
					2130903578,
					2130903579,
					2130903580,
					2130903581,
					2130903582,
					2130903583,
					2130903585,
					2130903586,
					2130903587,
					2130903588,
					2130903589,
					2130903590,
					2130903591,
					2130903592,
					2130903593,
					2130903594,
					2130903595,
					2130903596,
					2130903597,
					2130903598,
					2130903599,
					2130903600,
					2130903601,
					2130903602,
					2130903603,
					2130903604,
					2130903605,
					2130903606,
					2130903608,
					2130903609,
					2130903610,
					2130903611,
					2130903613,
					2130903614,
					2130903615,
					2130903616,
					2130903617,
					2130903618,
					2130903619,
					2130903620,
					2130903621,
					2130903624,
					2130903628,
					2130903729,
					2130903730,
					2130903777,
					2130903785,
					2130903790,
					2130903818,
					2130904068,
					2130904070};
			
			// aapt resource value: 15
			public const int ConstraintSet_android_alpha = 15;
			
			// aapt resource value: 28
			public const int ConstraintSet_android_elevation = 28;
			
			// aapt resource value: 1
			public const int ConstraintSet_android_id = 1;
			
			// aapt resource value: 4
			public const int ConstraintSet_android_layout_height = 4;
			
			// aapt resource value: 8
			public const int ConstraintSet_android_layout_marginBottom = 8;
			
			// aapt resource value: 26
			public const int ConstraintSet_android_layout_marginEnd = 26;
			
			// aapt resource value: 5
			public const int ConstraintSet_android_layout_marginLeft = 5;
			
			// aapt resource value: 7
			public const int ConstraintSet_android_layout_marginRight = 7;
			
			// aapt resource value: 25
			public const int ConstraintSet_android_layout_marginStart = 25;
			
			// aapt resource value: 6
			public const int ConstraintSet_android_layout_marginTop = 6;
			
			// aapt resource value: 3
			public const int ConstraintSet_android_layout_width = 3;
			
			// aapt resource value: 10
			public const int ConstraintSet_android_maxHeight = 10;
			
			// aapt resource value: 9
			public const int ConstraintSet_android_maxWidth = 9;
			
			// aapt resource value: 12
			public const int ConstraintSet_android_minHeight = 12;
			
			// aapt resource value: 11
			public const int ConstraintSet_android_minWidth = 11;
			
			// aapt resource value: 0
			public const int ConstraintSet_android_orientation = 0;
			
			// aapt resource value: 13
			public const int ConstraintSet_android_pivotX = 13;
			
			// aapt resource value: 14
			public const int ConstraintSet_android_pivotY = 14;
			
			// aapt resource value: 22
			public const int ConstraintSet_android_rotation = 22;
			
			// aapt resource value: 23
			public const int ConstraintSet_android_rotationX = 23;
			
			// aapt resource value: 24
			public const int ConstraintSet_android_rotationY = 24;
			
			// aapt resource value: 20
			public const int ConstraintSet_android_scaleX = 20;
			
			// aapt resource value: 21
			public const int ConstraintSet_android_scaleY = 21;
			
			// aapt resource value: 16
			public const int ConstraintSet_android_transformPivotX = 16;
			
			// aapt resource value: 17
			public const int ConstraintSet_android_transformPivotY = 17;
			
			// aapt resource value: 18
			public const int ConstraintSet_android_translationX = 18;
			
			// aapt resource value: 19
			public const int ConstraintSet_android_translationY = 19;
			
			// aapt resource value: 27
			public const int ConstraintSet_android_translationZ = 27;
			
			// aapt resource value: 2
			public const int ConstraintSet_android_visibility = 2;
			
			// aapt resource value: 30
			public const int ConstraintSet_animateCircleAngleTo = 30;
			
			// aapt resource value: 31
			public const int ConstraintSet_animateRelativeTo = 31;
			
			// aapt resource value: 32
			public const int ConstraintSet_barrierAllowsGoneWidgets = 32;
			
			// aapt resource value: 33
			public const int ConstraintSet_barrierDirection = 33;
			
			// aapt resource value: 34
			public const int ConstraintSet_barrierMargin = 34;
			
			// aapt resource value: 35
			public const int ConstraintSet_chainUseRtl = 35;
			
			// aapt resource value: 29
			public const int ConstraintSet_ConstraintRotate = 29;
			
			// aapt resource value: 36
			public const int ConstraintSet_constraint_referenced_ids = 36;
			
			// aapt resource value: 37
			public const int ConstraintSet_constraint_referenced_tags = 37;
			
			// aapt resource value: 38
			public const int ConstraintSet_deriveConstraintsFrom = 38;
			
			// aapt resource value: 39
			public const int ConstraintSet_drawPath = 39;
			
			// aapt resource value: 40
			public const int ConstraintSet_flow_firstHorizontalBias = 40;
			
			// aapt resource value: 41
			public const int ConstraintSet_flow_firstHorizontalStyle = 41;
			
			// aapt resource value: 42
			public const int ConstraintSet_flow_firstVerticalBias = 42;
			
			// aapt resource value: 43
			public const int ConstraintSet_flow_firstVerticalStyle = 43;
			
			// aapt resource value: 44
			public const int ConstraintSet_flow_horizontalAlign = 44;
			
			// aapt resource value: 45
			public const int ConstraintSet_flow_horizontalBias = 45;
			
			// aapt resource value: 46
			public const int ConstraintSet_flow_horizontalGap = 46;
			
			// aapt resource value: 47
			public const int ConstraintSet_flow_horizontalStyle = 47;
			
			// aapt resource value: 48
			public const int ConstraintSet_flow_lastHorizontalBias = 48;
			
			// aapt resource value: 49
			public const int ConstraintSet_flow_lastHorizontalStyle = 49;
			
			// aapt resource value: 50
			public const int ConstraintSet_flow_lastVerticalBias = 50;
			
			// aapt resource value: 51
			public const int ConstraintSet_flow_lastVerticalStyle = 51;
			
			// aapt resource value: 52
			public const int ConstraintSet_flow_maxElementsWrap = 52;
			
			// aapt resource value: 53
			public const int ConstraintSet_flow_verticalAlign = 53;
			
			// aapt resource value: 54
			public const int ConstraintSet_flow_verticalBias = 54;
			
			// aapt resource value: 55
			public const int ConstraintSet_flow_verticalGap = 55;
			
			// aapt resource value: 56
			public const int ConstraintSet_flow_verticalStyle = 56;
			
			// aapt resource value: 57
			public const int ConstraintSet_flow_wrapMode = 57;
			
			// aapt resource value: 58
			public const int ConstraintSet_layout_constrainedHeight = 58;
			
			// aapt resource value: 59
			public const int ConstraintSet_layout_constrainedWidth = 59;
			
			// aapt resource value: 60
			public const int ConstraintSet_layout_constraintBaseline_creator = 60;
			
			// aapt resource value: 61
			public const int ConstraintSet_layout_constraintBaseline_toBaselineOf = 61;
			
			// aapt resource value: 62
			public const int ConstraintSet_layout_constraintBaseline_toBottomOf = 62;
			
			// aapt resource value: 63
			public const int ConstraintSet_layout_constraintBaseline_toTopOf = 63;
			
			// aapt resource value: 64
			public const int ConstraintSet_layout_constraintBottom_creator = 64;
			
			// aapt resource value: 65
			public const int ConstraintSet_layout_constraintBottom_toBottomOf = 65;
			
			// aapt resource value: 66
			public const int ConstraintSet_layout_constraintBottom_toTopOf = 66;
			
			// aapt resource value: 67
			public const int ConstraintSet_layout_constraintCircle = 67;
			
			// aapt resource value: 68
			public const int ConstraintSet_layout_constraintCircleAngle = 68;
			
			// aapt resource value: 69
			public const int ConstraintSet_layout_constraintCircleRadius = 69;
			
			// aapt resource value: 70
			public const int ConstraintSet_layout_constraintDimensionRatio = 70;
			
			// aapt resource value: 71
			public const int ConstraintSet_layout_constraintEnd_toEndOf = 71;
			
			// aapt resource value: 72
			public const int ConstraintSet_layout_constraintEnd_toStartOf = 72;
			
			// aapt resource value: 73
			public const int ConstraintSet_layout_constraintGuide_begin = 73;
			
			// aapt resource value: 74
			public const int ConstraintSet_layout_constraintGuide_end = 74;
			
			// aapt resource value: 75
			public const int ConstraintSet_layout_constraintGuide_percent = 75;
			
			// aapt resource value: 76
			public const int ConstraintSet_layout_constraintHeight_default = 76;
			
			// aapt resource value: 77
			public const int ConstraintSet_layout_constraintHeight_max = 77;
			
			// aapt resource value: 78
			public const int ConstraintSet_layout_constraintHeight_min = 78;
			
			// aapt resource value: 79
			public const int ConstraintSet_layout_constraintHeight_percent = 79;
			
			// aapt resource value: 80
			public const int ConstraintSet_layout_constraintHorizontal_bias = 80;
			
			// aapt resource value: 81
			public const int ConstraintSet_layout_constraintHorizontal_chainStyle = 81;
			
			// aapt resource value: 82
			public const int ConstraintSet_layout_constraintHorizontal_weight = 82;
			
			// aapt resource value: 83
			public const int ConstraintSet_layout_constraintLeft_creator = 83;
			
			// aapt resource value: 84
			public const int ConstraintSet_layout_constraintLeft_toLeftOf = 84;
			
			// aapt resource value: 85
			public const int ConstraintSet_layout_constraintLeft_toRightOf = 85;
			
			// aapt resource value: 86
			public const int ConstraintSet_layout_constraintRight_creator = 86;
			
			// aapt resource value: 87
			public const int ConstraintSet_layout_constraintRight_toLeftOf = 87;
			
			// aapt resource value: 88
			public const int ConstraintSet_layout_constraintRight_toRightOf = 88;
			
			// aapt resource value: 89
			public const int ConstraintSet_layout_constraintStart_toEndOf = 89;
			
			// aapt resource value: 90
			public const int ConstraintSet_layout_constraintStart_toStartOf = 90;
			
			// aapt resource value: 91
			public const int ConstraintSet_layout_constraintTag = 91;
			
			// aapt resource value: 92
			public const int ConstraintSet_layout_constraintTop_creator = 92;
			
			// aapt resource value: 93
			public const int ConstraintSet_layout_constraintTop_toBottomOf = 93;
			
			// aapt resource value: 94
			public const int ConstraintSet_layout_constraintTop_toTopOf = 94;
			
			// aapt resource value: 95
			public const int ConstraintSet_layout_constraintVertical_bias = 95;
			
			// aapt resource value: 96
			public const int ConstraintSet_layout_constraintVertical_chainStyle = 96;
			
			// aapt resource value: 97
			public const int ConstraintSet_layout_constraintVertical_weight = 97;
			
			// aapt resource value: 98
			public const int ConstraintSet_layout_constraintWidth_default = 98;
			
			// aapt resource value: 99
			public const int ConstraintSet_layout_constraintWidth_max = 99;
			
			// aapt resource value: 100
			public const int ConstraintSet_layout_constraintWidth_min = 100;
			
			// aapt resource value: 101
			public const int ConstraintSet_layout_constraintWidth_percent = 101;
			
			// aapt resource value: 102
			public const int ConstraintSet_layout_editor_absoluteX = 102;
			
			// aapt resource value: 103
			public const int ConstraintSet_layout_editor_absoluteY = 103;
			
			// aapt resource value: 104
			public const int ConstraintSet_layout_goneMarginBaseline = 104;
			
			// aapt resource value: 105
			public const int ConstraintSet_layout_goneMarginBottom = 105;
			
			// aapt resource value: 106
			public const int ConstraintSet_layout_goneMarginEnd = 106;
			
			// aapt resource value: 107
			public const int ConstraintSet_layout_goneMarginLeft = 107;
			
			// aapt resource value: 108
			public const int ConstraintSet_layout_goneMarginRight = 108;
			
			// aapt resource value: 109
			public const int ConstraintSet_layout_goneMarginStart = 109;
			
			// aapt resource value: 110
			public const int ConstraintSet_layout_goneMarginTop = 110;
			
			// aapt resource value: 111
			public const int ConstraintSet_layout_marginBaseline = 111;
			
			// aapt resource value: 112
			public const int ConstraintSet_layout_wrapBehaviorInParent = 112;
			
			// aapt resource value: 113
			public const int ConstraintSet_motionProgress = 113;
			
			// aapt resource value: 114
			public const int ConstraintSet_motionStagger = 114;
			
			// aapt resource value: 115
			public const int ConstraintSet_pathMotionArc = 115;
			
			// aapt resource value: 116
			public const int ConstraintSet_pivotAnchor = 116;
			
			// aapt resource value: 117
			public const int ConstraintSet_polarRelativeTo = 117;
			
			// aapt resource value: 118
			public const int ConstraintSet_quantizeMotionSteps = 118;
			
			// aapt resource value: 119
			public const int ConstraintSet_transitionEasing = 119;
			
			// aapt resource value: 120
			public const int ConstraintSet_transitionPathRotate = 120;
			
			// aapt resource value: 13
			public const int Constraint_android_alpha = 13;
			
			// aapt resource value: 26
			public const int Constraint_android_elevation = 26;
			
			// aapt resource value: 1
			public const int Constraint_android_id = 1;
			
			// aapt resource value: 4
			public const int Constraint_android_layout_height = 4;
			
			// aapt resource value: 8
			public const int Constraint_android_layout_marginBottom = 8;
			
			// aapt resource value: 24
			public const int Constraint_android_layout_marginEnd = 24;
			
			// aapt resource value: 5
			public const int Constraint_android_layout_marginLeft = 5;
			
			// aapt resource value: 7
			public const int Constraint_android_layout_marginRight = 7;
			
			// aapt resource value: 23
			public const int Constraint_android_layout_marginStart = 23;
			
			// aapt resource value: 6
			public const int Constraint_android_layout_marginTop = 6;
			
			// aapt resource value: 3
			public const int Constraint_android_layout_width = 3;
			
			// aapt resource value: 10
			public const int Constraint_android_maxHeight = 10;
			
			// aapt resource value: 9
			public const int Constraint_android_maxWidth = 9;
			
			// aapt resource value: 12
			public const int Constraint_android_minHeight = 12;
			
			// aapt resource value: 11
			public const int Constraint_android_minWidth = 11;
			
			// aapt resource value: 0
			public const int Constraint_android_orientation = 0;
			
			// aapt resource value: 20
			public const int Constraint_android_rotation = 20;
			
			// aapt resource value: 21
			public const int Constraint_android_rotationX = 21;
			
			// aapt resource value: 22
			public const int Constraint_android_rotationY = 22;
			
			// aapt resource value: 18
			public const int Constraint_android_scaleX = 18;
			
			// aapt resource value: 19
			public const int Constraint_android_scaleY = 19;
			
			// aapt resource value: 14
			public const int Constraint_android_transformPivotX = 14;
			
			// aapt resource value: 15
			public const int Constraint_android_transformPivotY = 15;
			
			// aapt resource value: 16
			public const int Constraint_android_translationX = 16;
			
			// aapt resource value: 17
			public const int Constraint_android_translationY = 17;
			
			// aapt resource value: 25
			public const int Constraint_android_translationZ = 25;
			
			// aapt resource value: 2
			public const int Constraint_android_visibility = 2;
			
			// aapt resource value: 27
			public const int Constraint_animateCircleAngleTo = 27;
			
			// aapt resource value: 28
			public const int Constraint_animateRelativeTo = 28;
			
			// aapt resource value: 29
			public const int Constraint_barrierAllowsGoneWidgets = 29;
			
			// aapt resource value: 30
			public const int Constraint_barrierDirection = 30;
			
			// aapt resource value: 31
			public const int Constraint_barrierMargin = 31;
			
			// aapt resource value: 32
			public const int Constraint_chainUseRtl = 32;
			
			// aapt resource value: 33
			public const int Constraint_constraint_referenced_ids = 33;
			
			// aapt resource value: 34
			public const int Constraint_constraint_referenced_tags = 34;
			
			// aapt resource value: 35
			public const int Constraint_drawPath = 35;
			
			// aapt resource value: 36
			public const int Constraint_flow_firstHorizontalBias = 36;
			
			// aapt resource value: 37
			public const int Constraint_flow_firstHorizontalStyle = 37;
			
			// aapt resource value: 38
			public const int Constraint_flow_firstVerticalBias = 38;
			
			// aapt resource value: 39
			public const int Constraint_flow_firstVerticalStyle = 39;
			
			// aapt resource value: 40
			public const int Constraint_flow_horizontalAlign = 40;
			
			// aapt resource value: 41
			public const int Constraint_flow_horizontalBias = 41;
			
			// aapt resource value: 42
			public const int Constraint_flow_horizontalGap = 42;
			
			// aapt resource value: 43
			public const int Constraint_flow_horizontalStyle = 43;
			
			// aapt resource value: 44
			public const int Constraint_flow_lastHorizontalBias = 44;
			
			// aapt resource value: 45
			public const int Constraint_flow_lastHorizontalStyle = 45;
			
			// aapt resource value: 46
			public const int Constraint_flow_lastVerticalBias = 46;
			
			// aapt resource value: 47
			public const int Constraint_flow_lastVerticalStyle = 47;
			
			// aapt resource value: 48
			public const int Constraint_flow_maxElementsWrap = 48;
			
			// aapt resource value: 49
			public const int Constraint_flow_verticalAlign = 49;
			
			// aapt resource value: 50
			public const int Constraint_flow_verticalBias = 50;
			
			// aapt resource value: 51
			public const int Constraint_flow_verticalGap = 51;
			
			// aapt resource value: 52
			public const int Constraint_flow_verticalStyle = 52;
			
			// aapt resource value: 53
			public const int Constraint_flow_wrapMode = 53;
			
			// aapt resource value: 54
			public const int Constraint_layout_constrainedHeight = 54;
			
			// aapt resource value: 55
			public const int Constraint_layout_constrainedWidth = 55;
			
			// aapt resource value: 56
			public const int Constraint_layout_constraintBaseline_creator = 56;
			
			// aapt resource value: 57
			public const int Constraint_layout_constraintBaseline_toBaselineOf = 57;
			
			// aapt resource value: 58
			public const int Constraint_layout_constraintBaseline_toBottomOf = 58;
			
			// aapt resource value: 59
			public const int Constraint_layout_constraintBaseline_toTopOf = 59;
			
			// aapt resource value: 60
			public const int Constraint_layout_constraintBottom_creator = 60;
			
			// aapt resource value: 61
			public const int Constraint_layout_constraintBottom_toBottomOf = 61;
			
			// aapt resource value: 62
			public const int Constraint_layout_constraintBottom_toTopOf = 62;
			
			// aapt resource value: 63
			public const int Constraint_layout_constraintCircle = 63;
			
			// aapt resource value: 64
			public const int Constraint_layout_constraintCircleAngle = 64;
			
			// aapt resource value: 65
			public const int Constraint_layout_constraintCircleRadius = 65;
			
			// aapt resource value: 66
			public const int Constraint_layout_constraintDimensionRatio = 66;
			
			// aapt resource value: 67
			public const int Constraint_layout_constraintEnd_toEndOf = 67;
			
			// aapt resource value: 68
			public const int Constraint_layout_constraintEnd_toStartOf = 68;
			
			// aapt resource value: 69
			public const int Constraint_layout_constraintGuide_begin = 69;
			
			// aapt resource value: 70
			public const int Constraint_layout_constraintGuide_end = 70;
			
			// aapt resource value: 71
			public const int Constraint_layout_constraintGuide_percent = 71;
			
			// aapt resource value: 72
			public const int Constraint_layout_constraintHeight = 72;
			
			// aapt resource value: 73
			public const int Constraint_layout_constraintHeight_default = 73;
			
			// aapt resource value: 74
			public const int Constraint_layout_constraintHeight_max = 74;
			
			// aapt resource value: 75
			public const int Constraint_layout_constraintHeight_min = 75;
			
			// aapt resource value: 76
			public const int Constraint_layout_constraintHeight_percent = 76;
			
			// aapt resource value: 77
			public const int Constraint_layout_constraintHorizontal_bias = 77;
			
			// aapt resource value: 78
			public const int Constraint_layout_constraintHorizontal_chainStyle = 78;
			
			// aapt resource value: 79
			public const int Constraint_layout_constraintHorizontal_weight = 79;
			
			// aapt resource value: 80
			public const int Constraint_layout_constraintLeft_creator = 80;
			
			// aapt resource value: 81
			public const int Constraint_layout_constraintLeft_toLeftOf = 81;
			
			// aapt resource value: 82
			public const int Constraint_layout_constraintLeft_toRightOf = 82;
			
			// aapt resource value: 83
			public const int Constraint_layout_constraintRight_creator = 83;
			
			// aapt resource value: 84
			public const int Constraint_layout_constraintRight_toLeftOf = 84;
			
			// aapt resource value: 85
			public const int Constraint_layout_constraintRight_toRightOf = 85;
			
			// aapt resource value: 86
			public const int Constraint_layout_constraintStart_toEndOf = 86;
			
			// aapt resource value: 87
			public const int Constraint_layout_constraintStart_toStartOf = 87;
			
			// aapt resource value: 88
			public const int Constraint_layout_constraintTag = 88;
			
			// aapt resource value: 89
			public const int Constraint_layout_constraintTop_creator = 89;
			
			// aapt resource value: 90
			public const int Constraint_layout_constraintTop_toBottomOf = 90;
			
			// aapt resource value: 91
			public const int Constraint_layout_constraintTop_toTopOf = 91;
			
			// aapt resource value: 92
			public const int Constraint_layout_constraintVertical_bias = 92;
			
			// aapt resource value: 93
			public const int Constraint_layout_constraintVertical_chainStyle = 93;
			
			// aapt resource value: 94
			public const int Constraint_layout_constraintVertical_weight = 94;
			
			// aapt resource value: 95
			public const int Constraint_layout_constraintWidth = 95;
			
			// aapt resource value: 96
			public const int Constraint_layout_constraintWidth_default = 96;
			
			// aapt resource value: 97
			public const int Constraint_layout_constraintWidth_max = 97;
			
			// aapt resource value: 98
			public const int Constraint_layout_constraintWidth_min = 98;
			
			// aapt resource value: 99
			public const int Constraint_layout_constraintWidth_percent = 99;
			
			// aapt resource value: 100
			public const int Constraint_layout_editor_absoluteX = 100;
			
			// aapt resource value: 101
			public const int Constraint_layout_editor_absoluteY = 101;
			
			// aapt resource value: 102
			public const int Constraint_layout_goneMarginBaseline = 102;
			
			// aapt resource value: 103
			public const int Constraint_layout_goneMarginBottom = 103;
			
			// aapt resource value: 104
			public const int Constraint_layout_goneMarginEnd = 104;
			
			// aapt resource value: 105
			public const int Constraint_layout_goneMarginLeft = 105;
			
			// aapt resource value: 106
			public const int Constraint_layout_goneMarginRight = 106;
			
			// aapt resource value: 107
			public const int Constraint_layout_goneMarginStart = 107;
			
			// aapt resource value: 108
			public const int Constraint_layout_goneMarginTop = 108;
			
			// aapt resource value: 109
			public const int Constraint_layout_marginBaseline = 109;
			
			// aapt resource value: 110
			public const int Constraint_layout_wrapBehaviorInParent = 110;
			
			// aapt resource value: 111
			public const int Constraint_motionProgress = 111;
			
			// aapt resource value: 112
			public const int Constraint_motionStagger = 112;
			
			// aapt resource value: 113
			public const int Constraint_pathMotionArc = 113;
			
			// aapt resource value: 114
			public const int Constraint_pivotAnchor = 114;
			
			// aapt resource value: 115
			public const int Constraint_polarRelativeTo = 115;
			
			// aapt resource value: 116
			public const int Constraint_quantizeMotionInterpolator = 116;
			
			// aapt resource value: 117
			public const int Constraint_quantizeMotionPhase = 117;
			
			// aapt resource value: 118
			public const int Constraint_quantizeMotionSteps = 118;
			
			// aapt resource value: 119
			public const int Constraint_transformPivotTarget = 119;
			
			// aapt resource value: 120
			public const int Constraint_transitionEasing = 120;
			
			// aapt resource value: 121
			public const int Constraint_transitionPathRotate = 121;
			
			// aapt resource value: 122
			public const int Constraint_visibilityMode = 122;
			
			// aapt resource value: { 0x7F0301FF,0x7F030366 }
			public static int[] CoordinatorLayout = new int[] {
					2130903551,
					2130903910};
			
			// aapt resource value: 0
			public const int CoordinatorLayout_keylines = 0;
			
			// aapt resource value: { 0x10100B3,0x7F030209,0x7F03020A,0x7F03020B,0x7F03023C,0x7F030246,0x7F030247 }
			public static int[] CoordinatorLayout_Layout = new int[] {
					16842931,
					2130903561,
					2130903562,
					2130903563,
					2130903612,
					2130903622,
					2130903623};
			
			// aapt resource value: 0
			public const int CoordinatorLayout_Layout_android_layout_gravity = 0;
			
			// aapt resource value: 1
			public const int CoordinatorLayout_Layout_layout_anchor = 1;
			
			// aapt resource value: 2
			public const int CoordinatorLayout_Layout_layout_anchorGravity = 2;
			
			// aapt resource value: 3
			public const int CoordinatorLayout_Layout_layout_behavior = 3;
			
			// aapt resource value: 4
			public const int CoordinatorLayout_Layout_layout_dodgeInsetEdges = 4;
			
			// aapt resource value: 5
			public const int CoordinatorLayout_Layout_layout_insetEdge = 5;
			
			// aapt resource value: 6
			public const int CoordinatorLayout_Layout_layout_keyline = 6;
			
			// aapt resource value: 1
			public const int CoordinatorLayout_statusBarBackground = 1;
			
			// aapt resource value: { 0x7F03003E,0x7F030118,0x7F030119,0x7F03011A,0x7F03011B,0x7F03011C,0x7F03011D,0x7F03011F,0x7F030120,0x7F030121,0x7F03028C }
			public static int[] CustomAttribute = new int[] {
					2130903102,
					2130903320,
					2130903321,
					2130903322,
					2130903323,
					2130903324,
					2130903325,
					2130903327,
					2130903328,
					2130903329,
					2130903692};
			
			// aapt resource value: 0
			public const int CustomAttribute_attributeName = 0;
			
			// aapt resource value: 1
			public const int CustomAttribute_customBoolean = 1;
			
			// aapt resource value: 2
			public const int CustomAttribute_customColorDrawableValue = 2;
			
			// aapt resource value: 3
			public const int CustomAttribute_customColorValue = 3;
			
			// aapt resource value: 4
			public const int CustomAttribute_customDimension = 4;
			
			// aapt resource value: 5
			public const int CustomAttribute_customFloatValue = 5;
			
			// aapt resource value: 6
			public const int CustomAttribute_customIntegerValue = 6;
			
			// aapt resource value: 7
			public const int CustomAttribute_customPixelDimension = 7;
			
			// aapt resource value: 8
			public const int CustomAttribute_customReference = 8;
			
			// aapt resource value: 9
			public const int CustomAttribute_customStringValue = 9;
			
			// aapt resource value: 10
			public const int CustomAttribute_methodName = 10;
			
			// aapt resource value: { 0x10101F2,0x10101F3,0x10101F4,0x10101F5,0x10101F6,0x10101F7,0x7F030132,0x7F030133,0x7F030134,0x7F030138,0x7F0302BF,0x7F0302F7 }
			public static int[] DialogPreference = new int[] {
					16843250,
					16843251,
					16843252,
					16843253,
					16843254,
					16843255,
					2130903346,
					2130903347,
					2130903348,
					2130903352,
					2130903743,
					2130903799};
			
			// aapt resource value: 2
			public const int DialogPreference_android_dialogIcon = 2;
			
			// aapt resource value: 5
			public const int DialogPreference_android_dialogLayout = 5;
			
			// aapt resource value: 1
			public const int DialogPreference_android_dialogMessage = 1;
			
			// aapt resource value: 0
			public const int DialogPreference_android_dialogTitle = 0;
			
			// aapt resource value: 4
			public const int DialogPreference_android_negativeButtonText = 4;
			
			// aapt resource value: 3
			public const int DialogPreference_android_positiveButtonText = 3;
			
			// aapt resource value: 6
			public const int DialogPreference_dialogIcon = 6;
			
			// aapt resource value: 7
			public const int DialogPreference_dialogLayout = 7;
			
			// aapt resource value: 8
			public const int DialogPreference_dialogMessage = 8;
			
			// aapt resource value: 9
			public const int DialogPreference_dialogTitle = 9;
			
			// aapt resource value: 10
			public const int DialogPreference_negativeButtonText = 10;
			
			// aapt resource value: 11
			public const int DialogPreference_positiveButtonText = 11;
			
			// aapt resource value: { 0x7F03003C,0x7F03003D,0x7F030055,0x7F0300D4,0x7F030147,0x7F0301AE,0x7F03034E,0x7F0303D0 }
			public static int[] DrawerArrowToggle = new int[] {
					2130903100,
					2130903101,
					2130903125,
					2130903252,
					2130903367,
					2130903470,
					2130903886,
					2130904016};
			
			// aapt resource value: 0
			public const int DrawerArrowToggle_arrowHeadLength = 0;
			
			// aapt resource value: 1
			public const int DrawerArrowToggle_arrowShaftLength = 1;
			
			// aapt resource value: 2
			public const int DrawerArrowToggle_barLength = 2;
			
			// aapt resource value: 3
			public const int DrawerArrowToggle_color = 3;
			
			// aapt resource value: 4
			public const int DrawerArrowToggle_drawableSize = 4;
			
			// aapt resource value: 5
			public const int DrawerArrowToggle_gapBetweenBars = 5;
			
			// aapt resource value: 6
			public const int DrawerArrowToggle_spinBars = 6;
			
			// aapt resource value: 7
			public const int DrawerArrowToggle_thickness = 7;
			
			// aapt resource value: { 0x7F030156 }
			public static int[] DrawerLayout = new int[] {
					2130903382};
			
			// aapt resource value: 0
			public const int DrawerLayout_elevation = 0;
			
			// aapt resource value: { 0x7F030411 }
			public static int[] EditTextPreference = new int[] {
					2130904081};
			
			// aapt resource value: 0
			public const int EditTextPreference_useSimpleSummaryProvider = 0;
			
			// aapt resource value: { 0x7F0300CF,0x7F030156,0x7F03017A,0x7F0301BB,0x7F03033E,0x7F030343 }
			public static int[] ExtendedFloatingActionButton = new int[] {
					2130903247,
					2130903382,
					2130903418,
					2130903483,
					2130903870,
					2130903875};
			
			// aapt resource value: { 0x7F030059,0x7F03005A }
			public static int[] ExtendedFloatingActionButton_Behavior_Layout = new int[] {
					2130903129,
					2130903130};
			
			// aapt resource value: 0
			public const int ExtendedFloatingActionButton_Behavior_Layout_behavior_autoHide = 0;
			
			// aapt resource value: 1
			public const int ExtendedFloatingActionButton_Behavior_Layout_behavior_autoShrink = 1;
			
			// aapt resource value: 0
			public const int ExtendedFloatingActionButton_collapsedSize = 0;
			
			// aapt resource value: 1
			public const int ExtendedFloatingActionButton_elevation = 1;
			
			// aapt resource value: 2
			public const int ExtendedFloatingActionButton_extendMotionSpec = 2;
			
			// aapt resource value: 3
			public const int ExtendedFloatingActionButton_hideMotionSpec = 3;
			
			// aapt resource value: 4
			public const int ExtendedFloatingActionButton_showMotionSpec = 4;
			
			// aapt resource value: 5
			public const int ExtendedFloatingActionButton_shrinkMotionSpec = 5;
			
			// aapt resource value: { 0x101000E,0x7F030050,0x7F030051,0x7F030067,0x7F030156,0x7F030164,0x7F030182,0x7F030183,0x7F0301BB,0x7F0301C5,0x7F030285,0x7F030305,0x7F03031D,0x7F030333,0x7F030336,0x7F03033E,0x7F03040F }
			public static int[] FloatingActionButton = new int[] {
					16842766,
					2130903120,
					2130903121,
					2130903143,
					2130903382,
					2130903396,
					2130903426,
					2130903427,
					2130903483,
					2130903493,
					2130903685,
					2130903813,
					2130903837,
					2130903859,
					2130903862,
					2130903870,
					2130904079};
			
			// aapt resource value: 0
			public const int FloatingActionButton_android_enabled = 0;
			
			// aapt resource value: 1
			public const int FloatingActionButton_backgroundTint = 1;
			
			// aapt resource value: 2
			public const int FloatingActionButton_backgroundTintMode = 2;
			
			// aapt resource value: { 0x7F030059 }
			public static int[] FloatingActionButton_Behavior_Layout = new int[] {
					2130903129};
			
			// aapt resource value: 0
			public const int FloatingActionButton_Behavior_Layout_behavior_autoHide = 0;
			
			// aapt resource value: 3
			public const int FloatingActionButton_borderWidth = 3;
			
			// aapt resource value: 4
			public const int FloatingActionButton_elevation = 4;
			
			// aapt resource value: 5
			public const int FloatingActionButton_ensureMinTouchTargetSize = 5;
			
			// aapt resource value: 6
			public const int FloatingActionButton_fabCustomSize = 6;
			
			// aapt resource value: 7
			public const int FloatingActionButton_fabSize = 7;
			
			// aapt resource value: 8
			public const int FloatingActionButton_hideMotionSpec = 8;
			
			// aapt resource value: 9
			public const int FloatingActionButton_hoveredFocusedTranslationZ = 9;
			
			// aapt resource value: 10
			public const int FloatingActionButton_maxImageSize = 10;
			
			// aapt resource value: 11
			public const int FloatingActionButton_pressedTranslationZ = 11;
			
			// aapt resource value: 12
			public const int FloatingActionButton_rippleColor = 12;
			
			// aapt resource value: 13
			public const int FloatingActionButton_shapeAppearance = 13;
			
			// aapt resource value: 14
			public const int FloatingActionButton_shapeAppearanceOverlay = 14;
			
			// aapt resource value: 15
			public const int FloatingActionButton_showMotionSpec = 15;
			
			// aapt resource value: 16
			public const int FloatingActionButton_useCompatPadding = 16;
			
			// aapt resource value: { 0x7F0301F5,0x7F030251 }
			public static int[] FlowLayout = new int[] {
					2130903541,
					2130903633};
			
			// aapt resource value: 0
			public const int FlowLayout_itemSpacing = 0;
			
			// aapt resource value: 1
			public const int FlowLayout_lineSpacing = 1;
			
			// aapt resource value: { 0x7F0301A0,0x7F0301A1,0x7F0301A2,0x7F0301A3,0x7F0301A4,0x7F0301A5,0x7F0301A6 }
			public static int[] FontFamily = new int[] {
					2130903456,
					2130903457,
					2130903458,
					2130903459,
					2130903460,
					2130903461,
					2130903462};
			
			// aapt resource value: { 0x1010532,0x1010533,0x101053F,0x101056F,0x1010570,0x7F03019E,0x7F0301A7,0x7F0301A8,0x7F0301A9,0x7F03040B }
			public static int[] FontFamilyFont = new int[] {
					16844082,
					16844083,
					16844095,
					16844143,
					16844144,
					2130903454,
					2130903463,
					2130903464,
					2130903465,
					2130904075};
			
			// aapt resource value: 0
			public const int FontFamilyFont_android_font = 0;
			
			// aapt resource value: 2
			public const int FontFamilyFont_android_fontStyle = 2;
			
			// aapt resource value: 4
			public const int FontFamilyFont_android_fontVariationSettings = 4;
			
			// aapt resource value: 1
			public const int FontFamilyFont_android_fontWeight = 1;
			
			// aapt resource value: 3
			public const int FontFamilyFont_android_ttcIndex = 3;
			
			// aapt resource value: 5
			public const int FontFamilyFont_font = 5;
			
			// aapt resource value: 6
			public const int FontFamilyFont_fontStyle = 6;
			
			// aapt resource value: 7
			public const int FontFamilyFont_fontVariationSettings = 7;
			
			// aapt resource value: 8
			public const int FontFamilyFont_fontWeight = 8;
			
			// aapt resource value: 9
			public const int FontFamilyFont_ttcIndex = 9;
			
			// aapt resource value: 0
			public const int FontFamily_fontProviderAuthority = 0;
			
			// aapt resource value: 1
			public const int FontFamily_fontProviderCerts = 1;
			
			// aapt resource value: 2
			public const int FontFamily_fontProviderFetchStrategy = 2;
			
			// aapt resource value: 3
			public const int FontFamily_fontProviderFetchTimeout = 3;
			
			// aapt resource value: 4
			public const int FontFamily_fontProviderPackage = 4;
			
			// aapt resource value: 5
			public const int FontFamily_fontProviderQuery = 5;
			
			// aapt resource value: 6
			public const int FontFamily_fontProviderSystemFontFamily = 6;
			
			// aapt resource value: { 0x1010109,0x1010200,0x7F0301AB }
			public static int[] ForegroundLinearLayout = new int[] {
					16843017,
					16843264,
					2130903467};
			
			// aapt resource value: 0
			public const int ForegroundLinearLayout_android_foreground = 0;
			
			// aapt resource value: 1
			public const int ForegroundLinearLayout_android_foregroundGravity = 1;
			
			// aapt resource value: 2
			public const int ForegroundLinearLayout_foregroundInsidePadding = 2;
			
			// aapt resource value: { 0x1010003,0x10100D0,0x10100D1 }
			public static int[] Fragment = new int[] {
					16842755,
					16842960,
					16842961};
			
			// aapt resource value: { 0x1010003,0x10100D1 }
			public static int[] FragmentContainerView = new int[] {
					16842755,
					16842961};
			
			// aapt resource value: 0
			public const int FragmentContainerView_android_name = 0;
			
			// aapt resource value: 1
			public const int FragmentContainerView_android_tag = 1;
			
			// aapt resource value: 1
			public const int Fragment_android_id = 1;
			
			// aapt resource value: 0
			public const int Fragment_android_name = 0;
			
			// aapt resource value: 2
			public const int Fragment_android_tag = 2;
			
			// aapt resource value: { 0x101019D,0x101019E,0x10101A1,0x10101A2,0x10101A3,0x10101A4,0x1010201,0x101020B,0x1010510,0x1010511,0x1010512,0x1010513 }
			public static int[] GradientColor = new int[] {
					16843165,
					16843166,
					16843169,
					16843170,
					16843171,
					16843172,
					16843265,
					16843275,
					16844048,
					16844049,
					16844050,
					16844051};
			
			// aapt resource value: { 0x10101A5,0x1010514 }
			public static int[] GradientColorItem = new int[] {
					16843173,
					16844052};
			
			// aapt resource value: 0
			public const int GradientColorItem_android_color = 0;
			
			// aapt resource value: 1
			public const int GradientColorItem_android_offset = 1;
			
			// aapt resource value: 7
			public const int GradientColor_android_centerColor = 7;
			
			// aapt resource value: 3
			public const int GradientColor_android_centerX = 3;
			
			// aapt resource value: 4
			public const int GradientColor_android_centerY = 4;
			
			// aapt resource value: 1
			public const int GradientColor_android_endColor = 1;
			
			// aapt resource value: 10
			public const int GradientColor_android_endX = 10;
			
			// aapt resource value: 11
			public const int GradientColor_android_endY = 11;
			
			// aapt resource value: 5
			public const int GradientColor_android_gradientRadius = 5;
			
			// aapt resource value: 0
			public const int GradientColor_android_startColor = 0;
			
			// aapt resource value: 8
			public const int GradientColor_android_startX = 8;
			
			// aapt resource value: 9
			public const int GradientColor_android_startY = 9;
			
			// aapt resource value: 6
			public const int GradientColor_android_tileMode = 6;
			
			// aapt resource value: 2
			public const int GradientColor_android_type = 2;
			
			// aapt resource value: { 0x7F030034,0x7F030064,0x7F030078,0x7F030101,0x7F030115,0x7F0301D3,0x7F0301D4,0x7F0301D5,0x7F0301D6,0x7F0302D0,0x7F03031F,0x7F030320,0x7F030321,0x7F03041B }
			public static int[] ImageFilterView = new int[] {
					2130903092,
					2130903140,
					2130903160,
					2130903297,
					2130903317,
					2130903507,
					2130903508,
					2130903509,
					2130903510,
					2130903760,
					2130903839,
					2130903840,
					2130903841,
					2130904091};
			
			// aapt resource value: 0
			public const int ImageFilterView_altSrc = 0;
			
			// aapt resource value: 1
			public const int ImageFilterView_blendSrc = 1;
			
			// aapt resource value: 2
			public const int ImageFilterView_brightness = 2;
			
			// aapt resource value: 3
			public const int ImageFilterView_contrast = 3;
			
			// aapt resource value: 4
			public const int ImageFilterView_crossfade = 4;
			
			// aapt resource value: 5
			public const int ImageFilterView_imagePanX = 5;
			
			// aapt resource value: 6
			public const int ImageFilterView_imagePanY = 6;
			
			// aapt resource value: 7
			public const int ImageFilterView_imageRotate = 7;
			
			// aapt resource value: 8
			public const int ImageFilterView_imageZoom = 8;
			
			// aapt resource value: 9
			public const int ImageFilterView_overlay = 9;
			
			// aapt resource value: 10
			public const int ImageFilterView_round = 10;
			
			// aapt resource value: 11
			public const int ImageFilterView_roundPercent = 11;
			
			// aapt resource value: 12
			public const int ImageFilterView_saturation = 12;
			
			// aapt resource value: 13
			public const int ImageFilterView_warmth = 13;
			
			// aapt resource value: { 0x7F0300EB }
			public static int[] include = new int[] {
					2130903275};
			
			// aapt resource value: 0
			public const int include_constraintSet = 0;
			
			// aapt resource value: { 0x7F0302D2,0x7F0302D4,0x7F0302D5,0x7F0302D8 }
			public static int[] Insets = new int[] {
					2130903762,
					2130903764,
					2130903765,
					2130903768};
			
			// aapt resource value: 0
			public const int Insets_paddingBottomSystemWindowInsets = 0;
			
			// aapt resource value: 1
			public const int Insets_paddingLeftSystemWindowInsets = 1;
			
			// aapt resource value: 2
			public const int Insets_paddingRightSystemWindowInsets = 2;
			
			// aapt resource value: 3
			public const int Insets_paddingTopSystemWindowInsets = 3;
			
			// aapt resource value: { 0x7F0300D3 }
			public static int[] ItemsViewRendererTheme = new int[] {
					2130903251};
			
			// aapt resource value: 0
			public const int ItemsViewRendererTheme_collectionViewStyle = 0;
			
			// aapt resource value: { 0x101031F,0x1010320,0x1010321,0x1010322,0x1010323,0x1010324,0x1010325,0x1010326,0x1010327,0x1010328,0x10103FA,0x1010440,0x7F030117,0x7F0301AD,0x7F0302B1,0x7F0302B3,0x7F030402,0x7F030404,0x7F030406 }
			public static int[] KeyAttribute = new int[] {
					16843551,
					16843552,
					16843553,
					16843554,
					16843555,
					16843556,
					16843557,
					16843558,
					16843559,
					16843560,
					16843770,
					16843840,
					2130903319,
					2130903469,
					2130903729,
					2130903731,
					2130904066,
					2130904068,
					2130904070};
			
			// aapt resource value: 0
			public const int KeyAttribute_android_alpha = 0;
			
			// aapt resource value: 11
			public const int KeyAttribute_android_elevation = 11;
			
			// aapt resource value: 7
			public const int KeyAttribute_android_rotation = 7;
			
			// aapt resource value: 8
			public const int KeyAttribute_android_rotationX = 8;
			
			// aapt resource value: 9
			public const int KeyAttribute_android_rotationY = 9;
			
			// aapt resource value: 5
			public const int KeyAttribute_android_scaleX = 5;
			
			// aapt resource value: 6
			public const int KeyAttribute_android_scaleY = 6;
			
			// aapt resource value: 1
			public const int KeyAttribute_android_transformPivotX = 1;
			
			// aapt resource value: 2
			public const int KeyAttribute_android_transformPivotY = 2;
			
			// aapt resource value: 3
			public const int KeyAttribute_android_translationX = 3;
			
			// aapt resource value: 4
			public const int KeyAttribute_android_translationY = 4;
			
			// aapt resource value: 10
			public const int KeyAttribute_android_translationZ = 10;
			
			// aapt resource value: 12
			public const int KeyAttribute_curveFit = 12;
			
			// aapt resource value: 13
			public const int KeyAttribute_framePosition = 13;
			
			// aapt resource value: 14
			public const int KeyAttribute_motionProgress = 14;
			
			// aapt resource value: 15
			public const int KeyAttribute_motionTarget = 15;
			
			// aapt resource value: 16
			public const int KeyAttribute_transformPivotTarget = 16;
			
			// aapt resource value: 17
			public const int KeyAttribute_transitionEasing = 17;
			
			// aapt resource value: 18
			public const int KeyAttribute_transitionPathRotate = 18;
			
			// aapt resource value: { 0x101031F,0x1010322,0x1010323,0x1010324,0x1010325,0x1010326,0x1010327,0x1010328,0x10103FA,0x1010440,0x7F030117,0x7F0301AD,0x7F0302B1,0x7F0302B3,0x7F030404,0x7F030406,0x7F03041D,0x7F03041E,0x7F03041F,0x7F030420,0x7F030421 }
			public static int[] KeyCycle = new int[] {
					16843551,
					16843554,
					16843555,
					16843556,
					16843557,
					16843558,
					16843559,
					16843560,
					16843770,
					16843840,
					2130903319,
					2130903469,
					2130903729,
					2130903731,
					2130904068,
					2130904070,
					2130904093,
					2130904094,
					2130904095,
					2130904096,
					2130904097};
			
			// aapt resource value: 0
			public const int KeyCycle_android_alpha = 0;
			
			// aapt resource value: 9
			public const int KeyCycle_android_elevation = 9;
			
			// aapt resource value: 5
			public const int KeyCycle_android_rotation = 5;
			
			// aapt resource value: 6
			public const int KeyCycle_android_rotationX = 6;
			
			// aapt resource value: 7
			public const int KeyCycle_android_rotationY = 7;
			
			// aapt resource value: 3
			public const int KeyCycle_android_scaleX = 3;
			
			// aapt resource value: 4
			public const int KeyCycle_android_scaleY = 4;
			
			// aapt resource value: 1
			public const int KeyCycle_android_translationX = 1;
			
			// aapt resource value: 2
			public const int KeyCycle_android_translationY = 2;
			
			// aapt resource value: 8
			public const int KeyCycle_android_translationZ = 8;
			
			// aapt resource value: 10
			public const int KeyCycle_curveFit = 10;
			
			// aapt resource value: 11
			public const int KeyCycle_framePosition = 11;
			
			// aapt resource value: 12
			public const int KeyCycle_motionProgress = 12;
			
			// aapt resource value: 13
			public const int KeyCycle_motionTarget = 13;
			
			// aapt resource value: 14
			public const int KeyCycle_transitionEasing = 14;
			
			// aapt resource value: 15
			public const int KeyCycle_transitionPathRotate = 15;
			
			// aapt resource value: 16
			public const int KeyCycle_waveOffset = 16;
			
			// aapt resource value: 17
			public const int KeyCycle_wavePeriod = 17;
			
			// aapt resource value: 18
			public const int KeyCycle_wavePhase = 18;
			
			// aapt resource value: 19
			public const int KeyCycle_waveShape = 19;
			
			// aapt resource value: 20
			public const int KeyCycle_waveVariesBy = 20;
			
			// aapt resource value: { 0xFFFFFFFF }
			public static int[] KeyFrame = new int[] {
					-1};
			
			// aapt resource value: { 0xFFFFFFFF }
			public static int[] KeyFramesAcceleration = new int[] {
					-1};
			
			// aapt resource value: { 0xFFFFFFFF }
			public static int[] KeyFramesVelocity = new int[] {
					-1};
			
			// aapt resource value: { 0x7F030117,0x7F030142,0x7F0301AD,0x7F0301FD,0x7F0302B3,0x7F0302E1,0x7F0302E3,0x7F0302E4,0x7F0302E5,0x7F0302E6,0x7F030348,0x7F030404 }
			public static int[] KeyPosition = new int[] {
					2130903319,
					2130903362,
					2130903469,
					2130903549,
					2130903731,
					2130903777,
					2130903779,
					2130903780,
					2130903781,
					2130903782,
					2130903880,
					2130904068};
			
			// aapt resource value: 0
			public const int KeyPosition_curveFit = 0;
			
			// aapt resource value: 1
			public const int KeyPosition_drawPath = 1;
			
			// aapt resource value: 2
			public const int KeyPosition_framePosition = 2;
			
			// aapt resource value: 3
			public const int KeyPosition_keyPositionType = 3;
			
			// aapt resource value: 4
			public const int KeyPosition_motionTarget = 4;
			
			// aapt resource value: 5
			public const int KeyPosition_pathMotionArc = 5;
			
			// aapt resource value: 6
			public const int KeyPosition_percentHeight = 6;
			
			// aapt resource value: 7
			public const int KeyPosition_percentWidth = 7;
			
			// aapt resource value: 8
			public const int KeyPosition_percentX = 8;
			
			// aapt resource value: 9
			public const int KeyPosition_percentY = 9;
			
			// aapt resource value: 10
			public const int KeyPosition_sizePercent = 10;
			
			// aapt resource value: 11
			public const int KeyPosition_transitionEasing = 11;
			
			// aapt resource value: { 0x101031F,0x1010322,0x1010323,0x1010324,0x1010325,0x1010326,0x1010327,0x1010328,0x10103FA,0x1010440,0x7F030117,0x7F0301AD,0x7F0302B1,0x7F0302B3,0x7F030404,0x7F030406,0x7F03041C,0x7F03041D,0x7F03041E,0x7F03041F,0x7F030420 }
			public static int[] KeyTimeCycle = new int[] {
					16843551,
					16843554,
					16843555,
					16843556,
					16843557,
					16843558,
					16843559,
					16843560,
					16843770,
					16843840,
					2130903319,
					2130903469,
					2130903729,
					2130903731,
					2130904068,
					2130904070,
					2130904092,
					2130904093,
					2130904094,
					2130904095,
					2130904096};
			
			// aapt resource value: 0
			public const int KeyTimeCycle_android_alpha = 0;
			
			// aapt resource value: 9
			public const int KeyTimeCycle_android_elevation = 9;
			
			// aapt resource value: 5
			public const int KeyTimeCycle_android_rotation = 5;
			
			// aapt resource value: 6
			public const int KeyTimeCycle_android_rotationX = 6;
			
			// aapt resource value: 7
			public const int KeyTimeCycle_android_rotationY = 7;
			
			// aapt resource value: 3
			public const int KeyTimeCycle_android_scaleX = 3;
			
			// aapt resource value: 4
			public const int KeyTimeCycle_android_scaleY = 4;
			
			// aapt resource value: 1
			public const int KeyTimeCycle_android_translationX = 1;
			
			// aapt resource value: 2
			public const int KeyTimeCycle_android_translationY = 2;
			
			// aapt resource value: 8
			public const int KeyTimeCycle_android_translationZ = 8;
			
			// aapt resource value: 10
			public const int KeyTimeCycle_curveFit = 10;
			
			// aapt resource value: 11
			public const int KeyTimeCycle_framePosition = 11;
			
			// aapt resource value: 12
			public const int KeyTimeCycle_motionProgress = 12;
			
			// aapt resource value: 13
			public const int KeyTimeCycle_motionTarget = 13;
			
			// aapt resource value: 14
			public const int KeyTimeCycle_transitionEasing = 14;
			
			// aapt resource value: 15
			public const int KeyTimeCycle_transitionPathRotate = 15;
			
			// aapt resource value: 16
			public const int KeyTimeCycle_waveDecay = 16;
			
			// aapt resource value: 17
			public const int KeyTimeCycle_waveOffset = 17;
			
			// aapt resource value: 18
			public const int KeyTimeCycle_wavePeriod = 18;
			
			// aapt resource value: 19
			public const int KeyTimeCycle_wavePhase = 19;
			
			// aapt resource value: 20
			public const int KeyTimeCycle_waveShape = 20;
			
			// aapt resource value: { 0x7F0301AD,0x7F0302B3,0x7F0302B4,0x7F0302B5,0x7F0302C6,0x7F0302C8,0x7F0302C9,0x7F030408,0x7F030409,0x7F03040A,0x7F030416,0x7F030417,0x7F030418 }
			public static int[] KeyTrigger = new int[] {
					2130903469,
					2130903731,
					2130903732,
					2130903733,
					2130903750,
					2130903752,
					2130903753,
					2130904072,
					2130904073,
					2130904074,
					2130904086,
					2130904087,
					2130904088};
			
			// aapt resource value: 0
			public const int KeyTrigger_framePosition = 0;
			
			// aapt resource value: 1
			public const int KeyTrigger_motionTarget = 1;
			
			// aapt resource value: 2
			public const int KeyTrigger_motion_postLayoutCollision = 2;
			
			// aapt resource value: 3
			public const int KeyTrigger_motion_triggerOnCollision = 3;
			
			// aapt resource value: 4
			public const int KeyTrigger_onCross = 4;
			
			// aapt resource value: 5
			public const int KeyTrigger_onNegativeCross = 5;
			
			// aapt resource value: 6
			public const int KeyTrigger_onPositiveCross = 6;
			
			// aapt resource value: 7
			public const int KeyTrigger_triggerId = 7;
			
			// aapt resource value: 8
			public const int KeyTrigger_triggerReceiver = 8;
			
			// aapt resource value: 9
			public const int KeyTrigger_triggerSlack = 9;
			
			// aapt resource value: 10
			public const int KeyTrigger_viewTransitionOnCross = 10;
			
			// aapt resource value: 11
			public const int KeyTrigger_viewTransitionOnNegativeCross = 11;
			
			// aapt resource value: 12
			public const int KeyTrigger_viewTransitionOnPositiveCross = 12;
			
			// aapt resource value: { 0x10100C4,0x10100F4,0x10100F5,0x10100F7,0x10100F8,0x10100F9,0x10100FA,0x10103B5,0x10103B6,0x7F030056,0x7F030057,0x7F030058,0x7F030098,0x7F0300EE,0x7F0300EF,0x7F03020E,0x7F03020F,0x7F030210,0x7F030211,0x7F030212,0x7F030213,0x7F030214,0x7F030215,0x7F030216,0x7F030217,0x7F030218,0x7F030219,0x7F03021A,0x7F03021B,0x7F03021C,0x7F03021D,0x7F03021E,0x7F03021F,0x7F030220,0x7F030221,0x7F030222,0x7F030223,0x7F030224,0x7F030225,0x7F030226,0x7F030227,0x7F030228,0x7F030229,0x7F03022A,0x7F03022B,0x7F03022C,0x7F03022D,0x7F03022E,0x7F03022F,0x7F030231,0x7F030232,0x7F030233,0x7F030234,0x7F030235,0x7F030236,0x7F030237,0x7F030238,0x7F030239,0x7F03023A,0x7F03023B,0x7F03023D,0x7F03023E,0x7F03023F,0x7F030240,0x7F030241,0x7F030242,0x7F030243,0x7F030244,0x7F030245,0x7F030248,0x7F03024C,0x7F030284,0x7F030288,0x7F03028F,0x7F030293 }
			public static int[] Layout = new int[] {
					16842948,
					16842996,
					16842997,
					16842999,
					16843000,
					16843001,
					16843002,
					16843701,
					16843702,
					2130903126,
					2130903127,
					2130903128,
					2130903192,
					2130903278,
					2130903279,
					2130903566,
					2130903567,
					2130903568,
					2130903569,
					2130903570,
					2130903571,
					2130903572,
					2130903573,
					2130903574,
					2130903575,
					2130903576,
					2130903577,
					2130903578,
					2130903579,
					2130903580,
					2130903581,
					2130903582,
					2130903583,
					2130903584,
					2130903585,
					2130903586,
					2130903587,
					2130903588,
					2130903589,
					2130903590,
					2130903591,
					2130903592,
					2130903593,
					2130903594,
					2130903595,
					2130903596,
					2130903597,
					2130903598,
					2130903599,
					2130903601,
					2130903602,
					2130903603,
					2130903604,
					2130903605,
					2130903606,
					2130903607,
					2130903608,
					2130903609,
					2130903610,
					2130903611,
					2130903613,
					2130903614,
					2130903615,
					2130903616,
					2130903617,
					2130903618,
					2130903619,
					2130903620,
					2130903621,
					2130903624,
					2130903628,
					2130903684,
					2130903688,
					2130903695,
					2130903699};
			
			// aapt resource value: 2
			public const int Layout_android_layout_height = 2;
			
			// aapt resource value: 6
			public const int Layout_android_layout_marginBottom = 6;
			
			// aapt resource value: 8
			public const int Layout_android_layout_marginEnd = 8;
			
			// aapt resource value: 3
			public const int Layout_android_layout_marginLeft = 3;
			
			// aapt resource value: 5
			public const int Layout_android_layout_marginRight = 5;
			
			// aapt resource value: 7
			public const int Layout_android_layout_marginStart = 7;
			
			// aapt resource value: 4
			public const int Layout_android_layout_marginTop = 4;
			
			// aapt resource value: 1
			public const int Layout_android_layout_width = 1;
			
			// aapt resource value: 0
			public const int Layout_android_orientation = 0;
			
			// aapt resource value: 9
			public const int Layout_barrierAllowsGoneWidgets = 9;
			
			// aapt resource value: 10
			public const int Layout_barrierDirection = 10;
			
			// aapt resource value: 11
			public const int Layout_barrierMargin = 11;
			
			// aapt resource value: 12
			public const int Layout_chainUseRtl = 12;
			
			// aapt resource value: 13
			public const int Layout_constraint_referenced_ids = 13;
			
			// aapt resource value: 14
			public const int Layout_constraint_referenced_tags = 14;
			
			// aapt resource value: 15
			public const int Layout_layout_constrainedHeight = 15;
			
			// aapt resource value: 16
			public const int Layout_layout_constrainedWidth = 16;
			
			// aapt resource value: 17
			public const int Layout_layout_constraintBaseline_creator = 17;
			
			// aapt resource value: 18
			public const int Layout_layout_constraintBaseline_toBaselineOf = 18;
			
			// aapt resource value: 19
			public const int Layout_layout_constraintBaseline_toBottomOf = 19;
			
			// aapt resource value: 20
			public const int Layout_layout_constraintBaseline_toTopOf = 20;
			
			// aapt resource value: 21
			public const int Layout_layout_constraintBottom_creator = 21;
			
			// aapt resource value: 22
			public const int Layout_layout_constraintBottom_toBottomOf = 22;
			
			// aapt resource value: 23
			public const int Layout_layout_constraintBottom_toTopOf = 23;
			
			// aapt resource value: 24
			public const int Layout_layout_constraintCircle = 24;
			
			// aapt resource value: 25
			public const int Layout_layout_constraintCircleAngle = 25;
			
			// aapt resource value: 26
			public const int Layout_layout_constraintCircleRadius = 26;
			
			// aapt resource value: 27
			public const int Layout_layout_constraintDimensionRatio = 27;
			
			// aapt resource value: 28
			public const int Layout_layout_constraintEnd_toEndOf = 28;
			
			// aapt resource value: 29
			public const int Layout_layout_constraintEnd_toStartOf = 29;
			
			// aapt resource value: 30
			public const int Layout_layout_constraintGuide_begin = 30;
			
			// aapt resource value: 31
			public const int Layout_layout_constraintGuide_end = 31;
			
			// aapt resource value: 32
			public const int Layout_layout_constraintGuide_percent = 32;
			
			// aapt resource value: 33
			public const int Layout_layout_constraintHeight = 33;
			
			// aapt resource value: 34
			public const int Layout_layout_constraintHeight_default = 34;
			
			// aapt resource value: 35
			public const int Layout_layout_constraintHeight_max = 35;
			
			// aapt resource value: 36
			public const int Layout_layout_constraintHeight_min = 36;
			
			// aapt resource value: 37
			public const int Layout_layout_constraintHeight_percent = 37;
			
			// aapt resource value: 38
			public const int Layout_layout_constraintHorizontal_bias = 38;
			
			// aapt resource value: 39
			public const int Layout_layout_constraintHorizontal_chainStyle = 39;
			
			// aapt resource value: 40
			public const int Layout_layout_constraintHorizontal_weight = 40;
			
			// aapt resource value: 41
			public const int Layout_layout_constraintLeft_creator = 41;
			
			// aapt resource value: 42
			public const int Layout_layout_constraintLeft_toLeftOf = 42;
			
			// aapt resource value: 43
			public const int Layout_layout_constraintLeft_toRightOf = 43;
			
			// aapt resource value: 44
			public const int Layout_layout_constraintRight_creator = 44;
			
			// aapt resource value: 45
			public const int Layout_layout_constraintRight_toLeftOf = 45;
			
			// aapt resource value: 46
			public const int Layout_layout_constraintRight_toRightOf = 46;
			
			// aapt resource value: 47
			public const int Layout_layout_constraintStart_toEndOf = 47;
			
			// aapt resource value: 48
			public const int Layout_layout_constraintStart_toStartOf = 48;
			
			// aapt resource value: 49
			public const int Layout_layout_constraintTop_creator = 49;
			
			// aapt resource value: 50
			public const int Layout_layout_constraintTop_toBottomOf = 50;
			
			// aapt resource value: 51
			public const int Layout_layout_constraintTop_toTopOf = 51;
			
			// aapt resource value: 52
			public const int Layout_layout_constraintVertical_bias = 52;
			
			// aapt resource value: 53
			public const int Layout_layout_constraintVertical_chainStyle = 53;
			
			// aapt resource value: 54
			public const int Layout_layout_constraintVertical_weight = 54;
			
			// aapt resource value: 55
			public const int Layout_layout_constraintWidth = 55;
			
			// aapt resource value: 56
			public const int Layout_layout_constraintWidth_default = 56;
			
			// aapt resource value: 57
			public const int Layout_layout_constraintWidth_max = 57;
			
			// aapt resource value: 58
			public const int Layout_layout_constraintWidth_min = 58;
			
			// aapt resource value: 59
			public const int Layout_layout_constraintWidth_percent = 59;
			
			// aapt resource value: 60
			public const int Layout_layout_editor_absoluteX = 60;
			
			// aapt resource value: 61
			public const int Layout_layout_editor_absoluteY = 61;
			
			// aapt resource value: 62
			public const int Layout_layout_goneMarginBaseline = 62;
			
			// aapt resource value: 63
			public const int Layout_layout_goneMarginBottom = 63;
			
			// aapt resource value: 64
			public const int Layout_layout_goneMarginEnd = 64;
			
			// aapt resource value: 65
			public const int Layout_layout_goneMarginLeft = 65;
			
			// aapt resource value: 66
			public const int Layout_layout_goneMarginRight = 66;
			
			// aapt resource value: 67
			public const int Layout_layout_goneMarginStart = 67;
			
			// aapt resource value: 68
			public const int Layout_layout_goneMarginTop = 68;
			
			// aapt resource value: 69
			public const int Layout_layout_marginBaseline = 69;
			
			// aapt resource value: 70
			public const int Layout_layout_wrapBehaviorInParent = 70;
			
			// aapt resource value: 71
			public const int Layout_maxHeight = 71;
			
			// aapt resource value: 72
			public const int Layout_maxWidth = 72;
			
			// aapt resource value: 73
			public const int Layout_minHeight = 73;
			
			// aapt resource value: 74
			public const int Layout_minWidth = 74;
			
			// aapt resource value: { 0x10100AF,0x10100C4,0x1010126,0x1010127,0x1010128,0x7F03013B,0x7F03013D,0x7F030289,0x7F03033D }
			public static int[] LinearLayoutCompat = new int[] {
					16842927,
					16842948,
					16843046,
					16843047,
					16843048,
					2130903355,
					2130903357,
					2130903689,
					2130903869};
			
			// aapt resource value: 2
			public const int LinearLayoutCompat_android_baselineAligned = 2;
			
			// aapt resource value: 3
			public const int LinearLayoutCompat_android_baselineAlignedChildIndex = 3;
			
			// aapt resource value: 0
			public const int LinearLayoutCompat_android_gravity = 0;
			
			// aapt resource value: 1
			public const int LinearLayoutCompat_android_orientation = 1;
			
			// aapt resource value: 4
			public const int LinearLayoutCompat_android_weightSum = 4;
			
			// aapt resource value: 5
			public const int LinearLayoutCompat_divider = 5;
			
			// aapt resource value: 6
			public const int LinearLayoutCompat_dividerPadding = 6;
			
			// aapt resource value: { 0x10100B3,0x10100F4,0x10100F5,0x1010181 }
			public static int[] LinearLayoutCompat_Layout = new int[] {
					16842931,
					16842996,
					16842997,
					16843137};
			
			// aapt resource value: 0
			public const int LinearLayoutCompat_Layout_android_layout_gravity = 0;
			
			// aapt resource value: 2
			public const int LinearLayoutCompat_Layout_android_layout_height = 2;
			
			// aapt resource value: 3
			public const int LinearLayoutCompat_Layout_android_layout_weight = 3;
			
			// aapt resource value: 1
			public const int LinearLayoutCompat_Layout_android_layout_width = 1;
			
			// aapt resource value: 7
			public const int LinearLayoutCompat_measureWithLargestChild = 7;
			
			// aapt resource value: 8
			public const int LinearLayoutCompat_showDividers = 8;
			
			// aapt resource value: { 0x7F0301D7,0x7F0301DB }
			public static int[] LinearProgressIndicator = new int[] {
					2130903511,
					2130903515};
			
			// aapt resource value: 0
			public const int LinearProgressIndicator_indeterminateAnimationType = 0;
			
			// aapt resource value: 1
			public const int LinearProgressIndicator_indicatorDirectionLinear = 1;
			
			// aapt resource value: { 0x10102AC,0x10102AD }
			public static int[] ListPopupWindow = new int[] {
					16843436,
					16843437};
			
			// aapt resource value: 0
			public const int ListPopupWindow_android_dropDownHorizontalOffset = 0;
			
			// aapt resource value: 1
			public const int ListPopupWindow_android_dropDownVerticalOffset = 1;
			
			// aapt resource value: { 0x10100B2,0x10101F8,0x7F030166,0x7F030167,0x7F030411 }
			public static int[] ListPreference = new int[] {
					16842930,
					16843256,
					2130903398,
					2130903399,
					2130904081};
			
			// aapt resource value: 0
			public const int ListPreference_android_entries = 0;
			
			// aapt resource value: 1
			public const int ListPreference_android_entryValues = 1;
			
			// aapt resource value: 2
			public const int ListPreference_entries = 2;
			
			// aapt resource value: 3
			public const int ListPreference_entryValues = 3;
			
			// aapt resource value: 4
			public const int ListPreference_useSimpleSummaryProvider = 4;
			
			// aapt resource value: { 0x7F030049,0x7F03004A,0x7F03004B,0x7F03004C }
			public static int[] MaterialAlertDialog = new int[] {
					2130903113,
					2130903114,
					2130903115,
					2130903116};
			
			// aapt resource value: { 0x7F030264,0x7F030265,0x7F030266,0x7F030267,0x7F030268 }
			public static int[] MaterialAlertDialogTheme = new int[] {
					2130903652,
					2130903653,
					2130903654,
					2130903655,
					2130903656};
			
			// aapt resource value: 0
			public const int MaterialAlertDialogTheme_materialAlertDialogBodyTextStyle = 0;
			
			// aapt resource value: 1
			public const int MaterialAlertDialogTheme_materialAlertDialogTheme = 1;
			
			// aapt resource value: 2
			public const int MaterialAlertDialogTheme_materialAlertDialogTitleIconStyle = 2;
			
			// aapt resource value: 3
			public const int MaterialAlertDialogTheme_materialAlertDialogTitlePanelStyle = 3;
			
			// aapt resource value: 4
			public const int MaterialAlertDialogTheme_materialAlertDialogTitleTextStyle = 4;
			
			// aapt resource value: 0
			public const int MaterialAlertDialog_backgroundInsetBottom = 0;
			
			// aapt resource value: 1
			public const int MaterialAlertDialog_backgroundInsetEnd = 1;
			
			// aapt resource value: 2
			public const int MaterialAlertDialog_backgroundInsetStart = 2;
			
			// aapt resource value: 3
			public const int MaterialAlertDialog_backgroundInsetTop = 3;
			
			// aapt resource value: { 0x1010220 }
			public static int[] MaterialAutoCompleteTextView = new int[] {
					16843296};
			
			// aapt resource value: 0
			public const int MaterialAutoCompleteTextView_android_inputType = 0;
			
			// aapt resource value: { 0x10100D4,0x10101B7,0x10101B8,0x10101B9,0x10101BA,0x10101E5,0x7F030050,0x7F030051,0x7F030109,0x7F030156,0x7F0301C6,0x7F0301C8,0x7F0301C9,0x7F0301CA,0x7F0301CD,0x7F0301CE,0x7F03031D,0x7F030333,0x7F030336,0x7F030369,0x7F03036A }
			public static int[] MaterialButton = new int[] {
					16842964,
					16843191,
					16843192,
					16843193,
					16843194,
					16843237,
					2130903120,
					2130903121,
					2130903305,
					2130903382,
					2130903494,
					2130903496,
					2130903497,
					2130903498,
					2130903501,
					2130903502,
					2130903837,
					2130903859,
					2130903862,
					2130903913,
					2130903914};
			
			// aapt resource value: { 0x7F03009B,0x7F030330,0x7F030347 }
			public static int[] MaterialButtonToggleGroup = new int[] {
					2130903195,
					2130903856,
					2130903879};
			
			// aapt resource value: 0
			public const int MaterialButtonToggleGroup_checkedButton = 0;
			
			// aapt resource value: 1
			public const int MaterialButtonToggleGroup_selectionRequired = 1;
			
			// aapt resource value: 2
			public const int MaterialButtonToggleGroup_singleSelection = 2;
			
			// aapt resource value: 0
			public const int MaterialButton_android_background = 0;
			
			// aapt resource value: 5
			public const int MaterialButton_android_checkable = 5;
			
			// aapt resource value: 4
			public const int MaterialButton_android_insetBottom = 4;
			
			// aapt resource value: 1
			public const int MaterialButton_android_insetLeft = 1;
			
			// aapt resource value: 2
			public const int MaterialButton_android_insetRight = 2;
			
			// aapt resource value: 3
			public const int MaterialButton_android_insetTop = 3;
			
			// aapt resource value: 6
			public const int MaterialButton_backgroundTint = 6;
			
			// aapt resource value: 7
			public const int MaterialButton_backgroundTintMode = 7;
			
			// aapt resource value: 8
			public const int MaterialButton_cornerRadius = 8;
			
			// aapt resource value: 9
			public const int MaterialButton_elevation = 9;
			
			// aapt resource value: 10
			public const int MaterialButton_icon = 10;
			
			// aapt resource value: 11
			public const int MaterialButton_iconGravity = 11;
			
			// aapt resource value: 12
			public const int MaterialButton_iconPadding = 12;
			
			// aapt resource value: 13
			public const int MaterialButton_iconSize = 13;
			
			// aapt resource value: 14
			public const int MaterialButton_iconTint = 14;
			
			// aapt resource value: 15
			public const int MaterialButton_iconTintMode = 15;
			
			// aapt resource value: 16
			public const int MaterialButton_rippleColor = 16;
			
			// aapt resource value: 17
			public const int MaterialButton_shapeAppearance = 17;
			
			// aapt resource value: 18
			public const int MaterialButton_shapeAppearanceOverlay = 18;
			
			// aapt resource value: 19
			public const int MaterialButton_strokeColor = 19;
			
			// aapt resource value: 20
			public const int MaterialButton_strokeWidth = 20;
			
			// aapt resource value: { 0x101020D,0x7F030124,0x7F030125,0x7F030126,0x7F030127,0x7F0302C2,0x7F03030F,0x7F03042D,0x7F03042E,0x7F03042F }
			public static int[] MaterialCalendar = new int[] {
					16843277,
					2130903332,
					2130903333,
					2130903334,
					2130903335,
					2130903746,
					2130903823,
					2130904109,
					2130904110,
					2130904111};
			
			// aapt resource value: { 0x10101B7,0x10101B8,0x10101B9,0x10101BA,0x7F0301E5,0x7F0301EE,0x7F0301EF,0x7F0301F6,0x7F0301F7,0x7F0301FB }
			public static int[] MaterialCalendarItem = new int[] {
					16843191,
					16843192,
					16843193,
					16843194,
					2130903525,
					2130903534,
					2130903535,
					2130903542,
					2130903543,
					2130903547};
			
			// aapt resource value: 3
			public const int MaterialCalendarItem_android_insetBottom = 3;
			
			// aapt resource value: 0
			public const int MaterialCalendarItem_android_insetLeft = 0;
			
			// aapt resource value: 1
			public const int MaterialCalendarItem_android_insetRight = 1;
			
			// aapt resource value: 2
			public const int MaterialCalendarItem_android_insetTop = 2;
			
			// aapt resource value: 4
			public const int MaterialCalendarItem_itemFillColor = 4;
			
			// aapt resource value: 5
			public const int MaterialCalendarItem_itemShapeAppearance = 5;
			
			// aapt resource value: 6
			public const int MaterialCalendarItem_itemShapeAppearanceOverlay = 6;
			
			// aapt resource value: 7
			public const int MaterialCalendarItem_itemStrokeColor = 7;
			
			// aapt resource value: 8
			public const int MaterialCalendarItem_itemStrokeWidth = 8;
			
			// aapt resource value: 9
			public const int MaterialCalendarItem_itemTextColor = 9;
			
			// aapt resource value: 0
			public const int MaterialCalendar_android_windowFullscreen = 0;
			
			// aapt resource value: 1
			public const int MaterialCalendar_dayInvalidStyle = 1;
			
			// aapt resource value: 2
			public const int MaterialCalendar_daySelectedStyle = 2;
			
			// aapt resource value: 3
			public const int MaterialCalendar_dayStyle = 3;
			
			// aapt resource value: 4
			public const int MaterialCalendar_dayTodayStyle = 4;
			
			// aapt resource value: 5
			public const int MaterialCalendar_nestedScrollable = 5;
			
			// aapt resource value: 6
			public const int MaterialCalendar_rangeFillColor = 6;
			
			// aapt resource value: 7
			public const int MaterialCalendar_yearSelectedStyle = 7;
			
			// aapt resource value: 8
			public const int MaterialCalendar_yearStyle = 8;
			
			// aapt resource value: 9
			public const int MaterialCalendar_yearTodayStyle = 9;
			
			// aapt resource value: { 0x10101E5,0x7F030089,0x7F03009D,0x7F03009F,0x7F0300A0,0x7F0300A1,0x7F03031D,0x7F030333,0x7F030336,0x7F030363,0x7F030369,0x7F03036A }
			public static int[] MaterialCardView = new int[] {
					16843237,
					2130903177,
					2130903197,
					2130903199,
					2130903200,
					2130903201,
					2130903837,
					2130903859,
					2130903862,
					2130903907,
					2130903913,
					2130903914};
			
			// aapt resource value: 0
			public const int MaterialCardView_android_checkable = 0;
			
			// aapt resource value: 1
			public const int MaterialCardView_cardForegroundColor = 1;
			
			// aapt resource value: 2
			public const int MaterialCardView_checkedIcon = 2;
			
			// aapt resource value: 3
			public const int MaterialCardView_checkedIconMargin = 3;
			
			// aapt resource value: 4
			public const int MaterialCardView_checkedIconSize = 4;
			
			// aapt resource value: 5
			public const int MaterialCardView_checkedIconTint = 5;
			
			// aapt resource value: 6
			public const int MaterialCardView_rippleColor = 6;
			
			// aapt resource value: 7
			public const int MaterialCardView_shapeAppearance = 7;
			
			// aapt resource value: 8
			public const int MaterialCardView_shapeAppearanceOverlay = 8;
			
			// aapt resource value: 9
			public const int MaterialCardView_state_dragged = 9;
			
			// aapt resource value: 10
			public const int MaterialCardView_strokeColor = 10;
			
			// aapt resource value: 11
			public const int MaterialCardView_strokeWidth = 11;
			
			// aapt resource value: { 0x7F030084,0x7F030410 }
			public static int[] MaterialCheckBox = new int[] {
					2130903172,
					2130904080};
			
			// aapt resource value: 0
			public const int MaterialCheckBox_buttonTint = 0;
			
			// aapt resource value: 1
			public const int MaterialCheckBox_useMaterialThemeColors = 1;
			
			// aapt resource value: { 0x7F030084,0x7F030410 }
			public static int[] MaterialRadioButton = new int[] {
					2130903172,
					2130904080};
			
			// aapt resource value: 0
			public const int MaterialRadioButton_buttonTint = 0;
			
			// aapt resource value: 1
			public const int MaterialRadioButton_useMaterialThemeColors = 1;
			
			// aapt resource value: { 0x7F030333,0x7F030336 }
			public static int[] MaterialShape = new int[] {
					2130903859,
					2130903862};
			
			// aapt resource value: 0
			public const int MaterialShape_shapeAppearance = 0;
			
			// aapt resource value: 1
			public const int MaterialShape_shapeAppearanceOverlay = 1;
			
			// aapt resource value: { 0x10104B6,0x101057F,0x7F030250 }
			public static int[] MaterialTextAppearance = new int[] {
					16843958,
					16844159,
					2130903632};
			
			// aapt resource value: 0
			public const int MaterialTextAppearance_android_letterSpacing = 0;
			
			// aapt resource value: 1
			public const int MaterialTextAppearance_android_lineHeight = 1;
			
			// aapt resource value: 2
			public const int MaterialTextAppearance_lineHeight = 2;
			
			// aapt resource value: { 0x1010034,0x101057F,0x7F030250 }
			public static int[] MaterialTextView = new int[] {
					16842804,
					16844159,
					2130903632};
			
			// aapt resource value: 1
			public const int MaterialTextView_android_lineHeight = 1;
			
			// aapt resource value: 0
			public const int MaterialTextView_android_textAppearance = 0;
			
			// aapt resource value: 2
			public const int MaterialTextView_lineHeight = 2;
			
			// aapt resource value: { 0x7F0300C3,0x7F0301FE }
			public static int[] MaterialTimePicker = new int[] {
					2130903235,
					2130903550};
			
			// aapt resource value: 0
			public const int MaterialTimePicker_clockIcon = 0;
			
			// aapt resource value: 1
			public const int MaterialTimePicker_keyboardIcon = 1;
			
			// aapt resource value: { 0x7F0302BB,0x7F03036E,0x7F0303E3 }
			public static int[] MaterialToolbar = new int[] {
					2130903739,
					2130903918,
					2130904035};
			
			// aapt resource value: 0
			public const int MaterialToolbar_navigationIconTint = 0;
			
			// aapt resource value: 1
			public const int MaterialToolbar_subtitleCentered = 1;
			
			// aapt resource value: 2
			public const int MaterialToolbar_titleCentered = 2;
			
			// aapt resource value: { 0x101000E,0x10100D0,0x1010194,0x10101DE,0x10101DF,0x10101E0 }
			public static int[] MenuGroup = new int[] {
					16842766,
					16842960,
					16843156,
					16843230,
					16843231,
					16843232};
			
			// aapt resource value: 5
			public const int MenuGroup_android_checkableBehavior = 5;
			
			// aapt resource value: 0
			public const int MenuGroup_android_enabled = 0;
			
			// aapt resource value: 1
			public const int MenuGroup_android_id = 1;
			
			// aapt resource value: 3
			public const int MenuGroup_android_menuCategory = 3;
			
			// aapt resource value: 4
			public const int MenuGroup_android_orderInCategory = 4;
			
			// aapt resource value: 2
			public const int MenuGroup_android_visible = 2;
			
			// aapt resource value: { 0x1010002,0x101000E,0x10100D0,0x1010106,0x1010194,0x10101DE,0x10101DF,0x10101E1,0x10101E2,0x10101E3,0x10101E4,0x10101E5,0x101026F,0x7F030011,0x7F030025,0x7F030027,0x7F030033,0x7F0300F2,0x7F0301CD,0x7F0301CE,0x7F0302C5,0x7F03033B,0x7F0303F5 }
			public static int[] MenuItem = new int[] {
					16842754,
					16842766,
					16842960,
					16843014,
					16843156,
					16843230,
					16843231,
					16843233,
					16843234,
					16843235,
					16843236,
					16843237,
					16843375,
					2130903057,
					2130903077,
					2130903079,
					2130903091,
					2130903282,
					2130903501,
					2130903502,
					2130903749,
					2130903867,
					2130904053};
			
			// aapt resource value: 13
			public const int MenuItem_actionLayout = 13;
			
			// aapt resource value: 14
			public const int MenuItem_actionProviderClass = 14;
			
			// aapt resource value: 15
			public const int MenuItem_actionViewClass = 15;
			
			// aapt resource value: 16
			public const int MenuItem_alphabeticModifiers = 16;
			
			// aapt resource value: 9
			public const int MenuItem_android_alphabeticShortcut = 9;
			
			// aapt resource value: 11
			public const int MenuItem_android_checkable = 11;
			
			// aapt resource value: 3
			public const int MenuItem_android_checked = 3;
			
			// aapt resource value: 1
			public const int MenuItem_android_enabled = 1;
			
			// aapt resource value: 0
			public const int MenuItem_android_icon = 0;
			
			// aapt resource value: 2
			public const int MenuItem_android_id = 2;
			
			// aapt resource value: 5
			public const int MenuItem_android_menuCategory = 5;
			
			// aapt resource value: 10
			public const int MenuItem_android_numericShortcut = 10;
			
			// aapt resource value: 12
			public const int MenuItem_android_onClick = 12;
			
			// aapt resource value: 6
			public const int MenuItem_android_orderInCategory = 6;
			
			// aapt resource value: 7
			public const int MenuItem_android_title = 7;
			
			// aapt resource value: 8
			public const int MenuItem_android_titleCondensed = 8;
			
			// aapt resource value: 4
			public const int MenuItem_android_visible = 4;
			
			// aapt resource value: 17
			public const int MenuItem_contentDescription = 17;
			
			// aapt resource value: 18
			public const int MenuItem_iconTint = 18;
			
			// aapt resource value: 19
			public const int MenuItem_iconTintMode = 19;
			
			// aapt resource value: 20
			public const int MenuItem_numericModifiers = 20;
			
			// aapt resource value: 21
			public const int MenuItem_showAsAction = 21;
			
			// aapt resource value: 22
			public const int MenuItem_tooltipText = 22;
			
			// aapt resource value: { 0x10100AE,0x101012C,0x101012D,0x101012E,0x101012F,0x1010130,0x1010131,0x7F030304,0x7F03036B }
			public static int[] MenuView = new int[] {
					16842926,
					16843052,
					16843053,
					16843054,
					16843055,
					16843056,
					16843057,
					2130903812,
					2130903915};
			
			// aapt resource value: 4
			public const int MenuView_android_headerBackground = 4;
			
			// aapt resource value: 2
			public const int MenuView_android_horizontalDivider = 2;
			
			// aapt resource value: 5
			public const int MenuView_android_itemBackground = 5;
			
			// aapt resource value: 6
			public const int MenuView_android_itemIconDisabledAlpha = 6;
			
			// aapt resource value: 1
			public const int MenuView_android_itemTextAppearance = 1;
			
			// aapt resource value: 3
			public const int MenuView_android_verticalDivider = 3;
			
			// aapt resource value: 0
			public const int MenuView_android_windowAnimationStyle = 0;
			
			// aapt resource value: 7
			public const int MenuView_preserveIconSpacing = 7;
			
			// aapt resource value: 8
			public const int MenuView_subMenuArrow = 8;
			
			// aapt resource value: { 0x7F030294,0x7F030295,0x7F030296,0x7F030297,0x7F030298,0x7F030299 }
			public static int[] MockView = new int[] {
					2130903700,
					2130903701,
					2130903702,
					2130903703,
					2130903704,
					2130903705};
			
			// aapt resource value: 0
			public const int MockView_mock_diagonalsColor = 0;
			
			// aapt resource value: 1
			public const int MockView_mock_label = 1;
			
			// aapt resource value: 2
			public const int MockView_mock_labelBackgroundColor = 2;
			
			// aapt resource value: 3
			public const int MockView_mock_labelColor = 3;
			
			// aapt resource value: 4
			public const int MockView_mock_showDiagonals = 4;
			
			// aapt resource value: 5
			public const int MockView_mock_showLabel = 5;
			
			// aapt resource value: { 0x7F030035,0x7F030036,0x7F030142,0x7F0302B0,0x7F0302B2,0x7F0302E1,0x7F030308,0x7F030309,0x7F03030A,0x7F030404 }
			public static int[] Motion = new int[] {
					2130903093,
					2130903094,
					2130903362,
					2130903728,
					2130903730,
					2130903777,
					2130903816,
					2130903817,
					2130903818,
					2130904068};
			
			// aapt resource value: { 0x7F0302A6,0x7F0302A7,0x7F0302A8,0x7F0302A9,0x7F0302AA,0x7F0302AB,0x7F0302AC,0x7F0302AD }
			public static int[] MotionEffect = new int[] {
					2130903718,
					2130903719,
					2130903720,
					2130903721,
					2130903722,
					2130903723,
					2130903724,
					2130903725};
			
			// aapt resource value: 0
			public const int MotionEffect_motionEffect_alpha = 0;
			
			// aapt resource value: 1
			public const int MotionEffect_motionEffect_end = 1;
			
			// aapt resource value: 2
			public const int MotionEffect_motionEffect_move = 2;
			
			// aapt resource value: 3
			public const int MotionEffect_motionEffect_start = 3;
			
			// aapt resource value: 4
			public const int MotionEffect_motionEffect_strict = 4;
			
			// aapt resource value: 5
			public const int MotionEffect_motionEffect_translationX = 5;
			
			// aapt resource value: 6
			public const int MotionEffect_motionEffect_translationY = 6;
			
			// aapt resource value: 7
			public const int MotionEffect_motionEffect_viewTransition = 7;
			
			// aapt resource value: { 0x7F0302C7,0x7F0302CA }
			public static int[] MotionHelper = new int[] {
					2130903751,
					2130903754};
			
			// aapt resource value: 0
			public const int MotionHelper_onHide = 0;
			
			// aapt resource value: 1
			public const int MotionHelper_onShow = 1;
			
			// aapt resource value: { 0x1010095,0x1010096,0x1010097,0x1010098,0x10100AF,0x101014F,0x1010164,0x10103AC,0x1010535,0x7F030065,0x7F030066,0x7F030322,0x7F0303B9,0x7F0303BA,0x7F0303BB,0x7F0303BC,0x7F0303BD,0x7F0303C5,0x7F0303C6,0x7F0303C7,0x7F0303C8,0x7F0303CA,0x7F0303CB,0x7F0303CC,0x7F0303CD }
			public static int[] MotionLabel = new int[] {
					16842901,
					16842902,
					16842903,
					16842904,
					16842927,
					16843087,
					16843108,
					16843692,
					16844085,
					2130903141,
					2130903142,
					2130903842,
					2130903993,
					2130903994,
					2130903995,
					2130903996,
					2130903997,
					2130904005,
					2130904006,
					2130904007,
					2130904008,
					2130904010,
					2130904011,
					2130904012,
					2130904013};
			
			// aapt resource value: 8
			public const int MotionLabel_android_autoSizeTextType = 8;
			
			// aapt resource value: 7
			public const int MotionLabel_android_fontFamily = 7;
			
			// aapt resource value: 4
			public const int MotionLabel_android_gravity = 4;
			
			// aapt resource value: 6
			public const int MotionLabel_android_shadowRadius = 6;
			
			// aapt resource value: 5
			public const int MotionLabel_android_text = 5;
			
			// aapt resource value: 3
			public const int MotionLabel_android_textColor = 3;
			
			// aapt resource value: 0
			public const int MotionLabel_android_textSize = 0;
			
			// aapt resource value: 2
			public const int MotionLabel_android_textStyle = 2;
			
			// aapt resource value: 1
			public const int MotionLabel_android_typeface = 1;
			
			// aapt resource value: 9
			public const int MotionLabel_borderRound = 9;
			
			// aapt resource value: 10
			public const int MotionLabel_borderRoundPercent = 10;
			
			// aapt resource value: 11
			public const int MotionLabel_scaleFromTextSize = 11;
			
			// aapt resource value: 12
			public const int MotionLabel_textBackground = 12;
			
			// aapt resource value: 13
			public const int MotionLabel_textBackgroundPanX = 13;
			
			// aapt resource value: 14
			public const int MotionLabel_textBackgroundPanY = 14;
			
			// aapt resource value: 15
			public const int MotionLabel_textBackgroundRotate = 15;
			
			// aapt resource value: 16
			public const int MotionLabel_textBackgroundZoom = 16;
			
			// aapt resource value: 17
			public const int MotionLabel_textOutlineColor = 17;
			
			// aapt resource value: 18
			public const int MotionLabel_textOutlineThickness = 18;
			
			// aapt resource value: 19
			public const int MotionLabel_textPanX = 19;
			
			// aapt resource value: 20
			public const int MotionLabel_textPanY = 20;
			
			// aapt resource value: 21
			public const int MotionLabel_textureBlurFactor = 21;
			
			// aapt resource value: 22
			public const int MotionLabel_textureEffect = 22;
			
			// aapt resource value: 23
			public const int MotionLabel_textureHeight = 23;
			
			// aapt resource value: 24
			public const int MotionLabel_textureWidth = 24;
			
			// aapt resource value: { 0x7F030039,0x7F030116,0x7F030206,0x7F03029A,0x7F0302B1,0x7F03033F }
			public static int[] MotionLayout = new int[] {
					2130903097,
					2130903318,
					2130903558,
					2130903706,
					2130903729,
					2130903871};
			
			// aapt resource value: 0
			public const int MotionLayout_applyMotionScene = 0;
			
			// aapt resource value: 1
			public const int MotionLayout_currentState = 1;
			
			// aapt resource value: 2
			public const int MotionLayout_layoutDescription = 2;
			
			// aapt resource value: 3
			public const int MotionLayout_motionDebug = 3;
			
			// aapt resource value: 4
			public const int MotionLayout_motionProgress = 4;
			
			// aapt resource value: 5
			public const int MotionLayout_showPaths = 5;
			
			// aapt resource value: { 0x7F030128,0x7F030207 }
			public static int[] MotionScene = new int[] {
					2130903336,
					2130903559};
			
			// aapt resource value: 0
			public const int MotionScene_defaultDuration = 0;
			
			// aapt resource value: 1
			public const int MotionScene_layoutDuringTransition = 1;
			
			// aapt resource value: { 0x7F03039F,0x7F0303A0,0x7F0303A1 }
			public static int[] MotionTelltales = new int[] {
					2130903967,
					2130903968,
					2130903969};
			
			// aapt resource value: 0
			public const int MotionTelltales_telltales_tailColor = 0;
			
			// aapt resource value: 1
			public const int MotionTelltales_telltales_tailScale = 1;
			
			// aapt resource value: 2
			public const int MotionTelltales_telltales_velocityMode = 2;
			
			// aapt resource value: 0
			public const int Motion_animateCircleAngleTo = 0;
			
			// aapt resource value: 1
			public const int Motion_animateRelativeTo = 1;
			
			// aapt resource value: 2
			public const int Motion_drawPath = 2;
			
			// aapt resource value: 3
			public const int Motion_motionPathRotate = 3;
			
			// aapt resource value: 4
			public const int Motion_motionStagger = 4;
			
			// aapt resource value: 5
			public const int Motion_pathMotionArc = 5;
			
			// aapt resource value: 6
			public const int Motion_quantizeMotionInterpolator = 6;
			
			// aapt resource value: 7
			public const int Motion_quantizeMotionPhase = 7;
			
			// aapt resource value: 8
			public const int Motion_quantizeMotionSteps = 8;
			
			// aapt resource value: 9
			public const int Motion_transitionEasing = 9;
			
			// aapt resource value: { 0x10100B2,0x10101F8,0x7F030166,0x7F030167 }
			public static int[] MultiSelectListPreference = new int[] {
					16842930,
					16843256,
					2130903398,
					2130903399};
			
			// aapt resource value: 0
			public const int MultiSelectListPreference_android_entries = 0;
			
			// aapt resource value: 1
			public const int MultiSelectListPreference_android_entryValues = 1;
			
			// aapt resource value: 2
			public const int MultiSelectListPreference_entries = 2;
			
			// aapt resource value: 3
			public const int MultiSelectListPreference_entryValues = 3;
			
			// aapt resource value: { 0x10100D0,0x7F030130,0x7F030165,0x7F03016F,0x7F030204,0x7F0302EF,0x7F0302F0,0x7F0302F1,0x7F0302F2 }
			public static int[] NavAction = new int[] {
					16842960,
					2130903344,
					2130903397,
					2130903407,
					2130903556,
					2130903791,
					2130903792,
					2130903793,
					2130903794};
			
			// aapt resource value: 0
			public const int NavAction_android_id = 0;
			
			// aapt resource value: 1
			public const int NavAction_destination = 1;
			
			// aapt resource value: 2
			public const int NavAction_enterAnim = 2;
			
			// aapt resource value: 3
			public const int NavAction_exitAnim = 3;
			
			// aapt resource value: 4
			public const int NavAction_launchSingleTop = 4;
			
			// aapt resource value: 5
			public const int NavAction_popEnterAnim = 5;
			
			// aapt resource value: 6
			public const int NavAction_popExitAnim = 6;
			
			// aapt resource value: 7
			public const int NavAction_popUpTo = 7;
			
			// aapt resource value: 8
			public const int NavAction_popUpToInclusive = 8;
			
			// aapt resource value: { 0x1010003,0x10101ED,0x7F03003B,0x7F0302C3 }
			public static int[] NavArgument = new int[] {
					16842755,
					16843245,
					2130903099,
					2130903747};
			
			// aapt resource value: 1
			public const int NavArgument_android_defaultValue = 1;
			
			// aapt resource value: 0
			public const int NavArgument_android_name = 0;
			
			// aapt resource value: 2
			public const int NavArgument_argType = 2;
			
			// aapt resource value: 3
			public const int NavArgument_nullable = 3;
			
			// aapt resource value: { 0x10104EE,0x7F030003,0x7F03028D,0x7F03040E }
			public static int[] NavDeepLink = new int[] {
					16844014,
					2130903043,
					2130903693,
					2130904078};
			
			// aapt resource value: 1
			public const int NavDeepLink_action = 1;
			
			// aapt resource value: 0
			public const int NavDeepLink_android_autoVerify = 0;
			
			// aapt resource value: 2
			public const int NavDeepLink_mimeType = 2;
			
			// aapt resource value: 3
			public const int NavDeepLink_uri = 3;
			
			// aapt resource value: { 0x7F03035A }
			public static int[] NavGraphNavigator = new int[] {
					2130903898};
			
			// aapt resource value: 0
			public const int NavGraphNavigator_startDestination = 0;
			
			// aapt resource value: { 0x7F0302B8 }
			public static int[] NavHost = new int[] {
					2130903736};
			
			// aapt resource value: 0
			public const int NavHost_navGraph = 0;
			
			// aapt resource value: { 0x7F030050,0x7F030156,0x7F0301E4,0x7F0301E9,0x7F0301EA,0x7F0301ED,0x7F0301F9,0x7F0301FA,0x7F0301FB,0x7F030202,0x7F03028A }
			public static int[] NavigationBarView = new int[] {
					2130903120,
					2130903382,
					2130903524,
					2130903529,
					2130903530,
					2130903533,
					2130903545,
					2130903546,
					2130903547,
					2130903554,
					2130903690};
			
			// aapt resource value: 0
			public const int NavigationBarView_backgroundTint = 0;
			
			// aapt resource value: 1
			public const int NavigationBarView_elevation = 1;
			
			// aapt resource value: 2
			public const int NavigationBarView_itemBackground = 2;
			
			// aapt resource value: 3
			public const int NavigationBarView_itemIconSize = 3;
			
			// aapt resource value: 4
			public const int NavigationBarView_itemIconTint = 4;
			
			// aapt resource value: 5
			public const int NavigationBarView_itemRippleColor = 5;
			
			// aapt resource value: 6
			public const int NavigationBarView_itemTextAppearanceActive = 6;
			
			// aapt resource value: 7
			public const int NavigationBarView_itemTextAppearanceInactive = 7;
			
			// aapt resource value: 8
			public const int NavigationBarView_itemTextColor = 8;
			
			// aapt resource value: 9
			public const int NavigationBarView_labelVisibilityMode = 9;
			
			// aapt resource value: 10
			public const int NavigationBarView_menu = 10;
			
			// aapt resource value: { 0x7F0301B4,0x7F03028B }
			public static int[] NavigationRailView = new int[] {
					2130903476,
					2130903691};
			
			// aapt resource value: 0
			public const int NavigationRailView_headerLayout = 0;
			
			// aapt resource value: 1
			public const int NavigationRailView_menuGravity = 1;
			
			// aapt resource value: { 0x10100D4,0x10100DD,0x101011F,0x7F030156,0x7F0301B4,0x7F0301E4,0x7F0301E6,0x7F0301E8,0x7F0301E9,0x7F0301EA,0x7F0301EB,0x7F0301EE,0x7F0301EF,0x7F0301F0,0x7F0301F1,0x7F0301F2,0x7F0301F3,0x7F0301F4,0x7F0301F8,0x7F0301FB,0x7F03028A,0x7F030333,0x7F030336 }
			public static int[] NavigationView = new int[] {
					16842964,
					16842973,
					16843039,
					2130903382,
					2130903476,
					2130903524,
					2130903526,
					2130903528,
					2130903529,
					2130903530,
					2130903531,
					2130903534,
					2130903535,
					2130903536,
					2130903537,
					2130903538,
					2130903539,
					2130903540,
					2130903544,
					2130903547,
					2130903690,
					2130903859,
					2130903862};
			
			// aapt resource value: 0
			public const int NavigationView_android_background = 0;
			
			// aapt resource value: 1
			public const int NavigationView_android_fitsSystemWindows = 1;
			
			// aapt resource value: 2
			public const int NavigationView_android_maxWidth = 2;
			
			// aapt resource value: 3
			public const int NavigationView_elevation = 3;
			
			// aapt resource value: 4
			public const int NavigationView_headerLayout = 4;
			
			// aapt resource value: 5
			public const int NavigationView_itemBackground = 5;
			
			// aapt resource value: 6
			public const int NavigationView_itemHorizontalPadding = 6;
			
			// aapt resource value: 7
			public const int NavigationView_itemIconPadding = 7;
			
			// aapt resource value: 8
			public const int NavigationView_itemIconSize = 8;
			
			// aapt resource value: 9
			public const int NavigationView_itemIconTint = 9;
			
			// aapt resource value: 10
			public const int NavigationView_itemMaxLines = 10;
			
			// aapt resource value: 11
			public const int NavigationView_itemShapeAppearance = 11;
			
			// aapt resource value: 12
			public const int NavigationView_itemShapeAppearanceOverlay = 12;
			
			// aapt resource value: 13
			public const int NavigationView_itemShapeFillColor = 13;
			
			// aapt resource value: 14
			public const int NavigationView_itemShapeInsetBottom = 14;
			
			// aapt resource value: 15
			public const int NavigationView_itemShapeInsetEnd = 15;
			
			// aapt resource value: 16
			public const int NavigationView_itemShapeInsetStart = 16;
			
			// aapt resource value: 17
			public const int NavigationView_itemShapeInsetTop = 17;
			
			// aapt resource value: 18
			public const int NavigationView_itemTextAppearance = 18;
			
			// aapt resource value: 19
			public const int NavigationView_itemTextColor = 19;
			
			// aapt resource value: 20
			public const int NavigationView_menu = 20;
			
			// aapt resource value: 21
			public const int NavigationView_shapeAppearance = 21;
			
			// aapt resource value: 22
			public const int NavigationView_shapeAppearanceOverlay = 22;
			
			// aapt resource value: { 0x1010001,0x10100D0 }
			public static int[] Navigator = new int[] {
					16842753,
					16842960};
			
			// aapt resource value: 1
			public const int Navigator_android_id = 1;
			
			// aapt resource value: 0
			public const int Navigator_android_label = 0;
			
			// aapt resource value: { 0x7F0301B1 }
			public static int[] NavInclude = new int[] {
					2130903473};
			
			// aapt resource value: 0
			public const int NavInclude_graph = 0;
			
			// aapt resource value: { 0x7F0300C0,0x7F03039D }
			public static int[] OnClick = new int[] {
					2130903232,
					2130903965};
			
			// aapt resource value: 0
			public const int OnClick_clickAction = 0;
			
			// aapt resource value: 1
			public const int OnClick_targetId = 1;
			
			// aapt resource value: { 0x7F03003F,0x7F03013F,0x7F030140,0x7F030141,0x7F03024F,0x7F030280,0x7F030287,0x7F0302B6,0x7F0302C0,0x7F0302CC,0x7F03031E,0x7F030352,0x7F030353,0x7F030354,0x7F030355,0x7F030356,0x7F0303F6,0x7F0303F7,0x7F0303F8 }
			public static int[] OnSwipe = new int[] {
					2130903103,
					2130903359,
					2130903360,
					2130903361,
					2130903631,
					2130903680,
					2130903687,
					2130903734,
					2130903744,
					2130903756,
					2130903838,
					2130903890,
					2130903891,
					2130903892,
					2130903893,
					2130903894,
					2130904054,
					2130904055,
					2130904056};
			
			// aapt resource value: 0
			public const int OnSwipe_autoCompleteMode = 0;
			
			// aapt resource value: 1
			public const int OnSwipe_dragDirection = 1;
			
			// aapt resource value: 2
			public const int OnSwipe_dragScale = 2;
			
			// aapt resource value: 3
			public const int OnSwipe_dragThreshold = 3;
			
			// aapt resource value: 4
			public const int OnSwipe_limitBoundsTo = 4;
			
			// aapt resource value: 5
			public const int OnSwipe_maxAcceleration = 5;
			
			// aapt resource value: 6
			public const int OnSwipe_maxVelocity = 6;
			
			// aapt resource value: 7
			public const int OnSwipe_moveWhenScrollAtTop = 7;
			
			// aapt resource value: 8
			public const int OnSwipe_nestedScrollFlags = 8;
			
			// aapt resource value: 9
			public const int OnSwipe_onTouchUp = 9;
			
			// aapt resource value: 10
			public const int OnSwipe_rotationCenterId = 10;
			
			// aapt resource value: 11
			public const int OnSwipe_springBoundary = 11;
			
			// aapt resource value: 12
			public const int OnSwipe_springDamping = 12;
			
			// aapt resource value: 13
			public const int OnSwipe_springMass = 13;
			
			// aapt resource value: 14
			public const int OnSwipe_springStiffness = 14;
			
			// aapt resource value: 15
			public const int OnSwipe_springStopThreshold = 15;
			
			// aapt resource value: 16
			public const int OnSwipe_touchAnchorId = 16;
			
			// aapt resource value: 17
			public const int OnSwipe_touchAnchorSide = 17;
			
			// aapt resource value: 18
			public const int OnSwipe_touchRegionId = 18;
			
			// aapt resource value: { 0x1010176,0x10102C9,0x7F0302CF }
			public static int[] PopupWindow = new int[] {
					16843126,
					16843465,
					2130903759};
			
			// aapt resource value: { 0x7F030360 }
			public static int[] PopupWindowBackgroundState = new int[] {
					2130903904};
			
			// aapt resource value: 0
			public const int PopupWindowBackgroundState_state_above_anchor = 0;
			
			// aapt resource value: 1
			public const int PopupWindow_android_popupAnimationStyle = 1;
			
			// aapt resource value: 0
			public const int PopupWindow_android_popupBackground = 0;
			
			// aapt resource value: 2
			public const int PopupWindow_overlapAnchor = 2;
			
			// aapt resource value: { 0x1010002,0x101000D,0x101000E,0x10100F2,0x10101E1,0x10101E6,0x10101E8,0x10101E9,0x10101EA,0x10101EB,0x10101EC,0x10101ED,0x10101EE,0x10102E3,0x101055C,0x1010561,0x7F03002E,0x7F030030,0x7F03012B,0x7F03012E,0x7F030159,0x7F03015B,0x7F0301AC,0x7F0301C6,0x7F0301CB,0x7F0301E3,0x7F0301FC,0x7F030205,0x7F0302CD,0x7F0302E8,0x7F03032D,0x7F030339,0x7F030346,0x7F030376,0x7F0303E2,0x7F030422 }
			public static int[] Preference = new int[] {
					16842754,
					16842765,
					16842766,
					16842994,
					16843233,
					16843238,
					16843240,
					16843241,
					16843242,
					16843243,
					16843244,
					16843245,
					16843246,
					16843491,
					16844124,
					16844129,
					2130903086,
					2130903088,
					2130903339,
					2130903342,
					2130903385,
					2130903387,
					2130903468,
					2130903494,
					2130903499,
					2130903523,
					2130903548,
					2130903557,
					2130903757,
					2130903784,
					2130903853,
					2130903865,
					2130903878,
					2130903926,
					2130904034,
					2130904098};
			
			// aapt resource value: { 0x10100F2,0x1010129,0x101012A,0x7F03002F }
			public static int[] PreferenceFragment = new int[] {
					16842994,
					16843049,
					16843050,
					2130903087};
			
			// aapt resource value: { 0x10100F2,0x1010129,0x101012A,0x7F03002F }
			public static int[] PreferenceFragmentCompat = new int[] {
					16842994,
					16843049,
					16843050,
					2130903087};
			
			// aapt resource value: 3
			public const int PreferenceFragmentCompat_allowDividerAfterLastItem = 3;
			
			// aapt resource value: 1
			public const int PreferenceFragmentCompat_android_divider = 1;
			
			// aapt resource value: 2
			public const int PreferenceFragmentCompat_android_dividerHeight = 2;
			
			// aapt resource value: 0
			public const int PreferenceFragmentCompat_android_layout = 0;
			
			// aapt resource value: 3
			public const int PreferenceFragment_allowDividerAfterLastItem = 3;
			
			// aapt resource value: 1
			public const int PreferenceFragment_android_divider = 1;
			
			// aapt resource value: 2
			public const int PreferenceFragment_android_dividerHeight = 2;
			
			// aapt resource value: 0
			public const int PreferenceFragment_android_layout = 0;
			
			// aapt resource value: { 0x10101E7,0x7F0301DF,0x7F0302CE }
			public static int[] PreferenceGroup = new int[] {
					16843239,
					2130903519,
					2130903758};
			
			// aapt resource value: 0
			public const int PreferenceGroup_android_orderingFromXml = 0;
			
			// aapt resource value: 1
			public const int PreferenceGroup_initialExpandedChildrenCount = 1;
			
			// aapt resource value: 2
			public const int PreferenceGroup_orderingFromXml = 2;
			
			// aapt resource value: { 0x101011F,0x1010120,0x7F030284,0x7F030288 }
			public static int[] PreferenceImageView = new int[] {
					16843039,
					16843040,
					2130903684,
					2130903688};
			
			// aapt resource value: 1
			public const int PreferenceImageView_android_maxHeight = 1;
			
			// aapt resource value: 0
			public const int PreferenceImageView_android_maxWidth = 0;
			
			// aapt resource value: 2
			public const int PreferenceImageView_maxHeight = 2;
			
			// aapt resource value: 3
			public const int PreferenceImageView_maxWidth = 3;
			
			// aapt resource value: { 0x7F030099,0x7F030135,0x7F030150,0x7F030154,0x7F0302F8,0x7F0302F9,0x7F0302FA,0x7F0302FB,0x7F0302FC,0x7F0302FD,0x7F0302FE,0x7F0302FF,0x7F030300,0x7F03032B,0x7F03037C,0x7F03037D }
			public static int[] PreferenceTheme = new int[] {
					2130903193,
					2130903349,
					2130903376,
					2130903380,
					2130903800,
					2130903801,
					2130903802,
					2130903803,
					2130903804,
					2130903805,
					2130903806,
					2130903807,
					2130903808,
					2130903851,
					2130903932,
					2130903933};
			
			// aapt resource value: 0
			public const int PreferenceTheme_checkBoxPreferenceStyle = 0;
			
			// aapt resource value: 1
			public const int PreferenceTheme_dialogPreferenceStyle = 1;
			
			// aapt resource value: 2
			public const int PreferenceTheme_dropdownPreferenceStyle = 2;
			
			// aapt resource value: 3
			public const int PreferenceTheme_editTextPreferenceStyle = 3;
			
			// aapt resource value: 4
			public const int PreferenceTheme_preferenceCategoryStyle = 4;
			
			// aapt resource value: 5
			public const int PreferenceTheme_preferenceCategoryTitleTextAppearance = 5;
			
			// aapt resource value: 6
			public const int PreferenceTheme_preferenceFragmentCompatStyle = 6;
			
			// aapt resource value: 7
			public const int PreferenceTheme_preferenceFragmentListStyle = 7;
			
			// aapt resource value: 8
			public const int PreferenceTheme_preferenceFragmentStyle = 8;
			
			// aapt resource value: 9
			public const int PreferenceTheme_preferenceInformationStyle = 9;
			
			// aapt resource value: 10
			public const int PreferenceTheme_preferenceScreenStyle = 10;
			
			// aapt resource value: 11
			public const int PreferenceTheme_preferenceStyle = 11;
			
			// aapt resource value: 12
			public const int PreferenceTheme_preferenceTheme = 12;
			
			// aapt resource value: 13
			public const int PreferenceTheme_seekBarPreferenceStyle = 13;
			
			// aapt resource value: 14
			public const int PreferenceTheme_switchPreferenceCompatStyle = 14;
			
			// aapt resource value: 15
			public const int PreferenceTheme_switchPreferenceStyle = 15;
			
			// aapt resource value: 16
			public const int Preference_allowDividerAbove = 16;
			
			// aapt resource value: 17
			public const int Preference_allowDividerBelow = 17;
			
			// aapt resource value: 11
			public const int Preference_android_defaultValue = 11;
			
			// aapt resource value: 10
			public const int Preference_android_dependency = 10;
			
			// aapt resource value: 2
			public const int Preference_android_enabled = 2;
			
			// aapt resource value: 13
			public const int Preference_android_fragment = 13;
			
			// aapt resource value: 0
			public const int Preference_android_icon = 0;
			
			// aapt resource value: 15
			public const int Preference_android_iconSpaceReserved = 15;
			
			// aapt resource value: 6
			public const int Preference_android_key = 6;
			
			// aapt resource value: 3
			public const int Preference_android_layout = 3;
			
			// aapt resource value: 8
			public const int Preference_android_order = 8;
			
			// aapt resource value: 1
			public const int Preference_android_persistent = 1;
			
			// aapt resource value: 5
			public const int Preference_android_selectable = 5;
			
			// aapt resource value: 12
			public const int Preference_android_shouldDisableView = 12;
			
			// aapt resource value: 14
			public const int Preference_android_singleLineTitle = 14;
			
			// aapt resource value: 7
			public const int Preference_android_summary = 7;
			
			// aapt resource value: 4
			public const int Preference_android_title = 4;
			
			// aapt resource value: 9
			public const int Preference_android_widgetLayout = 9;
			
			// aapt resource value: 18
			public const int Preference_defaultValue = 18;
			
			// aapt resource value: 19
			public const int Preference_dependency = 19;
			
			// aapt resource value: 20
			public const int Preference_enableCopying = 20;
			
			// aapt resource value: 21
			public const int Preference_enabled = 21;
			
			// aapt resource value: 22
			public const int Preference_fragment = 22;
			
			// aapt resource value: 23
			public const int Preference_icon = 23;
			
			// aapt resource value: 24
			public const int Preference_iconSpaceReserved = 24;
			
			// aapt resource value: 25
			public const int Preference_isPreferenceVisible = 25;
			
			// aapt resource value: 26
			public const int Preference_key = 26;
			
			// aapt resource value: 27
			public const int Preference_layout = 27;
			
			// aapt resource value: 28
			public const int Preference_order = 28;
			
			// aapt resource value: 29
			public const int Preference_persistent = 29;
			
			// aapt resource value: 30
			public const int Preference_selectable = 30;
			
			// aapt resource value: 31
			public const int Preference_shouldDisableView = 31;
			
			// aapt resource value: 32
			public const int Preference_singleLineTitle = 32;
			
			// aapt resource value: 33
			public const int Preference_summary = 33;
			
			// aapt resource value: 34
			public const int Preference_title = 34;
			
			// aapt resource value: 35
			public const int Preference_widgetLayout = 35;
			
			// aapt resource value: { 0x10100DC,0x101031F,0x7F030230,0x7F0302B1,0x7F030419 }
			public static int[] PropertySet = new int[] {
					16842972,
					16843551,
					2130903600,
					2130903729,
					2130904089};
			
			// aapt resource value: 1
			public const int PropertySet_android_alpha = 1;
			
			// aapt resource value: 0
			public const int PropertySet_android_visibility = 0;
			
			// aapt resource value: 2
			public const int PropertySet_layout_constraintTag = 2;
			
			// aapt resource value: 3
			public const int PropertySet_motionProgress = 3;
			
			// aapt resource value: 4
			public const int PropertySet_visibilityMode = 4;
			
			// aapt resource value: { 0x7F03027B }
			public static int[] RadialViewGroup = new int[] {
					2130903675};
			
			// aapt resource value: 0
			public const int RadialViewGroup_materialCircleRadius = 0;
			
			// aapt resource value: { 0x7F030291,0x7F030412 }
			public static int[] RangeSlider = new int[] {
					2130903697,
					2130904082};
			
			// aapt resource value: 0
			public const int RangeSlider_minSeparation = 0;
			
			// aapt resource value: 1
			public const int RangeSlider_values = 1;
			
			// aapt resource value: { 0x7F0302D1,0x7F0302D7 }
			public static int[] RecycleListView = new int[] {
					2130903761,
					2130903767};
			
			// aapt resource value: 0
			public const int RecycleListView_paddingBottomNoButtons = 0;
			
			// aapt resource value: 1
			public const int RecycleListView_paddingTopNoTitle = 1;
			
			// aapt resource value: { 0x10100C4,0x10100EB,0x10100F1,0x7F030184,0x7F030185,0x7F030186,0x7F030187,0x7F030188,0x7F030208,0x7F03031C,0x7F03034D,0x7F030358 }
			public static int[] RecyclerView = new int[] {
					16842948,
					16842987,
					16842993,
					2130903428,
					2130903429,
					2130903430,
					2130903431,
					2130903432,
					2130903560,
					2130903836,
					2130903885,
					2130903896};
			
			// aapt resource value: 1
			public const int RecyclerView_android_clipToPadding = 1;
			
			// aapt resource value: 2
			public const int RecyclerView_android_descendantFocusability = 2;
			
			// aapt resource value: 0
			public const int RecyclerView_android_orientation = 0;
			
			// aapt resource value: 3
			public const int RecyclerView_fastScrollEnabled = 3;
			
			// aapt resource value: 4
			public const int RecyclerView_fastScrollHorizontalThumbDrawable = 4;
			
			// aapt resource value: 5
			public const int RecyclerView_fastScrollHorizontalTrackDrawable = 5;
			
			// aapt resource value: 6
			public const int RecyclerView_fastScrollVerticalThumbDrawable = 6;
			
			// aapt resource value: 7
			public const int RecyclerView_fastScrollVerticalTrackDrawable = 7;
			
			// aapt resource value: 8
			public const int RecyclerView_layoutManager = 8;
			
			// aapt resource value: 9
			public const int RecyclerView_reverseLayout = 9;
			
			// aapt resource value: 10
			public const int RecyclerView_spanCount = 10;
			
			// aapt resource value: 11
			public const int RecyclerView_stackFromEnd = 11;
			
			// aapt resource value: { 0x7F0301E0 }
			public static int[] ScrimInsetsFrameLayout = new int[] {
					2130903520};
			
			// aapt resource value: 0
			public const int ScrimInsetsFrameLayout_insetForeground = 0;
			
			// aapt resource value: { 0x7F030060 }
			public static int[] ScrollingViewBehavior_Layout = new int[] {
					2130903136};
			
			// aapt resource value: 0
			public const int ScrollingViewBehavior_Layout_behavior_overlapTop = 0;
			
			// aapt resource value: { 0x7F030326 }
			public static int[] ScrollViewRendererTheme = new int[] {
					2130903846};
			
			// aapt resource value: 0
			public const int ScrollViewRendererTheme_scrollViewStyle = 0;
			
			// aapt resource value: { 0x10100DA,0x101011F,0x1010220,0x1010264,0x7F0300C5,0x7F0300EA,0x7F030129,0x7F0301B0,0x7F0301CF,0x7F030205,0x7F03030B,0x7F03030C,0x7F030327,0x7F030328,0x7F03036C,0x7F030375,0x7F03041A }
			public static int[] SearchView = new int[] {
					16842970,
					16843039,
					16843296,
					16843364,
					2130903237,
					2130903274,
					2130903337,
					2130903472,
					2130903503,
					2130903557,
					2130903819,
					2130903820,
					2130903847,
					2130903848,
					2130903916,
					2130903925,
					2130904090};
			
			// aapt resource value: 0
			public const int SearchView_android_focusable = 0;
			
			// aapt resource value: 3
			public const int SearchView_android_imeOptions = 3;
			
			// aapt resource value: 2
			public const int SearchView_android_inputType = 2;
			
			// aapt resource value: 1
			public const int SearchView_android_maxWidth = 1;
			
			// aapt resource value: 4
			public const int SearchView_closeIcon = 4;
			
			// aapt resource value: 5
			public const int SearchView_commitIcon = 5;
			
			// aapt resource value: 6
			public const int SearchView_defaultQueryHint = 6;
			
			// aapt resource value: 7
			public const int SearchView_goIcon = 7;
			
			// aapt resource value: 8
			public const int SearchView_iconifiedByDefault = 8;
			
			// aapt resource value: 9
			public const int SearchView_layout = 9;
			
			// aapt resource value: 10
			public const int SearchView_queryBackground = 10;
			
			// aapt resource value: 11
			public const int SearchView_queryHint = 11;
			
			// aapt resource value: 12
			public const int SearchView_searchHintIcon = 12;
			
			// aapt resource value: 13
			public const int SearchView_searchIcon = 13;
			
			// aapt resource value: 14
			public const int SearchView_submitBackground = 14;
			
			// aapt resource value: 15
			public const int SearchView_suggestionRowLayout = 15;
			
			// aapt resource value: 16
			public const int SearchView_voiceIcon = 16;
			
			// aapt resource value: { 0x10100F2,0x1010136,0x7F030029,0x7F03028E,0x7F03032A,0x7F030340,0x7F03040D }
			public static int[] SeekBarPreference = new int[] {
					16842994,
					16843062,
					2130903081,
					2130903694,
					2130903850,
					2130903872,
					2130904077};
			
			// aapt resource value: 2
			public const int SeekBarPreference_adjustable = 2;
			
			// aapt resource value: 0
			public const int SeekBarPreference_android_layout = 0;
			
			// aapt resource value: 1
			public const int SeekBarPreference_android_max = 1;
			
			// aapt resource value: 3
			public const int SeekBarPreference_min = 3;
			
			// aapt resource value: 4
			public const int SeekBarPreference_seekBarIncrement = 4;
			
			// aapt resource value: 5
			public const int SeekBarPreference_showSeekBarValue = 5;
			
			// aapt resource value: 6
			public const int SeekBarPreference_updatesContinuously = 6;
			
			// aapt resource value: { 0x7F0300F9,0x7F0300FA,0x7F0300FB,0x7F0300FC,0x7F0300FD,0x7F0300FE,0x7F0300FF,0x7F030333,0x7F030336,0x7F030369,0x7F03036A }
			public static int[] ShapeableImageView = new int[] {
					2130903289,
					2130903290,
					2130903291,
					2130903292,
					2130903293,
					2130903294,
					2130903295,
					2130903859,
					2130903862,
					2130903913,
					2130903914};
			
			// aapt resource value: 0
			public const int ShapeableImageView_contentPadding = 0;
			
			// aapt resource value: 1
			public const int ShapeableImageView_contentPaddingBottom = 1;
			
			// aapt resource value: 2
			public const int ShapeableImageView_contentPaddingEnd = 2;
			
			// aapt resource value: 3
			public const int ShapeableImageView_contentPaddingLeft = 3;
			
			// aapt resource value: 4
			public const int ShapeableImageView_contentPaddingRight = 4;
			
			// aapt resource value: 5
			public const int ShapeableImageView_contentPaddingStart = 5;
			
			// aapt resource value: 6
			public const int ShapeableImageView_contentPaddingTop = 6;
			
			// aapt resource value: 7
			public const int ShapeableImageView_shapeAppearance = 7;
			
			// aapt resource value: 8
			public const int ShapeableImageView_shapeAppearanceOverlay = 8;
			
			// aapt resource value: 9
			public const int ShapeableImageView_strokeColor = 9;
			
			// aapt resource value: 10
			public const int ShapeableImageView_strokeWidth = 10;
			
			// aapt resource value: { 0x7F030104,0x7F030105,0x7F030106,0x7F030107,0x7F030108,0x7F03010A,0x7F03010B,0x7F03010C,0x7F03010D,0x7F03010E }
			public static int[] ShapeAppearance = new int[] {
					2130903300,
					2130903301,
					2130903302,
					2130903303,
					2130903304,
					2130903306,
					2130903307,
					2130903308,
					2130903309,
					2130903310};
			
			// aapt resource value: 0
			public const int ShapeAppearance_cornerFamily = 0;
			
			// aapt resource value: 1
			public const int ShapeAppearance_cornerFamilyBottomLeft = 1;
			
			// aapt resource value: 2
			public const int ShapeAppearance_cornerFamilyBottomRight = 2;
			
			// aapt resource value: 3
			public const int ShapeAppearance_cornerFamilyTopLeft = 3;
			
			// aapt resource value: 4
			public const int ShapeAppearance_cornerFamilyTopRight = 4;
			
			// aapt resource value: 5
			public const int ShapeAppearance_cornerSize = 5;
			
			// aapt resource value: 6
			public const int ShapeAppearance_cornerSizeBottomLeft = 6;
			
			// aapt resource value: 7
			public const int ShapeAppearance_cornerSizeBottomRight = 7;
			
			// aapt resource value: 8
			public const int ShapeAppearance_cornerSizeTopLeft = 8;
			
			// aapt resource value: 9
			public const int ShapeAppearance_cornerSizeTopRight = 9;
			
			// aapt resource value: { 0x101000E,0x1010024,0x1010146,0x10102DE,0x10102DF,0x7F0301B2,0x7F0301B3,0x7F030200,0x7F030201,0x7F0303D1,0x7F0303D2,0x7F0303D3,0x7F0303D4,0x7F0303D5,0x7F0303D9,0x7F0303DA,0x7F0303DB,0x7F0303DF,0x7F0303FA,0x7F0303FB,0x7F0303FC,0x7F0303FE }
			public static int[] Slider = new int[] {
					16842766,
					16842788,
					16843078,
					16843486,
					16843487,
					2130903474,
					2130903475,
					2130903552,
					2130903553,
					2130904017,
					2130904018,
					2130904019,
					2130904020,
					2130904021,
					2130904025,
					2130904026,
					2130904027,
					2130904031,
					2130904058,
					2130904059,
					2130904060,
					2130904062};
			
			// aapt resource value: 0
			public const int Slider_android_enabled = 0;
			
			// aapt resource value: 2
			public const int Slider_android_stepSize = 2;
			
			// aapt resource value: 1
			public const int Slider_android_value = 1;
			
			// aapt resource value: 3
			public const int Slider_android_valueFrom = 3;
			
			// aapt resource value: 4
			public const int Slider_android_valueTo = 4;
			
			// aapt resource value: 5
			public const int Slider_haloColor = 5;
			
			// aapt resource value: 6
			public const int Slider_haloRadius = 6;
			
			// aapt resource value: 7
			public const int Slider_labelBehavior = 7;
			
			// aapt resource value: 8
			public const int Slider_labelStyle = 8;
			
			// aapt resource value: 9
			public const int Slider_thumbColor = 9;
			
			// aapt resource value: 10
			public const int Slider_thumbElevation = 10;
			
			// aapt resource value: 11
			public const int Slider_thumbRadius = 11;
			
			// aapt resource value: 12
			public const int Slider_thumbStrokeColor = 12;
			
			// aapt resource value: 13
			public const int Slider_thumbStrokeWidth = 13;
			
			// aapt resource value: 14
			public const int Slider_tickColor = 14;
			
			// aapt resource value: 15
			public const int Slider_tickColorActive = 15;
			
			// aapt resource value: 16
			public const int Slider_tickColorInactive = 16;
			
			// aapt resource value: 17
			public const int Slider_tickVisible = 17;
			
			// aapt resource value: 18
			public const int Slider_trackColor = 18;
			
			// aapt resource value: 19
			public const int Slider_trackColorActive = 19;
			
			// aapt resource value: 20
			public const int Slider_trackColorInactive = 20;
			
			// aapt resource value: 21
			public const int Slider_trackHeight = 21;
			
			// aapt resource value: { 0x7F03034A,0x7F03034B,0x7F03034C }
			public static int[] Snackbar = new int[] {
					2130903882,
					2130903883,
					2130903884};
			
			// aapt resource value: { 0x101011F,0x7F030026,0x7F030037,0x7F03004D,0x7F030050,0x7F030051,0x7F030156,0x7F030281 }
			public static int[] SnackbarLayout = new int[] {
					16843039,
					2130903078,
					2130903095,
					2130903117,
					2130903120,
					2130903121,
					2130903382,
					2130903681};
			
			// aapt resource value: 1
			public const int SnackbarLayout_actionTextColorAlpha = 1;
			
			// aapt resource value: 0
			public const int SnackbarLayout_android_maxWidth = 0;
			
			// aapt resource value: 2
			public const int SnackbarLayout_animationMode = 2;
			
			// aapt resource value: 3
			public const int SnackbarLayout_backgroundOverlayColorAlpha = 3;
			
			// aapt resource value: 4
			public const int SnackbarLayout_backgroundTint = 4;
			
			// aapt resource value: 5
			public const int SnackbarLayout_backgroundTintMode = 5;
			
			// aapt resource value: 6
			public const int SnackbarLayout_elevation = 6;
			
			// aapt resource value: 7
			public const int SnackbarLayout_maxActionInlineWidth = 7;
			
			// aapt resource value: 0
			public const int Snackbar_snackbarButtonStyle = 0;
			
			// aapt resource value: 1
			public const int Snackbar_snackbarStyle = 1;
			
			// aapt resource value: 2
			public const int Snackbar_snackbarTextViewStyle = 2;
			
			// aapt resource value: { 0x10100B2,0x1010176,0x101017B,0x1010262,0x7F0302F5 }
			public static int[] Spinner = new int[] {
					16842930,
					16843126,
					16843131,
					16843362,
					2130903797};
			
			// aapt resource value: 3
			public const int Spinner_android_dropDownWidth = 3;
			
			// aapt resource value: 0
			public const int Spinner_android_entries = 0;
			
			// aapt resource value: 1
			public const int Spinner_android_popupBackground = 1;
			
			// aapt resource value: 2
			public const int Spinner_android_prompt = 2;
			
			// aapt resource value: 4
			public const int Spinner_popupTheme = 4;
			
			// aapt resource value: { 0x10100D0,0x7F0300F0 }
			public static int[] State = new int[] {
					16842960,
					2130903280};
			
			// aapt resource value: { 0x101011C,0x1010194,0x1010195,0x1010196,0x101030C,0x101030D }
			public static int[] StateListDrawable = new int[] {
					16843036,
					16843156,
					16843157,
					16843158,
					16843532,
					16843533};
			
			// aapt resource value: { 0x1010199 }
			public static int[] StateListDrawableItem = new int[] {
					16843161};
			
			// aapt resource value: 0
			public const int StateListDrawableItem_android_drawable = 0;
			
			// aapt resource value: 3
			public const int StateListDrawable_android_constantSize = 3;
			
			// aapt resource value: 0
			public const int StateListDrawable_android_dither = 0;
			
			// aapt resource value: 4
			public const int StateListDrawable_android_enterFadeDuration = 4;
			
			// aapt resource value: 5
			public const int StateListDrawable_android_exitFadeDuration = 5;
			
			// aapt resource value: 2
			public const int StateListDrawable_android_variablePadding = 2;
			
			// aapt resource value: 1
			public const int StateListDrawable_android_visible = 1;
			
			// aapt resource value: { 0x7F03012A }
			public static int[] StateSet = new int[] {
					2130903338};
			
			// aapt resource value: 0
			public const int StateSet_defaultState = 0;
			
			// aapt resource value: 0
			public const int State_android_id = 0;
			
			// aapt resource value: 1
			public const int State_constraints = 1;
			
			// aapt resource value: { 0x7F030379 }
			public static int[] SwipeRefreshLayout = new int[] {
					2130903929};
			
			// aapt resource value: 0
			public const int SwipeRefreshLayout_swipeRefreshLayoutProgressSpinnerBackgroundColor = 0;
			
			// aapt resource value: { 0x1010124,0x1010125,0x1010142,0x7F030341,0x7F030351,0x7F03037A,0x7F03037B,0x7F03037F,0x7F0303D6,0x7F0303D7,0x7F0303D8,0x7F0303F9,0x7F030400,0x7F030401 }
			public static int[] SwitchCompat = new int[] {
					16843044,
					16843045,
					16843074,
					2130903873,
					2130903889,
					2130903930,
					2130903931,
					2130903935,
					2130904022,
					2130904023,
					2130904024,
					2130904057,
					2130904064,
					2130904065};
			
			// aapt resource value: 1
			public const int SwitchCompat_android_textOff = 1;
			
			// aapt resource value: 0
			public const int SwitchCompat_android_textOn = 0;
			
			// aapt resource value: 2
			public const int SwitchCompat_android_thumb = 2;
			
			// aapt resource value: 3
			public const int SwitchCompat_showText = 3;
			
			// aapt resource value: 4
			public const int SwitchCompat_splitTrack = 4;
			
			// aapt resource value: 5
			public const int SwitchCompat_switchMinWidth = 5;
			
			// aapt resource value: 6
			public const int SwitchCompat_switchPadding = 6;
			
			// aapt resource value: 7
			public const int SwitchCompat_switchTextAppearance = 7;
			
			// aapt resource value: 8
			public const int SwitchCompat_thumbTextPadding = 8;
			
			// aapt resource value: 9
			public const int SwitchCompat_thumbTint = 9;
			
			// aapt resource value: 10
			public const int SwitchCompat_thumbTintMode = 10;
			
			// aapt resource value: 11
			public const int SwitchCompat_track = 11;
			
			// aapt resource value: 12
			public const int SwitchCompat_trackTint = 12;
			
			// aapt resource value: 13
			public const int SwitchCompat_trackTintMode = 13;
			
			// aapt resource value: { 0x7F030410 }
			public static int[] SwitchMaterial = new int[] {
					2130904080};
			
			// aapt resource value: 0
			public const int SwitchMaterial_useMaterialThemeColors = 0;
			
			// aapt resource value: { 0x10101EF,0x10101F0,0x10101F1,0x101036B,0x101036C,0x7F030139,0x7F030377,0x7F030378,0x7F030380,0x7F030381 }
			public static int[] SwitchPreference = new int[] {
					16843247,
					16843248,
					16843249,
					16843627,
					16843628,
					2130903353,
					2130903927,
					2130903928,
					2130903936,
					2130903937};
			
			// aapt resource value: { 0x10101EF,0x10101F0,0x10101F1,0x101036B,0x101036C,0x7F030139,0x7F030377,0x7F030378,0x7F030380,0x7F030381 }
			public static int[] SwitchPreferenceCompat = new int[] {
					16843247,
					16843248,
					16843249,
					16843627,
					16843628,
					2130903353,
					2130903927,
					2130903928,
					2130903936,
					2130903937};
			
			// aapt resource value: 2
			public const int SwitchPreferenceCompat_android_disableDependentsState = 2;
			
			// aapt resource value: 1
			public const int SwitchPreferenceCompat_android_summaryOff = 1;
			
			// aapt resource value: 0
			public const int SwitchPreferenceCompat_android_summaryOn = 0;
			
			// aapt resource value: 4
			public const int SwitchPreferenceCompat_android_switchTextOff = 4;
			
			// aapt resource value: 3
			public const int SwitchPreferenceCompat_android_switchTextOn = 3;
			
			// aapt resource value: 5
			public const int SwitchPreferenceCompat_disableDependentsState = 5;
			
			// aapt resource value: 6
			public const int SwitchPreferenceCompat_summaryOff = 6;
			
			// aapt resource value: 7
			public const int SwitchPreferenceCompat_summaryOn = 7;
			
			// aapt resource value: 8
			public const int SwitchPreferenceCompat_switchTextOff = 8;
			
			// aapt resource value: 9
			public const int SwitchPreferenceCompat_switchTextOn = 9;
			
			// aapt resource value: 2
			public const int SwitchPreference_android_disableDependentsState = 2;
			
			// aapt resource value: 1
			public const int SwitchPreference_android_summaryOff = 1;
			
			// aapt resource value: 0
			public const int SwitchPreference_android_summaryOn = 0;
			
			// aapt resource value: 4
			public const int SwitchPreference_android_switchTextOff = 4;
			
			// aapt resource value: 3
			public const int SwitchPreference_android_switchTextOn = 3;
			
			// aapt resource value: 5
			public const int SwitchPreference_disableDependentsState = 5;
			
			// aapt resource value: 6
			public const int SwitchPreference_summaryOff = 6;
			
			// aapt resource value: 7
			public const int SwitchPreference_summaryOn = 7;
			
			// aapt resource value: 8
			public const int SwitchPreference_switchTextOff = 8;
			
			// aapt resource value: 9
			public const int SwitchPreference_switchTextOn = 9;
			
			// aapt resource value: { 0x1010002,0x10100F2,0x101014F }
			public static int[] TabItem = new int[] {
					16842754,
					16842994,
					16843087};
			
			// aapt resource value: 0
			public const int TabItem_android_icon = 0;
			
			// aapt resource value: 1
			public const int TabItem_android_layout = 1;
			
			// aapt resource value: 2
			public const int TabItem_android_text = 2;
			
			// aapt resource value: { 0x7F030382,0x7F030383,0x7F030384,0x7F030385,0x7F030386,0x7F030387,0x7F030388,0x7F030389,0x7F03038A,0x7F03038B,0x7F03038C,0x7F03038D,0x7F03038E,0x7F03038F,0x7F030390,0x7F030391,0x7F030392,0x7F030393,0x7F030394,0x7F030395,0x7F030396,0x7F030397,0x7F030398,0x7F03039A,0x7F03039B,0x7F03039C }
			public static int[] TabLayout = new int[] {
					2130903938,
					2130903939,
					2130903940,
					2130903941,
					2130903942,
					2130903943,
					2130903944,
					2130903945,
					2130903946,
					2130903947,
					2130903948,
					2130903949,
					2130903950,
					2130903951,
					2130903952,
					2130903953,
					2130903954,
					2130903955,
					2130903956,
					2130903957,
					2130903958,
					2130903959,
					2130903960,
					2130903962,
					2130903963,
					2130903964};
			
			// aapt resource value: 0
			public const int TabLayout_tabBackground = 0;
			
			// aapt resource value: 1
			public const int TabLayout_tabContentStart = 1;
			
			// aapt resource value: 2
			public const int TabLayout_tabGravity = 2;
			
			// aapt resource value: 3
			public const int TabLayout_tabIconTint = 3;
			
			// aapt resource value: 4
			public const int TabLayout_tabIconTintMode = 4;
			
			// aapt resource value: 5
			public const int TabLayout_tabIndicator = 5;
			
			// aapt resource value: 6
			public const int TabLayout_tabIndicatorAnimationDuration = 6;
			
			// aapt resource value: 7
			public const int TabLayout_tabIndicatorAnimationMode = 7;
			
			// aapt resource value: 8
			public const int TabLayout_tabIndicatorColor = 8;
			
			// aapt resource value: 9
			public const int TabLayout_tabIndicatorFullWidth = 9;
			
			// aapt resource value: 10
			public const int TabLayout_tabIndicatorGravity = 10;
			
			// aapt resource value: 11
			public const int TabLayout_tabIndicatorHeight = 11;
			
			// aapt resource value: 12
			public const int TabLayout_tabInlineLabel = 12;
			
			// aapt resource value: 13
			public const int TabLayout_tabMaxWidth = 13;
			
			// aapt resource value: 14
			public const int TabLayout_tabMinWidth = 14;
			
			// aapt resource value: 15
			public const int TabLayout_tabMode = 15;
			
			// aapt resource value: 16
			public const int TabLayout_tabPadding = 16;
			
			// aapt resource value: 17
			public const int TabLayout_tabPaddingBottom = 17;
			
			// aapt resource value: 18
			public const int TabLayout_tabPaddingEnd = 18;
			
			// aapt resource value: 19
			public const int TabLayout_tabPaddingStart = 19;
			
			// aapt resource value: 20
			public const int TabLayout_tabPaddingTop = 20;
			
			// aapt resource value: 21
			public const int TabLayout_tabRippleColor = 21;
			
			// aapt resource value: 22
			public const int TabLayout_tabSelectedTextColor = 22;
			
			// aapt resource value: 23
			public const int TabLayout_tabTextAppearance = 23;
			
			// aapt resource value: 24
			public const int TabLayout_tabTextColor = 24;
			
			// aapt resource value: 25
			public const int TabLayout_tabUnboundedRipple = 25;
			
			// aapt resource value: { 0x1010095,0x1010096,0x1010097,0x1010098,0x101009A,0x101009B,0x1010161,0x1010162,0x1010163,0x1010164,0x10103AC,0x1010585,0x7F03019F,0x7F0301A8,0x7F0303A2,0x7F0303C4 }
			public static int[] TextAppearance = new int[] {
					16842901,
					16842902,
					16842903,
					16842904,
					16842906,
					16842907,
					16843105,
					16843106,
					16843107,
					16843108,
					16843692,
					16844165,
					2130903455,
					2130903464,
					2130903970,
					2130904004};
			
			// aapt resource value: 10
			public const int TextAppearance_android_fontFamily = 10;
			
			// aapt resource value: 6
			public const int TextAppearance_android_shadowColor = 6;
			
			// aapt resource value: 7
			public const int TextAppearance_android_shadowDx = 7;
			
			// aapt resource value: 8
			public const int TextAppearance_android_shadowDy = 8;
			
			// aapt resource value: 9
			public const int TextAppearance_android_shadowRadius = 9;
			
			// aapt resource value: 3
			public const int TextAppearance_android_textColor = 3;
			
			// aapt resource value: 4
			public const int TextAppearance_android_textColorHint = 4;
			
			// aapt resource value: 5
			public const int TextAppearance_android_textColorLink = 5;
			
			// aapt resource value: 11
			public const int TextAppearance_android_textFontWeight = 11;
			
			// aapt resource value: 0
			public const int TextAppearance_android_textSize = 0;
			
			// aapt resource value: 2
			public const int TextAppearance_android_textStyle = 2;
			
			// aapt resource value: 1
			public const int TextAppearance_android_typeface = 1;
			
			// aapt resource value: 12
			public const int TextAppearance_fontFamily = 12;
			
			// aapt resource value: 13
			public const int TextAppearance_fontVariationSettings = 13;
			
			// aapt resource value: 14
			public const int TextAppearance_textAllCaps = 14;
			
			// aapt resource value: 15
			public const int TextAppearance_textLocale = 15;
			
			// aapt resource value: { 0x1010095,0x1010096,0x1010097,0x101014F,0x1010161,0x1010162,0x1010163,0x1010164,0x10103AC,0x7F030065,0x7F030066,0x7F0303C1,0x7F0303C5,0x7F0303C6 }
			public static int[] TextEffects = new int[] {
					16842901,
					16842902,
					16842903,
					16843087,
					16843105,
					16843106,
					16843107,
					16843108,
					16843692,
					2130903141,
					2130903142,
					2130904001,
					2130904005,
					2130904006};
			
			// aapt resource value: 8
			public const int TextEffects_android_fontFamily = 8;
			
			// aapt resource value: 4
			public const int TextEffects_android_shadowColor = 4;
			
			// aapt resource value: 5
			public const int TextEffects_android_shadowDx = 5;
			
			// aapt resource value: 6
			public const int TextEffects_android_shadowDy = 6;
			
			// aapt resource value: 7
			public const int TextEffects_android_shadowRadius = 7;
			
			// aapt resource value: 3
			public const int TextEffects_android_text = 3;
			
			// aapt resource value: 0
			public const int TextEffects_android_textSize = 0;
			
			// aapt resource value: 2
			public const int TextEffects_android_textStyle = 2;
			
			// aapt resource value: 1
			public const int TextEffects_android_typeface = 1;
			
			// aapt resource value: 9
			public const int TextEffects_borderRound = 9;
			
			// aapt resource value: 10
			public const int TextEffects_borderRoundPercent = 10;
			
			// aapt resource value: 11
			public const int TextEffects_textFillColor = 11;
			
			// aapt resource value: 12
			public const int TextEffects_textOutlineColor = 12;
			
			// aapt resource value: 13
			public const int TextEffects_textOutlineThickness = 13;
			
			// aapt resource value: { 0x7F0303C2 }
			public static int[] TextInputEditText = new int[] {
					2130904002};
			
			// aapt resource value: 0
			public const int TextInputEditText_textInputLayoutFocusedRectEnabled = 0;
			
			// aapt resource value: { 0x101000E,0x101009A,0x101011F,0x101013F,0x1010150,0x7F03006D,0x7F03006E,0x7F03006F,0x7F030070,0x7F030071,0x7F030072,0x7F030073,0x7F030074,0x7F030075,0x7F030076,0x7F030077,0x7F03010F,0x7F030110,0x7F030111,0x7F030112,0x7F030113,0x7F030114,0x7F03015C,0x7F03015D,0x7F03015E,0x7F03015F,0x7F030160,0x7F030161,0x7F030168,0x7F030169,0x7F03016A,0x7F03016B,0x7F03016C,0x7F03016D,0x7F03016E,0x7F030172,0x7F0301B6,0x7F0301B7,0x7F0301B8,0x7F0301B9,0x7F0301BE,0x7F0301BF,0x7F0301C0,0x7F0301C1,0x7F0302DC,0x7F0302DD,0x7F0302DE,0x7F0302DF,0x7F0302E0,0x7F0302EA,0x7F0302EB,0x7F0302EC,0x7F030301,0x7F030302,0x7F030303,0x7F030333,0x7F030336,0x7F03035B,0x7F03035C,0x7F03035D,0x7F03035E,0x7F03035F,0x7F030372,0x7F030373,0x7F030374 }
			public static int[] TextInputLayout = new int[] {
					16842766,
					16842906,
					16843039,
					16843071,
					16843088,
					2130903149,
					2130903150,
					2130903151,
					2130903152,
					2130903153,
					2130903154,
					2130903155,
					2130903156,
					2130903157,
					2130903158,
					2130903159,
					2130903311,
					2130903312,
					2130903313,
					2130903314,
					2130903315,
					2130903316,
					2130903388,
					2130903389,
					2130903390,
					2130903391,
					2130903392,
					2130903393,
					2130903400,
					2130903401,
					2130903402,
					2130903403,
					2130903404,
					2130903405,
					2130903406,
					2130903410,
					2130903478,
					2130903479,
					2130903480,
					2130903481,
					2130903486,
					2130903487,
					2130903488,
					2130903489,
					2130903772,
					2130903773,
					2130903774,
					2130903775,
					2130903776,
					2130903786,
					2130903787,
					2130903788,
					2130903809,
					2130903810,
					2130903811,
					2130903859,
					2130903862,
					2130903899,
					2130903900,
					2130903901,
					2130903902,
					2130903903,
					2130903922,
					2130903923,
					2130903924};
			
			// aapt resource value: 0
			public const int TextInputLayout_android_enabled = 0;
			
			// aapt resource value: 4
			public const int TextInputLayout_android_hint = 4;
			
			// aapt resource value: 2
			public const int TextInputLayout_android_maxWidth = 2;
			
			// aapt resource value: 3
			public const int TextInputLayout_android_minWidth = 3;
			
			// aapt resource value: 1
			public const int TextInputLayout_android_textColorHint = 1;
			
			// aapt resource value: 5
			public const int TextInputLayout_boxBackgroundColor = 5;
			
			// aapt resource value: 6
			public const int TextInputLayout_boxBackgroundMode = 6;
			
			// aapt resource value: 7
			public const int TextInputLayout_boxCollapsedPaddingTop = 7;
			
			// aapt resource value: 8
			public const int TextInputLayout_boxCornerRadiusBottomEnd = 8;
			
			// aapt resource value: 9
			public const int TextInputLayout_boxCornerRadiusBottomStart = 9;
			
			// aapt resource value: 10
			public const int TextInputLayout_boxCornerRadiusTopEnd = 10;
			
			// aapt resource value: 11
			public const int TextInputLayout_boxCornerRadiusTopStart = 11;
			
			// aapt resource value: 12
			public const int TextInputLayout_boxStrokeColor = 12;
			
			// aapt resource value: 13
			public const int TextInputLayout_boxStrokeErrorColor = 13;
			
			// aapt resource value: 14
			public const int TextInputLayout_boxStrokeWidth = 14;
			
			// aapt resource value: 15
			public const int TextInputLayout_boxStrokeWidthFocused = 15;
			
			// aapt resource value: 16
			public const int TextInputLayout_counterEnabled = 16;
			
			// aapt resource value: 17
			public const int TextInputLayout_counterMaxLength = 17;
			
			// aapt resource value: 18
			public const int TextInputLayout_counterOverflowTextAppearance = 18;
			
			// aapt resource value: 19
			public const int TextInputLayout_counterOverflowTextColor = 19;
			
			// aapt resource value: 20
			public const int TextInputLayout_counterTextAppearance = 20;
			
			// aapt resource value: 21
			public const int TextInputLayout_counterTextColor = 21;
			
			// aapt resource value: 22
			public const int TextInputLayout_endIconCheckable = 22;
			
			// aapt resource value: 23
			public const int TextInputLayout_endIconContentDescription = 23;
			
			// aapt resource value: 24
			public const int TextInputLayout_endIconDrawable = 24;
			
			// aapt resource value: 25
			public const int TextInputLayout_endIconMode = 25;
			
			// aapt resource value: 26
			public const int TextInputLayout_endIconTint = 26;
			
			// aapt resource value: 27
			public const int TextInputLayout_endIconTintMode = 27;
			
			// aapt resource value: 28
			public const int TextInputLayout_errorContentDescription = 28;
			
			// aapt resource value: 29
			public const int TextInputLayout_errorEnabled = 29;
			
			// aapt resource value: 30
			public const int TextInputLayout_errorIconDrawable = 30;
			
			// aapt resource value: 31
			public const int TextInputLayout_errorIconTint = 31;
			
			// aapt resource value: 32
			public const int TextInputLayout_errorIconTintMode = 32;
			
			// aapt resource value: 33
			public const int TextInputLayout_errorTextAppearance = 33;
			
			// aapt resource value: 34
			public const int TextInputLayout_errorTextColor = 34;
			
			// aapt resource value: 35
			public const int TextInputLayout_expandedHintEnabled = 35;
			
			// aapt resource value: 36
			public const int TextInputLayout_helperText = 36;
			
			// aapt resource value: 37
			public const int TextInputLayout_helperTextEnabled = 37;
			
			// aapt resource value: 38
			public const int TextInputLayout_helperTextTextAppearance = 38;
			
			// aapt resource value: 39
			public const int TextInputLayout_helperTextTextColor = 39;
			
			// aapt resource value: 40
			public const int TextInputLayout_hintAnimationEnabled = 40;
			
			// aapt resource value: 41
			public const int TextInputLayout_hintEnabled = 41;
			
			// aapt resource value: 42
			public const int TextInputLayout_hintTextAppearance = 42;
			
			// aapt resource value: 43
			public const int TextInputLayout_hintTextColor = 43;
			
			// aapt resource value: 44
			public const int TextInputLayout_passwordToggleContentDescription = 44;
			
			// aapt resource value: 45
			public const int TextInputLayout_passwordToggleDrawable = 45;
			
			// aapt resource value: 46
			public const int TextInputLayout_passwordToggleEnabled = 46;
			
			// aapt resource value: 47
			public const int TextInputLayout_passwordToggleTint = 47;
			
			// aapt resource value: 48
			public const int TextInputLayout_passwordToggleTintMode = 48;
			
			// aapt resource value: 49
			public const int TextInputLayout_placeholderText = 49;
			
			// aapt resource value: 50
			public const int TextInputLayout_placeholderTextAppearance = 50;
			
			// aapt resource value: 51
			public const int TextInputLayout_placeholderTextColor = 51;
			
			// aapt resource value: 52
			public const int TextInputLayout_prefixText = 52;
			
			// aapt resource value: 53
			public const int TextInputLayout_prefixTextAppearance = 53;
			
			// aapt resource value: 54
			public const int TextInputLayout_prefixTextColor = 54;
			
			// aapt resource value: 55
			public const int TextInputLayout_shapeAppearance = 55;
			
			// aapt resource value: 56
			public const int TextInputLayout_shapeAppearanceOverlay = 56;
			
			// aapt resource value: 57
			public const int TextInputLayout_startIconCheckable = 57;
			
			// aapt resource value: 58
			public const int TextInputLayout_startIconContentDescription = 58;
			
			// aapt resource value: 59
			public const int TextInputLayout_startIconDrawable = 59;
			
			// aapt resource value: 60
			public const int TextInputLayout_startIconTint = 60;
			
			// aapt resource value: 61
			public const int TextInputLayout_startIconTintMode = 61;
			
			// aapt resource value: 62
			public const int TextInputLayout_suffixText = 62;
			
			// aapt resource value: 63
			public const int TextInputLayout_suffixTextAppearance = 63;
			
			// aapt resource value: 64
			public const int TextInputLayout_suffixTextColor = 64;
			
			// aapt resource value: { 0x1010034,0x7F030162,0x7F030163 }
			public static int[] ThemeEnforcement = new int[] {
					16842804,
					2130903394,
					2130903395};
			
			// aapt resource value: 0
			public const int ThemeEnforcement_android_textAppearance = 0;
			
			// aapt resource value: 1
			public const int ThemeEnforcement_enforceMaterialTheme = 1;
			
			// aapt resource value: 2
			public const int ThemeEnforcement_enforceTextAppearance = 2;
			
			// aapt resource value: { 0x10100AF,0x1010140,0x7F03007F,0x7F0300CD,0x7F0300CE,0x7F0300F3,0x7F0300F4,0x7F0300F5,0x7F0300F6,0x7F0300F7,0x7F0300F8,0x7F030262,0x7F030263,0x7F030282,0x7F03028A,0x7F0302B9,0x7F0302BA,0x7F0302F5,0x7F03036D,0x7F03036F,0x7F030370,0x7F0303E2,0x7F0303E6,0x7F0303E7,0x7F0303E8,0x7F0303E9,0x7F0303EA,0x7F0303EB,0x7F0303EC,0x7F0303ED }
			public static int[] Toolbar = new int[] {
					16842927,
					16843072,
					2130903167,
					2130903245,
					2130903246,
					2130903283,
					2130903284,
					2130903285,
					2130903286,
					2130903287,
					2130903288,
					2130903650,
					2130903651,
					2130903682,
					2130903690,
					2130903737,
					2130903738,
					2130903797,
					2130903917,
					2130903919,
					2130903920,
					2130904034,
					2130904038,
					2130904039,
					2130904040,
					2130904041,
					2130904042,
					2130904043,
					2130904044,
					2130904045};
			
			// aapt resource value: 0
			public const int Toolbar_android_gravity = 0;
			
			// aapt resource value: 1
			public const int Toolbar_android_minHeight = 1;
			
			// aapt resource value: 2
			public const int Toolbar_buttonGravity = 2;
			
			// aapt resource value: 3
			public const int Toolbar_collapseContentDescription = 3;
			
			// aapt resource value: 4
			public const int Toolbar_collapseIcon = 4;
			
			// aapt resource value: 5
			public const int Toolbar_contentInsetEnd = 5;
			
			// aapt resource value: 6
			public const int Toolbar_contentInsetEndWithActions = 6;
			
			// aapt resource value: 7
			public const int Toolbar_contentInsetLeft = 7;
			
			// aapt resource value: 8
			public const int Toolbar_contentInsetRight = 8;
			
			// aapt resource value: 9
			public const int Toolbar_contentInsetStart = 9;
			
			// aapt resource value: 10
			public const int Toolbar_contentInsetStartWithNavigation = 10;
			
			// aapt resource value: 11
			public const int Toolbar_logo = 11;
			
			// aapt resource value: 12
			public const int Toolbar_logoDescription = 12;
			
			// aapt resource value: 13
			public const int Toolbar_maxButtonHeight = 13;
			
			// aapt resource value: 14
			public const int Toolbar_menu = 14;
			
			// aapt resource value: 15
			public const int Toolbar_navigationContentDescription = 15;
			
			// aapt resource value: 16
			public const int Toolbar_navigationIcon = 16;
			
			// aapt resource value: 17
			public const int Toolbar_popupTheme = 17;
			
			// aapt resource value: 18
			public const int Toolbar_subtitle = 18;
			
			// aapt resource value: 19
			public const int Toolbar_subtitleTextAppearance = 19;
			
			// aapt resource value: 20
			public const int Toolbar_subtitleTextColor = 20;
			
			// aapt resource value: 21
			public const int Toolbar_title = 21;
			
			// aapt resource value: 22
			public const int Toolbar_titleMargin = 22;
			
			// aapt resource value: 23
			public const int Toolbar_titleMarginBottom = 23;
			
			// aapt resource value: 24
			public const int Toolbar_titleMarginEnd = 24;
			
			// aapt resource value: 27
			public const int Toolbar_titleMargins = 27;
			
			// aapt resource value: 25
			public const int Toolbar_titleMarginStart = 25;
			
			// aapt resource value: 26
			public const int Toolbar_titleMarginTop = 26;
			
			// aapt resource value: 28
			public const int Toolbar_titleTextAppearance = 28;
			
			// aapt resource value: 29
			public const int Toolbar_titleTextColor = 29;
			
			// aapt resource value: { 0x1010034,0x10100D5,0x10100F6,0x101013F,0x1010140,0x101014F,0x7F030050 }
			public static int[] Tooltip = new int[] {
					16842804,
					16842965,
					16842998,
					16843071,
					16843072,
					16843087,
					2130903120};
			
			// aapt resource value: 2
			public const int Tooltip_android_layout_margin = 2;
			
			// aapt resource value: 4
			public const int Tooltip_android_minHeight = 4;
			
			// aapt resource value: 3
			public const int Tooltip_android_minWidth = 3;
			
			// aapt resource value: 1
			public const int Tooltip_android_padding = 1;
			
			// aapt resource value: 5
			public const int Tooltip_android_text = 5;
			
			// aapt resource value: 0
			public const int Tooltip_android_textAppearance = 0;
			
			// aapt resource value: 6
			public const int Tooltip_backgroundTint = 6;
			
			// aapt resource value: { 0x1010320,0x1010321,0x1010322,0x1010323,0x1010324,0x1010325,0x1010326,0x1010327,0x1010328,0x10103FA,0x1010440,0x7F030402 }
			public static int[] Transform = new int[] {
					16843552,
					16843553,
					16843554,
					16843555,
					16843556,
					16843557,
					16843558,
					16843559,
					16843560,
					16843770,
					16843840,
					2130904066};
			
			// aapt resource value: 10
			public const int Transform_android_elevation = 10;
			
			// aapt resource value: 6
			public const int Transform_android_rotation = 6;
			
			// aapt resource value: 7
			public const int Transform_android_rotationX = 7;
			
			// aapt resource value: 8
			public const int Transform_android_rotationY = 8;
			
			// aapt resource value: 4
			public const int Transform_android_scaleX = 4;
			
			// aapt resource value: 5
			public const int Transform_android_scaleY = 5;
			
			// aapt resource value: 0
			public const int Transform_android_transformPivotX = 0;
			
			// aapt resource value: 1
			public const int Transform_android_transformPivotY = 1;
			
			// aapt resource value: 2
			public const int Transform_android_translationX = 2;
			
			// aapt resource value: 3
			public const int Transform_android_translationY = 3;
			
			// aapt resource value: 9
			public const int Transform_android_translationZ = 9;
			
			// aapt resource value: 11
			public const int Transform_transformPivotTarget = 11;
			
			// aapt resource value: { 0x10100D0,0x7F030046,0x7F0300EC,0x7F0300ED,0x7F030151,0x7F030207,0x7F0302AE,0x7F0302E1,0x7F030359,0x7F030403,0x7F030405 }
			public static int[] Transition = new int[] {
					16842960,
					2130903110,
					2130903276,
					2130903277,
					2130903377,
					2130903559,
					2130903726,
					2130903777,
					2130903897,
					2130904067,
					2130904069};
			
			// aapt resource value: 0
			public const int Transition_android_id = 0;
			
			// aapt resource value: 1
			public const int Transition_autoTransition = 1;
			
			// aapt resource value: 2
			public const int Transition_constraintSetEnd = 2;
			
			// aapt resource value: 3
			public const int Transition_constraintSetStart = 3;
			
			// aapt resource value: 4
			public const int Transition_duration = 4;
			
			// aapt resource value: 5
			public const int Transition_layoutDuringTransition = 5;
			
			// aapt resource value: 6
			public const int Transition_motionInterpolator = 6;
			
			// aapt resource value: 7
			public const int Transition_pathMotionArc = 7;
			
			// aapt resource value: 8
			public const int Transition_staggered = 8;
			
			// aapt resource value: 9
			public const int Transition_transitionDisable = 9;
			
			// aapt resource value: 10
			public const int Transition_transitionFlags = 10;
			
			// aapt resource value: { 0x7F0300F0,0x7F030318,0x7F030319,0x7F03031A,0x7F03031B }
			public static int[] Variant = new int[] {
					2130903280,
					2130903832,
					2130903833,
					2130903834,
					2130903835};
			
			// aapt resource value: 0
			public const int Variant_constraints = 0;
			
			// aapt resource value: 1
			public const int Variant_region_heightLessThan = 1;
			
			// aapt resource value: 2
			public const int Variant_region_heightMoreThan = 2;
			
			// aapt resource value: 3
			public const int Variant_region_widthLessThan = 3;
			
			// aapt resource value: 4
			public const int Variant_region_widthMoreThan = 4;
			
			// aapt resource value: { 0x1010000,0x10100DA,0x7F0302D3,0x7F0302D6,0x7F0303CE }
			public static int[] View = new int[] {
					16842752,
					16842970,
					2130903763,
					2130903766,
					2130904014};
			
			// aapt resource value: { 0x10100D4,0x7F030050,0x7F030051 }
			public static int[] ViewBackgroundHelper = new int[] {
					16842964,
					2130903120,
					2130903121};
			
			// aapt resource value: 0
			public const int ViewBackgroundHelper_android_background = 0;
			
			// aapt resource value: 1
			public const int ViewBackgroundHelper_backgroundTint = 1;
			
			// aapt resource value: 2
			public const int ViewBackgroundHelper_backgroundTintMode = 2;
			
			// aapt resource value: { 0x10100C4 }
			public static int[] ViewPager2 = new int[] {
					16842948};
			
			// aapt resource value: 0
			public const int ViewPager2_android_orientation = 0;
			
			// aapt resource value: { 0x10100D0,0x10100F2,0x10100F3 }
			public static int[] ViewStubCompat = new int[] {
					16842960,
					16842994,
					16842995};
			
			// aapt resource value: 0
			public const int ViewStubCompat_android_id = 0;
			
			// aapt resource value: 2
			public const int ViewStubCompat_android_inflatedId = 2;
			
			// aapt resource value: 1
			public const int ViewStubCompat_android_layout = 1;
			
			// aapt resource value: { 0x10100D0,0x7F030001,0x7F030002,0x7F0300BF,0x7F030151,0x7F0301D0,0x7F0301D1,0x7F0302AE,0x7F0302B3,0x7F0302CB,0x7F0302E1,0x7F030332,0x7F030403,0x7F03040C,0x7F030415 }
			public static int[] ViewTransition = new int[] {
					16842960,
					2130903041,
					2130903042,
					2130903231,
					2130903377,
					2130903504,
					2130903505,
					2130903726,
					2130903731,
					2130903755,
					2130903777,
					2130903858,
					2130904067,
					2130904076,
					2130904085};
			
			// aapt resource value: 0
			public const int ViewTransition_android_id = 0;
			
			// aapt resource value: 3
			public const int ViewTransition_clearsTag = 3;
			
			// aapt resource value: 4
			public const int ViewTransition_duration = 4;
			
			// aapt resource value: 5
			public const int ViewTransition_ifTagNotSet = 5;
			
			// aapt resource value: 6
			public const int ViewTransition_ifTagSet = 6;
			
			// aapt resource value: 7
			public const int ViewTransition_motionInterpolator = 7;
			
			// aapt resource value: 8
			public const int ViewTransition_motionTarget = 8;
			
			// aapt resource value: 9
			public const int ViewTransition_onStateTransition = 9;
			
			// aapt resource value: 10
			public const int ViewTransition_pathMotionArc = 10;
			
			// aapt resource value: 11
			public const int ViewTransition_setsTag = 11;
			
			// aapt resource value: 1
			public const int ViewTransition_SharedValue = 1;
			
			// aapt resource value: 2
			public const int ViewTransition_SharedValueId = 2;
			
			// aapt resource value: 12
			public const int ViewTransition_transitionDisable = 12;
			
			// aapt resource value: 13
			public const int ViewTransition_upDuration = 13;
			
			// aapt resource value: 14
			public const int ViewTransition_viewTransitionMode = 14;
			
			// aapt resource value: 1
			public const int View_android_focusable = 1;
			
			// aapt resource value: 0
			public const int View_android_theme = 0;
			
			// aapt resource value: 2
			public const int View_paddingEnd = 2;
			
			// aapt resource value: 3
			public const int View_paddingStart = 3;
			
			// aapt resource value: 4
			public const int View_theme = 4;
			
			static Styleable()
			{
				global::Android.Runtime.ResourceIdManager.UpdateIdValues();
			}
			
			private Styleable()
			{
			}
		}
		
		public partial class Xml
		{
			
			// aapt resource value: 0x7F110000
			public const int image_share_filepaths = 2131820544;
			
			// aapt resource value: 0x7F110001
			public const int standalone_badge = 2131820545;
			
			// aapt resource value: 0x7F110002
			public const int standalone_badge_gravity_bottom_end = 2131820546;
			
			// aapt resource value: 0x7F110003
			public const int standalone_badge_gravity_bottom_start = 2131820547;
			
			// aapt resource value: 0x7F110004
			public const int standalone_badge_gravity_top_start = 2131820548;
			
			// aapt resource value: 0x7F110005
			public const int standalone_badge_offset = 2131820549;
			
			// aapt resource value: 0x7F110006
			public const int xamarin_essentials_fileprovider_file_paths = 2131820550;
			
			static Xml()
			{
				global::Android.Runtime.ResourceIdManager.UpdateIdValues();
			}
			
			private Xml()
			{
			}
		}
	}
}
#pragma warning restore 1591
